#!/usr/bin/env python3
""" Utility for diagnosing the gssw proxy message bus """

from datetime                   import datetime
import logging

from gssw.common                import initGssw
from gssw.lib.arrays.gridded    import griddedArrayFromBinMessage
from gssw.lib.errors            import GsswTimeoutError
from gssw.lib.misc              import createGsswArgumentParser
from gssw.lib.sockets           import destroyAllSocketsInContext, SubSocket, tagName
from gssw.proto.array_pb2       import GriddedArrayMsg
from gssw.proto.hwregister_pb2  import HwBlockRep

logger = logging.getLogger(__name__)

def niceTime():
    """ get timestamp with milliseconds """
    return datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]

def createDataSubSocket(config, scienceDaemon, topicfilter):
    """ Factory method for creating a new socket for receiving data published by the proxy """
    dataReceiveSocket = SubSocket()
    dataReceiveSocket.createZmqSocket()
    # subscribe to all data (or specified with filter)
    dataReceiveSocket.subscribe(topicfilter)

    try:
        if scienceDaemon:
            # A science daemon was specified, connect to that one
            print("Connecting to science data bus of component '%s'" % scienceDaemon)
            scienceConfig     = config.componentParameters(scienceDaemon)
            scienceHost       = scienceConfig['hostname']
            sciencePubport    = scienceConfig['sockets']['device-pub']
            dataReceiveAddress = "tcp://{}:{}".format(scienceHost, sciencePubport)
        else:
            # Connect to the proxy bus
            print("Connecting to message bus of component 'proxy'")
            proxyConfig     = config.componentParameters('proxy')
            proxyHost       = proxyConfig['hostname']
            proxyPubport    = proxyConfig['sockets']['xpub']
            dataReceiveAddress = "tcp://{}:{}".format(proxyHost, proxyPubport)
        dataReceiveSocket.connectSocket(dataReceiveAddress)
        return dataReceiveSocket

    except KeyError as ex:
        logger.error("Could not find the component in the configuration file: %s", ex)

def main():
    """ Main logging function """

    # Parse the command line arguments containing the config server address or config file path
    parser = createGsswArgumentParser(__doc__)
    parser.add_argument('--science',
                        type=str,
                        metavar='science-daemon',
                        nargs='?',
                        const='science-ch0',
                        default=None,
                        help='Connect to a science daemon instead of the proxy, in order to monitor'
                             ' block data received from the hardware (default: %(const)s).')
    parser.add_argument('--filter', type=str, default='',
                        help='Filter topics with the give prefix, i.e. only display messages '
                             'starting with the given string.')

    # Parse the command line arguments
    args = parser.parse_args()

    # Retrieve configuration for connecting to PHC
    config = initGssw(args)
    print('Starting message bus logger')

    # Create socket for listening to the message bus
    dataReceiveSocket = createDataSubSocket(config, args.science, args.filter)
    if not dataReceiveSocket:
        print ('Could not connect to the message bus, stopping...')
        return
    dataReceiveSocket.setReceiveTimeout(5000)

    try:
        while True:
            try:
                # Wait until data appears on the message bus
                topic, tag, msg = dataReceiveSocket.receiveProtoBufWithTopic()

                # Log some information about the data
                if tag == tagName(GriddedArrayMsg):
                    gArray = griddedArrayFromBinMessage(msg).data
                    print('%s Topic: %-10s tag: %-18s shape: %-12s name: %s' % (niceTime(),
                        topic, tag, gArray.shape, gArray.name))  #pylint: disable=no-member
                elif tag == tagName(HwBlockRep):
                    # Decode science block data and show meta information
                    blockRep = HwBlockRep()
                    blockRep.ParseFromString(msg)
                    dataLength = len(blockRep.data.block_data)/(blockRep.data.bits_per_value/8)
                    print('%s Topic: %-10s tag: %-18s addr: %-10s datalen: %s' % (niceTime(),
                        topic, tag, hex(blockRep.addr), dataLength))
                else:
                    print('%s Topic: %-10s tag: %-18s' % (niceTime(), topic, tag))
            except GsswTimeoutError:
                # No data received within timeout
                pass

    except KeyboardInterrupt:
        print('Stopping message bus logger')

if __name__=='__main__':
    try:
        main()
    finally:
        destroyAllSocketsInContext()
