#! /usr/bin/env python3

""" Configuration server module.

    Other modules can request the config file from this daemon through a
    ZMQ reply socket using a FileReq protobuffer object.
"""

import os
import logging
import sys

from gssw.common              import initGssw
from gssw.lib.application     import NonQtApplication
from gssw.lib.sockets         import RepSocket, tagName
from gssw.proto.common_pb2    import ErrorRep
from gssw.proto.file_pb2      import FileRep, FileReq
from gssw.lib.misc            import createGsswArgumentParser

logger = logging.getLogger(__name__)

class ConfigServer():
    """ Encapsulates all functionality of the config server. """

    def __init__(self, app, configPath, port):

        self.configPath = configPath
        assert os.path.isfile(self.configPath), \
            "Config file does not exist or is a directory: {}".format(self.configPath)

        # connect linux signal handler
        self.signalHandler = app.getSignalHandler()
        self.signalHandler.sigInt.connect(self.stop)
        self.signalHandler.sigTerm.connect(self.stop)

        # Set up the REP socket for sending the configuration
        self.repSocket = RepSocket()
        self.repSocket.createZmqSocket()
        self.repSocket.bindSocket('tcp://*:%d' % port)


    def start(self):
        """ Main loop of the configuration server. """

        while True:
            # Wait infinitely for a request to come in
            self.repSocket.poll()

            # receive the reply including the header string
            parts = self.repSocket.receiveProtoBuf()

            # Verify the request type
            if parts.tag == tagName(FileReq):
                logger.debug("Request received")

                try:
                    # Open the config file and wrap it in a protobuf message
                    with open(self.configPath, "r") as f:
                        reply           = FileRep()
                        reply.data      = f.read()

                        # Find the directory of the local config file, to support relative paths
                        # in the config file
                        reply.filename  = os.path.normpath(os.path.realpath(self.configPath))
                        logger.debug("Sending config file: {}".format(repr(self.configPath)))
                        # Send the file via protobuf
                        self.repSocket.sendProtoBuf(tagName(FileRep), reply.SerializeToString())

                except (IOError, FileNotFoundError) as e:
                    errorMsg = "File '{}' could not be opened. {}".format(self.configPath, e)
                    logger.warning(errorMsg)
                    self.sendErrorRep(errorMsg, ErrorRep.ERR_FILEIO)
            else:
                # Unsupported request type
                errorMsg = "Received unsupported request: {}".format(repr(parts))
                self.sendErrorRep(errorMsg)

    def stop(self):
        """ Shutdown the server. Can be used for any cleanup actions. """
        logger.info("Shutting down")
        sys.exit(1)

    def sendErrorRep(self, msg, errtype=None):
        """ Send an ErrorRep message in response to an incorrect request. """

        logger.warning(msg)
        rep = ErrorRep()
        rep.errtype = ErrorRep.ERR_PROTO if errtype is None else errtype
        rep.errmsg = msg
        self.repSocket.sendProtoBuf(tagName(ErrorRep), rep.SerializeToString())


def main():
    """ Parse command line arguments and start the server. """

    parser = createGsswArgumentParser(__doc__)
    parser.add_argument('--port', type=int, required=True,
                        help='port to bind the config rep socket to')
    args = parser.parse_args()

    config = initGssw(args)
    application = NonQtApplication(config)
    server = ConfigServer(application, config.configFile, args.port)
    server.start()

if __name__ == "__main__":
    main()
