#! /usr/bin/env python3

""" Component that creates an archive (in the form HDF-5 file) and to which other components
    can publish data.
"""

import logging

from gssw.common import initGssw
from gssw.datawriter.h5worker import H5ArchiverWorker
from gssw.datawriter.h5server import H5ArchiverServer

from gssw.lib.sockets import destroyAllSocketsInContext
from gssw.lib.startup import makeAddress
from gssw.lib.misc    import createGsswArgumentParser
    

logger = logging.getLogger(__name__)


def main():
    """ Starts a H5Archiver component.
    """
    logger.info('Starting h5archiver')

    # Parse the command line argument(s)
    parser = createGsswArgumentParser(__doc__, enableOptionName=True,
                                      optionNameDefault='archiver')

    parser.add_argument('-b', '--bind-sub-socket', dest='bindTo',
                        help="Also binds the subsocket to this address. Useful for testing. "
                             "'Can be a full address such as 'tcp://*:5555', or an integer "
                             "in which case it is interpreted as a port number and the sub-socket "
                             "is bound to 'tcp://*:<portnr>'")

    args = parser.parse_args()
    config = initGssw(args)

    params = config.componentParameters(args.name)

    socketParams = params['sockets']

    archiver = H5ArchiverWorker()
    archiveServer = H5ArchiverServer(archiver,
        repSocketAddress=makeAddress(socketParams['arch-rep'], host='*'),
        subSocketAddress=makeAddress(socketParams['arch-sub']))

    if args.bindTo:
        try:
            port = int(args.bindTo)
            address = "tcp://*:{:d}".format(port)
        except ValueError:
            address = args.bindTo
        logger.debug("Binding sub-socket to: {}".format(address))
        archiveServer.bindSubSocket(address)


    try:
        archiveServer.start()
    except KeyboardInterrupt:
        logger.info("Archiver stopped, closing resources.")
    finally:
        archiveServer.finalize()
        destroyAllSocketsInContext()


if __name__ == '__main__':
    main()
