#!/usr/bin/env python3

""" GSSw Hardware Abstraction Layer Server

"""
#pylint: disable=too-many-statements
#pylint: disable=bad-whitespace
#pylint: disable=trailing-whitespace

###################################################################################################
# P Y T H O N   M O D U L E S
###################################################################################################
import logging
import os
import socket
import sys
import threading

from watchdog.observers             import Observer
from watchdog.events                import PatternMatchingEventHandler

import gssw.proto.halserver_pb2     as serverPb
from gssw.proto.common_pb2          import OkRep

from gssw.common                    import initGssw, diagnosticlogger
from gssw.lib.application           import NonQtApplication
from gssw.lib.config                import Config
from gssw.proto.halconfig_pb2       import HalConfig
from gssw.proto.halserver_pb2       import NotifyHalUpdate
from gssw.hal.parser                import HalParser
from gssw.lib.sockets               import destroyAllSocketsInContext, RepSocket, PubSocket, tagName
from gssw.lib.encryption            import PrivateKey
from gssw.lib.errors                import GsswError, PeerException
from gssw.lib.misc                  import createGsswArgumentParser
from gssw.lib.typing                import check_class
from gssw.hal.hal                   import Hal
from gssw.hal.versioncontrol.halgit import VersionControlGit

logger = logging.getLogger(__name__)


# Update delay for periodical git pull in the background
GIT_UPDATE_PERIOD = 60

# Only watching for Hal file changes (outside of server) on these types of files
HAL_FILE_PATTERNS = ['*.yaml']


###################################################################################################
# H A L S E R V E R
###################################################################################################

class HalServer:
    """ Server process for the Hardware Abstraction Layer

        HAL Server takes care of:

        - creation, reading, updating, deleting of yaml file
        - destribution of hal file
    """

    def __init__(self, config, app):
        """ Constructor """

        check_class(config, Config)

        # connect linux signal handler
        self.signalHandler = app.getSignalHandler()
        self.signalHandler.sigInt.connect(self.stop)
        self.signalHandler.sigTerm.connect(self.stop)

        # get the config parameters for this component
        halCfg = config.componentParameters(app.name)
        hostname    = halCfg['hostname']
        repPort     = halCfg['sockets']['hal-rep']
        pubPort     = halCfg['sockets']['hal-pub']

        # do some logging
        logger.info('mypath     = %s', os.getcwd())
        self.hal_path = config.asAbsolutePath(halCfg['hal_path'])
        logger.info('configpath = %s', self.hal_path)

        self.parser = None

        # create a REP socket
        self.repSock = RepSocket()
        self.repSock.createZmqSocket(timeoutMs=-1)

        # bind the REP socket to the host
        logger.info('Binding REP socket to %s:%d', hostname, repPort)
        self.repSock.bindSocket('tcp://*:%d' %  repPort)

        # create a PUB socket
        self.pubSock = PubSocket()
        self.pubSock.createZmqSocket()

        # bind the PUB socket to the host
        logger.info('Binding PUB socket to %s:%d', hostname, pubPort)
        self.pubSock.bindSocket('tcp://*:%d' % pubPort)

        self._privateKey = None # Placeholder for decryption key for git credentials
        try:
            self._versionControl = VersionControlGit(self.hal_path)
        except GsswError as ex:
            logger.error('Could not create hal git client, synchronization is disabled: %s', ex)
            self._versionControl = None

        # Object for watching hal file change (outside of the hal server)
        self.halFileChangeHandler = HalFileChangedHandler(self, HAL_FILE_PATTERNS)
        self.halFileObserver = Observer()
        self.halFileObserver.schedule(
            self.halFileChangeHandler, os.path.dirname(self.hal_path), recursive=True)
        self.halFileObserver.start()

        # Background thread for periodically pulling updates from git server
        autoUpdate  = halCfg.get('auto_update', True) # background update can be disabled in config
        if self._versionControl is not None and autoUpdate:
            self._backgroundUpdateStopEvent = threading.Event()
            self._backgroundUpdateThread    = threading.Thread(target=self.backgroundUpdateWorker,
                                                               name='Git-background-updater')
            self._backgroundUpdateThread.start()
        else:
            logger.info("HAL auto-updates are disabled")
            if self._versionControl is not None: # No notification if local HAL is used
                diagnosticlogger.info("HAL auto-updates are disabled",
                    "Your setup is fixed to one HAL version, and will not receive automatic "
                    "updates. To change edit 'auto_update' for the '{}' in the configuration "
                    "file".format(app.name))
            #TODO: Maybe notify user that database is used off-line without synchronization in GUI

        self.reloadHal()

    def reloadHal(self):
        """ Load HAL yaml file into memory. """
        # extra step to restore database
        hal = Hal()
        hal.setHalFromFile(self.hal_path)
        configData = hal.createYamlData()

        # Create a HAL parser and parse the YAML file
        self.parser = HalParser()
        self.parser.parseFromConfigData(configData)

        self.notifyClients()

    def backgroundUpdateWorker(self):
        """ Threaded git pull function, to update the HAL yaml file. """
        logger.info("Starting background git-updater")
        while not self._backgroundUpdateStopEvent.is_set():
            logger.debug("Checking for remote HAL updates")
            try:
                if self._versionControl:
                    # Check if there is a new HAL available, and if so, update
                    numNewCommits, _ = self._versionControl.checkUpdateAvailable()
                    if numNewCommits:
                        logger.info("The remote server contains %s new commits, pulling them now",
                                    numNewCommits)
                        self.halFileChangeHandler.suspend()
                        try:
                            self._versionControl.update()
                            self.halFileChangeHandler.trigger()
                        finally:
                            self.halFileChangeHandler.resume()
            except GsswError as ex:
                logger.exception("Could not update HAL from remote server: {}".format(ex))

            # No client update notification is required here;
            # the file-change observer will take care of this.
            self._backgroundUpdateStopEvent.wait(GIT_UPDATE_PERIOD)

    def start(self):  #pylint: disable=too-many-locals
        """ Worker function """

        while True:

            req = self.repSock.receiveProtoBuf()
            logger.debug("Request received")

            if req.tag == 'GetHalConfig' :

                logger.info('Received request for HAL config protobuf')

                configPb = self.parser.protobuf
                logger.debug('GetHalConfig protobuf from server')
                self.repSock.sendProtoBuf(tagName(HalConfig), configPb.SerializeToString())

            elif req.tag == 'SetHalConfig' :

                logger.info('Received request to set HAL config via protobuf')

                okRep = OkRep()
                self.repSock.sendProtoBuf(tagName(OkRep)
                                          , okRep.SerializeToString())

                setRequest = serverPb.SetHalConfig()
                setRequest.ParseFromString(req.message)
                newProtobuf = setRequest.config
                logger.debug("SetHalConfig protobuf from client = %s", newProtobuf)

                self.parser.parseFromProtobuf(newProtobuf)

                storedFile = self._storeHALConfiguration(self.parser.configData)

                if storedFile :
                    # Notify HAL clients that a new HAL is available
                    self.notifyClients()

            elif req.tag == 'GetHalDeviceTypeConfig' :

                getDevTypeReq = serverPb.GetHalDeviceTypeConfig()
                getDevTypeReq.ParseFromString(req.message)

                logger.debug('Received device type config from %s' , getDevTypeReq.name)

                # TODO: Fill protobuffer, send empty response for now
                devTypeResp = serverPb.HalDeviceTypeConfig()
                self.repSock.sendProtoBuf(tagName(serverPb.HalDeviceTypeConfig)
                                          , devTypeResp.SerializeToString())

            elif req.tag == 'GetServerPublicKey':
                # Create new encryption keys
                self._privateKey = PrivateKey()
                publicKeyRep = serverPb.GetServerPublicKeyRep()

                # Send the public key to the client
                publicKeyRep.public_key = self._privateKey.publicKey.serialize()
                self.repSock.sendProtoBuf(tagName(publicKeyRep), publicKeyRep.SerializeToString())
            elif req.tag == 'CommitHal':
                try:
                    assert self._privateKey,\
                        "Call GetServerPublicKey first to retrieve the public key"
                    assert self._versionControl is not None,\
                        "HAL cannot be uploaded to git, synchronization is disabled"

                    commitReq = serverPb.CommitHal()
                    commitReq.ParseFromString(req.message)
                    logger.info('Received commit request (comment: %s)', commitReq.comment)

                    # Add hostname to logmessage for later info
                    comment = commitReq.comment
                    if comment:
                        comment += "\n\n"
                    comment += "Uploaded from: {}".format(socket.gethostname())

                    # Get list of file paths to commit
                    commitFiles = [self.hal_path]
                    basepath = os.path.dirname(self.hal_path)
                    for directory in (d for d in os.scandir(basepath) if d.is_dir() and
                                                                         d.name in Hal.YAML_DIRS):
                        commitFiles += [f.path for f in os.scandir(directory.path) if f.is_file()]

                    # Now we can decrypt and use the credentials to make the commit and push
                    self._versionControl.commitPush(commitFiles, comment,
                        self._privateKey.decrypt(commitReq.username_encrypted).decode('utf-8'),
                        self._privateKey.decrypt(commitReq.password_encrypted).decode('utf-8'))

                    okRep = OkRep()
                    self.repSock.sendProtoBuf(tagName(okRep), okRep.SerializeToString())

                    # Notify clients of updated version
                    self.notifyClients()
                except Exception as ex:
                    # Prevent the server from crashing, and send an exception reply instead
                    logger.exception('An error occurred while handling the commit request')
                    replyTag, binReplyMessage = PeerException.exceptionToMessage(ex)
                    self.repSock.sendProtoBuf(replyTag, binReplyMessage)
            elif req.tag == tagName(serverPb.CheckHalUploadAvailable):
                # Check whether the upload function is applicable at the moment
                repMsg = serverPb.CheckHalUploadAvailableRep()
                if self._versionControl:
                    # Only if version control is used 
                    repMsg.upload_available = self._versionControl.checkLocalChanges()
                self.repSock.sendProtoBuf(tagName(repMsg), repMsg.SerializeToString())
            elif req.tag == tagName(serverPb.RevertHalReq):
                repMsg = OkRep()
                if self._versionControl:
                    # Ignore RPC call when not using version control for the HAL
                    self._versionControl.discardLocalChanges()
                self.repSock.sendProtoBuf(tagName(repMsg), repMsg.SerializeToString())
            else:
                # request does not have the correct syntax
                logger.error('Received unknown request type: %r', req.tag)
                peerEx = PeerException('Received unknown request type: %r' % req.tag)
                replyTag, binReplyMessage = PeerException.exceptionToMessage(peerEx)
                self.repSock.sendProtoBuf(replyTag, binReplyMessage)

    def stop(self):
        """ Shutdown the server. Can be used for any cleanup actions. """
        logger.info("Shutting down")
        sys.exit(1)

    def _storeHALConfiguration(self, halConfig):
        """ Takes HAL as python dictionary and stores as YAML
            Returns True for success.
        """
        hal = Hal()
        return hal.storeHal(self.hal_path, hal=halConfig)

    def notifyClients(self):
        """ Sends out an update notification to HalClient (new Hal version available on server) """

        pub = NotifyHalUpdate()
        self.pubSock.publishProtoBuf(""
                                    , tagName(pub)
                                    , pub.SerializeToString())
        logger.debug("Publish availability of updated HAL dict")


class HalFileChangedHandler(PatternMatchingEventHandler):
    """ Handles file system events for the HAL yaml file.

    :note: The filesystem can send many events when a file is changed, so a delay timer is used to
           avoid hysteresis and only reload the file once. """

    RELOAD_HYSTERESIS_DELAY = 2     # Delay before reloading the file after a file change event

    def __init__(self, halServer, patterns):
        """ Constructor """
        super().__init__(patterns=patterns)
        self._halServer = halServer
        self._hysteresisTimer = None
        self._threadLock = threading.Lock()
        self._suspendedEvent = threading.Event()

    def suspend(self):
        """ Pauses the file watcher, used when pulling new updates """
        self._suspendedEvent.set()

    def resume(self):
        """ Resumes the file watcher, used when pulling new updates is done """
        self._suspendedEvent.clear()

    def on_any_event(self, event):
        """ Event function """
        def dirName(path):
            """ Get name of lowest directory in path. """
            return os.path.basename(os.path.dirname(path))

        if self._suspendedEvent.is_set():
            logger.debug('Ignoring changed HAL file, the reloader is suspended')
            return

        halPath = self._halServer.hal_path
        # determine if changes to either source or destination directory are relevant
        src = event.src_path == halPath or dirName(event.src_path) in Hal.YAML_DIRS
        dst = hasattr(event, 'dest_path') and (event.dest_path == halPath or
                                               dirName(event.dest_path) in Hal.YAML_DIRS)
        if src or dst:
            logger.debug("HAL file has changed, schedule reload")
            self.trigger()

    def trigger(self):
        """ Schedule reload of yaml file, with hysteresis applied. """
        with self._threadLock:
            if self._hysteresisTimer:
                # There was another file event during the hysteresis delay, reset the delay
                self._hysteresisTimer.cancel()
                self._hysteresisTimer = None
            self._hysteresisTimer = threading.Timer(self.RELOAD_HYSTERESIS_DELAY,
                                                    self._reloadFile)
            self._hysteresisTimer.start()

    def _reloadFile(self):
        """ Called after reload hysteresis timer expires. """
        with self._threadLock:
            logger.info("HAL file has changed, reloading...")
            self._halServer.reloadHal()


###################################################################################################
# H E L P E R   F U N C T I O N S
###################################################################################################

###################################################################################################
# M A I N
###################################################################################################

def main():
    """Main function"""

    try:
        args = createGsswArgumentParser(__doc__, enableOptionName=True,
                                        optionNameDefault='hal_server').parse_args()

        config      = initGssw(args)
        application = NonQtApplication(config, args.name)
        server      = HalServer(config, application)
        server.start()
    finally:
        destroyAllSocketsInContext()

if __name__=='__main__':
    main()
