#! /usr/bin/env python3

""" This module is monitors house keeping (HK) values and publishes them.

    House keeping items can be registered with this module after which their
    value is updated periodically. The latest values are published
    asynchronously. House keeping variables that do not correspond to a
    physical register can also be registered but these need to be actively
    updated by the module they originate from.

    This class uses two threads. One updates the HK items and publishes them.
    The other is used to listen for configuration requests. If this would be handled
    in a single thread, timeouts from any device could propagate to the request
    latency.
"""

import logging
import os
import subprocess
from threading import Event, Thread, RLock
from datetime import datetime, timedelta
from collections import namedtuple, deque

import numpy as np

from gssw.common              import initGssw
from gssw.lib.application     import NonQtApplication
from gssw.hk.hklib            import getUnion, setUnion
from gssw.hal.halclient       import HalClient
from gssw.hal.hal             import Hal
from gssw.lib.config          import Config
from gssw.lib.device          import Device
from gssw.lib.arrays.gridded  import GriddedArray
from gssw.lib.arrays.semantic import SemanticArray
from gssw.lib.sockets         import PubSocket, RepSocket, tagName, destroyAllSocketsInContext
from gssw.lib.misc            import createGsswArgumentParser
from gssw.lib.typing          import check_class

from gssw.proto.array_pb2     import GriddedArrayMsg
from gssw.proto.common_pb2    import ErrorRep, OkRep
import gssw.proto.hk_pb2      as hkProto

logger = logging.getLogger(__name__)

# element type used in the update queue
UpdateItem = namedtuple('UpdateItem', ['name', 'timestamp'])

# collection holding the parameters of auto update hk items icluding limit state and actions
AutoParams = namedtuple('AutoParams', ['device', 'register', 'period_ms', 'engineering',
                                       'warning_action', 'warning_repeat', 'error_action',
                                       'error_repeat'])



class HkServer:
    """ This class encapsulates all house keeping monitoring functionality. """


    def __init__(self, config, app):

        check_class(config, Config)

        # connect linux signal handler
        self.signalHandler = app.getSignalHandler()
        self.signalHandler.sigInt.connect(self.stop)
        self.signalHandler.sigTerm.connect(self.stop)

        # initialize variables
        self.config = config
        self.hkValues = {}        # contains the most recent values
        self.hkAutoParams = {}    # dictionary of name - AutoParams pairs
        self.hkUnit = {}          # dictionary of name - unit pairs
        self.devices = {}         # contains device objects
        self.warningStickies = {} # dictionary of warning limit crossing sticky bits
        self.errorStickies = {}   # dictionary of error limit crossing sticky bits
        self.hkQueue = deque()    # queue of UpdateItems, sorted by timestamp, lowest left
        self.enabled = False      # determines whether HK items are actively updated
        self.lock = RLock()
        self.updateEvent = Event()
        self.enableEvent = Event()
        self.threadStopEvent = Event()

        # Get the config parameters for this component
        self.moduleConfig = config.componentParameters(app.name)

        # Get a HAL object to pass to device classes.
        self.hal       = Hal()
        self.halClient = HalClient(self.hal, self.config)
        self.halClient.requestHal()
        
        # register to halclient for HAL updates from the server
        self.halClient.registerCallBackFunction(self.reloadHal)

        # Set up the REP socket for control and get/set requests
        self.repSocket = RepSocket()
        self.repSocket.createZmqSocket()
        self.repSocket.bindSocket('tcp://*:%d' % self.moduleConfig['sockets']['hk-rep'])

        # Set up the PUB socket for asynchronous HK updates.
        self.pubSocket = PubSocket()
        self.pubSocket.createZmqSocket()
        self.pubSocket.bindSocket('tcp://*:%d' % self.moduleConfig['sockets']['hk-pub'])
        
        # subscribe to the proxy (for plotting) iff it is defined in the config file
        try:
            proxyConfig = config.componentParameters('proxy')
            proxyAddress = 'tcp://%s:%d' % (proxyConfig['hostname'], proxyConfig['sockets']['xsub'])
            self.pubSocket.connectSocket(proxyAddress)
            logger.info('Connected HK server to proxy @ %s', proxyAddress)
        except:
            logger.warning(
                'Cannot plot HK data because the proxy is not defined in the configuration')

        # Thread (placeholder) that gets values and publishes updates
        self.updateThread = None

    def reloadHal(self):
        """ Called when the HAL client fires a signal """
        # reload the HAL
        logger.info("ReloadHal requested by HKServer")
        self.halClient.requestHal()

    def run(self):
        """ Main loop (blocking). Also starts a thread that gets values and publishes updates. """
        self.updateThread = Thread(target=self.updateHkItems)
        logger.info('Starting HK server update thread')
        self.updateThread.start()

        # handle requests
        self.receiveRequest()


    def stop(self):
        """ Stops background update thread. """
        self.halClient.unregisterCallBackFunction() #Stop Hal Updater Callback loop
        logger.info('Stopping HK server update thread')
        self.threadStopEvent.set() # Set flag stop background thread
        self.enableEvent.set()     # Within the thread, there are some wait()'s for other flags
        self.updateEvent.set()     #  so set those as well
        if self.updateThread:
            self.updateThread.join()
        self.updateThread = None
                
    def insertQueue(self, name, timestamp):
        """ Insert an hk item into the update queue.

            name      - the name of the hk item that is inserted
            timestamp - timestamp of the next update of the item

            The update queue is sorted by timestamp from low to high.
        """

        with self.lock:

            # Insert the updated item in the empty queue.
            if not self.hkQueue:
                self.hkQueue.append(UpdateItem(name, timestamp))
                logger.debug('(a) new queue: %s', str(self.hkQueue))
            else:
                # Insert the item back in the sorted queue with the new timestamp.
                index = 0
                for (_, qtimestamp) in self.hkQueue:
                    if timestamp < qtimestamp:
                        # Pylint has a false positive for collections.deque.insert
                        # see: https://github.com/PyCQA/pylint/issues/1476
                        self.hkQueue.insert(index, UpdateItem(name, timestamp)) #pylint: disable=no-member
                        logger.debug('(b) new queue: %s', str(self.hkQueue))
                        return

                # Insert the item at the end of the queue
                self.hkQueue.append(UpdateItem(name, timestamp))
                logger.debug('(c) new queue: %s', str(self.hkQueue))


    def roundTimestamp(self, periodMs):
        """ round up the current time to the next nearest whole second or minute.

            periodMs - rounding is performed based on whether the period is a multiple of
                       minutes, seconds or milliseconds
        """

        timeStart = datetime.now()

        if periodMs % 60000 == 0:
            timeStart -= timedelta(seconds=timeStart.second, microseconds=timeStart.microsecond)
            timeStart += timedelta(minutes=1)
        else:
            timeStart -= timedelta(microseconds=timeStart.microsecond)
            timeStart += timedelta(seconds=1)

        return timeStart


    def receiveRequest(self):
        """ Wait for requests and dispatch the handler. """

        while True: 
            if self.repSocket.poll(timeout=None):
                self.handleRequest()
            
            
    def getOutdatedItems(self):
        """ Return a list of hk item names that need to be updated """

        now = datetime.now()
        updateItems = []

        # get immanent HK updates and their properties using a threading lock
        with self.lock:

            if not self.hkQueue: # check if queue empty
                return [] # wait indefinitely until something is put in the queue

            # pop outdated hk items from the queue into an update list
            while self.hkQueue and self.hkQueue[0].timestamp < now:
                updateItems.append(self.hkQueue.popleft())

            if updateItems == []:
                return []

            # re-insert the items into the queue with a new timestamp
            for name, timestamp in updateItems:
                tDelta = timedelta(milliseconds=self.hkAutoParams[name].period_ms)
                while timestamp + tDelta < now:
                    # this is in case HK has been disabled for some time
                    tDelta += timedelta(milliseconds=self.hkAutoParams[name].period_ms)
                self.insertQueue(name, timestamp + tDelta)

        return [x.name for x in updateItems]


    def getItemValues(self, names):
        """ Get the values of the hk items in list names.

            This function groups the items per device and updates them in a single call per device.
            returns a dict of name, value pairs
        """

        groups = {}
        values = {}

        # group names according to device
        for name in names:
            dev = self.hkAutoParams[name].device
            if dev not in groups:
                groups[dev] = [name]
            else:
                groups[dev].append(name)

        logger.debug('updating hk registers: %s', str(groups))

        # perform the updates for each device
        for dev in groups:

            registers = [self.hkAutoParams[x].register for x in groups[dev]]

            try:
                deviceValues = self.devices[dev].getMultiple(registers)
            except Exception as e:
                logger.error(
                    'Communication error getting HK values from device %s:\n %s', dev, e)
                # Nan values indicate an error
                deviceValues = [float('nan')] * len(groups[dev])

            else:
                # convert to engineering values
                for i, name in enumerate(groups[dev]):
                    if self.hkAutoParams[name].engineering:

                        # get the register name without offset
                        registerName = self.hkAutoParams[name].register.split(':')[0]
                        conversion = None

                        try:
                            conversion = self.hal.getRegisterConversion(
                                dev, None, None, registerName)
                        except Exception as e:
                            logger.error(
                                'Error getting conversion for register %s:%s:\n %s',
                                dev, registerName, e)

                        if conversion:
                            deviceValues[i] = conversion.applyConversion(deviceValues[i])
                        else:
                            deviceValues[i] = float('nan') # Nan value indicates an error
                            logger.warning('Device %s has no conversion for register'
                                           ' %s, cannot send the engineering value',
                                           dev, registerName)
            finally:
                for name, value in zip(groups[dev], deviceValues):
                    values[name] = value

        return values


    def checkLimits(self, valueDict):
        """ Check limit crossings, perform actions if defined and update sticky bits. """

        for name, value in valueDict.items():

            # limits are only applicable on hw items
            if name not in self.hkAutoParams:
                continue

            # get limits from hal
            device = self.hal.getDevice(self.hkAutoParams[name].device)
            limits = self.hal.getRegisterLimits(
                device.name, device.deviceType, device.calibration,
                self.hkAutoParams[name].register.split(':')[0]) # remove offset for HAL

            if limits is None:
                continue

            # check warning limit crossing
            if not limits.lowerWarn <= value <= limits.upperWarn:

                self.warningStickies[name] = True

                if (self.hkValues[name] > limits.upperWarn and
                        (self.hkAutoParams[name].warning_repeat or not self.warningStickies[name])):

                    logger.warning('upper warning limit crossed for hk item %s', name)
                    self.performLimitAction(self.hkAutoParams[name].warning_action)

                if (self.hkValues[name] < limits.lowerWarn and
                        (self.hkAutoParams[name].warning_repeat or not self.warningStickies[name])):

                    logger.warning('lower warning limit crossed for hk item %s', name)
                    self.performLimitAction(self.hkAutoParams[name].warning_action)

            # check error limit crossings
            if not limits.lowerError <= value <= limits.upperError:

                self.errorStickies[name] = True

                if (self.hkValues[name] > limits.upperError and
                        (self.hkAutoParams[name].error_repeat or not self.errorStickies[name])):

                    logger.error('upper error limit crossed for hk item %s', name)
                    self.performLimitAction(self.hkAutoParams[name].error_action)

                if (self.hkValues[name] < limits.lowerError and
                        (self.hkAutoParams[name].error_repeat or not self.errorStickies[name])):

                    logger.error('lower error limit crossed for hk item %s', name)
                    self.performLimitAction(self.hkAutoParams[name].error_action)


    def performLimitAction(self, command):
        """ Perform a limit action in a safe way """

        # limit checking is always performed even if not actions are defined
        if command is None:
            return

        try:
            # execute action in new thread
            subprocess.Popen(command.split(' '), preexec_fn=os.setpgrp)
        except Exception as e:
            logger.error('error during limit crossing action: %s', str(e))


    def updateHkItems(self):
        """ Update outdated hk items (runs in thread).

            Multiple reads are combined if they are scheduled at the same time.
        """

        waitTime = 0

        while not self.threadStopEvent.is_set():
            try:
                # wait here until housekeeping is enabled
                self.enableEvent.wait()

                # check time it will take for the next hk item to expire
                if not self.hkQueue: # check if queue empty -> wait indefinitely
                    waitTime = None
                elif self.hkQueue[0].timestamp < datetime.now(): # don't wait
                    waitTime = 0
                else: # wait until next item expires
                    waitTime = (self.hkQueue[0].timestamp - datetime.now()).total_seconds()

                # wait until the next event or until the event queue is updated
                self.updateEvent.wait(waitTime)
                self.updateEvent.clear()

                # get all items from the queue that are outdated
                names = self.getOutdatedItems()

                # get values of outdated items from devices
                valueDict = self.getItemValues(names)

                # store values internally
                for name, value in valueDict.items():
                    self.hkValues[name] = value

                # check limit crossings and execute actions if necessary
                self.checkLimits(valueDict)

                # Publish all updates at once
                self.publishUpdates(valueDict)

            except Exception as e:
                # Prevent the update thread's main loop from crashing in case of exceptions
                logger.exception('An exception occurred in the update thread: %s', e)


    def publishUpdates(self, valueDict):
        """ Publish an array of HK values.

            names  - a list of names of hk items to publish
            values - a list of corresponding hk values. None if an error occurred.
        """

        for name, value in valueDict.items():

            stickyBits = {'warning': self.warningStickies[name],
                          'error':   self.errorStickies[name]}

            dtype = ('<f8' if isinstance(value, float) else
                     '<i8' if isinstance(value, int) else
                     '<a8')

            data = np.array(value, dtype=dtype)
            gArray = GriddedArray([data], name=('HK_%s' % name), notes=repr(stickyBits))
            gArray.coords[0] = SemanticArray(
                np.array([np.datetime64(datetime.now())], dtype='datetime64[ms]'))
            pBuffer = gArray.asProtoBuf()

            if name in self.hkAutoParams:
                topic = '/hk/%s/%s' % (self.hkAutoParams[name].device, name)
            else:
                topic = '/hk/%s' % name


            self.pubSocket.publishProtoBuf(
                topic, tagName(GriddedArrayMsg),pBuffer.SerializeToString())

        logger.debug('auto update HK items %s', str(valueDict.keys()))


    def handleRequest(self):
        """ Handle a request received on the REP socket.

            The request is either a control, data get/set or hk item request.
            This function dispatches the corresponding handling function.
        """

        # receive the reply including the header string
        parts = self.repSocket.receiveProtoBuf()

        logger.debug('received: %s', parts.tag)

        # check the header string is to determine the request type
        if parts.tag == tagName(hkProto.HkItemReq):
            req = hkProto.HkItemReq()
            req.ParseFromString(parts.message)
            self.addHkItem(req)

        elif parts.tag == tagName(hkProto.HkRemoveItemsReq):
            req = hkProto.HkRemoveItemsReq()
            req.ParseFromString(parts.message)
            self.removeHkItems(req)

        elif parts.tag == tagName(hkProto.HkGetDataReq):
            req = hkProto.HkGetDataReq()
            req.ParseFromString(parts.message)
            self.sendHkValues(req)

        elif parts.tag == tagName(hkProto.HkGetAllDataReq):
            self.sendAllHkValues()

        elif parts.tag == tagName(hkProto.HkSetDataReq):
            req = hkProto.HkSetDataReq()
            req.ParseFromString(parts.message)
            self.recvHkValue(req)

        elif parts.tag == tagName(hkProto.HkControlReq):
            req = hkProto.HkControlReq()
            req.ParseFromString(parts.message)
            self.handleControlRequest(req)

        elif parts.tag == tagName(hkProto.HkGetItemsReq):
            self.sendHkItems()

        elif parts.tag == tagName(hkProto.HkResetLimitsReq):
            req = hkProto.HkResetLimitsReq()
            req.ParseFromString(parts.message)
            self.resetLimits(req)

        else:
            # Unknown request type, send an error reply.
            logger.error('unknown protobuf request received: %s', parts.tag)
            self.sendErrorRep('Unknown message type received: %s' % parts.tag)


    def resetLimits(self, req):
        """ Reset the limit crossing sticky bits for one or more hk items

            :param message: a HkResetLimitsReq protobuf object
        """

        with self.lock:
            for name in req.name:
                logger.info('clearing limits of HK item %s', name)
                self.warningStickies[name] = False
                self.errorStickies[name] = False

        # Send reply.
        rep = OkRep()
        self.repSocket.sendProtoBuf(tagName(OkRep), rep.SerializeToString())


    def sendHkItems(self):
        """ Send a list of all registered hk items.

            req - a HkGetItemsReq protobuf object

            This funtion sends a HkGetItemsRep containing all registered items.
        """

        rep = hkProto.HkGetItemsRep()

        # step through all hk items
        for itemName in self.hkValues:

            # Add an item to the protobuf and fill the fields
            hkItem = rep.item.add()
            hkItem.name = itemName
            if itemName in self.hkAutoParams: # optional fields
                item = self.hkAutoParams[itemName]
                hkItem.auto.address = item.device + ':' + item.register
                hkItem.auto.period_ms = item.period_ms
                hkItem.auto.engineering = item.engineering
            if itemName in self.hkUnit: # optional unit
                hkItem.unit = self.hkUnit[itemName]

        self.repSocket.sendProtoBuf(tagName(hkProto.HkGetItemsRep), rep.SerializeToString())


    def addHkItem(self, req):
        """ Add a new HK item.

            req - an HkItemReq protobuf object
        """

        # Check if the name is not already in use.
        if req.name in self.hkValues:
            self.sendErrorRep('Hk item name already exists (%s)' % req.name)
            return

        # Create the new item.
        self.hkValues[req.name]        = None
        self.warningStickies[req.name] = False
        self.errorStickies[req.name]   = False

        # Optionally set the item's unit.
        if req.HasField('unit'):
            self.hkUnit[req.name] = req.unit

        # Set the items parameters if the item is register based
        if req.HasField('auto'):

            # Split the device and register name.
            deviceName = req.auto.address.split(':', 1)[0]
            registerName = req.auto.address.split(':', 1)[1]

            # Get limit crossing action from protobuf. Set to None if they are not defined
            warningAction = req.warning_action.action if req.HasField('warning_action') else None
            warningRepeat = req.warning_action.repeat if req.HasField('warning_action') else False
            errorAction   = req.error_action.action   if req.HasField('error_action') else None
            errorRepeat   = req.error_action.repeat   if req.HasField('error_action') else False

            # Store the parameters as a named tuple.
            # No lock needes here because this only adds an element atomically
            self.hkAutoParams[req.name] = AutoParams(deviceName, registerName, req.auto.period_ms,
                req.auto.engineering, warningAction, warningRepeat, errorAction, errorRepeat)

            # Create a device object if it does not already exist.
            if deviceName not in self.devices:
                self.devices[deviceName] = Device(deviceName, self.config,
                    hal=self.hal, timeout=5)

            # The next update timestamp is shifted to the next whole ms, s, or min.
            # This facilitates updating multiple items at the same time.
            self.insertQueue(req.name, self.roundTimestamp(req.auto.period_ms))

            # signal the update thread that something has changed
            self.updateEvent.set()

        # Send reply.
        rep = OkRep()
        self.repSocket.sendProtoBuf(tagName(OkRep), rep.SerializeToString())

        # Publish state change
        self.sendHkStateChangedPub()

        logger.info('Added HK item %s', req.name)


    def removeHkItems(self, req):
        """ Remove one or multiple HK items.

            req - an HkRemoveItemsReq protobuf object
        """

        with self.lock:
            for name in req.name:

                if name in self.hkValues:
                    del self.hkValues[name]

                if name in self.hkAutoParams:
                    del self.hkAutoParams[name]

                if name in self.hkUnit:
                    del self.hkUnit[name]

                for i, (n, _) in enumerate(self.hkQueue):
                    if n == name:
                        del self.hkQueue[i]
                        break
        # Send reply.
        rep = OkRep()
        self.repSocket.sendProtoBuf(tagName(OkRep), rep.SerializeToString())

        # Publish state change
        self.sendHkStateChangedPub()

        logger.info('Removed HK items %s', req.name)


    def sendAllHkValues(self):
        """ Send the value of all registered hk items. """

        logger.info('Get all HK values')

        # Create a protobuf reply object.
        rep = hkProto.HkGetAllDataRep()

        # Iterate through all hk items.
        for hkitem in self.hkValues:

            # Create a new name value pair.
            repitem = rep.item.add()

            # Fill in the name.
            repitem.name = hkitem

            # Fill in the value.
            setUnion(repitem, self.hkValues[hkitem], logger)

        # Send the reply.
        self.repSocket.sendProtoBuf(tagName(hkProto.HkGetAllDataRep), rep.SerializeToString())


    def sendHkValues(self, req):
        """ Send the latest value of a HK item.

            req - an HkDataGetReq protobuf object
        """

        # Create the reply message.
        rep = hkProto.HkData()

        for name in req.name:

            # Check if the hk item exists.
            if name not in self.hkValues:
                errorMsg = 'HK get request received for an unregistered'
                errorMsg += ' hk item. (%s)' % name
                self.sendErrorRep(errorMsg)
                return

            # Add the value to the reply message.
            item = rep.item.add()
            setUnion(item, self.hkValues[name], logger)

        # Send the reply.
        self.repSocket.sendProtoBuf(tagName(hkProto.HkData), rep.SerializeToString())


    def recvHkValue(self, req):
        """ Update the hk value from an external source.

            req - an HkSetDataReq protobuf object
        """

        # It is not allowed to set the value of hk items that are automatically updated.
        if req.name in self.hkAutoParams:
            errorMsg = 'HK set request received for an item '
            errorMsg += 'that is automatically updated (%s)' % req.name
            self.sendErrorRep(errorMsg)
            return

        # Check if the hk item exists.
        if req.name not in self.hkValues:
            errorMsg = 'HK set request received for an unregistered'
            errorMsg += ' hk item. (%s)' % req.name
            self.sendErrorRep(errorMsg)

        rep = OkRep()

        # Try to update the value. Send an error when the item is not registered.
        value = getUnion(req)
        if value is None:
            logger.warning('No value found in HkSetDataReq object')
            rep = ErrorRep()
        else:
            self.hkValues[req.name] = value

        if not isinstance(rep, ErrorRep):
            logger.debug('Set of HK item %s to value %s',
                              req.name, str(self.hkValues[req.name]))

        # Send the reply message.
        self.repSocket.sendProtoBuf(tagName(OkRep), rep.SerializeToString())

        # Publish the HK update
        self.publishUpdates({req.name: value})

        logger.debug('update HK item %s = %f', req.name, float(self.hkValues[req.name]))


    def handleControlRequest(self, req):
        """ Handle a control request.

            req - an HkControlReq protobuf object
        """
        # Create reply message
        rep = hkProto.HkControlRep()

        if self.enabled:
            rep.state = hkProto.HK_RUNNING
        else:
            rep.state = hkProto.HK_STOPPED

        # Clear all registered hk items (devices are kept open.
        # removing them here leads to threading issues)
        if req.command == hkProto.HK_CLEAR:
            with self.lock:
                self.hkValues = {} # contains the most recent values
                self.hkAutoParams = {} # contains auto update parameters
                self.hkUnit = {} # contains units
                self.hkQueue = deque() # sorted list tuples of hk item names and update timestamps
                logger.info('Cleared all HK items')

        # Start housekeeping
        elif req.command == hkProto.HK_START:
            for i, (name, _) in enumerate(self.hkQueue):
                # Update outdated timestamps
                timeStart = self.roundTimestamp(self.hkAutoParams[name].period_ms)
                with self.lock:
                    self.hkQueue[i] = UpdateItem(name, timeStart)
            self.enabled = True
            self.updateEvent.set()
            self.enableEvent.set()
            logger.info('Housekeeping started')

        elif req.command == hkProto.HK_STOP:
            self.enabled = False
            self.enableEvent.clear()
            logger.info('Housekeeping stopped')

        # send ACK
        self.repSocket.sendProtoBuf(tagName(hkProto.HkControlRep), rep.SerializeToString())

        # notify all clients of change
        self.sendHkStateChangedPub()


    def sendErrorRep(self, msg):
        """ Send an errorRep message in response to an incorrect request. """

        logger.warning(msg)
        rep = ErrorRep()
        rep.errtype = ErrorRep.ERR_PROTO
        rep.errmsg = msg
        self.repSocket.sendProtoBuf(tagName(ErrorRep), rep.SerializeToString())


    def sendHkStateChangedPub(self):
        """ Publish a message to inform all clients they need to sync with the server. """

        pub = hkProto.HkStateChangedPub()
        if self.enabled:
            pub.state = hkProto.HK_RUNNING
        else:
            pub.state = hkProto.HK_STOPPED

        self.pubSocket.publishProtoBuf(tagName(hkProto.HkStateChangedPub), # topic
                                       tagName(hkProto.HkStateChangedPub), # tag
                                       pub.SerializeToString())            # message


def main():
    """ Parse command line arguments and run this module. """

    args = createGsswArgumentParser(__doc__, enableOptionName=True,
                                    optionNameDefault='hk_server').parse_args()

    config = initGssw(args)
    application = NonQtApplication(config, args.name)
    server = HkServer(config, application)

    # Start the HK manager
    try:
        server.run()
    finally:
        server.stop() # wait for background thread to stop.
        destroyAllSocketsInContext()


if __name__ == '__main__':
    main()
