#!/usr/bin/env python3

""" GSSw Event Logger Server

"""

###################################################################################################
# P Y T H O N   M O D U L E S
###################################################################################################

import os.path
import logging
import logging.handlers
import sys

from datetime                 import datetime
from gssw.common              import initGssw
from gssw.lib.config          import Config, outputDirectory
from gssw.lib.sockets         import PubSocket, SubSocket, tagName
from gssw.lib.misc            import createGsswArgumentParser
from gssw.lib.typing          import check_class
from gssw.lib.errors          import CommunicationError
from gssw.proto.logger_pb2    import SyncMsg, LoggerMsg

VERBOSE = 0
def vprint(*args, **kwargs): 
    """ Print if VERBOSE = 1 or do nothing. """
    if VERBOSE:
        print(*args, file=sys.stderr, flush = True, **kwargs)


###################################################################################################
# E L S E R V E R
###################################################################################################

class LogServer():
    """ 
    Event Logger Server, takes care of event logging from several clients (the loggers). 
    The logserver is subscribed to the PUB sockets called by the loggers. 
    Apart from this, the logserver has its own PUB socket for sending synchronization messages to 
    all loggers (handshaking). The handshaking is meant to mitigate the 'slow subscriber' behaviour 
    annex to the ZMQ PUB-SUB pattern.
    """

    def __init__(self, name, config):
        """ Constructor. """   
        vprint("(logserver) Creating LogServer. Working dir: %s @ %s" % (os.getcwd(),
            str(datetime.now().time())))  

        check_class(config, Config)

        # get the module configuration
        moduleCfg     = config.componentParameters(name)
        fileformat    = moduleCfg['format']
        if 'filename' in moduleCfg:
            filename      = config.asAbsolutePath(moduleCfg['filename'])
        else:
            filename = os.path.join(outputDirectory(), 'events.log')

        filemode      = moduleCfg['filemode']
        self.host     = moduleCfg['hostname']
        self.pubPort  = moduleCfg['sockets']['log-rep']
        self.subPort  = moduleCfg['sockets']['log-pub']

        try:
            filesize  = moduleCfg['filesize']
        except KeyError:
            filesize  = 25 * 1024 * 1024        # default log file 25 MByte before rotation

        try:
            filecount = moduleCfg['filecount']  
        except KeyError:
            filecount = 5                       # default max 5 rotated log files

        # Check if log directory exists, otherwise create it
        logDir = os.path.dirname(filename)
        if not os.path.exists(logDir):
            vprint("(logserver) creating directory for log file: %s" % logDir)
            os.mkdir(logDir)

        vprint("(logserver) opening file: %s @ %s" % (filename, str(datetime.now().time())))
        
        # logger object only used for printing the received log records into a file
        fileHandler = logging.handlers.RotatingFileHandler(filename, 
                                                           mode        = filemode, 
                                                           maxBytes    = filesize, 
                                                           backupCount = filecount)
        formatter = logging.Formatter(fileformat)
        fileHandler.setFormatter(formatter)  
        fileHandler.setLevel(logging.DEBUG)              
        self.slog = logging.getLogger(__name__)           
        self.slog.addHandler(fileHandler) 
        self.slog.setLevel(logging.DEBUG) 
        self.slog.propagate = False                                
                 
        # create a SUB socket for handshaking and message logging 
        self.subSock = SubSocket()
        self.subSock.createZmqSocket(timeoutMs=-1)

        # create a PUB socket for handshaking only
        self.pubSock = PubSocket()
        self.pubSock.createZmqSocket(timeoutMs=-1)
                                
    def start(self):
        """ Worker function. """ 

        # bind the SUB socket       
        self.subSock.subscribe(topic="")          
        self.subSock.bindSocket('tcp://*:%d' % self.subPort)    
        vprint('(logserver) Binding SUB socket to *:%d @ %s' % (self.subPort, 
            str(datetime.now().time())))         

        # bind the PUB socket                       
        self.pubSock.bindSocket('tcp://*:%d' % self.pubPort)     
        vprint('(logserver) Binding PUB socket to *:%d @ %s' % (self.pubPort, 
            str(datetime.now().time())))
                
        while True:
            vprint('(logserver) Waiting for a message @ %s' % str(datetime.now().time()))
            message = self.subSock.receiveProtoBufWithTopic(flags=0)
            
            # first determine the message type
            if message.tag == 'SyncMsg':
                # received message is a synchronization message (type 'SyncMsg')
                # first parse the message
                source = SyncMsg()
                source.ParseFromString(message.message)
                vprint('(logserver) SyncMsg received from %s @ %s' % (source.name, 
                    str(datetime.now().time())))                

                # after that, send the same message to the PUB socket (handshaking)
                reply = SyncMsg()
                reply.name = source.name
                protoString = reply.SerializeToString() 
                self.pubSock.publishProtoBuf("", tagName(SyncMsg), protoString, flags=0) 
                vprint('(logserver) handshake complete for %s @ %s' % (source.name, 
                    str(datetime.now().time())))                                                                     
                                       
            elif message.tag == 'LoggerMsg':
                # received message is a logging report (type 'LoggerMsg')
                # first parse the message
                report = LoggerMsg()
                report.ParseFromString(message.message) 
                vprint('(logserver) LoggerMsg received from %s @ %s' % (report.pathname, 
                    str(datetime.now().time())))                

                # after that, re-compose the log record which will be stored in the log file             
                mylogRecord = logging.LogRecord(
                    report.name, 
                    report.level, 
                    report.pathname, 
                    report.lineno, 
                    report.msg, 
                    None, 
                    None, 
                    report.func, 
                    report.sinfo) 
                mylogRecord.created = report.created
                mylogRecord.msecs = report.msecs  
                mylogRecord.process = report.processid
                mylogRecord.func = report.func
                mylogRecord.ipaddress_port = report.ipaddress_port   
                mylogRecord.script = report.componentname  
                
                # finally, write the log record into the log file                       
                self.slog.handle(mylogRecord) 
                vprint('(logserver) printing to slog @ %s' % str(datetime.now().time()))
                                
            else:
                # message type is unknown
                vprint('(logserver) received unknown message format @ %s' % 
                    str(datetime.now().time()))
                raise CommunicationError('received unknown message format')
        
                                    
###################################################################################################
# M A I N
###################################################################################################

def main():
    """ Main function. """   
    
    args = createGsswArgumentParser(__doc__, enableOptionName=True).parse_args()
    config = initGssw(args, logToSocketHandler=False)
    server = LogServer(args.name, config)
    server.start()


if __name__ == "__main__":
    main()    
