#! /usr/bin/env python3

"""
    Proxy daemon, provides a data publishing bus

    The proxy creates and binds two sockets (XSUB and XPUB) with configured port numbers,
    to which publishers and subscribers connect (respectively).
    All data received of the XSUB socket is forwarded via the XPUB socket
    (and thus to all subscribers).

    This way, the proxy closely resembles a bus architecture.
"""

import logging
import zmq

from gssw.common              import initGssw
from gssw.lib.typing          import check_class
from gssw.lib.misc            import createGsswArgumentParser

logger = logging.getLogger(__name__)

def main():
    """ Do what a daemon might do. """

    # Set up command line argument parsing
    args = createGsswArgumentParser(__doc__, enableOptionName=True,
                                    optionNameDefault='proxy').parse_args()

    config = initGssw(args)


    # Try to print the configuration parameters for this daemon
    proxyCfg = config.componentParameters(args.name)
    logger.debug(proxyCfg)

    # Set up the proxy
    ctx = zmq.Context.instance()
    xsub_socket = ctx.socket(zmq.XSUB)
    xpub_socket = ctx.socket(zmq.XPUB)

    xsub_port = proxyCfg['sockets']['xsub']
    xpub_port = proxyCfg['sockets']['xpub']
    check_class(xsub_port, int)
    check_class(xpub_port, int)

    xsub_socket.bind("tcp://*:{}".format(xsub_port))
    xpub_socket.bind("tcp://*:{}".format(xpub_port))

    # Start the proxy
    logger.info("Proxy is running (XSUB: %d, XPUB: %d)", xsub_port, xpub_port)
    zmq.proxy(xsub_socket, xpub_socket)
    logger.info("Proxy has stopped")

if __name__ == '__main__':
    main()
