#! /usr/bin/env python3

"""
    The runGSSw module is used to start and stop the different components of
    the GSSw including HW daemons and other programs which are specific to a
    setup. Additionally, this program also starts a configServer instance from
    which each program can request the configuration file containing the
    setup's configuration parameters.

    .. program-output:: gssw_run.py -h
"""

import logging
import os.path
import signal
import socket
import subprocess
import threading
import time
import sys

from collections import OrderedDict

import paramiko
import psutil
import yaml

from gssw.lib.config          import outputDirectory
from gssw.lib.colorprint      import ColorPrint
from gssw.lib.errors          import ConfigError
from gssw.lib.sockets         import destroyAllSocketsInContext
from gssw.config.configclient import ConfigClient
from gssw.lib.misc            import createGsswArgumentParser, normRealPath

logger = logging.getLogger("gssw.run_gssw")

class RunGSSw:
    """ This class encapsulates all functions to start and stop the GSSw.

        It can be used either via the command line executable gssw_run, or imported and used as a
        regular object in other modules. In the latter case, a thread lock ensures that the
        .pids file is atomic.
    """

    def __init__(self, configPath, verbose=True):
        """ Make a controller for starting/stopping GSSw components.

        :param str configPath: GSSw configuration file.
        :param bool verbose:   Print component status to stdout.
        """
        self.configPath = configPath
        self.hostname = socket.gethostname()
        self.colorPrinter = ColorPrint()
        self.pidsFile = os.path.join(outputDirectory(), 'gssw_run.pids')
        self.verbose  = verbose

        # get a configuration client
        try:
            configClient = ConfigClient(self.configPath)
        except ConfigError as e:
            sys.exit(e)

        # retrieve the system configuration
        self.config = configClient.config

        # set-up logging
        self.config.initLogging(logToSocketHandler=False)

        # get the config server port from the config file and create the socket address
        self.configHost = self.config.componentParameters('config_server')['hostname']
        if self.configHost == 'localhost' or self.configHost == '127.0.0.1':
            self.configHost = socket.gethostname()
        self.configPort = self.config.componentParameters('config_server')['sockets']['config-rep']
        self.configSocket = 'tcp://' + self.configHost + ':' + str(self.configPort)


        # When the RunGSSw is used as a module, a thread lock is required to ensure that the pids
        # file is not changed simultaneously from concurring threads
        self._thread_lock = threading.RLock()

        # Make sure the output directory exists so we can load/create a PIDS file
        self.cmdCreateDirPath(outputDirectory())

        self.pids = {}
        self.lockFile = None # Path to pids lock file, only set when there is a lock file in-place
        self.loadPidsFile()


    def loadPidsFile(self):
        """ (Re)load the file containing the process list """
        with self._thread_lock:
            try:
                self._acquirePidsLock()
                # load the pids file which holds the active processes pids and hosts
                try:
                    with open(self.pidsFile, "r") as pidsFile:
                        self.pids = yaml.load(pidsFile, Loader=yaml.SafeLoader)
                except yaml.YAMLError as ex:
                    logger.warning("Can't read PIDS file: %s", ex)
                    self.pids = {}
                except FileNotFoundError as ex:
                    logger.debug("Assuming that Egress is not running, no PID file exists: %s", ex)
                    self.pids = {}
            finally:
                self.releasePidsLock()


    def execute(self, command, components=None):
        """ Executes a command (e.g. start, stop)

            :param str command:     The command to execute [start, stop, restart or status].
            :param list components: An optional list of components to execute the command on.
        """

        # Check if specified component names are valid
        if components is not None:
            for componentName in components:
                if componentName not in self.config['components']:
                    logger.error("Component '%s' does not exist, aborting...", componentName)
                    # Stop on invalid components
                    return

        with self._thread_lock:
            logger.debug('Starting command: %s', command)

            # perform the action specified by the command line argument
            if command == 'start':
                self.startModules(components)
                time.sleep(2)
                self.getStatus()

            elif command == 'status':
                self.getStatus()

            elif command == 'stop':
                self.stopModules(components)
                time.sleep(2)
                self.getStatus()

            elif command == 'restart':
                self.stopModules(components)
                self.startModules(components)
                time.sleep(2)
                self.getStatus()

            else:
                raise ValueError('Invalid command: %s' % command)

            # write the active modules and their PIDs to a file
            self.writePidsFile()


    def startModules(self, selection=None):
        """ Start the modules listed in the config file on their respective hosts. """

        # make a sorted dict containing all components that must be started
        components = self.selectComponents(selection)

        # start the config server first, then the logserver and then the rest
        if 'log_server' in components:
            components.move_to_end('log_server', False)
        if 'config_server' in components:
            components.move_to_end('config_server', False)

        for name, parameters in components.items():

            # Only start components that are enabled
            # Also allow starting disabled component with -c ... start option
            if parameters['enable'] or selection is not None:

                # check if the process is already running
                if self.cmdIsRunning(name, parameters):
                    logger.info('Component %s is already running on %s',
                        name, parameters['hostname'])
                    continue

                logger.info('Starting component %s on host %s',
                    name, parameters['hostname'])

                # concatenate the executable with possible command line arguments
                cmd = parameters['exe']

                if name == 'config_server':
                    cmd += ' --config-path {} --port {}'.format(self.configPath, self.configPort)
                else:
                    # concatenate the configuration client command line arguments
                    cmd += ' --name {} --config-path {}'.format(name, self.configSocket)

                # concatenate arguments from the config file
                if 'args' in parameters:
                    cmd = cmd + ' ' + parameters['args']

                # create the path to the log file
                logFileName = self.logFileName(name, parameters)
                logDirName  = os.path.dirname(logFileName)

                self.cmdCreateDirPath(logDirName)
                self.cmdEnsureFileExists(logFileName) # Is this still necessary?
                pid, _output = self.performCommand(parameters, cmd, outputFile=logFileName)

                # Add the process to the pids file
                if pid != -1:
                    self.pids[name] = {'host' : parameters['hostname'],
                                       'pid'  : pid}


    def stopModules(self, selection=None):
        """ Stop the GSSw modules and daemons listed in the config file. """

        if not os.path.isfile(self.pidsFile):
            logger.info('No active processes to stop')
            return

        # make a sorted dict with the components of the system
        components = self.selectComponents(selection)

        if 'config_server' in components:
            components.move_to_end('config_server', True)
        if 'log_server' in components:
            components.move_to_end('log_server', True)

        # first politely ask all components to stop
        for name, parameters in components.items():
            if name in self.pids:

                host = self.pids[name]['host']
                pid = int(self.pids[name]['pid'])

                # check if the process was already stopped
                if not self.cmdIsRunning(name, parameters):
                    continue

                self.cmdKillProcess(pid)

                logger.info('Stopping daemon %s on host %s', parameters['exe'], host)

        # give the processes some time to exit cleanly
        logger.info("Waiting 1 second for the components to quit.")
        time.sleep(1)

        # Check if any component is still running and force them to quit
        for name, parameters in components.items():
            if name in self.pids:

                pid = int(self.pids[name]['pid'])

                if self.cmdIsRunning(name, parameters):
                    logger.warning("Component %s (PID=%s) still running. Using kill -9 now...",
                                   name, pid)
                    self.cmdKillProcess(pid, killSignal=signal.SIGKILL)

                del self.pids[name]


    def selectComponents(self, selection=None):
        """ Get a dictionary with a sub selection of the components.

            :param list selection: A list of names of components to select. If selection is None
                                   all components are returned.
        """

        components = OrderedDict(sorted(self.config['components'].items()))

        if selection is not None:
            # Create a subset of the components
            keys = set(selection).intersection(components)
            return OrderedDict({k: components[k] for k in keys})

        return components


    def getStatus(self):
        """ Print whether the daemons in the config file are running. """

        # make a sorted dict with the components of the system
        components = OrderedDict(sorted(self.config['components'].items()))

        for name, parameters in components.items():

            prefix = '{0:<15s} @ {1:s}'.format(name, parameters['hostname'])

            if self.cmdIsRunning(name, parameters):
                if self.verbose:
                    print('{0:<40s}: {1}'.format(prefix, self.colorPrinter.green('running')))
            else:
                if not parameters['enable']:
                    if self.verbose:
                        print('{0:<40s}: {1}'.format(prefix, self.colorPrinter.yellow('disabled')))
                else:
                    if name in self.pids:
                        del self.pids[name]
                    if self.verbose:
                        print('{0:<40s}: {1}'.format(prefix, self.colorPrinter.red('not running')))



    def cmdIsRunning(self, name, parameters):
        """ Check if a component is running on its host.

            This function checks if the component is still running under the PID it was started.

            :param str name:        The name of the component
            :param dict parameters: The config parameters for the component
        """
        with self._thread_lock:
            if self.pids == {}:
                return False

            if name not in self.pids:
                return False # The component was never started or was already stopped

            pid = self.pids[name]['pid']
            exeName = parameters['exe']

            if self._componentIsLocal(parameters):
                logger.debug("Checking if process is still running: PID=%s", pid)
                try:
                    process = psutil.Process(pid)
                except psutil.Error as ex:
                    # Catch psutils exceptions such as NoSuchProcess. This can happen if a
                    # component was closed without using run_gssw and is still listed in the PID
                    # file
                    logger.debug("Exception while checking component %s is running: %s", name, ex)
                    return False

                # Check if the process with that PID is still running. Since the PID might have been 
                # reused by another process, we also check that the command that was used to start 
                # the process contains the executable name of the component. 
                # (Another idea would be to store the process creation time in the PIDs file.)
                with process.oneshot():
                    try:
                        cmdLine = "" if process.cmdline() is None else " ".join(process.cmdline())
                    except Exception as ex:
                        logger.exception('Exception while checking component %s reused pid: %s',
                                         name, ex)
                        return False
                    else:
                        return process.is_running() and exeName in cmdLine
            else:
                cmd = 'ps -p {} -f'.format(pid)
                _pid, lines = self.performCommand(parameters, cmd)

                # Return True if the executable name is in the output of the ps command.
                return bool(len(lines) >= 2 and parameters['exe'] in lines[1])


    def cmdCreateDirPath(self, dirPath, parameters=None):
        """ Creates a directory. Equivalent to; mkdir -p <dirName>

            Can be executed local or remote (depending on the parameters (see _componentIsLocal)

            :param str dirPath     : Path to a directory that will be created.
            :param dict parameters : The config parameters for the component.
                If None this command is executed locally.
        """
        assert os.path.isabs(dirPath), "Directory should be an absolute path: {}".format(dirPath)
        dirPath = normRealPath(dirPath)
        with self._thread_lock: # for when run_gssw is called as a module
            if self._componentIsLocal(parameters):
                logger.debug("Creating directory path: %s", dirPath)
                os.makedirs(dirPath, exist_ok=True)
            else:
                cmd = 'mkdir -p "{}"'.format(dirPath)
                _ = self.performCommand(parameters, cmd)



    def cmdEnsureFileExists(self, fileName, parameters=None):
        """ Make sure a file exists.

            Can be executed local or remote (depending on the parameters (see _componentIsLocal)

            :param str dirPath     : Path to a directory that will be created.
            :param dict parameters : The config parameters for the component.
                If None this command is executed locally.
        """
        assert os.path.isabs(fileName), "fileName should be an absolute path: {}".format(fileName)
        fileName = normRealPath(fileName)
        with self._thread_lock: # for when run_gssw is called as a module
            if self._componentIsLocal(parameters):
                logger.debug("Ensuring file exists: %s", fileName)
                with open(fileName, 'a'):
                    pass
            else:
                # 2018-11-01 PTK, this function might be necessary if you start compoments remotely, 
                # but I can't test this at the moment. I therefore left it in so that I don't 
                # accidentally break things (for local components this is not necessary, but I just 
                # added this case to be consistent.) # TODO: still necessary?
                cmd = 'touch "{}"'.format(fileName)
                _ = self.performCommand(parameters, cmd)


    def cmdKillProcess(self, pid, killSignal=signal.SIGTERM, killChildren=True, parameters=None):
        """ Kills a process given its PID.

            Can be executed local or remote (depending on the parameters (see _componentIsLocal)

            :param int pid:         PID of the process
            :param int killSignal:  The signal used for killing (e.g. SIGTERM, SIGKILL)
            :param bool killChildren: If true child processes are killed as well. This is necessar
                on Windows because there the gssw command line tools spawn a python process.
                Only implemented for local commands.
            :param dict parameters: The config parameters for the component.
                If None this command is executed locally.
        """
        with self._thread_lock: # for when run_gssw is called as a module
            if self._componentIsLocal(parameters):
                try: 
                    proc = psutil.Process(pid)
                except psutil.Error as ex:
                    # Catch psutils exceptions such as NoSuchProcess. This can happen if a
                    # component was closed without using run_gssw and is still listed in the PID
                    # file
                    logger.debug("Error when trying to kill PID %s : %s", pid, ex)
                else:
                    #Only try to kill children if PID exists 
                    if killChildren:
                        # Need to kill children in Windows
                        # because the components spawn a Python proc.
                        procsToKill = proc.children(recursive=True)
                    else:
                        procsToKill = []
                    procsToKill.append(proc)
                    for ptk in procsToKill:
                        logger.debug("Killing process (signal=%s): %s", killSignal, ptk.pid)
                        ptk.send_signal(killSignal)
            else:
                cmd = 'kill -{} {}'.format(int(killSignal), pid)
                _ = self.performCommand(parameters, cmd)


    def _componentIsLocal(self, parameters):
        """ Uses the component's parameter dict to determine if the component is running locally.

            This is the case if the hostname is '127.0.0.1', 'localhost' or the actual hostname
        """
        return (parameters is None or
                parameters['hostname'] == 'localhost' or
                parameters['hostname'] == '127.0.0.1' or
                parameters['hostname'] == self.hostname)


    def performCommand(self, parameters, command, outputFile=None):
        """ Perform a shell command on a host.

        :param dict parameters: The config parameters for the module the command is related to
        :param str command:     String containing the command.
        :param str outputFile:  Optional path to redirect the stdout and stderr to.

        :return: Tuple containing the pid as an integer and a list containing the lines printed on
                 stdout. If a outputFile was specified, an empty list is returned.
        """
        if self._componentIsLocal(parameters):
            return self.localCommand(command, outputFile=outputFile)

        client = self.createClient(parameters)
        if client is not None:
            ret = self.remoteCommand(client, command, outputFile=outputFile)
            client.close()
            return ret
        return [], []


    def remoteCommand(self, client, command, outputFile=None):
        """ Perform a shell command on a remote machine.

        :param client:         A paramiko remote ssh client instance.
        :param str command:    A string containing the shell command.
        :param str outputFile: Optional path to redirect the stdout and stderr to.

        :return: Tuple containing the pid as an integer and a list containing the lines printed on
                 stdout. If a outputFile was specified, an empty list is returned.
        """

        # Append the command with output redirection if a outputFile was specified
        if outputFile:
            command = command + ' &> ' + outputFile

        logger.debug('remote: %s', command)

        # Get the PID and perform the command
        try:
            _stdin, stdout, _stderr = client.exec_command('echo $$; exec ' + command)
        except:
            logger.warning('Failed to execute command: %s', command)
            return (-1, [])

        # Get the PID and the stdout if no outputFile was specified
        pid = int(stdout.readline())
        output = stdout.readlines() if outputFile is None else []

        return (pid, output)


    def localCommand(self, command, outputFile=None):
        """ Perform a shell command on this machine or a remote machine.

        :param str command:    A string containing the shell command to be executed.
        :param str outputFile: Optional path to the outputFile to which the stdout and stderr
                               are redirected.

        :return: A tuple containing the pid as an integer and a list containing the lines
                 printed on stdout. If a outputFile was specified, an empty list is returned.
        """

        # The subprocess call expects the command to be formatted as a list of strings
        logger.debug('Executing local command: %s', command)

        # Set the output redirect location
        if outputFile is None:
            redirect = subprocess.PIPE
        else:
            logger.debug("Redirecting output to: %s", outputFile)
            try:
                # Don't use 'a' to append here as this yield large output files. It also makes
                # debugging more difficult. Typically you only are interested in the latest run.
                # (The real log file can be used to store log message for a longer time)
                redirect = open(outputFile, 'w')
            except IOError:
                logger.warning('Cannot open file: %s', outputFile)  # PTK: exception:
                redirect = subprocess.PIPE

        # perform the command
        split_cmd = command.split()
        try:
            process = subprocess.Popen(split_cmd,
                                       stdout=redirect,
                                       stderr=redirect,
                                       shell=False,
                                       preexec_fn=None if sys.platform == 'win32' else os.setpgrp)
        except Exception as ex:
            logger.error('Failed to execute command: %s', command)
            logger.exception(ex)
            return (-1, [])

        # return the output if it was not redirected to a file
        output = []
        if outputFile is None:
            for line in process.stdout.readlines():
                output.append(line.decode()) # covert from bytes to string

        return (process.pid, output)


    def createClient(self, parameters):
        """ Open a SSh connection to the machine hosting the module.

            :param dict parameters: Dictionary of config parameters for a module.
                                    Contains at least the hostname and optionally login credentials.
        """

        # set up the SSH connection
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.load_system_host_keys()
        try:
            if 'username' in parameters and 'password' in parameters:
                client.connect(parameters['hostname'],
                               username=parameters['username'],
                               password=parameters['password'])
            else:
                client.connect(parameters['hostname'])
            return client
        except socket.error:
            logger.exception('Could not connect to %r', parameters['hostname'])
            return None


    def logFileName(self, componentName, parameters):
        """ Constructs the full path for the log file (directory + filename).

            The log file is not realy a log file but is the redirected output of the componenent.

            :param dict parameters: Dictionary of config parameters for a module.
        """
        if 'log_dir' in parameters:
            logDir = parameters['log_dir'].strip()

            if not os.path.isabs(logDir):
                logDir = os.path.join(os.path.expanduser('~'), logDir)
        else:
            logDir = outputDirectory()

        return os.path.join(logDir, "{}.log".format(componentName))


    def writePidsFile(self):
        """ write the latest state to the pids file. """

        try:
            self._acquirePidsLock()
            if not self.pids: # delete the pids file if there are no running processes
                logger.debug("Removing pids file: %s", self.pidsFile)
                try:
                    os.remove(self.pidsFile)
                except FileNotFoundError:
                    pass
            else:
                with open(self.pidsFile, 'w') as pidFile:
                    logger.debug("Writing pids file: %s", self.pidsFile)
                    pidFile.write(yaml.dump(self.pids))
        finally:
            self.releasePidsLock()


    def _acquirePidsLock(self):
        """ Acquire exclusive lock for (writing) the pids file.

        :note: The function blocks until lock is acquired. """
        lockFile = self.pidsFile + ".lock"

        while True:
            try:
                # Create the lock file ('x' mode raises an error if the file already exists)
                with open(lockFile, 'x') as fLock:
                    logger.debug('Locked pids file (pid=%s)', os.getpid())
                    fLock.write("{}".format(os.getpid()))
                self.lockFile = lockFile
                break
            except FileExistsError:
                logger.debug('pids file is locked, retrying...')
                time.sleep(.05)


    def releasePidsLock(self):
        """ Release the exclusive lock, such that other application can write the pids file. """
        if not self.lockFile:
            logger.warning('No lock file to remove')
            return
        lockFile      = self.lockFile
        self.lockFile = None
        try:
            os.remove(lockFile)
            logger.debug('Unlocked pids file')
        except FileNotFoundError:
            logger.warning('Hmm, weird... '
                           'I tried to remove my own lock file, but it was already gone (%s)',
                           lockFile)



def main():
    """ Run the runGSSW script. """

    logger.debug("Starting runGSSw")

    parser = createGsswArgumentParser(__doc__)
    parser.add_argument('command', type=str, choices=['start', 'stop', 'restart', 'status'])
    parser.add_argument('-c', '--components', dest='components', nargs='+', type=str,
                        help='List of components to perform the command on.')
    parser.add_argument('-l', '--log-level', dest="logLevel", type=str, default="info",
                        choices=['critical', 'error', 'warning', 'info', 'debug'],
                        help="Filter messages with this level and lower. "
                             "Applies only to run_gssw, not to the components it starts.")

    args = parser.parse_args()

    if args.logLevel:
        # Using getLevelName to get the level number. This undocumented behavior has been upgraded
        # to documented behavior in Python 3.4.2.
        # See https://docs.python.org/3.4/library/logging.html#logging.getLevelName
        levelNr = logging.getLevelName(args.logLevel.upper()) - 1
        logging.disable(levelNr)

    runGSSw = RunGSSw(args.configPath)
    runGSSw.execute(args.command, args.components)

    logger.debug("Done runGSSw")


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info("Archiver stopped, closing resources.")
    finally:
        destroyAllSocketsInContext()
