""" Functionality that is common to all components and ATPs

    This module is not part of the gssw.lib package because it uses modules outside the libs class.
"""
import logging
import os.path
import sys
import traceback

from gssw.config.configclient import ConfigClient
from gssw.lib.diagnostics     import DiagnosticLogger
from gssw.lib.errors          import ConfigError
from gssw.lib.misc            import enableLargeProtobuffers
from gssw.lib.typing          import check_class

logger = logging.getLogger(__name__)

_config = None  #pylint: disable=invalid-name

#: Use this logger to send diagnostic messages to users from a script or daemon program.
#: The messages will be displayed from the Traymonitor.
#:
#: :note: You must first call :meth:`initGssw` before using this logger.
#: :see: :class:`.DiagnosticLogger` for the available functions (info, warning, error).
diagnosticlogger = DiagnosticLogger()  #pylint: disable=invalid-name

def getGsswConfig():
    """ Globally access the Egress configuration.

    :return: Egress configuration object (:class:`.Config`).
    """
    assert _config is not None, \
        "Egress is not initialized, cannot retrieve configuration (use initGssw first)"
    return _config

def initGssw(args, logToSocketHandler=True):
    """ Initializes the GSSw

        :note: After using this function, the Egress configuration object is globally accessible via
               the :meth:`getGsswConfig`.

        :param args: argparse.Namespace object (i.e. the result of ArgumentParser.parse_args())
        :param logToSocketHandler: if False the 'socketHandler' from the logging configuration
            will not be used. This is necessary to prevent the eventLog server to send log
            message to itself. The default is True.
        :return: Egress configuration object (:class:`.Config`).
    """

    # Adding data logging level
    # Warning: Must be done before the Logger initialization if you 
    # want to use the added levels in the config.yaml file 
    addLoggingLevel("DATA", 15)
    
    try:
        configClient = ConfigClient(args.configPath)
    except ConfigError as e:
        sys.exit(e)
    global _config  #pylint: disable=global-statement, invalid-name
    _config = configClient.config
    _config.initLogging(logToSocketHandler=logToSocketHandler)
    
    #Requires _config.initLogging to actually log
    logger.info("Added logging.data() with Level 15 to logging Facility")

    # Find the name of the python script that started this process.
    scriptpath = traceback.extract_stack(limit=-1)[0].filename
    scriptName = os.path.basename(scriptpath)

    msg = "Initialized GGSw for script: {}".format(scriptName)
    try:
        componentName = args.name
    except AttributeError:
        componentName = scriptName # --name parameter not present, use script name
    else:
        msg += ". Component: {}".format(componentName)

    logger.info(msg)

    # Connect diagnostic user logger to server (and remember component name)
    diagnosticlogger.initialize(_config, componentName)

    # Now, override protobuf message size limit, if requested in configuration file
    # Please see Redmine issue #3197. The default limit is 64MB, which is sometimes too small.
    miscParameters = _config.getGlobalMiscParameters()
    if miscParameters:
        enableLargeProtobufs = miscParameters.get('enable-large-protobufs')
        if enableLargeProtobufs:
            enableLargeProtobuffers()

    return _config

##################################################################################################
# Logging facility level expansion
##################################################################################################

def addLoggingLevel(levelName, levelValue):
    """ Function to add additional logging levels
    
    :param str levelName:  Name to be used for the new logging level (i.e. data)
    :param int levelValue: Level of specified logging level (i.e. 15)
        
    Usage: >>> addLoggingLevel('Data',15) 
        {LEVELNAME} = logging.levelValue (i.e. 15) i.e. logging.DATA = 15,         #Case: Uppercase
        logger.{levelname}("Log this string")      i.e. logger.data('Logs data!')  #Case: Lowercase
    """
    
    check_class(levelName, str)
    check_class(levelValue, int)
    
    # DATA = 15, but logging.data('Logs data!') #Case sensitive
    levelName = levelName.upper()
    
    # Add {LEVELNAME} = logging.levelValue (i.e. DATA = 15)
    logging.addLevelName(levelValue,levelName)
    # Add level name as module parameter
    setattr(logging, levelName, levelValue) # needed for logger.isEnabledFor(logging.DATA)
    
    def logToRoot(message, *args, **kwargs):
        """ Function used by setattr below to add "logger.{levelname}" method (Added for PyLint)
        """
        logging.log(levelValue, message, *args, **kwargs)
    setattr(logging, levelName.lower(),logToRoot)
    
    def logForLevel(self,message, *args, **kwargs):
        """ Function used by setattr below to add "logger.{levelname}" method (Added for PyLint)
        """
        if self.isEnabledFor(levelValue):
            self._log(levelValue, message, args, **kwargs)  #pylint: disable=protected-access
    setattr(logging.getLoggerClass(), levelName.lower(),logForLevel)
