""" Module containing the configuration client class. """

import logging
import yaml

from gssw.lib.config import Config
from gssw.lib.errors import GsswError, CommunicationError, GsswTimeoutError
from gssw.lib.sockets import ReqSocket, tagName
from gssw.lib.typing import check_class
from gssw.proto.file_pb2 import FileReq, FileRep
from gssw.proto.common_pb2 import ErrorRep

logger = logging.getLogger(__name__)


class ConfigClient:
    """ This class can be instantiated by programs that need information from
        the config file. This class can retreive the config file from the
        configServer module using a ZMQ req-rep socket pair. It is also
        possible to read the config file directly from memory.
    """

    def __init__(self, configPath=None):
        """ Constructor for the ConfigClient class.

            :param str configPath: Configuration file path, or address of the configuration server.

            The configPath can either be a filename, in which case the file is read, or it can be
            the address of the configuration server, in which case the configuration is retrieved
            from the server.
        """
        check_class(configPath, str)
        self._configPath = configPath
        self._config = None
        # If the configPath contains a '://' (e.g. as in 'tcp://localhost:1234'), it is assumed to
        #   be a server address
        if self._configPath.count("://") == 1:
            self._requestConfig()
        else:
            self._readConfig()
        if self._config is None:
            raise ValueError("Config is not initialized.")


    @property
    def config(self):
        """ Return the configuration as a dictionary.
            Will raise a ValueError if the configuration is None (i.e. not initialized)
        """
        if self._config is None:
            raise ValueError("Config is not initialized.")

        return self._config


    # def getConfigDir(self):
    #     """ Returns the directory of the configuration file (including trailing /).
    #         Is used as a prefix for relative paths in the configuration file.
    #     """
    #     return self._configDir


    def _readConfig(self):
        """ Read the configuration from a file. """
        assert self._configPath, "configPath undefined"

        self._config = Config.createFromYamlFile(self._configPath)


    def _requestConfig(self):
        """ Request the configuration file through a zmq socket. """

        logger.debug("Requesting config from {}".format(self._configPath))
        assert self._configPath, "configPath undefined"

        # create a REQ socket
        requestSocket = ReqSocket()
        requestSocket.createZmqSocket()

        # connect the REQ socket to the host
        requestSocket.connectSocket("{}".format(self._configPath))

        # Send the request for the config file
        fileReq = FileReq()
        requestSocket.sendProtoBuf(tagName(FileReq), fileReq.SerializeToString())

        try:
            # receive the reply
            rep = requestSocket.receiveProtoBuf()
        except GsswTimeoutError as ex:
            logger.error('Could not reach config server: %s', ex)
            return

        # Verify the reply type
        if rep.tag == tagName(FileRep):
            reply = FileRep()
            reply.ParseFromString(rep.message)
            dct = yaml.load(reply.data, Loader=yaml.CLoader)
            self._config = Config(dct)
            self.config.configFile = reply.filename

        elif rep.tag == tagName(ErrorRep):
            reply = ErrorRep()
            reply.ParseFromString(rep.message)
            logger.error("Server had an error: {}".format(reply.errmsg))
            CommunicationError.raiseFromProtoBuffer(reply)

        else:
            err = "Unexpected message type: {}".format(rep.tag)
            logger.error(err)
            raise GsswError(err)
