""" Common functions to read temperature from a thermistor. """

import logging
import time

from gssw.cryolib.thermistors   import getThermistor
from gssw.cryolib.utils         import ThermistorError

logger = logging.getLogger(__name__)

class TemperatureLib:
    """
    Class to read a resistance from a specified interface (a multimeter or a resistance bridge)
    and to convert the resistance to a temperature.


    Usage:

    >>> import gssw.ucelib
    >>>
    >>> multimeter = gssw.ucelib.HP3478A()
    >>> tempLib = TemperatureLib(multimeter)
    >>>
    >>> tempLib.selectThermistor(thermistorName)
    >>> resistance, temperature = tempLib.readResistanceTemperature()
    """

    def __init__(self, measHandler):
        """ Connect to TemperatureLib object.

        :param measHandler: Measurement object that will be used to read resistance.
                            This object must have at least one method 'measure' which
                            read the resistance of thermistor in Ohm.
                            This object should be a AVS47B or Keithley2000 object.
        """

        self.thermistorName = None
        self.thermistor = None
        self.handler = None

        #if isinstance(measHandler, (ucelib.Keithley2000, ucelib.HP3478A, ucelib.AVS47B)):
        if hasattr(measHandler, 'measure') and callable(measHandler.measure):
            self.handler = measHandler
            logger.info('Connect to temperature lib with device: %s', self.handler.getID())
        else:
            logger.error(
                'Measurement device (%s) is not a HP3478A multimeter or a AVS47B resistance bridge',
                str(measHandler))

    def selectThermistor(self, name):
        """ Select the thermistor that will be used by readTemperature method.

        :param str name: Name of the thermistor.
        """
        self.thermistorName = name

        try:
            self.thermistor = getThermistor(self.thermistorName)
            logger.info('Thermistor selected: %s', name)
        except ThermistorError as e:
            self.thermistor = None
            logger.error(e)


    def getResistance(self, nMeas=None):
        """
        Read resistance from the reading device..
        This function can also return the average resistance of a number of measurements.

        :param int nMeas: Number of measurement for averaging (optional).

        :return: Resistance in Ohm.
        """

        if not nMeas:
            # Do only one measurement
            resistance = self.handler.measure()
            logger.debug('Get Resistance: %f Ohm', resistance)
        else:
            # Do multiple readings and return average value
            resistance = 0.0
            for counter in range(nMeas):
                if counter:
                    time.sleep(0.5)

                # Do one reading
                newResistance = self.handler.measure()
                logger.debug('Get Resistance(%i/%i): %f Ohm', counter+1, nMeas, newResistance)
                resistance += newResistance

            resistance /= nMeas
            logger.debug('Get average Resistance of %i readings: %f Ohm', nMeas, resistance)

        return resistance

    def waitStable(self, nMeas=None):
        """ Wait until the readout value of the AVS is stable.

        :param int nMeas: number of measurement to read the resistance after stabilisation.

        :return: Resistance in Ohm after stabilisation.
        """

        logger.info('Wait stable')

        prevRes = self.getResistance()
        time.sleep(2)
        currentRes = self.getResistance()

        # Calculate absolute value of derivative (some sort of time derivative)
        prevChange = currentRes - prevRes
        prevChangeSign = 'neg' if prevChange < 0 else 'pos'
        prevChange = abs(prevChange)

        # Take another reading
        prevRes = self.getResistance()
        time.sleep(2)
        currentRes = self.getResistance()

        # Calculate absolute derivative again
        currentChange = currentRes - prevRes
        currentChangeSign = 'neg' if currentChange < 0 else 'pos'
        currentChange = abs(currentChange)

        logger.debug('Wait stable. currentRes = %f ohm, prevRes = %f ohm', currentRes, prevRes)
        logger.debug('Wait stable. currentChange = %f ohm, prevChange = %f ohm',
                     currentChange, prevChange)
        logger.debug('Wait stable. currentChangeSign = %s, prevChangeSign = %s',
                     currentChangeSign, prevChangeSign)

        avsStable = bool(currentChange > prevChange and currentChangeSign != prevChangeSign)

        numReadIterations = 0

        # Iterate while all of these conditions are true:
        #
        # * time-'derivative' keeps decreasing
        # * Sign of 'derivative' stays the same
        # * Maximum number of iterations is not reached.
        #
        # When any of these conditions fails, we consider the readout
        # to be stable.

        while (not avsStable) and (numReadIterations < 25):
            logger.info('Wait stable, iteration %i', numReadIterations)

            prevChange = currentChange
            prevChangeSign = currentChangeSign
            prevRes = currentRes

            time.sleep(1)

            currentRes = self.getResistance()
            currentChange = currentRes - prevRes
            currentChangeSign = 'neg' if currentChange < 0 else 'pos'
            currentChange = abs(currentChange)

            logger.debug('Wait stable. currentRes = %f ohm, prevRes = %f ohm', currentRes, prevRes)
            logger.debug('Wait stable. currentChange = %f ohm, prevChange = %f ohm',
                         currentChange, prevChange)
            logger.debug('Wait stable. currentChangeSign = %s, prevChangeSign = %s',
                         currentChangeSign, prevChangeSign)

            avsStable = bool(currentChange > prevChange and currentChangeSign != prevChangeSign)
            numReadIterations += 1

        if numReadIterations == 25:
            logger.warning('AVS failed to stabilize properly.')
        else:
            logger.info('Reading stabilized after %i reads', numReadIterations)

        # Now return a resistance value
        if not nMeas:
            # User did not request mean value, but we can use the last two
            # readings for computing a more accurate resistance anyway
            meanResistance = (currentRes + prevRes) / 2.0
        else:
            # Compute total average of last two readings and additional readings
            # requested by the user
            resistance = self.getResistance(nMeas)
            meanResistance = (resistance * nMeas + currentRes + prevRes) / (nMeas + 2.0)

        return meanResistance

    def waitStableGetResistance(self, nMeas=None):
        """
        Waits until the AVS readout is stable and returns the stabilized resistance.

        :param int nMeas: number of measurement to read the resistance after stabilisation

        :return: Resistance in Ohm after stabilisation.
        """

        resistance = self.waitStable(nMeas)

        if not nMeas:
            logger.info(
                'Wait stable and get resistance. Average of 2 readings is %f Ohm', resistance)
        else:
            logger.info('Wait stable and get resistance. Average of %i readings if %f Ohm',
                        1+nMeas, resistance)

        return resistance

    def readResistanceTemperature(self, nMeas=None):
        """ Read current temperature from thermistor.

        This procedure assumes that the AVS is already reading the desired thermistor
        and that the reading is stable.

        :param int nMeas: number of measurement to use for averaging (optional).

        :return: (Resistance, Temperature): Resistance in Ohm, Temperature in K.
        """

        resistance = self.getResistance(nMeas)

        if self.thermistor:
            temperature = self.thermistor.calcTemperature(resistance)
            if nMeas:
                logger.info(
                    'Read average temperature: %f K (resistance = %f ohm) after %i measurements',
                    temperature, resistance, nMeas)
            else:
                logger.info('Read temperature: %f K (resistance = %f ohm)', temperature, resistance)
        else:
            temperature = None
            logger.warning(
                'Invalid thermistor. Cannot convert resistance %f ohm to a temperature', resistance)

        return resistance, temperature

    def readTemperatureImm(self, nMeas=None):
        """
        Read current temperature from thermistor.

        This procedure assumes that the AVS is already reading the desired thermistor
        and that the reading is stable.

        :param int nMeas: Number of measurement to use for averaging (optional)

        :return: Temperature in K.
        """

        _, temperature = self.readResistanceTemperature(nMeas)

        return temperature

    def readTemperature(self, nMeas=None):
        """
        Wait till temperature in thermistor is stable and read current temperature from thermistor.

        :param int nMeas: Number of measurement for averaging (Optional)

        :return: Temperature in K.
        """

        logger.info('Waiting for thermistor %s to stabilize ...', self.thermistorName)
        self.waitStable()

        temperature = self.readTemperatureImm(nMeas)
        return temperature

    def readSamplesFile(self, sampleFilename):
        """ Extract list of (thermistor, channel, range, excitation) from a text file. """

        sampleLst = []

        # check if the filename exists

        # open filename
        with open(sampleFilename) as f:

            # Read all lines from the file
            sampleLines = f.readlines()

            # Filter only lines starting with '|'
            sampleLines = [el for el in sampleLines if el[0] == '|']

            sampleKeys = ['thermistorName', 'channel', 'range', 'excitation']

            for sampleLine in sampleLines:

                # Extract sample's parameters into a list
                sampleInfo = sampleLine.split('|')
                sampleInfo = [el.strip() for el in sampleInfo[1:-1]]

                sampleInfo[1:] = [int(el) for el in sampleInfo[1:]]

                # Convert sampleInfo into a dict
                sampleInfo = dict(zip(sampleKeys, sampleInfo))
                sampleLst.append(sampleInfo)

        return sampleLst

