""" This module contains all the calibrated thermistors available. """
#pylint: disable=too-many-lines

import os, sys, csv, inspect
import numpy as np
from scipy import interpolate, optimize
from gssw.cryolib.utils import ThermistorModel, ThermistorError, ThermistorModelResistanceLookup

__all__ = ['R0_9918', 'K15', 'ROV', 'K2still_and_CP', 'L45', 'RuO2_10k', 'RuO2_1500',
           'Pt1000', 'S0930', 'S0931', 'GeXFDM', 'GeProbe', 'Ruox_5', 'R500_1', 'Dt670',
           'Ge_40px_A', 'Ge_31278', 'Ge_31281', 'getThermistor', 'isValidThermistor']

#
# utils.convRT1 resistance to temperature formula
#

class Ge_40px_A(ThermistorModelResistanceLookup):  #pylint: disable=invalid-name
    """ Thermistor used for the 40 pixel A setup (TES) """

    def __init__(self):
        super().__init__('Ge_40px_A', '27 11 2017', self.__doc__,
                         R0        = 0.57080789, # kOhm
                         coefArray = [1.77620397,  0.29317274, -0.01163221])

        self.calcTemperature = self.convRT1


class Ge_31278(ThermistorModelResistanceLookup):  #pylint: disable=invalid-name
    """ Thermistor Ge 31278. """

    def __init__(self):
        super().__init__('Ge-thermistor #31278', '04 2017', self.__doc__,
                         R0        = 0.715330, # kOhm
                         coefArray = [1.899881, 0.219144, 0.004039])

        self.calcTemperature = self.convRT1


class Ge_31281(ThermistorModelResistanceLookup):  #pylint: disable=invalid-name
    """ Thermistor Ge 31281. """

    def __init__(self):
        super().__init__('Ge-thermistor #31281', '04 2017', self.__doc__,
                         R0        = 0.897310, # kOhm
                         coefArray = [1.931366, 0.209261, 0.004159])

        self.calcTemperature = self.convRT1


class R0_9918(ThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor R0 9918. """

    def __init__(self):
        super().__init__(
            'R0 9918',
            '07 19 2005 (Run 310)',
            'Temperatures outside range 55-380 mK are extrapolated.',
            'Accuracy of fit is within 1 mK for temperatures below 300 mK.',

            2.2,
            [-1.105674, 1.380146, -0.085304])
        self.calcTemperature = self.convRT1
        self.calcResistance = None


class K15(ThermistorModel):
    """ Thermistor K15. """

    def __init__(self):
        super().__init__(
            'RO K15 calibrated with respect to ROG',
            '08 10 2005 (Run 104)',
            'Temperatures outside range 30-200 mK are extrapolated.',
            'Accuracy of fit is within 3 mK for temperatures between 50 and 100 mK.',

            2.2,
            [-1.648257, 1.503540, -0.095732])
        self.calcTemperature = self.convRT1
        self.calcResistance = None


class ROV(ThermistorModel):
    """ Thermistor ROV. """

    def __init__(self):
        super().__init__(
            'ROV',
            '06 22 2004',
            'Calibrated with respect to sensor L45 during K2 Run 74',
            'Values below 33 mK and above 200 mK are extrapolated.',

            2.2,     # in kOhm
            [1.127985, 0.769435, 0.030524])
        self.calcTemperature = self.convRT1
        self.calcResistance = None


class K2still_and_CP(ThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor K2still_and_CP. """

    def __init__(self):
        super().__init__(
            'K2Still_and_CP',
            'January-2001',
            'Calibrated by Oxford Instruments RUO22004',
            'Error is 10 mK in range 50-250 mK',

            2.2,     # in kOhm
            [-0.98320, 1.16685, 0.02943, -0.02729, 0.00243])
        self.calcTemperature = self.convRT1
        self.calcResistance = None


class L45(ThermistorModel):
    """ Thermistor L45. """

    def __init__(self):
        super().__init__(
            'L45',
            '05 01 2003',
            'L45 Factory calibration',
            'Systematic factory error is +/- 5 mK between 50 and 150 mK',

            2.21,      # in kOhm
            [-0.70054423, 1.09232869, 0.06280054, -0.01192616])
        self.calcTemperature = self.convRT1
        self.calcResistance = None

#
# utils.convRT2 resistance to temperature forumula
#

class RuO2_10k(ThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor RuO2_10k. """

    def __init__(self):
        super().__init__(
            'RuO2_10k',
            date = None,
            description1 = self.__doc__,

            coefArray = [
                79362.66285435,
               -134213.2086187999,
                101026.7304322,
               -44413.273683345,
                12564.3214665816,
               -2371.588181102,
                298.641277814,
               -24.188904628,
                1.143358358,
               -0.024026573 ])

        self.calcTemperature = self.convRT2
        self.calcResistance = None


class RuO2_1500(ThermistorModel):  #pylint: disable=invalid-name
    """Thermistor RuO2_10k"""

    def __init__(self):
        super().__init__(
            'Ru02_1500',
            '29 08 2014',
            self.__doc__,

            coefArray = [
                -1.09406225453069e8,
                 2.64570367275206e8,
                -2.84130767924156e8,
                 1.77858095325872e8,
                -7.15164300477208e7,
                 1.91562166038666e7,
                -3.41810854721648e6,
                 391778.17083849304,
                -26174.321297539838,
                 776.59077708274708 ])

        self.calcTemperature = self.convRT2
        self.calcResistance = None


class Pt1000(ThermistorModel):
    """Thermistor Pt1000"""

    def __init__(self):
        super().__init__(
            'Pt1000',
            '29 08 2014',
            self.__doc__,

            coefArray = [
                -469.544790033,
                 2142.429105073,
                -4278.355519358,
                 4917.129912473,
                -3583.140137551,
                 1717.059717072,
                -541.269035711,
                 108.277676198,
                -12.478784052,
                 0.631579503 ])

        self.calcTemperature = self.convRT2
        self.calcResistance = None


class S0930(ThermistorModel):
    """Thermistor S0930"""

    def __init__(self):
        super().__init__(
            'S0930',
            '29 08 2014',
            self.__doc__,

            coefArray = [
                1.2984086888338e6,
               -3.08935089865459e6,
                3.25663587979261e6,
                784254.21799103729,
               -204770.48046283511,
                35536.83686322545,
               -3952.9152682310819,
                255.74532449375445,
               -7.33285017805232 ])

        self.calcTemperature = self.convRT2
        self.calcResistance = None


class S0931(ThermistorModel):
    """Thermistor S0931"""

    def __init__(self):
        super().__init__(
            'S0931',
            '29 08 2014',
            self.__doc__,

            coefArray = [
                988781.9678376466,
                -2.34319672344125e6,
                2.4596949546024e6,
                -1.50114428998402e6,
                587001.99250064464,
                -152524.26551352543,
                26335.068913128416,
                -2913.7175534480721,
                187.45512866778719,
                -5.34322191809326 ])

        self.calcTemperature = self.convRT2
        self.calcResistance = None

#
# micro-kelvin precision (Utrecht XFDM Setup)
#

class GeXFDM(ThermistorModel):
    """GeXFDM thermistor of the XFDM setup."""

    def __init__(self):
        super().__init__('GeXFDM', '13 01 2015', self.__doc__)

        path = os.path.dirname(os.path.realpath(__file__))
        with open(path + '/lookup/GeXFDM.csv') as csvfile:
            r = []  #pylint: disable=invalid-name
            t = []  #pylint: disable=invalid-name
            reader = csv.DictReader(csvfile)
            for row in reader:
                r.append(float(row['resistance']))
                t.append(float(row['temperature']))

            self.minRes = r[0]
            self.maxRes = r[len(r) - 1]
            self.tempAtMinRes = t[0]
            self.tempAtMaxRes = t[len(t) - 1]

            self.interp1dFunc = interpolate.interp1d(x=r, y=t, kind='linear')

    def calcTemperature(self, resistance):
        """convert <res> [Ohm] to temperature [K]"""
        if resistance <= self.minRes:
            return self.tempAtMinRes
        if resistance >= self.maxRes:
            return self.tempAtMaxRes
        return float(self.interp1dFunc(resistance))

    def calcResistance(self, temperature):
        """Convert <temperature> [K] to resistance [Ohm]"""
        if temperature <= 0.025:
            temperature = 0.003
        elif temperature >= 0.160:
            temperature = 0.160
        fit = np.array([-129.18140034, -193.35868382, -106.40849981, -25.34411052, -2.11918938])
        tlog = np.log(temperature) # natural logarithm of the temperature in mK
        return 1000.0 * np.exp(
            fit[0] + fit[1] * tlog + fit[2] * tlog**2 + fit[3] * tlog**3 + fit[4] * tlog**4)


class GeProbe(ThermistorModelResistanceLookup):
    """ GeProbe thermistor of the Probe setup.

    Thermistor calibration (XFDM-Probe)
    type: GR-50-AA (Ge thermistor)
    measured resistance with DualFEE12.

    Calibration reference thermistor used: ROV.
    Measured resistance with DualFEE12.

    :note: Thermistor(log(R)) is used to fit in log-scale using log(R_Ohm) and log(T_mK).

    :note: Calibrated by Kevin Ravensberg.
    """

    def __init__(self):
        super().__init__('GeProbe', '04 feb 2019', self.__doc__)

    def calcTemperature(self, resistance):
        """ Convert resistance [Ohm] to temperature [K]"""
        resistance = np.clip(resistance, 500.0, 450000.0) # clip to data range [~48 mK, 200mK] 
        fit = np.array( # Calibration data received from Kevin Ravensberg
            [6.93734815e-01, 5.32864127e-01, -9.56868322e-02, 6.75855910e-03, -1.72942864e-04])
        rlog = np.log(resistance) # natural logarithm of the resistance in Ohm
        return 1e-3 * np.e**np.exp(
            fit[0] + fit[1] * rlog + fit[2] * rlog**2 + fit[3] * rlog**3 + fit[4] * rlog**4)


#
# thermistors Groningen
#

class Ruox_5(ThermistorModel):  #pylint: disable=invalid-name
    """Thermistor Ruox_5 Groningen"""

    def __init__(self):
        super().__init__('Ruox_5', '23 10 2014', self.__doc__)

        self.calcTemperature = self.convRuox_5_res2temp
        self.calcResistance  = self.convRuox_5_temp2res


class R500_1(ThermistorModel):  #pylint: disable=invalid-name
    """Thermistor R500_1 (Groningen)"""

    def __init__(self):
        super().__init__('Ruox_5', '23 10 2014', self.__doc__)

        # old
        #self.calcTemperature = self.convR500_1_res2temp
        #self.calcResistance = self.convR500_1_temp2res

        path = os.path.dirname(os.path.realpath(__file__))
        with open(path + '/lookup/R500_1_rt.csv') as csvfile:
            r = []  #pylint: disable=invalid-name
            t = []  #pylint: disable=invalid-name
            reader = csv.DictReader(csvfile)
            for row in reader:
                r.append(float(row['resInOhm']))
                t.append(0.001 * float(row['tempInMilliK']))

            self.rtMinR = r[0]
            self.rtMaxR = r[len(r) - 1]
            self.rtTAtMinR = t[0]
            self.rtTAtMaxR = t[len(t) - 1]
            self.rtInterpolate = interpolate.interp1d(x=r, y=t, kind='linear')

        path = os.path.dirname(os.path.realpath(__file__))
        with open(path + '/lookup/R500_1_tr.csv') as csvfile:
            t = []  #pylint: disable=invalid-name
            r = []  #pylint: disable=invalid-name
            reader = csv.DictReader(csvfile)
            for row in reader:
                t.append(0.001 * float(row['tempInMilliK']))
                r.append(float(row['resInOhm']))

            self.trMinT = t[0]
            self.trMaxT = t[len(t) - 1]
            self.trRAtMinT = r[0]
            self.trRAtMaxT = r[len(r) - 1]
            self.trInterpolate = interpolate.interp1d(x=t, y=r, kind='linear')

    def calcTemperature(self, resInOhm):
        """ Convert resistance to temperature """
        if resInOhm <= self.rtMinR:
            return self.rtTAtMinR
        if resInOhm >= self.rtMaxR:
            return self.rtTAtMaxR
        return float(self.rtInterpolate(resInOhm))

    def calcResistance(self, tempInK):
        """ Convert temperature to resistance """
        if tempInK <= self.trMinT:
            return self.trRAtMinT
        if tempInK >= self.trMaxT:
            return self.trRAtMaxT
        return float(self.trInterpolate(tempInK))

class Dt670(ThermistorModel):
    """Calibration source thermistor (Groningen)"""

    def __init__(self):
        super().__init__('Dt670', '02 03 2015', self.__doc__)

        path = os.path.dirname(os.path.realpath(__file__))
        with open(path + '/lookup/dt670.csv') as csvfile:
            v = []  #pylint: disable=invalid-name
            t = []  #pylint: disable=invalid-name
            reader = csv.DictReader(csvfile)
            for row in reader:
                v.append(float(row['voltInV']))
                t.append(float(row['tempInK']))

            self.minVolt = v[0]
            self.maxVolt = v[len(v) - 1]
            self.tempAtMinVolt = t[0]
            self.tempAtMaxVolt = t[len(t) - 1]

            self.interp1dFunc = interpolate.interp1d(x=v, y=t, kind='linear')

        self.calcResistance = None

    def calcTemperature(self, voltInV):
        """Convert <voltInV> to temperature [Ohm]."""
        if voltInV <= self.minVolt:
            return float(self.tempAtMinVolt)
        if voltInV >= self.maxVolt:
            return float(self.tempAtMaxVolt)
        return float(self.interp1dFunc(voltInV))



class MottHoppingThermistorModel(ThermistorModel):
    """ 
    Base class for thermistors that are parameterized by Mott-Hopping coefficients:
        
        T^(1/c[0]) = c[1] + c[2]*LN(R) + c[3]*(LN(R))^2 + c[4]*(LN(R))^3 + ... 
    
    Depending on the coefficients, there is a range on which this function is 
    well-behaved, i.e. does not explode to NaN. The Mott-Hopping fit can only be inverted
    within this interval. It is important that care is taken that the interval [minR, maxR]
    constitutes a section of the curve that is well-behaved!
    """
    def __init__(self, sensorName, date, description, coefficients, minR, maxR):
        super().__init__(sensorName, date, description)
        self.minR = minR
        self.maxR = maxR
        self.coefficients = coefficients
        
        # calculate boundary temperatures
        self.minT = self.calcTemperature(self.maxR)
        self.maxT = self.calcTemperature(self.minR)
    
    
    def calcTemperature(self, resistance):
        """
        calculates temperature based on resistance using Mott-Hopping coefficients
        """
        # check that resistance is within valid range
        if resistance < self.minR:
            return 0.0
        elif resistance > self.maxR:
            return 0.0
        
        # perform the calculation
        rightHandPart = 0
        for exponent, coeff in enumerate(self.coefficients[1:]):
            rightHandPart += coeff * np.log(resistance) ** exponent
        temperature = rightHandPart ** self.coefficients[0]
        
        return temperature
    
    
    def calcResistance(self, temperature):
        """
        Calculates thermistor resistance based on temperature
        
        Using scipy.optimize.brentq to invert the temperature function. Brent's method finds
        the value for which a function is zero on the interval [minR, maxR]. 
        """
        # check that temperature is within valid range
        if temperature < self.minT:
            return 0.0
        elif temperature > self.maxT:
            return 0.0
        
        def tempError(resistance):
            """ error function used by scipy optimize """
            return self.calcTemperature(resistance) - temperature
        
        return optimize.brentq(tempError, self.minR, self.maxR)


class Ge_30511(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for Ge - 30511 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 6.0
        maxR = 1e6
        coefficients = [-5.00000, 6.46416, -6.99232, 3.34561, -0.827814,
                        0.118344, -0.00984415, 0.000442369, -8.29889e-006]

        super().__init__('Ge-30511', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class Ge_A_31314(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for Ge - A 31314 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 17
        maxR = 1e6
        coefficients = [-1.20000, -29.1548, 35.1832, -18.0088, 5.04742,
                             -0.824065, 0.0789078, -0.00411077, 9.00818e-005 ]
        
        super().__init__('Ge-A-31314', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class Ge_B_31315(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for Ge - B 31315 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 19
        maxR = 1e6
        coefficients = [-5.00000, -10.4865, 12.3582, -5.83760, 1.53656, -0.239050, 
                        0.0219074, -0.00109443, 2.30054e-005]
        
        super().__init__('Ge-B-31315', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class Ge_C_31316(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for Ge - C 31316 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 17
        maxR = 1e6
        coefficients = [-5.0, -8.87901, 10.413, -4.85246, 1.2662, -0.195814,
                        0.0178806, -0.000892101, 1.87732e-05]
        
        super().__init__('Ge-C-31316', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class Ge_D_31318(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for Ge - D 31311 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 18.3
        maxR = 1e6
        coefficients = [-5.0, -12.8446, 15.8412, -7.95362, 2.22753, -0.369999,
                        0.0363421, -0.00195346, 4.43541e-05]
        
        super().__init__('Ge-D-31318', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class Ge_E_31320(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for Ge - E 31320 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 17.3
        maxR = 1e6
        coefficients = [-5.0, -13.9788, 17.2301, -8.63142, 2.39722,
                        -0.393218, 0.0380019, -0.00200312, 4.44586e-05]
        
        super().__init__('Ge-E-31320', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class Ge_F_31326(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for Ge - F 31326 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 16.6
        maxR = 1e6
        coefficients = [-5.0, -8.07349, 9.59753, -4.49459, 1.17939,
                        -0.183173, 0.0167694, -0.000837282, 1.75993e-05]
        
        super().__init__('Ge-F-31326', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class Ge_TaMNT00(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for Ge - TaMNT00 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 13.0
        maxR = 1e6
        coefficients = [-1.40000, -1.23299, 0.321401, 0.0689722, -0.00186578]
        
        super().__init__('Ge-TaMNT00', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)

        
class Ge_TaMNT02(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for Ge - TaMNT02 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 12.0
        maxR = 1e6
        coefficients = [-1.40000, -1.06096, 0.281680, 0.0660156, -0.00176639]
        
        super().__init__('Ge-TaMNT02', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)
        

class RuOx_G_U06935(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - G - U06935 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1200
        maxR = 1e4
        coefficients = [-7.0, -81.5314, 28.8629, -3.40071, 0.135841]
        
        super().__init__('RuOx-G-U06935', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_H_U07016(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - H - U07016 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1295
        maxR = 1e4
        coefficients = [-7.0, -116.89, 42.5504, -5.16974, 0.212231]
        
        super().__init__('RuOx-H-U07016', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_I_U07024(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - I - U07024 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1280
        maxR = 1e4
        coefficients = [-7.0, -92.2474, 33.1538, -3.9772, 0.161871]
        
        super().__init__('RuOx-I-U07024', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_J_U07138(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - J - U07138 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1280
        maxR = 1e4
        coefficients = [-7.0, -90.7921, 32.4445, -3.86636, 0.156253]
        
        super().__init__('RuOx-J-U07138', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_K_U07139(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - K - U07139 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1306
        maxR = 1e4
        coefficients = [-7.0, -118.196, 42.9396, -5.20797, 0.21351]
        
        super().__init__('RuOx-K-U07139', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_L_U07140(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - L - U07140 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1322
        maxR = 1e4
        coefficients = [-7.0, -116.484, 42.3856, -5.15506, 0.212226]
        
        super().__init__('RuOx-L-U07140', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_M_U07141(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - M - U07141 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1290
        maxR = 1e4
        coefficients = [-7.0, -84.1832, 29.8811, -3.53612, 0.142115]
        
        super().__init__('RuOx-M-U07141', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_N_U07142(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - N - U07142 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1295
        maxR = 1e4
        coefficients = [-7.0, -77.0052, 27.1427, -3.18319, 0.126672]
        
        super().__init__('RuOx-N-U07142', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_O_U07143(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - O - U07143 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1301
        maxR = 1e4
        coefficients = [-7.0, -69.3303, 24.1643, -2.79927, 0.110226]
        
        super().__init__('RuOx-O-U07143', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_P_U07144(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - P - U07144 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1291
        maxR = 1e4
        coefficients = [-7.0, -81.6742, 28.9973, -3.43052, 0.137767]
        
        super().__init__('RuOx-P-U07144', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_Q_U07145(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - Q - U07145 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1282
        maxR = 1e4
        coefficients = [-7.0, -85.1589, 30.3213, -3.59858, 0.144927]
        
        super().__init__('RuOx-Q-U07145', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_R_U07146(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - R - U07146 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1300
        maxR = 1e4
        coefficients = [-7.0, -84.688, 30.0189, -3.54494, 0.142009]
        
        super().__init__('RuOx-R-U07146', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_S_U07148(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - S - U07148 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1279
        maxR = 1e4
        coefficients = [-7.0, -85.3336, 30.4382, -3.62108, 0.146271]
        
        super().__init__('RuOx-S-U07148', '01-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_Average(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - average thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-7.00000, -88.6333, 31.6018, -3.75617, 0.151397]
        
        super().__init__('RuOx-Average', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_Bare(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - bare thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-1.40000, -33.7438, 16.5583, -2.81877, 0.163656]
        
        super().__init__('RuOx-Bare', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_TaSQD05(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TaSQD05 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-1.40000, -30.5154, 15.5668, -2.73536, 0.162537]
        
        super().__init__('RuOx-TaSQD05', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_TaSQD21(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TaSQD21 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-1.80000, -307.346, 121.523, -16.1446, 0.723064]
        
        super().__init__('RuOx-TaSQD21', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_TaEIF20(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TaEIF20 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-1.80000, -335.652, 132.763, -17.6283, 0.788132]
        
        super().__init__('RuOx-TaEIF20', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)
        
class RuOx_TaSTR11(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TaSTR11 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1100
        maxR = 1e6
        coefficients = [-5.20000, -370.983, 146.115, -19.2778, 0.854392]
        
        super().__init__('RuOx-TaSTR11', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)
        

class RuOx_TaTIF10(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TaTIF10 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-2.40000, -392.040, 155.018, -20.5103, 0.910541]
        
        super().__init__('RuOx-TaTIF10', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)

class RuOx_TaTIF06(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TaTIF06 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-1.40000, -33.7438, 16.5583, -2.81877, 0.163656]
        
        super().__init__('RuOx-TaTIF06', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)
        

class RuOx_TaTIF22(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TaTIF22 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-1.80000, -302.725, 119.627, -15.8856, 0.711276]
        
        super().__init__('RuOx-TaTIF22', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)

class RuOx_TaSHD04(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TaSHD04 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-2.40000, -164.002, 61.9081, -7.85414, 0.337801]
        
        super().__init__('RuOx-TaSHD04', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_TaMNT01(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TaMNT01 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1100
        maxR = 1e6
        coefficients = [-5.20000, -70.6998, 24.4716, -2.84659, 0.114361]
        
        super().__init__('RuOx-TaMNT01', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_TaSHD03(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TaSHD03 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-2.20000, -146.322, 55.3296, -7.05965, 0.306916]
        
        super().__init__('RuOx-TaSHD03', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_TbSQD00(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TbSQD00 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-1.40000, -34.7133, 17.3144, -2.98121, 0.174241]
        
        super().__init__('RuOx-TbSQD00', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_TbSQD21(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TbSQD21 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-1.80000, -276.557, 109.081, -14.4684, 0.647743]
        
        super().__init__('RuOx-TbSQD21', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_TxYYYzz(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TxYYYzz thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-1.80000, -310.448, 122.780, -16.3136, 0.730580]
        
        super().__init__('RuOx-TxYYYzz', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_Tfing00(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - Tfing00 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1200
        maxR = 1e6
        coefficients = [-5.20000, -399.830, 152.183, -19.3254, 0.821276]
        
        super().__init__('RuOx-Tfing00', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_Tfing10(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - Tfing10 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1200
        maxR = 1e6
        coefficients = [-5.20000, -386.033, 150.152, -19.5317, 0.852122]
        
        super().__init__('RuOx-Tfing10', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_Tfpad00(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - Tfpad00 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1200
        maxR = 1e6
        coefficients = [-5.20000, -62.0542, 20.9461, -2.36314, 0.0919897]
        
        super().__init__('RuOx-Tfpad00', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_Tfpad10(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - Tfpad10 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1200
        maxR = 1e6
        coefficients = [-5.20000, -408.894, 159.850, -20.9021, 0.916648]
        
        super().__init__('RuOx-Tfpad10', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_TcSHD21(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TcSHD21 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-1.80000, -304.790, 121.180, -16.1920, 0.729433]
        
        super().__init__('RuOx-TcSHD21', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_TcTIF22(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TcTIF22 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1100
        maxR = 1e6
        coefficients = [-5.20000, -273.893, 105.252, -13.5402, 0.585677]
        
        super().__init__('RuOx-TcTIF22', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_TcEIF20(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - TcEIF20 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1000
        maxR = 1e6
        coefficients = [-1.80000, -344.057, 136.130, -18.0775, 0.808091]
        
        super().__init__('RuOx-TcEIF20', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_Tfpad20(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - Tfpad20 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1100
        maxR = 1e6
        coefficients = [-5.20000, -262.277, 100.913, -13.0106, 0.564714]
        
        super().__init__('RuOx-Tfpad20', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_Tfpad21(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - Tfpad21 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1100
        maxR = 1e6
        coefficients = [-5.20000, -305.116, 118.674, -15.4676, 0.678120]
        
        super().__init__('RuOx-Tfpad21', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class RuOx_Tfpad22(MottHoppingThermistorModel):  #pylint: disable=invalid-name
    """ Thermistor model for RuOx - Tfpad22 thermistor, using a Mott-Hopping fit """
    
    def __init__(self):
        minR = 1100
        maxR = 1e6
        coefficients = [-5.20000, -308.292, 119.626, -15.5454, 0.679089]
        
        super().__init__('RuOx-Tfpad22', '17-12-2019', self.__doc__, coefficients,
             minR, maxR)


class DummyThermistor(ThermistorModel):
    """ Dummy thermistor model with configurable conversion factor (linear) """
    def __init__(self):
        super().__init__('Dummy', 'Jun 26th, 2019, by S.Rosman', self.__doc__)
        self.dummyRes2TempFact = 1.0

    def calcResistance(self, temperature):
        """ Convert temperature to resistance """
        return temperature * self.dummyRes2TempFact

    def calcTemperature(self, resistance):
        """ Convert resistance to temperature """
        return resistance * self.dummyRes2TempFact


# Get list of all classes in this file
_classLst = inspect.getmembers(sys.modules[__name__], inspect.isclass)  #pylint: disable=invalid-name
thermistorLst = dict(  #pylint: disable=invalid-name
    [el for el in _classLst if el[0] not in (
        'ThermistorModel', 'ThermistorModelResistanceLookup', 'ThermistorError',
        'DummyThermistor', 'RuOxThermistorModel', 'MottHoppingThermistorModel') ])


def isValidThermistor(thermistorName):
    """ Check if 'thermistorName' is a valid thermistor. """

    allNames = [tName for tName, _ in thermistorLst.items()]
    isValid = thermistorName in allNames
    return isValid


def getThermistor(thermistorName):
    """
    Factory function that convert a thermistor name
    to a thermistor object.

    :param str thermistorName: Name of the thermistor.

    :return: Return the thermistor object.
    """

    if isValidThermistor(thermistorName):
        return thermistorLst[thermistorName]()
    raise ThermistorError("Thermistor '%s' not found" % thermistorName)
