""" Common functions for the thermistors. """

import csv
import math
import os
from scipy import interpolate

################  class ThermistorError  ################

class ThermistorError(Exception):
    """ Raised when thermistor error is received. """


class ThermistorModel:
    """ Generic class used by a thermistor class. """

    def __init__(self, sensorName,  date,
                 description1=None, description2=None,
                 R0=None,           coefArray=None):
        """ Initialize a model for a thermistor. """

        self.sensorName = sensorName
        self.date = date
        self.description1 = description1
        self.description2 = description2
        self.R0 = R0  #pylint: disable=invalid-name
        self.coefArray = coefArray

    def convRT1(self, resistance):
        """
        Convert resistance to temperature.
        Formula 1:
        ln(1/T) = A0 + A1 * ln(R - R0) + A2 * (ln(R - R0)^2) + ...

        :Param float resistance: Resistance in ohm.

        :return: Temperature in K.
        """

        # Return room temperature when the supplied resistance
        # is above room temperature resistance (R0)

        if (resistance / 1000.0) <= self.R0:
            return 273.15

        logOneDivT = 0.0
        x = math.log(resistance / 1000.0 - self.R0)
        for i, coef in enumerate(self.coefArray):
            logOneDivT += coef * math.pow(x, i)

        temperature = 1.0 / math.exp(logOneDivT)
        return temperature

    def convRT2(self, resistance):
        """
        Convert resistance to temperature.
        Formula 2:
        T = 10 ^ (A0 + A1 * log(R) + A2 * (log(R)^2) + ...

        :param float resistance: Resistance in ohm.

        :returns: Temperature in K.
        """

        tempVal = 0.0
        x = math.log10(resistance)
        for i, coef in enumerate(self.coefArray):
            tempVal += coef * math.pow(x, i)

        temperature = 10 ** tempVal

        return temperature / 1000.0

    def convRuox_5_temp2res(self, temperature):
        """Convert <temperature> [K] to resistance [Ohm] for thermistor-model Ruox_5 (Groningen)."""
        Tlog = math.log10(temperature)  #pylint: disable=invalid-name
        return (983.0113
                + 5131.63725 * math.exp(-(Tlog + 1.90161) / 0.54362)
                + 4302.42353 * math.exp(-(Tlog + 1.90161) / 0.22802)
                + 22976.26543 * math.exp(-(Tlog + 1.90161) / 0.54338))

    def convRuox_5_res2temp(self, resistance):
        """Convert <resistance> [Ohm] to temperature [K] for thermistor-model Ruox_5 (Groningen)."""
        Rn = (resistance - 1000.8) / (1262.5 - 1000.8)  #pylint: disable=invalid-name
        lnRn = math.log(Rn)
        expArg = -8.6182 + 1.74902 * lnRn - 0.3211 * lnRn**2 + 0.08213 * lnRn**3 - 0.0075 * lnRn**4
        temp_mK = 1.0 / math.exp(expArg)
        return temp_mK / 1000.0

    def convR500_1_temp2res(self, temperature):
        """Convert <temperature> [K] to resistance [Ohm] for thermistor-model R500_1 (Groningen)."""
        tempInMilliKelv = temperature * 1000.0
        return (10**(3.70975
                + 0.46097 * math.exp(-tempInMilliKelv / 97.77192)
                + 0.46135 * math.exp(-tempInMilliKelv / 97.77191)
                + 0.54031 * math.exp(-tempInMilliKelv / 12.94997)))

    def convR500_1_res2temp(self, resistance):
        """Convert <resistance> [Ohm] to temperature [K] for thermistor-model R500_1 (Groningen)."""
        # TODO: something is wrong with the resistance calculation for > 1k
        tempBypass = True
        if resistance < 1000 or tempBypass:
            Rn = (resistance - 1005.0) / (1395.0 - 1005.0)  #pylint: disable=invalid-name
            temp_mK = (
                2.98889
                + 353.97036 * math.exp(-Rn / 8.95711)
                + 3396.70647 * math.exp(-Rn / 1.33247)
                + 963.31672 * math.exp(-Rn / 3.61394)
                + 59.23845 * math.exp(-Rn / 96.21063)
                + 120.38707 * math.exp(-Rn / 27.8328)
                + 22.96051 * math.exp(-Rn / 947.16112)
                + 280.88933 * math.exp(-Rn / 0.00000000773076))
            return temp_mK / 1000.0
        else:
            Rn = (resistance - 1000.0) / (6996.0 - 1000.0)  #pylint: disable=invalid-name
            return (71.92619 * math.exp(-Rn / 0.00837)
                    + 13.53003 * math.exp(-Rn / 0.04438)
                    + 0.97572)

    def __str__(self):
        """ Return thermistor details. """
        res = "# ==============================================================\n"
        res += "#        THERMISTOR AND TEMPERATURE CALIBRATION DETAILS\n"
        res += "# ==============================================================\n"
        res += "# \n"
        res += "# Thermistor Name:           %s\n" % str(self.sensorName)
        res += "# Calibration Date:          %s\n" % str(self.date)
        res += "# Calibration Description 1: %s\n" % str(self.description1)
        res += "# Calibration Description 2: %s\n" % str(self.description2)
        res += "# \n"
        res += "# ==============================================================\n"
        return res


class ThermistorModelResistanceLookup(ThermistorModel):
    """ Thermistor type where the temperature to resistance conversion is performed using a 
        lookup and interpolation function (piece-wise-linear-interpolation).

        The lookup table is loaded from a .csv file /lookup/classname.csv."""
    def __init__(self, sensorName,  date,
                 description1=None, description2=None,
                 R0=None,           coefArray=None):
        super().__init__(sensorName, date, description1, description2, R0, coefArray)
        path = os.path.dirname(os.path.realpath(__file__))
        # Note: the lookup table is created as a reverse from the original fitting
        with open(path + '/lookup/' + self.__class__.__name__ + '.csv') as csvfile:
            t = []  #pylint: disable=invalid-name
            r = []  #pylint: disable=invalid-name
            reader = csv.DictReader(csvfile)
            for row in reader:
                t.append(float(row['temperature']))
                r.append(float(row['resistance']))

            self.minTemp = t[-1]
            self.maxTemp = t[0]
            self.resAtMinTemp = r[-1]
            self.resAtMaxTemp = r[0]
 
            self.interp1dFunc = interpolate.interp1d(x=t, y=r, kind='linear')

    def calcResistance(self, temperature):
        """Convert temperature [K] to resistance [Ohm]"""
        if temperature <= self.minTemp:
            return self.resAtMinTemp
        if temperature >= self.maxTemp:
            return self.resAtMaxTemp
        return float(self.interp1dFunc(temperature))

