""" 
HDF5 Writer Base Client 
Base client class for both Python and C++ HDF5 writer clients.
"""

import collections
import logging
import struct
from gssw.lib.config                import  Config
from gssw.lib.rpcbase               import  BaseClient
from gssw.lib.sockets               import  ReqSocket
from gssw.lib.typing                import  check_class
import gssw.proto.hdf5writer_pb2    as      h5proto
import gssw.proto.common_pb2        as      commonproto

logger = logging.getLogger(__name__)

################################################################################
# Helpers
################################################################################

def getFormatCharacter(encoding, bitWidth):
    """ Return the format character for the specified encoding and bit width.

    :param encoding: The encoding (ENC_UINT, ENC_SINT, ENC_FLOAT)
    :param bitWidth: The number of bits used to encode each word (8, 16, 32, 64)
    """
    if encoding in (commonproto.ENC_SINT, commonproto.ENC_UINT):
        if bitWidth == 8:
            formatChar = 'b'    # char
        elif bitWidth == 16:
            formatChar = 'h'    # short
        elif bitWidth == 32:
            formatChar = 'l'    # long
        elif bitWidth == 64:
            formatChar = 'q'    # long long
        else:
            raise TypeError("Unsupported word size")
        if encoding == commonproto.ENC_UINT:
            formatChar = formatChar.upper()     # unsigned integers use uppercase characters
        return formatChar
    elif encoding == commonproto.ENC_FLOAT:
        if bitWidth == 32:
            return 'f'          # float
        elif bitWidth == 64:
            return 'd'          # double
        else:
            raise TypeError("Unsupported word size")
    else:
        raise TypeError("Undefined encoding")

def flatten(item):
    """ Flatten any non-string or byte iterables nested within another iterable.
        e.g. (0, 1, [2, 3]) -> (0, 1, 2, 3)
    """
    for i in item:
        if isinstance(i, collections.Iterable) and not isinstance(i, (str, bytes)):
            yield from flatten(i)
        else:
            yield i

################################################################################
# class Hdf5BaseClient
################################################################################

class Hdf5BaseClient(BaseClient):
    """ Base client class for HDF5 writers. """

    def __init__(self, address=None, reqSocket=None):
        """ Constructor.

            The address and reqSocket parameters cannot be used both!

            :param address: if set, a ReqSocket is created and connected to this address.
            :param reqSocket: sets the request socket if given. If None, a socket will be created.
        """
        super().__init__(address, reqSocket)
        self.dataFormats = {}    # store dataformat for h5 paths to table data

    def connectToScienceDaemon(self, daemonName, config):
        """ Connect to a hdf5 writer daemon.

        :param daemonName: The name of the hardware daemon in the config file.
        :param config:     Egress configuration object.
        """
        check_class(config, Config)
        daemonConfig = config.componentParameters(daemonName)

        # check if the config file defines a datawriter for this daemon
        assert 'datawriter' in daemonConfig, 'No datawriter defined in the config file'
        assert isinstance(daemonConfig['datawriter'], dict), \
            "Datawriter configuration is unreadable"
        assert 'fileformat' in daemonConfig['datawriter'], \
            'No datawriter file format specified in the config file'
        assert daemonConfig['datawriter']['fileformat'] == 'hdf5', \
            'Wrong datawriter file format specified in the config file'

        # Create a req socket to message the writer.
        reqSocket = ReqSocket()
        reqSocket.createZmqSocket()

        # Connect the socket to the writers rep port.
        host = daemonConfig['hostname']
        port = daemonConfig['sockets']['daemon-rep']
        repAddress = 'tcp://%s:%d' % (host, port)
        reqSocket.connectSocket(repAddress)
        self.setReqSocket(reqSocket)

    def openFile(self, filePath, rotationTime=None):
        """ Open a HDF5 file.

            :param filepath: the path to the file that must be opened
        """
        self._fileOperations(True, filePath=filePath, rotationTime=rotationTime)

    def closeFile(self):
        """ Close the currently open HDF5 file.
        """
        self._fileOperations(False)

    def _fileOperations(self, openReq, filePath=None, rotationTime=None):
        """ Unified Function for file open and close requests.
        """
        requestMsg = h5proto.FileOperationReq()
        if openReq:
            requestMsg.path = filePath
            requestMsg.action = h5proto.FileOperationReq.OPEN
            if rotationTime:
                requestMsg.rotationTime = rotationTime
        else:
            requestMsg.action = h5proto.FileOperationReq.CLOSE

        self.requestReply(requestMsg, commonproto.OkRep())

    def fileStatus(self):
        """ Check whether a file is open.

            :return: Name of the file, if a file is opened, empty string otherwise.
        """
        repMsg = self.requestReply(h5proto.FileStatusReq(), h5proto.FileStatusRep())
        return repMsg.current_open_file

    def createGroup(self, h5Path):
        """ Create a new group

            :param h5Path: The HDF5 path where the group will be created. 
        """
        requestMsg = h5proto.CreateGroupReq()
        requestMsg.path = h5Path
        self.requestReply(requestMsg, commonproto.OkRep())

    def createDataSet(self, h5Path, dataType, wordSize, rank, shape=None):
        """ Create a new dataset.

            Datasets can be multidimensional but only grow in one dimension.
            It is required to specify the size of each fixed dimension.

            :param h5Path: The name of the new dataset
            :param dataType: The datasets type ['UINT', 'SINT' or 'FLOAT']
            :param wordSize: The number of bits per word
            :param rank: The number of dimensions
            :param shape: The shape of all static dimensions (= rank-1)
        """
        shape = shape or []

        assert dataType in ['UINT', 'SINT', 'FLOAT'], 'Undefined type'
        assert len(shape) is (rank - 1), 'The shape should define the size of all but one dimension'

        requestMsg = h5proto.CreateDataSetReq()
        requestMsg.path = h5Path
        requestMsg.rank = rank
        requestMsg.word_size = wordSize
        if rank > 1:
            requestMsg.shape.extend(shape)
        if dataType == 'UINT':
            requestMsg.type = commonproto.ENC_UINT
        elif dataType == 'SINT':
            requestMsg.type = commonproto.ENC_SINT
        else:
            requestMsg.type = commonproto.ENC_FLOAT

        self.requestReply(requestMsg, commonproto.OkRep())

    def addAttribute(self, h5Path, attrName, attrVal):
        """ Add a single attribute (name - value pair) to an object
            An attribute can only be a scalar integer, float or string.

            :param h5Path: Path of the group/dataset for which the attribute need to written
            :param attrName: Name of the arrtibute
            :param attrVal: Attribute value

            :note: If an attribute already exists in the HDF5 file, its value is replaced.
        """
        attribute = {}
        attribute[attrName] = attrVal
        self.addAttributes(h5Path, attribute)

    def addAttributes(self, h5Path, attrDict):
        """ Add multiple attributes (name - value pairs) to an object.
            An attribute can only be a scalar integer, float or string.

            :param h5Path: Path of the group/dataset for which the attributes need to written
            :param attrDict: Dictionary of the attibute names and values.

            :note: If an attribute already exists in the HDF5 file, its value is replaced.
        """
        requestMsg = h5proto.AddAttributesReq()
        requestMsg.object_name = h5Path
        for name, value in attrDict.items():
            attr = h5proto.Attribute()
            attr.name = name
            if isinstance(value, int):
                attr.vint = value
            elif isinstance(value, float):
                attr.vdbl = value
            elif isinstance(value, str):
                attr.vstr = value
            elif value is None:
                logger.warning('Skipping attribute %s, cannot save None as value', name)
                continue
            else:
                raise TypeError('Attribute {} has unknown type for value: {}'.format(name, value))
            requestMsg.attributes.extend([attr])        #Append does not work.

        self.requestReply(requestMsg, commonproto.OkRep())

    def createTable(self, h5Path, members, numRecords=None):
        """ Create a new table (i.e. a compound dataset).

            In a table, each record is one sample of a compound type. An initial table size can
            be specified but is not necessary as the table will be extended as necessary.

            :param h5Path: Path and name of the new dataset in the HDF5 File
            :param members: The name, type, and size of all the dataset's members
            :param numRecords: The initial length of the table
        """
        requestMsg = h5proto.CreateTableReq()
        requestMsg.path = h5Path

        if numRecords is not None:
            requestMsg.num_records = numRecords

        formatChars = ['>']     # use big-endian byte order
        for member in members:
            name, encoding, wordSize = member[0:3]

            try:
                formatChar = getFormatCharacter(encoding, wordSize)
            except TypeError as e:
                logger.error("%s for %s. Aborting table creation.", e, name)
                return

            memberMsg = requestMsg.members.add()
            memberMsg.name      = name
            memberMsg.type      = encoding
            memberMsg.word_size = wordSize

            # If num_value defined (array):
            if len(member) > 3:
                memberMsg.num_values = member[3]
                formatChar *= member[3]

            formatChars.append(formatChar)

        # Create data format string and store mapping
        self.dataFormats[h5Path] = ''.join(formatChars)

        self.requestReply(requestMsg, commonproto.OkRep())

    def writeTable(self, h5Path, data, dataFormat=None, recordIndex=None):
        """ Writes a record to an existing table. If the recordIndex is not specified, the
            data will be appended to the end of the table.

            :param str   h5Path:            Name of the table
            :param tuple data:              Values that will be sent.
            :param str   dataFormat:        Structure for serializing the data (deprecated)
            :param int   recordIndex:       Index to which the data is written (optional).

            :note: Specifying the data format is no longer necessary as it is auto-generated on
                table creation. If specified, it will be ignored.
        """
        if dataFormat is not None:
            logger.warning("Specifying data format is deprecated as it is now auto-generated.")

        try:
            dataFormat = self.dataFormats[h5Path]
        except KeyError:
            logger.error("Could not find stored data format for h5 path: %s", h5Path)
            return

        # flatten any lists, arrays, etc...
        data = list(flatten(data))

        requestMsg = h5proto.WriteTableReq()
        requestMsg.path = h5Path
        requestMsg.data = struct.pack(dataFormat, *data)
        requestMsg.data_format = dataFormat
        if recordIndex is not None:
            requestMsg.record_index = recordIndex

        self.requestReply(requestMsg, commonproto.OkRep())
