""" Contains H5ArchiverServer

"""
import logging

from gssw.datawriter.h5worker import H5ArchiverWorker
from gssw.lib.arrays.gridded import GriddedArrayMsg
from gssw.lib.rpcbase import BaseServer
from gssw.lib.rpcmsg import DelegatingReqRepService, DelegatingPubSubService
from gssw.lib.sockets import tagName, SubSocket, RepSocket

from gssw.lib.typing import check_class

from gssw.lib.arrays.gridded import griddedArrayFromBinMessage

from gssw.proto.hdf5writer_pb2 import Hdf5DataWriterCommandServiceSpec, \
                                        Hdf5DataWriterDataServiceSpec

from gssw.proto.common_pb2 import OkRep

logger = logging.getLogger(__name__)


class _H5ArchiverPubSubService(DelegatingPubSubService):
    """ Class that overrides the processBinaryMessage of the DelegatingReqRepService

        Is typically not necessary for other servers (add DelegatingReqRepService directly instead.)
    """
    def processBinaryMessage(self, tag, binMessage):
        """ Processes a message.

            In case the message consists of a single array the array is appended
        """
        logger.debug("Processing message: {}".format(tag))
        if tag == tagName(GriddedArrayMsg):
            array = griddedArrayFromBinMessage(binMessage)
            self.worker.appendArray(array)
            replyMsg = OkRep()

            binReplyMessage = replyMsg.SerializeToString()
            return tagName(replyMsg), binReplyMessage
        
        return super().processBinaryMessage(tag, binMessage)



# PTK: same for socket. Can't we make bindRepSocket and bindSubSocket methods, just like the
# Quicklooker?

class H5ArchiverServer(BaseServer):
    """ Server that allows remote procedure calls on a H5ArchiverWorker object
    """
    def __init__(self, worker, repSocketAddress=None, subSocketAddress=None):
        """ Constructor
        """
        super().__init__()
        check_class(worker, H5ArchiverWorker)
        self._worker = worker

        self._repSocket = RepSocket()
        self._repSocket.createZmqSocket()
        if repSocketAddress is not None:
            self.bindRepSocket(repSocketAddress)

        self.addService(DelegatingReqRepService(self._worker,
                                                Hdf5DataWriterCommandServiceSpec(),
                                                self._repSocket))

        self._subSocket = SubSocket()
        self._subSocket.createZmqSocket()
        self._subSocket.subscribe('') # Subscribe to all topics
        if subSocketAddress is not None:
            self.connectSubSocket(subSocketAddress)

        self.addService(_H5ArchiverPubSubService(self._worker,
                                                 Hdf5DataWriterDataServiceSpec(),
                                                 self._subSocket))

    def finalize(self):
        """ Will close the archive. Must be called explicitly.
        """
        self._worker.finalize()
        super().finalize()

    @property
    def worker(self):
        """ The underlying worker (a H5ArchiverWorker)
        """
        return self._worker


    def bindRepSocket(self, address):
        """ Binds the Reply socket to the address
        """
        self._repSocket.bindSocket(address)


    def connectSubSocket(self, address):
        """ Connects the Subscribe socket to the address
        """
        self._subSocket.connectSocket(address)


    # Keep this to be able to use the quicklooker without a proxy (in future projects and testing).
    def bindSubSocket(self, address):
        """ Binds the Subscribe socket to the address
        """
        self._subSocket.bindSocket(address)

