# pylint: disable=bad-whitespace

""" Module for logging data to HDF-5
"""
from collections import namedtuple
import logging, os.path, time, struct
from threading import Timer, RLock
import h5py
import numpy as np

from gssw.lib.arrays.gridded import GriddedArray
from gssw.lib.arrays.semantic import SemanticArray
from gssw.lib.typing import check_class, check_is_a_numpy_array, check_is_a_string
import gssw.proto.hdf5writer_pb2    as      h5proto
import gssw.proto.common_pb2        as      commonproto


logger = logging.getLogger(__name__)

DatasetSpec = namedtuple('DatasetSpec', ['path', 'concatDim'])
DatasetSpec.__doc__ = "Specifies how the H5ArchiverServer creates an HDF dataset for an array"
try:
    DatasetSpec.path.__doc__     = "The full path of the created dataset "
    DatasetSpec.concatDim.__doc__= "The dimension to which newly received data is appended"
except AttributeError:
    pass # Property docstrings are only writeable in Python 3.5 and higher

def copyAttributesToDataset(dataset, semanticArray, attrNames):
    """ Takes the attributes from the semanticArray and puts them in the HDF-5 dataset.

        If the attribute already exists the function checks if the values are the same, and gives
        a warning it this is not the case. The dataset attribute is always overridden.

        :param dataset: h5py.dataset
        :param semanticArray: a SemanticArray
        :param attrNames: list of attribute names to copy

    """
    check_class(semanticArray, SemanticArray)

    for attrName in attrNames:
        attrValue = getattr(semanticArray, attrName)
        if attrValue is not None: # Check if attribute is present
            if attrName in dataset.attrs and attrValue != dataset.attrs[attrName]:
                logger.warning("Overriding {} attribute with: {}".format(attrName, attrValue))
            dataset.attrs[attrName] = attrValue


def appendArrayToDataset(dataset, array, axis):
    """ Appends the array to the h5py dataset along the appending (concatenating) dimension

        :param dataset: h5py.dataset to which the data is appended.
        :param semanticArray: a numpy array (including SemanticArrays)
        :param axis: axis to which the array is appended
    """
    check_is_a_numpy_array(array)

    oldSize = dataset.shape[axis]
    newSize = oldSize + array.shape[axis] if axis <= (len(array.shape) - 1) else oldSize + 1
    dataset.resize(newSize, axis=axis)

    # Set: dataset[..., oldSize:newSize, ...] = array (for the appending axis)
    indices = [slice(None)] * len(dataset.shape)
    indices[axis] = slice(oldSize, newSize)
    dataset[tuple(indices)] = array if axis <= (len(array.shape) - 1) else \
                              np.expand_dims(array, axis)



class H5ArchiverWorker():
    """ Component that archives data to an HDF-5 file.
    """
    ACTION_IGNORE  = 'ignore'
    ACTION_WARNING = 'warn'
    ACTION_RAISE   = 'raise'

    ENUM_ACTIONS = (ACTION_IGNORE, ACTION_WARNING, ACTION_RAISE)

    def __init__(self, unknownArrayAction=ACTION_WARNING):
        """ Constructor
        """
        if unknownArrayAction not in H5ArchiverWorker.ENUM_ACTIONS:
            raise ValueError("unknownArrayAction must be one of {}, got: {}"
                             .format(H5ArchiverWorker.ENUM_ACTIONS, unknownArrayAction))

        self.unknownArrayAction = unknownArrayAction

        self._arraySpecs = {} # dictionary that specifies how to create a dataset for arrays
        self._rootGroup = None  # An h5py File object.

        self.basefileName = None
        self.fileMode = None
        self.rotationTime = None
        self.timer = Timer(86400, self._updateFileName)
        self.lock = RLock()

        # Archive all attributes except name
        self.attrNamesToSave = [name for name in SemanticArray.METADATA_ATTRIBUTES
                                if name != 'name']

    def finalize(self):
        """ Can be used to cleanup resources. Current implementation only closes the file.
            Must be called explicitly.
        """
        with self.lock:
            self.closeFile()


    def fileName(self):
        """ The of the file to which the worker saves the data.

            :returns: the  absolute, normalized path of the HDF file.
                If the archive is closed, None is returned.
        """
        with self.lock:
            if self._rootGroup is None:
                return None
            
            return os.path.normpath(os.path.realpath(self._rootGroup.filename))


    def isOpen(self):
        """ Returns True if the archive is currently open.
        """
        with self.lock:
            return self._rootGroup is not None

    def fileOperations(self, action, path=None, rotationTime=0):
        """ Unified Function for file open and close requests.

        :param action:          Protobuf request for file operaion
        :param path:            Path of the HDF5 file. Optional. If not given, the file is 
                                assumed to be in the current direction
        :param rotationTime:    HDF5 File rotation time
        """
        logger.info("the action is: {}".format(action))
        if action == h5proto.FileOperationReq.OPEN:
            self.openFile(path, rotationTime, 'w')
        elif action == h5proto.FileOperationReq.CLOSE:
            self.closeFile()

    def fileStatus(self):
        """ Request whether a file is open """
        return self.fileName()

    def openFile(self, fileName, rotationTime=None, mode='w'):
        """ Opens a new HDF5 archive.

            Closes the old archive if one is still open.

        :param fileName:        Name of the HDF5 file
        :param rotationTime:    HDF5 File rotation time
        :pram  mode:            File open mode
        """
        if self.isOpen():
            self.closeFile()

        if rotationTime:
            logger.info("Started rotation logger function")
            with self.lock:
                self.rotationTime = rotationTime
                self.fileMode = mode
                self.basefileName = fileName
                saveFileName = self.basefileName[0:-3] + "_" + time.strftime("%Y%m%d-%H%M") + \
                                self.basefileName[-3:]
                logger.info("reached inside lock")
                self.timer = Timer(self.rotationTime, self._updateFileName)
                self.timer.start() 
                self._rootGroup = h5py.File(saveFileName, mode=self.fileMode)
        else:
            logger.info("Opening archive: {}".format(os.path.realpath(fileName)))
            self._rootGroup = h5py.File(fileName, mode=mode)

    def _updateFileName(self):
        """ Timer function that is called at the timer event so change the filename.
        """
        with self.lock:
            self.closeFile()
            saveFileName = self.basefileName[0:-3] + "_" + time.strftime("%Y%m%d-%H%M") + \
                                self.basefileName[-3:]
            self._rootGroup = h5py.File(saveFileName, mode=self.fileMode)
            self.timer = Timer(self.rotationTime, self._updateFileName)
            self.timer.start()

    def closeFile(self):
        """ Closes the underlying HDF-5 file
        """
        with self.lock:
            self.timer.cancel()
            if self._rootGroup is not None:
                logger.info("Closing archive: {}" \
                    .format(os.path.realpath(self._rootGroup.filename)))
                self._rootGroup.close()
                self._rootGroup = None
            else:
                logger.info("Archive was already closed. Ignored call.")


    def createGroup(self, path):
        """ Create a new group

            :param path: String. The HDF5 path where the group will be created. 
        """

        check_is_a_string(path)
        with self.lock:
            self._rootGroup.create_group(path)


    def createDataset(self, path, type=None, word_size=None, rank=None, shape=None):
        # pylint: disable=W0622, W0613
        """ Create a new empty dataset

            :param path: String. The HDF5 path where the dataset will be created.
            :param type: Type of encoding of the data in the dataset. 
                        Eg.: UINT/SINT/FLOAT/ASCII STRING
            :param word_size: The number of bits used to encode each word must be 8, 16, 32 or 64
            :param rank: Not used for python datawriter.
            :param shape: Shape of the dataset.
        """

        check_is_a_string(path)

        if type == commonproto.ENC_UINT:
            dtype = 'u'
        elif type == commonproto.ENC_SINT:
            dtype = 'i'
        else:
            dtype = 'f'

        # The protobufs define size in bits whereas h5py uses NumPy convention of bytes.
        dtype = '>' + dtype + str(int(word_size / 8))

        # Append a time dimension to the dataset shape.
        shape.insert(0, 0)
        with self.lock:
            self._rootGroup.create_dataset(path, dtype = dtype, shape = shape)


    def createTable(self, path, members, num_records):
        """ Create a new table (i.e. a compound dataset).

            In a table, each record is one sample of a compound type.

            :param path: Path and name of the new dataset in the HDF5 File.
            :param members: The name, type, and size of all the dataset's members
            :param num_records: The length of the table (for future data write actions).
        """
        dtype = []
        for item in members:
            name = item["name"]
            if item['type'] == commonproto.ENC_UINT:
                dataType = '>' + 'u' + str(int(item['word_size']/8))
            elif item['type'] == commonproto.ENC_SINT:
                dataType = '>' + 'i' + str(int(item['word_size']/8))
            elif item['type'] == commonproto.ENC_STRING:
                # For string `word_size` is string length.
                dataType = '>' + 'S' + str(int(item['word_size']))
            else:
                dataType = '>' + 'f' + str(int(item['word_size']/8))
            itemTypeStruct  = (name, dataType)
            dtype.append(itemTypeStruct)
        with self.lock:
            self._rootGroup.create_dataset(path, dtype = dtype, shape = [num_records])


    def writeTable(self, path, data, data_format, record_index=None):
        """ Writes a record to an existing table.

            :param str   path:        Name of the table
            :param bytes data:        Data tuple packed using Struct and given data format. 
            :param str   data_format: Format used for packing the data to bytes using Struct.
            :param int   recordIndex: Index to which the data is written (optional).

            :note: Specify the endianness in the format, otherwise unexpected padding bytes might be
               inserted.
        """

        # Unpack data 
        unpackedData = struct.unpack(data_format, data)
        with self.lock:
            table = self._rootGroup.get(path)
            # If there is no record_index, the data is written to the first entry.
            if record_index is None:
                table[0] = unpackedData
            # else the data at the record index is overwritten.
            else:
                table[record_index] = unpackedData


    def registerArray(self, arrayName, h5Path, concatDim=0):
        """ Makes the archiver listen for GriddedArrays of this name.

            :param arrayName: any received array having this name will be written to the archive
            :param h5Path: the result will be appended to this HDF-5 dataset. The full path
                name should be specified.
            :param concatDim: dimension along which newly received data will be concatenated.
        """
        check_is_a_string(arrayName)
        check_is_a_string(h5Path)
        logger.debug("Registering array name {} to {}".format(arrayName, h5Path))
        spec = DatasetSpec(path=h5Path, concatDim=concatDim)
        self._arraySpecs[arrayName] = spec


    def unregisterArray(self, arrayName):
        """ Unregisters the Gridded array with name arrayName

        :param arrarName: Name of array to unmap from the HDF5 dataset.
        """
        logger.debug("Unregistering array name {}, which was {})"
                     .format(arrayName, self._arraySpecs[arrayName]))
        del self._arraySpecs[arrayName]


    def addAttributes(self, object_name, attributes):
        """
        Add attributes (name-value pairs) to HDF5 obejcts.

        :param object_name: HDF5 object with link to which the attributes are applied to
        :param attributes: Google Protobuf reapeated Attribute message type
        """
        check_is_a_string(object_name)
        assert self.isOpen(), "HDF5 File not Open."
        with self.lock:
            location = self._rootGroup.get(object_name)
            assert location is not None, "Group/Dataset not found."
            for item in attributes:
                for key in item:
                    if key == 'name':
                        attrName = item[key]
                    else:
                        attrValue = item[key]
                location.attrs[attrName] = attrValue


    def appendArray(self, array):
        """ Appends a gridded array to the corresponding dataset.

            :param GriddedArray array: A gridded array with data to append.

            Raises KeyError if the array name is not registered.
        """
        if not self.isOpen():
            logger.warning("No open archive: data discarded.")
            return

        check_class(array, GriddedArray)
        data = array.data

        # Check if the array is registered
        if data.name not in self._arraySpecs:
            if self.unknownArrayAction == H5ArchiverWorker.ACTION_IGNORE:
                return
            elif self.unknownArrayAction == H5ArchiverWorker.ACTION_WARNING:
                logger.warning("Received unregistered array name: {}".format(data.name))
                return
            elif self.unknownArrayAction == H5ArchiverWorker.ACTION_RAISE:
                raise TypeError("Received unregistered array name: {!r}".format(data.name))
            else:
                raise ValueError("Invalid unknownArrayAction: {}"
                                 .format(self.unknownArrayAction))

        # Add new array or append to an existing one.
        spec = self._arraySpecs[array.data.name]
        with self.lock:
            dataset = self._rootGroup.get(spec.path)
            if dataset is None:
                # Data set does not yet exist, create it.
                logger.debug("Dataset not found, creating one: {}".format(spec.path))

                # Extend 1D array to at least `spec.concatDim`.
                if (len(data.shape) - 1) < spec.concatDim:
                    data.shape = tuple(list(data.shape) +
                                 [1] * (spec.concatDim - len(data.shape) + 1))

                maxShape = list(data.shape)
                maxShape[spec.concatDim] = None
                dataset = self._rootGroup.create_dataset(spec.path, data.shape, data.dtype,
                                                            maxshape=maxShape)
                dataset[:] = data
                copyAttributesToDataset(dataset, data, self.attrNamesToSave)

                # Add dimension scales for the coordinate vectors
                for dimNr, coordVector in array.coords.items():
                    if coordVector.name is None:
                        logger.warning("No dimension scale for dim {}, \
                            coordinate vector has no name".format(dimNr))
                        continue

                    assert coordVector.ndim == 1, \
                        "Coordinate vector should be 1-dimensional. \
                        Got {}D".format(coordVector.ndim)

                    # Label the dimensions of the dataset
                    dataset.dims[dimNr].label = coordVector.name

                    # The dimension scale is resizable only for the concatenating dimension
                    maxShape = (None, ) if dimNr == spec.concatDim else None

                    dimScale = dataset.parent.create_dataset(coordVector.name, coordVector.shape,
                                                             coordVector.dtype, maxshape=maxShape)
                    dimScale[:] = coordVector
                    copyAttributesToDataset(dimScale, coordVector, self.attrNamesToSave)

                    # Turn the dimScale into a HDF dimension scale and attach it to the dimension
                    dataset.dims.create_scale(dimScale, coordVector.name)
                    dataset.dims[dimNr].attach_scale(dimScale)
            else:
                # Data set already exists. Append data.
                appendArrayToDataset(dataset, data, spec.concatDim)
                copyAttributesToDataset(dataset, data, self.attrNamesToSave)

                # Append the coordinate vector for the concatenating dimension. The other dimensions
                # are checked: a ValueError is raised if they are not the same as in the existing
                # dataset.
                for dimNr, coordVector in array.coords.items():
                    if coordVector.name is None:
                        logger.warning("No dimension scale for dim {}, \
                            coordinate vector has no name".format(dimNr))
                        continue

                    assert coordVector.ndim == 1, \
                        "Coordinate vector should be 1-dimensional. \
                            Got {}D".format(coordVector.ndim)

                    dimScale = dataset.parent.get(coordVector.name)
                    assert dimScale, "dimension scale not found for: {!r}".format(coordVector.name)

                    if dimNr == spec.concatDim:
                        # Append coordinate vector for the concatenating dimension.
                        appendArrayToDataset(dimScale, coordVector, 0)
                        copyAttributesToDataset(dimScale, coordVector, self.attrNamesToSave)
                    else:
                        # For other dimensions check that the coordvector remains the same.
                        # Perhaps we should make this check optional.
                        if not np.equal(dimScale[:], coordVector):
                            raise ValueError("Coordinate vector changed for dim: {}".format(dimNr))
                        copyAttributesToDataset(dimScale, coordVector, self.attrNamesToSave)

