""" KIDS datawriter client """

import logging
import threading
import time

from gssw.datawriter.scienceclient  import Hdf5ScienceClient
from gssw.proto.common_pb2          import OkRep
from gssw.proto.kidswriter_pb2      import ChannelType

import gssw.proto.kidswriter_pb2 as kidsproto

logger = logging.getLogger(__name__)

###############################################################################
#  Constants
###############################################################################

HDF5FILE_FORMAT_VERSION = "KID_1.0"

###############################################################################
# Helper functions
###############################################################################

def getHDF5Group(rack_index, channel_index):
    """ Return a string representation of the group path for a given channel. """
    return "/rack_{:02d}/channel_{:02d}".format(rack_index, channel_index)

def timeInMillis():
    """ Returns timestamp in millis as int (to match timestamp in HDF5 dataset). """
    return int(round(time.time() * 1000))

###############################################################################
# Class Hdf5KidsClient
###############################################################################

class Hdf5KidsClient(Hdf5ScienceClient):
    """ Client class for KIDs HDF5 writer """

    def __init__(self, rack_index, hal, address=None, reqSocket=None):
        super().__init__(hal, address, reqSocket)

        self.lock           = threading.Lock()

        self.rnum           = rack_index
        self.filename       = None
        self.autoClose      = True
        self.mappedChannels = 0
        self.status         = {}    # Channel status

    def initializeWriter(self, rackIndex, filename, attributes=None, autoClose=True):
        """ Initialize writer in HW daemon

        :param rackIndex:   The index of the rack
        :param filename:    The path to the file
        :param attributes:  Optional attributes to write to group '/'
        :param autoClose:   Close HDF5 file when last channel mapping is removed
        """
        assert rackIndex == self.rnum, "Datawriter rack ({}) does not match given rack ({})".format(
            rackIndex, self.rnum)

        logger.info("\tHDF5: opening file: '%s'", filename)
        self.openFile(filename)

        logger.debug("\tHDF5: writing attributes to '/'")
        if attributes is None:
            attributes = {}
        attributes.update({"time_start_utc":    timeInMillis(),
                           "time_end_utc":      0,                  # set when measurement is done
                           "end_status":        "OK"})              # updated when finished
        self.addAttributes("/", attributes)

        group = "/rack_{:02d}".format(rackIndex)
        logger.debug("\tHDF5: creating group: '%s'", group)
        self.createGroup(group)

        self.filename  = filename
        self.autoClose = autoClose

    def mapChannel(self, channelIndex, boardSlot, skdpChannel, bins,
                   groupAttributes=None, dataAttributes=None):
        """ Maps the output from an ADC board to the specified rack and channel.

        This will create a group in the HDF5 file (/rack_###/channel_###), add attributes to the
        group, add the bin mapping as a dataset, and add the block data from any HWBlockRep messages
        received from the specified board slot to a dataset (/rack_###/channel_###/data).

        :param channelIndex:    The index of the channel
        :param boardSlot:       The slot of the (ADC) board
        :param skdpChannel      The ChannelType to map
        :param bins:            The bin mapping
        :param groupAttributes: Dictionary of attributes to write to channel group (hardware info)
        :param dataAttributes:  Dictionary of attributes to write to dataset (measurement info)
        """
        assert self.filename, "Datawriter is not initialized"
        assert skdpChannel in ChannelType.values(), "SKDP channel type not valid"

        group = getHDF5Group(self.rnum, channelIndex)
        logger.debug("\tHDF5: creating group: '%s'", group)
        self.createGroup(group)

        if groupAttributes:
            logger.debug("\tHDF5: adding channel attributes to: '%s'", group)
            self.addAttributes(group, groupAttributes)

        path        = "{}/data".format(group)
        logger.info("\tHDF5: mapping board in slot %d to: '%s'", boardSlot, path)

        req         = kidsproto.MapChannelDataReq()
        req.board   = boardSlot
        req.path    = path
        req.channel = skdpChannel
        req.bins.extend(bins)
        self.requestReply(req, OkRep())

        if dataAttributes:
            logger.debug("\tHDF5: adding measurement attributes to: '%s'", path)
            self.addAttributes(path, dataAttributes)

        self.status[channelIndex] = "OK"
        self.mappedChannels += 1

    def unMapChannel(self, channelIndex, boardSlot, skdpChannel):
        """ Unmaps the output from an ADC board to the specified rack and channel.

        This removes the mapping from the specified channel which direc

        :param channelIndex:    The index of the science channel
        :param boardSlot:       The slot of the (ADC) board
        :param skdpChannel      The ChannelType to map
        """
        assert skdpChannel in ChannelType.values(), "SKDP channel type not valid"
        assert channelIndex in self.status, "Channel {} was not mapped".format(channelIndex)
        logger.info("\tHDF5: Unmapping channel %d, slot %d", channelIndex, boardSlot)

        req         = kidsproto.UnMapChannelDataReq()
        req.board   = boardSlot
        req.channel = skdpChannel
        self.requestReply(req, OkRep())
        self.mappedChannels -= 1

        # write channel status
        group = getHDF5Group(self.rnum, channelIndex)
        logger.debug("\tHDF5: adding final channel attributes to: '%s'", group)
        self.addAttributes(group, {"end_status": self.status[channelIndex]})

        if self.autoClose and self.mappedChannels <= 0:
            self.stopWriter()

    def stopWriter(self):
        """ Stop writer (close file) in HW daemon and write final attributes to root group. """
        assert self.filename, "The datawriter is not active (no file open)"

        logger.debug("\tHDF5: writing final attributes to '/'")
        attributes = {
            "end_status":   "ABORTED" if any(s != "OK" for s in self.status.values()) else "OK",
            "time_end_utc": timeInMillis()
        }
        self.addAttributes("/", attributes)

        logger.info("\tHDF5: closing file: '%s'", self.filename)
        self.closeFile()

        # reset status
        self.filename = None
        self.status   = {}
