""" HDF5 client for recording science data using the C++ datawriter in the HWDaemon
"""
import logging
import gssw.proto.hdf5writer_pb2    as      h5proto
from gssw.proto.common_pb2          import  OkRep
from gssw.lib.typing                import  check_class
from gssw.hal.hal                   import  Hal
from gssw.datawriter.baseclient import Hdf5BaseClient

logger = logging.getLogger(__name__)

class Hdf5ScienceClient(Hdf5BaseClient):
    """ Hdf5ScienceClient client class """

    def __init__(self, hal, address=None, reqSocket=None):

        check_class(hal, Hal)
        self.hal        = hal
        super().__init__(address=address, reqSocket=reqSocket)


    def mapDataSet(self, registerName, dataSetPath):
        """ Map a data stream from register to data set.

            registerName - the name of the source register (device:register:offset)
            dataSetPath  - the path of the destination data set in the output file
        """

        # Get register address from the HAL
        device, register = registerName.split(':')
        registerAddress = self.hal.getRegisterInfo(device, register).address

        # Create request protobuffer
        req = h5proto.MapDataStreamReq()
        req.register_address = registerAddress
        req.dataset_path = dataSetPath

        _replyMsg = self.requestReply(req, OkRep())


    def mapMultipleDataSets(self, registerName, dataSetPaths):
        """ Map a data stream from registers to multiple data sets.

            registerName - the name of the source register (device:register:offset)
            dataSetPaths - list of paths of the destination data sets in the output file

            This mapping is used for registers that generate composite data (records).
            The datawriter de-interleaves the data stream and writes the individual elements
            to the corresponding datasets. It is the resposibility of the user to ensure the number
            and order of elements in the composite data stream matches the number and order
            of datasets it is mapped to.
        """

        # Get register address from the HAL
        device, register = registerName.split(':')
        registerAddress = self.hal.getRegisterInfo(device, register).address

        # Create request protobuffer
        req = h5proto.MapCompositeDataStreamReq()
        req.register_address = registerAddress
        req.dataset_paths.extend(dataSetPaths)

        _replyMsg = self.requestReply(req, h5proto.MapCompositeDataStreamRep())
