#!/usr/bin/env python3

""" test script for the TES/FDM HDF5 datawriter

    To use this test script, the port the writer listens for science data must be changed.
    In filewriter.cc, in the FileWriter constructor:

        change: ``zmqsub_->Connect(SocketAddrTcp("localhost", sciencePort));``
        to:     ``zmqsub_->Connect(SocketAddrTcp("localhost", 9991));``
"""

import getpass
import os
import struct
from time                           import sleep

from gssw.hal.hal                   import Hal
from gssw.hal.halclient             import HalClient
from gssw.lib.sockets               import PubSocket
from gssw.datawriter.tesfdmclient   import TesFdmClient
from gssw.proto.hwregister_pb2      import HwBlockRep
from gssw.proto.common_pb2          import ENC_UINT, ENC_SINT
from gssw.lib.misc                  import createGsswArgumentParser
from gssw.common                    import initGssw




def sendData(socket, encoding, bits):
    """ send data to hardware daemon """

    assert bits in [16, 32]
    assert encoding in [ENC_SINT, ENC_UINT]

    rep = HwBlockRep()
    rep.write = False
    rep.addr = 0x0
    rep.data.bits_per_value = bits
    rep.data.value_encoding = encoding

    if bits == 16 and encoding == ENC_SINT:
        rep.data.block_data = struct.pack('16h', *list(range(16)))
    if bits == 16 and encoding == ENC_UINT:
        rep.data.block_data = struct.pack('16H', *list(range(16)))
    if bits == 32 and encoding == ENC_SINT:
        rep.data.block_data = struct.pack('16i', *list(range(16)))
    if bits == 32 and encoding == ENC_UINT:
        rep.data.block_data = struct.pack('16I', *list(range(16)))

    # print(rep)
    # print('msg size = %d' % len(rep.SerializeToString()))
    # print('data size = %d' % len(rep.data.block_data))
    socket.publishProtoBuf('', 'HwBlockRep', rep.SerializeToString())
    sleep(4)




CHANNEL_INDEX = 0
NUM_PIXELS = 4
SAMPLE_RATE = 1000


TESTFILE = '/data/' + getpass.getuser() + '/pythontests/tesfdm_test.h5'

def main():
    """ Test function """
    args = createGsswArgumentParser(__doc__).parse_args()

    # Retrieve Egress configuration
    config = initGssw(args)

    # Create a dummy science socket
    socket = PubSocket()
    socket.createZmqSocket()
    socket.bindSocket('tcp://*:9991')
    sleep(1) # slow joiner

    # remove old output file
    try:
        os.remove(TESTFILE)
    except:
        pass

    # Get the HAL
    hal = Hal()
    halClient = HalClient(hal, config)
    halClient.requestHal()

    # create a datawriter client
    client = TesFdmClient(hal, "tcp://localhost:40300")

    # open output file
    client.openFile(TESTFILE)

    # initialize writer
    client.initializeWriter(0, 4, 1000, 22, 1.0, [1.0], 1000)

    # Create file structure defined in the specification
    client.createGroup('/mux')
    client.createGroup('/mux/channel_%03d' % CHANNEL_INDEX)
    client.createGroup('/mux/channel_%03d/ADC1' % (CHANNEL_INDEX))
    client.createGroup('/mux/channel_%03d/DAC1' % (CHANNEL_INDEX))
    client.createGroup('/mux/channel_%03d/DAC2' % (CHANNEL_INDEX))
    for i in range(NUM_PIXELS):
        client.createGroup('/mux/channel_%03d/freq_%03d' % (CHANNEL_INDEX, i))


    # create a mapping for a single pixel and send data (event 0)
    client.mapPixel('demux5', 'DUMMY', 0, 'ampl')
    sendData(socket, ENC_SINT, 32)

    # create a mapping for a single pixel and send data (event 1)
    client.mapPixel('demux5', 'DUMMY', 1, 'IQ')
    sendData(socket, ENC_SINT, 32)

    # create a mapping for raw ADC data and send data (event 2)
    client.mapRaw('demux5', 'DUMMY', 'ADC1')
    sendData(socket, ENC_UINT, 16)

    # create a mapping for raw ADC data and send data (event 3)
    client.mapAllPixels('demux5', 'DUMMY', 'ampl-phase')
    sendData(socket, ENC_SINT, 32)

    # close the output file
    client.closeFile()

if __name__=='__main__':
    main()

