""" TES-FDM datawriter client """

import time
from gssw.datawriter.scienceclient import Hdf5ScienceClient
import gssw.proto.tesfdmwriter_pb2 as tesfdmproto
from gssw.proto.common_pb2 import OkRep

class TesFdmClient(Hdf5ScienceClient):
    """ Client class for TES FDM HDF5 writer """

    SCIENCE_TYPE_MAP = {'ampl'      : tesfdmproto.TYPE_AMPL,
                        'phase'     : tesfdmproto.TYPE_PHASE,
                        'ampl-phase': tesfdmproto.TYPE_AMPL_PHASE,
                        'IQ'        : tesfdmproto.TYPE_IQ}

    RAW_SOURCE_MAP = {'ADC1': tesfdmproto.SOURCE_ADC1,
                      'DAC1': tesfdmproto.SOURCE_DAC1,
                      'DAC2': tesfdmproto.SOURCE_DAC2}

    def initializeWriter(self, channelIndex, numPixels, sampleRate, wordSize,
                         scale_phase_rad, scale_volt_demux, pretrig_samples):
        """ Initialze writer in HW daemon """

        # create request protobuf message
        req = tesfdmproto.InitializeReq()
        req.channel_index         = channelIndex
        req.num_pixels            = numPixels
        req.sample_rate           = sampleRate
        req.word_size             = wordSize
        req.scale_phase_rad       = scale_phase_rad
        req.scale_volt_demux.extend(scale_volt_demux)
        req.pretrig_samples       = pretrig_samples
        req.start_time            = float(time.time())

        self.requestReply(req, OkRep())


    def mapPixel(self, device, register, pixelIndex, dataType):
        """ Create pixel mapping in HW Daemon """

        assert dataType in self.SCIENCE_TYPE_MAP

        # Get register address from the HAL
        registerAddress = self.hal.getRegisterInfo(device, register).address

        # create request protobuf message
        req = tesfdmproto.MapPixelReq()
        req.register_address = registerAddress
        req.pixel_index = pixelIndex
        req.data_type = self.SCIENCE_TYPE_MAP[dataType]

        _replyMsg = self.requestReply(req, OkRep())


    def mapAllPixels(self, device, register, dataType, activePixels=None):
        """
         Map all-pixel data in HW Daemon.

        In this case the hardware generates data for all pixels in case a trigger is detected.

        However, many pixels do not contain data (because they are not used), so we should not store
        them, to prevent generating huge files.

        :param str  device:       Hal device name of device being mapped.
        :param str  register:     Name of mapped register, used to look up the hardware address.
        :param enum dataType:     Specifies what kind of data (ampl, phase, etc.).
        :param list activePixels: Pixels which generate actual data an must be stored (optional).
                                  When omitted, all pixels will be stored, incuding inactive ones.
        """

        assert dataType in self.SCIENCE_TYPE_MAP

        # Get register address from the HAL
        registerAddress = self.hal.getRegisterInfo(device, register).address

        # create request protobuf message
        req = tesfdmproto.MapAllPixelsReq()
        req.register_address = registerAddress
        req.data_type = self.SCIENCE_TYPE_MAP[dataType]

        if activePixels is not None:
            req.active_pixels.extend(activePixels)

        _replyMsg = self.requestReply(req, OkRep())


    def mapRaw(self, device, register, dataSource):
        """ Map raw data in HW Daemon """

        assert dataSource in self.RAW_SOURCE_MAP

        # Get register address from the HAL
        registerAddress = self.hal.getRegisterInfo(device, register).address

        # create request protobuf message
        req = tesfdmproto.MapRawReq()
        req.register_address = registerAddress
        req.data_source = self.RAW_SOURCE_MAP[dataSource]

        _replyMsg = self.requestReply(req, OkRep())
