#!/usr/bin/env python3

""" GSSw Logging Handler """

###################################################################################################
# P Y T H O N   M O D U L E S
###################################################################################################

import os
import inspect
import sys
import socket
import logging
import logging.handlers
import zmq

from gssw.proto.logger_pb2 import SyncMsg, LoggerMsg

VERBOSE = 0
def vprint(*args, **kwargs): 
    """ Print if VERBOSE = 1 or do nothing. """
    if VERBOSE:
        print(*args, file=sys.stderr, flush = True, **kwargs)
        

###################################################################################################
# L O G G I N G H A N D L E R
###################################################################################################

class ZmqSocketHandler(logging.Handler):
    """
    A handler class which writes logging records, in Google Protocol buffer format,
    to a ZMQ PUB socket connected to the logserver (being the only subscriber). 
    The socket is kept open across logging calls.
    If the PUB socket is not yet present, an attempt is made to create it.
    
    This handler is based on logging.SocketHandler.
    (https://hg.python.org/cpython/file/3.5/Lib/logging/handlers.py)
    In order to mitigate the 'slow subscriber' phenomenon, a handshaking mechanism has been 
    implemented which is loosely based on the Node Coordinations pattern from the ZMQ Guide.
    (http://zguide.zeromq.org/page:chapter2#Node-Coordination, in particular "A more robust model
    could be...")
    
    Handshaking mechanism: 
    Send PUB messages of type 'SyncMsg' to the logserver.
    A 'SyncMsg', has just one field: the pathname of the calling logger. The logserver
    should echo this 'SyncMsg' using its own PUB socket. The echo is sent to the loggers
    which are all subscribed to that socket. However, only the specific logger which has sent 
    the original 'SyncMsg' message should react on it. 
    After each transmission of a 'SyncMsg', the logger waits for the response from the 
    logserver. The timeout for this wait loop has been set to DEFAULT_POLL_TIMEOUT, defined 
    above. The number of attempts has been set to DEFAULT_POLL_RETRIES (also defined above).

    The wait loop is realized with help of a poller. This poller is registered with the SUB
    socket and is called with timeout set to DEFAULT_POLL_TIMEOUT. The poller fires either 
    when activity is detectected on the SUB socket or after DEFAULT_POLL_TIMEOUT, whichever 
    is first. 
    When activity has been detected (scenario 1), a non blocking read operation from the SUB 
    socket is carried out. If the received message corresponds to the original message sent 
    by the logger, then the self._connected variable is set. From that moment on, actual 
    logging messages (type 'loggerMsg') can be sent via the PUB socket to the logserver. 
    
    After completion of the handshaking loop, the SUB socket is closed and the poller is 
    unregistered. Moreover, if self._connected is not set, also the PUB socket is closed.
    When the handler is called again and the self._connected variable has not been set, the 
    handshaking loop starts all over again.       
    """

    DEFAULT_POLL_TIMEOUT = 1    # handshaking timeout time i.e. 1 ms
    DEFAULT_POLL_RETRIES = 10   # number of handshaking attempts

    def __init__(self, host, repPort, pubPort):
        """ Initializes the handler with a specific host address and port numbers. """
        super().__init__()   

        self._host      = host        
        self._pubPort   = pubPort # Port for sending log messages
        self._subPort   = repPort # Port for handshaking
        self._context   = zmq.Context.instance()
        self._pubSock   = None
        self._subSock   = None
        self._connected = False

        # get the name of the calling module for printing                              
        self.__module = os.path.basename(inspect.stack()[5][1])       
            
    def _connectSockets(self, syncMsg):  
        """
        Make a ZeroMQ PUB connection to the logserver. A handshaking mechanism is used to mitigate 
        the 'slow subscriber' phenomenon mentioned above.
        """
        
        # PUB socket for sending 'SyncMsg' and 'loggerMsg' type messages with the logserver being 
        # the only subscriber.
        vprint('(%s --> handler) creating PUB socket' % self.__module)
        self._pubSock = self._context.socket(zmq.PUB)       
        self._pubSock.connect('tcp://%s:%d' % (self._host, self._pubPort))
                    
        # SUB socket belonging to the handshaking mechanism 
        vprint('(%s --> handler) creating SUB socket' % self.__module)
        self._subSock = self._context.socket(zmq.SUB)
        self._subSock.setsockopt_string(zmq.SUBSCRIBE,"")  
        self._subSock.connect('tcp://%s:%d' % (self._host, self._subPort))

        # Poller belonging to the handshaking mechanism
        vprint('(%s --> handler) creating and registering poller' % self.__module)
        poll = zmq.Poller()
        poll.register(self._subSock, zmq.POLLIN)
        
        # Handshaking mechanism        
        self._connected = False
        retries_left = self.DEFAULT_POLL_RETRIES
        protoString = syncMsg.SerializeToString()
        multiPart = ["".encode('ascii'), type(syncMsg).__name__.encode('ascii'), protoString]
        reply = SyncMsg()        
        while retries_left:
            vprint('(%s --> handler) sending the SyncMsg via the PUB socket' % self.__module)
            self._pubSock.send_multipart(multiPart, flags=0, copy=True, track=False)
            events = dict(poll.poll(self.DEFAULT_POLL_TIMEOUT))
            retries_left -= 1

            if events.get(self._subSock) == zmq.POLLIN:
                vprint('(%s --> handler) detected activity on the SUB socket' % self.__module)
                rep = self._subSock.recv_multipart(flags=zmq.NOBLOCK, copy=True, track=False)
                reply.ParseFromString(rep[2])
                if reply.name == syncMsg.name:
                    self._connected = True  
                    vprint('(%s --> handler) connection with logserver established after' \
                        % self.__module, self.DEFAULT_POLL_RETRIES - retries_left, 'attempts')
                    break
            else:
                vprint('(%s --> handler) poller has timed out;' % self.__module, \
                    retries_left, 'retries left')                

        # the poller can be unregistered now
        vprint('(%s --> handler) unregister the poller' % self.__module)
        poll.unregister(self._subSock)        
        
        # the SUB socket can closed now 
        vprint('(%s --> handler) close the SUB socket' % self.__module)
        self._subSock.close(linger=0)        
        
        # if not connected, also close the PUB socket      
        if not self._connected:
            vprint('(%s --> handler) close the PUB socket' % self.__module)
            self._pubSock.close(linger=0)
            self._pubSock = None
            
    def _send(self, record):
        """
        Send a serialized string to the PUB socket.
        """
        # first create Google Protocol buffer messages    
        syncMsg   = self._makeSyncMsg(record)  
        loggerMsg = self._makeLoggerMsg(record)
        
        # check if the PUB socket already exists
        if self._pubSock is None:
            # if not, try to create a PUB socket
            vprint('(%s --> handler) try to get connection with logserver' % self.__module)
            self._connectSockets(syncMsg)

        if self._connected:
            # PUB socket did already exist or has been successfully created
            protoString = loggerMsg.SerializeToString()
            multiPart = ["".encode('ascii'), type(loggerMsg).__name__.encode('ascii'), protoString]
            self._pubSock.send_multipart(multiPart, flags=0, copy=True, track=False)
            vprint('(%s --> handler) sending log message to logserver' % self.__module)
        else:
            vprint('(%s --> handler) log message could not be sent to logserver' % self.__module)

    def _makeSyncMsg(self, record):
        """
        Create a syncMsg from record.
        """
        syncMsg = SyncMsg()
        syncMsg.name = record.pathname
        return syncMsg

    def _makeLoggerMsg(self, record):
        """
        Create a loggerMsg from record.
        """                
        loggerMsg                = LoggerMsg()
        loggerMsg.name           = record.name
        loggerMsg.level          = record.levelno
        loggerMsg.pathname       = record.pathname
        loggerMsg.lineno         = record.lineno
        loggerMsg.msg            = record.getMessage()
        loggerMsg.func           = record.funcName
        if record.stack_info is None:
            loggerMsg.sinfo      = ""
        else:
            loggerMsg.sinfo      = record.stack_info
        loggerMsg.processid      = record.process
        loggerMsg.ipaddress_port = socket.gethostbyname(self._host) + ':' + str(self._pubPort)
        loggerMsg.created        = record.created
        loggerMsg.msecs          = record.msecs
        # strip .py and possible / from component name
        loggerMsg.componentname  = os.path.basename(sys.argv[0])
        return loggerMsg

    def emit(self, record):
        """
        Emit a record.        
        """
        self._send(record)

    def close(self):
        """
        Closes the handlers. Should NOT close the sockets here!!!
        The program should do this by calling gssw.lib.sockets.destroyAllSocketsInContext at exit.
        """
        vprint('(%s --> handler) closing handler' % self.__module)
        self.acquire()
        try:
            logging.Handler.close(self)
        finally:
            self.release()
            vprint('(%s --> handler) closed handler' % self.__module)
            
