""" Module with dialog for Hal manager git upload function. """

import logging
import os

from PyQt5                  import uic
from PyQt5.QtCore           import pyqtSignal, pyqtSlot, QObject, QThread
from PyQt5.QtGui            import QIcon, QMovie

from gssw.hal.halclient     import HalClient
from gssw.lib.widgets.misc  import gsswUiError, gsswUiInfo, gsswUiWarning

__all__ = ['UploadDialog']

logger = logging.getLogger(__name__)

class UploadDialog(QObject):
    """ Creates a modal dialog with functionality to upload the changed HAL yaml file to git.

        The HAL client is used for the commit.
    """

    def __init__(self, guiPath, hal, config):
        """ Create a commit dialog.

        :param guiPath:   Directory which contains the root of the HAL manager gui files (``*.ui``).
        :param halClient: Client for committing the HAL.
        """

        super().__init__()
        # Create our own HalClient, since it is used from the background thread
        self.halClient = HalClient(hal, config)
        self._ui       = uic.loadUi(os.path.join(guiPath, 'HalManagerUploadDialog.ui'))

        # Prepare gif animation for displaying during uploading
        self.animationWaiting = QMovie(os.path.join(guiPath + "/images/ajax-loader.gif"))
        self.iconError        = QIcon( os.path.join(guiPath + "/images/icon-error.png"))
        self.iconOk           = QIcon( os.path.join(guiPath + "/images/icon-ok.png"))

        self._ui.buttonUpload.clicked.connect(self._slotButtonUploadClicked)
        self._ui.buttonCancel.clicked.connect(self._slotCancel)
        self.backgroundUploader = None

    def _startSaveButtonAnimation(self):
        """ Adds an animated waiting icon to the upload button """
        self.animationWaiting.start()
        self.animationWaiting.frameChanged.connect(self._updateSaveButtonAnimation)
        self._ui.buttonUpload.setEnabled(False)

    def _stopSaveButtonAnimation(self, success):
        """ Stops the animated upload button icon, and sets a succes/failure icon. """
        self.animationWaiting.frameChanged.disconnect(self._updateSaveButtonAnimation)
        self.animationWaiting.stop()
        if success:
            self._ui.buttonUpload.setIcon(self.iconOk)
        else:
            self._ui.buttonUpload.setIcon(self.iconError)
        self._ui.buttonUpload.setEnabled(True)

    def _updateSaveButtonAnimation(self, _):
        """ Runs the gif animation on the upload button icon """
        # Set the gif frame as the button icon
        self._ui.buttonUpload.setIcon(QIcon(self.animationWaiting.currentPixmap()))


    def _checkInput(self):
        """ Check whether the input fields are filled """
        if not self._ui.inputUsername.text():
            gsswUiWarning('Enter your username')
            return False
        if not self._ui.inputPassword.text():
            gsswUiWarning('Enter your password')
            return False
        if not self._ui.inputComment.toPlainText():
            gsswUiWarning('Enter a suitable description')
            return False
        return True

    def _slotButtonUploadClicked(self):
        """ Upload button starts the upload sequence (oh really?) """
        if not self._checkInput():
            # User didn't fill in all fields, so don't bother trying to send the request
            return
        self._startSaveButtonAnimation()

        # Do the update in the background (so the gui doesn't block)
        self.backgroundUploader = BackgroundUploader(self.halClient,
                                                     self._ui.inputUsername.text(),
                                                     self._ui.inputPassword.text(),
                                                     self._ui.inputComment.toPlainText())
        self.backgroundUploader.sigUploadCompleted.connect(self._slotUploadCompleted)
        self.backgroundUploader.sigUploadFailed.connect(self._slotUploadFailed)
        self.backgroundUploader.start()

    @pyqtSlot()
    def _slotCancel(self):
        """ Cancel button returns to the main screen . """
        self._ui.inputPassword.clear()
        self._ui.reject()

    @pyqtSlot()
    def _slotUploadCompleted(self):
        """ Slot called from background update when commit was successfull. """
        self._ui.inputPassword.clear()
        self._stopSaveButtonAnimation(True)
        gsswUiInfo("The HAL database was successfully uploaded.")
        self._ui.accept()

    @pyqtSlot(Exception)
    def _slotUploadFailed(self, ex):
        """ Slot called from background update when commit failed.

        :param ex: Exception details of the commit action.
        """
        self._stopSaveButtonAnimation(False)
        self._ui.inputPassword.clear()
        gsswUiError("Could not upload the HAL:\n\n{}".format(ex))
        self._ui.buttonUpload.setIcon(QIcon())


    def runDialog(self):
        """ Opens the dialog for uploading (blocking). """
        # Reset Upload button status icon
        self._ui.buttonUpload.setIcon(QIcon())

        # Show the dialog and wait until closed
        self._ui.exec()


class BackgroundUploader(QThread):
    """ Separate thread to do the HAL uploading in the background.

        :note: This is done in a thread because otherwise the gui is unresponsive, and the waiting
               icon animation freezes.
    """

    sigUploadCompleted = pyqtSignal()
    sigUploadFailed    = pyqtSignal(Exception)

    def __init__(self, halClient, username, password, comment):
        super().__init__()

        self._halClient = halClient

        # User input
        self._username = username
        self._password = password
        self._comment  = comment

    def run(self):
        """ Do the upload in a different thread. """
        try:
            # First update the version number and save the file
            #TODO: Move version nr update to server?
            hal = self._halClient.hal
            hal.updateVersionNumber()
            self._halClient.setHal(hal.createYamlData())

            # Then do the commit
            logger.info('Start upload')
            self._halClient.commit(self._username, self._password, self._comment)
            logger.info('Upload completed with success')
            self.sigUploadCompleted.emit()
        except Exception as ex:
            logger.exception('Upload failed')
            self.sigUploadFailed.emit(ex)
        finally:
            del self._username
            del self._password
