#!/usr/bin/env python3
# pylint: disable=C0302
# (too-many-lines)
# pylint: disable=R0912
# (too_many_branches)
# pylint: disable=R0913
# (too_many_arguments)
# pylint: disable=R0914
# (too-many-locals)
# pylint: disable=R0915
# (too-many-statements)
# pylint: disable=R0916
# (too-many-boolean-expressions)

""" GSSw Hardware Abstraction Layer

    Hal:         object with dict{ deviceName,      Device      } and
                             dict{ deviceTypeName,  DeviceType  } and
                             dict{ calibrationName, Calibration ) and
                             str versionNumber


    Device:      object with deviceTypeName and calirationName

    DeviceType   object with dict{ registerName,    Register    } and
                             dict{ constantName,    Constant    } and

    Register:    object with register info (name, address, access, unit, type, ...)

    Constant:    object with constant info (name, unit, type, ...)

    Calibration: object with dict{ constantName,    value       } and
                             dict{ registerName,    Limits      } and
                             dict{ registerName,    Conversion  }

    Limits:      object with limit info (valid range, hk limits, ...)

    Conversion:  object with conversion info (polynome, function, lut)
"""

###################################################################################################
# P Y T H O N   M O D U L E S
###################################################################################################
import collections
import logging
from operator        import itemgetter
from pathlib         import Path

import yaml

from sympy           import sympify, symbols
from gssw.lib.errors import DataError, ItemNotFoundError
from gssw.lib.typing import check_class, check_is_a_string, is_a_string, check_is_a_number


logger = logging.getLogger(__name__)

###################################################################################################
# H E L P E R  F U N C T I O N S
###################################################################################################

def conv2String(conv):
    """ Translate the access mode (integer) to a string. """
    if conv == Conversion.LUT :
        return "Lut"
    if conv == Conversion.POLYNOME:
        return "Polynome"
    if conv == Conversion.FUNCTION:
        return "Function"

def string2Conv(conv):
    """ Translate the access mode (string) to an integer. """
    if conv.lower() == "lut":
        return Conversion.LUT
    if conv.lower() == "polynome":
        return Conversion.POLYNOME
    if conv.lower() == "function":
        return Conversion.FUNCTION


###################################################################################################
# Y A M L  L O A D E R
###################################################################################################

class IncludeLoader(yaml.CLoader):
    """ Custom constructor for YAML CLoader to allow loading of several files in a directory.
            e.g. halDict = yaml.load(infile, Loader=IncludeLoader)

        https://stackoverflow.com/a/9577670
    """

    def __init__(self, stream):
        self._root = Path(stream.name).parent
        super().__init__(stream)

    def include_dir(self, node):
        """ Parse all yaml files in the specified node (directory name) and attach
            each one as an item in a list at the specified node.
              e.g. calibrations: !include_dir calibrations

            :param node: The yaml node containing the directory name
        """
        # assume that directory is at the same relative path as the yaml file
        path = self._root / self.construct_scalar(node)

        items = []
        for filename in sorted(path.glob('*.yaml')):
            with filename.open('r') as f:
                items.append(yaml.load(f, IncludeLoader))
        return items

# Register constructors
# Tells the yaml loader how to handle the yaml tag !include_dir (i.e. IncludeLoader.include_dir)
IncludeLoader.add_constructor('!include_dir', IncludeLoader.include_dir)


###################################################################################################
# Y A M L  D U M P E R
###################################################################################################

class include_dir(str):     # pylint: disable=invalid-name
    """ Class to identify !include_dir node to yaml dumper.

        This class is essentially a string, but is used to identify to the yaml dumper that it
        should be treated differently than a string when dumping it to a yaml file.

        In this case, it is the name of a sub-directory that contains separate yaml files for
        each list item when the dumper encounters an object of this type.
            e.g. hal['hal']['calibrations'] = include_dir('calibrations')

        NOTE: The sub-directory and files must be explicitly created when dumping the file.
    """
    pass

def repr_include_dir(dumper, dirName):
    """ Return the node as: !include_dir 'dirName' to be dumped to yaml

        This is used instead of a string, because a string node would be dumped as:
            '!include_dir dirName'
        However, the loader would interpret this as a normal string and not as a tag.

        Instead, this creates a yaml (scalar) node with only the directory name as a string:
            !include_dir 'dirName'
        The !include_dir tag can be picked up by the yaml loader as a custom tag and directed to
        the appropiate constructor (IncludeLoader.include_dir)
    """
    return dumper.represent_scalar('!include_dir', dirName)

# Register representers
# Tells the yaml dumper how to handle objects of type include_dir (i.e. run repr_include_dir)
yaml.add_representer(include_dir, repr_include_dir)


###################################################################################################
#  B A S E  C L A S S E S
###################################################################################################

class UpperCaseHalItem():
    """ HAL object that has an upper case name"""
    def __init__(self, name) :
        self._name = name.upper()

    @property
    def name(self):
        """ return upper case name"""
        return self._name

    @name.setter
    def name(self, value):
        """ convert to upper case"""
        self._name = value.upper()

class LowerCaseHalItem():
    """ HAL object that has an lower case name"""
    def __init__(self, name) :
        self._name = name.lower()

    @property
    def name(self):
        """ return lower case name"""
        return self._name

    @name.setter
    def name(self, value):
        """ convert to lower case"""
        self._name = value.lower()

class ConstantReference(UpperCaseHalItem):
    """ Object that has a 'name' attribute that references a constant """
    pass

class RegisterReference(UpperCaseHalItem):
    """ Object that has a 'name' attribute that references a register """
    pass

class DeviceTypeReference(LowerCaseHalItem):
    """ Object that has a 'name' attribute that references a device type """
    pass

class DeviceReference(LowerCaseHalItem):
    """ Object that has a 'name' attribute that references a device """
    pass

class CalibrationReference(LowerCaseHalItem):
    """ Object that has a 'name' attribute that references a calibration """
    pass



###################################################################################################
#  C O N S T A N T
###################################################################################################

class Constant(ConstantReference):
    """ One constant of a device object in the hardware abstraction layer"""

    # List of allowed Constant types
    TYPES = ['float', 'int', 'uint', 'string', 'lut']

    def __init__(self, constName, constUnit, constType, constDescription):
        """ Constructor """
        assert None not in {constName, constUnit, constType}, 'Missing parameter for Constant'
        assert constType in self.TYPES, 'Incorrect constant type'
        super().__init__(constName)

        self.unit         = constUnit
        self.type         = constType
        self.description  = constDescription

    def __eq__(self, other):
        """ Equals methode """
        if not isinstance(other, Constant):
            return False
        return self.__dict__ == other.__dict__

    def __str__(self):
        """ Give a readable string of the constant data """
        string  = "Constant    %s\n"           % self.name
        string += "    type        = %s\n"     % self.type
        string += "    unit        = %s\n"     % self.unit
        string += "    description = %s\n"     % self.description
        return string

###################################################################################################
# R E G I S T E R
###################################################################################################

class Register(RegisterReference):
    """ One register of a device object in the hardware abstraction layer """
    READ_ONLY   = 0
    READ_WRITE  = 1
    WRITE_ONLY  = 2

    def __init__(self, regName, regAddress, regAccess, regType, regWidth,
                       regUnit, regBlock, regMaxOffset, regDescription):
        """ Constructor """

        assert None not in {regName, regAddress, regAccess, regType, regWidth, regUnit, regBlock,
                            regMaxOffset},'Incomplete parameter list for Register'

        super().__init__(regName)

        self.address      = regAddress
        self.access       = regAccess
        self.type         = regType
        self.width        = regWidth
        self.unit         = regUnit
        self.block        = regBlock
        self.maxOffset    = regMaxOffset
        self.description  = regDescription

        # possible additions
        """
        self.dataBits     = 32
        self....
        """

    def __eq__(self, other):
        """ Equals methode """
        if not isinstance(other, Register):
            return False
        return self.__dict__ == other.__dict__

    def __str__(self):
        """ Give a readable sting of the register data """
        string  = "Register    %s\n"           % self.name
        string += "    address     = 0x%08x\n" % self.address
        string += "    access      = %s\n"     % self.accessToString(self.access)
        string += "    type        = %s\n"     % self.type
        string += "    width       = %d\n"     % self.width
        string += "    unit        = %s\n"     % self.unit
        string += "    block       = %s\n"     % self.block
        string += "    max offset  = %d\n"     % self.maxOffset
        string += "    description = %s\n"     % self.description
        return string

    @classmethod
    def accessToString(cls, access):
        """ Gives the access mode in readable form """
        if access == cls.READ_ONLY:
            return "read only"
        if access == cls.READ_WRITE:
            return "read-write"
        if access == cls.WRITE_ONLY:
            return "write only"
        return "unknown"

    @classmethod
    def stringToAccess(cls, access):
        """ Gives the access mode based on a string """
        if access in ["ro", "read only" ]:
            return cls.READ_ONLY
        if access in ["rw", "read-write"]:
            return cls.READ_WRITE
        if access in ["wo", "write-only"]:
            return cls.WRITE_ONLY
        return None


###################################################################################################
# D E V I C E _ T Y P E
###################################################################################################

class DeviceType(DeviceTypeReference):
    """ One device object of the hardware abstraction layer """

    def __init__(self, deviceTypeName):

        assert deviceTypeName is not None, 'Incomplete parameter list for DeviceType'

        super().__init__(deviceTypeName)
        self.registersN   = {}        # empty dictionary for registers, based on name
        self.registersA   = {}        # empty dictionary for registers, based on address
        self.constantsN   = {}        # empty dictionary for constants, based on name

    def __eq__(self, other):
        """ Equals methode """
        if not isinstance(other, DeviceType):
            return False
        return self.__dict__ == other.__dict__

    def __str__(self):
        """ Dumps the registers and constants for this devicetype """
        string  = "DeviceType\t%s\n" % self.name
        for key in self.registersN:
            string += str(self.registersN[key])
        for key in self.constantsN:
            string += str(self.constantsN[key])
        return string

    def addRegister(self, register):
        """ Adds a register to the register dictionary

            :param register: A register object
        """

        check_class(register, Register)
        assert register.name not in self.registersN, \
            'addRegister fails: name {} already present in {}'.format(register.name, self.name)

        if register.address not in self.registersA:
            self.registersN[register.name] = register
            self.registersA[register.address] = register
        else:
            logger.info('addRegister fails: address %s with name %s already present in %s',
                        register.address, register.name, self.name)

# problematic with DEMUX database
#        if register.address in self.registersA:
#            raise DataError("addRegister fails: key 0x%08x already present in %s" %
#                    (register.address, self.name))

    def addConstant(self, constant):
        """ Adds a constant to the constant dictionary

            :param constant: A constant object
        """

        check_class(constant, Constant)
        assert constant.name not in self.constantsN, \
            'addConstant fails: key {} already present in {}'.format(constant.name, self.name)

        self.constantsN[constant.name]    = constant

    def registerExists(self, registerName):
        """ Checks is registerName exists as a register object in the deviceType dict """

        if is_a_string(registerName):
            ref = RegisterReference(registerName)
            return ref.name in self.registersN.keys()

        if isinstance(registerName, int):
            return registerName in self.registersA.keys()

        return False

    def constantExists(self, constantName):
        """ Checks if constantName exists as a constant object in the deviceType dict """
        ref = ConstantReference(constantName)
        return ref.name in self.constantsN.keys()

    def getRegister(self, registerName):
        """ Returns the register object for the given register name/address """
        try:
            if is_a_string(registerName):
                ref = RegisterReference(registerName)
                return self.registersN[ref.name]

            if isinstance(registerName, int):
                return self.registersA[registerName]
        except KeyError:
            pass    # raise DataError below

        raise DataError("getRegister fails: register name/address '%s' incorrect" % registerName)

    def getRegisters(self):
        """ Returns all register objects """
        return self.registersN

    def getConstant(self, constantName):
        """ Returns the  constant object for the given constant name """
        if not self.constantExists(constantName):
            raise DataError('getConstant fails: constant name %s incorrect' % constantName)

        ref = ConstantReference(constantName)
        return self.constantsN[ref.name]

    def getConstants(self):
        """ Returns all constant objects """
        return self.constantsN

    def getConstantType(self, constantName):
        """ Returns the type of the constant """
        constant = self.getConstant(constantName)

        if constant is None:
            return float
        elif constant.type == 'float':
            return float
        elif constant.type == 'int':
            return int
        elif constant.type == 'uint':
            return int
        elif constant.type == 'string':
            return str
        elif constant.type == 'lut':
            return dict
        else:
            raise TypeError("Constant {} has unexpected type: {}".format(
                constantName, constant.type))

    def removeRegister(self, registerName):
        """ Removes the register object from its dict """
        if is_a_string(registerName):
            ref = RegisterReference(registerName)
            address = self.registersN[ref.name].address
            del self.registersN[ref.name]
            del self.registersA[address]

        if isinstance(registerName, int):
            name = self.registersA[registerName].name
            del self.registersA[registerName]
            del self.registersN[name]

    def removeConstant(self, constantName):
        """ Removes the constant object from its dict """
        ref = ConstantReference(constantName)
        del self.constantsN[ref.name]

###################################################################################################
# L I M I T
###################################################################################################

class Limits(RegisterReference):
    """ One limit object of the hardware abstraction layer """

    def __init__(self, registerName, upperValid, upperError, upperWarn,
                                     lowerWarn,  lowerError, lowerValid):
        """ Constructor """
        assert None not in {registerName, upperValid, upperError, upperWarn, lowerWarn,
                            lowerError, lowerValid}, 'Incomplete parameter list for Limits'

        super().__init__(registerName)

        self.upperValid   = upperValid
        self.upperError   = upperError
        self.upperWarn    = upperWarn
        self.lowerWarn    = lowerWarn
        self.lowerError   = lowerError
        self.lowerValid   = lowerValid

    def __eq__(self, other):
        """ Equals methode """
        if not isinstance(other, Limits):
            return False
        return self.__dict__ == other.__dict__

    def __str__(self):
        """ Give a readable sting of the limit data """
        string  = "Limit       %s\n"       % self.name
        string += "    upperError  = %8.2f\n"  % self.upperError
        string += "    upperWarn   = %8.2f\n"  % self.upperWarn
        string += "    upperValid  = %8.2f\n"  % self.upperValid
        string += "    lowerValid  = %8.2f\n"  % self.lowerValid
        string += "    lowerWarn   = %8.2f\n"  % self.lowerWarn
        string += "    lowerError  = %8.2f\n"  % self.lowerError
        return string

    def isValid(self):
        """ Return a boolean value indicating whether valid limits are set.

            If all limits are set to zero, the limits are assumed to be invalid.
        """

        return not all(item == 0.0 for item in {self.upperValid, self.upperError, self.upperWarn,
                                                self.lowerValid, self.lowerError, self.lowerWarn})

###################################################################################################
# L O O K U P   T A B L E
###################################################################################################

class LUT(RegisterReference):
    """ One look-up table for the converstion object in the hardware abstraction layer """

    def __init__(self, registerName):
        """ Constructor """
        assert registerName is not None, 'Incomplete parameter list for LUT'
        check_is_a_string(registerName)

        super().__init__(registerName)
        self.lut          = {}

    def __eq__(self, other):
        """ Equals methode """
        if not isinstance(other, LUT):
            return False
        return self.__dict__ == other.__dict__

    def addLutEntry(self, key, value):
        """ Add an entry to the lookup table """
        # add code to check the validity of the key and value
        if key in self.lut:
            logger.warning("addLutEntry: key %s already in Lut for %s. Reassigning...",
                    str(key), self.name)

        self.lut[key] = value

    def removeLutEntry(self, key):
        """ Remove entry from the lookup table """
        # add code to check the validity of the key
        if key in self.lut:
            del self.lut[key]

    def __str__(self):
        """ Give a readable sting of the look-up table """
        orderedDict = collections.OrderedDict(sorted(self.lut.keys()))
        string  = ""
        for key, value in orderedDict:
            string += str(key) + ": " + str(value)

        return string


###################################################################################################
# C O N V E R S I O N
###################################################################################################

class GrowingPolynome(list):
    """ List variant for polynomes, which automatically grows when higher-orders are filled in.
        The advantage over regular lists is that no attribute error is raised.
        The maximum nr of elements is 5 (as defined in MAXORDER).

        It also shrinks (down to linear (ax+b), as defined in MINORDER) when higher orders are 0.

        :example: Consider polynome x=[0,1]. When operation x[3]=7 is performed, the polynome
                  becomes [0,1,0,7].
                  Consider polynome x=[0,0,2,0,7]. When operation x[4]=0 is performed, the polynome
                  becomes [0,0,2].
    """
    MAXORDER = 5
    MINORDER = 2

    def _cleanArray(self):
        # Remove highest order elements which are 0
        for value in reversed(self):
            if not value and len(self) > self.MINORDER:
                # Remove highest order, since it is zero
                self.pop()
            else:
                break

    def __setitem__(self, key, value):
        if key >= self.MAXORDER:
            raise TypeError('Polynome can have {} orders at maximum'.format(self.MAXORDER))
        if key >= len(self):
            # Grow the array when higher order values are inserted
            self.extend([0] * (1 + key - len(self)))
        super().__setitem__(key, value)
        self._cleanArray()


class Conversion(RegisterReference):
    """ One conversion object of the hardware abstraction layer """

    POLYNOME = 0
    LUT      = 1
    FUNCTION = 2

    def __init__(self, registerName = None,
                       getLut       = None,
                       setLut       = None,
                       getPolynome  = None,
                       setPolynome  = None,
                       getFunction  = None,
                       setFunction  = None,
                       getConv      = None,
                       setConv      = None,
                       limits       = None):
        assert registerName is not None, 'Incomplete parameter list for Conversion'

        super().__init__(registerName)

        if getLut is None:
            self.getLut = {}
        else:
            self.getLut = getLut

        if setLut is None:
            self.setLut = {}
        else:
            self.setLut = setLut

        if getPolynome is None:
            self.getPolynome = GrowingPolynome([0, 1])
        else:
            self.getPolynome = GrowingPolynome(getPolynome)

        if setPolynome is None:
            self.setPolynome = GrowingPolynome([0, 1])
        else:
            self.setPolynome = GrowingPolynome(setPolynome)

        if getFunction is None:
            self.getFunction = "x"
        else:
            self.getFunction = getFunction

        if setFunction is None:
            self.setFunction = "x"
        else:
            self.setFunction = setFunction

        if getConv is None:
            self.getConv = Conversion.POLYNOME
        else:
            self.getConv = getConv

        if setConv is None:
            self.setConv = Conversion.POLYNOME
        else:
            self.setConv = setConv

        self.limits  = limits

    def __str__(self):
        """ Give a readable sting of the conversion data """
        string  = "Conversion  %s\n"           % self.name
        string += "    getLut      = %s\n"     % self.getLut
        string += "    setLut      = %s\n"     % self.setLut
        string += "    getPoly     = %s\n"     % self.getPolynome
        string += "    setPoly     = %s\n"     % self.setPolynome
        string += "    getFunc     = %s\n"     % self.getFunction
        string += "    setFunc     = %s\n"     % self.setFunction
        string += "    getConv     = %s\n"     % conv2String(self.getConv)
        string += "    setConv     = %s\n"     % conv2String(self.setConv)
        string += "    upperError  = %8.2f\n"  % self.limits.upperError
        string += "    upperWarn   = %8.2f\n"  % self.limits.upperWarn
        string += "    upperValid  = %8.2f\n"  % self.limits.upperValid
        string += "    lowerValid  = %8.2f\n"  % self.limits.lowerValid
        string += "    lowerWarn   = %8.2f\n"  % self.limits.lowerWarn
        string += "    lowerError  = %8.2f\n"  % self.limits.lowerError
        return string

    def __eq__(self, other):
        """ Equals methode """
        if not isinstance(other, Conversion):
            return False
        return self.__dict__ == other.__dict__

    def updateLut(self, lut, get=True):
        """ Update get look-up table """

        check_class(lut, dict)

        if get:
            self.getLut = lut
        else:
            self.setLut = lut

    def updatePolynome(self, polynome, get=True):
        """ Update the get polynome """

        check_class(polynome, list)

        poly = []
        for idx, val in enumerate(polynome):
            try:
                value = float(val)
            except ValueError:
                # Set default polynome to 0 + 1x
                if idx == 1:
                    value = 1.0
                else:
                    value = 0.0

            poly.append(value)

        if get:
            self.getPolynome = GrowingPolynome(poly)
        else:
            self.setPolynome = GrowingPolynome(poly)

    def updateFunction(self, function, get=True):
        """ Update get functions """

        check_is_a_string(function)

        if get:
            self.getFunction = function
        else:
            self.setFunction = function

    def updateConv(self, conv, get=True):
        """ Update conversion type """

        if conv not in [Conversion.POLYNOME, Conversion.LUT, Conversion.FUNCTION]:
            conv = Conversion.POLYNOME

        if get:
            self.getConv = conv
        else:
            self.setConv = conv

    def updateLimits(self, limits):
        """ Update the limits object """

        check_class(limits, Limits)

        self.limits = limits

    def getLimits(self):
        """ Returns the limits for given registerName """

        return self.limits

    def applyConversion(self, value, get=True):
        """ Applies the desired conversion, by default for a get action.

            Luts take precedence over polynomes, polynomes take precedence over functions
        """

        if get:
            lut  = self.getLut
            poly = self.getPolynome
            func = self.getFunction
            conv = self.getConv
        else:
            lut  = self.setLut
            poly = self.setPolynome
            func = self.setFunction
            conv = self.setConv

        if conv == Conversion.LUT:
            try:
                return lut[value]
            except KeyError:
                return None

        if conv == Conversion.POLYNOME:
            power = 0.0
            val   = 0.0
            for item in poly:
                val = val + item * value**power
                power += 1.0
            return val

        if conv == Conversion.FUNCTION:
            try:
                x = symbols("x")
                expr = sympify(func)
                result = expr.subs(x, value) # inject the register value into the formulae as 'x'
                return float(result.evalf(10)) # 10 is the decimal accuracy for numerical evaluation
                #                       # see: http://docs.sympy.org/latest/modules/evalf.html
            except:
                return None


###################################################################################################
# C A L I B R A T I O N
###################################################################################################

class Calibration(CalibrationReference):
    """ One calibration object of the hardware abstraction layer """

    def __init__(self, calibrationName, deviceType):
        """ Construction """
        assert None not in {calibrationName, deviceType}, \
            'Incomplete parameter list for Calibration'

        super().__init__(calibrationName)
        ref = DeviceTypeReference(deviceType)

        self.deviceType   = ref.name
        self.conversions  = {}  # dict with (name, conversion object)
        self.constants    = {}  # dict with (name, value)

    def __eq__(self, other):
        """ Equals methode """
        if not isinstance(other, Calibration):
            return False
        return self.__dict__ == other.__dict__

    def __str__(self):
        """ Dumps the registers and constants for this calibration """
        string  = "\nCalibration = %s\n" % self.name
        string += "deviceType = %s\n" % self.deviceType
        for key in self.conversions:
            string += str(self.conversions[key])
        for key in self.constants:
            string += str(self.constants[key])
        return string

    def addConstant(self, constantName, value):
        """ Add a constant to the calibration data.

        :note: Checking if the constant exists in the corresponding deviceType
               must be done by the caller.
        :note: Value is not checked, can be anything.
        """
        check_is_a_string(constantName)

        ref = ConstantReference(constantName)
        if ref.name in self.constants.keys():
            logger.warning('addConstant: %s already in calibration object. \
                    Reassigning...', ref.name)

        self.constants[ref.name] = value

    def removeConstant(self, constantName):
        """ Removes the constant object from its dict """
        ref = ConstantReference(constantName)

        try:
            del self.constants[ref.name]
        except KeyError:
            pass

    def getConstant(self, constantName):
        """ Returns the constant (value) for given constantName """
        ref = ConstantReference(constantName)

        try:
            return self.constants[ref.name]
        except KeyError:
            logger.warning(
                'Constant %s does not exist for device type %s', ref.name, self.deviceType)
        return None

    def updateConstantName(self, oldName, newName):
        """ Updates the constant in its dict """
        oldRef = ConstantReference(oldName)
        newRef = ConstantReference(newName)

        try:
            self.constants[newRef.name]      = self.constants.pop(oldRef.name)
        except KeyError:
            pass

    def addConversion(self, registerName, conversion):
        """ Add a conversion object.

            Note: checking of the register exists in the corresponding deviceType
                  must be done by the caller.
        """

        check_is_a_string(registerName)
        check_class(conversion, Conversion)

        ref = RegisterReference(registerName)
        if ref.name in self.conversions.keys():
            logger.warning('addConversion: %s already in calibration object, \
                    Reassigning...', ref.name)

        self.conversions[ref.name] = conversion

    def removeConversion(self, conversionName):
        """ Removes the conversion object from its dict """
        ref = RegisterReference(conversionName)

        try:
            del self.conversions[ref.name]
        except KeyError:
            pass

    def getConversion(self, registerName):
        """ Returns the conversion object for given registerName """

        ref = RegisterReference(registerName)

        try:
            conversion = self.conversions[ref.name]
        except KeyError:
            conversion = None
            logger.debug('getConversion fails: no conversion available for %s', ref.name)
        return conversion

    def updateConversionName(self, oldName, newName):
        """ Updates the conversion in its dict """
        oldRef = RegisterReference(oldName)
        newRef = RegisterReference(newName)

        try:
            self.conversions[newRef.name]      = self.conversions.pop(oldRef.name)
            self.conversions[newRef.name].name = newRef.name
        except KeyError:
            pass

    def getLimits(self, registerName):
        """ Returns the limits for given registerName """

        ref = RegisterReference(registerName)

        try:
            limits = self.conversions[ref.name].getLimits()
        except KeyError:
            limits = None
            logger.debug('getLimits fails: no conversion available for %s', ref.name)
        return limits


###################################################################################################
# D E V I C E
###################################################################################################

class Device(DeviceReference):
    """ One device object of the hardware abstraction layer """

    def __init__(self, deviceName, deviceDaemon, scienceDaemon, deviceApid, devicePort,
                 deviceCalibration, deviceType):
        assert None not in {deviceName, deviceDaemon, deviceApid, devicePort}, \
            'Incomplete parameter list for Device'

        super().__init__(deviceName)

        self.daemon         = deviceDaemon
        self.scienceDaemon  = scienceDaemon
        self.apid           = deviceApid
        self.port           = devicePort

        self.calibration = CalibrationReference(deviceCalibration).name
        self.deviceType = DeviceTypeReference(deviceType).name

    def __eq__(self, other):
        """ Equals methode """
        if not isinstance(other, Device):
            return False
        return self.__dict__ == other.__dict__

    def __str__(self):
        """ Give a readable string of the device data """
        string  = "Device    %s\n"                % self.name
        string += "    daemon         = %s\n"     % self.daemon
        string += "    scienceDaemon  = %s\n"     % self.scienceDaemon
        string += "    apid           = %d\n"     % self.apid
        string += "    port           = %d\n"     % self.port
        string += "    calibration    = %s\n"     % self.calibration
        string += "    deviceType     = %s\n"     % self.deviceType
        return string


###################################################################################################
# H A R D W A R E   A B S T R A C T I O N   L A Y E R
###################################################################################################

class Hal():
    """ Hardware Abstraction Layer """

    YAML_DIRS = ['calibrations', 'deviceTypes']     # split these sections into separate directories

    def __init__(self):
        self._calibrations  = {}    # key is a calibrationName
        self._deviceTypes   = {}    # key is a deviceTypeName
        self._devices       = {}    # key is a deviceName
        self._versionNumber = ''    # String with major.minor.patch
        self._yamlFile      = None  # yaml path/filename

    def __eq__(self, other):
        """ Equals methode """
        if not isinstance(other, Hal):
            return False
        return self.__dict__ == other.__dict__

    def setYamlFilename(self, yamlFile):
        """ Stores the path/filename of the yaml file in use. This is needed such that
            the HAL with the HalManager can easily save the data upon edits.
        """
        self._yamlFile = yamlFile

    def setHalFromDict(self, data) :
        """ Set the dict containing HAL settings directly """
        self.processHalSettings(data)

    def setHalFromFile(self, yamlFile):
        """ Loads a yaml file with the hardware abstraction layer """

        logger.debug("Loading hardware abstraction layer file %s", yamlFile)
        check_is_a_string(yamlFile)

        with open(yamlFile, "r") as infile:
            halData = yaml.load(infile, Loader=IncludeLoader)
        self.processHalSettings(halData)

    def setHalFromString(self, yamlData):
        """ Sets yaml data """
        assert yamlData is not None, 'No YAML data given'
        halData = yaml.load(yamlData, Loader=yaml.CLoader)
        self.processHalSettings(halData)

    def storeHal(self, filename=None, hal=None):
        """ Saves the current HAL as YAML.

            The HAL will be saved as one main file at the specified location and possibly several
            subdirectories (one for each item in YAML_DIRS) containing a file for each item in that
            category.

            e.g. YAML_DIRS = ['calibrations', 'deviceTypes'] will result in one main HAL file, with
            a 'calibration' subdirectory containing a file for each calibration and a 'devicetypes'
            subdirectory containing a file for each device type (containing register definitions).

            :param filename: Where to store
        """

        # Get HAL file location
        filename = self._yamlFile if filename is None else filename
        if filename is None:
            logger.error("HAL path for yaml file is undefined.")
            return False
        halPath = Path(filename)

        if hal is None:
            hal = self.createYamlData()

        try:
            # Split following sections into separate files in directories with same name
            for tag in self.YAML_DIRS:
                # Create hal subdirectory if does not exist
                itemPath = halPath.parent / tag
                itemPath.mkdir(exist_ok=True)
                logger.debug("Adding items from '%s' to dir: %s", tag, itemPath)

                # Write each item into separate file in subdirectory
                for item in hal['hal'][tag]:
                    itemFilename = Path(itemPath, "{}.yaml".format(item['name']))
                    logger.debug("  Writing to: %s", itemFilename)

                    with itemFilename.open('w') as outfile:
                        yaml.dump(item, outfile, Dumper=yaml.CDumper)

                # Replace dictionary item with directive that points to subdirectory
                hal['hal'][tag] = include_dir(itemPath.name)

            # Write the main hal file
            with halPath.open("w") as outfile:
            # do not use "default_flow_style" in the dump command below: this changes the
            # lists in the dict, after which these lists cannot be read anymore
                logger.info("Storing yaml data at: %s", halPath)
                yaml.dump(hal, outfile)
        except (OSError, IOError) as ex:
            logger.error('File %s could not be opened: %s', halPath, ex)
            return False
        else:
            return True

    def processHalSettings(self, halDict):
        """ Processes the information in the dict.

        :param halDict: The dictionary containing the Hal config.
        """

        self._calibrations = {}    # key is a calibrationName
        self._deviceTypes  = {}    # key is a deviceTypeName
        self._devices      = {}    # key is a deviceName

        assert halDict is not None, 'No YAML dict available'

        # Read the HAL database version number from the YAML dict
        try:
            self._versionNumber = halDict['hal']['versionNumber'] # major.minor.patch
        except KeyError:
            # Insert a version number if it doesn't exist
            #   This is done to make it easier to work with stand-alone HAL yaml files.
            logger.exception('HAL file does not contain a version number, inserted')
            self._versionNumber = '0.0.0'

        # Add all deviceTypes with their registers
        for item in halDict['hal']['deviceTypes']:
            deviceTypeName = item['name']
            logger.debug("Processing deviceTypes[%s]", deviceTypeName)
            deviceType = DeviceType(deviceTypeName)
            self.addDeviceType(deviceType)

            try:
                for reg in item['registers']:
                    logger.debug("    Adding register %s", reg['name'])
                    register = Register(     reg['name'],
                                        int( reg['address']),
                                             reg['access'],
                                             reg['type'],
                                        int( reg['width']),
                                             reg['unit'],
                                             reg['block'],
                                        int( reg['maxOffset']),
                                             reg['description'])
                    deviceType.addRegister(register)
            except KeyError as e:
                logger.warning("DeviceTypes-Register: key %s not found for [%s]",
                        e, deviceTypeName)

            try:
                for const in item['constants']:
                    logger.debug("    Adding constant %s", const['name'])
                    constant = Constant(const['name'], const['unit'], const['type'],
                                        const['description'])
                    deviceType.addConstant(constant)
            except KeyError as e:
                logger.warning("DeviceType-Constant: key %s not found for [%s]",
                        e, deviceTypeName)

        #
        # Add all calibrations with conversions, limits, etc
        try:
            calibrations = halDict['hal']['calibrations']
        except KeyError:
            calibrations = None

        try:
            devices = halDict['hal']['devices']
        except KeyError:
            devices = None

        if calibrations is not None:
            for item in calibrations:
                calibrationName = item['name']
                deviceTypeName  = item['deviceType']
                logger.debug("Processing calibrations[%s][%s]",
                        deviceTypeName, calibrationName)

                deviceType = self._deviceTypes[deviceTypeName]
                cal        = Calibration(calibrationName, deviceTypeName)
                self._calibrations[calibrationName] = cal

                try:
                    for constantName, constantValue in item['constants'].items():
                        logger.debug("Processing calibrations[%s][%s]",
                                deviceTypeName, calibrationName)

                        # find constant in deviceType
                        # if present, cast value to correct type
                        if deviceType is not None:
                            dtype = deviceType.getConstantType(constantName)
                            constantValue = dtype(constantValue)

                        cal.addConstant(constantName, constantValue)

                except KeyError as e:
                    logger.warning("Calibration-Constant: key %s not found for [%s]",
                            e, calibrationName)

                try:
                    for conv in item['conversions']:
                        conversionName = conv['name']
                        logger.debug("Processing calibrations[%s][%s][%s]",
                                deviceTypeName, calibrationName, conversionName)
                        con = Conversion(conversionName)
                        if not self.registerExists(deviceTypeName, conversionName):
                            logger.info('(conversion) deviceType %s has no register with name %s',
                                        deviceTypeName, conversionName)
                            continue
                        #assert self.registerExists(deviceTypeName, conversionName), \
                        #    'deviceType {} has no register with name {}'.\
                        #        format(deviceTypeName, conversionName)

                        # use if statements here: not all conversions are always present
                        if 'getlut'      in conv.keys():
                            con.updateLut(conv['getlut'])
                        if 'setlut'      in conv.keys():
                            con.updateLut(conv['setlut'], False)
                        if 'getpolynome' in conv.keys():
                            con.updatePolynome(conv['getpolynome'])
                        if 'setpolynome' in conv.keys():
                            con.updatePolynome(conv['setpolynome'], False)
                        if 'getfunction' in conv.keys():
                            con.updateFunction(conv['getfunction'])
                        if 'setfunction' in conv.keys():
                            con.updateFunction(conv['setfunction'], False)
                        if 'getconversion' in conv.keys():
                            con.updateConv(string2Conv(conv['getconversion']))
                        if 'setconversion' in conv.keys():
                            con.updateConv(string2Conv(conv['setconversion']), False)
                        if 'limits' in conv.keys():
                            lim   = conv['limits']
                            limit = Limits(conv['name'],
                                   lim['upperValid'],   lim['upperError'], lim['upperWarning'],
                                   lim['lowerWarning'], lim['lowerError'], lim['lowerValid'])
                            con.updateLimits(limit)
                        cal.addConversion(conversionName, con)
                except KeyError as e:
                    logger.warning("Calibration-Conversion: key %s not found for %s",
                            e, calibrationName)

        if devices is not None:
            #
            # Add all actual devices
            for item in halDict['hal']['devices']:
                try:
                    logger.debug("Processing device %s", item['name'])
                    device = Device(    item['name'],
                                        item['daemon'],
                                        item['scienceDaemon'] if 'scienceDaemon' in item else None,
                                    int(item['apid']),
                                    int(item['port']),
                                        item['calibration'],
                                        item['deviceType'])
                    self.addDevice( device)
                except KeyError as e:
                    logger.warning("Device: key %s not found for devices", e)

    def createYamlData(self):
        """ Dumps the HAL data to a yaml string """
        logger.debug("creating dict")
        deviceList      = []
        deviceTypeList  = []
        calibrationList = []

        # make a new device list
        for _, device in sorted(self._devices.items(), key=itemgetter(0)):
            deviceDict = {'name'            : device.name,
                          'daemon'          : device.daemon,
                          'apid'            : device.apid,
                          'port'            : device.port,
                          'calibration'     : device.calibration,
                          'deviceType'      : device.deviceType}
            if device.scienceDaemon:
                # scienceDaemon is omitted from the Yaml file in case it is not used
                deviceDict['scienceDaemon'] = device.scienceDaemon
            deviceList.append(deviceDict)

        # make a new device type list
        for key in self._deviceTypes:
            registerList = []
            constantList = []

            # put registers in a new dict
            regDict = self._deviceTypes[key].registersN
            for reg in sorted(regDict):  # keys are reg names, so sort them
                register = {'name'          : regDict[reg].name,
                            'address'       : regDict[reg].address,
                            'access'        : regDict[reg].access,
                            'type'          : regDict[reg].type,
                            'width'         : regDict[reg].width,
                            'unit'          : regDict[reg].unit,
                            'block'         : regDict[reg].block,
                            'maxOffset'     : regDict[reg].maxOffset,
                            'description'   : regDict[reg].description}
                registerList.append(register)

            # put constants in a new dict
            conDict = self._deviceTypes[key].constantsN
            for con in sorted(conDict): # keys are constant names, so sort them
                constant = {'name'          : conDict[con].name,
                            'unit'          : conDict[con].unit,
                            'type'          : conDict[con].type,
                            'description'   : conDict[con].description}
                constantList.append(constant)


            deviceType   = {'name'          : self._deviceTypes[key].name,
                            'registers'     : registerList,
                            'constants'     : constantList}

            # put deviceType on a list
            deviceTypeList.append(deviceType)

        # make a new calibration list
        for key in sorted(self._calibrations):  # keys are calibration names, so sort them

            # put constants (with actual value!) in a new dict
            constantList = self._calibrations[key].constants

            # put conversions in a new list of dicts
            conversiList = []
            conDict = self._calibrations[key].conversions
            for con in sorted(conDict): # sort names
                conversion = {'name'          : con,
                              'getlut'        : conDict[con].getLut,
                              'setlut'        : conDict[con].setLut,
                              'getpolynome'   : conDict[con].getPolynome,
                              'setpolynome'   : conDict[con].setPolynome,
                              'getfunction'   : conDict[con].getFunction,
                              'setfunction'   : conDict[con].setFunction,
                              'getconversion' : conv2String(conDict[con].getConv),
                              'setconversion' : conv2String(conDict[con].setConv)}

                # Only save the limits if they are specified
                if conDict[con].limits is not None:
                    limits     = {'upperValid'  : conDict[con].limits.upperValid,
                                  'upperError'  : conDict[con].limits.upperError,
                                  'upperWarning': conDict[con].limits.upperWarn,
                                  'lowerWarning': conDict[con].limits.lowerWarn,
                                  'lowerError'  : conDict[con].limits.lowerError,
                                  'lowerValid'  : conDict[con].limits.lowerValid }
                    conversion['limits'] = limits

                conversiList.append(conversion)

            # put all calibrations on a dict of its own
            calibration = {'name'           : self._calibrations[key].name,
                           'deviceType'     : self._calibrations[key].deviceType,
                           'constants'      : constantList,
                           'conversions'    : (conversiList) }

            # make a list with all calibrations
            calibrationList.append(calibration)

        # create entire yaml model
        hal          = {'hal'               : { 'deviceTypes'  : deviceTypeList,
                                                'calibrations' : calibrationList,
                                                'devices'      : deviceList,
                                                'versionNumber': self._versionNumber } }
        return hal

    def addDeviceType(self, deviceType):
        """ Adds a deviceType object to the hardware abstraction layer

            :param deviceType: A deviceType object
        """

        check_class(deviceType, DeviceType)

        deviceTypeName = deviceType.name
        assert deviceTypeName not in self._deviceTypes, \
            'deviceType {} already exists'.format(deviceTypeName)
        self._deviceTypes[deviceTypeName] = deviceType
        logger.debug('Added deviceType: %s', deviceTypeName)

    def addRegisterInfo(self, deviceName, register):
        """ Adds a register to its dict in the deviceType object """
        self.getDeviceType(deviceName).addRegister(register)

    def addRegisterConversion(self, deviceName, deviceTypeName, calibrationName,
                              registerName, conversion):
        """ Add a register conversion to the calibration object
        """
        calibration = self.getCalibration(deviceName, deviceTypeName, calibrationName)
        calibration.addConversion(registerName, conversion)

    def addRegisterLimits(self, deviceName, deviceTypeName, calibrationName,
                          registerName, limits):
        """ Add a register limit object to the conversion object
        """
        conversion = self.getRegisterConversion(deviceName, deviceTypeName, calibrationName,
                                                registerName)
        conversion.updateLimits(limits)

    def addConstantInfo(self, deviceName, constant):
        """ Adds a constant to its dict in the deviceType object """
        self.getDeviceType(deviceName).addConstant(constant)

    def addCalibration(self, deviceTypeName, calibration):
        """ Adds a calibraton object to the hardware abstraction layer

            :param deviceTypeName: The device type for which the calibration is (case insensitive)
            :param calibration: A calibration object
        """

        check_is_a_string(deviceTypeName)
        check_class(calibration, Calibration)

        deviceTypeRef = DeviceTypeReference(deviceTypeName)

        assert deviceTypeRef.name in self._deviceTypes, 'deviceType does not exist'
        assert calibration.name not in self._calibrations, 'calibration already exists'

        self._calibrations[calibration.name] = calibration
        logger.debug('Added Calibration: %s to deviceType: %s', calibration.name, deviceTypeName)

    def addDevice(self, device):
        """ Adds a device to the hardware abstraction layer

            :param device: A device object
        """
        check_class(device, Device)
        deviceName = device.name
        assert deviceName not in self._devices, 'device {} already exists'.format(deviceName)
        self._devices[deviceName] = device
        logger.debug('Added Device: %s', deviceName)

    def getDeviceType(self, deviceName=None, apid=None):
        """ Returns the deviceType object of the deviceName or APID

            :param deviceName: Search for this device name or device type name
            :param apid: Search for this APID
            :note: If both arguments are supplied, deviceName is used!
        """

        if deviceName is not None:
            ref1 = DeviceReference(deviceName)
            if ref1.name in self._devices:
                deviceTypeName = self._devices[ref1.name].deviceType
                return self._deviceTypes[deviceTypeName]

            ref2 = DeviceTypeReference(deviceName)
            if ref2.name in self._deviceTypes:
                return self._deviceTypes[ref2.name]

            if apid is None:
                raise DataError('Device name %s not exist in the HAL' % deviceName)
            else:
                logger.info('device name %s not in device or deviceType dict, trying with apid %d',
                        deviceName, apid)

        if apid is not None:
            for dev in self._devices:
                device = self._devices[dev]
                if device.apid == apid or device.apid+1 == apid:
                    return self._deviceTypes[device.deviceType]

            raise DataError('device name with apid 0x%08x (%d) not in device or deviceType dict'
                    % (apid, apid))

        raise DataError('no deviceName or apid given')

    def getDeviceTypeNames(self):
        """ Get the deviceTypes in the current HAL

            :returns: A list of deviceType names (strings)
        """
        return list(self._deviceTypes.keys())

    def getRegistersInfo(self, deviceName):
        """ Returns the dict with all registers for the given device """
        return self.getDeviceType(deviceName).getRegisters()

    def getRegisterInfo(self, deviceName, registerName, registerAddress=None):
        """ Returns the requested register

            :param deviceName: An actual deviceName.
            :param registerName: The name or address of a register.
            :param registerAddress: The address of a register.
            :note: If both registerName and registerAddress are supplied, then
                   registerName will first be used, followed by registerAddress.
                   This is useful if you want to search by both (as one may have changed).
            :raises DataError: the device or register could not be found.
            :returns: The object representing the specified register.
        """
        # Try first with registerName (which could be a name or address)
        try:
            return self.getDeviceType(deviceName).getRegister(registerName)
        except DataError as error:
            # Raise error if address is not defined, otherwise try again below
            if registerAddress is None:
                raise error

        # If not found, then use explicity defined address (if specified)
        return self.getDeviceType(deviceName).getRegister(registerAddress)

    def getRegisterConversion(self, deviceName, deviceTypeName, calibrationName, registerName):
        """ Returns the requested conversion object for given device and register

            device is an acutal deviceName
            registerName is a string (not an address!)
        """
        calibration = self.getCalibration(deviceName, deviceTypeName, calibrationName)
        return calibration.getConversion(registerName)

    def getRegisterLimits(self, deviceName, deviceTypeName, calibrationName, registerName):
        """ Returns the requested limits

            device is an acutal deviceName
            registerName is a string (not an address!)
        """
        calibration = self.getCalibration(deviceName, deviceTypeName, calibrationName)
        return calibration.getLimits(registerName)

    def getConstantsInfo(self, deviceName):
        """ Returns the dict with all constant for the given device """
        return self.getDeviceType(deviceName).getConstants()

    def getConstantInfo(self, deviceName, constantName):
        """ Returns the requested constant

            device is an actual deviceName
            constantName is a constant name
        """
        return self.getDeviceType(deviceName).getConstant(constantName)

    def getConstantValue(self,
                         constantName,
                         deviceName     = None,
                         deviceTypeName = None, calibrationName = None):
        """ Returns the value of the constant for the given calibration scheme """
        assert deviceName is not None or \
            (deviceTypeName is not None and calibrationName is not None),\
            "Retrieve constant with either a deviceName or by device type+calibration"

        calibration = self.getCalibration(deviceName, deviceTypeName, calibrationName)
        return calibration.getConstant(constantName)

    def getCalibrations(self, deviceTypeName):
        """ Returns a list with calibration objects for the given deviceType

            :param deviceTypeName: Name of the device type
        """
        ref = DeviceTypeReference(deviceTypeName)
        calibrations = [cal for cal in self._calibrations.values() if cal.deviceType == ref.name]
        return calibrations

    def getCalibration(self, deviceName = None, deviceTypeName = None, calibrationName = None):
        """ Get a calibration for a given device or calibration. If deviceType is specified,
            then, will also make sure the calibration belongs to that deviceType.

            :param deviceName:      Name of the device
            :param deviceTypeName:  Name of the deviceType
            :param calibrationName: Name of the calibration
            :returns: A calibration object
        """
        # If device defined, get calibration for device
        if deviceName is not None:
            devRef = DeviceReference(deviceName)
            assert devRef.name in self._devices.keys(), \
                'Invalid device name {}'.format(devRef.name)
            calibrationName = self._devices[devRef.name].calibration
            return self._calibrations[calibrationName]

        # Else get calibration by name
        assert calibrationName is not None, "getCalibration fails: missing 'calibrationName'"
        calRef = CalibrationReference(calibrationName)
        assert calRef.name in self._calibrations.keys(), \
            "Invalid calibration name {}".format(calRef.name)
        calibration = self._calibrations[calRef.name]

        # If deviceType is specified, check that calibration belongs to that device
        if deviceTypeName is not None:
            devTypeRef = DeviceTypeReference(deviceTypeName)
            assert devTypeRef.name in self._deviceTypes.keys(), \
                "Invalid deviceType name {}".format(devTypeRef.name)
            assert calibration.deviceType == devTypeRef.name, \
                "Invalid calibration '{}' for deviceType '{}'".format(calRef.name, devTypeRef.name)

        return calibration

    def getCalibrationNames(self, deviceTypeName = None):
        """ Get the calibrations in the current HAL (no deviceTypeName specified)
            or in the deviceType (deviceTypeName specified)

            :param deviceTypeName: Name of the deviceType.

            :returns: A list of calibration names.
        """
        if deviceTypeName is not None:
            ref = DeviceTypeReference(deviceTypeName)
            calibrationNames = [calName for calName, cal in self._calibrations.items()
                                if cal.deviceType == ref.name]
            return calibrationNames

        return list(self._calibrations.keys())

    def getDevices(self, deviceTypeName, calibrationName = None):
        """ Returns a list with device objects for the given deviceType;
            if calibrationName is given, then also the calibration is checked

            :param deviceTypeName: Name of the device type
            :param calibrationName: Name of the calibration
        """
        devices = []
        ref1 = DeviceTypeReference(deviceTypeName)

        if calibrationName is not None:
            ref2 = CalibrationReference(calibrationName)
            for dev in self._devices:
                if self._devices[dev].deviceType == ref1.name and \
                                self._devices[dev].calibration == ref2.name:
                    devices.append(self._devices[dev])
        else:
            for dev in self._devices:
                if self._devices[dev].deviceType == ref1.name:
                    devices.append(self._devices[dev])

        return devices

    def getDevice(self, deviceName=None, apid=None):
        """ Get the device based on the device name or APID

            :param deviceName: Device name to look up (case-insensitive)
            :param apid: APID to look up
            :note: If both arguments are supplied, APID is used
            :note: If device is not present in the HAL, an ItemNotFoundError exception is raised.
        """
        if apid is None:
            ref = DeviceReference(deviceName)
            try:
                return self._devices[ref.name]
            except KeyError:
                raise ItemNotFoundError("Device: %s not found in HAL" % deviceName) from None

        for dev in self._devices:
            if dev.apid == apid or dev.apid+1 == apid:
                return dev
        raise ItemNotFoundError("Device with apid: %s not found in HAL" % apid)

    def getDeviceNames(self):
        """ Get the devices in the current HAL

            :returns: A list of device names (strings)
        """
        return list(self._devices.keys())

    def getVersionNumber(self):
        """ Get the HAL version number.

            :returns: String 'major.minor.patch'
        """
        return self._versionNumber

    def updateVersionNumber(self):
        """ Increases the patch number of the HAL version (by 1).

            :note: Major and minor version number must be updated by hand in the yaml file.
        """

        (major, minor, patch) = self._versionNumber.split('.')
        patch = int(patch)
        patch += 1

        self._versionNumber = '{}.{}.{}'.format(major, minor, patch)

    def registerExists(self, deviceType, registerName):
        """ Checks if registerName exists as a register object in the deviceType dict

            :param deviceType: Either the name of the device type[string] or a device type object
            :returns: True or False
        """

        if isinstance(deviceType, DeviceType):
            return deviceType.registerExists(registerName)

        if is_a_string(deviceType):
            ref = DeviceTypeReference(deviceType)
            return self._deviceTypes[ref.name].registerExists(registerName)

        raise DataError('registerExists fails: device type is of incorrect type')

    def constantExists(self, deviceType, constantName):
        """ Checks if constantName exists as a constant object in the deviceType dict

            :param deviceType: Either a device type name[string] or a DeviceType object
            :param constantName: The name[string] of the constant.
        """
        if isinstance(deviceType, DeviceType):
            return deviceType.constantExists(constantName)

        if is_a_string(deviceType):
            ref = DeviceTypeReference(deviceType)
            return self._deviceTypes[ref.name].constantExists(constantName)

        raise DataError('constantExists fails: device type is of incorrect type')

    def removeRegisterInfo(self, deviceName, registerName):
        """ Removes registerName for deviceName from its dict in the deviceType object """
        self.getDeviceType(deviceName).removeRegister(registerName)

    def removeRegisterConversion(self, deviceName, deviceTypeName, calibrationName, registerName):
        """ Removes a conversion from the hardware abstraction layer

            :param deviceName: Name of a device object
            :param deviceTypeName: Name of a deviceType object
            :param calibrationName: Name of a calibration object
            :param registerName: Name of a conversion object
        """
        calibration = self.getCalibration(deviceName, deviceTypeName, calibrationName)
        calibration.removeConversion(registerName)

    def removeConstantInfo(self, deviceName, constantName):
        """ Removes constantName for deviceName from its dict in the deviceType object """
        self.getDeviceType(deviceName).removeConstant(constantName)

    def removeDevice(self, deviceName):
        """ Removes a device from the hardware abstraction layer

            :param deviceName: Name of a device object
        """
        check_is_a_string(deviceName)
        ref = DeviceReference(deviceName)
        assert ref.name in self._devices, \
            'remove fails: device {} does not exist'.format(ref.name)
        del self._devices[ref.name]
        logger.debug('Removed Device: %s', ref.name)

    def removeDeviceType(self, deviceTypeName):
        """ Removes a device from the hardware abstraction layer

            :param deviceTypeName: Name of a deviceType object
        """
        check_is_a_string(deviceTypeName)
        ref = DeviceTypeReference(deviceTypeName)
        assert ref.name in self._deviceTypes, \
            'remove fails: deviceType {} does not exist'.format(ref.name)
        del self._deviceTypes[ref.name]
        logger.debug('Removed deviceType: %s', ref.name)

    def removeCalibration(self, deviceTypeName, calibrationName):
        """ Removes a calibration from the hardware abstraction layer

            :param (deviceTypeName, calibrationName): Name of a calibration object
        """
        check_is_a_string(deviceTypeName)
        check_is_a_string(calibrationName)
        devTypeRef = DeviceTypeReference(deviceTypeName)
        calRef = CalibrationReference(calibrationName)
        assert calRef.name in self._calibrations, \
            "remove fails: calibration {} does not exist".format(calRef.name)
        assert self._calibrations[calRef.name].deviceType == devTypeRef.name, \
            'remove fails: calibration ({}, {}) does not exist'.format(devTypeRef.name, calRef.name)
        del self._calibrations[calRef.name]
        logger.debug('Removed Calibration: %s', calRef.name)

    def applyConversion(self, deviceName, registerName, value, get = True):
        """ Applies a conversion for the given register of the given device
            :param deviceName: is a string
            :param registerName: is a string
            :param value: is not None
            :param get: True for 'Get'
        """
        check_is_a_string(deviceName)
        check_is_a_string(registerName)
        check_is_a_number(value)

        conversion  = self.getCalibration(deviceName=deviceName).getConversion(registerName)
        assert conversion is not None,\
            "Register {} has no conversion polynomials".format(registerName)
        value = conversion.applyConversion(value, get)
        return value
