# pylint: disable=too-many-lines
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-branches
# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
# pylint: disable=invalid-name

""" GSSw Hardware Abstraction Layer Server

"""

###################################################################################################
# P Y T H O N   M O D U L E S
###################################################################################################

import sys
import os.path
import logging
import pprint

from copy import deepcopy
from ast import literal_eval

from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal, pyqtSlot, Qt, QSettings, QSize, QPoint
from PyQt5.QtGui import QKeySequence
from PyQt5.QtWidgets import QWidget, QMenu, QTableWidgetItem, QMessageBox, QLabel, \
    QStyledItemDelegate, QComboBox, QVBoxLayout, QInputDialog, QPlainTextEdit, QHeaderView

from gssw.hal.frontend.uploaddialog import UploadDialog
from gssw.hal.halclient import HalClient
from gssw.hal.hal import Hal, Device, DeviceType, Calibration, Register, Constant, \
    Conversion, Limits

from gssw.lib.config import Config
from gssw.lib.socketsqt import NotifyingSubSocket
from gssw.lib.typing import check_class
from gssw.lib.widgets.misc import gsswUiError, gsswUiWarning, gsswUiMessage, gsswUiYesNoQuestion
from gssw.lib.version import VERSION_STRING
from gssw.lib.errors import DataError

logger = logging.getLogger(__name__)


###################################################################################################
# H E L P E R  F U N C T I O N S
###################################################################################################

def access2String(access):
    """ Translate the access mode (integer) to a string. """
    if access == Register.READ_ONLY:
        return "ro"
    if access == Register.READ_WRITE:
        return "rw"
    if access == Register.WRITE_ONLY:
        return "wo"
    raise AssertionError('Illegal access type: {}'.format(access))


def string2Access(access):
    """ Translate the access mode (string) to an integer. """
    if access == "ro":
        return Register.READ_ONLY
    if access == "rw":
        return Register.READ_WRITE
    if access == "wo":
        return Register.WRITE_ONLY
    raise AssertionError('Illegal access type: {}'.format(access))


###################################################################################################
# H A L  M A N A G E R
###################################################################################################

class FrontEnd(QWidget):
    """ User Interface for the Hardware Abstraction Layer. """

    STACKSIZE = 10

    def __init__(self, config, app=None):
        """ Constructor of this class. """
        super(FrontEnd, self).__init__(flags=Qt.Widget)

        check_class(config, Config)

        # get linux signal handler
        self.signalHandler = app.getSignalHandler() if app is not None else None

        # load the hardware abstraction layer
        self.halOriginal = Hal()
        self.halClient = HalClient(self.halOriginal, config)
        self.halClient.requestHal()
        self.hal = deepcopy(self.halOriginal)

        # create a notifying SUB socket for listening to HAL updates from the server
        self.subSock = NotifyingSubSocket()
        self.subSock.createZmqSocket()
        hostname, _, pubPort = self.halClient.getServerAddress() 
        self.subSock.subscribe(topic="")
        self.subSock.connectSocket('tcp://%s:%d' % (hostname, pubPort))
        self.subSock.sigReadyReceive.connect(self.slotReloadHal)

        # get the directory of this module
        modulePath = os.path.dirname(os.path.abspath(__file__))

        # initialize the GUIs
        guiPath = os.path.join(modulePath, 'gui')
        self.gui = uic.loadUi(os.path.join(guiPath, 'HalManagerMainWindow.ui'))

        # Allow editing of constant type (in table) using combobox
        self.constantTypeEditorDelegate = ComboBoxItemDelegate(
            self.gui.constantTable, Constant.TYPES)

        # Allow editing of constant value in popup, for LUTs
        self.constantValueEditorDelegate = PopupItemDelegate(self.gui.constantTable)

        # Apply the delegates to the table columns
        self.gui.constantTable.setItemDelegateForColumn(1, self.constantValueEditorDelegate)
        self.gui.constantTable.setItemDelegateForColumn(3, self.constantTypeEditorDelegate)
        self.gui.registerTable.setItemDelegateForColumn(8, DescriptionEditDelegate())

        self.newRegisterUi = uic.loadUi(os.path.join(guiPath, 'HalManagerNewRegister.ui'))
        self.newConstantUi = uic.loadUi(os.path.join(guiPath, 'HalManagerNewConstant.ui'))
        self.newConstantUi.type.addItems(Constant.TYPES) # Fill combobox in popup dialog
        self.conversionsUi = uic.loadUi(os.path.join(guiPath, 'HalManagerConversions.ui'))
        self.limitsUi = uic.loadUi(os.path.join(guiPath, 'HalManagerLimits.ui'))
        self.newDeviceUi = uic.loadUi(os.path.join(guiPath, 'HalManagerNewDevice.ui'))
        self.newDeviceTypeUi = uic.loadUi(os.path.join(guiPath, 'HalManagerNewDeviceType.ui'))
        self.newCalibrationUi  = uic.loadUi(os.path.join(guiPath, 'HalManagerNewCalibration.ui'))
        self.editDeviceUi = uic.loadUi(os.path.join(guiPath, 'HalManagerEditDevice.ui'))
        self.editDeviceTypeUi = uic.loadUi(os.path.join(guiPath, 'HalManagerEditDeviceType.ui'))
        self.editCalibrationUi = uic.loadUi(os.path.join(guiPath, 'HalManagerEditCalibration.ui'))
        self.saveUploadToServerDialog = UploadDialog(guiPath, self.halOriginal, config)

        # define shortcuts for (existing) main GUI actions
        self.gui.actionSave.setShortcut(QKeySequence.Save)
        self.gui.actionClose.setShortcut(QKeySequence.Close)
        self.gui.actionUndo.setShortcut(QKeySequence.Undo)

        # Vertical fit (for line breaks in register descriptions)
        verticalHeader = self.gui.registerTable.verticalHeader()
        verticalHeader.setDefaultSectionSize(35)
        verticalHeader.sectionResizeMode(QHeaderView.ResizeToContents)

        # create and populate the context menu of the main GUI register Tab
        self.regMenu = QMenu()
        self.regMenu.addAction(self.gui.actionNewRegister)
        self.regMenu.addAction(self.gui.actionDeleteRegisters)
        self.regMenu.addAction(self.gui.actionEditConversions)
        self.regMenu.addAction(self.gui.actionEditLimits)

        # create and populate the context menu of the main GUI constant Tab
        self.constMenu = QMenu()
        self.constMenu.addAction(self.gui.actionNewConstant)
        self.constMenu.addAction(self.gui.actionRenameConstant)
        self.constMenu.addAction(self.gui.actionDeleteConstants)

        # connect the handlers
        self.connectHandlers()  # connect the handlers for the main GUI
        self.connectNewRegisterHandlers()  # connect the handlers for the new register GUI
        self.connectNewConstantHandlers()  # connect the handlers for the constant GUI
        self.connectConversionHandlers()  # connect the handlers for the conversions GUI
        self.connectLimitHandlers()  # connect the handlers for the limits GUI
        self.connectNewDeviceHandlers()  # connect the handlers for the new device GUI
        self.connectNewDeviceTypeHandlers()  # connect the handlers for the new deviceType GUI
        self.connectNewCalibrationHandlers()  # connect the handlers for the new calibration GUI
        self.connectEditDeviceHandlers()  # connect the handlers for the edit device GUI
        self.connectEditDeviceTypeHandlers()  # connect the handlers for the edit deviceType GUI
        self.connectEditCalibrationHandlers()  # connect the handlers for the edit calibration GUI

        self.windowSize = QSize(1000, 1000)
        self.windowPos = QPoint(0, 0)
        self.tabIndex = 0
        self.sortColumnRegister = 1
        self.sortColumnConstant = 1
        self.sortOrderRegister = Qt.AscendingOrder
        self.sortOrderConstant = Qt.AscendingOrder
        self.regColumnWidth = [200, 95, 50, 50, 60, 50, 50, 55, 285]
        self.conColumnWidth = [200, 100, 50, 50, 400]
        self.deviceName = ""
        self.deviceTypeName = ""
        self.calibrationName = ""
        self.halStack = []
        self.conversion = None

        # get settings
        self.readSettings()
        logger.info("Started HAL Manager, version %s", VERSION_STRING)

    def connectHandlers(self):
        """ Connect signals to slots. """

        # combo box changes
        self.gui.device.currentIndexChanged.connect(self.slotDeviceChanged)
        self.gui.deviceType.currentIndexChanged.connect(self.slotDeviceTypeChanged)
        self.gui.calibration.currentIndexChanged.connect(self.slotCalibrationChanged)

        # selection changes
        self.gui.registerTable.itemSelectionChanged.connect(self.slotRegisterSelectionChanged)
        self.gui.constantTable.itemSelectionChanged.connect(self.slotConstantSelectionChanged)

        # right-click menu
        self.gui.registerTable.customContextMenuRequested.connect(self.slotRegisterMenu)
        self.gui.constantTable.customContextMenuRequested.connect(self.slotConstantMenu)

        # sorting
        rHeader = self.gui.registerTable.horizontalHeader()
        cHeader = self.gui.constantTable.horizontalHeader()
        rHeader.sortIndicatorChanged.connect(self.slotSortRegisterTable)
        cHeader.sortIndicatorChanged.connect(self.slotSortConstantTable)

        # main GUI actions
        self.gui.actionNewRegister.triggered.connect(self.slotNewRegister)
        self.gui.actionNewConstant.triggered.connect(self.slotNewConstant)
        self.gui.actionRenameConstant.triggered.connect(self.slotRenameConstant)
        self.gui.actionDeleteRegisters.triggered.connect(self.slotDeleteRegisters)
        self.gui.actionDeleteConstants.triggered.connect(self.slotDeleteConstants)
        self.gui.actionEditConversions.triggered.connect(self.slotEditConversions)
        self.gui.actionEditLimits.triggered.connect(self.slotEditLimits)
        self.gui.actionNewDevice.triggered.connect(self.slotNewDevice)
        self.gui.actionNewDeviceType.triggered.connect(self.slotNewDeviceType)
        self.gui.actionNewCalibration.triggered.connect(self.slotNewCalibration)
        self.gui.actionEditDevice.triggered.connect(self.slotEditDevice)
        self.gui.actionEditDeviceType.triggered.connect(self.slotEditDeviceType)
        self.gui.actionEditCalibration.triggered.connect(self.slotEditCalibration)
        self.gui.actionUndo.triggered.connect(self.slotUndo)
        self.gui.actionSave.triggered.connect(self.slotSave)
        self.gui.actionSaveUploadToServer.triggered.connect(self.slotSaveUploadToServer)
        self.gui.actionRestore.triggered.connect(self.slotRestore)
        self.gui.actionRestoreToServerVersion.triggered.connect(self.slotRestoreToServerVersion)
        self.gui.actionClose.triggered.connect(self.slotClose)
        #self.gui.closeEvent = self.closeEvent

        # attach signal/slot for changes to the tables
        self.gui.registerTable.itemChanged.connect(self.slotUpdateRegister)
        self.gui.constantTable.itemChanged.connect(self.slotUpdateConstant)

        # change tab
        self.gui.tabs.currentChanged.connect(self.slotTabsIndexChanged)

        # Linux signal handlers
        if self.signalHandler:
            self.signalHandler.sigInt.connect(self.slotSaveClose)
            self.signalHandler.sigTerm.connect(self.slotSaveClose)


    def connectNewRegisterHandlers(self):
        """ Connects the signal/slot handlers for the new register GUI. """

    def connectNewConstantHandlers(self):
        """ Connects the signal/slot handlers for the constant GUI. """

    def connectConversionHandlers(self):
        """ Connects the signal/slot handlers for the conversions GUI. """
        self.conversionsUi.getA0.editingFinished.connect(self.slotUpdateConvGetA0)
        self.conversionsUi.getA1.editingFinished.connect(self.slotUpdateConvGetA1)
        self.conversionsUi.getA2.editingFinished.connect(self.slotUpdateConvGetA2)
        self.conversionsUi.getA3.editingFinished.connect(self.slotUpdateConvGetA3)
        self.conversionsUi.getA4.editingFinished.connect(self.slotUpdateConvGetA4)
        self.conversionsUi.getFunction.editingFinished.connect(self.slotUpdateConvGetFunction)
        self.conversionsUi.getLut.editingFinished.connect(self.slotUpdateConvGetLut)
        self.conversionsUi.getCheckBox.clicked.connect(self.slotUpdateConvGetCheckBox)
        self.conversionsUi.getSelLut.clicked.connect(self.slotUpdateConvGetSelection)
        self.conversionsUi.getSelPoly.clicked.connect(self.slotUpdateConvGetSelection)
        self.conversionsUi.getSelFunc.clicked.connect(self.slotUpdateConvGetSelection)
        self.conversionsUi.setA0.editingFinished.connect(self.slotUpdateConvSetA0)
        self.conversionsUi.setA1.editingFinished.connect(self.slotUpdateConvSetA1)
        self.conversionsUi.setA2.editingFinished.connect(self.slotUpdateConvSetA2)
        self.conversionsUi.setA3.editingFinished.connect(self.slotUpdateConvSetA3)
        self.conversionsUi.setA4.editingFinished.connect(self.slotUpdateConvSetA4)
        self.conversionsUi.setFunction.editingFinished.connect(self.slotUpdateConvSetFunction)
        self.conversionsUi.setLut.editingFinished.connect(self.slotUpdateConvSetLut)
        self.conversionsUi.setCheckBox.clicked.connect(self.slotUpdateConvSetCheckBox)
        self.conversionsUi.setSelLut.clicked.connect(self.slotUpdateConvSetSelection)
        self.conversionsUi.setSelPoly.clicked.connect(self.slotUpdateConvSetSelection)
        self.conversionsUi.setSelFunc.clicked.connect(self.slotUpdateConvSetSelection)
        self.conversionsUi.discardButton.clicked.connect(self.slotDiscardConversions)

    def connectLimitHandlers(self):
        """ Connects the signal/slot handlers for the limits GUI. """
        self.limitsUi.upperError.valueChanged.connect(self.slotUpdateLimits)
        self.limitsUi.upperWarning.valueChanged.connect(self.slotUpdateLimits)
        self.limitsUi.upperValid.valueChanged.connect(self.slotUpdateLimits)
        self.limitsUi.lowerError.valueChanged.connect(self.slotUpdateLimits)
        self.limitsUi.lowerWarning.valueChanged.connect(self.slotUpdateLimits)
        self.limitsUi.lowerValid.valueChanged.connect(self.slotUpdateLimits)
        self.limitsUi.discardButton.clicked.connect(self.slotDiscardLimits)

    def connectNewDeviceHandlers(self):
        """ Connects the signal/slot handlers for the new device GUI. """
        self.newDeviceUi.createFrom.clicked.connect(self.slotUpdateNewDevice)
        self.newDeviceUi.oldDevice.currentIndexChanged.connect(self.slotUpdateNewDevice)
        self.newDeviceUi.deviceType.currentIndexChanged.connect(self.slotUpdateNewDevice)

    def connectNewDeviceTypeHandlers(self):
        """ Connects the signal/slot handlers for the new deviceType GUI. """
        self.newDeviceTypeUi.createFrom.clicked.connect(self.slotUpdateNewDeviceType)

    def connectNewCalibrationHandlers(self):
        """ Connects the signal/slot handlers for the new calibration GUI. """
        self.newCalibrationUi.deviceType.currentIndexChanged.connect(self.slotUpdateNewCalibration)
        self.newCalibrationUi.createFrom.clicked.connect(self.slotUpdateNewCalibration)

    def connectEditDeviceHandlers(self):
        """ Connects the signal/slot handlers for the edit device GUI. """
        self.editDeviceUi.oldName.currentIndexChanged.connect(self.slotUpdateEditDevice1)
        self.editDeviceUi.deviceType.currentIndexChanged.connect(self.slotUpdateEditDevice2)
        self.editDeviceUi.erase.toggled.connect(self.slotUpdateEditDevice3)

    def connectEditDeviceTypeHandlers(self):
        """ Connects the signal/slot handlers for the edit deviceType GUI. """
        self.editDeviceTypeUi.oldName.currentIndexChanged.connect(self.slotUpdateEditDeviceType1)
        self.editDeviceTypeUi.erase.toggled.connect(self.slotUpdateEditDeviceType2)

    def connectEditCalibrationHandlers(self):
        """ Connects the signal/slot handlers for the edit calibration GUI. """
        self.editCalibrationUi.deviceType.currentIndexChanged.\
            connect(self.slotUpdateEditCalibration1)
        self.editCalibrationUi.oldName.currentIndexChanged.connect(self.slotUpdateEditCalibration2)
        self.editCalibrationUi.erase.toggled.connect(self.slotUpdateEditCalibration3)


###################################################################################################
# S E T T I N G S
###################################################################################################

    def readSettings(self):
        """ Reads the settings file """
        try:
            settings = QSettings('SRON', 'HalManager')

            self.windowSize = settings.value('windowSize', QSize(1000, 1000))
            self.windowPos  = settings.value('windowPos', QPoint(0, 0))
            self.tabIndex   = int(settings.value('tabIndex', 0))

            self.sortColumnRegister = int(settings.value('regSortColumn', 1))
            self.sortOrderRegister  = int(settings.value('regSortOrder', Qt.AscendingOrder))
            self.regColumnWidth = [int(settings.value('regNameSize', 200)),
                                   int(settings.value('regAddrSize',  95)),
                                   int(settings.value('regTypeSize',  50)),
                                   int(settings.value('regUnitSize',  50)),
                                   int(settings.value('regAccsSize',  60)),
                                   int(settings.value('regWdthSize',  50)),
                                   int(settings.value('regBlckSize',  50)),
                                   int(settings.value('regOffSize',   55)),
                                   int(settings.value('regDescSize', 285))]

            self.sortColumnConstant = int(settings.value('conSortColumn', 1))
            self.sortOrderConstant  = int(settings.value('conSortOrder', Qt.AscendingOrder))
            self.conColumnWidth = [int(settings.value('conNameSize',  200)),
                                   int(settings.value('conValueSize', 100)),
                                   int(settings.value('conUnitSize',   50)),
                                   int(settings.value('conTypeSize',   50)),
                                   int(settings.value('conDescSize',  400))]

            self.deviceName = settings.value('device')
            self.deviceTypeName = settings.value('deviceType')
            self.calibrationName = settings.value('calibration')
        except:
            logger.warning("readSettings exception occurred")

    def writeSettings(self):
        """ Writes settings file """
        settings = QSettings('SRON', 'HalManager')
        regTableHdr = self.gui.registerTable.horizontalHeader()
        conTableHdr = self.gui.constantTable.horizontalHeader()

        settings.setValue('windowSize', self.gui.size())
        settings.setValue('windowPos', self.gui.pos())
        settings.setValue('tabIndex', self.gui.tabs.currentIndex())

        settings.setValue('regSortColumn', regTableHdr.sortIndicatorSection())
        settings.setValue('regSortOrder', regTableHdr.sortIndicatorOrder())
        settings.setValue('regNameSize', regTableHdr.sectionSize(0))
        settings.setValue('regAddrSize', regTableHdr.sectionSize(1))
        settings.setValue('regTypeSize', regTableHdr.sectionSize(2))
        settings.setValue('regUnitSize', regTableHdr.sectionSize(3))
        settings.setValue('regAccsSize', regTableHdr.sectionSize(4))
        settings.setValue('regWdthSize', regTableHdr.sectionSize(5))
        settings.setValue('regBlckSize', regTableHdr.sectionSize(6))
        settings.setValue('regOffSize', regTableHdr.sectionSize(7))
        settings.setValue('regDescSize', regTableHdr.sectionSize(8))

        settings.setValue('conSortColumn', conTableHdr.sortIndicatorSection())
        settings.setValue('conSortOrder', conTableHdr.sortIndicatorOrder())
        settings.setValue('conNameSize', conTableHdr.sectionSize(0))
        settings.setValue('conValueSize', conTableHdr.sectionSize(1))
        settings.setValue('conUnitSize', conTableHdr.sectionSize(2))
        settings.setValue('conTypeSize', conTableHdr.sectionSize(3))
        settings.setValue('conDescSize', conTableHdr.sectionSize(4))

        settings.setValue('device', self.deviceName)
        settings.setValue('deviceType', self.deviceTypeName)
        settings.setValue('calibration', self.calibrationName)


###################################################################################################
# M A I N   U S E R   I N T E R F A C E
###################################################################################################

    def show(self):
        """ Shows the user interface. """

        # main GUI properties
        self.gui.resize(self.windowSize)
        self.gui.move(self.windowPos)

        self.gui.tabs.blockSignals(True)
        self.gui.tabs.setCurrentIndex(self.tabIndex)
        self.gui.tabs.blockSignals(False)

        regLabels = ['Name', 'Address', 'Type', 'Unit', 'Access', 'Width', 'Block',
                     'Offset', 'Description']
        self.gui.registerTable.setColumnCount(len(regLabels))
        self.gui.registerTable.setRowCount(1)
        self.gui.registerTable.setHorizontalHeaderLabels(regLabels)
        self.gui.registerTable.horizontalHeader().setFixedHeight(40)
        self.gui.registerTable.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)
        self.gui.registerTable.horizontalHeader().setSortIndicatorShown(True)
        self.gui.registerTable.horizontalHeader().setStretchLastSection(True)
        self.gui.registerTable.sortItems(0)
        self.gui.registerTable.setContextMenuPolicy(Qt.CustomContextMenu)
        self.gui.registerTable.horizontalHeader().setSortIndicator(self.sortColumnRegister,
                                                                   self.sortOrderRegister)
        for col in range(len(regLabels)):
            self.gui.registerTable.setColumnWidth(col, self.regColumnWidth[col])

        conLabels = ['Name', 'Value', 'Unit', 'Type', 'Description']
        self.gui.constantTable.setColumnCount(len(conLabels))
        self.gui.constantTable.setRowCount(1)
        self.gui.constantTable.setHorizontalHeaderLabels(conLabels)
        self.gui.constantTable.horizontalHeader().setFixedHeight(40)
        self.gui.constantTable.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)
        self.gui.constantTable.horizontalHeader().setSortIndicatorShown(True)
        self.gui.constantTable.horizontalHeader().setStretchLastSection(True)
        self.gui.constantTable.sortItems(0)
        self.gui.constantTable.setContextMenuPolicy(Qt.CustomContextMenu)
        self.gui.constantTable.horizontalHeader().setSortIndicator(self.sortColumnConstant,
                                                                   self.sortOrderConstant)
        for col in range(len(conLabels)):
            self.gui.constantTable.setColumnWidth(col, self.conColumnWidth[col])

        self.updateGui()
        self.gui.show()

    def updateGui(self):
        """ Fill the all parts of the GUI with data. """
        self.updateDeviceComboBox()
        self.updateDeviceTypeComboBox()
        self.updateCalibrationComboBox()
        self.fillTables()
        self.updateMenu()
        self.updateInfo()
        self.writeSettings()

    def updateDeviceComboBox(self):
        """ Fill the device combobox with known devices. """

        # block the signals
        self.gui.device.blockSignals(True)

        # clear the device combobox
        self.gui.device.clear()

        # fill the device combobox
        deviceList = self.hal.getDeviceNames()
        if deviceList:
            deviceList.sort()
            for item in deviceList:
                self.gui.device.addItem(item)
            if self.deviceName in deviceList:
                self.gui.device.setCurrentText(self.deviceName)
            else:
                self.gui.device.setCurrentIndex(0)
                self.deviceName = self.gui.device.currentText()
            self.gui.device.setEnabled(True)
        else:
            self.gui.device.setEnabled(False)
            self.deviceName = ""

        # de-block the signals
        self.gui.device.blockSignals(False)

    def updateDeviceTypeComboBox(self):
        """ Updates the deviceType combobox. """

        # block the signals
        self.gui.deviceType.blockSignals(True)

        # clear the deviceType combobox
        self.gui.deviceType.clear()

        # fill the deviceType combobox
        if self.deviceName != "":
            self.deviceTypeName = self.hal.getDevice(self.deviceName).deviceType
            self.gui.deviceType.addItem(self.deviceTypeName)
            self.gui.deviceType.setCurrentIndex(0)
            self.gui.deviceType.setEnabled(False)
        else:
            deviceTypeList = self.hal.getDeviceTypeNames()
            if deviceTypeList:
                deviceTypeList.sort()
                for item in deviceTypeList:
                    self.gui.deviceType.addItem(item)
                if self.deviceTypeName in deviceTypeList:
                    self.gui.deviceType.setCurrentText(self.deviceTypeName)
                else:
                    self.gui.deviceType.setCurrentIndex(0)
                    self.deviceTypeName = self.gui.deviceType.currentText()
                self.gui.deviceType.setEnabled(True)
            else:
                self.gui.deviceType.setEnabled(False)
                self.deviceTypeName = ""

        # de-block the signals
        self.gui.deviceType.blockSignals(False)

    def updateCalibrationComboBox(self):
        """ Updates the calibration combobox. """

        # block the signals
        self.gui.calibration.blockSignals(True)

        # clear the calibration combobox
        self.gui.calibration.clear()

        # fill the calibration combobox
        if self.deviceName != "":
            self.calibrationName = self.hal.getDevice(self.deviceName).calibration
            self.gui.calibration.addItem(self.calibrationName)
            self.gui.calibration.setCurrentIndex(0)
            self.gui.calibration.setEnabled(False)
        else:
            if self.deviceTypeName != "":
                calibrationList = self.hal.getCalibrationNames(self.deviceTypeName)
                if calibrationList:
                    calibrationList.sort()
                    for item in calibrationList:
                        self.gui.calibration.addItem(item)
                    if self.calibrationName in calibrationList:
                        self.gui.calibration.setCurrentText(self.calibrationName)
                    else:
                        self.gui.calibration.setCurrentIndex(0)
                        self.calibrationName = self.gui.calibration.currentText()
                    self.gui.calibration.setEnabled(True)
                else:
                    self.gui.calibration.setEnabled(False)
                    self.calibrationName = ""
            else:
                self.gui.calibration.setEnabled(False)
                self.calibrationName = ""

        # de-block the signals
        self.gui.calibration.blockSignals(False)

    def fillTables(self):
        """ Fill the register and constant table widgets with data. """

        # block the signals
        self.gui.registerTable.blockSignals(True)
        self.gui.constantTable.blockSignals(True)

        # remove old contents
        self.gui.registerTable.clearContents()
        self.gui.constantTable.clearContents()

        # check if there is a deviceType
        if self.deviceTypeName == "":
            return

        # set the register info
        registers = self.hal.getRegistersInfo(self.deviceTypeName)
        self.gui.registerTable.setRowCount(len(registers))
        row = 0
        for key in registers:
            self.fillRegister(registers[key], row)
            row += 1

        # set the constant info
        constants = self.hal.getConstantsInfo(self.deviceTypeName)
        self.gui.constantTable.setRowCount(len(constants))
        row = 0
        for key in constants:
            if self.calibrationName == "":
                value = None
            else:
                value = self.hal.getConstantValue(key, deviceTypeName=self.deviceTypeName,
                                                       calibrationName=self.calibrationName)
            self.fillConstant(constants[key], value, row)
            self.gui.constantTable.setRowHeight(row, 20)
            row += 1

        # sort tables
        self.slotSortRegisterTable()
        self.slotSortConstantTable()

        # Resize rows (for multi-line register descriptions)
        self.gui.registerTable.resizeRowsToContents()

        # de-block the signals
        self.gui.registerTable.blockSignals(False)
        self.gui.constantTable.blockSignals(False)

    def fillRegister(self, register, row):
        """ Fill one row with register info. """
        self.fillRegisterCell(register.name, row, 0)
        self.fillRegisterCell("0x%08x" % register.address, row, 1)
        self.fillRegisterCell(register.type, row, 2)
        self.fillRegisterCell(register.unit, row, 3)
        self.fillRegisterCell(access2String(register.access), row, 4)
        self.fillRegisterCell(register.width, row, 5)
        self.fillRegisterCell(register.block, row, 6)
        self.fillRegisterCell(register.maxOffset, row, 7)
        self.fillRegisterCell(register.description, row, 8, tooltip=register.description)

    def fillRegisterCell(self, content, row, col, tooltip=None):
        """ Fills the contents of one table cell. """
        tableItem = QTableWidgetItem()
        tableItem.setText(str(content))
        if tooltip is not None:
            tableItem.setToolTip(tooltip)
        self.gui.registerTable.setItem(row, col, tableItem)

    def fillConstant(self, constant, value, row):
        """ Fill one row with constant info. """
        self.fillConstantCell(constant.name, row, 0, readOnly=True)
        self.fillConstantCell(value, row, 1)
        self.fillConstantCell(constant.unit, row, 2)
        self.fillConstantCell(constant.type, row, 3)
        self.fillConstantCell(constant.description, row, 4)

    def fillConstantCell(self, content, row, col, readOnly=False):
        """ Fills the contents of one table cell. """
        tableItem = QTableWidgetItem()
        if content is None:
            tableItem = EmptyConstantValueCell()
        else:
            tableItem.setFlags(tableItem.flags() | Qt.ItemIsEnabled)
            tableItem.setFlags(tableItem.flags() | Qt.ItemIsSelectable)
            tableItem.setFlags(tableItem.flags() | Qt.ItemIsEditable)
            tableItem.setText(str(content))

        if readOnly:
            tableItem.setFlags(tableItem.flags() & ~Qt.ItemIsEditable)

        self.gui.constantTable.setItem(row, col, tableItem)

    def updateMenu(self):
        """ Update enable state of the main GUI menu items. """
        selectedRegisterCount = len(self.gui.registerTable.selectedIndexes())/\
                                self.gui.registerTable.columnCount()
        selectedConstantCount = len(self.gui.constantTable.selectedIndexes())/\
                                self.gui.constantTable.columnCount()
        self.gui.actionNewRegister.setEnabled(len(self.hal.getDeviceTypeNames()) > 0)
        self.gui.actionDeleteRegisters.setEnabled(selectedRegisterCount > 0)
        self.gui.actionNewConstant.setEnabled(len(self.hal.getDeviceTypeNames()) > 0)
        self.gui.actionRenameConstant.setEnabled(selectedConstantCount == 1)
        self.gui.actionDeleteConstants.setEnabled(selectedConstantCount > 0)
        self.gui.actionEditConversions.setEnabled(selectedRegisterCount == 1)
        self.gui.actionEditLimits.setEnabled(selectedRegisterCount == 1)
        self.gui.actionNewDevice.setEnabled((len(self.hal.getDeviceTypeNames()) > 0) and
                                            (len(self.hal.getCalibrationNames()) > 0))
        self.gui.actionNewDeviceType.setEnabled(True)
        self.gui.actionNewCalibration.setEnabled(len(self.hal.getDeviceTypeNames()) > 0)
        self.gui.actionEditDevice.setEnabled(len(self.hal.getDeviceNames()) > 0)
        self.gui.actionEditDeviceType.setEnabled(len(self.hal.getDeviceTypeNames()) > 0)
        self.gui.actionEditCalibration.setEnabled(len(self.hal.getCalibrationNames()) > 0)
        self.gui.actionSave.setEnabled(self.hal != self.halOriginal)

        # Enable only when there are changes
        uploadAvailable = self.halClient.checkHalUploadAvailable()
        self.gui.actionSaveUploadToServer.setEnabled(uploadAvailable)
        self.gui.actionRestore.setEnabled(self.hal != self.halOriginal)

        # If the user can upload, he can also revert 
        self.gui.actionRestoreToServerVersion.setEnabled(uploadAvailable)
        self.gui.actionUndo.setEnabled(len(self.halStack) != 0)

    def updateInfo(self):
        """ Updates the Hal version info displayed in the gui """

        self.gui.versionDisplay.setText(self.hal.getVersionNumber())

###################################################################################################
# R E G I S T E R   U S E R   I N T E R F A C E
###################################################################################################

    def getRegisterName(self):
        """ Get the name of the selected register from the main GUI. """
        currentRow = self.gui.registerTable.currentRow()
        try:
            registerName = self.gui.registerTable.item(currentRow, 0).text()
        except AttributeError:
            registerName = ""
        return registerName

    def getRegisterUpdates(self):
        """ Slot: called when an item in the register table is changed. """
        row = self.gui.registerTable.currentRow()
        col = self.gui.registerTable.currentColumn()
        value = self.gui.registerTable.item(row, col).text()
        logger.debug("item changed: row,col = %d,%d", row, col)

        # get all all calibrations belonging to the actual deviceType
        calibrations = self.hal.getCalibrations(self.deviceTypeName)

        # take appropriate actions depending on the selected column
        if col == 0: # name changed

            # new name must be a string
            try:
                registerName = str(value)
            except:
                gsswUiError("Name must be a string")
                return

            # new name must be an identifier
            if not registerName.isidentifier():
                gsswUiError("Name must be an identifier")
                return

            # new name should not be in use
            if self.hal.registerExists(self.deviceTypeName, registerName):
                gsswUiError("Name already in use")
                return

            # pop actual register object based on address field
            registerAddress = int(self.gui.registerTable.item(row, 1).text(), 0)
            register = self.hal.getRegisterInfo(self.deviceTypeName, registerAddress)
            self.hal.removeRegisterInfo(self.deviceTypeName, registerAddress)

            # update the conversion name in applicable calibration objects
            for cal in calibrations:
                try:
                    cal.updateConversionName(register.name, registerName)  # old name, new name
                except DataError:
                    pass

            # update name field and push the updated register object
            register.name = registerName
            self.hal.addRegisterInfo(self.deviceTypeName, register)

        elif col == 1: # address changed

            # new address must be an integer
            try:
                registerAddress = int(value, 0)
            except ValueError:
                gsswUiError("Address must be integer")
                return

            # new address should not be in use
            if self.hal.registerExists(self.deviceTypeName, registerAddress):
                gsswUiError("Address already in use")
                return

            # pop actual register object based on name field
            registerName = self.gui.registerTable.item(row, 0).text()
            register = self.hal.getRegisterInfo(self.deviceTypeName, registerName)
            self.hal.removeRegisterInfo(self.deviceTypeName, registerName)

            # update address field and push the updated register object
            register.address = registerAddress
            self.hal.addRegisterInfo(self.deviceTypeName, register)

            # make sure the address is displayed as hex string
            self.gui.registerTable.blockSignals(True)
            self.gui.registerTable.item(row, 1).setText("0x%08x" % registerAddress)
            self.gui.registerTable.blockSignals(False)

        elif col == 2: # type changed

            # type must be one of int, unsigned int or float
            registerType = value.lower()
            if registerType not in ["int", "uint", "float"]:
                gsswUiError('Type must be one of "int", "uint" or "float"')
                return

            # get actual register object based on name field and update type field
            registerName = self.gui.registerTable.item(row, 0).text()
            register = self.hal.getRegisterInfo(self.deviceTypeName, registerName)
            register.type = registerType

        elif col == 3: # unit changed

            # unit must be a string
            try:
                registerUnit = str(value)
            except:
                gsswUiError("Unit must be a string")
                return

            # get actual register object based on name field and update unit field
            registerName = self.gui.registerTable.item(row, 0).text()
            register = self.hal.getRegisterInfo(self.deviceTypeName, registerName)
            register.unit = registerUnit

        elif col == 4: # access mode changed

            # access must be one of rw, ro or wo
            registerAccess = value.lower()
            if registerAccess not in ["rw", "ro", "wo"]:
                gsswUiError('Access mode must be one of "rw", "ro" or "wo"')
                return

            # get actual register object based on name field and update access field
            registerName = self.gui.registerTable.item(row, 0).text()
            register = self.hal.getRegisterInfo(self.deviceTypeName, registerName)
            register.access = string2Access(registerAccess)

        elif col == 5: # width changed

            # width must be an integer
            try:
                registerWidth = int(value)
            except:
                gsswUiError("Width must be integer")
                return

            # get actual register object based on name field and update width field
            registerName = self.gui.registerTable.item(row, 0).text()
            register = self.hal.getRegisterInfo(self.deviceTypeName, registerName)
            register.width = registerWidth

        elif col == 6: # block changed

            # block must be one of true or false
            registerBlock = value.lower()
            if registerBlock not in ["true", "false"]:
                gsswUiError('Block must be one of "True" or "False"')
                return

            # get actual register object based on name field and update block field
            registerName = self.gui.registerTable.item(row, 0).text()
            register = self.hal.getRegisterInfo(self.deviceTypeName, registerName)
            register.block = bool(registerBlock == 'true')

        elif col ==7: # offset changed

            # offset must be integer
            try:
                registerMaxOffset = int(value)
            except:
                gsswUiError("Offset must be integer")
                return

            # get actual register object based on name field and update offset field
            registerName = self.gui.registerTable.item(row, 0).text()
            register = self.hal.getRegisterInfo(self.deviceTypeName, registerName)
            register.maxOffset = registerMaxOffset

        elif col == 8: # description changed

            # description must be a string
            try:
                registerDescription = str(value)
            except:
                gsswUiError("Description must be a string")
                return

            # get actual register object based on name field and update description field
            registerName = self.gui.registerTable.item(row, 0).text()
            register = self.hal.getRegisterInfo(self.deviceTypeName, registerName)
            register.description = registerDescription

    def deleteRegisters(self):
        """ Delete the selected register(s). """

        # determine which register(s) need(s) to be deleted
        indexes = []
        for idx in self.gui.registerTable.selectedIndexes():
            indexes.append(idx.row())
        rowsToDelete = list(set(indexes))
        rowsToDelete.sort(reverse=True)

        # get all calibrations belonging to the active deviceType
        calibrations = self.hal.getCalibrations(self.deviceTypeName)

        # delete the selected register(s) including the annexed conversion(s)
        for row in rowsToDelete:
            registerName = self.gui.registerTable.item(row, 0).text()
            self.hal.removeRegisterInfo(self.deviceTypeName, registerName)
            for calibration in calibrations:
                try:
                    calibration.removeConversion(registerName)
                except DataError:
                    pass

    def saveNewRegister(self):
        """ Gets the data from the newRegisterUi and assigns this to the register objects. """

        # first, check if all required QLineEdit widgets have been filled in
        name = self.newRegisterUi.name.text()
        address = self.newRegisterUi.address.text()
        if not (bool(name) and bool(address)):
            gsswUiError("Can not create register: empty field(s)")
            return False

        # check the name (must be an identifier)
        if not name.isidentifier():
            gsswUiError("Name must be an identifier")
            return False

        # check the address (must be integer)
        try:
            address = int(address,0)
        except ValueError:
            gsswUiError("Address input must be integer")
            return False

        # check if there is already an existing register with the same name in the deviceType
        if self.hal.registerExists(self.deviceTypeName, name):
            gsswUiError("Register name must be unique")
            return False
        if self.hal.registerExists(self.deviceTypeName, address):
            gsswUiError("Register address must be unique")
            return False

        # create a new register (blockData must be stored as a boolean)
        register = Register(name,
                            address,
                            string2Access(self.newRegisterUi.access.currentText()),
                            self.newRegisterUi.type.currentText(),
                            self.newRegisterUi.width.value(),
                            self.newRegisterUi.unit.text(),
                            bool(self.newRegisterUi.blockData.currentText().lower() == 'true'),
                            self.newRegisterUi.maxOffset.value(),
                            self.newRegisterUi.description.toPlainText())

        # add the new register to the HAL
        self.hal.addRegisterInfo(self.deviceTypeName, register)
        return True

###################################################################################################
# C O N S T A N T  U S E R   I N T E R F A C E
###################################################################################################

    def getConstantUpdates(self):
        """ Slot: called when an item in the constant table is changed. """
        row = self.gui.constantTable.currentRow()
        col = self.gui.constantTable.currentColumn()
        value = self.gui.constantTable.item(row, col).text()
        logger.debug("item changed: row,col = %d,%d", row, col)

        # take appropriate actions depending on the selected column
        if col == 1:  # constant value changed

            # get actual constant object based on name field
            constantName = self.gui.constantTable.item(row, 0).text()
            constant = self.hal.getConstantInfo(self.deviceTypeName, constantName)
            constantType = constant.type

            # check new value against the type
            if constantType == "float":
                try:
                    constantValue = float(value)
                except ValueError:
                    gsswUiError("Constant value must be double")
                    return
            elif constantType == "int":
                try:
                    constantValue = int(value)
                except ValueError:
                    gsswUiError("Constant value must be integer")
                    return
            elif constantType == "uint":
                try:
                    constantValue = int(value)
                except ValueError:
                    gsswUiError("Constant value must be unsigned integer")
                    return
                if constantValue < 0:
                    gsswUiError("Constant value must be unsigned integer")
                    return
            elif constantType == "lut":
                try:
                    # Parse the text entered by the user into a valid python object
                    constantValue = literal_eval(value)
                except (SyntaxError, ValueError):
                    gsswUiError("Constant value must be a valid dictionary")
                    return
                # Check that parsed python object is actually a dict (no float, int, etc)
                if not isinstance(constantValue, dict):
                    gsswUiError("Constant value must be a valid dictionary "
                                "(instead of a {}).".format(type(constantValue).__name__))
                    return
            else:
                try:
                    constantValue = str(value)
                except ValueError:
                    gsswUiError("Constant value type must be a string")
                    return

            # update or add constant name/value pair to the constants dict of the actual calibration
            try:
                calibration = self.hal.getCalibration(None, self.deviceTypeName,
                                                      self.calibrationName)
                # if constant already exists, first remove it from the actual calibration
                if constantName in calibration.constants:
                    calibration.removeConstant(constantName)
                # add constant to the actual calibration provided that value is not an empty string
                if constantValue != "":
                    calibration.addConstant(constantName, constantValue)
            except DataError:
                logger.warning("Error retrieving calibrationName: %s", self.calibrationName)

        elif col == 2:  # unit changed

            # unit must be a string
            try:
                constantUnit = str(value)
            except ValueError:
                gsswUiError("Unit must be a string")
                return

            # get actual constant object based on name field and update unit field
            constantName = self.gui.constantTable.item(row, 0).text()
            constant = self.hal.getConstantInfo(self.deviceTypeName, constantName)
            constant.unit = constantUnit

        elif col == 3:  # type changed

            # type must be one of float, int, unsigned int or string
            constantType = value.lower()
            if constantType not in Constant.TYPES:
                gsswUiError('Type must be one of "{}"'.format('", "'.join(Constant.TYPES)))
                return

            # get actual constant object based on name field and update type field
            constantName = self.gui.constantTable.item(row, 0).text()
            constant = self.hal.getConstantInfo(self.deviceTypeName, constantName)
            constant.type = constantType

            # warning
            gsswUiWarning("Calibration constants are NOT automatically adapted to new type!!")

        elif col == 4:  # description changed

            # description must be a string
            try:
                constantDescription = str(value)
            except ValueError:
                gsswUiError("Description must be a string")
                return

            # get actual constant object based on name field and update description field
            constantName = self.gui.constantTable.item(row, 0).text()
            constant = self.hal.getConstantInfo(self.deviceTypeName, constantName)
            constant.description = constantDescription

    def renameConstant(self):
        """ Renames the selected constant. """
        row = self.gui.constantTable.currentRow()
        constName = self.gui.constantTable.item(row, 0).text()

        while True:
            # Popup dialog box to get new name
            newName, response = QInputDialog.getText(self, "Rename '{}'".format(constName),
                                                     "Enter a new constant name:", text=constName)
            if not response:
                return

            # new name must be an identifier
            if not newName.isidentifier():
                gsswUiError("Name must be an identifier")
                continue

            # new name should not be in use
            if self.hal.constantExists(self.deviceTypeName, newName):
                gsswUiError("Constant name '{}' already in use".format(newName))
                continue

            constant = self.hal.getConstantInfo(self.deviceTypeName, constName)
            self.hal.removeConstantInfo(self.deviceTypeName, constName)

            # update constants dict in all calibrations belonging to the actual deviceType
            for cal in self.hal.getCalibrations(self.deviceTypeName):
                try:
                    cal.updateConstantName(constant.name, newName)  # old name, new name
                except DataError:
                    pass

            # update name field and push the updated constant object
            constant.name = newName
            self.hal.addConstantInfo(self.deviceTypeName, constant)
            return

    def deleteConstants(self):
        """ Delete the selected constant(s). """

        # determine which constant(s) need(s) to be deleted
        indexes = []
        for idx in self.gui.constantTable.selectedIndexes():
            indexes.append(idx.row())
        rowsToDelete = list(set(indexes))
        rowsToDelete.sort(reverse=True)

        # determine which calibration(s) belong(s) to the deviceType
        calibrations = self.hal.getCalibrations(self.deviceTypeName)

        # delete the selected constant(s) including the calibration value(s)
        for row in rowsToDelete:
            constantName = self.gui.constantTable.item(row, 0).text()

            # remove the selected constant(s) from the deviceType
            self.hal.removeConstantInfo(self.deviceTypeName, constantName)

            # also remove the selected constant(s) from the calibration(s)
            for calibration in calibrations:
                try:
                    calibration.removeConstant(constantName)
                except DataError:
                    pass

    def saveNewConstant(self):
        """ Gets the data from the newConstantUi and assigns this to the constant objects. """

        # first, check if all required QLineEdit widgets have been filled in
        name = self.newConstantUi.name.text()
        if not bool(name):
            gsswUiError("Can not create constant: empty field(s)")
            return False

        # check the name (must be an identifier)
        if not name.isidentifier():
            gsswUiError("Name must be an identifier")
            return False

        # get the deviceType
        deviceType = self.hal.getDeviceType(self.deviceTypeName)

        # check if there is already an existing constant with the same name in the deviceType
        if deviceType.constantExists(name):
            gsswUiError("Constant name must be unique")
            return False

        # get the actual calibration (if any)
        try:
            calibration = self.hal.getCalibration(None, self.deviceTypeName, self.calibrationName)
        except DataError:
            calibration = None

        # read other fields belonging to the deviceType constant
        unit = self.newConstantUi.unit.text()
        ctype = self.newConstantUi.type.currentText()
        description = self.newConstantUi.description.toPlainText()

        # create a new constant
        constant = Constant(name, unit, ctype, description)

        # add the new constant to the HAL
        deviceType.addConstant(constant)

        # if valid value has been entered, add the constant to the actual calibration (if any)
        if calibration is not None:
            value = self.newConstantUi.value.text()
            if value != "":
                if ctype == 'float':
                    try:
                        constantValue = float(value)
                        calibration.addConstant(name, constantValue)
                    except ValueError:
                        gsswUiWarning("Constant value type must be double\n"
                                    "Creating constant object without value")
                        return False
                elif ctype == "int":
                    try:
                        constantValue = int(value)
                        calibration.addConstant(name, constantValue)
                    except ValueError:
                        gsswUiWarning("Constant value must be integer\n"
                                    "Creating constant object without value")
                        return False
                elif ctype == "uint":
                    try:
                        constantValue = int(value)
                        if constantValue < 0:
                            gsswUiWarning("Constant value must be unsigned integer\n"
                                        "Creating constant object without value")
                        else:
                            calibration.addConstant(name, constantValue)
                    except ValueError:
                        gsswUiWarning("Constant value must be unsigned integer\n"
                                    "Creating constant object without value")
                        return False
                else:
                    try:
                        constantValue = str(value)
                        calibration.addConstant(name, constantValue)
                    except ValueError:
                        gsswUiWarning("Constant value type must be a string\n"
                                    "Creating constant object without value")
                        return False
            else:
                gsswUiWarning("No value given\n"
                              "Creating constant object without value")
                return False
        return True


###################################################################################################
# C O N V E R S I O N S   U S E R   I N T E R F A C E
###################################################################################################

    def clearConversionCheckboxes(self):
        """ Clears the get and set checkboxes in the conversions GUI."""

        # block the signals
        self.conversionsUi.getCheckBox.blockSignals(True)
        self.conversionsUi.setCheckBox.blockSignals(True)
        self.conversionsUi.getLut.blockSignals(True)
        self.conversionsUi.setLut.blockSignals(True)

        # clear the checkboxes
        self.conversionsUi.getCheckBox.setEnabled(True)
        self.conversionsUi.setCheckBox.setEnabled(True)
        self.conversionsUi.getCheckBox.setChecked(False)
        self.conversionsUi.setCheckBox.setChecked(False)
        self.conversionsUi.getLut.setEnabled(True)
        self.conversionsUi.setLut.setEnabled(True)

        # de-block the signals
        self.conversionsUi.getCheckBox.blockSignals(False)
        self.conversionsUi.setCheckBox.blockSignals(False)
        self.conversionsUi.getLut.blockSignals(False)
        self.conversionsUi.setLut.blockSignals(False)

    def fillConversions(self, conversion):
        """ Fills the conversion GUI. """

        # block the signals
        self.conversionsUi.getA0.blockSignals(True)
        self.conversionsUi.getA1.blockSignals(True)
        self.conversionsUi.getA2.blockSignals(True)
        self.conversionsUi.getA3.blockSignals(True)
        self.conversionsUi.getA4.blockSignals(True)
        self.conversionsUi.getFunction.blockSignals(True)
        self.conversionsUi.getLut.blockSignals(True)
        self.conversionsUi.getCheckBox.blockSignals(True)
        self.conversionsUi.getSelLut.blockSignals(True)
        self.conversionsUi.getSelPoly.blockSignals(True)
        self.conversionsUi.getSelFunc.blockSignals(True)
        self.conversionsUi.setA0.blockSignals(True)
        self.conversionsUi.setA1.blockSignals(True)
        self.conversionsUi.setA2.blockSignals(True)
        self.conversionsUi.setA3.blockSignals(True)
        self.conversionsUi.setA4.blockSignals(True)
        self.conversionsUi.setFunction.blockSignals(True)
        self.conversionsUi.setLut.blockSignals(True)
        self.conversionsUi.setCheckBox.blockSignals(True)
        self.conversionsUi.setSelLut.blockSignals(True)
        self.conversionsUi.setSelPoly.blockSignals(True)
        self.conversionsUi.setSelFunc.blockSignals(True)

        # fill in the register name
        self.conversionsUi.registerName.setText(conversion.name)

        # Clear the Gui
        for convField in [self.conversionsUi.getA0, self.conversionsUi.setA0,
                         self.conversionsUi.getA1, self.conversionsUi.setA1,
                         self.conversionsUi.getA2, self.conversionsUi.setA2,
                         self.conversionsUi.getA3, self.conversionsUi.setA3,
                         self.conversionsUi.getA4, self.conversionsUi.setA4]:
            convField.clear()

        # fill in the get polynomial (only the orders that are used)
        try:
            self.conversionsUi.getA0.setText("%g" % conversion.getPolynome[0])
            self.conversionsUi.getA1.setText("%g" % conversion.getPolynome[1])
            self.conversionsUi.getA2.setText("%g" % conversion.getPolynome[2])
            self.conversionsUi.getA3.setText("%g" % conversion.getPolynome[3])
            self.conversionsUi.getA4.setText("%g" % conversion.getPolynome[4])
        except IndexError:
            pass

        # fill in the set polynomial (only the orders that are used)
        try:
            self.conversionsUi.setA0.setText("%g" % conversion.setPolynome[0])
            self.conversionsUi.setA1.setText("%g" % conversion.setPolynome[1])
            self.conversionsUi.setA2.setText("%g" % conversion.setPolynome[2])
            self.conversionsUi.setA3.setText("%g" % conversion.setPolynome[3])
            self.conversionsUi.setA4.setText("%g" % conversion.setPolynome[4])
        except IndexError:
            pass

        # fill in the get and set functions
        self.conversionsUi.getFunction.setText(conversion.getFunction)
        self.conversionsUi.setFunction.setText(conversion.setFunction)

        # fill in the get LUT
        if conversion.getLut is not None:
            self.conversionsUi.getLut.setText(str(conversion.getLut))
        else:
            self.conversionsUi.getLut.clear()

        # fill in the set LUT
        if conversion.setLut is not None:
            self.conversionsUi.setLut.setText(str(conversion.setLut))
        else:
            self.conversionsUi.setLut.clear()

        # fill in the get radio buttons
        self.conversionsUi.getSelLut.setChecked(conversion.getConv == Conversion.LUT)
        self.conversionsUi.getSelPoly.setChecked(conversion.getConv == Conversion.POLYNOME)
        self.conversionsUi.getSelFunc.setChecked(conversion.getConv == Conversion.FUNCTION)

        # fill in the set radio buttons
        self.conversionsUi.setSelLut.setChecked(conversion.setConv == Conversion.LUT)
        self.conversionsUi.setSelPoly.setChecked(conversion.setConv == Conversion.POLYNOME)
        self.conversionsUi.setSelFunc.setChecked(conversion.setConv == Conversion.FUNCTION)


        # Enable Polynomial Textboxes
        self.conversionsUi.getA0.setEnabled(True)
        self.conversionsUi.getA1.setEnabled(True)
        self.conversionsUi.getA2.setEnabled(True)
        self.conversionsUi.getA3.setEnabled(True)
        self.conversionsUi.getA4.setEnabled(True)
        self.conversionsUi.setA0.setEnabled(True)
        self.conversionsUi.setA1.setEnabled(True)
        self.conversionsUi.setA2.setEnabled(True)
        self.conversionsUi.setA3.setEnabled(True)
        self.conversionsUi.setA4.setEnabled(True)


        # get the actual conversion from the Hal and compare with modified conversion
        # in order to determine the enable state of the save and discard buttons
        registerName = self.getRegisterName()
        conversionHal = self.hal.getRegisterConversion(None,
            self.deviceTypeName, self.calibrationName, registerName)
        self.enableSaveCancelConversions(conversion != conversionHal)

        # de-block the signals
        self.conversionsUi.getA0.blockSignals(False)
        self.conversionsUi.getA1.blockSignals(False)
        self.conversionsUi.getA2.blockSignals(False)
        self.conversionsUi.getA3.blockSignals(False)
        self.conversionsUi.getA4.blockSignals(False)
        self.conversionsUi.getFunction.blockSignals(False)
        self.conversionsUi.getLut.blockSignals(False)
        self.conversionsUi.getCheckBox.blockSignals(False)
        self.conversionsUi.getSelLut.blockSignals(False)
        self.conversionsUi.getSelPoly.blockSignals(False)
        self.conversionsUi.getSelFunc.blockSignals(False)
        self.conversionsUi.setA0.blockSignals(False)
        self.conversionsUi.setA1.blockSignals(False)
        self.conversionsUi.setA2.blockSignals(False)
        self.conversionsUi.setA3.blockSignals(False)
        self.conversionsUi.setA4.blockSignals(False)
        self.conversionsUi.setFunction.blockSignals(False)
        self.conversionsUi.setLut.blockSignals(False)
        self.conversionsUi.setCheckBox.blockSignals(False)
        self.conversionsUi.setSelLut.blockSignals(False)
        self.conversionsUi.setSelPoly.blockSignals(False)
        self.conversionsUi.setSelFunc.blockSignals(False)


###################################################################################################
# L I M I T S   U S E R   I N T E R F A C E
###################################################################################################

    def fillLimits(self, registerName, unit, limits):
        """ Fills the limits GUI. """

        # block the signals
        self.limitsUi.upperError.blockSignals(True)
        self.limitsUi.upperWarning.blockSignals(True)
        self.limitsUi.upperValid.blockSignals(True)
        self.limitsUi.lowerError.blockSignals(True)
        self.limitsUi.lowerWarning.blockSignals(True)
        self.limitsUi.lowerValid.blockSignals(True)

        # fill in the table
        self.limitsUi.registerName.setText(registerName)
        self.limitsUi.upperErrorUnitLabel.setText(unit)
        self.limitsUi.upperWarningUnitLabel.setText(unit)
        self.limitsUi.upperValidUnitLabel.setText(unit)
        self.limitsUi.lowerErrorUnitLabel.setText(unit)
        self.limitsUi.lowerWarningUnitLabel.setText(unit)
        self.limitsUi.lowerValidUnitLabel.setText(unit)
        self.limitsUi.upperError.setValue(limits.upperError)
        self.limitsUi.upperWarning.setValue(limits.upperWarn)
        self.limitsUi.upperValid.setValue(limits.upperValid)
        self.limitsUi.lowerError.setValue(limits.lowerError)
        self.limitsUi.lowerWarning.setValue(limits.lowerWarn)
        self.limitsUi.lowerValid.setValue(limits.lowerValid)

        # de-block the signals
        self.limitsUi.upperError.blockSignals(False)
        self.limitsUi.upperWarning.blockSignals(False)
        self.limitsUi.upperValid.blockSignals(False)
        self.limitsUi.lowerError.blockSignals(False)
        self.limitsUi.lowerWarning.blockSignals(False)
        self.limitsUi.lowerValid.blockSignals(False)

    def getLimitUpdates(self, limits):
        """ Gets the data from the registerLimitsUi and assigns this to the limit objects. """
        limits.upperError = self.limitsUi.upperError.value()
        limits.upperWarn = self.limitsUi.upperWarning.value()
        limits.upperValid = self.limitsUi.upperValid.value()
        limits.lowerError = self.limitsUi.lowerError.value()
        limits.lowerWarn = self.limitsUi.lowerWarning.value()
        limits.lowerValid = self.limitsUi.lowerValid.value()


###################################################################################################
# D E V I C E   U S E R   I N T E R F A C E
###################################################################################################

    def saveNewDevice(self):
        """ Store the new device into the HAL. """

        # first, check if all QLineEdit widgets have been filled in
        newDeviceName = self.newDeviceUi.newDevice.text()
        daemon        = self.newDeviceUi.daemon.text()
        scienceDaemon = self.newDeviceUi.scienceDaemon.text()
        if not(bool(newDeviceName) and bool(daemon)):
            gsswUiError("Can not create device: empty field(s)")
            return

        # check the name (must be an identifier)
        if not newDeviceName.isidentifier():
            gsswUiError("Name must be an identifier")
            return

        # check if there is already an existing device with the same name
        deviceList = self.hal.getDeviceNames()
        if newDeviceName in deviceList:
            gsswUiError("Can not create device: name already exists")
            return

        try:
            newApid = int(self.newDeviceUi.apid.text())
        except ValueError:
            gsswUiError("'{}' is not a valid APID".format(self.newDeviceUi.apid.text()))
            return

        # create a new device
        newDevice = Device(newDeviceName,
                           daemon,
                           scienceDaemon,
                           newApid,
                           self.newDeviceUi.port.value(),
                           self.newDeviceUi.calibration.currentText(),
                           self.newDeviceUi.deviceType.currentText())

        # add the new device to the HAL
        self.hal.addDevice(newDevice)

    def saveEditDevice(self):
        """ Modify or delete the device. """

        # get the current name of the device from the GUI
        oldName = self.editDeviceUi.oldName.currentText()

        # check if the 'delete' radio button has been selected
        if self.editDeviceUi.erase.isChecked():
            # remove the indicated device from the HAL
            self.hal.removeDevice(oldName)

        # check if the 'edit' radio button has been selected
        elif self.editDeviceUi.edit.isChecked():
            # edit the device and update the GUI

            # first, check if all QLineEdit widgets have been filled in
            newName       = self.editDeviceUi.newName.text()
            daemon        = self.editDeviceUi.daemon.text()
            scienceDaemon = self.editDeviceUi.scienceDaemon.text()
            if not(bool(newName) and bool(daemon)):
                gsswUiError("Can not edit device: empty field(s)")
                return

            # check the name (must be an identifier)
            if not newName.isidentifier():
                gsswUiError("Name must be an identifier")
                return

            # check if there is already an existing device with the new name
            # if so, return without modifying the device
            deviceList = self.hal.getDeviceNames()
            if newName != oldName and newName in deviceList:
                gsswUiError("Can not rename device: name already exists")
                return

            try:
                newApid = int(self.editDeviceUi.apid.text())
            except ValueError:
                gsswUiError("'{}' is not a valid APID".format(self.editDeviceUi.apid.text()))
                return

            # now modify the device
            device = self.hal.getDevice(oldName)
            self.hal.removeDevice(oldName)
            device.name = newName
            device.daemon = daemon
            device.scienceDaemon = scienceDaemon
            device.apid = newApid
            device.port = self.editDeviceUi.port.value()
            device.calibration = self.editDeviceUi.calibration.currentText()
            device.deviceType = self.editDeviceUi.deviceType.currentText()
            self.hal.addDevice(device)


###################################################################################################
# D E V I C E T Y P E   U S E R   I N T E R F A C E
###################################################################################################

    def saveNewDeviceType(self):
        """ Store the new device type into the HAL. """

        # first, check if all QLineEdit widgets have been filled in
        newDeviceTypeName = self.newDeviceTypeUi.newDeviceType.text()
        if not bool(newDeviceTypeName):
            gsswUiError("Can not create deviceType: empty field(s)")
            return

        # check the name (must be an identifier)
        if not newDeviceTypeName.isidentifier():
            gsswUiError("Name must be an identifier")
            return

        # check if there is already an existing device type with the same name
        deviceTypeList = self.hal.getDeviceTypeNames()
        if newDeviceTypeName in deviceTypeList:
            gsswUiError("Can not create deviceType: name already exists")
            return

        # check if the user wants to create the new device type from an existing device type
        # if so, deep copy the existing device type and replace the name field;
        # else, create a new device type using the constructor
        if self.newDeviceTypeUi.createFrom.isChecked():
            oldDeviceTypeName = self.newDeviceTypeUi.oldDeviceType.currentText()
            oldDeviceType = self.hal.getDeviceType(oldDeviceTypeName)
            newDeviceType = deepcopy(oldDeviceType)
            newDeviceType.name = newDeviceTypeName
            
        else:
            newDeviceType = DeviceType(newDeviceTypeName)

        # add the new device type to the HAL
        self.hal.addDeviceType(newDeviceType)
        
        # Also check if the user wants to create the new device type from an existing device type
        # to prevent deepcopying calibrations without copying the deviceType 
        if self.newDeviceTypeUi.createFrom.isChecked():
        # check if user also wants to copy the calibration(s)
            if self.newDeviceTypeUi.copyCal.isChecked():
                oldCalibrations = self.hal.getCalibrations(oldDeviceTypeName)
                for oldCalibration in oldCalibrations:
                    newCalibration = deepcopy(oldCalibration)
                    newCalibration.deviceType = newDeviceTypeName

                    #Add (sequential) number to newCalibration.name 
                    # looks for the highest number in the list of names and adds +1.
                    # This way, each new copy gets it's own (new) unique number and with that an
                    # unique name (even if it copies a copy) format used:
                    # Calibration.name_(#) where # is the number

                    # Explanation: newCopyNumber gets all the numbers of the Calibrations with the
                    # same name (without numbers). # Take the highest value of those (this is to
                    # prevent errors when copying a copy, which already has a number)
                    # The new name als cuts off any number of the old one, to avoid __#___#_, 
                    # Additionally, this will make each Calibration with the same name
                    # (excluding the number) a copy of the original
                    # i.e. each Calibration named "default__#_" is a copy of "default", 
                    #  even if copied from "default__#_" which already was a copy of "default"
                    # (or "default__#_")

                    #Code explanation:
                    # First, a list comprehension to get all the numbers (and names) from the
                    # Calibrations List
                    # --- [int(i.split(SepStart)[1].split(SepEnd)[0]) \
                    #      for i in self.hal._calibrations.keys()] ---

                    # that actually have a number (i.e. "default__2_" but not "default"
                    # (hence the split)
                    # --- if len(i.split(SepStart)) > 1 --- 

                    # and have the same name as the old CalibrationName
                    # (and turn the numbers to ints)
                    # --- and i.split(SepStart)[0] == oldCalibration.name.split(SepStart)[0]] ---

                    # We add an empty list with 0 ( [0] ) so that there's always an entry in the
                    # list when we take the max() value and add 1 to it, so when it's the first copy
                    # it becomes Name__1_)
                    # --- max(AboveCode + [0]) + 1 ---

                    #Seperators for the number
                    SepStart = "__"
                    SepEnd = "_"
                    newCopyNumber = max(
                        [int(
                            i.split(SepStart)[1].split(
                                SepEnd)[0]) for i in self.hal.getCalibrationNames() 
                                if len(i.split(SepStart)) > 1 
                                and i.split(SepStart)[0] == oldCalibration.name.split(
                                    SepStart)[0]] + [0]) + 1
                    #Formatting the new CalibrationName
                    newCalibration.name = oldCalibration.name.split(
                        SepStart)[0] + "{}{}{}".format(SepStart,newCopyNumber,SepEnd)

                    # add the new calibration(s) to the HAL
                    self.hal.addCalibration(newDeviceTypeName, newCalibration)

    def saveEditDeviceType(self):
        """ Modify or delete the device type. """

        # get the current name of the device type from the GUI
        oldName = self.editDeviceTypeUi.oldName.currentText()

        # check if the 'delete' radio button has been selected
        if self.editDeviceTypeUi.erase.isChecked():
            # remove the indicated deviceType from the HAL and update the GUI
            self.hal.removeDeviceType(oldName)

        # check if the 'rename' radio button has been selected
        elif self.editDeviceTypeUi.rename.isChecked():
            # rename the indicated deviceType and update the GUI

            # first, check if all QLineEdit widgets have been filled in
            newName = self.editDeviceTypeUi.newName.text()
            if not bool(newName):
                gsswUiError("Can not edit device type: empty field(s)")
                return

            # check the name (must be an identifier)
            if not newName.isidentifier():
                gsswUiError("Name must be an identifier")
                return

            # check if there is already an existing device type with the new name
            # if so, return without modifying the device type name
            deviceTypeList = self.hal.getDeviceTypeNames()
            if newName != oldName and newName in deviceTypeList:
                gsswUiError("Can not rename device type: name already exists")
                return

            # now modify the device type
            deviceType = self.hal.getDeviceType(oldName)
            self.hal.removeDeviceType(oldName)
            deviceType.name = newName
            self.hal.addDeviceType(deviceType)

            # modify the device type field in the existing device(s)
            deviceList = self.hal.getDevices(oldName)
            if deviceList:
                for device in deviceList:
                    device.deviceType = newName

            # modify the device type field in the existing calibrations(s)
            # note that also the calibration dict keys need to be changed
            calibrationList = self.hal.getCalibrations(oldName)
            if calibrationList:
                for calibration in calibrationList:
                    calibration.deviceType = newName
                    self.hal.removeCalibration(oldName, calibration.name)
                    self.hal.addCalibration(newName, calibration)


###################################################################################################
# C A L I B R A T I O N   U S E R   I N T E R F A C E
###################################################################################################

    def saveNewCalibration(self):
        """ Store the new calibration into the HAL. """

        # first, check if all QLineEdit widgets have been filled in
        newCalibrationName = self.newCalibrationUi.newCalibration.text()
        if not bool(newCalibrationName):
            gsswUiError("Can not create calibration: empty field(s)")
            return

        # check the name (must be an identifier)
        if not newCalibrationName.isidentifier():
            gsswUiError("Name must be an identifier")
            return

        # check if there is already an existing calibration with the same name
        calibrationList = self.hal.getCalibrationNames()
        if newCalibrationName in calibrationList:
            oldCalibration = self.hal.getCalibration(calibrationName=newCalibrationName)
            gsswUiError("Can not create calibration: name must be unique.\n"
                        "Calibration name already exists for device type '{}'."
                        .format(oldCalibration.deviceType))
            return

        # check if the user wants to create the new calibration from an existing calibration
        # if so, deep copy the existing calibration and replace the name field;
        # else, create a new calibration using the constructor
        deviceTypeName = self.newCalibrationUi.deviceType.currentText()
        if self.newCalibrationUi.createFrom.isChecked():
            oldCalibrationName = self.newCalibrationUi.oldCalibration.currentText()
            oldCalibration = self.hal.getCalibration(None, deviceTypeName, oldCalibrationName)
            newCalibration = deepcopy(oldCalibration)
            newCalibration.name = newCalibrationName
        else:
            newCalibration = Calibration(newCalibrationName, deviceTypeName)

        # add the new calibration to the HAL
        self.hal.addCalibration(deviceTypeName, newCalibration)


    def saveEditCalibration(self):
        """ Modify or delete the calibration. """

        # get the deviceType and old calibration name from the GUI
        deviceTypeName = self.editCalibrationUi.deviceType.currentText()
        oldName = self.editCalibrationUi.oldName.currentText()

        # check if the 'delete' radio button has been selected
        if self.editCalibrationUi.erase.isChecked():
            # remove the indicated calibration from the HAL and update the GUI
            self.hal.removeCalibration(deviceTypeName, oldName)

        # check if the 'edit' radio button has been selected
        elif self.editCalibrationUi.rename.isChecked():
            # edit the calibration and update the GUI

            # first, check if all QLineEdit widgets have been filled in
            newName = self.editCalibrationUi.newName.text()
            if not bool(newName):
                gsswUiError("Can not edit calibration: empty field(s)")
                return

            # check the name (must be an identifier)
            if not newName.isidentifier():
                gsswUiError("Name must be an identifier")
                return

            # check if there is already an existing calibration with the new name
            # if so, return without modifying the calibration
            calibrationList = self.hal.getCalibrationNames()
            if newName != oldName and newName in calibrationList:
                calibration = self.hal.getCalibration(calibrationName=newName)
                gsswUiError("Can not rename calibration: name must be unique.\n"
                            "Calibration name already exists for device type '{}'."
                            .format(calibration.deviceType))
                return

            # now modify the calibration
            calibration = self.hal.getCalibration(None, deviceTypeName, oldName)
            self.hal.removeCalibration(deviceTypeName, oldName)
            calibration.name = newName
            self.hal.addCalibration(deviceTypeName, calibration)

            # modify the calibration field in the existing device(s)
            deviceList = self.hal.getDevices(deviceTypeName)
            if deviceList:
                for device in deviceList:
                    device.calibration = newName


###################################################################################################
# S T A C K  H A N D L I N G
###################################################################################################

    def stackPush(self, hal):
        """ Push the current hal on the stack. """
        self.halStack.append(hal)
        if len(self.halStack) == self.STACKSIZE+1:
            self.halStack.pop(0)

    def stackPull(self):
        """ Pull the previous hal from the stack. """
        if self.halStack:
            hal = self.halStack.pop()
        else:
            hal = None
        return hal


###################################################################################################
# E N A B L E S
###################################################################################################

    def enableSaveCancelConversions(self, enable=True):
        """ Slot: Changes enable state of save and discard button of the conversions GUI. """
        self.conversionsUi.saveButton.setEnabled(enable)
        self.conversionsUi.discardButton.setEnabled(enable)

    def enableSaveCancelLimits(self, enable=True):
        """ Slot: Changes enable state of save and discard button of limit GUI. """
        self.limitsUi.saveButton.setEnabled(enable)
        self.limitsUi.discardButton.setEnabled(enable)

    def enableEditDeviceFields(self, enable=True):
        """ Slot: Changes enable state of Edit fields in the editDevice GUI. """
        self.editDeviceUi.newName.setEnabled(enable)
        self.editDeviceUi.deviceType.setEnabled(enable)
        self.editDeviceUi.calibration.setEnabled(enable)
        self.editDeviceUi.apid.setEnabled(enable)
        self.editDeviceUi.port.setEnabled(enable)
        self.editDeviceUi.daemon.setEnabled(enable)
        self.editDeviceUi.scienceDaemon.setEnabled(enable)

    def enableEditDeviceTypeFields(self, enable=True):
        """ Slot: Changes enable state of Edit fields in the editDeviceType GUI. """
        self.editDeviceTypeUi.newName.setEnabled(enable)

    def enableEditCalibrationFields(self, enable=True):
        """ Slot: Changes enable state of Edit fields in the editCalibration GUI. """
        self.editCalibrationUi.newName.setEnabled(enable)


###################################################################################################
# S L O T S
###################################################################################################

    def slotDeviceChanged(self):
        """ Slot: Update function for when the device comboBox changes value. """
        self.deviceName = self.gui.device.currentText()
        self.updateGui()

    def slotDeviceTypeChanged(self):
        """ Slot: Update function for when the deviceType comboBox changes value. """
        self.deviceTypeName = self.gui.deviceType.currentText()
        self.updateGui()

    def slotCalibrationChanged(self):
        """ Slot: Update function for when the calibration comboBox changes value. """
        self.calibrationName = self.gui.calibration.currentText()
        self.updateGui()

    def slotTabsIndexChanged(self):
        """ Slot: Update function when tabs index changes (register/constant). """
        self.updateGui()

    def slotRegisterSelectionChanged(self):
        """ Slot (main GUI): called when register table selection changed.  """
        self.updateMenu()

    def slotConstantSelectionChanged(self):
        """ Slot (main GUI): called when constant table selection changed.  """
        self.updateMenu()

    def slotRegisterMenu(self, position):
        """ Slot: Process right-click on register tab widget. """
        self.regMenu.exec(self.gui.registerTable.mapToGlobal(position))

    def slotConstantMenu(self, position):
        """ Slot: Process right-click on constant tab widget. """
        self.constMenu.exec(self.gui.constantTable.mapToGlobal(position))

    def slotSortRegisterTable(self):
        """ Slot: Sorts the register table based on the column clicked. """
        col = self.gui.registerTable.horizontalHeader().sortIndicatorSection()
        order = self.gui.registerTable.horizontalHeader().sortIndicatorOrder()
        self.gui.registerTable.sortItems(col, order)
        self.gui.registerTable.resizeRowsToContents()

    def slotSortConstantTable(self):
        """ Slot: Sorts the constant table based on the column clicked. """
        col = self.gui.constantTable.horizontalHeader().sortIndicatorSection()
        order = self.gui.constantTable.horizontalHeader().sortIndicatorOrder()
        self.gui.constantTable.sortItems(col, order)

    def slotNewRegister(self):
        """ Slot (main Gui): create a new register. """

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # execute the dialog
        while True:
            accepted = self.newRegisterUi.exec()
            if accepted:
                # User clicked save, now validate input
                if self.saveNewRegister():
                    # Auto-increment register address
                    oldAddress = self.newRegisterUi.address.text()
                    if oldAddress:
                        nextAddress = int(oldAddress, 0) + 1
                        if 'x' in oldAddress.lower():
                            self.newRegisterUi.address.setText("0x{:08x}".format(nextAddress))
                        else:
                            self.newRegisterUi.address.setText(str(nextAddress))
                    break
            else:
                # User clicked cancel
                break

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotUpdateRegister(self):
        """ Slot: Update function for when a register changes value. """

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # get the update
        self.getRegisterUpdates()

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotDeleteRegisters(self):
        """ Slot (main GUI): delete the selected register(s). """

        # First, ask for confirmation
        if gsswUiMessage('Do you want to delete the selected register(s)?') == QMessageBox.Ok:
            # store the current version of the HAL
            halSave = deepcopy(self.hal)

            # perform the delete operation
            self.deleteRegisters()

            # if HAL has changed, push the previous version onto the stack
            if self.hal != halSave:
                self.stackPush(halSave)

            self.updateGui()

    def slotNewConstant(self):
        """ Slot (main Gui): create a new constant. """

        # initialization
        self.newConstantUi.name.clear()
        self.newConstantUi.value.clear()
        self.newConstantUi.unit.clear()
        self.newConstantUi.type.setCurrentIndex(0)
        self.newConstantUi.description.clear()

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # execute the dialog
        while True:
            accepted = self.newConstantUi.exec()
            if accepted:
                if self.saveNewConstant():
                    break
            else:
                # User clicked cancel
                break

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotUpdateConstant(self):
        """ Slot: Update function for when a constant changes value. """

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # get the update
        self.getConstantUpdates()

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotRenameConstant(self):
        """ Slot (main GUI): rename the selected constant. """

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # perform the rename
        self.renameConstant()

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotDeleteConstants(self):
        """ Slot (main GUI): delete the selected constant(s). """

        # First, ask for confirmation
        if gsswUiMessage('Do you want to delete the selected constant(s)?') == QMessageBox.Ok:
            # store the current version of the HAL
            halSave = deepcopy(self.hal)

            # perform the delete operation
            self.deleteConstants()

            # if HAL has changed, push the previous version onto the stack
            if self.hal != halSave:
                self.stackPush(halSave)

            self.updateGui()

    def slotEditConversions(self):
        """ Slot (main Gui): edit the conversion of the selected register. """

        # get the actual conversion object from the Hal
        registerName = self.getRegisterName()
        conversionHal = self.hal.getRegisterConversion(None,
            self.deviceTypeName, self.calibrationName, registerName)

        # if no conversion object available then create one
        if conversionHal is None:
            # Create with empty conversion polynomial, otherwise the gui will display [0, 1]
            conversionHal = Conversion(registerName, setPolynome=[], getPolynome=[])

        # create working copy of actual (or created) conversion object
        self.conversion = deepcopy(conversionHal)

        self.clearConversionCheckboxes()
        self.fillConversions(self.conversion)

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # execute the dialog
        accepted = self.conversionsUi.exec()

        # replace existing conversion object (if any) with modified conversion object
        if accepted:
            self.hal.removeRegisterConversion(None, self.deviceTypeName, self.calibrationName,
                    registerName)
            self.hal.addRegisterConversion(None, self.deviceTypeName, self.calibrationName,
                    registerName, self.conversion)

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def _updateConvGet(self, index, rawValue):
        """ Updates the conversion order value at the given index. """
        if not rawValue: # Handle empty textbox
            rawValue = 0
        try:
            getA = float(rawValue)
        except ValueError:
            gsswUiError("Polynomial must be double")
            return
        self.conversion.getPolynome[index] = getA
        self.fillConversions(self.conversion)

    def slotUpdateConvGetA0(self):
        """ Slot: update get A0 polynomial. """
        self._updateConvGet(0, self.conversionsUi.getA0.text())

    def slotUpdateConvGetA1(self):
        """ Slot: update get A1 polynomial. """
        self._updateConvGet(1, self.conversionsUi.getA1.text())

    def slotUpdateConvGetA2(self):
        """ Slot: update get A2 polynomial. """
        self._updateConvGet(2, self.conversionsUi.getA2.text())

    def slotUpdateConvGetA3(self):
        """ Slot: update get A3 polynomial. """
        self._updateConvGet(3, self.conversionsUi.getA3.text())

    def slotUpdateConvGetA4(self):
        """ Slot: update get A4 polynomial. """
        self._updateConvGet(4, self.conversionsUi.getA4.text())

    def slotUpdateConvGetFunction(self):
        """ Slot: update get function. """
        self.conversion.getFunction = self.conversionsUi.getFunction.text()
        self.fillConversions(self.conversion)

    def slotUpdateConvGetLut(self):
        """ Slot: update get LUT. """

        try:
            getLut = literal_eval(str(self.conversionsUi.getLut.text()))
        except SyntaxError:
            getLut = None
        if isinstance(getLut, dict):
            self.conversion.getLut = getLut
        else:
            self.conversion.getLut = {}
        self.fillConversions(self.conversion)

    def slotUpdateConvGetCheckBox(self):
        """ Slot: update get checkbox. """
        
        if self.conversionsUi.getCheckBox.isChecked():
            self.conversionsUi.setCheckBox.setEnabled(False)
            # Disable editing for the Textboxes and Block Signals
            self._changeConversionUiTextFieldsStatus("get", False, True)
            # Fill Values
            if len(self.conversion.setPolynome) > 2:
                gsswUiWarning("Attempted to Invert non-linear function. Check Input.")
                self.conversionsUi.getCheckBox.setChecked(False)
            try:
                self.conversion.getPolynome[0] = (-1)*self.conversion.setPolynome[0] \
                                                /  self.conversion.setPolynome[1]
                self.conversion.getPolynome[1] = 1/self.conversion.setPolynome[1]
                self.conversion.getLut = {v: k for k, v in self.conversion.setLut.items()}
            except ZeroDivisionError:
                gsswUiWarning("Attempted Divion by Zero! Check Input.")
                self.conversionsUi.getCheckBox.setChecked(False)
            except IndexError:
                gsswUiWarning("Insufficient input defined! Check Input.")
                self.conversionsUi.getCheckBox.setChecked(False)
                # Un-block Signals from TextBoxes
            self._changeConversionUiTextFieldsStatus("get", False, False)
        #Note: Don't use else, value can be changed during the above if statement.
        if not self.conversionsUi.getCheckBox.isChecked():
            self.conversionsUi.setCheckBox.setEnabled(True)
            self._changeConversionUiTextFieldsStatus("get", True, False)
        self.fillConversions(self.conversion)

    def slotUpdateConvGetSelection(self):
        """ Slot: update get selection. """

        if self.conversionsUi.getSelPoly.isChecked():
            self.conversion.getConv = Conversion.POLYNOME
        elif self.conversionsUi.getSelFunc.isChecked():
            self.conversion.getConv = Conversion.FUNCTION
        else:
            self.conversion.getConv = Conversion.LUT
        self.fillConversions(self.conversion)

    def _updateConvSet(self, index, rawValue):
        """ Updates the conversion order value at the given index. """
        if not rawValue: # Handle empty textbox
            rawValue = 0
        try:
            setA = float(rawValue)
        except ValueError:
            gsswUiError("Polynomial must be double")
            return
        self.conversion.setPolynome[index] = setA
        self.fillConversions(self.conversion)

    def slotUpdateConvSetA0(self):
        """ Slot: update set A0 polynomial"""
        self._updateConvSet(0, self.conversionsUi.setA0.text())

    def slotUpdateConvSetA1(self):
        """ Slot: update set A1 polynomial"""
        self._updateConvSet(1, self.conversionsUi.setA1.text())

    def slotUpdateConvSetA2(self):
        """ Slot: update set A2 polynomial"""
        self._updateConvSet(2, self.conversionsUi.setA2.text())

    def slotUpdateConvSetA3(self):
        """ Slot: update set A3 polynomial"""
        self._updateConvSet(3, self.conversionsUi.setA3.text())

    def slotUpdateConvSetA4(self):
        """ Slot: update set A4 polynomial"""
        self._updateConvSet(4, self.conversionsUi.setA4.text())

    def slotUpdateConvSetFunction(self):
        """ Slot: update set function. """
        self.conversion.setFunction = self.conversionsUi.setFunction.text()
        self.fillConversions(self.conversion)

    def slotUpdateConvSetLut(self):
        """ Slot: update get LUT. """

        try:
            setLut = literal_eval(str(self.conversionsUi.setLut.text()))
        except SyntaxError:
            setLut = None
        if isinstance(setLut, dict):
            self.conversion.setLut = setLut
        else:
            self.conversion.setLut = {}
        self.fillConversions(self.conversion)

    def slotUpdateConvSetCheckBox(self):
        """ Slot: update set checkbox. """

        if self.conversionsUi.setCheckBox.isChecked():
            self.conversionsUi.getCheckBox.setEnabled(False)
            # Disable editing for the Textboxes and Block Signals
            self._changeConversionUiTextFieldsStatus("set", False, True)
            # Fill Values
            if len(self.conversion.getPolynome) > 2:
                gsswUiWarning("Attempted to Invert non-linear function. Check Input.")
            try:
                self.conversion.setPolynome[0] = (-1)*self.conversion.getPolynome[0] \
                                                    / self.conversion.getPolynome[1]
                self.conversion.setPolynome[1] = 1/self.conversion.getPolynome[1]
                self.conversion.setLut = {v: k for k, v in self.conversion.getLut.items()}
            except ZeroDivisionError:
                gsswUiWarning("Attempted Divion by Zero! Check Input.")
                self.conversionsUi.setCheckBox.setChecked(False)
            except IndexError:
                gsswUiWarning("Insufficient input defined! Check Input.")
                self.conversionsUi.setCheckBox.setChecked(False)
            # Un-block Signals from TextBoxes
            self._changeConversionUiTextFieldsStatus("set", False, False)
        #Note: Don't use else, value can be changed during the above if statement.
        if not self.conversionsUi.setCheckBox.isChecked():
            self.conversionsUi.getCheckBox.setEnabled(True)
            self._changeConversionUiTextFieldsStatus("set", True, False)
        self.fillConversions(self.conversion)


    def _changeConversionUiTextFieldsStatus(self, field, enabledStatus, blockSignalStatus):
        """
        Enable/Disable editing and Qt Signals for the Textboxes
        The field input specific the get/set fields in the conversions. 
        enabledStatus and bloackSignalStatus are boolean.
        """

        if field == "get":
            for fld in [self.conversionsUi.getA0,
                        self.conversionsUi.getA1,
                        self.conversionsUi.getA2,
                        self.conversionsUi.getA3,
                        self.conversionsUi.getA4,
                        self.conversionsUi.getLut]:

                # Set Signal and Editing
                fld.setEnabled(enabledStatus)
                fld.blockSignals(blockSignalStatus)
        elif field == "set":
            for fld in [self.conversionsUi.setA0,
                        self.conversionsUi.setA1,
                        self.conversionsUi.setA2,
                        self.conversionsUi.setA3,
                        self.conversionsUi.setA4,
                        self.conversionsUi.setLut]:

                # Set Signal and Editing
                fld.setEnabled(enabledStatus)
                fld.blockSignals(blockSignalStatus)
        else:
            logger.warning("Function field input not valid.")

    def slotUpdateConvSetSelection(self):
        """ Slot: update set selection. """

        if self.conversionsUi.setSelPoly.isChecked():
            self.conversion.setConv = Conversion.POLYNOME
        elif self.conversionsUi.setSelFunc.isChecked():
            self.conversion.setConv = Conversion.FUNCTION
        else:
            self.conversion.setConv = Conversion.LUT
        self.fillConversions(self.conversion)

    def slotDiscardConversions(self):
        """ Slot: Discards edits in the conversions GUI. """

        # get the actual conversion object from the Hal
        registerName = self.getRegisterName()
        conversionHal = self.hal.getRegisterConversion(None,
            self.deviceTypeName, self.calibrationName, registerName)

        # if no conversion object available then create one
        if conversionHal is None:
            conversionHal = Conversion(registerName)

        # create working copy of actual (or created) conversion object
        self.conversion = deepcopy(conversionHal)

        self.clearConversionCheckboxes()
        self.fillConversions(self.conversion)

    def slotEditLimits(self):
        """ Slot (main Gui): edit the limits of the selected register. """

        # first, get the actual conversion from the Hal
        new = False
        registerName = self.getRegisterName()
        conversion = self.hal.getRegisterConversion(None,
            self.deviceTypeName, self.calibrationName, registerName)

        if conversion is None:
            gsswUiError("Can not create Limits object, first create a Conversion")
            return

        # now get the actual limits from the Hal
        limits = self.hal.getRegisterLimits(None, self.deviceTypeName,
            self.calibrationName, registerName)

        # if no limits available then create them
        if limits is None:
            limits = Limits(registerName, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
            new = True

        unit = self.hal.getRegisterInfo(self.deviceTypeName, registerName).unit
        self.fillLimits(registerName, unit, limits)
        self.enableSaveCancelLimits(new)

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # show the dialog
        accepted = self.limitsUi.exec()

        # get updated values and if applicable, add the new limits object to the Hal.
        if accepted:
            self.getLimitUpdates(limits)  # get updated values
            if new:
                conversion.updateLimits(limits)

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotUpdateLimits(self):
        """ Slot: Update enable states of Save and Discard buttons in the limits GUI. """

        # get the actual limits from the Hal
        registerName = self.getRegisterName()
        limitsHal = self.hal.getRegisterLimits(None,
            self.deviceTypeName, self.calibrationName, registerName)

        # create an empty limits object in order to store the GUI values
        limitsGui = Limits(registerName, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
        self.getLimitUpdates(limitsGui)

        # enable Save and Discard buttons if new limits object or existing limits changed
        if limitsHal is None:
            self.enableSaveCancelLimits(True)
        else:
            self.enableSaveCancelLimits(limitsHal != limitsGui)

    def slotDiscardLimits(self):
        """ Slot: Discard edits in the limits GUI. """

        # get the actual limits from the Hal
        new = False
        registerName = self.getRegisterName()
        limits = self.hal.getRegisterLimits(None,
            self.deviceTypeName, self.calibrationName, registerName)

        # if no limits available then create them
        if limits is None:
            limits = Limits(registerName, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
            new = True

        unit = self.hal.getRegisterInfo(self.deviceTypeName, registerName).unit
        self.fillLimits(registerName, unit, limits)
        self.enableSaveCancelLimits(new)

    def slotNewDevice(self):
        """ Slot (main Gui): create a new device. """

        # block the signals
        self.newDeviceUi.createFrom.blockSignals(True)
        self.newDeviceUi.oldDevice.blockSignals(True)
        self.newDeviceUi.deviceType.blockSignals(True)

        # initialization
        self.newDeviceUi.newDevice.clear()
        self.newDeviceUi.oldDevice.clear()
        self.newDeviceUi.deviceType.clear()
        self.newDeviceUi.calibration.clear()
        self.newDeviceUi.apid.clear()
        self.newDeviceUi.port.setValue(0)
        self.newDeviceUi.daemon.clear()
        self.newDeviceUi.scienceDaemon.clear()
        self.newDeviceUi.createFrom.setChecked(False)
        self.newDeviceUi.oldDevice.setEnabled(False)

        # fill in the 'oldDevice' combobox
        deviceList = self.hal.getDeviceNames()
        if deviceList:
            deviceList.sort()
            for deviceName in deviceList:
                self.newDeviceUi.oldDevice.addItem(deviceName)
            if self.deviceName in deviceList:
                self.newDeviceUi.oldDevice.setCurrentText(self.deviceName)
            else:
                self.newDeviceUi.oldDevice.setCurrentIndex(0)
            self.newDeviceUi.createFrom.setEnabled(True)
        else:
            self.newDeviceUi.createFrom.setEnabled(False)

        # fill in the deviceType combobox (always at least one deviceType present)
        deviceTypeList = self.hal.getDeviceTypeNames()
        deviceTypeList.sort()
        shortList = []
        for deviceTypeName in deviceTypeList:
            # only add deviceTypes with calibrations to the menu
            if self.hal.getCalibrationNames(deviceTypeName):
                self.newDeviceUi.deviceType.addItem(deviceTypeName)
                shortList.append(deviceTypeName)
        if self.deviceTypeName in shortList:
            self.newDeviceUi.deviceType.setCurrentText(self.deviceTypeName)
        else:
            self.newDeviceUi.deviceType.setCurrentIndex(0)

        # fill in the calibration combobox (always at least one calibration present)
        deviceTypeName = self.newDeviceUi.deviceType.currentText()
        calibrationList = self.hal.getCalibrationNames(deviceTypeName)
        for calibrationName in calibrationList:
            self.newDeviceUi.calibration.addItem(calibrationName)
        if self.calibrationName in calibrationList:
            self.newDeviceUi.calibration.setCurrentText(self.calibrationName)
        else:
            self.newDeviceUi.calibration.setCurrentIndex(0)

        # de-block the signals
        self.newDeviceUi.createFrom.blockSignals(False)
        self.newDeviceUi.oldDevice.blockSignals(False)
        self.newDeviceUi.deviceType.blockSignals(False)

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # execute the dialog
        accepted = self.newDeviceUi.exec()
        if accepted:
            self.saveNewDevice()

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotUpdateNewDevice(self):
        """ Slot: Update the combobox in the newDevice menu. """
        if self.newDeviceUi.createFrom.isChecked():
            self.newDeviceUi.oldDevice.setEnabled(True)
            oldDeviceName = self.newDeviceUi.oldDevice.currentText()
            oldDevice = self.hal.getDevice(oldDeviceName)
            self.newDeviceUi.deviceType.setCurrentText(oldDevice.deviceType)
            self.newDeviceUi.deviceType.setEnabled(False)
            calibrationList = self.hal.getCalibrationNames(oldDevice.deviceType)
            self.newDeviceUi.calibration.clear()
            for calibrationName in calibrationList:
                self.newDeviceUi.calibration.addItem(calibrationName)
            self.newDeviceUi.calibration.setCurrentText(oldDevice.calibration)
            self.newDeviceUi.calibration.setEnabled(False)
            self.newDeviceUi.apid.setValue(oldDevice.apid)
            self.newDeviceUi.port.setValue(oldDevice.port)
            self.newDeviceUi.daemon.setText("%s" % oldDevice.daemon)
            self.newDeviceUi.scienceDaemon.setText("%s" % oldDevice.scienceDaemon)
        else:
            self.newDeviceUi.oldDevice.setEnabled(False)
            self.newDeviceUi.deviceType.setEnabled(True)
            self.newDeviceUi.calibration.clear()
            deviceTypeName = self.newDeviceUi.deviceType.currentText()
            calibrationList = self.hal.getCalibrationNames(deviceTypeName)
            for calibrationName in calibrationList:
                self.newDeviceUi.calibration.addItem(calibrationName)
            self.newDeviceUi.calibration.setCurrentIndex(0)
            self.newDeviceUi.calibration.setEnabled(True)

    def slotNewDeviceType(self):
        """ Slot (main Gui): create a new device type. """

        # block the signals
        self.newDeviceTypeUi.createFrom.blockSignals(True)

        # initialization
        self.newDeviceTypeUi.newDeviceType.clear()
        self.newDeviceTypeUi.oldDeviceType.clear()
        self.newDeviceTypeUi.createFrom.setChecked(False)
        self.newDeviceTypeUi.oldDeviceType.setEnabled(False)
        self.newDeviceTypeUi.copyCal.setChecked(True)
        self.newDeviceTypeUi.copyCal.setEnabled(False)

        # fill in the 'oldDeviceType' combobox
        deviceTypeList = self.hal.getDeviceTypeNames()
        if deviceTypeList:
            deviceTypeList.sort()
            for deviceTypeName in deviceTypeList:
                self.newDeviceTypeUi.oldDeviceType.addItem(deviceTypeName)
            if self.deviceTypeName in deviceTypeList:
                self.newDeviceTypeUi.oldDeviceType.setCurrentText(self.deviceTypeName)
            else:
                self.newDeviceTypeUi.oldDeviceType.setCurrentIndex(0)
            self.newDeviceTypeUi.createFrom.setEnabled(True)
        else:
            self.newDeviceTypeUi.createFrom.setEnabled(False)

        # de-block the signals
        self.newDeviceTypeUi.createFrom.blockSignals(False)

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # execute the dialog
        accepted = self.newDeviceTypeUi.exec()
        if accepted:
            self.saveNewDeviceType()

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotUpdateNewDeviceType(self):
        """ Slot: Update the combobox in the newDeviceType menu. """

        if self.newDeviceTypeUi.createFrom.isChecked():
            self.newDeviceTypeUi.oldDeviceType.setEnabled(True)
            self.newDeviceTypeUi.copyCal.setEnabled(True)
        else:
            self.newDeviceTypeUi.oldDeviceType.setEnabled(False)
            self.newDeviceTypeUi.copyCal.setEnabled(False)

    def slotNewCalibration(self):
        """ Slot (main Gui): create a new calibration. """

        # block the signals
        self.newCalibrationUi.createFrom.blockSignals(True)
        self.newCalibrationUi.deviceType.blockSignals(True)
        self.newCalibrationUi.oldCalibration.blockSignals(True)

        # initialization
        self.newCalibrationUi.newCalibration.clear()
        self.newCalibrationUi.oldCalibration.clear()
        self.newCalibrationUi.deviceType.clear()
        self.newCalibrationUi.createFrom.setChecked(False)
        self.newCalibrationUi.oldCalibration.setEnabled(False)

        # fill in the deviceType combobox (always at least one deviceType present)
        deviceTypeList = self.hal.getDeviceTypeNames()
        deviceTypeList.sort()
        for deviceTypeName in deviceTypeList:
            self.newCalibrationUi.deviceType.addItem(deviceTypeName)
        if self.deviceTypeName in deviceTypeList:
            self.newCalibrationUi.deviceType.setCurrentText(self.deviceTypeName)
        else:
            self.newCalibrationUi.deviceType.setCurrentIndex(0)

        # fill in the 'oldCalibration' combobox
        deviceTypeName = self.newCalibrationUi.deviceType.currentText()
        calibrationList = self.hal.getCalibrationNames(deviceTypeName)
        if calibrationList:
            calibrationList.sort()
            for calibrationName in calibrationList:
                self.newCalibrationUi.oldCalibration.addItem(calibrationName)
            if self.calibrationName in calibrationList:
                self.newCalibrationUi.oldCalibration.setCurrentText(self.calibrationName)
            else:
                self.newCalibrationUi.oldCalibration.setCurrentIndex(0)
            self.newCalibrationUi.createFrom.setEnabled(True)
        else:
            self.newCalibrationUi.createFrom.setEnabled(False)

        # de-block the signals
        self.newCalibrationUi.createFrom.blockSignals(False)
        self.newCalibrationUi.deviceType.blockSignals(False)
        self.newCalibrationUi.oldCalibration.blockSignals(False)

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # execute the dialog
        accepted = self.newCalibrationUi.exec()
        if accepted:
            self.saveNewCalibration()

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotUpdateNewCalibration(self):
        """ Slot: Update the combobox in the newDevice menu. """

        # get the selected deviceType
        deviceTypeName = self.newCalibrationUi.deviceType.currentText()

        # fill in the 'oldCalibration' combobox
        self.newCalibrationUi.oldCalibration.clear()
        calibrationList = self.hal.getCalibrationNames(deviceTypeName)
        if calibrationList:
            calibrationList.sort()
            for calibrationName in calibrationList:
                self.newCalibrationUi.oldCalibration.addItem(calibrationName)
            if self.calibrationName in calibrationList:
                self.newCalibrationUi.oldCalibration.setCurrentText(self.calibrationName)
            else:
                self.newCalibrationUi.oldCalibration.setCurrentIndex(0)
            self.newCalibrationUi.createFrom.setEnabled(True)
        else:
            self.newCalibrationUi.createFrom.setEnabled(False)

        if self.newCalibrationUi.createFrom.isChecked():
            self.newCalibrationUi.oldCalibration.setEnabled(True)
        else:
            self.newCalibrationUi.oldCalibration.setEnabled(False)

    def slotEditDevice(self):
        """ Slot (main Gui): edit the selected device. """

        # block the signals
        self.editDeviceUi.oldName.blockSignals(True)
        self.editDeviceUi.deviceType.blockSignals(True)
        self.editDeviceUi.erase.blockSignals(True)

        # get the active device (selected in the main GUI)
        device = self.hal.getDevice(self.deviceName)

        # initialization
        self.editDeviceUi.erase.setEnabled(True)
        self.editDeviceUi.edit.setChecked(True)
        self.enableEditDeviceFields(True)

        # fill in the 'oldName' combobox and set current index to the active device
        # all available devices are filled in
        deviceList = self.hal.getDeviceNames()
        deviceList.sort()
        self.editDeviceUi.oldName.clear()
        for deviceName in deviceList:
            self.editDeviceUi.oldName.addItem(deviceName)
        self.editDeviceUi.oldName.setCurrentText(self.deviceName)

        # fill in the 'deviceType' combobox and set current index to the active deviceType
        # all available deviceTypes are filled in
        deviceTypeList = self.hal.getDeviceTypeNames()
        deviceTypeList.sort()
        self.editDeviceUi.deviceType.clear()
        for deviceTypeName in deviceTypeList:
            # only deviceTypes which have at least one calibration are considered
            if self.hal.getCalibrationNames(deviceTypeName):
                self.editDeviceUi.deviceType.addItem(deviceTypeName)
        self.editDeviceUi.deviceType.setCurrentText(self.deviceTypeName)

        # fill in the 'calibration' combobox and set current index to the active calibration
        # all available calibrations for the selected deviceType are filled in
        calibrationList = self.hal.getCalibrationNames(self.deviceTypeName)
        self.editDeviceUi.calibration.clear()
        for calibrationName in calibrationList:
            self.editDeviceUi.calibration.addItem(calibrationName)
        self.editDeviceUi.calibration.setCurrentText(self.calibrationName)

        # fill in the remaining edit fields
        self.editDeviceUi.newName.setText(self.deviceName)
        self.editDeviceUi.apid.setText(str(device.apid))
        self.editDeviceUi.port.setValue(device.port)
        self.editDeviceUi.daemon.setText(device.daemon)
        self.editDeviceUi.scienceDaemon.setText(device.scienceDaemon)

        # de-block the signals
        self.editDeviceUi.oldName.blockSignals(False)
        self.editDeviceUi.deviceType.blockSignals(False)
        self.editDeviceUi.erase.blockSignals(False)

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # execute the dialog
        accepted = self.editDeviceUi.exec()
        if accepted:
            self.saveEditDevice()

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotUpdateEditDevice1(self):
        """ Slot: Update edit device menu. OldName combobox changed. """

        # get the name of the selected device from the dialog combobox
        oldName = self.editDeviceUi.oldName.currentText()
        device = self.hal.getDevice(oldName)

        # update the 'deviceType' combobox
        self.editDeviceUi.deviceType.setCurrentText(device.deviceType)

        # update the 'calibration' combobox
        calibrationList = self.hal.getCalibrationNames(device.deviceType)
        self.editDeviceUi.calibration.clear()
        for calibrationName in calibrationList:
            self.editDeviceUi.calibration.addItem(calibrationName)
        self.editDeviceUi.calibration.setCurrentText(device.calibration)

        # update the remaining edit fields
        self.editDeviceUi.newName.setText(oldName)
        self.editDeviceUi.apid.setText(str(device.apid))
        self.editDeviceUi.port.setValue(device.port)
        self.editDeviceUi.daemon.setText(device.daemon)
        self.editDeviceUi.scienceDaemon.setText(device.scienceDaemon)

    def slotUpdateEditDevice2(self):
        """ Slot: Update edit device menu. DeviceType combobox changed. """

        # get the name of the selected deviceType from the dialog combobox
        deviceTypeName = self.editDeviceUi.deviceType.currentText()

        # update the 'calibration' combobox and set current index to 0
        # all available calibrations for the selected deviceType are filled in
        calibrationList = self.hal.getCalibrationNames(deviceTypeName)
        self.editDeviceUi.calibration.clear()
        for calibrationName in calibrationList:
            self.editDeviceUi.calibration.addItem(calibrationName)
        if self.calibrationName in calibrationList:
            self.editDeviceUi.calibration.setCurrentText(self.calibrationName)
        else:
            self.editDeviceUi.calibration.setCurrentIndex(0)

    def slotUpdateEditDevice3(self):
        """ Slot: Update edit device menu. Radio buttons toggled. """
        self.enableEditDeviceFields(not self.editDeviceUi.erase.isChecked())

    def slotEditDeviceType(self):
        """ Slot (main Gui): edit the selected deviceType. """

        # block the signals
        self.editDeviceTypeUi.oldName.blockSignals(True)
        self.editDeviceTypeUi.erase.blockSignals(True)

        # initialization
        self.editDeviceTypeUi.erase.setEnabled(True)
        self.editDeviceTypeUi.rename.setChecked(True)
        self.enableEditDeviceTypeFields(True)

        # fill in the 'oldName' combobox and set current index to the active deviceType
        deviceTypeList = self.hal.getDeviceTypeNames()
        deviceTypeList.sort()
        self.editDeviceTypeUi.oldName.clear()
        for item in deviceTypeList:
            self.editDeviceTypeUi.oldName.addItem(item)
        self.editDeviceTypeUi.oldName.setCurrentText(self.deviceTypeName)

        # fill in the remaining edit fields
        self.editDeviceTypeUi.newName.setText(self.deviceTypeName)

        # disable the 'delete' radio button if the selected deviceType is a member of a device
        deviceList = self.hal.getDevices(self.deviceTypeName)
        if deviceList:
            self.editDeviceTypeUi.erase.setEnabled(False)

        # disable the 'delete' radio button if the selected deviceType is a member of a calibration
        calibrationList = self.hal.getCalibrationNames(self.deviceTypeName)
        if calibrationList:
            self.editDeviceTypeUi.erase.setEnabled(False)

        # de-block the signals
        self.editDeviceTypeUi.oldName.blockSignals(False)
        self.editDeviceTypeUi.erase.blockSignals(False)

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # execute the dialog
        accepted = self.editDeviceTypeUi.exec()
        if accepted:
            self.saveEditDeviceType()

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotUpdateEditDeviceType1(self):
        """ Slot: Update edit deviceType menu. Oldname combobox changed. """

        # get the name of the selected deviceType from the dialog combobox
        oldName = self.editDeviceTypeUi.oldName.currentText()

        # update the edit fields
        self.editDeviceTypeUi.newName.setText(oldName)

        # default, the 'delete' ratio button is enabled
        self.editDeviceTypeUi.erase.setEnabled(True)

        # disable the 'delete' radio button if the selected deviceType is a member of a device
        deviceList = self.hal.getDevices(oldName)
        if deviceList:
            self.editDeviceTypeUi.erase.setEnabled(False)

        # disable the 'delete' radio button if the selected deviceType is a member of a calibration
        calibrationList = self.hal.getCalibrationNames(oldName)
        if calibrationList:
            self.editDeviceTypeUi.erase.setEnabled(False)

    def slotUpdateEditDeviceType2(self):
        """ Slot: Update edit deviceType menu. Radiobuttons toggled. """
        self.enableEditDeviceTypeFields(not self.editDeviceTypeUi.erase.isChecked())

    def slotEditCalibration(self):
        """ Slot (main Gui): edit the selected calibration. """

        # block the signals
        self.editCalibrationUi.deviceType.blockSignals(True)
        self.editCalibrationUi.oldName.blockSignals(True)
        self.editCalibrationUi.erase.blockSignals(True)

        # initialization
        self.editCalibrationUi.rename.setChecked(True)
        self.enableEditCalibrationFields(True)

        # fill in the 'deviceType' combobox with deviceTypes having at least one calibration
        deviceTypeList = self.hal.getDeviceTypeNames()
        deviceTypeList.sort()
        shortList = []
        self.editCalibrationUi.deviceType.clear()
        for deviceTypeName in deviceTypeList:
            calibrationList = self.hal.getCalibrationNames(deviceTypeName)
            if calibrationList:
                self.editCalibrationUi.deviceType.addItem(deviceTypeName)
                shortList.append(deviceTypeName)
        if self.deviceTypeName in shortList:
            self.editCalibrationUi.deviceType.setCurrentText(self.deviceTypeName)
        else:
            self.editCalibrationUi.deviceType.setCurrentIndex(0)

        # fill in the 'oldName' combobox
        deviceTypeName = self.editCalibrationUi.deviceType.currentText()
        calibrationList = self.hal.getCalibrationNames(deviceTypeName)
        calibrationList.sort()
        self.editCalibrationUi.oldName.clear()
        for calibrationName in calibrationList:
            self.editCalibrationUi.oldName.addItem(calibrationName)
        if self.calibrationName in calibrationList:
            self.editCalibrationUi.oldName.setCurrentText(self.calibrationName)
        else:
            self.editCalibrationUi.oldName.setCurrentIndex(0)

        # fill in the remaining edit fields
        calibrationName = self.editCalibrationUi.oldName.currentText()
        self.editCalibrationUi.newName.setText(calibrationName)

        # disable the 'delete' radio button if the selected calibration is a member of a device
        deviceList = self.hal.getDevices(deviceTypeName, calibrationName)
        self.editCalibrationUi.erase.setEnabled(len(deviceList) == 0)

        # de-block the signals
        self.editCalibrationUi.deviceType.blockSignals(False)
        self.editCalibrationUi.oldName.blockSignals(False)
        self.editCalibrationUi.erase.blockSignals(False)

        # store the current version of the HAL
        halSave = deepcopy(self.hal)

        # execute the dialog
        accepted = self.editCalibrationUi.exec()
        if accepted:
            self.saveEditCalibration()

        # if HAL has changed, push the previous version onto the stack
        if self.hal != halSave:
            self.stackPush(halSave)

        self.updateGui()

    def slotUpdateEditCalibration1(self):
        """ Slot: Update edit calibration menu. deviceType combobox changed. """

        # get the name of the selected deviceType from the dialog combobox
        deviceTypeName = self.editCalibrationUi.deviceType.currentText()

        # fill in the 'oldName' combobox
        calibrationList = self.hal.getCalibrationNames(deviceTypeName)
        calibrationList.sort()
        self.editCalibrationUi.oldName.clear()
        for calibrationName in calibrationList:
            self.editCalibrationUi.oldName.addItem(calibrationName)
        if self.calibrationName in calibrationList:
            self.editCalibrationUi.oldName.setCurrentText(self.calibrationName)
        else:
            self.editCalibrationUi.oldName.setCurrentIndex(0)
        calibrationName = self.editCalibrationUi.oldName.currentText()

        # fill in the remaining edit fields
        self.editCalibrationUi.newName.setText(calibrationName)

        # disable the 'delete' radio button if the selected calibration is a member of a device
        deviceList = self.hal.getDevices(deviceTypeName, calibrationName)
        self.editCalibrationUi.erase.setEnabled(len(deviceList) == 0)

    def slotUpdateEditCalibration2(self):
        """ Slot: Update edit calibration menu. oldName combobox changed. """

        # get the names of the selected deviceType and calibration from the dialog comboboxes
        deviceTypeName = self.editCalibrationUi.deviceType.currentText()
        calibrationName = self.editCalibrationUi.oldName.currentText()

        # update the edit fields
        self.editCalibrationUi.newName.setText(calibrationName)

        # disable the 'delete' radio button if the selected calibration is a member of a device
        deviceList = self.hal.getDevices(deviceTypeName, calibrationName)
        self.editCalibrationUi.erase.setEnabled(len(deviceList) == 0)

    def slotUpdateEditCalibration3(self):
        """ Slot: Update edit calibration menu. Radio buttons toggled. """
        self.enableEditCalibrationFields(not self.editCalibrationUi.erase.isChecked())

    def slotUndo(self):
        """ Slot (main GUI): called when 'Undo' is clicked. """
        # restore to most recent version of the hal
        self.hal = self.stackPull()
        self.updateGui()

    def slotSave(self):
        """ Slot: Saving HAL to YAML file. """
        hal = self.hal.createYamlData()
        self.halClient.setHal(hal)
        self.halClient.requestHal()
        # since self.halClient has been instantiated with argument self.halOriginal,
        # the output of the request operation ends up in self.halOriginal (see __init__)
        self.hal = deepcopy(self.halOriginal)
        self.halStack = []
        self.updateGui()

    def slotSaveClose(self):
        """ Save and close """
        logger.info("Save and close")
        hal = self.hal.createYamlData()
        self.halClient.setHal(hal)
        self.writeSettings()
        sys.exit(1)

    def slotRestore(self):
        """ Slot: Restore HAL from YAML file. """
        self.halClient.requestHal()
        # since self.halClient has been instantiated with argument self.halOriginal,
        # the output of the request operation ends up in self.halOriginal (see __init__)
        self.hal = deepcopy(self.halOriginal)
        self.halStack = []
        self.updateGui()

    @pyqtSlot()
    def slotRestoreToServerVersion(self):
        """ Slot: Confirm whether to discard local modifications. """
        if gsswUiYesNoQuestion("Do you want to remove any local modifications to the HAL database, "
                               "and revert to the centralized server version?"):
            self.halClient.revertHal()

    def slotReloadHal(self):
        """ Slot: A new HAL is available on the server. """
        logger.debug("Received hal update notification")
        if self.subSock.receiveProtoBufWithTopic().tag == 'NotifyHalUpdate':
            self.slotRestore()

    def slotSaveUploadToServer(self):
        """ Slot: Commit modified HAL YAML file to git (via HAL server). """
        self.saveUploadToServerDialog.runDialog()

    def slotClose(self):
        """ Slot (main GUI): called when 'Close' is clicked. """
        if self.hal != self.halOriginal:
            returnValue = gsswUiMessage('Do you want to save the HAL first?')
            if returnValue == QMessageBox.Ok:
                self.slotSave()
        self.writeSettings()
        self.newRegisterUi.close()
        self.newConstantUi.close()
        self.conversionsUi.close()
        self.limitsUi.close()
        self.newDeviceUi.close()
        self.newDeviceTypeUi.close()
        self.newCalibrationUi.close()
        self.editDeviceUi.close()
        self.editDeviceTypeUi.close()
        self.editCalibrationUi.close()
        self.gui.close()


###################################################################################################
# C L O S I N G
###################################################################################################

    def closeEvent(self, event):
        """ CLose the window """
        returnValue = gsswUiMessage('About to quit, are you sure?')

        if returnValue == QMessageBox.Ok:
            event.accept()
        else:
            event.ignore()

###################################################################################################
# EDITING HELPERS
###################################################################################################

class ComboBoxItemDelegate(QStyledItemDelegate):
    """ Delegate for in-table-comboboxes (dropdown list).
        By using this delegate, the table cell will change into a ComboBox when you click it.

    :note: Inspired by https://wiki.qt.io/Combo_Boxes_in_Item_Views
    """

    def __init__(self, parent, types):
        """ Create a editing combobox for the cells in one column.

        :param parent: Table on which this delegate operates
        :param types:  Entries of the ComboBox
        """
        super().__init__(parent)
        self.types  = types

    def createEditor(self, parent, option, index):  #pylint: disable=unused-argument
        """ Qt callback function, creates the combobox for the table cells in the configured column.
        """
        # Add ComboBox
        # Create the combobox and populate it
        cellComboBox = QComboBox(parent)
        cellComboBox.addItems(self.types)
        return cellComboBox

    def setEditorData(self, editor, index):
        """ Sets the selected option in the combobox when starting editing. """
        check_class(editor, QComboBox)
        # get the index of the text in the combobox that matches the current value of the item
        currentText = index.data(Qt.EditRole)
        cbIndex = editor.findText(currentText)
        # if it is valid, adjust the combobox
        if cbIndex >= 0:
            editor.setCurrentIndex(cbIndex)

    def setModelData(self, editor, model, index):
        """ Saves the selected option from the combobox to the table data. """
        check_class(editor, QComboBox)
        # save the current text of the combo box as the current value of the item
        model.setData(index, editor.currentText(), Qt.EditRole)


class ConstantLUTValuePopup(QWidget):
    """ Widget with popup editing dialog. """
    sigSave   = pyqtSignal()
    sigCancel = pyqtSignal()

    def __init__(self, parentCell):
        super().__init__(parentCell)

        self.originalValue = None
        self.newValue      = None
        self.accepted      = False

        # Provide a placeholder text that will be displayed in the cell (editing is done in popup)
        self.layout = QVBoxLayout()
        label = QLabel('<i>&nbsp;Editing...</i>')
        label.setStyleSheet("background: opaque;")
        self.layout.addWidget(label)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        # Make the popup (where editing is done)
        self.dialog = QInputDialog(self)
        self.dialog.setOption(QInputDialog.UsePlainTextEditForTextInput)
        self.dialog.setModal(True)

        self.dialog.accepted.connect(self.slotSave)
        self.dialog.rejected.connect(self.sigCancel)
        self.dialog.show()

    def setTitle(self, title):
        """ Set the popup dialog title. """
        self.dialog.setWindowTitle('Update {}'.format(title))

    def setLabel(self, label):
        """ Set the description in the popup. """
        self.dialog.setLabelText('Enter new value for {}:'.format(label))

    def setValue(self, value):
        """ Set the original text in the popup. """
        try:
            # Try displaying dicts in a user-readable way (not collapsed on one line)
            value = literal_eval(value)
            value = pprint.pformat(value, width=1)
        except (SyntaxError, ValueError):
            pass # No dict, plain string
        self.dialog.setTextValue(value)
        self.originalValue = value

    @pyqtSlot()
    def slotSave(self):
        """ Fired when OK button is clicked """
        self.accepted = True
        self.newValue = self.dialog.textValue()
        if self.newValue != self.originalValue:
            # Save the new value
            self.sigSave.emit()
        else:
            # Emit cancel, otherwise the delegate will not be removed
            self.sigCancel.emit()

class PopupItemDelegate(QStyledItemDelegate):
    """ Delegate for editing table cell values in a popup.
        By using this delegate, the table cell is blocked, and an editor pops up when you click it.
    """

    STRING_MINIMUM_LENGTH_POPUP = 15

    def __init__(self, parentTable):
        super().__init__(parentTable)
        self.popup = None

    def createEditor(self, parent, option, index):
        """ Create the popup editor (for lut and long strings only). """
        model = index.model()
        constantType = model.data(model.index(index.row(), 3))
        if constantType == 'lut' \
        or (constantType == 'string' and len(index.data()) > self.STRING_MINIMUM_LENGTH_POPUP):
            self.popup = ConstantLUTValuePopup(parent)
            self.popup.sigSave.connect(self.saveAndCloseEditor)
            self.popup.sigCancel.connect(self.cancelAndCloseEditor)
            return self.popup
        return super().createEditor(parent, option, index)

    def setEditorData(self, editor, index):
        """ Put the original values in the popup editor. """
        if isinstance(editor, ConstantLUTValuePopup):
            model = index.model()
            constantName = model.data(model.index(index.row(), 0))
            self.popup.setTitle(constantName)
            self.popup.setLabel(constantName)
            self.popup.setValue(index.data())
            self.popup.dialog.adjustSize()
        else:
            super().setEditorData(editor, index)

    def setModelData(self, editor, model, index):
        """ Save the new values from the editor into the model (table). """
        if isinstance(editor, ConstantLUTValuePopup):
            if self.popup and self.popup.accepted:
                model.setData(index, self.popup.newValue)

            # Close and remove popup
            del self.popup
            self.popup = None
        else:
            super().setModelData(editor, model, index)

    @pyqtSlot()
    def cancelAndCloseEditor(self):
        """ Stop editing """
        self.closeEditor.emit(self.popup)

    @pyqtSlot()
    def saveAndCloseEditor(self):
        """ Editing is done, emit save signal """
        self.commitData.emit(self.popup)


class EmptyConstantValueCell(QTableWidgetItem):
    """ Cell widget to show constants without a value. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.touched = False

    def data(self, role):
        """ Overrides the text shown when the cell is empty. """
        if not self.touched:
            if role == Qt.DisplayRole:
                # Value shown in cell
                return '-'
            elif role in (Qt.ToolTipRole, Qt.StatusTipRole):
                # Text shown in mouse tooltip, or window status bar
                return 'There is no value set for this constant.'
            elif role == Qt.WhatsThisRole:
                # Text shown in help message of the cell
                return "This constant doesn't have a value yet for this device, " \
                       "click to enter a value."
        # A value was set, show the real data
        return super().data(role)

    def setData(self, role, value):
        """ Clear the empty flag. """
        self.touched = True
        super().setData(role, value)

class DescriptionEditDelegate(QStyledItemDelegate):
    """ Delegate to handle autocompletion of a table cell """

    # pylint: disable=unused-argument
    def createEditor(self, parent, _styleOption, index):
        """ Create a multi-line editor for description field. """
        return QPlainTextEdit(parent)

    def updateEditorGeometry(self, editor, option, index):
        """ Expand editor by ~1 line when editing. """
        option.rect.setHeight(option.rect.height() + 30)    # add extra line in editor
        super().updateEditorGeometry(editor, option, index)
