""" GSSw Hardware Abstraction Layer Client """

###################################################################################################
# P Y T H O N   M O D U L E S
###################################################################################################
import logging
from threading                import Thread, Event

from gssw.proto.common_pb2    import PeerExceptionRep, OkRep
from gssw.proto.halserver_pb2 import GetHalConfig, SetHalConfig
from gssw.proto.halserver_pb2 import GetServerPublicKey, GetServerPublicKeyRep
from gssw.proto.halserver_pb2 import CommitHal, RevertHalReq
from gssw.proto.halserver_pb2 import CheckHalUploadAvailable, CheckHalUploadAvailableRep
from gssw.proto.halconfig_pb2 import HalConfig
from gssw.hal                 import HALSERVER_COMPONENT
from gssw.hal.parser          import HalParser
from gssw.lib.encryption      import PublicKey
from gssw.lib.rpcbase         import BaseClient
from gssw.lib.rpcmsg          import Packetizer
from gssw.lib.sockets         import SubSocket, ReqSocket, tagName
from gssw.lib.socketsqt       import NotifyingSubSocket
from gssw.lib.errors          import CommunicationError, GsswTimeoutError, PeerException
from PyQt5.QtWidgets          import QApplication


logger = logging.getLogger(__name__)
###################################################################################################
# H A L C L I E N T
###################################################################################################

class HalClient(BaseClient):
    """ Client object to be used by python scripts or daemons which
        require the hardware abstraction layer.
    """

    def __init__(self, hal, config):

        """ Constructor """
        self.hal     = hal
        self.config  = config

        # get the config parameters for this component
        hostname, repPort, _ = self.getServerAddress()

        # create a REQ socket
        self.reqSock = ReqSocket()
        self.reqSock.createZmqSocket()

        # connect the REQ socket to the hal server
        self.reqSock.connectSocket('tcp://%s:%d' % (hostname, repPort))

        # Create a HAL parser object
        self.parser = HalParser()

        super().__init__(reqSocket=self.reqSock)
        self.packetizer = Packetizer()

        #Init Callback Function Materials
        self.callbackFunc = None
        self.halUpdaterSubSocket = None
        self.updateHalThread = None
        self.threadStopEvent = Event()

    def __del__(self):
        """ Finalizer function to stop the HAL background thread
        in case the component did not unregister the Callback function."""
        self.unregisterCallBackFunction()
        
    def unregisterCallBackFunction(self):
        """ Stops HAL-background-updater thread """

        logger.info('Stopping HAL update thread')
        self.threadStopEvent.set() #Set flag to end Update Loop
        if self.updateHalThread:
            self.updateHalThread.join() # Wait for therad to complete
        self.updateHalThread = None

    def registerCallBackFunction(self,callbackFunc):
        """ Registers the CallBack Function which will be called if the HAL receives an Update.
            :param function callbackFunc: Function to be called if Hal update signal is received.
        """

        #Error Checking
        assert callable(callbackFunc), "Callback argument must be a function"

        #Check if Func is NonQt (True) or QT App (False)
        non_qt_check = QApplication.instance() is None 

        #Make (Notifying) SubSocket for HalClient PubServer, which will send a signal when updating
        hostname, _, pubPort = self.getServerAddress()
        halTcp = 'tcp://{}:{}'.format(hostname, pubPort)
        self.halUpdaterSubSocket = SubSocket() if non_qt_check else NotifyingSubSocket()
        self.halUpdaterSubSocket.createZmqSocket() 
        self.halUpdaterSubSocket.subscribe(topic="")
        self.halUpdaterSubSocket.connectSocket(halTcp)
        logger.debug('Connecting (%s App) to hal pub socket on %s',
                     "Non-Qt" if non_qt_check else "Qt", halTcp)

        self.callbackFunc = callbackFunc

        if non_qt_check: #Method for Non QT App (uses a thread) 
            #This part uses a new thread to check whether the SubSocket has received a message,
            #but will have to do this indefinitely since there is no signal like the PyQT signal
            # system used below. This thread will check the subSocket until it is actually ready
            # and run the callback function (after checking if it has the 'NotifyHalUpdate' tag)
            self.updateHalThread = Thread(target=self.halUpdaterCallBackStartThread,
                                     name='HAL-background-updater')
            self.updateHalThread.start()
            logger.info('Starting HAL update thread')

        else: #Method for Qt App (uses PyQT signal)
            #This part doesn't need a thread since the signal only fires when the SubSocket is ready
            # to receive and subsequently calls the "halUpdaterCallBackQT()" function. (after
            # checking if it has the 'NotifyHalUpdate' tag)
            self.halUpdaterSubSocket.sigReadyReceive.connect(self.halUpdaterCallBackQT)

    def halUpdaterCallBackQT(self):
        """ Function for QT Apps; will be called after the NotifyHalUpdate(PyQT) signal is received
        """
        if self.halUpdaterSubSocket.receiveProtoBufWithTopic().tag == 'NotifyHalUpdate':
            self.callbackFunc()

    def halUpdaterCallBackStartThread(self):
        """ Function for NonQT Apps; will be called after the NotifyHalUpdate signal is received,
            but will have to check for this manually (with a while True loop) and .readyReceive().
        """
        while not self.threadStopEvent.wait(2): #HAL Updating Loop: 2 secs before checking again
            # Check for package
            if self.halUpdaterSubSocket.readyReceive():
                if self.halUpdaterSubSocket.receiveProtoBufWithTopic().tag == 'NotifyHalUpdate':
                    self.callbackFunc()

    def getServerAddress(self):
        """ Retrieve the connection details of the HAL server. Can be used when creating sockets
        to the HAL server from outside the HalClient class, e.g. for making a subsocket.

        :return: tuple containing hostname, repPort, and pubPort. """
        halServerConfig = self.config.componentParameters(HALSERVER_COMPONENT)
        hostname = halServerConfig['hostname']
        repPort  = halServerConfig['sockets']['hal-rep']
        pubPort  = halServerConfig['sockets']['hal-pub']
        return hostname, repPort, pubPort

    def requestHal(self):
        """ Requests the HAL configuration as a protobuf message """
        request = GetHalConfig()
        protoString = request.SerializeToString()
        self.reqSock.sendProtoBuf(tagName(GetHalConfig), protoString, flags=0)

        try:
            rep = self.reqSock.receiveProtoBuf(flags=0)
            if rep.tag == 'HalConfig' :
                configPb = HalConfig.FromString(rep.message)
                self.parser.parseFromProtobuf(configPb)
                self.hal.setHalFromDict(self.parser.configData)
            else :
                raise CommunicationError('received unknown reply type: {}'.format(rep.tag))
        except GsswTimeoutError as ex:
            logger.error("Could not retrieve the HAL from the HAL server: %s", ex)
            raise GsswTimeoutError(
                "Could not contact the HAL server, check if it is running.") from None

    def setHal(self, hal):
        """ Set the HAL through a protobuffer request

            :param hal: The HAL as a Python dictionary
        """
        self.parser.parseFromConfigData(hal)

        request = SetHalConfig()
        request.config.CopyFrom(self.parser.protobuf)

        protoString = request.SerializeToString()
        self.reqSock.sendProtoBuf(tagName(SetHalConfig), protoString, flags=0)

        # Saving can be quite slow, so increase timeout
        oldTimeout = self.reqSock.zmqSocket.RCVTIMEO
        self.reqSock.setReceiveTimeout(-1)
        rep = self.reqSock.receiveProtoBuf(flags=0)
        self.reqSock.setReceiveTimeout(oldTimeout)
        if rep.tag != tagName(OkRep):
            # message does not have the correct syntax
            raise CommunicationError('received unknown reply type: {}'.format(rep.tag))

    def commit(self, username, password, comment):
        """
         Commit the updated HAL file to the versioning system (i.e. git).

        :param username: Username for versioning system (will be sent to hal server encrypted).
        :param password: Password for versioning system (will be sent to hal server encrypted).
        :param comment:  Message with reason or description for the commit.
        """

        # Commit + push can be quite slow, so increase timeout
        oldTimeout = self.reqSock.getReceiveTimeout()
        self.reqSock.setReceiveTimeout(-1)

        try:
            # We will encrypt the username & password using public key encryption
            #    First obtain the public key from the server
            pubkeyReq = GetServerPublicKey()
            self.reqSock.sendProtoBuf(tagName(pubkeyReq), pubkeyReq.SerializeToString())

            rep = self.reqSock.receiveProtoBuf()
            pubkeyRep = GetServerPublicKeyRep()
            if rep.tag != tagName(pubkeyRep):
                raise CommunicationError('Received unknown reply type while receiving encryption '
                                         'key: {}'.format(rep.tag))

            pubkeyRep.ParseFromString(rep.message)
            publicKey = PublicKey.deserialize(pubkeyRep.public_key)

            # Now encrypt the credentials, and do the commit request
            username_encrypted = publicKey.encrypt(username.encode('utf-8'))
            password_encrypted = publicKey.encrypt(password.encode('utf-8'))

            commitReq = CommitHal()
            commitReq.username_encrypted = username_encrypted
            commitReq.password_encrypted = password_encrypted
            commitReq.comment            = comment

            self.reqSock.sendProtoBuf(tagName(commitReq), commitReq.SerializeToString())

            rep = self.reqSock.receiveProtoBuf()

            # Contains a PeerExceptionRep protocol buffer.
            if rep.tag == tagName(PeerExceptionRep):
                raise PeerException(rep.message)

            if rep.tag != tagName(OkRep):
                raise CommunicationError(
                    'Received unknown reply type while performing commit: {}'.format(rep.tag))
        finally:
            # Try to clean credentials from memory
            del username
            del password

            self.reqSock.setReceiveTimeout(oldTimeout)

    def checkHalUploadAvailable(self):
        """ Ask the Hal server whether there are changes which can be uploaded. """
        reqMsg = CheckHalUploadAvailable()
        self.reqSock.sendProtoBuf(tagName(reqMsg), reqMsg.SerializeToString())
        rep = self.reqSock.receiveProtoBuf()

        repMsg = CheckHalUploadAvailableRep()
        if rep.tag != tagName(repMsg):
            raise CommunicationError(
                'Received unknown reply type while checking upload status: {}'.format(rep.tag))
        repMsg.ParseFromString(rep.message)
        return bool(repMsg.upload_available)

    def revertHal(self):
        """ Discard local changes (can already be saved), and revert the HAL to the 
            centralized version.

            :note: You can check whether this option is applicable by calling
                   :meth:`checkHalUploadAvailable`."""

        replyMsg = self.requestReply(RevertHalReq(), OkRep())
        return self.packetizer.getMessageAttrs(replyMsg)
