#!/usr/bin/env python3
#pylint: disable=too-many-return-statements
"""
Module for importing a list of registers into the HAL from an Excel file.
The devicetype can be specified.

Note: Registers which already exist in the HAL are ignored.
"""

import logging
import os.path
import openpyxl


from gssw.common        import initGssw
from gssw.lib.misc      import createGsswArgumentParser
from gssw.lib.typing    import check_class, check_is_a_sequence
from gssw.hal.hal       import Hal, Register
from gssw.hal.halclient import HalClient
from gssw.lib.sockets   import destroyAllSocketsInContext
from gssw.lib.errors    import DataError


logger = logging.getLogger(__name__)

__all__ = ['HalImporter']

class HalImporter:
    """ Class for importing an excel sheet into a devicetype """

    def __init__(self, hal):
        check_class(hal, Hal)

        self.hal = hal
        self._workbook = None

        self._columnIndexes = {'name':          None,
                               'address':       None,
                               'type':          None,
                               'unit':          None,
                               'access':        None,
                               'width':         None,
                               'block':         None,
                               'offset':        None,
                               'description':   None}

    def open(self, filename):
        """
        Opens an excel file for reading (read-only)

        :param str filename: name of excel file
        """
        assert os.path.exists(filename), "File '{}' does not exist".format(filename)

        try:
            self._workbook = openpyxl.load_workbook(filename, read_only=True)
            return True
        except Exception:
            logger.error("Cannot open excel file '%s'", filename, exc_info=True)
        return False


    def __safeStrip(self, value):
        """ Strips whitespace of strings, non-string objects are left untouched """
        if isinstance(value, str):
            return value.strip()
        return value


    def _parseRegister(self, row):  #pylint: disable=too-many-locals
        """ Parses an excel sheet row as a HAL Register """
        check_is_a_sequence(row)

        # Read register properties from row cells
        registerNameRaw        = row[self._columnIndexes['name']].value
        registerAddressRaw     = row[self._columnIndexes['address']].value
        registerTypeRaw        = row[self._columnIndexes['type']].value
        registerUnitRaw        = row[self._columnIndexes['unit']].value
        registerAccessRaw      = row[self._columnIndexes['access']].value
        registerWidthRaw       = row[self._columnIndexes['width']].value
        registerBlockRaw       = row[self._columnIndexes['block']].value
        registerDescriptionRaw = row[self._columnIndexes['description']].value

        # Trim off any surrounding whitespace in the cells' text
        registerName        = self.__safeStrip(registerNameRaw)
        registerAddress     = self.__safeStrip(registerAddressRaw)
        registerType        = self.__safeStrip(registerTypeRaw)
        registerUnit        = self.__safeStrip(registerUnitRaw)
        registerAccess      = self.__safeStrip(registerAccessRaw)
        registerWidth       = self.__safeStrip(registerWidthRaw)
        registerBlock       = self.__safeStrip(registerBlockRaw)
        registerDescription = self.__safeStrip(registerDescriptionRaw)

        if registerName and registerWidth != '#':
            # Defaults for properties not present in file
            if not registerType:
                registerType = 'uint' # Default to uint when the column is empty
            elif registerType not in ['int', 'uint', 'float']:
                logger.error("Register type value '%s' for register '%s' is not recognized",
                             registerTypeRaw,
                             registerName)
                return None

            # The blockdata column can be set to Yes/No
            if isinstance(registerBlock, str) and registerBlock.lower()=="no":
                registerBlock = False
            else:
                registerBlock = bool(registerBlock)
            registerMaxOffset   = 0

            # Interpret properties
            registerAccess = Register.stringToAccess(registerAccess)
            if registerAccess is None:
                logger.error("Register read/write value '%s' for register '%s' is not recognized",
                             registerAccessRaw,
                             registerName)
                return None

            try:
                registerAddress = int(registerAddress, 0)
            except ValueError:
                logger.error("Register address '%s' for register '%s' is not a valid",
                             registerAddressRaw,
                             registerName)
                return None

            if registerWidth is None:
                logger.error("Register width value '%s' for register '%s' is not recognized",
                             registerWidthRaw,
                             registerName)
                return None
            try:
                registerWidth = int(registerWidth)
            except ValueError:
                logger.error("Register width '%s' for register '%s' is not a valid",
                             registerWidthRaw,
                             registerName)
                return None

            # Convert empty cells to empty strings
            if registerUnit is None:
                registerUnit = ''
            if registerDescription is None:
                registerDescription = ''

            # Create register from parsed properties
            return Register(registerName,
                            registerAddress,
                            registerAccess,
                            registerType,
                            registerWidth,
                            registerUnit,
                            registerBlock,
                            registerMaxOffset,
                            registerDescription)
        return None


    def _parseHeader(self, row):
        """ Finds the columns in which the relevant register information is located """
        check_is_a_sequence(row)

        # Parse header
        for columnIndex, cell in enumerate(row):
            headerName = cell.value
            if headerName:
                if isinstance(headerName, str):
                    headerName = headerName.lower()

                # Find columns in header
                if   headerName=='db name':
                    self._columnIndexes['name']=columnIndex
                elif headerName=='address':
                    self._columnIndexes['address']=columnIndex
                elif headerName=='type':
                    self._columnIndexes['type']=columnIndex
                elif headerName=='unit':
                    self._columnIndexes['unit']=columnIndex
                elif headerName=='r/w':
                    self._columnIndexes['access']=columnIndex
                elif headerName=='bits':
                    self._columnIndexes['width']=columnIndex
                elif headerName=='blockdata':
                    self._columnIndexes['block']=columnIndex
                elif headerName=='notes':
                    self._columnIndexes['description']=columnIndex
                else:
                    logger.debug("Column '%s' is ignored", cell.value)


    def _addRegister(self, register, deviceTypeName):
        """ Adds a (previously parsed) register to the hal devicetype """
        check_class(register, Register)
        check_class(deviceTypeName, str)

        try:
            deviceType = self.hal.getDeviceType(deviceTypeName)
        except DataError:
            logger.error("Device type '%s' does not exist", deviceTypeName, exc_info=True)
            return False

        # Check if the register is not already present (by name or address)
        if deviceType.registerExists(register.name):
            logger.error("Skipping register with duplicate name: %s", register.name)
            return False
        if deviceType.registerExists(register.address):
            logger.error("Skipping register with duplicate address: %s for register '%s'",
                         hex(register.address), register.name)
            return False

        # Everything seems OK, add register to HAL
        deviceType.addRegister(register)
        logger.info("Added register '%s' to HAL", register.name)
        return True


    def loadRegistersIntoDeviceType(self, sheetName, deviceTypeName):
        """
        Starts the process of importing the registers from the excel file into the HAL

        :param HAL hal: HAL to which the imported registers are saved
        :param str sheetName: Name of the excel workbook sheet to import
        :param str deviceTypeName: Name of the device type into which the registers are imported 
        """
        check_class(sheetName, str)
        check_class(deviceTypeName, str)

        if self._workbook:
            logger.info("Loading registers for HAL device: %s", deviceTypeName)

            registerSheet = self._workbook[sheetName]

            nrOfRegisters = 0
            for rowIndex, row in enumerate(registerSheet):
                if rowIndex == 0:
                    self._parseHeader(row)
                else:
                    register = self._parseRegister(row)
                    if register:
                        if self._addRegister(register, deviceTypeName):
                            nrOfRegisters += 1
            logger.info("Imported %s registers from file into HAL", nrOfRegisters)
            return True
        return False


    def deleteAllRegisters(self, deviceTypeName):
        """
         Deletes all registers in the HAL for a devicetype

        :param str deviceTypeName: name of the devicetype
        """
        check_class(deviceTypeName, str)

        deviceType = self.hal.getDeviceType(deviceName = deviceTypeName)
        if deviceType:
            deviceRegistersDict = self.hal.getRegistersInfo(deviceTypeName)
            if deviceRegistersDict:
                deviceRegisters = list(deviceRegistersDict.keys())
                for register in deviceRegisters:
                    deviceType.removeRegister(register)
            else:
                logger.info("Devicetype '%s' does not contain any register to be deleted",
                            deviceTypeName)
        else:
            raise AttributeError("Devicetype {} was not found in the HAL".format(deviceTypeName))


def main():
    """ Reads an excel file and imports it into the HAL """
    deviceTypeName  = "dualfee"
    sheetName       = 'Registers'


    parser = createGsswArgumentParser(description=__doc__)

    parser.add_argument('filename', type=str,
                help='Name of the excel file to import ')
    parser.add_argument('--delete', action='store_true',
                help='Delete all registers from device before importing the file. WARNING: This '
                     'will delete all calibrations for the all boards of the given device type')
    parser.add_argument('devicetypename', nargs='?', type=str,
                help='Name of the device type for which to create the registers '
                     '(default: %(default)s)',
                default=deviceTypeName)
    parser.add_argument('sheetname', nargs='?', type=str,
                help='Name of the sheet in the excel file which contains the registers '
                     '(default: %(default)s)',
                default=sheetName)
    args = parser.parse_args()
    config = initGssw(args)

    hal              = Hal()
    halClient        = HalClient(hal, config)
    halClient.requestHal()

    halImporter = HalImporter(hal)
    if args.delete:
        halImporter.deleteAllRegisters(args.devicetypename)
        halYaml = hal.createYamlData()
        halClient.setHal(halYaml)
        print ("Deleted all registers for device: {}".format(args.devicetypename))
    if halImporter.open(args.filename):
        print ("Loading registers from file")
        if halImporter.loadRegistersIntoDeviceType(args.sheetname, args.devicetypename):
            halYaml = hal.createYamlData()
            halClient.setHal(halYaml)
            print ("Saved new HAL, please restart GSSW...")
        else:
            print("Importing the registers failed")
    else:
        print ("Loading aborted")


if __name__ == '__main__':
    try:
        main()
    finally:
        destroyAllSocketsInContext()
