#!/usr/bin/env python3

""" Parses HAL from file into protobuffers and vice versa
"""

import logging
from   operator import itemgetter   # Used for sorting HAL items

from gssw.proto.halconfig_pb2 import HalConfig, Register
from gssw.lib.errors import CommunicationError

logger = logging.getLogger(__name__)


###################################################################################################
# H A L  P A R S E R
###################################################################################################

class HalParser:
    """ Class to parse HAL configuration files into protobuffer messages and vice versa.
        Internally it parses either a YAML or JSON file into a dictionary.
        This dictionary is translated into a protobuf message.
        When using a protobuf message as the input, the dictionary is generated from
        the protobuf message.
        The result can be a protobuf message  or a YAML/JSON file
    """
    def __init__(self):
        """ Constructs internal representations of HAL
        """
        self._halConfigPb = HalConfig()
        self._configData = {'hal': {
                'devices': [{}],'calibrations': [{}], 'deviceTypes': [{}], 'versionNumber': ''}}
        self._initialized = False

    @property
    def protobuf(self):
        """ Get HAL configuration as a protobuf message
        """
        if not self._initialized:
            logger.warning('Returning protobuffer while HAL is not initialized')

        return self._halConfigPb

    @property
    def configData(self):
        """ Get HAL configuration as a dictionary
        """
        if not self._initialized:
            logger.critical('Returning dictionary while HAL is not initialized')
            raise CommunicationError('HAL not initialized')

        return self._configData

    def parseFromConfigData(self, configData):
        """ To set the dictionary directly

            :param configData: Python dictionary with HAL configuration
        """
        self._configData = configData
        self._fillProtobuf()
        self._initialized = True

    def parseFromProtobuf(self, halConfigPb):
        """ Set the protobuf message and translate in internal
            dictionary.

            :note: The HAL items are sorted alphabetically. This prevents the (exported) yaml file
                   to be written in a totally different order after a saving. Succesively, this
                   minimized the changes in the saved yaml file and allows for a meaningful
                   comparision (diff) between two different yaml files.

            :param halConfigPb: The HAL configuration as protobuffer
        """

        logger.info('Parsing protobuffer')

        # Set internal values
        self._halConfigPb = halConfigPb
        self._configData = {}
        self._initialized = True

        # Create empty lists
        self._configData['hal']                  = {}
        self._configData['hal']['versionNumber'] = self._halConfigPb.version_number
        self._configData['hal']['devices']       = []
        self._configData['hal']['calibrations']  = []
        self._configData['hal']['deviceTypes']   = []

        # Note: All the sorting of the HAL items is required since the protobuf uses lots of map
        #       fields (e.g. for storing the register list), which do not guarantee ordering.
        for deviceName, devicePb in sorted(self._halConfigPb.devices.items(), key=itemgetter(0)):
            # Fill 'devices' list, sorted by device name
            logger.debug('Parsing device %s', deviceName)

            self._configData['hal']['devices'].append(
                self._parseDevicePb(deviceName, devicePb))

        for deviceTypeName, deviceTypePb in sorted(self._halConfigPb.device_types.items(),
                                                   key=itemgetter(0)):

            logger.debug('Parsing device type %s', deviceTypeName)

            # Fill 'deviceTypes' list, sorted by device type name
            registers = self._parseRegistersPb(deviceTypePb.registers)
            constants = self._parseConstantsPb(deviceTypePb.constants)

            self._configData['hal']['deviceTypes'].append(
                {  'name'      : deviceTypeName
                 , 'constants' : constants
                 , 'registers' : registers
                })

            # Fill 'calibrations' list, one device type can have multiple calibrations
            #   sort list by calibration name
            for cName, calibrationPb in sorted(deviceTypePb.calibrations.items(),
                                               key=itemgetter(0)):

                logger.debug('Got calibration %s for deviceType %s'
                    , cName, deviceTypeName)

                conversions = self._parseConversionsPb(calibrationPb.conversions)
                calibrationConstants = self._parseCalibrationConstantsPb(
                        calibrationPb.constant_values)

                self._configData['hal']['calibrations'].append (
                    {  'deviceType'  : deviceTypeName
                     , 'name'        : cName
                     , 'constants'   : calibrationConstants 
                     , 'conversions' : conversions
                    }
                )

    def _parseDevicePb(self, deviceName, devicePb):
        """ Parse a device protobuffer message and return it as a dictionary
        """ 

        deviceEntry = {  'name'        : deviceName
                       , 'apid'        : devicePb.apid 
                       , 'calibration' : devicePb.calibration
                       , 'daemon'      : devicePb.daemon
                       , 'deviceType'  : devicePb.device_type
                       , 'port'        : devicePb.port
                      }
        if devicePb.scienceDaemon:
            # omit the scienceDaemon key if it is not used
            deviceEntry['scienceDaemon'] = devicePb.scienceDaemon   

        return deviceEntry
    
    def _parseConversionsPb(self, conversionsPb):
        """ Build up a list of conversions from a protobuf message
        """
        conversions = []
        # Fill the list of conversions, sorted by register name
        for registerName, conversionPb in sorted(conversionsPb.items(), key=itemgetter(0)):

            conDict = {   'name'          : registerName
                        , 'setfunction'   : conversionPb.set_function
                        , 'getfunction'   : conversionPb.get_function
                        , 'setlut'        : dict(conversionPb.set_lut.lut_map)
                        , 'getlut'        : dict(conversionPb.get_lut.lut_map)
                        , 'setpolynome'   : conversionPb.set_polynome.coefficients[:]
                        , 'getpolynome'   : conversionPb.get_polynome.coefficients[:]
                        , 'setconversion' : conversionPb.set_conversion
                        , 'getconversion' : conversionPb.get_conversion
                      }
            if conversionPb.limits is not None:
                limits = {  'lowerWarning' : conversionPb.limits.lower_warning
                          , 'upperWarning' : conversionPb.limits.upper_warning
                          , 'upperError'   : conversionPb.limits.upper_error
                          , 'lowerError'   : conversionPb.limits.lower_error
                          , 'upperValid'   : conversionPb.limits.upper_valid
                          , 'lowerValid'   : conversionPb.limits.lower_valid
                         }
                conDict['limits'] = limits
            conversions.append(conDict)

        return conversions

    def _parseCalibrationConstantsPb(self, constantValuesPb):
        """ Build up a dict of constant names and assigned values
        """
        calibrationConstants = {}

        # Fill the list of constant values, sorted by constant name
        for constName, constValPb in sorted(constantValuesPb.items(), key=itemgetter(0)):
            setVal = constValPb.WhichOneof('value')

            if setVal == 'lut_value':
                calibrationConstants[constName] = self._parseCalibrationConstantLutPb(
                    constValPb.lut_value.items)
            else:
                calibrationConstants[constName] = getattr(constValPb, setVal)

        return calibrationConstants

    def _parseCalibrationConstantLutPb(self, calibrationConstantLutValuePb):
        """ Returns a dictionary of parsed constant lut value protobuf message. """
        lutValueDict = {} 
        for lutValueItemPb in calibrationConstantLutValuePb:
            key    = getattr(lutValueItemPb, lutValueItemPb.WhichOneof('key'))
            value  = getattr(lutValueItemPb, lutValueItemPb.WhichOneof('value'))

            lutValueDict[key] = value
        return lutValueDict
    
    def _parseRegistersPb(self, registersPb):
        """ Returns a dictionary of the registers parsed from a protobuffer
        """
        registers = []
        # Fill the list of reigsters, sorted by register name
        for registerName, registerPb in sorted(registersPb.items(), key=itemgetter(0)):

            registers.append(
                        {  'name'        : registerName
                         , 'block'       : registerPb.block
                         , 'type'        : self._registerTypeToString(registerPb.type)
                         , 'access'      : registerPb.access
                         , 'unit'        : registerPb.unit
                         , 'address'     : registerPb.address
                         , 'width'       : registerPb.width
                         , 'description' : registerPb.description
                         , 'maxOffset'   : registerPb.max_offset
                        }
                    )
        return registers 
    
    def _parseConstantsPb(self, constantsPb):
        """ Returns a dictionary of the constants parsed from a protobuffer
        """

        constants = []
        # Fill the list of constants, sorted by constant name
        for constantName, constantPb in sorted(constantsPb.items(), key=itemgetter(0)):

            constants.append(
                        {  'name' : constantName
                         , 'unit' : constantPb.unit
                         , 'type' : constantPb.type
                         , 'description' : constantPb.description
                        }
                    )
        return constants

    def _fillProtobuf(self) :
        """ Creates protobuffer from dictionary.

            :note: Converts daemon names, calibrations, device types and devices to lower case,
                   converts register names and constant names to upper case.
        """
        self._halConfigPb.Clear()

        if self._configData is not None:
            self._halConfigPb.version_number = self._configData['hal']['versionNumber']
            if self._configData['hal']['devices'] is not None:
                for dev in self._configData['hal']['devices']:
                    # Create the device
                    logger.debug('Read device %s', dev['name'])

                    name = dev['name'].lower()

                    # Create a new devices map entry
                    devicePb = self._halConfigPb.devices[name]

                    devicePb.apid        = dev['apid']
                    devicePb.daemon      = dev['daemon'].lower()
                    if 'scienceDaemon' in dev and dev['scienceDaemon']:
                        # scienceDaemon can be None, in which case it is omitted from the protobuf
                        devicePb.scienceDaemon = dev['scienceDaemon'].lower()  
                    devicePb.port        = dev['port']
                    devicePb.calibration = dev['calibration'].lower()
                    devicePb.device_type = dev['deviceType'].lower()

            if self._configData['hal']['deviceTypes'] is not None:
                for devType in self._configData['hal']['deviceTypes']:
                    devTypeName = devType['name'].lower()

                    logger.debug('Read device type %s', devTypeName)

                    deviceTypePb = self._halConfigPb.device_types[devTypeName]

                    self._fillDeviceType(  deviceTypePb, devTypeName, devType)

    def _fillDeviceType(self, deviceTypePb, deviceTypeName, deviceTypeConfig):
        self._fillRegisters(deviceTypePb, deviceTypeConfig)
        self._fillDeviceTypeConstants(deviceTypePb, deviceTypeConfig)
        self._fillDeviceTypeCalibrations(deviceTypePb, deviceTypeName)

    def _fillRegisters(self, deviceTypePb, deviceTypeConfig):
        if deviceTypeConfig['registers'] is not None:
            for register in deviceTypeConfig['registers']:
                registerPb = deviceTypePb.registers[register['name'].upper() ]

                registerPb.block           = register['block']
                registerPb.access          = register['access']
                try:
                    registerPb.unit        = register['unit']
                except ValueError:
                    registerPb.unit        = '' # Some weird binary in some YAML files
                    logger.warning('Can not read unit for register %s', register['name'])

                registerPb.address         = register['address']
                registerPb.width           = register['width']
                try:
                    registerPb.description = register['description']
                except ValueError:
                    registerPb.description = '' # Some weird binary in some YAML files
                    logger.warning('Can not read description for register %s', register['name'])

                registerPb.max_offset  = register['maxOffset']
                registerPb.type        = self._stringToRegisterType(register['type'])

    def _fillDeviceTypeConstants(self, deviceTypePb, deviceTypeConfig):
        if deviceTypeConfig['constants'] is not None:
            for constant in deviceTypeConfig['constants']:
                cName = constant['name'].upper()

                constantPb = deviceTypePb.constants[cName]

                constantPb.name        = cName
                constantPb.unit        = constant['unit']
                constantPb.type        = constant['type']
                constantPb.description = constant['description']
            # Value is set in calibration

    def _fillDeviceTypeCalibrations(self, deviceTypePb, deviceTypeName):
        if self._configData['hal']['calibrations'] is not None:
            for calibration in self._configData['hal']['calibrations']:
                if calibration['deviceType'].lower() == deviceTypeName:
                    cName = calibration['name'].lower()
                    # Add a calibration
                    deviceTypePb.calibrations[cName].name = cName
                    self._fillConversions(calibration, deviceTypePb.calibrations[cName])
                    self._fillCalibrationConstants(calibration, deviceTypePb.calibrations[cName])

    def _fillConversions(self, configDataCalibration, calibrationPb):
        if configDataCalibration['conversions'] is not None:
            for conv in configDataCalibration['conversions']:
                # Conversion Protobuffer for this register name
                convPb = calibrationPb.conversions[conv['name'].upper()]
                if 'getlut' in conv and conv['getlut'] is not None:
                    for (key,value) in conv['getlut'].items():
                        convPb.get_lut.lut_map[key] = value
                if 'setlut' in conv  and conv['setlut'] is not None:
                    for (key,value) in conv['setlut'].items():
                        convPb.set_lut.lut_map[key] = value
                if 'getfunction' in conv and conv['getfunction'] is not None:
                    convPb.get_function = conv['getfunction']
                if 'setfunction' in conv and conv['setfunction'] is not None:
                    convPb.set_function = conv['setfunction']
                if 'getpolynome' in conv and conv['getpolynome'] is not None:
                    convPb.get_polynome.coefficients[:] = conv['getpolynome']
                if 'setpolynome' in conv and conv['setpolynome'] is not None:
                    convPb.set_polynome.coefficients[:] = conv['setpolynome']
                if 'getconversion' in conv and conv['getconversion'] is not None:
                    convPb.get_conversion = conv['getconversion']
                if 'setconversion' in conv and conv['setconversion'] is not None:
                    convPb.set_conversion = conv['setconversion']
                if 'limits' in conv and conv['limits' ] is not None:
                    convPb.limits.lower_warning = conv['limits']['lowerWarning']
                    convPb.limits.upper_warning = conv['limits']['upperWarning']
                    convPb.limits.lower_error   = conv['limits']['lowerError']
                    convPb.limits.upper_error   = conv['limits']['upperError']
                    convPb.limits.lower_valid   = conv['limits']['lowerValid']
                    convPb.limits.upper_valid   = conv['limits']['upperValid']

    def _fillCalibrationConstants(self, configDataCalibration, calibrationPb):
        """ Convert calibration constants to protobuffer format
        """
        if configDataCalibration['constants'] is not None:
            for (key, value) in configDataCalibration['constants'].items():
                newConstName = key.upper()
                if isinstance(value, float):
                    calibrationPb.constant_values[newConstName].double_value = value
                elif isinstance(value, int):
                    if value >= 0:
                        calibrationPb.constant_values[newConstName].uint_value = value
                    else:
                        calibrationPb.constant_values[newConstName].int_value = value
                elif isinstance(value, dict):
                    # Make sure Pb repeated field is set (in case of empty lut dict)
                    calibrationPb.constant_values[newConstName].lut_value.Clear()

                    # Now put the dict items in the Pb
                    self._fillCalibrationConstantLut(
                        value, calibrationPb.constant_values[newConstName].lut_value.items)
                else :
                    calibrationPb.constant_values[newConstName].string_value = value

    def _fillCalibrationConstantLut(self, constantLut, calibrationConstantLutValuePb):
        """ Converts a calibration constant LUT value (which is a dict) into protobuf format
        """
        for (key, value) in constantLut.items():
            pbLut = calibrationConstantLutValuePb.add()
            if isinstance(key, float):
                pbLut.double_key     = key
            elif isinstance(key, int):
                if key >= 0:
                    pbLut.uint_key   = key
                else:
                    pbLut.int_key    = key
            else:
                pbLut.string_key     = str(key)

            if isinstance(value, float):
                pbLut.double_value   = value
            elif isinstance(value, int):
                if value >= 0:
                    pbLut.uint_value = value
                else:
                    pbLut.int_value  = value
            else:
                pbLut.string_value   = str(value)

    def _registerTypeToString(self, registerType):
        """ Convert a register type enum representation to a string
        """
        result = 'unknown'

        if registerType == Register.TYPE_INT:
            result = 'int'
        elif registerType == Register.TYPE_UINT:
            result = 'uint'
        elif registerType == Register.TYPE_FLOAT:
            result = 'float'

        return result

    def _stringToRegisterType(self, regTypeString):
        """ Convert a register type string representation to an enum
        """
        result = Register.TYPE_UNKNOWN

        if regTypeString == 'uint':
            result = Register.TYPE_UINT
        elif regTypeString == 'int':
            result = Register.TYPE_INT
        elif regTypeString == 'float':
            result = Register.TYPE_FLOAT

        return result
