#!/usr/bin/env python3
"""
Module for duplicating related registers in the kids HAL.
The purpose is that only the rflogicboard miscellaneous, osc1 and conv2 registers + conversions
need to be filled in the kids HAL using the HAL manager. The other devices (osc3, osc5, osc7, conv4,
conv6 and conv8) are filled in by the script. The script also copies all rflogicboard related
registers + conversions (miscellaneous, osc1/3/5/7 and conv2/4/6/8) into the adc address space.
First all rflogicboard osc3/5/7 and conv4/6/8 registers + conversions and all adc registers +
conversions related to the rflogicboard (i.e. miscellaneous, osc1/3/5/7 and conv2/4/6/8) are
deleted. After this, the rflogicboard osc1 and conv2 registers + conversions are copied into
osc3/5/7 and conv4/6/8 respectively (rflogicboard address space). Finally, the rflogicboard
miscellaneous, osc1/3/5/7 and conv2/4/6/8 registers are copied into the adc address space.

Note: the HAL server needs to be running when this script is run!
"""

import logging

from copy               import deepcopy
from gssw.common        import initGssw
from gssw.lib.misc      import createGsswArgumentParser
from gssw.lib.typing    import check_class
from gssw.hal.hal       import Hal
from gssw.hal.halclient import HalClient
from gssw.lib.sockets   import destroyAllSocketsInContext


logger = logging.getLogger(__name__)

__all__ = ['HalUpdater']

class HalUpdater:
    """ Class for updating the kids HAL with RF logicboard registers """

    def __init__(self, hal):
        check_class(hal, Hal)

        self.hal = hal
        self.baseAddress = 0x00350000


    def deleteRegisters(self, deviceTypeName): # pylint: disable=too-many-locals
        """
         Delete all osc3/5/7 and conv4/6/8 registers+conversions (if deviceType is 'rflogicboard')
         or delete all osc1/3/5/7, conv2/4/6/8 and miscellaneous registers+conversions (if
         deviceTypeName is not 'rflogicboard').

        :note: should be run (both for rflogicboard and for adc) before self.copyRegisters.

        :param str deviceTypeName: name of the devicetype
        """
        check_class(deviceTypeName, str)

        deviceType = self.hal.getDeviceType(deviceName = deviceTypeName)
        calList  = self.hal.getCalibrations(deviceTypeName)

        if deviceType:
            registersDict = self.hal.getRegistersInfo(deviceTypeName)
            if registersDict:
                registerNames = list(registersDict.keys())
                rflbRegs  = [x for x in registerNames if x.startswith('RFLB')]
                osc1Regs  = [x for x in registerNames if x.startswith('OSC1')]
                osc3Regs  = [x for x in registerNames if x.startswith('OSC3')]
                osc5Regs  = [x for x in registerNames if x.startswith('OSC5')]
                osc7Regs  = [x for x in registerNames if x.startswith('OSC7')]
                conv2Regs = [x for x in registerNames if x.startswith('CONV2')]
                conv4Regs = [x for x in registerNames if x.startswith('CONV4')]
                conv6Regs = [x for x in registerNames if x.startswith('CONV6')]
                conv8Regs = [x for x in registerNames if x.startswith('CONV8')]
                if deviceTypeName != 'rflogicboard':
                    for registerName in rflbRegs+osc1Regs+conv2Regs:
                        deviceType.removeRegister(registerName)
                        for cal in calList:
                            cal.removeConversion(registerName)
                for registerName in osc3Regs+osc5Regs+osc7Regs+conv4Regs+conv6Regs+conv8Regs:
                    deviceType.removeRegister(registerName)
                    for cal in calList:
                        cal.removeConversion(registerName)
            else:
                logger.info("Devicetype '%s' does not contain any register to be deleted",
                            deviceTypeName)
        else:
            raise AttributeError("Devicetype {} was not found in the HAL".format(deviceTypeName))


    def copyRegisters(self): # pylint: disable=too-many-locals,too-many-statements
        """
         Copy the rflogicboard osc1 and conv2 registers+conversions into osc3/5/7 and conv4/6/8
         (rflogicboard map) and copy the rflogicboard osc1/3/5/7, conv2/4/6/8 and miscellanous
         registers+conversions into the adc address map.
         Note: this function should be run after self.deleteRegisters (which should be run twice:
         first for the rflogicboard and second for the adc).
        """
        rflogicboard = self.hal.getDeviceType(deviceName = 'rflogicboard') 
        adc = self.hal.getDeviceType(deviceName = 'adc')

        if rflogicboard and adc:
            rfRegistersDict = self.hal.getRegistersInfo('rflogicboard')
            if rfRegistersDict:
                rfRegisters = list(rfRegistersDict.keys())
                osc1Regs  = [x for x in rfRegisters if x.startswith('OSC1')]
                conv2Regs = [x for x in rfRegisters if x.startswith('CONV2')]
                rflbRegs  = [x for x in rfRegisters if x not in osc1Regs and x not in conv2Regs]
            else:
                return False
        else:
            return False

        rfCalList  = self.hal.getCalibrations('rfLogicboard')
        # it is supposed that rflogicboard has just one calibration
        rfCal = rfCalList[0]
        adcCalList = self.hal.getCalibrations('adc')

        # The rflogicboard miscellaneous registers are copied into the adc address space;
        # Since some registers have the same name as the existing adc registers, the register name
        # is prefixed with 'RFLB_'
        # In the adc address map the miscellaneous registers are placed in the first 256 registers
        # block following the rflogicboard base address whilst in the rflogicboard these registers
        # are in repetitive 256 register blocks starting at address 0 through the first 256
        # registers block following the rflogicboard base address.
        for registerName in rflbRegs:
            register = rflogicboard.getRegister(registerName)
            newRegister = deepcopy(register)
            newRegister.name = 'RFLB_' + register.name
            newRegister.address = register.address + self.baseAddress
            adc.addRegister(newRegister)
            if rfCal:
                conversion = rfCal.getConversion(registerName)
                if conversion:
                    newConversion = deepcopy(conversion)
                    newConversion.name = 'RFLB_' + conversion.name
                    for adcCal in adcCalList:
                        adcCal.addConversion(newConversion.name, deepcopy(newConversion))

        for registerName in osc1Regs:
            # the osc1 registers are copied into the adc address map
            register = rflogicboard.getRegister(registerName)
            adc.addRegister(deepcopy(register))
            if rfCal:
                conversion = rfCal.getConversion(registerName)
                if conversion:
                    for adcCal in adcCalList:
                        adcCal.addConversion(registerName, deepcopy(conversion))
            # the osc1 registers are copied into osc3, osc5 and osc7 (both rflogicboard and adc)
            for i in ['3', '5', '7']:
                newRegister = deepcopy(register)
                nameList = list(newRegister.name)
                nameList[3] = i
                newRegister.name = ''.join(nameList)
                newRegister.address = register.address + (int(i)-1)*0x100
                rflogicboard.addRegister(newRegister)
                adc.addRegister(deepcopy(newRegister))
                if rfCal:
                    if conversion:
                        newConversion = deepcopy(conversion)
                        nameList = list(newConversion.name)
                        nameList[3] = i
                        newConversion.name = ''.join(nameList)
                        rfCal.addConversion(newConversion.name, newConversion)
                        for adcCal in adcCalList:
                            adcCal.addConversion(newConversion.name, deepcopy(newConversion))

        for registerName in conv2Regs:
            # the conv2 registers are copied into the adc address map
            register = rflogicboard.getRegister(registerName)
            adc.addRegister(deepcopy(register))
            if rfCal:
                conversion = rfCal.getConversion(registerName)
                if conversion:
                    for adcCal in adcCalList:
                        adcCal.addConversion(registerName, deepcopy(conversion))
            # the conv2 registers are copied into conv4, conv6 and conv8 (both rflogicboard and adc)
            for i in ['4', '6', '8']:
                newRegister = deepcopy(register)
                nameList = list(newRegister.name)
                nameList[4] = i
                newRegister.name = ''.join(nameList)
                newRegister.address = register.address + (int(i)-2)*0x100
                rflogicboard.addRegister(newRegister)
                adc.addRegister(deepcopy(newRegister))
                if rfCal:
                    if conversion:
                        newConversion = deepcopy(conversion)
                        nameList = list(newConversion.name)
                        nameList[4] = i
                        newConversion.name = ''.join(nameList)
                        rfCal.addConversion(newConversion.name, newConversion)
                        for adcCal in adcCalList:
                            adcCal.addConversion(newConversion.name, deepcopy(newConversion))


def main():
    """ Update kids HAL with rflogicboard registers. """

    parser = createGsswArgumentParser(description=__doc__)
    args = parser.parse_args()
    config = initGssw(args)

    hal       = Hal()
    halClient = HalClient(hal, config)
    halClient.requestHal()

    halUpdater = HalUpdater(hal)
    # always delete first!
    halUpdater.deleteRegisters('rflogicboard')
    halUpdater.deleteRegisters('adc')
    # copy after delete!
    halUpdater.copyRegisters()

    # update the HAL
    halYaml = hal.createYamlData()
    halClient.setHal(halYaml)


if __name__ == '__main__':
    try:
        main()
    finally:
        destroyAllSocketsInContext()
