#!/usr/bin/env python3
""" Module for managing hal.yaml files under version control.

    Supported actions:
      - Pulling updates from git remote.
      - Making commits with specified authentication credentials, and pushing to remote.
"""

import logging
import os
import pwd
import urllib.parse

try:
    import git
    from   git.exc import InvalidGitRepositoryError
except ImportError as ex: # git module might not be installed, when not installed via package. 
    git = None
    InvalidGitRepositoryError = Exception

import gssw.hal.versioncontrol.git_askpass as git_askpass
from gssw.lib.errors import GsswError

__all__ = ['VersionControlGit']

logger = logging.getLogger(__name__)


# SRON internal GIT server credentials (read-only access to git)
GIT_PULL_USERNAME = 'egse_hal'
GIT_PULL_PASSWORD = 'egse_hal'

class VersionControlGit:
    """ Class for interfacing with git. """


    def __init__(self, repoPath):
        """ Create a git interface object.

        :param str repoPath: Path to the git repository, or any file or subdirectory within.
        """

        # Credentials for background-pulling
        self.pullUsername  = GIT_PULL_USERNAME
        self.pullPassword  = GIT_PULL_PASSWORD

        if git is None:
            raise GsswError("Git module is not available, cannot use synchronization")

        try:
            self._repo   = git.Repo(repoPath, search_parent_directories=True)
        except InvalidGitRepositoryError:
            # raise ... from none will remove the original exception 
            raise GsswError("Path of hal file '{}' is not a git repository, " \
                            "cannot be used for synchronization".format(repoPath)) from None

        try:
            self._origin = self._repo.remotes[0]
        except IndexError:
            # raise ... from none will remove the original exception
            raise GsswError("Git repository for hal file in '{}' has no remote server, " \
                            "cannot be used for synchronization".format(repoPath)) from None

        # Derive git server url from repo origin (credentials required for git cli tools < 2.7)
        self._prepareCredentials(remoteInfo=False)
        try:
            originUrl = list(self._origin.urls)[0]
        finally:
            self._cleanCredentials()

        logger.info('Repository %s is updated from remote repository at: %s',
                    self._repo.working_dir, originUrl)
        parsedUrl = urllib.parse.urlparse(originUrl)

        # Extracted protocol and server address are stored for use in later authentication
        self.remoteServerProtocol  = '{}://'.format(parsedUrl.scheme)
        self.remoteServerUri = '{}'.format(parsedUrl.netloc)

    def _getLocalBranches(self):
        """ Compiles a list of branche names on the local repository clone. """
        return [branch.name for branch in self._repo.branches]

    def _getActiveBranch(self):
        """ Retrieve the name of the active local branch. """
        return self._repo.active_branch.name

    def checkLocalChanges(self):
        """ Checks whether there are any modifications that can be uploaded. """
        return self._repo.is_dirty()

    def checkUpdateAvailable(self):
        """ Check for updates on remote server. """
        self._prepareCredentials()
        try:
            trackingBranch = self._repo.active_branch.tracking_branch()
            if trackingBranch:
                referenceSpecifier = '{}:{}'.format(
                    self._repo.active_branch.path,
                    trackingBranch.path)
                fetches = self._origin.fetch(referenceSpecifier)
                fetch = fetches[0]

                revisionSpecifier = '{}..{}'.format(self._getActiveBranch(),
                                                    fetch.ref.name)
                commitsBehind = self._repo.iter_commits(revisionSpecifier)
                count = len(list(commitsBehind))
                return count, fetch.ref.name
        except git.GitCommandError as ex:  #pylint: disable=no-member
            logger.exception("Could not check for updates: %s", ex)
            raise GsswError("Could not check for updates: {}".format(ex)) from None
        finally:
            self._cleanCredentials()
        return None, None

    def update(self):
        """ Download updates from git remote """
        self._prepareCredentials()
        try:
            oldUser, oldEmail = self._getCommitAuthor()
            try:
                # For stashing, a committer is required, so configure one in the git configation
                self._setCommitAuthor(*self._generateUsernameEmail())
                try:
                    if self._repo.is_dirty():
                        if git.Git().version_info < (2,6,0): # autostash only available on newer git
                            logger.error(
                                "Cannot perform pull, there are local uncommitted changes.")
                            return
                        logger.warning(
                            "There are local uncomitted changes, pulling with autostash and rebase")
                        # Pulling with rebase + autostash retrieves all the updates, and re-applies
                        #   local changes
                        # Pruning is used to prevent lots of cache objects building up.
                        # This can happen because the hal server extensively pulls in the background
                        pullInfo = self._origin.pull(rebase=True, autostash=True, prune=True)
                    else:
                        # See note above about pruning
                        pullInfo = self._origin.pull(prune=True)
                    if not bool(pullInfo):
                        raise GsswError(
                            "Pulling updates from remote server failed (no details available)")
                    pullInfo = pullInfo[0]
                    if pullInfo.flags & pullInfo.ERROR:
                        raise GsswError("Pulling updates from remote server failed: {!r}".format(
                            pullInfo.summary.strip()))
                except git.GitCommandError as ex:  #pylint: disable=no-member
                    logger.exception("Could not pull updates")
                    raise GsswError(
                        "Pulling updates from remote server failed: {}".format(ex)) from ex
            finally:
                self._setCommitAuthor(oldUser, oldEmail)
        finally:
            self._cleanCredentials()

    def _prepareCredentials(self, username=None, password=None, remoteInfo=True):
        """ Prepares credentials for the git askpass utility.

        Custom credentials can be provided, otherwise the SRON read-only user is used. """

        # Git doesn't provide a direct API for supplying the credentials
        # However, via the GIT_ASKPASS environment variable, we can use our own utility for
        # suppyling the credentials, which we pass to this utility via other environment variables.
        # See: https://stackoverflow.com/questions/44784828/gitpython-git-authentication-using-user-
        #      and-password
        # See: https://git-scm.com/docs/gitcredentials.html
        os.environ['GIT_ASKPASS']  = git_askpass.__file__
        if remoteInfo:
            os.environ['GIT_URI']      = self.remoteServerUri
            os.environ['GIT_PROTOCOL'] = self.remoteServerProtocol
        os.environ['GIT_USERNAME'] = self.pullUsername if username is None else username
        os.environ['GIT_PASSWORD'] = self.pullPassword if password is None else password

    def _cleanCredentials(self):
        """ Cleans up the credentials from the environment for the git askpass utility. """
        # Remove credentials environment variables
        for variable in ('GIT_ASKPASS',
                         'GIT_URI',
                         'GIT_PROTOCOL',
                         'GIT_USERNAME',
                         'GIT_PASSWORD'):
            if variable in os.environ:
                del os.environ[variable]

    def _getCommitAuthor(self):
        """ Reads the current committer name and email from the git configuration. """
        configWriter = self._repo.config_writer('repository')
        configUser  = None
        configEmail = None
        with configWriter:
            # Try to set the full name (and email) in the local git repository config,
            # so they will appear in the git log

            #pylint: disable=no-member
            if configWriter.has_section('user'):
                if configWriter.has_option('user', 'name'):
                    configUser = configWriter.get_value('user', 'name')
                if configWriter.has_option('user', 'email'):
                    configEmail = configWriter.get_value('user', 'email')
            #pylint: enable=no-member
            return configUser, configEmail

    def _setCommitAuthor(self, user, email):
        """ It is mandatory to have an author configured in the git config, for some git actions.
        (e.g. stashing), so here you can set them. """
        configWriter = self._repo.config_writer('repository')
        with configWriter:
            #pylint: disable=no-member
            if not configWriter.has_section('user'):
                configWriter.add_section('user')
            # Try to set the full name (and email) in the local git repository config,
            if user is None:
                configWriter.remove_option('user', 'name')
            else:
                configWriter.set_value('user', 'name', user)
            if email is None:
                configWriter.remove_option('user', 'email')
            else:
                configWriter.set_value('user', 'email', email)
            configWriter.write()
            #pylint: enable=no-member

    def _generateUsernameEmail(self, username=None):
        """ Try to resolve the full name of the provided username (ask the operating system). """
        commitUser  = self.pullUsername if username is None else username # default username
        commitEmail = '{}@sron.nl'.format(commitUser)
        try:
            commitUser = pwd.getpwnam(commitUser).pw_gecos.strip()
        except KeyError:
            # git user is not available on this host, no full name available (use username)
            pass
        return commitUser, commitEmail


    def commitPush(self, files, comment, username=None, password=None):
        """ Commits files to git, and pushes them to remote. """
        logger.info('Uploading %r to repository (%s)', files, comment)

        assert self._repo.is_dirty(), "No changes available to commit"

        # Stage the mofidied files 
        self._repo.index.add(files)

        oldUser, oldEmail = self._getCommitAuthor()
        try:
            # Try to set the full name (and email) in the local git repository config,
            # so they will appear in the git log
            commitUser, commitEmail = self._generateUsernameEmail(username)
            self._setCommitAuthor(commitUser, commitEmail)

            # Now do the commit
            self._repo.index.commit(comment)
        finally:
            # Restore the git user + email in the local git repository config
            self._setCommitAuthor(oldUser, oldEmail)

        # No do the push (to the remote git server)
        self._prepareCredentials(username, password)
        try:
            try:
                pushInfo = self._origin.push()
            except git.GitCommandError as ex:  #pylint: disable=no-member
                logger.exception("Could not push commits")
                raise GsswError("Pushing commits to remote server failed: {}".format(ex))
            if not bool(pushInfo):
                raise GsswError("Pushing changes to remote server failed (no details available)")
            pushInfo = pushInfo[0]
            if pushInfo.flags & pushInfo.ERROR:
                raise GsswError("Pushing changes to remote server failed: {!r}".format(
                    pushInfo.summary.strip()))
            logger.info("Pushed changes to remote server: %r", pushInfo.summary.strip())
        finally:
            self._cleanCredentials()

    def discardLocalChanges(self):
        """ Undo any local uncommitted changes """

        if self._repo.is_dirty():
            logger.info("Git repository is dirty, reverting all files...")
            self._repo.index.checkout(force=True) # Revert all files
        else:
            logger.debug("Git repository is clean, nothing to do")

def main():
    """ Test function """
    path = '/data/simonr/hal/egse_hal/'

    versionControl  = VersionControlGit(path)
    newCommits, ref = versionControl.checkUpdateAvailable()
    if newCommits:
        print("Behind by {} commit(s) ({})".format(newCommits, ref))
        if input('Update? (y/n) ').lower()=='y':
            versionControl.update()
    else:
        print("Up to date ({!r})".format(ref))

    if input('Commit + push changes? (y/n) ').lower()=='y':
        files = ['{}/testfile.txt'.format(path)]
        versionControl.commitPush(files, "test commit")

if __name__=='__main__':
    main()

