""" This module contains the graphical elements used in the housekeeping manager.
"""

import re
import math

from PyQt5.QtCore import QMimeData, QPoint, QRect, QSize, Qt, QByteArray, QDataStream, QIODevice, \
                         QMargins
from PyQt5.QtGui import QDrag, QPalette, QPixmap, QPainter, QFont, QColor, QIcon, QRegion, QPolygon
from PyQt5.QtWidgets import QFrame, QLabel, QHBoxLayout, QVBoxLayout, QLineEdit, QMenu, QMainWindow


class DraggableMixin():
    """ This is a mixin class for items that can be dragged themselves (HK items, HK groups). """

    # All sizes of widgets in the hk manager can be adjusted using these constants.
    # These can not just be changed freely since that would probably mess up alignment
    # between different widgets
    LABEL_WIDTH = 200
    LABEL_HEIGHT = 20
    GROUPNAME_HEIGHT = 31
    MARGIN = 2 # space around widgets inside groups and between labels and groups
    BORDER = 2 # width of outline around groups
    GROUPNAME_BORDER = 1
    X_GRID = LABEL_WIDTH + 3 * MARGIN + 2 * BORDER
    Y_GRID = LABEL_HEIGHT + MARGIN
    GROUP_WIDTH = LABEL_WIDTH + 2 * MARGIN + 2 * BORDER
    # This should equal Y_GRID
    GROUP_HEIGHT_OVERHEAD = GROUPNAME_HEIGHT + 2 * MARGIN + 2 * BORDER + 2 * GROUPNAME_BORDER
    # The space inside the group that is reserved for the name and margins
    GROUP_HEADER = GROUPNAME_HEIGHT + 2 * MARGIN + 2 * GROUPNAME_BORDER
    GROUP_HEIGHT_INIT = GROUP_HEIGHT_OVERHEAD + Y_GRID


    def mousePressEvent(self, event):
        """ event handler for when a mouse click is performed on the widget.
            This prepares a drag operation.
        """

        if not event.buttons() == Qt.LeftButton:
            event.ignore()
            return

        hotSpot = event.pos()
        mimeData = QMimeData()

        # Serialize hotspot to use as MIME data
        itemData = QByteArray()
        dataStream = QDataStream(itemData, QIODevice.WriteOnly)
        dataStream << hotSpot  #pylint: disable=pointless-statement
        mimeData.setData('hotspot', itemData)

        pixmap = QPixmap(self.size())
        self.render(pixmap)

        drag = QDrag(self)
        drag.setMimeData(mimeData)
        drag.setPixmap(pixmap)
        drag.setHotSpot(hotSpot)

        dropAction = drag.exec_(Qt.CopyAction | Qt.MoveAction, Qt.CopyAction)

        if dropAction == Qt.MoveAction:
            self.close()
            self.update()


class PlaceableMixin():
    """ This is a mixin class for objects where other objects can be placed upon
        (HK group, backplane).
    """

    def updateValue(self, name, value, warningSticky, errorSticky):
        """ Update the value of an hk item that is a child of this object. """

        if name in self.itemMap:
            self.itemMap[name].setValue(value, warningSticky, errorSticky)
        else:
            # if the item is not a direct child, do a depth first search
            for childName in self.itemMap:
                child = self.itemMap[childName]
                if isinstance(child, DragGroup):
                    child.updateValue(name, value, warningSticky, errorSticky)


    def setInactive(self):
        """ Set all hk items state (label color) to inactive.

            This is a recursive function. When called in a BackPlane, it calls the
            same function for all its DragGroup children and a similarly named funciton
            for DragLabel children.
        """

        for item in self.itemMap.values():
            item.setInactive()


    # pylint: disable=unused-argument
    def dragLeaveEvent(self, event):
        """ This event occurs when a drag operation leaves the objects airspace. """

        self.landingZone = QRect()
        self.update()


    def printHierarchy(self, indentation=0):
        """ Print all children of this object hierarchically. """

        if isinstance(self, BackPlane):
            print('\nHierarchy:')

        for name in self.itemMap:
            item = self.itemMap[name]

            if isinstance(item, DragLabel):
                print('%s- %s [%3d, %3d] (%d)' % (' ' * indentation, item.name,
                    item.position.x(), item.position.y(), item.rawValue))
            else:
                print('%s- %s [%3d, %3d]' % (' ' * indentation, item.name,
                    item.position.x(), item.position.y()))

            if issubclass(type(item), PlaceableMixin):
                item.printHierarchy(indentation + 1)


class DragLabel(DraggableMixin, QFrame):
    """ This class is used to represent hk items.
        They are draggable and can be docked in hk groups or the backplane.
        This class also holds the latest value of the hk item.
    """

    # color codes used to change the background color
    # corresponding to the value and the items limits
    COLOR_INACTIVE  = 'rgb(200, 200, 200)'
    COLOR_TIMEOUT   = 'rgb(120, 120, 120)'
    COLOR_NO_CHECK  = '#AED6F1' #'rgb(30, 144, 255)'
    COLOR_OK        = '#ABEBC6' #'rgb(50, 205, 50)'
    COLOR_INVALID   = '#FFFF00' #'rgb(255, 255, 0)'
    COLOR_WARNING   = '#FF8000' #'rgb(255, 140, 0)'
    COLOR_ERROR     = '#FF0000' #'rgb(255, 0, 0)'
    COLOR_BORDER    = 'rgb(0,0,0)'
    COLOR_TEXT      = 'rgb(0,0,0)'

    def __init__(self, name, value, parent, item, hal, inactive=True):
        """ DragLabel constructor
            name   - the name of the HK item the label represents
            value  - the initial raw value of the HK item
            parent - the GUI parent of the label. Either a group or the backplane
            item   - pointer to the corresponding HkItem
            hal    - reference to hal object
        """
        QFrame.__init__(self, parent)

        self.name = name
        self.rawValue = value
        self.parent = parent
        self.item = item
        self.hal = hal
        self.inactive = inactive
        self.warningSticky = False
        self.errorSticky = False
        self.position = None

        # initialize the contents
        self.font = QFont("Arial", 8)
        self.layout = QHBoxLayout()
        self.labelName = QLabel(name)
        self.labelName.setFont(self.font)
        self.labelValue = QLabel(str(value))
        self.labelValue.setFont(self.font)
        self.layout.setContentsMargins(5, 1, 5, 1)
        self.layout.addWidget(self.labelName)
        self.layout.addStretch(1)
        self.layout.addWidget(self.labelValue)
        self.setLayout(self.layout)

        # get main window handle
        while not issubclass(type(parent), QMainWindow):
            parent = parent.parent
        self.hkManager = parent

        # visual stuff
        self.setMinimumSize(self.LABEL_WIDTH, self.LABEL_HEIGHT)
        self.setMaximumSize(self.LABEL_WIDTH, self.LABEL_HEIGHT)
        self.setFocusPolicy(Qt.ClickFocus | Qt.TabFocus)
        self.setAutoFillBackground(True)
        self.updateAppearance()


    @property
    def index(self):
        """ Get the index in the parent group. if not in the group, return -1. """

        layout = self.parent.innerLayout

        for i in range(layout.count()):
            if layout.itemAt(i).widget() is self:
                return i

        return -1


    def setPosition(self, position):
        """ Move the widget to a new position.

            The input position must be normailized to the grid.
            This funtion also adds an offset to convert grid coordinates to a nice looking layout.
        """

        self.position = position
        self.move(self.position + QPoint((self.GROUP_WIDTH - self.LABEL_WIDTH) / 2, self.MARGIN))

        self.show()


    def setInactive(self):
        """ Change the label color to inactive. """

        self.inactive = True
        self.updateAppearance()


    def setValue(self, rawValue, warningSticky, errorSticky):
        """ Update the value of the hk parameter. """

        self.rawValue = rawValue
        self.warningSticky = warningSticky
        self.errorSticky = errorSticky
        self.inactive = False
        self.updateAppearance()


    def updateAppearance(self):
        """ Update the color and displayed value. """

        # get the engineering value (for hw engineering items)
        if self.item.address is not None and self.item.dataFormat in [0,1]:
            engValue = self.rawValue # Conversion is already applied by the HK server

        # check the limits and change the background color if a limit is violated
        if self.inactive:
            color = self.COLOR_INACTIVE
        elif math.isnan(self.rawValue):
            color = self.COLOR_TIMEOUT
        elif not self.item.engineering:
            color = self.COLOR_NO_CHECK
        else:
            # get limits from hal
            device = self.hal.getDevice(self.item.getDeviceName())
            limits = self.hal.getRegisterLimits(
                device.name, device.deviceType, device.calibration, self.item.getRegisterName())

            if limits is None:
                color = self.COLOR_NO_CHECK
            elif engValue < limits.lowerError or engValue > limits.upperError:
                color = self.COLOR_ERROR
            elif engValue < limits.lowerWarn or engValue > limits.upperWarn:
                color = self.COLOR_WARNING
            elif engValue < limits.lowerValid or engValue > limits.upperValid:
                color = self.COLOR_INVALID
            else:
                color = self.COLOR_OK

        # color the label's border according to the limit sticky bits
        if self.errorSticky:
            borderColor = self.COLOR_ERROR
        elif self.warningSticky:
            borderColor = self.COLOR_WARNING
        else:
            borderColor = self.COLOR_BORDER

        # Set QSS of PlaceableMixin and sub-label. Use border color for text and border.
        self.setStyleSheet("""
            DragLabel {
                background-color: %s;
                color: %s;
                border: 2px solid %s;
            }
            DragLabel QLabel {
                background-color: %s;
                color: %s;
                border: none;
            }
        """ % (color, self.COLOR_TEXT, borderColor, color, self.COLOR_TEXT))

        self.labelName.setStyleSheet("color")

        # update the value on screen
        if math.isnan(self.rawValue):
            value = '---'
        else:
            # display the value in the selected format
            if self.item.dataFormat == 0:
                value = '{1:.{0}F}'.format(self.item.decimals, engValue)
            elif self.item.dataFormat == 1:
                value = '{1:.{0}E}'.format(self.item.decimals, engValue)
            elif self.item.dataFormat == 2:
                value = '0x{:X}'.format(self.rawValue)
            elif self.item.dataFormat == 3:
                value = '{:d}'.format(self.rawValue)
            elif self.item.dataFormat == 4:
                value = '0o{:o}'.format(self.rawValue)
            elif self.item.dataFormat == 5:
                value = '0b{:b}'.format(self.rawValue)
            elif self.item.dataFormat == 6:
                value = '{1:.{0}F}'.format(self.item.decimals, self.rawValue)

        # Differentiate between register/software hk; using `self.item.address`.
        if self.item.unit and (self.item.dataFormat in [0, 1] or \
           (not self.item.address and self.item.dataFormat == 6)):
            value += ' ' + self.item.unit

        self.labelValue.setText(value)


    # pylint: disable=unused-argument
    def focusInEvent(self, event):
        """ event handler for when the widget comes into focus. """

        # change the color of the widget slightly
        palette = QPalette()
        palette.setColor(QPalette.Window, Qt.gray)
        self.setAutoFillBackground(True)
        self.setPalette(palette)


    # pylint: disable=unused-argument
    def focusOutEvent(self, event):
        """ event handler for when the widget loses focus. """

        palette = QPalette()
        palette.setColor(QPalette.Window, Qt.lightGray)
        self.setAutoFillBackground(True)
        self.setPalette(palette)


    def contextMenuEvent(self, event):
        """ This handler handles right clicks on the widget by showing a menu. """

        menu = QMenu(self)
        editAction = menu.addAction("Edit HK Item")
        deleteAction = menu.addAction("Delete HK Item")
        resetAction = menu.addAction("Reset Limit Crossings")
        menuAction = menu.exec_(self.mapToGlobal(event.pos()))

        if menuAction == editAction:
            self.hkManager.launchEditHkItem(self.name)
        elif menuAction == deleteAction:
            self.hkManager.removeHkItems([self])
            self.hkManager.hkClient.removeHkItem(self.name)
        elif menuAction == resetAction:
            self.hkManager.resetHkLimits([self])


class DragGroup(PlaceableMixin, DraggableMixin, QFrame):
    """ This class represent hk groups in the manager.

        Hk items can be dragged into and docked in these groups. Groups automatically resize
        to fit their contents. Groups can also be dragged themselves but they are not nestable.
    """

    # text and border colors
    COLOR_BACKGROUND = '#DCDCDC' # this is the default
    COLOR_BORDER = 'rgb(0, 0, 0)'
    COLOR_NAME = 'rgb(0, 0, 0)'
    COLOR_NAME_BACKGROUND = 'rgb(255, 255, 255)'
    COLOR_GROUPNAME_BORDER = 'rgb(0, 0, 0)'

    # colors available for group backgrounds
    N_COLORS = 9
    COLOR_0 = '#FF9999'
    COLOR_1 = '#FF9966'
    COLOR_2 = '#99FF99'
    COLOR_3 = '#66FFCC'
    COLOR_4 = '#99FFFF'
    COLOR_5 = '#66CCFF'
    COLOR_6 = '#9999FF'
    COLOR_7 = '#FF99FF'
    COLOR_8 = COLOR_BACKGROUND # default

    def __init__(self, parent, name=None, itemMap=None, order=None, color=None):
        QFrame.__init__(self, parent)

        self.name = None
        self.order = [] if order is None else order # hold names of items in descending order
        self.itemMap = {} if itemMap is None else itemMap
        self.parent = parent
        self.position = None
        self.landingZone = QRect()
        self.color = None

        # The outer layout contains the inner layout and a single stretcher
        self.outerLayout = QVBoxLayout()
        self.outerLayout.setContentsMargins(0,0,0,0)
        self.outerLayout.setSpacing(0)

        # the inner layout holds all the hk items (DragLabels) and the group name (QlineEdit)
        self.innerLayout = QVBoxLayout()
        self.innerLayout.setContentsMargins(self.MARGIN, self.MARGIN, self.MARGIN, self.MARGIN)
        self.innerLayout.setSpacing(self.MARGIN)

        # create a mask to cut a corner of the group name
        mask = QRegion(QPolygon([QPoint(0, 0),
                                 QPoint(0, self.GROUPNAME_HEIGHT),
                                 QPoint(self.LABEL_WIDTH, self.GROUPNAME_HEIGHT),
                                 QPoint(self.LABEL_WIDTH - self.GROUPNAME_HEIGHT, 0),
                                 QPoint(0, 0)]))
        self.nameEdit = QLineEdit()
        self.nameEdit.setMask(mask)
        self.nameEdit.setFont(QFont("Arial", 12))
        self.nameEdit.setTextMargins(QMargins(4,4,4,4))
        self.nameEdit.setFrame(True)
        self.nameEdit.setStyleSheet("border: 1px solid black")
        self.nameEdit.setMinimumSize(self.LABEL_WIDTH, self.GROUPNAME_HEIGHT)
        self.nameEdit.setMaximumSize(self.LABEL_WIDTH, self.GROUPNAME_HEIGHT)
        self.innerLayout.addWidget(self.nameEdit)
        self.outerLayout.addLayout(self.innerLayout)
        self.outerLayout.addStretch(1)
        self.setLayout(self.outerLayout)

        # set the name
        if not name is None:
            self.setName(name)

        # set the color
        self.setColor(color if not color is None else self.COLOR_BACKGROUND)

        # add children
        for cName in self.order:
            item = self.itemMap[cName]
            item.parent = self
            self.innerLayout.addWidget(item)
            item.show()
        self.autoResize()

        # get main window handle
        parent = self.parent
        while not issubclass(type(parent), QMainWindow):
            parent = parent.parent
        self.hkManager = parent

        # visual stuff
        newPalette = self.palette()
        newPalette.setColor(QPalette.Window, Qt.white)
        self.setPalette(newPalette)
        self.setAcceptDrops(True)
        self.setFocusPolicy(Qt.ClickFocus | Qt.TabFocus)
        self.setAutoFillBackground(True)

        # connect signals
        self.nameEdit.textEdited.connect(self.nameEditUpdated)


    def setColor(self, color):
        """ Change the background color of this group. """

        self.color = color

        self.setStyleSheet("""
            DragGroup{
                background-color: %s;
                color: %s;
                border: %dpx solid %s;
            }
            DragGroup QLineEdit {
                background-color: %s;
                color: %s;
                border-width: %dpx;
                border-style: solid;
                border-color: %s;
                font-weight: bold;
            }
        """ % (color, self.COLOR_BORDER, self.BORDER, self.COLOR_BORDER,
               self.COLOR_NAME_BACKGROUND, self.COLOR_NAME,
               self.GROUPNAME_BORDER, self.COLOR_GROUPNAME_BORDER))


    def gridPos(self, position):
        """ Return the position discretized to the grid. """

        if position.y() < self.GROUP_HEIGHT_OVERHEAD:
            return QPoint(0, self.GROUP_HEADER)

        return QPoint(0, (position.y() - self.GROUP_HEADER) // DraggableMixin.Y_GRID *
                          DraggableMixin.Y_GRID + self.GROUP_HEADER)


    def halfGridPos(self, position):
        """ Return the position discretized to the grid but with a half phase shift. """

        if position.y() < self.GROUP_HEIGHT_OVERHEAD:
            return QPoint(0, self.GROUP_HEADER)

        return QPoint(0, (position.y() - self.GROUP_HEADER) // DraggableMixin.Y_GRID *
                          DraggableMixin.Y_GRID + self.GROUP_HEADER + DraggableMixin.Y_GRID / 2)


    def setPosition(self, position):
        """ Move the group to a new position. The location is normalized to the grid. """

        self.position = position
        self.move(position + QPoint(
            (self.X_GRID - self.GROUP_WIDTH) // 2 - self.BORDER, self.MARGIN))
        self.show()


    def newName(self):
        """ Create an unique new name for this group. """

        name = 'new group'

        # avoid duplicate names by adding a number to the end of the name
        while name in self.parent.itemMap:
            match = re.search(r'\d+$', name)
            if match is None:
                name += ' 2'
            else:
                name = name[:-len(match.group())] + str(int(match.group()) + 1)

        self.name = name
        self.nameEdit.setText(self.name)
        self.parent.itemMap[self.name] = self


    def setName(self, name):
        """ Change the name of the group. """

        # update parents itemMap
        if self.name in self.parent.itemMap:
            del self.parent.itemMap[self.name]
        self.parent.itemMap[name] = self

        self.name = name
        self.nameEdit.setText(self.name)


    def nameEditUpdated(self, name):
        """ Called after user has changed the name box in the GUI """

        # update parents itemMap
        if self.name in self.parent.itemMap:
            del self.parent.itemMap[self.name]
        self.parent.itemMap[name] = self

        self.name = name


    def insertItem(self, hkItem, index=None):
        """ Insert a new hk item into the group.

            The new item can be inserted into a specific location or by default, at the bottom.
        """

        if index is None:
            index = len(self.itemMap) + 1 # the name qLineEdit is also in the layout so + 1

        hkItem.parent = self

        self.innerLayout.insertWidget(index, hkItem) # The group name is also in the layout so + 1
        self.order.insert(index, hkItem.name)
        self.itemMap[hkItem.name] = hkItem

        hkItem.show()
        self.autoResize()

        # to avoid overlap with other widgets due to resizing,
        # the group is reinserted to do collision detection
        del self.parent.itemMap[self.name]
        self.parent.insertItem(self, position=self.position)


    def removeItem(self, name):
        """ remove an item from the group. """

        # remove the item from the gui and the internal data structure
        if name in self.itemMap:
            self.innerLayout.removeWidget(self.itemMap[name])
            self.itemMap[name].setParent(None)
            del self.itemMap[name]
            self.order.remove(name)

        self.autoResize()
        self.update()


    def autoResize(self):
        """ Resize the group based on the number of items it contains. The minimum size is 1. """

        heightUnits = max(1, len(self.itemMap))
        size = QSize(self.GROUP_WIDTH, self.GROUP_HEIGHT_OVERHEAD + heightUnits * self.Y_GRID)

        self.resize(size)


    def dragMoveEvent(self, event):
        """ Event handler invoked when something is dragged somewhere over the group. """

        oldZone = self.landingZone

        if self.itemMap == {}:
            gPos = self.gridPos(QPoint())
            self.landingZone = QRect(gPos +
                QPoint(self.MARGIN + self.BORDER, 0), QSize(self.LABEL_WIDTH, self.LABEL_HEIGHT))
        else:
            gPos = self.halfGridPos(event.pos())
            self.landingZone = QRect(gPos + QPoint(self.MARGIN + self.BORDER, -self.MARGIN),
                QSize(self.LABEL_WIDTH, self.Y_GRID))

        updateRect = self.landingZone.united(oldZone)

        event.setDropAction(Qt.MoveAction)
        event.accept()
        self.update(updateRect)


    def dropEvent(self, event):
        """ Event handler for drop events.

            When a item is dropped, it is added to the group.
        """

        item = event.source()
        name = item.name

        index = self.halfGridPos(event.pos()).y() // self.Y_GRID

        if isinstance(event.source(), DragLabel):

            # This is required for logical behavior when moving an item downward within a group
            if item.parent is self and item.parent.order.index(name) < index:
                index -= 1

            value = item.rawValue
            newItem = DragLabel(name, value, self, item.item, item.hal, inactive=item.inactive)
        else:
            event.ignore()
            return

        event.accept()

        # remove from previous parent
        event.source().parent.removeItem(name)

        # add the item
        self.insertItem(newItem, index)

        self.landingZone = QRect()

        event.setDropAction(Qt.MoveAction)
        self.update()
        newItem.setFocus()


    def dragEnterEvent(self, event):
        """ Event handler for when someone drags something into the groups airspace. """

        if isinstance(event.source(), DragLabel):
            event.setDropAction(Qt.MoveAction)
            event.accept()
        else:
            event.ignore()


    def contextMenuEvent(self, event):
        """ Event handler for right clicks on the group. """

        mainMenu = QMenu(self)
        colorMenu = QMenu('Change Group Color', self)
        addAction = mainMenu.addAction("Add HK Item")
        deleteAction = mainMenu.addAction("Delete Group")

        # generate menu items for all available colors
        colorActions = {}
        for i in range(self.N_COLORS):
            colorCode = QColor(getattr(self, 'COLOR_' + str(i)))
            pixmap = QPixmap(20, 20)
            pixmap.fill(colorCode)
            icon = QIcon(pixmap)
            colorActions[i] = colorMenu.addAction(icon, getattr(self, 'COLOR_' + str(i)))

        mainMenu.addMenu(colorMenu)
        menuAction = mainMenu.exec_(self.mapToGlobal(event.pos()))
        if menuAction == addAction:
            self.hkManager.launchAddHkItem(groupName=self.name)
        if menuAction == deleteAction:
            l = list(self.itemMap.values())
            l.append(self)
            self.hkManager.removeHkItems(l)
            self.hkManager.hkClient.removeHkItems([x.name for x in l])
        else:
            for i in range(self.N_COLORS):
                if menuAction == colorActions[i]:
                    self.setColor(getattr(self, 'COLOR_' + str(i)))


    # pylint: disable=unused-argument
    def paintEvent(self, event):
        """ The paint event is overridden to draw drag and drop guides. """

        painter = QPainter()
        painter.begin(self)

        painter.setBrush(Qt.black)
        painter.setPen(Qt.NoPen)
        painter.setOpacity(0.5)

        if self.landingZone.isValid():
            painter.drawRect(self.landingZone)

        painter.end()


class BackPlane(PlaceableMixin, QFrame):
    """ This class represents the backplane of the hk manager.

        The backplane can hold hk items and groups which can be moved with drag and drop operations.
    """

    def __init__(self, parent):
        QFrame.__init__(self, parent)

        self.setStyleSheet("""
            BackPlane{
                background-color: rgb(150, 150, 150);
            }
        """)

        self.parent = parent
        self.itemMap = {} # dictionary with names and DragLabel instances
        self.landingZone = QRect()
        newPalette = self.palette()
        newPalette.setColor(QPalette.Window, Qt.white)
        self.setPalette(newPalette)
        self.setAcceptDrops(True)


    def gridPos(self, position):
        """ Return the position normalized to the grid. """

        # avoid placement of items outside the window
        if position.x() < 0:
            position.setX(0)
        if position.y() < 0:
            position.setY(0)

        return QPoint(position.x() // DraggableMixin.X_GRID * DraggableMixin.X_GRID,
                      position.y() // DraggableMixin.Y_GRID * DraggableMixin.Y_GRID)


    def gridRect(self, position, hUnits=1):
        """ Return the rectangle the position takes on the grid.

        :param position: Input position
        :param hUnits:   the number of grid steps in the vertical direction
                         (the height on the rectangle)
        """

        pos = self.gridPos(position)
        return QRect(pos, QSize(DraggableMixin.X_GRID, DraggableMixin.Y_GRID * hUnits))


    def removeItem(self, name):
        """ Remove an item from the backplane. """

        if name in list(self.itemMap):
            self.itemMap[name].setParent(None)
            del self.itemMap[name]

        else: # if the item is not a direct child. do a depth first search
            for childName in self.itemMap:
                child = self.itemMap[childName]
                if isinstance(child, DragGroup):
                    if name in child.itemMap:
                        child.removeItem(name)

        self.update()


    def isOccupied(self, position):
        """ Check if position is occupied.

            Iterate through all items to check if they occupy the location on the grid.
            Return the occupant and otherwise None.
        """

        yPos = position.y() // DraggableMixin.Y_GRID * DraggableMixin.Y_GRID

        for name in self.itemMap:
            item = self.itemMap[name]

            if item.position is None:
                continue

            # check if the item is in the right column
            if position.x() // DraggableMixin.X_GRID != item.position.x() // DraggableMixin.X_GRID:
                continue

            itemYPos = item.position.y()
            if itemYPos <= yPos < itemYPos + item.height() + DraggableMixin.MARGIN:
                return item

        return None


    def placeItem(self, item, position):
        """ Place a widget on the plane. The position must be free. """

        self.itemMap[item.name] = item
        item.setParent(self)
        item.setPosition(position)
        item.show()
        area = self.gridRect(position)
        self.update(area)


    def insertItem(self, item, position=None):
        """ Insert an hk item or group into the backplane.

            This function also takes care of displacing items that are in the way.
        """

        y = 0

        # find an empty space if no position is specified
        while position is None:
            if self.isOccupied(QPoint(0, y)):
                y += DraggableMixin.Y_GRID
            else:
                position = QPoint(0, y)

        displaced = []

        # collect and remove widgets that must be relocated
        for p in self.getOccupiedPositions(item, position):
            widget = self.isOccupied(p)
            if widget is not None:
                displaced.append(widget)
                del self.itemMap[widget.name]

        # place new widget
        self.placeItem(item, position)

        # place back the displaced widgets pushing down other widgets
        pos = position + self.getGridHeight(item)
        while displaced != []:

            # pop a widget from the to place this loop iteration
            placedWidget = displaced.pop(0)

            # collect all conflicting widgets from this placement
            for p in self.getOccupiedPositions(placedWidget, pos):
                widget = self.isOccupied(p)
                if widget is not None:
                    displaced.append(widget)
                    del self.itemMap[widget.name]

            # place the widget and update the position
            self.placeItem(placedWidget, QPoint(pos))
            pos += self.getGridHeight(placedWidget)


    def getGridHeight(self, item):
        """ Get the height of a hk item or group discretized to the grid. """

        return QPoint(0, (item.height() // DraggableMixin.Y_GRID + 1) * DraggableMixin.Y_GRID)


    def getOccupiedPositions(self, item, position):
        """ Get a list of all the grid slots occupied by a widget given its position. """

        l = []
        for i in range((item.height() + DraggableMixin.MARGIN) // DraggableMixin.Y_GRID):
            l.append(position + QPoint(0, i * DraggableMixin.Y_GRID))

        return l


    def getGroup(self, name):
        """ Return the group object with the given name. Return None if not present. """

        if not name in self.itemMap:
            return None
        if not isinstance(self.itemMap[name], DragGroup):
            return None

        return self.itemMap[name]


    def getItem(self, name):
        """ Return the dragLabel object (hk item) with the given name. Return None if nto found"""

        # check direct children
        if name in self.itemMap:
            if isinstance(self.itemMap[name], DragLabel):
                return self.itemMap[name]

        # check inside groups
        for childName in self.itemMap:
            if isinstance(self.itemMap[childName], DragGroup):
                if name in self.itemMap[childName].itemMap:
                    return self.itemMap[childName].itemMap[name]

        return None


    def dropEvent(self, event):
        """ Event handler for drop events.

            Dropping a hk item or group inserts it in the backplane.
            The widget is not moved but instead a new one is created and the old one is destroyed.
        """

        oldWidget = event.source()
        name = oldWidget.name
        parent = oldWidget.parent
        newPosition = self.landingZone.topLeft()

        # Create a new widget
        if isinstance(oldWidget, DragGroup):
            newWidget = DragGroup(self, name=name, itemMap=oldWidget.itemMap,
                                  order=oldWidget.order, color=oldWidget.color)
        elif isinstance(oldWidget, DragLabel):
            newWidget = DragLabel(name, oldWidget.rawValue, self, oldWidget.item, oldWidget.hal,
                                  inactive=oldWidget.inactive)
        else:
            event.ignore()
            return

        # remove the old widget from its parent
        parent.removeItem(name)

        event.accept()
        self.landingZone = QRect()

        if isinstance(newWidget, DragGroup):
            newWidget.setName(name)

        # insert the new widget
        self.insertItem(newWidget, position=newPosition)

        event.setDropAction(Qt.MoveAction)
        self.update(self.gridRect(event.pos()))
        newWidget.setFocus()


    def dragEnterEvent(self, event):
        """ Event handler for dragging something into the backplanes airspace. """

        if issubclass(type(event.source()), DraggableMixin): # only accept hk items and groups
            event.setDropAction(Qt.MoveAction)
            event.accept()
        else:
            event.ignore()


    def dragMoveEvent(self, event):
        """ Event handler for dragging something over the backplane. """

        updateRect = self.landingZone.united(self.landingZone).united(self.gridRect(event.pos()))
        item = event.source()

        # Create drop shadow
        height = (item.size().height() + DraggableMixin.MARGIN) // DraggableMixin.Y_GRID
        self.landingZone = self.gridRect(event.pos(), hUnits=height)

        event.setDropAction(Qt.MoveAction)
        event.accept()
        self.update(updateRect)


    def contextMenuEvent(self, event):
        """ Event handler for right clicks. """

        menu = QMenu(self)
        addItem = menu.addAction("Add Item...")
        addGroup = menu.addAction("Add Group")
        menuAction = menu.exec_(self.mapToGlobal(event.pos()))
        if menuAction == addItem:
            self.parent.launchAddHkItem()
        elif menuAction == addGroup:
            self.parent.createHkGroup()


    # pylint: disable=unused-argument
    def paintEvent(self, event):
        """ The paint event is overridden to draw drag and drop guides. """

        painter = QPainter()
        painter.begin(self)

        painter.setBrush(Qt.black)
        painter.setPen(Qt.NoPen)
        painter.setOpacity(0.2)

        if self.landingZone.isValid():
            painter.drawRect(self.landingZone)

        painter.end()
