""" HK Client module """

import logging
import numpy

from gssw.lib.sockets           import ReqSocket, SubSocket, tagName
from gssw.lib.errors            import CommunicationError
from gssw.hk.hklib              import getUnion
from gssw.proto.common_pb2      import ErrorRep
import gssw.proto.hk_pb2        as hkProto


logger = logging.getLogger(__name__)


class HkClient():
    """ This class can be instantiated by modules that want to interact with
        the house keeping manager. This module supports creating new hk items
        and reading and/or updating their values.
    """

    def __init__(self, hkServerRep, hkServerPub, socketContext=None):
        """ Constructor for the house keeping client.

            :param str hkserverRep: address of the hk server reply socket
            :param str hkServerPub: address of the hk server publish socket
            :param socketContext: zmq context created by a higher layer
        """

        # Set up logging.
        logger.info('GSSw HkClient module started')
        self.callback = None

        # Create a REQ socket to communicate with the hk server.
        self.pubSocket = ReqSocket(context=socketContext)
        self.pubSocket.createZmqSocket(timeoutMs=60000)
        self.pubSocket.connectSocket(hkServerRep)

        # Create a sub socket to communicate with the hk server.
        self.subSocket = SubSocket()
        self.subSocket.createZmqSocket()
        self.subSocket.connectSocket(hkServerPub)


    def addHkItem(self, name, address=None, periodMs=1000, engineering=True, unit=None,
                  warningAction=None, warningRepeat=False, errorAction=None, errorRepeat=False):
        """ Register an hk item.

            :param str name:            the name of the hk item
            :param str address:         the address of the corresponding register in the form of:
                                        deviceName:registerName:offset
            :param int periodMs:        the update period in milliseconds
            :param bool engineering:    monitor the raw or the engineering value
            :param str unit:            the unit of the hk item
            :param str warningAction:   action upon warning limit crossing
            :param bool warningRepeat:  repeat warning action on every crossing
            :param str errorAction:     action upon error limit crossing
            :param bool errorRepeat:    repeat error action on every crossing
            :raises CommunicationError: if the server returns an error reply

            If the address parameter is set the hk manager will automatically update
            the hk value.
        """

        # Create the request protobuf object.
        req = hkProto.HkItemReq()
        req.name = name

        # add optional fields
        if address:
            req.auto.address = address
            req.auto.period_ms = periodMs
            req.auto.engineering = engineering
        if warningAction:
            req.warning_action.action = warningAction
            req.warning_action.repeat = warningRepeat
        if errorAction:
            req.error_action.action = errorAction
            req.error_action.repeat = errorRepeat
        if unit:
            req.unit = unit

        # Send request.
        self.pubSocket.sendProtoBuf(tagName(hkProto.HkItemReq), req.SerializeToString())

        # Receive reply and check for errors.
        parts = self.pubSocket.receiveProtoBuf()
        if parts.tag == tagName(ErrorRep):
            logger.error('Received an error reply when registering hk item')
            CommunicationError.raiseFromProtoBuffer(parts.message)
        else:
            logger.info('Added HK item %s', name)


    def removeHkItem(self, name):
        """ Unregister a hk item.

            :param str name:            the name of the item to remove.
            :raises CommunicationError: if the server returns an error reply
        """

        self.removeHkItems([name])


    def removeHkItems(self, names):
        """ Unregister multiple hk item.

            :param list[str] names: list of names of the items to remove.
            :raises CommunicationError: if the server returns an error reply
        """

        # Create protobuf object
        req = hkProto.HkRemoveItemsReq()
        req.name.extend(names)

        # Send request.
        self.pubSocket.sendProtoBuf(tagName(hkProto.HkRemoveItemsReq), req.SerializeToString())

        # Receive reply and check for errors.
        parts = self.pubSocket.receiveProtoBuf()
        if parts.tag == tagName(ErrorRep):
            logger.error('Error reply received during removeHkItems')
            CommunicationError.raiseFromProtoBuffer(parts.message)


    def resetHkLimits(self, names):
        """ Reset the limit crossing sticky bits for multiple HK items
            :param list[str] names: list of names of items to reset
            :raises CommunicationError: if the server returns an error reply
        """

        # Create protobuf object
        req = hkProto.HkResetLimitsReq()
        req.name.extend(names)

        # Send request.
        self.pubSocket.sendProtoBuf(tagName(hkProto.HkResetLimitsReq), req.SerializeToString())

        # Receive reply and check for errors.
        parts = self.pubSocket.receiveProtoBuf()
        if parts.tag == tagName(ErrorRep):
            logger.error('Error reply received during resetHkLimits')
            CommunicationError.raiseFromProtoBuffer(parts.message)


    def getHkState(self):
        """ Query the HK server's state.

            :return: indication whether the server is running.
            :rtype: bool
            :raises CommunicationError: if the server returns an error reply
        """

        # request the state from the server
        state = self._sendControlReq(hkProto.HK_STATE)

        return state == hkProto.HK_RUNNING


    def startHk(self):
        """ Start house keeping.

            :return: the state of the server after processing the request
            :rtype: HkControlRep.State
            :raises CommunicationError: if the server returns an error reply
        """

        logger.info('Start HK.')
        self._sendControlReq(hkProto.HK_START)


    def stopHk(self):
        """ Start house keeping.

            :return: the state of the server after processing the request
            :rtype: HkControlRep.State
            :raises CommunicationError: if the server returns an error reply
        """

        logger.info('Stop HK.')
        self._sendControlReq(hkProto.HK_STOP)


    def clearHkItems(self):
        """ Remove all registered hk items.

            :return: the state of the server after processing the request
            :rtype: HkControlRep.State
            :raises CommunicationError: if the server returns an error reply
        """

        self._sendControlReq(hkProto.HK_CLEAR)
        logger.info('Cleared all HK items.')


    def _sendControlReq(self, command):
        """ Send a command request to the hk server.

            :param HkControlReq.Command command: command to send to hk server
            :return: the state of the server after processing the request
            :rtype: HkControlRep.State
            :raises CommunicationError: if the server returns an error reply
        """

        # Create request message
        req = hkProto.HkControlReq()
        req.command = command

        # Send request
        self.pubSocket.sendProtoBuf(tagName(hkProto.HkControlReq), req.SerializeToString())

        # Receive reply and check for errors.
        parts = self.pubSocket.receiveProtoBuf()
        if parts.tag == tagName(ErrorRep):
            logger.error('Received an error reply when sending control request')
            CommunicationError.raiseFromProtoBuffer(parts.message)

        rep = hkProto.HkControlRep()
        rep.ParseFromString(parts.message)

        return rep.state


    def getAllHkValues(self):
        """ Get the value of all registerd hk items.

            :return: dictionary containing (name, (value, warning_sticky, error_sticky)) pairs
            :rtype: dict(str, list[str, bool, bool])
            :raises CommunicationError: if the server returns an error reply
        """

        # Create request protobuf object
        req = hkProto.HkGetAllDataReq()

        # Send request.
        self.pubSocket.sendProtoBuf(tagName(hkProto.HkGetAllDataReq), req.SerializeToString())

        # Receive reply, check for errors and format return
        return self._recvHkValues()


    def getHkValues(self, names):
        """ Request the most recent value of one or more HK items.

            :param list[str]: names a name or a list of names of hk items to read.
            :return: dictionary containing (name, (value, warning_sticky, error_sticky)) pairs
            :rtype: dict(str, list[str, bool, bool])
            :raises CommunicationError: if the server returns an error reply
        """

        # place a single name in a list.
        if isinstance(names, str):
            names = [names]

        # Create the protobuf request object.
        req = hkProto.HkGetDataReq()
        req.name.extend(names)

        # Send request.
        self.pubSocket.sendProtoBuf(tagName(hkProto.HkGetDataReq), req.SerializeToString())

        # Receive reply, check for errors and format return
        return self._recvHkValues()


    def _recvHkValues(self):
        """ Handle reply from server after sending HkGetDataReq or HkGetAllDataReq.

            :return: dictionary containing (name, (value, warning_sticky, error_sticky)) pairs
            :rtype: dict(str, list[str, bool, bool])
            :raises CommunicationError: if the server returns an error reply
        """

        parts = self.pubSocket.receiveProtoBuf()
        if parts.tag == tagName(ErrorRep):
            logger.error('Received an error reply when requesting hk values')
            CommunicationError.raiseFromProtoBuffer(parts.message)
        else:
            # Place the return values in a list
            rep = hkProto.HkData()
            rep.ParseFromString(parts.message)
            ret = {}
            for item in rep.item:
                ret[item.name] = (
                    getUnion(item, logger=logger), item.warning_sticky, item.error_sticky)

            return ret


    def getHkItems(self):
        """ Request a list of all registered HK items.

            :return: a dict with containing all hk items and their parameters
            :rtype: dict(str, dict)
            :raises CommunicationError: if the server returns an error reply

            This function returns a dictionary of all items. The dictionary is indexed by
            name and each value is a dictionary containing all the items parameters
        """

        # Create request buffer
        req = hkProto.HkGetItemsReq()

        # Send request
        self.pubSocket.sendProtoBuf(tagName(hkProto.HkGetItemsReq), req.SerializeToString())

        # Receive reply and check for errors.
        parts = self.pubSocket.receiveProtoBuf()
        if parts.tag == tagName(ErrorRep):
            logger.error('Received an error reply when requesting all hk item')
            CommunicationError.raiseFromProtoBuffer(parts.message)

        # Parse reply message
        rep = hkProto.HkGetItemsRep()
        rep.ParseFromString(parts.message)

        # Create return dict
        ret = {}
        for item in rep.item:
            itemDict = {}

            if item.HasField('auto'):
                itemDict['address'] = item.auto.address
                itemDict['period_ms'] = item.auto.period_ms
                itemDict['engineering'] = item.auto.engineering
            else:
                itemDict['address'] = None
                itemDict['period_ms'] = None
                itemDict['engineering'] = None

            if item.HasField('warning_action'):
                itemDict['warning_action'] = item.warning_action.action
                itemDict['warning_repeat'] = item.warning_action.repeat
            else:
                itemDict['warning_action'] = None
                itemDict['warning_repeat'] = None

            if item.HasField('error_action'):
                itemDict['error_action'] = item.error_action.action
                itemDict['error_repeat'] = item.error_action.repeat
            else:
                itemDict['error_action'] = None
                itemDict['error_repeat'] = None

            ret[item.name] = itemDict

        return ret


    def setHkValue(self, name, value):
        """ Update the value of an hk item.

            :param str name: the name of the hk item.
            :param value: the value to set the hk item to
            :type value: int or float
            :raises CommunicationError: if the server returns an error reply
        """

        if isinstance(value, int):

            # Use to a signed int if safe to cast.
            if numpy.can_cast(value, numpy.int64):
                self._setHkAsType(name, value, numpy.int64)
            else:
                self._setHkAsType(name, value, numpy.uint64)

        elif isinstance(value, float):
            self._setHkAsType(name, value, numpy.float64)
        else:
            logger.warning('Unsupported type for HK value assignment (%s)', str(type(value)))


    def setHkUnsigned(self, name, value):
        """ Update an HK item using an unsigned integer.

            :param str name: the name of the hk item
            :param int value: the unsingned integer to write to the hk item
            :raises CommunicationError: if the server returns an error reply
        """

        self._setHkAsType(name, value, numpy.uint64)


    def setHkSigned(self, name, value):
        """ Update an HK item using a signed integer.

            :param str name: the name of the hk item
            :param int value: the signed integer to write to the hk item
            :raises CommunicationError: if the server returns an error reply
        """

        self._setHkAsType(name, value, numpy.int64)


    def setHkDouble(self, name, value):
        """ Update an HK item using a double.

            :param str name: the name of the hk item
            :param float value: the double to write to the hk item
            :raises CommunicationError: if the server returns an error reply
        """

        self._setHkAsType(name, value, numpy.float64)


    def _setHkAsType(self, name, value, dtype):
        """ Update a hk item with a value of a specific data type.

            :param str name: the name of the hk item.
            :param value: the value to write to the hk item
            :param dtype: the numpy data type of the value. [uint64, int64, and float64]
            :type dtype: numpy.dtype
            :raises CommunicationError: if the server returns an error reply
        """

        # Create the protobuf request object.
        req = hkProto.HkSetDataReq()
        req.name = name

        if dtype == numpy.uint64:
            req.vuint = value

        elif dtype == numpy.int64:
            req.vint = value

        elif dtype == numpy.float64:
            req.vdbl = value

        else:
            logger.warning('Unsupported type for HK value assignment (%s)', str(dtype))

        # Send request
        self.pubSocket.sendProtoBuf(tagName(hkProto.HkSetDataReq), req.SerializeToString())

        # Receive reply and check for errors.
        parts = self.pubSocket.receiveProtoBuf()
        if parts.tag == tagName(ErrorRep):
            logger.error('Received an error reply when setting hk value')
            CommunicationError.raiseFromProtoBuffer(parts.message)
