""" This module contains functions that are used in multiple house keeping
    related modules.
"""

import numpy


def setUnion(item, value, logger=None):
    """ Assign a value to the right field in a protobuf oneof construct.

        :param item:                A protobuf message object with a value field that is a union
                                    (oneof) of an int64 (vint), uint64 (vuint) and a float64 (vdbl).
        :param int of float value:  The value to assign.
        :param Logger logger:       Optional logger instance to log failure.

        When the value is not suitable for any of the fields, no assignment is done,
        leaving the union with its previous value or None.
    """

    if isinstance(value, int):
        if numpy.can_cast(value, numpy.int64):
            item.vint = value
        else:
            item.vuint = value

    elif isinstance(value, float):
        item.vdbl = value

    elif logger is not None:
        logger.warning('The hk item has an unsupported type (%s)', type(value))


def getUnion(item, logger=None):
    """ Retreive the value from the right field in a protobuf oneof construct.

        :param item:   A protobuf message object with a value field that is a union (oneof)
                       of an int64 (vint), uint64 (vuint) and a float64 (vdbl).
        :param logger: Optional logger instance to log failure.

        :return:       The value of the field that was set. If none of the fields
                       is set, this function returns None.
        :rtype:        int or float or None
    """

    if item.HasField('vint'):
        return item.vint
    elif item.HasField('vuint'):
        return item.vuint
    elif item.HasField('vdbl'):
        return item.vdbl
    elif logger is not None:
        logger.warning('No value found in HkGetAllDataRep object for item %s', item.name)
        return None
