#!/usr/bin/env python3
# pylint: disable=too-many-lines
""" House keeping manager graphical user interface.

"""

import os
import sys
import logging
import functools
from datetime                 import datetime, timedelta

import yaml
from PyQt5                    import QtWidgets, uic
from PyQt5.QtCore             import QSize, QSettings, QRegExp, Qt
from PyQt5.QtWidgets          import QLabel
from PyQt5.QtGui              import QIntValidator, QRegExpValidator

from gssw.hk.dragwidget       import DragLabel, DragGroup, BackPlane
from gssw.hk.hkclient         import HkClient
from gssw.hal.halclient       import HalClient
from gssw.hal.hal             import Hal
from gssw.lib.sockets         import tagName
from gssw.lib.socketsqt       import NotifyingSubSocket
from gssw.lib.arrays.gridded  import GriddedArray
from gssw.lib.widgets.misc    import gsswUiWarning
from gssw.lib.errors          import CommunicationError
from gssw.proto.array_pb2     import GriddedArrayMsg
import gssw.proto.hk_pb2      as hkProto


logger = logging.getLogger(__name__)


class HkItem():
    """ This class holds the parameters of an HK item. """

    FORMATS = ['eng:gen', 'eng:sci', 'raw:hex', 'raw:dec', 'raw:oct', 'raw:bin']

    def __init__(self,  #pylint: disable=too-many-arguments
                 name, dataFormat, decimals, address=None, periodMs=None, engineering=False,
                 unit=None, warningAction=None, warningRepeat=None, errorAction=None,
                 errorRepeat=None):

        # attributes visible in the tree view
        self.name = name

        # other attributes
        self.dataFormat = dataFormat # (index in the representation combobox)
        self.decimals = decimals

        # optional attributes
        self.address = address
        self.periodMs = periodMs
        self.engineering = engineering
        self.unit = unit
        self.warningAction = warningAction
        self.warningRepeat = warningRepeat
        self.errorAction = errorAction
        self.errorRepeat = errorRepeat

    def splitName(self):
        """ Split the name into a devicename, registername and offset. """

        return self.address.split(':')

    def getDeviceName(self):
        """ return the devicename. """

        return self.splitName()[0]

    def getRegisterName(self):
        """ return the registername. """

        return self.splitName()[1]

    def getOffset(self):
        """ return the registername. """

        # if the offset is zero it might be omitted in the address
        try:
            return self.splitName()[2]
        except:
            return 0


class HkManager(QtWidgets.QMainWindow):
    """ HK Manager GUI class. """

    def __init__(self, name, config, app=None):
        """ Constructor.

            :param name: name of the component in the config file
            :param config: a gssw.lib.config.Config object
        """
        #pylint: disable=unused-argument
        # the name argument is included to give all GSSw components the same command line interface.

        QtWidgets.QMainWindow.__init__(self)

        # get linux signal handler and connect to slots
        self.signalHandler = None
        if app is not None:
            self.signalHandler = app.getSignalHandler()
            self.signalHandler.sigInt.connect(self.saveCloseConfiguration)
            self.signalHandler.sigTerm.connect(self.saveCloseConfiguration)

        # Variables
        self.items = {} # dict of HkItems stored in HkGuiItem. indexed by hk item name
        self.oldName = ""
        self.newItem = None
        self.addItem = None
        self.browse = None
        
        # used to ignore hk updates for 1s after stopping the server
        self.stopTime = datetime.now() - timedelta(seconds=1)

        logger.debug('Starting HK manager')

        # retrieve the system configuration
        self.config = config

        # initialize the HAL client
        self.hal       = Hal()
        self.halClient = HalClient(self.hal, self.config)
        self.halClient.requestHal()

        # Initialize the HK client
        serverCfg = self.config.componentParameters('hk_server')
        hkHost = serverCfg['hostname']
        hkRepPort = serverCfg['sockets']['hk-rep']
        hkPubPort = serverCfg['sockets']['hk-pub']
        hkPubAddress = 'tcp://%s:%d' % (hkHost, hkPubPort)

        self.hkClient = HkClient('tcp://%s:%d' % (hkHost, hkRepPort), hkPubAddress)

        settingsFileName = self.config.componentParameters(name).get('settings')
        if settingsFileName is None:
            settingsFileName = '{}.yaml'.format(name)
            logger.warning("No 'settings' file in the configuration file, using default '%s'",
                           settingsFileName)

        self.settingsFile = self.config.asAbsolutePath(settingsFileName)

        # Retrieve the hk server address from the config
        hkPubPort = serverCfg['sockets']['hk-pub']
        hkPubAddress = 'tcp://%s:%d' % (hkHost, hkPubPort)

        # Create a sub socket and connect the signal to the handler
        self.subSocket = NotifyingSubSocket()
        self.subSocket.createZmqSocket()
        self.subSocket.connectSocket(hkPubAddress)
        self.subSocket.subscribe() # subscribe to all
        self.subSocket.sigReadyReceive.connect(self.hkUpdateReceived)

        # Initialize the gui
        self.initMainGui()
        self.initAddItemGui()
        self.initBrowseGui()
        self.initLegendGui()


    def show(self):
        """ Show user interface """
        settings = QSettings()
        geom = settings.value('geometry')
        if geom:
            self.ui.restoreGeometry(geom)
        else:
            logger.debug('Unable to restore window geometry.')
        self.ui.show()


    def closeEvent(self, event): # pylint: disable=unused-argument
        """ This function is called when the main window is closed. """

        logger.info('Closing house keeping manager')
        settings = QSettings()
        settings.setValue('geometry', self.ui.saveGeometry())


    def initMainGui(self):
        """ Initialize the main window GUI. """

        # Gets the directory of this module
        modulePath = os.path.dirname(os.path.abspath(__file__))
        guiPath = os.path.join(modulePath, 'gui')

        logger.debug('Starting user interface')
        self.ui        = uic.loadUi(os.path.join(guiPath, 'hkMainWindow.ui'))
        self.addItemUi = uic.loadUi(os.path.join(guiPath, 'hkAddItem.ui'))
        self.browseUi  = uic.loadUi(os.path.join(guiPath, 'hkBrowse.ui'))
        self.legendUi  = uic.loadUi(os.path.join(guiPath, 'hkLegend.ui'))
        self.addItemUi.setModal(True)
        self.browseUi.setModal(True)
        self.ui.closeEvent = self.closeEvent

        # Add custom widget
        self.backPlane = BackPlane(self)
        self.backPlane.setMinimumSize(QSize(1920, 1200))
        self.ui.scrollArea.setWidget(self.backPlane)
        self.ui.scrollArea.setWidgetResizable(True)

        if self.hkClient.getHkState():
            self.ui.act_stop.setEnabled(True)
            self.ui.act_start.setEnabled(False)
        else:
            self.ui.act_stop.setEnabled(False)
            self.ui.act_start.setEnabled(True)

        # ui listeners
        self.ui.act_save.triggered.connect(self.saveHkConfiguration)
        self.ui.act_load.triggered.connect(self.loadHkConfiguration)
        self.ui.act_quit.triggered.connect(sys.exit)
        self.ui.act_start.triggered.connect(self.startHkClicked)
        self.ui.act_stop.triggered.connect(self.stopHkClicked)
        self.ui.act_limit.triggered.connect(self.resetLimitsClicked)
        self.ui.act_add_item.triggered.connect(
            functools.partial(self.launchAddHkItem, groupName=None))
        self.ui.act_edit_item.triggered.connect(self.launchEditHkItem)
        self.ui.act_add_group.triggered.connect(self.createHkGroup)
        self.ui.act_sync.triggered.connect(self.syncWithServer)
        self.ui.act_legend.triggered.connect(self.legendUi.exec_)

        # check if the server is active
        if self.hkClient.getHkState():
            self.statusLabel = QLabel('Housekeeping: active')
        else:
            self.statusLabel = QLabel('Housekeeping: not active')
        self.ui.statusbar.addWidget(self.statusLabel)
        self.loadHkConfiguration()


    def initAddItemGui(self):
        """ Initialize the add item dialog. """

        # Dialog variables
        self.newItem = True # Differentiate between creating a new item
                            # and editing an existing one

        # Clear fields
        self.addItemUi.line_item.setText('')
        self.addItemUi.line_device.setText('')
        self.addItemUi.line_register.setText('')
        self.addItemUi.line_offset.setText('0')

        # Validators
        offsetValidator = QIntValidator(0, 1000000)
        self.addItemUi.line_offset.setValidator(offsetValidator)
        itemValidator = QRegExpValidator(QRegExp('[A-Za-z0-9_:]*'))
        self.addItemUi.line_item.setValidator(itemValidator)

        # ui listeners
        self.addItemUi.btn_browse.clicked.connect(self.browseUi.exec_)
        self.addItemUi.btn_ok.clicked.connect(self.createHkItem)
        self.addItemUi.btn_cancel.clicked.connect(self.addItemUi.close)
        self.addItemUi.line_device.textChanged.connect(self.checkFormComplete)
        self.addItemUi.line_register.textChanged.connect(self.checkFormComplete)
        self.addItemUi.line_item.textChanged.connect(self.checkFormComplete)
        self.addItemUi.line_offset.textChanged.connect(self.lineOffsetChanged)
        self.addItemUi.rbtn_register.clicked.connect(self.addHkItemTypeSelect)
        self.addItemUi.rbtn_software.clicked.connect(self.addHkItemTypeSelect)
        self.addItemUi.rbtn_register.clicked.connect(self.checkFormComplete)
        self.addItemUi.rbtn_software.clicked.connect(self.checkFormComplete)
        self.addItemUi.cbox_format.currentIndexChanged.connect(self.formatChanged)


    def initBrowseGui(self):
        """ Initialize the browse GUI. """

        # fill the drop-down menu with devices from the hal
        deviceNames = self.hal.getDeviceNames()
        deviceNames.sort()
        availableComponents = list(self.config["components"].keys())

        for deviceNum, deviceName in enumerate(deviceNames):

            self.browseUi.box_devices.addItem(deviceName, deviceName)
            deviceObject = self.hal.getDevice(deviceName)

            # annotate devices that are unaccessible in the combobox
            if deviceObject.daemon not in availableComponents:
                self.browseUi.box_devices.model().item(deviceNum).setText(
                    self.browseUi.box_devices.model().item(deviceNum).text() +
                    " (daemon not in configuration)")

        # Set table properties
        #    sort 1st column by default
        self.browseUi.table_registers.horizontalHeader().setSortIndicator(0, Qt.AscendingOrder)
        self.browseUi.table_registers.setSelectionBehavior(
            QtWidgets.QAbstractItemView.SelectRows)
        self.browseUi.table_registers.setSelectionMode(
            QtWidgets.QAbstractItemView.SingleSelection)
        self.browseUi.table_registers.setEditTriggers(
            QtWidgets.QAbstractItemView.NoEditTriggers)

        # Initialize the table
        self.fillRegisterTable(self.browseUi.box_devices.itemData(0))

        # ui listeners
        self.browseUi.box_devices.currentIndexChanged[int].connect(self.deviceIndexChanged)
        self.browseUi.btn_ok.clicked.connect(self.browseOk)
        self.browseUi.table_registers.doubleClicked.connect(self.browseOk)
        self.browseUi.btn_cancel.clicked.connect(self.browseUi.close)
        self.browseUi.table_registers.itemSelectionChanged.connect(self.browserSelection)


    def initLegendGui(self):
        """ Initialize the color code legend. """

        self.legendUi.lbl_inactive.setStyleSheet('QLabel {color: %s;}' % DragLabel.COLOR_INACTIVE)
        self.legendUi.lbl_timeout.setStyleSheet('QLabel {color: %s;}' % DragLabel.COLOR_TIMEOUT)
        self.legendUi.lbl_noLimits.setStyleSheet('QLabel {color: %s;}' % DragLabel.COLOR_NO_CHECK)
        self.legendUi.lbl_ok.setStyleSheet('QLabel {color: %s;}' % DragLabel.COLOR_OK)
        self.legendUi.lbl_warning.setStyleSheet('QLabel {color: %s;}' % DragLabel.COLOR_WARNING)
        self.legendUi.lbl_error.setStyleSheet('QLabel {color: %s;}' % DragLabel.COLOR_ERROR)
        self.legendUi.lbl_invalid.setStyleSheet('QLabel {color: %s;}' % DragLabel.COLOR_INVALID)


    def deviceIndexChanged(self, index):
        """ slot to deal with change of device index in the UI """
        deviceName = self.browseUi.box_devices.itemData(index)
        self.fillRegisterTable(deviceName)


    def fillRegisterTable(self, deviceName):
        """ Fill the browser table with registers for a certain device. """

        table = self.browseUi.table_registers

        registers = sorted(self.hal.getRegistersInfo(deviceName))
        deviceType = self.hal.getDeviceType(deviceName)

        # Remove any rows that are already present in the table
        table.setRowCount(0)

        # Disable sorting while creating the table, otherwise the records are immediately sorted
        # when they are created, which messes up the index
        table.setSortingEnabled(False)

        # Create a row in the table for each register
        i = 0
        for registerName in registers:

            register = deviceType.getRegister(registerName)

            # No house keeping for block registers
            if register.block:
                continue

            table.insertRow(i)

            table.setItem(i, 0, QtWidgets.QTableWidgetItem(register.name))
            table.setItem(i, 1, QtWidgets.QTableWidgetItem(register.type))
            table.setItem(i, 2, QtWidgets.QTableWidgetItem(register.unit))
            table.setItem(i, 3, QtWidgets.QTableWidgetItem(str(register.access)))
            table.setItem(i, 4, QtWidgets.QTableWidgetItem(str(register.width)))
            table.setItem(i, 5, QtWidgets.QTableWidgetItem(str(register.block)))
            table.setItem(i, 6, QtWidgets.QTableWidgetItem(str(hex(register.address))))
            table.setItem(i, 7, QtWidgets.QTableWidgetItem(str(register.maxOffset)))
            table.setItem(i, 8, QtWidgets.QTableWidgetItem(str(register.description)))

            i += 1

        # Re-enable sorting, after all the inserting is done
        table.setSortingEnabled(True)

        table.resizeColumnsToContents()

        # disable the entire table if the device is not accessible
        availableComponents = list(self.config["components"].keys())
        if self.hal.getDevice(deviceName).daemon not in availableComponents:
            table.setEnabled(False)
        else:
            table.setEnabled(True)


    def createHkGroup(self):
        """ Add an empty hk group. """

        hkGroup = DragGroup(self.backPlane)
        hkGroup.newName()
        self.backPlane.insertItem(hkGroup)


    def createHkItem(self, groupName=None):
        """ Create a new HK item.

            This function gets the hk items parameters from the addHkItem dialog.
        """

        # get the HK items parameters from the corresponding fields
        (name, address, periodMs, engineering, unit, warningAction, warningRepeat,
            errorAction, errorRepeat) = self.getItemParameters()

        if name in self.items:
            gsswUiWarning("The name %s already exists." % name)
            return

        hkItem = HkItem(name, self.addItemUi.cbox_format.currentIndex(),
                        self.addItemUi.sbox_decimals.value(), address=address,
                        periodMs=periodMs, engineering=engineering, unit=unit,
                        warningAction=warningAction, warningRepeat=warningRepeat,
                        errorAction=errorAction, errorRepeat=errorRepeat)

        self.addHkItem(hkItem, groupName=groupName)
        self.registerItem(hkItem)

        # Close the dialog
        self.addItemUi.close()


    def modifyHkItem(self, oldName):
        """ Modify an existing HK item. """

        # Get the new item parameters from the dialog fields
        (newName, address, periodMs, engineering, unit, warningAction, warningRepeat,
            errorAction, errorRepeat) = self.getItemParameters()

        if newName != oldName and newName in self.items:
            gsswUiWarning("The name %s already exists." % newName)
            return

        newItem = HkItem(newName, self.addItemUi.cbox_format.currentIndex(),
                         self.addItemUi.sbox_decimals.value(), address=address,
                         periodMs=periodMs, engineering=engineering, unit=unit,
                         warningAction=warningAction, warningRepeat=warningRepeat,
                         errorAction=errorAction, errorRepeat=errorRepeat)

        oldItem = self.backPlane.getItem(oldName)

        # if the item is in a group collect the group name and the index in the group
        if isinstance(oldItem.parent, DragGroup):
            groupName = oldItem.parent.name
            index = oldItem.index
            position = None
        else:
            groupName = None
            index = None
            position = oldItem.position

        # temporary disable listening to derver state changes
        #self.subSocket.unsubscribe('HkStateChangedPub')
        self.subSocket.unsubscribe()

        # remove the old item
        self.removeHkItems([oldItem])
        self.hkClient.removeHkItem(oldItem.name)

        # Create the modified item and register with the server
        self.addHkItem(newItem, groupName=groupName, index=index, position=position)
        self.registerItem(newItem)

        # re-disable listening to derver state changes
        #self.subSocket.subscribe('HkStateChangedPub')
        self.subSocket.subscribe('')

        # sync in case server updated were missed
        self.syncWithServer()

        # Close the dialog
        self.addItemUi.close()


    def registerItem(self, hkItem):
        """ Register a hk item with the server. """

        try:
            if hkItem.address is None:
                self.hkClient.addHkItem(hkItem.name)
            else:
                self.hkClient.addHkItem(hkItem.name, address=hkItem.address,
                    periodMs=hkItem.periodMs, engineering=hkItem.engineering, unit=hkItem.unit,
                    warningAction=hkItem.warningAction, warningRepeat=hkItem.warningRepeat,
                    errorAction=hkItem.errorAction, errorRepeat=hkItem.errorRepeat)
        except CommunicationError as e:
            logger.error('failed to register a HK item with the server: %s', str(e))
            return


    def addHkItem(self, hkItem, groupName=None, index=None, position=None, noPlace=False):
        """ Add a HK item to the manager.

            :param hkItem:    HkItem object
            :param groupName: name of the group this HK item belongs to
            :param position:  place of this HK item on the UI
            :param noPlace:   True if the paramater is not to be added to the UI
                              (only register at server)
        """

        # Add the item to the dictionary
        self.items[hkItem.name] = hkItem

        # insert in the widget
        if groupName is None:
            parent = self.backPlane
            arg    = position
        else:
            parent = self.backPlane.getGroup(groupName)
            arg    = index

        if not noPlace:
            newLabel = DragLabel(hkItem.name, 0, parent, hkItem, self.hal)
            parent.insertItem(newLabel, arg)


    def removeHkItems(self, items):
        """ Remove HK items and/or groups from the manager only.

            items - a list of DragLabel or DragGroup objects
        """

        for item in items:

            name = item.name
            self.backPlane.removeItem(name)

            if isinstance(item, DragLabel):
                del self.items[name]


    def resetHkLimits(self, items):
        """ Clear the limit crossing sticky bits for HK items at the server.

            items - a list of DragLabel objects
        """

        names = [x.name for x in items]
        self.hkClient.resetHkLimits(names)

        # immediately update appearance
        for item in items:
            item.warningSticky = False
            item.errorSticky = False
            item.updateAppearance()


    def getItemParameters(self):
        """ Get the HK item parameters from the addHkItem screen fields.

            This function returns the name, address, update period, representation, value
            limits and unit. For software based hk items, all except the name are None.
        """

        address = None
        periodMs = None
        engineering = None
        warningAction = None
        errorAction = None
        warningRepeat = False
        errorRepeat = False
        name = self.addItemUi.line_item.text()

        # For register based hk items, determine the parameters from the dialog field
        if self.addItemUi.rbtn_register.isChecked():

            # concatenate address
            address = (self.addItemUi.line_device.text() + ':' +  # register address
                       self.addItemUi.line_register.text())
            offset = self.addItemUi.line_offset.text()

            # Always have offset field
            if not offset:
                offset = '0'
            address += ':' + offset

            # determine update period
            unit = self.addItemUi.cbox_period.currentText()
            period = self.addItemUi.sbox_period.value()
            if unit == 'miliseconds':
                periodMs =  period
            elif unit == 'seconds':
                periodMs = period * 1000
            elif unit == 'minutes':
                periodMs = period * 60 * 1000
            else:
                logger.error('Invalid update period unit: %s', unit)
                sys.exit(1)

            # determine representation
            engineering = (self.addItemUi.cbox_format.currentText() == 'Eng: Generic' or
                           self.addItemUi.cbox_format.currentText() == 'Eng: Scientific')

            # get the items limits from the HAL. and the limit actions if available
            if self.addItemUi.grp_limitActions.isEnabled():

                if self.addItemUi.line_warning.text() != '':
                    warningAction = self.addItemUi.line_warning.text()

                if self.addItemUi.line_error.text() != '':
                    errorAction = self.addItemUi.line_error.text()

                warningRepeat = self.addItemUi.btn_warningRepeat.isChecked()
                errorRepeat = self.addItemUi.btn_errorRepeat.isChecked()

            unit = self.hal.getRegisterInfo(self.addItemUi.line_device.text(),
                                            self.addItemUi.line_register.text()).unit

        else: # This is a software-housekeeping item.
            if self.addItemUi.cbox_format.currentText() == 'Raw: Float':
                unit = self.addItemUi.line_unit.text()
            else:
                unit = None

        return (name, address, periodMs, engineering, unit, warningAction, warningRepeat,
                errorAction, errorRepeat)


    def launchAddHkItem(self, groupName=None):
        """ GUI listener for the add hk item button. """

        self.newItem = True
        self.addItemUi.line_item.setText("")
        self.addItemUi.line_device.setText("")
        self.addItemUi.line_register.setText("")
        self.addItemUi.line_offset.setText("")

        self.addItemUi.setWindowTitle("New housekeeping item")
        self.addItemUi.btn_ok.clicked.disconnect()
        self.addItemUi.btn_ok.clicked.connect(
            functools.partial(self.createHkItem, groupName))
        self.addItem = self.addItemUi.exec_()


    def launchEditHkItem(self, name):
        """ Change the parameters of an HK item. """

        self.newItem = False

        # Get the hk item to edit
        item = self.items[name]

        # store the old name for removal
        self.oldName = item.name

        # fill the dialog fields with the current hk parameters
        self.addItemUi.line_item.setText(item.name)
        self.addItemUi.sbox_decimals.setValue(item.decimals)
        self.addItemUi.cbox_format.setCurrentIndex(item.dataFormat)
        self.addItemUi.line_unit.setText(item.unit)

        if item.address is not None: # register based

            self.addItemUi.rbtn_register.setChecked(True)

            device, register, offset = item.address.split(':')
            self.addItemUi.line_device.setText(device)
            self.addItemUi.line_register.setText(register)
            self.addItemUi.line_offset.setText(offset)

            if item.periodMs % 60000 == 0: # set period unit to minutes
                self.addItemUi.cbox_period.setCurrentIndex(2)
                self.addItemUi.sbox_period.setValue(item.periodMs / 60 / 1000)
            elif item.periodMs % 1000 == 0: # set period unit to seconds
                self.addItemUi.cbox_period.setCurrentIndex(1)
                self.addItemUi.sbox_period.setValue(item.periodMs / 1000)
            else: # set period unit to microseconds
                self.addItemUi.cbox_period.setCurrentIndex(0)
                self.addItemUi.sbox_period.setValue(item.periodMs)

            # enable the limit actions group if the register has limits defined in the HAL
            device = self.hal.getDevice(item.getDeviceName())
            limits = self.hal.getRegisterLimits(item.getDeviceName(), device.deviceType,
                                                device.calibration, item.getRegisterName())
            self.addItemUi.grp_limitActions.setEnabled(not limits is None)

        else:
            self.addItemUi.rbtn_software.setChecked(True)

        # Adapt dialog fields.
        self.formatChanged()
        self.addHkItemTypeSelect(False)
        self.checkFormComplete()

        self.addItemUi.setWindowTitle("Edit housekeeping item")
        self.addItemUi.btn_ok.clicked.disconnect()
        self.addItemUi.btn_ok.clicked.connect(functools.partial(self.modifyHkItem, name))
        self.addItem = self.addItemUi.exec_()


    def browseOk(self):
        """ GUI listener for the ok button in the browse dialog. """

        deviceName = self.browseUi.box_devices.currentText()
        registerName = self.browseUi.table_registers.item(
            self.browseUi.table_registers.selectionModel().selectedRows()[0].row(), 0).text()
        maxOffset =self.browseUi.table_registers.item(
            self.browseUi.table_registers.selectionModel().selectedRows()[0].row(), 7).text()

        self.addItemUi.line_device.setText(self.browseUi.box_devices.currentText())
        self.addItemUi.line_offset.setText('0')
        self.addItemUi.line_register.setText(registerName)

        # update the offset validator with the max offset for this specific register
        self.addItemUi.line_offset.validator().setTop(int(maxOffset))

        if int(maxOffset) > 0:
            self.addItemUi.line_item.setText('%s:%d' % (registerName, 0))
        else:
            self.addItemUi.line_item.setText(registerName)

        # if no conversions are available for this register, disable engeneering values
        if self.hal.getRegisterConversion(deviceName, None, None, registerName):
            self.addItemUi.cbox_format.model().item(0).setEnabled(True)
            self.addItemUi.cbox_format.model().item(1).setEnabled(True)
            self.addItemUi.cbox_format.setCurrentIndex(0)
        else:
            self.addItemUi.cbox_format.model().item(0).setEnabled(False)
            self.addItemUi.cbox_format.model().item(1).setEnabled(False)
            self.addItemUi.cbox_format.setCurrentIndex(2)

        # fetch the register unit from the HAL and fill in the unit field
        unit = self.hal.getRegisterInfo(deviceName, registerName).unit
        self.addItemUi.line_unit.setText(unit)

        # enable the limits groupbox if the selected register has limits defined in the hal
        device = self.hal.getDevice(deviceName)
        limits = self.hal.getRegisterLimits(deviceName, device.deviceType, device.calibration,
                                            registerName)

        if limits is None:
            self.addItemUi.grp_limitActions.setEnabled(False)
            self.addItemUi.grp_limitActions.setTitle('Limit Crossing Actions '
                                                     '(no limits defined in HAL)')
        else:
            self.addItemUi.grp_limitActions.setEnabled(True)
            self.addItemUi.grp_limitActions.setTitle('Limit Crossing Actions')

        self.browseUi.close()


    def lineOffsetChanged(self):
        """ Slot: deal with line offset changes """

        # append the register offset to the HK items name if it is not 0
        if self.addItemUi.line_offset.text() != '':
            if self.addItemUi.line_offset.validator().top() == 0:
                self.addItemUi.line_item.setText(
                    self.addItemUi.line_item.text().split(':')[0])
            else:
                self.addItemUi.line_item.setText(
                    self.addItemUi.line_item.text().split(':')[0]
                    + ':' + self.addItemUi.line_offset.text())

        self.checkFormComplete()


    def checkFormComplete(self):
        """ Check if all field are filled in the add hk item dialog. """

        # Make the OK button enabled when the form is filled in completely.
        if (self.addItemUi.rbtn_software.isChecked() and
                self.addItemUi.line_item.text() != ''):

            self.addItemUi.btn_ok.setEnabled(True)

        elif (self.addItemUi.line_device.text() != '' and
                self.addItemUi.line_register.text() != '' and
                self.addItemUi.line_item.text() != ''):

            self.addItemUi.btn_ok.setEnabled(True)

        else:
            self.addItemUi.btn_ok.setEnabled(False)


    def formatChanged(self):
        """ Adapt decimal combobox and unit line-edit based on floating-point value. """

        # determine representation
        decimals = ('Eng' in self.addItemUi.cbox_format.currentText() or
                    self.addItemUi.cbox_format.currentText() == 'Raw: Float')

        # decimal combobox
        self.addItemUi.label_6.setEnabled(decimals)
        self.addItemUi.sbox_decimals.setEnabled(decimals)

        # unit line-edit
        if self.addItemUi.rbtn_register.isChecked():
            eng = 'Eng' in self.addItemUi.cbox_format.currentText()

            self.addItemUi.label_9.setEnabled(eng)
            self.addItemUi.line_unit.setEnabled(eng)
        else:
            rfloat = self.addItemUi.cbox_format.currentText() == 'Raw: Float'

            self.addItemUi.label_9.setEnabled(rfloat)
            self.addItemUi.line_unit.setEnabled(rfloat)


    def browserSelection(self):
        """ Adapt GUI based on what is selected in the register browse table. """

        # The ok button becomes availale only when a register is selected.
        if self.browseUi.table_registers.selectedItems() != []:
            self.browseUi.btn_ok.setEnabled(True)

        else:
            self.browseUi.btn_ok.setEnabled(False)


    def addHkItemTypeSelect(self, setIndex=True):
        """ Adapt the add hk item dialog based on what type of hk item is selected.

            (register or non-register based type)
        """

        # Some fields only apply to register based HK items.
        # Additionally, SW items do not have engineering values because they have no conversions
        if self.addItemUi.rbtn_register.isChecked():
            self.addItemUi.grp_registerSelect.setEnabled(True)
            self.addItemUi.sbox_period.setEnabled(True)
            self.addItemUi.cbox_period.setEnabled(True)
            self.addItemUi.grp_limitActions.setEnabled(True)

            # Don't show register engineering values without conversion.
            deviceName = self.addItemUi.line_device.text()
            registerName = self.addItemUi.line_register.text()
            if deviceName and registerName and \
               self.hal.getRegisterConversion(deviceName, None, None, registerName):
                self.addItemUi.cbox_format.model().item(0).setEnabled(True)
                self.addItemUi.cbox_format.model().item(1).setEnabled(True)
            else:
                self.addItemUi.cbox_format.model().item(0).setEnabled(False)
                self.addItemUi.cbox_format.model().item(1).setEnabled(False)
            if setIndex:
                self.addItemUi.cbox_format.setCurrentIndex(0)

            self.addItemUi.line_unit.setReadOnly(True)
        else:
            self.addItemUi.grp_registerSelect.setEnabled(False)
            self.addItemUi.sbox_period.setEnabled(False)
            self.addItemUi.cbox_period.setEnabled(False)
            self.addItemUi.grp_limitActions.setEnabled(False)
            self.addItemUi.cbox_format.model().item(0).setEnabled(False)
            self.addItemUi.cbox_format.model().item(1).setEnabled(False)
            if setIndex:
                self.addItemUi.cbox_format.setCurrentIndex(2)
            self.addItemUi.line_unit.setReadOnly(False)


    def saveHkConfiguration(self):
        """ Save the current HK configuration to a YAML file. """

        logger.info("Saving configuration to %s", self.settingsFile)
        groups = {} # dict of group name, position pairs
        labels = {} # dict of (top level) item name, position pairs
        order = {}  # dict of group name, child order pairs
        colors = {} # dict of groups colors

        for name in self.backPlane.itemMap:
            item = self.backPlane.itemMap[name]
            if isinstance(item, DragLabel):
                labels[name] = item.parent.gridPos(item.position)
            elif isinstance(item, DragGroup):
                groups[name] = item.parent.gridPos(item.position)
                order[name] = item.order
                colors[name] = item.color

        d = {}
        d['items']  = self.items
        d['labels'] = labels
        d['groups'] = groups
        d['order']  = order
        d['colors'] = colors

        with open(self.settingsFile, 'w') as outfile:
            yaml.dump(d, outfile, default_flow_style=True)


    def saveCloseConfiguration(self):
        """ Save and close """

        settings = QSettings()
        settings.setValue('geometry', self.ui.saveGeometry())

        self.saveHkConfiguration()
        sys.exit(1)


    def loadHkConfiguration(self):
        """ Load a HK configuration from a YAML file. """

        logger.info("Loading configuration from %s", self.settingsFile)

        # Clear all items from GUI and internal data structure.
        self.removeHkItems(self.backPlane.itemMap.copy().values())
        self.items = {}

        try:
            with open(self.settingsFile) as infile:
                d = yaml.load(infile, Loader=yaml.Loader)
        except Exception as e:
            logger.warning('Could not load YAML file (%s)', str(e))
            d = {}

        try:
            self.items = d['items']
            labels     = d['labels']
            groups     = d['groups']
            order      = d['order']
            colors     = d['colors']
        except Exception as e:
            logger.warning('Could not parse YAML file (%s)', str(e))
            return

        # Add the loaded hk items
        for name in self.items:
            self.addHkItem(self.items[name], noPlace=True)

        # Restore top level hk items
        for name in labels:
            logger.debug('load label %s @ (%d, %d)', name, labels[name].x(), labels[name].y())

            dragLabel = DragLabel(name, 0, self.backPlane, self.items[name], self.hal)
            self.backPlane.placeItem(dragLabel, labels[name])

        # restore groups including their child items
        for groupName in groups:
            logger.debug('load group %s @ (%f, %f)',
                groupName, groups[groupName].x(), groups[groupName].y())

            # create group
            group = DragGroup(self.backPlane, name=groupName, color=colors[groupName])
            self.backPlane.placeItem(group, groups[groupName])

            # add children
            for childName in order[groupName]:
                group.insertItem(DragLabel(childName, 0, group, self.items[childName], self.hal))

        self.syncWithServer(initialize=True)


    def syncWithServer(self, initialize=False):
        """ Synchronize hk items between the server and manager.

            initialize - when true the manager copies its state to the server. This is used
                         when the manager boots. When false the manager follows the servers state
                         and deletes items when the server does so, This is done for all subsequent
                         server updates.
        """

        # request all items from server
        serverItems = self.hkClient.getHkItems()

        # update all items with parameters defined by the server
        for serverItemName, serverItem in serverItems.items():

            if serverItemName in self.items:# known items
                dataFormat = self.items[serverItemName].dataFormat
                decimals   = self.items[serverItemName].decimals
                unit       = self.items[serverItemName].unit
            else: # unknown items
                dataFormat = 0 if serverItem['engineering'] else 2
                decimals   = 3
                unit       = None
                address    = serverItem['address'] if serverItem['address'] else 'software HK item'
                logger.info('Add hk item from server: %s (%s)', serverItemName, address)

            newItem = HkItem(serverItemName, dataFormat, decimals,
                             address=serverItem['address'],
                             periodMs=serverItem['period_ms'],
                             engineering=serverItem['engineering'],
                             unit=unit,
                             warningAction=serverItem['warning_action'],
                             warningRepeat=serverItem['warning_repeat'],
                             errorAction=serverItem['error_action'],
                             errorRepeat=serverItem['error_repeat'])

            if serverItemName in self.items:
                self.items[serverItemName] = newItem # update known items
            else:
                self.addHkItem(newItem) # add unkown items

        # register items unknown to the server
        for itemName, item in [(n, i) for n, i in self.items.items() \
                               if n not in serverItems.keys()]:
            if initialize:
                self.registerItem(item)
            else:
                self.removeHkItems([item])
                del self.items[itemName]


    def hkUpdateReceived(self):
        """ Slot that handles hk updates.

            This slot is triggered by the sub socket.
            When an hk update is received from the server, this function reads
            the message and updates the table.

            Note that the manager does not use the HkClient callback interface because that filters
            out HkStateChangedPub messages.
        """

        # Receive the topic, tag and protobuf
        parts = self.subSocket.receiveProtoBufWithTopic()

        if parts.tag == tagName(hkProto.HkStateChangedPub):

            logger.info('HK server state change')
            pub = hkProto.HkStateChangedPub()
            pub.ParseFromString(parts.message)

            self.syncWithServer()

            # change item appearance if server was stopped
            if pub.state == hkProto.HK_STOPPED:
                self.hkStopped()

        elif parts.tag == tagName(GriddedArrayMsg):

            # Parse the stream and extract the value.
            pBuffer = GriddedArrayMsg()
            pBuffer.ParseFromString(parts.message)
            gArray = GriddedArray.createFromProtoBuf(pBuffer)

            name = gArray.name[3:] # ignore HK_ prefix
            value = gArray.data[0]
            stickyBits = yaml.safe_load(gArray.notes)
            logger.debug('Hk update %s = %s', name, str(value))

            if name in self.items:
                # discard updates for 1s after stopping the server (for correct drawing)
                if (datetime.now() - self.stopTime).total_seconds() < 1:
                    return
                self.backPlane.updateValue(name, value, stickyBits['warning'], stickyBits['error'])
            else:
                logger.warning('received a HK update for an item unknown HK item %s', name)
                self.syncWithServer()
        else:
            raise CommunicationError('received unknown protobuf message: %s' % parts.tag)


    def startHkClicked(self):
        """ Start housekeeping at the server. """

        self.hkClient.startHk()
        self.ui.act_start.setEnabled(False)
        self.ui.act_stop.setEnabled(True)
        self.statusLabel.setText('Housekeeping: active')


    def stopHkClicked(self):
        """ Stop housekeeping at the server. """

        self.hkClient.stopHk()


    def hkStopped(self):
        """ Handle GUI changes due to hk being stopped """

        self.stopTime = datetime.now()
        self.backPlane.setInactive() # set all GUI items color to inactive
        self.ui.act_start.setEnabled(True)
        self.ui.act_stop.setEnabled(False)
        self.statusLabel.setText('Housekeeping: not active')


    def resetLimitsClicked(self):
        """ Reset all limit crossings at the server """

        self.hkClient.resetHkLimits(self.items.keys())
        logger.info('Cleared all limit crossings')
