""" Python wrapper for QApplication. Sets the style sheet and the window icon. """
#pylint: disable=too-many-statements

import sys
import signal
import logging
import os.path

from PyQt5.QtWidgets          import QApplication, QStyleFactory
from PyQt5.QtGui              import QIcon
from PyQt5.QtCore             import QObject, pyqtSignal, QTimer

from gssw.lib.config          import Config
from gssw.lib.misc            import normRealPath
from gssw.lib.typing          import check_class
from gssw.lib.version         import VERSION_STRING
from gssw.lib.widgets.colors  import parsePalette
from gssw.lib.widgets.misc    import gsswExceptHook

logger = logging.getLogger(__name__)

##################################################################################################
# Linux signal handler
##################################################################################################

class SignalHandler(QObject):
    """ Deals with Linux signals SIGINT and SIGTERM """

    sigInt  = pyqtSignal()
    sigTerm = pyqtSignal()

    def handle(self, sig, _frame):
        """ Deals with the actual signal """
        logger.info("Received signal %d", sig)

        if sig == signal.SIGINT:
            logger.debug("emitting sigInt")
            self.sigInt.emit()
        elif sig == signal.SIGTERM:
            logger.debug("emitting sigTerm")
            self.sigTerm.emit()
        elif sig == signal.SIGSEGV:
            print ("Application segfaulted!")
            sys.exit(1) # In case of segfault, there is not much we can do other than exit
        else:
            # not interessed in other signals
            pass


##################################################################################################
# Qt Application base class
##################################################################################################

class Application(QApplication):
    """ Application class encapsulates all functions to interface with the hardware daemon. """

    def __init__(self,
                 config,
                 name="",
                 org="SRON",
                 domain="www.sron.nl",
                 style=True):
        """ Constructor application wrapper

            :param config: gssw.lib.config.Config object. If style is True, the location
                of the style sheet is take from this configuration.
            :param name:   application name
            :param org:    organization
            :param domain: organization domain
            :param style: boolean indicating whether to apply a stylesheet
        """
        check_class(config, Config)

        # Even though we don't use Qt parameters, qt requires that args contains at least one valid
        # character string (see:  http://doc.qt.io/qt-5/qapplication.html#QApplication)
        #
        # Note: I've encountered a case where not having arguments causes core dumps when loading
        #       gui themes.
        qAppArgs = [name if name else 'python3']
        QApplication.__init__(self, qAppArgs)

        # Check that the name is not empty. They might be used to get
        # a QSettings object, but if the name is empty the settings will apply to all application
        # that have the same organization.
        assert name, "Application name undefined: {!r}".format(name)

        self.setApplicationName(name)
        self.setOrganizationName(org)
        self.setOrganizationDomain(domain)
        self.setApplicationDisplayName("Egress %s" % VERSION_STRING)

        # If the component receives SIGINT or SIGTERM, handle these signals gracefully
        self.signalHandler = SignalHandler()

        signal.signal(signal.SIGINT,  self.signalHandler.handle)
        signal.signal(signal.SIGTERM, self.signalHandler.handle)
        signal.signal(signal.SIGSEGV, self.signalHandler.handle)
        logger.info("Installed Linux signal handlers")

        # Start a void timer to make sure that Qt gives the context back to the Python interpreter
        # at least 10 times every second. This is done so that Linux signals are caught while the
        # Qt event loop is busy waiting for events. The Python signal handler is only called between
        # atomic python operations. If the Qt event queue is empty the Python interpreter will
        # not be activated and the Linux signal will not be handled until an event arrives that
        # calls a Python signal handler. We make this happen with this timer.
        # (see https://stackoverflow.com/a/4939113)
        self.timer = QTimer(self)
        self.timer.start(100)
        self.timer.timeout.connect(lambda: None)

        # Override the Python exception hook so that a message box is shown to user,
        # when not in DEBUG mode.
        try:
            if not config.getLoggingParameters()['loggers']['gssw']['level'] == 'DEBUG':
                sys.excepthook = gsswExceptHook
                logger.debug('Installed exception hook')
        except KeyError:
            if not config.getLoggingParameters()['root']['level'] == 'DEBUG':
                sys.excepthook = gsswExceptHook
                logger.debug('Installed exception hook')

        self._iconPath = None
        if style and 'style' in config.keys():

            styleCfg = config['style']
            stylePath = config.asAbsolutePath(styleCfg.get('stylepath', ''))

            if not styleCfg.get('stylepath', '') or not os.path.isdir(stylePath):
                logger.warning("Stylepath not defined, does not exist, or is not a directory: {}"
                               .format(stylePath))
                # Take stylepath relative to this file "../qss" if "stylepath" is not in config.
                currentModuleDir = os.path.dirname(os.path.abspath(__file__))
                stylePath = os.path.join(os.path.dirname(currentModuleDir), 'qss')
                logger.debug("Using default stylepath: {}".format(stylePath))

            if not os.path.isdir(stylePath):
                raise OSError("stylepath does not exist or is not a directory: {}"
                              .format(stylePath))
            
            self._iconPath = os.path.join(stylePath, 'icons')

            # Set the Qt style. If not specified we use Fusion because its colors can be fully
            # customized by means of setting the palette.
            styleName = styleCfg.get('style', 'Fusion')
            if styleName.lower() == 'os-default':
                logger.debug("Using default style of OS: {}".format(self.style().objectName()))
            else:
                logger.debug("Set Qt style to: {}".format(styleName))
                self.setStyle(QStyleFactory.create(styleName))
                if self.style().objectName().lower() != styleName.lower():
                    logger.warning(
                            "Setting style failed: actual style {} is not the specified style {}"
                            .format(self.style().objectName(), styleName))

            # Palette
            paletteFileName = styleCfg.get('palette', None)
            if paletteFileName:
                if not os.path.isabs(paletteFileName):
                    paletteFileName = os.path.join(stylePath, paletteFileName)

                logger.debug("Loading palette from: {}".format(normRealPath(paletteFileName)))
                try:
                    with open(paletteFileName, 'r') as paletteFile:
                        pal = parsePalette(paletteFile)
                except OSError as ex:
                    logger.warning("Unable to palette file sheet because: {}".format(ex))
                else:
                    self.setPalette(pal)

            # Style sheet
            styleSheetFileName = styleCfg.get('stylesheet', '')
            if styleSheetFileName:
                if not os.path.isabs(styleSheetFileName):
                    styleSheetFileName = os.path.join(stylePath, styleSheetFileName)

                logger.debug("Reading stylesheet: {}".format(styleSheetFileName))
                try:
                    with open(styleSheetFileName, 'r') as styleFile:
                        lines = styleFile.read()
                except OSError as ex:
                    logger.warning("Unable to read style sheet because: {}".format(ex))
                else:
                    self.setStyleSheet(lines)

            # Logo
            if 'logo' in styleCfg:
                logoFileName = os.path.normpath(
                    os.path.join(stylePath, "icons/logos/", styleCfg['logo']))

                if os.path.exists(logoFileName):
                    icon = QIcon(logoFileName)
                    self.setWindowIcon(icon)
                else:
                    logger.warning("Logo file not found: {}".format(logoFileName))

    def getSignalHandler(self):
        """ Returns the Linux signal handler """
        return self.signalHandler


    @property
    def iconPath(self):
        """ The stylePath followed by '/icons'
        """
        assert self._iconPath, \
            "No iconDir specified. Please configure style/stylepath. " \
            "This should contain an 'iconds' directory."
        return self._iconPath


##################################################################################################
# Non-Qt Application base class
##################################################################################################
class NonQtApplication:
    """ Application class encapsulates all functions to interface with the hardware daemon. """

    def __init__(self, config, name=""):
        """ Constructor application wrapper

            :param config: gssw.lib.config.Config object. If style is True, the location
                of the style sheet is take from this configuration.
            :param name:   application name
        """
        check_class(config, Config)

        self.name = name

        # If the component receives SIGINT or SIGTERM, handle these signals gracefully
        self.signalHandler = SignalHandler()

        signal.signal(signal.SIGINT,  self.signalHandler.handle)
        signal.signal(signal.SIGTERM, self.signalHandler.handle)
        signal.signal(signal.SIGSEGV, self.signalHandler.handle)
        logger.info("Installed Linux signal handlers")

    def getSignalHandler(self):
        """ Returns the Linux signal handler """
        return self.signalHandler
