#!/usr/bin/env python3
# pylint: disable=no-member, invalid-name
""" ArgParse GUI module """

import argparse
import textwrap
import html

from PyQt5 import QtWidgets, QtGui, QtCore


class ArgParseGui(argparse.ArgumentParser):
    """ ArgParse GUI class """

    def __init__(self, *arg, **kwargs):
        super().__init__(*arg, **kwargs)
        self._app = QtWidgets.QApplication([])

        self._widgetMain = QtWidgets.QDialog()
        self._layoutMain = QtWidgets.QVBoxLayout(self._widgetMain)
        self._widgetMain.setLayout(self._layoutMain)

        self._widgetButtons = QtWidgets.QWidget(self._widgetMain)
        self._layoutButtons = QtWidgets.QHBoxLayout(self._widgetButtons)
        self._widgetButtons.setLayout(self._layoutButtons)

        def _runClose():
            QtWidgets.QApplication.quit()
        self._widgetRun = QtWidgets.QPushButton('Run', self._widgetButtons)
        self._widgetRun.clicked.connect(_runClose)
        self._widgetRun.setAutoDefault(False)

        self._layoutButtons.addWidget(self._widgetRun)
        self._layoutButtons.addSpacerItem(
            QtWidgets.QSpacerItem(20, 1, QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Minimum))
        self._parseActions = {}
        self._widgetParsers = None
        self._widgetTab = None
        self._layoutTab = None
        self._widgetScroll = None
        self._widgetGroups = None
        self._layoutGroups = None
        self._widgetArgs = None
        self._layoutArgs = None


    def _showDialog(self):
        self._createDialogLayout()
        self._widgetMain.setWindowIcon(self._iconBase64(self._iconSron))
        self._widgetMain.show()
        self._app.exec_()


    def _createDialogLayout(self):
        self._widgetParsers = QtWidgets.QTabWidget(self._widgetMain)

        self._createTabLayout('Main', self._action_groups)
        for agroup in self._action_groups:
            #pylint: disable=protected-access
            for action in agroup._group_actions:
                if isinstance(action, argparse._SubParsersAction):
                    for aparser in action.choices:
                        self._createTabLayout(aparser, action.choices[aparser]._action_groups)

        self._layoutMain.addWidget(self._widgetButtons)


    def _createTabLayout(self, name, groups):
        self._widgetTab = QtWidgets.QWidget(self._widgetParsers)
        self._layoutTab = QtWidgets.QGridLayout(self._widgetParsers)
        self._widgetTab.setLayout(self._layoutTab)

        self._widgetScroll = QtWidgets.QScrollArea(self._widgetTab)
        self._widgetScroll.setWidgetResizable(True)

        self._widgetGroups = QtWidgets.QWidget(self._widgetScroll)
        self._layoutGroups = QtWidgets.QVBoxLayout(self._widgetGroups)
        self._widgetGroups.setLayout(self._layoutGroups)
        self._widgetScroll.setWidget(self._widgetGroups)

        self._widgetParsers.addTab(self._widgetTab, name)
        self._layoutTab.addWidget(self._widgetScroll)

        self._createGroupLayout(groups)

        self._layoutMain.addWidget(self._widgetParsers)


    def _createGroupLayout(self, groups):
        for lgroup in groups:
            self._widgetArgs = QtWidgets.QGroupBox(self._widgetGroups)
            self._layoutArgs = QtWidgets.QFormLayout(self._widgetArgs)
            self._layoutGroups.addWidget(self._widgetArgs)
            if lgroup.title:
                self._widgetArgs.setTitle(lgroup.title.capitalize())
                if 'Egress configuration' in lgroup.title:
                    self._widgetArgs.setVisible(False)
            else:
                self._widgetArgs.setStyleSheet("QGroupBox{padding-top:15px; margin-top:-15px}")

            #pylint: disable=protected-access
            for arg in lgroup._group_actions:
                argDict = self._parseArgument(arg)

                {'parser': lambda argDict: None,
                 'help': lambda argDict: None,
                 'version': lambda argDict: None,
                 'store': self._createStoreAction,
                 'store_parse': self._createStoreParseAction,
                 'store_const': self._createStoreBoolAction,
                 'store_true': self._createStoreBoolAction,
                 'store_false': self._createStoreBoolAction,
                 'count': self._createCountAction,
                 'append': self._createAppendAction,
                 'append_const': lambda: None}[argDict['action']](argDict)


    def _parseArgument(self, arg):
        argDict = {'option_strings': arg.option_strings}
        argDict['required'] = arg.required
        argDict['dest'] = arg.dest
        argDict['nargs'] = arg.nargs
        argDict['const'] = arg.const
        argDict['default'] = arg.default
        argDict['type'] = arg.type
        argDict['choices'] = arg.choices
        #pylint: disable=protected-access
        argDict['action'] = {
            argparse._HelpAction : 'help',
            argparse._VersionAction: 'version',
            argparse._SubParsersAction: 'parser',
            argparse._StoreAction: 'store',
            argparse._StoreConstAction: 'store_const',
            argparse._StoreTrueAction: 'store_true',
            argparse._StoreFalseAction: 'store_false',
            argparse._CountAction: 'count',
            argparse._AppendAction: 'append',
            argparse._AppendConstAction: 'append_const'}[type(arg)]

        if argDict['action'] == 'store_true' or argDict['action'] == 'store_false':
            argDict['default'] = not argDict['default']

        if (isinstance(argDict['nargs'], str) and argDict['nargs'] in '*+') or \
           (isinstance(argDict['nargs'], int) and argDict['nargs'] > 1):
            argDict['action'] = 'store_parse'

        # Label help text.
        argDict['help'] = '<b>' + argDict['dest'].replace('_', ' ').capitalize() + ':</b>'
        if argDict['type']:
            argDict['help'] += ' ['
            if isinstance(argDict['nargs'], int):
                argDict['help'] += str(argDict['nargs']) + ' '
            argDict['help'] += argDict['type'].__name__
            if isinstance(argDict['nargs'], int):
                argDict['help'] += 's; use comma'
            if argDict['nargs'] in ['*', '+']:
                argDict['help'] += argDict['nargs'] + '; use comma'
            argDict['help'] += ']'
        if arg.help:
            hstring = '<br>&nbsp;&nbsp;<i>' + html.escape(arg.help.capitalize()) + '</i>'
            argDict['help'] += '<br>&nbsp;&nbsp;'.join(textwrap.wrap(hstring, 40,
                                                                     break_long_words=False))

        return argDict


    def _createStoreAction(self, argDict):
        if argDict['choices']:
            widgetEdit = QtWidgets.QComboBox(self._widgetArgs)
            for choice in argDict['choices']:
                widgetEdit.addItem("{}".format(choice))
            if argDict['default']:
                widgetEdit.setCurrentIndex(widgetEdit.findText('{}'.format(argDict['default'])))
            def _parseAction():
                text = widgetEdit.itemText(widgetEdit.currentIndex())
                return argDict['type'](text) if text else None
            self._parseActions[argDict['dest']] = _parseAction
        else:
            widgetEdit = QtWidgets.QLineEdit()
            if argDict['default']:
                widgetEdit.setText('{}'.format(argDict['default']))
            def _parseAction():
                text = widgetEdit.text()
                return argDict['type'](text) if text else None
            self._parseActions[argDict['dest']] = _parseAction

        self._createArgumentRow(argDict, widgetEdit)


    def _createStoreParseAction(self, argDict):
        widgetEdit = QtWidgets.QLineEdit()
        if argDict['default']:
            widgetEdit.setText('{}'.format(argDict['default']))
        def _parseAction():
            text = widgetEdit.text()
            lst = []  # Parse comma-sperated list.
            for itm in text.split(','):
                lst.append(argDict['type'](itm) if text else None)
            return lst[0] if len(lst) == 1 else lst
        self._parseActions[argDict['dest']] = _parseAction

        self._createArgumentRow(argDict, widgetEdit)


    def _createAppendAction(self, argDict):
        widgetEdit = QtWidgets.QWidget(self._widgetArgs)
        layoutEdit = QtWidgets.QHBoxLayout(widgetEdit)
        widgetEdit.setLayout(layoutEdit)

        widgetCombo = QtWidgets.QComboBox(self._widgetArgs)
        widgetCombo.setEditable(True)
        widgetCombo.setInsertPolicy(QtWidgets.QComboBox.InsertAfterCurrent)
        layoutEdit.addWidget(widgetCombo)

        if argDict['default']:
            for item in argDict['default']:
                widgetCombo.addItem(str(item))

        widgetIcon = QtWidgets.QPushButton(self._iconBase64(self._iconRowInsert, True),
                                           '', widgetEdit)
        widgetIcon.setIconSize(QtCore.QSize(16, 16))
        widgetIcon.setFixedSize(QtCore.QSize(22, 22))

        def _insertBefore():
            widgetCombo.setInsertPolicy(QtWidgets.QComboBox.InsertBeforeCurrent)
        widgetIcon.clicked.connect(_insertBefore)
        layoutEdit.addWidget(widgetIcon)

        widgetIcon = QtWidgets.QPushButton(self._iconBase64(self._iconRowInsert),
                                           '', widgetEdit)
        widgetIcon.setIconSize(QtCore.QSize(16, 16))
        widgetIcon.setFixedSize(QtCore.QSize(22, 22))

        def _insertAfter():
            widgetCombo.setInsertPolicy(QtWidgets.QComboBox.InsertAfterCurrent)
        widgetIcon.clicked.connect(_insertAfter)
        layoutEdit.addWidget(widgetIcon)

        widgetIcon = QtWidgets.QPushButton(self._iconBase64(self._iconRowDelete),
                                           '', widgetEdit)
        widgetIcon.setIconSize(QtCore.QSize(16, 16))
        widgetIcon.setFixedSize(QtCore.QSize(22, 22))

        def _removeItem():
            widgetCombo.removeItem(widgetCombo.currentIndex())
        widgetIcon.clicked.connect(_removeItem)
        layoutEdit.addWidget(widgetIcon)

        def _parseAction():
            itms = [widgetCombo.itemText(idx) for idx in range(widgetCombo.count())]
            return itms
        self._parseActions[argDict['dest']] = _parseAction

        self._createArgumentRow(argDict, widgetEdit)


    def _createStoreBoolAction(self, argDict):
        widgetCheckbox = self._createArgumentRow(argDict, None)
        self._parseActions[argDict['dest']] = widgetCheckbox.isChecked


    def _createCountAction(self, argDict):
        widgetEdit = QtWidgets.QSpinBox(self._widgetArgs)
        widgetEdit.setRange(0, 100)
        widgetEdit.setValue(argDict['nargs'])
        self._parseActions[argDict['dest']] = widgetEdit.value
        self._createArgumentRow(argDict, widgetEdit)


    def _createArgumentRow(self, argDict, widgetEdit):

        if argDict['required']:
            widgetDescr = QtWidgets.QLabel(argDict['help'], self._widgetArgs)
        else:
            widgetDescr = QtWidgets.QWidget(self._widgetArgs)
            layoutDescr = QtWidgets.QHBoxLayout(widgetDescr)
            layoutDescr.setContentsMargins(0, 0, 8, 16)
            widgetDescr.setLayout(layoutDescr)
            widgetCheckbox = QtWidgets.QCheckBox('', widgetDescr)
            widgetCheckbox.setChecked(not argDict['default'])
            widgetLabel = QtWidgets.QLabel(argDict['help'], widgetDescr)

            if widgetEdit:
                widgetEdit.setEnabled(not argDict['default'])
                widgetCheckbox.clicked.connect(\
                    lambda: widgetEdit.setEnabled(widgetCheckbox.isChecked()))

            layoutDescr.addWidget(widgetCheckbox)
            layoutDescr.addWidget(widgetLabel)

        self._layoutArgs.addRow(widgetDescr, widgetEdit)
        return widgetCheckbox if not argDict['required'] else None


    def parse_args(self, args=None, namespace=None):  #pylint: disable=unused-argument
        ## super().parse_args(args, namespace)
        self._showDialog()
        namespace = argparse.Namespace()
        for key in self._parseActions:
            setattr(namespace, key, self._parseActions[key]())
        return namespace


    def add_subparsers(self, **kwargs):
        return super().add_subparsers(**dict(**kwargs, parser_class=argparse.ArgumentParser))


    @staticmethod
    def _iconBase64(base64, flip=False):
        pixmap = QtGui.QPixmap()
        pixmap.loadFromData(QtCore.QByteArray.fromBase64(base64))
        if flip:
            transform = QtGui.QTransform()
            transform.scale(1, -1)
            icon = QtGui.QIcon(pixmap.transformed(transform))
        else:
            icon = QtGui.QIcon(pixmap)

        return icon


    _iconSron = b"iVBORw0KGgoAAAANSUhEUgAAABYAAAAWCAYAAADEtGw7AAAABHNCSVQICAgIfAhkiAAAAAlwSFlz"\
        b"AAAOxAAADsQBlSsOGwAAAmRJREFUOI2d1U+oVHUUwPHPjPOMq+iDZ4QVgjDzpge28M+iQuyfoLsiUPgtKly1"\
        b"TVzYTl1p2TaCINJV/ECKQCJqUVaQCxE3ljXMiLYIFWeGVLrxfD5b3DvP23Dv+OfAhXvP75zvOff8zvn9au1m"\
        b"S6fXBe1mayWewZvYjlkkuIM/cQbf4Ftc7/S6t1VIbfTSbra24gB2VBkX5C52d3rdL6oMGjl0Oz7HEw8AHcmT"\
        b"kxYb7WZrGh+WQBfwC/7IE9iAjZgqZF0NziNvGtOfQuj0uleLynaztQp78VYhQKks+3446J+Ymbkmq+0yzOOd"\
        b"Tq97vt1s6Q8HS8b94WC+Pxz8tGZm5mtc6g8H16vA9edDfLnT636MdTiG5ViBpW4pZCzvoou4MCnjWhLiQUyl"\
        b"618/4P1kcXZ2bqq2uPA0Lnd63Yl1vB94M87iOA6lMVx+VFhR6jifQ/fgTBLi4STE5SODJERJiA8NruXOq/El"\
        b"XpJ1yiI+w6f4Hbcxn8awUAwIaQzl4CTEpcUkxHfxNjYXbAayUv2KLq7hShrDz0XfqoyL8MfxHPbh1RKfBdzA"\
        b"rjSGH6rgtXHFWJA5fIJtJbbfpTHsLE23DFwR4APsHzPpYEdVF9VHkPGdT2MobtB7Jb6PyY7UUqknIb6SxvA/"\
        b"UEn2T5Wo/8GtKnAtCfEGzuEofsNN/CvbpCmslbXis2O+J9MYXqsCN2S1ejF//navpW5hGi9gVYnvkfxvSrui"\
        b"IbtytuTf04X3SfJGGsNpqgekjosPACKbvkvYlsbw1f3GvIGPcAVzWIOVsqOzJrslUvyFH3EijSGdNHEj+Q8I"\
        b"49mCoQA0+QAAAABJRU5ErkJggg=="

    _iconRowInsert = b"iVBORw0KGgoAAAANSUhEUgAAABYAAAAWCAYAAADEtGw7AAAABHNCSVQICAgIfAhkiAAAAAl"\
        b"wSFlzAAAOvAAADrwBlbxySQAAAs9JREFUOI2tlM9rVFcUxz/n/niT9yYxKSZWiBJBC112V0qhK6EtQl0U6cJ"\
        b"CIQuloqALF4qguy66aBeNCNIWSn8RofQPsJvSP6AtyaKgplbFRqWpyYyTmXfPvV3MZDITtRmi39W93/ve555"\
        b"7zuEIPdp3kdkiyw4ZhFqreePaMfayRZnezeIycebN9/nh0HGWH5G2CgVwAN9/89PZcbdr6dTt/S//9uev3Bk"\
        b"qKNLQ2M+zfxy9G64XBsvd2k3SAHellMjzHDf1kf9x5taZgyLCndoiVibxrsL9h6vbzy28dymkgDElF964TMX"\
        b"lm4KttczPz+O8yMHPjrzOSnMVa17l/r3rqJZcPXECjYFqlnHyyreM7xhnOBshAbIxyp619565uTlcaETuLS5"\
        b"Qb5XdQxHP4t83AKh7hzYipPYzN4KekAtSiriVmt6c/uKXqbzIWXhwi88/eJudoyO89eks+yZ2E2Kg8MMYMf+"\
        b"H61OMCffg6/jS7MzV1rsH9vPOxQMM5VWGq2PsGZnk99N/8bDRIiaot5afWjwRQaQNtNagqrhPLnyZv1COUVd"\
        b"oNksmJ3YytWOC1dWSlZay0lrupuBpEoG1T1JKRFVco1GXbSPDlKWiocXhjy9RVCoUdjuqSkyJbsVS6lDWiG1"\
        b"PdX0fY0I14EKppJiI0uTy9HcEbRfRe0/gEXnFd0IBYwwxahve58Vuq1SyjKgRF1SJKWKtpTpU3bQw1toneKb"\
        b"vPETFqQaiKgIYWe/QmFLfflDPiLRzrCEQY8QY6T5HACvyWA8M4omRdldoJxXeWVQt0oli7fben1Inut7axQ1"\
        b"e5kw7YquOChWKzJK7okMAjeBsfwOUCn4zz4BPHmGaUeBfnrMcNdKeVzLOn3mRxrONYAAKDB8ev910REgKEdD"\
        b"nAFYSKeIdVeSfpcBXV5YI5bODvRfKmIzIa2RJOUKD5mODditKQHV9Zg8+EwfUf0pDWARllWh+AAAAAElFTkS"\
        b"uQmCC"

    _iconRowDelete = b"iVBORw0KGgoAAAANSUhEUgAAABYAAAAWCAYAAADEtGw7AAAABHNCSVQICAgIfAhkiAAAAAl"\
        b"wSFlzAAAOvgAADr4B6kKxwAAAAu9JREFUOI3VlV1oHFUYhp/vzJnZXcmm7Ua0JFqL2roGSmxR19IIFapCob3"\
        b"wjwgJFKtIr9oLb0xLiRWkVShoK8FSEFS8CBgJXnjjhVgidW3BCrEiUrzQZlPJVl27m5kzM58Xm9h2+8OGeOM"\
        b"LH+fMw+E9Z/jemcNbsGdI5PzHxvQyuEOk937vyAPrc6/BgR0iZz+C1TRlFlPygufpaLHInqmp36prikcv3Xv"\
        b"fhbs+n9i6JpN5+smeHl6szEyEDz0yYZyztCFFyeVyyO4VK46/vG7dzq5z53j3ljy1lT1s+OoLBvv7GctmNXn"\
        b"qGfG674A0bccXay1fnyzD2Pk/O155YqtOdRX0V9CToAr6fqmkU998q6qqcarqUtVEm2Oc3pipqr468oaayok"\
        b"vM6wt8vNslZ7Nmynt2gXALzMX6H34QS41QubCiDCMiCJHGEbMzVcYRrgWpoAYg+kqnyj0vXeE7QMDfDY7y+u"\
        b"jo7B/P7ujkA+3bSfNZohcTORiGqH7d75Q9RYWuhTPM5j6O2//NDgyopP1OhdL/fR9MMbRw4fJDw+zqTrLqb3"\
        b"7sLkcxphmNIy5qq7HjBhsxtozZ2dm+mqFW9n25kEyyzuprO5m/KWdrN24keWrVpHEMWmaIkDS0sRWlqQJnud"\
        b"h/+rID03G+v2zx46jvhAqdD26idLYOBdPn+LOgefRrI9kfEQAvToF1zAByeeRQ8Xejmya1Jbl86RJMn8GEM9"\
        b"DPEPqXDMnLDjINcm9kokR/qhWsd6PP3grOzt57p67iZxrK6s3U+D7HDvzHba5iwHfJ1iyLRAEIII1QNRo8Hu"\
        b"lQhTHS/e1Fo1jrDy2hVo2y2R3N2mbn+3NZIzBLStg3cAQhdtvo/T4FlwUXV4hMt80FsWCIOCT8U+xNk0Q5wh"\
        b"rf+OuaN51ktUWUz/CSxLMjV6p1aBdtvBsgyDA2uavVqQ1o4uXiOD7PtYYg3OO6elpkiRZsrHneagq1lrbqNf"\
        b"rw+Vyee6/OLFq8wb5/+kf6ll2LThMBRAAAAAASUVORK5CYII="


def main():
    """ GUI argparser test function """

    parser = ArgParseGui()
    ##parser = argparse.ArgumentParser()
    parser.add_argument('stepsize', nargs='?', default=1, type=float, \
           help='Stepsize [mm] (default: %(default)s [mm]).')
    parser.add_argument("-m", "--make-argument-true", help="optional boolean argument",
                        action="store_true")
    parser.add_argument("-n","--number", help="an optional number", type=int, nargs=4)
    parser.add_argument("-r","--restricted-number", help="one of a few possible numbers", type=int,
                        choices=[1,2,3], default=2)
    parser.add_argument("-s", type=str, choices=['X', 'Y'], default='X', required=True)
    parser.add_argument("-l", type=int, nargs=2, choices=range(0, 10))
    parser.add_argument("-c", "--counting-argument", help="counting #occurrences", action="count")
    parser.add_argument("-d", "--default-value-argument", help="default value argument" \
                        " with a very very very very very very very very very very very" \
                        " very very very very very very very very very very very very" \
                        " very very very very very very very very very very very" \
                        " very very very very very very long description", type=float,
                        default="3.14")
    parser.add_argument("-a", "--append-args", help="append arguments to list", type=str,
                        action="append", default=["dish", "dash"])
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-v", "--verbose", action="store_true")
    group.add_argument("-q", "--quiet", action="store_true")

    group = parser.add_argument_group()
    group.add_argument("-b", "--boolean", action="store_true")
    group.add_argument("integer", nargs='?', type=int)

    subparsers = parser.add_subparsers(help='sub-command help')

    # create the parser for the "a" command
    parser_a = subparsers.add_parser('Sub', help='sub-parser help')
    parser_a.add_argument('bar', type=int, help='bar help')

    args = parser.parse_args()
    print(args)

if __name__ == "__main__":
    main()
