""" Defines the GriddedArray class.

"""
import logging, copy, os
import numpy as np

from gssw.lib.typing          import check_class, check_is_binary
from gssw.lib.arrays.semantic import SemanticArray
from gssw.lib.arrays.semantic import concatenate as concatenateSemArr
from gssw.lib.sockets         import tagName
from gssw.lib.errors          import ConsistencyError

from gssw.proto.array_pb2 import GriddedArrayMsg

logger = logging.getLogger(__name__)

LOG_FMT = '%(asctime)s %(filename)25s:%(lineno)-4d : %(levelname)-7s: pid=%(process)d: %(message)s'


# If a setter property is defined, the getter property is not correctly regognized by pylint.
# This seems a bug so we disable the no-member. See: https://github.com/PyCQA/pylint/issues/870
# pylint: disable=no-member

class GriddedArray():
    """ Contains a SemanticArray plus optional coordinate vectors that span up a rectilinear grid.

        The GriddedArray class resembles a HDF-5 dataset plus dimension scales.
    """
    def __init__(self, data, name=None, label=None, notes=None, units=None, coords=None):
        """ Constructor.

            The coords parameter is a dict of coordinate vectors. This constructor, together with
            the getter and setter properties, guarantee that it is a dict. However, it is impossible
            in Python to guarantee that it a dict of the correct length and data types (because
            a dict is mutable and contains references). Therefore the consistency should be checked
            after the array is updated, or at least before it is used. This can be done by calling
            checkConsistency(), which raises a ConsistencyError if the coordinates vectors are not
            correct. This function called by the createdFromProtoBuf and asProtoBuf methods, so
            serialization is only done with consistent arrays.

            A coordinate vector is always a one-dimensional SemanticArray. If there is no coordinate
            vector specified for a dimension it is assumed to be the index number (ranging from 0 to
            the number of elements in that dimension). See also makeCoordinateVector()

            :param data:        The central SemanticArray (when a numpy array is given as input it
                                is converted to a SemanticArray)
            :param name:        Name that identifies the variable.
                                Could, for instance, be used as dataset name in HDF5 files
            :param label:       String that describes the variable. E.g. for use in plot titles.
                                Comparable to the long_name attribute in the NetCDF-CF conventions.
            :param notes:       Free form string for additional notes that don't fit elsewhere.
            :param units:       String containing the units of the data. Take a look at the
                                NetCDF-CF for inspiration. Use '1' for dimensionless quantities.
            :param coords:      A dict of coordinate vectors (1D SemanticArray). Default empty dict.
        """
        self._data = SemanticArray(data, name=name, label=label, notes=notes, units=units)

        self._coords = {}
        self.coords = {} if coords is None else coords


    @property
    def data(self):
        """ Returns the (semantic) array
        """
        return self._data


    @data.setter
    def data(self, value):
        """ Sets the semantic array
            Checks type (TODO: convert?)
        """
        check_class(value, SemanticArray)
        self._data = value


    @property
    def name(self):
        """ Convenience method: returns the name of the data.
        """
        return self._data.name


    @property
    def label(self):
        """ Convenience method: returns the label of the data.
        """
        return self._data.label


    @property
    def labelOrName(self):
        """ Convenience method: returns the label of the data or the name if it doesn't exist
        """
        return self._data.label or self._data.name


    @property
    def notes(self):
        """ Convenience method: returns the notes of the data.
        """
        return self._data.notes


    @property
    def units(self):
        """ Convenience method: returns the unites of the data.
        """
        return self._data.units


    @property
    def coords(self):
        """ Returns the dict of coordinate vectors (dimNr -> vector)
        """
        return self._coords


    @coords.setter
    def coords(self, value):
        """ Sets the dict of coordinate vectors to value.

            The keys of the dictionary should be dimension numbers.
            This not checked here but can be checked by self.checkConsistency()

            Checks that every element is a SemanticArray
        """
        coordVectors = dict(value)
        for dimNr, coordVec in coordVectors.items():
            check_class(coordVec, SemanticArray) # TODO: coerce to semantic array?
            if isinstance(coordVec, GriddedArray):
                raise TypeError("dimNr {}: ".format(dimNr) +
                    "Coordinate vectors should not be a SemanticArray, not a GriddedArray.")
        self._coords = coordVectors


    def coordSizes(self):
        """ Returns a list with the size of all coordVectors for all dimensions.

            If there is no vector for a dimension the list will contain None at that position.
        """
        result = []
        for dimNr in range(self.data.ndim):
            if dimNr in self.coords:
                result.append(self.coords[dimNr].size)
            else:
                result.append(None)
        return result


    def makeCoordinateVector(self, dimNr, start=0):
        """ Returns the coordinate vector for dimension dimNr

            If no coordinate vector is defined for that dimension, a default coordinate vector
            will be constructed, having range(dimSize) as data and having emtpy labels as
            metadata. This way, clients can handle all dimension the same way

            :param dimNr: the dimension number (dimension numbers start at 0)
            :param start: the value to start the coordinate vector (for appending)
            :return: 1D SemanticArray
        """
        if dimNr in self._coords:
            return self._coords[dimNr]

        return SemanticArray(np.arange(start, start + self.data.shape[dimNr]))


    def __repr__(self):
        assert self.data is not None, "sanity check"
        arrStr = 'dtype={}'.format(self.data.dtype)
        coordsStr = ", ".join(['_' if s is None else str(s) for s in self.coordSizes()])

        return "<GriddedArray at {:#x}: {!r}, {}, coordDims=({})>"\
            .format(id(self), self.data.name, arrStr, coordsStr)


    def checkConsistency(self):
        """ Checks if the dimensions of the coordinate vectors are consistent with the central array

            Raises a ConsistencyError if the array coordinate vectors are inconsistent.
        """
        if self.data is None:
            raise ConsistencyError("GriddedArray.array is undefined")

        for dimNr, coordVec in self.coords.items():

            if dimNr < 0 or dimNr >= self.data.ndim:
                raise ConsistencyError("dimNr must be >= 0 and < {}. Got: {}"
                                       .format(self.data.ndim, dimNr))

            dimSize = self.data.shape[dimNr]

            if not isinstance(coordVec, SemanticArray):
                raise ConsistencyError("dim {}: expected SemanticArray, got {}"
                                       .format(dimNr, type(coordVec)))

            if coordVec.ndim != 1:
                raise ConsistencyError("dim {}: coordinate vector should be 1D vector, got {}D"
                                       .format(dimNr, coordVec.ndim))

            if len(coordVec) != dimSize:
                raise ConsistencyError("dim {}: len(coordVec) != dim size: {} != {}"
                                       .format(dimNr, len(coordVec), dimSize))


    @classmethod
    def createFromSemanticArray(cls, semanticArray):
        """ Factory that creates a GriddedArray initialized from a SemanticArray

            Does NOT set the coords attribute; it will be an empty dict.

            :param semanticArray: SemanticArray
            :return: GriddedArray
        """
        griddedArray = cls(semanticArray,
                           name  = semanticArray.name,
                           label = semanticArray.label,
                           notes = semanticArray.notes,
                           units = semanticArray.units)

        check_class(griddedArray, GriddedArray)
        return griddedArray


    @classmethod
    def createFromProtoBuf(cls, msg):
        """ Factory that creates a GriddedArray object given the corresponding protocol buffer.

            Checks that the created GriddedArray is consistent before returning it.

            :param msg: GriddedArrayMsg protocol buffer message
            :return: SemanticArray
        """
        assert msg.HasField('data'), "Sanity check failed"
        array = SemanticArray.createFromProtoBuf(msg.data)

        # The GriddedArray inherits from SemanticArray. There is no inheritance in Protocol Buffers,
        # so the protobuf.GriddedArray has a field called central, which is a protobuf.SemanticArray
        griddedArray = cls.createFromSemanticArray(array)

        for coordMapItem in msg.coords:
            coordVector = SemanticArray.createFromProtoBuf(coordMapItem.vector)
            griddedArray.coords[coordMapItem.dim_nr] = coordVector

        griddedArray.checkConsistency()

        return griddedArray


    def asProtoBuf(self):
        """ Serializes itself to the corresponding protocol buffer message.

            :return: GriddedArrayMsg protocol buffer message
        """
        msg = GriddedArrayMsg()
        self.addToProtoBuf(msg)
        return msg


    def addToProtoBuf(self, msg):
        """ Serializes itself ands itself to an existing GriddedArrayMsg protocol buffer message.
            Checks itself for consistency before serializing.

            :param msg: GriddedArrayMsg protocol buffer message
        """
        check_class(msg, GriddedArrayMsg)
        self.checkConsistency()

        self.data.addToProtoBuf(msg.data)

        for dimNr, coordVector in self.coords.items():
            coordMapItemMsg = msg.coords.add()
            coordMapItemMsg.dim_nr = dimNr
            coordVector.addToProtoBuf(coordMapItemMsg.vector)

        return msg


    @staticmethod
    def addArrayToProtoBuf(griddedArrayMsg, griddedArray):
        """ Adds a GriddedArray to a GriddedArrayMsg protocol buffer message.

            This static method that can be used in packetizer addTo dictionary
        """
        griddedArray.addToProtoBuf(griddedArrayMsg)


    def __getitem__(self, index):
        """ Called when using the GriddedArray with an index (e.g. arr[0:3])

            Applies the index to the data and the relevelant coordinate vectors.

            Not all types of indexes are implemented. For instance Ellipsis are not allowed.
        """
        #logger.debug("GriddedArray.__getitem__[{}]".format(index))
        result = GriddedArray(self.data[index])
        result.data.copyMetadata(self.data)

        if isinstance(index, tuple):
            # A tuple index happens when the index is multi-dimensional. E.g. arr[3, 2:7]
            assert len(index) == self.data.ndim, "Index must have on element per data dimension"
            indices = index
        else:
            # A non-tuple index occurs with a single-dimensional index. E.g. arr[3] or arr[2:7]
            # Make an indices list so that we can process the single and multi-dimensional cases the
            # same way. The first element of the index list will be the scalare index, the rest is
            # padded with Nones until the length of the list is the number of data dimensions.
            indices = [None] * self.data.ndim
            indices[0] = index

        # If the index for this dimension is a not a slice (but a scalar), this dimension
        # will be squeezed (the result will have less dimensions). So only increase the
        # dimension index of the result if this dimension index is a slice.
        resDimNr = 0
        for dimNr, dimIdx in enumerate(indices):
            if dimIdx is None:
                # Padded dimension when the index was single-dimensional. Must be copied as is.
                if dimNr in self.coords:
                    result.coords[resDimNr] = self.coords[dimNr]
                resDimNr += 1
            elif isinstance(dimIdx, (slice, list)):
                result.coords[resDimNr] = self.makeCoordinateVector(dimNr)[dimIdx]
                resDimNr += 1
            elif isinstance(dimIdx, int):
                pass # dimension is squeezed, coordinate vector should not be copied
            else:
                raise TypeError("index of type {} is not (yet?) implemented"
                                .format(type(dimIdx)))

        return result


    def exportToCsv(self, fileName):
        """ Export the array to a Comma Separated Vaflues file.
            The array must be one-dimensional
        """
        assert self.data.ndim == 1, "1D array expected. Got: {}D".format(self.data.ndim)
        coords = self.coords.get(0, None)
        nCols = 1
        headerElems = []
        if coords is not None:
            headerElems.append("Column {:d}: {}".format(
                nCols, coords.label if coords.label else 'X-Coordinates'))
            nCols += 1
        headerElems.append("Column {:d}: {}".format(nCols, self.labelOrName))
        header = "\n".join(headerElems)

        formatSpec = '%.18e'
        if coords is None:
            allData = self.data
        else:
            # Check if the X axis are timestamps
            if coords.dtype in (np.datetime64, np.dtype('datetime64[ms]')) \
            or coords.dtype.type == np.datetime64:
                # Because we need to store timestamps differently (X dtype differs from Y type) 
                allData = np.vstack(np.array((coords, self.data), dtype=object))
                formatSpec = ('%-26s', '%.18e')
            else:
                allData = np.vstack((coords, self.data))

        logger.info("Exporting data to CSV file: {}".format(os.path.realpath(fileName)))
        np.savetxt(fileName, allData.T, delimiter=', ', header=header, fmt=formatSpec)



def concatenate(griddedArrayList, axis=0):
    """ Concatenates one or more gridded arrays along an axis

        The axis parameter specifies the axis (dimension) number along which the array is
        concatenated. For this dimension the coordinate vectors must all be specified, or all
        be unspecified.

        :param griddedArrayList: list (or list-like object) of GriddedArray
        :param axis: axis (dimension) number along which the arrays are concatenated
        :return: GriddedArray
    """
    if not griddedArrayList:
        raise ValueError("need at least one array to concatenate")

    for arr in griddedArrayList:
        check_class(arr, GriddedArray)
        check_class(arr.data, SemanticArray) # sanity check

    data = concatenateSemArr([arr.data for arr in griddedArrayList], axis=axis)

    # Concatenate and check the coordinate vectors.
    coords = {}
    for dimNr in range(data.ndim):
        if all([dimNr not in arr.coords for arr in griddedArrayList]):
            # All coordinate vectors are unspecified. Use an unspecified coordinate vector for
            # the result.
            pass
        else:
            if dimNr == axis:
                # Concatenate the coordinate Vectors along this axis
                start = 0     # number to start coords on (for appending)
                vectors = []
                for arr in griddedArrayList:
                    vectors.append(arr.makeCoordinateVector(dimNr, start))
                    start = vectors[-1][-1] + 1     # get last item in last array + 1
                # vectors = [arr.makeCoordinateVector(dimNr) for arr in griddedArrayList]
                coords[dimNr] = concatenateSemArr(vectors)
            else:
                # Create new coordinates vector to merge coordinate-vector metadata into
                newCoord = copy.deepcopy(griddedArrayList[0].makeCoordinateVector(dimNr))

                # Check that all coordinate vectors are equal (for this dimension)
                for arr in griddedArrayList[1:]:
                    coord = arr.makeCoordinateVector(dimNr)
                    newCoord.mergeMetadata([newCoord, coord])

                    if np.any(np.not_equal(coord, newCoord)):
                        raise ValueError("dimNr: {}. Coordinate mismatch: {} != {}"
                            .format(dimNr, griddedArrayList[0].coords[dimNr],arr.coords[dimNr]))

                coords[dimNr] = newCoord

    result = GriddedArray.createFromSemanticArray(data)
    result.coords = coords
    return result



def griddedArrayFromMessageParts(parts):
    """ Deserializes multi parts message that contains a gridded array

        :param parts: ReqRepParts or PubSubParts named tuple.
            Must have tag and message attribute.
        :returns GriddedArray:
    """
    if parts.tag == tagName(GriddedArrayMsg):
        return griddedArrayFromBinMessage(parts.message)
    else:
        raise TypeError("Unexpected protocol buffer {!r}, got {!r}"
                        .format(tagName(GriddedArrayMsg), parts.tag))


def griddedArrayFromBinMessage(binMessage):
    """ Deserializes message that contains a gridded array

        :binMessage: bytes
        :returns GriddedArray:
    """
    check_is_binary(binMessage)
    protobuf = GriddedArrayMsg()
    protobuf.ParseFromString(binMessage)
    return GriddedArray.createFromProtoBuf(protobuf)


