""" Contains classes for interfacing with the quicklooker.

"""
import logging

import numpy as np

from gssw.lib.arrays.gridded import GriddedArray
from gssw.lib.typing import check_class
from gssw.lib.sockets import tagName, PubSocket

logger = logging.getLogger(__name__)


class ArrayPublisherClient():
    """ Client class that publishes GriddedArrays
    """
    def __init__(self, address=None, pubSocket=None):
        """ Constructor

            The address and reqSocket parameters cannot be used both!

            :param address: if set, a ReqSocket is created and connected to this address.
            :param reqSocket: sets the request socket if given. If None, a socket will be created.
        """
        if address is not None:
            assert pubSocket is None, "address and pubSocket can not both be set."
            pubSocket = PubSocket()
            pubSocket.createZmqSocket()
            pubSocket.connectSocket(address)

        self._pubSocket = None
        if pubSocket:
            self.setPubSocket(pubSocket)


    @property
    def pubSocket(self):
        """ The PubSocket"""
        return self._pubSocket


    def setPubSocket(self, pubSocket):
        """ Sets the ReqSocket.

            :param: must be a socket.ReqSocket object
        """
        check_class(pubSocket, PubSocket)
        self._pubSocket = pubSocket


    def publishArray(self, griddedArray, topic=''):
        """ Publishes the gridded array over the socket

            :param GriddedArray griddedArray: the gridded array that contains the data
            :param str topic: optional topic key for ZMQ filtering
        """
        check_class(griddedArray, GriddedArray)
        msg = griddedArray.asProtoBuf()
        tag = tagName(msg)
        self.pubSocket.publishProtoBuf(topic, tag, msg.SerializeToString())


    def publishAsGriddedArray(self, numpyArray, name,
                              label=None, notes=None, units=None, coords=None, topic=''):
        """ Convenience function that wraps a numpy array in a gridded array and publishes it.

            :param GriddedArray griddedArray: the gridded array that contains the data
            :param str name: name of the gridded array that will be published
            :param str label: label of the gridded array that will be published
            :param str notes: notes of the gridded array that will be published
            :param str units: units of the gridded array that will be published
            :param str coords: coords of the gridded array that will be published
            :param str topic: optional topic key for ZMQ filtering
        """
        check_class(numpyArray, np.ndarray)
        griddedArray = GriddedArray(numpyArray, name=name, label=label,
                                    notes=notes, units=units, coords=coords)
        self.publishArray(griddedArray, topic=topic)


