""" Defines:
        MethodsArray serialization.
        SemanticArray class that contains arrays plus metadata such as units and dimensions

    There are unit tests available. These can be found in utest_array.py

"""
import sys
import logging, re
import numpy as np

from gssw.lib.typing import (check_is_a_numpy_array, array_is_structured, check_class,
                             check_is_a_string)
from gssw.proto.array_pb2 import NumpyArrayMsg, SemanticArrayMsg

logger = logging.getLogger(__name__)


#################################
# Serialization of Numpy arrays #
#################################


# We cannot simply say msg.array = arrayToMessage(self.array), we would have to use
# msg.array.CopyFrom(arrayToMessage(self.array)). This may be slow so we need the setArrayInMessage
# function.
# https://developers.google.com/protocol-buffers/docs/reference/python-generated#embedded_message
#
def setArrayInMessage(msg, array):
    """ Serializes a numpy array and adds it to an existing gssw.proto.array_pb2.NumpyArray message
    """
    check_class(msg, NumpyArrayMsg)
    check_is_a_numpy_array(array)

    del msg.shape[:] # Clear any previous data
    msg.shape.extend(array.shape)
    msg.typestr = array.dtype.str
    msg.buffer = array.tobytes()

    # Fill in the field descriptions if the array is structured,
    if array_is_structured(array):
        for fieldDescr in array.dtype.descr:
            assert len(fieldDescr) in (2, 3), \
                "Unexpected field description length: {}".format(len(fieldDescr))

            # Create new FieldDescriptionMsg message and add it to msg.descr
            descrMsg = msg.descr.add()
            descrMsg.name = fieldDescr[0]
            descrMsg.typestr = fieldDescr[1]
            if len(fieldDescr) == 3:
                del descrMsg.shape[:]
                descrMsg.shape.extend(fieldDescr[2])


def arrayToMessage(array):
    """ Serializes a numpy array into a new gssw.proto.array_pb2.NumpyArray message
    """
    msg = NumpyArrayMsg()
    setArrayInMessage(msg, array)
    return msg


def messageToArray(msg):
    """ Deserializes a gssw.proto.array_pb2.NumpyArray message into a numpy array
    """
    check_class(msg, NumpyArrayMsg)
    # Reconstruct field description list of tuples
    description_list = []
    for descr_msg in msg.descr:
        if not descr_msg.shape:
            # Regular field, the tuple only has field name and type description
            field_descr = (descr_msg.name, descr_msg.typestr)
        else:
            # Sub array's field description contain a shape
            field_descr = (descr_msg.name, descr_msg.typestr, descr_msg.shape)
        description_list.append(field_descr)

    if not description_list:
        # A regular, non-structured array. The typestr completely describes the dtype
        dataType = np.dtype(msg.typestr)
    else:
        # The array is structured, it has fields. The dtype is constructed with the field
        # descriptions. The typestr of the message is ignored here, it is redundant. (It does
        # specify the total number the number of bytes. This should be equal to the sum of the
        # byte sizes of the fields but we don't check this.)
        dataType = np.dtype(description_list)

    # Fill the array from the buffer. Does not copy the data but uses a pointer.
    #logger.debug("dtype={}".format(dataType))
    array_out = np.frombuffer(msg.buffer, dtype=dataType)

    # Reshape the array. This should be possible without a data copy so we  assigning the shape
    # instead of using the reshape function. This will raise an exception if a copy is needed.
    # See: docs.scipy.org/doc/numpy/reference/generated/numpy.reshape.html
    array_out.shape = msg.shape

    return array_out


###################
# Semantic arrays #
###################

REGEXP_VALID_NAME = re.compile(r"\A[A-Za-z][_:A-Za-z0-9]*\Z")


def nameIsValid(string):
    """ Returns True if the string can be used as an identifier

        Identifiers start with a letter and contain only letters, digits and underscores.
    """
    return REGEXP_VALID_NAME.match(string) is not None


# Deriving from Numpy ndarray goes with a __new__ constructor instead of __init__. Therefore we
# disabled checking that all attributes have to be defined in __init__
# pylint: disable=attribute-defined-outside-init


class SemanticArray(np.ndarray):
    """ Numpy array with optional attributes for storing meta-data

        This extra information consists of common attributes (name, title, etc). Some of the
        attribute names have been inspired by the NetCDF-CF conventions (http://cfconventions.org/).
        See the constructor for the complete list.

        (The SemanticArray does not include a free-form attribute dictionary. This would be too much
        effort to serialize into protocol buffers. Therefore, only a fixed set of attributes is
        defined. Perhaps a future version could include an attribute dictionary.)

        All metadata attributes are optional and default to None. When two semantic arrays are
        used in a binary operation, for instance two semantic arrays are added, their metadata
        will be merged. This case means that: if one of them is None, the other will be used.
        If they are equal the metadata attribute of the results will be unchanged, If they are not
        None and not equal, the metadata attribute of self will be set to None.

        Example use:
            arr1 = SemanticArray(np.ones(shape=(1, 3)),
                                 name='same_name', label='different_label', notes='arr1_notes')
            arr2 = SemanticArray(np.arange(3).reshape((1, 3)),
                                 name='same_name', label='another_label', units='arr2_units')

            result = arr1 + arr2
            print("result: {}".format(result))
            print("result.name : {}".format(result.name))  # both have the same name
            print("result.label: {}".format(result.label)) # but have different labels
            print("result.notes: {}".format(result.notes)) # notes is only defined for arr1
            print("result.units: {}".format(result.units)) # units is only defined for arr2

        Example output:
            result: [[ 1.  2.  3.]]
            result.name : same_name
            result.label: None
            result.notes: arr1_notes
            result.units: arr2_units

        This function derives from numpy.ndarray. Unfortunately not all numpy functions are
        subclass aware. E.g. if you multiply two SemanticArrays the result will be a SemanticArray,
        as expected. However, if you concatenate two SemanticArrays the result will be a ndarray.
        Therefore this modules defines its own concatenate function. Be carefull to verify the
        result type when using SemanticArrays in numpy functions.

    """
    # Tuple with all metadata attribute names.
    METADATA_ATTRIBUTES = ('name', 'label', 'notes', 'units')

    def __new__(cls, array, name=None, label=None, notes=None, units=None):
        """ Constructs a new SemanticArray, initialized from a regular Numpy array or another
            SemanticArray.

            If array is a semantic array, the metadata is copied. If the name, label or other meta
            data parameters are given, they override the metadata.

            Deriving from a Numpy array goes a little bit different than regular object inheritance.
            See the "Subclassing ndarray" page from the numpy documentation:
                http://docs.scipy.org/doc/numpy/user/basics.subclassing.html

            The metadata attributes below (name, label, etc) may also be None. This means they're
            not set.

            :param array:       A numpy array that is used to initialize the array values.
            :param name:        Name that identifies the variable. Must start with a letter and may
                                only contain letters, digits and underscores. Could, for instance,
                                be used as dataset name in HDF5 files.
            :param label:       String that describes the variable. E.g. for use in plot titles.
                                Comparable to the long_name attribute in the NetCDF-CF conventions.
            :param notes:       Free form string for additional notes that don't fit elsewhere.
            :param units:       String containing the units of the data. Take a look at the
                                NetCDF-CF for inspiration. Use '1' for dimensionless quantities.
        """
        check_is_a_string(name,  allow_none=True)
        check_is_a_string(label, allow_none=True)
        check_is_a_string(notes, allow_none=True)
        check_is_a_string(units, allow_none=True)

        semanticArray = np.asarray(array).view(cls)
        semanticArray.name  = None
        semanticArray.label = None
        semanticArray.notes = None
        semanticArray.units = None

        # Initialize metadata from another semantic array
        if isinstance(array, SemanticArray):
            semanticArray.copyMetadata(array)

        #pylint: disable=multiple-statements
        if name  is not None: semanticArray.name  = name
        if label is not None: semanticArray.label = label
        if notes is not None: semanticArray.notes = notes
        if units is not None: semanticArray.units = units
        #pylint: enable=multiple-statements

        return semanticArray


    @property
    def name(self):
        """ Returns the name that identifies the variable.
        """
        return self._name


    @name.setter
    def name(self, name):
        """ Name that identifies the variable.

            Must start with a letter and may only contain letters, digits and underscores.
        """
        if name is not None:
            if REGEXP_VALID_NAME.match(name) is None:
                raise ValueError("Invalid name: {!r}".format(name))

        self._name = name



    @property
    def axisLabel(self):
        """ Concatenates the label and units properties into an string for use in plot axes.
            Returns "label [units]" if the units is defined, otherwise just the label
            Always returns a string (can be the emtpy string the label is None)
        """
        labelStr = '' if self.label is None else self.label
        if self.units:
            return "{} [{}]".format(labelStr, self.units)

        return labelStr


    def __array_finalize__(self, obj):
        """ Copies the metadata from obj to self.

            The __new__ method is not always called when a new SemanticArray is created so numpy
            provides this function that _is_ called in all casses. It can therefore be used to
            initialize the metadata.

            For more info see:
                http://docs.scipy.org/doc/numpy/user/basics.subclassing.html

            :param obj: object from which the metadata is copied (if present)
        """
        if obj is None:
            pass # obj already initialized in __new__
        else:
            # Called when view casting, e.g arr.view(SemanticArray).
            # Or called when new-from-template, e.g semanticArr[:3].
            self.name  = getattr(obj, 'name',  None)
            self.label = getattr(obj, 'label', None)
            self.notes = getattr(obj, 'notes', None)
            self.units = getattr(obj, 'units', None)


    def __array_wrap__(self, out_arr, context=None):
        """ At the end of every ufunc, this method is called on the input object with the highest
            array priority, or the output object if one was specified. The ufunc-computed array is
            passed in and whatever is returned is passed to the user. Subclasses inherit a default
            implementation of this method, which transforms the array into a new instance of the
            object’s class. Subclasses may opt to use this method to transform the output array
            into an instance of the subclass and update metadata before returning the array to the
            user.

            See; http://docs.scipy.org/doc/numpy/user/basics.subclassing.html#array-wrap-for-ufuncs
        """
        # Merge metadata with the SemanticArrays in the operands
        if context is None: # context can be None since numpy 1.16, so avoid unpacking None below
            return out_arr # argument is no Semantic array, so return it directly

        _operator, opterands, _ = context
        out_arr.mergeMetadata([arr for arr in opterands if isinstance(arr, SemanticArray)])

        # Call the parent
        return np.ndarray.__array_wrap__(self, out_arr, context)


    @classmethod
    def createFromProtoBuf(cls, msg):
        """ Factory that creates a SemanticArray object given the corresponding protocol buffer

            :param cls: the class of this object (a SemanticArray or descendant)
            :param msg: SemanticArrayMsg protocol buffer message
            :return: SemanticArray (or descendant)
        """
        npArr = messageToArray(msg.np_array)
        semanticArray = cls(npArr)

        for attr in SemanticArray.METADATA_ATTRIBUTES:
            if msg.HasField(attr):
                value = getattr(msg, attr)
                setattr(semanticArray, attr, value)

        return semanticArray


    def asProtoBuf(self):
        """ Serializes itself into a newly created SemanticArrayMsg protocol buffer message.
            Needed to prevent CopyFrom(self.asProtobuf())

            :return: SemanticArrayMsg protocol buffer message
        """
        msg = SemanticArrayMsg()
        self.addToProtoBuf(msg)
        return msg


    def addToProtoBuf(self, msg):
        """ Serializes itself and adds itself to an existing SemanticArrayMsg protocol buffer
            message.

            :param msg: SemanticArrayMsg protocol buffer message
        """
        check_class(msg, SemanticArrayMsg)
        setArrayInMessage(msg.np_array, self)
        for attr in SemanticArray.METADATA_ATTRIBUTES:
            value = getattr(self, attr)
            if value is not None:
                setattr(msg, attr, value)

        return msg


    @staticmethod
    def addArrayToProtoBuf(semanticArrayMsg, semanticArray):
        """ Adds a SemanticArray to a SematnicArrayMsg protocol buffer message.

            This static method that can be used in packetizer addTo dictionary
        """
        semanticArray.addToProtoBuf(semanticArrayMsg)


    def copyMetadata(self, semanticArray):
        """ Copies the metadata of semanticArray into self
        """
        for attr in SemanticArray.METADATA_ATTRIBUTES:
            setattr(self, attr, getattr(semanticArray, attr))


    def mergeMetadata(self, semanticArrayList):
        """ Merges the metadata attributes of all SemanticArrays in the list, into self.

            Merging two metadata attributes(name, units, etc) in this case means that: if one of
            them is None, the other will be used. If they are equal the metadata attribute of self
            will be unchanged,If they are not None and not equal, the metadata attribute of
            self will be set to None.

            :param semanticArrayList: list (or list-like object) of SemanticArray
            :return: True if all metadata was equal (self remains unchanged), False otherwise
        """
        for arr in semanticArrayList:
            check_class(arr, SemanticArray)

        for arr in semanticArrayList:
            for attr in SemanticArray.METADATA_ATTRIBUTES:
                val1 = getattr(self, attr)
                val2 = getattr(arr, attr)
                if val1 is None:
                    setattr(self, attr, val2)
                elif val2 is None:
                    pass # Nothing to be done: self.attr is already val1
                elif val1 != val2:
                    setattr(self, attr, None)
                else:
                    pass # Nothing to be done: self.attr is already val1



def concatenate(semanticArrayList, axis=0):
    """ Concatenates one or more semantic arrays along an axis,

        Unfortunately the numpy concatenate function is not subclass aware so we need to write our
        own version.

        The axis parameter specifies the axis (dimension) number along which the array is
        concatenated.

        :param semanticArrayList: list (or list-like object) of SemanticArray
        :param axis: axis (dimension) number along which the arrays are concatenated
        :return: SemanticArray
    """
    if not semanticArrayList:
        raise ValueError("need at least one array to concatenate")

    for arr in semanticArrayList:
        check_class(arr, SemanticArray)

    if 'pyqtgraph' in sys.modules and hasattr(np, 'concatenate_orig'):
        # Ughh, PyQtGraph has a 'numpy fix' that overrides the numpy.concatenate function.
        # This breaks when concatenating arrays with datetime64 elements.
        data = np.concatenate_orig(semanticArrayList, axis=axis)
    else:
        data = np.concatenate(semanticArrayList, axis=axis)

    result = SemanticArray(data)
    result.copyMetadata(semanticArrayList[0])
    result.mergeMetadata(semanticArrayList)

    return result
