#pylint: disable=invalid-name
#pylint: disable=too-many-locals
"""
Python module with utility functions for calculations.
"""

# Import Python system modules
import numpy
import scipy.signal

# Declare exported symbols
__all__ = [ 'deriv', 'movingAverage', 'polyFitUnwrap',
            'hft116d', 'Spectrum', 'logSmooth',
            'makeStepList', 'makeStepListN',
            'shellquote']


def deriv(x, y):
    """Calculate the derivative dY/dX of a sequence of data points (X,Y).

    x                   -- 1D numpy array of X values
    y                   -- 1D numpy array of Y values, same length as xs

    Returns a 1D numpy array of same length as xs, containing the
    local derivatives dY/dX.

    The calculation is based on 3-point interpolation.
    """

    n = len(x)
    assert n >= 3
    assert len(y) == n

    x   = x.astype(numpy.float64)

    dx  = x[1:n] - x[0:n-1]
    ddx = x[2:n] - x[0:n-2]

    ret = numpy.zeros(n)

    ret[0]     = ( - y[0] * (dx[0] + ddx[0]) / (dx[0] * ddx[0])
                   + y[1] * ddx[0] / (dx[0] * dx[1])
                   - y[2] * dx[0] / (dx[1] * ddx[0]) )

    ret[1:n-1] = ( - y[0:n-2] * dx[1:n-1] / (dx[0:n-2] * ddx)
                   + y[1:n-1] / dx[0:n-2]
                   - y[1:n-1] / dx[1:n-1]
                   + y[2:n]   * dx[0:n-2] / (dx[1:n-1] * ddx) )

    ret[n-1]   = ( + y[n-3] * dx[n-2] / (dx[n-3] * ddx[n-3])
                   - y[n-2] * ddx[n-3] / (dx[n-3] * dx[n-2])
                   + y[n-1] * (dx[n-2] + ddx[n-3]) / (dx[n-2] * ddx[n-3]) )

    return ret


def movingAverage(data, width, axis=0):
    """Compute a moving average and return an averaged array
    of identical dimensions.

    data            -- nD numpy array
    width           -- width of moving average kernel, must be odd
    axis            -- axis along which the average is computed
    """

    assert width > 0 and width <= len(data) and width % 2 == 1

    t = data.cumsum(axis=axis)  #pylint: disable=invalid-name

    a = t[width//2:width]  #pylint: disable=invalid-name
    a = a / numpy.arange(width//2+1, width+1, dtype=numpy.float64)  #pylint: disable=invalid-name

    b = (t[width:] - t[:-width]) / float(width)  #pylint: disable=invalid-name

    c = t[-1] - t[-width:-width//2]  #pylint: disable=invalid-name
    c = c / numpy.arange(width-1, (width-1)//2, -1, dtype=numpy.float64)  #pylint: disable=invalid-name

    return numpy.concatenate((a, b, c))


def polyFitUnwrap(x, y, degree, wrap=360.0, outlier=2.5):  #pylint: disable=too-many-locals
    """Least squares polynomial fit to wrapped data.

    Find the polynomial
      poly(x) = p[0] * x**degree + p[1] * x**(degree-1) + ... + p[degree]

    which is a least-squares fit to points (x[i], y[i]) where the
    fit errors are calculated modulo "wrap".

    The algorithm is heuristic and may fail if the data wraps too often
    or wraps more than once between points.

    x               -- 1D Numpy array of x values
    y               -- 1D Numpy array of y values
    degree          -- degree of the polynomial
    wrap            -- modulus to use when calculating fit error
    outlier         -- ignore points further than (outlier * stddev) from early fit

    Return a 1D array of length (degree+1) containing the coefficients,
    starting with the highest power, ending with the constant term.
    (Same order as numpy.polyfit).
    """

    assert degree > 0
    assert len(x) == len(y)

    # Calculate deltas between y values.
    ydelta = y[1:] - y[:-1]

    # Unwrap the places where the y value jumps more than 0.5 * wrap.
    ydelta = (ydelta + 0.5 * wrap) % wrap - 0.5 * wrap

    # Calculate local derivative for each interval.
    dydx = ydelta / (x[1:] - x[:-1])

    # Fit a polynomial to the derivatives.
    dpoly = numpy.polyfit(0.5 * (x[1:] + x[:-1]), dydx, degree-1)

    # Calculate primitive of fitted polynomial.
    # This is an early estimate of the polynomial, with offset term unknown.
    coeff      = numpy.zeros(degree + 1, dtype=numpy.float64)
    coeff[:-1] = dpoly / numpy.arange(degree, 0, -1, dtype=numpy.float64)

    # Subtract polynomial from the data points.
    # The residual dataset is approximately constant, but points may still
    # be displaced by arbitrary multiples of "wrap".
    ytmp = y - numpy.polyval(coeff, x)

    # Unwrap data points assuming offset is close to zero.
    # Estimate offset from median of unwrapped points.
    a1 = numpy.median((ytmp + 0.5 * wrap) % wrap - 0.5 * wrap)  #pylint: disable=invalid-name

    # Unwrap data points assuming offset is close to 0.5 * wrap.
    # Estimate offset from median of unwrapped data points.
    a2 = numpy.median(ytmp % wrap)  #pylint: disable=invalid-name

    # Calculate standard deviation for both offset estimates.
    s1 = numpy.std((ytmp - a1 + 0.5 * wrap) % wrap - 0.5 * wrap)  #pylint: disable=invalid-name
    s2 = numpy.std((ytmp - a2 + 0.5 * wrap) % wrap - 0.5 * wrap)  #pylint: disable=invalid-name

    # Take the offset that gives lowest standard devation.
    coeff[-1] = a2 if s2 < s1 else a1

    # Subtract estimated offset, then unwrap data.
    ytmp = ytmp - coeff[-1]
    ytmp = (ytmp + 0.5 * wrap) % wrap - 0.5 * wrap

    # Reject outliers further than (outlier * stddev) from the early fit.
    if outlier:
        s   = numpy.std(ytmp)
        idx = numpy.where(numpy.abs(ytmp) <= outlier * s)
    else:
        idx = Ellipsis

    # Normal polynomial fit on the unwrapped residual data.
    rpoly = numpy.polyfit(x[idx], ytmp[idx], degree)

    # Add residual polynomial to early estimated polynomial.
    coeff += rpoly

    # Normalize the constant term.
    coeff[-1] = coeff[-1] % wrap

    return coeff


def hft116d(n):
    """Return the HFT116D window.

    The HFT116D window is optimized to achieve the lowest sidelobe level
    achieveable with 5 cosine terms while maintaining a sidelobe drop rate
    of f**(-3).

    NENBW =  4.2186 bins
    W-3dB =  4.1579 bins
    emax  = -0.0028 dB
    PSLL  = -116.8 dB

    Ref: G. Heinzel, A. Rudiger, R. Schilling, Spectrum and spectral density 
            estimation by the Discrete Fourier Transform, 2002.
    """

    if n < 2:
        return numpy.ones(n)
    
    a = numpy.array([ 1.0, -1.9575375,  1.4780705, -0.6367431,
                             0.1228389, -0.0066288 ])
    a /= (a[0] - a[1] + a[2] - a[3] + a[4] - a[5])
    xs = 2 * numpy.pi / (n - 1) * numpy.arange(n)
    return ( a[0] + a[1] * numpy.cos(xs)
                  + a[2] * numpy.cos(2*xs)
                  + a[3] * numpy.cos(3*xs)
                  + a[4] * numpy.cos(4*xs)
                  + a[5] * numpy.cos(5*xs) )

def maskSigned(rawvalue, bits):
    """ Mask signed value """
    m = 1 << (bits-1)
    return (rawvalue & (m - 1)) - (rawvalue & m)

def maskUnsigned(value, bits):
    """ Mask unsigned value """
    return value & ((1 << bits) - 1)

def nearbyPrime(n):
    """Return the prime number closest to n."""
    n = int(n)

    if n <= 2:
        return 2

    # consider only odd numbers
    d = (n + 1) % 2 

    while 1:
        ts = (n+d, n-d) if (d > 0 and n - d > 2) else (n+d,)
        for t in ts: 

            # trial factoring to test if t is prime
            p = 3 
            while p * p <= t and t % p != 0:
                p += 2

            if p * p > t:
                # t is prime
                return t

        d += 2



class Spectrum:
    """Calculate power spectrum or PSD of time-domain data."""

    windowFunctions = {
        'rect':     numpy.ones,
        'hann':     numpy.hanning,
        'hanning':  numpy.hanning,
        'hamming':  numpy.hamming,
        'blackman': numpy.blackman,
        'blackmanharris': scipy.signal.blackmanharris,
        'flattop':  scipy.signal.flattop,
        'hft116d':  hft116d }

    def __init__(self, fftlen, fsample, windowfunc, removemean=True, iscomplex=False):
        """Initialize PSD calculation.

        :param fftlen:     number of samples per spectrum
        :param fsample:    sample rate
        :param windowfunc: name of window for power density estimation;
                           one of 'rect', 'hann', 'hamming', 'blackman',
                           'blackmanharris', 'flattop', 'hft116d'
        :param removemean: True to remove the mean value before
                           calculating the spectrum (default True)
        :param iscomplex:  True to support complex data
        """

        self.fftlen     = fftlen
        self.fsample    = fsample
        self.navg       = 0
        self.removemean = removemean
        self.iscomplex  = iscomplex

        windowFunction  = self.windowFunctions[windowfunc]
        self.window     = windowFunction(fftlen)
        self.window     /= numpy.sum(self.window) # Normalize window
        self.winnbw     =  numpy.mean(self.window**2) / numpy.mean(self.window)**2

        self.sumspec    = None

    def add(self, data):
        """Add a series of samples to the PSD."""

        assert len(data) == self.fftlen,\
            "Incorrect data length (must be: {}, was {})".format(self.fftlen, len(data))

        # Optionally remove mean from data.
        if self.removemean:
            data = data - data.mean()

        if self.iscomplex:

            # Apply window; calculate FFT.
            spec = numpy.fft.fft(data * self.window)

            # Convert to power spectrum.
            spec = numpy.real(spec * spec.conj())

            # Shift zero frequency to center of spectrum.
            spec = numpy.fft.fftshift(spec)
 
        else:

            # Apply window; calculate FFT.
            spec = numpy.fft.rfft(data * self.window)

            # Convert to power spectrum.
            spec = numpy.real(spec * spec.conj())

            # Convert to single-sided spectrum.
            if self.fftlen % 2 == 0:
                spec[1:-1] *= 2
            else:
                spec[1:] *= 2

        # Add to running total.
        if self.navg == 0:
            self.sumspec = spec
        else:
            self.sumspec += spec
        self.navg += 1

    def clear(self):
        """Remove sample data."""

        self.sumspec = None
        self.navg = 0

    def powerspectrum(self):
        """Return power spectrum."""
        assert self.sumspec is not None, "Spectrum contains no data"

        return self.sumspec * (1.0 / self.navg)

    def psd(self):
        """Return power spectral density."""
        assert self.sumspec is not None, "Spectrum contains no data"

        df = self.fsample / float(self.fftlen)  #pylint: disable=invalid-name
        return self.sumspec * (1.0 / df / self.winnbw / self.navg)

    def asd(self):
        """Return RMS amplitude spectral density."""

        return numpy.sqrt(self.psd())

    def freqs(self):
        """Return array of frequency labels."""

        if self.iscomplex:
            dt = 1.0 / self.fsample  #pylint: disable=invalid-name
            return numpy.fft.fftshift(numpy.fft.fftfreq(self.fftlen, dt))
        else:
            assert self.sumspec is not None, "Spectrum contains no data"
            df = self.fsample / float(self.fftlen)  #pylint: disable=invalid-name
            return numpy.arange(len(self.sumspec)) * df


def logSmooth(data, finalWidth):
    """Apply logarithmic smoothing to an array of samples, such that
    the first samples are not smoothed at all while the last samples
    are heavily smoothed.

    data                -- 1D array of sample values
    finalWidth          -- maximum width of smoothing window

    Only suitable for noise-floor analysis, not for peaks and harmonics.

    The smoothing is skewed: Samples are smeared asymmetrically from left
    to right only.
    """

    finalWidth = min(len(data), finalWidth)
    width = 1 + numpy.arange(len(data)) * finalWidth // len(data)

    #
    # Calculate result[i] = mean( data[i-width[i]+1:i+1] )
    #

    result = 1.0 * data

    p = 0
    while p < len(data):
        w = width[p]  #pylint: disable=invalid-name
        p += 1
        while p < len(data) and width[p] == w:
            p += 1
        if p < len(data):
            result[p:] += data[p-w:-w]

    return result / width


def makeStepList(start, stop, stepSize,
                 fineStart=None, fineStop=None, fineStep=None):
    """Build a list of steps to scan a range.

    start               -- start value of scan range
    stop                -- end value of scan range
    stepSize            -- interval between successive steps
    fineStart           -- (optional) enable fine steps past this value
    fineStop            -- (optional) end fine steps past this value
    fineStep            -- interval between fine steps

    Returns a 1D array of step values.

    A negative step size is automatically used if start < stop.
    """

    assert (start == stop) or (stepSize != 0)

    # shortcut for single step
    if start == stop:
        return numpy.array([ start ])

    # generate coarse steps

    if start > stop:
        stepSize = -abs(stepSize)
    else:
        stepSize = abs(stepSize)

    p  = int(numpy.floor(start / float(stepSize)))
    q  = int(numpy.ceil(stop / float(stepSize)))
    xs = numpy.arange(p, q+1) * stepSize

    # generate fine steps
    if (    (fineStart is not None or fineStop is not None)
        and (fineStep  is not None) 
        and (fineStep != 0)):

        assert fineStep is not None

        if start > stop:
            fineStep = -abs(fineStep)
        else:
            fineStep = abs(fineStep)

        if fineStart is None:
            fineStart = start
        if fineStop is None:
            fineStop = stop

        pp = int(numpy.ceil(fineStart / float(fineStep)))
        qq = int(numpy.floor(fineStop / float(fineStep)))

        if pp <= qq:
            p1 = int(numpy.ceil(pp * fineStep / float(stepSize))) - 1
            q1 = int(numpy.floor(qq * fineStep / float(stepSize))) + 1
            xs = numpy.concatenate((
                     numpy.arange(p, p1 + 1) * stepSize,
                     numpy.arange(pp, qq + 1) * fineStep,
                     numpy.arange(q1, q + 1) * stepSize))

    return xs


def makeStepListN(start, stop, nsteps):
    """Build a list of steps to scan a range.

    start               -- start value of scan range
    stop                -- end value of scan range
    nsteps              -- number of points (including start and stop)

    Returns a 1D array of step values.
    A negative step size is automatically used if start < stop.
    """

    if nsteps == 0:
        return [ ]

    if nsteps == 1:
        return [ start ]

    d = (stop - start) / float(nsteps - 1)
    return start + d * numpy.arange(nsteps)


def shellquote(s):
    """Simple shell-like quoting."""

    if s.isalnum():
        return s

    return "'" + s.replace("'", "'\"'\"'") + "'"

