"""
Python module to support printing colored text to ANSI/VT100 terminals.
"""

import logging
import os as _os


class ColorPrint():
    """Wrap text in ANSI/VT100 color codes."""

    # Set to True to enable colored printing.
    enabled = False

    # ANSI code to reset colors
    resetColors = '\x1b[m'

# RED=   "\[\033[31;1m\]"
# GREEN= "\[\033[32;1m\]"
# YELLOW="\[\033[33;1m\]"
# BLUE== "\[\033[34;1m\]"
# PURPLE="\[\033[35;1m\]"

    colorTableDarkbg = {
        'red':      '\x1b[31;1m',
        'green':    '\x1b[32m',
        'yellow':   '\x1b[33;1m',
        'blue':     '\x1b[36m',
        'bold':     '\x1b[1m',
    }

    colorTableLightbg = {
        'red':      '\x1b[31m',
        'green':    '\x1b[32m',
        'yellow':   '\x1b[33m',
        'blue':     '\x1b[34m',
        'bold':     '\x1b[1m',
    }

    def __init__(self, force=False):
        """Detect terminal type."""

        if force:
            self.enabled = True
        else:
            # Enable colored printing only if stdout is a terminal
            # and TERM variable is properly set.
            term = _os.getenv('TERM')
            if _os.isatty(1) and term is not None and term.lower() != 'dumb':
                self.enabled = True

        # Try to autodetect light background; only works in Konsole.
        fgbg = _os.getenv('COLORFGBG')
        if fgbg is not None and fgbg.endswith(';15'):
            self.colorTable = dict(self.colorTableLightbg)
        else:
            self.colorTable = dict(self.colorTableDarkbg)

    def color(self, msg, color):
        """Wrap string in color codes for specified color name."""

        if self.enabled:
            return self.colorTable[color] + msg + self.resetColors
        
        return msg

    def normal(self, msg):
        """Prepend reset color command before the string."""

        if self.enabled:
            return self.resetColors + msg
        
        return msg

    def bold(self, msg):
        """Wrap string in highlight/bold color codes."""

        if self.enabled:
            return self.colorTable['bold'] + msg + self.resetColors
        
        return msg

    def red(self, msg):
        """Wrap string in red color codes."""

        if self.enabled:
            return self.colorTable['red'] + msg + self.resetColors
        
        return msg

    def green(self, msg):
        """Wrap string in green color codes."""

        if self.enabled:
            return self.colorTable['green'] + msg + self.resetColors
        
        return msg

    def yellow(self, msg):
        """Wrap string in yellow color codes."""

        if self.enabled:
            return self.colorTable['yellow'] + msg + self.resetColors
        
        return msg

    def blue(self, msg):
        """Wrap string in blue color codes."""

        if self.enabled:
            return self.colorTable['blue'] + msg + self.resetColors
        return msg


class ColorLogFormatter(logging.Formatter):
    """ Wrapper around a logging.Formatter object that adds colors
    depending on the log level of the message.

    Use in your config file in the logging section, as follows:..
        formatters:
            streamFormatter:
                format: '... some formatting string ...'
                class: gssw.lib.colorprint.ColorLogFormatter
    """

    def __init__(self, fmt=None, datefmt=None, style='%'):
        super().__init__(fmt, datefmt, style)
        self.colorprint = ColorPrint()

    def format(self, record):
        """ Formats the log message, with a relevant color. """
        message       = super().format(record)
        logLevelColor = self.levelColor(record.levelname)
        return self.colorprint.color(message, logLevelColor)

    def levelColor(self, level):
        """ Looks up the color corresponding to an error level. """
        if level in ('WARNING', 'ERROR', 'CRITICAL'):
            return 'red'

        if level == 'DEBUG':
            return 'blue'

        return 'green'
