""" Module containing the Config class. """
import getpass
import logging
import logging.config
import tempfile

import copy
import os.path
import yaml
from yaml.scanner import ScannerError

from gssw.lib.errors import GsswError, ConfigError
from gssw.lib.typing import check_class

logger = logging.getLogger(__name__)

def tmpDirectory():
    """ Returns the temporary directory in which an output directory can be created.
    """
    return tempfile.gettempdir()


def outputDirectory():
    """ Returns directory for output (log and PID files.

        Will be located in the temp directory.
        Has the form : <temp-dir>/gssw_<user-name>
    """
    return os.path.join(tmpDirectory(), 'gssw_{}'.format(getpass.getuser()))



class Config(dict):
    """ Class that contains the configuration for the GSSw.

        It extents a dictionary for convenience.
        The configuration is typically loaded by a ConfigClient. The original location of the
        config file is stored by this class. This makes it possible to have relative paths in
        the configuration.
    """

    def __init__(self, dct=None):
        """ Constructor for the ConfigClient class.

            :param dct: An optional existing dictionary.with initial values.
        """
        super().__init__({} if dct is None else dct)
        self._configFile = None


    @classmethod
    def createFromYamlFile(cls, fileName):
        """ Creates a Config from a YAML file
        """
        config = cls()
        config.readFromYaml(fileName)
        return config


    def readFromYaml(self, fileName):
        """ Reads the configuration from a YAML file.

            The Config object must be empty before calling it.
        """
        assert not self, "Config must be empty before reading from YAML"

        logger.info('Reading the config file from: {}'.format(fileName))

        try:
            with open(fileName, 'r') as stream:
                valueDict = yaml.load(stream, Loader=yaml.SafeLoader)
        except (FileNotFoundError, ScannerError) as e:
            raise ConfigError("Error reading config file: %s" % e)

        for key, value in valueDict.items():
            self[key] = value

        self.configFile = fileName


    def checkConfigFileDefined(self):
        """ Raises a ValueError if the configFile is not defined."""
        if self._configFile is None:
            raise ValueError("configFile property is undefined.")


    @property
    def configFile(self):
        """ The config file that was used to fill this configuration"""
        self.checkConfigFileDefined()
        return self._configFile


    @configFile.setter
    def configFile(self, value):
        """ Sets the config file that was used to fill this configuration.
            Makes an absolute path from it and replaces any '../' occurrences in the path.
        """
        self._configFile = os.path.normpath(os.path.realpath(value))


    def getConfigDir(self):
        """ Returns the directory of the configuration file (including trailing /).
            Is used as a prefix for relative paths in the configuration file.
        """
        self.checkConfigFileDefined()
        return os.path.join(os.path.dirname(self.configFile), '')   # adds trailing slash


    def asAbsolutePath(self, fileName):
        """ Returns fileName as absolute path with respect to the config file location.

            If fileName is an absolute path already it will remain unchanged.
        """
        return os.path.join(self.getConfigDir(), fileName)


    def componentParameters(self, name):
        """ Get the configuration parameters for a specific module/component.

            :param str name: The name of the component in the config file.
            :return: A dict containing the configuration parameters.
        """
        check_class(name, str)
        if name not in self['components']:
            raise GsswError(
                "Component '{}' is not defined in the Egress configuration file.".format(name))
        return self['components'][name]


    def getLoggingParameters(self):
        """ Get the logging parameters.

            This function returns a dict containing the logging parameters.
            If the config file is not loaded or does not contain the information,
            None is returned.
        """
        if 'logging' in self:
            return self['logging']

        return None


    def initLogging(self, logToSocketHandler=True):
        """ Uses the 'logging' sub config to initialize the global logging.

            Should be called only once per program or GSSw component.

            :param bool logToSocketHandler: if False the 'socketHandler' from the logging
                configuration will not be used. This is necessary to prevent the eventLog server to
                send log messages to itself. The default is True.

            Does anybody ever read docstrings? #2

            >>> s = 'Qvq lbh rire jngpu Ebpxrgzra? Guvf vf n zneiryybhf OOP qbphzragnel bs gur '\
                    'qnja bs fcnpr ntr.\nVg vf irel zhpu yvxr Gur evtug fghss, rkprcg gung vg '\
                    'pbafvfgf bs uvfgbevp sbbgntr.'
            >>> import codecs
            >>> codecs.decode(s, ...) # ...=? Back to Roman times
        """
        logParms = copy.deepcopy(self.getLoggingParameters())

        if logToSocketHandler:
            logging.config.dictConfig(logParms)
        else:
            try:
                items = logParms['root']['handlers']
                items.remove('socketHandler')
            except ValueError:
                logger.debug('No socketHandler connected to root. Not removed.')

            logging.config.dictConfig(logParms)

        # We log this here as well because earlier log statements may have been lost because the
        # logging was not yet configured.
        if self.configFile is not None:
            logger.info("Config file (including log cfg) was read from: {}".format(self.configFile))

    def getGlobalMiscParameters(self):
        """ Get the global miscellaneous parameters.

            This function returns a dict containing global miscellaneous parameters.
            If the config file is not loaded or does not contain the information,
            None is returned.
        """
        if 'global-parameters' in self:
            return self['global-parameters']

        return None



