""" Python interface to hardware devices.

    This module is the python interface to communicate with the hardware daemon
    itself for configurting properties related with the hardware daemon itself.
"""

from   collections             import namedtuple
import logging
import threading

import gssw.proto.hwd_pb2      as hwd_proto

from gssw.proto.common_pb2     import ErrorRep, OkRep
from gssw.lib.sockets          import ReqSocket, tagName
from gssw.lib.errors           import CommunicationError
from gssw.hal.halclient        import HalClient
from gssw.hal.hal              import Hal
from gssw.lib.misc             import toProtobufEncoding


logger = logging.getLogger(__name__)

class Daemon:
    """ The Daemon class encapsulates all functions to interface with the hardware daemon. """

    def __init__(self, deviceName, config, hal=None, timeout=5, scienceDaemon=False,
                 daemonName=None):
        """ Constructor for the Daemon class.

        :param deviceName: Name of the device in the HAL.
        :param hal:        Hardware abstraction layer object (optional).
        :param config:     System configuration (a dictionary loaded from the config.yaml file).
        :param timeout:    [s] Communication timeout for daemon operations.
                           set to -1 for infinite.
        :param science:    If set, this daemon class connects to a science daemon
        :param daemonName: Override name via which to connect to the daemon.
                           Can be used to force deviceName's traffic to custom daemon.

        """

        # Initialize variables.
        self.deviceName = deviceName

        # Hardware abstraction layer
        if hal is None:
            self.hal       = Hal()
            self.halClient = HalClient(self.hal, config)
            self.halClient.requestHal()
        else:
            self.hal = hal

        logger.info('Initializing daemon interface for device: %s', self.deviceName)

        # Get the configuration for the HW daemon the device class will communicate with
        try:
            if daemonName is not None:
                self.daemonName = daemonName
            elif scienceDaemon:
                self.daemonName = self.hal.getDevice(deviceName=deviceName).scienceDaemon
            else:
                self.daemonName = self.hal.getDevice(deviceName=deviceName).daemon
        except KeyError:
            raise AssertionError("The device '{}' was not found in the HAL, "
                                 "check your configuration file".format(deviceName)) from None

        logger.info("Using daemon %s for device %s", self.daemonName, self.deviceName)
        deviceConfig = config.componentParameters(self.daemonName)

        # Create a REQ socket to communicate with the device's hw_daemon.
        # The socket is opened just before usage
        self._sockets = {}
        self.timeout = timeout
        self.repAddress = 'tcp://%s:%d' % (deviceConfig['hostname'],
                              deviceConfig['sockets']['daemon-rep'])

    @property
    def socket(self):
        """ To make this class thread-safe, the underlying ReqSocket objects are duplicated 
            per thread (The ReqSocket class is not thread safe because of the zmq sockets). """
        threadId = threading.get_ident()
        if threadId not in self._sockets:
            logger.debug('Creating new ReqSocket for device: %s (thread %s)',
                         self.deviceName, threadId)
            self.setupSocket()
        return self._sockets[threadId]

    @socket.setter
    def socket(self, value):
        threadId = threading.get_ident()
        self._sockets[threadId] = value

    def setupSocket(self):
        """ Sets up the REQ-REP socket for hardware daemon transactions """
        self.socket = ReqSocket()
        self.socket.createZmqSocket(timeoutMs=1000 * self.timeout)
        self.socket.connectSocket(self.repAddress)

    def hwdTransaction(self, request):
        """ Perform a communication transaction (request, reply) with the hardware daemon
            with the intention to configure the hardware daemon itself.

            :param request: HwRegisterReq or HwBlockReq protobuf object.

            If an error occurred, an ErrorRep object is returned.
        """

        # first transaction or after a timeout
        if self.socket is None:
            self.setupSocket()

        # receive the reply
        try:
            # send the header string and the serialized request
            self.socket.sendProtoBuf(tagName(request), request.SerializeToString())

            # receive the reply
            tag, message = self.socket.receiveProtoBuf()

        except Exception as e:
            self.socket.close()
            self.socket = None
            raise CommunicationError('Error receiving reply from daemon {}, '
                                     'probably due to a timeout.'.format(self.daemonName)) from e

        # check that the correct header string is send
        if tag == 'ErrorRep':
            reply = ErrorRep()
            reply.ParseFromString(message)
            logger.error("Received an error reply during a daemon transaction (%s)",
                    ErrorRep.ErrorType.Name(reply.errtype))
            CommunicationError.raiseFromProtoBuffer(reply, 'Daemon {}'.format(self.daemonName))
        elif tag == 'GetConfigRep':
            reply = hwd_proto.GetConfigRep()
        elif tag == 'GetInfoRep':
            reply = hwd_proto.GetInfoRep()
        elif tag == 'OkRep':
            reply = OkRep()
        else:
            raise TypeError('Wrong message header received from device:\n%s' % tag)

        # deserialize the received reply
        reply.ParseFromString(message)

        # print the response
#        logger.debug("received reply:\n%s", reply)

        return reply

    def setTimeout(self, timeout):
        """ Sets the timeout value on the req/rep socket """
        if self.socket is None:
            self.setupSocket()

        self.timeout = timeout
        self.socket.setReceiveTimeout(self.timeout)

    def getHwTimeout(self):
        """ Gets the timeout to the HW Daemon to be used in its communication
            with the hardware
        """
        # create and send request protobuf object
        reply = self.hwdTransaction(hwd_proto.GetConfigReq())
        return reply.config.timeout

    def setHwTimeout(self, timeout):
        """ Sets the timeout to the HW Daemon to be used in its communication
            with the hardware

            :param timeout: timeout value in seconds to be used on a hardware transaction
        """
        request = hwd_proto.SetConfigReq()
        request.config.timeout = timeout

        # perform the message transaction
        self.hwdTransaction(request)

    def setPacketSieve(self, skip=0, total=-1):
        """ Configures the HWDaemon to skip a certain amount of packets when
            publishing science data.

            :param skip: number of packets to skip before passing on 1 science packet
            :param total: number of packets to forward. -1 to keep going until measurment stops
        """
        request = hwd_proto.SetConfigReq()
        request.config.sieve.packets_to_skip    = skip
        request.config.sieve.packets_to_forward = total

        # perform the message transaction
        self.hwdTransaction(request)

    def getPacketSieve(self):
        """ Gets the packet sieve settings """
        reply = self.hwdTransaction(hwd_proto.GetConfigReq())
        return reply.config.sieve

    def getInterfaceInfo(self):
        """ Gets the low-level interface information from the daemon. """
        reply = self.hwdTransaction(hwd_proto.GetInfoReq())
        device_info = reply.device

        # Translate the protobuf object into a namedtuple
        DeviceInfo = namedtuple(device_info.__class__.__name__,  #pylint: disable=invalid-name
                                ['translator', 'physical', 'info_items'])
        info_items = {item.key: item.value for item in device_info.info_items} # key-values to dict

        return DeviceInfo(device_info.translator, device_info.physical, info_items)

    def setHalExcerpt(self, board, address = None, channel = None, bits = 32, encoding = "uint"): 
        """ Sets a part of the HAL to the HWDaemon, for filling in the
            protocol buffers upon reception of SKDP packets from the 
            hardware. Information which is configured:

            :param board:    board identifier
            :param address:  address of the registers
            :param channel:  channel of the science data (SKDP)
            :param bits:     number of bits to be used for this parameter
            :param encoding: encoding to be used for this parameter
        """
        assert address is not None or channel is not None 
        request = hwd_proto.SetHalExcerptReq()

        if address is not None:
            request.address = address
        if channel is not None:
            request.channel  = channel

        request.board    = board
        request.bits     = bits
        request.encoding = toProtobufEncoding(encoding)

        # perform the message transaction
        self.hwdTransaction(request)

    def fireTimeCodePulse(self, tcp = None):
        """ Fires a time code pulse on the (SpaceWire) interface

            :param tcp:   value to use in the Time Code Pulse
        """
        request = hwd_proto.FireTimeCodePulseReq()
        if tcp is not None:
            request.tcp = tcp

        # perform the message transaction
        self.hwdTransaction(request)

