""" Python interface to hardware devices.

    This module is the python interface to communicate with hardware devices
    through the GSSw. This module facilitates performing register reads and
    write to and from physical hardware. Besides regular reads and writes,
    this module also support block reads and writes.
"""

from   collections            import namedtuple
import logging
import threading
import numpy as np

import gssw.proto.hwregister_pb2 as hw_proto

from gssw.proto.common_pb2    import ErrorRep
from gssw.lib.sockets         import ReqSocket, SubSocket, tagName
from gssw.lib.errors          import CommunicationError, OutOfBoundsError, DataError
from gssw.lib.misc            import toProtobufEncoding
from gssw.lib.typing          import check_class

logger = logging.getLogger(__name__)

SetGetData = namedtuple('SetGetData', 'values statuses timestamp')
SetGetData.__doc__ = """
Extended information returned from an extended read operation.

Below are some examples of typical usage:

        ::

          # Query hardware register, and check status field
          getExt = device.get('ADC', True)
          print("The ADC value was: {}".format(getExt.values))
          if getExt.statuses != 0:
              # Note: the status field interpretation is application-specific
              print("The ADC status field was: {}".format(getExt.statuses))
"""
try:
    # Create Sphinx docstrings (appear in generated documentation)
    #pylint: disable=no-member
    SetGetData.values.__doc__    = "List of values, or single value for single read."
    SetGetData.statuses.__doc__  = "List of status fields, or single status field for a single " \
                                   "read, or None when not available.\n\n" \
                                   ":note: The status field interpretation is application-specific."
    SetGetData.timestamp.__doc__ = "Timestamp sent by the hardware, or None when not available.\n" \
                                   "\n:note: The timestamp interpretation is application-specific."
    #pylint: enable=no-member
finally:
    pass

    
class Device:
    """ The Device class encapsulates all functions to interface with the hardware device. """

    def __init__(self, deviceName, config, hal, timeout=5, daemonName=None, strict=True):
        """ Constructor for the Device class.

            :param str deviceName: name of the device in the HAL.
            :param hal:            hardware abstraction layer object.
            :param config:         system configuration
                                   (a dictionary loaded from the config.yaml file).
            :param float timeout:  [s] communication timeout for device read and write operations.
                                   set to -1 for infinite.
            :param daemonName:     Override daemon via which to connect to the Device.
                                   Can be used to force deviceName's traffic via custom daemon.
            :param strict:         Whether to raise exceptions for received packets with unexpected
                                   values. Can be used when an unknown device is connected.
        """

        # Initialize variables.
        self.deviceName = deviceName
        self.strict     = strict

        # Hardware abstraction layer
        assert hal is not None, "Hal must be given"
        self.hal = hal

        logger.debug('Initializing python device interface')

        # Get some device parameters from the HAL
        deviceObject = self.hal.getDevice(deviceName=self.deviceName)
        self.apid = deviceObject.apid
        self.board = deviceObject.port # This is the device's physical address (e.g. backplane slot)

        # Set HwDaemon to send all register requests to (optionally overridden via argument)
        self.daemonName = deviceObject.daemon if daemonName is None else daemonName
        logger.debug('daemonName: %s', self.daemonName)

        # Get the configuration for the HW daemon the device class will comunicate with
        deviceConfig = config.componentParameters(self.daemonName)
        logger.debug('configuration: %s', str(deviceConfig))

        # Create a REQ socket to communicate with the device's hw_daemon.
        # The socket is opened just before usage
        self._sockets = {}
        self.timeout = timeout
        self.repAddress = 'tcp://%s:%d' % (deviceConfig['hostname'],
                           deviceConfig['sockets']['device-rep'])


    def logTCTM(self, tctmType, registerName, address, addressOffset, rawValue, itemWrite):
        # pylint: disable=line-too-long
        # (I'm keeping the long lines below for clarity, sorry pylint)
        """ Logs collected request info from getSetMultiple into logging.data

            :param string TCTMype:      Type for log: TC (TeleCommand) or TM (TeleMetry)
            :param str registerName:    Register Name
            :param int Address:         Register Address
            :param int addressOffset:   Address offset (TC Only, None for TM)
            :param int value:           (Raw) Value (before changes)
            :param bool item.write:     To check if Read/Write 

            Example output:
            TC/TM: Read/Write  request/reply   deviceName          Board     ItemName           (RegAddress, hex)      Address Offset       (Raw) Value       
            TC:    Read         reqst          vp780_fmc126_adc1,  Board: 4, OSC3_ADC_PLL_BASE  (3474208, 0x00350320), offset = 0x00000003                    
            TM:    Read         reply          vp780_fmc126_adc1,  Board: 4, OSC3_ADC_PLL_BASE  (3474211, 0x00350323),                      value = 0x00000002
            TC:    Write        reqst          sron_3ghz_lo1,      Board: 2, LED                (      1, 0x00000001), offset = 0x00000000, value = 0x00000001
            TM:    Write        reply          sron_3ghz_lo1,      Board: 2, LED                (      1, 0x00000001)

        """
        # pylint: enable=line-too-long

        if not (hasattr(logging, "DATA") and logger.isEnabledFor(logging.DATA)):  # pylint: disable=no-member
            return # Skip expensive logging
        # Error Checks
        if tctmType not in ["TC","TM"]:
            raise ValueError('tctmType should be TC or TM, but is ({})'.format(tctmType))
        check_class(registerName,str)
        check_class(address,int)
        check_class(addressOffset,int,allow_none=True)
        check_class(itemWrite,bool)
        
        reqType = "Write" if itemWrite else "Read "
        rpqType = "reqst" if tctmType == "TC" else "reply"
        
        # Log Output Generation
        msg = ""
        msg += tctmType + ": "
        msg += reqType + " "
        msg += rpqType + " "
        msg += str(self.deviceName) + ", " 
        msg += "Board: " + str(self.board) + ", "
        msg += str(registerName) + " ("
        msg += str(address) + ", "
        msg += "0x{:08x}".format(address) + ")"
        if addressOffset is not None:
            msg += ", offset = " + "0x{:08x}".format(addressOffset)
        if (tctmType == "TC" and itemWrite) or (tctmType == "TM" and not itemWrite):
            if isinstance(rawValue, int):
                msg += ", value = " + "0x{:08x}".format(rawValue) + ""
            else:
                msg += ", value = {}".format(rawValue)
        
        # Logging at the data level
        logger.data(msg)


    @property
    def socket(self):
        """ To make this class thread-safe, the underlying ReqSocket objects are duplicated 
            per thread (The ReqSocket class is not thread safe because of the zmq sockets). """
        threadId = threading.get_ident()
        if threadId not in self._sockets:
            logger.debug('Creating new ReqSocket for device: %s (thread %s)',
                         self.deviceName, threadId)
            self.setupSocket()
        return self._sockets[threadId]

    @socket.setter
    def socket(self, value):
        threadId = threading.get_ident()
        self._sockets[threadId] = value

    def setupSocket(self):
        """ Sets up the REQ-REP socket """
        self.socket = ReqSocket()
        self.socket.createZmqSocket(timeoutMs=1000 * self.timeout)
        self.socket.connectSocket(self.repAddress)


    def get(self, identifier, extended=False):
        """ Read the value of a register in the HW device.

            :param str identifier: The name of the register to read, possibly extended with
                                   an address offset. Example: "name[:offset]"
            :param bool extended:  Whether to return extended information or not. Set to False to
                                   only return register values.
        """

        getResult = self.getSetMultiple([(identifier, None)], extended)
        if extended:
            # unpack sample value which is still in a list (and status field)
            return SetGetData(getResult.values[0], getResult.statuses[0], getResult.timestamp)
        return getResult[0] # send sample only


    def set(self, identifier, value):
        """ Set the value of a register in the HW device.

            :param str identifier: The name of the register to read, possibly extended
                                   with an address offset.
                                   Example: "name[:offset]"
            :param int value:      The value that is written to the register.
        """

        return self.getSetMultiple([(identifier, value)])


    def getEng(self, identifier, extended=False):
        """ Read the engineering value of a register.

            The raw register value is read and converted to
            an engineering value using the get polynome.

            :param str identifier: The name of the register to read, possibly extended
                                   with an address offset.
                                   Example: "name[:offset]"
            :param int offset:     An offset to the register address corresponding to 'name'.
            :param bool extended:  Whether to return extended information or not. Set to False to
                                   only return register values.
        """

        # split the name and offset
        registerName, _ = self._splitRegisterIdentifier(identifier)

        # perform the read from the device
        getResult = self.getSetMultiple([(identifier, None)], extended)
        raw = getResult.values[0] if extended else getResult[0]

        # retrieve the conversion object from the HAL
        conversion = self.hal.getRegisterConversion(self.deviceName, None, None, registerName)

        # Show a nice error in case no conversion is defined
        assert conversion is not None, \
            "Device {} register {} has no get conversion".format(self.deviceName, registerName)

        # apply the polynome and return
        eng = conversion.applyConversion(raw, True)

        if extended:
            # unpack sample value which is still in a list
            return SetGetData(eng, getResult.statuses[0], getResult.timestamp)
        return eng # send sample only


    def setEng(self, identifier, eng):
        """ Write to a register using an engineering value.

            The set polynome is first applied to the value
            before writing to the register.

            :param str identifier: The name of the register to read, possibly extended
                                   with an address offset. Example: "name[:offset]"
            :param int offset:     An offset to the register address corresponding to 'name'.
            :param float value:    The engineering value to write.
            :note: The value is clipped to the digital range of the register.

            :return: Returns the actual engineering value written to the register
                     (after converting to a raw value, and back).
        """

        # split the name and offset
        registerName, _ = self._splitRegisterIdentifier(identifier)

        # retrieve the set polynome from the HAL
        conversion = self.hal.getRegisterConversion(self.deviceName, None, None, registerName)

        # Show a nice error in case no conversion is defined
        assert conversion is not None, \
            "Device {} register {} has no set conversion".format(self.deviceName, registerName)

        # apply the polynome to the data
        raw = conversion.applyConversion(eng, False)

        def _clip(value, minLimit, maxLimit):
            """ Clipping function for clipping to the digital range of the register. """
            return min(max(value, minLimit), maxLimit)

        # Clip the raw value to the digital range
        regInfo = self.hal.getRegisterInfo(self.deviceName, registerName)
        if   regInfo.type == 'uint':
            raw = _clip(raw, 0, 2 ** regInfo.width - 1)
        elif regInfo.type == 'int':
            raw = _clip(raw, -(2 ** (regInfo.width - 1)), 2 ** (regInfo.width - 1) - 1)
        # else: double, don't clip

        # write the data to the device
        # note: is raw is None, then read operation is performed, result is discarded
        self.getSetMultiple([(identifier, raw)])

        return conversion.applyConversion(raw, True)


    def getMultiple(self, identifiers, extended=False):
        """ Perform multiple register reads on the device.

            :param list identifier: A lsist of names of the registers to read, possibly extended
                                    with an address offsets.
                                    Example: ["name[:offset]", ...]
            :param bool extended:   Whether to return extended information or not. Set to False to
                                    only return register values.
        """

        return self.getSetMultiple([(i, None) for i in identifiers], extended)


    def setMultiple(self, operations):
        """ Perform multiple register writes on the device.

            :param list operations: A list containing tuples of register identifiers and
                                    the values to write to them.
                                    Example: [("name[:offset]", value), ...].
        """

        # check if all operations have a valid value field
        _identifiers, values = zip(*operations)
        if None in values:
            raise TypeError('One of the write operations has no data.')

        return self.getSetMultiple(operations)


    def getSetMultiple(self, operations, extended=False):  #pylint: disable=too-many-branches, too-many-statements
        """ Perform multiple read or write operations on the device.

            This function handles register reads and writes in batch mode.
            The return value is dependent on the type of operations. A single value (in a list) is
            returned in case of a single read. Multiple read requests returns a list of results.

            Additionally, the timestamp and status fields from the operation can be returned using
            the ``extended`` parameter.

            Writes return nothing. If multiple reads and writes are performed, a list is
            returned with the results of the reads only.

            :return: :class:`SetGetData`. Where entries may become None in case of absent fields.

            :param list operations: A list containing tuples of the names of registers and
                                    the value to write to them. set value to None for reads.
                                    Example: [("name[:offset]", value), ...,("name[:offset]", None)]
            :param bool extended:   Whether to return extended information or not. Set to False to
                                    only return register values.
        """

        # create a request protobuf object
        request = hw_proto.HwRegisterReq()
        request.apid = self.apid
        request.board = self.board

        # add an item to the HwRegisterReq for each operation
        for identifier, value in operations:

            # split the identifier into a name and offset
            registerName, addressOffset = self._splitRegisterIdentifier(identifier)

            # get the register properties from the HAL
            regInfo = self.hal.getRegisterInfo(self.deviceName, registerName)

            # check if the register is not a block register
            if regInfo.block is True:
                raise TypeError('Cannot use regular read or write on blockregister %s.' %
                    registerName)

            # check if the offset is within the bounds for this register
            if addressOffset > regInfo.maxOffset:
                raise OutOfBoundsError(
                    'Register address offset is too large (%d > %d) for register %s.'
                    % (addressOffset, regInfo.maxOffset, registerName))

            # collects data for (TC) Data logger (Below)
            # regInfo, address, addressOffset, rawValue, (itemWrite: added later)
            tcInfo = [regInfo.name, regInfo.address, addressOffset, value, None]
            
            # perform some checks on the write value
            if value is not None:

                # check if the raw value fits in the physical register
                errMsg = None
                if   regInfo.type == 'uint' and int(value) < 0:
                    errMsg = 'Raw value %s too small for %s bit %s register %s:%s' % (
                        value, regInfo.width, regInfo.type, registerName, addressOffset)
                elif regInfo.type == 'uint' and int(value) > 2 ** regInfo.width - 1:
                    errMsg = 'Raw value %s too large for %s bit %s register %s:%s' % (
                        value, regInfo.width, regInfo.type, registerName, addressOffset)
                elif regInfo.type == 'int' and int(value) < -(2 ** (regInfo.width - 1)):
                    errMsg = 'Raw value %s too small for %s bit %s register %s:%s' % (
                        value, regInfo.width, regInfo.type, registerName, addressOffset)
                elif regInfo.type == 'int' and int(value) > 2 ** (regInfo.width - 1) - 1:
                    errMsg = 'Raw value %s too large for %s bit %s register %s:%s' % (
                        value, regInfo.width, regInfo.type, registerName, addressOffset)

                # retrieve the set polynome from the HAL
                conversion = self.hal.getRegisterConversion(self.deviceName, 
                                                            None, 
                                                            None, 
                                                            registerName)

                # check if the value falls within the limits defined in the HAL
                if conversion: # if no set polynome is defined, limits cannot be checked

                    # get the register limits
                    limits = self.hal.getRegisterLimits(self.deviceName, None, None, registerName)

                    # check if the limits are valid
                    if limits and (limits.lowerValid != 0 or limits.upperValid != 0):

                        # calculate the engineering value (value is raw as we would read it from the
                        #   device, so convert using GET-conversion
                        engValue = conversion.applyConversion(value, get=True)

                        # In case of LUTs, there is no eng value that can be checked against limits
                        if engValue is not None and isinstance(engValue, (int, float)):
                            # compare with the limts and throw an exception if it is out of bounds
                            if engValue < limits.lowerValid or engValue > limits.upperValid:
                                errMsg = ('Engineering value %f falls outside the limits [%f, %f] '
                                    'set for register %s in the Hardware Abstraction Layer' %
                                    (engValue, limits.lowerValid, limits.upperValid, registerName))
                        
                if errMsg:
                    logger.error(errMsg)
                    raise DataError(errMsg)

            item = request.item.add()
            item.addr = regInfo.address + addressOffset
            item.bits = regInfo.width 
            if value is None: # value argument is None for read requests
                item.write = False
                value = 0 # Send a value in requests, enabling the Hardware daemon to know the type
            else:
                item.write = True

            # Always put a value of the correct type in the request (even for read request)
            #   This enables the HwDaemon to put the type in CCSDS/SKDP packets (same as item.bits).
            if regInfo.type == 'int':
                # cast to int because the polynome might convert the value to a float
                item.vint = int(value)
            elif regInfo.type == 'uint':
                item.vuint = int(value)
            elif regInfo.type == 'float':
                item.vdbl = float(value)
            else:
                raise ValueError('Unknown data type (%s) defined in the HAL for register: %s'
                    % (regInfo.type, registerName))

            # Add if req is read/write to TCInfo list
            tcInfo[-1] = item.write
            # Call for logTCTM which will log TC Data into logger.data()
            self.logTCTM("TC",*tcInfo) #unpack to function args

        # perform the message transaction
        reply = self.hwRegisterTransaction(request)

        # check for errors
        if not isinstance(reply, hw_proto.HwRegisterRep):
            raise CommunicationError('Received incorrect protobuf type from device')
#        if reply.apid != self.apid:
#            raise CommunicationError('Received protobuf with incorrect apid')
        if self.strict and reply.board != self.board:
            raise CommunicationError('Received protobuf with incorrect slot id (expected: {}, '
                'received: {})'.format(self.board, reply.board))

        # create a list with read values to return
        # (also check the data type for each against the HAL)
        values   = []
        statuses = []

        for item in reply.item:
            if not item.write:

                # get the name of the register
                regInfo = self.hal.getRegisterInfo(self.deviceName, regInfo.address)

                if item.WhichOneof('value') == 'vint':
                    # if regInfo.type != 'int':
                    #     raise TypeError("Expected %s but received an int for register %s"
                    #         % (regInfo.type, regInfo.name))
                    # else:
                    values.append(item.vint)
                elif item.WhichOneof('value') == 'vuint':
                    # if regInfo.type != 'uint':
                    #     raise TypeError("Expected %s but received an uint for register %s"
                    #         % (regInfo.type, regInfo.name))
                    # else:
                    values.append(item.vuint)
                elif item.WhichOneof('value') == 'vdbl':
                    # if regInfo.type != 'float':
                    #     raise TypeError("Expected %s but received an float for register %s"
                    #         % (regInfo.type, regInfo.name))
                    # else:
                    values.append(item.vdbl)
                else:
                    raise ValueError('Data field empty for read reply')

                statuses.append(item.status if item.HasField('status') else None)

            # collects data for (TM) Data logger (Below)
            #regInfo, address, addressOffset = None, rawValue, itemWrite
            tmInfo = [regInfo.name, item.addr, None, values[0] if values else None, item.write]
            # Call for logTCTM which will log TM Data into logger.data()
            self.logTCTM("TM",*tmInfo) #unpack to function args

        if extended:
            return SetGetData(values, statuses,
                              reply.timestamp if reply.HasField('timestamp') else None)
        # return only the values, no timestamps or status info 
        return values

    def getBlock(self, identifier, samples):
        """ Read a blockdata device register.

            The register must be a block register.
            A list of read values is returned.

            :param str identifier: The name of the register to read, possibly extended
                                   with an address offset.
                                   Example: "name[:offset]"
            :param int offset:     An offset to the register address corresponding to 'name'.
            :param int samples:    The number of samples to read from the register.
        """

        # split the identifier into a name and offset
        (registerName, addressOffset) = self._splitRegisterIdentifier(identifier)

        # get the register properties from the HAL
        regInfo = self.hal.getRegisterInfo(self.deviceName, registerName)

        # check if the register really is a block register
        if not regInfo.block:
            raise TypeError('Block read performed on non-block register.')

        # check if the offset is within the bounds for this register
        if addressOffset > regInfo.maxOffset:
            raise OutOfBoundsError('Register address offset is too large (%d > %d).'
                % (addressOffset, regInfo.maxOffset))

        # create a request protobuf object
        request = hw_proto.HwBlockReq()

        # fill in the request protobuf objects fields.
        request.apid = self.apid
        request.board = self.board
        request.addr = regInfo.address + addressOffset
        request.write = False
        request.nsamples = samples

        # perform the message transaction
        reply = self.hwRegisterTransaction(request)

        # Reply error checking.
        # The datatype is not always fixed for block registers so no type checking
        if not isinstance(reply, hw_proto.HwBlockRep):
            raise TypeError('received wrong type of protobuf type from device: %s.'
                % type(reply).__name__)
        if reply.write:
            raise TypeError('Received a write reply in response to a read request.')
        if reply.HasField("apid") and reply.apid != self.apid+1:
            raise CommunicationError('Received reply with incorrect apid (%d instead of %d).' %
                (reply.apid, self.apid+1))
        if reply.board != self.board:
            raise CommunicationError('Received reply with incorrect board (%d instead of %d).' %
                (reply.board, self.board))

        # check the bitwidth of the reply's data against the HAL
        if reply.data.bits_per_value != regInfo.width:
            raise TypeError("expected blockdata of %d bits wide but received %d bits wide data"
                % (regInfo.width, reply.data.bits_per_value))

        # extract the read data from the reply
        values = self._deserialize(reply.data.block_data, regInfo.width, regInfo.type)
        logger.debug("Received blockdata: %s", str(values))

        return values


    def setBlock(self, identifier, data):
        """ Write blockdata to a device register

            The register must be a block register.

            :param str identifier: The name of the register to read, possibly extended
                                   with an address offset.
                                   Example: "name[:offset]"
            :param int offset:     An offset to the register address corresponding to 'name'.
            :param int samples:    The number of samples to read from the register.
            :param list data:      A list containing the values to write.
        """

        if not isinstance(data, (list, np.ndarray)):
            raise TypeError('SetBlock was called without a list of values to write.')

        # split the identifier into a name and offset
        (registerName, addressOffset) = self._splitRegisterIdentifier(identifier)

        # get the register properties from the HAL
        regInfo = self.hal.getRegisterInfo(self.deviceName, registerName)

        # check if the register is indeed a block register
        if not regInfo.block:
            raise ValueError('Register {!r} is not a block register.'.format(registerName))

        # check if the offset is within the bounds for this register
        if addressOffset > regInfo.maxOffset:
            raise OutOfBoundsError('Register address offset is too large (%d > %d).'
                % (addressOffset, regInfo.maxOffset))

        # create a request protobuf object
        request = hw_proto.HwBlockReq()

        # enforce alignment: necessary for HwDaemon also
        assert regInfo.width in [ 8, 16, 32, 64 ]

        # fill in the request protobuf objects fields.
        request.apid = self.apid
        request.board = self.board
        request.addr = regInfo.address + addressOffset
        request.write = True
        request.data.bits_per_value = regInfo.width
        request.data.block_data = self._serialize(data, regInfo.width, regInfo.type)
        request.nsamples = len(data)
        request.data.value_encoding = toProtobufEncoding(regInfo.type)

        # perform the message transaction
        reply = self.hwRegisterTransaction(request)

        # reply error checking.
        if not isinstance(reply, hw_proto.HwBlockRep):
            raise TypeError('received wrong type of protobuf type from device: %s'
                % type(reply).__name__)
        if reply.HasField("apid") and reply.apid != self.apid:
            raise CommunicationError('Received reply with incorrect apid.')
        if reply.board != self.board:
            raise CommunicationError('Received reply with incorrect board.')
        if not reply.write:
            raise TypeError('Received a read reply in response to a write request.')



    def getBlockEng(self, identifier, samples):
        """ Read the engineering value from a device blockdata register.

            This function reads the raw value from the register and converts
            this to an engineering value using the get polynome.
            The register must be a block register. A list of read values is returned.

            :param str identifier: The name of the register to read, possibly extended
                                   with an address offset.
                                   Example: "name[:offset]"
            :param int offset:     An offset to the register address corresponding to 'name'.
            :param int samples:    The number of samples to read from the register
        """

        # split the name and offset
        registerName, _ = self._splitRegisterIdentifier(identifier)

        # get the get polynome from the HAL
        conversion = self.hal.getRegisterConversion(self.deviceName, None, None, registerName)

        # perform a block read from the device
        data = self.getBlock(identifier, samples)

        # perform the conversion using the polynome
        for value in data:
            value = conversion.applyConversion(value, True)

        return data


    def setBlockEng(self, identifier, data):
        """ Write blockdata to a device register.

            The data is converted to raw values using the set polynome before
            it is written to the register. The register must be a block register.

            :param str identifier: The name of the register to read, possibly extended
                                   with an address offset.
                                   Example: "name[:offset]"
            :param int offset:     An offset to the register address corresponding to 'name'.
            :param list data:      A list containing the (engineering) values to write.
        """

        # split the name and offset
        registerName, _ = self._splitRegisterIdentifier(identifier)

        # get the get polynome from the HAL
        conversion = self.hal.getRegisterConversion(self.deviceName, None, None, registerName)

        # convert the write data using the set polynome
        for value in data:
            value = conversion.applyConversion(value, False)

        # write the block data to the device
        self.setBlock(identifier, data)


    def _deserialize(self, data, bitWidth, datatype):
        """ Deserialize a string of bytes into a list of elements of datatype and bitWidth. """

        dtype = self._getNumpyType(bitWidth, datatype)
        array = np.fromstring(data, dtype)

        return array


    def _serialize(self, data, bitWidth, datatype):
        """ Serialize an array of elements of datatype and bitWidth to a string. """

        dtype = self._getNumpyType(bitWidth, datatype)
        array = np.array(data, dtype=dtype)
        string = array.tostring()

        return string


    def _getNumpyType(self, bitWidth, datatype):
        """ Get the numpy datatype for a bitWidth and datatype identifier. """

        # the bitwidth must be a multiple of 8 bits with a maximum of 64
        if bitWidth > 64 or bitWidth & 0x7 != 0:
            raise TypeError('Unsupported bitwith (%d). Should be: 8, 16, 32, or 64.'
                % bitWidth)

        elif datatype == 'int':
            dtype = np.dtype('>i' + str(int(bitWidth) >> 3))

        elif datatype == 'uint':
            dtype = np.dtype('>u' + str(int(bitWidth) >> 3))

        elif datatype == 'string':
            dtype = np.dtype('a' + str(int(bitWidth) >> 3))

        elif datatype == 'float':
            if bitWidth & 0x1F != 0:
                raise TypeError('unsupported bitwidth (%d) for float. Should be: 32 or 64'
                    % bitWidth)
            else:
                dtype = np.dtype('<f' + str(int(bitWidth) >> 3))

        else:
            raise TypeError('unsupported datatype')

        return dtype

    def hwRegisterTransaction(self, request):
        """ Perform a communication transaction (request, reply) with the hardware device.

            :param request: HwRegisterReq or HwBlockReq protobuf object.

            :return: Normally returns a protobuf :class:`HwRegisterRep` or :class:`HwBlockRep`
                     object. Alternatively, if an error occurred, an ErrorRep object is returned.
                     In case of a delivery acknowledge (OkRep), None is returned.
        """

        # first transaction or after a timeout
        if self.socket is None:
            self.setupSocket()

        # print the request
        logger.debug("sending request:\n%s", request)

        # receive the reply
        try:

            # send the header string and the serialized request
            self.socket.sendProtoBuf(tagName(request), request.SerializeToString())

            # receive the reply
            tag, message = self.socket.receiveProtoBuf()

        except Exception as e:
            self.socket.close()
            self.socket = None
            logger.warning('Error when requesting values from device: %s', e)
            raise CommunicationError('Error receiving reply from daemon {}, '
                                     'probably due to a timeout.'.format(self.daemonName)) from e

        # check that the correct header string is send
        if tag == 'ErrorRep':
            reply = ErrorRep()
            reply.ParseFromString(message)
            logger.error("Received an error reply from deamon '%s' during a device register for "
                         "device %s transaction (%s)", self.daemonName, self.deviceName,
                         ErrorRep.ErrorType.Name(reply.errtype))
            CommunicationError.raiseFromProtoBuffer(reply, 'Device {}'.format(self.deviceName))
        elif tag == 'HwRegisterRep':
            reply = hw_proto.HwRegisterRep()
        elif tag == 'HwBlockRep':
            reply = hw_proto.HwBlockRep()
        elif tag == 'OkRep':    
            # reply from HwStreamReq
            return
        else:
            raise TypeError('Wrong message header received from device:\n%s' % tag)

        # deserialize the received reply
        reply.ParseFromString(message)

        # print the response
        logger.debug("received reply:\n%s", reply)

        return reply

    def _splitRegisterIdentifier(self, identifier):
        """ Split of the offset of a register identifier

            :param str identifier: Register name plus optional offset.
                                   Formatted as: "register:1" or "register"

            :return: Tuple of the register name (string) and the offset (int).
                     The result for the input examples would be:
                     ("register", 1) or ("register", 0)
        """

        lst = identifier.split(':')

        if len(lst) == 1:
            return (lst[0], 0)
        elif len(lst) == 2 and bool(lst[1]):
            return (lst[0], int(lst[1]))
        raise ValueError('Incorrect register name format.\n \
                          Should be: \"register[:offset]\".\n \
                          Is: %s' % identifier)

    def setTimeout(self, timeout):
        """ Sets the timeout value on the req/rep socket """

        if self.socket is None:
            self.setupSocket()

        self.timeout = timeout
        self.socket.setReceiveTimeout(self.timeout)


class StreamDevice(Device):
    """ The Device class encapsulates all functions to interface with the hardware device. 
        This derived class is intended for (raw) CCSDS communication with a hardware device,
        thus not via register access. """


# pylint: disable=super-init-not-called
    def __init__(self, daemonName, config, timeout=5):
        """ Constructor for the Device class.

            :param str deviceName: name of the device in the HAL.
            :param hal:            hardware abstraction layer object.
            :param config:         system configuration
                                   (a dictionary loaded from the config.yaml file).
            :param float timeout:  [s] communication timeout for device read and write operations.
                                   set to -1 for infinite.
            :param daemonName:     Override daemon via which to connect to the Device.
                                   Can be used to force deviceName's traffic via custom daemon.
            :param strict:         Whether to raise exceptions for received packets with unexpected
                                   values. Can be used when an unknown device is connected.
        """
        logger.debug('Initializing python stream device interface')

        self.daemonName = daemonName
        self.deviceName = ""

        # Get the configuration for the HW daemon the device class will comunicate with
        deviceConfig = config.componentParameters(daemonName)
        logger.debug('configuration: %s', str(deviceConfig))

        # Create a REQ socket to communicate with the device's hw_daemon.
        # The socket is opened just before usage
        self._sockets = {}
        self.timeout = timeout

        # Build the REP socket address
        self.repAddress = 'tcp://%s:%d' % (deviceConfig['hostname'], 
                                           deviceConfig['sockets']['device-rep'])

        # Build the PUB socket address
        pubAddress      = 'tcp://%s:%d' % (deviceConfig['hostname'],
                                           deviceConfig['sockets']['device-pub'])

        # Create a sub socket
        self.subSocket = SubSocket()
        self.subSocket.createZmqSocket()
        logger.debug('Connecting to PUB socket @ %s', pubAddress)
        self.subSocket.connectSocket(pubAddress)
        self.subSocket.subscribe()

    def poll(self):
        """ Poll the subscribe socket to check for available messages.
            Returns a boolean indicating whether new messages are available.
        """
        return self.subSocket.readyReceive()


    def getStream(self):
        """ Receive one or more packet(s) from the subscribe socket"""

        msgList = []

        while self.poll():

            # get one message from the sub socket
            _, tag, msg = self.subSocket.receiveProtoBufWithTopic()

            # type checking
            if tag !=  tagName(hw_proto.HwStreamData):
                raise CommunicationError('Unexpected message type. (%s)' % tag)

            # parsing ZMQ message to protocol buffer
            reply = hw_proto.HwStreamData()
            reply.ParseFromString(msg)
            msgList.append(reply)

        return msgList


    def get(self, identifier, extended=False):
        """ Overloaded from base class, not used."""
        return None

    def set(self, identifier, value):
        """ Overloaded from base class, not used."""

    def getEng(self, identifier, extended=False):
        """ Overloaded from base class, not used."""
        return None

    def setEng(self, identifier, eng):
        """ Overloaded from base class, not used."""

    def getSetMultiple(self, operations, extended=False):  
        """ Overloaded from base class, not used."""
        return None

    def getBlock(self, identifier, samples):
        """ Overloaded from base class, not used."""
        return None

    def setBlock(self, identifier, data):
        """ Overloaded from base class, not used."""

    def getBlockEng(self, identifier, samples):
        """ Overloaded from base class, not used."""
        return None

    def setBlockEng(self, identifier, data):
        """ Overloaded from base class, not used."""

    def setStream(self, apid, data, pre=None, sec_header=None, seq_counter=None):
        """ Write a stream of (raw) data to a device.

            :param int  apid:        The apid to be used.
            :param list data:        The raw byte stream to sent to a device.
            :param list pre:         Send (routing) bytes before the actual packet.
            :param bool sec_header:  Control value of secondary header flag in packet header.
            :param int  seq_counter: Reset packet sequence counter to this value.
        """

        if not isinstance(data, (np.ndarray, list, bytes)):
            raise TypeError('SetStream was called without a numpy array with values to write.')

        # create a request protobuf object
        request = hw_proto.HwStreamReq()

        # fill in the request protobuf objects fields.
        request.apid = apid
        if isinstance(data, np.ndarray):
            request.data = self._serialize(data, 8, 'int')
        elif isinstance(data, bytes):
            request.data = data
        elif isinstance(data, list):
            request.data = ''.join(chr(item) for item in data).encode()
        if pre is not None:
            assert isinstance(pre, list), "Pre packet data must be list of bytes"
            request.pre_packet_data = ''.join(chr(item) for item in pre).encode()
        if sec_header is not None:
            request.sec_header = sec_header
        if seq_counter is not None:
            request.seq_counter = seq_counter

        # perform the message transaction
        reply = self.hwRegisterTransaction(request)

        # reply error checking.
        if reply is None:
            # OkRep received, all done now
            return

        if isinstance(reply, ErrorRep):
            logger.error("Received an error reply from deamon '%s' during a stream request "
                         "transaction (%s)", self.daemonName, 
                         ErrorRep.ErrorType.Name(reply.errtype))
            CommunicationError.raiseFromProtoBuffer(reply, 'Daemon {}'.format(self.daemonName))
        else:
            raise TypeError('received wrong type of protobuf type from daemon: %s'
                % type(reply).__name__)

        if reply.HasField("primheader"):
            return (reply.primheader, reply.data)
        else:
            return reply.data

