""" This module provides the class for transporting diagnostic user messages. """
__all__ = ['DiagnosticLogger']

import logging
import time
import traceback
from gssw.lib.errors            import GsswError
from gssw.lib.sockets           import PubSocket, tagName
from gssw.proto.diagnostics_pb2 import DiagMsg

logger = logging.getLogger(__name__)

class DiagnosticLogger:
    """ Using this class, a script or daemon can send messages which will be shown to the user. """

    def __init__(self):
        """ Make a new logger for diagnostic user messages.

        :note: Do not create directly. Instead use :obj:`.diagnosticlogger` from :mod:`.common`.
        """

        self.pubSocket      = None
        self.componentName  = ''

    def initialize(self, config, componentName):
        """ 
        :param Config config: Egress configuration object.
        """
        self.componentName = componentName

        try:
            proxyConfig  = config.componentParameters('proxy')
            proxyHost    = proxyConfig['hostname']
            proxyPubport = proxyConfig['sockets']['xsub']
            self.pubSocket = PubSocket()
            self.pubSocket.createZmqSocket()
            self.pubSocket.connectSocket("tcp://{}:{}".format(proxyHost, proxyPubport))
        except GsswError as ex:
            # e.g., when there is no 'proxy' in the config file
            logger.exception("Diagnostic logging could not be initialized: %s", ex)

    def _log(self, severity, title, details):
        if not self.pubSocket:
            logger.warning("Could not send diagnostic message to user (message was: '%s')", title)
            return
        msg = DiagMsg()
        msg.title           = title
        msg.details         = details
        msg.timestamp       = time.time()
        msg.severity        = severity
        stacksourcepath     = traceback.extract_stack(limit=3)[0].filename
        msg.sourceinfo      = stacksourcepath

        msg.componentname   = self.componentName
        binMsg = msg.SerializeToString()
        self.pubSocket.publishProtoBuf('/diagnostics/', tagName(msg), binMsg)

    def info(self, title, details=''):
        """ Log user message as INFO.

        :param str title:   Title of the message, or short description.
        :param str details: More details of the message (optional).
        """
        self._log(DiagMsg.INFO, title, details)

    def warning(self, title, details=''):
        """ Log user message as WARNING.

        :param str title:   Title of the message, or short description.
        :param str details: More details of the message (optional).
        """
        self._log(DiagMsg.WARNING, title, details)

    def error(self, title, details=''):
        """ Log user message as ERROR.

        :param str title:   Title of the message, or short description.
        :param str details: More details of the message (optional).
        """
        self._log(DiagMsg.ERROR, title, details)
