#!/usr/bin/env python3
""" Module for performing public key encryption.

:note: The module is inspired by by
       https://www.pythonsheets.com/notes/python-crypto.html#simple-rsa-encrypt-with-oaep
"""

import base64

from Crypto.PublicKey   import RSA
from Crypto.Cipher      import PKCS1_OAEP
from gssw.lib.typing    import check_class

# RSA key length (bits), minimum is 1024
KEYLENGTH = 2048

class PrivateKey():
    """ Class for asymmetrical decryption (on server end).

    Typically, a public key is created (from this private key) on the server.
    When encrypted data is received on the server, it can be decrypted using this key. """

    def __init__(self):
        """ Create a random RSA key (2048 bits). """
        self._privateKey = RSA.generate(KEYLENGTH)
        self._publicKey  = PublicKey(self._privateKey.publickey())

    @property
    def publicKey(self):
        """ Retrieve the public key from this private key, to be used for encryption. """
        return self._publicKey

    def decrypt(self, text):
        """ Decrypt data received from a client.

        :param bytes text: Data to be decrypted.
        """
        check_class(text, bytes)

        # Create a decipherer which uses this private key for decryption 
        decipher = PKCS1_OAEP.new(self._privateKey)

        # decode base64
        decipher_text = base64.b64decode(text)

        # Do the decryption
        return decipher.decrypt(decipher_text)


class PublicKey():
    """ Class for asymmetrical encryption (on client end).

    Typically, the public key is created (from the private key) by the server.
    Then, the public key is serialized and sent to the client. The client can deserialize the public
    key, and then use it for encrypting the data which is to be sent to the server. """

    def __init__(self, rsaPublicKey):
        """ Construct a PublicKey object from an RSA key. """
        check_class(rsaPublicKey, RSA._RSAobj)  #pylint: disable=protected-access
        assert not rsaPublicKey.has_private(), \
            "PublicKey can only be constructed from an RSA public key object"
        self._publicKey = rsaPublicKey

    def serialize(self):
        """ Serialize the key (e.g. for sending to a client).

        :return: bytes object with serialized key.
        """
        return self._publicKey.exportKey('PEM')

    @staticmethod
    def deserialize(serializedPublicKey):
        """ Create a PublicKey object for further used, from a (received) serialized PublicKey.

        :param bytes serializedPublicKey: bytes with serialized key.
        """
        check_class(serializedPublicKey, bytes)
        rsaPublicKey = RSA.importKey(serializedPublicKey)
        return PublicKey(rsaPublicKey)

    def encrypt(self, text):
        """ Encrypt text using asymmetric cryptographic padding

        :param bytes text: Input to be encrypted.
        :return: bytes object with encrypted text.
        """
        check_class(text, bytes)
        # create a cipher object
        cipher = PKCS1_OAEP.new(self._publicKey)
        # encrypt plain text
        cipher_text = cipher.encrypt(text)

        # encode encrypted text via base64
        return base64.b64encode(cipher_text)
