""" Common errors or exceptions for the GSSw """

import re
import traceback

try:
    from google.protobuf.message import DecodeError
except ImportError:
    DecodeError = Exception
from gssw.lib.typing         import type_name, is_binary


class GsswError(Exception):
    """Base class for GSSw exceptions."""
    pass

class ConfigError(GsswError):
    """ Raised when error reading config file. """
    pass

class DataError(GsswError):
    """ Raised upon incorrect or unexpected HAL data """
    pass

class ConversionError(GsswError):
    """ Raised upon unexpected conversion """
    pass

class OutOfBoundsError(GsswError):
    """Raised when an address is out of bounds."""
    pass

class ConsistencyError(GsswError):
    """ Raised when the GriddedArray coordinate vector are inconsistent with the central array """
    pass

class DuplicateItemError(GsswError):
    """ Raised when trying to add an item that already exists in the model."""
    pass

class ItemNotFoundError(GsswError):
    """ Raised when trying to read that does not exists in the model."""
    pass

class ServiceError(GsswError):
    """ Raised when the AbstractService descendant encounters an issue """
    pass

class ServiceConfigError(ServiceError):
    """ Raised when a protobuf service contains two methods having the same input message type """
    pass

class ServiceProcessingError(ServiceError):
    """ Raised when the AbstractService descendant cannot process a message """
    pass

class UnsupportedServiceError(ServiceProcessingError):
    """ Raised when the an unsupported message type is received """

class GsswExportError(GsswError):
    """ Raised when data cannot be exported to CSV"""
    pass


###################################################################################################

class GsswTimeoutError(GsswError):
    """ Raised when a time out occurs. Typically when using a ZMQ socket"""
    pass

class GsswSocketError(GsswError):
    """ Raised when a non-timeout exceptions occur. Typically when using a ZMQ socket"""

class PeerException(GsswError):
    """ Raised when the peer (ZMQ client/server) raised an exception, while processing a
        RPC message.

        The exception info from the peer is included in the PeerExceptionRep. The PeerExceptionRep
        protocol buffer and this exception is handled in RPC base classes.

        Typically, when you encounter a PeerException the actual issue occurred in the Peer, and you
        want to look in the log output of the peer for the corresponding stack trace.
    """
    def __init__(self, *args, **kwargs):
        """ Constructor.

            :param args:
            :param kwargs:
        """
        # import here to prevent circular deps
        from gssw.proto.common_pb2 import PeerExceptionRep

        if is_binary(args[0]):
            # Assumes args[0] contains a PeerExceptionRep
            binMsg = args[0]
            msg = PeerExceptionRep()
            msg.ParseFromString(binMsg)
            super().__init__(msg.type_name, msg.err_msg, *args[1:], **kwargs)
        elif isinstance(args[0], PeerExceptionRep):
            msg = args[0]
            super().__init__(msg.type_name, msg.err_msg, *args[1:], **kwargs)
        else:
            super().__init__(*args, **kwargs)

    @staticmethod
    def exceptionToMessage(ex):
        """ Serializes an Exception to a PeerExceptionRep protocol buffer message

            :param ex: Exception
            :return: (replyTag, binMessage) tuple, where replyTag='PeerExceptionRep' 
                and the binMessage is the exception message serialized to a bytestring.
        """
        # import here to prevent circular deps
        from gssw.proto.common_pb2 import PeerExceptionRep
        from gssw.lib.sockets import tagName  # pylint: disable=cyclic-import

        msg = PeerExceptionRep()
        msg.type_name = type_name(ex)
        msg.err_msg = str(ex)
        if ex.__context__:

            # try adding relevant traceback.
            tbcks = traceback.format_tb(ex.__context__.__traceback__)
            tbck = [tb for tb in tbcks] # show both gssw and project stack. if 'gssw' in tb]
            tbck = tbck[-1] if bool(tbck) else tbcks[-1]
            tbck = re.sub(r'\n\s*', ', ', tbck)
            tbck = re.sub(r'\s*File', 'file', tbck)

            msg.err_msg += " caused by " + tbck + type_name(ex.__context__) + \
                           ": " + str(ex.__context__)

        binMessage = msg.SerializeToString()
        replyTag = tagName(PeerExceptionRep)

        return replyTag, binMessage


###################################################################################################


class CommunicationError(GsswError):
    """ Raised when something went wrong in hwdaemon with communication to a device.
        Raised when the Hardware Daemon cannot process a request.
    """

    @staticmethod
    def raiseFromProtoBuffer(error_reply, extra_msg=''):
        """ Raise error-type specific CommunicationError based on ErrorRep protocol buffer.

        :param error_reply:   Protobuffer object containing the error. If a serialized protobuf is
                              passed, it will be deserialized.
        :param str extra_msg: Extra user-understandable text to be included in the raised exception.
        """
        # Import here to prevent circular dependencies
        from gssw.proto.common_pb2   import ErrorRep
        from gssw.lib.sockets import ReqRepParts  # pylint: disable=cyclic-import

        if isinstance(error_reply, ReqRepParts):
            protobuf = ErrorRep()
            try:
                protobuf.ParseFromString(error_reply.message)
                error_reply = protobuf
            except DecodeError:
                pass
        if not isinstance(error_reply, ErrorRep):
            raise CommunicationError("Received unsupported message: " + str(error_reply))

        # Add extra user details, when given
        if extra_msg:
            errorMsg = '{}: {}'.format(extra_msg, error_reply.errmsg)
        else:
            errorMsg = error_reply.errmsg

        if error_reply.errtype==ErrorRep.ERR_ROUTE:
            # can not route request to destination
            raise CommunicationErrorRoute(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_DATATYPE:
            # unknown request data type
            raise CommunicationErrorDataType(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_HARDWARE:
            # hardware error or hardware not connected
            raise CommunicationErrorHardware(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_TIMEOUT:
            # request timed out
            raise CommunicationErrorTimeout(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_APID:
            # unknown APID
            raise CommunicationErrorAPID(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_BOARD:
            # unknown board
            raise CommunicationErrorBoard(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_SYSTEM:
            # System/OS failure.
            raise CommunicationErrorSystem(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_STATE:
            # request not supported in current state
            raise CommunicationErrorState(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_TRANSLATE:
            # can not translate to hardware protocol
            raise CommunicationErrorTranslate(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_FILEIO:
            # file IO error
            raise CommunicationErrorFileIO(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_PROTO:
            # Unknown or unsupported request type
            raise CommunicationErrorProto(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_VALUE:
            # invalid or out-of-range value
            raise CommunicationErrorValue(errorMsg)

        elif error_reply.errtype==ErrorRep.ERR_BUSY:
            # request rejected due to ongoing activity
            raise CommunicationErrorBusy(errorMsg)

        else:
            raise CommunicationError("Error of unkown type({}) '{}'".
                    format(error_reply.errtype, error_reply.errmsg))


class CommunicationErrorRoute(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorDataType(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorHardware(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorTimeout(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorAPID(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorBoard(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorSystem(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorState(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorTranslate(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorFileIO(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorProto(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorValue(CommunicationError):
    """ communication error """
    pass

class CommunicationErrorBusy(CommunicationError):
    """ communication error """
    pass


class SocketError(GsswError):
    """ Network / communication failure """
