'''
This module provides a lock-key mechanism which is used by a PHC server and client.

It is used for locking the PHC server during measurements.

Created on May 23, 2017

@author: simonr
'''

from   datetime import datetime
import inspect
import logging
import ntpath
import socket
import threading
import time
import traceback

from gssw.lib.errors    import GsswError
from gssw.lib.rpcbase   import BaseClient, ThreadedBaseServer
from gssw.lib.rpcmsg    import Packetizer, DelegatingReqRepService
from gssw.lib.sockets   import ReqSocket, RepSocket
from gssw.lib.typing    import check_class
from gssw.proto.lock_pb2    import  LockKeepaliveServiceSpec,\
                                    SystemLockReq,            SystemLockRep,\
                                    SystemLockKeepaliveReq
from gssw.proto.common_pb2  import  OkRep

logger = logging.getLogger(__name__)

#: The lock is kept for this duration (by default), after which it is released, if not kept alive.
LOCK_TIMEOUT_DEFAULT = 10   # seconds
LOCK_TIMEOUT_THRESHOLD = 1  # additional threshold before a lock times out
                            # (avoids a race condition between timeout and keep-alive)

class PhcLock():
    """ This class provides a statemachine on the PhcServer

        The lock is set and released via the handle_request method.

        The lock also implements a time-out mechanism, which releases the lock after a timeout.
        This prevents clients which do not release the lock when exitting from locking up the PHC.
        The timeout mechanism runs in a separate thread.
        The lock must be kept alive via ACTIVATE requests.

        The PhcLockKey runs in the PhcServer context.
     """

    def __init__(self, timeout = LOCK_TIMEOUT_DEFAULT):
        """ Creates the lock

        :param timeout: Timeout after which the lock is released (in seconds)
        """
        self._user_id = None # User for which the PHC is locked
        self._key     = None # Token that must be used by the client
        self._keepalive_lastupdate = 0
        self._keepalive_timeout    = timeout + LOCK_TIMEOUT_THRESHOLD
        self._keepalive_timer      = None
        self._thread_lock = threading.RLock()

    def is_active(self):
        """ Reports whether the lock is closed by a user (i.e. whether someone is using the PHC)

        :return:        bool
        """
        with self._thread_lock:
            return self._user_id is not None

    def handle_request(self, lock_type, user_id, key):
        """ Activates or releases the lock for a PhcClient

            The parameters are specified in the lock protobuf

        :param lock_type: ACTIVATE or RELEASE
        :param user_id:   identifier for which user is locking the PHC
        :param key:       'password' (required for activating the lock or for keep-alives)
        :return: dict with the folowing contents:

            * 'result':
              Status code (as specified in the protobuf),
            * 'user_id':
              Identifier of who has locked the system
              (useful in case the sysem is occupied by some other script),
            * 'key':
              ‛Password’ that must be used for further locking requests.
        """
        logger.info("Received lock request: %s for %s, with key: '%s'",
                     SystemLockReq.LockType.Name(lock_type), user_id, key)
        if lock_type == SystemLockReq.ACTIVATE:
            result, user_id, key = self._activate(user_id)
        elif lock_type == SystemLockReq.RELEASE:
            result               = self._release(user_id, key)
        logger.info("Lock result: %s for %s, with key: %s",
                     SystemLockRep.Result.Name(result), user_id, key)
        return {'result': result, 'user_id': user_id, 'key': key}

    def _activate(self, user_id):
        """ Activates, keeps-alive, or blocks the lock.

        :param user_id: See handle_request. Reported as OCCUPIED if already locked.
        :return:        tuple (result, user_id, key), see handle_request function for details
        """
        with self._thread_lock:
            if not self.is_active():
                # System was unlocked, lock it
                generatedkey = time.ctime() + " " + str(datetime.now().microsecond / 1e6)

                self._close(user_id, generatedkey) # Close the lock with timestamp as key
                logger.debug("System lock is set (user: %s)", self._user_id)
                return SystemLockRep.ALLYOURS, self._user_id, self._key
            else:
                # Lock is already used
                return SystemLockRep.OCCUPIED, self._user_id, "no-key"

    def _release(self, user_id, key):
        """ Releases or blocks the lock

        :param user_id: See handle_request. Reported as OCCUPIED if the user_id doesn't match.
        :param key:     See handle_request. Reported as OCCUPIED if the key doesn't match.
        :return:        RELEASED or OCCUPIED (as specified in protobuf)
        """
        with self._thread_lock:
            if user_id == self._user_id and key == self._key:
                # Only release lock when the given credentials are right
                self._open()

            # Report status of lock
            if not self.is_active():
                return SystemLockRep.RELEASED
            else:
                return SystemLockRep.OCCUPIED

    def _close(self, user_id, key):
        """ Close the lock, thus blocking the PHC.
            Also starts the timeout mechanism

        :param user_id: Is stored for later matching
        :param key:     Is stored for later matching
        """
        with self._thread_lock:
            self._user_id   = user_id
            self._key       = key
            self._update_timeout()

    def _open(self):
        """ Opens the lock, thus unlocking the PHC.
            Also stops the timeout mechanisms
        """
        with self._thread_lock: # we must use a thread-lock here,
                                # otherwise the timeout thread might cause a race condition
                                # (i.e. the timer could go off even though we're disabling it)
            self._user_id              = None
            self._key                  = None
            self._keepalive_lastupdate = 0
            if self._keepalive_timer:
                self._keepalive_timer.cancel()
                if threading.current_thread() is threading.main_thread():
                    # Wait for timeout thread to stop (this function is also used from within
                    #      that thread, so only wait from the main thread)
                    self._keepalive_timer.join()

    def keepalive(self):
        """ Updates the keepalive timer.
            Must be called regularly (via a request) to keep the lock active
        """
        with self._thread_lock:
            self._keepalive_lastupdate = time.time()
            logger.debug("System lock alive timestamp updated (user: %s)", self._user_id)

    def _update_timeout(self):
        """ Timeout mechanism.
            Releases the lock when the timer expires or starts a new timer.
        """
        with self._thread_lock: # we must use a thread-lock here to avoid race conditions
                                # (between main thread and the timeout thread)
            if self.is_active() and self._keepalive_timeout:
                if self._keepalive_lastupdate \
                and time.time() > self._keepalive_lastupdate + self._keepalive_timeout:
                    # The timeout has expired, release the lock
                    logger.warning("System lock timed out after %d s and is released (user: %s)",
                                self._keepalive_timeout, self._user_id)
                    self._open()
                else:
                    # A keep-alive message was received in time, start a new timeout timer
                    self.keepalive()    # Update the keepalive timestamp here as well.
                                        # This prevents a timeout due to clock drifting
                                        # (between the PhcKey and PhcLock)
                    self._keepalive_timer = threading.Timer(
                        self._keepalive_timeout, self._update_timeout)
                    self._keepalive_timer.name = "Phc server lock timeout"
                    self._keepalive_timer.setDaemon(True)
                    self._keepalive_timer.start()



class KeepAliveClient(BaseClient):
    """ Facilitates threaded PHC lock keepalives """

    def __init__(self, keepAliveHost, keepAlivePort, keepAliveInterval, userId, key):
        """ Constructor """

        self.userId = userId
        self.key    = key

        # Create a REQ socket used to connect to the PHC server
        reqSocket = ReqSocket()
        reqSocket.createZmqSocket()
        keepAliveAddress = "tcp://{}:{}".format(keepAliveHost, keepAlivePort)
        reqSocket.connectSocket(keepAliveAddress)

        super().__init__(reqSocket=reqSocket)
        self.packetizer = Packetizer()

        # Keepalive objects 
        self._keepaliveTimer = None
        self._keepAliveInterval = keepAliveInterval

        # Objects for thread-safe stopping of the keepalive timer
        self.stopEvent  = threading.Event()
        self.threadLock = threading.RLock()

    def sendkeepalive(self):
        """ Send a keepalive message to the PHC lock keepalive server """
        with self.threadLock:
            if not self.stopEvent.is_set():
                logger.debug("Sending keepalive")
                reqMsg               = SystemLockKeepaliveReq()
                reqMsg.user_id       = self.userId
                reqMsg.key           = self.key
                self.requestReply(reqMsg, OkRep())
                self.start() # restart next keepalive

    def start(self):
        """ Starts/restart the keepalive timer, which sends a keepalive message after it expires """
        with self.threadLock:
            self._keepaliveTimer = threading.Timer(self._keepAliveInterval,
                                                    self.sendkeepalive)
            self._keepaliveTimer.name = "Phc lock client keepalive timer"
            self._keepaliveTimer.setDaemon(True)
            self._keepaliveTimer.start()

    def stop(self):
        """ Stops the keep alive mechanism """
        if self._keepaliveTimer is not None:
            with self.threadLock:
                self.stopEvent.set()
                self._keepaliveTimer.cancel()
                self._keepaliveTimer.join()
                self._keepaliveTimer = None
            logger.debug("Stopped keepalive client")


class PhcKey():
    """ This class provides interaction from a PhcClient to the PhcServer
        The key must be used in a 'with ...:' context.
        This ensures that the lock is released when the context is left.

        The key also implements a keep-alive mechanism, which updates the lock with an interval.
        The keep-alive mechanism runs in a separate thread, and uses a separate client for sending
        keepalive messages to the PHC server.
        The PhcKey runs in the PhcClient context.
     """
    def __init__(self, keepAliveHost, lockFunction,
                 keepAliveInterval = LOCK_TIMEOUT_DEFAULT/2, label=None):
        """
        Creates a lock, suitable for use in 'with' context

        :param keepAliveHost:      Hostname of the server for sending lock keepalive messages.
        :param lockFunction:       Function via which we can reach the Phc server
                                   for setting/releasing the lock.
        :param keepAliveInterval:  The lock will be updated with this interval. By default the
                                   client will update two times in the server timeout interval.
        :param str label:          Override default label shown in lock log messages.
        """
        self.keepAliveHost = keepAliveHost

        check_class(keepAliveInterval, (float, int))
        assert callable(lockFunction)
        lock_function_signature = inspect.signature(lockFunction)
        for argument in ('lock_type', 'user_id', 'key'):
            assert argument in lock_function_signature.parameters,\
                "Lock function misses argument: {}".format(argument)

        self._keepaliveInterval = keepAliveInterval
        self._lockFunction = lockFunction

        # Create user id
        scriptpath = traceback.extract_stack(limit=-1)[0].filename

        # set scriptname@hostname as user_id
        if label is None:
            self._user_id = ntpath.basename(scriptpath)
        else:
            self._user_id = label
        self._user_id += "@" + socket.gethostname()
        self._key     = ""

        # Set up keep alive bookkeeping
        self.keepAliveClient = None

    def __enter__(self):
        """ Activates the lock when entering the 'with' context """
        self._lock()
        return self

    def __exit__(self, *args):
        """ Releases the lock when leaving the 'with' context"""
        try:
            self._unlock()
        except Exception as ex:
            logger.exception("An error occurred while unlocking: %s", ex)

    def _lock(self):
        """ Instruct the Phc server to close the lock
            Also starts the keep-alive mechanism """
        reply = self._lockFunction(lock_type = SystemLockReq.ACTIVATE,
                                   user_id   = self._user_id,
                                   key       = self._key)

        if reply['result'] == SystemLockRep.ALLYOURS:
            # Locking the PHC was successfull, start a keepalive client thread
            # This will keep the PHC locked, while our caller conducts business (e.g. measuring)
            if 'key' in reply:
                self._key = reply['key']
            else:
                logger.warning("No lock key received after PHC lock request.")
                self._key = ''
            self.keepAliveClient = KeepAliveClient(userId            = self._user_id,
                                                   key               = self._key,
                                                   keepAliveHost     = self.keepAliveHost,
                                                   keepAlivePort     = reply['keep_alive_port'],
                                                   keepAliveInterval = self._keepaliveInterval)
            self.keepAliveClient.start()
        else:
            raise GsswError(
                "Could not acquire system lock (Status: {}, by user: {})".format(
                    SystemLockRep.Result.Name(reply['result']), reply['user_id']))

    def _unlock(self):
        """ Instruct the Phc server to release the lock """

        if self.keepAliveClient is not None:
            self.keepAliveClient.stop()

        reply = self._lockFunction(lock_type= SystemLockReq.RELEASE,
                                   user_id  = self._user_id,
                                   key      = self._key)
        if reply['result'] != SystemLockRep.RELEASED:
            logger.warning("The PHC unlock result was unexpected (Status: %s, user: %s)",
                           SystemLockRep.Result.Name(reply['result']), reply['user_id'])



class LockServer(ThreadedBaseServer):
    """ Threaded RPC server, which is can handle PHC lock keepalive messages.
        The server listens on a separate port for keepalives.
    """

    def __init__(self):
        """ Constructor """
        super().__init__('LockServer')

        repSocket    = RepSocket()
        repSocket.createZmqSocket()

        keepAlivePort    = repSocket.bindSocketToRandomPort("tcp://*")
        self.lockControl = LockControl(keepAlivePort)

        controlService   = DelegatingReqRepService(self.lockControl,
                                                   LockKeepaliveServiceSpec(),
                                                   repSocket)
        self.addService(controlService)


class LockControl():
    """ Lock control worker (PHC server-side) """

    def __init__(self, keepAlivePort):
        self._lock         = PhcLock()
        self.keepAlivePort = keepAlivePort

    def is_active(self):
        """ Reports whether the lock is closed by a user (i.e. whether someone is using the PHC)

        :return:        bool
        """
        return self._lock.is_active()

    def handle_request(self, lock_type, user_id, key):
        """ Handles lock/unlock requests.
            (The requests are forwarded from the regular PHC control worker).

        :param lock_type: ACTIVATE or RELEASE
        :param user_id:   User id
        :param key:       Password, for RELEASE
        """
        result = self._lock.handle_request(lock_type, user_id, key)

        # Send the PHC client the port via which to send the keepalive messages  
        result['keep_alive_port'] = self.keepAlivePort
        return result

    def keepalive(self, user_id, key):
        """ Handles keepalive requests. 

        :param user_id: User id
        :param key:     Password
        """
        logger.debug("KEEPALIVE for user: %s (key: %s)", user_id, key)
        self._lock.keepalive()
