#pylint: disable=redefined-outer-name
""" Helper functions"""

import os
import re
import socket
import sys
import time
import logging
import argparse
import itertools
import threading

from datetime import datetime

from gssw.lib.argpgui import ArgParseGui
import gssw.proto.common_pb2 as common_proto 

logger = logging.getLogger(__name__)

###########################################################################
# C O N S T A N T S
###########################################################################

###########################################################################
# M I S C
###########################################################################

def versionStr(version):
    """ Convert 32-bit version numbers to major.minor.revision string."""
    major    = version >> 16
    minor    = (version >> 8) & 0xff
    revision = version & 0xff
    return '%d.%d.%d' % (major, minor, revision)

def gitVersion(path):
    """ Try to get GIT revision for a specific file or directory.

    :return: String with revision number ('-modified' suffix if path is modified locally),
             and posix timestamp.
             None if path not under git version control.
             e.g.: 'hash: 92088053(...)-modified, utc: 1527869035.0'
    """
    # Place import statement here because git is an optional dependency. That is, not all
    # components will use it and you should be able to run them without installing git.
    try:
        import git
    except ImportError:
        logger.warning("Git not installed, cannot determine version of %r", path)
        return None

    try:
        repo   = git.Repo(path, search_parent_directories=True)

        # Find the most recent commit for the path
        commit = next(repo.iter_commits(paths=path))
        revisionString = '{}'.format(commit.hexsha)

        if repo.is_dirty(path=path):
            # File is modified locally, mark with suffix
            revisionString += '-modified'

        return 'hash: {}, utc: {}'.format(revisionString , commit.committed_datetime.timestamp())
    except Exception:
        # Path not under git version control 
        return None


def stringToIdentifier(s, white_space_becomes='_'):  #pylint: disable=invalid-name
    """ Takes a string and makes it suitable for use as an identifier

        Replaces white space by the white_space_becomes character (default=underscore).
        Removes and punctuation.
    """
    s = re.sub(r"\s+", white_space_becomes, s) # replace whitespace with underscores
    s = re.sub(r"-", "_", s) # replace hyphens with underscores
    s = re.sub(r"[^A-Za-z0-9_]", "", s) # remove everything that's not a character, a digit or a _
    return s


def normRealPath(fileName):
    """ Returns the normalized absolute path of fileName, with symlinks expanded.
    """
    return os.path.normpath(os.path.realpath(fileName))


def pbEnumName(enum, value):
    """ Function for retrieving name of a protobuf enum value (str), e.g. for logging.
        This prevent exceptions in case the integer value is not defined in the enum.
    """
    if value in enum.values():
        return '{} ({})'.format(enum.Name(value), value)
    else:
        return 'Unknown ({})'.format(value)


def same(lst):
    """ If all list elements are the same the element is returned. Otherwise returns an empty str.
    """
    result = None
    for elem in lst:
        if result is None:  # the first element
            result = elem
        else:
            if elem != result:
                return ''
    return result

def enableLargeProtobuffers():
    """ This function increases the maximum size of serialized protobuf messages. The size is
        increased from 64MB to INT_MAX (i.e. 2GB).

        :note: This is a little bit dangerous piece of code. It can easily induce OOM conditions
               when generating larges messages.

        :note: Do not call this function directly. Instead add the following to your config file:

            ::

              global-parameters:
                  enable-large-protobufs: true # Increase message size limit from 64MB to 2GB.

            """
    logger.info("Increasing protobuf message size limit from 64MB to 2GB...")
    from google.protobuf.pyext._message import SetAllowOversizeProtos  #pylint: disable=import-error,no-name-in-module
    SetAllowOversizeProtos(True)


###########################################################################
# L O G G I N G
###########################################################################


def logDict(dictionary, msg=None, logger=None, level='debug', item_prefix='  '):
    """ Writes a log message with key and value for each item in the dictionary.

        :param dictionary: the dictionary to be logged
        :param name: An optional message that is logged before the contents
        :param logger: A logging.Logger object to log to. If not set, the 'main' logger is used.
        :param level: log level. String or int as described in the logging module documentation.
            Default: 'debug'.
        :param item_prefix: String that will be prefixed to each line. Default: two spaces.
    """
    level_nr = logging.getLevelName(level.upper())

    if logger is None:
        logger = logging.getLogger(__name__)

    if msg is None:
        logger.log(level_nr, "logging dictionary...")
    elif msg:
        logger.log(level_nr, str(msg))

    if not dictionary:
        logger.log(level_nr,"{}<empty dictionary>".format(item_prefix))
        return

    max_key_len = max([len(k) for k in dictionary.keys()])

    for key, value in sorted(dictionary.items()):
        logger.log(level_nr, "{0}{1:<{2}s} = {3}".format(item_prefix, key, max_key_len, value))



###########################################################################
# D A T E   A N D   T I M E
###########################################################################

def getDate():
    """get the current data as string"""
    return datetime.utcnow().strftime("%Y%m%d")

def getYear():
    """get the current year as string"""
    return datetime.utcnow().strftime("%Y")

def getTime():
    """get the current time as string"""
    return datetime.utcnow().strftime("%H%M%S")

def getDateTime():
    """get current date en time as string"""
    return datetime.utcnow().strftime("%Y%m%d-%H%M%S")

###########################################################################
# T W O S   C O M P L E M E N T
###########################################################################

def twosComp(val, bits):
    """compute the 2's compliment of int value val"""
    if (val & (1 << (bits - 1))) != 0: # if sign bit is set e.g., 8bit: 128-255
        val = val - (1 << bits)        # compute negative value
    return val                         # return positive value as is


###########################################################################
# ZeroMQ socket addresses
###########################################################################

def zmqTcpBindAddress(tcpport):
    """zmq bind"""
    return 'tcp://*:%u' % tcpport


def zmqTcpConnectAddress(host, tcpport):
    """zmq connect"""
    return 'tcp://%s:%u' % (host, tcpport)


def zmqIpcAddress(name):
    """zmq ipc address"""
    return 'ipc:///tmp/gssw_socket_' + name


def zmqAutoConnectAddress(name, host, tcpport):
    """zmq auto connect address"""
    if host == 'localhost':
        return zmqIpcAddress(name)
    
    return zmqTcpConnectAddress(host, tcpport)


###########################################################################
# A R G U M E N T   P A R S I N G
###########################################################################

#: GSSW configuration file environment variable name.
GSSW_CONFIGURATION_ENVIRONMENT_VARIABLE = 'GSSW_CONFIGURATION_FILE'
#: Default file name searched for GSSW configuration.
GSSW_CONFIGURATION_FILE_PATTERN         = '~/etc/egress/$HOSTNAME/config.yaml'

def createGsswArgumentParser(description, enableOptionName=False, optionNameDefault=None):
    """ Creates an argument parser with default arguments for specifying a configuration
        file/server.
        Additionally, it searches for a configuration file specified in an environment variable,
        which is used as fallback.
        As last resort, a default location is searched, GSSW_CONFIGURATION_FILE_PATTERN.

        You still need to call parser.parse_args(). This is done so that optional arguments
        can be added to the parser.

        :param str  description:       The help text for a script, use __doc__ for the script's own
                                       help text from the top of the file.
        :param bool enableOptionName:  The argument for deamon name is added to the parser.
        :param str  optionNameDefault: When the name option is enabled, use this as default value.
        :note: Example usage: parser = createGsswArgumentParser(__doc__)

        :return: argparse.ArgumentParser
    """
    # Remove Sphinx marker from description string, so that the script output shows only the actual
    # help itself
    if description is not None and '.. program-output::' in description:
        description = '\n'.join(description.splitlines()[:-2]) # remove 2 last lines from docstring

    if '--gui' in sys.argv:
        parser = ArgParseGui()
    else:
        parser = argparse.ArgumentParser(description=description,
                                         formatter_class=argparse.RawDescriptionHelpFormatter)

    # Try discovering the configuration file via the environment variable
    configPath = None
    gssw_configuration_file = os.getenv(GSSW_CONFIGURATION_ENVIRONMENT_VARIABLE)
    if gssw_configuration_file is not None:
        if os.path.isfile(gssw_configuration_file):
            logger.debug("A configuration file was found: %s", gssw_configuration_file)
            configPath = gssw_configuration_file 
        else:
            logger.error(
                "Configuration file '%s' specified in environment variable '%s' was not found",
                gssw_configuration_file, GSSW_CONFIGURATION_ENVIRONMENT_VARIABLE)
    else: # Use fallback pattern: ~/etc/egress/$HOSTNAME/config.yaml
        patternFile = GSSW_CONFIGURATION_FILE_PATTERN.replace('$HOSTNAME', socket.gethostname())
        patternFile = os.path.expanduser(patternFile)

        if os.path.isfile(patternFile):
            configPath = patternFile
            logger.debug("A configuration file was found in the default directory: %s", configPath)

    configPathHelp = """
        The path to the configuration file, or the address of the configuration server.
        A default configuration file can also be set by
        creating an environment variable {} or by using the default location '{}'.""".format(
            GSSW_CONFIGURATION_ENVIRONMENT_VARIABLE, GSSW_CONFIGURATION_FILE_PATTERN)

    if configPath is not None:
        configPathHelp += " Currently, the following file is used: '{}',"\
                          " but it can be overridden via this argument.".format(configPath)
    else:
        configPathHelp += " If the environment variable is set and points to an existing file,"\
                          " or a file is placed in the default location,"\
                          " this argument becomes optional."

    # Use the file from the environment variable as default, but override it with the optional
    # argument.
    group = parser.add_argument_group('Egress configuration')
    group.add_argument('-p', '--config-path', dest='configPath', type=str, default=configPath,
                       help=configPathHelp,
                       required=configPath is None) # Require that a configuration is specified via
    #                                               # arguments if no file was found via the
    #                                               # environment variable

    if enableOptionName:
        group.add_argument('-n', '--name', type=str, required=not bool(optionNameDefault),
                        help='The identifier of this program in the config file' + \
                             (' (default: %(default)s).' if optionNameDefault else '.'),
                        default=optionNameDefault)
    return parser


def validate_port_number(value):
    """ Raises argparse.ArgumentTypeError if the port number is negative.
        Returns the port number if its valid.
    """
    ivalue = int(value)
    if ivalue < 0:
        raise argparse.ArgumentTypeError("port number must be a positive integer. got: {}"
                                         .format(value))
    return ivalue


class RangeCheckedType:
    """ Helper class for argparse module to check the range of integer and float arguments.

    Below are some examples of typical usage:

    ::

      parser = createGsswArgumentParser(__doc__)

      # Add some '--ampl' argument with a range of [-1.0, 1.0]
      parser.add_argument('--ampl', type=RangeCheckedType(float, -1, 1), help='Set amplitude')

      # Add some '--period' argument with a range of [1, 100]
      parser.add_argument('--period', type=RangeCheckedType(int, 1, 100), help='Set period (in µs)')

      # (...)

      args = parser.parse_args() # Now the ranges of the arguments are checked automatically
    """

    def __init__(self, argType, start, end):
        """ Used to define the range of a new script argument.

        :param argType: Type of the value eventually supplied to the script after parsing the
                        argument. Either `float` or `int`.
        :param start:   Lowest allowed value.
        :param end:     Highest allowed value.
        """
        assert argType in [int, float]
        self.argType  = argType
        self.start = start
        self.end   = end

    def __call__(self, value):
        """ Parsing function, called by argparse module """
        value = self.argType(value) # Check if the type is correct will raise an exception otherwise
        if self.start <= value <= self.end:
            return value # Everything is fine, return the casted value

        # Given value is out of range
        raise argparse.ArgumentTypeError(
            "{} must be in the range [{}, {}]".format(value, self.start, self.end))

    def __repr__(self):
        """ Used for printing the type, called by argparse module """
        if self.argType is int:
            return "integer"
        elif self.argType is float:
            return "float"
        return repr(self.argType)


###########################################################################
# F I L E   A N D   F OL D E R   H A N D L I N G
###########################################################################
def ensureDir(dirname):
    """Creates the specified directory if it does not already exist."""
    if dirname:
        os.makedirs(dirname, exist_ok=True)

###########################################################################
# T I M E   E X E C U T I O N
###########################################################################
class DebugTimer :
    """Debug timer"""
    def __init__(self, event):
        self._event = event
        self._startTime = None

    def __enter__(self) :
        """enter function"""
        self._startTime = time.time()

    def __exit__(self, *args) :
        """exit function"""
        print('{event} took {elapsed} seconds'.format(
            event = self._event, elapsed = time.time() - self._startTime))

###########################################################################
# A S C I I   L O A D I N G   I C O N
###########################################################################

class AsciiLoadingIcon(threading.Thread):
    """ Simple class to use as ascii loading icon 

        Example use:

        .. code-block:: python

            print('Doing stuff..')
            with AsciiLoadingIcon(updatetime=0.4):
                time.sleep(10)
                # Do some stuff
            print('Done!')

            Advanced use:

            .. code-block:: python

            li = AsciiLoadingIcon()
            li.characters='LOADING'
            print('Doing stuff')
            li.start()
            # Do stuff ...
            li.stop_and_wait()

    """

    def __init__(self, updatetime=0.1):
        """ Constructor """
        super(AsciiLoadingIcon, self).__init__()
        self._stop_event = threading.Event()
        self.characters = "/-\\|"
        self.updatetime = updatetime

    def __enter__(self):
        """ Start loop """
        self.start()

    def __exit__(self, *_): # *_ ignore all input parameters
        """ Stop and wait loop """
        self.stop_and_wait()

    def stop(self):
        """ Stop """
        self._stop_event.set()

    def stop_and_wait(self):
        """ Stop and wait """
        self.stop()
        self.join()

    def stopped(self):
        """ Stopped """
        return self._stop_event.is_set()

    def run(self):
        """ Thread run function """
        for character in itertools.cycle(self.characters):
            sys.stdout.write(character + '\r')
            sys.stdout.flush()
            if self.stopped():
                break
            time.sleep(self.updatetime)

###########################################################################
# G E N E R I C  T Y P E  C O N V E R S I O N
###########################################################################

def toProtobufEncoding(halType):
    """ converts the type as given in the HAL to the types as used in 
        the protocol puffers 
    """
    if halType == 'uint':
        return common_proto.ENC_UINT
    elif halType== 'int':
        return common_proto.ENC_SINT
    elif halType == 'float':
        return common_proto.ENC_FLOAT
    elif halType == 'string':
        return common_proto.ENC_STRING
    else:
        raise TypeError('Unknown type (%s)' % str(halType))

