""" Base classes for clients and servers that use ZMQ messaging.
"""
import logging
from   datetime                 import datetime, timedelta
from   queue                    import PriorityQueue
import threading
import time
from   google.protobuf.message  import Message as PbMessage

from gssw.proto.common_pb2      import ErrorRep, PeerExceptionRep
from gssw.lib.errors            import CommunicationError, GsswSocketError, GsswTimeoutError, \
                                       PeerException, UnsupportedServiceError
from gssw.lib.sockets           import tagName, SubSocket, ReqSocket, RepSocket, Poller
from gssw.lib.sockets           import SubSocketMixin, RepSocketMixin, PubSocketMixin
from gssw.lib.typing            import check_class

logger = logging.getLogger(__name__)


class AbstractService():
    """ Defines the methods and properties that a service class must implement so that it can be
        added to a Server.
    """
    def __init__(self):
        """ Constructor.
        """
        pass


    def finalize(self):
        """ Can be used to cleanup resources. Must be called explicitly.
            The base implementation does nothing, descendants may override it.
        """
        pass


    @property
    def inputSocket(self):
        """ The socket from where the Service gets its input.
        """
        raise NotImplementedError("Abstract method.")


    @property
    def outputSocket(self):
        """ The socket to where Service sends its output.
        """
        raise NotImplementedError("Abstract method.")


    def receiveAndProcessMessage(self):
        """ Reads a message and processes it.

            May assume that there is a message to receive on on the input socket.
        """
        raise NotImplementedError("Abstract method.")


    def processBinaryMessage(self, tag, binMessage):
        """ Processes a binary message (i.e. a message that is not yet parsed from a byte string)

            Can use the tag to determine the message type. Should then parse it, process it and
            return the reply (plus reply tag)

            :param tag: string with the name of the message
            :binMessage: bytes containing an encoded protocol buffer message
            :returns: (replyTag, binReplyMessage) tuple
        """
        raise NotImplementedError("Abstract class, please override this method")


# Pylint doesn't recognize AbstractReqRepService as an abstract class, Probably because it doesn't
# contain any NotImplementedError directly. https://stackoverflow.com/a/22224042/
# pylint: disable=abstract-method


class AbstractReqRepService(AbstractService):
    """ A service that implements a request/reply pattern.

        It has a single REP socket that it uses for both input and output.

        Is an abstract class, the descendant needs to implement processBinaryMessage.
    """

    def __init__(self, repSocket):
        """ Constructor.
        """
        super().__init__()
        check_class(repSocket, RepSocketMixin)
        self._repSocket = repSocket


    @property
    def inputSocket(self):
        """ The socket from where the Service gets its input. This is the REP socket.
        """
        return self._repSocket


    @property
    def outputSocket(self):
        """ The socket to where Service send its output. This is the REP socket.
        """
        return self._repSocket


    def receiveAndProcessMessage(self):
        """ Receives a message and processes it (assumes that there is a message ready to receive).

            Receives the message on the REP socket, calls processBinaryMessage to process it, and
            sends the reply message back on the REP socket. If processBinaryMessage raises an
            exception, the exception is put in a PeerExceptionRep message and send back as answer.
            This way the service always sends a reply for every request.

            The processBinaryMessage is not implemented by the AbstractReqRepService class, it
            should be implemented by a descendant.
        """
        requestTag, binRequestMessage = self._repSocket.receiveProtoBuf()
        try:
            replyTag, binReplyMessage = self.processBinaryMessage(requestTag, binRequestMessage)

        except Exception as ex:
            logger.error("Sending a PeerExceptionRep cause of the exception below")
            logger.exception(ex)
            replyTag, binReplyMessage = PeerException.exceptionToMessage(ex)

        # Always send a reply.
        self._repSocket.sendProtoBuf(replyTag, binReplyMessage)



class AbstractPubSubService(AbstractService):
    """ A service that implements a Pub-Sub pattern.

        Has a SUB socket and an optional PUB socket. Listens on the SUB socket for an input message,
        calls receiveAndProcessMessage to read and process it, and optionally publishes the result
        of receiveAndProcessMessage on the pubSocket. If processBinaryMessage raises an exception,
        the exception is put in a PeerExceptionRep message and published if a PUB socket is present.

        This class can act as data sink or as a pipe in a messaging network. If it is a data sink,
        the service doesn't publish data and the PUB socket is optional. If it is a pipe, the
        PUB socket is mandatory.

        Is an abstract class, the descendant needs to implement processBinaryMessage.

        :param subSocket: a SubSocket (or QtSubSocket)
        :param pubSocket: a pubSocket (or QtPubSocket). Can also be None. In that case the result
            of processBinaryMessage is NOT published.
    """

    def __init__(self, subSocket, pubSocket):
        """ Constructor.
        """
        super().__init__()
        check_class(subSocket, SubSocketMixin)
        check_class(pubSocket, PubSocketMixin, allow_none=True)
        self._subSocket = subSocket
        self._pubSocket = pubSocket

        self.topic = '' # Always publish without topic (key)


    @property
    def inputSocket(self):
        """ The socket from where the Service gets its input. This is the SUB socket.
        """
        return self._subSocket


    @property
    def outputSocket(self):
        """ The socket to where Service send its output. This is the PUB socket.
        """
        return self._pubSocket


    def receiveAndProcessMessage(self):
        """ Receives a message and processes it (assumes that there is a message ready to receive)

            Receives the message on the SUB socket, calls processBinaryMessage to process it, and
            publishes the result on the PUB socket (if this socket is defined).

            If processBinaryMessage raises an exception, the exception is put in a PeerExceptionRep
            and published as well.

            The processBinaryMessage is not implemented by the AbstractReqRepService class, it
            should be implemented by a descendant.
        """
        _topic, tag, binMessage = self._subSocket.receiveProtoBufWithTopic()
        try:
            replyTag, binReplyMessage = self.processBinaryMessage(tag, binMessage)
        except UnsupportedServiceError as ex:
            logger.info("Ignoring message with tag %r: %s", tag, ex)
            return
        except Exception as ex:
            logger.error("Unable to process {!r} message because of the following"
                         .format(tag))
            logger.exception(ex)
            replyTag, binReplyMessage = PeerException.exceptionToMessage(ex)

        # If a pub socket is defined, the reply is published.
        if self._pubSocket:
            self._pubSocket.publishProtoBuf(self.topic, replyTag, binReplyMessage)
            logger.debug("Published message with tag: {}".format(replyTag))
        else:
            logger.debug("Server has no PUB socket. Discarded message with tag: {}"
                         .format(replyTag))


class BaseServer:
    """ Base class for all server classes.

        A server has a collection of services. Every service has a unique input socket. The
        server will poll all input sockets for messages. If it detects a message on an input socket
        it will call the receiveAndProcessMessage method of the corresponding service to read and
        process the message.

        # TODO: rewrite
        # The AbstractServer can also periodically execute messages (with the repeatEvery() method).
        # The reply of the message will be published on the PubSocket.
    """
    def __init__(self):
        """ Constructor.

            :param sockets: an optional list of sockets can also be added later with addSocket()
            :param debugging: if True the server quits when an exception occurs
        """
        self._poller = Poller()
        self._periods = {} # (tagName, binMessage) -> seconds dictionary to store repeat schedule
        self._priorityQueue = PriorityQueue() # ordered list of (timestamp, key) tuples

        self._services = {} # Socket -> ServiceInterface dictionary

        # Used to stop the main loop, when using a threaded BaseServer
        self._stopEvent = threading.Event()


    def finalize(self):
        """ Can be used to cleanup resources. Current implementation does nothing.
            Must be called explicitly.
        """


    def addService(self, service):
        """ Adds a service to the server. Registers the input socket of the service for poll-in.

            :param service: a descendant of ServiceInterface
        """
        check_class(service, AbstractService)

        # The service's inputSocket is the key in the _services dictionary.
        # Check that it doesn't occur yet.
        socket = service.inputSocket
        check_class(socket, (RepSocket, SubSocket))

        if socket in self._services:
            raise KeyError("Input socket {} already used by another service: {}"
                           .format(socket, self._services[socket]))

        self._poller.registerPollIn(socket)
        self._services[socket] = service


    # def getServiceByInputMessage(self, inputMsg):
    #     """ Returns the service that can process this message.
    #         Raises KeyError if no service found.
    #     """
    #
    #     # for service in self._services.values():
    #     #     if s


    def repeatEvery(self, msg, seconds):
        """ Executes a protocol buffer message periodically. The reply is published on the
            PubSocket.

            The message will be added to a message->period dictionary, so if repeatEvery is called
            twice with the same msg, the second call overrides the first.

            If seconds is None the message will be removed from the dict of periodically executed
            messages. (Note: the currently scheduled call will still be executed!)

            For debugging purposes the msg can also be callable (i.e. a function of method). In
            that case the result of the function/method is only logged.

            :param seconds: repetition period in seconds. Must be > 0, can be a float.
        """
        if isinstance(msg, PbMessage):
            binMessage = msg.SerializeToString()
            tag = tagName(msg)
            key = tag, binMessage
        elif callable(msg):
            key = msg
        else:
            raise TypeError("msg must be a google.protobuf.message.Message or must be callable")

        if seconds is None:
            del self._periods[key]
        else:
            self._periods[key] = seconds
            self._scheduleNextCall(key)


    def _scheduleNextCall(self, key):
        """ Schedule an execution of message in the number of seconds as specified by repeatEvery
        """
        timestamp = datetime.now() + timedelta(seconds=self._periods[key])
        logger.debug("Scheduling next call at: {}".format(timestamp))
        self._priorityQueue.put_nowait((timestamp, key))


    def _execScheduledMessage(self, key):
        """ Processes the message and schedules the next call.

            :param key: dictionary key from the ._periods dictionary. This will be a
                (tag, binary-message) tuple in case a message was scheduled (the regular case), or
                just a callable in case a regular function or method was scheduled (typically only
                for debugging). If a message was scheduled, the reply message is published on the
                PubSocket. If a callable was scheduled the result is logged..
        """
        self._scheduleNextCall(key)

        if isinstance(key, tuple):

            tag, binMessage = key
            logger.debug("Auto executing message of tag: {}".format(tag))
            try:
                replyTag, binReplyMessage = self.processBinaryMessage(tag, binMessage)  #pylint: disable=no-member
            except Exception:
                raise # TODO: what to do here? Only warn for consistency?

            self._pubSocket.publishProtoBuf('', replyTag, binReplyMessage)  #pylint: disable=no-member

        elif callable(key):
            # Primarily for testing and debugging
            logger.debug("Executing periodical function: {}".format(key))
            result = key()  # key is a function or method
            logger.debug("Successfully executed: {}. Result: {}".format(key, result))
        else:
            raise TypeError("msg must be a google.protobuf.message.Message or must be callable")


    def start(self):
        """ Starts the indefinite poll-in loop that polls the registered sockets for messages.
        """
        logger.info("Starting polling loop")
        try:
            while not self._stopEvent.is_set():

                # Listen indefinitely if no function literals are registered for periodic execution.
                if self._priorityQueue.qsize() == 0:
                    timeoutMs = None
                    timestamp, msg = None, None
                else:
                    timestamp, msg = self._priorityQueue.get_nowait()
                    timeoutMs = int(1000 * (timestamp - datetime.now()).total_seconds())

                    # The scheduled message must be called immediately
                    if timeoutMs <= 0:
                        self._execScheduledMessage(msg)
                        continue

                # Block until a socket has data or the timeout occurs because a msg was scheduled.
                logger.debug("Polling sockets with timeout {} ms".format(timeoutMs))
                pollingEvents = self._poller.poll(timeout=timeoutMs)

                if pollingEvents == 0:
                    # A timeout has occurred.
                    self._execScheduledMessage(msg)

                else:
                    # A socket has input. Place the message back in the queue
                    if timestamp is not None:
                        self._priorityQueue.put_nowait((timestamp, msg))

                    # Process the sockets' input.
                    for socket, _ in pollingEvents:
                        service = self._services[socket]
                        service.receiveAndProcessMessage()
        except GsswSocketError as ex:
            # Ignore the exception in case of a stop event. Because the stop is enforced by closing
            # the sockets, a socket exception is expected 
            if not self._stopEvent.is_set():
                raise
            logger.info('Socket exception (%s) ignored, a stop event was received...', ex)


    def stop(self):
        """ Stops the RPC server's main poll-in loop. """
        logger.info("Stopping polling loop")
        self._stopEvent.set()
        for socket in self._services:
            # Closing the socket interrupts the socket poller in the 'start' method
            # otherwise, the self._poller.poll method will not return
            socket.close()

class ThreadedBaseServer(BaseServer):
    """ Base class for all server classes, but running inside a thread instead of the main thread.

    :see: Also see the documentation of :class:`BaseServer`.
    """
    def __init__(self, threadName=None):
        """ Create a new server object, for starting an Rpc server in a thread.

        :param str threadName: Name given to the CPU thread, usefull during logging, optional.
        """
        super().__init__()
        self._serverThread = None
        self._threadName   = threadName if threadName else "ThreadedBaseServer"

    def start(self):
        """ Starts a threaded server. """
        self._serverThread = threading.Thread(target=super().start, name=self._threadName)
        self._serverThread.start()

    def stop(self):
        """ Stops the server thread.

        :note: Function blocks until thread is finished.
        """
        super().stop()
        self._serverThread.join()
        self._serverThread = None

class BaseClient():
    """ Base class for client component that make request/reply queries to servers.
    """
    def __init__(self, address=None, reqSocket=None):
        """ Constructor.

            The address and reqSocket parameters cannot be used both!

            :param address: if set, a ReqSocket is created and connected to this address.
            :param reqSocket: sets the request socket if given. If None, a socket will be created.
        """
        if address is not None:
            assert reqSocket is None, "address and reqSocket can not both be set."
            reqSocket = ReqSocket()
            reqSocket.createZmqSocket()
            reqSocket.connectSocket(address)

        self._reqSocket = None
        if reqSocket:
            self.setReqSocket(reqSocket)

    def resetSocket(self):
        """ Reset socket so that the client is still usable after a timeout """
        # Create a REQ socket used to connect to the RPC server
        reqSocket = ReqSocket()
        reqSocket.createZmqSocket()

        assert self._reqSocket is not None, "Cannot auto-reset socket, no old socket was available"
        # We need the information from the old socket, to reconnect
        oldTimeout  = self._reqSocket.getReceiveTimeout()
        connectedTo = self._reqSocket.connectedTo

        # Connect new socket to same server(s) and with same timeout
        reqSocket.setReceiveTimeout(oldTimeout)
        for address in connectedTo:
            reqSocket.connectSocket(address)

        self.setReqSocket(reqSocket)
        logger.debug("A new socket has been created")

    def useTemporaryTimeout(self, temporaryTimeoutMs):
        """ Temporarily uses a different timeout, e.g. longer than usual for set of following 
        requests. This function requires that you use it in a context.

        :param int temporaryTimeoutMs: New timeout [ms].

        :return: Context manager object, which sets and restores the timeout automatically.

        Use as follows:

        ::

            with rpcClient.useTemporaryTimeout(10000):
                # We now have a long timeout, usefull in case RPC server might not be ready
                rpcClient.someFunction()

            # Now, the old timeout is restored, and will be used for remaining functions.
            rpcClient.otherFunction()
        """
        return SocketTimeoutChanger(self, temporaryTimeoutMs)


    def finalize(self):
        """ Can be used to cleanup resources. Current implementation does nothing.
            Must be called explicitly.
        """


    def getReqSocket(self):
        """ The ReqSocket. Can be None if not defined.

            :return: socket.ReqSocket object
        """
        return self._reqSocket


    def setReqSocket(self, reqSocket):
        """ Sets the ReqSocket.

            :param: must be a socket.ReqSocket object
        """
        check_class(reqSocket, ReqSocket)
        self._reqSocket = reqSocket


    def requestReply(self, requestMsg, replyMsg):
        """ Sends a protocol buffer request message via the ZeroMQ req/rep socket pair.

            Takes an empty protocol buffer as the second parameter and populates it with the reply.

            If the reply message contains a PeerExceptionRep, this means that the server encountered
            an exception while calculating the reply. In this case this function raises a
            PeerException.

            :param requestMsg: Protocol buffer message that defines the request
            :param replyMsg: Empty protocol buffer that will be filled with the reply.
            :return: The replyMsg after it has been populated.

            :note: In case of communication failure, the socket is reset, and an exception is raised
        """
        check_class(requestMsg, PbMessage)
        check_class(replyMsg, PbMessage)
        assert self._reqSocket, "reqSocket is undefined"

        logger.debug("Sending request: {}".format(tagName(requestMsg)))
        self._reqSocket.sendProtoBuf(tagName(requestMsg), requestMsg.SerializeToString())

        logger.debug("Awaiting reply...")
        try:
            replyTag, binReplyMessage = self._reqSocket.receiveProtoBuf()
        except (KeyboardInterrupt, GsswSocketError, GsswTimeoutError) as ex:
            if isinstance(ex, KeyboardInterrupt):
                time.sleep(1)
            self.resetSocket()
            raise

        # Contains a PeerExceptionRep protocol buffer.
        if replyTag == tagName(PeerExceptionRep):
            raise PeerException(binReplyMessage)
        elif replyTag == tagName(ErrorRep):
            reply = ErrorRep()
            reply.ParseFromString(binReplyMessage)
            CommunicationError.raiseFromProtoBuffer(reply)
        elif replyTag != tagName(replyMsg):
            raise TypeError("Expected reply tag {}, got {}".format(tagName(replyMsg), replyTag))

        logger.debug("Received reply: {}".format(replyTag))
        replyMsg.ParseFromString(binReplyMessage)

        return replyMsg


class SocketTimeoutChanger:
    """ Context manager class for temporarily using a different timeout.

    When the context is left, the old timeout is restored.

    :note: Do not use this class directly, instead use the :meth:`.useTemporaryTimeout` function.
    """
    def __init__(self, rpcClient, temporaryTimeoutMs):
        """ Creates the context manager. """
        self.rpcClient = rpcClient
        self.temporaryTimeoutMs = temporaryTimeoutMs
        self.oldTimeoutMs = None

    def __enter__(self):
        """ Start using new timeout when entering context """
        clientSocket = self.rpcClient.getReqSocket()
        self.oldTimeoutMs = clientSocket.getReceiveTimeout()
        clientSocket.setReceiveTimeout(self.temporaryTimeoutMs)

    def __exit__(self, *args):
        """ Restore old timeout when leaving context """
        self.rpcClient.getReqSocket().setReceiveTimeout(self.oldTimeoutMs)
