""" Service classes that send Google Protocol buffers over ZMQ messages
    This module handles the protobuf functionality, the ZMQ stuff is mainly implemented in
    rpcbase.py.

    Typically you need to use the DelegatingReqRepService and DelegatingPubSubService classes when
    you create a program that function as a server (i.e. add these classes to a BaseServer)

    If you make a Qt GUI you'll want to use the DelegatingQtReqRepService or
    DelegatingQtPubSubService classes. You can add them directly into your program.
"""

import inspect, logging

from google.protobuf.message import Message as PbMessage
from google.protobuf.service import Service as PbService
from google.protobuf.descriptor import FieldDescriptor as FD
from google.protobuf.descriptor import Descriptor

from gssw.lib.arrays.semantic import SemanticArray, messageToArray, setArrayInMessage
from gssw.lib.arrays.gridded import GriddedArray
from gssw.lib.sockets import tagName
from gssw.lib.socketsqt import NotifyingRepSocket, NotifyingSubSocket
from gssw.lib.rpcbase import AbstractReqRepService, AbstractPubSubService
from gssw.lib.typing import check_class, check_is_a_dict, check_is_a_string, check_is_binary
from gssw.lib.typing import is_a_dict, check_is_a_sequence, get_symbol_from_module
from gssw.lib.errors import ServiceConfigError, ServiceProcessingError, UnsupportedServiceError

from gssw.proto.array_pb2 import NumpyArrayMsg, SemanticArrayMsg, GriddedArrayMsg

logger = logging.getLogger(__name__)

#############################################
# Protocol buffer message types and labels. #
#############################################


# pylint: disable=line-too-long
# For the list of scalar types and their mapping to Python see:
# https://developers.google.com/protocol-buffers/docs/proto#scalar
# https://developers.google.com/protocol-buffers/docs/reference/python/google.protobuf.descriptor.FieldDescriptor-class

TYPE_NAMES = {
    1: 'TYPE_DOUBLE',
  	2: 'TYPE_FLOAT',
  	3: 'TYPE_INT64',
  	4: 'TYPE_UINT64',
  	5: 'TYPE_INT32',
  	6: 'TYPE_FIXED64',
  	7: 'TYPE_FIXED32',
  	8: 'TYPE_BOOL',
  	9: 'TYPE_STRING',
  	10: 'TYPE_GROUP',
  	11: 'TYPE_MESSAGE',
  	12: 'TYPE_BYTES',
  	13: 'TYPE_UINT32',
  	14: 'TYPE_ENUM',
  	15: 'TYPE_SFIXED32',
  	16: 'TYPE_SFIXED64',
  	17: 'TYPE_SINT32',
  	18: 'TYPE_SINT64'
}
# pylint: enable=line-too-long

# Sanity check to see if we enumerate all types here.
assert len(TYPE_NAMES) == FD.MAX_TYPE, "Not all type constants are enumerated"

LABEL_NAMES = {
    1: 'LABEL_OPTIONAL',
  	2: 'LABEL_REQUIRED',
  	3: 'LABEL_REPEATED',
}
# Sanity check to see if we enumerate all labels here.
assert len(LABEL_NAMES) == FD.MAX_LABEL, "Not all label constants are enumerated"

SCALAR_TYPES = (FD.TYPE_DOUBLE, FD.TYPE_FLOAT, FD.TYPE_INT64, FD.TYPE_UINT64, FD.TYPE_INT32,
                FD.TYPE_FIXED64, FD.TYPE_FIXED32, FD.TYPE_BOOL, FD.TYPE_STRING, FD.TYPE_BYTES,
                FD.TYPE_UINT32, FD.TYPE_ENUM, FD.TYPE_SFIXED32, FD.TYPE_SFIXED64, FD.TYPE_SINT32,
                FD.TYPE_SINT64)


#####################
# Auxiliary Classes #
#####################


class Packetizer():
    """ Class for automatically packing an object into a protocol buffer message, and vice versa.
    """
    def __init__(self, convertTo=None, addTo=None):
        """ Constructor.

            To unpack a message use the getMessageAttrs method. It returns a dictionary with the
            keys having the same name as the message field names.

            To add object data to a message use the setMessageAttrs method. It takes a protocol
            buffer message and an object, and fills in the message fields from the object.
            If the object is a dictionary, it is assumed that the message fields names match the
            dictionary keys. The object can also be a scalar or a tuple. See the setMessageAttrs
            doc-string for more details.

            The objects and messages can be nested.

            If you want a certain (sub)messages to be processed differently, for instance because
            its field names do not match the object's attribute names, you can specify this with
            the convertTo and addTo dictionaries.

            The convertTo parameter can be a dictionary that assigns a constructor or factory
            function to certain message types . The getMessageAttrs uses this constructor to create
            an object from a (sub)message, whenever it encounters a (sub)message of this type.

            For example, set convertTo = {GriddedArrayMsg: GriddedArray.createFromProtoBuf}
            to use the GriddedArray.createFromProtoBuf method for converting sub messages of type
            GriddedArrayMsg to GriddedArray objects. Note that GriddedArray.createFromProtoBuf
            must be a function that takes a GriddedArrayMsg message as input and returns a
            GriddedArray.

            The addTo dictionary is for converting in the opposite direction. It specifies for
            message types a function that adds an object to a message of that message type. For
            instance set addTo = {GriddedArrayMsg: GriddedArray.addArrayToProtoBuf} to use the
            GriddedArray.addArrayToProtoBuf for adding GriddedArray objects to a protocol buffer
            message of type GriddedArrayMsg.

            :param convertTo: dict(MessageType = <function(Message)-> object)>
            :param addTo: dict(MessageType = <function(Message, object)>
        """
        check_is_a_dict(convertTo, allow_none=True)
        self._convertTo = {} if convertTo is None else convertTo
        self._addTo = {} if addTo is None else addTo


    @property
    def convertTo(self):
        """ Returns the dictionary that defines converting (sub)messages of a certain type to
            objects. See constructor doc string for details.
        """
        return self._convertTo


    @convertTo.setter
    def convertTo(self, value):
        """ Sets the dictionary that defines converting (sub)messages of a certain type to
            objects. See constructor doc string for details.
        """
        self._convertTo = value


    @property
    def addTo(self):
        """ Returns the dictionary that defines converting objects to (sub)messages of a certain
            type. See constructor doc string for details.
        """
        return self._addTo


    @addTo.setter
    def addTo(self, value):
        """ Sets the dictionary that defines converting objects to (sub)messages of a certain
            type. See constructor doc string for details.
        """
        self._addTo = value


    def getMessageAttrs(self, msg):
        """ Creates a {name: value} dictionary from the message fields.

            This dictionary can then be used as ``**kwargs`` in a worker method (assuming that the
            worker method has keyword parameters with matching names).

            :param msg: google.protobuf.message.Message descendant
            :return: attr dictionary
        """
        check_class(msg, PbMessage)
        logger.debug("Unpacking message of type: {}".format(msg.DESCRIPTOR.full_name))

        # Use the conversion function for message types present in the convertTo dict, and return.
        if type(msg) in self.convertTo:  #pylint: disable=unidiomatic-typecheck
            return self.convertTo[type(msg)](msg)

        result = {}
        for field in msg.DESCRIPTOR.fields:
            value = getattr(msg, field.name)

            # logger.debug("Field {}: type={}, label={}"
            #              .format(field.name, TYPE_NAMES[field.type], LABEL_NAMES[field.label]))

            if field.label != FD.LABEL_REPEATED:

                # Singular fields. Only add them to the dict if they are defined in the message
                if msg.HasField(field.name):

                    if field.type in SCALAR_TYPES:
                        result[field.name] = value

                    elif field.type == FD.TYPE_MESSAGE:
                        # Recursively process the nested message
                        result[field.name] = self.getMessageAttrs(value)
                    else:
                        raise AssertionError("Type not (yet?) implemented: {}".format(field.type))
            else:
                # Repeated field
                if field.type in SCALAR_TYPES:
                    result[field.name] = list(value)

                elif field.type == FD.TYPE_MESSAGE:
                    # Recursively process the nested message
                    result[field.name] = [self.getMessageAttrs(elem) for elem in value]

                else:
                    raise AssertionError("Type not (yet?) implemented: {}".format(field.type))


        return result
        # if convertTo is None:
        #     return result
        # else:
        #     return convertTo(**result)


    # This just is a complicated function. Feels wrong to split it up.
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    def setMessageAttrs(self, msg, obj):
        """ Sets the message attributes given a python object (obj).

            The most common use case is for obj to be a dictionary. (In all other cases a dict is
            created internally).

            * If obj is a dict, the keys must correspond to the field names of msg. This dictionary
              is used to populate the keys.
            * If obj is None the msg must have zero fields
            * If obj is a tuple, msg will be populated by field order. It is checked that msg has
              the same number of fields as the length of the tuple
            * In all other cases obj is considered a scalar and msg must have only one field, which
              is populated from the scalar.

            For convenience, if a value in the dictionary is None, setMessageAttr treats it
            equivalent to a dictionary without that key. This allows us to represent Nones in
            protocol buffers. Not that getMessageAttribute will NOT map missing optional fields to
            Nones, but will just omit that field from the dictionary.

            If the message contains sub messages, this function is called recursively. The obj must
            then also be nested (e.g. be a dictionary that contains another dictionary)

            :param msg: google.protobuf.message.Message descendant
            :param obj: a python object.
        """
        check_class(msg, PbMessage)
        logger.debug("Packing message of type: %s", msg.DESCRIPTOR.full_name)

        fields = msg.DESCRIPTOR.fields

        # If the message type is in the addToFn dict, use the conversion function and return
        if type(msg) in self.addTo:  #pylint: disable=unidiomatic-typecheck
            addToFn = self.addTo[type(msg)]
            logger.debug("Calling {} with a {!r} message and the {!r} object as parameters"
                         .format(addToFn, type(msg), obj))
            addToFn(msg, obj)
            return

        # Create a dictionary that maps field-name to object attributes
        if obj is None:
            # None will be mapped to an empty message
            if fields:
                raise ValueError("If obj is None, the msg should have 0 fields, got: {}"
                                 .format(len(fields)))
            dct = {}

        elif isinstance(obj, tuple):
            # If obj is a tuple the fields are populated by order of occurrence
            if len(fields) != len(obj):
                raise ValueError("If obj is a tuple of len {}, the msg should have the same "
                                 "number of fields, got: {}"
                                 .format(len(obj), len(fields)))
            dct = {}
            for idx, item in enumerate(obj):
                dct[msg.DESCRIPTOR.fields[idx].name] = item

        elif is_a_dict(obj):
            dct = obj

        else:
            if len(fields) != 1:
                raise ValueError("If obj is a scalar, the msg should have one field, got: {}"
                                 .format(len(fields)))
            dct = {msg.DESCRIPTOR.fields[0].name: obj}

        # Use this dictionary to populate the message (recursively if needed)
        for name, value in dct.items():
            field = msg.DESCRIPTOR.fields_by_name[name]

            # Leave values equal to None out of the message.
            if value is None:
                continue

            if field.label != FD.LABEL_REPEATED:
                # Singular fields (i.e. optional/required)

                if field.type in SCALAR_TYPES:
                    try:
                        setattr(msg, name, value)
                    except TypeError as e:
                        raise TypeError(
                            "Could not set '{}' field to '{}'".format(name, value)) from e
                    # Checking whether all required fields are present is done by protobuf itself

                elif field.type == FD.TYPE_MESSAGE:

                    # A sub message must be filled from a dictionary
                    check_is_a_dict(value)

                    # Recursively process the nested message
                    subMsg = getattr(msg, name)
                    self.setMessageAttrs(subMsg, value)

                else:
                    raise AssertionError("Not (yet?) implemented")
            else:
                # Repeated fields
                try:
                    check_is_a_sequence(value)
                except TypeError as e:
                    # Add the name of the current variable to the exception
                    raise TypeError("Incorrect type for argument '{}': {}".format(name, e)) from e

                sequence = value

                if field.type in SCALAR_TYPES:
                    # You can't say msg.name = value when value is a list, so we call
                    # msg.name.extend(value)
                    subMsg = getattr(msg, name)
                    try:
                        subMsg.extend(sequence)
                    except TypeError as e:
                        raise TypeError(
                            "Could not set '{}' field to '{}'".format(name, value)) from e

                elif field.type == FD.TYPE_MESSAGE:
                    # A sub message must be filled recursively from a list of dictionaries
                    subMsg = getattr(msg, name)
                    for idx, elem in enumerate(sequence):
                        try:
                            check_is_a_dict(elem)
                        except TypeError as ex:
                            # Add the name of the current variable and element to the exception
                            raise TypeError(
                                "Incorrect type of element '{}' in argument '{}': {}".format(
                                    elem, name, ex)) from ex
                        subMsg.add()
                        self.setMessageAttrs(subMsg[idx], elem)

                else:
                    raise AssertionError("Not (yet?) implemented")

    # pylint: enable=too-many-locals
    # pylint: enable=too-many-statements

    @classmethod
    def createDefaultPacketizer(cls):
        """ Creates packetizer with defaults for the GSSw. This is a packetizer that can
            (un)pack numpy arrays, SemanticArrays and GriddedArrays.
        """
        convertTo = {NumpyArrayMsg: messageToArray,
                     SemanticArray: SemanticArray.createFromProtoBuf,
                     GriddedArrayMsg: GriddedArray.createFromProtoBuf}

        addTo     = {NumpyArrayMsg: setArrayInMessage,
                     SemanticArrayMsg: SemanticArray.addToProtoBuf,
                     GriddedArrayMsg: GriddedArray.addArrayToProtoBuf}

        return Packetizer(convertTo=convertTo, addTo=addTo)



class ServiceSpecWrapper():
    """ Gives access to a Google protocol buffer service (specification).

        Wraps around a google.protobuf.service.Service object, which is typically defined by
        specifying a service in a .proto file, and can be imported from the generated _pb2.py
        module.

        The wrapper construct a methodInfo dictionary from the specification. The methodInfo
        dictionary maps the tag name of the input message to a tuple with (method-name,
        inputMessageType, outputMessageType) information. This can then be used to find and call
        the corresponding method of a worker class.

        See https://developers.google.com/protocol-buffers/docs/proto#services on how to define
        a service in a .proto file.

        See https://developers.google.com/protocol-buffers/docs/reference/python/ for more info
        about the classes that describe a protocol buffer RPC service.

        This class is typically not used directly but is used indirectly by a Delegator. It
        is implemented as a separate class so that it can be reused in other classes (e.g. a
        hypothetical DelegatingClient implementation).
    """
    def __init__(self, pbService):
        """ Constructor

            :param pbService: a google.protobuf.service.Service that describes the service.
        """
        check_class(pbService, PbService)

        self.pbService = pbService # protocol buffer service (specification)
        self._methodInfo = {}

        self._populateMethodInfo()


    @property
    def methodInfo(self):
        """ Returns method info dictionary.

            This dictionary maps an input_message_type to a (method_name, output_message_type)
            tuple.
        """
        return self._methodInfo


    def _concreteClass(self, msgDescriptor):
        """ Returns the class that this message descriptor describes.

            This should be equivalent to msgDescriptor._concrete_class, but unfortunately that
            method is private so we better not use it.

            :param msgDescriptor: a google.protobuf.descriptor.Descriptor
            :return: a class
        """
        check_class(msgDescriptor, Descriptor)
        moduleName, className = msgDescriptor.full_name.rsplit('.', 1)

        # Sanity check
        assert className == msgDescriptor.name, \
            "last part of full_name != name ({} != {})".format(className, msgDescriptor.name)

        result = get_symbol_from_module(moduleName + '_pb2', className)

        # Disable the pylint warning in the following sanity check
        # pylint: disable=protected-access
        if result != msgDescriptor._concrete_class:
            raise AssertionError("Unexpected concreteClass result: {} != {}"
                                 .format(result, msgDescriptor._concrete_class))
        # pylint: enable=protected-access

        return result


    def _populateMethodInfo(self):
        """ Populates the dictionary that specifies which method, input and output given the tag
            of the input msg.

            Raises an ServiceConfigError if the service contains two methods having the same input
            message type.

            :returns: dictionary {tag : (methodName, inputType, outputType)}
                that is, a { str: (str, type, type) } dictionary.
        """
        servDescr = self.pbService.GetDescriptor()
        for methDescr in servDescr.methods:

            tag = methDescr.input_type.name
            if tag in self._methodInfo:
                raise ServiceConfigError("Service {} contains duplicate input message {}"
                                         .format(servDescr.name, tag))

            #inputMessageType = methDescr.input_type._concrete_class
            #outputMessageType = methDescr.output_type._concrete_class
            inputMessageType = self._concreteClass(methDescr.input_type)
            outputMessageType = self._concreteClass(methDescr.output_type)

            self._methodInfo[tag] = (methDescr.name, inputMessageType, outputMessageType)


    def checkWorkerMethodsPresent(self, worker):
        """ Checks that all methods in the service are present as an attribute of the worker object.

            Does nothing is this is the case. Raises an AttributeError if it is not.
        """
        logger.debug("Checking that the worker implements all service methods.")
        for methodName, argumentSpec, _ in self.methodInfo.values():
            logger.debug("  Checking {} has attribute: {}".format(worker, methodName))

            if not hasattr(worker, methodName):
                raise AttributeError(
                    "Worker {} does not have a method called: {}. "
                    "Perhaps the protocol buffers need to be updated or (re)compiled."
                        .format(worker, methodName))
            else:
                expectedArguments = [f.name for f in argumentSpec.DESCRIPTOR.fields]
                self.checkMethodParametersPresent(worker, getattr(worker, methodName),
                                                                expectedArguments)


    @staticmethod
    def checkMethodParametersPresent(worker, method, expectedArguments):
        """ Checks that a worker method accepts the arguments specified.

            Does nothing if this is the case. Raises an AttributeError if it is not.

            :param worker: Worker object being checked
            :param method: Callable worker function
            :param list expectedArguments: Names of the arguments that the function must have
        """

        methodArguments = inspect.signature(method).parameters

        for argument in expectedArguments:
            logger.debug("  Checking {} method has argument: '{}'".format(
                method.__name__, argument))
            if argument not in methodArguments:
                raise AttributeError(
                    "Worker {} method {!r} does not have an argument called: '{}'. "
                    "Perhaps the protocol buffers need to be updated or (re)compiled."
                        .format(type(worker), method.__name__, argument))


    # I'd like to write InputType and OutputType starting with a capital, since they ar types.
    # pylint: disable=invalid-name

    def printSummary(self):
        """ Prints a summary of this service description.

            Prints "method(InputType) -> OutputType" of all methods (sorted by method name)
        """
        print("Service: {}".format(self.pbService.GetDescriptor().name))
        for tag, info in sorted(self._methodInfo.items(), key = lambda item: item[1][0]):
            methodName, InputType, OutputType = info

            print("{:15s}:  {}({}) -> {}".format(tag, methodName,
                                                 InputType.__name__, OutputType.__name__))

    def logSummary(self, logFunc=None):
        """ Logs a summary of this service description.

            :param logFunc: e.g. logger.debug, logger.info or print (default).
        """
        logFunc = print if logFunc is None else logFunc

        logFunc("ServiceWrapper for: {}".format(self.pbService.GetDescriptor().name))

        # Log "method(InputType) -> OutputType" of all rpc methods (sorted by method name)
        for tag, info in sorted(self._methodInfo.items(), key = lambda item: item[1][0]):
            methodName, InputType, OutputType = info
            logFunc("  {:15s}:  {}({}) -> {}".format(tag, methodName,
                                                     InputType.__name__, OutputType.__name__))



class Delegator():
    """ Is used by a delegating service to delegate all actual work to a worker class.

        The service class does the ZMQ communication. It then calls its own processBinaryMessage
        method which calls the processBinaryMessage method of the _delegator. Here the _delegator
        determines which worker method to call, constructs the parameters from the message fields,
        calls the method and packs the result back into a return message.

        To be able to know which worker method to call, the Delegator has a reference a service
        specification which is defined in .proto file and imported from the generated _pb2 module.

        The Delegator has a Packetizer object to be able to automatically pack and unpack the
        protocol buffer messages automatically.
    """
    def __init__(self, worker, serviceSpec, packetizer=None):
        """ Constructor.

            :param worker: a regular Python object that does the actual work. It must implemented
                the methods as specified in the serviceSpec. The method parameters must correspond
                to the message fields (i.e. have the same name and compatible type).
            :param serviceSpec: a google.protobuf.service.Service. This is defined in a .proto file
                and can be imported from the generated _pb2.py module.
            :param packetizer: a Packetizer, which specifies how to unpack messages. If None, the
                default GSSw packetizer is used, which know how to (un)pack arrays.
        """
        check_class(serviceSpec, PbService)
        check_class(packetizer, Packetizer, allow_none=True)

        self.worker = worker
        self.serviceWrapper = ServiceSpecWrapper(serviceSpec)
        self.packetizer = Packetizer.createDefaultPacketizer() if packetizer is None else packetizer

        self.serviceWrapper.logSummary(logFunc=logger.debug)
        self.serviceWrapper.checkWorkerMethodsPresent(worker)

    # I'd like to write InputType and OutputType starting with a capital, since they ar types.
    # pylint: disable=invalid-name

    def processBinaryMessage(self, tag, binMessage):
        """ Processes a binary message (i.e. a message that is not yet parsed from a byte string)

            Uses the tag to determine the message type and parses it. It uses the service
            specification to determine which worker method to call, and uses the packetizer to
            create the method arguments from the message field. It then calls the worker method
            and creates a return message from the result.

            Raises a ServiceProcessingError if the message cannot be processed.

            Also see the ServiceInterface.processBinaryMessage doc-string for more details.

            :param str tag: the name of the message
            :param bytes binMessage: binary data containing an encoded protocol buffer message
            :returns: (replyTag, binReplyMessage) tuple
        """
        logger.debug("Processing message: {}".format(tag))
        check_is_a_string(tag)
        check_is_binary(binMessage)

        # Find which method of the worker to call
        try:
            methodName, InputType, OutputType = self.serviceWrapper.methodInfo[tag]
        except KeyError:
            # Typically this happens if the service in the .proto file does not contain an rpc
            # method that takes a message of type tag as input.
            raise UnsupportedServiceError(
                "No {!r} message found in service description".format(tag)) from None

        # Unpack the parameters from the input message, call the message and pack the result into
        # the output message.

        try:
            logger.debug("methodName: %r, InputType: %r, OutputType: %r",
                         methodName, InputType, OutputType)

            workerMethod = getattr(self.worker, methodName)
            assert callable(workerMethod), "workerMethod is not callable: {!r}".format(workerMethod)

            requestMessage = InputType()
            requestMessage.ParseFromString(binMessage)
            kwargs = self.packetizer.getMessageAttrs(requestMessage)

            # Call the worker method. Don't catch any exceptions so that the stack trace is clearer.
            logger.debug("calling %r with args", workerMethod)
            result = workerMethod(**kwargs)
            logger.debug("called  %r got result of type: %s", workerMethod, type(result))

            # Create the reply message and populate it with the result of the worker method call.
            replyMessage = OutputType()

            # result can be None for one-way rpc methods (which do not return anything)
            if result is not None: 
                self.packetizer.setMessageAttrs(replyMessage, result)

            replyTag = tagName(replyMessage)
            if replyMessage.IsInitialized(): # Messages can be incomplete for one-way rpc methods
                binReplyMessage = replyMessage.SerializeToString()
            else:
                binReplyMessage = b''

        except Exception as ex:
            raise ServiceProcessingError("Unable to process input message of tag: {!r}"
                                         .format(tag)) from ex
        return replyTag, binReplyMessage


###################
# Service classes #
###################


class DelegatingReqRepService(AbstractReqRepService):
    """ A service that implements a request-reply pattern and delegates the work to a worker class.

        It has an internal _delegator object that automatically unpacks the message and calls the
        corresponding worker method with parameters constructed from the message fields.
    """
    def __init__(self, worker, serviceSpec, repSocket, packetizer=None):
        """ Constructor.
        """
        super().__init__(repSocket)
        self._delegator = Delegator(worker, serviceSpec, packetizer=packetizer)


    @property
    def worker(self):
        """ The worker object
        """
        return self._delegator.worker


    @property
    def packetizer(self):
        """ The packetizer object
        """
        return self._delegator.packetizer


    def processBinaryMessage(self, tag, binMessage):
        """ Processes a binary message by simply calling self._delegator.processBinaryMessage.
            See the ServiceInterface.processBinaryMessage doc-string for more details.
        """
        return self._delegator.processBinaryMessage(tag, binMessage)



class DelegatingPubSubService(AbstractPubSubService):
    """ A service that implements a pub-sub pattern and delegates the work to a worker class.

        It has an internal _delegator object that automatically unpacks the message and calls the
        corresponding worker method with parameters constructed from the message fields.
    """
    def __init__(self, worker, serviceSpec, subSocket, pubSocket=None, packetizer=None):
        """ Constructor.
        """
        super().__init__(subSocket, pubSocket)
        self._delegator = Delegator(worker, serviceSpec, packetizer=packetizer)


    @property
    def worker(self):
        """ The worker object
        """
        return self._delegator.worker


    @property
    def packetizer(self):
        """ The packetizer object
        """
        return self._delegator.packetizer


    def processBinaryMessage(self, tag, binMessage):
        """ Processes a binary message by simply calling self._delegator.processBinaryMessage.
            See the ServiceInterface.processBinaryMessage doc-string for more details.
        """
        return self._delegator.processBinaryMessage(tag, binMessage)


######################
# Qt Service classes #
######################


class DelegatingQtReqRepService(DelegatingReqRepService):
    """ A service that implements a request-reply pattern and delegates the work to a worker class.

        Specialized service for use in a Qt application. Connects the sigReadyReceive signal of its
        (notifying) REP socket to the receiveAndProcessMessage slot. This makes that worker methods
        are automatically are called when messages are recevied.

        It has an internal _delegator object that automatically unpacks the message and calls the
        corresponding worker method with parameters constructed from the message fields.
    """
    def __init__(self, worker, serviceSpec, repSocket, packetizer=None):
        """ Constructor.
        """
        super().__init__(worker, serviceSpec, repSocket, packetizer=packetizer)

        check_class(repSocket, NotifyingRepSocket)
        self._repSocket.sigReadyReceive.connect(self.receiveAndProcessMessage)


    def finalize(self):
        """ Used to cleanup resources (disconnect signals). Must be called explicitly.
        """
        self._repSocket.sigReadyReceive.disconnect(self.receiveAndProcessMessage)



class DelegatingQtPubSubService(DelegatingPubSubService):
    """ A service that implements a pub-sub pattern and delegates the work to a worker class.

        Specialized service for use in a Qt application. Connects the sigReadyReceive signal of its
        (notifying) SUB socket to the receiveAndProcessMessage slot. This makes that worker methods
        are automatically are called when messages are recevied.

        It has an internal _delegator object that automatically unpacks the message and calls the
        corresponding worker method with parameters constructed from the message fields.
    """
    def __init__(self, worker, serviceSpec, subSocket, pubSocket=None, packetizer=None):
        """ Constructor.
        """
        super().__init__(worker, serviceSpec, subSocket, pubSocket=pubSocket, packetizer=packetizer)

        check_class(subSocket, NotifyingSubSocket)
        self._subSocket.sigReadyReceive.connect(self.receiveAndProcessMessage)


    def finalize(self):
        """ Used to cleanup resources (disconnect signals). Must be called explicitly.
        """
        self._subSocket.sigReadyReceive.disconnect(self.receiveAndProcessMessage)
