""" This module can be used to receive asynchronous (science) data from a hardware device. """

import logging
from collections import namedtuple

import numpy as np

import gssw.proto.hwregister_pb2 as proto
import gssw.proto.common_pb2     as common_proto
import gssw.lib.sockets          as sockets

from gssw.hal.hal             import Hal
from gssw.hal.halclient       import HalClient
from gssw.lib.errors          import CommunicationError

logger = logging.getLogger(__name__)


class Science:
    """ This class is the interface to the asynchronous (science) data port of a HW device. """

    def __init__(self, deviceName, config, hal=None):
        """ Constructor.

            deviceName -- name of the device in the HAL
            hal        -- hardware abstraction layer object
            config     -- system configuration, a dictionary loaded from the config.yaml file
        """

        # Initialize variables.
        self.deviceName = deviceName

        # Hardware abstraction layer
        if hal is None:
            self.hal       = Hal()
            self.halClient = HalClient(self.hal, config)
            self.halClient.requestHal()
        else:
            self.hal = hal

        logger.info('Initializing python science interface')

        # Get the name of the daemon associated with the device from the HAL
        self.daemonName = self.hal.getDevice(deviceName=self.deviceName).scienceDaemon

        if not self.daemonName:
            self.daemonName = self.hal.getDevice(deviceName=self.deviceName).daemon
        logger.debug('daemonName: %s', self.daemonName)

        # Get the configuration for the HW daemon this instance communicates with
        deviceConfig = config.componentParameters(self.daemonName)
        logger.debug('configuration: %s', str(deviceConfig))

        # Build the devices' PUB socket address.
        pubAddress = 'tcp://%s:%d' % (deviceConfig['hostname'],
                                      deviceConfig['sockets']['device-pub'])

        # Create a SUB socket
        self.socket = sockets.SubSocket()
        self.socket.createZmqSocket()
        logger.debug('Connecting to PUB socket @ %s', pubAddress)
        self.socket.connectSocket(pubAddress)
        self.socket.subscribe()


    def poll(self):
        """ Poll the subscribe socket to check for available messages.

            Returns a boolean indicating whether new messages are available.
        """

        return self.socket.readyReceive()


    def getData(self):
        """ Receive all available science data.

            This funtion returns a list of all blockdata messages.
            A named tuple is created for each message containing the fields of the proto buffer.
            The data field is converted to a numpy array.
        """

        # Raise an error when no data is available.
        #if not self.poll():
        #    raise  CommunicationError('No messages available.')

        l = []

        while self.poll():

            msg = self.socket.receiveMultiMessage()
            if msg[1] != b'HwBlockRep':
                raise  CommunicationError('Unexpected message type. (%s)' % msg[1])
            pBuffer = proto.HwBlockRep()
            pBuffer.ParseFromString(msg[2])

            tupleType = namedtuple('BlockData', 'apid board write addr data status timestamp')

            print(pBuffer.data.bits_per_value)

            dtype = np.dtype(self.buildTypeString(pBuffer.data.value_encoding,
                                                  pBuffer.data.bits_per_value))
            dataArray = np.frombuffer(pBuffer.data.block_data, dtype=dtype)

            dataTuple = tupleType(apid=pBuffer.apid, board=pBuffer.board, write=pBuffer.write,
                                  addr=pBuffer.addr, data=dataArray, status=pBuffer.status,
                                  timestamp=pBuffer.timestamp)
            l.append(dataTuple)

        return l


    def buildTypeString(self, encoding, bits):
        """ Build the numpy type string form the relevant BlockData protobuffer fields.

            encoding -- BlockData protobuf encoding field
            bits     -- BlockData protobuf bitwidth field
        """

        if bits > 64 or bits % 8 != 0:
            raise ValueError('Bitwidth must be a 8, 16, 32 or 64.')

        typeStr = '<'

        if encoding == common_proto.ENC_UINT:
            typeStr += 'u'
        elif encoding == common_proto.ENC_SINT:
            typeStr += 'i'
        elif encoding == common_proto.ENC_FLOAT:
            typeStr += 'f'
        elif encoding == common_proto.ENC_STRING:
            typeStr += 'a'
        else:
            raise TypeError('Unknown type (%s)' % str(encoding))

        return typeStr + str(int(bits) >> 3)


