""" Wrapper classes for ZMQ sockets.

    The purpose of these wrapper classes is to harmonize the way how ZMQ is used within the GSSw.
    For a more detailed explaination see the :ref:`section_socket_classes` section in the
    developers guide.

    This module consists of the following classes:

    * AbstractSocket: Functionality common to all sockets
    * ReqSocketMixin: Functionality that is used only with REQ sockets
    * RepSocketMixin: Functionality that is used only with REP sockets
    * PubSocketMixin: Functionality that is used only with PUB sockets
    * SubSocketMixin: Functionality that is used only with SUB sockets
    * ReqSocket: Combines the functionality of the AbstractSocket and the ReqSocketMixin
    * RepSocket: Combines the functionality of the AbstractSocket and the RepSocketMixin
    * PubSocket: Combines the functionality of the AbstractSocket and the PubSocketMixin
    * SubSocket: Combines the functionality of the AbstractSocket and the SubSocketMixin
    * Poller: Simple wrapper arround zmq.Poller class that can handle AbstractSockets.

    The AbstractSockets and Mixins are abstract classes which should not be used directly. Only
    objects of one the last four classes (ReqSocket, RepSocket, PubSocket, SubSocket) are to be
    created directly.

    A poor man's inheritance diagram would look as follows (only for the SUB socket): ::

        +----------------+   +----------------+
        | AbstractSocket |   | SubSocketMixin |
        +----------------+   +----------------+
                |                     |
                +----------+----------+
                           |
                     +-----^-----+
                     | SubSocket |
                     +-----------+

    Socket wrappers that depend on QtCore.QObject are placed in socketsqt.py
"""
import logging, collections, uuid
import functools
import zmq

from gssw.lib.errors import GsswSocketError, GsswTimeoutError

from gssw.lib.typing import check_class, check_is_a_string, check_is_binary, type_name, is_a_type

logger = logging.getLogger(__name__)


# Sets the default read timeout. If the value is 0, receiveMessage will raise a zmq.Again
# exception immediately if there is no message to receive. If the value is -1, it will block
# until a message is available. For all other values, it will wait for a message for that
# amount of time before raising a zmq.Again exception.
DEFAULT_RECEIVE_TIMEOUT_MS=5000
DEFAULT_LINGERING_DURATION=100  # Delay before closing a socket while messages are pending
INFINITE_TIMEOUT = -1

# Dictionary for converting socket type integer constants to string
SOCKET_TYPE_STRING = {zmq.PUB: 'PUB', zmq.SUB: 'SUB', zmq.REQ: 'REQ', zmq.REP: 'REP'}


# Named tuples with message parts for PUB-SUB and REQ-REP communication.
# pylint: disable=bad-whitespace

ReqRepParts = collections.namedtuple('ReqRepParts', ('tag', 'message'))
ReqRepParts.__doc__ = """
    A namedtuple of message parts (tag, message), which are transferred during REQ/REP
    communication.

    The tag is typically decoded (i.e. regular string). The message typically is not decoded but
    binary data (a Python bytes object) that can be converted to a Protocol Buffer.

    You can use it as a tuple like this: ::

       (tag, message) = receiveProtoBuf()
       print(message)

    Or you can use named attributes as with regular python objects: ::

       parts = receiveProtoBuf()
       print(parts.message)

    See also https://docs.python.org/3.5/library/collections.html#collections.namedtuple
"""
try:
    ReqRepParts.tag.__doc__    = "Identifies the type of protocol buffer used to encode the message"
    ReqRepParts.message.__doc__= "A message encoded with a protocol buffer"
except AttributeError:
    pass # Property docstrings are only writeable in Python 3.5 and higher


PubSubParts = collections.namedtuple('PubSubParts', ('topic', 'tag', 'message'))
PubSubParts.__doc__ = """
    A namedtuple of message parts (topic, tag, message) which are transferred during PUB/SUB
    communication.

    A namedtuple of message parts (tag, message), which are transferred during REQ/REP
    communication.

    The topic and tag are typically decoded (i.e. regular string). The message typically is not
    decoded but binary data (a Python bytes object) that can be converted to a Protocol Buffer.

    You can use it as a tuple like this: ::

       (topic, tag, message) = receiveProtoBuf()
       print(message)

    Or you can use named attributes as with regular python objects: ::

       parts = receiveProtoBuf()
       print(parts.message)

    See also https://docs.python.org/3.5/library/collections.html#collections.namedtuple
"""
try:
    PubSubParts.topic.__doc__  = "Message topic on which subscribers can filter"
    PubSubParts.tag.__doc__    = "Identifies the type of protocol buffer used to encode the message"
    PubSubParts.message.__doc__= "A message encoded with a protocol buffer"
except AttributeError:
    pass # Property docstrings are only writeable in Python 3.5 and higher

# pylint: enable=bad-whitespace



def tagName(msg):
    """ Returns the tag name given a Protocol buffer message or message class.

        All GSSw protocol buffer messages contain a tag name. This tells receivers which message
        type the payload contains.
    """
    if is_a_type(msg):
        # msg is a protocol buffer class
        return msg.__name__
    
    # msg is a protocol buffer object
    return type_name(msg)


# # The pattern is not very strict because it can be used for partial addresses. Therefore, it
# # cannot be used for validation!
# # If you update this pattern please also update the unit tests in testscripts/utest_sockets.py
# REGEXP_ADDRESS = re.compile(r'\A([a-z]+://)?([*.a-z0-9]+:)?([0-9]+)\Z')
#
# def parseAddress(address):
#     """ Parses a socket address.
#
#         The address is expected to be a string of the form: protocol://host:port
#         A ValueError is raised if this is not the case.
#
#         :param address: string containing an address. (Can be an int with just a port number, this
#             is usefull when reading from Yaml files)
#         :return: (protocol, host, port) tuple. The port number is converted to int.
#     """
#     matchObj = REGEXP_ADDRESS.match(str(address))
#     if matchObj is None:
#         raise ValueError("Address has not the correct format: {!r}".format(address))
#
#     protocol = matchObj.group(1)
#     host= matchObj.group(2)
#     port = matchObj.group(3)
#
#     protocol = protocol[:-3] if protocol is not None else None  # strip '://' from end of match
#     host = host[:-1] if host is not None else None  # strip colon from end of match
#     port = int(port) if port is not None else None
#
#     return protocol, host, port


# The pattern is not very strict because it can be used for partial addresses. Therefore, it
# cannot be used for validation!
# If you update this pattern please also update the unit tests in testscripts/utest_sockets.py
# REGEXP_ADDRESS = re.compile(r'\A([a-z]+://)?([*.a-z0-9]+:)?([0-9]+)\Z')

def parseAddress(address):
    """ Parses a socket address.

        The address is expected to be a string of the form: protocol://endpoint

        In case the protocol is left out, tcp is assumed. If the protocol is tcp, the endpoint
        consists of a host:port combination. Otherwise the endpoint is just a string.

        A ValueError is raised if this is not the case.

        :param address: string containing an address. (Can be an int with just a port number, this
            is useful when reading from Yaml files)
        :return: (protocol, host, port) tuple if the protocol equals 'tcp'. The port number is
            converted to int. If the protocol is not equal to tcp (protocol, endpoint, None) is
            returned.
    """
    check_class(address, (str, int))

    valid_protocols = ['tcp', 'inproc', 'ipc']

    parts = str(address).strip().split('://', 1)
    if len(parts) == 2:
        protocol, endpoint = parts
    else:
        protocol, endpoint = None, parts[0]

    if protocol == 'tcp' or protocol is None:
        parts = endpoint.split(':', 1)
        if len(parts) == 2:
            host, port = parts
        else:
            host, port = None, parts[0]

        try:
            port = int(port)
        except ValueError:
            print("address: {}".format(address))
            raise
    elif protocol in valid_protocols:
        host, port = endpoint, None
    else:
        raise ValueError("Protocol must be one of {}, got: {}".format(valid_protocols, protocol))
    return protocol, host, port



def destroyAllSocketsInContext(context=None, linger=0):
    """ Close all sockets associated with this context, and then terminate the context.

        A well written program should always call this function before exiting. The reason is that
        any ZQM exception (e.g. a zmq.Again exception after a time out) may leave libzmq hanging.
        See: http://zguide.zeromq.org/page:chapter1#Making-a-Clean-Exit

        Good practice is therefore to use:

            try:
                myMainRoutine()
            finally:
                destroyAllSocketsInContext()

        WARNING: if you call this twice with context=None, it will create a new context and
        destroy it. This will break if you log to eventlog-server afterwards. Therefore just
        call it once at program exit, as in the example above.

        WARNING from http://pyzmq.readthedocs.io/en/latest/api/zmq.html#zmq.Context.destroy
            [Context]destroy involves calling zmq_close(), which is NOT threadsafe. If there are
            active sockets in other threads, this must not be called.

        Therefore we should use a separate ZMQ context per thread?

        :param context: the context to destroy. If None, the global singleton context is destroyed.
        :param linger: if specified, the LINGER sockopt of the sockets will be set prior to closing.
    """
    logger.debug("Destroying ZMQ context to make a clean exit. Lingering for {} seconds"
                 .format(linger))

    context = zmq.Context.instance() if context is None else context
    context.destroy(linger=linger)


def catchZmqErrors(func):
    """ Wrap low-level zmq exceptions, such as timeouts.

        This avoids the (ridiculously) long exception stacktrace of ZMQ."""

    @functools.wraps(func) # preserve original function docstrings
    def catchZmqErrorsWrapper(*args, **kwargs):
        """ The decorated function is wrapped in this function. """
        try:
            return func(*args, **kwargs)
        except zmq.Again as ex:
            # HERE we can (optionally) reset the underlying zmq socket, such that our own socket is
            # still usable after a timeout.
            raise GsswTimeoutError('Timeout on blocking operation in function {}: {}'
                .format(func.__name__, ex)) from None # Discard original stack trace
        except zmq.ZMQError as ex:
            raise GsswSocketError('Socket error on blocking operation in function {}: {}'
                .format(func.__name__, ex)) from None # Discard original stack trace
    return catchZmqErrorsWrapper


class AbstractSocket():
    """ Abstract base class for GSSw ZMQ socket wrapper classes.

        Every AbstractSocket has an identifier and remembers the addresses to which it is connected
        or bound.

        All send and receive commands are done with the copy parameter set to True. In the future
        we may examine if we need non-copying manner as well for possible speed increase.
    """
    def __init__(self, identifier=None, context=None):
        """ Constructor

            :param identifier: Should be a unique ID. If not given a GUID will be autogenerated
                Must be immutable so it can be used as a key in a dictionary.
            :param context: ZMQ context that will be used to create the underlying ZMQ socket
                If None the global context singleton will be used.
        """
        super().__init__()

        # The identifier is necessary for knowing which socket the message originates from. Perhaps
        # we don't need this, or perhaps we can use the file descriptors instead. TBD.
        # For now I made the identifier parameter optional and generated a GUID if it is not given.
        # This allows us to create sockets without the identifier parameter.
        self._identifier = uuid.uuid4() if identifier is None else identifier
        self._context = zmq.Context.instance() if context is None else context
        self._boundTo = []     # list of addresses this socket is bound to
        self._connectedTo = [] # list of addresses this socket is connected to

        self._zmqSocket = None # must be created by create socket


    @catchZmqErrors
    def close(self):
        """ Close the zmq socket.
        """

        self._zmqSocket.close()


    def finalize(self):
        """ Can be used to cleanup resources (like a destructor)
        """
        pass


    def __repr__(self):
        return "<{} id={}>".format(type_name(self), self.identifier)


    @property
    def identifier(self):
        """ Unique socket ID.
        """
        return self._identifier


    @property
    def context(self):
        """ The ZMQ context that was used to create the underlying ZMQ socket.
        """
        return self._context


    @property
    def zmqSocket(self):
        """ The underlying ZMQ Socket.

            :return: a zmq.Socket object
        """
        if self._zmqSocket:
            return self._zmqSocket
        else:
            raise ValueError("ZMQ socket undefined (use createZmqSocket to create)")


    @catchZmqErrors
    def createZmqSocket(self, timeoutMs=None):
        """ Creates the socket. The type will be determined by the descendant.

            Uses self.socketType to determine the ZMQ socket type. This method should be
            provided by the mixin/descendant so createZmqSocket cannot be called in the
            constructor. We therefore have to call it explicitly.

            :param timeoutMs: set the receive-timeout to this number of milliseconds.
                Use -1 for infinite timeout. If None, the receive-timeout set to
                DEFAULT_RECEIVE_TIMEOUT_MS for request sockets, and not set for other types of
                sockets (which effectively makes other types of sockets have an infinite timeout)

            TODO: Or make the getter create if it doesn't exist?
        """
        # pylint: disable=no-member
        #   self.socketType should be provided by the mixin class.

        logger.debug("Creating socket of type: {}".format(SOCKET_TYPE_STRING[self.socketType]))
        self._zmqSocket = self.context.socket(self.socketType)

        # Set the socket lingering to allow closing sockets while messages are pending
        self.zmqSocket.linger = DEFAULT_LINGERING_DURATION

        if timeoutMs is None and self.socketType == zmq.REQ:
            timeoutMs = DEFAULT_RECEIVE_TIMEOUT_MS

        if timeoutMs is not None:
            logger.debug("Setting timeout to: {} ms".format(timeoutMs))
            self.setReceiveTimeout(timeoutMs)

        # pylint: enable=no-member


    @property
    def connectedTo(self):
        """ List of addresses this socket is connected to.

            Returns a copy so you can modify the result.
        """
        return list(self._connectedTo)


    @catchZmqErrors
    def connectSocket(self, address):
        """ Connect to the address.

            Not to be confused with the connect method, which connects QtSignals.
        """
        check_is_a_string(address)
        logger.debug("Connecting socket {!r} to: {}".format(self.identifier, address))
        self.zmqSocket.connect(address)
        self._connectedTo.append(address)


    @catchZmqErrors
    def disconnectSocket(self, address):
        """ Disconnects from the address

            Not to be confused with the disconnect method, which disconnects QtSignals.
        """
        check_is_a_string(address)
        logger.debug("Disconnecting socket {!r} from: {!r}".format(self.identifier, address))
        self.zmqSocket.disconnect(address)
        self._connectedTo.remove(address)


    @property
    def boundTo(self):
        """ List of addresses this socket is bound to.

            Returns a copy so you can modify the result.
        """
        return list(self._boundTo)


    @catchZmqErrors
    def bindSocket(self, address):
        """ Binds the ZMQ socket to the address.

            The address must be an IP address or a "*" to indicate that connections are accepted
            from all hosts. You cannot use a hostname or 'localhost'. This results in the following
            error: zmq.error.ZMQError: No such device
        """
        check_is_a_string(address)
        logger.debug("Binding socket to: {!r}".format(address))
        self.zmqSocket.bind(address)
        self._boundTo.append(address)


    @catchZmqErrors
    def bindSocketToRandomPort(self, address):
        """ Binds the ZMQ socket to the address.

            :param str address: The address must be an IP address or a "*" to indicate that
                                connections are accepted from all hosts. You cannot use a hostname
                                or 'localhost'. This results in the following error:
                                zmq.error.ZMQError: No such device
                                The address must not contain a port number.

            :return: Port number on which the socket is listening.
        """
        check_is_a_string(address)
        port = self.zmqSocket.bind_to_random_port(address)
        logger.debug("Socket is bound to: '{}:{}'".format(address, port))
        self._boundTo.append(address)
        return port


    @catchZmqErrors
    def unbindSocket(self, address):
        """ Unbinds the ZMQ socket from the address
        """
        check_is_a_string(address)
        logger.debug("Unbinding socket from: {}".format(address))
        self.zmqSocket.unbind(address)
        self._boundTo.remove(address)


    @catchZmqErrors
    def setReceiveTimeout(self, timeoutMs):
        """ Make the socket raise a zmq.Again exception when a receiveMessage does not return
            after timeoutMs milli seconds.

            Sets the RCVTIMEO socket option on the underlying ZMQ socket. If the value is 0,
            receiveMessage will raise a zmq.Again exception immediately if there is no message
            to receive. If the value is -1, it will block until a message is available. For all
            other values, it will wait for a message for that amount of time before raising a
            zmq.Again exception. This is the default.

            There is no corresponding setSendTimeout function because we don't expect timeouts on
            sendMessage calls. If a the connection fails in, for instance, a REQ socket the sent
            message will remain in the output buffer of libzmq but the call to sendMessage will
            return successfully. The timeout will occur in the subsequent call to receiveMessage.

            :param timeoutMs: timeout in milli seconds
        """
        logger.debug("Setting {} receive timeout to {} ms".format(self, timeoutMs))
        self.zmqSocket.setsockopt(zmq.RCVTIMEO, timeoutMs)


    @catchZmqErrors
    def getReceiveTimeout(self):
        """ Gets the RCVTIMEO socket option on the underlying ZMQ socket. If the value is 0,
            receiveMessage will raise a zmq.Again exception immediately if there is no message
            to receive. If the value is -1, it will block until a message is available. For all
            other values, it will wait for a message for that amount of time before raising a
            zmq.Again exception. This is the default.

            :see: Related function: setReceiveTimeout.
            :return: Timeout (in ms).
        """
        return self.zmqSocket.getsockopt(zmq.RCVTIMEO)


    @catchZmqErrors
    def readyReceive(self):
        """ Returns True if the socket can read at least one message

            Checks the ZMQ_EVENTS socket options masked with zmq.POLLIN
        """
        return bool(self.zmqSocket.getsockopt(zmq.EVENTS) & zmq.POLLIN)


    @catchZmqErrors
    def __not_used__receiveMessage(self, flags=0):
        """ Reads a message from the ZMQ socket.

            All communication within the GSSw is done with multi-part messages so this function
            is not expected to be used much. Use receiveMultiMessage instead.

            :param flags: Any valid recv flag (default=0)
            :return: the read message as a binary (bytes) string
        """
        logger.debug("receiveMessage called")
        assert self._connectedTo or self._boundTo, "Socket not connected or bound {}".format(self)
        return self.zmqSocket.recv(flags=flags)


    @catchZmqErrors
    def __not_used__sendMessage(self, message, flags=0):
        """ Sends a message to the ZMQ socket.

            All communication within the GSSw is done with multi-part messages so this function
            is not expected to be used much. Use sendMultiMessage instead.

            :param mesage: bytes to be send
            :param flags: Any valid send flag (default=0)
                It is typically not necessary to set this flag.
                Look in https://pyzmq.readthedocs.io/en/latest/api/zmq.html#zmq.Socket.send
                for the possible values.
        """
        logger.debug("sendMessage called")
        check_is_binary(message)
        assert self._connectedTo or self._boundTo, "Socket not connected or bound {}".format(self)
        self.zmqSocket.send(message, flags=flags)


    @catchZmqErrors
    def receiveMultiMessage(self, flags=0):
        """ Reads a multipart message from the ZMQ socket.

        If there is no data available this function will block until there is data or until
        the timeout has elapsed. See also the setReveiveTimeout doc-string.

        :param flags: Any valid recv flag (default=0). Look in
                      https://pyzmq.readthedocs.io/en/latest/api/zmq.html#zmq.Socket.send_multipart
                      for the possible values.
        :return: tuple of bytes (binary strings) containing the message parts.
        """
        #logger.debug("receiveMultiMessage called")
        assert self._connectedTo or self._boundTo, "Socket not connected or bound {}".format(self)
        return self.zmqSocket.recv_multipart(flags=flags)


    @catchZmqErrors
    def sendMultiMessage(self, parts, flags=0):
        """ Sends a multipart message to the ZMQ socket.

        :param parts: tuple of bytes to be send
        :param flags: Any valid send flag (default=0). Look in
                      https://pyzmq.readthedocs.io/en/latest/api/zmq.html#zmq.Socket.send_multipart
                      for the possible values.
        """
        #logger.debug("sendMultiMessage called")
        assert self._connectedTo or self._boundTo, "Socket not connected or bound {}".format(self)
        self.zmqSocket.send_multipart(parts, flags=flags)


######################
# Socket type mixins #
######################

# The receiveProtoBuf and sendProtoBuf of the ReqSocketMixin and RepSocketMixin are exactly the
# same. I didn't add a common ancestor because didn't want to make the class hierarchy more
# complicated than it already is. TODO: think of best solution.

class ReqSocketMixin():
    """ Mixin class that provides specific functionality for REQ sockets
    """

    @property
    def socketType(self):
        """ ZMQ constant representing the socket type. I.e. zmq.REQ
        """
        return zmq.REQ


    def receiveProtoBuf(self, flags=0):
        """ Receives a protocol buffer encoded message.

            :param flags: Any valid read flag (default=0)
            :return: ReqRepParts(tag, message) named tuple where:
                tag identifies the type of protocol buffer in the message,
                message is a protocol buffer encoded message (bytes),
        """
        binTag, message = self.receiveMultiMessage(flags=flags)
        tag = binTag.decode(encoding='utf-8')
        return ReqRepParts(tag, message)


    def sendProtoBuf(self, tag, message, flags=0):
        """ Send a protocol buffer encoded message

            The tag and message all both be of type bytes. Prepend string literals with a 'b'
            to achieve this. E.g. sendProtoBuf(b"my-tag", message)
            :param tag: identifies the type of protocol buffer in the message
            :param message: a protocol buffer encoded message (bytes)
            :param flags: Any valid send flag (default=0)
        """

        check_is_a_string(tag)
        check_is_binary(message) ###
        binTag = tag.encode(encoding='utf-8')
        self.sendMultiMessage(ReqRepParts(binTag, message), flags=flags)


class RepSocketMixin():
    """ Mixin class that provides specific functionality for REP sockets
    """

    @property
    def socketType(self):
        """ ZMQ constant representing the socket type. I.e. zmq.REP
        """
        return zmq.REP


    def receiveProtoBuf(self, flags=0):
        """ Receives a protocol buffer encoded message.

            :param flags: Any valid read flag (default=0)
            :return: ReqRepParts(tag, message) named tuple where:
                tag identifies the type of protocol buffer in the message,
                message is a protocol buffer encoded message (bytes),
        """
        binTag, message = self.receiveMultiMessage(flags=flags)
        tag = binTag.decode(encoding='utf-8')
        return ReqRepParts(tag, message)


    def sendProtoBuf(self, tag, message, flags=0):
        """ Send a protocol buffer encoded message

            The tag and message all both be of type bytes. Prepend string literals with a 'b'
            to achieve this. E.g. sendProtoBuf(b"my-tag", message)
            :param tag: identifies the type of protocol buffer in the message
            :param message: a protocol buffer encoded message (bytes)
            :param flags: Any valid send flag (default=0)
        """
        check_is_a_string(tag)
        check_is_binary(message)
        binTag = tag.encode(encoding='utf-8')
        self.sendMultiMessage(ReqRepParts(binTag, message), flags=flags)


    @catchZmqErrors
    def poll(self, timeout=None):
        """ Poll the reply socket

        :param timeout: the maximum amount of time spent polling in ms.
                        zero or None means an infinite amount of time.

        returns events, The events that are ready and waiting.
        Will be 0 if no events were ready by the time timeout was reached.
        """

        return self._zmqSocket.poll(timeout=timeout)



class PubSocketMixin():
    """ Mixin class that provides specific functionality for PUB sockets
    """

    @property
    def socketType(self):
        """ ZMQ constant representing the socket type. I.e. zmq.PUB
        """
        return zmq.PUB


    def publishProtoBuf(self, topic, tag, message, flags=0):
        """ Publishes a protocol buffer encoded message

            The topic, tag and message all must be of type bytes. Prepend string literals with a 'b'
            to achieve this. E.g. publishProtoBuf(b"my-topic", b"my-tag", message)
            :param topic: topic on which subscribers can filter.
            :param tag: identifies the type of protocol buffer in the message
            :param message: a protocol buffer encoded message (bytes)
            :param flags: Any valid send flag (default=0)
        """
        check_is_a_string(topic)
        check_is_a_string(tag)
        check_is_binary(message)
        binTopic = topic.encode(encoding='utf-8')
        binTag = tag.encode(encoding='utf-8')
        self.sendMultiMessage(PubSubParts(binTopic, binTag, message), flags=flags)



class SubSocketMixin():
    """ Mixin class that provides specific functionality for SUB sockets
    """

    @property
    def socketType(self):
        """ ZMQ constant representing the socket type. I.e. zmq.SUB
        """
        return zmq.SUB


    @catchZmqErrors
    def subscribe(self, topic=""):
        """ Subscribes to a topic.

            :param topic: topic filter string. Use empty string (default) all topics.
        """
        check_is_a_string(topic)
        logger.debug("Subscribing {} to {!r}".format(self, topic))
        self.zmqSocket.setsockopt_string(zmq.SUBSCRIBE, topic)


    @catchZmqErrors
    def unsubscribe(self, topic=""):
        """ Unsubscribes from a topic.

            :param topic: topic filter string. Use empty string (default) for all topics.
        """
        check_is_a_string(topic)
        logger.debug("Unsubscribing {} from {!r}".format(self, topic))
        self.zmqSocket.setsockopt_string(zmq.UNSUBSCRIBE, topic)


    # Named differently because it returns a different type than receiveProtoBuf
    def receiveProtoBufWithTopic(self, flags=0):
        """ Receives a protocol buffer encoded message.

            :param flags: Any valid read flag (default=0)
            :return: (topic, tag, message) tuple where:
                message is a protocol buffer encoded message (bytes),
                tag identifies the type of protocol buffer in the message,
                topic message topic on which subscribers can filter.
        """
        binTopic, binTag, message = self.receiveMultiMessage(flags=flags) # checks msg has 3 parts
        topic = binTopic.decode(encoding='utf-8')
        tag = binTag.decode(encoding='utf-8')
        return PubSubParts(topic, tag, message)


#######################
# Socket type classes #
#######################

# Non-Qt sockets
class ReqSocket(AbstractSocket, ReqSocketMixin):
    """ A basic wrapper class for a REQ socket"""
    pass

class RepSocket(AbstractSocket, RepSocketMixin):
    """ A basic wrapper class for a REP socket"""
    pass

class PubSocket(AbstractSocket, PubSocketMixin):
    """ A basic wrapper class for a PUB socket"""
    pass

class SubSocket(AbstractSocket, SubSocketMixin):
    """ A basic wrapper class for a SUB socket"""
    pass



########################
# Poller Wrapper class #
########################


class Poller():
    """ Wrapper class around the PyZMQ poller class.

        The poll method returns a list of tuples, just like the zmq.Poller class. However, the
        first element of each tuple is a AbstractSocket instead of a bare PyZMQ socket.
    """

    def __init__(self):
        """ Constructor

        """
        self._poller = zmq.Poller()
        self._registeredSockets = {}


    @property
    def registeredSockets(self):
        """ Returns the dictionary with registered sockets.

            This is a (zmq.Socket -> AbstractSocket) dictionary
        """
        return self._registeredSockets


    def registerPollIn(self, socket):
        """ Registers a GSSw socket for polling for input.

            :param socket: An AbstractSocket descendant.
        """
        check_class(socket, AbstractSocket)
        if socket.zmqSocket in self.registeredSockets:
            if self.registeredSockets[socket.zmqSocket] != socket:
                raise ValueError("ZMQ socket {} allready registered to another GSSW socket: {}"
                                 .format(self.registeredSockets[socket.zmqSocket], socket))

        self._poller.register(socket.zmqSocket, zmq.POLLIN)
        self._registeredSockets[socket.zmqSocket] = socket


    def unregister(self, socket):
        """ Registers a GSSw socket
        """
        check_class(socket, AbstractSocket)
        del self._registeredSockets[socket.zmqSocket]
        self._poller.unregister(socket.zmqSocket)


    @catchZmqErrors
    def poll(self, timeout=None):
        """ Poll the registered GSSw sockets for input.

            Calls zqm.poll but and returns the same list of tuples, except that the first element
            is not a zmq.Socket but the registered AbstractSocket descendant.

            :param timeout: the timeout in milliseconds. If None, no timeout (infinite).
            :return: events - The list of events that are ready to be processed. This is a list
                of tuples of the form (AbstractSocket, event), and the poll event mask
                (POLLIN, POLLOUT) is the second.
                If case of a timeout an empty list if returned.

                It is common to call events = dict(poller.poll()), which turns the list of tuples
                into a mapping of socket : event. But I (PTK) don't see the point of this.
        """
        events = self._poller.poll(timeout=timeout) # list of (zmq.socket, mask) tuples

        if events:
            logger.debug("A socket has data.")
        else:
            logger.debug("Polling time out passed ({} ms)".format(timeout))

        result = []
        for zmqSocket, eventMask in events:

            # Sanity check. Only POLLIN is implemented so eventMask must be this.
            assert eventMask == zmq.POLLIN, "Unexpected eventMask: {}".format(eventMask)

            result.append((self._registeredSockets[zmqSocket], eventMask))

        return result
