""" Wrapper classes (deriving from QObject) for ZMQ sockets.

    This module depends on QtCore. If you are not making a Qt user interface (or a Qt program that
    has it's own event loop) you can probably use the clasess in sockets.py instead.

    All classes in this module derive from the AbstractSocket class so they have at least the same
    the same functionality as their non-Qt counterparts. They also derive from QtCore.QObject so
    they can emit signals.

    This module contains the following classes:

    * AbstractQtSignalingSocket: Emits a signal when a message was received.
    * QtSubSocket: AbstractQtSignalingSocket for SUB sockets.

    * AbstractQtNotifyingSocket: Emits a signal when data is ready for receiving.
    * NotifyingSubSocket: AbstractQtNotifyingSocket for SUB sockets
    * NotifyingRepSocket: AbstractQtNotifyingSocket for REP sockets

    A poor man's inheritance diagram would look as follows (only for the SUB socket): ::

               +----------------+
               | AbstractSocket |
               +----------------+
                       |
        +---------------------------+     +----------------+
        | AbstractQtNotifyingSocket |     | SubSocketMixin |
        +---------------------------+     +----------------+
                       |                           |
                       +--------------+------------+
                                      |
                           +----------^---------+
                           | NotifyingSubSocket |
                           +--------------------+

    The abstract classes are not meant to be instantiated directly so you would typically only use
    the NotifyingSubSocket and NotifyingRepSocket classes (the QtSubSocket is only used in some
    example code for a threaded solutions)

"""
import logging

from PyQt5 import QtCore

from gssw.lib.sockets import AbstractSocket, SubSocketMixin, RepSocketMixin


logger = logging.getLogger(__name__)



class AbstractQtSignalingSocket(AbstractSocket, QtCore.QObject):
    """ Abstract class that can will emit a sigMessageReceived signal after a message was received.

        Typically not used within the GSSw, only for experiments.
    """
    sigMultiMessageReceived = QtCore.pyqtSignal(tuple) # tuple of bytes (binary strings)

    def __init__(self, identifier=None, context=None, parent=None):
        """ Constructor

            :param parent: Qt parent (QObject)
        """
        AbstractSocket.__init__(self, identifier=identifier, context=context)
        QtCore.QObject.__init__(self, parent=parent)


    def receiveMessage(self, flags=0):
        """ Read a message from the ZMQ socket.

            Emits the sigMessageReceived signal with the message as signal parameter

            :param flags: Any valid recv flag (default=0)
            :return: the read message as a binary (bytes) string
        """
        msg = super().receiveMessage(flags=flags)

        logger.debug("Emitting sigMessageReceived with message of {} bytes".format(len(msg)))
        self.sigMessageReceived.emit(msg)

        return msg


    def receiveMultiMessage(self, flags=0):
        """ Read a message from the ZMQ socket.

            Emits the sigMultiMessageReceived signal with the message parts list as signal parameter

            :param flags: Any valid recv flag (default=0)
            :return: list of bytes (binary strings) containing the message parts.
        """
        parts = super().receiveMultiMessage(flags=flags)

        logger.debug("Emitting sigMultiMessageReceived with {} parts".format(len(parts)))
        self.sigMultiMessageReceived.emit(tuple(parts))

        return parts



class AbstractQtNotifyingSocket(AbstractSocket, QtCore.QObject):
    """ GSSw socket class that wraps a ZMQ socket and uses a QtSocketNotifier to monitor input.

        The QSocketNotifier.activated signal is emitted when data is available for receiving.
        However it may trigger false alarms, therefore this class checks (in self.checkReadyReceive)
        if data is actually available and then emits the sigReadyReceive signal. The user of this
        class then can safely call the receiveMultiMessage method.

        The `zmq_getsockopt-ZMQ_FD in the ZMQ API <http://api.zeromq.org/4-0:zmq-getsockopt>`_
        says the following::

            The ability to read from the returned file descriptor does not necessarily indicate
            that messages are available to be read from, or can be written to, the underlying
            socket; applications must retrieve the actual event state with a subsequent
            retrieval of the ZMQ_EVENTS option.

            The returned file descriptor is also used internally by the zmq_send and zmq_recv
            functions. As the descriptor is edge triggered, applications must update the state
            of ZMQ_EVENTS after each invocation of zmq_send or zmq_recv.To be more explicit:
            after calling zmq_send the socket may become readable (and vice versa) without
            triggering a read event on the file descriptor.

        The second paragraph is a bit unclear but states that the socket events option may
        change if new data becomes available while a zmq_send or zmq_recv is in progress, This
        will then NOT trigger a state change and thus no QSocketNotifier.activated signal is
        emitted. Therefore we must always check if data is available for polling after
        a zmq_send or zmq_recv. Here we do this by scheduling a new call to checkReadyReceive,
        which is achieved by emitting the _sigScheduleCheckReadyReceive signal. This uses a queued
        connection so the checkReadyReceive is not executed directly but later, via the event loop.

        This class emits the sigReadyReceive signal when data is available. The user of this class
        is then responsible for reading the data. It would seem simpler to read the data and then
        emit a sigMultiMessageReceived signal. There are few reasons for the current choice:

        1) There are different kind of possible messages (single part, ReqRepParts,
           PubSubParts) and we should then make read methods and corresponding signals for all of
           them.
        2) The messages are potentially large and Qt will make a copy of the message for every
           emitted signal. If at a later stage this has to be optimized it would probably involve
           rewriting the functionality as we have done it now.
        3) To prevent fast growing buffers if the receiver of the signal cannot keep up with the
           sender. However, this applies only for queued connections.
    """

    # Signal that is emitted when data is available for reading.
    sigReadyReceive = QtCore.pyqtSignal(object) # object parameter will be self

    # Internal signal that will schedule a call to checkReadyReceive using a queued connection.
    _sigScheduleCheckReadyReceive = QtCore.pyqtSignal(int)

    def __init__(self, identifier=None, context=None, parent=None):
        """ Constructor

            :param context: ZMQ context that will be used to create the underlying ZMQ socket
                If None the global context singleton will be used.
            :param parent: Qt parent (QObject)
        """
        AbstractSocket.__init__(self, identifier=identifier, context=context)
        QtCore.QObject.__init__(self, parent=parent)

        self._notifier = None # will be created when the ZMQ socket is created.


    def finalize(self):
        """ Can be used to cleanup resources
        """
        self._notifier.activated.disconnect(self.receiveMultiMessagesIfReady)
        self._notifier = None


    def createZmqSocket(self, timeoutMs=None):
        """ Creates the socket and the socketNotifyer that alerts for new input.
        """
        super().createZmqSocket(timeoutMs=timeoutMs)
        self._notifier = QtCore.QSocketNotifier(self.zmqSocket.FD, QtCore.QSocketNotifier.Read,
                                                parent=self)
        self._notifier.activated.connect(self.checkReadyReceive)
        self._sigScheduleCheckReadyReceive.connect(self.checkReadyReceive,
                                                   type=QtCore.Qt.QueuedConnection)


    @QtCore.pyqtSlot(int)
    def checkReadyReceive(self, socketFD=None):
        """ Checks that the socket can read a data. If so it emits the sigReadyReceive signal.

            This slot is typically connected to the QSocketNotifier.activated signal. However the
            signal may be emitted even when there is not message data available, Therefore we first
            must check if a message can be read, and only emit the sigReadyReceive if this is true.

            :param socketFD: socket file descriptor. If not None it is used to check the underlying
                file-descript of the ZMQ socket is the same.
        """
        # Sanity check on the socket file descriptor (if given).
        if socketFD is not None:
            logger.debug("socketReadyRead: FD={}".format(socketFD))
            assert self.zmqSocket.FD == socketFD, \
                "File descriptor mismatch: {} != {}".format(self.zmqSocket.FD, socketFD)
        else:
            logger.debug("socketReadyRead: FD=auto")

        # Only emit sigReadyReceive if there is actual data
        if self.readyReceive():
            logger.debug("Emitting sigReadyReceive")
            self.sigReadyReceive.emit(self)


    def receiveMultiMessage(self, flags=0):
        """ Reads a multipart message from the ZMQ socket.

            If there is no data available this function will block until there is data or until
            the timeout has elapsed. See also the setReveiveTimeout doc-string.
            You can check for available data with the self.readyReceive() method, or only call
            this function after a sigReadyReceive signal was emitted.

            :param flags: Any valid recv flag (default=0)
            :return: tuple of bytes (binary strings) containing the message parts.
        """

        # Schedule a new call to checkReadyReceive because, during receiving the message there may
        # become new data available which then won't trigger a new socket notification.
        self._sigScheduleCheckReadyReceive.emit(self.zmqSocket.FD)

        # Read the message.
        wasEnabled = self._notifier.isEnabled()
        self._notifier.setEnabled(False)
        try:
            if self.readyReceive(): # see docstring
                return super().receiveMultiMessage()
        finally:
            self._notifier.setEnabled(wasEnabled)


    def sendMultiMessage(self, parts, flags=0):
        """ Sends a message to the ZMQ socket.

            :param part: tuple bytes to be send
            :param flags: Any valid send flag (default=0)
        """

        # Schedule a new call to checkReadyReceive because, during sending of the message there may
        # become new data available which then won't trigger a new socket notification.
        self._sigScheduleCheckReadyReceive.emit(self.zmqSocket.FD)

        wasEnabled = self._notifier.isEnabled()
        self._notifier.setEnabled(False)
        try:
            super().sendMultiMessage(parts, flags=flags)
        finally:
            self._notifier.setEnabled(wasEnabled)


#######################
# Socket type classes #
#######################

# Qt sockets, which will emit a sigMultiMessageReceived signal when a message was read.
# This is only used/necessary when the socket is in another thread.
# Typically not used within the GSSw, only for experiments.

# class QtReqSocket(AbstractQtSignalingSocket, ReqSocketMixin): # not used yet
#     pass
#
# class QtRepSocket(AbstractQtSignalingSocket, RepSocketMixin): # not used yet
#     pass
#
# class QtPubSocket(AbstractQtSignalingSocket, PubSocketMixin): # not used yet
#     pass

class QtSubSocket(AbstractQtSignalingSocket, SubSocketMixin):
    """ A Qt wrapper class for a SUB socket. Will emit a signal on received messages"""
    pass


# Qt sockets, which have a QSocketNotifier that alerts for new input
class NotifyingSubSocket(AbstractQtNotifyingSocket, SubSocketMixin):
    """ A wrapper class for a SUB socket that has a socket notifyer monitoring the socket"""
    pass

class NotifyingRepSocket(AbstractQtNotifyingSocket, RepSocketMixin):
    """ A wrapper class for a REP socket that has a socket notifyer monitoring the socket"""
    pass

