#pylint: disable=invalid-name
"""
Text-based user interface for FDM software.

Use as follows:

  >>> from fdmui import *
  >>> a = inputNumber('Number of events', int, defaultValue=3, minValue=0, maxValue=10)
  >>> b = inputYesNo('Delete everything', defaultValue=True)

"""

# Import Python system modules
import sys      as _sys
import logging  as _logging
import os       as _os
import os.path  as _os_path
import readline as _readline
from gssw.lib.units import SmartUnit


# ##########################################################
#
#  Ask user input.
#

def _formatUnit(value, smartunit):
    """ Formats a value as SmartUnit if a smartunit is available

    :param value:               Value to be formatted
    :param SmartUnit smartunit: SmartUnit which is used for formatting
    """
    if isinstance(smartunit, SmartUnit):
        return smartunit(value)
    return value

def inputNumber(name,
                typ,
                defaultValue=None,
                minValue=None,
                maxValue=None,
                unit=None):
    """Ask user to enter a number.

    :param str name:     Initial part of prompt (range and default value
                         are added automatically)
    :param typ:          float to accept floating point numbers;
                         int to accept integers.
    :param defaultValue: value to return on empty input
    :param minValue:     minimum allowed value, or None if there is no minimum
    :param maxValue:     maximum allowed value, or None if there is no maximum
    :param unit:         Short string to display as unit name.
                         Or SmartUnit to accept value with SI-prefixes.

    :return: numeric value, or returns None if the user pressed Enter and there is no default.
    """

    assert typ in (int, float)
    assert defaultValue is None or isinstance(defaultValue, typ)

    while 1:

        # Prepare prompt.
        rangeStr = ''
        defaultStr = ''
        if minValue is not None and maxValue is not None:
            rangeStr = 'range %s .. %s' % (_formatUnit(minValue, unit), _formatUnit(maxValue, unit))
        elif minValue is not None:
            rangeStr = 'min %s' % _formatUnit(minValue, unit)
        elif maxValue is not None:
            rangeStr = 'max %s' % _formatUnit(maxValue, unit)
        if unit is not None:
            if isinstance(unit, SmartUnit):
                if rangeStr:
                    rangeStr += ', '
                rangeStr += 'input in {}'.format(unit)
            else:
                if rangeStr:
                    rangeStr += ' '
                rangeStr += unit
        if rangeStr:
            rangeStr = ' (' + rangeStr + ')'
        if defaultValue is not None:
            defaultStr = ' [%s]' % _formatUnit(defaultValue, unit)
        prompt = name + rangeStr + defaultStr + ': '

        # Disable TAB-completion.
        _readline.parse_and_bind('tab: tab-insert')

        # Show prompt and read input.
        s = input(prompt)
        s = s.strip()

        # Handle empty string.
        if s == '':
            return defaultValue

        if typ is int:

            # Convert to integer.
            try:
                if isinstance(unit, SmartUnit):
                    v = round(unit.parseInput(s)) # smartunit are always float, so round them
                    v = int(v)
                else:
                    v = int(s)
            except ValueError:
                # Not a decimal integer; try hex (0xabc).
                try:
                    v = int(s, 0)
                except ValueError:
                    # Still not a valid integer.
                    _errmsg('ERROR: Invalid integer format ' + repr(s))
                    # Ask again. The user must learn to accept our wisdom.
                    continue

        elif typ is float:
            try:
                if isinstance(unit, SmartUnit):
                    v = unit.parseInput(s)
                else:
                    # Convert to float.
                    v = float(s)
            except ValueError:
                # Not a floating point number.
                _errmsg('ERROR: Invalid number format ' + repr(s))
                # Keep asking. They will eventually overcome their stupidity.
                continue

        # Check min/max range.
        if minValue is not None and v < minValue:
            _errmsg('ERROR: Value %g too small (min %g)' % (v, minValue))
            continue

        if maxValue is not None and v > maxValue:
            _errmsg('ERROR: Value %g too large (max %g)' % (v, maxValue))
            continue

        # We got an answer.
        return v


def inputYesNo(name, defaultValue=None):
    """Ask user to say Yes or No.

    :param str name:     Initial part of prompt (choices are added automatically)
    :param defaultValue: Value to return on empty input

    :return: True or False, or returns None if the user pressed Enter and there is no default.
    """

    assert defaultValue is None or isinstance(defaultValue, bool)

    while 1:

        # Show prompt
        if defaultValue is not None:
            sy = 'Y' if defaultValue else 'y'
            sn = 'N' if not defaultValue else 'n'
            defaultStr = ' [Y]' if defaultValue else ' [N]'
        else:
            sy = 'y'
            sn = 'n'
            defaultStr = ''

        prompt = name + ' (%s/%s)%s: ' % (sy, sn, defaultStr)

        s = input(prompt)
        s = s.strip()

        # Handle empty string.
        if s == '':
            return defaultValue

        # Convert to boolean.
        if s.lower() == 'y' or s.lower() == 'yes':
            return True
        elif s.lower() == 'n' or s.lower() == 'no':
            return False

        # Will they ever learn?
        _errmsg('ERROR: Invalid input %r, enter Y or N' % s)


def inputChoice(name, items, defaultValue=None):
    """Ask user to choose one item from a list.

    :param str name:     Label of list of choices
    :parma list items:   list [ (value, label), ... ] of menu items; each item
                         specified as a tuple of return-value and display-label.
    :param defaultValue: value to return on empty input
    """

    defaultIndex = None

    # Show choices.
    print('== ' + name + ' ==')
    for (i, (val, label)) in enumerate(items):
        print('  %2d: ' % (i+1), label)
        if val == defaultValue:
            defaultIndex = i

    while 1:

        # Ask for input.
        if defaultIndex is not None:
            prompt = 'Choice [%d]: ' % (defaultIndex+1)
        else:
            prompt = 'Choice: '

        # Read input from user.
        s = input(prompt)
        s = s.strip()

        # Handle empty line.
        if s == '':
            return defaultValue

        # Convert integer.
        v = None
        try:
            v = int(s)
        except ValueError:
            pass

        if v is not None and v >= 1 and v <= len(items):
            # Valid selection.
            (val, label) = items[v-1]
            return val

        # User entered unknown option selector.
        _errmsg('ERROR: Unknown option ' + repr(s))

        # Go back to the top of the loop and ask again for input ...


def inputFileName(name, defaultValue=None, basedir=None,
                  mustExist=False, confirmOverwrite=False,
                  expectDir=False):
    """Ask user to enter a file name.

    :param str name:              Initial part of prompt.
    :param defaultValue:          Value to return if on empty input.
    :param str basedir:           Reference path in case input string is a relative path.
    :param bool mustExist:        True to accept only names of existing files.
    :param bool confirmOverwrite: True to warn user before accepting an existing file.
    :param bool expectDir:        Expect a directory name (False to expect a file name).

    :return: Path entered by user, with relative path resolved against basedir.
    """

    while 1:

        # Show prompt and read input line.
        if defaultValue is not None:
            prompt = name + ' [' + defaultValue + ']: '
        else:
            prompt = name + ': '

        # Temporarily switch to basedir to support TAB completion.
        origWorkDir = _os.getcwd()
        if basedir is not None:
            _os.chdir(basedir)

        try:
            # Enable TAB-completion.
            _readline.parse_and_bind('tab: complete')
            _readline.set_completer_delims(':;\\<>\t\n')

            # Show prompt and read input line.
            s = input(prompt)
            s = s.strip()   

        finally:
            if basedir is not None:
                _os.chdir(origWorkDir)
            _readline.parse_and_bind('tab: tab-insert')

        if s == '':
            # Empty string: use default value.
            if defaultValue is None:
                return defaultValue
            s = defaultValue
        else:
            # Interpret string as path relative to basedir.
            if basedir is not None:
                s = _os_path.join(basedir, s)

        # Check if file exists.
        if not expectDir and _os_path.isfile(s):
            if confirmOverwrite:
                if not inputYesNo('Overwrite existing file %r' % s, False):
                    continue
            return s
        elif expectDir and _os_path.isdir(s):
            return s

        if mustExist:
            if expectDir:
                _errmsg("ERROR: Directory %r does not exist" % s)
            else:
                _errmsg("ERROR: File %r does not exist" % s)
            continue

        # Check that file name is not empty or weird.
        b = _os_path.basename(s)
        if b == '' or b.startswith('.'):
            _errmsg("ERROR: Invalid file name %r" % b)
            continue

        # No weird symbols in file name.
        badsym = None
        for c in s:
            if c < ' ' or c > '~' or c in ':;<>\\':
                badsym = c
        if badsym:
            _errmsg("ERROR: Weird character %r in file name" % badsym)
            continue

        # Check if parent directory exists.
        d = _os_path.dirname(s)
        if d and not _os_path.isdir(d):
            _errmsg("ERROR: Invalid path name %r" % s)
            continue

        return s


# ##########################################################
#
#  Menu system.
#

class MenuOption:
    """Option in a menu."""

    def __init__(self, selector, label, action=None, value=None):
        """Create a menu option.

        :param selector:  Integer or short string to be typed by the user,
                          or None to automatically use the next available number.
        :param str label: String to show in the menu.
        :param action:    Python function to run when this option is selected.
        :param value:     Value to show in square brackets behind menu item.
        """

        self.selector   = selector
        self.label      = label
        self.action     = action
        self.value      = value


class MenuText:
    """Short text to be printed literally between menu options.

    The string should NOT end with "\\n" unless you want to print an empty
    line after the text.
    """

    def __init__(self, msg):

        self.msg = msg


def showMenu(name, items, defaultAction=None, defaultLabel=None, path=None):
    """Show menu and ask for user input.

    :param str name:      Name of this menu.
    :param list items:    List of MenuOption and/or MenuText objects.
    :param defaultAction: Optional value to return in case of empty input.
    :param defaultLabel:  Optional label to use for default action.
    :param list path:     Optional list of names used to show the menu hierarchy.

    :return: The action linked to the selected menu option.
    """

    # Show menu.

    if path is None:
        path = [ ]

    print("")
    print('====  ' + ' > '.join(path + [ name ]) + '  ====')
    print("")

    itemsBySelector = { }
    defaultSelector = None
    for item in items:
        if isinstance(item, MenuOption):
            if item.selector is not None:
                selector = str(item.selector).lower()
                assert selector not in itemsBySelector, \
                                        "The menu contains a duplicate option: {}".format(selector)
                itemsBySelector[selector] = item

    lastSelector = 0
    for item in items:
        if isinstance(item, MenuOption):
            selector = item.selector
            if selector is None:
                lastSelector += 1
                while str(lastSelector) in itemsBySelector:
                    lastSelector += 1
                selector = str(lastSelector)
                itemsBySelector[selector] = item
            if item.action is defaultAction:
                defaultSelector = selector
            if item.value is not None:
                print(' (%2s)  %-40s [%s]' % (selector, item.label, item.value))
            else:
                print(' (%2s) ' % selector, item.label)
        else:
            assert isinstance(item, MenuText)
            print(item.msg)

    print("")

    # Wait for input from the user.

    while 1:

        # Ask for input.
        if defaultLabel is not None:
            prompt = 'Choice or press Enter to '  + defaultLabel + ': '
        elif defaultSelector is not None:
            prompt = 'Choice [%s]: ' % defaultSelector
        else:
            prompt ='Choice: '

        # Read input from user.
        s = input(prompt) #'input' method allows cycling through input history via 'readline' module
        s = s.strip()

        selectedItem = None

        # Handle empty line.
        if s == '':
            return defaultAction

        # Convert integer to canonical form.
        try:
            s = str(int(s))
        except ValueError:
            pass

        # Look up selected option.
        selectedItem = itemsBySelector.get(s.lower())
        if selectedItem is not None:
            # Return action of selected menu item.
            return selectedItem.action

        # User entered unknown option selector.
        _errmsg('ERROR: Unknown option ' + repr(s))

        # Go back to the top of the loop and ask again for input ...


# ##########################################################
#
#  Decorators.
#

def catch_exceptions(function):
    """Decorator which catches exceptions and reports them as a warnings
    without stopping the program.
    """

    def newfn(*args, **kwargs):
        """ The decorated function is wrapped in this function. """
        try:
            return function(*args, **kwargs)
        except Exception as e:
            # Suppress exceptions (but not KeyboardInterrupt)
            _logging.getLogger(__name__).error(e.__class__.__name__ + ': ' + str(e),
                                               exc_info=_sys.exc_info())
            _errmsg('')
            _errmsg('ERROR: ' + e.__class__.__name__ + ': ' + str(e))
            _errmsg('')

    return newfn


# ##########################################################
#
#  Private support functions.
#

def _errmsg(s):
    """Print error message."""

    print(s, file=_sys.stderr)


# ##########################################################
