# -*- coding: utf-8 -*-

""" Module that contains functions that have to do with type checking, importing, etc.

    They follow the reqular python names (with underscores) so they can easily be reused
    in other projects.
"""
import re, logging, importlib

import numpy as np

logger = logging.getLogger(__name__)


def type_name(var):
    """ Returns the name of the type of var"""
    return type(var).__name__


def full_type_name(var):
    """ Returns the full class name of an object. This includes packages and module names.

        It depends on where the class is imported, so be aware of this!
    """
    module = var.__class__.__module__
    if module is None:
        module = "__main__"

    check_class(module, str)

    return "{}.{}".format(module, var.__class__.__name__)


def is_a_type(var, allow_none=False):
    """ Returns True if var is a type (i.e. a class)
    """
    return isinstance(var, type) or (var is None and allow_none)


def is_a_string(var, allow_none=False):
    """ Returns True if var is a string (ascii or unicode)
    """
    return isinstance(var, str) or (var is None and allow_none)


def check_is_a_string(var, allow_none=False):
    """ Calls is_a_string and raises a type error if the check fails.
    """
    if not is_a_string(var, allow_none=allow_none):
        raise TypeError("var must be a string, however type(var) is {}"
                        .format(type(var)))

def is_a_number(var, allow_none=False):
    """ Returns True if var is a number (int or float)
    """
    return isinstance(var, (int, float, np.number, np.ndarray)) or (var is None and allow_none)


def check_is_a_number(var, allow_none=False):
    """ Calls is_a_number and raises a type error if the check fails.
    """
    if not is_a_number(var, allow_none=allow_none):
        raise TypeError("var must be a number, however type(var) is {}"
                        .format(type(var)))

def is_binary(var, allow_none=False):
    """ Returns True if var is a byte string
    """
    return isinstance(var, (bytes, np.bytes_)) or (var is None and allow_none)


def check_is_binary(var, allow_none=False):
    """ Calls is_binary and raises a type error if the check fails.
    """
    if not is_binary(var, allow_none=allow_none):
        raise TypeError("var must be a binary string (bytes), however type(var) is {}"
                        .format(type(var)))


def is_a_sequence(var, allow_none=False):
    """ Returns True if var is a list or a tuple (but not a string!)
    """
    return isinstance(var, (list, tuple)) or (var is None and allow_none)


def check_is_a_sequence(var, allow_none=False):
    """ Calls is_a_sequence and raises a type error if the check fails.
    """
    if not is_a_sequence(var, allow_none=allow_none):
        raise TypeError("var must be a list or tuple, however type(var) is {}"
                        .format(type(var)))


def is_a_dict(var, allow_none=False):
    """ Returns True if var is a dictionary or an ordered dict
    """
    return isinstance(var, dict) or (var is None and allow_none)


def check_is_a_dict(var, allow_none=False):
    """ Calls is_a_dict and raises a type error if the check fails.
    """
    if not is_a_dict(var, allow_none=allow_none):
        raise TypeError("var must be a dict, however type(var) is {}"
                        .format(type(var)))


def is_a_numpy_array(var, allow_none=False):
    """ Returns True if var is a numpy array.
    """
    return isinstance(var, np.ndarray) or (var is None and allow_none)


def check_is_a_numpy_array(var, allow_none=False):
    """ Calls is_a_numpy_array and raises a type error if the check fails.
    """
    if not is_a_numpy_array(var, allow_none=allow_none):
        raise TypeError("var must be a NumPy array, however type(var) is {}"
                        .format(type(var)))


def array_is_structured(array):
    """ Returns True if the array has a structured data type (i.e. has fields)
    """
    return bool(array.dtype.names)


def array_has_real_numbers(array):
    """ Uses the dtype kind of the numpy array to determine if it represents real numbers.

        That is, the array kind should be one of: i u f

        Possible dtype.kind values.
        b     boolean
        i     signed integer
        u     unsigned integer
        f     floating-point
        c     complex floating-point
        O     object
        S     (byte-)string
        U     Unicode
        V     void
    """
    kind = array.dtype.kind
    assert kind in 'biufcOSUV', "Unexpected array kind: {}".format(kind)
    return kind in 'iuf'


def check_class(obj, target_class, allow_none=False):
    """ Checks that the  obj is a (sub)type of target_class.
        Raises a TypeError if this is not the case.

        :param obj: object whos type is to be checked
        :type obj: any type
        :param target_class: target type/class
        :type target_class: any class or type
        :param allow_none: if true obj may be None
        :type allow_none: boolean
    """
    if not isinstance(obj, target_class):
        if not (allow_none and obj is None):
            raise TypeError("obj must be a of type {}, got: {}"
                            .format(target_class, type(obj)))


REGEXP_IDENTIFIER = re.compile(r"\A[_A-Za-z][_A-Za-z0-9]*\Z")
def is_identifier(string):
    """ Returns True if the string can be used as an identifier

        Identifiers contain only letters, digits and underscores. They do not start with a digit.
    """
    return REGEXP_IDENTIFIER.match(string) is not None


#############
# Importing #
#############

def get_symbol_from_module(module_name, symbol_name):
    """ Imports the module and returns a class that is defined in that module.

        For example

            MyClass = getSymbolFromModule('yourpackage.mymodule', 'MyClass')

        is equivalent to

            from yourpackage.mymodule import MyClass

        :param module: string with the module (can include packages separated with dots
        :param symbol_name: the name of any symbol that is defined in the module
        :return: the symbol that is defined in the module
    """
    logger.debug("Importing {} from {}".format(symbol_name, module_name))
    module = importlib.import_module(module_name)
    try:
        symbol = getattr(module, symbol_name)
    except AttributeError as ex:
        raise ImportError("{} not found in module: {}.".format(symbol_name, module)) from ex
    return symbol


#############
# Date-time #
#############




