""" Module for reporting measurement units with automatically determined SI symbol. """

import itertools

__all__ = ['SmartUnit']

class SmartUnit:
    """
    Adds SI-symbols for a specified unit
    """
    SCALES = [ #Lookup table for determining the order of magnitude and corresponding prefix
        (1e+24, 'Y'),
        (1e+21, 'Z'),
        (1e+18, 'E'),
        (1e+15, 'P'),
        (1e+12, 'T'),
        (1e+09, 'G'),
        (1e+06, 'M'),
        (1e+03, 'k'),
        (1e+00, ' '),
        (1e-03, 'm'),
        (1e-06, 'µμu'), # Micro can be entered both as 'µ' (in two variants) and 'u'
        (1e-09, 'n'),
        (1e-12, 'p'),
        (1e-15, 'f'),
        (1e-18, 'a'),
        (1e-21, 'z'),
        (1e-24, 'y'),
        ]

    def __init__(self, unit, baseMagnitude=None):
        """
        :param unit: Base unit string suffix for value=1
        :param baseMagnitude: Magnitude when the value=1
                              E.g. when a variable is always specified in µA, the baseMagnitude can
                              be set to 1e-6, so it will be displayed with the correct SI prefix
        """
        self.unit          = unit
        self.baseMagnitude = 1e+00 if baseMagnitude is None else baseMagnitude

    def searchPrefix(self, value):
        """ Finds the matching factor and symbol for a value """
        value = value * self.baseMagnitude
        mFactor = 1
        mSymbol = ' '
        if value: # Report 0 without prefix
            for factor, symbol in self.SCALES:
                mFactor = factor
                mSymbol = symbol[0] # Do not return the prefix aliases
                if abs(value) >= factor:
                    break
        return mFactor, mSymbol

    @classmethod
    def searchMagnitude(cls, prefix):
        """ Finds the matching magnitude for a given SI-prefix """
        for magnitude, symbols in cls.SCALES:
            if prefix in symbols:
                return magnitude
        return 1.0

    def parseInput(self, rawInput):
        """ Parses a string and returns the value as a float in the correct order of magnitude
        :param rawValue: str, e.g. 123mV
        :return: float, e.g. 0.123
        """
        rawInput = rawInput.strip()

        # Split the number from the unit+prefix, e.g. "100 mV" -> ['100', 'mV']
        rawSplit = [''.join(g) for _, g in itertools.groupby(rawInput, str.isalpha)]
        #print(rawSplit)

        if len(rawSplit) == 2:
            # A value and a unit
            rawValue = rawSplit[0]
            rawPrefixUnit  = rawSplit[1]

            # Parse the number from the raw input
            value = float(rawValue)

            if rawPrefixUnit.lower() == self.unit.lower():
                # Only the unit is specified, e.g. "1V"
                # Hence the order of magnitude is 1
                return value / self.baseMagnitude
            elif rawPrefixUnit[1:].lower() == self.unit.lower():
                # Determine the order of magnitude
                prefix = rawPrefixUnit[0] # Strip the prefix from the unit, e.g. "mV" -> "m"

                try:
                    # Look up the magnitude for the prefix, e.g. "m" -> 0.001
                    magnitude, _ = next(filter(lambda scale: prefix in scale[1], self.SCALES))
                    return value * magnitude / self.baseMagnitude
                except StopIteration:
                    # The input prefix is not valid
                    pass # An error is raised below
        else:
            # Perhaps no units were specified, try parsing directly as float
            return float(rawInput)

        raise ValueError("Invalid input: {}".format(rawInput))


    def __call__(self, value):
        """
        Generates a string in the format '000.000 xX', where 'x' is the SI-prefix, and 'X' the unit.
        For example, SmartUnit('V').__call__(0.123456) yields "123.456 mV"
        :param value:
        :return: str, formatted value
        """
        if value is not None:
            factor, prefix = self.searchPrefix(value)
            return "{:.3f} {}{}".format(value * self.baseMagnitude / factor, prefix, self.unit)

    def __str__(self):
        _, prefix = self.searchPrefix(1)
        return prefix + self.unit

