""" Qt colors and palette functionality.
"""
import enum
import logging
import sys

from PyQt5           import QtCore, QtGui
from PyQt5.QtCore    import Qt
from PyQt5.QtGui     import QColor, QPalette
from gssw.lib.typing import check_class

logger = logging.getLogger(__name__)

QCOLOR_SRON_BLUE = QColor('#115DAA')   # Blue from the SRON logo


def createDarkTestPalette(pal=None):
    """ Create a dark palette. Only used for testing.

        Best to use in combination with the Fusion style of Qt as some styles (Windows, OSX) use
        the native rendering engine.

        https://gist.github.com/QuantumCD/6245215
        https://gist.github.com/lschmierer/443b8e21ad93e2a2d7eb
    """

    pal = QPalette() if pal is None else pal

    pal.setColor(QPalette.Window, QColor(53, 53, 53))
    pal.setColor(QPalette.WindowText, Qt.white)
    pal.setColor(QPalette.Base, QColor(25, 25, 25))
    pal.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
    pal.setColor(QPalette.ToolTipBase, Qt.white)
    pal.setColor(QPalette.ToolTipText, Qt.white)
    pal.setColor(QPalette.Text, Qt.white)
    pal.setColor(QPalette.Button, QColor(53, 53, 53))
    pal.setColor(QPalette.ButtonText, Qt.white)
    pal.setColor(QPalette.BrightText, Qt.red)
    pal.setColor(QPalette.Link, QColor(42, 130, 218))
    pal.setColor(QPalette.Highlight, QColor(42, 130, 218))
    pal.setColor(QPalette.HighlightedText, Qt.black)

    return pal


# Define some Enums that allow use to iterate over the Qt enum values (therefore only unique vals.)

@enum.unique
class PaletteGroupEnum(enum.Enum):
    """ Enumeration for the group in a QPalette.

        # See http://doc.qt.io/archives/qt-5.6/qpalette.html for the full descriptions.
    """
    Active   = QPalette.Active     # The window that has keyboard focus.
    Inactive = QPalette.Inactive   # Other windows.
    Disabled = QPalette.Disabled   # Widgets (not windows) that are disabled for some reason



@enum.unique
class PaletteRoleEnum(enum.Enum):
    """ Enumeration for the role in a QPalette.

        # See http://doc.qt.io/archives/qt-5.6/qpalette.html for the full descriptions.
    """
    # Central roles
    Window        = QPalette.Window         # A general background color.
    WindowText    = QPalette.WindowText     # A general foreground color.
    Base          = QPalette.Base           # Typically the background color for text entry widgets.
    AlternateBase = QPalette.AlternateBase  # Alternate background color when alternating rows
    ToolTipBase   = QPalette.ToolTipBase	# Used as the background color for QToolTip
    ToolTipText   = QPalette.ToolTipText    # Used as the foreground color for QToolTip
    Text          = QPalette.Text           # The foreground color used with Base.
    Button        = QPalette.Button	        # The general button background color.
    ButtonText    = QPalette.ButtonText     # A foreground color used with the Button color.
    BrightText    = QPalette.BrightText     # A text color that is very different from WindowText

    Light         = QPalette.Light          # Lighter than Button color.
    Midlight      = QPalette.Midlight       # Between Button and Light.
    Dark          = QPalette.Dark	        # Darker than Button.
    Mid           = QPalette.Mid            # Between Button and Dark.
    Shadow        = QPalette.Shadow         # A very dark color. By default it's Qt::black.

    Highlight       = QPalette.Highlight        # Indicates a selected item or the current item.
    HighlightedText = QPalette.HighlightedText	# A text color that contrasts with Highlight.
    Link          = QPalette.Link           # A text color used for unvisited hyperlinks.
    LinkVisited   = QPalette.LinkVisited    # A text color used for already visited hyperlinks.

# Bevel and shadowing roles. These are normally derived from Window, and used in ways that
# depend on that relationship.
SHADOWING_ROLES = (PaletteRoleEnum.Light, PaletteRoleEnum.Midlight, PaletteRoleEnum.Dark,
                   PaletteRoleEnum.Mid, PaletteRoleEnum.Shadow)

# Currently only used in PaletteTable
@enum.unique
class ColorFormatEnum(enum.Enum):
    """ How colors are represented as a string.
    """
    RGB_HEX = 0
    ARGB_HEX = 1
    RGB_INT = 2
    RGBA_INT = 3
    HSL_INT = 4
    HSL_FLOAT = 5


def colorToString(color, colorFormat):
    """ Returns a string that describes the color in a certain format.
    """
    if colorFormat == ColorFormatEnum.RGB_HEX:
        return color.name()
    elif colorFormat == ColorFormatEnum.ARGB_HEX:
        return color.name(QColor.HexArgb)
    elif colorFormat == ColorFormatEnum.RGB_INT:
        return "rgb({:3d}, {:3d}, {:3d})".format(*color.getRgb())
    elif colorFormat == ColorFormatEnum.RGBA_INT:
        return "rgba({:3d}, {:3d}, {:3d}, {:3d})".format(*color.getRgb())
    elif colorFormat == ColorFormatEnum.HSL_INT:
        return "hsl({:3d}, {:3d}, {:3d})".format(*color.getHsl())
    elif colorFormat == ColorFormatEnum.HSL_FLOAT:
        return "hsl({:5.2f}, {:5.2f}, {:5.2f})".format(*color.getHslF())
    else:
        raise ValueError("Unexpected colorFormat: {}".format(colorFormat))



def printPalette(pal, colorFormat=ColorFormatEnum.RGB_HEX, file=sys.stdout):
    """ Prints a palette to text
    """
    print("# Role name : Role nr : Active : Inactive : Disabled", file=file)

    for role in PaletteRoleEnum:
        colActive   = pal.color(PaletteGroupEnum.Active.value, role.value)
        colInactive = pal.color(PaletteGroupEnum.Inactive.value, role.value)
        colDisabled = pal.color(PaletteGroupEnum.Disabled.value, role.value)
        line = "{:17} : {:2d} : {} : {} : {}".format(role.name, role.value,
            colorToString(colActive, colorFormat=colorFormat),
            colorToString(colInactive, colorFormat=colorFormat),
            colorToString(colDisabled, colorFormat=colorFormat))
        print(line, file=file)


def parsePalette(file=sys.stdin):
    """ Reads a palette from text.

        If must have been printed with colorFormat=ColorFormatEnum.RGB_HEX!
    """
    pal = QtGui.QPalette()
    for line in file:
        line = line.strip()
        logger.debug("  {}".format(line))
        if line.startswith('#'):
            continue

        parts = line.split(':')
        assert len(parts) == 5, "Expected 5 columns, got {} in: {}".format(len(parts), line)
        _roleName, roleNr, colNameActive, colNameInactive, coleNameDisabled = parts
        roleNr = int(roleNr)
        colorActive = QColor(colNameActive.strip())
        colorInactive = QColor(colNameInactive.strip())
        colorDisabled = QColor(coleNameDisabled.strip())

        pal.setColor(PaletteGroupEnum.Active.value, roleNr, colorActive)
        pal.setColor(PaletteGroupEnum.Inactive.value, roleNr, colorInactive)
        pal.setColor(PaletteGroupEnum.Disabled.value, roleNr, colorDisabled)

    return pal


def getWidgetColor(widget, role, group=None):
    """ Retrieves the foreground color of a widget.

        :param QWidget widget: the Qt Widget
        :param PaletteRoleEnum role: the role (foreground color, back ground color,etc)
        :param PaletteGroupEnum group: the Palette group that is used (None = current)
        :rtype QtGui.QColor:
    """
    check_class(role, PaletteRoleEnum)
    check_class(group, PaletteGroupEnum, allow_none=True)
    if group is None:
        return widget.palette().color(role.value)
    else:
        return widget.palette().color(group.value, role.value)


def setWidgetColor(widget, color, role, group=None):
    """ Sets the background color of a widget.

        Note: if the widget remains transparent use: widget.setAutoFillBackground(True)

        :param QWidget widget: the Qt Widget
        :param QtGui.QColor color: the color to set
        :param PaletteGroupEnum group: the Palette group that is used.(None = current):
        :param PaletteRoleEnum role: the role (foreground color, back ground color,etc)
    """
    check_class(color, (QColor, QtCore.Qt.GlobalColor))
    check_class(role, PaletteRoleEnum)
    check_class(group, PaletteGroupEnum, allow_none=True)

    pal = widget.palette()
    if group is None:
        pal.setColor(role.value, color)
    else:
        pal.setColor(group.value, role.value, color)
    widget.setPalette(pal)


def widgetHasDarkBackGround(widget, threshold=0.6):
    """ Returns True if the widget has a dark backgroun color.

        Convenience function
        :param widget:
        :param int threshold: If background color lightness is lower it is considered a dark.
    """
    bgColor = getWidgetColor(widget, PaletteRoleEnum.Window)
    return bgColor.lightnessF() < threshold
