""" Miscellaneous Qt stuff

"""

import sys, traceback, logging, os.path, inspect, threading

from datetime import datetime
import functools
from PyQt5 import QtWidgets, QtGui, QtCore, QtPrintSupport

from gssw.lib.typing import check_class
from gssw.lib.sockets import destroyAllSocketsInContext

logger = logging.getLogger(__name__)

# Remember directory where files were saved
_LAST_FILE_DIRECTORY = None


if sys.platform == 'linux':
    MONO_FONT_NAME = 'Monospace'
    MONO_FONT_SIZE = 10
elif sys.platform == 'win32' or sys.platform == 'cygwin':
    MONO_FONT_NAME = 'Courier'
    MONO_FONT_SIZE = 10
elif sys.platform == 'darwin':
    MONO_FONT_NAME = 'Courier'
    MONO_FONT_SIZE = 13
else:
    MONO_FONT_NAME = 'Courier'
    MONO_FONT_SIZE = 13


MONO_QFONT = QtGui.QFont(MONO_FONT_NAME, MONO_FONT_SIZE)
MONO_QFONT.setFixedPitch(True)


# This function was copied with permission from
# https://github.com/titusjan/astviewer/blob/v1.1.1/astviewer/misc.py
class ResizeDetailsMessageBox(QtWidgets.QMessageBox):
    """ Message box that enlarges when the 'Show Details' button is clicked.
        Can be used to better view stack traces. I could't find how to make a resizeable message
        box but this it the next best thing.

        Taken from:
        http://stackoverflow.com/questions/2655354/how-to-allow-resizing-of-qmessagebox-in-pyqt4
    """
    def __init__(self, detailsBoxWidth=700, detailBoxHeight=300, *args, **kwargs):
        """ Constructor
            :param detailsBoxWidht: The width of the details text box (default=700)
            :param detailBoxHeight: The heights of the details text box (default=700)
        """
        super(ResizeDetailsMessageBox, self).__init__(*args, **kwargs)
        self.detailsBoxWidth = detailsBoxWidth
        self.detailBoxHeight = detailBoxHeight

        self.setStandardButtons(QtWidgets.QMessageBox.Ok)
        self.setEscapeButton(QtWidgets.QMessageBox.Ok)


    def resizeEvent(self, event):
        """ Resizes the details box if present (i.e. when 'Show Details' button was clicked)
        """
        result = super(ResizeDetailsMessageBox, self).resizeEvent(event)

        details_box = self.findChild(QtWidgets.QTextEdit)
        if details_box is not None:
            #details_box.setFixedSize(details_box.sizeHint())
            details_box.setFixedSize(QtCore.QSize(self.detailsBoxWidth, self.detailBoxHeight))

        return result


# This function was copied with permission from
# https://github.com/titusjan/astviewer/blob/v1.1.1/astviewer/misc.py

def gsswExceptHook(exc_type, exc_value, exc_traceback):
    """ Causes the application to quit in case of an unhandled exception.
        Shows an error dialog and quits the application. There is no 'continue' button. This is by
        design (it would also be very technically challenging, if not impossible, to make this)
        The 'finaly' code-block following a 'try' will not be executed, because
        the exception hook will exit the application. Without this exception hook
        the PyQt except hook will result in the same.
    """

    #Getting the extra trace info (See Bug #3430)
    extraData = ["\n","=== Extra Info ===","\n"]
    for eData in inspect.stack()[1:]: #[1:] => Exclude this (gsswExceptHook) Function
        #Edata contains the following information from inspect.stack():
        # eData[0] = frame=<frame object at 0x7f253c17f248>, 
        # eData[1] = filename='/home/robbertg/egress/python/gssw/lib/widgets/misc.py',
        # eData[2] = lineno=101, 
        # eData[3] = function='gsswExceptHook', 
        # eData[4] = code_context=['    lst = lst + [repr(inspect.stack())]\n'], 
        # eData[5] = index=0), 
        if not "threading.py" in repr(eData[1]): #Exclude traceback from threading module itself
            #Make information readable
            extraData.append("{} ({}) (line {}) \n".format(repr(eData[3]),
                                                           repr(eData[1]),
                                                           repr(eData[2])).replace("'",""))
    extraData += ["\n"]
    
    _app = QtWidgets.QApplication.instance()
    # Constructing a QApplication in case this hasn't been done yet.
    if not _app:
        _app = QtWidgets.QApplication([])

    #Also getting the main Application (like HalServer, SetGetManager etc.)    
    mainAppName = _app.applicationName()
    
    #And get the threadname of which the Error is raised:
    mainThreadName = threading.current_thread().name
    if isinstance(mainThreadName,str): 
        mainThreadName = mainThreadName.replace("'","")
    else:
        mainThreadName = "Unidentified"
        
    traceback.format_exception(exc_type, exc_value, exc_traceback)

    logger.exception("Component: %s, Thread: %s, Bug: uncaught %s : %s ",
                     mainAppName, mainThreadName, exc_type.__name__, exc_value,
                     exc_info=(exc_type, exc_value, exc_traceback))
    
    #msgBox = QtWidgets.QMessageBox()
    msgBox = ResizeDetailsMessageBox()
    msgBox.setText("Bug: uncaught {} \nComponent: {} \nThread: {}".format(
        exc_type.__name__, mainAppName, mainThreadName))
    msgBox.setInformativeText(
        str(exc_value) + "\n\nPlease send the details of this message to the development team.")
    lst = ["Component: {} \nThread: {} \n\n".format(mainAppName, mainThreadName)] \
        + traceback.format_exception(exc_type, exc_value, exc_traceback) \
        + extraData
    msgBox.setDetailedText("".join(lst))
    msgBox.setIcon(QtWidgets.QMessageBox.Critical)
    msgBox.setEscapeButton(QtWidgets.QMessageBox.Ok)
    logger.debug("escapeButton after2: %s",msgBox.escapeButton())

    msgBox.exec_()

    exitCode = 1
    logger.debug("Aborting program with exit code: %s", exitCode)
    destroyAllSocketsInContext()
    sys.exit(exitCode)


###############################################################################
# Decorator for catching exceptions
###############################################################################

def gsswShowError(fn):  #pylint: disable=invalid-name
    """Decorator for catching exceptions.

    If an exception occurs in a function with this decorator,
    the exception is reported in the log and displayed in a messagebox
    for the user. The function then returns None.

    To be used as follows:

    ::

      @pyqtSlot()
      @showError
      def yourSlotFunction(self):
          # Function which might raise error that can be displayed
          pass
    """

    # The functools.wraps decorator ensures that the outside interface
    # of the newfn looks like that of fn (e.g. name, signature, docstring).
    # This is necessary because the pyqtSlot decorator uses the function name
    # to do its magic.

    @functools.wraps(fn)
    def newfn(*args, **kwargs):  #pylint: disable=missing-docstring

        try:
            return fn(*args, **kwargs)
        except Exception as e:
            exname = getattr(type(e), '__name__', None)
            msg = exname + ': ' if exname is not None else ''

            try:
                # Try to emphasize actual error message, which is typically at the end 
                # This is particularly useful when the error message was generated on an RPC server
                # and is displayed in a client.
                exMsg = str(e)
                pos = exMsg.rfind(',')
                assert pos != -1 # trap into silent except below if no comma was found
                msg += "{}<br /><br /><strong>{}</strong>".format(
                    exMsg[:pos + 1], exMsg[pos + 1:].strip())
            except Exception: # Could not emphasize error message, just display exception directly
                msg += str(e)
            logger.exception(msg)
            QtWidgets.QMessageBox.warning(None, 'Exception', msg)
        except KeyboardInterrupt as e:
            logger.exception(str(e))
            QtWidgets.QMessageBox.warning(None, 'KeyboardInterrupt', 'Action canceled via Ctrl-C.')

    return newfn


def gsswUiMessageBox_(string):
    """ Pops an message box, showing the given string"""
    msg = QtWidgets.QMessageBox()
    msg.resize(1600,100)
    msg.setWindowTitle("Error")
    msg.setText(string)
    msg.setStandardButtons(QtWidgets.QMessageBox.Close)
    return msg


def gsswUiYesNoQuestion(string):
    """ Pops a yes/no question. Returns True if the user clicked yes, False otherwise."""
    msg = gsswUiMessageBox_(string)
    msg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
    reply = msg.exec_()
    return reply == QtWidgets.QMessageBox.Yes


def gsswUiError(string):
    """ Pops an error model dialog, showing the given string"""
    msg = gsswUiMessageBox_(string)
    msg.setIcon(QtWidgets.QMessageBox.Critical)
    return msg.exec_()

def gsswUiWarning(string):
    """ Pops a warning model dialog, showing the given string"""
    msg = gsswUiMessageBox_(string)
    msg.setIcon(QtWidgets.QMessageBox.Warning)
    msg.setWindowTitle("Warning")
    return msg.exec_()

def gsswUiInfo(string):
    """ Pops an info model dialog, showing the given string"""
    msg = gsswUiMessageBox_(string)
    msg.setIcon(QtWidgets.QMessageBox.Information)
    msg.setWindowTitle("Info")
    return msg.exec_()

def gsswUiMessage(string):
    """ Pops a message box, with OK/Cancel buttons """
    msg = gsswUiMessageBox_(string)
    msg.setWindowTitle("Careful now")
    msg.setIcon(QtWidgets.QMessageBox.Question)
    msg.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
    msg.setDefaultButton(QtWidgets.QMessageBox.Ok) # Ok button is focused by default
    return msg.exec_()



def promptForFileName(parent, baseName, extension, directory='', 
                      suggestDateTime=True, 
                      fileFilter=''):
    """ Asks the user for a file name to export data to.

        :param QWidget parent: Parent widget of the dialog
        :param str baseName: proposed file name without path and extension
        :param str directory: save-directory suggested to the user. If left empty, the same
            directory will be used as where previous files were saved. Defaults to the user's home
            directory when for the first time.
        :param bool suggestDateTime: if True, the current date-time will be included in
            the suggested file name.
        :param str extension: extension of the file name that will be suggested to the user.
            default 'png'
        :param str fileFilter: file type filter used in the dialog box (see Qt.QFileDialog docs).
            E.g. "Images (*.png *.jpg *.bmp);; All files (*)"
    """
    check_class(parent, QtWidgets.QWidget)
    global _LAST_FILE_DIRECTORY  #pylint: disable=global-statement

    if suggestDateTime:
        now = datetime.now()
        nowStr = now.strftime("_%Y-%m-%d_%H-%M-%S")
    else:
        nowStr = ''

    if not directory:
        if _LAST_FILE_DIRECTORY is None:
            directory = QtCore.QDir.homePath()
        else:
            directory = _LAST_FILE_DIRECTORY

    suggestedFile = os.path.join(directory,
                                 "{}{}.{}".format(baseName, nowStr, extension))

    fileName, _ = QtWidgets.QFileDialog.getSaveFileName(
            parent, "Save as", suggestedFile, fileFilter)
    if fileName:
        # Remember save directory for next time
        _LAST_FILE_DIRECTORY = os.path.dirname(fileName)

    return fileName


def saveWidgetAsImage(parent, widget, directory=None,
                      baseName="plot", suggestDateTime=True, extension="png"):
    """ Asks the user for a file name and saves a screen grab of this widget to file.


        :param QWidget parent: Parent widget of the dialog
        :param QWidget widget: The widget that is saved as image
        :param str baseName: proposed file name without path and extension
        :param str directory: save-directory suggested to the user. If None (default) the users home
            directory will be used. Use '' for the current dir.
        :param bool suggestDateTime: if True, the current date-time will be included in
            the suggested file name.
        :param str extension: extension of the file name that will be suggested to the user.
            default 'png'
    """
    check_class(widget, QtWidgets.QWidget)

    fileName = promptForFileName(parent, baseName, extension, directory, suggestDateTime,
                                 fileFilter="Images (*.png *.jpg *.bmp);; All files (*)")
    if not fileName:
        logger.info("Saving image file canceled")
        return

    logger.info("Saving screen grab to: {}".format(os.path.realpath(fileName)))
    widget.grab().save(fileName)



def printWidgetAsImage(parent, widget, printer):
    """ Asks the user for a file name and saves a screen grab of this widget to file.

        :param QWidget parent: Parent widget of the dialog
        :param QWidget widget: The widget that is saved as image
        :param QtPrintSupport.QPrinter printer: printer to which the snapshot is sent.
    """
    check_class(parent, QtWidgets.QWidget)
    check_class(widget, QtWidgets.QWidget)
    check_class(printer, QtPrintSupport.QPrinter)  #pylint: disable=no-member

    dialog = QtPrintSupport.QPrintDialog(printer, widget)  #pylint: disable=no-member
    dialog.setWindowTitle('Print Plot')

    if not dialog.exec_():
        logger.info("Saving image file canceled")
        return False

    paintPixmapOnPrinter(widget.grab(), printer)



def paintPixmapOnPrinter(pixmap, printer):
    """ Print a pixmap on the top of the page.

        :param QtGui.QPixmap pixmap: pixmap that will be printed
        :param QtPrintSupport.QPrinter printer: printer to which the snapshot is sent.
    """
    check_class(pixmap, QtGui.QPixmap)
    check_class(printer, QtPrintSupport.QPrinter)  #pylint: disable=no-member

    xScale = printer.pageRect().width() / pixmap.width()
    yScale = printer.pageRect().height() / pixmap.height()
    scale = min(xScale, yScale)

    # Draw pixmap with painter
    painter = QtGui.QPainter()
    if not painter.begin(printer):
        raise AssertionError("Unable to begin painting on: {}".format(painter))

    painter.drawPixmap(0, 0,
                       pixmap.width() * scale,
                       pixmap.height() * scale,
                       pixmap)
    painter.end()



def qRectToDict(rect):
    """ Converts a QRect to a dictionary

        :param rect: Qt rectangle object.
    """
    #check_class(rect, (QtCore.QRect, QtCore.QRectF))
    check_class(rect, QtCore.QRect)
    dct = {}
    dct['x'] = rect.x()
    dct['y'] = rect.y()
    dct['width'] = rect.width()
    dct['height'] = rect.height()

    return dct


def qRectFromDict(dct):
    """ Create a QRect from a dictionary

        :param dict dct: dictionary that has x, y, width and height items.
    """
    check_class(dct, dict)
    rect = QtCore.QRect()
    rect.setX(dct['x'])
    rect.setY(dct['y'])
    rect.setWidth(dct['width'])
    rect.setHeight(dct['height'])

    return rect

