""" Plot widgets for use in the GSSw.
"""
from datetime import datetime
import logging

import numpy as np

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt, pyqtSignal, QModelIndex, pyqtSlot

from silx.gui.colors import Colormap
from silx.gui.plot._utils import addMarginsToLimits
from silx.gui.plot.ColorBar import ColorBarWidget
from silx.gui.plot.PlotWidget import PlotWidget
from silx.gui.plot.actions import PlotAction
from silx.gui.plot.actions.control import ResetZoomAction, ZoomBackAction, GridAction
from silx.gui.plot.actions.control import XAxisLogarithmicAction, YAxisLogarithmicAction
from silx.gui.plot.actions.control import XAxisAutoScaleAction, YAxisAutoScaleAction
from silx.gui.plot.items.axis import TickMode
from silx.gui.plot.LegendSelector import LegendIcon as SilxLegendIcon

try:
    from silx.gui.widgets.LegendIconWidget import LegendIconWidget   #pylint: disable=unused-import
    def isEmptyLineStyle(linestyle):
        """" Wrapper function for silx > 0.10.0 """
        return LegendIconWidget.isEmptyLineStyle(linestyle)
    def isEmptySymbol(symbol):
        """" Wrapper function for silx > 0.10.0 """
        return LegendIconWidget.isEmptySymbol(symbol)
# compatibility for silx 0.10.0
except ImportError as ex:
    from silx.gui.plot.LegendSelector import NoLineStyle, NoSymbols  #pylint: disable=unused-import
    def isEmptyLineStyle(linestyle):
        """" Wrapper function for silx ≤ 0.10.0 """
        return linestyle in NoLineStyle
    def isEmptySymbol(symbol):
        """" Wrapper function for silx ≤ 0.10.0 """
        return symbol in NoLineStyle

try:
    from silx.gui.printer import getDefaultPrinter
except ImportError as ex:
    from PyQt5.Qt import QPrinter  #pylint: disable=ungrouped-imports
    def getDefaultPrinter():
        """ Function is only available as of silx 0.8, so mimic it here. """
        return QPrinter()

from gssw.lib.widgets.colors import setWidgetColor, PaletteRoleEnum
from gssw.lib.widgets.misc import saveWidgetAsImage, printWidgetAsImage
from gssw.lib.typing import check_is_a_sequence, check_class

__all__ = ["GsswPlotWidget", "GsswImageWidget", "BACKEND_MPL", "BACKEND_GL", "DEFAULT_BACKEND"]

logger = logging.getLogger(__name__)

BACKEND_MPL = "mpl"  # Matplotlib
BACKEND_GL = "gl"    # OpenGl

DEFAULT_BACKEND = BACKEND_GL

# Acording to http://www.silx.org/doc/silx/0.8.0/modules/gui/plot/getting_started.html#colormap
# the following color maps are always available in silx
# To choose a good colormap see https://matplotlib.org/users/colormaps.html
# To see which are actually available use Colormap.getSupportedColormaps()
COLOR_MAP_NAMES = [
    'gray', 'reversed gray', 'temperature',
    'red', 'green', 'blue',
    'viridis', 'magma', 'inferno', 'plasma'
]


class GsswLegendIcon(SilxLegendIcon):
    """Object displaying a curve linestyle and symbol.

        Inherits from silx.gui.plot.LegendSelector.LegendIcon and overrides the sizehint to make it
        less wide.
    """
    def __init__(self, *args, **kwargs):
        """ Initialization """
        super().__init__(*args, **kwargs)
        self.showSymbol = None
        self.showLine = None
        self.symbolOutlineBrush = None

    def sizeHint(self):
        """ Holds the recommended size for of legend icon
        """
        return QtCore.QSize(24, 16)



class LegendListItemWidget(QtWidgets.QItemDelegate):
    """ Object displaying a single item (i.e., a row) in the list.

        Forked from silx.gui.plot.LegendSelector.LegendListItemWidget.

        Note: LegendListItem does NOT inherit from QObject, it cannot emit signals!
    """
    def __init__(self, parent=None):

        super().__init__(parent)

        # Dictionary to render checkboxes
        self.labelDict = {}
        self.iconDict = {}

        # Keep legend to get sizeHint
        self.legend = QtWidgets.QLabel()
        self.icon = GsswLegendIcon()


    def paint(self, painter, option, modelIndex):
        """ Override the paint event to draw the legend item.

            :param QPainter painter:
            :param QStyleOptionViewItem option:
            :param QModelIndex modelIndex:
        """
        # Code taken from silx. Won't fix too many locals
        # pylint: disable=too-many-locals
        painter.save()
        rect = option.rect

        # Calculate the icon rectangle
        iconSize = self.icon.sizeHint()

        # Calculate icon position
        x = rect.left() + 2
        y = rect.top() + int(.5 * (rect.height() - iconSize.height()))
        iconRect = QtCore.QRect(QtCore.QPoint(x, y), iconSize)

        # Calculate label rectangle
        legendSize = QtCore.QSize(rect.width() - iconSize.width() - 30,
                              rect.height())
        # Calculate label position
        x = rect.left() + iconRect.width()
        y = rect.top()
        labelRect = QtCore.QRect(QtCore.QPoint(x, y), legendSize)
        labelRect.translate(QtCore.QPoint(10, 0))

        # Remember the rectangles
        idx = modelIndex.row()
        self.iconDict[idx] = iconRect
        self.labelDict[idx] = labelRect

        # Draw background first! Don't use model's BackgroundRole, it has alternate colors.
        backgroundBrush = QtGui.QBrush(Qt.white)
        painter.fillRect(rect, backgroundBrush)

        # Draw label
        label = modelIndex.data(GsswLegendModel.labelRole)
        legendText = modelIndex.data(Qt.DisplayRole) if label is None else label
        #textBrush = modelIndex.data(Qt.ForegroundRole)
        textBrush = QtGui.QBrush(Qt.black)
        textAlign = modelIndex.data(Qt.TextAlignmentRole)
        painter.setBrush(textBrush)
        painter.setFont(self.legend.font())
        painter.setPen(textBrush.color())
        painter.drawText(labelRect, textAlign, legendText)

        # Draw icon
        iconColor = modelIndex.data(GsswLegendModel.iconColorRole)
        iconLineWidth = modelIndex.data(GsswLegendModel.iconLineWidthRole)
        iconLineStyle = modelIndex.data(GsswLegendModel.iconLineStyleRole)
        iconSymbol = modelIndex.data(GsswLegendModel.iconSymbolRole)
        icon = GsswLegendIcon()
        icon.resize(iconRect.size())
        icon.move(iconRect.topRight())
        icon.showSymbol = modelIndex.data(GsswLegendModel.showSymbolRole)
        icon.showLine = modelIndex.data(GsswLegendModel.showLineRole)
        icon.setSymbolColor(iconColor)
        icon.setLineColor(iconColor)
        icon.setLineWidth(iconLineWidth)
        icon.setLineStyle(iconLineStyle)
        icon.setSymbol(iconSymbol)
        icon.symbolOutlineBrush = backgroundBrush
        icon.paint(painter, iconRect, option.palette)

        painter.restore()
        # pylint: enable=too-many-locals


    def sizeHint(self, _option, _idx):
        """ Holds the recommended size for the widget
        """
        # return qt.QSize(68,24)
        iconSize = self.icon.sizeHint()
        legendSize = self.legend.sizeHint()
        height = max([iconSize.height(), legendSize.height()]) + 4
        width = iconSize.width() + legendSize.width()
        return QtCore.QSize(width, height)


class GsswLegendModel(QtGui.QStandardItemModel):
    """ Model that stores the legend information. This info is used by the LegendListItemWidget.

        This model is used instead of the silx LegendModel class because we needed to add an extra
        role (the labelRole) to make it possible to add the same label twice in a legend.
    """
    iconColorRole = Qt.UserRole + 0
    iconLineWidthRole = Qt.UserRole + 1
    iconLineStyleRole = Qt.UserRole + 2
    showLineRole = Qt.UserRole + 3
    iconSymbolRole = Qt.UserRole + 4
    showSymbolRole = Qt.UserRole + 5

    labelRole = Qt.UserRole + 1000 # override label in legend.

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.setColumnCount(1)
        self._palette = QtGui.QPalette()


    def insertLegendList(self, row, llist):
        """ Fills the legend from a list of dictionaries.

            Adapted from silx.gui.plot.LegendSelector.insertLegendList.

            :param int row: Determines after which row the items are inserted
            :param list llist: Carries the new legend information.
                see LegendWidget.setLegendList for more details.
        """
        for (legend, dct) in llist:

            linestyle = dct.get('linestyle', None)
            symbol = dct.get('symbol', None)

            if isEmptyLineStyle(linestyle):
                # Curve had no line, give it one and hide it
                # So when toggle line, it will display a solid line
                showLine = False
                linestyle = '-'
            else:
                showLine = True

            if isEmptySymbol(symbol):
                # Curve had no symbol, give it one and hide it
                # So when toggle symbol, it will display 'o'
                showSymbol = False
                symbol = 'o'
            else:
                showSymbol = True

            brush = self._palette.brush(QtGui.QPalette.Normal, QtGui.QPalette.WindowText)
            alignment = (Qt.AlignVCenter | Qt.AlignLeft)

            item = QtGui.QStandardItem()
            item.setData(legend, role=Qt.DisplayRole)
            item.setData(brush, role=Qt.ForegroundRole)
            item.setData(alignment, role=Qt.TextAlignmentRole)

            item.setData(dct['color'], role=GsswLegendModel.iconColorRole)
            item.setData(dct['linewidth'], role=GsswLegendModel.iconLineWidthRole)
            item.setData(linestyle, role=GsswLegendModel.iconLineStyleRole)
            item.setData(showLine, role=GsswLegendModel.showLineRole)
            item.setData(symbol, role=GsswLegendModel.iconSymbolRole)
            item.setData(showSymbol, role=GsswLegendModel.showSymbolRole)

            item.setData(dct.get('label'), role=GsswLegendModel.labelRole)

            self.insertRow(row, item)
        return True



class LegendWidget(QtWidgets.QListView):
    """ Widget displaying a list of curve legends, line style and symbol.

        Forked from silx.gui.plot.LegendSelector.LegendListView.

        The GsswPlotWidget includes a LegendWidget that is updated automatically.
    """

    sigItemClicked = pyqtSignal(str) # String is the reference label
    sigItemHover = pyqtSignal(str, QtCore.QPoint)

    @pyqtSlot(QModelIndex)
    def processClick(self, index):
        """ Get the identifier name of clicked curve, and re-emit """ 
        item = self.model().itemFromIndex(index)
        strRef = item.data(Qt.DisplayRole)
        self.sigItemClicked.emit(strRef)

    @pyqtSlot(QModelIndex)
    def processHover(self, index):
        """ Get the identifier name and position of clicked curve and re-emit """
        item = self.model().itemFromIndex(index)
        strRef = item.data(Qt.DisplayRole)
        position = self.visualRect(index).center()
        position.setY(self.visualRect(index).top())
        globalPosition = self.mapToGlobal(position)
        self.sigItemHover.emit(strRef, globalPosition)

    def __init__(self, parent=None, model=None):
        super().__init__(parent)

        self.setCursor(Qt.ArrowCursor)
        self.setItemDelegate(LegendListItemWidget())
        self.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        self.setMouseTracking(True)

        if model is None:
            #model = LegendModel(parent=self)
            model = GsswLegendModel(parent=self)
        self.setModel(model)
        self.clicked.connect(self.processClick)
        self.entered.connect(self.processHover)


    def setLegendList(self, legendList):
        """ Sets the legend contents from a list of dictionaries

            :param list legendList: a list with (curveName, dict) tuples where
                every dictionary contains information on how to plot the curve.

                E.g. legendList = [
                    ('first-curve', {
                        'color': PyQt5.QtGui.QColor.red,
                        'symbol': 'o',
                        'linewidth': 2.3,
                        'active': False,
                        'selected': False,
                        'linestyle': '-'
                    }),
                    ('second-curve', { ... }),
                ]
        """
        self.clear()
        model = self.model()
        model.insertLegendList(0, legendList)


    def clear(self):
        """ Clears the legend
        """
        model = self.model()
        model.removeRows(0, model.rowCount())



# Define actions as separate classes, in line with the silx style
class SeparatorAction(QtWidgets.QAction):
    """A QAction that serves as a separator in a context menu
    """
    def __init__(self, parent):  # using **kwargs, somehow it doesn't work on OSX otherwise
        #logger.debug("QAction : {} {}".format(parent))
        super().__init__(parent)
        self.setSeparator(True)


class ClearPlotAction(PlotAction):
    """A QAction that can be added to PlotWidget toolbar to clear the plot"""

    def __init__(self, plot, parent=None):
        super().__init__(
            plot, icon='profile1D', text='Clear plot',
            tooltip='Clear the plotted data in this panel',
            triggered=self._clear,
            checkable=False, parent=parent)

    def _clear(self):
        """Handle action triggered and clear the plot"""
        self.plot.clear()

#
# class ResetMarginZoomAction(ResetZoomAction):
#     """ Reset zoom with margins"""
#
#     def __init__(self, plot, dataMargins=None, parent=None):
#         """ Constructor """
#         super().__init__(plot, parent=parent)
#         self.dataMargins = dataMargins
#
#     def _actionTriggered(self, checked=False):
#         """ Resets the plot with the data margins"""
#         self.plot.resetZoom(dataMargins=self.plot.getZoomDataMargins())


class SavePlotAction(PlotAction):
    """A QAction that can be added to PlotWidget toolbar to save the plot"""

    def __init__(self, plot, parent=None):
        super().__init__(
            plot, icon='document-save', text='Save as image...',
            tooltip='Save curve/image/plot snapshot dialog',
            triggered=self._saveAsImage,
            checkable=False, parent=parent)

        # self.setShortcut(QtGui.QKeySequence.Save) # doesn't seem to work
        # self.setShortcutContext(Qt.WidgetShortcut)

    def _saveAsImage(self):
        """Handle triggered action. """

        self.plot.askNameAndSaveAsImage(
            baseName="plot" if not self.plot.getGraphTitle() else self.plot.getGraphTitle())


class PrintPlotAction(PlotAction):
    """A QAction that can be added to PlotWidget toolbar to print the plot"""

    def __init__(self, plot, parent=None):
        super().__init__(
            plot, icon='document-print', text='Print...',
            tooltip='Print curve/image/plot snapshot dialog',
            triggered=self._printAsImage,
            checkable=False, parent=parent)

        # self.setShortcut(QtGui.QKeySequence.Print)  doesn't seem to work
        # self.setShortcutContext(Qt.WidgetShortcut)

    def _printAsImage(self):
        """Handle triggered action. """
        self.plot.askNameAndPrintAsImage()



class GsswSilxPlotWidget(PlotWidget):
    """ Derives from a silx PlotWidget,
        Sets default values for a consistent style. Does NOT have a legend, for that use
        the GsswPlotWidget class.
    """
    def __init__(self, *args, **kwargs):
        if 'backend' not in kwargs:
            logger.debug("Setting backend to: {}".format(DEFAULT_BACKEND))
            kwargs['backend'] = DEFAULT_BACKEND

        super().__init__(*args, **kwargs)

        self.showPositionTooltip = True # If True a tooltip shows the cursor position in data coords

        # Configure default look and feel
        self.setInteractiveMode(mode='zoom',      # zoom in on an area by drawing a box around it
                                zoomOnWheel=True  # allow zooming with mouse wheel
                                )
        self.setGraphGrid(which='both')           # Enable drawing a grid

        # The list of actions in the context menu.
        # Users can modify it via the corresponding setter
        self._contextMenuActions = [
            ZoomBackAction(plot=self, parent=self),
            ResetZoomAction(plot=self, parent=self),
            SeparatorAction(parent=self),
            XAxisAutoScaleAction(plot=self, parent=self),
            YAxisAutoScaleAction(plot=self, parent=self),
            XAxisLogarithmicAction(plot=self, parent=self),
            YAxisLogarithmicAction(plot=self, parent=self),
            GridAction(plot=self, parent=self),
            SeparatorAction(parent=self),
            SavePlotAction(plot=self, parent=self),
            PrintPlotAction(plot=self, parent=self),
            SeparatorAction(parent=self),
            ClearPlotAction(plot=self, parent=self),
        ]

        # Retrieve PlotWidget's plot area widget
        plotArea = self.getWidgetHandle()

        # Set plot area custom context menu
        plotArea.setContextMenuPolicy(Qt.CustomContextMenu)
        plotArea.customContextMenuRequested.connect(self._showContextMenu)


    @property
    def contextMenuActions(self):
        """ The list of actions for in the context menu. Can be safely modified.
        """
        return self._contextMenuActions


    @contextMenuActions.setter
    def contextMenuActions(self, value):
        """ The list of actions for in the context menu. Can be safely modified.
        """
        check_is_a_sequence(value)
        self._contextMenuActions = value


    def _showContextMenu(self, pos):
        """ Shows the context menu at position pos.
        """
        contextMenu = QtWidgets.QMenu()

        for action in self._contextMenuActions:
            contextMenu.addAction(action)

        contextMenu.exec_(self.mapToGlobal(pos))


    def onMouseMove(self, xPixel, yPixel):
        """ Overrides tne onMouse move to show a tooltip
        """
        if self.showPositionTooltip:
            dataPos = self.pixelToData(xPixel, yPixel, check=True)
            globalMousePos = self.mapToGlobal(QtCore.QPoint(xPixel, yPixel))

            if dataPos is None:
                toolTipText = ""
            else:
                xData, yData = dataPos
                if self.getXAxis().getTickMode()==TickMode.TIME_SERIES:
                    # Display timestamps as human-readable time
                    xData = str(datetime.fromtimestamp(xData))
                toolTipText = "X = {}\nY = {!r}".format(xData, yData)

            # Shows Tooltip immediately. See https://stackoverflow.com/a/13720606/625350
            QtWidgets.QToolTip.showText(globalMousePos, toolTipText)

        super().onMouseMove(xPixel, yPixel)


    def isZoomed(self):
        """ Checks whether the plot is zoomed or panned in (by the user).

            Used to determine whether to autoscale in case of appending data.
        """
        # Add margin to data limits and see if this is equal to the view limits
        xView = self.getGraphXLimits()
        yView = self.getGraphYLimits()

        if xView is None or yView is None:
            return False
        else:
            xViewMin, xViewMax = xView
            yViewMin, yViewMax = yView

        # Code copied from Silx PlotPanel._forceResetZoom
        ranges = self.getDataRange()
        xmin, xmax = (1., 100.) if ranges.x is None else ranges.x
        ymin, ymax = (1., 100.) if ranges.y is None else ranges.y

        # Add margin to data using the same routine as Silx
        xDataMin, xDataMax, yDataMin, yDataMax = addMarginsToLimits(
            self.getDataMargins(),
            self._xAxis._isLogarithmic(),  #pylint: disable=protected-access
            self._yAxis._isLogarithmic(),  #pylint: disable=protected-access
            xmin, xmax, ymin, ymax)

        # In case of identical values data max and min are identical, but not the same as the
        # view limits, yet the user is not zoomed in
        xZoomed = (xViewMin != xDataMin or xDataMax != xViewMax) and not xDataMin == xDataMax
        yZoomed = (yViewMin != yDataMin or yDataMax != yViewMax) and not yDataMin == yDataMax

        return xZoomed or yZoomed


    def getXAxisLogarithmic(self):
        """ Returns True if the X Axis is logarithmic. """
        axis = self.getXAxis()
        return axis.getScale() == axis.LOGARITHMIC


    def setXAxisLogarithmic(self, flag):
        """ Set the X axis to logarithmic or linear display. """
        axis = self.getXAxis()
        axis.setScale(axis.LOGARITHMIC if flag else axis.LINEAR)


    def getYAxisLogarithmic(self):
        """ Returns True if the X Axis is logarithmic. """
        axis = self.getYAxis()
        return axis.getScale() == axis.LOGARITHMIC


    def setYAxisLogarithmic(self, flag):
        """ Set the Y axis to logarithmic or linear display. """
        axis = self.getYAxis()
        axis.setScale(axis.LOGARITHMIC if flag else axis.LINEAR)


    def askNameAndSaveAsImage(self, baseName="plot"):
        """ Asks the user for a name by means of a dialog, and exports the plot as an image.

            :param str baseName: suggested base name
        """
        saveWidgetAsImage(self, self, baseName=baseName)


    def askNameAndPrintAsImage(self):
        """ Shows the Print dialog and lets the user print the plot.
        """
        printWidgetAsImage(self, self, getDefaultPrinter())


    ### Overloaded functions from PlotWidget ###

    # We use *args and **kwargs as to not have to maintain a long argument list.
    # Pylint should not give an error on this. Is probably a pylint bug anyway.
    # See https://github.com/PyCQA/pylint/issues/1553
    # pylint: disable=arguments-differ
    def addCurve(self, x, y, *args, **kwargs):
        """ Add a 1D curve given by x an y to the graph.

            Replaces infinite values with NaNs (otherwise silx gives an error
            when calculating the plot ranges). After that the addCurve of the
            super class is called.
        """
        # First convert to array. The silx addCurve docs say x and y should be
        # arrays so this can be done here.
        x = np.array(x)
        y = np.array(y)

        if x.dtype.kind == 'f':
            logger.debug("Replacing infinite values in x by NaNs before plot")
            x[np.isinf(x)] = np.nan

        if y.dtype.kind == 'f':
            logger.debug("Replacing infinite values in y by NaNs before plot")
            y[np.isinf(y)] = np.nan

        #logger.debug("addCurve (kind={}): y: {}".format(y.dtype.kind, y))

        super().addCurve(x, y, *args, **kwargs)

    # pylint: enable=arguments-differ


class AbstractGsswSilxPlotWidget(QtWidgets.QWidget):
    """ Combines a GsswSilxPlotWidget with a legend or colorbar in a single widget.

        Contains a splitter, and on the left side it contains a GsswSilxPlotWidget. 
        On the right side it can contain a legend or color bar. This is to be determined 
        by the descendant.
    """
    IDX_PLOT = 0
    IDX_LEGEND = 1

    def __init__(self, parent=None, *args, **kwargs):
        """ Constructor.

            :param parent: Qt parent widget
            :param args: will be forwarded to the underlying plot widget
            :param kwargs: will be forwarded to the underlying plot widget
        """
        super().__init__(parent=parent)

        # Layout with legend
        self.contentsLayout = QtWidgets.QHBoxLayout()
        self.contentsLayout.setSpacing(0)
        self.contentsLayout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.contentsLayout)

        self.plot = GsswSilxPlotWidget(*args, **kwargs)

        # By default silx sets 'X' and 'Y' as labels. Override this.
        self.plot.getXAxis().setLabel('')
        self.plot.getYAxis().setLabel('')

        self.splitter = QtWidgets.QSplitter()
        self.contentsLayout.addWidget(self.splitter)
        self.splitter.addWidget(self.plot)

        # Replace SavePlotAction by Action that also saves the legend.
        # Idem for PrintAction. Assumes no duplicates Save or Print actions.
        actions = self.plot.contextMenuActions
        for idx, oldAction in enumerate(actions):
            if isinstance(oldAction, SavePlotAction):
                actions[idx] = SavePlotAction(plot=self, parent=self)
            elif isinstance(oldAction, PrintPlotAction):
                actions[idx] = PrintPlotAction(plot=self, parent=self)


    def paintEvent(self, _event):
        """ Reimplementation of paintEvent to allow for style sheets
            See: http://qt-project.org/wiki/How_to_Change_the_Background_Color_of_QWidget
        """
        opt = QtWidgets.QStyleOption()
        opt.initFrom(self)
        painter = QtGui.QPainter(self)
        self.style().drawPrimitive(QtWidgets.QStyle.PE_Widget, opt, painter, self)
        painter.end()


    # Approach taken from PyQtGraph. I don't like hacks like this because of unexpected
    # complications, but the alternative is for users to have to use plot.plot everywhere.
    # This hack fails, for instance, if you inherit from GsswPlotWidget and try to call
    # super().myMethod. This is because super() does not return a regular object (but some
    # wrapper) and therefore __getattr__ is not called. It also doesn't work with setter
    # properties.
    def __getattr__(self, attr):
        """ Forwards all other public methods and properties to in the underlying silx plot.
        """
        # logger.debug("Looking for attr: {}".format(attr))
        if hasattr(self.plot, attr):
            method = getattr(type(self.plot), attr)
            methodResult = getattr(self.plot, attr)
            # logger.debug("  Found for attr: {}, is callable: {}, is property:{}"
            #              .format(attr, callable(method), isinstance(method, property)))
            if (callable(method) or isinstance(method, property)) and not attr.startswith("__"):
                return methodResult

        raise AttributeError(attr)


    def splitterSizes(self):
        """ Returns the sizes of the plot and the legend in the splitter.

            :returns: [plotSize, legendSize]
        """
        plotSize, legendSize = self.splitter.sizes()
        return [plotSize, legendSize]


    def askNameAndSaveAsImage(self, baseName="plot"):
        """ Asks the user for a name by means of a dialog, and exports the plot as an image.
        """
        # Temporary set the splitter width to zero so that it doesn't show in the exported image.
        oldWidth = self.splitter.handleWidth()
        self.splitter.setHandleWidth(0)
        try:
            saveWidgetAsImage(self, self, baseName=baseName)
        finally:
            self.splitter.setHandleWidth(oldWidth)


    def askNameAndPrintAsImage(self):
        """ Shows the Print dialog and lets the user print the plot
        """
        # Temporary set the splitter width to zero so that it doesn't show in the exported image.
        oldWidth = self.splitter.handleWidth()
        self.splitter.setHandleWidth(0)
        try:
            printWidgetAsImage(self, self, getDefaultPrinter())
        finally:
            self.splitter.setHandleWidth(oldWidth)



class GsswPlotWidget(AbstractGsswSilxPlotWidget):
    """ Combines a GsswSilxPlotWidget with a legend in a single widget.

        The silx library has some legend functionality but this is placed in
        an external dock widget. We believe the legend should be an integral
        part of a plot. For example, when making screen shots for in reports a
        legend should not look like an external widget.

        This class wraps the GsswSilxPlotWidget so that any methods that are
        not part of the GsswPlotWidget are forwarded to its GsswSilxPlotWidget.
        The GsswSilxPlotWidget inherits from the silx PlotWidget, which is
        documented here:
        http://www.silx.org/doc/silx/latest/modules/gui/plot/plotwidget.html#plotwidget
    """
    IDX_PLOT = 0
    IDX_LEGEND = 1

    # Initial widths when setting the splitter position. These are target value, the actual
    # values depend also on the total plot width. See QSplitter.setSizes() documentation.
    SPLITTER_PLOT_WIDTH = 600
    SPLITTER_LEGEND_WIDTH = 150

    def __init__(self, parent=None, *args, **kwargs):
        """ Constructor.

            :param parent: Qt parent widget
            :param args: will be forwarded to the underlying plot widget
            :param kwargs: will be forwarded to the underlying plot widget
        """
        super().__init__(parent=parent, *args, **kwargs)

        self.legend = LegendWidget()
        self._labels = {} # legendId -> label dictionary

        # We set the background color to white here. Otherwise the legend and
        # the plot will have different background colors, which is visible if
        # the legends isn't completely filled (i.e. almost always). Also using
        # white makes it easy to print. Also disabled border.
        self.legend.setStyleSheet('border: 0px; background: white')
        #self.setStyleSheet('background: magenta') # For testing

        self.splitter.addWidget(self.legend)
        self.splitter.setSizes([self.SPLITTER_PLOT_WIDTH, self.SPLITTER_LEGEND_WIDTH])

        self.plot.sigContentChanged.connect(self.updateLegend)
        self.legend.sigItemClicked.connect(self.toggleCurveVisible)
        self.legend.sigItemHover.connect(self.displayTooltip)

    @pyqtSlot(str)
    def toggleCurveVisible(self, refLabel):
        """ Toggles whether a curve is visible.

        :param int index: Row index of the item.
        """
        logger.info("Toggling item visibility: %s", refLabel)
        item = self.plot.getCurve(refLabel)

        if item is not None:
            item.setVisible(not item.isVisible())

    @pyqtSlot(str, QtCore.QPoint)
    def displayTooltip(self, refLabel, position):
        """ Display a tooltip with the name and label of a legend item.

        :param str refLabel: The label of the item in the legend.
        :param QtCore.Qpoint position: The position of the item in the legend.
        """
        toolTipText = "{} : {}".format(self._labels.get(refLabel), refLabel)
        QtWidgets.QToolTip.showText(position, toolTipText)

    def addCurve(self, x, y, *args, label=None, **kwargs):
        """ Add a 1D curve given by x an y to the graph.

            Calls the underlying GsswSilxPlotWidget.addCurve but accepts an extra label parameter.

            :param str label: if not None this string is used as label in the legend. If None the
                value of the legend parameter is used. This allows the plot to have multiple curves
                with the same label in the legend (the legend parameter identifies the curve and
                must be unique in the plot).
        """
        legendId = kwargs.get('legend')
        if label is not None and legendId:
            self._labels[legendId] = label
        self.plot.addCurve(x, y, *args, **kwargs)


    def updateLegend(self):
        """ Syncs the legend with the info of the plot.
        """
        # Based on silx LegendsDockWidget.updateLegends
        legendList = []
        for curve in self.plot.getAllCurves(withhidden=True):
            legend = curve.getLegend()
            # Use active color if curve is active
            if legend == self.plot.getActiveCurve(just_legend=True):
                color = QtGui.QColor(self.plot.getActiveCurveColor())
                isActive = True
            else:
                color = QtGui.QColor.fromRgbF(*curve.getColor())
                isActive = False

            curveInfo = {
                'label': self._labels.get(legend),
                'color': color,
                'linewidth': curve.getLineWidth(),
                'linestyle': curve.getLineStyle(),
                'symbol': curve.getSymbol(),
                'selected': not self.plot.isCurveHidden(legend),
                'active': isActive}
            legendList.append((legend, curveInfo))

        self.legend.setLegendList(legendList)

    # Doesn't work yet
    # def legendIsVisible(self):
    #     """ Returns True if the legend is visible, False if it's hidden.
    #     """
    #     return self.legend.isVisible()
    #
    #
    # def setLegendVisible(self, boolean):
    #     """ Shows or hides the legend.
    #     """
    #     self.legend.setVisible(boolean)



class GsswImageWidget(AbstractGsswSilxPlotWidget):
    """ Combines a GsswSilxPlotWidget with a colorbar in a single widget.

        Simply a convenience class, useful to plot a single image with a color bar.
    """
    def __init__(self, parent=None, colormap=None, *args, **kwargs):
        """ Constructor.

            :param parent: Qt parent widget
            :param ColorMap None colormap: color map that is used.
            :param args: will be forwarded to the underlying plot widget
            :param kwargs: will be forwarded to the underlying plot widget
        """
        super().__init__(parent=parent)

        check_class(colormap, Colormap, allow_none=True)

        self._barTitle = ''
        self.plot.setGraphGrid(False)

        self.colormap = Colormap() if colormap is None else colormap

        self.colorbar = ColorBarWidget(plot=self.plot)
        self.colorbar.setAutoFillBackground(True)
        setWidgetColor(self.colorbar, Qt.white, PaletteRoleEnum.Window)
        self.splitter.addWidget(self.colorbar)


    def addImage(self, imageData, *args, **kwargs):
        """ Calls add image of the underlying plog
        """
        colormap = kwargs.get('colormap', self.colormap)
        return self.plot.addImage(imageData, colormap=colormap, *args, **kwargs)


    def setBarTitle(self, title):
        """ Sets the label (title) in the colorbar
        """
        self._barTitle = title
        self.colorbar.setLegend(title)


    def getBarTitle(self):
        """ Returns the label (title) in the colorbar
        """
        # Using a stored title because self.colorbar.getLegend() has a bug in silx < 0.8.0
        # See https://github.com/silx-kit/silx/pull/2072
        return self._barTitle




# This is temporary test code. It will be removed in time.
#pylint: disable=using-constant-test,ungrouped-imports
if __name__ == "__main__":

    LOG_FMT = '%(asctime)s %(filename)25s:%(lineno)-4d : %(levelname)-7s: %(message)s'
    logging.basicConfig(level='INFO', format=LOG_FMT)
    logger.setLevel('DEBUG')

    from PyQt5.QtWidgets import QApplication 

    if 0:
        BACKEND = 'mpl'
    else:
        BACKEND = 'gl'

    def makePlot1D():
        """ test function """

        # Just some random numbers
        if 0:
            y = np.random.normal(size=int(1.2e4))
            y[9999] = 23
            x = np.arange(len(y)) + 9000
        else:
            y = np.random.normal(size=int(20))
            y[9] = 23
            x = np.arange(len(y)) + 9000

        plot1D = GsswPlotWidget(backend=BACKEND)
        margin = 0.05
        plot1D.setDataMargins(xMinMargin=margin, xMaxMargin=margin,
                              yMinMargin=margin, yMaxMargin=margin)
        xAxis = plot1D.getXAxis()
        xAxis.setLimits(-11, 35)

        plot1D.addCurve(x=x, y=y, legend='blauw', linewidth=4, linestyle='-.')
        plot1D.addCurve(x=x, y=y+5, legend='rood', color='red', symbol='o', label='oranje')

        #xAxis.setAutoScale(False)
        return plot1D



    def makePlot2D():
        """ test function """

        # Just some random numbers
        imgData = np.random.randint(5, 1000, size=(1216, 1936), dtype=np.uint16)

        plot2D = GsswImageWidget(backend=BACKEND)
        plot2D.addImage(imgData)

        plot2D.colorbar.setLegend('narf')
        #print(plot2D.colorbar.getLegend())
        plot2D.colormap.setName('viridis')
        return plot2D


    def main():
        """ test code """
        app = QApplication([])
        plot = makePlot1D()
        #plot = makePlot2D()
        plot.show()
        app.exec_()

    main()
