#!/usr/bin/env python3
""" Utility to find the location of the Egress common `*.proto` source files, for use in projects.

Use as follows in your `Makefile`:
::

  # Path to Egress common protobuf files is detected by python module
  # Custom path can be set via GSSW_PROTO_PATH environment variable
  GSSW_PROTO_PATH := $(shell python3 -m gssw.proto.findprotofiles)
  $(info Using GSSW_PROTO_PATH: $(GSSW_PROTO_PATH) as root source directory for Egress proto files)

  # ... other Makefile stuff

  target:
      protoc ...... --proto_path="$GSSW_PROTO_PATH" ...... $<
 """
import os.path
import sys

#: The module searches for this file, and assumes the directory is correct when the file is found.
PROTO_SEARCH_FILE = 'gssw/proto/common.proto'

def main():
    """ Prints the location of the `*.proto` files to STDOUT, for use in project build scripts. """
    modulePath = os.path.abspath(os.path.dirname(__file__))

    protoSourcePaths = [os.environ.get('GSSW_PROTO_PATH', None), # Override via environment variable
                        os.path.join(modulePath, '../protobuf/'),       # Installed via package
                        os.path.join(modulePath, '../../../protobuf/')] # Installed from source

    for candidatePath in protoSourcePaths:
        if candidatePath is None or not candidatePath: # Skip when environment variable is not set
            continue
        candidatePath = os.path.abspath(candidatePath)
        protoSearchFilePath = os.path.join(candidatePath, PROTO_SEARCH_FILE)

        # Check if Egress protos are found in current candidate
        if os.path.exists(protoSearchFilePath):
            print("INFO: Found Egress protobuf files: {}".format(protoSearchFilePath),
                  file=sys.stderr) # Log found path to non-stdout
            print(candidatePath)   # Return found firectory via stdout
            return 0
    print("ERROR: Path to common protobuf files not found!", file=sys.stderr)

    # Return at least some output, although it will probably result in "File not found's" later.
    print('.')
    return 1

if __name__=='__main__':
    sys.exit(main())
