""" Configuration Dialog
"""
import logging
import os.path


from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt, pyqtSlot

from gssw.lib.defs import ActionEnum
from gssw.lib.typing import check_class
from gssw.lib.errors import DuplicateItemError
from gssw.lib.widgets.colors import widgetHasDarkBackGround
from gssw.quicklook.basemodel import BaseModel, BaseItem
from gssw.quicklook.shared import GlobalArrayInfoRepo

logger = logging.getLogger(__name__)


class BaseConfigItem(BaseItem):
    """ An item for in the Config Dialog's CurveModel or the PlotPanel curves list

        Represents a curve in the plot and the data necessary to display it (color, line-style, etc)

        Extends a PanelItem and thus maintains a reference to a RepoItem from the RepositoryModel.
    """
    def __init__(self, name, label=''):
        """ Constructor

            :param repoItem: the RepoItem where this PanelItem links to
            :param colorName: string containging the name of the color (can be #RRGGBB ?)
        """
        super().__init__()
        self._name = name
        self._label = '' if not label else label


    @property
    def name(self):
        """ Unique name of the item.
        """
        return self._name


    @name.setter
    def name(self, value):
        """ Unique name of the item.
        """
        check_class(value, str)
        self._name = value


    @property
    def label(self):
        """ String describing the item. E.g. for use in legends
        """
        return self._label


    @label.setter
    def label(self, value):
        """ String describing the item. E.g. for use in legends
        """
        check_class(value, str)
        self._label = value


    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        dct = {} # Configuration subdict

        dct['name']         = self.name
        dct['label']        = self.label

        return dct


    def setValuesFromDict(self, dct):
        """ Creates a panel from a dictionary with values
        """
        # Name should certainly be present in the dict. For the others we can use defaults if they
        # somehow are not present in the dictionary.
        self.name         = dct['name']
        self.label        = dct.get('label', '')



class BaseConfigModel(BaseModel):
    """ Base class for all config models.

        At the moment it does the same as BaseModel, but we add a BaseConfigModel to be consistent
        with BaseItem/BaseConfigItem
    """

    # All items in this model will be of this class
    ITEM_CLASS = BaseConfigItem


    # Extent the possible columns in the plot pannel model
    HEADERS    = ["Name", "Label"]
    PROPERTIES = ["name", "label"]
    COL_NAME, COL_LABEL = range(len(HEADERS))




class PanelConfigDialog(QtWidgets.QDialog):
    """ Dialog that contains a single PlotPanelConfigWidget
    """
    def __init__(self, configWidget, parent=None):
        """ Constructor
            :param repo: Refers to the repository. Needed to execute the SelectRepoItemDialog
            :param parent: Qt parent object
        """
        super().__init__(parent=parent)

        self.setWindowTitle("Configuration")

        self.mainLayout = QtWidgets.QVBoxLayout(self)
        self.setLayout(self.mainLayout)

        self.configWidget = configWidget
        self.mainLayout.addWidget(self.configWidget)

        self.buttonBox = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok |
            QtWidgets.QDialogButtonBox.Cancel |
            QtWidgets.QDialogButtonBox.Apply)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)

        self.resize(QtCore.QSize(400, 600)) # TODO: make persistent



class AbstractPanelConfigWidget(QtWidgets.QWidget):
    """ Contains widgets for configuring PlotPanel settings
    """
    # Many statements are inherit to Qt constructors.
    # pylint: disable=too-many-statements
    def __init__(self, configModel, parent=None):
        """ Constructor
        """
        super().__init__(parent=parent)
        check_class(configModel, BaseModel, allow_none=True)

        self._configModel = configModel

        self.controlLayout = QtWidgets.QVBoxLayout()
        self.setLayout(self.controlLayout)

        # Details common to all plot items
        self.commonLayout = QtWidgets.QFormLayout()
        self.commonLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.controlLayout.addLayout(self.commonLayout)

        self._addCommonWidgets()

        # -- Widgets for individual plot items --
        self.itemGroupBox = QtWidgets.QGroupBox("Items")
        self.controlLayout.addWidget(self.itemGroupBox)
        self.itemsGroupLayout = QtWidgets.QVBoxLayout(self.itemGroupBox)

        # The config list of data items.
        self.listLayout = QtWidgets.QHBoxLayout()
        self.itemsGroupLayout.addLayout(self.listLayout)

        self.listView = QtWidgets.QListView()
        self.listView.setModel(self._configModel)
        self.listView.setAlternatingRowColors(False)
        self.listLayout.addWidget(self.listView)


        # Buttons for adding/removing list items.
        self.listButtonsLayout = QtWidgets.QVBoxLayout()
        self.listLayout.addLayout(self.listButtonsLayout)

        self.itemAddNewButton = QtWidgets.QToolButton()
        self.itemAddNewButton.setText("+")
        self.itemAddNewButton.setToolTip("Add new curve item.")

        self.itemRemoveButton = QtWidgets.QToolButton()
        self.itemRemoveButton.setText("-")
        self.itemRemoveButton.setToolTip("Remove selected curve item.")

        self.listButtonsLayout.addWidget(self.itemAddNewButton)
        self.listButtonsLayout.addWidget(self.itemRemoveButton)
        self.listButtonsLayout.addStretch()

        # Set icon depending on color.
        app = QtWidgets.QApplication.instance()
        if widgetHasDarkBackGround(self):
            iconDir = os.path.join(app.iconPath, 'snipicons', 'white')
            addIconName = os.path.join(iconDir, 'plus-sign-i.svg')
            removeIconName = os.path.join(iconDir, 'minus-sign-i.svg')
        else:
            iconDir = os.path.join(app.iconPath, 'snipicons', 'black')
            addIconName = os.path.join(iconDir, 'plus-sign-m.svg')
            removeIconName = os.path.join(iconDir, 'minus-sign-m.svg')

        assert os.path.exists(addIconName), "File not found: {}".format(addIconName)
        assert os.path.exists(removeIconName), "File not found: {}".format(removeIconName)

        self.itemAddNewButton.setIcon(QtGui.QIcon(addIconName))
        self.itemRemoveButton.setIcon(QtGui.QIcon(removeIconName))

        # Widgets for showing properties of the individual plot items. They are linked to the
        # model by the widget mapper.
        self.mapper = QtWidgets.QDataWidgetMapper(parent=self)
        self.mapper.setModel(self._configModel)

        self.itemsFormLayout = QtWidgets.QFormLayout()
        self.itemsFormLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.itemsGroupLayout.addLayout(self.itemsFormLayout)

        self.nameComboBox = QtWidgets.QComboBox()
        self.nameComboBox.setFocusPolicy(Qt.StrongFocus)
        self.nameComboBox.setEditable(True)
        self.nameComboBox.setInsertPolicy(QtWidgets.QComboBox.InsertAlphabetically)
        self.nameComboBox.setModel(GlobalArrayInfoRepo.instance().dynamicModel)

        self.itemsFormLayout.addRow("Name", self.nameComboBox)
        self.mapper.addMapping(self.nameComboBox, 0) # Name should always be the first column

        self._addItemWidgets()

        # Emtpy space
        self.controlLayout.addStretch()

        self.updateWidgetsForIndex(QtCore.QModelIndex())

        # -- Connect signals --
        selectionModel = self.listView.selectionModel()
        selectionModel.currentRowChanged.connect(self.updateWidgetsForIndex)

        self.itemAddNewButton.clicked.connect(self.addNewItem)
        self.itemRemoveButton.clicked.connect(self.removeCurrentItem)

        self.setCurrentRow(0)


    # pylint: enable=too-many-statements

    def finalize(self):
        """ Cleanup resources. Must be called explicitly.
        """
        selectionModel = self.listView.selectionModel()
        selectionModel.currentRowChanged.connect(self.updateWidgetsForIndex)

        self.itemAddNewButton.clicked.disconnect(self.addNewItem)
        self.itemRemoveButton.clicked.disconnect(self.removeCurrentItem)

        super().finalize()


    def _addCommonWidgets(self):
        """ Adds common widgets to the config box.

            The base implementation does nothing. Should be implemented by the descendant.
        """
        pass


    def _addItemWidgets(self):
        """ Adds widget to the config item group box.

            The base implementation does nothing. Should be implemented by the descendant.
        """
        pass


    @pyqtSlot(str)
    def onItemChanged(self, _name):
        """ Is called when a panel item is updated in the model.
            Copies the values from the model to the widgets (i.e. reverts the data).
        """
        self.revertWidgetData()


    def setCurrentRow(self, row):
        """ Sets (and selects) the panel item by the row number in the item list
        """
        logger.debug("setCurrentRow: {}".format(row))
        idx = self.listView.model().index(row, 0)
        self.listView.setCurrentIndex(idx)


    def getCurrentRow(self):
        """ Gets the row number of the selected panel item in the item list.

           :return: int, Either the row number, or -1 in case of no selection.
        """
        return self.listView.currentIndex().row()


    def setCurrentRowByName(self, name):
        """ Sets (and selects) the panel item by the row number in the item list
        """
        logger.debug("setCurrentRowByName: {}".format(name))
        row = self._configModel.rowByName(name)
        self.setCurrentRow(row)


    def updateWidgetsForIndex(self, currentIndex, _previousIndex=QtCore.QModelIndex()):
        """ Updates the widgets for when current index has changed.
        """
        self.mapper.setCurrentModelIndex(currentIndex)
        self.nameComboBox.setEnabled(currentIndex.isValid())

        self._updateSpecialWidgetsForIndex(currentIndex)


    def _updateSpecialWidgetsForIndex(self, currentIndex):
        """ Updates the specialized widgets.
        """
        pass


    def addNewItem(self):
        """ Creates a new curve item and adds it to the model.

            Selects the new item in the list.

            Finds a unique item name by adding numbers to the end. E.g. item-1, item-2, etc.
        """
        name = "Item-0"
        row = self._configModel.rowByName(name, notFoundAction=ActionEnum.IGNORE)
        duplicateNr = 1
        while row is not None and duplicateNr < 1000:
            name = "Item-{}".format(duplicateNr)
            row = self._configModel.rowByName(name, notFoundAction=ActionEnum.IGNORE)
            duplicateNr += 1
        assert row is None, "Unable to add unique curve name."

        try:
            curveItem = self._createConfigItem(name)
            self._configModel.addItem(curveItem)

        except DuplicateItemError:
            logger.warning("Item already exists: {}".format(name))
        else:
            row = self._configModel.rowByName(name)
            self.setCurrentRow(row)

        self.nameComboBox.setFocus(Qt.OtherFocusReason)
        self.nameComboBox.lineEdit().selectAll()


    #@pyqtSlot(CurveItem)
    def _createConfigItem(self, name):
        """ Creates a new config item and adds it to the model.
        """
        raise NotImplementedError()


    @pyqtSlot()
    def removeCurrentItem(self):
        """ Removes the item that is currently selected in the listView
        """
        currentIndex = self.listView.selectionModel().currentIndex()

        if not currentIndex.isValid():
            logger.warning("No item selected (ignored)")
            return

        curveItem = self._configModel.itemByRow(currentIndex.row())
        assert curveItem is not None, "curveItem is None" # sanity check

        self._configModel.removeItemByName(curveItem.name)


    def asDict(self):
        """ Returns a dictionary representation for serialization purposes

            The descendant should call mapper.submit() as the last statement in its implementation.
            This will update the config model from the widgets.
        """
        # self.mapper.submit() # call this first in your implementation
        raise NotImplementedError()


    def setValuesFromDict(self, cfg):
        """ Sets the panel attributes from a dictionary with values

            The descendant should call mapper.revert() as the last statement in its implementation.
            This will update the widget from the config model.
        """
        raise NotImplementedError()
        # self.mapper.revert() # call this last in your implementation


