""" Base model for data repositories.
"""
import logging

from PyQt5 import QtCore
from PyQt5.QtCore import Qt, pyqtSignal

from gssw.lib.defs import ActionEnum
from gssw.lib.typing import type_name, check_class, check_is_a_string, check_is_a_sequence
from gssw.lib.errors import DuplicateItemError, ItemNotFoundError

logger = logging.getLogger(__name__)


class BaseItem():
    """ A registered array and its concatenating dimension
    """
    def __init__(self):
        """ Constructor
        """
        pass


    def __repr__(self):
        return "<{}: name={}>".format(type_name(self), self.name)


    @property
    def name(self):
        """ String representation of the item name.

            Abstract method. The descendant should implement it.
        """
        raise NotImplementedError()


    @name.setter
    def name(self, value):
        """ Unique name of the item.
        """
        raise NotImplementedError()


class BaseModel(QtCore.QAbstractTableModel):
    """ A table model that contains ArrayItems (or ArrayItem descendants).
    """

    # All items in this model will be of this class. BaseModel descendant will typically override
    # the ITEM_CLASS value
    ITEM_CLASS = BaseItem

    HEADERS    = [] # Column header labels. Should be overridden in descendants.
    PROPERTIES = [] # Item property name per column. Should be overridden in descendants

    sigModelAboutToBeCleared = pyqtSignal()
    sigItemAboutToBeRemoved  = pyqtSignal(str)  # param: item name
    sigItemAdded             = pyqtSignal(str)  # param: item name
    sigItemChanged           = pyqtSignal(str)  # param: item name

    def __init__(self, parent=None):
        """ Constructor
        """
        logger.debug("Constructing BaseModel, parent={}".format(parent))
        super().__init__(parent=parent)

        assert len(self.PROPERTIES) == len(self.HEADERS), "sanity check"

        logger.debug("Checking that {} has all properties in model's property list".format(self))
        for prop in self.PROPERTIES:
            if not hasattr(self.ITEM_CLASS, prop):
                raise AttributeError("Item class {} does not have the {!r} property"
                                     .format(self.ITEM_CLASS, prop))

        self._items = []


    def finalize(self):
        """ Cleanup resources.
        """
        self.clear()


    ######################
    # Overloaded methods #
    ######################

    def rowCount(self, _parent=None):
        """ Returns the number of rows under the given parent.
        """
        return len(self._items)


    def columnCount(self, _parent=None):
        """ Returns the number of columns for the children of the given parent.
        """
        return len(self.HEADERS)


    def flags(self, index):  #pylint: disable=unused-argument
        """ Returns the item flags for the given index

            The base implementation returns enabled, selectable and editable items.
        """
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable


    def data(self, index, role=Qt.DisplayRole):
        """ Returns the data stored under the given role for the item referred to by the index.
        """
        if not index.isValid():
            return None

        row, col = index.row(), index.column()

        if col < 0 or col >= self.columnCount():
            return None

        if row < 0 or row >= self.rowCount():
            return None

        if role == Qt.DisplayRole or role == Qt.EditRole:

            # Find and return the property of the item that corresponds to this column.
            item = self.items[row]
            itemProp = getattr(item, self.PROPERTIES[col])
            return itemProp

        return None


    def setData(self, index, value, role=Qt.EditRole):
        """ Sets the data of the item at the index to the given value.

            If the index is valid, setItemData is called to actually set the data. Descendants
            should typically override setItemData instead of setData

            Emits the dataChanged signal, but not the sigItemChanged signal (the latter is the
            responsibility of the caller). Otherwise the sigItemChanged could be called once per
            column and you want typically only one signal per complete item change.

            :param index: QModelIndex
            :param value: the item at the index will be set to this value
            :param role: typically the Qt.EditRole
            :returns: True for success, False otherwise
        """

        if role != Qt.EditRole:
            logger.warning("BaseModel.setData called with role other than EditRole {}".format(role))
            return False

        if not index.isValid():
            logger.warning("BaseModel.setData called with invalid index")
            return None

        row, col = index.row(), index.column()
        item = self.items[row]

        logger.debug("setDataCalled(row={}, col={}, value={}".format(row, col, value))

        # Find and return the property of the item that corresponds to this column and set it
        # to value.
        try:
            setattr(item, self.PROPERTIES[col], value)
        except AttributeError as ex:
            # If you get a 'can't set attribute' message, you are probably trying to set a
            # property without a setter.
            logger.debug("Unable to set {!r} property of row {} to {!r}. Reason: {}"
                         .format(self.PROPERTIES[col], row, value, ex))
            return False

        # Emit dataChanged signal for this cell (so that view can update themselves)
        # We don't call sigItemChanged, this is the responsibility of the caller.
        blockedStr = '(BLOCKED) ' if self.signalsBlocked() else ''
        logger.debug("self.signalsBlocked() {}".format(self.signalsBlocked()))
        logger.debug("{} emitting {}dataChanged signal for cell: ({}, {})"
                     .format(self, blockedStr, row, col))
        self.dataChanged.emit(index, index)

        return True


    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """ Data for the given role and section in the header with the specified orientation.

            :param section: row or column number, depending on orientation
            :param orientation: Qt.Horizontal or Qt.Vertical
        """
        if role != Qt.DisplayRole:
            return None

        if orientation == Qt.Horizontal:
            return self.HEADERS[section]
        
        return str(section)


    def blockSignals(self, block):
        """ If block is true, signals emitted by this object are blocked (i.e., emitting a signal
            will not invoke anything connected to it). If block is false, no such blocking will
            occur.

            Overloaded to add logging.

            :param block: New blocking value
            :return: the previous value of signalsBlocked().
        """
        logger.debug("Setting blocking of signal to {!r} for {!r}".format(block, self))
        result = super().blockSignals(block)
        assert self.signalsBlocked() == block, \
            "Sanity check failed ({} != {})".format(self.signalsBlocked(), block)
        return result


    ###############
    # New methods #
    ###############


    def itemByRow(self, row):
        """ Returns the TreeItem for the given row.
        """
        return self._items[row]


    @property
    def items(self):
        """ Returns (a shallow copy) of the list of items.
        """
        return list(self._items)


    def checkItemType(self, item):
        """ Raises TypeError if the item does not have self.ITEM_CLASS as type. If it does,
            nothing happens.

            Ensures the model only contains of type self.ITEM_CLASS
        """
        check_class(item, self.ITEM_CLASS)


    def rowByName(self, itemName, notFoundAction=ActionEnum.RAISE):
        """ Returns the array index in the repository having a certain name.

            Traverses the repository so is O(n). # TODO: index on name to make it faster?

            The notFoundAction value determines what happens if the item is not found. By default
            a ItemNotFoundError is raised. Other options are to log a warning or to ignore it (and
            return None).

            :param itemName: name of the item.
            :param notFoundAction: ActionEnum. Determines action if no item is found with that name.
        """
        ActionEnum.checkValueInEnum(notFoundAction)

        for idx, item in enumerate(self._items):
            if item.name == itemName:
                return idx

        msg = "No item found called {!r} in model: {}".format(itemName, self)

        if notFoundAction == ActionEnum.IGNORE:
            return None
        elif notFoundAction == ActionEnum.WARNING:
            logger.warning(msg)
            return
        elif notFoundAction == ActionEnum.RAISE:
            raise ItemNotFoundError(msg)
        else:
            raise ValueError("Invalid notFoundAction: {}".format(notFoundAction))


    def itemByName(self, itemName, notFoundAction=ActionEnum.RAISE):
        """ Returns the array in the repository having a certain name.

            Traverses the repository so is O(n). This may be optimized in the future.
            Throws ItemNotFoundError if not found.

            The notFoundAction value determines what happens if the item is not found. By default
            a ItemNotFoundError is raised. Other options are to log a warning or to ignore it (and
            return None).

            :param itemName: name of the item.
            :param notFoundAction: ActionEnum. Determines action if no item is found with that name.
        """
        return self._items[self.rowByName(itemName, notFoundAction=notFoundAction)]


    def clear(self):
        """ Removes all items from the model.
        """
        self.repopulate([])


    def asDictList(self):
        """ Returns a list of dictionaryies for serialization purposes
        """
        return [item.asDict() for item in self.items]


    def setValuesFromDictList(self, dictList):
        """ Clears and sets the model given a list of dictionaries
        """
        itemList = []
        for dct in dictList:
            check_class(dct, dict)
            item = self.ITEM_CLASS(**dct)
            #item.setValuesFromDict(dct)
            itemList.append(item)

        for elem in itemList:
            check_class(elem, self.ITEM_CLASS)

        self.repopulate(itemList)


    def repopulate(self, newItems=None):
        """ Removes all items from the model and fills it with new items
        """
        check_is_a_sequence(newItems, allow_none=True)

        # First give the panels the chance to clear their plots
        logger.debug("emitting sigModelAboutToBeCleared signal")
        self.sigModelAboutToBeCleared.emit()

        self.beginResetModel()
        if newItems is None:
            self._items = []
        else:
            self._items = list(newItems) # makes copy
        self.endResetModel()

        self.logContents(msg="Model {} after clearing all items".format(self))


    def addItem(self, item, row=None, removeExisting=False):
        """ Adds an item to the model

            Calls self.checkItemType to check that the item is of the correct type.

            Emits the sigItemAdded signal.

            :param item: RepoItem
            :param row: Where the item is inserted. If row=None, the item is appended at the end.
            :param removeExisting: If True, any existing array is unregistered . If False (the
                default), a DuplicateItemError is raised if the array is already registered.
            :param return: index to the added row (with column 0)
        """
        self.checkItemType(item)
        check_class(row, int, allow_none=True)

        # Check that the arrayName does not yet occur
        try:
            oldItem = self.itemByName(item.name)
        except ItemNotFoundError:
            oldItem = None

        if oldItem is not None:
            if removeExisting:
                self.removeItemByName(item.name)
            else:
                raise DuplicateItemError("Array with name {!r} already exists".format(item.name))

        row = len(self._items) if row is None else row

        rowsBefore = self.rowCount()
        logger.debug("Inserting item {} at row: {}".format(item.name, row))

        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        self._items.insert(row, item)
        self.endInsertRows()

        rowsAfter = self.rowCount()
        logger.debug("Rowcounts before {}, after {}".format(rowsBefore, rowsAfter))
        assert rowsBefore + 1 == rowsAfter, "{} + 1 is not {}".format(rowsBefore, rowsAfter)

        self.sigItemAdded.emit(item.name)

        return self.index(row, 0)


    def removeItemByName(self, name, notFoundAction=ActionEnum.RAISE):
        """ Removes the model item with that name.

            If the panel does not have an item with this name an exception is raised.

            :param name: array item name
            :param notFoundAction: ActionEnum. Determines action if no item is found with that name.
        """
        check_is_a_string(name)
        row = self.rowByName(name, notFoundAction=notFoundAction)

        if row is not None:
            # First give the panels the chance to remove the item from their plots.
            self.sigItemAboutToBeRemoved.emit(name)

            self.beginRemoveRows(QtCore.QModelIndex(), row, row)
            del self._items[row]
            self.endRemoveRows()

            self.logContents(msg="Model {} after removing: {}".format(self, name))


    def emitDataChangedForRow(self, row):
        """ Emits the dataChanged signal for the entire row.
        """
        blockedStr = '(BLOCKED) ' if self.signalsBlocked() else ''
        logger.debug("{} emitting {}dataChanged signal for row: {}".format(self, blockedStr, row),
                     stack_info=False)

        topLeftIndex = self.index(row, 0)
        bottomRightIndex = self.index(row, self.columnCount() - 1)
        assert topLeftIndex.isValid(), "topLeftIndex is not valid"
        assert bottomRightIndex.isValid(), "bottomRightIndex is not valid"
        self.dataChanged.emit(topLeftIndex, bottomRightIndex)


    def emitSigItemChanged(self, name):
        """ Emits the sigItemChanged signal with the item name name
        """
        blockedStr = '(BLOCKED) ' if self.signalsBlocked() else ''
        logger.debug("{} emitting {}sigItemChanged signal for array: {}"
                     .format(self, blockedStr, name),
                     stack_info=False)

        self.sigItemChanged.emit(name) # Triggers updates of the plots


    def logContents(self, msg=None, loggerName=None, level='debug', item_prefix='  '):
        """ Logs the contents of the model.
        """
        if loggerName is not None:
            localLogger = logging.getLogger(loggerName)
        else:
            localLogger = logger

        levelNr = logging.getLevelName(level.upper())
        if msg is None:
            localLogger.log(levelNr, "Logging {}".format(self))
        elif msg:
            localLogger.log(levelNr, str(msg))

        for item in self._items:
            localLogger.log(levelNr, "{}{}".format(item_prefix, item))
