""" Panels for in the Quicklook application.
"""
import copy, logging, os.path

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot
from PyQt5.QtGui import QColor

from gssw.lib.arrays.gridded import GriddedArray
from gssw.lib.typing import check_class, check_is_a_string, full_type_name, get_symbol_from_module
from gssw.lib.widgets.colors import setWidgetColor, widgetHasDarkBackGround, PaletteRoleEnum, \
         QCOLOR_SRON_BLUE
from gssw.lib.widgets.misc import saveWidgetAsImage, qRectToDict
from gssw.quicklook.baseconfig import PanelConfigDialog
from gssw.quicklook.basemodel import ItemNotFoundError
from gssw.quicklook.defs import MIME_GSSW_PANEL, MIME_ARRAYS_INFO
from gssw.quicklook.repo import RepositoryModel, AppendModeEnum
from gssw.quicklook.shared import ArrayInfoModel

logger = logging.getLogger(__name__)


# The default width and height of a plot panel (in grid points)
PANEL_SIZE_DEFAULT = 18



def encodePanelMimeData(edgesFlag, hotSpot, geometry):
    """ Extracts drag-drop  from the binary mime data

        :param edgesFlag: Qt.Edges flags containigg the edges near the hotspot.
        :param hotSpot: QPoint containing the drag hotspot
        :param geometry: QRect with the panel geometry
        :return: bytes with encoded mime data.
    """
    mimeStr = "{} {} {} {} {} {} {}" \
        .format(int(edgesFlag), hotSpot.x(), hotSpot.y(),
                geometry.x(), geometry.y(), geometry.width(), geometry.height()
                )
    mimeBytes = bytes(mimeStr, encoding='ascii')
    return mimeBytes


def decodePanelMimeData(mimeBytes):
    """ Extracts hotspot and widget geometry from the binary mime data

        :param mimeBytes: bytes
        :return: (edgesFlag, hotspot, geometry) tuple
        :rtype: (int, QPoint, QRect) tuple
    """
    mimeStr = str(mimeBytes, encoding='ascii')
    edgesFlag, hotX, hotY, geomX, geomY, geomWidth, geomHeight = mimeStr.split(' ')
    edgesFlag = int(edgesFlag)
    hotSpot = QtCore.QPoint(int(hotX), int(hotY))
    geometry = QtCore.QRect(int(geomX), int(geomY), int(geomWidth), int(geomHeight))
    return edgesFlag, hotSpot, geometry


def createPanelByClassName(fullClassName, panelName, parent=None):
    """ Factory function to create a panel given its full class name.

        :param fullClassName: String with the full class name.
            E.g. gssw.quicklook.plotpanel import PlotPanel
        :param panelName: name of the panel
        :param parent: Panel Qt parent
        :return: object of the desired panel class
    """
    check_class(panelName, str, allow_none=True)
    # PanelClass is really a type so it should be capitalized.
    # pylint: disable=invalid-name
    moduleName, className = fullClassName.rsplit('.', 1)
    PanelClass = get_symbol_from_module(moduleName, className)

    logger.debug("Creating object of type: {}".format(PanelClass))

    assert issubclass(PanelClass, AbstractPanel), \
        "Class should derive from AbstractPanel: {}".format(PanelClass)

    panel = PanelClass.createPanel(panelName, parent=parent)
    check_class(panel, PanelClass) # sanity check
    return panel

    # pylint: enable=invalid-name


class AbstractPanel(QtWidgets.QFrame):
    """ Base class for panels in the Quicklook GUI.

        Is an empty model. Has no associated model.
    """
    #pylint: disable=too-many-statements
    sigClosePanel = pyqtSignal(str)  # param: panel name
    sigPanelConfigChanged = pyqtSignal(str)  # param: panel name

    # Default geometry in grid points. Mainly for initialization, typically the panel is resized
    # and moved after construction.
    DEFAULT_GRID_GEOMETRY = QtCore.QRect(0, 0, PANEL_SIZE_DEFAULT, PANEL_SIZE_DEFAULT)

    def __init__(self, name, parent=None):
        """ Constructor

            :param name: panel name string
            :param parent: Qt parent object.
        """
        super().__init__(parent=parent)
        check_is_a_string(name)

        self.setObjectName("panel")
        self._page = None
        self._name = name
        self._title = ""
        self._gridGeometry = copy.copy(self.DEFAULT_GRID_GEOMETRY)
        self.dialog = None # Will be set in execConfigDialog
        self._contentMargin = 3  # Determines the area where the user can resize the panel

        # With Plain shadow the WindowText role will be used for coloring
        self.setFrameStyle(QtWidgets.QFrame.Panel | QtWidgets.QFrame.Plain)
        self.setLineWidth(1)

        # Prevent background from being transparent if no style sheet is set.
        self.setAutoFillBackground(True)

        # This key must be pressed, together with moving the mouse, for moving and resizing windows.
        self.panModifier = Qt.ControlModifier

        self.setMouseTracking(True)

        # The panel accepts drops from the array info list. In that case a rubber band will be
        # drawn with the same size as the panel to indicate that the array is added to the panel,
        self.setAcceptDrops(True)
        self.rubberBand = QtWidgets.QRubberBand(QtWidgets.QRubberBand.Rectangle, self)
        setWidgetColor(self.rubberBand, QColor("magenta"), PaletteRoleEnum.Highlight)

        ### Actions ###
        self.configAction = QtWidgets.QAction("Configure", self)
        self.configAction.setToolTip("Configure panel")
        #self.configAction.setIcon(QtGui.QIcon(ICON_DIR + "/cog-i.svg"))
        self.configAction.triggered.connect(self.execConfigDialog)

        self.clearAction = QtWidgets.QAction("Clear data", self)
        self.clearAction.setToolTip("Removes the data from the panel")

        self.renameAction = QtWidgets.QAction("Rename...", self)
        self.renameAction.setToolTip("Renames the panel")
        self.renameAction.triggered.connect(self.renamePanel)

        self.saveAsImageAction = QtWidgets.QAction("Save As Image...", self)
        self.saveAsImageAction.setToolTip("Save panel as image")
        self.saveAsImageAction.triggered.connect(self.saveAsImage)

        ### Context menu ###

        # Override the context menu policy of the main window so that panels have their own.
        self.setContextMenuPolicy(Qt.CustomContextMenu)

        self.menu = QtWidgets.QMenu(self.name, self)
        self.menu.insertAction(self.saveAsImageAction, self.configAction)
        self.menu.insertAction(self.saveAsImageAction, self.renameAction)
        self.menu.insertAction(self.saveAsImageAction, self.clearAction)
        self.menu.addAction(self.saveAsImageAction)

        ### Widgets ###

        self.mainLayout = QtWidgets.QVBoxLayout()
        #self.mainLayout.setContentsMargins(1, 1, 1, 1)
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.mainLayout.setSpacing(0)
        self.setLayout(self.mainLayout)

        self.titleBar = QtWidgets.QWidget()
        self.titleBar.setObjectName('panel-title-bar')

        sizePolicy = self.titleBar.sizePolicy()
        sizePolicy.setHorizontalPolicy(QtWidgets.QSizePolicy.Ignored)
        sizePolicy.setVerticalPolicy(QtWidgets.QSizePolicy.Fixed)
        self.titleBar.setSizePolicy(sizePolicy)

        self.mainLayout.addWidget(self.titleBar)

        self.titleBarLayout = QtWidgets.QHBoxLayout()
        self.titleBarLayout.setContentsMargins(0, 0, 0, 0)
        self.titleBarLayout.setSpacing(0)

        self.titleBar.setLayout(self.titleBarLayout)

        self.titleBarLabel = QtWidgets.QLabel(self._name)
        self.titleBarLabel.setObjectName('panel-title-bar-label')
        self.titleBarLabel.setWordWrap(False)
        self.titleBarLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.titleBarLabel.setCursor(Qt.OpenHandCursor)
        self.titleBarLabel.setMouseTracking(True)
        self.titleBarLabel.installEventFilter(self)
        titleFont = self.titleBarLabel.font()
        titleFont.setWeight(QtGui.QFont.Normal)
        self.titleBarLabel.setFont(titleFont)

        self.configButton = QtWidgets.QToolButton()
        self.configButton.setObjectName('panel-config-button')
        self.configButton.setDefaultAction(self.configAction)
        self.configButton.setMinimumSize(QtCore.QSize(16, 16))
        self.configButton.setIconSize(QtCore.QSize(20, 20))
        self.configButton.setCursor(Qt.ArrowCursor)
        # Use style sheet to get rid of shading form the Qt style
        self.configButton.setStyleSheet('border: none;')
        #self.configButton.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)

        self.closeButton = QtWidgets.QToolButton()
        self.closeButton.setObjectName('panel-close-button')
        self.closeButton.setCursor(Qt.ArrowCursor)
        self.closeButton.setText("Close")
        #self.closeButton.setIcon(QtGui.QIcon(ICON_DIR + "/remove-sign-i.svg"))
        self.closeButton.setToolTip("Close panel")
        self.closeButton.setStyleSheet('border: none;') # Use style sheet to get rid of shading

        self.titleBar.setAutoFillBackground(True)
        setWidgetColor(self.titleBar, QColor('red'), PaletteRoleEnum.Window) # Active and Inactive
        setWidgetColor(self.titleBar, QColor('yellow'), PaletteRoleEnum.WindowText)

        self.titleBarLayout.addWidget(self.configButton)
        self.titleBarLayout.addWidget(self.titleBarLabel)
        self.titleBarLayout.addWidget(self.closeButton)

        # Set title bar colors. Note that in the style sheets you need to set them for all widgets
        # In classic CSS, when font and color of an item is not explicitly set, it gets
        # automatically inherited from the parent. By default, when using Qt Style Sheets, a
        # widget does not automatically inherit its font and color setting from its parent widget.
        # From Qt 5.7 onwards you can set the Qt::AA_UseStyleSheetPropagationInWidgetStyles flag,
        # like this:
        #   QCoreApplication.setAttribute(Qt.AA_UseStyleSheetPropagationInWidgetStyles, True)
        # See: http://doc.qt.io/qt-5/stylesheet-syntax.html#inheritance
        # fgColor = getWidgetColor(self, PaletteRoleEnum.HighlightedText)
        # bgColor = getWidgetColor(self, PaletteRoleEnum.Highlight)
        fgColor = QColor("white")
        bgColor = QCOLOR_SRON_BLUE
        for widget in [self.titleBar, self.configButton, self.titleBarLabel, self.closeButton]:
            setWidgetColor(widget, fgColor, PaletteRoleEnum.WindowText)
            setWidgetColor(widget, bgColor, PaletteRoleEnum.Window)

        # Set icon depending on color.
        app = QtWidgets.QApplication.instance()
        if widgetHasDarkBackGround(self.titleBarLabel):
            iconDir = os.path.join(app.iconPath, 'snipicons', 'white')
            configIconName = os.path.join(iconDir, 'cog-i.svg')
            closeIconName = os.path.join(iconDir, 'remove-sign-i.svg')
        else:
            iconDir = os.path.join(app.iconPath, 'snipicons', 'black')
            configIconName = os.path.join(iconDir, 'cog-m.svg')
            closeIconName = os.path.join(iconDir, 'remove-sign-m.svg')

        assert os.path.exists(configIconName), "File not found: {}".format(configIconName)
        assert os.path.exists(closeIconName), "File not found: {}".format(closeIconName)
        self.configAction.setIcon(QtGui.QIcon(configIconName))
        self.closeButton.setIcon(QtGui.QIcon(closeIconName))

        # The contentsWidget is a container for all the contents of the panel.
        # Descendants should assign a layout to it and put their contents in it.
        self.contentsWidget = QtWidgets.QWidget()
        self.contentsWidget.setObjectName('panel-contents')
        self.contentsWidget.setMouseTracking(True)
        #self.contentsWidget.setStyleSheet('background: pink;') for testing
        self.mainLayout.addWidget(self.contentsWidget)

        self.contentsLayout = QtWidgets.QVBoxLayout()
        mar = self._contentMargin
        self.contentsLayout.setContentsMargins(mar, mar, mar, mar)

        self.contentsLayout.setSpacing(0)
        self.contentsWidget.setLayout(self.contentsLayout)

        # connect signals
        self.closeButton.clicked.connect(self.closePanel)


    #pylint: enable=too-many-statements

    def finalize(self):
        """ Cleanup resources.
        """
        self.configAction.triggered.disconnect(self.execConfigDialog)
        self.closeButton.clicked.disconnect(self.closePanel)


    @property
    def page(self):
        """ Returns the page on which this panel is located.
        """
        return self._page


    @page.setter
    def page(self, value):
        """ A the page on which this page is located.

            This value is typically set by Page.insertPanel. You should not set it yourself!
        """
        self._page = value


    @property
    def name(self):
        """ A name of this panel. Should be unique."""
        return self._name


    @name.setter
    def name(self, value):
        """ Sets the name of this panel.

            Should only be called by the page because it needs to be consistent!
        """
        self._name = value
        self.titleBarLabel.setText(value)


    def getTitle(self):
        """ Returns the panel title (as shown in the title bar)
        """
        return self.titleBarLabel.text()


    @property
    def gridGeometry(self):
        """ The geometry of this panel expressed in grid points.
            :rtype: QtCore.QRect
        """
        return self._gridGeometry


    @gridGeometry.setter
    def gridGeometry(self, geomGrid):
        """ Sets geometry of this panel expressed in grid points.

            Always use page.setPanelGridGeometry() to update this so that the geometry and
            gridGeometry are consistent. Don't resize this panel directly.
            TODO: make reference to the containing page and get the grid from there?

            :param QtCore.QRect geomGrid: rectangle with position and size in grid coordintes.
        """
        check_class(geomGrid, QtCore.QRect)
        assert geomGrid.isValid(), "Grid rectangle not valid: {}".format(geomGrid)
        self._gridGeometry = geomGrid



    def nearEdges(self, pos):
        """ Returns the edges that are near the mouse position.

            Pos is considered near an edge if it's distance is <= contents margin or on the frame
            border.

            :param pos: QPoint
            :return: Qt.Edges flags (or 0 if no flags)
        """
        if not self.rect().contains(pos):
            return 0

        flags = 0

        margins = self.contentsLayout.contentsMargins()
        frameWidth = self.frameWidth()

        if pos.x() <= margins.left() + frameWidth:
            flags |= Qt.LeftEdge

        if pos.y() <= frameWidth: # Only use the framewidth here because the title bar must override
            flags |= Qt.TopEdge

        if pos.x() >= self.rect().width() - margins.right() - frameWidth:
            flags |= Qt.RightEdge

        if pos.y() >= self.rect().height() - margins.bottom() - frameWidth:
            flags |= Qt.BottomEdge

        return flags


    @classmethod
    def createPanel(cls, name, parent=None):
        """ Common interface to create a Panel (i.e. always with the same signature)

            E.g. the repo parameter is ignored in AbstractPanel but is used in AbstractDataPanel and
            descendants.

            :return: panel object of type cls
        """
        logger.debug("Creating panel of type: {}".format(cls))
        panel = cls(name=name, parent=parent)
        check_class(panel, cls)
        return panel


    def closePanel(self):
        """ Closes the panel after asking for confirmation by the user


            Emits sigClosePanel signal so that the containing page can remove it.
        """
        reply = QtWidgets.QMessageBox.question(
            self, "Confirm Remove Panel {}".format(self.name),
            "Do you want to remove the {} panel?\nThere is no undo!".format(self.name),
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

        if reply != QtWidgets.QMessageBox.Yes:
            logger.info("Remove panel cancelled by user: {}".format(self.name))
            return

        logger.debug("Emitting sigCLosePanel({})".format(self.name))
        self.sigClosePanel.emit(self.name)


    def eventFilter(self, _watchedObject, event):
        """ Filters out the event of the title bar label to that we can drag from the title bar.
        """
        if event.type() == QtCore.QEvent.MouseButtonPress:

            assert isinstance(event, QtGui.QMouseEvent)
            posInPanel = self.titleBarLabel.mapTo(self, event.pos())

            self._startDrag(posInPanel)
            return True
        else:
            return False


    def mouseMoveEvent(self, event):
        """ Handles a mouse move.
        """
        edgesFlag = int(self.nearEdges(event.pos()))
        try:
            if edgesFlag == 0:
                # Cursor in the middle of the panel
                modifierPressed = event.modifiers() & self.panModifier
                if modifierPressed:
                    #self.setCursor(Qt.DragMoveCursor) # Doesn't seem to work on X11
                    self.setCursor(Qt.SizeAllCursor)
                else:
                    self.setCursor(Qt.ArrowCursor)

            elif edgesFlag == Qt.LeftEdge or edgesFlag == Qt.RightEdge:
                # Cursor on left or right edge
                self.setCursor(Qt.SizeHorCursor)
            elif edgesFlag == Qt.TopEdge or edgesFlag == Qt.BottomEdge:
                # Cursor on top or bottom edge
                self.setCursor(Qt.SizeVerCursor)
            elif (edgesFlag == int(Qt.TopEdge | Qt.RightEdge) or
                  edgesFlag == int(Qt.BottomEdge | Qt.LeftEdge)):
                # Cursor on top-right or bottom-left corner
                self.setCursor(Qt.SizeBDiagCursor)
            elif (edgesFlag == int(Qt.TopEdge | Qt.LeftEdge) or
                  edgesFlag == int(Qt.BottomEdge | Qt.RightEdge)):
                # Cursor on top-left or bottom-right corner
                self.setCursor(Qt.SizeFDiagCursor)
            else:
                self.setCursor(Qt.ArrowCursor)
                raise AssertionError("Unexpected edge flag combination: {}".format(edgesFlag))
        finally:
            event.ignore()


    def mousePressEvent(self, event):
        """ Handles a mouse event. Can start a drag.
        """
        if event.button() != Qt.LeftButton:
            return

        edgesFlag = int(self.nearEdges(event.pos()))
        modifierPressed = event.modifiers() & self.panModifier

        if modifierPressed or edgesFlag != 0:
            self._startDrag(event.pos())


    def _startDrag(self, eventPos):
        """ Starts a drag operation.

            The drag will move the panel (or resize it if the cursor is close to the edge).
            Note that the drag will end on the containing page, so the actual resize is done
            in FreeLayoutPage.dropEvent() which then calls panel.setGeomToGrid

            :param eventPos: QPoint with event.pos() data
        """
        self.raise_() # Raise to the top

        edgesFlag = int(self.nearEdges(eventPos))

        mimeBytes = encodePanelMimeData(edgesFlag, eventPos, self.geometry())
        mimeData = QtCore.QMimeData()
        mimeData.setData(MIME_GSSW_PANEL, mimeBytes)

        drag = QtGui.QDrag(self)
        drag.setMimeData(mimeData)
        drag.setPixmap(QtGui.QPixmap(1, 1)) # Hack to get rid of small rectangle on OS-X
        drag.setHotSpot(eventPos )

        logger.debug("Starting drag...")

        # This will typically be handled by FreeLayoutPage.dragMoveEvent()
        dropAction = drag.exec_(Qt.MoveAction, Qt.MoveAction)

        logger.debug("Drop action: {}".format(dropAction))

        if dropAction == Qt.IgnoreAction:
            logger.debug("Drag finished. Drop action: ignore")
        elif dropAction == Qt.MoveAction:
            logger.debug("Drag finished. Drop action: move")
        else:
            raise AssertionError("Unexpected drop action: {}".format(dropAction))



    def dragEnterEvent(self, event):
        """ Called when a drag is in progress and the mouse enters this widget.

            Just calls drag MoveEvent.
        """
        self.dragMoveEvent(event)


    def dragMoveEvent(self, event):
        """ Called when a drag is in progress and when any of the following conditions occur:
            the cursor enters this widget, the cursor moves within this widget, or a modifier key
            is pressed on the keyboard while this widget has the focus.
        """
        # logger.debug("Drag move event called at pos: {}".format(event.pos()))

        mimeData = event.mimeData()
        if mimeData.hasFormat(MIME_ARRAYS_INFO):
            self.rubberBand.show()
            self.rubberBand.setGeometry(0, 0, self.width(), self.height())
            event.acceptProposedAction()
        else:
            logger.debug("No GSSw array info drag event. Ignored.")
            self.rubberBand.hide()
            self.rubberBand.setGeometry(QtCore.QRect()) # invalidates geometry
            event.ignore()


    def dragLeaveEvent(self, _event):
        """ Called when a drag is in progress and the mouse leaves this widget.

            Hides and invalidates the rubberband.
        """
        self.rubberBand.hide()
        self.rubberBand.setGeometry(QtCore.QRect()) # invalidates geometry


    def dropEvent(self, event):
        """ Called when the drag is dropped on this widget.
        """
        logger.debug("Drop event called. pos: {}".format(event.pos()))
        mimeData = event.mimeData()
        if mimeData.hasFormat(MIME_ARRAYS_INFO):

            # Create panel from dragged array name(s) and open a config dialog
            arrInfoItems = ArrayInfoModel.decodeArraysInfoMimeData(mimeData.data(MIME_ARRAYS_INFO))
            self.configureFromArrayInfoList(arrInfoItems) # will open the config dialog
        else:
            event.ignore()

        self.rubberBand.hide()
        self.rubberBand.setGeometry(QtCore.QRect()) # invalidates geometry


    def configureFromArrayInfoList(self, arrInfoItems):
        """ Configures the panel from the items in the array info list.

           This is an abstract method, descendants should override this method.

            ;param list arrInfoItems: list of ArrayInfoItems
        """
        raise NotImplementedError("Not yet implemented for this type of panel.")


    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        # Pylint doesn't detect isValid is a member of a QRect
        #pylint: disable=no-member
        assert self.gridGeometry.isValid(), "Grid rectangle not valid: {}".format(self.gridGeometry)
        #pylint: enable=no-member
        return {'name': self.name,
                '_class_': full_type_name(self),
                'gridGeometry': qRectToDict(self.gridGeometry)}


    def setValuesFromDict(self, dct):
        """ Creates a panel from a dictionary with values
        """
        # Not setting gridGeometry here. This is done in Page.insertPanel()


    def saveAsImage(self):
        """ Asks the user for a file name and saves the panel as image to file
        """
        saveWidgetAsImage(self, self.contentsWidget, baseName=self.name)


    @pyqtSlot()
    def execConfigDialog(self, selectName=None):
        """ Shows the configuration dialog for this panel.
        """
        logger.debug("Executing config dialog")
        self.dialog = self._createConfigDialog(parent=self) # Can be local var if modal
        self.dialog.setWindowTitle("Config {}".format(self.name))
        if selectName:
            self.dialog.configWidget.setCurrentRowByName(selectName)

        applyButton = self.dialog.buttonBox.button(QtWidgets.QDialogButtonBox.Apply)
        applyButton.clicked.connect(self._applyDialogSettings)
        self.dialog.setModal(True)
        returnCode = self.dialog.exec_()

        if returnCode == QtWidgets.QDialog.Accepted:
            self._applyDialogSettings()
        else:
            logger.debug("User pressed cancel. Discarding dialog settings")
        applyButton.clicked.disconnect(self._applyDialogSettings)


    def _createConfigDialog(self, parent=None):
        """ Creates a configuration dialog that belongs to this panel.

            The default implementation creates a BasePanelConfigDialog, which is an empty dialog.
            Descendants should override this.
        """
        configWidget = self._createConfigWidget(parent=parent)
        dialog = PanelConfigDialog(configWidget, parent=parent)

        cfg = self.asDict()['cfg']
        dialog.configWidget.setValuesFromDict(cfg)

        return dialog


    def _createConfigWidget(self, parent=None):
        """ Creates a config widget that belongs to this

            The base implementation does nothing. Should be overridden in the descendant.

            :return: AbstractPaneConfigWidget descendant.
        """
        raise NotImplementedError()


    def _applyDialogSettings(self):
        """ Applies the settings from the config dialog to the panel.
        """
        logger.debug("Using new value from dialog")

        # The settings from the dialog only correspond to the 'cfg' subdict, so we replace it.
        dct = self.asDict()
        dct['cfg'] = self.dialog.configWidget.asDict()
        self.setValuesFromDict(dct)
        self.emitConfigChanged()


    def emitConfigChanged(self):
        """ Emits the sigPanelConfigChanged signal
        """
        logger.debug("Emitting sigPanelConfigChanged({})".format(self.name))
        self.sigPanelConfigChanged.emit(self.name)


    def arrayInConfig(self, _arrayName):
        """ Returns True if this panel is configured to show data of this array.

            Abstract method, descendants should override this.

            :param arrayName: gridded array name
            :return: False
        """
        raise NotImplementedError("Descendant should override this method")


    def getConfigArrayNames(self):
        """ Returns a list with the names of the array used in the configuration"""
        raise NotImplementedError("Descendant should override this method")


    def updatePanelData(self, griddedArray):
        """ Updates the panel data with a new gridded array.

            Will/should only be called with arrays that this panel is configured to show
            (i.e. self.showsArray(array.name) ==True). The panel class should override this and
            update or replace the array it stores.

            Abstract method, descendants should override this.
        """
        #check_class(griddedArray, GriddedArray)
        #assert self.showsArray(griddedArray.name)
        raise NotImplementedError("Descendants should override this method")


    @pyqtSlot()
    def clearData(self):
        """ Clears the panel's array data.
            Abstract method, descendants should override this.
        """
        raise NotImplementedError("Descendants should override this method")


    @pyqtSlot()
    def renamePanel(self, newName=None):
        """ Renames the panel.

            :param str None newName: the new panel name.
                If None the user will be prompted for a new name
        """
        currentName = self.name

        if newName is None:
            newName, okClicked = QtWidgets.QInputDialog.getText(
                self, "Rename Panel", "Panel name:", QtWidgets.QLineEdit.Normal, currentName)

            if not okClicked:
                logger.debug("Rename canceled")
                return

            newName = newName.strip()

            if newName == '':
                _reply = QtWidgets.QMessageBox.information(
                    self, "Panel Rename Failed", "Panel name cannot be empty.")
                return

            if newName in self.page.panels:
                _reply = QtWidgets.QMessageBox.information(
                    self, "Page Rename Failed", "A page with this name already exists")
                return

        check_class(newName, str)
        logger.debug("Renaming panel: {} -> {}".format(currentName, newName))

        # Rename key in the panels dict of the containting page.
        self.page.panels[newName] = self.page.panels.pop(currentName)

        self.name = newName
        self.menu.menuAction().setText(newName)



# Many statements are inherit to Qt constructors.
# pylint: disable=too-many-statements

class AbstractDataPanel(AbstractPanel):
    """ A panel displays data from a Repository.
    """
    def __init__(self, name, repo=None, parent=None):
        """ Constructor

            :param name: panel name string
            :param repo: a RepoModel. If None a new RepoModel will be created.
            :param parent: Qt parent object.
        """
        super().__init__(name, parent=parent)
        #check_class(model, PanelModel)
        check_class(repo, RepositoryModel, allow_none=True)

        if repo is None:
            self._repo = RepositoryModel(parent=self, appendMode=AppendModeEnum.replace)
        else:
            self._repo = repo
            assert self._repo.appendMode == AppendModeEnum.replace, \
                "Repo appendMode should be replacing. Got: {}".format(self._repo.appendMode)

        # Add widgets for the descendants.
        self._addSpecializedWidgets()

        # Connect signals
        self.clearAction.triggered.connect(self.clearData)

        self.repo.sigModelAboutToBeCleared.connect(self.onRepoAboutToBeCleared)
        self.repo.sigItemAboutToBeRemoved.connect(self.onRepoItemAboutToBeRemoved)
        self.repo.sigItemAdded.connect(self.onRepoItemChanged)
        self.repo.sigItemChanged.connect(self.onRepoItemChanged)


    def _addSpecializedWidgets(self):
        """ Adds specialized (i.e. the non-common) widgets to the config box.

            The base implementation does nothing. Should be implemented by the descendant.
        """
        pass


    def finalize(self):
        """ Cleanup resources. Must be called explicitly.
        """
        self.repo.sigModelAboutToBeCleared.disconnect(self.onRepoAboutToBeCleared)
        self.repo.sigItemAboutToBeRemoved.disconnect(self.onRepoItemAboutToBeRemoved)
        self.repo.sigItemAdded.disconnect(self.onRepoItemChanged)
        self.repo.sigItemChanged.disconnect(self.onRepoItemChanged)

        self.clearAction.triggered.disconnect(self.clearData)

        super().finalize()


    @property
    def repo(self):
        """ The RepositoryModel from which items can be taken and placed in the PanelModdel. """
        return self._repo


    @property
    def storeMaxSamples(self):
        """ The maximum samples that will be plotted and stored.
        """
        return self.repo.storeMaxSamples


    @storeMaxSamples.setter
    def storeMaxSamples(self, value):
        """ The maximum samples that will be plotted and stored.
        """
        check_class(value, int)
        self.repo.storeMaxSamples = value


    def updatePanelData(self, griddedArray):
        """ Updates the panel data with a new gridded array.

            Will/should only be called with arrays that this panel is configured to show
            (i.e. self.showsArray(array.name) == True).
            The panel class should override this and update or replace the array it stores.

            The base implementation returns False. Descendants should override this.
        """
        check_class(griddedArray, GriddedArray)
        assert self.arrayInConfig(griddedArray.name), \
            "updatePanelData called for array it doesn't show: {}".format(griddedArray.name)

        self.repo.appendArray(griddedArray)


    @pyqtSlot()
    def clearData(self):
        """ Clears the repository that holds the panel's array data.
        """
        self.repo.clear()


    @pyqtSlot()
    def onRepoAboutToBeCleared(self):
        """ Is called when the model is about to be cleared.

            The base implementation does nothing. Descendants should override this method so that
            it clears the panel contents
        """
        logger.debug("Repo cleared")


    @pyqtSlot()
    def onRepoItemAboutToBeRemoved(self):
        """ Is called an items will be removed from

            The base implementation does nothing. Descendants should override this method so that
            it clears the panel contents
        """
        pass


    @pyqtSlot(str)
    def onRepoItemChanged(self, name):
        """ Called when an array of this name has changed.

            If checks if the repository contains an repoItem of this name and calls
            self._updateFromRepoItem.
        """
        logger.debug("onRepoItemChanged: {} {}".format(self.name, name))
        if not self.arrayInConfig(name):
            logger.debug("No plot data item for: {}".format(name))
            return

        try:
            repoItem = self.repo.itemByName(name)
        except ItemNotFoundError:
            logger.warning("Panel {}: no repo item found with this name (ignored): {}"
                           .format(self.name, name))
        else:
            self._updateFromRepoItem(repoItem)


    def _updateFromRepoItem(self, repoItem):
        """ Called when a repo item has changed.

            The base implementation does nothing. Descendants should override this method.
        """
        pass


    def setCurrentRow(self, row):
        """ Sets (and selects) the panel item by the row number in the item list (if a config has
            multiple items.)
            This implementation does noting.
        """
        pass

