""" Definitions shared by the quicklook client and server.

"""

import enum
from PyQt5.QtCore import Qt

from gssw.lib.rpcmsg            import Packetizer
import gssw.proto.quicklook_pb2 as ql_proto

# Mime types for dragging an dropping
MIME_GSSW_PANEL = 'application/gssw-panel' # Move and resize quicklook panels.
MIME_ARRAYS_INFO = 'application/gssw-arrays-info'  # To create a panel by dragging array name(s)


@enum.unique
class LineStyleEnum(enum.Enum):
    """ Enumeration class that defines pen styles
    """
    # Please update the corresponding protocol buffer in case of changes.
    #no_pen = Qt.NoPen    # 0  Not used: does not draw a (boundary) line but does fill an area
    solid = Qt.SolidLine  # 1
    dashed = Qt.DashLine  # 2
    dotted = Qt.DotLine   # 3
    dash_dot = Qt.DashDotLine # 4
    #dash_dot_dot = Qt.DashDotDotLine # 5 # not implemented in silx
    no_line = 9999 # Don't draw any curve at all

# Dictionaries to map between the LineStyle and the order in which they occur in the combobox
LINE_STYLE_TO_IDX = dict([(ls, idx) for idx, ls in enumerate(LineStyleEnum)])
IDX_TO_LINE_STYLE = dict([(idx, ls) for idx, ls in enumerate(LineStyleEnum)])

# Short string that describes the linestyle in silx
# http://www.silx.org/doc/silx/latest/modules/gui/plot/plotwidget.html#plot-data
LINE_STYLE_STR = {
    LineStyleEnum.solid: '-',
    LineStyleEnum.dashed: '--',
    LineStyleEnum.dotted: ':',
    LineStyleEnum.dash_dot: '-.',
    LineStyleEnum.no_line: ''
}


#pylint: disable=line-too-long
@enum.unique
class ColorEnum(enum.Enum):
    """ Enumeration class that defines the possible line colors.

        Based on Kenneth Kelly's 22 colors of maximum contrast, but sorted to appear less
        cluttered (Kelly's colors are meant to be used in a certain order to maximize contrast.)

        https://gist.github.com/ollieglass/f6ddd781eeae1d24e391265432297538
        theory - https://eleanormaclure.files.wordpress.com/2011/03/colour-coding.pdf (page 5)
        kelly's colors - https://i.kinja-img.com/gawker-media/image/upload/1015680494325093012.JPG
        hex values - http://hackerspace.kinja.com/iscc-nbs-number-hex-r-g-b-263-f2f3f4-242-243-244-267-22-1665795040
    """
    # Please update the corresponding protocol buffer in case of changes.
    red              = ql_proto.RED
    buf              = ql_proto.BUF
    grey             = ql_proto.GREY
    green            = ql_proto.GREEN
    purplish_pink    = ql_proto.PURPLISH_PINK
    blue             = ql_proto.BLUE
    yellowish_pink   = ql_proto.YELLOWISH_PINK
    violet           = ql_proto.VIOLET
    orange_yellow    = ql_proto.ORANGE_YELLOW
    purplish_red     = ql_proto.PURPLISH_RED
    greenish_yellow  = ql_proto.GREENISH_YELLOW
    reddish_brown    = ql_proto.REDDISH_BROWN
    yellow_green     = ql_proto.YELLOW_GREEN
    yellowish_brown  = ql_proto.YELLOWISH_BROWN
    reddish_orange   = ql_proto.REDDISH_ORANGE
    olive_green      = ql_proto.OLIVE_GREEN
    #white            = ql_proto.WHITE
    black            = ql_proto.BLACK
    yellow           = ql_proto.YELLOW
    purple           = ql_proto.PURPLE
    orange           = ql_proto.ORANGE
    light_blue       = ql_proto.LIGHT_BLUE

    @classmethod
    def getColorOrDefault(cls, key):
        """ Retrieve enum for color key, or return a valid default color, when key is not found. """
        try:
            return cls[key]
        except (KeyError, ValueError):
            # Color 'key' is invalid, send default
            return cls.light_blue

#pylint: enable=line-too-long

# Dictionaries to map between the Color and the order in which they occur in the combobox
COLOR_TO_IDX = dict([(color, idx) for idx, color in enumerate(ColorEnum)])
IDX_TO_COLOR = dict([(idx, color) for idx, color in enumerate(ColorEnum)])


def getColorForCurve(curveIndex):
    """ Rotates through the available curve colors, based on the curve index. """
    return IDX_TO_COLOR[curveIndex % len(IDX_TO_COLOR)]



def createQuicklookPacketizer():
    """ Creates a packetizer for use in the Quicklook client and server

        Currently the same as the Packetizer.createDefaultPacketizer()
    """
    packetizer = Packetizer.createDefaultPacketizer()
    return packetizer




@enum.unique
class BinAlignmentEnum(enum.Enum):
    """ Specifies histogram bins value are used to draw the left, right or center of the bars.
    """
    # Please update the EnumBinAlignment in the quicklook.proto in case of changes.
    left   = 1
    center = 2
    right  = 3


