""" Plot Panels for the Quicklook application
"""
import logging

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt, pyqtSlot

from gssw.lib.errors import GsswExportError, DataError
from gssw.lib.typing import check_class, check_is_a_string
from gssw.lib.widgets.colors import getWidgetColor, PaletteRoleEnum
from gssw.lib.widgets.misc import promptForFileName, gsswUiWarning, gsswUiYesNoQuestion
from gssw.lib.widgets.plotting import GsswSilxPlotWidget, ClearPlotAction, SavePlotAction
from gssw.quicklook.basepanel import AbstractPanel
from gssw.quicklook.defs import ColorEnum, BinAlignmentEnum
from gssw.quicklook.shared import GlobalArrayInfoRepo, ArrayInfoItem

logger = logging.getLogger(__name__)

START_COLOR_IDX = 0 # Start at blue because light blue is a bit too light
SHOW_GRID_BY_DEFAULT = False



class HistPanel(AbstractPanel):
    """ A panel containing a central histogram.

        This class does NOT calculate the histogram. Making the bins and counts are the
        responsibility of the user. The bins and counts must be sent as 1D gridded array where the
        coordinate vector are the bins.
    """

    def __init__(self, name, parent=None):
        """ Constructor

            :param name: panel name string
            :param parent: Qt parent object.
        """
        super().__init__(name, parent=parent)

        self._griddedArray = None  # The gridded array that holds the data. Used for export.
        self._arrayName = ''       # The array name as configured by the users.
        self._color = None
        self._binAlignment = None

        # Set members using the setters so that types are checked.
        self.color = ColorEnum.red
        self.binAlignment = BinAlignmentEnum.center

        self.exportCsvAction = QtWidgets.QAction("Export to CSV...", self)
        self.exportCsvAction.setToolTip("Exports the plot data to file Comma Separated Values")
        self.exportCsvAction.triggered.connect(self.exportToCsv)
        self.exportCsvAction.setShortcut("Ctrl+E")
        self.menu.insertAction(self.saveAsImageAction, self.exportCsvAction)

        ###########
        # Widgets #
        ###########

        self.plot = GsswSilxPlotWidget(parent=self)
        self.plot.setGraphXLabel('')
        self.plot.setGraphYLabel('')
        self.showGrid = SHOW_GRID_BY_DEFAULT
        self.contentsLayout.addWidget(self.plot)

        saveAsImageIndex = 0
        # Override clearAction in the context menu
        for idx in range(len(self.plot.contextMenuActions)):
            if isinstance(self.plot.contextMenuActions[idx], ClearPlotAction):
                self.plot.contextMenuActions[idx] = self.clearAction
            elif isinstance(self.plot.contextMenuActions[idx], SavePlotAction):#SaveAsImageAction):
                # Locate the 'save as image' item, so we can put the csv option next to it
                saveAsImageIndex = idx

        # Put the 'export to csv' option next to the 'save as image' option
        self.plot.contextMenuActions.insert(saveAsImageIndex + 1, self.exportCsvAction)
        self.plot.contextMenuActions.append(self.configAction)


    def _createConfigWidget(self, parent=None):
        """ Creates a config widget that belongs to this class

            :return: a new HistPanelConfigWidget
        """
        return HistPanelConfigWidget(parent=parent)


    @property
    def plotTitle(self):
        """ The title of the plot
        """
        return self.plot.getGraphTitle()


    @plotTitle.setter
    def plotTitle(self, value):
        """ The title of the plot

            If set to the empty string, the plot title will be hidden for more screen space.
            If you don't want a visible title but want want the plot to expand you can use a space
            as the title.
        """
        check_class(value, str)
        self.plot.setGraphTitle(value)


    @property
    def xTitle(self):
        """ The X-axis title
        """
        return self.plot.getXAxis().getLabel()


    @xTitle.setter
    def xTitle(self, value):
        """ The X-axis title

            If set to the empty string, the X-axis title will be hidden for more screen space.
            If you don't want a visible title but want want the plot to expand you can use a space
            as the title.
        """
        check_class(value, str)
        self.plot.getXAxis().setLabel(value)


    @property
    def yTitle(self):
        """ The Y-axis title
        """
        return self.plot.getYAxis().getLabel()


    @yTitle.setter
    def yTitle(self, value):
        """ The Y-axis title

            If set to the empty string, the Y-axis title will be hidden for more screen space.
            If you don't want a visible title but want want the plot to expand you can use a space
            as the title.
        """
        check_class(value, str)
        self.plot.getYAxis().setLabel(value)


    @property
    def showGrid(self):
        """ True the grid lines are shown
        """
        return bool(self.plot.getGraphGrid())


    @showGrid.setter
    def showGrid(self, value):
        """ True the grid lines are shown
        """
        self.plot.setGraphGrid(value)


    @property
    def color(self):
        """ Color for the bars
        """
        return self._color


    @color.setter
    def color(self, value):
        """ Sets the color for the bars
        """
        check_class(value, ColorEnum)
        self._color = value


    @property
    def binAlignment(self):
        """ Specifies if the bin value denote the left, the right or the center of the bars.
        """
        return self._binAlignment


    @binAlignment.setter
    def binAlignment(self, value):
        """ Specifies if the bin value denote the left, the right or the center of the bars.
        """
        check_class(value, BinAlignmentEnum)
        self._binAlignment = value


    @property
    def arrayName(self):
        """ The gridded array name as configured by the user.
        """
        return self._arrayName


    @arrayName.setter
    def arrayName(self, name):
        """ The gridded array name as configured by the user.
        """
        check_is_a_string(name)
        if name != self._arrayName:
            logger.debug("Showing new array {} (old= {!r}). Clearing histogram plot"
                         .format(name, self._arrayName))
            self.clearData() # Clears the plote as well

        self._arrayName = name


    def getConfigArrayNames(self):
        """ Returns a list with the names of the array used in the configuration"""
        return [self.arrayName] if self._arrayName else []


    def arrayInConfig(self, arrayName):
        """ Returns True if this panel is configured to show data of this array.

            :param arrayName: gridded array name
            :return: True if the array name in the list of configured arrays
        """
        return arrayName == self._arrayName


    def updatePanelData(self, griddedArray):
        """ Replaces array and redraws the histogram
        """
        assert self.arrayInConfig(griddedArray.name), \
            "updatePanelData called for array it doesn't show: {}".format(griddedArray.name)

        try:
            self._griddedArray = griddedArray
            values = griddedArray.data
            if values.ndim != 1:
                raise DataError("Data ignored: expected 1D data, got: {}D".format(values.ndim))

            bins = griddedArray.makeCoordinateVector(0)
            if bins.dtype.kind == 'M':
                raise DataError("Data {} ignored: bins cannot be datatime values"
                                .format(griddedArray.name))

            counts = griddedArray.data
            if counts.ndim != 1:
                raise DataError("Data {} ignored: expected 1D counts data, got: {}D"
                                .format(griddedArray.name, counts.dim))

            self._plotHistogramData(bins, counts)

        except Exception as ex:
            logger.error("Unable to set histogram data: {}".format(ex))
            self.clearData()
            raise


    def _plotHistogramData(self, bins, counts):
        """ Sets the histogram data given two 1D numpy arrays: bins and coutns
        """
        # We can't use addCurve here because it doens't pass resetZoom to addHistogram in silx
        self.plot.addHistogram(
            edges = bins, histogram = counts,
            align = self.binAlignment.name,
            fill = True,
            legend = self._arrayName,
            color = "#{:06X}".format(self.color.value),
            resetzoom = not self.plot.isZoomed())

        # This shows how to set the line with if desired.
        # histo = self.plot.getHistogram(self._arrayName)
        # histo.setLineWidth(3)


    @pyqtSlot()
    def clearData(self):
        """ Is called when the model is about to be cleared.
            Clears the plot
        """
        self._griddedArray = None
        self.plot.clear()


    def exportToCsv(self):
        """ Prompts the user for a file name and exports the data to CSV
        """
        logger.debug("exporting: {}".format(self._griddedArray))
        if self._griddedArray is None:
            gsswUiWarning("No data in plot. Export canceled.")
            return

        fileName = promptForFileName(self, self.plotTitle if self.plotTitle else 'export', 'txt',
                                     directory='', suggestDateTime=True,
                                     fileFilter="CSV files (*.txt);; All files (*)")
        if not fileName:
            logger.info("Export canceled")
            return

        try:
            self._griddedArray.exportToCsv(fileName)
        except GsswExportError as ex:
            gsswUiWarning(str(ex))



    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        dct = super().asDict()

        cfg = dict()
        cfg['arrayName'] = self.arrayName
        cfg['plotTitle'] = self.plotTitle
        cfg['xTitle'] = self.xTitle
        cfg['yTitle'] = self.yTitle
        cfg['showGrid'] = self.showGrid
        cfg['color'] = self.color.name
        cfg['binAlignment'] = self.binAlignment.name
        dct['cfg'] = cfg

        return dct


    def setValuesFromDict(self, dct):
        """ Sets panel parameters given a dictionary with (config options)
        """
        super().setValuesFromDict(dct)

        cfg = dct['cfg']

        # We use default values (such as the empty strings) here so that we are robust
        # against changes and so can accept old config files
        self.arrayName = cfg.get('arrayName', '')
        self.plotTitle = cfg.get('plotTitle', '')
        self.xTitle = cfg.get('xTitle', '')
        self.yTitle = cfg.get('yTitle', '')
        self.showGrid = cfg.get('showGrid', SHOW_GRID_BY_DEFAULT)
        self.color = ColorEnum.getColorOrDefault(cfg.get('color', 'grey'))
        self.binAlignment = BinAlignmentEnum[cfg.get('binAlignment', 'center')]


    def configureFromArrayInfoList(self, arrInfoItems):
        """ Configures the panel from the items in the array info list

            ;param list arrInfoItems: list of ArrayInfoItems
        """
        if len(arrInfoItems) != 1:
            gsswUiWarning("You cannot drag multiple arrays onto this panel.")
            return

        arrInfoItem = arrInfoItems[0]
        check_class(arrInfoItem, ArrayInfoItem)

        if self.arrayName and self.arrayName != arrInfoItem.name:
            if not gsswUiYesNoQuestion("Replace current array {!r} with {!r}?"
                                      .format(self.arrayName, arrInfoItem.name)):
                return

        self.arrayName = arrInfoItem.name

        # Set plot labels if all dragged items have the same label values.
        if not self.plotTitle:
            self.plotTitle = arrInfoItem.labelOrName

        if not self.xTitle:
            self.xTitle = arrInfoItem.xLabel

        if not self.yTitle:
            self.yTitle = arrInfoItem.yLabel

        self.execConfigDialog()



class HistPanelConfigWidget(QtWidgets.QWidget):
    """ Contains widgets for configuring HistPanel settings
    """

    def __init__(self, parent=None):
        """ Constructor
        """
        super().__init__(parent=parent)

        self.controlLayout = QtWidgets.QVBoxLayout()
        self.setLayout(self.controlLayout)

        # Details common to all plot items
        self.commonLayout = QtWidgets.QFormLayout()
        self.commonLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.controlLayout.addLayout(self.commonLayout)

        self.titleEdit = QtWidgets.QLineEdit()
        self.commonLayout.addRow("Plot title", self.titleEdit)

        self.xTitleEdit = QtWidgets.QLineEdit()
        self.commonLayout.addRow("X-axis title", self.xTitleEdit)

        self.yTitleEdit = QtWidgets.QLineEdit()
        self.commonLayout.addRow("Y-axis title", self.yTitleEdit)

        self.showGridCheckBox = QtWidgets.QCheckBox()
        self.showGridCheckBox.setChecked(True)
        self.commonLayout.addRow("Show grid", self.showGridCheckBox)

        # Line color
        self.colorComboBox = QtWidgets.QComboBox()
        self.colorComboBox.setFocusPolicy(Qt.StrongFocus)
        self.colorComboBox.setEditable(False)

        for color in ColorEnum:
            # Draw rectangle with color inside combobox.
            iconSizeX, iconSizeY = 12, 12
            pixmap = QtGui.QPixmap(iconSizeX, iconSizeY)
            iconColor = QtGui.QColor(color.value)
            pixmap.fill(iconColor)
            painter = QtGui.QPainter(pixmap)
            painter.setPen(getWidgetColor(self, PaletteRoleEnum.WindowText))
            painter.setBrush(Qt.NoBrush)
            painter.drawRect(QtCore.QRect(0, 0, iconSizeX-1, iconSizeY-1))
            painter.end()
            colorIcon = QtGui.QIcon(pixmap)
            self.colorComboBox.addItem(colorIcon, color.name)

        self.commonLayout.addRow("Color", self.colorComboBox)

        self.binAlignmentComboBox = QtWidgets.QComboBox()
        self.binAlignmentComboBox.setFocusPolicy(Qt.StrongFocus)
        self.binAlignmentComboBox.setEditable(False)

        for binAlignment in BinAlignmentEnum:
            self.binAlignmentComboBox.addItem(binAlignment.name)

        self.commonLayout.addRow("Bin alignment", self.binAlignmentComboBox)

        self.nameComboBox = QtWidgets.QComboBox()
        self.nameComboBox.setFocusPolicy(Qt.StrongFocus)
        self.nameComboBox.setEditable(True)
        self.nameComboBox.setInsertPolicy(QtWidgets.QComboBox.InsertAlphabetically)
        self.nameComboBox.setModel(GlobalArrayInfoRepo.instance().dynamicModel)

        self.commonLayout.addRow("Array name", self.nameComboBox)


    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        cfg = dict()
        cfg['arrayName'] = self.nameComboBox.currentText()
        cfg['plotTitle'] = self.titleEdit.text()
        cfg['xTitle'] = self.xTitleEdit.text()
        cfg['yTitle'] = self.yTitleEdit.text()
        cfg['showGrid'] = self.showGridCheckBox.isChecked()
        cfg['color'] = self.colorComboBox.currentText()
        cfg['binAlignment'] = self.binAlignmentComboBox.currentText()

        return cfg


    def setValuesFromDict(self, cfg):
        """ Creates a panel from a dictionary with values
        """
        logger.debug("setValuesFromDict: {}".format(cfg))
        self.nameComboBox.setCurrentText(cfg.get('arrayName', ''))
        self.titleEdit.setText(cfg.get('plotTitle', ''))
        self.xTitleEdit.setText(cfg.get('xTitle', ''))
        self.yTitleEdit.setText(cfg.get('yTitle', ''))
        self.showGridCheckBox.setChecked(cfg.get('showGrid', SHOW_GRID_BY_DEFAULT))
        self.colorComboBox.setCurrentText(cfg.get('color', 'grey'))
        self.binAlignmentComboBox.setCurrentText(cfg.get('binAlignment', 'center'))


