""" Silx image plot Quicklook component.

"""
import logging

from PyQt5 import  QtWidgets
from PyQt5.QtCore import Qt, pyqtSlot

from silx.gui.colors import Colormap

from gssw.lib.typing import check_class, check_is_a_string
from gssw.lib.errors import DataError
from gssw.lib.widgets.misc import gsswUiWarning, gsswUiYesNoQuestion
from gssw.lib.widgets.plotting import GsswImageWidget
from gssw.lib.widgets.scientificspinbox import ScientificDoubleSpinBox
from gssw.quicklook.basepanel import AbstractPanel
from gssw.quicklook.shared import GlobalArrayInfoRepo, ArrayInfoItem


logger = logging.getLogger(__name__)


BACKEND_MPL = "mpl"  # Matplotlib
BACKEND_GL = "gl"    # OpenGl

DEFAULT_BACKEND = BACKEND_GL


class ImagePanel(AbstractPanel):
    """ A panel containing a central plot.
    """
    def __init__(self, name, parent=None):
        """ Constructor

            :param name: panel name string
            :param parent: Qt parent object.
        """
        super().__init__(name, parent=parent)

        self._arrayName = '' # the name of the array that is currently plotted

        self._colorAutoRange = True
        self._colorRange = (0, 1)
        self.colormap = Colormap(name='viridis') # viridis is a good default.
        self.imageWidget = GsswImageWidget(backend=DEFAULT_BACKEND, colormap=self.colormap)
        self.imageWidget.setGraphXLabel('')
        self.imageWidget.setGraphYLabel('')

        self.contentsLayout.addWidget(self.imageWidget)


    def _createConfigWidget(self, parent=None):
        """ Creates a config widget that belongs to this
            :return: a new PlotPanelConfigWidget
        """
        return ImagePanelConfigWidget(parent=parent)


    @property
    def plotTitle(self):
        """ The title of the plot
        """
        return self.imageWidget.getGraphTitle()


    @plotTitle.setter
    def plotTitle(self, value):
        """ The title of the plot
        """
        check_class(value, str)
        self.imageWidget.setGraphTitle(value)


    @property
    def xTitle(self):
        """ The x-axis title
        """
        return self.imageWidget.getGraphXLabel()


    @xTitle.setter
    def xTitle(self, value):
        """ The x-axis title
        """
        check_class(value, str)
        self.imageWidget.setGraphXLabel(value)


    @property
    def yTitle(self):
        """ The y-axis title
        """
        return self.imageWidget.getGraphYLabel()


    @yTitle.setter
    def yTitle(self, value):
        """ The y-axis title
        """
        check_class(value, str)
        self.imageWidget.setGraphYLabel(value)


    @property
    def barTitle(self):
        """ The label in the color bar
        """
        return self.imageWidget.getBarTitle()


    @barTitle.setter
    def barTitle(self, value):
        """ The label in the color bar
        """
        check_class(value, str)
        self.imageWidget.setBarTitle(value)


    @property
    def colormapName(self):
        """ The name of the current colorbar
        """
        return self.imageWidget.colormap.getName()


    @colormapName.setter
    def colormapName(self, name):
        """ Sets the colormap by name
        """
        self.imageWidget.colormap.setName(name)


    @property
    def colorAutoRange(self):
        """ If True the color range is calculated from the data. If false it is set by the user.
        """
        return self._colorAutoRange


    @colorAutoRange.setter
    def colorAutoRange(self, value):
        """ If True the color range is calculated from the data. If false it is set by the user.
        """
        self._colorAutoRange = bool(value) # Store locally to keep independent from color range.
        if self._colorAutoRange:
            self.colormap.setVRange(None, None)
        else:
            vMin, vMax = self._colorRange
            self.colormap.setVRange(vMin, vMax)


    @property
    def colorRange(self):
        """ The color range. A tuple with (vMin, vMax)
        """
        return self._colorRange


    @colorRange.setter
    def colorRange(self, value):
        """ The color range. A tuple with (vMin, vMax)
        """
        # Store in local member here as well, otherwise the values are lost during auto range.
        self._colorRange = value
        vMin, vMax = value  # Ensures value is list or tuple with 2 elements.
        if self._colorAutoRange:
            self.colormap.setVRange(None, None)
        else:
            self.colormap.setVRange(vMin, vMax)


    @colormapName.setter
    def colormapName(self, name):
        """ Sets the colormap by name
        """
        self.imageWidget.colormap.setName(name)


    @barTitle.setter
    def barTitle(self, value):
        """ The label in the color bar
        """
        check_class(value, str)
        self.imageWidget.setBarTitle(value)


    @property
    def arrayName(self):
        """ The gridded array name as configured by the user.
        """
        return self._arrayName


    @arrayName.setter
    def arrayName(self, name):
        """ The gridded array name as configured by the user.
        """
        check_is_a_string(name)
        if name != self._arrayName:
            logger.debug("Showing new array {} (old= {!r}). Clearing histogram plot"
                         .format(name, self._arrayName))
            self.clearData() # Clears the plote as well

        self._arrayName = name


    def arrayInConfig(self, arrayName):
        """ Returns True if this panel is configured to show data of this array.

            :param arrayName: gridded array name
            :return: True if the arrayName is equal to the array that is to be shown as configured
            by the user.
        """
        return arrayName == self._arrayName


    def getConfigArrayNames(self):
        """ Returns a list with the names of the array used in the configuration"""
        return [self.arrayName] if self._arrayName else []



    def updatePanelData(self, griddedArray):
        """ Replaces array and redraws the histogram
        """
        assert self.arrayInConfig(griddedArray.name), \
            "updatePanelData called for array it doesn't show: {}".format(griddedArray.name)

        try:
            img = griddedArray.data
            if img.ndim != 2:
                raise DataError("Array should be 2D. Got: {}D".format(img.ndim))

            self.imageWidget.addImage(img, resetzoom = not self.imageWidget.isZoomed())

        except Exception as ex:
            logger.error("Unable to set histogram data: {}".format(ex))
            self.clearData()
            raise



    @pyqtSlot()
    def clearData(self):
        """ Is called when the model is about to be cleared.
            Clears the plot
        """
        self.imageWidget.clear()


    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        dct = super().asDict()

        cfg = {} # Configuration subdict

        cfg['arrayName'] = self._arrayName
        cfg['plotTitle'] = self.plotTitle
        cfg['xTitle'] = self.xTitle
        cfg['yTitle'] = self.yTitle
        cfg['barTitle'] = self.barTitle
        cfg['colormapName'] = self.colormapName
        cfg['colorAutoRange'] = self.colorAutoRange
        cfg['colorRange'] = list(self.colorRange)

        dct['cfg'] = cfg

        return dct


    def setValuesFromDict(self, dct):
        """ Sets panel parameters given a dictionary with (config options)
        """
        super().setValuesFromDict(dct)

        cfg = dct['cfg']

        # We use default values (such as the empty strings) here so that we are robust
        # against changes and so can accept old config files
        self.arrayName = cfg.get('arrayName', '')
        self.plotTitle = cfg.get('plotTitle', '')
        self.xTitle = cfg.get('xTitle', '')
        self.yTitle = cfg.get('yTitle', '')
        self.barTitle = cfg.get('barTitle', '')
        self.colormapName = cfg.get('colormapName', 'gray')
        self.colorAutoRange = cfg.get('colorAutoRange', True)
        self.colorRange = cfg.get('colorRange', [0.0, 1.0])



    def configureFromArrayInfoList(self, arrInfoItems):
        """ Configures the panel from the items in the array info list

            ;param list arrInfoItems: list of ArrayInfoItems
        """
        if len(arrInfoItems) != 1:
            gsswUiWarning("You cannot drag multiple arrays onto this panel.")
            return

        arrInfoItem = arrInfoItems[0]
        check_class(arrInfoItem, ArrayInfoItem)

        if self.arrayName and self.arrayName != arrInfoItem.name:
            if not gsswUiYesNoQuestion("Replace current array {!r} with {!r}?"
                                      .format(self.arrayName, arrInfoItem.name)):
                return

        self.arrayName = arrInfoItem.name

        # Set plot labels if all dragged items have the same label values.
        if not self.plotTitle:
            self.plotTitle = arrInfoItem.labelOrName

        if not self.xTitle:
            self.xTitle = arrInfoItem.xLabel

        if not self.yTitle:
            self.yTitle = arrInfoItem.yLabel

        self.execConfigDialog()



class ImagePanelConfigWidget(QtWidgets.QWidget):
    """ Contains widgets for configuring PlotPanel settings
    """

    def __init__(self, parent=None):
        """ Constructor
        """
        super().__init__(parent=parent)

        self.controlLayout = QtWidgets.QVBoxLayout()
        self.setLayout(self.controlLayout)

        # Details common to all plot items
        self.commonLayout = QtWidgets.QFormLayout()
        self.commonLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.controlLayout.addLayout(self.commonLayout)

        self.titleEdit = QtWidgets.QLineEdit()
        self.commonLayout.addRow("Plot title", self.titleEdit)

        self.xTitleEdit = QtWidgets.QLineEdit()
        self.commonLayout.addRow("X-axis title", self.xTitleEdit)

        self.yTitleEdit = QtWidgets.QLineEdit()
        self.commonLayout.addRow("Y-axis title", self.yTitleEdit)

        self.barTitleEdit = QtWidgets.QLineEdit()
        self.commonLayout.addRow("Color Bar title", self.barTitleEdit)

        self.colormapComboBox = QtWidgets.QComboBox()
        self.colormapComboBox.setFocusPolicy(Qt.StrongFocus)
        self.colormapComboBox.setEditable(False)
        self.colormapComboBox.setInsertPolicy(QtWidgets.QComboBox.InsertAlphabetically)
        self.commonLayout.addRow("Color Map", self.colormapComboBox)
        for name in Colormap.getSupportedColormaps():
            self.colormapComboBox.addItem(name)

        self.colorAutoRangeCheckBox = QtWidgets.QCheckBox()
        self.commonLayout.addRow("Auto color range", self.colorAutoRangeCheckBox)
        self.colorAutoRangeCheckBox.stateChanged.connect(self.updateAutoRangeEnabled)

        self.colorRangeMinSpinBox = ScientificDoubleSpinBox()
        self.colorRangeMinSpinBox.setMinimum(-1e100)
        self.colorRangeMinSpinBox.setMaximum(1e100)
        self.colorRangeMinSpinBox.setSingleStep(1)
        self.colorRangeMinSpinBox.setKeyboardTracking(False)

        self.colorRangeMaxSpinBox = ScientificDoubleSpinBox()
        self.colorRangeMaxSpinBox.setMinimum(-1e100)
        self.colorRangeMaxSpinBox.setMaximum(1e100)
        self.colorRangeMaxSpinBox.setSingleStep(1)
        self.colorRangeMaxSpinBox.setKeyboardTracking(False)

        self.colorRangeLayout = QtWidgets.QHBoxLayout()
        self.colorRangeLayout.setContentsMargins(0, 0, 0, 0)
        self.colorRangeLayout.addWidget(self.colorRangeMinSpinBox)
        self.colorRangeLayout.addWidget(self.colorRangeMaxSpinBox)
        self.commonLayout.addRow("Color range", self.colorRangeLayout)

        self.nameComboBox = QtWidgets.QComboBox()
        self.nameComboBox.setFocusPolicy(Qt.StrongFocus)
        self.nameComboBox.setEditable(True)
        self.nameComboBox.setInsertPolicy(QtWidgets.QComboBox.InsertAlphabetically)
        self.nameComboBox.setModel(GlobalArrayInfoRepo.instance().dynamicModel)
        self.commonLayout.addRow("Array name", self.nameComboBox)


    def finalize(self):
        """ Cleanup resources. Must be called explicitly.
        """
        pass


    def updateAutoRangeEnabled(self, _checkBoxState=None):
        """ Enables or disables the color range spinboxes depending on if autorange is checked.
        """
        enabled = not self.colorAutoRangeCheckBox.isChecked()
        self.colorRangeMinSpinBox.setEnabled(enabled)
        self.colorRangeMaxSpinBox.setEnabled(enabled)


    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        cfg = {}
        cfg['arrayName'] = self.nameComboBox.currentText()
        cfg['plotTitle'] = self.titleEdit.text()
        cfg['xTitle'] = self.xTitleEdit.text()
        cfg['yTitle'] = self.yTitleEdit.text()
        cfg['barTitle'] = self.barTitleEdit.text()
        cfg['colormapName'] = self.colormapComboBox.currentText()
        cfg['colorAutoRange'] = self.colorAutoRangeCheckBox.isChecked()
        cfg['colorRange'] = [self.colorRangeMinSpinBox.value(), self.colorRangeMaxSpinBox.value()]

        return cfg


    def setValuesFromDict(self, cfg):
        """ Creates a panel from a dictionary with values
        """
        self.nameComboBox.setCurrentText(cfg.get('arrayName', ''))
        self.titleEdit.setText(cfg.get('plotTitle', ''))
        self.xTitleEdit.setText(cfg.get('xTitle', ''))
        self.yTitleEdit.setText(cfg.get('yTitle', ''))
        self.barTitleEdit.setText(cfg.get('barTitle', ''))
        self.colormapComboBox.setCurrentText(cfg.get('colormapName', ''))
        self.colorAutoRangeCheckBox.setChecked(cfg.get('colorAutoRange', True))
        colorRangeMin, colorRangeMax = cfg.get('colorRange', [0.0, 1.0])
        self.colorRangeMinSpinBox.setValue(colorRangeMin)
        self.colorRangeMaxSpinBox.setValue(colorRangeMax)
        self.updateAutoRangeEnabled()



