""" Main GUI Window """
# pylint: disable=too-many-lines
# Hard to make it shorter in this case

import collections
import os.path
import logging
import yaml

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtSlot, Qt


from gssw.lib.config import Config
from gssw.lib.arrays.gridded import griddedArrayFromBinMessage, GriddedArrayMsg
from gssw.lib.rpcmsg import DelegatingQtReqRepService, DelegatingQtPubSubService
from gssw.lib.sockets import tagName
from gssw.lib.socketsqt import NotifyingSubSocket, NotifyingRepSocket
from gssw.lib.startup import makeAddress
from gssw.lib.typing import check_class, check_is_a_string
from gssw.lib.version import VERSION_STRING
from gssw.lib.errors import DuplicateItemError, ItemNotFoundError
from gssw.lib.widgets.colors import widgetHasDarkBackGround
from gssw.lib.widgets.misc import gsswUiError, saveWidgetAsImage

from gssw.quicklook.basepanel import createPanelByClassName, AbstractPanel
from gssw.quicklook.defs import LineStyleEnum, ColorEnum, BinAlignmentEnum
from gssw.quicklook.pages import BasePage, FreeLayoutPage, PageOverviewWidget
from gssw.quicklook.plotpanel import CurveItem, PlotPanel
from gssw.quicklook.repo import AppendModeEnum
from gssw.quicklook.shared import GlobalArrayInfoRepo, ConfigArrayInfoWidget, LiveArrayInfoWidget

from gssw.proto.quicklook_pb2 import QuicklookCommandServiceSpec, QuicklookDataServiceSpec

logger = logging.getLogger(__name__)

_TEST_PANEL_NAME = 'plot-panel-1'

PROGRAM_NAME = 'GSSw-quicklooker'




class QuicklookerPubSubService(DelegatingQtPubSubService):
    """ Class that overrides the processBinaryMessage of the DelegatingQtPubSubService

        Is typically not necessary for other servers, but we override the processBinaryMessage here
        so that we can process GriddedArrayMsg
    """
    def processBinaryMessage(self, tag, binMessage):
        """ Processes a message.

            In case the message consists of a single array the array is appended
            This is useful for debugging, e.g. with the ts_puslisher
        """
        logger.debug("Processing message: {}".format(tag))
        if tag == tagName(GriddedArrayMsg):
            array = griddedArrayFromBinMessage(binMessage)
            logger.debug("Received array: {}".format(array))

            self.worker.appendArray(array)
            logger.debug("Appended array: {}".format(array))

            return None, None
        
        return super().processBinaryMessage(tag, binMessage)


# Many attributes are inherit to Qt constructors.
# pylint: disable=too-many-instance-attributes

# The constructor is split up in _setupViews, setupActions, setupMenu
# pylint: disable=attribute-defined-outside-init

class MainWindow(QtWidgets.QMainWindow):
    """ The main application.
    """
    def __init__(self, pagesConfigName, reset=False,
                 repSocketAddress=None, subSocketAddress=None):
        """ Constructor

            :param reset: If True, persistent settings (e.g. window position) will be reset to
                their default values.
        """
        super().__init__()

        # Set WA_DeleteOnClose to make sure that the objects are deleted in the correct order.
        # Otherwise I got the following error:
        #   QBasicTimer::start: QBasicTimer can only be used with threads started with QThread
        # See also https://stackoverflow.com/a/27178019/625350
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        ## Service ##

        self._repSocket = NotifyingRepSocket(parent=self)
        self._repSocket.createZmqSocket()
        if repSocketAddress is not None:
            self.bindRepSocket(repSocketAddress)

        self._subSocket = NotifyingSubSocket(parent=self)
        self._subSocket.createZmqSocket()
        self._subSocket.subscribe('') # Subscribe to all topics
        if subSocketAddress is not None:
            self.connectSubSocket(subSocketAddress)

        # Create a request-reply service that will listen to the sockets for messages that are
        # defined in the QuicklookService. It will look up the corresponding methods of the
        # MainWindow and call them, using the fields of the message as parameter, and send a reply
        # message from the result of the method call. For details see:
        # http://softwaredocs.sron.nl/gssw/devguide/communication.html#gssw-rpc-classes-python

        worker = self  # The service will call methods of the MainWindow to handle the message.
        self.reqRepService = DelegatingQtReqRepService(worker, QuicklookCommandServiceSpec(),
                                                       self._repSocket)

        self.pubSubService = QuicklookerPubSubService(worker, QuicklookDataServiceSpec(),
                                                      self._subSocket)

        ## Model ##

        self._currentPage = None
        self._pagesConfigName = os.path.realpath(pagesConfigName)
        self._pages = {}

        ## Views ##
        self._appendTestMethod = True

        self._setupActions()
        self._setupViews()
        self._setupMenu()
        self._updateWindowTitle()

        ## Dialogs ##

        nameListModel = GlobalArrayInfoRepo.instance().staticModel
        self.configNameListDialog = ConfigArrayInfoWidget(nameListModel, parent=None)

        ## Read persistent settings ##

        self._readViewSettings(reset)
        self._settingsSaved = False

        if not os.path.exists(self._pagesConfigName):
            # This can happen the first time you execute the quicklooker with a new config.yaml
            logger.warning("Config file not found. Creating empty config file: {}"
                           .format(self._pagesConfigName))

            # Try if we can write to a config file instead of failing at program exit.
            with open(self._pagesConfigName, 'w') as configFile:
                print("# Empty config file", file=configFile)
        else:
            self._readConfigFromYaml(self._pagesConfigName)

        self.onPageConfigChanged() # will update the live array model with the plot config counte

        # Select first page upon startup, if there are already pages configured
        #   otherwise, the user is presented with a button to create a new page,
        #   which can lead to many pages created by accident
        allPageNames = self.pageOverview.getAllPageNames()
        if not self.currentPage and allPageNames: # Only if there is no page remembered from session
            self.setCurrentPage(allPageNames[0])


        self.pageOverview.pageListWidget.currentItemChanged.connect(self.onCurrentPageItemChanged)

        # The persistent of the program is saved during exist. However, since the program can be
        # terminated by a SIGTERM or SIGKILL from run_gssw we save the persistent state every 5
        # seconds. Alternative solutions were considered: Using signal handlers is not robust in
        # practice (and possibly brittle under Windows). Using Qt signals/slots would require too
        # many connects to handle all cases where the state has changed.
        # 2018-06-26: The gssw.lib.application.Application object now has a timer that forces
        # the return to the Python interpreter every 0.1 seconds. This makes catching Linux signals
        # work even if the Qt application is idle. Perhaps now Linux signal handling is more robust
        # and we don't need to write the state every five seconds. For now we disable the timer
        # below. If this works we will remove it.
        # signals
        self._saveStateTimer = QtCore.QTimer(self)
        self._saveStateTimer.timeout.connect(self._writeViewSettings)
        self._saveStateTimer.timeout.connect(self._writeConfigToYaml)
        # self._saveStateTimer.start(5000) # disabled


    def finalize(self):
        """ Cleanup resources.
        """
        logger.debug("Cleaning up resources.")

        # Stop timer, otherwise it may fire after the window has closed and thus save settings,
        # This gives a RuntimeError: wrapped C/C++ object of type FreeLayoutPage has been deleted
        self._saveStateTimer.stop()

        # Disconnect signals
        self._saveStateTimer.timeout.disconnect(self._writeViewSettings)
        self._saveStateTimer.timeout.disconnect(self._writeConfigToYaml)
        self.pageOverview.pageListWidget.currentItemChanged.disconnect(
            self.onCurrentPageItemChanged)

        for page in self._pages.values():
            page.finalize()

            # Delete pages because they have no parent (probably not necessary but just to be sure).
            page.deleteLater()


    def _setupActions(self):
        """ Sets up the actions.
        """
        self._contextActionGroup = QtWidgets.QActionGroup(self)

        # Actions
        self.addNewPageAction = QtWidgets.QAction("New Page", self)
        self.addNewPageAction.setToolTip("Adds a new page to the list")
        #self.addNewPageAction.setShortcut(QtGui.QKeySequence("Ctrl+P")) #Disabled, can be confusing
        self.addNewPageAction.triggered.connect(self.addNewPage)
        self._contextActionGroup.addAction(self.addNewPageAction)
        self.addAction(self.addNewPageAction)

        self.removeCurrentPageAction = QtWidgets.QAction("Remove Current Page", self)
        self.removeCurrentPageAction.setToolTip("Remove selected page from the list")
        self.removeCurrentPageAction.triggered.connect(self.removePage)
        self._contextActionGroup.addAction(self.removeCurrentPageAction)

        self.renameCurrentPageAction = QtWidgets.QAction("Rename Current Page...", self)
        self.renameCurrentPageAction.setToolTip("Rename the selected page in the list")
        self.renameCurrentPageAction.triggered.connect(self.renameCurrentPage)
        self._contextActionGroup.addAction(self.renameCurrentPageAction)

        self.saveCurrentPageAsImageAction = QtWidgets.QAction("Save Page As Image...", self)
        self.saveCurrentPageAsImageAction.setToolTip("Save current page as image")
        self.saveCurrentPageAsImageAction.triggered.connect(self.saveCurrentPageAsImage)
        self._contextActionGroup.addAction(self.saveCurrentPageAsImageAction)

        self.addLinePlotAction = QtWidgets.QAction("New Line-Plot", self)
        self.addLinePlotAction.setToolTip("Adds a new line plot panel to the current page")
        self.addLinePlotAction.setShortcut(QtGui.QKeySequence("Ctrl+L"))
        self.addLinePlotAction.triggered.connect(self.addPlotPanel)
        # self.addLinePlotAction.triggered.connect(
        #     lambda: self._createAndAddPanel(fullClassName = 'gssw.quicklook.plotpanel.PlotPanel'))
        self._contextActionGroup.addAction(self.addLinePlotAction)

        self.addHistPlotAction = QtWidgets.QAction("New Histogram", self)
        self.addHistPlotAction.setToolTip("Adds a new histogram panel to the current page")
        self.addHistPlotAction.setShortcut(QtGui.QKeySequence("Ctrl+J"))
        self.addHistPlotAction.triggered.connect(self.addHistogramPanel)
        self._contextActionGroup.addAction(self.addHistPlotAction)

        self.addImagePlotAction = QtWidgets.QAction("New Image-Plot", self)
        self.addImagePlotAction.setToolTip("Adds a new Silx image plot panel to the current page")
        self.addImagePlotAction.setShortcut(QtGui.QKeySequence("Ctrl+I"))
        self.addImagePlotAction.triggered.connect(self.addImagePanel)
        self._contextActionGroup.addAction(self.addImagePlotAction)

        self.addTableAction = QtWidgets.QAction("New Table", self)
        self.addTableAction.setToolTip("Adds a new line table to the current page")
        self.addTableAction.setShortcut(QtGui.QKeySequence("Ctrl+T"))
        self.addTableAction.triggered.connect(
            lambda: self._createAndAddPanel(fullClassName ='gssw.quicklook.tablepanel.TablePanel'))
        self._contextActionGroup.addAction(self.addTableAction)

        # The 'known' arrays will always appear in the editable comboxes when the user is prompted
        # to add arrays to a panel.
        self.configKnownArrayNamesAction = QtWidgets.QAction("Edit Known Array Names...", self)
        self.configKnownArrayNamesAction.setToolTip("Configure the list of known arrays.")
        #self.configKnownArrayNamesAction.setShortcut(QtGui.QKeySequence("Ctrl+B"))
        self.configKnownArrayNamesAction.triggered.connect(self.execConfigureArraysDialog)


    def _setupViews(self):
        """ Creates the UI widgets. 
        """
        # Create an empty widget for holding the current page. We don't set the pages directly
        # as the central widget so that the widget size is retained when a new page is selected.
        widget = QtWidgets.QWidget()
        widget.setMinimumSize(100, 100)
        self.pagesLayout = QtWidgets.QStackedLayout() # This is a Qt stack where only one of its
        #                                             # widgets is visible at a time (selected one)
        widget.setLayout(self.pagesLayout)

        self.noPageButton = QtWidgets.QPushButton("No page selected.\nClick to create a new page.")
        self.noPageButton.clicked.connect(self.addNewPage)
        self.pagesLayout.addWidget(self.noPageButton)

        self.pagesLayout.addWidget(self.noPageButton)
        self.pagesLayout.setContentsMargins(0, 0, 0, 0)
        self.setCentralWidget(widget)
        self._widget = widget

        # Pages overview dock widget
        self.pageOverview = PageOverviewWidget()

        self.pageDock = QtWidgets.QDockWidget("Pages", self)
        self.pageDock.setObjectName("page_overview_dock") # needed for saveState
        self.pageDock.setWidget(self.pageOverview)
        self.pageDock.setAllowedAreas(QtCore.Qt.AllDockWidgetAreas)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.pageDock)
        self.pageDock.toggleViewAction().setChecked(True)

        self.pageOverview.pageAddButton.setDefaultAction(self.addNewPageAction)
        self.pageOverview.pageRemoveButton.setDefaultAction(self.removeCurrentPageAction)

        self.dataOverView = LiveArrayInfoWidget(model=GlobalArrayInfoRepo.instance().dynamicModel)

        self.dataDock = QtWidgets.QDockWidget("Data Arrays", self)
        self.dataDock.setObjectName("data_overview_dock") # needed for saveState
        self.dataDock.setWidget(self.dataOverView)
        self.dataDock.setAllowedAreas(QtCore.Qt.AllDockWidgetAreas)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.dataDock)
        self.dataDock.toggleViewAction().setChecked(True)

        # Set icons depending on the color. We set the Icons on the action because setting it on 
        # the QToolButton it won't show. Setting it via a stylesheet (on the border-image) will 
        # lose the button shading and make it flat:-(
        app = QtWidgets.QApplication.instance()
        if widgetHasDarkBackGround(self):
            iconDir = os.path.join(app.iconPath, 'snipicons', 'white')
            pageAddIconName = os.path.join(iconDir, 'plus-sign-i.svg')
            pageDelIconName = os.path.join(iconDir, 'minus-sign-i.svg')
        else:
            iconDir = os.path.join(app.iconPath, 'snipicons', 'black')
            pageAddIconName = os.path.join(iconDir, 'plus-sign-m.svg')
            pageDelIconName = os.path.join(iconDir, 'minus-sign-m.svg')

        assert os.path.exists(pageAddIconName), "File not found: {}".format(pageAddIconName)
        assert os.path.exists(pageDelIconName), "File not found: {}".format(pageDelIconName)
        self.addNewPageAction.setIcon(QtGui.QIcon(pageAddIconName))
        self.removeCurrentPageAction.setIcon(QtGui.QIcon(pageDelIconName))



    def _setupMenu(self):
        """ Sets up the main menu.
        """
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.showContextMenu)

        fileMenu = self.menuBar().addMenu("&File")
        fileMenu.addAction("E&xit", self.quitApplication, "Ctrl+Q")

        if self._appendTestMethod:
            fileMenu.addSeparator()
            fileMenu.addAction("&Test", self.myTest, "Shift+Ctrl+T")

        self.pageMenu = self.menuBar().addMenu("&Page")
        self.pageMenu.addAction(self.addNewPageAction)
        self.pageMenu.addAction(self.removeCurrentPageAction)
        self.pageMenu.addAction(self.renameCurrentPageAction)
        self.pageMenu.addSeparator()
        self.pageMenu.addAction(self.saveCurrentPageAsImageAction)
        self.pageMenu.addSeparator()
        self.pageMenu.addAction(self.addLinePlotAction)
        self.pageMenu.addAction(self.addHistPlotAction)
        self.pageMenu.addAction(self.addImagePlotAction)
        self.pageMenu.addAction(self.addTableAction)

        self.panelsMenu = self._emptyPanelsMenu = QtWidgets.QMenu("Panels", self)
        self._currentPanelsMenuAction = self.menuBar().addMenu(self.panelsMenu)

        self.viewMenu = self.menuBar().addMenu("&View")
        self.viewMenu.addAction(self.pageDock.toggleViewAction())
        self.viewMenu.addAction(self.dataDock.toggleViewAction())

        cfgMenu = self.menuBar().addMenu("&Configure")
        cfgMenu.addAction(self.configKnownArrayNamesAction)

        self.menuBar().addSeparator()
        self.helpMenu = self.menuBar().addMenu("&Help")
        self.helpMenu.addAction('&About...', self.about)


    def showContextMenu(self, pos):
        """ Shows the context menu at position pos.
        """
        logger.debug("showContextMenu")
        contextMenu = QtWidgets.QMenu()

        for action in self._contextActionGroup.actions():
            contextMenu.addAction(action)

        contextMenu.addSeparator()
        if self.currentPage:
            contextMenu.addAction(self.currentPage.panelsMenu.menuAction())

        contextMenu.exec_(self.mapToGlobal(pos))


    @classmethod
    def createFromConfig(cls, name, config, reset=False):
        """ Creates a main window given quicklooker dictionary from the config.yaml

            :param name: component name in the module
            :param cfg: gssw.lib.config.Config
            :param reset: reset parameter forwarded to MainWindow constructor
            :return: new MainWinndow object
        """
        check_class(config, Config)

        qlCfg = config.componentParameters(name)
        proxyCfg = config.componentParameters('proxy')
        socketParams = qlCfg['sockets']

        # Get the address of the data proxy, via which all data is received
        proxySocketParams = proxyCfg['sockets']
        repSocketAddress = makeAddress(socketParams['ql-rep'], host='*')
        subSocketAddress = makeAddress(proxySocketParams['xpub'], host=proxyCfg['hostname'])

        pagesConfigName = config.asAbsolutePath(qlCfg['pageConfig'])

        mainWindow = MainWindow(
            reset=reset,
            pagesConfigName=pagesConfigName,
            repSocketAddress=repSocketAddress,
            subSocketAddress=subSocketAddress)

        return mainWindow


    def sizeHint(self):
        """ Holds the recommended size for the widget.
        """
        return QtCore.QSize(800, 600)

    #########
    # Model #
    #########

    def bindRepSocket(self, address):
        """ Binds the Reply socket to the address
        """
        self._repSocket.bindSocket(address)


    def connectSubSocket(self, address):
        """ Connects the Subscribe socket to the address
        """
        self._subSocket.connectSocket(address)


    # Keep this to be able to use the quicklooker without a proxy (in future projects and testing).
    def bindSubSocket(self, address):
        """ Binds the Subscribe socket to the address
        """
        self._subSocket.bindSocket(address)


    ###########
    # Service #
    ###########


    def setCurrentPage(self, pageName):
        """ Selects the page with this name in the central window and the page overview list.

            If the pageName is empty, no page will be selected.

            This function is not a property because it does do a fair bit of work.

            Raises a ItemNotFoundError if no page with this name exists
        """
        logger.debug('setCurrentPage: {}'.format(pageName), stack_info=False)
        if pageName is None:
            pageName = ''

        if self._currentPage and pageName == self._currentPage.name:
            logger.debug("Page {!r} is already the current.".format(pageName))
            self._updateWindowTitle()  # Update title in case the page has just been renamed.
            return

        # Raises KeyError if page doesn't exist
        newPage = self._getPageByName(pageName) if pageName else None

        if newPage:
            logger.debug("Setting current page to: {!r}".format(pageName))

            self._currentPage = newPage

            item = self.pageOverview.pageListWidget.findItems(pageName, Qt.MatchExactly)[0]
            self.pageOverview.pageListWidget.setCurrentItem(item)
            row = self.pageOverview.pageListWidget.row(item)
        else:
            self._currentPage = None
            self.pageOverview.pageListWidget.setCurrentItem(None)
            row = -1

        self._updateWindowTitle()
        logger.debug("Setting page index to: {}".format(row))
        self.pagesLayout.setCurrentIndex(row + 1) # first page is a label

        # Update panels menu
        self.menuBar().removeAction(self._currentPanelsMenuAction)
        if self._currentPage is None:
            self.panelsMenu = self._emptyPanelsMenu
        else:
            self.panelsMenu = self._currentPage.panelsMenu
        self._currentPanelsMenuAction = self.menuBar().insertMenu(
            self.viewMenu.menuAction(), self.panelsMenu)

        # Disable redrawing in the other pages, to improve performance
        for page in self._pages.values():
            page.suspendDrawing()

        if self._currentPage is not None:
            # Start (re)drawing in the now active page
            self._currentPage.resumeDrawing()


    #### pages CRUD methods ####

    def makeUniquePageName(self):
        """ Creates a pagel name that does not exist yet.

            Returns 'Page-X' where is X is the first page number that does not exist yet.
        """
        pageNames = [page.name for page in self._pages.values()]
        for duplicateNr in range(1000):
            name = "Page-{}".format(duplicateNr)
            if name not in pageNames:
                return name

        raise RecursionError("Infinite loop detected") # sanity check


    @pyqtSlot()
    def addNewPage(self, pageName=None, row=None, select=True, errorIfExists=False):
        """ Creates an empty page and adds it to the page list.

            Makes the new page the current current page

            :param str pageName: name of the created page.
                In case a page with this name exists an warning is logged.
                If pageName is None, the new page will be named 'page-X', where X is the first
                number that yields a unique name.
            :param int row: optional row number at which the page will be inserted. If None, the
                page will be inserted at the end.
            :param bool select: If True (the default) it will be selected and set as the current.
            :param bool errorIfExists: if True a DuplicateItemError is raised if a page with this
                name already exists. If False (the default) the old page is removed and a warning
                is logged.
            :return: the name of the newly created page. The empty string if no new page has been
                created.
        """
        check_is_a_string(pageName, allow_none=True)

        if pageName is None:
            pageName = self.makeUniquePageName()

        if pageName in self._pages:
            msg = "A Page with this name already exists: {}".format(pageName)
            if errorIfExists:
                raise DuplicateItemError(msg)
            else:
                logger.info("%s (reusing it, using exiting one).", msg)
                createdPageName = ''
                #self.removePage(pageName)
        else:
            self._insertPage(FreeLayoutPage(pageName), row=row)
            createdPageName = pageName

        if select:
            self.setCurrentPage(pageName)

        return createdPageName


    def _insertPage(self, page=None, row=None):
        """ Inserts a page to the list,

            Does the actual work.

            Site effect: Will set the page.mainWindow object to this MainWindow (self).

            :param page: a BasePage descendant.
            :param row: row number at which the page will be inserted.
        """
        check_class(page, BasePage)
        logger.debug("Adding page: {}".format(page.name))
        assert page.name not in self._pages, "Page already exists"
        assert page.mainWindow is None, "Page is already part of: {}".format(page.mainWindow)

        listItem = QtWidgets.QListWidgetItem(page.name)
        if row is None:
            self.pageOverview.pageListWidget.addItem(listItem)
            self.pagesLayout.addWidget(page)
        else:
            self.pageOverview.pageListWidget.insertItem(row, listItem)
            self.pagesLayout.insertWidget(row + 1, page) # First item is a label

        self._pages[page.name] = page

        page.mainWindow = self

        page.sigPageConfigChanged.connect(self.onPageConfigChanged)
        #self.onPageConfigChanged()


    @pyqtSlot()
    def removePage(self, pageName=None):
        """ Deletes a page given its name.

            :param str pageName: the name of the page that is deleted. If None the current selected
                page (if any) is deleted. In that case the user is asked for confirmation.
        """
        if pageName is None:
            currentItem = self.pageOverview.pageListWidget.currentItem()

            if not currentItem:
                logger.debug("No item selected for removal")
                return

            pageName = currentItem.text()

            reply = QtWidgets.QMessageBox.question(
                self, "Confirm Remove Current Page",
                "Do you want to remove the current page: {}?\n\nThere is no undo!".format(pageName),
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

            if reply != QtWidgets.QMessageBox.Yes:
                logger.info("Remove current page cancelled by user.")
                return

            logger.debug("Removing page: {}".format(pageName))

        if pageName not in self._pages:
            logger.debug("Page that is to be deleted does not exist: {}".format(pageName))
            return

        items = self.pageOverview.pageListWidget.findItems(pageName, Qt.MatchExactly)
        assert len(items) == 1, \
            "Expected 1 list item with name {}, got {}".format(pageName, len(items))

        page = self._pages[pageName]
        page.mainWindow = None
        page.sigPageConfigChanged.disconnect(self.onPageConfigChanged)
        self.pagesLayout.removeWidget(page)
        del self._pages[pageName]

        row = self.pageOverview.pageListWidget.row(items[0])
        self.pageOverview.pageListWidget.takeItem(row)
        self.onPageConfigChanged()


    @pyqtSlot()
    def renameCurrentPage(self):
        """ Shows a dialog to rename currently selected page.
        """
        logger.debug("Rename current page")
        currentItem = self.pageOverview.pageListWidget.currentItem()

        if not currentItem:
            logger.info("No item selected for renaming")
            return

        page = self._pages[currentItem.text()]
        oldName = page.name
        newName, okClicked = QtWidgets.QInputDialog.getText(
            self, "Rename Page", "Page name:", QtWidgets.QLineEdit.Normal, oldName)

        if not okClicked:
            logger.debug("Rename canceled")
            return

        if newName == '':
            _reply = QtWidgets.QMessageBox.information(
                self, "Page Rename Failed", "Page name cannot be empty.")
            return

        if newName in self._pages:
            _reply = QtWidgets.QMessageBox.information(
                self, "Page Rename Failed", "A page with this name already exists")
            return

        currentItem.setText(newName)
        page.name = newName
        self._pages[newName] = page
        del self._pages[oldName]

        self.setCurrentPage(newName)
        self.onPageConfigChanged()


    #### pages CRUD methods ####

    def _updateWindowTitle(self):
        """ Sets the window title from on the currenly select page.
        """
        if self._currentPage is None:
            self.setWindowTitle('{}'.format(PROGRAM_NAME))
        else:
            self.setWindowTitle('{} -- {}'.format(self._currentPage.name, PROGRAM_NAME))


    def _getPageByName(self, pageName):
        """ Gets a page by its name.
            An ItemNotFoundError is raised if the page does not exist.
        """
        try:
            return self._pages[pageName]
        except KeyError:
            raise ItemNotFoundError("Page {!r} does not exist.".format(pageName))


    def _getPageAndPanelByName(self, pageName, panelName):
        """ Gets a page and the a panel on that page.

            An ItemNotFoundError is raised if the page or panel does not exist.

            :return: (page, panel) tuple
        """
        page = self._getPageByName(pageName)
        check_class(page, BasePage) # sanity check
        panel = page.getPanelByName(panelName)
        check_class(panel, AbstractPanel)
        return page, panel


    def _createAndAddPanel(self,
                           pageName = None,
                           panelName = None,
                           fullClassName = 'gssw.quicklook.plotpanel.PlotPanel',
                           errorIfExists=False,
                           gridGeometry=None):
        """ Creates a new panel and adds it to the current page.

            It uses a dynamic import mechanism. The modules for the panels are only imported when
            you load them. This allows you to write panel classes with dependencies that not need
            to be installed for all users, only for users that use the panel. This makes the
            Quicklooker more extensible.

            :param str pageName: name of the page where the panel will be created.
                An ItemNotFoundError is raised if this page does not exist.
            :param str panelName: name of the panel, If None, panel-N will be used where N will be
                the first integer that is not used.
            :param str fullClassName: class name of the panel 'gssw.quicklook.plotpanel.PlotPanel'
            :param bool errorIfExists: if True a DuplicateItemError is raised if a page with this
                name already exists. If False (the default) the old page is removed and a warning
                is logged.
            :param dict gridGeometry: panel geometry in grid coordinates. Grid coordinates range
                from 0 to the number of grid points (36 in the current implementation)
                The dict should have keys: x, y, width, height
                If None, the new panel will cover a quarter of the page.
            :returns: The panel (new one if created, existing object if a panel already existed).
        """
        if pageName is None:
            page = self.currentPage
            if not page:
                logger.warning("No page selected or while adding a new panel")
                return
        else:
            page = self._getPageByName(pageName)

        if panelName is None:
            panelName = page.makeUniquePanelName()

        if panelName in page.panels:
            msg = "A panel with this name already exists: {}".format(panelName)
            if errorIfExists:
                raise DuplicateItemError(msg)
            else:
                logger.info("%s (no new panel added, using existing one).", msg)
                return page.getPanelByName(panelName)
        else:
            panel = createPanelByClassName(fullClassName, panelName)
            page.insertPanel(panel, gridGeometry=gridGeometry)
            #self.onPageConfigChanged() will be done by page.insertPanel
            return panel


    @pyqtSlot()
    def addPlotPanel(self,  #pylint: disable=too-many-arguments
                     pageName = None,
                     panelName = None,
                     errorIfExists=False,
                     gridGeometry=None,
                     appendMode=None,
                     storeMaxSamples = None,
                     plotTitle = None,
                     xTitle = None,
                     yTitle = None,
                     xAxisLogarithmic = None,
                     yAxisLogarithmic = None,
                     showGrid = None):
        """ Adds a plot panel with name panelName on the page with name pageName.

            If a panel on the page already exists (and errorIfExists is False, the default), the
            new panel settings are applied. If errorIfExists is True, a DuplicateItemError
            exception is raised.

            If you want to ensure that a new panel is always used you must call removePanel
            beforehand.

            If the following parameters are omitted, defaults are used.
            See :meth:`_createAndAddPanel` for the other parameters.

            :param int appendMode: 0 for AppendModeEnum.Append, 1 for AppendModeEnum.replace
            :param int storeMaxSamples: maximum number of samples that are stored per array
            :param str plotTitle: optional plot title
            :param str xTitle: optional title for the x-axis
            :param str yTitle: optional title for the y-axis
            :param bool xAxisLogarithmic: Display x axis in logarithmic scale.
            :param bool yAxisLogarithmic: Display y axis in logarithmic scale.
            :param bool showGrid: if True the grid lines are displayed.
            :returns: string with the name of the newly created panel. Returns the emtpy string if
                the panel already existed.
        """
        check_class(appendMode, int, allow_none=True)

        plotPanel = self._createAndAddPanel(
            pageName = pageName, panelName = panelName,
            fullClassName = 'gssw.quicklook.plotpanel.PlotPanel',
            errorIfExists = errorIfExists, gridGeometry = gridGeometry)

        # Set the plot panel specific properties here. Saves us from making methods in the Page.
        # pylint: disable=multiple-statements
        if plotTitle is not None: plotPanel.plotTitle = plotTitle
        if xTitle is not None: plotPanel.xTitle = xTitle
        if yTitle is not None: plotPanel.yTitle = yTitle
        if appendMode is not None: plotPanel.plotMode = AppendModeEnum(appendMode)
        if storeMaxSamples is not None: plotPanel.storeMaxSamples = storeMaxSamples
        if xAxisLogarithmic is not None: plotPanel.xAxisLogarithmic = xAxisLogarithmic
        if yAxisLogarithmic is not None: plotPanel.yAxisLogarithmic = yAxisLogarithmic
        if showGrid is not None: plotPanel.showGrid = showGrid

        return plotPanel.name


    @pyqtSlot()
    def addHistogramPanel(self,  #pylint: disable=too-many-arguments
                          pageName = None,
                          panelName = None,
                          errorIfExists=False,
                          gridGeometry=None,
                          plotTitle = None,
                          xTitle = None,
                          yTitle = None,
                          showGrid = None,
                          color = None,
                          binAlignment=None,
                          arrayName=None):
        """ Adds a histogram panel with name panelName on the page with name pageName.

            If a panel on the page already exists (and errorIfExists is False, the default), the
            new panel settings are applied. If errorIfExists is True, a DuplicateItemError
            exception is raised.

            If you want to ensure that a new panel is always used you must call removePanel
            beforehand.

            If the following parameters are omitted, defaults are used.
            See :meth:`_createAndAddPanel` for the other parameters.

            :param str plotTitle: optional plot title
            :param str xTitle: optional title for the x-axis
            :param str yTitle: optional title for the y-axis
            :param bool showGrid: if True the grid lines are displayed.
            :param int color: a value attribute of a ColorEnum enumeration object
            :param string binAlignment: a value attribute of a BinAlignment enumeration object
            :param string arrayName: the name of the gridded array that is visualized

            :returns: string with the name of the newly created panel. Returns the emtpy string if
                the panel already existed.
        """
        check_class(color, int, allow_none=True)
        check_class(binAlignment, int, allow_none=True)

        histPanel = self._createAndAddPanel(
            pageName = pageName, panelName = panelName,
            fullClassName = 'gssw.quicklook.histpanel.HistPanel',
            errorIfExists = errorIfExists, gridGeometry = gridGeometry)

        # Set the plot panel specific properties here. Saves us from making methods in the Page.
        # pylint: disable=multiple-statements
        if plotTitle is not None: histPanel.plotTitle = plotTitle
        if xTitle is not None: histPanel.xTitle = xTitle
        if yTitle is not None: histPanel.yTitle = yTitle
        if showGrid is not None: histPanel.showGrid = showGrid
        if color is not None: histPanel.color = ColorEnum(color)
        if binAlignment is not None: histPanel.binAlignment = BinAlignmentEnum(binAlignment)
        if arrayName is not None: histPanel.arrayName = arrayName

        return histPanel.name


    @pyqtSlot()
    def addImagePanel(self,  #pylint: disable=too-many-arguments
                      pageName = None,
                      panelName = None,
                      errorIfExists=False,
                      gridGeometry=None,
                      plotTitle = None,
                      xTitle = None,
                      yTitle = None,
                      colorBarTitle = None,
                      colorMap = None,
                      colorAutoRange=None,
                      colorRange=None,
                      arrayName=None):
        """ Adds a 2D image plot panel with name panelName on the page with name pageName.

            If a panel on the page already exists (and errorIfExists is False, the default), the
            new panel settings are applied. If errorIfExists is True, a  DuplicateItemError
            exception is raised.

            If you want to ensure that a new panel is always used you must call removePanel
            beforehand.

            If the following parameters are omitted, defaults are used.
            See :meth:`_createAndAddPanel` for the other parameters.

            :param str plotTitle: optional plot title
            :param str xTitle: optional title for the x-axis
            :param str yTitle: optional title for the y-axis
            :param str colorBarTitle: optional title for in the color bar
            :param str colorMap: color scale. At least the following names are supported by the
                Silx plot library: gray, reversed gray, temperature, red, green, blue, jet,
                viridis, magma, inferno, plasma.
            :param bool colorAutoRange: if True the color range will be determined from the data.
            :param tuple colorRange: tuple with 2 elements defining the color map range.
            :param string arrayName: the name of the gridded array that is visualized

            :returns: string with the name of the newly created panel. Returns the emtpy string if
                the panel already existed.
        """
        imagePanel = self._createAndAddPanel(
            pageName = pageName, panelName = panelName,
            fullClassName = 'gssw.quicklook.imgpanel_silx.ImagePanel',
            errorIfExists = errorIfExists, gridGeometry = gridGeometry)

        logger.debug("colorBarTitle: {!r}".format(colorBarTitle))
        logger.debug("colorMap: {!r}".format(colorMap))
        logger.debug("colorAutoRange: {!r}".format(colorAutoRange))
        logger.debug("colorRange: {!r}".format(colorRange))

        # Set the plot panel specific properties here. Saves us from making methods in the Page.
        # pylint: disable=multiple-statements
        if plotTitle is not None: imagePanel.plotTitle = plotTitle
        if xTitle is not None: imagePanel.xTitle = xTitle
        if yTitle is not None: imagePanel.yTitle = yTitle
        if colorBarTitle is not None: imagePanel.barTitle = colorBarTitle
        if colorMap is not None: imagePanel.colormapName = colorMap
        if colorAutoRange is not None: imagePanel.colorAutoRange = colorAutoRange
        if colorRange: imagePanel.colorRange = colorRange # can be empty list if
        if arrayName is not None: imagePanel.arrayName = arrayName

        return imagePanel.name

    # pylint: enable=multiple-statements

    def removePanel(self, pageName, panelName):
        """ Removes a panel from the current page, given its name

            :param str pageName: name of the page where the panel will be created.
                An ItemNotFoundError is raised if this page does not exist.
            :param str panelName: name of the panel,
        """
        logger.debug("Removing panel {} on page {}".format(panelName, pageName))
        try:
            page = self._getPageByName(pageName)
        except ItemNotFoundError:
            logger.warning("Page not found in removePanel({}, {})".format(pageName, panelName))
        else:
            page.removePanel(panelName)


    def addCurveToPlotPanel(self, pageName, panelName, arrayName,
                            errorIfExists=False,
                            label = None,
                            lineStyle = None,
                            lineWidth = None,
                            color = None,
                            hasSymbol = None):
        """ Adds a curve to the plot in the panel on the current page.

            Note that if the panel is not a PlotPanel a TypeError is raised.

            :param str pageName: page name
            :param str panelName: panel name
            :param str arrayName: name of the array that the curve item visualizes.
            :param bool errorIfExists: if True a DuplicateItemError is raised if a curve with this
                name already exists. If False (the default) the old curve is replaced by the new
                curve. In that case parameter (e.g. lineStyle) that are None are take from the
                old curve.
            :param str label: label that is used in the plot (if None, the arrayName is used)
            :param float lineWidth: width of the line. Also infuences symbol size.
            :param int lineStyle: a value attribute of a LineStyleEnum object
            :param int color: a value attribute of a ColorEnum object
            :param bool hasSymbol: if True, a dot is drawn for every data point.
        """
        # TODO: adapt the rpcmsg.Packetizer class so that can return a LineStyleEnum and
        # ColorEnum objects.

        check_is_a_string(pageName)
        check_is_a_string(panelName)
        check_is_a_string(arrayName)

        # Convert Enums
        check_class(lineStyle, int, allow_none=True)
        lineStyle = None if lineStyle is None else LineStyleEnum(lineStyle)
        check_class(lineStyle, LineStyleEnum, allow_none=True)

        check_class(color, int, allow_none=True)
        color = None if color is None else ColorEnum(color)
        check_class(color, ColorEnum, allow_none=True)

        _, panel = self._getPageAndPanelByName(pageName, panelName)
        check_class(panel, PlotPanel)

        curveItem = (panel.getCurveItemByName(arrayName) if panel.arrayInConfig(arrayName)
                     else CurveItem(arrayName)) # default values if not old curve item.

        # Override with parameter
        curveItem.label = curveItem.label if label is None else label
        curveItem.lineStyle = curveItem.lineStyle if lineStyle is None else lineStyle
        curveItem.lineWidth = curveItem.lineWidth if lineWidth is None else lineWidth
        curveItem.hasSymbol = curveItem.hasSymbol if hasSymbol is None else hasSymbol

        # Auto-assign each trace a different color if no color is specified in the protobuf.
        # This needs to be done here, and not in panel.addCurveItem because CurveItem,color can
        # not be None to indicate an undefined value.
        if color is not None:
            curveItem.color = color
        elif panel.arrayInConfig(arrayName):
            pass # use old value if it exists.
        else:
            curveItem.color = panel.getNextCurveColor() # auto assigned color

        panel.addCurveItem(curveItem, errorIfExists=errorIfExists)



    def removeItemFromPanel(self, pageName, panelName, arrayName):
        """ Removes an item (e.g. a plot curveItem) from the panel

            Currently only implemented for the PlotPanel, not for the TablePanel (TODO)

            :param str pageName: page name
            :param str panelName: panel name
            :param str arrayName: name of the array that the curve item visualizes.
        """
        _, panel = self._getPageAndPanelByName(pageName, panelName)
        panel.removeCurveByName(arrayName)


    def clearPanelData(self, pageName, panelName):
        """ Removes all array data in the panel.

            Does not remove the curves from the configuration, just removes their array contents.

            :param str pageName: page where the panel should be present.
            :param str panelName: panel name
        """
        _, panel = self._getPageAndPanelByName(pageName, panelName)
        panel.clearData()


    def appendArray(self, array):
        """ Appends gridded array data to the corresponding array in the repository.
        """
        assert(len(array.data) > 0), "Cannot plot empty array: {}".format(array.name) #pylint: disable=len-as-condition

        # Update the array received statistice
        GlobalArrayInfoRepo.instance().updateArrayStatistics(array)

        # Append array to all plot that contain this name
        for page in self._pages.values():
            for panel in page.panels.values():
                if panel.arrayInConfig(array.name):
                    panel.updatePanelData(array)


    ##################################################
    # RPC methods for retrieving current configuration
    ##################################################

    def getPageNames(self):
        """ Retrieve the list of pages available in the Quicklooker. """
        return self.pageOverview.getAllPageNames()


    def getPanelNames(self, pageName):
        """ Retrieve the list of panels available in the Quicklooker pages.

        :param str pageName: Get list from this page.
        """
        page = self._getPageByName(pageName)
        return sorted([pan.name for pan in page.panels.values()])


    def getCurveNames(self, pageName, panelName):
        """ Retrieve the list of curves from a specific panel.

            :param str pageName:  Name of the page which contains the panel.
            :param str panelName: Name of the panel.
        """
        _, panel = self._getPageAndPanelByName(pageName, panelName)
        check_class(panel, PlotPanel) # Only implemented for PlotPanel classes.
        return panel.getCurveNames()


    #########################
    # Miscellaneous Methods #
    #########################

    @property
    def currentPage(self):
        """ Returns the current page
        """
        return self._currentPage


    def onPageConfigChanged(self, name=None):
        """ Updates main UI window when a page or panel config has changed.

            Will update the live data array model with the number of plots every array occurs in.
        """
        logger.debug("onPageConfigChanged called (name = {!r})".format(name))
        plotCounts = collections.Counter()
        for page in self._pages.values():
            for panel in page.panels.values():
                arrayNames = panel.getConfigArrayNames()
                logger.debug("Array names in plot: {}".format(arrayNames))
                plotCounts.update(arrayNames)

        GlobalArrayInfoRepo.instance().updatePlotStatistics(plotCounts)


    def onCurrentPageItemChanged(self, currentItem, previousItem):
        """ Called when the current item in the page overview list has changed.

            Does nothing if currentItem == previousItem (to prevent loops).<
            Otherwise it sets the current page.
        """
        if currentItem == previousItem:
            logger.debug("currentItem == previousItem. Ignored call.")
            return

        pageName = currentItem.text() if currentItem else ''
        self.setCurrentPage(pageName)


    def execConfigureArraysDialog(self):
        """ Executes the dialog that is used to administrate the list of 'known' arrays

            This list is used to populate the comboboxes for adding arrays to the panel. Next to
            these 'known-arrays' the comboboxes will contain all arrays that are currently in a
            panel or have been received as a message during current execution.
        """
        logger.debug("execConfigureArraysDialog.")
        self.configNameListDialog.show()


    def saveCurrentPageAsImage(self):
        """ Asks the user for a file name and saves the current page as image to file
        """
        if not self.currentPage:
            logger.warning("No image saved (no current page selected.")
            return

        saveWidgetAsImage(self, self.currentPage, baseName=self.currentPage.name)


    ###############
    # Persistence #
    ###############


    def _readConfigFromYaml(self, fileName):
        """ Reads the pages configuration from a YAML file.

            Also reads the list of configurable array names.

            The YAML file is a dict containing the following
                'array_names': list of known array names,
                'pages': list of page configurations.
        """
        logger.info("Loading pages configuration from YAML file {}"
                    .format(os.path.normpath(os.path.realpath(fileName))))
        cfgVersion = 'n/a'
        try:

            with open(fileName, 'r') as yamlFile:
                cfg = yaml.safe_load(yamlFile)

            if cfg is None:
                logger.info("Found empty configuration file. Using empty config.")
                cfg = {}
            elif isinstance(cfg, list):
                # Old config was only a list of pages
                logger.warning("Old config format detected. Will be updated.")
                cfg = {'pages': cfg}

            cfgVersion = cfg.get('version', 'unspecified')
            logger.info("Loading config file format version: %s", cfgVersion)

            # pages
            pagesCfg = cfg.get('pages', [])
            check_class(pagesCfg, list)

            for pageDict in pagesCfg:
                check_class(pageDict, dict)
                pageClassName = pageDict['_class_']
                assert pageClassName.endswith('FreeLayoutPage'), \
                    "Unknown page class: {}".format(pageClassName) # Others not (yet?) implemented

                page = FreeLayoutPage.createFromDict(pageDict)
                self._insertPage(page)

            curPage = cfg.get('currentPage', '')
            if curPage:
                self.setCurrentPage(curPage)

            # Fill the dynamic list of available array names from the page config
            #self._populateGlobalNameList()

            # Read the static list of array names
            if 'array_info' in cfg:
                arrayInfo = cfg['array_info']
            elif 'array_names' in cfg:
                # Convert data from old config files that consisted of a list of names.
                arrayInfo = [dict(name=name) for name in cfg['array_names']]
            else:
                logger.warning("No array_info found in config.")
                arrayInfo = []

            GlobalArrayInfoRepo.instance().setStaticData(arrayInfo)
        except Exception as ex:
            logger.exception("An exception occurred while loading the configuration file "
                             "(format version: %s): %s", cfgVersion, ex)
            gsswUiError("The configuration file could not be loaded correctly:<pre>{}</pre>"
                        "Not all pages and panels might have been loaded correctly.<br />"
                        "If problems persist, please remove the file ({}).".format(
                            ex, os.path.normpath(os.path.realpath(fileName))))

        # The configNameListDialog shows the static arrayNames so so it must be updated/reverted.
        self.configNameListDialog.revertChanges()



    def _writeConfigToYaml(self, fileName=None, silent=True):
        """ Reads the pages configuration from a YAML file.

            :param: silent: If True this write action will not be logged.
        """
        if fileName is None:
            fileName = self._pagesConfigName

        if not silent:
            logger.info("Writing page config for window: {}".format(os.path.realpath(fileName)))

        # Pages
        pagesCfg = []
        for pageName in self.pageOverview.getAllPageNames():
            page = self._pages[pageName]

            # So far only FreeLayoutPage persistence is supported.
            if isinstance(page, FreeLayoutPage):
                pagesCfg.append(page.asDict())
            else:
                logger.warning("Page cannot be saved (skipped): {}".format(page))

        # Static array info
        arrayInfo = GlobalArrayInfoRepo.instance().getStaticData()

        cfg = {'pages': pagesCfg, 'array_info': arrayInfo,
               'currentPage': self._currentPage.name if self._currentPage else '',
               'version':     VERSION_STRING}
        with open(fileName, 'w') as yamlFile:
            yaml.dump(cfg, yamlFile, default_flow_style=False, indent=4)


    def _readViewSettings(self, reset):
        """ Reads the persistent program settings.

            :param reset: If True, the program resets to its default settings
        """
        if reset:
            logger.debug("Resetting persistent view settings")
        else:
            logger.debug("Reading view settings")
            settings = QtCore.QSettings()
            settings.beginGroup('main_window')
            try:
                win_geom = settings.value("geometry")
                if win_geom:
                    self.restoreGeometry(win_geom)
                else: # This happens the first time the application is started
                    logger.debug("Unable to restore main window geometry.")

                win_state = settings.value("state")
                if win_state:
                    self.restoreState(win_state)
                else: # This happens the first time the application is started
                    logger.debug("Unable to restore main window state.")

                self.dataOverView.arrayTableView.read_view_settings(
                    "arrayTable", settings, reset=reset)
            finally:
                settings.endGroup()

            settings.beginGroup('names_dialog')
            try:
                dialog_geom = settings.value("geometry")
                if dialog_geom:
                    self.configNameListDialog.restoreGeometry(dialog_geom)
                else: # This happens the first time the application is started
                    logger.debug("Unable to restore name dialog geometry.")
            finally:
                settings.endGroup()


    def _writeViewSettings(self, duringExit=False, silent=True):
        """ Writes the view settings to the persistent store.

            The duringExit parameter can be used to prevent further saves. This is necessary
            because this function may be called twice during exit, depending on how the user exists
            the application (CTRL+Q vs clicking on the X-icon). The second time it is called not
            some dock-widgets may be closed already, which should not be saved in the QSettings.

            :param: silent: If True this write action will not be logged.
            :param duringExit: If True the settings will be saved for the last time and
                log message will indicate this.
        """
        if duringExit and self._settingsSaved:
            logger.debug("Settings have been saved before. Returning to caller.")
            return

        try:
            settings = QtCore.QSettings()
            if not silent:
                logger.info("Writing view settings for window in: {}".format(settings.fileName()))

            settings.beginGroup('main_window')
            try:
                settings.setValue("geometry", self.saveGeometry())
                settings.setValue("state", self.saveState())
                self.dataOverView.arrayTableView.write_view_settings("arrayTable", settings)
            finally:
                settings.endGroup()

            settings.beginGroup('names_dialog')
            try:
                settings.setValue("geometry", self.configNameListDialog.saveGeometry())
            finally:
                settings.endGroup()
        finally:
            if duringExit:
                self._settingsSaved = True


    def about(self):
        """ Shows the about message window.
        """
        logger.info("About: {!r} version: {!r}".format(PROGRAM_NAME, VERSION_STRING))
        repStr = ', '.join(self._repSocket.boundTo)
        subStr = ', '.join(self._subSocket.connectedTo + self._subSocket.boundTo)
        aboutMessage = ("{} {}\n\nRep-socket: {}\nSub-socket: {}"
                        .format(PROGRAM_NAME, VERSION_STRING, repStr, subStr))
        QtWidgets.QMessageBox.about(self, "About {}".format(PROGRAM_NAME), aboutMessage)


    def closeEvent(self, event):
        """ Called when the window is closed.
        """
        logger.debug("closeEvent")
        self._writeViewSettings(duringExit=True, silent=False)
        self._writeConfigToYaml(silent=False)
        self.finalize()
        self.close()
        event.accept()
        logger.debug("Closed {}".format(PROGRAM_NAME))


    def quitApplication(self):
        """ Closes all windows.
        """
        logger.info("Quicklooker quitApplication called")
        # Save the settings before closing any windows. Otherwise it may close the dock window
        # and then save the state. In _writeViewSettings it is checked if the settings have
        # already been saved.
        self._writeViewSettings(duringExit=True, silent=False)
        self._writeConfigToYaml(silent=False)
        app = QtWidgets.QApplication.instance()
        app.closeAllWindows()
        logger.debug("quitApplication done")


    def myTest(self):
        """ Function for testing.
        """
        logger.info("------- Test function called -------")
        _, panel = self._addTestData()

        panel.exportToCsv()



    def _addTestData(self):
        """ Adds a page with test data.
        """
        import numpy as np
        from gssw.lib.arrays.gridded import GriddedArray
        from gssw.lib.arrays.semantic import SemanticArray

        pageName = "myTest"
        panelName = "myPlot"
        self.addNewPage(pageName)

        self.addPlotPanel(pageName, panelName)

        for power in range(1, 4):
            arrayName = "array_{}".format(power)
            self.addCurveToPlotPanel(pageName, panelName, arrayName)
            xValues = np.linspace(-1.0, 1.0, 50)
            yValues = np.power(xValues, power)
            gArray = GriddedArray(yValues, coords={0:SemanticArray(xValues)}, name=arrayName)
            self.appendArray(gArray)

        return self._getPageAndPanelByName(pageName, panelName)
