""" Pages for in the Quicklook application.
"""
import collections
import logging, os.path
import threading
import numpy as np
import yaml

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot, QThread

from gssw.lib.widgets.colors import getWidgetColor, PaletteRoleEnum
from gssw.lib.typing import check_class, check_is_a_string, full_type_name
from gssw.lib.widgets.misc import qRectFromDict
from gssw.lib.errors import ItemNotFoundError
from gssw.quicklook.defs import MIME_GSSW_PANEL, MIME_ARRAYS_INFO
from gssw.quicklook.basepanel import PANEL_SIZE_DEFAULT, AbstractPanel, createPanelByClassName
from gssw.quicklook.basepanel import decodePanelMimeData
from gssw.quicklook.shared import ArrayInfoModel

logger = logging.getLogger(__name__)

#: The size of each Quicklooker page (number of 'cells' horizontally and vertically)
PAGE_SIZE = 36

#: Redraw frequency (delay in seconds), over all curves on a page
PAGE_REDRAW_DELAY = .05



class BasePage(QtWidgets.QFrame):
    """ Base class for pages in the Quicklook GUI.
    """
    sigPageConfigChanged = pyqtSignal(str) # page name

    def __init__(self, name, parent=None):
        """ Constructor

            :param name: page name
            :param parent: Qt parent object.
        """
        super().__init__(parent=parent)
        self.setWindowTitle(name) # Not necessary but just in case a page ends up without a parent.
        check_is_a_string(name)
        self._name = name
        self._mainWindow = None


    def finalize(self):
        """ Cleanup resources.
        """
        pass


    @property
    def name(self):
        """ A name of this page. Should be unique."""
        return self._name


    @name.setter
    def name(self, value):
        """ A name of this page. Should be unique."""
        check_class(value, str)
        assert value != "", "Name cannot be empty"
        self._name = value


    @property
    def mainWindow(self):
        """ Returns the (main) window on which this page is located.
        """
        return self._mainWindow


    @mainWindow.setter
    def mainWindow(self, value):
        """ A the (main) window on which this page is located.

            This value is typically set by MainWindow.insertPage. You should not set it yourself!
        """
        self._mainWindow = value


    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        #geom = self.geometry()
        return {'name': self.name,
                '_class_': full_type_name(self)}
                #'geometry': [geom.x(), geom.y(), geom.width(), geom.height()]}


    @classmethod
    def createFromDict(cls, dictionary):
        """ Creates a page from a dictionary with values.

            Is typically overriden by descendants.
        """
        logger.debug("createFromDict: {}".format(dictionary))
        page = cls(name=dictionary['name'])
        return page



    @classmethod
    def createFromYamlFile(cls, fileName):
        """ Loads a Yaml file and creates a Page from it.
        """
        logger.debug("Loading page from Yaml file {}".format(os.path.realpath(fileName)))
        with open(fileName, 'r') as yamlFile:
            pageDict = yaml.safe_load(yamlFile)
        return cls.createFromDict(pageDict)


    def saveToYaml(self, fileName):
        """ Saves the page state to a Yaml file.
        """
        logger.debug("Saving {} to {}".format(self, os.path.realpath(fileName)))
        with open(fileName, 'w') as yamlFile:
            yaml.dump(self.asDict(), yamlFile, default_flow_style=False, indent=4)



##################
# No Layout Page #
##################



class FreeLayoutPage(BasePage):
    """ Page that has no layout. The panels are free floating but snap to a grid.

        Has a PageUpdater which is a QThread that signals the Page to update periodically.
        This throttles the update frequency. Otherwise the page updates may not keep up
        if a lot of data is being added to long lines.

        If you create a FreeLayoutPage you MUST call finalize to stop the underlying thread.
    """

    # # The panel is about to be removed from the current page
    # sigPanelAboutToBeRemoved  = pyqtSignal(str)  # param: panel name

    def __init__(self, name, parent=None):
        """ Constructor.

            :param name: page name
            :param parent: Qt parent object.
        """
        super().__init__(name, parent=parent)

        self.setObjectName('page')

        self._panels = {}

        self.panelsMenu = QtWidgets.QMenu("Panels", self)

        # Create menu item for clearing all plots in the current panel
        self._clearPlotsAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("empty"),
                                                   "Clear all (on page)", self)
        self._clearPlotsAction.setToolTip("Clear all plots (on current page)")
        self._clearPlotsAction.triggered.connect(self.clearPlots)
        self.panelsMenu.addAction(self._clearPlotsAction)

        self.setAcceptDrops(True)
        self.rubberBand = QtWidgets.QRubberBand(QtWidgets.QRubberBand.Rectangle, self)

        # Grid and Background check-board pattern.
        self._gridStepX = None
        self._gridStepY = None
        self._gridStepsFixed = False

        self.setGrid(32, 32) # Start with cells of 32 by 32 pixels. Will change at window resize.
        self._numGridPoints = PAGE_SIZE

        # Create the thread that will manage the redrawing of all the panels on this page
        self.plotUpdateThread = PageUpdater(parent=self)
        self.plotUpdateThread.sigDoRedraw.connect(self.slotDoRedraw,
              type=Qt.BlockingQueuedConnection) # Use a blocking connection here so that new data
        #                                       # can be queued (and merged) in the meantime
        self.plotUpdateThread.start()


    def finalize(self):
        """ Cleanup resources.
        """
        logger.debug("FreeLayoutPage.finalize. Stopping updater QThread.")
        self.plotUpdateThread.stop()
        for panel in self.panels.values():
            panel.sigClosePanel.disconnect(self.removePanel)
            panel.sigPanelConfigChanged.disconnect(self.emitConfigChanged)
            if hasattr(panel, 'sigDataQueued'):
                panel.sigDataQueued.disconnect(self.plotUpdateThread.slotQueRedraw)
            panel.finalize()
        super().finalize()


    @property
    def panels(self):
        """ Returns the panels in this page as a dictionary. Should be considered read-only.
            :returns: {name: AbstractPanel} dict
        """
        return self._panels


    def getPanelByName(self, panelName):
        """ Gets a panel by its name.
            An ItemNotFoundError is raised if the page does not exist.
        """
        try:
            return self.panels[panelName]
        except KeyError:
            raise ItemNotFoundError("Panel {!r} does not exist on page {!r}."
                                    .format(panelName, self.name)) from None


    def makeUniquePanelName(self):
        """ Creates a panel name that does not exist yet.

            Returns 'Panel-X' where is X is the first panel number that does not exist yet.
        """
        panelNames = [pan.name for pan in self.panels.values()]
        for duplicateNr in range(1000):
            name = "Panel-{}".format(duplicateNr)
            if name not in panelNames:
                return name

        raise RecursionError("Infinite loop detected") # sanity check


    def insertPanel(self, panel, gridGeometry=None):
        """ Adds a panel to the list.

            RetRaises a DuplicateItemError if a panel of that name already is present.

            :param dict gridGeometry: panel geometry in grid coordinates. Grid coordinates range
                from 0 to the number of grid points (36 in the current implementation)
                The dict should have keys: x, y, width, height
                If None, the new panel will cover a quarter of the page.

            :returns: the pane None
        """
        check_class(panel, AbstractPanel)
        check_class(gridGeometry, dict, allow_none=True)
        logger.info("Adding panel {} to page {}".format(panel.name, self.name))

        self._panels[panel.name] = panel

        assert panel.page is None, "Panel is already part of: {}".format(panel.page)
        panel.page = self
        panel.setParent(self)
        panel.show()

        gridGeomDict = gridGeometry
        check_class(gridGeometry, dict, allow_none=True)

        if not gridGeomDict:
            logger.debug("Automatically positioning the panel in a 2x2 setup")
            numPanelsBeforeInsert = len(self.panels) - 1
            numPanelsHorizontal = PAGE_SIZE / PANEL_SIZE_DEFAULT
            numPanelsVertical   = PAGE_SIZE / PANEL_SIZE_DEFAULT
            positionX = numPanelsBeforeInsert % numPanelsHorizontal
            positionY = (numPanelsBeforeInsert // numPanelsVertical) % numPanelsVertical
            positionX *= PANEL_SIZE_DEFAULT
            positionY *= PANEL_SIZE_DEFAULT
            gridGeomRect = panel.gridGeometry # the default panel size
            gridGeomRect.moveTo(int(positionX), int(positionY)) # Put it on a new place
        else:
            logger.debug("_createAndAddPanel gridGeomDict: {!r}".format(gridGeomDict))
            gridGeomRect = qRectFromDict(gridGeomDict)
            #self.setPanelGridGeometry(panel.name, gridGeomRect)

        self.setPanelGridGeometry(panel.name, gridGeomRect)

        panel.sigClosePanel.connect(self.removePanel)
        panel.sigPanelConfigChanged.connect(self.emitConfigChanged)
        if hasattr(panel, 'sigDataQueued'):
            # Use a direct connection here so that new data overwrites old (cached) data as
            # quickly as possible, before a redraw occurs
            panel.sigDataQueued.connect(self.plotUpdateThread.slotQueRedraw,
                                        type=Qt.DirectConnection)

        self.update()
        self.panelsMenu.addMenu(panel.menu)
        self.emitConfigChanged()


    def removePanel(self, panelName):
        """ Removes the panel given its name.
        """
        logger.debug("Removing panel {} from page {}".format(panelName, self.name))

        if panelName not in self._panels:
            logger.debug("Trying to delete panel {!r} that does not exist in page {!r}"
                         .format(panelName, panelName))
            return

        panel = self._panels[panelName]

        self.panelsMenu.removeAction(panel.menu.menuAction())

        panel.sigClosePanel.disconnect(self.removePanel)
        panel.sigPanelConfigChanged.disconnect(self.emitConfigChanged)

        if hasattr(panel, 'sigDataQueued'):
            panel.sigDataQueued.disconnect(self.plotUpdateThread.slotQueRedraw)

        panel.page = None
        panel.setParent(None)
        panel.deleteLater()
        del self._panels[panelName]

        self.emitConfigChanged()


    def clearPlots(self):
        """ Clear the data in all plot panels on the current page  """
        for _panelName, panel in sorted(self.panels.items()):
            panel.clearData()


    def emitConfigChanged(self):
        """ Emits the sigPageConfigChanged signal
        """
        logger.debug("Emitting sigPageConfigChanged({})".format(self.name))
        self.sigPageConfigChanged.emit(self.name)


    @property
    def gridStepsFixed(self):
        """ If True the grid step sizes remain fixed in case the page is resized, but the number of
            grid points changes. If there the grid step sizes change but the number of grid points
            remains constant)
        """
        return False # Only False is implemented.
        #return self._gridStepsFixed


    @property
    def gridStepX(self):
        """ The grid size in the X direction in pixels"""
        return self._gridStepX


    @property
    def gridStepY(self):
        """ The grid size in the Y direction in pixels"""
        return self._gridStepY


    def setGrid(self, stepX, stepY):
        """ Sets the grid steps in pixels
        """
        self._gridStepX = stepX
        self._gridStepY = stepY
        return


    def setPanelGridGeometry(self, panelName, gridGeomRect):
        """ Sets the panel grid geometry.

            Als updates the panels geometry using the page grid. This means that the panel will
            be resize and moved.

            :param gridGeomRect: QRect
        """
        check_class(gridGeomRect, QtCore.QRect)
        assert gridGeomRect.isValid(), "gridGeomRect not valid: {}".format(gridGeomRect)

        panel = self.panels[panelName]
        panel.gridGeometry = gridGeomRect

        newX = gridGeomRect.x() * self.gridStepX
        newY = gridGeomRect.y() * self.gridStepY
        newWidth  = gridGeomRect.width() * self.gridStepX
        newHeight = gridGeomRect.height() * self.gridStepY
        panel.setGeometry(QtCore.QRect(newX, newY, newWidth, newHeight))


    def resizeEvent(self, _resizeEvent):
        """ Called when the window is resized.

            Will resize the grid while keeping the number of grid point the same.
        """
        geom = self.geometry()
        xGrid = np.linspace(0, geom.width(), self._numGridPoints + 1)
        yGrid = np.linspace(0, geom.height(), self._numGridPoints + 1)
        xSize = xGrid[1] - xGrid[0]
        ySize = yGrid[1] - yGrid[0]
        self.setGrid(xSize, ySize)

        # Resize all panels (apply panel's gridGeometry to the new grid)
        for panelName, panel in self.panels.items():
            self.setPanelGridGeometry(panelName, panel.gridGeometry)


    @property
    def tileColorBright(self):
        """ The color of the light tiles in the checkboard grid.

            :rtype: a QtGui.QColor
        """
        return getWidgetColor(self, PaletteRoleEnum.Window) # (Window == BackGround(


    @property
    def tileColorDark(self):
        """ The color of the dark tiles in the checkboard grid.

            :rtype: a QtGui.QColor
        """
        return getWidgetColor(self, PaletteRoleEnum.AlternateBase)


    def paintEvent(self, _event):
        """ Draws a checkerboard background with each tile the size of the grid.
        """
        geom = self.geometry()
        painter = QtGui.QPainter(self)
        painter.fillRect(0, 0, geom.width(), geom.height(), self.tileColorDark)
        painter.setRenderHint(QtGui.QPainter.Antialiasing, on=False)

        colorBright = self.tileColorBright
        for yIdx in range(self._numGridPoints):
            y = yIdx * self.gridStepY
            for xIdx in range(self._numGridPoints):
                x = xIdx * self.gridStepX
                useBright = (xIdx + yIdx) % 2
                if not useBright:
                    continue
                rect = QtCore.QRectF(x, y, self.gridStepX, self.gridStepY)
                painter.fillRect(rect, colorBright)

        painter.end()


    def snapToGrid(self, rect):
        """ Rounds the corners of a rectangle to the grid of the page.

            :param rect: QRect (so it contains integers)
            :returns: QRect (so it contains integers)
        """
        return QtCore.QRect(
            round(rect.x() / self.gridStepX) * self.gridStepX,
            round(rect.y() / self.gridStepY) * self.gridStepY,
            round(rect.width() / self.gridStepX) * self.gridStepX,
            round(rect.height() / self.gridStepY) * self.gridStepY)


    def pixelRectToGrid(self, rect):
        """ Converts a rectangle in pixel coordinates to grid points

            Uses the gridStepX and gridStepY to determine the current grid resolution

            :param rect: QRect (so it contains integers)
            :returns: QRect (so it contains integers)
        """
        relRect = QtCore.QRect(
            round(rect.x() / self.gridStepX),
            round(rect.y() / self.gridStepY),
            round(rect.width() / self.gridStepX),
            round(rect.height() / self.gridStepY)
        )
        logger.debug("snapToRelGrid: {} - {}".format(rect, relRect))
        return relRect


    def dragEnterEvent(self, event):
        """ Called when a drag is in progress and the mouse enters this widget.

            Just calls drag MoveEvent.
        """
        self.dragMoveEvent(event)


    def dragMoveEvent(self, event):
        """ Called when a drag is in progress and when any of the following conditions occur:
            the cursor enters this widget, the cursor moves within this widget, or a modifier key
            is pressed on the keyboard while this widget has the focus.
        """
        #logger.debug("Drag move event called at pos: {}".format(event.pos()))

        mimeData = event.mimeData()
        if mimeData.hasFormat(MIME_GSSW_PANEL):

            # Move or resize panel
            edgesFlag, hotSpot, orgRect = \
                decodePanelMimeData(mimeData.data(MIME_GSSW_PANEL))
            newGeom = QtCore.QRect(orgRect)

            if edgesFlag == 0:
                # Move the rubber band
                relCurPos = event.pos() - hotSpot
                newGeom.moveTo(relCurPos.x(), relCurPos.y())
            else:
                # Resize the rubber band
                if edgesFlag & Qt.LeftEdge:
                    newGeom.setX(event.pos().x())
                if edgesFlag & Qt.TopEdge:
                    newGeom.setY(event.pos().y())
                if edgesFlag & Qt.RightEdge:
                    newGeom.setRight(event.pos().x())
                if edgesFlag & Qt.BottomEdge:
                    newGeom.setBottom(event.pos().y())

            snapGeom = self.snapToGrid(newGeom)
            self.rubberBand.show()
            self.rubberBand.setGeometry(snapGeom)

            event.acceptProposedAction()

        elif mimeData.hasFormat(MIME_ARRAYS_INFO):
            # Suggest location when creating panel from dragging an array name
            # The page is divided in 3 by 3 areas. Every area results in a different suggested
            # panel size. E.g. the top left corner suggest a panel in that corner, the top center
            # area suggest a panel that takes up the top half of the screen, etc.
            relCurPos = event.pos() - self.pos()
            xArea = int(relCurPos.x() / self.width() * 3)
            yArea = int(relCurPos.y() / self.height() * 3)

            # Potential panel geometry values
            xMin, xMax = self.x(), self.width() - self.x()
            yMin, yMax = self.y(), self.height() - self.y()
            xCen = (xMax - xMin) // 2
            yCen = (yMax - yMin) // 2

            self.rubberBand.show()
            if xArea == 0 and yArea == 0:
                self.rubberBand.setGeometry(xMin, yMin, self.width() // 2, self.height() // 2)
            elif xArea == 1 and yArea == 0:
                self.rubberBand.setGeometry(xMin, yMin, self.width(), self.height() // 2)
            elif xArea == 2 and yArea == 0:
                self.rubberBand.setGeometry(xCen, yMin, self.width() // 2, self.height() // 2)
            elif xArea == 0 and yArea == 1:
                self.rubberBand.setGeometry(xMin, yMin, self.width() // 2, self.height())
            elif xArea == 1 and yArea == 1:
                self.rubberBand.setGeometry(xMin, yMin, self.width(), self.height())
            elif xArea == 2 and yArea == 1:
                self.rubberBand.setGeometry(xCen, yMin, self.width() // 2, self.height())
            elif xArea == 0 and yArea == 2:
                self.rubberBand.setGeometry(xMin, yCen, self.width() // 2, self.height() // 2)
            elif xArea == 1 and yArea == 2:
                self.rubberBand.setGeometry(xMin, yCen, self.width(), self.height() // 2)
            elif xArea == 2 and yArea == 2:
                self.rubberBand.setGeometry(xCen, yCen, self.width() // 2, self.height() // 2)
            else:
                self.rubberBand.hide()
                self.rubberBand.setGeometry(QtCore.QRect()) # invalidates geometry
                assert False, "Cursor at unexpected page area ({}, {})".format(xArea, yArea)

            event.acceptProposedAction()
        else:
            logger.debug("No GSSw mime data in drag event. Ignored.")
            self.rubberBand.hide()
            self.rubberBand.setGeometry(QtCore.QRect()) # invalidates geometry
            event.ignore()


    def dragLeaveEvent(self, _event):
        """ Called when a drag is in progress and the mouse leaves this widget.

            Hides and invalidates the rubberband.
        """
        self.rubberBand.hide()
        self.rubberBand.setGeometry(QtCore.QRect()) # invalidates geometry


    def dropEvent(self, event):
        """ Called when the drag is dropped on this widget.
        """
        logger.debug("Drop event called. pos: {}".format(event.pos()))
        mimeData = event.mimeData()
        if mimeData.hasFormat(MIME_GSSW_PANEL):

            # Move or resize panel
            newGeom = self.rubberBand.geometry()
            if newGeom.isValid():
                panel = event.source()
                self.setPanelGridGeometry(panel.name, self.pixelRectToGrid(newGeom))
            else:
                logger.debug("Invalid geometry (ignored).")

            event.acceptProposedAction()

        elif mimeData.hasFormat(MIME_ARRAYS_INFO):

            # Create panel from dragged array name(s)
            arrInfoItems = ArrayInfoModel.decodeArraysInfoMimeData(mimeData.data(MIME_ARRAYS_INFO))

            # TODO: make it possible to add other panel types by dragging the right-mouse button
            panelName = self.mainWindow.addPlotPanel()
            panel = self.getPanelByName(panelName)

            newGeom = self.rubberBand.geometry()
            if newGeom.isValid():
                self.setPanelGridGeometry(panel.name, self.pixelRectToGrid(newGeom))
            else:
                logger.debug("Invalid geometry (ignored).")

            panel.configureFromArrayInfoList(arrInfoItems) # will open the config dialog
        else:
            event.ignore()

        self.rubberBand.hide()
        self.rubberBand.setGeometry(QtCore.QRect()) # invalidates geometry


    @classmethod
    def createFromDict(cls, dictionary):
        """ Creates a page from a dictionary with values
        """
        logger.debug("createFromDict: {}".format(dictionary))
        page = cls(name=dictionary['name'])

        for panelDict in dictionary.get('panels', []):
            logger.debug("   panelDict: {}".format(panelDict))
            fullClassName= panelDict['_class_']
            panel = createPanelByClassName(fullClassName, panelDict['name'])
            panel.setValuesFromDict(panelDict)

            gridGeom = panelDict.get('gridGeometry', {})
            if isinstance(gridGeom, list):
                # backwards compatibility: gridGeometry used to be stored a s a list
                x, y, width, height = gridGeom
                gridGeomRect = dict(x=x, y=y, width=width, height=height)
            else:
                gridGeomRect = gridGeom
            page.insertPanel(panel, gridGeometry=gridGeomRect)

        return page


    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        dictionary = super().asDict()
        dictionary['panels'] = [panel.asDict() for _, panel in sorted(self.panels.items())]
        return dictionary


    @pyqtSlot(str, str)
    def slotDoRedraw(self, panelName, curveName):
        """ Function (in gui thread) that will forward the redraw event to the plot library. """
        # Get the panel object
        logger.debug("     Redrawing: panel {}, curve: {}".format(panelName, curveName))
        try:
            panel = self.getPanelByName(panelName)
        except ItemNotFoundError:
            # Due to race conditions, a panel might have been removed (by user), just as the redraw
            # thread is about to draw new data. In this case, ignore the draw call (issue #2856).
            logger.warning('Panel %r is no longer available for drawing curve %r',
                           panelName, curveName)
        else:
            # Do a draw (in gui thread)
            panel.doRedraw(curveName)


    def suspendDrawing(self):
        """ Stops incoming data from being drawn.

        Any data coming in during suspend state will be draw after resuming.
        """
        self.plotUpdateThread.suspendDrawing()

    def resumeDrawing(self):
        """ Resumes drawing incoming.

        Any data in the queue received before resuming will now be drawn.
        """
        self.plotUpdateThread.resumeDrawing()



class PageUpdater(QThread):
    """ Class for drawing all the panels on a page round-robin style.

        Is QThread that signals the coresponding Page to update itself periodically.
        This throttles the update frequency. Otherwise the page updates may not keep up
        if a lot of data is being added to long lines.
    """

    sigDoRedraw = pyqtSignal(str, str) # params: panel name, curve name

    def __init__(self, parent):
        super().__init__(parent=parent)

        logger.debug("Creating PageUpdater QThread: {}".format(PageUpdater))

        # Que with panels/curves (1 entry per curve)
        #     we use an ordered dict (instead of normal dict) to get FIFO behavior
        self._redrawQueue = collections.OrderedDict()
        self._stopEvent   = threading.Event() # For stopping redraw thread
        self._pauseEvent  = threading.Event() # For suspending page redraw while page is not active

    @pyqtSlot(str, str)
    def slotQueRedraw(self, panelName, curveName):
        """ Trigger/schedule a curve redraw """
        logger.debug("Scheduling redraw for panel {}, curve {}".format(panelName, curveName))
        queKey = (panelName, curveName)
        self._redrawQueue[queKey] = True # insertion is atomic, so no thread lock required

    def run(self):
        """ Main thread loop """
        while not self._stopEvent.is_set():
            # draw incoming data in round-robin style
            if self._redrawQueue:
                # Pop a redraw request from the queue (FIFO-style)
                #     (popitem is atomic, so no thread lock required)
                (panelName, curveName), _ = self._redrawQueue.popitem(last=False)

                # Send a redraw signal to the main (=gui) thread
                self.sigDoRedraw.emit(panelName, curveName)

            # Thread sleep (interruptible), to allow handling of new data in the background
            self._stopEvent.wait(PAGE_REDRAW_DELAY)
            self._pauseEvent.wait()

    def stop(self):
        """ Stops the thread (blocking). """
        logger.debug("Stopping PageUpdater QThread: {}".format(PageUpdater))
        self.blockSignals(True)
        self._stopEvent.set()
        self._pauseEvent.set()
        self.wait()

    def suspendDrawing(self):
        """ Stops incoming data from being drawn.

        Any data coming in during suspend state will be draw after resuming.
        """
        self._pauseEvent.clear()

    def resumeDrawing(self):
        """ Resumes drawing incoming.

        Any data in the queue received before resuming will now be drawn.
        """
        self._pauseEvent.set()

class PageOverviewWidget(QtWidgets.QWidget):
    """ Widget that allows the user to select, add and remove pages.
        Consists of a list widgets plus add/remove buttons.
    """
    def __init__(self, parent=None):
        """ Constructor
        """
        super().__init__(parent=parent)

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.setLayout(self.mainLayout)

        self.pageListWidget = QtWidgets.QListWidget()

        # Enable internal drag drop
        # See http://doc.qt.io/qt-5/model-view-programming.html#using-convenience-views
        self.pageListWidget.setDragEnabled(True)
        self.pageListWidget.viewport().setAcceptDrops(True)
        self.pageListWidget.setDropIndicatorShown(True)
        self.pageListWidget.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)

        # Buttons for adding/removing list items.
        self.pageButtonsLayout = QtWidgets.QHBoxLayout()

        iconSize = 20

        QtWidgets.QApplication.instance()
        self.pageAddButton = QtWidgets.QToolButton()
        self.pageAddButton.setObjectName("ql-button-add-new-page")
        self.pageAddButton.setIconSize(QtCore.QSize(iconSize, iconSize))
        self.pageAddButton.setMinimumSize(QtCore.QSize(16, 16))
        self.pageAddButton.setToolButtonStyle(Qt.ToolButtonIconOnly)

        self.pageRemoveButton = QtWidgets.QToolButton()
        self.pageRemoveButton.setIconSize(QtCore.QSize(iconSize, iconSize))
        self.pageRemoveButton.setMinimumSize(QtCore.QSize(16, 16))

        self.pageButtonsLayout.addStretch()
        self.pageButtonsLayout.addWidget(self.pageAddButton)
        self.pageButtonsLayout.addWidget(self.pageRemoveButton)

        self.mainLayout.addLayout(self.pageButtonsLayout)
        self.mainLayout.addWidget(self.pageListWidget)



    def getAllPageNames(self):
        """ Returns a list of all page names in the same order as they are in the item list
        """
        names = []
        for row in  range(self.pageListWidget.count()):
            names.append(self.pageListWidget.item(row).text())
        return names

