""" Plot Panels for the Quicklook application
"""
import copy, logging, numbers

from collections import OrderedDict

import silx
import numpy as np

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot

from gssw.common import diagnosticlogger
from gssw.lib.defs import ActionEnum
from gssw.lib.typing import check_class
from gssw.lib.errors import DuplicateItemError, ItemNotFoundError, GsswExportError
from gssw.lib.misc import same
from gssw.lib.widgets.colors import getWidgetColor, PaletteRoleEnum
from gssw.lib.widgets.misc import promptForFileName, gsswUiWarning
from gssw.lib.widgets.plotting import GsswPlotWidget, ClearPlotAction, SavePlotAction
from gssw.quicklook.baseconfig import BaseConfigItem, BaseConfigModel, AbstractPanelConfigWidget
from gssw.quicklook.basepanel import AbstractDataPanel
from gssw.quicklook.defs import ColorEnum, COLOR_TO_IDX, IDX_TO_COLOR, getColorForCurve
from gssw.quicklook.defs import LineStyleEnum, LINE_STYLE_STR, IDX_TO_LINE_STYLE, LINE_STYLE_TO_IDX
from gssw.quicklook.repo import RepositoryModel, AppendModeEnum
from gssw.quicklook.shared import ArrayInfoItem

logger = logging.getLogger(__name__)

START_COLOR_IDX = 0 # Start at blue because light blue is a bit too light


class CurveItem(BaseConfigItem):
    """ An item for in the Config Dialog's CurveModel or the PlotPanel curves list

        Represents a curve in the plot and the data necessary to display it (color, line-style, etc)
    """
    def __init__(self,
                 name,
                 label = None,
                 lineStyle = None,
                 lineWidth = None,
                 color = None,
                 hasSymbol = None):
        """ Constructor

            :param str name: array name
            :param LineStyleEnum lineStyle: used for drawing the line style.
                Use no_line for drawing scatter plots (in combination with hasSymbol == True).
            :param float lineWidth: width of the line
            :param ColorEnum color: color for the line and symbols
            :param bool hasSymbol: If True, a dot will be drawn at every data point.
                The dot size grows with the line Width
        """
        super().__init__(name, label=label)

        # Declare members
        self._lineStyle = None
        self._lineWidth = None
        self._color     = None
        self._hasSymbol = None

        # Set members using the setters so that types are checked.
        self.lineStyle = LineStyleEnum.solid if lineStyle is None else lineStyle
        self.lineWidth = 1.0 if lineWidth is None else lineWidth
        self.color = ColorEnum.blue if color is None else color
        self.hasSymbol = False if hasSymbol is None else hasSymbol

    @property
    def legendName(self):
        """ The name of the curve in the legend.

            This is equal to the label if it is defined, otherwise it reverts to the name of
            the curve item.
        """
        return self.label if self.label else self.name

    @property
    def color(self):
        """ Color for the line and symbols (returns a ColorEnum)
        """
        return self._color

    @color.setter
    def color(self, value):
        """ Sets the color for the line and symbols
        """
        check_class(value, ColorEnum)
        self._color = value

    @property
    def colorIdx(self):
        """ The index of the current color.
            Needed for setting from/to comboboxes.
        """
        return COLOR_TO_IDX[self.color]

    @colorIdx.setter
    def colorIdx(self, value):
        """ The index of the current color.
            Needed for setting from/to comboboxes.
        """
        self.color = IDX_TO_COLOR[value]

    @property
    def lineStyle(self):
        """ The current line style (a LineStyleEnum)
        """
        return self._lineStyle

    @lineStyle.setter
    def lineStyle(self, value):
        """ The current line style (a LineStyleEnum)
        """
        check_class(value, LineStyleEnum)
        self._lineStyle = value

    @property
    def lineStyleIdx(self):
        """ The index of the current line style.
            Needed for setting from/to comboboxes.
        """
        return LINE_STYLE_TO_IDX[self.lineStyle]

    @lineStyleIdx.setter
    def lineStyleIdx(self, value):
        """ The index of the current line style.
            Needed for setting from/to comboboxes.
        """
        self.lineStyle = IDX_TO_LINE_STYLE[value]

    @property
    def lineWidth(self):
        """ Line width (can be a float)
        """
        return self._lineWidth

    @lineWidth.setter
    def lineWidth(self, value):
        """ Line width (can be a float)
        """
        check_class(value, numbers.Real)
        self._lineWidth = value

    @property
    def hasSymbol(self):
        """ True if a symbol will be drawn
        """
        return self._hasSymbol

    @hasSymbol.setter
    def hasSymbol(self, value):
        """ True if a symbol will be drawn
        """
        check_class(value, bool)
        self._hasSymbol = value

    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        dct = {} # Configuration subdict

        dct['name']         = self.name
        dct['label']        = self.label
        dct['color']        = self.color.name
        dct['lineStyle']    = self.lineStyle.name
        dct['lineWidth']    = self.lineWidth
        dct['hasSymbol']    = self.hasSymbol

        return dct

    def setValuesFromDict(self, dct):
        """ Creates a panel from a dictionary with values
        """
        # Name should certainly be present in the dict. For the others we can use defaults if they
        # somehow are not present in the dictionary.
        self.name         = dct['name']
        self.label        = dct.get('label', '')
        self.color        = ColorEnum.getColorOrDefault(dct.get('color', 'grey'))
        self.lineStyle    = LineStyleEnum[dct.get('lineStyle', 'solid')]
        self.lineWidth    = dct.get('lineWidth')
        self.hasSymbol    = dct.get('hasSymbol')


class PlotCurvesModel(BaseConfigModel):
    """ A table model for storing the plot-panel items.
    """

    # All items in this model will be of this class
    ITEM_CLASS = CurveItem

    # Extent the possible columns in the plot pannel model
    HEADERS    = ["Name", "Label", "Color (idx)", "Style (idx)", "Width", "Symbol"]
    PROPERTIES = ["name", "label", "colorIdx", "lineStyleIdx", "lineWidth", "hasSymbol"]
    COL_NAME, COL_LABEL, COL_COLOR_NAME, COL_STYLE_IDX, COL_WIDTH, COL_DRAW_SYMBOL = \
        range(len(HEADERS))


class PlotPanel(AbstractDataPanel):
    """ A panel containing a central plot.

        Has a PlotCurvesModel to store al the data and a link to a RepositoryModel so that it can
        take data from the repository and put it into the plot model.
    """

    sigDataQueued = pyqtSignal(str, str)  # param: panel name, curve name

    def __init__(self, name, repo=None, parent=None):
        """ Constructor

            :param name: panel name string
            :param repo: a RepositoryModel. If None, an emtpy one will be created.
            :param parent: Qt parent object.
        """
        check_class(repo, RepositoryModel, allow_none=True)
        super().__init__(name, repo=repo, parent=parent)

        self.exportCsvAction = QtWidgets.QAction("&Export to CSV...", self)
        self.exportCsvAction.setToolTip("Exports the plot data to file Comma Separated Values")
        self.exportCsvAction.triggered.connect(self.exportToCsv)
        self.exportCsvAction.setShortcut("Ctrl+E")
        self.menu.insertAction(self.saveAsImageAction, self.exportCsvAction)

        self.exportCsvSepAction = QtWidgets.QAction("Export to CSV per c&urve...", self)
        self.exportCsvSepAction.setToolTip(
            "Exports the plot data to multiple Comma Separated Value file (one per curve)")
        self.exportCsvSepAction.triggered.connect(self.exportToCsvSeparate)
        self.menu.insertAction(self.saveAsImageAction, self.exportCsvSepAction)

        # Keep a dictionary CurveItems that stores which repo items will be plotted (if actually
        # present in the repo) and how (e.g. line color).
        # Note that doesn't always have to be a matching repo item (e.g. the repo may be empty at
        # start up).
        self._curveItemsDict = OrderedDict()

        # Data queue (with length 1 per curve)
        self._dataCache = {} # Dictionary uses the keys as plot labels and the values as x, y arrays

        ###########
        # Widgets #
        ###########

        self.plot = GsswPlotWidget(parent=self)
        mar = 0.02 # add 2% margins around plot
        self.plot.setDataMargins(xMinMargin=mar, xMaxMargin=mar, yMinMargin=mar, yMaxMargin=mar)

        self.contentsLayout.addWidget(self.plot)

        saveAsImageIndex = 0
        # Override clearAction in the context menu
        for idx in range(len(self.plot.contextMenuActions)):
            if isinstance(self.plot.contextMenuActions[idx], ClearPlotAction):
                self.plot.contextMenuActions[idx] = self.clearAction
            elif isinstance(self.plot.contextMenuActions[idx], SavePlotAction):#SaveAsImageAction):
                # Locate the 'save as image' item, so we can put the csv option next to it
                saveAsImageIndex = idx

        # Put the 'export to csv' option next to the 'save as image' option
        self.plot.contextMenuActions.insert(saveAsImageIndex + 1, self.exportCsvSepAction)
        self.plot.contextMenuActions.insert(saveAsImageIndex + 1, self.exportCsvAction)
        self.plot.contextMenuActions.append(self.configAction)

    def _createConfigWidget(self, parent=None):
        """ Creates a config widget that belongs to this

            The base implementation does nothing. Should be overridden in the descendant.

            :return: a new PlotPanelConfigWidget
        """
        return PlotPanelConfigWidget(parent=parent)

    @property
    def plotTitle(self):
        """ The title of the plot
        """
        return self.plot.getGraphTitle()

    @plotTitle.setter
    def plotTitle(self, value):
        """ The title of the plot

            If set to the empty string, the plot title will be hidden for more screen space.
            If you don't want a visible title but want want the plot to expand you can use a space
            as the title.
        """
        check_class(value, str)
        self.plot.setGraphTitle(value)

    @property
    def xTitle(self):
        """ The X-axis title
        """
        return self.plot.getXAxis().getLabel()

    @xTitle.setter
    def xTitle(self, value):
        """ The X-axis title

            If set to the empty string, the X-axis title will be hidden for more screen space.
            If you don't want a visible title but want want the plot to expand you can use a space
            as the title.
        """
        check_class(value, str)
        self.plot.getXAxis().setLabel(value)

    @property
    def yTitle(self):
        """ The Y-axis title
        """
        return self.plot.getYAxis().getLabel()

    @yTitle.setter
    def yTitle(self, value):
        """ The Y-axis title

            If set to the empty string, the Y-axis title will be hidden for more screen space.
            If you don't want a visible title but want want the plot to expand you can use a space
            as the title.
        """
        check_class(value, str)
        self.plot.getYAxis().setLabel(value)

    @property
    def showGrid(self):
        """ True the grid lines are shown
        """
        return bool(self.plot.getGraphGrid())

    @showGrid.setter
    def showGrid(self, value):
        """ True the grid lines are shown
        """
        self.plot.setGraphGrid(value)

    @property
    def xAxisLogarithmic(self):
        """ True if X-axis' scale is logarithmic, False if it's linear.
        """
        return self.plot.getXAxisLogarithmic()

    @xAxisLogarithmic.setter
    def xAxisLogarithmic(self, value):
        """ True if X-axis' scale is logarithmic, False if it's linear.
        """
        self.plot.setXAxisLogarithmic(value)

    @property
    def yAxisLogarithmic(self):
        """ True if Y-axis' scale is logarithmic, False if it's linear.
        """
        return self.plot.getYAxisLogarithmic()

    @yAxisLogarithmic.setter
    def yAxisLogarithmic(self, value):
        """ True if Y-axis' scale is logarithmic, False if it's linear.
        """
        self.plot.setYAxisLogarithmic(value)

    @property
    def xAxisAutoScale(self):
        """ True if X-axis' scale is auto-scaled,
        """
        return self.plot.isXAxisAutoScale()

    @xAxisAutoScale.setter
    def xAxisAutoScale(self, value):
        """ True if X-axis' scale is auto-scaled,
        """
        self.plot.setXAxisAutoScale(value)

    @property
    def yAxisAutoScale(self):
        """ True if Y-axis' scale is auto-scaled,
        """
        return self.plot.isYAxisAutoScale()

    @yAxisAutoScale.setter
    def yAxisAutoScale(self, value):
        """ True if Y-axis' scale is auto-scaled,
        """
        self.plot.setYAxisAutoScale(value)

    @property
    def plotMode(self):
        """ The current plot mode.
            See PlotModeEnum doc string for possible values and their explanation.
        """
        return self.repo.appendMode

    @plotMode.setter
    def plotMode(self, value):
        """ The current plot mode.
            See PlotModeEnum doc string for possible values and their explanation.
        """
        check_class(value, AppendModeEnum)
        self.repo.appendMode = value

    def setAxesLogarithmic(self, x, y):
        """ Configures the x and/or y axis in logarithmic/linear mode.
        """
        self.plot.setAxesLogarithmic(x, y)

    def addCurveItem(self, curveItem, errorIfExists=False):
        """ Adds a curve to the plot.

            :param bool errorIfExists: if True a DuplicateItemError is raised if a curve with this
                name already exists. If False (the default) the old curve is removed and a warning
                is logged.
        """
        check_class(curveItem, CurveItem)
        logger.debug("Adding plot curve {!r} to {}".format(curveItem.name, self.name))

        if errorIfExists and curveItem.name in self._curveItemsDict:
            raise DuplicateItemError("CurveItem for {} already exists".format(curveItem.name))

        self._curveItemsDict[curveItem.name] = curveItem

        # Set the data from the repo item.
        self.onRepoItemChanged(curveItem.name)

    def removeCurveByName(self, name):
        """ Removes a curve from the plot.

            :param str name: the name of the curve item.
        """
        check_class(name, str)

        if name in self._curveItemsDict:
            # Remove curve from plot widget
            self.plot.remove(name)
            del self._curveItemsDict[name]

    def getCurveNames(self):
        """ Returns a list of the names of the curves that are configured to be plotted.
            :return: list of strings
        """
        return list(self._curveItemsDict.keys())

    def getCurveItemByName(self, name):
        """ Returns a copy of the curve item given by Name.

            Raises ItemNotFoundError if the item is not found

            :rtype: CurveItem
        """
        try:
            return copy.copy(self._curveItemsDict[name])
        except KeyError:
            raise ItemNotFoundError("No CurveItem found having name: {}".format(name))

    def arrayInConfig(self, arrayName):
        """ Returns True if this panel is configured to show data of this array.

            :param arrayName: gridded array name
            :return: True if the array name in the list of configured arrays
        """
        return arrayName in self._curveItemsDict

    def getConfigArrayNames(self):
        """ Returns a list with the names of the array used in the configuration"""
        return list(self._curveItemsDict.keys())

    @pyqtSlot(str)
    def onRepoItemChanged(self, name):
        """ Called when a curveItem is updated in the model. Updates the curve in the plot.
        """
        logger.debug("onRepoItemChanged: {} {}".format(self.name, name))

        try:
            repoItem = self.repo.itemByName(name)
        except ItemNotFoundError:
            logger.debug("Panel {}: no repo item found with this name (ignored): {}"
                           .format(self.name, name))
        else:
            values = repoItem.griddedArray.data
            if values.ndim != 1:
                logger.warning("Data ignored: expected 1D data, got: {}D".format(values.ndim))
                return

            coords = repoItem.griddedArray.makeCoordinateVector(0)

            # Drawing itself is decoupled, here the data is only queued
            #self.plot.queueData(x=coords, y=values, curveName=name)
            self._dataCache[name] = (coords, values)
            self.sigDataQueued.emit(self.name, name) # send new data flag to updater thread

    def doRedraw(self, curveName):
        """ Does a blocking call to redraw the plot.

            :note: Triggered via signal received from plot update scheduler thread.
        """
        # Draw the scheduled data of the curve
        if curveName not in self._dataCache:
            # Data was already drawn, so queue is already cleared
            return
        xData, yData = self._dataCache.pop(curveName)
        noTimestamps = True

        if xData.dtype.kind == 'M':
            # The X coordinates contain datetime64 elements,
            # Convert to list of python datetime.datetime objects so silx can plot them.
            noTimestamps = False
            xData = np.array(xData.tolist())
            assert silx.version_info >= (0, 8, 0), \
                "Timestamps only supported for silx >= 0.8.0. Current = {}".format(silx.version)

            xAxis = self.plot.getXAxis()
            xAxis.setTickMode(silx.gui.plot.TickMode.TIME_SERIES)

        # Check whether to apply auto scaling
        curvePresent = curveName in self.plot.getAllCurves(just_legend=True)
        if not curvePresent:
            # Reset zoom always when receiving 1st data
            resetZoom = True
        else:
            # Scroll when data is appended, but only if the user has not zoomed in
            # Always resetZoom when autoScale is enabled
            resetZoom = self.xAxisAutoScale or self.yAxisAutoScale or not self.plot.isZoomed()

        # Do the actual silx drawing call
        curveItem = self._curveItemsDict.get(curveName)
        if curveItem is None:
            # This can happen when the user has just deleted the curve, but data is still plotting
            logger.warning(
                'Redraw requested on panel %s for non-existing curve: %s', self.name, curveName)
            return

        try:
            # Silx cannot deal with extremely small numbers (see GLPlotFrame:812)
            # and handles only float32 numbers
            # Therefore, clip our data to a safe value (small numbers are set to 0
            FLOAT32_SAFE_PLOT_MIN = -1e-35  #pylint: disable=invalid-name
            FLOAT32_SAFE_PLOT_MAX =  1e-35  #pylint: disable=invalid-name

            # Find occurrences of small numbers
            if noTimestamps:
                # Only if the X coordinates do not contain datetime64 elements,
                xDataSmallEps = (xData > 0) & (np.abs(xData) < FLOAT32_SAFE_PLOT_MAX)

                if np.any(xDataSmallEps):
                    # Zero the X data
                    xData = xData.copy()
                    xEpsPos = (xData > 0) & (xData < FLOAT32_SAFE_PLOT_MAX)
                    xEpsNeg = (xData < 0) & (xData > FLOAT32_SAFE_PLOT_MIN)
                    xData[xEpsPos] = 0.0
                    xData[xEpsNeg] = 0.0
 
            yDataSmallEps = (yData > 0) & (np.abs(yData) < FLOAT32_SAFE_PLOT_MAX)
            if np.any(yDataSmallEps):
                # Zero the Y data
                yData = yData.copy()
                yEpsPos = (yData > 0) & (yData < FLOAT32_SAFE_PLOT_MAX)
                yEpsNeg = (yData < 0) & (yData > FLOAT32_SAFE_PLOT_MIN)
                yData[yEpsPos] = 0.0
                yData[yEpsNeg] = 0.0

            self.plot.addCurve(
                xData, yData,
                legend = curveItem.name,
                label = curveItem.legendName,
                color = "#{:06X}".format(curveItem.color.value),
                linewidth = curveItem.lineWidth,
                linestyle = LINE_STYLE_STR[curveItem.lineStyle],
                symbol = 'o' if curveItem.hasSymbol else None, # no symbol size unfortunately.
                resetzoom = resetZoom,
                selectable = False) # Disable mouse-clicking curve (changes color)
        except Exception as ex: # In case of exceptions within silx, try to not crash hard
            # Notify user
            diagnosticlogger.error('Can not plot {} data'.format(curveName),
                                   "Exception while plotting {} data: {}".format(curveName, ex))

            # Log all details
            logger.exception("Critical exception while plotting %s: %s", curveName, ex)
            logger.error("Relevant plotted data: curveName: %s, xData: %s (%s), yData: %s (%s)",
                         curveName, xData, xData.dtype, yData, yData.dtype)

    @pyqtSlot()
    def onRepoAboutToBeCleared(self):
        """ Is called when the model is about to be cleared.
            Sets all curves to empty data
        """
        self._dataCache = {}
        self.plot.clear()

    def exportToCsv(self):
        """ Prompts the user for a file name and exports the data to CSV
        """
        if self.repo.rowCount() == 0:
            gsswUiWarning("No data in plot. Export canceled.")
            return

        fileName = promptForFileName(self, self.plotTitle if self.plotTitle else 'export', 'txt',
                                     directory='', suggestDateTime=True,
                                     fileFilter="CSV files (*.txt);; All files (*)")
        if not fileName:
            logger.info("Export canceled")
            return

        try:
            self.repo.exportToCsv(fileName)
        except (GsswExportError, Exception) as ex:
            gsswUiWarning(str(ex))

    def exportToCsvSeparate(self):
        """ Prompts the user for a file prefix and exports the data to separate CVS file (one
            per curve)
        """
        if self.repo.rowCount() == 0:
            gsswUiWarning("No data in plot. Export canceled.")
            return

        fileName = promptForFileName(self, self.plotTitle if self.plotTitle else 'export', 'txt',
                                     directory='', suggestDateTime=True,
                                     fileFilter="CSV files (*.txt);; All files (*)")
        if not fileName:
            logger.info("Export canceled")
            return

        try:
            self.repo.exportToCsvSeparate(fileName)
        except (GsswExportError, Exception) as ex:
            gsswUiWarning(str(ex))

    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        dct = super().asDict()

        cfg = {} # Configuration subdict

        cfg['legendSplitter'] = self.plot.splitterSizes()
        cfg['plotTitle'] = self.plotTitle
        cfg['xTitle'] = self.xTitle
        cfg['yTitle'] = self.yTitle
        cfg['xLogScale'] = self.xAxisLogarithmic
        cfg['yLogScale'] = self.yAxisLogarithmic
        cfg['xAxisAutoScale'] = self.xAxisAutoScale
        cfg['yAxisAutoScale'] = self.yAxisAutoScale
        cfg['showGrid'] = self.showGrid
        cfg['plotMode'] = self.plotMode.name
        cfg['maxSamples'] = self.storeMaxSamples
        cfg['curves'] = [curveItem.asDict() for curveItem in self._curveItemsDict.values()]
        dct['cfg'] = cfg

        return dct

    def setValuesFromDict(self, dct):
        """ Sets panel parameters given a dictionary with (config options)
        """
        super().setValuesFromDict(dct)

        cfg = dct['cfg']

        plotWidth, legendWidth = cfg.get('legendSplitter',
            [GsswPlotWidget.SPLITTER_PLOT_WIDTH, GsswPlotWidget.SPLITTER_LEGEND_WIDTH])
        self.plot.splitter.setSizes([plotWidth, legendWidth])

        # We use default values (such as the empty strings) here so that we are robust
        # against changes and so can accept old config files
        self.plotTitle = cfg.get('plotTitle', '')
        self.xTitle = cfg.get('xTitle', '')
        self.yTitle = cfg.get('yTitle', '')
        self.xAxisLogarithmic = cfg.get('xLogScale', False)
        self.yAxisLogarithmic = cfg.get('yLogScale', False)
        self.xAxisAutoScale = cfg.get('xAxisAutoScale', True)
        self.yAxisAutoScale = cfg.get('yAxisAutoScale', True)
        self.showGrid = cfg.get('showGrid', True)
        self.plotMode = AppendModeEnum(AppendModeEnum[cfg.get('plotMode', 'append')])
        self.storeMaxSamples = cfg.get('maxSamples', RepositoryModel.MAX_SAMPLES_BY_DEFAULT)

        # Remove old curveItems. First get a list of all names so that we don't iterate over the
        # _curveItemDict while removing items.
        oldCurveNames = list(self._curveItemsDict.keys())
        for name in oldCurveNames:
            self.removeCurveByName(name)

        self.plot.clearCurves()

        for curveDct in cfg.get('curves', []):
            curveItem = CurveItem(curveDct['name'])
            curveItem.setValuesFromDict(curveDct)
            self.addCurveItem(curveItem, errorIfExists=False)

        # Remove curves from repo that are no longer in the configuration
        configCurveNames = set([curveDct['name'] for curveDct in cfg.get('curves', [])])
        repoCurveNames = set([item.name for item in self.repo.items])
        delCurveNames = repoCurveNames - configCurveNames
        logger.debug("Removing curves from repo that are no longer present in the config: {}"
                     .format(delCurveNames))
        for delCurveName in delCurveNames:
            self.repo.removeItemByName(delCurveName)

    def configureFromArrayInfoList(self, arrInfoItems):
        """ Configures the panel from the items in the array info list

            ;param list arrInfoItems: list of ArrayInfoItems
        """

        appendModeHint = None
        wasEmpty       = not bool(self._curveItemsDict)
        for arrInfoItem in arrInfoItems:
            check_class(arrInfoItem, ArrayInfoItem)

            if self.arrayInConfig(arrInfoItem.name):
                logger.debug("An array with this name is already in the plot: {}"
                             .format(arrInfoItem.name))
            else:
                curveItem = CurveItem(arrInfoItem.name)
                curveItem.label = arrInfoItem.label
                curveItem.color = self.getNextCurveColor() # auto assigned color
                self.addCurveItem(curveItem)
                appendModeHint = arrInfoItem.appendModeHint

        # Automatically detect whether it makes sense to use Append mode rather than replace mode.
        # This is detected by checking for receiving 1 sample at a time instead of a block
        if wasEmpty: # Only change setting if plot panel had no curves in it
            if appendModeHint:
                self.plotMode = AppendModeEnum['append']

        # Set plot labels if all dragged items have the same label values.
        if not self.plotTitle:
            self.plotTitle = same([item.labelOrName for item in arrInfoItems])

        if not self.xTitle:
            self.xTitle = same([item.xLabel for item in arrInfoItems])

        if not self.yTitle:
            self.yTitle = same([item.yLabel for item in arrInfoItems])

        lastName = arrInfoItems[-1].name
        self.execConfigDialog(selectName=lastName)

    def getNextCurveColor(self):
        """ Determines a new color for a the next curve which is added to the panel.
            By using this function, each curve is assigned a different color.
        """
        numUsedColors = len(self._curveItemsDict)
        return getColorForCurve(numUsedColors + START_COLOR_IDX)


class PlotPanelConfigWidget(AbstractPanelConfigWidget):
    """ Contains widgets for configuring PlotPanel settings
    """

    def __init__(self, parent=None):
        """ Constructor
        """
        super().__init__(configModel=PlotCurvesModel(), parent=parent)

    def _addCommonWidgets(self):
        """ Adds specialized (i.e. the non-common) widgets to the config box.

            The base implementation does nothing. Should be implemented by the descendant.
        """
        # Details common to all plot items
        self.titleEdit = QtWidgets.QLineEdit()
        self.commonLayout.addRow("Plot title", self.titleEdit)

        self.xTitleEdit = QtWidgets.QLineEdit()
        self.xLogCheckBox = QtWidgets.QCheckBox("logarithmic")
        self.xAutoScaleCheckBox = QtWidgets.QCheckBox("autoscale")
        xAxis = QtWidgets.QHBoxLayout()
        xAxis.addWidget(self.xTitleEdit)
        xAxis.addWidget(self.xAutoScaleCheckBox)
        xAxis.addWidget(self.xLogCheckBox)
        self.commonLayout.addRow("X-axis", xAxis)

        self.yTitleEdit = QtWidgets.QLineEdit()
        self.yLogCheckBox = QtWidgets.QCheckBox("logarithmic")
        self.yAutoScaleCheckBox = QtWidgets.QCheckBox("autoscale")
        yAxis = QtWidgets.QHBoxLayout()
        yAxis.addWidget(self.yTitleEdit)
        yAxis.addWidget(self.yAutoScaleCheckBox)
        yAxis.addWidget(self.yLogCheckBox)
        self.commonLayout.addRow("Y-axis", yAxis)

        self.showGridCheckBox = QtWidgets.QCheckBox()
        self.showGridCheckBox.setChecked(True)
        self.commonLayout.addRow("Show grid", self.showGridCheckBox)

        self.plotModeComboBox = QtWidgets.QComboBox()
        self.plotModeComboBox.setFocusPolicy(Qt.StrongFocus)
        self.commonLayout.addRow("Plot mode", self.plotModeComboBox)

        for plotMode in AppendModeEnum:
            self.plotModeComboBox.addItem(plotMode.name)

        self.maxSamplesSpinBox = QtWidgets.QSpinBox()
        self.maxSamplesSpinBox.setValue(RepositoryModel.MAX_SAMPLES_BY_DEFAULT)
        self.maxSamplesSpinBox.setMinimum(1)
        self.maxSamplesSpinBox.setMaximum(10000000) # a plot with more samples seems unworkable
        self.maxSamplesSpinBox.setSingleStep(10)
        self.commonLayout.addRow("Max samples", self.maxSamplesSpinBox)

    def _addItemWidgets(self):
        """ Adds widget to the config item group box
        """
        self.labelEdit = QtWidgets.QLineEdit()
        self.labelEdit.setEnabled(True)
        self.itemsFormLayout.addRow("Label", self.labelEdit)
        self.mapper.addMapping(self.labelEdit, PlotCurvesModel.COL_LABEL)

        # Line color
        self.colorComboBox = QtWidgets.QComboBox()
        self.colorComboBox.setFocusPolicy(Qt.StrongFocus)
        self.colorComboBox.setEditable(False)

        for color in ColorEnum:
            # Draw rectangle with color inside combobox.
            iconSizeX, iconSizeY = 12, 12
            pixmap = QtGui.QPixmap(iconSizeX, iconSizeY)
            iconColor = QtGui.QColor(color.value)
            pixmap.fill(iconColor)
            painter = QtGui.QPainter(pixmap)
            painter.setPen(getWidgetColor(self, PaletteRoleEnum.WindowText))
            painter.setBrush(Qt.NoBrush)
            painter.drawRect(QtCore.QRect(0, 0, iconSizeX-1, iconSizeY-1))
            painter.end()
            colorIcon = QtGui.QIcon(pixmap)
            self.colorComboBox.addItem(colorIcon, color.name)

        self.itemsFormLayout.addRow("Color", self.colorComboBox)
        self.mapper.addMapping(self.colorComboBox, PlotCurvesModel.COL_COLOR_NAME, b'currentIndex')

        # Line style (solid, dotted, etc.)
        self.styleComboBox = QtWidgets.QComboBox()
        self.styleComboBox.setFocusPolicy(Qt.StrongFocus)

        for penStyle in LineStyleEnum:
            #self.styleComboBox.addItem(penStyle.name, userData=penStyle.value)
            self.styleComboBox.addItem(penStyle.name)

        self.itemsFormLayout.addRow("Style", self.styleComboBox)

        # We can't map to the currentData property because it's read-only
        self.mapper.addMapping(self.styleComboBox, PlotCurvesModel.COL_STYLE_IDX, b'currentIndex')

        # Line width
        self.widthSpinBox = QtWidgets.QDoubleSpinBox()
        self.widthSpinBox.setMinimum(0)
        self.widthSpinBox.setMaximum(100)
        self.widthSpinBox.setSingleStep(0.1)
        # Keyboard tracking must be True when using the mapper. Otherwise any data that the user
        # types in will be reverted after pressing tab.
        self.widthSpinBox.setKeyboardTracking(True)

        self.itemsFormLayout.addRow("Width", self.widthSpinBox)
        self.mapper.addMapping(self.widthSpinBox, PlotCurvesModel.COL_WIDTH)

        # Symbols
        self.symbolCheckBox = QtWidgets.QCheckBox()
        self.itemsFormLayout.addRow("Symbol", self.symbolCheckBox)
        self.mapper.addMapping(self.symbolCheckBox, PlotCurvesModel.COL_DRAW_SYMBOL)

    def _updateSpecialWidgetsForIndex(self, currentIndex):
        """ Updates the specialized widgets.
        """
        self.labelEdit.setEnabled(currentIndex.isValid())
        self.colorComboBox.setEnabled(currentIndex.isValid())
        self.styleComboBox.setEnabled(currentIndex.isValid())
        self.widthSpinBox.setEnabled(currentIndex.isValid())
        self.symbolCheckBox.setEnabled(currentIndex.isValid())

    def addNewItem(self):
        """ Creates a new curve item and adds it to the model.

            Finds a unique item name by adding numbers to the end. E.g. item-1, item-2, etc.
        """
        name = "Item-0"
        row = self._configModel.rowByName(name, notFoundAction=ActionEnum.IGNORE)
        duplicateNr = 1
        while row is not None and duplicateNr < 1000:
            name = "Item-{}".format(duplicateNr)
            row = self._configModel.rowByName(name, notFoundAction=ActionEnum.IGNORE)
            duplicateNr += 1
        assert row is None, "Unable to add unique curve name."

        self._addNewCurveToModel(name, '')
        self.nameComboBox.setFocus(Qt.OtherFocusReason)
        self.nameComboBox.lineEdit().selectAll()

    #@pyqtSlot(CurveItem)
    def _createConfigItem(self, name):  # Not used
        """ Creates a new config item and adds it to the model.
        """
        return CurveItem(name)

    @pyqtSlot(CurveItem)
    def _addNewCurveToModel(self, name, label):
        """ Creates a new curve item and adds it to the model. Selects the new item in the list.

            :param name: unique name of the new model item
            :param label: label (e.g. for use in legend).
            :param row: Where the item is inserted. If row=None, the item is appended at the end.
            :param removeExisting: If True, any existing curve item is removed. If False (the
                default), a KeyError is raised if an item with the same names already exists
            :returns: the newly added item.
        """
        try:
            curveItem = self._createConfigItem(name)
            curveItem.label = label

            # Auto-rotate through the available colors for each new curve
            curveItem.color = getColorForCurve(self._configModel.rowCount() + START_COLOR_IDX)

            self._configModel.addItem(curveItem)

        except DuplicateItemError:
            logger.warning("Item already exists: {}".format(name))
        else:
            row = self._configModel.rowByName(name)
            self.setCurrentRow(row)

            return curveItem

    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        self.mapper.submit()

        cfg = {}

        cfg['plotTitle'] = self.titleEdit.text()
        cfg['xTitle'] = self.xTitleEdit.text()
        cfg['yTitle'] = self.yTitleEdit.text()
        cfg['xLogScale'] = self.xLogCheckBox.isChecked()
        cfg['yLogScale'] = self.yLogCheckBox.isChecked()
        cfg['xAxisAutoScale'] = self.xAutoScaleCheckBox.isChecked()
        cfg['yAxisAutoScale'] = self.yAutoScaleCheckBox.isChecked()
        cfg['showGrid'] = self.showGridCheckBox.isChecked()
        cfg['plotMode'] = self.plotModeComboBox.currentText()
        cfg['maxSamples'] = self.maxSamplesSpinBox.value()

        cfg['curves'] = [item.asDict() for item in self._configModel.items]

        return cfg

    def setValuesFromDict(self, cfg):
        """ Creates a panel from a dictionary with values
        """
        logger.debug("setValuesFromDict: {}".format(cfg))
        self.titleEdit.setText(cfg.get('plotTitle', ''))
        self.xTitleEdit.setText(cfg.get('xTitle', ''))
        self.yTitleEdit.setText(cfg.get('yTitle', ''))
        self.xLogCheckBox.setChecked(cfg.get('xLogScale', False))
        self.yLogCheckBox.setChecked(cfg.get('yLogScale', False))
        self.xAutoScaleCheckBox.setChecked(cfg.get('xAxisAutoScale', False))
        self.yAutoScaleCheckBox.setChecked(cfg.get('yAxisAutoScale', False))
        self.showGridCheckBox.setChecked(cfg.get('showGrid', True))
        self.plotModeComboBox.setCurrentText(cfg.get('plotMode', 'append'))
        # self.maxSamplesSpinBox.setValue(cfg.get('maxSamples',
        #                                         RepositoryModel.MAX_SAMPLES_BY_DEFAULT))
        self.maxSamplesSpinBox.setValue(cfg['maxSamples'])

        firstItemName = None
        for curveDct in cfg.get('curves', []):
            if not firstItemName:
                firstItemName = curveDct['name']
            curveItem = CurveItem(curveDct['name'])
            curveItem.setValuesFromDict(curveDct)
            self._configModel.addItem(curveItem)

        if self.getCurrentRow()==-1:
            if firstItemName is not None:
                logger.debug('Select 1st one by name: %r', firstItemName)
                self.setCurrentRowByName(firstItemName)
            else:
                logger.debug('Select 1st record at index 0')
                self.setCurrentRow(0)

        self.mapper.revert()
