""" Contains classes for interfacing with the quicklooker.

Additionally, there are some convenience function(s) for quickly sending data fo plotting:

* :meth:`QlPlot.plot`
* :meth:`QlPlot.update`

.. note::
    If the documentation below is unclear, or when you do not understand how
    to use it in your script, please contact the software development team.
"""

import datetime
import logging

import numpy as np

from gssw.common               import getGsswConfig
from gssw.lib.arrays.gridded   import GriddedArray
from gssw.lib.arrays.semantic  import SemanticArray
from gssw.lib.config           import Config
from gssw.lib.errors           import GsswError
from gssw.lib.misc             import stringToIdentifier
from gssw.lib.rpcbase          import BaseClient
from gssw.lib.sockets          import ReqSocket
from gssw.lib.typing           import check_class, check_is_a_sequence

# Make sure not to import the plotpanel (or other modules that use Silx) here, neither directly or
# indirectly. This makes it possible to use the client without installing silx. For a justification
# see test_silxNotImported function in testscripts/ftests/ftest_lineplotclient.py
from gssw.quicklook.defs       import BinAlignmentEnum, ColorEnum, LineStyleEnum, LINE_STYLE_STR
from gssw.quicklook.defs       import createQuicklookPacketizer
from gssw.quicklook.repo       import AppendModeEnum

from gssw.proto.common_pb2     import OkRep
from gssw.proto.quicklook_pb2  import AddNewPageReq,             AddNewPageRep
from gssw.proto.quicklook_pb2  import RemovePageReq
from gssw.proto.quicklook_pb2  import SetCurrentPageReq
from gssw.proto.quicklook_pb2  import AddPlotPanelReq,           AddPlotPanelRep
from gssw.proto.quicklook_pb2  import AddHistogramPanelReq,      AddHistogramPanelRep
from gssw.proto.quicklook_pb2  import AddImagePanelReq,          AddImagePanelRep
from gssw.proto.quicklook_pb2  import RemovePanelReq
from gssw.proto.quicklook_pb2  import AddCurveToPlotPanelReq,    AddCurveToPlotPanelRep
from gssw.proto.quicklook_pb2  import RemoveItemFromPanelReq,    RemoveItemFromPanelRep
from gssw.proto.quicklook_pb2  import ClearPanelDataReq,         ClearPanelDataRep
from gssw.proto.quicklook_pb2  import AppendArrayReq,            AppendArrayRep
from gssw.proto.quicklook_pb2  import GetPanelNamesReq,          GetPanelNamesRep
from gssw.proto.quicklook_pb2  import GetPageNamesReq,           GetPageNamesRep
from gssw.proto.quicklook_pb2  import GetCurveNamesReq,          GetCurveNamesRep

# Indicates which functionality is typically imported by the client users.
__all__ = ['QlPlot', 'QuicklookClient',
           'ColorEnum', 'LineStyleEnum', 'AppendModeEnum', 'BinAlignmentEnum']


logger = logging.getLogger(__name__)

#: Default name of the Quicklooker component in the GSSW configuration file
#: The QlClients connect to this component
QUICKLOOKER_COMPONENT = 'quicklooker'


class QlPlot:
    """ Proxy class with convenience functions for scripts to easily plot to Quicklooker.

        :Warning:
            Before using these convenience funcitions, :func:`gssw.common.initGssw` must be
            called first. Below is a typical example of how to setup a script.

            .. code-block:: python

                # Necessary imports
                from gssw.lib.misc              import createGsswArgumentParser
                from gssw.common                import initGssw
                from gssw.quicklook.qlclient    import QlPlot

                # This part needs to be run at the beginning of scripts
                parser = createGsswArgumentParser(__doc__)
                args   = parser.parse_args()
                config = initGssw(args)

                # Now the convenience functions (and other gssw methods) can be used
                QlPlot.plot("Example", [0, 1, 2, 3], [0, 2, 4, 6])

    """

    _altLineStyle = {value: key for key, value in LINE_STYLE_STR.items()}
    _qlclient = None

    @staticmethod
    def _getAxisTitle(kwargs, axis, label, units):
        """ Get axis title from label and units. """
        label = kwargs.get(label)
        units = kwargs.get(units)
        if any((label, units)):
            title = label if label else axis
            title += ' [{}]'.format(units) if units else ''
            return title
        else:
            return None

    @staticmethod
    def _getEnum(value, enum):
        """ Ensure that color is defined as ColorEnum. If not than convert from string."""
        if isinstance(value, str):
            value = enum[value.lower()]
        check_class(value, enum, allow_none=True)
        return value

    @staticmethod
    def _getLinestyleEnum(lineStyle):
        """ Ensure that lineStyle is defined as LineStyleEnum. If not than convert from string."""
        if isinstance(lineStyle, str):
            if lineStyle in QlPlot._altLineStyle:
                lineStyle = QlPlot._altLineStyle[lineStyle]
            else:
                lineStyle = LineStyleEnum[lineStyle.lower()]
        check_class(lineStyle, LineStyleEnum, allow_none=True)
        return lineStyle

    @staticmethod
    def _getXYData(*args):
        """ Get the X and Y data from the args. """

        # Get x and y from arguments
        if len(args) == 1:
            yData = args[0]
        elif len(args) == 2:
            xData = args[0]
            yData = args[1]
        else:
            raise GsswError("Invalid number of values")

        # Ensure data is in iterable form
        try:
            assert len(xData) == len(yData)
        except AssertionError:
            raise GsswError("X and Y data must be the same length.")
        except NameError:  # x is not defined -> only y is defined
            xData = None
            yData = [yData]
        except TypeError:  # must be single numbers, defined as list
            xData = [xData]
            yData = [yData]

        return xData, yData

    @staticmethod
    def _getGriddedArray(name, xData, yData, **kwargs):
        """ Create a gridded aray from the x/y data and keyword arguments. """

        # Create Gridded Array
        label = kwargs.pop('label', None)
        units = kwargs.pop('units', None)
        gArray = GriddedArray(yData, name=name, label=label, units=units)

        # Set x-axis data if defined
        if xData is not None:
            if isinstance(xData[0], datetime.date):
                xData = np.array([np.datetime64(x) for x in xData], dtype='datetime64[ms]')
                xLabel = 'Time'
                xUnits = 's'
            else:
                xData = np.array(xData)
                xLabel = kwargs.pop('xlabel', None)
                xUnits = kwargs.pop('xunits', None)
            gArray.coords[0] = SemanticArray(xData, name='x', label=xLabel, units=xUnits)

        return gArray, kwargs

    @classmethod
    def plot(cls, name, *args, **kwargs):
        """ Plot the the value(s) to the Quicklooker. Creates a page/panel in Quicklooker (if page
            or panel does not exist), adds the curve to the plot (if curve does not already exist),
            and adds the data to the curve. If either :attr:`pageName` or :attr:`panelName` is not
            specified, a generic name will be generated. The :attr:`pageName` and :attr:`panelName`
            are returned and can be used to append data to the plot.

            The default mode is 'append' so that subsequent calls to this method with the same
            :attr:`name`, :attr:`pageName`, and :attr:`panelName` will result in the new data
            being appended to the array. Use :code:`appendMode='replace'` to overwrite the data
            on each plot call.

            :param name:    The name for the array in Quicklooker.

            Data for X and Y can be supplied as args as numbers, lists, or arrays as follows.

            :Args:
                * If one value is supplied, then it is taken as Y.
                * If two values (or two lists of values) then they are taken as (X, Y) in that order
                * X and Y can be single values or lists/array of values (length must be equal)
                * X can be numbers or datetime objects (for example datetime.now())

            :Example:

                .. code-block:: python

                    QlPlot.plot("MyArray", 5)                    # Plot y=5
                    QlPlot.plot("MyArray", 1, 5)                 # Plot x=1, y=5
                    QlPlot.plot("MyArray", [1,2,3], [10,20,30])  # Plot x=[1,2,3], y=[10,20,30]
                    QlPlot.plot("MyArray", datetime.now(), 5)    # Plot value 5 at timestamp

            Extra keywords can be given to supply extra information.
            Below are some of the more useful kwargs.

            :Kwargs:
                * pageName (str):   The page name (if not supplied then will use Page-#)
                * panelName (str):  The panel name (if not supplied then will use Panel-#)
                * appendMode (str): Whether to 'append' or 'replace' array values. Default: append
                * plotTitle (str):  The title of the plot
                * xTitle (str):     The title of the X-axis (overrides label + units)
                * yTitle (str):     The title of the Y-axis (overrides label + units)
                * color (str):      The color for the curve
                * lineStyle (str):  The style for the line ('solid' and '-' are equivalent)
                * lineWidth (int):  The thickness for the curve
                * hasSymbol (bool): Whether to draw a dot for every data point.
                * label (str):   The label for the Y-axis (and legend)
                * units (str):   The units of the Y-axis
                * xlabel (str):  The label for the X-axis
                * xunits (str):  The units of the X-axis

                * See :meth:`QuicklookClient.addPlotPanel` for other panel options.
                * See :meth:`QuicklookClient.addCurveToPlotPanel` for other plot options.

            :Note: If pageName exists, then the panel will be added to the page. If both pageName
                   and panelName exist, the curve will be added to the existing panel on the page.
                   If the curve exists on the specified panel, then the data will be appended or
                   replaced depending on appendMode.
            :Note: Subsequent calls to plot will retain style settings from previous calls unless
                   explicitly redefined.

            :return: the pageName and panelName are returned as strings.

            :Example:

                .. code-block:: python

                    # create plot with unspecified page/panel
                    page, panel = QlPlot.plot("MyArray", [0, 1, 2], [0, 20, 40], hasSymbol=True,
                                              color='blue', lineWidth=2, lineStyle='-')
                    # add data to the same plot (note that previous style options are retained)
                    QlPlot.plot("MyArray", 3, 60, pageName=page, panelName=panel)
                    # add another curve to same plot
                    QlPlot.plot("NewArray", [0, 1, 2], [0, 10, 20], pageName=page, panelName=panel)
                    # overwrite 'MyArray' data with new values, change line color to green
                    QlPlot.plot("MyArray", [0, 1, 2], [5, 10, 15], color='green',
                                pageName=page, panelName=panel, appendMode='replace')
        """
        if cls._qlclient is None:
            cls._qlclient = QuicklookClient(config=getGsswConfig())

        # Add Page
        pageName = kwargs.pop('pageName', None)
        newPage = cls._qlclient.addNewPage(pageName)
        pageName = newPage if newPage else pageName

        cls._qlclient.setCurrentPage(pageName)

        # Add Panel
        panelKwargs = {arg: kwargs.pop(arg, None) for arg in
                       ['panelName', 'gridGeometry', 'storeMaxSamples', 'showGrid',
                        'plotTitle', 'xTitle', 'yTitle', 'xAxisLogarithmic', 'yAxisLogarithmic']}
        panelKwargs['appendMode'] = cls._getEnum(kwargs.pop('appendMode', 'append'), AppendModeEnum)
        if panelKwargs['xTitle'] is None:
            panelKwargs['xTitle'] = cls._getAxisTitle(kwargs, 'x', 'xlabel', 'xunits')
        if panelKwargs['yTitle'] is None:
            panelKwargs['yTitle'] = cls._getAxisTitle(kwargs, 'y', 'label', 'units')
        panelName = cls._qlclient.addPlotPanel(pageName, **panelKwargs)

        # Get curve options
        plotKwargs = {arg: kwargs.pop(arg, None) for arg in ['lineStyle', 'lineWidth',
                                                             'color', 'hasSymbol']}
        plotKwargs['label'] = kwargs.get('label')  # don't pop since needed for publish

        # Convert strings to enums if needed
        plotKwargs['color'] = cls._getEnum(plotKwargs['color'], ColorEnum)
        plotKwargs['lineStyle'] = cls._getLinestyleEnum(plotKwargs['lineStyle'])

        # Add curve to plot
        cls._qlclient.addCurveToPlotPanel(pageName, panelName, name, **plotKwargs)

        # Update data
        cls.update(name, *args, **kwargs)

        return pageName, panelName

    @classmethod
    def update(cls, name, *args, **kwargs):
        """ Update the data in Quicklooker. If the array :attr:`name` has been previously added
            to a Quicklooker panel(s), then the plot(s) will be updated with the new data.
            This is useful to update an already existing plot(s) with new data.

            :param name:    The name for the array in Quicklooker.

            :Note: This is a blocking call. For a non-blocking publish, see
                   :class:`gssw.lib.arrays.pubclient`

            Data for X and Y can be supplied as args.

            :Args:
                * If one value is supplied, then it is taken as Y.
                * If two values (or two lists of values) then they are taken as (X, Y) in that order
                * X and Y can be single values or lists/array of values (length must be equal)
                * X can be numbers or datetime objects (for example datetime.now())

            :Example:

                .. code-block:: python

                    QlPlot.update("MyArray", 5)                    # Add y=5 to MyArray
                    QlPlot.update("MyArray", 1, 5)                 # Add x=1, y=5
                    QlPlot.update("MyArray", [1,2,3], [10,20,30])  # Add x=[1,2,3], y=[10,20,30]
                    QlPlot.update("MyArray", datetime.now(), 5)    # Add value 5 at timestamp

            Extra keywords can be given to supply extra information.

            :Kwargs:
                * label (str):   The label for the Y-axis (and legend)
                * units (str):   The units of the Y-axis
                * xlabel (str):  The label for the X-axis
                * xunits (str):  The units of the X-axis

            :Note: These values will not overwrite plot labels on existing plots.
        """

        if cls._qlclient is None:
            cls._qlclient = QuicklookClient(config=getGsswConfig())

        xData, yData = cls._getXYData(*args)
        gArray, kwargs = cls._getGriddedArray(name, xData, yData, **kwargs)
        cls._qlclient.appendArray(gArray)

        if kwargs:
            logger.warning("The following kwargs were not recognized: %s", kwargs)

class QuicklookClient(BaseClient):
    """ Client class for controlling with the Quicklooker via request-reply
    """

    # Exported enum aliases, for easy access from scripts which use a Quicklook Client
    MODE_APPEND  = AppendModeEnum.append
    MODE_REPLACE = AppendModeEnum.replace

    def __init__(self, address=None, reqSocket=None, config=None):
        """ Constructor

            The address and reqSocket and config parameters cannot be used together. Use either an
            addres, or supply a reqSocket, or provide the Egress config object.

            :param address: Connect to the Quicklooker via this address.
            :param reqSocket: Use the given socket for connecting to the Quicklooker.
            :param Config config: Look up the Quicklooker automatically in the config file.
        """

        # Check only one argument is set
        assert sum([int(arg is not None) for arg in [address, reqSocket, config]]) == 1,\
            "One argument allowed for creating a client"

        if config is not None:
            check_class(config, Config)
            self.config = config

            # Check if the Quicklooker name is specified in the configuration
            if QUICKLOOKER_COMPONENT not in self.config['components']:
                logger.critical("Quicklooker '%s' was not found in the configuration file",
                    QUICKLOOKER_COMPONENT)
                raise GsswError("The Quicklooker was not found in the configuration file")

            # Create a REQ socket used to connect to the Quicklooker
            reqSocket = ReqSocket()
            reqSocket.createZmqSocket()
            qlPort = config.componentParameters(QUICKLOOKER_COMPONENT)['sockets']['ql-rep']
            qlHost = config.componentParameters(QUICKLOOKER_COMPONENT)['hostname']
            qlAddress = "tcp://{}:{}".format(qlHost, qlPort)
            reqSocket.connectSocket(qlAddress)

        super().__init__(address=address, reqSocket=reqSocket)

        self.packetizer = createQuicklookPacketizer()

    def finalize(self):
        """ Can be used to cleanup resources. Current implementation does nothing.
            Must be called explicitly.
        """
        pass  # pylint: disable=unnecessary-pass

    #############################
    # Quicklooker functionality #
    #############################

    def addNewPage(self, pageName=None, row=None, select=True, errorIfExists=False):
        """ Creates an empty page and adds it to the page list.

            Remotely calls :meth:`gssw.quicklook.main.MainWindow.addNewPage`

            :param pageName: name of the created page. If None, the name will be 'page-X', where X
                is the first number that yields a unique name.
                In case a page with this name exists an warning is logged.

            :param row: optional row number at which the page will be inserted. If None, the page
                will be inserted at the end.

            :param bool select: If True (the default) it will be selected and set as the current.

            :param bool errorIfExists: if True a DuplicateItemError is raised if a page with this
                name already exists. If False (the default) the old page is removed and a warning
                is logged.

            :return: the name of the newly created panel. The empty string if no new page has been
                created.
        """
        requestMsg = AddNewPageReq(pageName=pageName, row=row, select=select,
                                   errorIfExists=errorIfExists)
        replyMsg = self.requestReply(requestMsg, AddNewPageRep())
        return replyMsg.pageName


    def removePage(self, pageName):
        """ Deletes a page given its name.

            Raises a KeyError if no page with that name exists.

            Remotely calls :meth:`gssw.quicklook.main.MainWindow.removePage`

            :param str pageName: the name of the page that is deleted. If None
        """
        requestMsg = RemovePageReq(pageName=pageName)
        self.requestReply(requestMsg, OkRep())


    def setCurrentPage(self, pageName):
        """ Selects the page with this name in the page overview.

            If the pageName is empty, no new page will be made, only the old one will be removed.

            Raises a key error if no page with this name exists

            Remotely calls :meth:`gssw.quicklook.main.MainWindow.setCurrentPage`
        """
        requestMsg = SetCurrentPageReq(pageName=pageName)
        self.requestReply(requestMsg, OkRep())


    def addPlotPanel(self,  #pylint: disable=too-many-arguments
                     pageName,
                     panelName = None,
                     errorIfExists=False,
                     gridGeometry=None,
                     appendMode=None,
                     storeMaxSamples=None,
                     plotTitle=None,
                     xTitle=None,
                     yTitle=None,
                     xAxisLogarithmic=None,
                     yAxisLogarithmic=None,
                     showGrid=None):
        """ Adds a plot panel

            Remotely calls :meth:`gssw.quicklook.main.MainWindow.addPlotPanel`

            :param str pageName: name of the page where the panel will be created.
                An ItemNotFoundError is raised if this page does not exist.
            :param str panelName: name of the panel, If None, panel-N will be used where N will be
                the first integer that is not used.
            :param bool errorIfExists: if True a DuplicateItemError is raised if a page with this
                name already exists. If False (the default) the old page is removed and a warning
                is logged.
            :param dict gridGeometry: panel geometry in grid coordinates. Grid coordinates range
                from 0 to the number of grid points (36 in the current implementation)
                The dict should have keys: x, y, width, height
                If None, the new panel will cover a quarter of the page.
            :param AppendModeEnum appendMode: indicates if new array data is appended to the old
                data or replaces the old data.
            :param int storeMaxSamples: maximum number of samples that are stored per array
            :param str plotTitle: optional plot title
            :param str xTitle: optional title for the x-axis
            :param str yTitle: optional title for the y-axis
            :param bool xAxisLogarithmic: Display x axis in logarithmic scale.
            :param bool yAxisLogarithmic: Display y axis in logarithmic scale.
            :param bool showGrid: if True the grid lines are displayed.

            :returns: the name of the newly created panel. Returns the emtpy string if
                the panel already existed.
        """
        check_class(gridGeometry, dict, allow_none=True)
        check_class(appendMode, AppendModeEnum, allow_none=True)
        appendModeValue = appendMode if appendMode is None else appendMode.value

        requestMsg = AddPlotPanelReq(pageName=pageName, panelName=panelName,
                                     errorIfExists=errorIfExists, gridGeometry=gridGeometry,
                                     appendMode=appendModeValue, storeMaxSamples=storeMaxSamples,
                                     plotTitle=plotTitle, xTitle=xTitle, yTitle=yTitle,
                                     xAxisLogarithmic=xAxisLogarithmic,
                                     yAxisLogarithmic=yAxisLogarithmic,
                                     showGrid=showGrid)

        replyMsg = self.requestReply(requestMsg, AddPlotPanelRep())
        return replyMsg.panelName


    def addHistogramPanel(self,  #pylint: disable=too-many-arguments
                          pageName = None,
                          panelName = None,
                          errorIfExists=False,
                          gridGeometry=None,
                          plotTitle = None,
                          xTitle = None,
                          yTitle = None,
                          showGrid = None,
                          color = None,
                          binAlignment=None,
                          arrayName=None):
        """ Adds a histogram panel with name panelName on the page with name pageName.

            Remotely calls :meth:`gssw.quicklook.main.MainWindow.addHistogramPanel`

            If a panel on the page already exists (and errorIfExists is False, the default), a
            message is logged and the emtpy string is returned. If errorIfExists is True, a
            DuplicateItemError exception is raised.

            If you want to ensure that a new panel is always used you must call removePanel
            beforehand.

            If the following parameters are omitted, defaults are used.
            See :meth:`_createAndAddPanel` for the other parameters.

            :param str plotTitle: optional plot title
            :param str xTitle: optional title for the x-axis
            :param str yTitle: optional title for the y-axis
            :param bool showGrid: if True the grid lines are displayed.
            :param ColorEnum color: a value attribute of a ColorEnum enumeration object
            :param BinAlignmentEnum binAlignment: a value attribute of a BinAlignment enum object
            :param string arrayName: the name of the gridded array that is visualized

            :returns: string with the name of the newly created panel. Returns the emtpy string if
                the panel already existed.
        """
        check_class(gridGeometry, dict, allow_none=True)
        color = None if color is None else color.value
        binAlignment = None if binAlignment is None else binAlignment.value

        requestMsg = AddHistogramPanelReq(
            pageName=pageName, panelName=panelName,
            errorIfExists=errorIfExists, gridGeometry=gridGeometry,
            plotTitle=plotTitle, xTitle=xTitle, yTitle=yTitle,
            showGrid=showGrid, color=color, binAlignment=binAlignment,
            arrayName=arrayName)

        replyMsg = self.requestReply(requestMsg, AddHistogramPanelRep())
        return replyMsg.panelName


    #pylint: disable=too-many-arguments
    def addImagePanel(self,
                      pageName = None,
                      panelName = None,
                      errorIfExists=False,
                      gridGeometry=None,
                      plotTitle = None,
                      xTitle = None,
                      yTitle = None,
                      colorBarTitle = None,
                      colorMap = None,
                      colorAutoRange=None,
                      colorRange=None,
                      arrayName=None):
        """ Adds a 2D image plot panel with name panelName on the page with name pageName.

            Remotely calls :meth:`gssw.quicklook.main.MainWindow.addImagePanel`

            If a panel on the page already exists (and errorIfExists is False, the default), a
            message is logged and the emtpy string is returned. If errorIfExists is True, a
            DuplicateItemError exception is raised.

            If you want to ensure that a new panel is always used you must call removePanel
            beforehand.

            If the following parameters are omitted, defaults are used.
            See :meth:`_createAndAddPanel` for the other parameters.

            :param str plotTitle: optional plot title
            :param str xTitle: optional title for the x-axis
            :param str yTitle: optional title for the y-axis
            :param str colorBarTitle: optional title for in the color bar
            :param str colorMap: color scale. At least the following names are supported by the
                Silx plot library: gray, reversed gray, temperature, red, green, blue, jet,
                viridis, magma, inferno, plasma.
            :param bool colorAutoRange: if True the color range will be determined from the data.
            :param tuple colorRange: tuple with 2 elements defining the color map range.
            :param string arrayName: the name of the gridded array that is visualized

            :returns: string with the name of the newly created panel. Returns the emtpy string if
                the panel already existed.
        """
        check_class(gridGeometry, dict, allow_none=True)
        check_is_a_sequence(colorRange, allow_none=True)

        if colorRange:
            assert len(colorRange) == 2, "colorRange must be of length 2. Got: {}"\
                .format(len(colorRange))

        requestMsg = AddImagePanelReq(
            pageName=pageName, panelName=panelName,
            errorIfExists=errorIfExists, gridGeometry=gridGeometry,
            plotTitle=plotTitle, xTitle=xTitle, yTitle=yTitle, colorBarTitle=colorBarTitle,
            colorMap=colorMap, colorAutoRange=colorAutoRange, colorRange=colorRange,
            arrayName=arrayName)

        replyMsg = self.requestReply(requestMsg, AddImagePanelRep())
        return replyMsg.panelName


    def removePanel(self, pageName, panelName):
        """ Removes a panel from the current page, given its name

            Remotely calls :meth:`gssw.quicklook.main.MainWindow.removePanel`

            :param str pageName: name of the page where the panel will be created.
                An ItemNotFoundError is raised if this page does not exist.
            :param str panelName: panel name
        """
        requestMsg = RemovePanelReq(pageName=pageName, panelName=panelName)
        self.requestReply(requestMsg, OkRep())


    def addCurveToPlotPanel(self, pageName, panelName, arrayName,
                            errorIfExists=False,
                            label = None,
                            lineStyle = None,
                            lineWidth = None,
                            color = None,
                            hasSymbol = None):
        """ Adds a curve to the plot in the panel on the current page.

            Note that if the panel is not a PlotPanel a TypeError is raised.

            Remotely calls :meth:`gssw.quicklook.main.MainWindow.addCurveToPanel`

            :param str pageName: page name
            :param str panelName: panel name
            :param str arrayName: name of the array that the curve item visualizes.
            :param bool errorIfExists: if True a DuplicateItemError is raised if a curve with this
                name already exists. If False (the default) the old curve is replaced by the new
                curve. In that case parameter (e.g. lineStyle) that are None are take from the
                old curve.
            :param str label: label that is used in the plot (if None, the arrayName is used)
            :param float lineWidth: width of the line. Also influences symbol size.
            :param LineStyleEnum lineStyle: a value attribute of a LineStyleEnum object
            :param ColorEnum color: a value attribute of a ColorEnum object
            :param bool hasSymbol: if True, a dot is drawn for every data point.
        """
        check_class(lineStyle, LineStyleEnum, allow_none=True)
        check_class(color, ColorEnum, allow_none=True)

        requestMsg = AddCurveToPlotPanelReq(
            pageName = pageName,
            panelName = panelName,
            arrayName = arrayName,
            errorIfExists = errorIfExists,
            label = label,
            lineStyle = None if lineStyle is None else lineStyle.value,
            lineWidth = lineWidth,
            color = None if color is None else color.value,
            hasSymbol = hasSymbol
        )
        self.requestReply(requestMsg, AddCurveToPlotPanelRep())
    #pylint: enable=too-many-arguments

    def removeItemFromPanel(self, pageName, panelName, arrayName):
        """ Removes an item (e.g. a plot curveItem) from the panel

            Currently only implemented for the PlotPanel, not for the TablePanel (TODO)

            Remotely calls :meth:`gssw.quicklook.main.MainWindow.removeItemFromPanel`

            :param str pageName: page name
            :param panelName: panel name
            :param arrayName: name of the array that the curve item visualizes.
        """
        requestMsg = RemoveItemFromPanelReq(
            pageName=pageName, panelName=panelName, arrayName=arrayName)
        self.requestReply(requestMsg, RemoveItemFromPanelRep())


    def clearPanelData(self, pageName, panelName):
        """ Appends gridded array data to the corresponding array in the repository.

            Remotely calls :meth:`gssw.quicklook.main.MainWindow.clearPanelData`

            :param str pageName: page where the panel should be present.
            :param str panelName: panel name
        """
        requestMsg = ClearPanelDataReq(pageName=pageName, panelName=panelName)
        self.requestReply(requestMsg, ClearPanelDataRep())


    def appendArray(self, griddedArray):
        """ Appends gridded array data to the corresponding array in the plots.

            Remotely calls :meth:`gssw.quicklook.main.MainWindow.appendArray`

            NOTE: this is done with request-reply and therefore blocking. If you want to use
            publish-subscribe, you can send GriddedArray message to the quicklooker. The
            ArrayPublisherClient can do this conveniently.
        """
        check_class(griddedArray, GriddedArray)
        requestMsg = AppendArrayReq()
        #griddedArray.addToProtoBuf(requestMsg.array)
        self.packetizer.setMessageAttrs(requestMsg.array, griddedArray)
        self.requestReply(requestMsg, AppendArrayRep())


    def getPageNames(self):
        """ Retrieve the list of pages available in the Quicklooker.
        """
        reqMsg = GetPageNamesReq()
        repMsg = self.requestReply(reqMsg, GetPageNamesRep())
        return self.packetizer.getMessageAttrs(repMsg)['pageNames']


    def getPanelNames(self, pageName):
        """ Retrieve the list of panels available in the Quicklooker pages.

            :param str page: Get list from this page.
        """
        reqMsg = GetPanelNamesReq()
        reqMsg.pageName = pageName
        repMsg = self.requestReply(reqMsg, GetPanelNamesRep())
        return self.packetizer.getMessageAttrs(repMsg)['panelNames']


    def getCurves(self, pageName, panelName):
        """ Retrieve the list of curves from a specific panelName.

        :param str pageName:  Name of the page which contains the panelName.
        :param str panelName: Name of the panelName.
        """
        reqMsg = GetCurveNamesReq()
        reqMsg.pageName = pageName
        reqMsg.panelName = panelName
        repMsg = self.requestReply(reqMsg, GetCurveNamesRep())
        return self.packetizer.getMessageAttrs(repMsg)['curveNames']


class LinePlotQlClient():
    """ An object oriented, high level client to create line plots on a single panel.
    """

    def __init__(self, qlClient, pageName, panelName,
                 forceNewPage = False, forceNewPanel = False,
                 prefix = None, overplot = False, overplotReplace = True):
        """ Constructor

            :param QuicklookClient qlClient: reference to a quicklook client object
            :param str pageName: name of the page where the line will be plotted
            :param str panelName: name of the panel the line will be plotted
            :param bool forceNewPage: if True the old page with that name is removed.
            :param bool forceNewPage: if True the old panel with that name is removed.
            :param str None prefix: unique string that identifies the panel. Is prepended to the
                array names to make sure that they only apprear on the panel. If None, a default
                prefix is constructed from the page and panel name.
            :param bool overplot: If True the old array data is not removed when plotting the
                new data. This makes a kind of 'waterfall' plot possible. This is achieved by
                appending a sequence counter to the array name. If False (the default) old array
                data is removed or appended.
            :param bool overplotReplace: Start new curve at every plot call, set to False when using
                                         the client with append-mode data, and call :meth:`.next`
                                         for every new curve.
        """
        check_class(qlClient, QuicklookClient)
        check_class(pageName, str)
        check_class(panelName, str)
        check_class(overplot, bool)
        check_class(prefix, str, allow_none=True)

        self.qlClient = qlClient
        self._overplot = overplot
        self._overplotReplace = overplotReplace
        self._numCurves = 0

        self.pageName = pageName
        self.panelName = panelName
        self.prefix = (prefix if prefix is not None else
                       stringToIdentifier('{}__{}'.format(pageName, panelName)))

        if forceNewPage:
            self.qlClient.removePage(pageName)

        if forceNewPanel:
            self.qlClient.removePanel(pageName, panelName)

        self.qlClient.addNewPage(self.pageName)
        self.qlClient.addPlotPanel(self.pageName, self.panelName)

    def plot(self, data, label,
             arrayName = None,
             lineStyle = None,
             lineWidth = None,
             color = None,
             hasSymbol = None):
        """ Plot x, y with a label to use in the legend.

            A curve will be created in the line plot. It's name will be automatically generated
            and given back as return value.

            :param data: Either a tuple (numpy.ndarray, numpy.ndarray) with x- and y-coordinates,
                         or a GriddedArray.
            :param str label: label to use in the legend.
            :param str arrayName: can be used to override the name of the created array in the
                plot configuration. Typically you don't need to do this, the array name will then
                be automatycally generated. This auto-generated name is the return value.
            :param float lineWidth: width of the line. Also influences symbol size.
            :param LineStyleEnum lineStyle: a value attribute of a LineStyleEnum object
            :param ColorEnum color: a value attribute of a ColorEnum object
            :param bool hasSymbol: if True, a dot is drawn for every data point.

            :returns: the (auto-generated) name of the array that i
            :rtype: str
        """
        if arrayName is None:
            if self._overplot:
                arrayName = "{}_{}".format(self.prefix, self._numCurves)
            else:
                arrayName = "{}_{}".format(self.prefix, stringToIdentifier(label))

        self.qlClient.addCurveToPlotPanel(self.pageName, self.panelName, arrayName, label=label,
                                          lineStyle=lineStyle, lineWidth=lineWidth,
                                          color=color, hasSymbol=hasSymbol)

        if isinstance(data, tuple):
            x, y = data
            check_class(x, np.ndarray)
            check_class(y, np.ndarray)

            xCoordsVec = SemanticArray(x, name=arrayName + "_coordinates")
            griddedArray = GriddedArray(y, name=arrayName, label=label, coords={0: xCoordsVec})
        else:
            # The data is already a gridded array, only update the name
            check_class(data, GriddedArray)
            griddedArray = data
            # Replace name with new name, so that it ends up in the correct panel
            griddedArray.data.name = arrayName

        self.qlClient.appendArray(griddedArray)

        if self._overplotReplace:
            self._numCurves += 1

        return arrayName

    def next(self):
        """ Move on to the next curve. """
        assert not self._overplotReplace, \
            "Cannot move to next curve in replace mode, this is done automatically"
        self._numCurves += 1
