""" Quicklook data repository
"""
import enum
import logging
import os

import numpy as np

from gssw.lib.arrays.gridded import GriddedArray, concatenate
from gssw.lib.defs import ActionEnum
from gssw.lib.typing import check_class
from gssw.lib.errors import GsswExportError

from gssw.quicklook.basemodel import BaseItem, BaseModel

logger = logging.getLogger(__name__)


def maxSizeHStack(dataArray, msgArray, maxArraySize):
    """ Concatenates the message array to the data array.

        Expects 1D GriddedArrays. The new result will be have a maximum length of maxArraySize.

        :param GriddedArray None dataArray: 1 dimensional array. If None only the msgArray is
            returned with a maximum length of maxArraySize

        :param GriddedArray None msgArray: 1 dimensional array.
        :param int maxArraySize: the maximum length of the result array
        :returns: 1D concatenated array
        :rtype: GriddedArray

    """
    # TODO: multi-dimensional arrays.
    if dataArray is not None:
        assert dataArray.data.ndim == 1, \
            "Expected 1D dataArray, got: {}D".format(dataArray.data.ndim)
    assert msgArray.data.ndim == 1, "Expected 1D msgArray, got: {}D".format(msgArray.data.ndim)

    if dataArray is None or len(msgArray.data) >= maxArraySize:
        startIdx = max(0, len(msgArray.data) - maxArraySize)
        return msgArray[startIdx:]
    
    # To avoid unnecessary storage first do a subset then concatenate
    startIdx = max(0, len(dataArray.data) + len(msgArray.data) - maxArraySize)
    return concatenate([dataArray[startIdx:], msgArray])


# Please update the corresponding protobuf in case of changes.
@enum.unique
class AppendModeEnum(enum.Enum):
    """ append: new data will be appended (along the concat dimension)
        replace: the old data will be replaced
    """
    append = 0 # the default
    replace = 1



class RepoItem(BaseItem):
    """ A registered array and its concatenating dimension
    """
    def __init__(self, griddedArray, concatDim=None):
        """ Constructor

            :param griddedArray: The underlying gridded array
            :param concatDim: The dimension/axis along which new data is appended.
        """
        super().__init__()
        check_class(griddedArray, GriddedArray)
        check_class(concatDim, int, allow_none=True)

        self._griddedArray = griddedArray
        self._concatDim = concatDim


    @property
    def griddedArray(self):
        """ The underlying GriddedArray object.
        """
        return self._griddedArray


    @griddedArray.setter
    def griddedArray(self, value):
        """ Sets GriddedArray object.
        """
        self._griddedArray = value


    # I think pylint gives some unjustified warnings here so I disabled them.
    # pylint: disable=no-member

    @property
    def name(self):
        """ String representation of the item name
        """
        return self.griddedArray.name


    @property
    def label(self):
        """ String representation of the item label
        """
        return self.griddedArray.label


    @property
    def shapeStr(self):
        """ String representation of the item shape
        """
        return " x ".join(str(elem) for elem in self.griddedArray.data.shape)


    @property
    def elemTypeStr(self):
        """ String representation of the item dtype
        """
        return str(self.griddedArray.data.dtype)


    @property
    def units(self):
        """ String representation of the item unit
        """
        return self.griddedArray.units


    @property
    def concatDim(self):
        """ Number of the dimension (axis) along which the data will be appended.
        """
        return self._concatDim

    # pylint: enable=no-member


class RepositoryModel(BaseModel):
    """ A table model for use in the repository
    """
    # All items in this model will be of this class.
    ITEM_CLASS = RepoItem

    HEADERS    = ["Name", "Label", "Shape", "Type", "Units", "Concat Dim"]
    PROPERTIES = ["name", "label", "shapeStr", "elemTypeStr", "units", "concatDim"]

    COL_NAME, COL_LABEL, COL_SHAPE, COL_ELEM_TYPE, COL_UNITS, COL_CONCAT_DIM = range(len(HEADERS))

    MAX_SAMPLES_BY_DEFAULT = 1000000 # The value for storeMaxSamples if this is not specified.

    def __init__(self, parent=None,
                 storeMaxSamples=MAX_SAMPLES_BY_DEFAULT,
                 unknownArrayAction=ActionEnum.WARNING,
                 appendMode=None):
        """ Constructor
        """
        super().__init__(parent=parent)
        check_class(appendMode, AppendModeEnum, allow_none=True)

        self._appendMode = AppendModeEnum.append if appendMode is None else appendMode

        self._storeMaxSamples = storeMaxSamples

        assert len(self.PROPERTIES) == len(self.HEADERS), "len(PROPERTIES) != len(HEADERS)"

        ActionEnum.checkValueInEnum(unknownArrayAction)
        self.unknownArrayAction = unknownArrayAction


    @property
    def storeMaxSamples(self):
        """ The number of samples that are stored per array.

            This is the maximum size along the concatenation dimension of the array
        """
        return self._storeMaxSamples


    @storeMaxSamples.setter
    def storeMaxSamples(self, value):
        """ The number of samples that are stored per array.

            This is the maximum size along the concatenation dimension of the array
        """
        self._storeMaxSamples = value


    @property
    def appendMode(self):
        """ The current append mode.
            See AppendModeEnum doc string for possible values and their explanation.
        """
        return self._appendMode


    @appendMode.setter
    def appendMode(self, value):
        """ The current plot mode.
            See AppendModeEnum doc string for possible values and their explanation.
        """
        check_class(value, AppendModeEnum)
        self._appendMode = value


    def register(self, griddedArray, concatDim=0, removeExisting=False):
        """ Makes a RepoItem that contains the griddedArray and adds it to the RepoModel

            Emits the sigItemAdded signal.

            Raises DuplicateItemError if an array with this name already exists.

            :param arrayName: any received array having this name will be processed
            :param concatDim: dimension along which newly received data will be concatenated.
            :param removeExisting: If True, any existing array is unregistered . If False (the
                default), a KeyError is raised if the array is already resigtered.
        """
        check_class(griddedArray, GriddedArray)
        logger.debug("Registering array name {}".format(griddedArray.name))

        self.addItem(RepoItem(griddedArray, concatDim=concatDim),
                     removeExisting=removeExisting)


    def unregister(self, arrayName):
        """ Unregisters the Gridded array with name arrayName.

            If the array wasn't registered, warning is logged.
        """
        self.removeItemByName(arrayName, notFoundAction=ActionEnum.WARNING)


    def appendArray(self, griddedArray):
        """ Appends a gridded array to the corresponding dataset.

            Raises KeyError if the array name is not registered.
        """
        check_class(griddedArray, GriddedArray)
        logger.debug("Appending array {} to repo: {}".format(griddedArray.name, self))

        if griddedArray.data.ndim != 1:
            if self.appendMode != AppendModeEnum.replace:
                logger.warning("For {}D the {} mode is invalid (replace used)"
                               .format(griddedArray.data.ndim, AppendModeEnum.replace))

        row = self.rowByName(griddedArray.name, notFoundAction=ActionEnum.IGNORE)
        if row is None:
            # Array not yet present in repo. Register (add) the array.
            if griddedArray.data.ndim == 1:
                griddedArray = maxSizeHStack(None, griddedArray, self.storeMaxSamples)
            self.register(griddedArray)
            row = self.rowByName(griddedArray.name, notFoundAction=ActionEnum.RAISE)

        else:
            # The repo contains this array. Append the data.
            item = self._items[row]

            if griddedArray.data.ndim != 1:
                item.griddedArray = griddedArray
            else:
                if self.appendMode == AppendModeEnum.append:
                    item.griddedArray = maxSizeHStack(item.griddedArray, griddedArray,
                                                      self.storeMaxSamples)

                elif self.appendMode == AppendModeEnum.replace:
                    # Received data (in Replace-mode) can be too large, so resize when required.
                    item.griddedArray = maxSizeHStack(None, griddedArray,
                                                      self.storeMaxSamples)
                else:
                    raise ValueError("Unexpected append mode: {}".format(self.appendMode))


        self.emitDataChangedForRow(row) # Triggers update of the table views.
        self.emitSigItemChanged(griddedArray.name) # Triggers updates of the plots
        logger.debug("appendArray done\n")


    def exportToCsv(self, fileName):
        """ Exports repo contents to a Comma Separated Values text file.
        """
        logger.debug("exportData: {}".format(self))

        # Gather data. The data will be exported in x1, y1, x2, y2, ..., xN, yN format
        # See Redmine issue #3065.
        headerElems = ['Columns:']
        dataElems = []
        colNr = 1
        for _, repoItem in enumerate(self.items, 1):
            gArr = repoItem.griddedArray
            assert gArr.data.ndim == 1, "1D array expected. Got: {}D".format(gArr.data.ndim)

            coords = repoItem.griddedArray.coords.get(0, None)
            if coords is not None:
                if coords.dtype in (np.datetime64, np.dtype('datetime64[ms]')) \
                or coords.dtype.type == np.datetime64:
                    # TODO: Handle coords with timestamps, for now point to workaround
                    raise GsswExportError(
                        "Use 'Export CSV per Curve' for saving plots with timestamps.")
                coordHeader = coords.name if coords.name else 'X-coordinates of {}'.format(
                    gArr.labelOrName)
                headerElems.append("column-{}: {}".format(colNr, coordHeader))
                dataElems.append(coords)
                colNr += 1

            headerElems.append("column-{}: {}".format(colNr, gArr.labelOrName))
            dataElems.append(gArr.data)
            colNr += 1

        logger.info("Exporting data to CSV file: {}".format(os.path.realpath(fileName)))
        try:
            allData = np.vstack(dataElems).T
            np.savetxt(fileName, allData, delimiter=', ', header="\n".join(headerElems))
        except Exception as ex:
            logger.error("Unable to export data: {}".format(ex))
            raise GsswExportError("Unable to export data: {}".format(ex) +
                "\n Try 'Export CSV per Curve' to export to separate files.") from None



    def exportToCsvSeparate(self, fileName):
        """ Exports repo contents to separate Comma Separated Values text files (one per item)
        """
        logger.debug("exportData: {}".format(self))

        for itenNr, repoItem in enumerate(self.items):
            itemFileName = "{}.{:02d}.{}".format(fileName, itenNr, repoItem.griddedArray.name)
            repoItem.griddedArray.exportToCsv(itemFileName)

