
""" Data that is shared between the panels
"""
import collections
import copy
import logging
import os.path
import yaml

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt, pyqtSlot
from PyQt5.QtWidgets import QMessageBox

from gssw.lib.errors import ItemNotFoundError
from gssw.lib.typing import check_class
from gssw.lib.widgets.colors import widgetHasDarkBackGround
from gssw.lib.widgets.toggle_column_mixin import ToggleColumnTableView
from gssw.quicklook.defs import MIME_ARRAYS_INFO
from gssw.quicklook.basemodel import BaseItem, BaseModel

logger = logging.getLogger(__name__)


##########
# Static #
##########

class ArrayInfoItem(BaseItem):
    """ An item that holds information of a row in the live array table

        Contains array names plus extra info per array such as label, suggested X and Y axis labels,
        number of dimensions.
    """
    def __init__(self, name, label='', xLabel='', yLabel='', appendModeHint=False):
        """ Constructor

            :param repoItem: the RepoItem where this PanelItem links to
            :param colorName: string containging the name of the color (can be #RRGGBB ?)
        """
        super().__init__()

        check_class(name, str)
        self._name = name
        self._label = '' if not label else label
        self._xLabel = '' if not xLabel else xLabel
        self._yLabel = '' if not yLabel else yLabel
        self._numMessages = 0
        self._numPanels = 0
        self._appendModeHint = appendModeHint

    @property
    def appendModeHint(self):
        """ bool, indicates that it makes sense to display this array using append mode. """
        return self._appendModeHint

    @appendModeHint.setter
    def appendModeHint(self, value):
        check_class(value, bool)
        self._appendModeHint = value


    @property
    def name(self):
        """ Unique name of the item.
        """
        return self._name


    @name.setter
    def name(self, value):
        """ Unique name of the item.
        """
        check_class(value, str)
        self._name = value


    @property
    def label(self):
        """ String describing the item. E.g. for use in legends
        """
        return self._label


    @label.setter
    def label(self, value):
        """ String describing the item. E.g. for use in legends
        """
        check_class(value, str, allow_none=True) # Can be None in case incoming array has no label
        self._label = value


    @property
    def xLabel(self):
        """ String describing the suggested X-label. E.g. for use in plots
        """
        return self._xLabel


    @xLabel.setter
    def xLabel(self, value):
        """ String describing the suggested X-label. E.g. for use in plots
        """
        check_class(value, str)
        self._xLabel = value


    @property
    def yLabel(self):
        """ String describing the suggested Y-label. E.g. for use in plots
        """
        return self._yLabel


    @yLabel.setter
    def yLabel(self, value):
        """ String describing the suggested Y-label. E.g. for use in plots
        """
        check_class(value, str)
        self._yLabel = value


    @property
    def labelOrName(self):
        """ The label if defined, otherwise the name (used for display purposes)
        """
        return self._label if self._label else self._name


    @property
    def numMessages(self):
        """ Number of received message with this array name.
        """
        return self._numMessages


    @numMessages.setter
    def numMessages(self, value):
        """ Number of received message with this array name.
        """
        check_class(value, int)
        self._numMessages = value


    @property
    def numPanels(self):
        """ Number panels that this panel occues in.
        """
        return self._numPanels


    @numPanels.setter
    def numPanels(self, value):
        """ Number panels that this panel occues in.
        """
        check_class(value, int)
        self._numPanels = value


    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        .
            The number of message and panels are not serialized.
        """
        dct = {} # Configuration subdict

        dct['name'] = self.name
        dct['label'] = self.label
        dct['xLabel'] = self.xLabel
        dct['yLabel'] = self.yLabel
        dct['appendModeHint'] = self.appendModeHint

        return dct


    def setValuesFromDict(self, dct):
        """ Creates a panel from a dictionary with values

            The number of message and panels are not serialized.
        """
        # Name should certainly be present in the dict. For the others we can use defaults if they
        # somehow are not present in the dictionary.
        self.name  = dct['name']
        self.label = dct.get('label', '')
        self.xLabel = dct.get('xLabel', '')
        self.yLabel = dct.get('yLabel', '')
        self.appendModeHint = dct.get('self.appendModeHint', False)



class ArrayInfoModel(BaseModel):
    """ Table model with an underlying ArrayInfoItem per row.
    """

    # All items in this model will be of this class
    ITEM_CLASS = ArrayInfoItem

    # Extent the possible columns in the plot pannel model
    HEADERS    = ["Name", "Label", "X-Label", "Y-Label", "Panels", "Messages", "Append data"]
    PROPERTIES = ["name", "label", "xLabel", "yLabel", "numPanels", "numMessages", "appendModeHint"]
    COL_NAME, COL_LABEL, COL_X_LABEL, COL_Y_LABEL, COL_NUM_PANEL, COL_NUM_MSG, COL_NUM_APPEND_HINT = range(len(HEADERS))  #pylint: disable=line-too-long


    def __init__(self, parent=None):
        """ Constructor
        """
        super().__init__(parent=parent)

        self.textAlign = Qt.AlignLeft | Qt.AlignVCenter
        self.numAlign = Qt.AlignRight | Qt.AlignVCenter
        self.newDataColor = QtGui.QColor("red")
        self.noDataColor = QtGui.QColor("gray")
        self.plottedDataColor = QtGui.QColor("black")


    def flags(self, index):
        """ Returns the item flags for the given index

            Items can be dragged.
        """
        flags = int(super().flags(index)) | Qt.ItemIsDragEnabled

        if not index.isValid():
            return None

        col = index.column()
        if col in (ArrayInfoModel.COL_NUM_PANEL, ArrayInfoModel.COL_NUM_MSG):
            return flags
        else:
            return flags | Qt.ItemIsEditable


    def data(self, index, role=Qt.DisplayRole):
        """ Returns the data stored under the given role for the item referred to by the index.
        """
        if not index.isValid():
            return None

        row, col = index.row(), index.column()

        if col < 0 or col >= self.columnCount():
            return None

        if row < 0 or row >= self.rowCount():
            return None

        if role == Qt.TextColorRole:

            item = self.itemByRow(row)
            if item.numMessages > 0:
                if item.numPanels == 0:
                    return self.newDataColor
                else:
                    return self.plottedDataColor
            else:
                return self.noDataColor

        elif role == Qt.TextAlignmentRole:
            if col == self.COL_NUM_PANEL or col == self.COL_NUM_MSG:
                return self.numAlign
            else:
                return self.textAlign
        else:
            return super().data(index, role)



    def repopulateFromModel(self, other):
        """ Clears and repopulates the model with data from another static model

        :param ArrayInfoModel other: the source model
        """
        check_class(other, ArrayInfoModel)
        self.setValuesFromDictList(other.asDictList())


    def mimeData(self, indexes):
        """ Returns a QMime type with binary encoded array names of the selected cells.

            See http://doc.qt.io/qt-5/qtwidgets-itemviews-puzzle-example.html
            Not to be confused with
                http://doc.qt.io/qt-5/qtwidgets-draganddrop-puzzle-example.html
            which reimplements startDrag on the view.

            :param indexes: list of indices of cells that the user is dragging
            :return: binary encoded yaml string represntation of the selected items.
            :rtype: QMimeData
        """
        rows = [index.row() for index in indexes if index.isValid and index.column() == 0]

        items = [self.itemByRow(row) for row in rows]
        return self.encodeArraysInfoMimeData(items)


    @classmethod
    def encodeArraysInfoMimeData(cls, arrayInfoItems):
        """ Returns a QMime type with binary encoded yaml represetation of the array info items
        """
        itemDicts = []
        for item in arrayInfoItems:
            check_class(item, cls.ITEM_CLASS)
            itemDicts.append(item.asDict())

        mimeStr = yaml.safe_dump(itemDicts)
        mimeBytes = bytes(mimeStr, encoding='ascii')

        mimeData = QtCore.QMimeData()
        mimeData.setData(MIME_ARRAYS_INFO, mimeBytes)
        return mimeData


    @classmethod
    def decodeArraysInfoMimeData(cls, mimeBytes):
        """ Constructs a list of ArrayInfoItems from the QMime type with binary encoded array names.
        """
        mimeStr = str(mimeBytes, encoding='ascii')
        itemDicts = yaml.safe_load(mimeStr)

        items = []
        for dct in itemDicts:
            logger.debug("Creating object of type {} from dict: {}".format(cls.ITEM_CLASS, dct))
            items.append(cls.ITEM_CLASS(**dct))

        return items


    def updateArrayStatistics(self, griddedArray):
        """ If the array name does not yet occur it is added to the model.

            Increases the message count by one.
        """
        arrayName = griddedArray.name
        try:
            row = self.rowByName(arrayName)
        except ItemNotFoundError:
            item = ArrayInfoItem(name=griddedArray.name)
            self.addItem(item)
            row = None
        else:
            item = self.itemByRow(row)

        item.numMessages += 1
        item.label = griddedArray.label

        # When drag-dropping arrays, auto-select append mode (instead of replace) when
        # single samples are being received, rather than blocks of samples
        appendModeHint = len(griddedArray.data) == 1
        item.appendModeHint = appendModeHint

        # Add labels of coordinate arrays (if present) to the item xLabel and yLabel
        xCoords = griddedArray.coords.get(0, None)
        if xCoords is not None:
            item.xLabel = xCoords.axisLabel

        yCoords = griddedArray.coords.get(1, None)
        if yCoords is not None:
            # Set the Y axis label from the coordinate vector
            item.yLabel = yCoords.axisLabel
        else:
            # Try setting Y axis label from array itself (i.o. the coordinate vector)
            if griddedArray.data is not None:
                item.yLabel = griddedArray.data.axisLabel

        if row is None:
            # New item, (what to do here? Sort the model by name?)
            pass
        else:
            # Existing item, signal that this row is updated.
            self.emitSigItemChanged(item.name) # use custom signal only, otherwise GUI selections
            #                                  # using the same data model are invalidated


    def updatePlotStatistics(self, counts):
        """ Updates the numPlots column with the value from the plotCount dictionary.

            :param collections.Counter counts: arrayName -> plotCount dict
        """
        # Add arrays that occur in plot panels but do not yet occur in the live model
        check_class(counts, collections.Counter)
        panelArrayNames = set(list(counts.keys()))
        modelArrayNames = set([item.name for item in self.items])
        for name in panelArrayNames - modelArrayNames:
            item = ArrayInfoItem(name=name)
            self.addItem(item)

        # Update the panel count for all arrays in the model
        for row, item in enumerate(self.items):
            oldCount = item.numPanels
            item.numPanels = counts.get(item.name, 0)
            if item.numPanels != oldCount:
                self.emitDataChangedForRow(row)


class ConfigArrayInfoWidget(QtWidgets.QWidget):
    """ Widget to edit the static array info model.

        Behaves like a (non-modal) dialog but does not derive from QDialog because that class
        sets behaviour for the Return ane Escape keys. In this widget the Escape key cancels,
        editing but in a QDialog it would close the dialog window. The Return key would close the
        dialog. Therefore we derive from QWidget, Unless you want a modal dialog, inheriting from
        QDialog would add nothing. If you want a modal dialog you probably need to install
        Event filters.
    """
    # Many statements are inherit to Qt constructors.
    # pylint: disable=too-many-statements

    def __init__(self, staticModel=None, parent=None):
        """
            :param parent: Qt parent object
            :model: a RepositoryModel
        """
        super().__init__(parent=parent)

        # The window title will show an asterisk if the contents has not been saved.
        self.setWindowTitle("Configure arrays [*]")

        # The table model that is updated after 'save' or 'ok'
        check_class(staticModel, ArrayInfoModel)
        self._actualModel = staticModel

        # The table that is updated. After 'save' or Ok this will be copied to the _actualModel
        self._cacheModel = ArrayInfoModel()
        self._cacheModel.repopulateFromModel(self._actualModel)

        self._cacheModel.dataChanged.connect(self.onDataChanged)

        # Define actions

        # Do not use Qt.Key_Return as the edit action short cut. It will interfere when returning
        # from the editing mode and give  a 'edit: editing failed' on stderr.
        # See https://stackoverflow.com/questions/12146364/
        self.editNameAction = QtWidgets.QAction("Edit", self)
        self.editNameAction.setToolTip("Edit selected name.")
        self.editNameAction.setShortcut(QtGui.QKeySequence(Qt.Key_Space))
        self.editNameAction.triggered.connect(self.editCurrentItem)

        self.addNewNameAction = QtWidgets.QAction("Add", self)
        self.addNewNameAction.setToolTip("Add new name.")
        self.addNewNameAction.setShortcut(QtGui.QKeySequence(Qt.Key_Insert))
        self.addNewNameAction.triggered.connect(self.addNewItem)

        self.removeNameAction = QtWidgets.QAction("Remove", self)
        self.removeNameAction.setToolTip("Remove selected array.")
        self.removeNameAction.setShortcut(QtGui.QKeySequence(Qt.Key_Delete))
        self.removeNameAction.triggered.connect(self.removeCurrentItem)

        self.saveChangesAction = QtWidgets.QAction("Save", self)
        self.saveChangesAction.setToolTip("Save changes.")
        self.saveChangesAction.setShortcut(QtGui.QKeySequence("CTRL+S"))
        self.saveChangesAction.triggered.connect(self.saveChanges)
        self.addAction(self.saveChangesAction) # Needed because it's not added elsewhere

        self.revertChangesAction = QtWidgets.QAction("Revert", self)
        self.revertChangesAction.setToolTip("Revert changes.")
        self.revertChangesAction.setShortcut(QtGui.QKeySequence("CTRL+R"))
        self.revertChangesAction.triggered.connect(self.revertChanges)
        self.addAction(self.revertChangesAction) # Needed because it's not added elsewhere

        self.closeAction = QtWidgets.QAction("Close", self)
        self.closeAction.setToolTip("Close dialog.")
        self.closeAction.setShortcut(QtGui.QKeySequence("CTRL+W"))
        self.closeAction.triggered.connect(self.closeWindow)
        self.addAction(self.closeAction) # Needed because it's not added elsewhere

        # Set icon depending on color.
        app = QtWidgets.QApplication.instance()
        if widgetHasDarkBackGround(self):
            iconDir = os.path.join(app.iconPath, 'snipicons', 'white')
            editIconName = os.path.join(iconDir, 'pencil-i.svg')
            addIconName = os.path.join(iconDir, 'plus-sign-i.svg')
            removeIconName = os.path.join(iconDir, 'minus-sign-i.svg')
        else:
            iconDir = os.path.join(app.iconPath, 'snipicons', 'black')
            editIconName = os.path.join(iconDir, 'pencil-m.svg')
            addIconName = os.path.join(iconDir, 'plus-sign-m.svg')
            removeIconName = os.path.join(iconDir, 'minus-sign-m.svg')

        assert os.path.exists(editIconName), "File not found: {}".format(editIconName)
        assert os.path.exists(addIconName), "File not found: {}".format(addIconName)
        assert os.path.exists(removeIconName), "File not found: {}".format(removeIconName)

        self.editNameAction.setIcon(QtGui.QIcon(editIconName))
        self.addNewNameAction.setIcon(QtGui.QIcon(addIconName))
        self.removeNameAction.setIcon(QtGui.QIcon(removeIconName))

        # Define widgets
        self.centralLayout = QtWidgets.QVBoxLayout(self)

        self.helpLabel = QtWidgets.QLabel(
            "This table contains data array names that are 'known' to the Quicklooker. "
            "Those arrays are always included in the array table and the drop-down boxes that "
            "select arrays by name.")
        self.helpLabel.setWordWrap(True)
        self.centralLayout.addWidget(self.helpLabel)

        hline = QtWidgets.QFrame()
        hline.setFrameShape(QtWidgets.QFrame.HLine)
        hline.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.centralLayout.addWidget(hline)

        self.tableLayout = QtWidgets.QHBoxLayout()
        self.centralLayout.addLayout(self.tableLayout)

        self.tableView = QtWidgets.QTableView()
        self.tableView.setModel(self._cacheModel)
        self.tableView.setAlternatingRowColors(True)

        self.tableView.setEditTriggers(QtWidgets.QAbstractItemView.DoubleClicked |
                                       QtWidgets.QAbstractItemView.EditKeyPressed) # F2 in Linux

        header = self.tableView.horizontalHeader()

        for col in range(header.count()):
            header.setSectionHidden(col,
                col in (ArrayInfoModel.COL_NUM_MSG, ArrayInfoModel.COL_NUM_PANEL))

        self.tableLayout.addWidget(self.tableView)

        # Buttons for adding/removing list items.
        self.listButtonsLayout = QtWidgets.QVBoxLayout()
        self.tableLayout.addLayout(self.listButtonsLayout)

        self.itemAddNewButton = QtWidgets.QToolButton()
        self.itemAddNewButton.setDefaultAction(self.addNewNameAction)
        self.listButtonsLayout.addWidget(self.itemAddNewButton)

        self.itemRemoveButton = QtWidgets.QToolButton()
        self.itemRemoveButton.setDefaultAction(self.removeNameAction)
        self.listButtonsLayout.addWidget(self.itemRemoveButton)

        self.itemEditButton = QtWidgets.QToolButton()
        self.itemEditButton.setDefaultAction(self.editNameAction)
        self.listButtonsLayout.addWidget(self.itemEditButton)
        self.listButtonsLayout.addStretch()

        # Buttons for saving/discarding/closing
        self.bottomButtonsLayout = QtWidgets.QHBoxLayout()
        self.centralLayout.addLayout(self.bottomButtonsLayout)

        self.saveButton = QtWidgets.QPushButton("Save")
        self.saveButton.clicked.connect(self.saveChangesAction.trigger)
        self.saveButton.setToolTip("Save changes.")
        self.bottomButtonsLayout.addWidget(self.saveButton)

        self.revertButton = QtWidgets.QPushButton("Revert")
        self.revertButton.clicked.connect(self.revertChangesAction.trigger)
        self.revertButton.setToolTip("Revert changes.")
        self.bottomButtonsLayout.addWidget(self.revertButton)

        self.bottomButtonsLayout.addStretch()

        self.closeButton = QtWidgets.QPushButton("Close")
        self.closeButton.clicked.connect(self.closeAction.trigger)
        self.closeButton.setToolTip("Close dialog.")
        self.bottomButtonsLayout.addWidget(self.closeButton)

        self.onDataChanged(cacheDirty=self.isWindowModified())

        #self.resize(QtCore.QSize(400, 400))

    def sizeHint(self):
        """ The prefered size of this window"""
        return QtCore.QSize(500, 600)


    @pyqtSlot()
    def editCurrentItem(self):
        """ Will let the user edit the current item.
        """
        currentIndex = self.tableView.selectionModel().currentIndex()

        if not currentIndex.isValid():
            logger.warning("No item selected (ignored)")
            return

        logger.debug("editing: row {}, col {}".format(currentIndex.row(), currentIndex.column()))
        self.tableView.edit(currentIndex)


    @pyqtSlot()
    def addNewItem(self, name="New Item"):
        """ Adds a new item at the current position
        """
        currentIndex = self.tableView.selectionModel().currentIndex()

        # Append to last row if no element is selected
        row = currentIndex.row() if currentIndex.isValid() else None
        #itemIndex = self._cacheNameListModel.insertName(name, row=row)
        newItem = ArrayInfoItem(name=name)
        itemIndex = self._cacheModel.addItem(newItem, row=row)

        self.tableView.setCurrentIndex(itemIndex)
        self.editCurrentItem()
        self.onDataChanged(cacheDirty=True)

        return itemIndex


    @pyqtSlot()
    def removeCurrentItem(self):
        """ Removes the item that is currently selected in the listView
        """
        currentIndex = self.tableView.selectionModel().currentIndex()

        if not currentIndex.isValid():
            logger.warning("No item selected (ignored)")
            return

        item = self._cacheModel.itemByRow(currentIndex.row())

        logger.debug("Removing item with name: {}".format(item.name))
        self._cacheModel.removeItemByName(item.name)
        self.onDataChanged(cacheDirty=True)


    @pyqtSlot()
    def onDataChanged(self, cacheDirty=True):
        """ Updates the window when data has changed.

            Will enable/disabled the Revert and Save button.

            :param cacheDirty: Set to True if there there are unsaved changes. Set to False after
                               the data has been saved.
        """
        logger.debug("onDataChanged")
        self.setWindowModified(cacheDirty)

        self.saveChangesAction.setEnabled(cacheDirty)
        self.revertChangesAction.setEnabled(cacheDirty)

        self.saveButton.setEnabled(cacheDirty)
        self.revertButton.setEnabled(cacheDirty)


    @pyqtSlot()
    def saveChanges(self):
        """ Will save changes.
        """
        logger.debug("Saving changes")
        self._actualModel.repopulateFromModel(self._cacheModel)
        self.onDataChanged(cacheDirty=False)


    @pyqtSlot()
    def revertChanges(self):
        """ Will discard the edits made by the user since the last save.
        """
        logger.debug("Reverting changes")
        self._cacheModel.repopulateFromModel(self._actualModel)
        self.onDataChanged(cacheDirty=False)


    @pyqtSlot()
    def closeWindow(self):
        """ Closes this window
        """
        if self.isWindowModified():
            answer = QMessageBox.question(self, # parent
                "Close window",
                "There are unsaved changes. Do you want to save them?",
                 QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel,
                 QMessageBox.Cancel)

            if answer == QMessageBox.Cancel:
                return
            elif answer == QMessageBox.Yes:
                self.saveChanges()
            elif answer == QMessageBox.No:
                self.revertChanges()
            else:
                raise ValueError("Unexpected answer button clicked: {}".format(answer))

        logger.debug("Closing window")
        self.close()


###########
# Dynamic #
###########




class LiveArrayInfoWidget(QtWidgets.QWidget):
    """ Widget that contains a table with dynamically updated array info.

        Consists of a table with array info plus perhaps some extra buttons (e.g. config, find)
    """
    def __init__(self, model, **kwargs):
        """ Constructor
        """
        super().__init__(**kwargs)

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.setLayout(self.mainLayout)

        self.arrayTableView = ToggleColumnTableView()
        self.arrayTableView.setModel(model)
        self.arrayTableView.setAlternatingRowColors(True)
        self.arrayTableView.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.arrayTableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.arrayTableView.setDragEnabled(True)
        self.arrayTableView.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.arrayTableView.horizontalHeader().setDefaultSectionSize(120)
        self.arrayTableView.verticalHeader().setVisible(False) # Hide meaningless row numbers

        # Add context menu for the table header to show/hide the columns,
        checkable = {}
        checked = {}
        for name in ArrayInfoModel.HEADERS:
            checkable[name] = (name != 'Name') # Cannot hide Name column
            checked[name] = (name in ['Name', 'Label']) # visable by default

        self.arrayTableView.add_header_context_menu(checkable=checkable, checked=checked)

        self.helpLabel = QtWidgets.QLabel("Drag to page to create a line plot.")

        self.mainLayout.addWidget(self.arrayTableView)
        self.mainLayout.addWidget(self.helpLabel)





############
# Combined #
############


class GlobalArrayInfoRepo(QtCore.QObject):
    """ Maintains a table of array info (name, and metadata).

        There are actually two tables, a static and a dynamic list. The static table is typically
        configured up front, depending on the project. However the user can edit the static table
        from within the GUI with a special dialog.

        The dynamic table is the actual list. It is the union of
            1) the static table
            2) all array names that are present in one of the plots
            3) all arrays that have been received since start up.

        The table are always alphabetical by name.
    """
    __singleton = None

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._staticModel = ArrayInfoModel(parent=self)
        self._dynamicModel = ArrayInfoModel(parent=self)

        # The modelReset signal will be emitted when setStringList is called.
        self._staticModel.modelReset.connect(self._mergeStaticIntoDynamic)


    @staticmethod
    def instance():
        """ Returns the instance singleton, which is shared by the complete application
        """
        if GlobalArrayInfoRepo.__singleton is None:
            parent = QtGui.QGuiApplication.instance()
            GlobalArrayInfoRepo.__singleton = GlobalArrayInfoRepo(parent=parent)
        return GlobalArrayInfoRepo.__singleton


    @property
    def dynamicModel(self):
        """ The dynamic model.

            It contains all array names that are present in one of the plots, plus the names of all
            arrays that have been received during since start up.
        """
        return self._dynamicModel


    @property
    def staticModel(self):
        """ The static model

            Contains all array names entered by the user via the GUI (File menu, Edit Array Names)
        """
        return self._staticModel


    def getStaticData(self):
        """ Gets the static array info as a list of dictionaries.
        """
        return self._staticModel.asDictList()


    def setStaticData(self, dictList):
        """ Sets the static array info from a list of dictionaries.
        """
        self._staticModel.setValuesFromDictList(dictList)


    def getDynamicNames(self):
        """ Gets the dynamic array info as a list of dictionaries.
        """
        return self._dynamicModel.asDictList()


    def setDynamicData(self, dictList):
        """ Sets the dynamic array info from a list of dictionaries.
        """
        self._dynamicModel.setValuesFromDictList(dictList)


    def updateArrayStatistics(self, griddedArray):
        """ Updates dynamic array info.
        """
        self._dynamicModel.updateArrayStatistics(griddedArray)


    def updatePlotStatistics(self, counts):
        """ Updates the numPlots column with the value from the plotCount dictionary
        """
        self._dynamicModel.updatePlotStatistics(counts)


    def _mergeStaticIntoDynamic(self):
        """ Copies all names from the static model into the dynamic model
        """
        logger.debug("_mergeStaticIntoDynamic", stack_info=False)
        #dynamicNames = [item.name for item in self._dynamicModel.items]

        # We just reset/invalidate the entire model. Perhaps a more fine grained approach is needed
        # later with emitDataChangedForRow.
        self._dynamicModel.beginResetModel()

        for staticItem in self._staticModel.items:
            logger.debug("Merging: {}".format(staticItem))
            try:
                dynamicItem = self._dynamicModel.itemByName(staticItem.name)
            except ItemNotFoundError:
                # Add item to the dynamic model.
                check_class(staticItem, ArrayInfoItem)
                newDynItem = copy.deepcopy(staticItem)
                logger.debug("Adding item: {}".format(newDynItem))
                self._dynamicModel.addItem(newDynItem)
            else:
                dynamicItem.setValuesFromDict(staticItem.asDict())

        self._dynamicModel.endResetModel()




