""" Table panels for the Quicklook application
"""
import logging

from PyQt5 import QtGui, QtWidgets
from PyQt5.QtCore import pyqtSlot

from gssw.quicklook.baseconfig import AbstractPanelConfigWidget, BaseConfigItem,  BaseConfigModel
from gssw.quicklook.basepanel import AbstractDataPanel
from gssw.quicklook.repo import RepositoryModel

logger = logging.getLogger(__name__)



class TablePanel(AbstractDataPanel):
    """ A panel containing a central plot.

        Has a PlotCurvesModel to store al the data and a link to a RepositoryModel so that it can
        take data from the repository and put it into the plot model.

    """
    def __init__(self, name, repo=None, parent=None):
        """ Constructor

            :param name: panel name string
            :param repo: a RepositoryModel. If None, an emtpy one will be created.
            :param parent: Qt parent object.
        """
        super().__init__(name, repo=repo, parent=parent)

        self._arrayNames = set([])  # names of the arrays this panel visualizes

        self._tableModel = QtGui.QStandardItemModel(self)
        self._tableModel.setHorizontalHeaderLabels(["Name", "Last Value"])
        self.tableView.setModel(self._tableModel)


    def _addSpecializedWidgets(self):
        """ Adds specialized (i.e. the non-common) widgets to the config box.

            The base implementation does nothing. Should be implemented by the descendant.
        """
        self.tableView = QtWidgets.QTableView()
        self.tableView.setColumnWidth(0, 150)
        self.tableView.setColumnWidth(1, 150)
        self.tableView.setObjectName('pg-table')

        self.contentsLayout.addWidget(self.tableView)


    def _createConfigWidget(self, parent=None):
        """ Creates a new TablePanelConfigWidget

            :param parent:
            :return: new TablePanelConfigWidget
        """
        return TablePanelConfigWidget(parent=parent)


    def _tableRowByName(self, arrayName):
        """ Returns the row that contains the array name (in the first column)

            Raises an KeyError if the name isn't present or occurs more than once.
            :param arrayName: array name
            :return: int
        """
        items = self._tableModel.findItems(arrayName, column=0)
        if len(items) == 1:
            return items[0].row()
        else:
            raise KeyError("Expected 1 item called: {}, got {}".format(arrayName, len(items)))


    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        dct = super().asDict()
        cfg = dct['cfg'] = {}

        cfg['maxSamples'] = self.storeMaxSamples

        cfgItems = []
        for row in range(self._tableModel.rowCount()):
            cfgItem = dict(
                name = self._tableModel.item(row, 0).text(),
                label = '')
            cfgItems.append(cfgItem)

        cfg['cfgItems'] = cfgItems

        return dct


    def setValuesFromDict(self, dct):
        """ Sets panel parameters given a dictionary with (config options)
        """
        super().setValuesFromDict(dct)
        logger.debug("setValuesFromDict: {}".format(dct))

        cfg = dct.get('cfg')
        if not cfg:
            logger.warning("No cfg field in dict. Ignored.")
            return

        self.storeMaxSamples = cfg.get('maxSamples', RepositoryModel.MAX_SAMPLES_BY_DEFAULT)

        # Remove old items
        self._tableModel.removeRows(0, self._tableModel.rowCount())

        for cfgItem in cfg['cfgItems']:
            tableItems = [QtGui.QStandardItem(cfgItem['name']), QtGui.QStandardItem()]
            self._tableModel.appendRow(tableItems)

        # Store the names of the arrays that this panel knows of
        self._arrayNames = set([cfgItem['name'] for cfgItem in cfg['cfgItems']])


    def arrayInConfig(self, arrayName):
        """ Returns True if this panel is configured to show data of this array.

            :param arrayName: gridded array name
            :return: True if the array name in the list of configured arrays
        """
        return arrayName in self._arrayNames


    def getConfigArrayNames(self):
        """ Returns a list with the names of the array used in the configuration"""
        return self._arrayNames


    def _updateFromRepoItem(self, repoItem):
        """ Called when a repo item has changed.

            The base implementation does nothing. Descendants should override this method.
        """
        values = repoItem.griddedArray.data
        if values.ndim == 1:
            lastValueStr = repr(values[-1])
        else:
            lastValueStr = "Not a 1D array"

        row = self._tableRowByName(repoItem.name)
        lastValueTableItem = self._tableModel.item(row, 1) # last value column
        lastValueTableItem.setText(lastValueStr)


    @pyqtSlot()
    def onRepoAboutToBeCleared(self):
        """ Is called when the model is about to be cleared.
            Sets the current image to None
        """
        for row in range(self._tableModel.rowCount()):
            lastValueTableItem = self._tableModel.item(row, 1) # last value column
            lastValueTableItem.setText('')



class TableConfigItem(BaseConfigItem):
    """ A config item
    """



class TableEntryModel(BaseConfigModel):
    """ A table model for storing the table entry items.
    """
    ITEM_CLASS = TableConfigItem




class TablePanelConfigWidget(AbstractPanelConfigWidget):
    """ Contains widgets for configuring PlotPanel settings

        :param repo: Refers to the repository. Needed to execute the SelectRepoItemDialog
    """

    def __init__(self, parent=None):
        """ Constructor
        """
        super().__init__(configModel=TableEntryModel(), parent=parent)


    #@pyqtSlot(CurveItem)
    def _createConfigItem(self, name):
        """ Creates a new config item and adds it to the model.
        """
        configItem = TableConfigItem(name)
        return configItem


    def _addCommonWidgets(self):
        """ Adds specialized (i.e. the non-common) widgets to the config box.

            The base implementation does nothing. Should be implemented by the descendant.
        """
        self.maxSamplesSpinBox = QtWidgets.QSpinBox()
        self.maxSamplesSpinBox.setValue(RepositoryModel.MAX_SAMPLES_BY_DEFAULT)
        self.maxSamplesSpinBox.setMinimum(0)
        self.maxSamplesSpinBox.setMaximum(100000) # a plot with more samples seems unworkable
        self.maxSamplesSpinBox.setSingleStep(10)
        self.commonLayout.addRow("Max samples", self.maxSamplesSpinBox)


    def _updateSpecialWidgetsForIndex(self, currentIndex):
        """ Updates the specialized widgets.
        """


    def asDict(self):
        """ Returns a dictionary representation for serialization purposes
        """
        self.mapper.submit()
        cfg = {}

        cfg['maxSamples'] = self.maxSamplesSpinBox.value()
        cfg['cfgItems'] = [item.asDict() for item in self._configModel.items]

        return cfg


    def setValuesFromDict(self, cfg):
        """ Creates a panel from a dictionary with values
        """
        self.maxSamplesSpinBox.setValue(cfg.get('maxSamples',
                                                RepositoryModel.MAX_SAMPLES_BY_DEFAULT))

        for itemDct in cfg.get('cfgItems', []):
            cfgItem = TableConfigItem(itemDct['name'])
            cfgItem.setValuesFromDict(itemDct)
            self._configModel.addItem(cfgItem)

        self.mapper.revert()
