#!/usr/bin/env python3
# pylint: disable=C0302
# (too-many-lines)
# pylint: disable=R0902
# (too-many-instance-attributes)
# pylint: disable=R0911
# (too-many-return-statements)
# pylint: disable=R0914
# (too-many-locals)
# pylint: disable=R0915
# (too-many-statements)
# pylint: disable=E0611
# (No name 'QKeySequence' in module 'PyQt5.QtGui')
# pylint: disable=R0911
# (too-many-return-statements)

""" GSSw SetGet Manager """

###################################################################################################
# P Y T H O N   M O D U L E S
###################################################################################################

import sys
import logging
import os.path
from datetime import datetime

import yaml
import numpy as np

from PyQt5.QtCore             import Qt, QSettings, QEvent, QSize, QPoint
from PyQt5.QtGui              import QGuiApplication, QKeySequence
from PyQt5.QtWidgets          import (QWidget, QTableWidgetItem, QLineEdit, QComboBox, QCheckBox,
                                      QPushButton, QMenu, QAction, QMessageBox, QFileDialog,
                                      QStyle, QDialogButtonBox, QCompleter, QStyledItemDelegate)
from PyQt5                    import uic

from gssw.hal.hal             import Hal, Register, Conversion
from gssw.hal.halclient       import HalClient
from gssw.lib.arrays.gridded  import GriddedArray, SemanticArray
from gssw.lib.config          import Config
from gssw.lib.device          import Device
from gssw.lib.sockets         import PubSocket, tagName
from gssw.lib.typing          import check_class
from gssw.lib.version         import VERSION_STRING
from gssw.lib.widgets.misc    import (promptForFileName, gsswUiWarning, gsswUiError, gsswUiInfo,
                                      gsswUiMessageBox_)
from gssw.lib.errors          import DataError, GsswError

logger = logging.getLogger(__name__)


###################################################################################################
# H E L P E R  F U N C T I O N S
###################################################################################################

def access2String(access):
    """ Translate the access mode (integer) to a string. """
    if access == Register.READ_ONLY :
        return "ro"
    if access == Register.READ_WRITE:
        return "rw"
    if access == Register.WRITE_ONLY:
        return "wo"
    raise AssertionError('Illegal access type: {}'.format(access))

def setReadOnly(item, switch = True):
    """ Wrapper """
    if isinstance(item, QTableWidgetItem):
        if switch:
            item.setFlags(item.flags() | Qt.ItemIsEnabled)
            item.setFlags(item.flags() | Qt.ItemIsSelectable)
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
        else:
            item.setFlags(item.flags() | Qt.ItemIsEnabled)
            item.setFlags(item.flags() | Qt.ItemIsSelectable)
            item.setFlags(item.flags() | Qt.ItemIsEditable)
    elif isinstance(item, QLineEdit):
        item.setReadOnly(switch)
    elif isinstance(item, QComboBox):
        item.setEditable(not switch)
    else:
        raise ValueError

def formatRowData(row, columns):
    """ Return formatted row string for exporting to the clipboard/csv file. """

    row['Address'] = "0x{:08x}".format(row['Address'])
    row['Offset'] = "0x{:08x}".format(row['Offset'])
    if row['Hex']:
        row['Value'] = "0x{:08x}".format(row['Value'])

    rowString = ','.join([str(row[col]) for col in columns])

    return rowString

###################################################################################################
# O V E R R I D D E N  C L A S S E S
###################################################################################################

class TableItemCompleter(QStyledItemDelegate):
    """ Delegate to handle autocompletion of a table cell """

    def createEditor(self, parent, _styleOption, index):
        """ Creates the editor when a non-readonly cell is clicked.
            Sets up and adds autocompletion information if data exists.
        """
        editor = QLineEdit(parent)

        # autocomplete data is stored as a list in Qt.UserRole, so setup if defined
        if index.data(Qt.UserRole):
            completionList = sorted(index.data(Qt.UserRole), key=str.lower)
            autoComplete = QCompleter(completionList)
            autoComplete.setModelSorting(QCompleter.CaseInsensitivelySortedModel)
            autoComplete.setCompletionMode(QCompleter.UnfilteredPopupCompletion)
            autoComplete.setCaseSensitivity(Qt.CaseInsensitive)
            editor.setCompleter(autoComplete)

        return editor

class QTableWidgetItemSortable(QTableWidgetItem):
    """ Overriden QTableWidgetItem to allow custom keys to be used for table sorting """

    def __init__(self, sortKey):
        self._sortKey = float(sortKey) if sortKey is not None else 0.0
        super().__init__()

    def __lt__(self, other):
        try:
            return self.sortKey < other.sortKey
        except AttributeError:  # when compared to other objects without sortkey (QTableWidgetItem)
            return super().__lt__(other)

    @property
    def sortKey(self):
        """ Retrieve the sort key. """
        return self._sortKey

    @sortKey.setter
    def sortKey(self, value):
        """ Set the key to be used for sorting. Must be castable to float. """
        self._sortKey = float(value)

class QTableWidgetItemValue(QTableWidgetItemSortable):
    """ Overriden QTableWidgetItemSortable to allow a raw value to be stored. """

    def __init__(self, sortKey):
        super().__init__(sortKey)       # set sortKey to (raw) value
        self._rawValue = self.sortKey   # store raw value

    @property
    def rawValue(self):
        """ Retrieve the raw value. """
        return self._rawValue

    @rawValue.setter
    def rawValue(self, value):
        """ Set the raw value so it can be retrieved later. """
        self._rawValue = value
        self.sortKey = value


###################################################################################################
# S E T G E T  M A N A G E R
###################################################################################################

class FrontEnd(QWidget):
    """ User Interface for the Set Get Manager """

    STACKSIZE = 10


    def __init__(self, name, config, app=None):
        """ Constructor of this class. """
        QWidget.__init__(self, flags=Qt.Widget)

        check_class(config, Config)

        # get linux signal handler
        self.signalHandler = app.getSignalHandler() if app is not None else None

        # initialization of global variables
        self.settingsStack    = []
        self.newRegisters     = []
        self.deletedRegisters = []
        self.devices          = {}
        self.deviceName       = ""
        self.dragRows         = None # List of dragged registers during a drag-drop operation
        self.rowCache         = None

        # main GUI default settings
        self.regLabels       = ['Device', 'Device type', 'Name', 'Address', 'Offset',
                                'Hex', 'Eng', '', 'Value', 'Unit', 'Get', 'Set']
        self.columnsToStore  = ['Device', 'Device type', 'Name', 'Address', 'Offset',
                                'Hex', 'Eng']
        self.columnsToExport = ['Device', 'Device type', 'Name', 'Address', 'Offset',
                                'Value', 'Unit', 'Hex', 'Eng']
        self.mainWindowSize  = QSize(1000, 1000)
        self.mainWindowPos   = QPoint(0,0)
        self.mainColumnWidth = [100, 100, 200, 100, 100, 50, 50, 20, 100, 50, 50, 50]
        self.mainSortColumn  = 1
        self.mainSortOrder   = Qt.AscendingOrder

        # new register GUI default settings
        self.newRegLabels    = ['Name', 'Address', 'Type',  'Unit', 'Access',
                                'Width', 'Block', 'Offset', 'Description' ]
        self.newrWindowSize  = QSize(1000, 1000)
        self.newrWindowPos   = QPoint(0,0)
        self.newrColumnWidth = [200, 95, 50, 50, 60, 50, 50, 55, 285]
        self.newrSortColumn  = 1
        self.newrSortOrder   = Qt.AscendingOrder

        # info GUI default settings
        self.infoLabels      = ['Attribute', 'Value']
        self.infoColumnWidth = [200, 300]

        # retrieve the system configuration
        self.config = config

        # load the hardware abstraction layer
        self.hal = Hal()
        self.halClient = HalClient(self.hal, self.config)
        self.halClient.requestHal()
        # register to halclient for HAL updates from the server
        self.halClient.registerCallBackFunction(self.slotReloadHal)
        
        # create a PUB socket to publish queried vales
        self.pubSocket = PubSocket()
        self.pubSocket.createZmqSocket()
        try:
            proxyConfig     = config.componentParameters('proxy')
            proxyHost       = proxyConfig['hostname']
            proxySubport    = proxyConfig['sockets']['xsub']
            proxyAddress = "tcp://{}:{}".format(proxyHost, proxySubport)
            self.pubSocket.connectSocket(proxyAddress)
            logger.info("Connected to proxy @ %s", proxyAddress)
        except (GsswError, KeyError) as e:
            logger.warning("Cannot connect to proxy, incorrect proxy configuration: %s", e)
            self.pubSocket = None

        # read the Yaml file containing the initial register list
        settingsFileRelative = self.config.componentParameters(name).get('settings')

        # if not defined then use default value
        if settingsFileRelative is None:
            settingsFileRelative = '{}.yaml'.format(name)
            logger.warning("No 'settings' file in the configuration file, using default '%s'",
                           settingsFileRelative)

        # get the absolute path of the settings file (in the config file is a relative path)
        self.settingsFile  = self.config.asAbsolutePath(settingsFileRelative)

        # read the settings (YAML format) and convert to list of registers
        settings          = self._readYaml(self.settingsFile)
        self.registerList = self._yamlToRegList(settings)

        # initialize the GUIs
        modulePath  = os.path.dirname(os.path.abspath(__file__))
        guiPath     = os.path.join(modulePath, 'gui')
        self.mainUi = uic.loadUi(os.path.join(guiPath, 'SetGetManagerMainWindow.ui'))
        self.newrUi = uic.loadUi(os.path.join(guiPath, 'SetGetManagerAddRegister.ui'))
        self.infoUi = uic.loadUi(os.path.join(guiPath, 'info.ui'))

        # define shortcuts for (existing) main GUI actions
        self.mainUi.actionOpen.setShortcut(QKeySequence.Open)
        self.mainUi.actionSave.setShortcut(QKeySequence.Save)
        self.mainUi.actionSaveAs.setShortcut(QKeySequence.SaveAs)
        self.mainUi.actionClose.setShortcut(QKeySequence.Close)
        self.mainUi.actionNewReg.setShortcut(QKeySequence.New)
        self.mainUi.actionCut.setShortcut(QKeySequence.Cut)
        self.mainUi.actionPaste.setShortcut(QKeySequence.Paste)
        self.mainUi.actionBack.setShortcut(QKeySequence.Undo)
        self.mainUi.actionInfoShow.setShortcut(QKeySequence.WhatsThis)
        self.mainUi.actionGetAll.setShortcut(QKeySequence.SelectAll)

        # create and populate the context menu of the main GUI
        self.mainUiContextMenu = QMenu()
        self.mainUiContextMenu.addAction(self.mainUi.actionNewReg)
        self.mainUiContextMenu.addAction(self.mainUi.actionCut)
        self.mainUiContextMenu.addAction(self.mainUi.actionPaste)
        self.mainUiContextMenu.addAction(self.mainUi.actionBack)
        self.mainUiContextMenu.addAction(self.mainUi.actionInfoShow)

        # create new register GUI action including shortcut
        # and add this action to new register GUI else shortcut does not work
        self.actionAdd = QAction("&Add", self)
        self.actionAdd.setShortcut(QKeySequence.New)
        self.newrUi.addAction(self.actionAdd)

        # create and populate the context menu of the new register GUI
        self.newrUiContextMenu = QMenu()
        self.newrUiContextMenu.addAction(self.actionAdd)

        # connect the handlers
        self.connectHandlers()

        # get GUI settings as stored in Qsettings (if available)
        self.readSettings()

        # we are started now
        logger.info("Started %s, version %s", name, VERSION_STRING)

    def connectHandlers(self):
        """ Connect signals to slots. """

        # main GUI actions
        self.mainUi.actionOpen.triggered.connect(self.slotOpen)
        self.mainUi.actionSave.triggered.connect(self.slotSave)
        self.mainUi.actionSaveAs.triggered.connect(self.slotSaveAs)
        self.mainUi.actionExToClipboard.triggered.connect(self.slotExportToClipboard)
        self.mainUi.actionExToFile.triggered.connect(self.slotExportToFile)
        self.mainUi.actionClose.triggered.connect(self.slotCloseMainUi)
        self.mainUi.actionNewReg.triggered.connect(self.slotNewReg)
        self.mainUi.actionCut.triggered.connect(self.slotCut)
        self.mainUi.actionPaste.triggered.connect(self.slotPaste)
        self.mainUi.actionBack.triggered.connect(self.slotBack)
        self.mainUi.actionInfoShow.triggered.connect(self.slotInfoShow)
        self.mainUi.actionGetAll.triggered.connect(self.slotGetAll)
        self.mainUi.closeEvent = self.closeEvent

        if self.signalHandler:
            self.signalHandler.sigInt.connect(self.slotSaveClose)
            self.signalHandler.sigTerm.connect(self.slotSaveClose)

        # main GUI context menu
        self.mainUi.registerTable.customContextMenuRequested.connect(self.slotMainUiContextMenu)

        # main GUI selections
        self.mainUi.registerTable.itemSelectionChanged.connect(self.slotUpdateMainUi)

        # main GUI sorting
        self.mainUi.registerTable.horizontalHeader(). \
            sortIndicatorChanged.connect(self.slotSortMainUi)

        # main GUI signal/slot for changes to the table
        self.mainUi.registerTable.itemPressed.connect(self.slotCacheItemData)
        self.mainUi.registerTable.itemChanged.connect(self.slotUpdateRegister)

        # main GUI auto-complete delegate
        self.mainUi.registerTable.setItemDelegate(TableItemCompleter())

        # new register GUI combo box
        self.newrUi.device.currentIndexChanged[int].connect(self.slotDeviceChanged)

        # new register GUI action
        self.actionAdd.triggered.connect(self.slotAdd)
        applyButton = self.newrUi.buttonBox.button(QDialogButtonBox.Apply)
        applyButton.setText("Add")
        applyButton.clicked.connect(self.slotAdd)

        # new register GUI context menu
        self.newrUi.registerTable.customContextMenuRequested.connect(self.slotNewrUiContextMenu)

        # new register GUI selections
        self.newrUi.registerTable.itemSelectionChanged.connect(self.slotUpdateNewrUi)

        # new register GUI sorting
        self.newrUi.registerTable.horizontalHeader(). \
            sortIndicatorChanged.connect(self.slotSortNewrUi)

    def show(self):
        """ Shows the user interface. """

        # main GUI properties
        self.mainUi.resize(self.mainWindowSize)
        self.mainUi.move(self.mainWindowPos)
        self.mainUi.registerTable.setColumnCount(len(self.regLabels))
        self.mainUi.registerTable.setRowCount(1)
        self.mainUi.registerTable.setHorizontalHeaderLabels(self.regLabels)
        self.mainUi.registerTable.horizontalHeader().setFixedHeight(40)
        self.mainUi.registerTable.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)
        self.mainUi.registerTable.horizontalHeader().setSortIndicatorShown(True)
        self.mainUi.registerTable.horizontalHeader().setStretchLastSection(False)
        self.mainUi.registerTable.setContextMenuPolicy(Qt.CustomContextMenu)
        self.mainUi.registerTable.horizontalHeader().blockSignals(True)
        self.mainUi.registerTable.horizontalHeader(). \
            setSortIndicator(self.mainSortColumn, self.mainSortOrder)
        self.mainUi.registerTable.horizontalHeader().blockSignals(False)
        for col in range(len(self.regLabels)):
            self.mainUi.registerTable.setColumnWidth(col, self.mainColumnWidth[col])
        self.mainUi.registerTable.viewport().installEventFilter(self)

        # new register GUI properties
        self.newrUi.resize(self.newrWindowSize)
        self.newrUi.move(self.newrWindowPos)
        self.newrUi.registerTable.setColumnCount(len(self.newRegLabels))
        self.newrUi.registerTable.setRowCount(1)
        self.newrUi.registerTable.setHorizontalHeaderLabels(self.newRegLabels)
        self.newrUi.registerTable.horizontalHeader().setFixedHeight(40)
        self.newrUi.registerTable.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)
        self.newrUi.registerTable.horizontalHeader().setSortIndicatorShown(True)
        self.newrUi.registerTable.horizontalHeader().setStretchLastSection(True)
        self.newrUi.registerTable.setContextMenuPolicy(Qt.CustomContextMenu)
        self.newrUi.registerTable.horizontalHeader().blockSignals(True)
        self.newrUi.registerTable.horizontalHeader(). \
            setSortIndicator(self.newrSortColumn, self.newrSortOrder)
        self.newrUi.registerTable.horizontalHeader().blockSignals(False)
        for col in range(len(self.newRegLabels)):
            self.newrUi.registerTable.setColumnWidth(col, self.newrColumnWidth[col])

        # info GUI properties
        self.infoUi.infoTable.setColumnCount(len(self.infoLabels))
        self.infoUi.infoTable.setHorizontalHeaderLabels(self.infoLabels)
        self.infoUi.infoTable.setRowCount(10)
        self.infoUi.infoTable.horizontalHeader().setFixedHeight(40)
        self.infoUi.infoTable.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)
        self.infoUi.infoTable.horizontalHeader().setStretchLastSection(True)
        for col in range(len(self.infoLabels)):
            self.infoUi.infoTable.setColumnWidth(col, self.infoColumnWidth[col])

        # fill and show the main GUI
        self.createDevices()
        self.fillRegisterTable()
        self.mainUi.show()

        # fill but do not yet show the new register GUI
        self.fillDeviceComboBox()
        self.fillNewRegisterTable()
        self.sortNewRegisterTable()

        # PM: info GUI is filled and shown when main GUI info button is pressed


###################################################################################################
# M A I N   U S E R   I N T E R F A C E
###################################################################################################
    @property
    def colDevice(self):
        """ Return the column index for the Device column """
        return self.regLabels.index('Device')

    @property
    def colDeviceType(self):
        """ Return the column index for the Device Type column """
        return self.regLabels.index('Device type')

    @property
    def colName(self):
        """ Return the column index for the Register Name column """
        return self.regLabels.index('Name')

    @property
    def colAddress(self):
        """ Return the column index for the Address column """
        return self.regLabels.index('Address')

    @property
    def colOffset(self):
        """ Return the column index for the Offset column """
        return self.regLabels.index('Offset')

    @property
    def colHex(self):
        """ Return the column index for the Hex column """
        return self.regLabels.index('Hex')

    @property
    def colEng(self):
        """ Return the column index for the Eng column """
        return self.regLabels.index('Eng')

    @property
    def colStatus(self):
        """ Return the column index for the Status column """
        return self.colValue-1

    @property
    def colValue(self):
        """ Return the column index for the Value column """
        return self.regLabels.index('Value')

    @property
    def colUnit(self):
        """ Return the column index for the Unit column """
        return self.regLabels.index('Unit')

    @property
    def colGet(self):
        """ Return the column index for the Get column """
        return self.regLabels.index('Get')

    @property
    def colSet(self):
        """ Return the column index for the Set column """
        return self.regLabels.index('Set')

    def createDevices(self):
        """ Create the devices necessary for set/get operations """
        for device in self.hal.getDeviceNames():
            try:
                self.devices[device] = Device(device, self.config, self.hal)
            except GsswError:
                # if Egress component for the Device is not configured, ignore the error
                pass

    def getRegisterInfoByRow(self, row):
        """ Return register info for the current row. """
        deviceName = self.getCellValue(row, self.colDevice)
        registerName = self.getCellValue(row, self.colName)
        registerAddress = self.getCellValue(row, self.colAddress)
        try:
            return self.hal.getRegisterInfo(deviceName, registerName, registerAddress)
        except DataError:
            return

    def updateRegisterValues(self, register):
        """ Update the register information with the most recent from the HAL or fill missing
            fields with default values.

        :param dict register:   A dictionary with register information as entries. At the minimum,
                                must contain an entry for Device, Name, and Address (can be None).
                                    e.g. {'Device': 'Dev01, 'Name' : 'Reg01', 'Address': None}
        """

        try:
            registerInfo = self.hal.getRegisterInfo(register['Device'], register['Name'],
                                                    register['Address'])
        except DataError:
            if register:
                logger.warning('Cannot find %s:%s, skipping', register['Device'], register['Name'])
            return

        # ignore block-type registers
        if registerInfo.block is True:
            logger.warning('Register %s/%s is block register, ignored',
                           register['Device'], register['Name'])
            return

        register['Device type'] = self.hal.getDeviceType(register['Device']).name

        # one of Name and Address may diffur from YAML dict, so redefine
        register['Name'] = registerInfo.name
        register['Address'] = registerInfo.address
        register['Unit'] = registerInfo.unit

        # determine read/write access
        register['Get'] = registerInfo.access in (Register.READ_WRITE, Register.READ_ONLY)
        register['Set'] = registerInfo.access in (Register.READ_WRITE, Register.WRITE_ONLY)

        # if not stored (on load or open) then initialize:
        register['Hex'] = register.get('Hex', False)
        register['Eng'] = register.get('Eng', False)
        register['Value'] = register.get('Value', 0)
        register['RawValue'] = register.get('RawValue', 0)
        register['Offset'] = register.get('Offset', 0)

        return register

    def getValue(self, row, eng=None):
        """ Return the number value of of the register at specified row. """

        # if eng not specified, then get current row values
        if eng is None:
            eng = self.getCellValue(row, self.colEng)

        # get register info
        registerInfo = self.getRegisterInfoByRow(row)
        if registerInfo is None:
            return

        textValue = self.mainUi.registerTable.item(row, self.colValue).text()
        try:
            if isinstance(self.mainUi.registerTable.cellWidget(row, self.colValue), QComboBox):
                return self.mainUi.registerTable.cellWidget(row, self.colValue).currentText()
            elif eng or registerInfo.type == 'float':
                return float(textValue)
            else:
                return int(textValue, 0)
        except ValueError:
            gsswUiWarning("Invalid input: should be type '{}'".format(registerInfo.type))

    def getCellValue(self, row, col):
        """ Return the cell value. """

        # if column name given, then retrieve column index
        if isinstance(col, str):
            try:
                col = self.regLabels.index(col)
            except ValueError:
                logger.error("Column '%s' not defined", col)
                return

        cellWidget = self.mainUi.registerTable.cellWidget(row, col)
        if isinstance(cellWidget, QCheckBox):
            return cellWidget.isChecked()
        elif isinstance(cellWidget, QComboBox):
            return cellWidget.currentText()
        elif col == self.colGet or col == self.colSet:
            return isinstance(cellWidget, QPushButton)
        elif col == self.colOffset or col == self.colAddress:
            try:
                return int(self.mainUi.registerTable.item(row, col).text(), 0)
            except ValueError:
                logger.warning("Invalid offset at row=%s", row)
                return
        elif col == self.colValue:
            return self.getValue(row)
        else:
            return self.mainUi.registerTable.item(row, col).text()

    def getTableRow(self, row):
        """ Get all the values in the current row. """

        register = {}
        for i, col in enumerate(self.regLabels):
            register[col] = self.getCellValue(row, i)
        try:
            register['RawValue'] = self.mainUi.registerTable.item(row, self.colValue).rawValue
        except AttributeError:
            register['RawValue'] = 0
        return register

    def rowExists(self, row):
        """ Checks if an entry (address and offset) alerady exists for given device. """

        device = self.getCellValue(row, self.colDevice)
        address = self.getCellValue(row, self.colAddress) + self.getCellValue(row, self.colOffset)

        for tableRow in range(self.mainUi.registerTable.rowCount() - 1):
            rowDevice = self.getCellValue(tableRow, self.colDevice)
            rowAddress = self.getCellValue(tableRow, self.colAddress) \
                         + self.getCellValue(tableRow, self.colOffset)

            if tableRow != row and rowDevice == device and rowAddress == address:
                return True
        return False

    def fillRegisterTable(self, registerList=None):
        """ Fill the main GUI register table with data. """

        # first, remove old contents
        self.mainUi.registerTable.blockSignals(True)
        self.mainUi.registerTable.clearContents()

        if registerList is None:
            registerList = self.registerList

        try:
            if registerList is None:
                # Do the 'finally' thing below
                return

            # then, fill the table according to the SetGet settings file
            self.mainUi.registerTable.setRowCount(len(registerList))
            for row, register in enumerate(registerList):
                self.fillRegister(row, register)
            self.fillDummyRegister()
            self.mainUi.registerTable.resizeRowsToContents()
        finally:
            self.mainUi.registerTable.blockSignals(False)

    def fillDummyRegister(self):
        """ Fill last row with empty QTableItemWidget objects. """
        row = self.mainUi.registerTable.rowCount()
        self.mainUi.registerTable.setRowCount(row + 1)

        for col in range(len(self.regLabels)):
            tableItem = QTableWidgetItem()
            setReadOnly(tableItem, True)
            self.mainUi.registerTable.setItem(row, col, tableItem)

    def fillRegister(self, row, register):
        """ Fill one row with register info. """

        # get register info from HAL
        try:
            regInfo = self.hal.getRegisterInfo(register['Device'], register['Name'],
                                               register['Address'])

            # Get list of registers, filtering out block registers
            regDict = self.hal.getRegistersInfo(register['Device'])
            regList = [reg for reg in regDict if not regDict[reg].block]

        except DataError:
            logger.warning('Cannot find %s: %s[0x%08x], skipping',
                           register['Device'], register['Name'], register['Address'])
            return

        regConv = self.hal.getRegisterConversion(register['Device'], None, None, register['Name'])

        # fill values
        self.fillRegisterCell(row, self.colDevice, register['Device'])
        self.fillRegisterCell(row, self.colDeviceType, register['Device type'])
        self.fillRegisterCell(row, self.colName, register['Name'],
                              readOnly=False, autoComplete=regList)
        self.fillRegisterCell(row, self.colAddress, "0x{:08x}".format(register['Address']))
        self.fillRegisterCell(row, self.colOffset, "0x{:08x}".format(register['Offset']),
                              readOnly=False)
        self.fillRegisterCell(row, self.colUnit, register['Unit'],
                              tooltip='Unit may not apply to Raw value.'
                                      '\nUse Eng checkbox to see converted value.')

        # fill checkboxes
        self.fillCheckBox(row, self.colHex, register['Hex'], self.hexCheckBoxAction,
                          disabled=register['Eng'] or regInfo.type == 'float')
        self.fillCheckBox(row, self.colEng, register['Eng'], self.engCheckBoxAction,
                          disabled=regConv is None)

        # fill buttons
        self.fillButton(row, self.colGet, register['Get'], True, self.getButtonAction)
        self.fillButton(row, self.colSet, register['Set'], False, self.setButtonAction)

        # fill value
        self.fillValueCell(row, register['RawValue'], regInfo.access == Register.READ_ONLY)
        self.fillValue(row, register['Value'])

        # create status indicator
        self.mainUi.registerTable.setItem(row, self.colStatus, QTableWidgetItem())
        self.setValueStatus(row, "NEW")

    def fillRegisterCell(self, row, col, content, readOnly=True, tooltip=None, autoComplete=None):
        """ Fill the contents of one table cell with a QTableWidgetItem object. """
        # Use overriden QTableWidgetItem if possible (i.e. if castable to float)
        tableItem = QTableWidgetItem()

        tableItem.setText(str(content))
        setReadOnly(tableItem, readOnly)
        if tooltip is not None:
            tableItem.setToolTip(tooltip)

        # Set autocomplete data (list) for cell
        if autoComplete is not None:
            tableItem.setData(Qt.UserRole, autoComplete)

        self.mainUi.registerTable.setItem(row, col, tableItem)

    def fillCheckBox(self, row, col, checked, slot, disabled=False):
        """ Fill the contents of one table cell with QTableItemWidget and QCheckBox objects. """
        tableItem = QTableWidgetItemSortable(checked)   # sortable on checked
        setReadOnly(tableItem, True)
        tableCell = QCheckBox()
        tableCell.setChecked(checked)
        tableCell.clicked.connect(slot)
        tableCell.setDisabled(disabled)

        # insert the item and the cell widget into the QTableWidget
        self.mainUi.registerTable.setItem(row, col, tableItem)
        self.mainUi.registerTable.setCellWidget(row, col, tableCell)

    def fillButton(self, row, col, registerAccess, get, buttonAction):
        """ Fill the contents of the button with QTableItemWidget and QPushButton objects. """
        tableItem = QTableWidgetItemSortable(True)      # sortable if button present
        setReadOnly(tableItem, True)
        self.mainUi.registerTable.setItem(row, col, tableItem)

        self.mainUi.registerTable.item(row, col).sortKey = registerAccess
        # only add button if register has read/write access
        if registerAccess:
            tableCell = QPushButton()
            arrowIcon = QStyle.SP_ArrowDown if get else QStyle.SP_ArrowUp
            arrow = self.style().standardIcon(arrowIcon)
            tableCell.setIcon(arrow)
            tableCell.setIconSize(QSize(20, 20))
            tableCell.clicked.connect(buttonAction)
            self.mainUi.registerTable.setCellWidget(row, col, tableCell)

    def fillValueCell(self, row, value, readOnly=False):
        """ Fill the contents of the value table cell with a QTableWidgetItemValue object. """
        tableItem = QTableWidgetItemValue(value)
        tableItem.setText(str(value))
        setReadOnly(tableItem, readOnly)
        self.mainUi.registerTable.setItem(row, self.colValue, tableItem)

    def fillValue(self, row, value):
        """ Fill the contents of the data field with a QTableItemWidget;
            if engineering selected and LUT is present, also instantiate a QComboBox object.
        """
        # get register info
        deviceName = self.getCellValue(row, self.colDevice)
        registerName = self.getCellValue(row, self.colName)
        conversion = self.hal.getRegisterConversion(deviceName, None, None, registerName)

        # instantiate a QComboBox if engineering mode and set LUT selected
        if self.getCellValue(row, self.colEng) and conversion.setConv == Conversion.LUT:
            tableCell = QComboBox()
            for item in conversion.setLut:
                tableCell.addItem(item)
            tableCell.setCurrentIndex(0)
            # do NOT set to read-only else QComboBox lineEdit mode disabled
            setReadOnly(tableCell, False)
            self.mainUi.registerTable.setCellWidget(row, self.colValue, tableCell)
        # else, if applicable, remove QComboBox
        elif isinstance(self.mainUi.registerTable.cellWidget(row, self.colValue), QComboBox):
            self.mainUi.registerTable.removeCellWidget(row, self.colValue)

        # update value
        self.updateValue(row, value)

    def setValueStatus(self, row, status):
        """ Set the status icon to show if the value has been changed by user. """

        if status.upper() == "GET":
            icon = self.style().standardIcon(QStyle.SP_ArrowDown)
        elif status.upper() == "SET":
            icon = self.style().standardIcon(QStyle.SP_ArrowUp)
        else:
            icon = self.style().standardIcon(QStyle.SP_BrowserStop)

        cellItem = self.mainUi.registerTable.item(row, self.colStatus)
        cellItem.setIcon(icon)
        cellItem.setToolTip("{} @ {}".format(status, datetime.now().strftime("%Y-%m-%d %H:%M:%S")))

    def updateOffset(self, row, offset):
        """ Update the offset value."""
        # prevent infinite recursion by disabling signals (itemChanged)
        self.mainUi.registerTable.blockSignals(True)
        newOffset = self.rowCache['Offset']

        try:
            # get register info
            registerInfo = self.getRegisterInfoByRow(row)
            if registerInfo is None:
                return

            # check that new offset is valid
            assert isinstance(offset, int), "Offset must be an integer!"
            assert offset <= registerInfo.maxOffset, \
                "Address offset is too large %d > %d for register %s" \
                % (offset, registerInfo.maxOffset, registerInfo.name)
            assert not self.rowExists(row), "Offset=0x%08x already exists for %s!" \
                                            % (offset, registerInfo.name)
            newOffset = offset
        except AssertionError as e:
            gsswUiError(str(e))
        finally:
            offsetStr = "0x{:08x}".format(newOffset)
            self.mainUi.registerTable.item(row, self.colOffset).setText(offsetStr)
            self.mainUi.registerTable.blockSignals(False)

    def updateValue(self, row, value):
        """ Update the data field in the correct format """

        # get register info
        registerInfo = self.getRegisterInfoByRow(row)
        if registerInfo is None:
            return

        if value is None:
            logger.warning("Could not update value, reverting to stored value: %s",
                           self.rowCache['Value'])
            self.updateValue(row, self.rowCache['Value'])
            return

        cellItem = self.mainUi.registerTable.item(row, self.colValue)
        # Value should be ENG
        if self.getCellValue(row, self.colEng):
            # update table cell
            cellWidget = self.mainUi.registerTable.cellWidget(row, self.colValue)
            if isinstance(cellWidget, QComboBox):
                cellWidget.lineEdit().setText("%s" % value)
                cellWidget.lineEdit().setReadOnly(True)
            else:
                cellItem.setText("%.04f" % float(value))
        # Value should be HEX
        elif self.getCellValue(row, self.colHex):
            hexval = (int(value) + (1 << 32)) % (1 << 32)
            cellItem.setText("0x%08x" % hexval)
        # Value should be RAW
        else:
            if registerInfo.type == "float":
                cellItem.setText("%.04f" % float(value))
            else:
                cellItem.setText("%d" % int(value))

    def hexCheckBoxAction(self):
        """ Slot: action for the HEX checkbox in main GUI register table. """
        # get row info
        row = self.getMainUiRow()
        cellValue = self.getValue(row)
        hexIsChecked = self.sender().isChecked()

        self.sender().clearFocus() # Otherwise focus causes empty cell when table is cleared

        # Toggle sort key
        self.mainUi.registerTable.item(row, self.colHex).sortKey = hexIsChecked

        # update the data field
        self.updateValue(row, cellValue)

    def engCheckBoxAction(self):
        """ Slot: action for the ENG checkbox in main GUI register table. """
        # get row info
        row = self.getMainUiRow()
        engIsChecked = self.sender().isChecked()
        deviceName = self.getCellValue(row, self.colDevice)
        registerName = self.getCellValue(row, self.colName)
        registerInfo = self.getRegisterInfoByRow(row)
        if registerInfo is None:
            return

        self.sender().clearFocus()  # Otherwise focus causes empty cell when table is cleared

        # Toggle sort key
        self.mainUi.registerTable.item(row, self.colEng).sortKey = engIsChecked

        rawValue = self.mainUi.registerTable.item(row, self.colValue).rawValue
        # it was RAW, now want ENG
        if engIsChecked:
            newValue = self.convertRawToEng(rawValue, deviceName, registerName)
        # it was ENG, now want RAW
        else:
            newValue = rawValue

        # if input is invalid, then display error and revert to previous state
        if newValue is None:
            gsswUiError("Could not get Eng value. Enter valid raw input.")
            self.mainUi.registerTable.cellWidget(row, self.colEng).setCheckState(not engIsChecked)
            self.updateValue(row, self.rowCache['Value'])
            return

        # if appropriate, insert or remove data field QComboBox
        self.mainUi.registerTable.blockSignals(True)
        self.fillValue(row, newValue)
        self.mainUi.registerTable.blockSignals(False)

        # enable/disable the Hex checkbox when eng checkbox is enabled
        # unless float type since should always remain disabled
        if registerInfo.type != 'float':
            self.mainUi.registerTable.cellWidget(row, self.colHex).setDisabled(engIsChecked)

    def getButtonAction(self):
        """ Slot: action for GET push button in main GUI register table. """
        currentRow = self.getMainUiRow()
        try:
            self.sender().clearFocus() # Otherwise focus causes empty cell when table is cleared
        except AttributeError:
            pass

        deviceName = self.getCellValue(currentRow, self.colDevice)
        registerName = self.getCellValue(currentRow, self.colName)

        # determine the 'identifier' by combining the register name and the address offset
        try:
            registerOffset = self.getCellValue(currentRow, self.colOffset)
        except ValueError:
            gsswUiWarning("Address offset must be an integer.")
            return False
        identifier = "%s:%s" % (registerName, registerOffset)

        # get the raw value from the device
        try:
            value = self.devices[deviceName].get(identifier)
            logger.debug('getButton raw value = %s', value)
        except Exception as ex: # Show a popup with a message instead of quitting the application
            logger.error("An error occurred while reading the register: %s", ex)
            gsswUiError("An error occurred while reading the register: \n{}".format(ex))
            return False

        # store raw value for later switching between eng/raw display
        self.mainUi.registerTable.item(currentRow, self.colValue).rawValue = value

        # if Eng is selected, convert to engineering value
        if self.getCellValue(currentRow, self.colEng):
            value = self.convertRawToEng(value, deviceName, registerName)
            if value is None:
                gsswUiWarning("Could not convert to Eng value.")
                return False
            publishName = "{}_{}_{}_ENG".format(deviceName, registerName, registerOffset)
            units = self.getCellValue(currentRow, self.colUnit)
        else:
            publishName = "{}_{}_{}_RAW".format(deviceName, registerName, registerOffset)
            units = None

        # publish the raw/eng value
        self.publishValue(publishName, value, units)

        # update the data field
        self.updateValue(currentRow, value)
        self.setValueStatus(currentRow, "GET")
        return True

    def setButtonAction(self):
        """ Action for SET push button in main GUI register table. """
        currentRow = self.getMainUiRow()
        self.sender().clearFocus() # Otherwise focus causes empty cell when table is cleared

        # get row info
        deviceName = self.getCellValue(currentRow, self.colDevice)
        registerName = self.getCellValue(currentRow, self.colName)

        # determine the 'identifier' by combining the register name and the address offset
        try:
            registerOffset = self.getCellValue(currentRow, self.colOffset)
        except ValueError:
            gsswUiWarning("Address offset must be an integer.")
            return
        identifier = "%s:%s" % (registerName, registerOffset)

        # convert current cell value to raw form
        cellValue = self.getValue(currentRow)
        if self.getCellValue(currentRow, self.colEng):
            rawValue = self.convertEngToRaw(cellValue, deviceName, registerName)
        else:
            rawValue = cellValue

        if rawValue is None:
            gsswUiError("An error occurred while setting the register: "
                        "Could not get raw value from current cell value.")
            return

        # store raw value for later switching between eng/raw display
        self.mainUi.registerTable.item(currentRow, self.colValue).rawValue = rawValue

        # send the raw value to the device
        logger.debug('setButton raw value = %s', rawValue)
        try:
            self.devices[deviceName].set(identifier, rawValue)
        except Exception as ex:
            # Handle exceptions during setting
            logger.error("An error occurred while setting the register: %s", ex)
            gsswUiError("An error occurred while setting the register: \n{}".format(ex))
            return
        self.setValueStatus(currentRow, "SET")

    def convertEngToRaw(self, value, deviceName, registerName):
        """ Convert the given eng value to the raw value. """
        # get register information
        conversion = self.hal.getRegisterConversion(deviceName, None, None, registerName)
        registerInfo = self.hal.getRegisterInfo(deviceName, registerName)

        # convert
        try:
            # if is LUT, then already in correct format for applyConversion
            engValue = value if conversion.setConv == Conversion.LUT else float(value)
            rawValue = conversion.applyConversion(engValue, False)
            assert rawValue is not None
        except ValueError:
            logger.warning("convertEngToRaw: Value '%s' cannot be converted to float.", value)
            return
        except AssertionError:
            logger.warning("convertEngToRaw: Value '%s' cannot be converted to raw.", value)
            return

        if registerInfo.type in ['int', 'uint']:
            rawValue = int(rawValue)

        return rawValue

    def convertRawToEng(self, rawValue, deviceName, registerName):
        """ Convert the given raw value to an engineering value """
        # get register info
        conversion = self.hal.getRegisterConversion(deviceName, None, None, registerName)

        # convert
        try:
            engValue = conversion.applyConversion(rawValue, True)
            assert engValue is not None
            return engValue
        except AttributeError:
            logger.warning("convertRawToEng: No conversion for %s:%s.", deviceName, registerName)
        except (AssertionError, DataError):
            logger.warning("convertRawToEng: Can not convert '%s' to eng value.", rawValue)

    def addRegisters(self, registers, row=None, overwrite=False):
        """ Insert register(s) into the main GUI register table.

        :param list registers:  The list of register(s) to add. At the minimum, must contain
                                an entry for Device, Name, and Address (can be None).
                                    e.g. [{'Device': 'Dev01, 'Name' : 'Reg01', 'Address': None}]
        :param int row:         The row to insert the register(s).
                                Defaults to the end of the table if not specified.
        :param bool overwrite:  Whether to overwrite the register(s) at the specified row.
                                Note: can not be specified for multiple registers.
        """

        if overwrite and len(registers) > 1:
            logger.warning("Can not overwrite more than one register at a time!")
            return

        # Add to end of table if row not specified
        if row is None:
            row = self.mainUi.registerTable.rowCount()

        self.mainUi.registerTable.blockSignals(True)

        # save the current settings
        settingsSave = self._regListToYaml(saveData=True)

        addedRegisters = 0
        for register in registers:

            # update register info from HAL
            register = self.updateRegisterValues(register)
            if register is None:
                # updateRegisterValues already displays log error
                continue

            # increment register offset if already in main GUI
            register['Offset'] = self.getNextRegisterOffset(register)
            if register['Offset'] < 0:
                # Can only add one register for each offset
                logger.warning("Register %s:%s defined for all offsets... skipping",
                               register['Device'], register['Name'])
                continue

            # Add register
            if overwrite:
                self.mainUi.registerTable.removeRow(row)
            self.mainUi.registerTable.insertRow(row)
            self.fillRegister(row, register)
            logger.debug("Added %s:%s at row %d", register['Device'], register['Name'], row)
            addedRegisters += 1
            row += 1

        logger.debug("Added %d of %d registers", addedRegisters, len(registers))
        self.mainUi.registerTable.blockSignals(False)

        # if the settings have changed, store the saved version on the stack
        if addedRegisters > 0:
            self.stackPush(settingsSave)

        if addedRegisters == len(registers):
            return True
        else:
            gsswUiWarning("Not all registers could be added. See log for details.")
            return False

    def deleteRegister(self):
        """ Delete the selected register(s). """

        # first, save the current settings
        settingsSave = self._regListToYaml(saveData=True)

        # determine which register(s) need(s) to be deleted
        indexes = {idx.row() for idx in self.mainUi.registerTable.selectedIndexes()}
        rowsToDelete = sorted(list(indexes), reverse=True)

        # delete the selected rows and store them into self.deletedRegisters
        self.deletedRegisters = []
        for row in rowsToDelete:
            self.deletedRegisters.append(self.getTableRow(row))
            self.mainUi.registerTable.removeRow(row)

        # if the settings have changed, store the saved version on the stack
        # and enable the paste action (for copy paste operations)
        if rowsToDelete:
            self.stackPush(settingsSave)
            self.actionPasteEnable(True)

    def pasteRegister(self):
        """ Paste the deleted register(s). """

        self.mainUi.registerTable.blockSignals(True)

        # first, save the current settings
        settingsSave = self._regListToYaml(saveData=True)

        # determine the position where the register(s) will be inserted
        # when no row is selected, currentRow = -1, in that case currentRow is set to rowCount
        currentRow = self.mainUi.registerTable.currentRow()
        if not 0 <= currentRow < self.mainUi.registerTable.rowCount():
            currentRow = self.mainUi.registerTable.rowCount()

        # insert the deleted register(s) on the current position
        currentRegs = self.getCurrentRegs()
        registersPasted = False
        self.mainUi.registerTable.clearSelection()
        for reg in self.deletedRegisters:
            register = self.updateRegisterValues(reg)
            if register is None:
                continue

            # first check if selected register already in main window
            if (register['Device'], register['Name'], register['Offset']) in currentRegs:
                logger.warning('Register %s/%s/%s is already in current list, ignored',
                    register['Device'], register['Name'], register['Offset'])
            # self.deletedRegisters is in reverse order
            # therefore, no need to increment currentRow after each insert
            else:
                self.mainUi.registerTable.insertRow(currentRow)
                self.fillRegister(currentRow, register)
                self.mainUi.registerTable.selectRow(currentRow)
                registersPasted = True

        # if the settings have changed, store the saved version on the stack
        if registersPasted:
            self.stackPush(settingsSave)

        # nothing to paste anymore so disable the paste action
        self.actionPasteEnable(False)

        self.mainUi.registerTable.blockSignals(False)

    def getCurrentRegs(self):
        """ Create a list of existing (deviceName,regName) tuples. """
        currentRegs = set()
        for row in range(self.mainUi.registerTable.rowCount() - 1):
            register = self.getTableRow(row)
            currentRegs.add((register['Device'], register['Name'], register['Offset']))
        return currentRegs

    def getMainUiRow(self):
        """ Alternative for self.mainUi.registerTable.currentRow(). """

        # self.mainUi.registerTable.currentRow() does not work with cell widgets
        # other than QTableItemWidget
        mywidget = self.mainUi.focusWidget()
        index = self.mainUi.registerTable.indexAt(mywidget.pos())
        return index.row()

    def getNextRegisterOffset(self, register):
        """ Return the next highest offset of a register defined in main GUI or -1 """

        try:
            regInfo = self.hal.getRegisterInfo(register['Device'], register['Name'],
                                               register['Address'])
        except DataError:
            return -1

        # get set of offsets for given register already in main ui table
        offsets = {self.getCellValue(row, 'Offset')
                   for row in range(self.mainUi.registerTable.rowCount())
                   if register['Device'] == self.getCellValue(row, 'Device')
                   and register['Address'] == self.getCellValue(row, 'Address')}

        # find first available offset
        for offset in range(regInfo.maxOffset + 1):
            if offset not in offsets:
                return offset
        return -1

    def eventFilter(self, obj, event):
        """ Eventfilter used for drag and drop operations. """
        if event.type() == QEvent.DragEnter and \
           (obj is self.mainUi.registerTable.viewport()):
            indexes = {idx.row() for idx in self.mainUi.registerTable.selectedIndexes()}
            self.dragRows = sorted(list(indexes), reverse=True)
            logger.debug("dragRows = %s", self.dragRows)
            event.accept()
            return True

        elif event.type() == QEvent.Drop and \
            (obj is self.mainUi.registerTable.viewport()) and self.dragRows is not None:
            # Determine correct row to drop to
            dropRow = self.mainUi.registerTable.rowAt(event.pos().y())
            for row in self.dragRows:
                if row < dropRow:
                    dropRow += 1
            if not 0 <= dropRow < self.mainUi.registerTable.rowCount():
                dropRow = self.mainUi.registerTable.rowCount() - 1
            logger.debug("dropRow = %s", dropRow)

            registers = [self.getTableRow(row) for row in self.dragRows]
            self.mainUi.registerTable.blockSignals(True)
            for register in registers:
                self.mainUi.registerTable.insertRow(dropRow)
                self.fillRegister(dropRow, register)
            self.mainUi.registerTable.blockSignals(False)

            self.dragRows = None
            event.accept()
            return True

        return super().eventFilter(obj, event)


###################################################################################################
# N E W   R E G I S T E R   U S E R   I N T E R F A C E
###################################################################################################

    def fillDeviceComboBox(self):
        """ Fill the new register GUI device combo box with data. """

        # block the signals
        self.newrUi.device.blockSignals(True)

        # create a list of devices present in the hardware abstraction layer (HAL)
        deviceList = self.hal.getDeviceNames()
        deviceList.sort()
        availableComponents = list(self.config["components"].keys())

        # populate the device combo box
        self.newrUi.device.clear()
        firstValidDeviceIndex = -1
        for deviceNum, device in enumerate(deviceList):
            self.newrUi.device.addItem(device)
            deviceObject = self.hal.getDevice(device)
            # The daemon in HAL is called component in the config file.
            if deviceObject.daemon not in availableComponents:
                self.newrUi.device.model().item(deviceNum).setText(
                    self.newrUi.device.model().item(deviceNum).text() + \
                    " (Daemon not in Config)")
                self.newrUi.device.model().item(deviceNum).setEnabled(False)
            else:
                # Remember 1st valid device in list, so that we can select it by default
                if firstValidDeviceIndex == -1:
                    firstValidDeviceIndex = deviceNum

        if self.deviceName in deviceList:
            self.newrUi.device.setCurrentText(self.deviceName)
        else:
            # Select first valid device (where daemon is in config)
            self.newrUi.device.setCurrentIndex(firstValidDeviceIndex)
            self.deviceName = self.newrUi.device.currentText()

        # de-block the signals
        self.newrUi.device.blockSignals(False)

    def fillNewRegisterTable(self):
        """ Fill the new register GUI register table with data. """

        # first, remove old contents of the new register table
        self.newrUi.registerTable.clearContents()

        # fill newRegisters list with all registers belonging to the selected device
        if self.deviceName not in self.devices:
            return
        newRegisters = self.hal.getRegistersInfo(self.deviceName)

        # filter the registers with block set to False
        self.newRegisters = {name:reg for (name, reg) in newRegisters.items() if not reg.block}

        # set the row count of the new register table
        self.newrUi.registerTable.setRowCount(len(self.newRegisters))

        # fill the new register table with the newRegisters list
        row = 0
        for reg in self.newRegisters:
            self.fillNewRegister(self.newRegisters[reg], row)
            #self.newrUi.registerTable.setRowHeight(row, 20)
            row += 1
        self.newrUi.registerTable.resizeRowsToContents()

        # sort the new register table
        self.sortNewRegisterTable()

    def fillNewRegister(self, register, row):
        """ Fill one row with register info. """

        colName        = self.newRegLabels.index('Name')
        colAddress     = self.newRegLabels.index('Address')
        colType        = self.newRegLabels.index('Type')
        colUnit        = self.newRegLabels.index('Unit')
        colAccess      = self.newRegLabels.index('Access')
        colWidth       = self.newRegLabels.index('Width')
        colBlock       = self.newRegLabels.index('Block')
        colOffset      = self.newRegLabels.index('Offset')
        colDescription = self.newRegLabels.index('Description')

        self.fillNewRegisterCell(              register.name,        row, colName)
        self.fillNewRegisterCell("0x%08x" %    register.address,     row, colAddress)
        self.fillNewRegisterCell(              register.type,        row, colType)
        self.fillNewRegisterCell(              register.unit,        row, colUnit)
        self.fillNewRegisterCell(access2String(register.access),     row, colAccess)
        self.fillNewRegisterCell(              register.width,       row, colWidth)
        self.fillNewRegisterCell(              register.block,       row, colBlock)
        self.fillNewRegisterCell(              register.maxOffset,   row, colOffset)
        self.fillNewRegisterCell(              register.description, row, colDescription,
                                 tooltip=register.description)

    def fillNewRegisterCell(self, content, row, col, tooltip=None):
        """ Fill the contents of one table cell with a QTableWidgetItem object. """
        tableItem = QTableWidgetItem()
        tableItem.setText(str(content))
        setReadOnly(tableItem, True)
        if tooltip is not None:
            tableItem.setToolTip(tooltip)
        self.newrUi.registerTable.setItem(row, col, tableItem)

    def addNewRegisters(self):
        """ Insert selected register(s) into the main GUI register table. """

        # determine the active row (if any) in the main GUI
        # when no row is selected, currentRow = -1, in that case currentRow is set to rowCount
        currentRow = self.mainUi.registerTable.currentRow()
        if not 0 <= currentRow < self.mainUi.registerTable.rowCount():
            currentRow = self.mainUi.registerTable.rowCount()

        # create list of registers to add
        colAddr = self.newRegLabels.index('Address')
        colName = self.newRegLabels.index('Name')
        rowsToInsert = {idx.row() for idx in self.newrUi.registerTable.selectedIndexes()}
        registers = []
        for row in rowsToInsert:
            # Get register info from GUI and fill in data from HAL
            registers.append({'Device'      : self.deviceName,
                               'Name'        : self.newrUi.registerTable.item(row, colName).text(),
                               'Address'     : self.newrUi.registerTable.item(row, colAddr).text()
                              })
        self.addRegisters(registers, currentRow)

        # update the new register GUI
        self.newrUi.registerTable.clearSelection()
        self.actionAddEnable(False)
        self.mainUi.registerTable.blockSignals(False)

    def sortNewRegisterTable(self):
        """ Sort the new register GUI register table based on the column clicked. """
        col   = self.newrUi.registerTable.horizontalHeader().sortIndicatorSection()
        order = self.newrUi.registerTable.horizontalHeader().sortIndicatorOrder()
        self.newrUi.registerTable.sortItems(col, order)


###################################################################################################
# I N F O   U S E R   I N T E R F A C E
###################################################################################################

    def fillInfoTable(self):
        """ Fill the info GUI register table with data. """

        # first, remove old contents
        self.infoUi.infoTable.clearContents()

        # cached row was last selected register, so get info for this
        deviceName = self.rowCache['Device']
        regName    = self.rowCache['Name']
        regAddress = self.rowCache['Address']

        try:
            register = self.hal.getRegisterInfo(deviceName, regName, regAddress)
        except DataError:
            self.infoUi.close()
            return

        self.rowCache['Name'] = register.name
        self.rowCache['Address'] = register.address
        conversion = self.hal.getRegisterConversion(deviceName, None, None, register.name)

        # fill the info GUI
        self.fillInfoLine(0,  'Name'       , register.name)
        self.fillInfoLine(1,  'Address'    , "0x%08x" % register.address)
        self.fillInfoLine(2,  'Type'       , register.type)
        self.fillInfoLine(3,  'Unit'       , register.unit)
        self.fillInfoLine(4,  'Access'     , access2String(register.access))
        self.fillInfoLine(5,  'Width'      , register.width)
        self.fillInfoLine(6,  'Block'      , register.block)
        self.fillInfoLine(7,  'Max Offset' , register.maxOffset)
        self.fillInfoLine(8,  'Description', register.description)
        self.fillInfoLine(9,  'Conversion' , conversion)
        self.infoUi.infoTable.resizeRowsToContents()

    def fillInfoLine(self, row, label, content):
        """ Fill the contents of one table row with two QTableWidgetItem objects. """

        attribute = QTableWidgetItem()
        attribute.setText(label)
        setReadOnly(attribute, True)
        self.infoUi.infoTable.setItem(row, 0, attribute)
        value = QTableWidgetItem()
        value.setText(str(content))
        setReadOnly(value, True)
        self.infoUi.infoTable.setItem(row, 1, value)


###################################################################################################
# G U I  S E T T I N G S
###################################################################################################

    def writeSettings(self):
        """ write the main and new register GUI settings into Qsettings """

        logger.info("now in writeSettings")

        settings     = QSettings('SRON', 'SetGetManager')
        mainTableHdr = self.mainUi.registerTable.horizontalHeader()
        newrTableHdr = self.newrUi.registerTable.horizontalHeader()

        # get the mainUi column numbers
        mainColDevice      = self.regLabels.index('Device')
        mainColDeviceType  = self.regLabels.index('Device type')
        mainColName        = self.regLabels.index('Name')
        mainColAddress     = self.regLabels.index('Address')
        mainColOffset      = self.regLabels.index('Offset')
        mainColValue       = self.regLabels.index('Value')
        mainColUnit        = self.regLabels.index('Unit')
        mainColHexCheckBox = self.regLabels.index('Hex')
        mainColEngCheckBox = self.regLabels.index('Eng')
        mainColGetButton   = self.regLabels.index('Get')
        mainColSetButton   = self.regLabels.index('Set')

        # get the newrUi column numbers
        newrColName        = self.newRegLabels.index('Name')
        newrColAddress     = self.newRegLabels.index('Address')
        newrColType        = self.newRegLabels.index('Type')
        newrColUnit        = self.newRegLabels.index('Unit')
        newrColAccess      = self.newRegLabels.index('Access')
        newrColWidth       = self.newRegLabels.index('Width')
        newrColBlock       = self.newRegLabels.index('Block')
        newrColOffset      = self.newRegLabels.index('Offset')
        newrColDescript    = self.newRegLabels.index('Description')

        settings.setValue('mainWindowSize'     , self.mainUi.size())
        settings.setValue('mainWindowPos'      , self.mainUi.pos())
        settings.setValue('mainSortColumn'     , mainTableHdr.sortIndicatorSection())
        settings.setValue('mainSortOrder'      , mainTableHdr.sortIndicatorOrder())
        settings.setValue('mainSizeDevice'     , mainTableHdr.sectionSize(mainColDevice))
        settings.setValue('mainSizeDeviceType' , mainTableHdr.sectionSize(mainColDeviceType))
        settings.setValue('mainSizeName'       , mainTableHdr.sectionSize(mainColName))
        settings.setValue('mainSizeAddress'    , mainTableHdr.sectionSize(mainColAddress))
        settings.setValue('mainSizeOffset'     , mainTableHdr.sectionSize(mainColOffset))
        settings.setValue('mainSizeValue'      , mainTableHdr.sectionSize(mainColValue))
        settings.setValue('mainSizeUnit'       , mainTableHdr.sectionSize(mainColUnit))
        settings.setValue('mainSizeHex'        , mainTableHdr.sectionSize(mainColHexCheckBox))
        settings.setValue('mainSizeEng'        , mainTableHdr.sectionSize(mainColEngCheckBox))
        settings.setValue('mainSizeGet'        , mainTableHdr.sectionSize(mainColGetButton))
        settings.setValue('mainSizeSet'        , mainTableHdr.sectionSize(mainColSetButton))

        settings.setValue('newrWindowSize'     , self.newrUi.size())
        settings.setValue('newrWindowPos'      , self.newrUi.pos())
        settings.setValue('newrSortColumn'     , newrTableHdr.sortIndicatorSection())
        settings.setValue('newrSortOrder'      , newrTableHdr.sortIndicatorOrder())
        settings.setValue('newrSizeName'       , newrTableHdr.sectionSize(newrColName))
        settings.setValue('newrSizeAddress'    , newrTableHdr.sectionSize(newrColAddress))
        settings.setValue('newrSizeType'       , newrTableHdr.sectionSize(newrColType))
        settings.setValue('newrSizeUnit'       , newrTableHdr.sectionSize(newrColUnit))
        settings.setValue('newrSizeAccess'     , newrTableHdr.sectionSize(newrColAccess))
        settings.setValue('newrSizeWidth'      , newrTableHdr.sectionSize(newrColWidth))
        settings.setValue('newrSizeBlock'      , newrTableHdr.sectionSize(newrColBlock))
        settings.setValue('newrSizeOffset'     , newrTableHdr.sectionSize(newrColOffset))
        settings.setValue('newrSizeDescription', newrTableHdr.sectionSize(newrColDescript))

        settings.setValue('device', self.deviceName)

    def readSettings(self):
        """ read the main and new register GUI settings from Qsettings """

        logger.debug("now in readSettings")

        # get the mainUi column numbers
        mainColDevice      = self.regLabels.index('Device')
        mainColDeviceType  = self.regLabels.index('Device type')
        mainColName        = self.regLabels.index('Name')
        mainColAddress     = self.regLabels.index('Address')
        mainColOffset      = self.regLabels.index('Offset')
        mainColValue       = self.regLabels.index('Value')
        mainColUnit        = self.regLabels.index('Unit')
        mainColHexCheckBox = self.regLabels.index('Hex')
        mainColEngCheckBox = self.regLabels.index('Eng')
        mainColGetButton   = self.regLabels.index('Get')
        mainColSetButton   = self.regLabels.index('Set')

        # get the newrUi column numbers
        newrColName        = self.newRegLabels.index('Name')
        newrColAddress     = self.newRegLabels.index('Address')
        newrColType        = self.newRegLabels.index('Type')
        newrColUnit        = self.newRegLabels.index('Unit')
        newrColAccess      = self.newRegLabels.index('Access')
        newrColWidth       = self.newRegLabels.index('Width')
        newrColBlock       = self.newRegLabels.index('Block')
        newrColOffset      = self.newRegLabels.index('Offset')
        newrColDescript    = self.newRegLabels.index('Description')

        try:
            settings     = QSettings('SRON', 'SetGetManager')

            self.mainWindowSize = settings.value('mainWindowSize', QSize(1000, 1000))
            self.mainWindowPos  = settings.value('mainWindowPos', QPoint(0,0))

            self.mainSortColumn = int(settings.value('mainSortColumn', mainColDevice))
            self.mainSortOrder  = int(settings.value('mainSortOrder', Qt.AscendingOrder))
            self.mainColumnWidth[mainColDevice     ] = int(settings.value('mainSizeDevice',    100))
            self.mainColumnWidth[mainColDeviceType ] = int(settings.value('mainSizeDeviceType',100))
            self.mainColumnWidth[mainColName       ] = int(settings.value('mainSizeName',      200))
            self.mainColumnWidth[mainColAddress    ] = int(settings.value('mainSizeAddress',   100))
            self.mainColumnWidth[mainColOffset     ] = int(settings.value('mainSizeOffset',    100))
            self.mainColumnWidth[mainColValue      ] = int(settings.value('mainSizeValue',     100))
            self.mainColumnWidth[mainColUnit       ] = int(settings.value('mainSizeUnit',       50))
            self.mainColumnWidth[mainColHexCheckBox] = int(settings.value('mainSizeHex',        50))
            self.mainColumnWidth[mainColEngCheckBox] = int(settings.value('mainSizeEng',        50))
            self.mainColumnWidth[mainColGetButton  ] = int(settings.value('mainSizeGet',        50))
            self.mainColumnWidth[mainColSetButton  ] = int(settings.value('mainSizeSet',        50))

            self.newrWindowSize = settings.value('newrWindowSize', QSize(1000, 1000))
            self.newrWindowPos  = settings.value('newrWindowPos', QPoint(0,0))

            self.newrSortColumn = int(settings.value('newrSortColumn', newrColName))
            self.newrSortOrder  = int(settings.value('newrSortOrder', Qt.AscendingOrder))
            self.newrColumnWidth[newrColName    ] = int(settings.value('newrSizeName',       200))
            self.newrColumnWidth[newrColAddress ] = int(settings.value('newrSizeAddress',     95))
            self.newrColumnWidth[newrColType    ] = int(settings.value('newrSizeType',        50))
            self.newrColumnWidth[newrColUnit    ] = int(settings.value('newrSizeUnit',        50))
            self.newrColumnWidth[newrColAccess  ] = int(settings.value('newrSizeAccess',      60))
            self.newrColumnWidth[newrColWidth   ] = int(settings.value('newrSizeWidth',       50))
            self.newrColumnWidth[newrColBlock   ] = int(settings.value('newrSizeBlock',       50))
            self.newrColumnWidth[newrColOffset  ] = int(settings.value('newrSizeOffset',      55))
            self.newrColumnWidth[newrColDescript] = int(settings.value('newrSizeDescription',285))

            self.deviceName = settings.value('device')
        except:
            logger.warning("readSettings exception occurred")


###################################################################################################
# R E G I S T E R  S E T T I N G S
###################################################################################################

    def _readYaml(self, settingsFile):
        """ Read the register settings from a YAML file. """
        settings = []

        try:
            stream = open(settingsFile, 'r')
        except OSError:
            logger.warning('Could not read the Register settings')
            return settings

        logger.info('Reading the Register settings from: %s', settingsFile)
        try:
            settings = yaml.load(stream, Loader=yaml.SafeLoader)
        except yaml.YAMLError:
            logger.warning('Invalid Register settings file: %s', settingsFile)

        return settings

    def _yamlToRegList(self, settings):
        """ Convert YAML dict to register list. """
        registerList = []

        if not settings or settings['setget']['registers'] is None:
            return registerList

        for item in settings['setget']['registers']:

            # Old style YAML - convert to newer style
            if 'device' in item:
                logger.warning("Old YAML format detected, attempting to read data")
                register = {col:item[col.lower()] for col in self.regLabels if col.lower() in item}
            # New style YAML, continue
            else:
                register = item

            # load register information from HAL
            register = self.updateRegisterValues(register)

            if register is not None:
                registerList.append(register)

        return registerList

    def _regListToYaml(self, saveData=False):
        """ Convert the actual register settings to a yaml dict.  """

        registers = []
        logger.info("rows = %s", self.mainUi.registerTable.rowCount())
        for row in range(self.mainUi.registerTable.rowCount() - 1):
            rowData = self.getTableRow(row)
            # only save relevant info if storing to file
            if not saveData:
                register = {col: rowData[col] for col in self.columnsToStore}
            else:
                register = rowData
            registers.append(register)

        # create entire yaml model
        settings = {'setget' : { 'registers' : registers } }
        return settings

    def _saveYaml(self, settingsFile, settings):
        """ Store the settings to a YAML file. """
        with open(settingsFile, "w") as outfile:
            # do not use "default_flow_style" in the dump command below:
            # this changes the lists in the dict, afterwhich these lists cannot be
            # read anymore
            yaml.dump(settings, outfile)
            logger.info('Writing the Register settings to: %s', settingsFile)
        outfile.close()

    def _checkSave(self):
        """ Check if configuration has changed before open or close. """

        if self.mainUi.actionBack.isEnabled():
            messageBox = gsswUiMessageBox_("There are unsaved changes.\n"
                                           "Would you like save them?")
            messageBox.setWindowTitle("Are you sure?")
            messageBox.setIcon(QMessageBox.Question)
            messageBox.setStandardButtons(QMessageBox.Save | QMessageBox.Discard |
                                          QMessageBox.Cancel)
            choice = messageBox.exec_()
            if choice == QMessageBox.Save:
                self.slotSave()
            return choice != QMessageBox.Cancel
        return True

    def publishValue(self, name, value, units=None):
        """ Publish the value with a timestamp set to time of method call.

        :param name: a unique identifier for the published value (will be prepended with SG_)
        :param value: the value to be published
        """
        if self.pubSocket is None:
            return

        gArray = GriddedArray([value], name="SG_{}".format(name), units=units)

        timestamp = np.array([np.datetime64(datetime.now())], dtype='datetime64[ms]')
        gArray.coords[0] = SemanticArray(timestamp, name='x', label='Time', units='s')

        # create protobuffer and publish
        msg = gArray.asProtoBuf()
        self.pubSocket.publishProtoBuf("/sg/{}".format(name), tagName(msg), msg.SerializeToString())

        logger.debug("Publish %s = %s", gArray.name, gArray.data)

###################################################################################################
# S T A C K  H A N D L I N G
###################################################################################################

    def stackPush(self, settings):
        """ Push the current register settings on the configuration stack. """
        self.settingsStack.append(settings)
        if len(self.settingsStack) == self.STACKSIZE+1:
            self.settingsStack.pop(0)

        # enable the back action
        self.actionBackEnable(True)

    def stackPull(self):
        """ Pull the previous register settings from the configuration stack. """

        if self.settingsStack:
            settings = self.settingsStack.pop()
        else:
            settings = None

        # disable the back action if stack empty
        if self.settingsStack:
            self.actionBackEnable(True)
        else:
            self.actionBackEnable(False)

        return settings


###################################################################################################
# E N A B L E S
###################################################################################################

    def actionCutEnable(self, enable=True):
        """ Change enable state of main GUI 'Cut' action. """
        self.mainUi.actionCut.setEnabled(enable)

    def actionPasteEnable(self, enable=True):
        """ Change enable state of main GUI 'Paste' action. """
        self.mainUi.actionPaste.setEnabled(enable)

    def actionBackEnable(self, enable=True):
        """ Change enable state of main GUI 'Back' action. """
        self.mainUi.actionBack.setEnabled(enable)

    def actionInfoShowEnable(self, enable=True):
        """ Change enable state of main GUI 'Info' action. """
        self.mainUi.actionInfoShow.setEnabled(enable)

    def actionAddEnable(self, enable=True):
        """ Change enable state of new register GUI 'Add' action. """
        self.actionAdd.setEnabled(enable)


###################################################################################################
# S L O T S
###################################################################################################

    def slotOpen(self):
        """ Slot (main GUI): called when 'Open' is clicked. """
        if self._checkSave():
            fileDialog = QFileDialog(directory=os.path.dirname(self.settingsFile))
            fileDialog.setAcceptMode(QFileDialog.AcceptOpen)
            fileDialog.setNameFilter("YAML files (*.yaml)")
            if fileDialog.exec_():
                fileNames = fileDialog.selectedFiles()
                settings = self._readYaml(fileNames[0])
                registerList = self._yamlToRegList(settings)
                self.fillRegisterTable(registerList)

    def slotSave(self):
        """ Slot (main GUI): called when 'Save' is clicked. """
        settings = self._regListToYaml(saveData=False)
        self._saveYaml(self.settingsFile, settings)

    def slotSaveClose(self):
        """ Save and close """
        logger.info("Save and close")
        self.writeSettings()
        self.slotSave()
        sys.exit(1)

    def slotSaveAs(self):
        """ Slot (main GUI): called when 'Save as' is clicked. """
        fileDialog = QFileDialog(directory=os.path.dirname(self.settingsFile))
        fileDialog.setAcceptMode(QFileDialog.AcceptSave)
        fileDialog.setNameFilter("YAML files (*.yaml)")
        if fileDialog.exec_():
            fileNames = fileDialog.selectedFiles()
            fileName = fileNames[0]
            if not fileName.endswith('.yaml'):
                fileName += '.yaml' # Add .yaml suffix if the users did not enter it themselves
            settings = self._regListToYaml(saveData=False)
            self._saveYaml(fileName, settings)

    def slotCloseMainUi(self):
        """ Slot (main GUI): called when 'Close' is clicked. """
        self.mainUi.close() # Will be caught by self.closeEvent (so can close other windows)

    def slotNewReg(self):
        """ Slot (main GUI): called when 'New' is clicked.  """
        # call new register GUI with show (non-blocking) rather than exec (blocking)
        self.newrUi.show()
        self.newrUi.raise_() # Make sure window pops up (again, if it was minimized)

    def slotCut(self):
        """ Slot (main GUI): called when 'Cut' is clicked. """
        # cut (i.e. delete and store) the selected register(s)
        self.deleteRegister()

    def slotPaste(self):
        """ Slot (main GUI): called when 'Paste' is clicked. """
        # paste the previously deleted register(s)
        self.pasteRegister()

    def slotBack(self):
        """ Slot (main GUI): called when 'Back' is clicked. """
        # restore most recent main GUI register table
        registerList = self._yamlToRegList(self.stackPull())
        self.fillRegisterTable(registerList)

    def slotInfoShow(self):
        """ Slot (main GUI): called when 'Show' is clicked. """
        # present all HAL data of selected register via a pop-up GUI
        self.fillInfoTable()
        # call info GUI with show (non-blocking) rather than exec (blocking)
        self.infoUi.show()

    def slotGetAll(self):
        """ slot (main GUI): called when 'Get all' is clicked. """
        mainColGetButton   = self.regLabels.index('Get')
        for idx in range(0,self.mainUi.registerTable.rowCount()):
            self.mainUi.registerTable.setCurrentCell(idx, mainColGetButton)
            if not self.getButtonAction():
                break

    def slotMainUiContextMenu(self, position):
        """ Slot (main GUI): called when the context menu is requested. """
        self.mainUiContextMenu.exec_(self.mainUi.registerTable.mapToGlobal(position))

    def slotUpdateMainUi(self):
        """ Slot (main GUI): called when at least one row is selected.  """
        selectedRows = {idx.row() for idx in self.mainUi.registerTable.selectedIndexes()}

        try:
            # if dummy row in selection, deselect
            selectedRows.remove(self.mainUi.registerTable.rowCount() - 1)
            self.mainUi.registerTable.clearSelection()
            for row in selectedRows:
                self.mainUi.registerTable.selectRow(row)
        except KeyError:
            pass    # dummy row was not in the selection, can continue

        # update enable state of main GUI 'Cut' and 'Show' actions
        if selectedRows:
            self.actionCutEnable(True)
            self.actionInfoShowEnable(True)
        else:
            self.actionCutEnable(False)
            self.actionInfoShowEnable(False)

    def slotCacheItemData(self, item):
        """ Slot: Store current cell data on user click. """
        # Cache row data except dummy row (last table row) so can return to on invalid user entry
        if 0 <= item.row() < self.mainUi.registerTable.rowCount() - 1:
            self.rowCache = self.getTableRow(item.row())
        logger.debug("slotCacheItemData: %s", self.rowCache)

    def slotUpdateRegister(self, item):
        """ Slot: Update function for when a register changes value. """
        row = item.row()
        col = item.column()

        if col == self.colOffset:
            value = self.getCellValue(row, self.colOffset)
            self.updateOffset(row, value)
            self.setValueStatus(row, "NEW")
        elif col == self.colValue:
            value = self.getValue(row)
            self.updateValue(row, value)
            self.setValueStatus(row, "NEW")
        elif col == self.colName:
            # Get Device and Register from table row
            rowData = self.getTableRow(row)
            register = {'Device'    : rowData['Device'],
                        'Name'      : rowData['Name'],
                        'Address'   : None}            # Only want lookup by name, not old address

            # Add register, if not successful then return to cached text
            if not self.addRegisters([register], row, overwrite=True):
                self.mainUi.registerTable.blockSignals(True)
                item.setText(self.rowCache['Name'])
                self.mainUi.registerTable.blockSignals(False)
            else:
                self.setValueStatus(row, "NEW")

        # self.mainUi.registerTable.blockSignals(True)
        # self.setValueStatus(row, "NEW")
        # self.mainUi.registerTable.blockSignals(False)

    def slotSortMainUi(self):
        """ Slot (main GUI): called when a sort indicator has changed. """
        col   = self.mainUi.registerTable.horizontalHeader().sortIndicatorSection()
        order = self.mainUi.registerTable.horizontalHeader().sortIndicatorOrder()

        # Remove dummy row
        self.mainUi.registerTable.removeRow(self.mainUi.registerTable.rowCount() - 1)
        self.mainUi.registerTable.sortItems(col, order)

        # Add in dummy row
        self.mainUi.registerTable.blockSignals(True)
        self.fillDummyRegister()
        self.mainUi.registerTable.blockSignals(False)

    def slotReloadHal(self):
        """ Slot (main GUI): called when the HAL client fires a signal """
        # reload the HAL
        logger.info("slotReloadHal requested by SetGetManager")
        self.halClient.requestHal()
        settings = self._regListToYaml(saveData=True)
        registerList = self._yamlToRegList(settings)
        self.createDevices()
        self.fillDeviceComboBox()
        self.fillRegisterTable(registerList)
        self.fillNewRegisterTable()
        if self.rowCache is not None:
            self.fillInfoTable()

    def slotDeviceChanged(self, index):
        """ Slot (new register GUI): called when the 'Device' combobox has changed. """
        # re-draw the new register GUI register table provided that the combo box is not
        # empty (clearing the combo box fires the currentIndexChanged signal as well!)
        if index != -1:
            self.deviceName = self.newrUi.device.currentText()
            self.fillNewRegisterTable()
            self.writeSettings()

    def slotAdd(self):
        """ Slot (new register GUI): called when 'Add' is clicked. """
        # add selected register(s) to main GUI register table
        self.addNewRegisters()

    def slotNewrUiContextMenu(self, position):
        """ Slot (new register GUI): called when the context menu is requested. """
        self.newrUiContextMenu.exec_(self.newrUi.registerTable.mapToGlobal(position))

    def slotUpdateNewrUi(self):
        """ Slot (new register GUI): called when at least one row is selected.  """
        indexes = []
        for idx in self.newrUi.registerTable.selectedIndexes():
            indexes.append(idx.row())
        selectedRows = list(set(indexes))

        # update enable state of new register GUI 'Add' action
        if not selectedRows:
            self.actionAddEnable(False)
        else:
            self.actionAddEnable(True)

    def slotSortNewrUi(self):
        """ Slot (new register GUI): called when a sort indicator has changed. """
        # sort the new register table
        self.sortNewRegisterTable()
        self.writeSettings()

    def slotExportToClipboard(self):
        """ Export data to clipboard. (actionExToClipboard) """
        strInfo = self.getExportData()
        maxChars = 2000 #Set max (2000) to prevent data from flooding the box/screen.

        # Copy to clip board
        clipboard = QGuiApplication.clipboard()
        clipboard.setText(strInfo)

        gsswUiInfo(
            "The data has been exported to the clipboard:<br />"
            "<pre>{}{}</pre>".format(strInfo[:maxChars], " ..." if len(strInfo) > maxChars else ""))

    def slotExportToFile(self):
        """ Export data to txt/csv file. (actionExToFile) """
        strInfo = self.getExportData()

        #Check if there is data to save.
        if not strInfo:
            gsswUiWarning("Data string empty. Export canceled.")
            return

        #Getting filename from user.
        fileName = promptForFileName(self, 'export', 'csv',
                                     directory='', suggestDateTime=True,
                                     fileFilter="CSV files (*.csv);; All files (*)")
        if not fileName:
            logger.info("A Filename is required. Export canceled")
            return

        try:
            #Saving to file
            with open(fileName, 'w') as openFile:
                logger.debug("Exporting data to file: %s", fileName)
                openFile.write(strInfo)
        except Exception as ex:
            gsswUiWarning(str(ex))

    def getExportData(self):
        """ Gets the data to be exported (to file or clipboard). """

        # Add header
        tableData = [','.join(self.columnsToExport)]

        # Add row data
        for rowInd in range(self.mainUi.registerTable.rowCount() - 1):
            try:
                rowData = self.getTableRow(rowInd)
                rowStr = formatRowData(rowData, self.columnsToExport)
            except Exception as ex:
                # Ignore row data
                logger.debug("Error when trying to access rowdata, %s", ex)
            else:
                tableData.append(rowStr)

        return '\n'.join(tableData)

###################################################################################################
# C L O S I N G
###################################################################################################

    def closeEvent(self, event):
        """ Close the window """

        if self._checkSave():
            self.writeSettings()
            self.newrUi.close()
            self.infoUi.close()
            event.accept()
        else:
            event.ignore()
