'''
Created on Aug 28, 2019

@author: simonr
'''
from   datetime                 import datetime
import                          logging
import                          os
import                          textwrap

from PyQt5                      import uic
from PyQt5.QtCore               import pyqtSlot, pyqtSignal, Qt
from PyQt5.QtGui                import QGuiApplication, QIcon
from PyQt5.QtWidgets            import QDialog, QTableWidgetItem, QPushButton, QHeaderView

from gssw.proto.diagnostics_pb2 import DiagMsg
from gssw.lib.errors            import GsswError
from gssw.lib.misc              import pbEnumName
from gssw.lib.sockets           import tagName
from gssw.lib.socketsqt         import NotifyingSubSocket
from gssw.lib.widgets.misc      import gsswShowError, gsswUiInfo

logger = logging.getLogger(__name__)


#: Maximum number of messages kept in the diagnostics window. 
DIAG_MAX_MESSAGES = 25

class DiagnosticsHandler(QDialog):
    """ Window which receives and displays diagnostic user messages """

    #: New message reveiced (severity, title, details)
    sigNewMessage = pyqtSignal(int, str, str)

    def __init__(self, config):
        """
        :param Config config: Egress configuration object.
        """
        super().__init__()

        uiDir = os.path.dirname(__file__)
        uiFile = os.path.join(uiDir, 'gui/DiagWindow.ui')

        try:
            uic.loadUi(uiFile, self)
        except FileNotFoundError as ex:
            logger.exception("Could not load ui: %s", ex)
            return
        self.tableWidget.verticalHeader().setVisible(False)
        self.tableWidget.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.tableWidget.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        header = self.tableWidget.horizontalHeader()
        header.setCascadingSectionResizes(True)
        header.setSectionResizeMode(0, QHeaderView.Interactive)
        header.resizeSection(0, 170)
        header.setSectionResizeMode(1, QHeaderView.ResizeToContents)
        #Icon
        header.resizeSection(2, 60)
        header.setSectionResizeMode(2, QHeaderView.Fixed)
        header.setSectionResizeMode(3, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(4, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(5, QHeaderView.Stretch)
        # Copy button
        header.resizeSection(6, 45)
        header.setSectionResizeMode(6, QHeaderView.Fixed)
        # Clear button
        header.resizeSection(7, 45)
        header.setSectionResizeMode(7, QHeaderView.Fixed)

        # Set up receiving socket for user diagnostic events
        self.diagMsgs = []
        self.userDiagSocket = NotifyingSubSocket()
        self.userDiagSocket.createZmqSocket()
        self.userDiagSocket.subscribe('/diagnostics/')

        try:
            proxyConfig         = config.componentParameters('proxy')
            proxyHost           = proxyConfig['hostname']
            proxyPubport        = proxyConfig['sockets']['xpub']
            self.userDiagSocket.connectSocket("tcp://{}:{}".format(proxyHost, proxyPubport))
        except GsswError as ex:
            # In case of errors, e.g. when there is no 'proxy' defined
            # log a message in the diagnostics window
            logger.exception("Diagnostic logging could not be initialized: %s", ex)
            noInit = DiagMsg()
            noInit.timestamp = datetime.now().timestamp()
            noInit.severity = DiagMsg.WARNING
            noInit.title = 'Diagnostic logging is unavailable'
            noInit.details = "Diagnostic logging could not be initialized: {}".format(ex)
            noInit.componentname = 'traymonitor'
            self.diagMsgs.append(noInit)
            self.updateOverviewTable()
        self.userDiagSocket.sigReadyReceive.connect(self.slotReceiveUserDiagnostics)
        self.btnRemoveAll.clicked.connect(self.slotRemoveAllMsg)

        # Add dummy messages for test
#         for i in range(4):
#             x = DiagMsg()
#             x.timestamp = time.time()
#             x.severity = i
#             x.title = 'Short {}'.format(i)
#             self.diagMsgs.append(x)
#         self.updateOverviewTable()


    @pyqtSlot()
    @gsswShowError
    def slotReceiveUserDiagnostics(self):
        """ Process User diagnostic messages received via the SubSocket """
        _, tag, msg = self.userDiagSocket.receiveProtoBufWithTopic()
        if tagName(DiagMsg) != tag:
            logger.error('Ignoring unsupported user diag message: %s', tag)
            return

        diagMsg = DiagMsg()
        diagMsg.ParseFromString(msg)
        self.diagMsgs.append(diagMsg)

        # Slice off old messages, to avoid flooding the log
        del self.diagMsgs[:-DIAG_MAX_MESSAGES]

        severity = diagMsg.severity
        emitTitle = diagMsg.componentname if diagMsg.componentname else diagMsg.sourceInfo
        logger.info("Received message: %s, %s", emitTitle, diagMsg.title)

        self.updateOverviewTable()
        self.sigNewMessage.emit(severity, emitTitle, diagMsg.title)

    @staticmethod
    def formatTime(timestamp):
        """ Convert epoch timestamp into human readable format. """
        return datetime.fromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]

    @gsswShowError
    def updateOverviewTable(self):
        """ Refreshes the table with events """

        table = self.tableWidget
        table.setRowCount(0)

        levelIconNameMap = {DiagMsg.INFO:    'dialog-information',
                            DiagMsg.WARNING: 'dialog-warning',
                            DiagMsg.ERROR:   'dialog-error'}

        for msgNum, diagMsg in enumerate(self.diagMsgs):
            table.insertRow(msgNum)

            table.setItem(msgNum, 0, QTableWidgetItem(self.formatTime(diagMsg.timestamp)))
            table.setItem(msgNum, 1, QTableWidgetItem(diagMsg.title))

            logLevel = diagMsg.severity if diagMsg.severity in range(4) else 0
            icon = QIcon.fromTheme(levelIconNameMap.get(logLevel, 'dialog-question'))
            logLevelItem = QTableWidgetItem('')
            #table.setCellWidget(msgNum, 2, icon)
            logLevelItem.setIcon(icon)
            logLevelItem.setFlags(logLevelItem.flags() ^ Qt.ItemIsEditable)
            table.setItem(msgNum, 2, logLevelItem)
            table.setItem(msgNum, 3, QTableWidgetItem(diagMsg.componentname))
            table.setItem(msgNum, 4, QTableWidgetItem(diagMsg.details))
            table.setItem(msgNum, 5, QTableWidgetItem(diagMsg.sourceinfo))

            # Add button for copying message
            copyMsgButton = QPushButton()
            copyMsgButton.setIcon(QIcon.fromTheme('edit-copy'))
            copyMsgButton.setFlat(True)
            copyMsgButton.setToolTip("Copy this message to the clipboard")
            copyMsgButton.clicked.connect(self.slotCopyMsg)
            table.setCellWidget(msgNum,6, copyMsgButton)

            # Add button for deleting message
            removeMsgButton = QPushButton()
            removeMsgButton.setIcon(QIcon.fromTheme('window-close'))
            removeMsgButton.setFlat(True)
            removeMsgButton.setToolTip("Delete this message from the list")
            removeMsgButton.clicked.connect(self.slotRemoveMsg)
            table.setCellWidget(msgNum,7, removeMsgButton)

            # Make columns read-only
            for col in range(8):
                readOnlyItem = table.item(msgNum, col)
                if readOnlyItem:
                    readOnlyItem.setFlags(readOnlyItem.flags() & ~Qt.ItemIsEditable)

    @pyqtSlot()
    def slotCopyMsg(self):
        """ Copies the selected message to the clip board """
        button = self.sender()
        index = self.tableWidget.indexAt(button.pos())
        try:
            # Compile string with all relevant info to copy into clip board
            diagMsg = self.diagMsgs[index.row()]
            # Add time in human readable format (non-epoch)
            strInfo = 'timestamp:     {}\n'.format(self.formatTime(diagMsg.timestamp))
            strInfo += 'severity:      {}\n'.format(pbEnumName(diagMsg.Icon, diagMsg.severity))

            # Format protobuf message
            for field, value in diagMsg.ListFields():
                if field.name == 'details' and len(value) > 40:
                    # Word-wrap longer texts, otherwise they will be invisible
                    strInfo += textwrap.fill(value, 60,
                                             initial_indent="{:<14} ".format(field.name + ':'),
                                             subsequent_indent=' ' * 15)
                    strInfo += '\n'
                elif field.name not in ['timestamp', 'severity']:
                    strInfo += "{:<14} {}\n".format(field.name + ":", value)

            # Copy to clip board
            clipboard = QGuiApplication.clipboard()
            clipboard.setText(strInfo)
            gsswUiInfo(
                "The message details have been copied into the clipboard:<br />"
                "<pre>{}</pre>".format(strInfo))
        except IndexError:
            # Somehow, gone from the list
            pass

    @pyqtSlot()
    def slotRemoveMsg(self):
        """ Removes the selected message from the overview """
        button = self.sender()
        index = self.tableWidget.indexAt(button.pos())
        try:
            self.diagMsgs.pop(index.row())
        except IndexError:
            # Somehow, already gone from the list
            pass
        self.updateOverviewTable()

    @pyqtSlot()
    def slotRemoveAllMsg(self):
        """ Removes all messages from the overview """
        self.diagMsgs.clear()
        self.updateOverviewTable()

    @pyqtSlot()
    def openWindow(self):
        """ Displays (and puts window in front) the overview window with the diagnostic messages """
        self.show()
        self.activateWindow()
        self.raise_()
