#!/usr/bin/env python3
""" This is a GSSw monitoring utility which runs as a system-tray application.
"""

import getpass
import logging
import os
import socket
import sys
import threading
import webbrowser

# Import Qt5 classes
#pylint: disable=no-name-in-module
from PyQt5.Qt                 import QIcon, QMessageBox, QMenu, QMovie, QSystemTrayIcon
from PyQt5.QtCore             import pyqtSignal, pyqtSlot, QFileSystemWatcher, QObject, QThread
from PyQt5.QtGui              import QKeySequence
from PyQt5.QtWidgets          import QWidgetAction, QLabel
from gssw.bin.gssw_run        import RunGSSw
from gssw.common              import initGssw
from gssw.lib.application     import Application
from gssw.lib.widgets.misc    import gsswUiError
from gssw.traymonitor.splash  import TraySplashScreen
from gssw.traymonitor.diagnostics import DiagnosticsHandler


logger = logging.getLogger(__name__)

APP_NAME = "Egress monitor"
TOOLTIP_TIMEOUT = 2500
MODULE_PATH = os.path.dirname(os.path.abspath(__file__))

ICONS = {}

class FileOpener():
    """ Wrapper for opening files in the system-default editor """
    @classmethod
    def open(cls, file):
        """
        Due to a bug in xdg-open (used by webbrowser.open), the command must be sent twice.
        Otherwise the file doesn't actually open
        """

        if os.path.isfile(file):
            # Open file
            # Notice: this might not work when gedit is used as a default file editor (xdg-open bug)
            cls._open(file)
        else:
            logger.error("File '%s' doesn't exist", file)
            gsswUiError("The file '{}' doesn't exist.".format(file))

    @classmethod
    def _open(cls, file):
        return webbrowser.open(url="file://" + file, new=1, autoraise=True)

class ComponentEntry(QMenu):
    """ Menu entry for a single Gssw component """

    # Signal-slot is used for updating the GUI from other threads
    _updateTrigger   = pyqtSignal()

    # External signals
    sigStartSingle   = pyqtSignal(str)
    sigStopSingle    = pyqtSignal(str)
    sigRestartSingle = pyqtSignal(str)

    def __init__(self, name, enabled, app):
        assert isinstance(app, App)

        self.name = name
        super().__init__(self.name)
        self.enabled = enabled
        self.run_gssw = app.run_gssw
        self.config = app.config

        icon = ICONS['disabled']
        self.setIcon(icon)

        # Show the component's description in the menu
        entryConfig = self.config.componentParameters(self.name)
        entryDescription = entryConfig.get('description')
        if entryDescription:
            # The description can be a long text
            # Normally, I would use self.addSection("description") here
            # However, the long text is weirdly cut off
            # Instead create a QAction with a custom widget in it (the QLabel)
            # The QLabel is nicely formatted
            descAction = QWidgetAction(self)
            descLabel = QLabel(entryDescription)
            descLabel.setMargin(6)
            descLabel.setWordWrap(True)
            descAction.setDefaultWidget(descLabel)
            descAction.setEnabled(False) # Disable the action so it cannot be clicked
            self.addAction(descAction)
            self.addSeparator() # For line in menu below description

        # Sub-menu items, which will be modified when the status changes
        self.actionReStart = self.addAction("S&tart") # Start / Restart
        self.actionStop    = self.addAction("St&op")
        self.actionLog     = self.addAction("View &log")

        # Add optional menu items (specific to some components)
        if 'filename' in entryConfig: # this is the case for the log-server
            self.actionEventLog = self.addAction("&View central Egress event log")
            self.actionEventLog.triggered.connect(self.viewEventLog)

        self.actionStop.triggered.connect(self.stop)
        self.actionLog.triggered.connect(self.viewLog)

        self.updateStatus()
        # Connect the trigger for updating the status from other threads
        self._updateTrigger.connect(self.updateStatus)

    def updateStatus(self):
        """ Retrieves the actual status of the current Gssw component, and updates the tray menu """
        is_running = self.isComponentRunning() 
        if is_running:
            icon = ICONS['ok']
        else:
            if self.enabled:
                icon = ICONS['error']
            else:
                icon = ICONS['disabled']
        logger.debug("Updating menu item of '%s' (Running: %s)", self.name, is_running)

        self.setIcon(icon)
        self.actionReStart.triggered.disconnect() # Disconnect old (Re)start action
        # Update sub-menu options
        if is_running:
            self.actionReStart.setText("&Restart")
            self.actionReStart.triggered.connect(self.restart)
            self.actionStop.setEnabled(True)
        else:
            if self.enabled:
                self.actionReStart.setText("S&tart")
                self.actionReStart.triggered.connect(self.start)
            else:
                self.actionReStart.setText("S&tart manually")
                self.actionReStart.triggered.connect(self.start)
            self.actionStop.setEnabled(False)

    def setWaiting(self):
        """ Sets the menu item's icon to the waiting icon. """
        self.setIcon(ICONS['waiting'])

    def triggerUpdateStatus(self):
        """ Send a signal to the GUI thread to update its menu """
        self._updateTrigger.emit()

    def isComponentRunning(self):
        """ Checks whether the current Gssw component is running """
        component_parameters = self.config.componentParameters(self.name)
        return self.run_gssw.cmdIsRunning(self.name, component_parameters)

    def viewLog(self):
        """ Opens the current Gssw component's log file in the default editor """
        parameters = self.config.componentParameters(self.name)
        log_path = self.run_gssw.logFileName(self.name, parameters)
        logger.info("Opening component log file: %s", log_path)
        FileOpener.open(log_path)

    def viewEventLog(self):
        """ Opens the Gssw central event log file in the default editor """
        parameters = self.config.componentParameters(self.name)
        eventLogFile = parameters['filename']
        eventLog = self.config.asAbsolutePath(eventLogFile)
        logger.info("Opening Egress event log file: %s (%s)", eventLog, eventLogFile)
        FileOpener.open(eventLog)

    def restart(self):
        """ Restarts the current Gssw component """
        self.sigRestartSingle.emit(self.name)

    def start(self):
        """ Starts the current Gssw component """
        self.sigStartSingle.emit(self.name)

    def stop(self):
        """ Stops the current Gssw component """
        self.sigStopSingle.emit(self.name)


class StatusUpdater(QThread):
    """ Threaded timer class which updates the status of the Gssw components in the background """
    def __init__(self, callback_function):
        super().__init__()
        self.updateInterval = 1
        self._stopEvent = threading.Event()
        self.callback_function = callback_function
        self._threadLock = threading.Lock()
        self._paused     = False

    def run(self):
        """ Main thread loop for updating component status periodically. """
        while True:
            self._stopEvent.wait(timeout=self.updateInterval)
            if self._stopEvent.is_set():
                break
            with self._threadLock:
                if not self._paused:
                    self.callback_function()

    def stop(self):
        """ Stops the background updater (blocking) """
        # Set the stop flag
        self._stopEvent.set()
        logger.info("Stopping status updater thread")
        self.wait()

    @pyqtSlot()
    def pause(self):
        """ Suspends the background updater """
        with self._threadLock:
            self._paused = True

    @pyqtSlot()
    def resume(self):
        """ Resumes the background updater """
        with self._threadLock:
            self._paused = False


class AppController(QObject):
    """ Actual actions are put in this object, which is run in a separate thread.
        This is required in order to make the gui responsive during blocking calls. """

    sigActionStarted = pyqtSignal()
    sigActionStopped = pyqtSignal()

    def __init__(self, app):
        super().__init__()
        self.app = app

    @pyqtSlot()
    def startAllEnabled(self):
        """ Start all Gssw components that are enabled in the configuration file """
        self.sigActionStarted.emit()
        for _, componentEntry in self.app.component_actions.items():
            componentEntry['menu'].setWaiting()
        self.app.run_gssw.execute('start')
        self.sigActionStopped.emit()

    @pyqtSlot()
    def stopAll(self):
        """ Stops all running Gssw components """
        self.sigActionStarted.emit()
        for _, componentEntry in self.app.component_actions.items():
            componentEntry['menu'].setWaiting()
        self.app.run_gssw.execute('stop')
        self.sigActionStopped.emit()

    @pyqtSlot(str)
    def startSingle(self, componentName):
        """ Start a single Gssw component """
        self.sigActionStarted.emit()
        self.app.run_gssw.execute('start', [componentName])
        self.sigActionStopped.emit()


    @pyqtSlot(str)
    def stopSingle(self, componentName):
        """ Stops a single Gssw component """
        self.sigActionStarted.emit()
        self.app.run_gssw.execute('stop', [componentName])
        self.sigActionStopped.emit()

    @pyqtSlot(str)
    def restartSingle(self, componentName):
        """ Restart a single Gssw component """
        self.sigActionStarted.emit()
        self.app.run_gssw.execute('restart', [componentName])
        self.sigActionStopped.emit()


class App(QObject):
    """ Main class of tray monitoring utiliy """

    def __init__(self, args):  #pylint: disable=too-many-statements
        super().__init__()
        self.config_path = args.configPath # We must store the path for monitoring file changes
        # We must store the args because initGssw requires them upon reloading
        self.args        = args
        self.config = initGssw(self.args)

        # Create background thread which monitors the status of the gssw components
        self._backgroundStatusUpdater = StatusUpdater(self.updateComponentStatuses)

        # Create a Qt application (wih Egress stylesheet)
        self.app = Application(self.config, name=APP_NAME)
        # Prevent the tray application from closing when popup-dialogs are closed
        self.app.setQuitOnLastWindowClosed(False)

        # Show the Egress splash screen
        if sys.platform != 'darwin' and not args.nosplash: # Splash screen doesn't work on mac OSX
            splash = TraySplashScreen()
            splash.start(3.0)

        ICONS['tray']      = QIcon(MODULE_PATH + "/images/gssw_traymonitor.png")
        ICONS['error']     = QIcon(MODULE_PATH + "/images/icon_red.png")
        ICONS['ok']        = QIcon(MODULE_PATH + "/images/icon_green.png")
        ICONS['disabled']  = QIcon(MODULE_PATH + "/images/icon_grey.png")
        self.movie_waiting = QMovie(MODULE_PATH + "/images/icon_waiting_green.gif")
        ICONS['waiting']   = QIcon(self.movie_waiting.currentPixmap())

        self.controller = AppController(self)
        self.controllerThread = QThread()
        self.controller.moveToThread(self.controllerThread)
        self.controllerThread.start() # Start separate event loop in controller thread

        self.controller.sigActionStarted.connect(self.lockMenu)
        self.controller.sigActionStarted.connect(self.startAnimation)
        self.controller.sigActionStarted.connect(self._backgroundStatusUpdater.pause)

        self.controller.sigActionStopped.connect(self.stopAnimation)
        self.controller.sigActionStopped.connect(self._backgroundStatusUpdater.resume)
        self.controller.sigActionStopped.connect(self.reload)
        self.controller.sigActionStopped.connect(self.unlockMenu)

        # Connect signal event handler for gracefull shutdown when receiving kill signal
        self.app.signalHandler.sigInt.connect(self.slotShutdown)
        self.app.signalHandler.sigTerm.connect(self.slotShutdown)

        self.diagnosticsHandler = DiagnosticsHandler(self.config)

        # Create the application's main menu
        menuTitle = "{} ({}@{})".format(APP_NAME, getpass.getuser(), socket.gethostname())
        self.menu = QMenu()
        self.menu.addSection(menuTitle)
        self.menu.addAction("About and &help", self.about)
        configEditAction = self.menu.addAction("&Edit configuration file")
        userDiagnosticsAction = self.menu.addAction("Show &user diagnostics")
        userDiagnosticsAction.triggered.connect(self.diagnosticsHandler.openWindow)
        exitAction = self.menu.addAction("E&xit")
        exitAction.triggered.connect(lambda: self.app.exit(0))

        # On OSX the underlined keys don't exist, use keyboard shortcuts instead
        if sys.platform == 'darwin':
            configEditAction.setShortcut(QKeySequence('Ctrl+E'))
            exitAction.setShortcut(QKeySequence('Ctrl+X'))
            if hasattr(userDiagnosticsAction, 'setShortcutVisibleInContextMenu'):
                userDiagnosticsAction.setShortcutVisibleInContextMenu(True)
            userDiagnosticsAction.setShortcut(QKeySequence("Ctrl+U"))

        self.menu.addSeparator()
        self.menu.addAction(
            "&Start all enabled modules").triggered.connect(self.controller.startAllEnabled)
        self.menu.addAction("Stop &all modules").triggered.connect(self.controller.stopAll)

        configEditAction.triggered.connect(lambda: FileOpener.open(self.config_path))

        self.allowedMenuItems = [configEditAction]
        self.menu.addSection("Modules:")
        self.component_actions = {}

        self.reload()

        # Create the system tray menu
        self.tray = QSystemTrayIcon()
        self.tray.setIcon(ICONS['tray'])
        self.tray.setContextMenu(self.menu)
        self.tray.setToolTip(menuTitle)

        # Make left-clicking the icon also open the context menu
        if sys.platform != 'darwin': # Not on OSX, because left click is default there
            self.tray.activated.connect(lambda: self.menu.popup(
                self.tray.geometry().center()))
        self.tray.show()
        self.app.processEvents() # This is required to make the message below show up in the right
        # place (see: https://bugreports.qt.io/browse/QTBUG-1544) 
        if QSystemTrayIcon.supportsMessages():
            self.tray.showMessage(APP_NAME,
                                  "Started GSSw tray utility (on {}).\n"
                                  "Right click for options...".format(socket.gethostname()),
                                  msecs=TOOLTIP_TIMEOUT)

            self.diagnosticsHandler.sigNewMessage.connect(self.slotShowTrayMessage)
            # Open the list of messages whenever a message will be clicked
            self.tray.messageClicked.connect(self.diagnosticsHandler.openWindow)

        self.setupConfigurationFileWatcher()

        self._backgroundStatusUpdater.start()


    @pyqtSlot(int, str, str)
    def slotShowTrayMessage(self, severity, source, title):
        """ A new user message was received, display it in the system tray """
        if QSystemTrayIcon.supportsMessages(): 
            popupMsg = "{}: {}".format(source, title)
 
            msgLevel = severity if severity in range(4) else QSystemTrayIcon.Warning
 
            self.tray.showMessage(APP_NAME,
                                  popupMsg,
                                  msecs=TOOLTIP_TIMEOUT * 3,
                                  icon=msgLevel)

    def about(self):
        """ Shows the about message window (with links to Egress documentation). """
        urlManual      = "http://softwaredocs.sron.nl/gssw/"
        urlTraymonitor = "http://softwaredocs.sron.nl/gssw/usrman/running.html#taskbar-utility"
        aboutMessage = """
            <h3>About {}</h3>
            Egress is the new Ground Support Software: a software package for readout and control of
            electronic instruments. Egress provides a common software platform which is used in
            several different projects, e.g. for TES and KID sensor readout.<br />
            <br />
            For help on how to use Egress, developing your own scripts, and upgrading scripts from
            the old EGSE, check the online documentation at: <a href="{}">{}</a>.<br />
            <br />
            <i>Egress is developed by <a href="https://www.sron.nl">SRON</a> and
            <a href="https://www.stcorp.nl">S[&amp;]T</a></i>.
            <br />
            <h3>About {}</h3>
            The Egress monitor utility provides a modernized method of controlling Egress. With this
            utility you can control either all or indivual Egress components quickly
            (e.g., start or stop them), as well as viewing log files and opening the Egress
            configuration file.<br />
            <br />
            For help in using the Egress monitor, check its user manual at: <a href="{}">{}</a>.
            """.format(self.app.applicationDisplayName(), urlManual, urlManual, APP_NAME,
                       urlTraymonitor, urlTraymonitor)
        QMessageBox.about(self.app.activeWindow(), "About {}".format(APP_NAME), aboutMessage)

    @pyqtSlot()
    def startAnimation(self):
        """ Starts rotating icons for querying status """
        self.movie_waiting.start()
        self.movie_waiting.frameChanged.connect(self.updateGifFrame)

    @pyqtSlot()
    def stopAnimation(self):
        """ Stops icon animations """
        self.movie_waiting.stop()
        self.movie_waiting.frameChanged.disconnect(self.updateGifFrame)

    @pyqtSlot(int)
    def updateGifFrame(self, _):  #pylint: disable=unused-argument
        """
        Callback for updating gif of 'waiting icon'.

        :param frameNr: Ignored
        """

        ICONS['waiting'].swap(QIcon(self.movie_waiting.currentPixmap()))
        for _, componentEntry in self.component_actions.items():
            componentEntry['menu'].setWaiting()

    @pyqtSlot()
    def lockMenu(self):
        """ Disable all the menu items which control gssw components, wait for action completed. """
        for item in self.menu.actions():
            if not item.isSeparator() and item not in self.allowedMenuItems:
                item.setEnabled(False)

    @pyqtSlot()
    def unlockMenu(self):
        """ Disable all the menu items which control gssw components, wait for action completed. """
        for item in self.menu.actions():
            item.setEnabled(True)

    def setupConfigurationFileWatcher(self):
        """ Watch the configuration file for changes
            (so that we can reload the module list when the configuration file is changed) """

        # Some text editors don't write the file itself, but replace it with a new file
        # In that case, we won't be notified if we are watching only the configuration file itself
        # Therefore, also watch the directory of the configuration file
        directories = [os.path.dirname(self.config_path),
                       self.config_path]
        logger.info ("Watching config file in: %r", directories)
        self.fs_watcher = QFileSystemWatcher(directories)

        self.fs_watcher.directoryChanged.connect(self.directory_changed)
        self.fs_watcher.fileChanged.connect(self.file_changed)

    @pyqtSlot()
    def reload(self):
        """ Reloads the Gssw configuration file """
        logger.info("Reloading configuration...")
        self._backgroundStatusUpdater.pause()
        for _, component_action_item in self.component_actions.items():
            self.menu.removeAction(component_action_item['action'])

        self.component_actions = {}
        self.config            = initGssw(self.args)
        self.run_gssw          = RunGSSw(configPath=self.config_path, verbose=False)

        components = self.config['components'] if  'components' in self.config else {}


        # Create a menu item for each gssw component
        for component_name, component_config in sorted(components.items()):
            component_enable = component_config['enable'] if 'enable' in component_config else False

            component_menu = ComponentEntry(component_name, component_enable, self)
            component_action = self.menu.addMenu(component_menu)

            # Connect component menu actions to worker
            component_menu.sigStartSingle.connect(self.controller.startSingle)
            component_menu.sigStopSingle.connect(self.controller.stopSingle)
            component_menu.sigRestartSingle.connect(self.controller.restartSingle)

            self.component_actions[component_name] = {'menu': component_menu,
                                                      'action': component_action}
        self._backgroundStatusUpdater.resume()

    def updateComponentStatuses(self):
        """ Refreshes the status of all components """
        # First reload this PIDS file in RunGSSw
        # If processes have been restarted, they will probably have new PIDs
        self.run_gssw.loadPidsFile()

        # For each component, send a signal to the GUI thread to update the corresponding menu item
        for _, componentEntry in self.component_actions.items():
            componentEntry['menu'].triggerUpdateStatus()

    def directory_changed(self, path):
        """ Callback function of directory watcher """
        logger.info("Configuration directory '%s' changed, reloading configuration file", path)
        self.reload()

        # Re-add the configuration file path in case the config file is replaced
        self.fs_watcher.removePath(self.config_path)
        self.fs_watcher.addPath(self.config_path)

    def file_changed(self, path):
        """ Callback function of configuration file watcher """
        logger.info("Configuration file '%s' changed, reloading configuration file", path)
        self.reload()

    @pyqtSlot()
    def slotShutdown(self):
        """ Called in case of receiving kill signals """
        logger.info("Release pids lock and close")
        self.run_gssw.releasePidsLock()
        sys.exit(1)

    def run(self):
        """ Starts the tray application """
        # Some systems don't have a notification area in their system tray by default
        # In this case a tray icon will not show up.
        # Instead warn the user and quit
        if QSystemTrayIcon.isSystemTrayAvailable():
            # Enter Qt application main loop
            self.app.exec_()
        else:
            # Cannot run the app if there is no system tray
            QMessageBox(QMessageBox.Critical,
                APP_NAME,
                "Notification area not found, cannot run the {} application.\n\n"
                "Please add a notification area to your system tray "
                "and restart the application.".format(APP_NAME),
                QMessageBox.Ok).exec()
        # Stop background threads
        self._backgroundStatusUpdater.stop()
        sys.exit()


if __name__=='__main__':
    print(__doc__)
    print("Don't start this script directly, use 'gssw_traymonitor'")
