#!/usr/bin/env python3
#pylint: disable=invalid-name
""" Module for interfacing with an Agilent 33120A 15MHz Function/Arbitrary Waveform Generator. """

import argparse
import logging
import gssw.lib.typing  as typing
import gssw.ucelib.gpib as gpib

logger = logging.getLogger(__name__)

__all__ = ['Agilent33120A']

DEVICE_IDENTITY_PREFIX = "HEWLETT-PACKARD,33120A"

# Defaults
DEFAULT_HOST = "192.168.1.201"
DEFAULT_PORT = 1234
DEFAULT_GPIB_ADDRESS = 11

#pylint: disable=too-many-public-methods
class Agilent33120A(gpib.GpibTcpIp):
    """ Provides a communication interface to an Agilent 33120A 15MHz Function/Arbitrary Waveform
        Generator.

    :note: The waveform generator must be connected via a Prologix Gpib to Ethernet converter.
    :note: Some function docstrings are excerpts from the Agilent 33120A programming manual.
           Capital words correspond to (soft) keys on the instrument's front panel.
    """

    def __init__(self, host, port, gpibAddr):
        """ Constructor, for LAN communication. """

        super().__init__(host, port, "analyzer", commandInterval=0.01, processingDelay=0.01,
                         explicitRead=True,       # Request readings from the Prologix converter
                         timeout=4)               # Receiving measurement data can take a while

        if self.isConnected:
            if gpibAddr:
                # Set up the Prologix converter
                self.tx('++auto 0')                   # Don't read response after sending commands
                #                                       Otherwise commands which do not send a
                #                                       response will generate a -420 error on the
                #                                       instrument.
                self.tx('++mode 1')                   # Use Prologix converter in Controller mode
                self.tx('++addr {}'.format(gpibAddr)) # Configure the GPIB address
                self.tx('++eos 2')                    # Append LF to instrument commands
                self.tx('++eoi 1')                    # Use EOI signals after all last characters

            self._checkIdentification()


    def finalize(self):
        """ Cleans up the connection to the waveform generator and the Prologix converter. """
        if self.isConnected:
            pass

    def _checkIdentification(self):
        """ Verifies that we are connected to an Agilent 33120A. """
        identification = self.identification().strip()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        assert deviceCorrect,\
            "Device with identity '{}' is not supported.".format(identification)


    def setApply(self, waveform, frequency=None, amplitude=None, offset=None):
        """ Output a sine(SIN), square(SQU), triangle(TRI), ramp(RAMP) or user(USER)
            waveform with the specified frequency, amplitude, and dc-offset. The
            waveform is output as soon as the command is executed.

        :param waveform: Waveform, SIN, SQU, TRI, RAMP, USER.
        :param frequency: float, Frequency.
        :param amplitude: float, Amplitude.
        :param offset: float, Offset.
        :see: manual p143

        """
        assert waveform in ['SIN', 'SQU', 'TRI', 'RAMP', 'USER']
        if frequency:
            typing.check_class(frequency, float)
        if amplitude:
            typing.check_class(amplitude, float)
        if offset:
            typing.check_class(offset, float)
        if frequency and amplitude and offset:
            self.tx("APPL:{} {},{},{}".format(waveform, frequency, amplitude, offset))
        elif frequency and amplitude:
            self.tx("APPL:{} {},{}".format(waveform, frequency, amplitude))
        elif frequency:
            self.tx("APPL:{} {}".format(waveform, frequency))
        else:
            self.tx("APPL:{}")


    def setApplyNoise(self, amplitude=None, offset=None):
        """ Output noise with the specified amplitude, and dc offset.

        :param amplitude: float, Amplitude.
        :param offset: float, Offset.
        :see: manual p144
        """
        if amplitude:
            typing.check_class(amplitude, float)
        if offset:
            typing.check_class(offset, float)
        if amplitude and offset:
            self.tx("APPL:NOIS DEF,{},{}".format(amplitude, offset))
        elif amplitude:
            self.tx("APPL:NOIS DEF,{}".format(amplitude))
        else:
            self.tx("APPL:NOIS DEF")


    def setApplyDcOffset(self, offset=None):
        """ Output a dc voltage with the level specified by the offset parameter.
            The dc voltage is output as soon as the command is executed.

        :param offset: float, Offset.
        :see: manual p144
        """
        if offset:
            typing.check_class(offset, float)
        if offset:
            self.tx("APPL:DC DEF,DEF,{}".format(offset))
        else:
            self.tx("APPL:DC")


    def getApply(self):
        """ Query the function generator's present configuration and return a quoted string.

        :return: (str, float, float, float), Waveform, frequency, amplitude and offset values.
        :see: manual p144
        """
        strg = self.ask("APPL?").strip().strip('"').split()
        ret = [float(x) for x in strg[1].split(',')]
        return tuple([strg[0]] + ret)


    def setFunctionShape(self, shape):
        """ Select the output function. The selected waveform is output using the
            previously selected frequency, amplitude, and dc offset settings.

        :param shape: str, Waveform shape, SIN, SQU, TRI, RAMP, NOIS, DC, USER.
        :see: manual p145

        """
        assert shape in ['SIN', 'SQU', 'TRI', 'RAMP', 'NOIS', 'DC', 'USER']
        self.tx("FUNC:SHAP {}".format(shape))


    def getFunctionShape(self):
        """ Query the output function.

        :return: str, Waveform shape, SIN, SQU, TRI, RAMP, NOIS, DC, USER.
        :see: manual p145
        """
        return self.ask("FUNC:SHAP?")


    def setFrequency(self, frequency):
        """ Set the output frequency. MIN selects the lowest frequency allowed.
            MAX selects the highest frequency allowed.

        :param frequency: Output frequency, float, MIN, MAX.
        :see: manual p146
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
        self.tx("FREQ {}".format(frequency))


    def getFrequency(self, arg=None):
        """ Query the frequency setting for the function currently active.

        :param arg: str, Argument, None, MIN, MAX.
        :return: float, Frequency.
        :see: manual p146
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("FREQ? {}".format(arg)))
        else:
            return float(self.ask("FREQ?"))


    def setDutyCycle(self, percent):
        """ Set the duty cycle in percent for square waves only. Duty cycle
            represents the amount of time per cycle that the square wave is high.

        :param percent: Percent duty cycle, float, MIN, MAX.
        :see: manual p147
        """
        if percent not in ['MIN', 'MAX']:
            typing.check_class(percent, float)
        self.tx("PULS:DCYC {}".format(percent))


    def getDutyCycle(self, arg=None):
        """ Query the duty cycle setting. Returns a value in percent.

        :param arg: str, Argument, None, MIN, MAX.
        :return: float, Duty cycle.
        :see: manual p147
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("PULS:DCYC? {}".format(arg)))
        else:
            return float(self.ask("PULS:DCYC?"))


    def setVoltage(self, amplitude):
        """ Set the output amplitude for the current active function.

        :param amplitude: Output amplitude, float, MIN, MAX.
        :see: manual p148
        """
        if amplitude not in ['MIN', 'MAX']:
            typing.check_class(amplitude, float)
        self.tx("VOLT {}".format(amplitude))


    def getVoltage(self, arg=None):
        """ Query the output amplitude for the current selected function.

        :param arg: str, Argument, None, MIN, MAX.
        :return: float, Voltage.
        :see: manual p149
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("VOLT? {}".format(arg)))
        else:
            return float(self.ask("VOLT?"))


    def setVoltageOffset(self, offset):
        """ Set the dc offset voltage for the current active function.

        :param offset: Offset, float, MIN, MAX.
        :see: manual p149
        """
        if offset not in ['MIN', 'MAX']:
            typing.check_class(offset, float)
        self.tx("VOLT:OFF {}".format(offset))


    def getVoltageOffset(self, arg=None):
        """ Query the output amplitude for the current selected function.

        :param arg: str, Argument, None, MIN, MAX.
        :return: float, Offset voltage.
        :see: manual p150
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("VOLT:OFF? {}".format(arg)))
        else:
            return float(self.ask("VOLT:OFF?"))


    def setVoltageUnit(self, unit=None):
        """ Select the output units for amplitude only.

        :param: unit, NONE, VPP, VRMS, DBM
        :see: manual p150
        """
        assert unit in [None, 'VPP', 'VRMS', 'DBM']
        if unit is None:
            unit = 'DEF'
        self.tx("VOLT:UNIT {}".format(unit))


    def getVoltageUnit(self):
        """ Query the units selected.

        :return: str, Unit, VPP, VRMS, DBM.
        :see: manual p150
        """
        return float(self.ask("VOLT:UNIT?"))


    def setOutputLoad(self, load=None):
        """ Select the output termination for output amplitude and offset voltage.

        :param load: Output termination, 50, INF, MIN, MAX
        :see: manual p151
        """
        assert load in [None, 'INF', 'MIN', 'MAX']
        if load is None:
            load = 50
        self.tx("OUTP:LOAD {}".format(load))


    def getOutputLoad(self, arg=None):
        """ Query the output impedance.

        :param arg: str, Argument, None, MIN, MAX.
        :return: float, Output termination.
        :see: manual p151
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("OUTP:LOAD? {}".format(arg)))
        else:
            return float(self.ask("OUTP:LOAD?"))


    def setOutputSyncOn(self, state):
        """ Disable or enable output from the SYNC terminal.

        :param state: bool, Output sync state.
        :see: manual p151
        """
        typing.check_class(state, bool)
        self.tx("OUTP:SYNC {}".format(['OFF', 'ON'][state]))


    def getOutputSyncOn(self):
        """ Query the state of the SYNC terminal.

        :return: bool, Output sync state.
        :see: manual p151
        """
        return 'ON' if int(self.ask("OUTP:SYNC?")) else 'OFF'


    def setAMDepth(self, percent):
        """ Set the internal modulation depth in percent.

        :param percent: int, Modulation depth.
        :see: manual p155
        """
        if percent not in ['MIN', 'MAX']:
            typing.check_class(percent, float)
            assert percent >= 0 and percent <= 120
        self.tx("AM:DEPT {}".format(percent))


    def getAMDepth(self, arg=None):
        """ Query the modulation depth.

        :param arg: Argument, None, MIN, MAX
        :return: float, Modulation depth.
        :see: manual p155
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("AM:DEPT? {}".format(arg)))
        else:
            return float(self.ask("AM:DEPT?"))


    def setAMInternalFunction(self, shape):
        """ Select the shape of the modulating wavefrorm.

        :param shape: Modulating waveform shape, SIN, SQU, TRI, RAMP, NOIS, USER.
        :see: manual p155
        """
        assert shape in ['SIN', 'SQU', 'TRI', 'RAMP', 'NOIS', 'DC', 'USER']
        self.tx("AM:INT:FUNC {}".format(shape))


    def getAMInternalFunction(self):
        """ Query the shape of the internal modulating waveform.

        :return: Modulating waveform shape.
        :see: manual p155
        """
        return self.ask("AM:INT:FUNC?")


    def setAMInternalFrequency(self, frequency):
        """ Set the frequency of the modulating waveform.

        :param frequency: Modulating frequency.
        :see: manual p155
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
        self.tx("AM:INT:FREQ {}".format(frequency))


    def getAMInternalFrequency(self, arg=None):
        """ Query the internal modulating frequency.

        :param arg: Argument, None, MIN, MAX.
        :return: float, Modulating frequency.
        :see: manual p155
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("AM:INT:FREQ? {}".format(arg)))
        else:
            return float(self.ask("AM:INT:FREQ?"))


    def setAMSource(self, signal):
        """ Select the source of the modulating signal.

        :param signal: Source, BOTH, EXT.
        :see: manual p156
        """
        assert signal in ['BOTH', 'EXT']
        self.tx("AM:SOUR {}".format(signal))


    def getAMSource(self):
        """ Query the modulating source.

        :return: Source, BOTH, EXT.
        :see: manual p156
        """
        return float(self.ask("AM:SOUR?"))


    def setAMState(self, state):
        """ Disable or enable AM.

        :param state: bool, AM state.
        :see: manual p156
        """
        typing.check_class(state, bool)
        self.tx("AM:STAT {}".format(['OFF', 'ON'][state]))


    def getAMState(self):
        """ Query the state of AM.

        :return: bool, AM state.
        :see: manual p156
        """
        return 'ON' if int(self.ask("AM:STAT?")) else 'OFF'


    def setFMPeakDeviation(self, deviation):
        """ Set the peak frequency deviation.

        :param deviation: Frequency deviation, float, MIN, MAX.
        :see: manual p158
        """
        if deviation not in ['MIN', 'MAX']:
            typing.check_class(deviation, float)
            assert deviation >= 10e-3 and deviation <= 75e5
        self.tx("FM:DEV {}".format(deviation))


    def getFMPeakDeviation(self, arg):
        """ Query the peak frequency deviation.

        :param arg: Argument, None, MIN, MAX.
        :return: float, Frequency deviation.
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("FM:DEV? {}".format(arg)))
        else:
            return float(self.ask("FM:DEV?"))


    def setFMInternalFunction(self, shape):
        """ Select the shape of the modulating waveform.

        :param shape: Modulating waveform shape, SIN, SQU, TRI, RAMP, NOIS, USER.
        :see: manual p159
        """
        assert shape in ['SIN', 'SQU', 'TRI', 'RAMP', 'NOIS', 'USER']
        self.tx("FM:INT:FUNC {}".format(shape))


    def getFMInternalFunction(self):
        """ Query the shape of the modulating waveform.

        :return: Modulating waveform shape.
        :see: manual p159
        """
        return self.ask("FM:INT:FUNC?")


    def setFMInternalFrequency(self, frequency):
        """ Set the frequency of the modulating waveform.

        :param frequency: float, Modulating waveform frequency.
        :see: manual p159
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
            assert frequency >= 10e-3 and frequency <= 10e3
        self.tx("FM:INT:FREQ {}".format(frequency))


    def getFMInternalFrequency(self, arg=None):
        """ Query the modulating frequency.

        :param arg: Argument, None, MIN, MAX.
        :return: float, Modulating waveform frequency.
        :see: manual p159
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("FM:INT:FREQ? {}".format(arg)))
        else:
            return float(self.ask("FM:INT:FREQ?"))


    def setFMState(self, state):
        """ Disable or enable FM.

        :params state: bool, FM state.
        :see: manual p159
        """
        typing.check_class(state, bool)
        self.tx("FM:STAT {}".format(['OFF', 'ON'][state]))


    def getFMState(self):
        """ Query the state of FM.

        :return: bool, FM state.
        :see: manual p159
        """
        return 'ON' if int(self.ask("FM:STAT?")) else 'OFF'


    def setBurstCycles(self, cycles):
        """ Set the number for cycles to be output per burst.

        :param cycles: Cycles per burst, int, INF, MIN, MAX
        :see: manual p162
        """
        if cycles not in ['INF', 'MIN', 'MAX']:
            typing.check_class(cycles, int)
            assert cycles >= 1 and cycles <= 50000
        self.tx("BM:NCYC {}".format(cycles))


    def getBurstCycles(self, arg):
        """ Query the burst count.

        :return: int, Cycles per burst.
        :see: manual p163
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return int(self.ask("BM:NCYC? {}".format(arg)))
        else:
            return int(self.ask("BM:NCYC?"))


    def setBurstPhase(self, degrees):
        """ Set starting phase for the burst.

        :param degrees: int, Degrees
        :see: manual p163
        """
        if degrees not in ['MIN', 'MAX']:
            typing.check_class(degrees, float)
            assert degrees >= -360 and degrees <= 360
        self.tx("BM:NCYC {}".format(degrees))


    def getBurstPhase(self, arg):
        """ Query the starting phase.

        :param arg: Argument, None, MIN, MAX.
        :return: int, Degrees.
        :see: manual p163
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("BM:PHAS? {}".format(arg)))
        else:
            return float(self.ask("BM:PHAS?"))


    def setBMInternalRate(self, rate):
        """ Set the burst rate for internally triggered bursts.

        :param rate: float, Burst Rage.
        :see: manual p164
        """
        if rate not in ['MIN', 'MAX']:
            typing.check_class(rate, float)
            assert rate >= 10e-3 and rate <= 555550e3
        self.tx("BM:INT:RATE {}".format(rate))


    def getBMInternalRate(self, arg=None):
        """ Query the burst rate.

        :param arg: Argument, None, MIN, MAX.
        :return: float, Burst Rage.
        :see: manual p154
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("BM:INT:RATE? {}".format(arg)))
        else:
            return float(self.ask("BM:INT:RATE?"))


    def setBMSource(self, source):
        """ Select the burst modulation source.

        :param source: str, Modulation source, INT, EXT.
        :see: manual p164
        """
        assert source in ['INT', 'EXT']
        self.tx("BM:SOUR {}".format(source))


    def getBMSource(self):
        """ Query the present burst modulation source.

        :return: str, Modulation source.
        :see: manual p164
        """
        return float(self.ask("BM:SOUR?"))


    def setBMState(self, state):
        """ Disable or enable burst modulation.

        :param state: bool, Burst modulation state.
        :see: manual p165
        """
        typing.check_class(state, bool)
        self.tx("BM:STAT {}".format(['OFF', 'ON'][state]))


    def getBMState(self):
        """ Query the state of burst modulation.

        :return: bool, Burst modulation state.
        :see: manual p165
        """
        return 'ON' if int(self.ask("BM:STAT?")) else 'OFF'


    def setFSKFrequency(self, frequency):
        """ Set the FSK "hop" frequency.

        :param frequency: FSK "hop" frequency.
        :see: manual p168
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
            assert frequency >= 10e-3 and frequency <= 15e6
        self.tx("FSK:FREQ {}".format(frequency))


    def getFSKFrequency(self, arg=None):
        """ Query the FSK "hop" frequency.

        :param arg: Argument, None, MIN, MAX.
        :return: FSK "hop" frequency.
        :see: manual p168
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("FSK:FREQ? {}".format(arg)))
        else:
            return float(self.ask("FSK:FREQ?"))


    def setFSKInternalRate(self, rate):
        """ Set the rate at which the output frequency "shifts" between carrier
            and hop frequency.

        :param rate: float, FSK rate.
        :see: manual p168
        """
        if rate not in ['MIN', 'MAX']:
            typing.check_class(rate, float)
            assert rate >= 10e-3 and rate <= 50e3
        self.tx("FSK:INT:RATE {}".format(rate))


    def getFSKInternalRate(self, arg=None):
        """ Query the FSK rate.

        :return: float, FSK rate.
        :see: manual p168
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("FSK:INT:RATE? {}".format(arg)))
        else:
            return float(self.ask("FSK:INT:RATE?"))


    def setFSKSource(self, source):
        """ Select an internal or external FSK source.

        :param source: str, FSK source, INT, EXT.
        :see: manual p169
        """
        assert source in ['INT', 'EXT']
        self.tx("FSK:SOUR {}".format(source))


    def getFSKSource(self):
        """ Query the internal FSK source.

        :return: FSK source.
        :see: manual p169
        """
        return float(self.ask("FSK:SOUR?"))


    def setFSKState(self, state):
        """ Disable or enable FSK modulation.

        :param state: bool, FSK modulation sate.
        :see: manual p169
        """
        typing.check_class(state, bool)
        self.tx("FSK:STAT {}".format(['OFF', 'ON'][state]))


    def getFSKState(self):
        """ Query the state of the FSK mode.

        :return: FSK modulation sate.
        :see: manual p169
        """
        return 'ON' if int(self.ask("FSK:STAT?")) else 'OFF'


    def setSweepStartFrequency(self, frequency):
        """ Set the start frequency.

        :param frequency: Sweep start frequency, float, MIN, MAX.
        :see: manual p171
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
            assert frequency >= 10e-3 and frequency <= 15e6
        self.tx("FREQ:STAR {}".format(frequency))


    def getSweepStartFrequency(self, arg=None):
        """ Query the start frequency.

        :param arg: Argument, None, MIN, MAX.
        :return: float, Sweep start frequency.
        :see: manual p171
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("FREQ:STAR? {}".format(arg)))
        else:
            return float(self.ask("FREQ:STAR?"))


    def setSweepStopFrequency(self, frequency):
        """ Set the stop frequency.

        :param frequency: Sweep stop frequency, float, MIN, MAX.
        :see: manual p171
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
            assert frequency >= 10e-3 and frequency <= 15e6
        self.tx("FREQ:STOP {}".format(frequency))


    def getSweepStopFrequency(self, arg=None):
        """ Query the stop frequency.

        :return: float, Sweep stop frequency.
        :see: manual p171
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("FREQ:STOP? {}".format(arg)))
        else:
            return float(self.ask("FREQ:STOP?"))


    def setSweepSpacing(self, spacing):
        """ Set linear or logarithmic spacing for the sweep.

        :param spacing: str, Sweep spacing, LIN, LOG.
        :see: manual p171
        """
        assert spacing in ['LIN', 'LOG']
        self.tx("SWE:SPAC {}".format(spacing))


    def getSweepSpacing(self):
        """ Query the sweep mode.

        :return: str, Sweep spacing, LIN, LOG.
        :see: manual p171
        """
        return float(self.ask("SWE:SPAC?"))


    def setSweepTime(self, time):
        """ Set the number of seconds required to sweep from start
            frequency to stop frequency.

        :param time: Sweep time, float, MIN, MAX.
        :see: manual p171

        """
        if time not in ['MIN', 'MAX']:
            typing.check_class(time, float)
            assert time >= 1e-3 and time <= 500
        self.tx("SWE:TIME {}".format(time))


    def getSweepTime(self, arg=None):
        """ Query the sweep time.

        :return: float, Sweep time.
        :see: manual p172
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SWE:TIME? {}".format(arg)))
        else:
            return float(self.ask("SWE:TIME?"))


    def setSweepState(self, state):
        """ Disable or enable the sweep mode.

        :param state: bool, Sweep mode state.
        :see: manual p172
        """
        typing.check_class(state, bool)
        self.tx("SWE:STAT {}".format(['OFF', 'ON'][state]))


    def getSweepState(self):
        """ Query the state of the sweep mode.

        :return: bool, Sweep mode state.
        :see: manual p172
        """
        return 'ON' if int(self.ask("SWE:STAT?")) else 'OFF'


    def setTriggerSource(self, source):
        """ Select the trigger source.

        :param source: str, Trigger source, IMM, EXT, BUS.
        :see: manual p186
        """
        assert source in ['IMM', 'EXT', 'BUS']
        self.tx("TRIG:SOUR {}".format(source))


    def getTriggerSource(self):
        """ Query the present trigger source.

        :return: str, Trigger source.
        :see: manual p186
        """
        return float(self.ask("TRIG:SOUR?"))


    def setTriggerSlope(self, slope):
        """ Select the polarity of the Ext Trig terminal.

        :param slope: str, Trigger slope, POS, NEG.
        :see: manual p187
        """
        assert slope in ['POS', 'NEG']
        self.tx("TRIG:SLOP {}".format(slope))


    def getTriggerSlope(self):
        """ Query the present polarity of the Ext Trig terminal.

        :return: str, Trigger slope.
        :see: manual p187
        """
        return float(self.ask("TRIG:SLOP?"))


    def setDisplayOn(self, state):
        """ Turn the front-panel display off or on.

        :param state: bool, Display sate.
        :see: manual p187
        """
        typing.check_class(state, bool)
        self.tx("DISP {}".format(['OFF', 'ON'][state]))


    def getDisplayOn(self):
        """ Query the front-panel display settings.

        :return: bool, Display sate.
        """
        return 'ON' if int(self.ask("DISP?")) else 'OFF'


    def setDisplayText(self, text):
        """ Display a message containing up to 11 characters
            on the front panel.

        :param text: str, Display text.
        :see: manual p188
        """
        typing.check_class(text, str)
        self.tx('DISP:TEXT "{}"'.format(text))


    def getDisplayText(self):
        """ Query the message sent to the front panel and return
            a quoted string.

        :return: str, Display text.
        :see: manual p188
        """
        return self.tx("DISP:TEXT?")


    def setDisplayClear(self):
        """ Clear the message displayed on the front panel. """
        self.tx("DISP:TEXT:CLE")


def main():
    """ User interface program for configuring a waveform generator. """

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('host', nargs='?',
                        type=str,
                        help='hostname of the waveform generator '
                             '(default: %(default)s)',
                        default=DEFAULT_HOST)
    parser.add_argument('port', nargs='?',
                        type=int,
                        help='SCPI port number of the waveform generator '
                             '(default: %(default)s)',
                        default=DEFAULT_PORT)
    parser.add_argument('gpibAddress', nargs='?',
                        type=int,
                        help='GPIB address of te waveform generator '
                             '(default: %(default)s)',
                        default=DEFAULT_GPIB_ADDRESS)

    args = parser.parse_args()

    agilent = Agilent33120A(host=args.host,
                            port=args.port,
                            gpibAddr=args.gpibAddress)

    print("Identification: {}".format(agilent.identification().rstrip()))
    print("Prologix version: {}".format(agilent.converterVersion().rstrip()))

    agilent.setApply('SIN', 1.0, 0.1)
    print(agilent.getApply())

    agilent.finalize()


if __name__ == '__main__':
    main()
