#!/usr/bin/env python3
#pylint: disable=invalid-name
""" Module for interfacing with an Agilent 33522A 30MHz Function/Arbitrary Waveform Generator. """

import argparse
import logging
import gssw.lib.typing  as typing
import gssw.ucelib.gpib as gpib

logger = logging.getLogger(__name__)

__all__ = ['Agilent33522A']

DEVICE_IDENTITY_PREFIX = "Agilent Technologies,33522A"

# Defaults
DEFAULT_HOST = "192.168.1.6"
DEFAULT_PORT = 5025

#pylint: disable=too-many-public-methods
class Agilent33522A(gpib.GpibTcpIp):
    """ Provides a communication interface to an Agilent 33120A 15MHz Function/Arbitrary Waveform
        Generator.

    :note: The waveform generator must be connected via a Prologix Gpib to Ethernet converter.
    :note: Some function docstrings are excerpts from the Agilent 33120A programming manual.
           Capital words correspond to (soft) keys on the instrument's front panel.
    """

    def __init__(self, host, port):
        """ Constructor, for LAN communication. """

        super().__init__(host, port, "analyzer")
        self._checkIdentification()


    def finalize(self):
        """ Cleans up the connection to the waveform generator and the Prologix converter. """
        if self.isConnected:
            pass

    def _checkIdentification(self):
        """ Verifies that we are connected to an Agilent 33552A. """
        identification = self.identification().strip()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        assert deviceCorrect,\
            "Device with identity '{}' is not supported.".format(identification)


    def setApply(self, source, waveform, frequency=None, amplitude=None, offset=None):
        """ Output a sine(SIN), square(SQU), triangle(TRI), ramp(RAMP) or user(USER)
            waveform with the specified frequency, amplitude, and dc-offset. The
            waveform is output as soon as the command is executed.

        :param source: Source 1 or 2.
        :param waveform: Waveform, SIN, SQU, TRI, RAMP, USER.
        :param frequency: float, Frequency.
        :param amplitude: float, Amplitude.
        :param offset: float, Offset.
        :see: manual p143

        """
        assert waveform in ['SIN', 'SQU', 'TRI', 'RAMP', 'USER']
        if frequency:
            typing.check_class(frequency, float)
        if amplitude:
            typing.check_class(amplitude, float)
        if offset:
            typing.check_class(offset, float)
        if frequency and amplitude and offset:
            self.tx("SOUR{}:APPL:{} {},{},{}".format(
                source, waveform, frequency, amplitude, offset))
        elif frequency and amplitude:
            self.tx("SOUR{}:APPL:{} {},{}".format(source, waveform, frequency, amplitude))
        elif frequency:
            self.tx("SOUR{}:APPL:{} {}".format(source, waveform, frequency))
        else:
            self.tx("SOUR{}:APPL:{}".format(source, None))


    def setApplyNoise(self, source, amplitude=None, offset=None):
        """ Output noise with the specified amplitude, and dc offset.

        :param source: Source 1 or 2.
        :param amplitude: float, Amplitude.
        :param offset: float, Offset.
        :see: manual p144
        """
        if amplitude:
            typing.check_class(amplitude, float)
        if offset:
            typing.check_class(offset, float)
        if amplitude and offset:
            self.tx("SOUR{}:APPL:NOIS DEF,{},{}".format(source, amplitude, offset))
        elif amplitude:
            self.tx("SOUR{}:APPL:NOIS DEF,{}".format(source, amplitude))
        else:
            self.tx("SOUR{}:APPL:NOIS DEF")


    def setApplyDcOffset(self, source, offset=None):
        """ Output a dc voltage with the level specified by the offset parameter.
            The dc voltage is output as soon as the command is executed.

        :param offset: float, Offset.
        :see: manual p144
        """
        if offset:
            typing.check_class(offset, float)
        if offset:
            self.tx("SOUR{}:APPL:DC DEF,DEF,{}".format(source, offset))
        else:
            self.tx("SOUR{}:APPL:DC".format(source))


    def getApply(self, source):
        """ Query the function generator's present configuration and return a quoted string.

        :param: Source 1 or 2.
        :return: (str, float, float, float), Waveform, frequency, amplitude and offset values.
        :see: manual p144
        """
        strg = self.ask("SOUR{}:APPL?".format(source)).strip().strip('"').split()
        ret = [float(x) for x in strg[1].split(',')]
        return tuple([strg[0]] + ret)


    def setFunctionShape(self, source, shape):
        """ Select the output function. The selected waveform is output using the
            previously selected frequency, amplitude, and dc offset settings.

        :param source: Source 1 or 2.
        :param shape: str, Waveform shape, SIN, SQU, TRI, RAMP, NOIS, DC, USER.
        :see: manual p145

        """
        assert shape in ['SIN', 'SQU', 'TRI', 'RAMP', 'NOIS', 'DC', 'USER']
        self.tx("SOUR{}:FUNC:SHAP {}".format(source, shape))


    def getFunctionShape(self, source):
        """ Query the output function.

        :param source: Source 1 or 2.
        :return: str, Waveform shape, SIN, SQU, TRI, RAMP, NOIS, DC, USER.
        :see: manual p145
        """
        return self.ask("SOUR{}:FUNC:SHAP?".format(source))


    def setFrequency(self, source, frequency):
        """ Set the output frequency. MIN selects the lowest frequency allowed.
            MAX selects the highest frequency allowed.

        :param source: Source 1 or 2.
        :param frequency: Output frequency, float, MIN, MAX.
        :see: manual p146
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
        self.tx("SOUR{}:FREQ {}".format(source, frequency))


    def getFrequency(self, source, arg=None):
        """ Query the frequency setting for the function currently active.

        :param source: Source 1 or 2.
        :param arg: str, Argument, None, MIN, MAX.
        :return: float, Frequency.
        :see: manual p146
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{},FREQ? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}, FREQ?".format(source)))


    def setVoltage(self, source, amplitude):
        """ Set the output amplitude for the current active function.

        :param source: Source 1 or 2.
        :param amplitude: Output amplitude, float, MIN, MAX.
        :see: manual p148
        """
        if amplitude not in ['MIN', 'MAX']:
            typing.check_class(amplitude, float)
        self.tx("SOUR{}:VOLT {}".format(source, amplitude))


    def getVoltage(self, source, arg=None):
        """ Query the output amplitude for the current selected function.

        :param source: Source 1 or 2.
        :param arg: str, Argument, None, MIN, MAX.
        :return: float, Voltage.
        :see: manual p149
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:VOLT? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:VOLT?".format(source)))


    def setVoltageOffset(self, source, offset):
        """ Set the dc offset voltage for the current active function.

        :param source: Source 1 or 2.
        :param offset: Offset, float, MIN, MAX.
        :see: manual p149
        """
        if offset not in ['MIN', 'MAX']:
            typing.check_class(offset, float)
        self.tx("SOUR{}:VOLT:OFF {}".format(source, offset))


    def getVoltageOffset(self, source, arg=None):
        """ Query the output amplitude for the current selected function.

        :param source: Source 1 or 2.
        :param arg: str, Argument, None, MIN, MAX.
        :return: float, Offset voltage.
        :see: manual p150
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:VOLT:OFF? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:VOLT:OFF?".format(source)))


    def setVoltageUnit(self, source, unit=None):
        """ Select the output units for amplitude only.

        :param source: Source 1 or 2.
        :param: unit, NONE, VPP, VRMS, DBM
        :see: manual p150
        """
        assert unit in [None, 'VPP', 'VRMS', 'DBM']
        if unit is None:
            unit = 'DEF'
        self.tx("SOUR{}:VOLT:UNIT {}".format(source, unit))


    def getVoltageUnit(self, source):
        """ Query the units selected.

        :param source: Source 1 or 2.
        :return: str, Unit, VPP, VRMS, DBM.
        :see: manual p150
        """
        return float(self.ask("SOUR{}:VOLT:UNIT?".format(source)))


    def setOutputLoad(self, source, load=None):
        """ Select the output termination for output amplitude and offset voltage.

        :param source: Source 1 or 2.
        :param load: Output termination, 50, INF, MIN, MAX
        :see: manual p151
        """
        assert load in [None, 'INF', 'MIN', 'MAX']
        if load is None:
            load = 50
        self.tx("OUTP{}:LOAD {}".format(source, load))


    def getOutputLoad(self, source, arg=None):
        """ Query the output impedance.

        :param source: Source 1 or 2.
        :param arg: str, Argument, None, MIN, MAX.
        :return: float, Output termination.
        :see: manual p151
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("OUTP{}:LOAD? {}".format(source, arg)))
        else:
            return float(self.ask("OUTP{}:LOAD?".format(source)))


    def setOutputSyncOn(self, state):
        """ Disable or enable output from the SYNC terminal.

        :param state: bool, Output sync state.
        :see: manual p151
        """
        typing.check_class(state, bool)
        self.tx("OUTP:SYNC {}".format(['OFF', 'ON'][state]))


    def getOutputSyncOn(self):
        """ Query the state of the SYNC terminal.

        :return: bool, Output sync state.
        :see: manual p151
        """
        return 'ON' if int(self.ask("OUTP:SYNC?")) else 'OFF'


    def setAMDepth(self, source, percent):
        """ Set the internal modulation depth in percent.

        :param source: Source 1 or 2.
        :param percent: int, Modulation depth.
        :see: manual p155
        """
        if percent not in ['MIN', 'MAX']:
            typing.check_class(percent, float)
            assert percent >= 0 and percent <= 120
        self.tx("SOUR{}:AM:DEPT {}".format(source, percent))


    def getAMDepth(self, source, arg=None):
        """ Query the modulation depth.

        :param source: Source 1 or 2.
        :param arg: Argument, None, MIN, MAX
        :return: float, Modulation depth.
        :see: manual p155
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:AM:DEPT? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:AM:DEPT?".format(source)))


    def setAMInternalFunction(self, source, shape):
        """ Select the shape of the modulating wavefrorm.

        :param source: Source 1 or 2.
        :param shape: Modulating waveform shape, SIN, SQU, TRI, RAMP, NOIS, USER.
        :see: manual p155
        """
        assert shape in ['SIN', 'SQU', 'TRI', 'RAMP', 'NOIS', 'DC', 'USER']
        self.tx("SOUR{}:AM:INT:FUNC {}".format(source, shape))


    def getAMInternalFunction(self, source):
        """ Query the shape of the internal modulating waveform.

        :param source: Source 1 or 2.
        :return: Modulating waveform shape.
        :see: manual p155
        """
        return self.ask("SOUR{}:AM:INT:FUNC?".format(source))


    def setAMInternalFrequency(self, source, frequency):
        """ Set the frequency of the modulating waveform.

        :param source: Source 1 or 2.
        :param frequency: Modulating frequency.
        :see: manual p155
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
        self.tx("SOUR{}:AM:INT:FREQ {}".format(source, frequency))


    def getAMInternalFrequency(self, source, arg=None):
        """ Query the internal modulating frequency.

        :param source: Source 1 or 2.
        :param arg: Argument, None, MIN, MAX.
        :return: float, Modulating frequency.
        :see: manual p155
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:AM:INT:FREQ? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:AM:INT:FREQ?".format(source)))


    def setAMSource(self, source, signal):
        """ Select the source of the modulating signal.

        :param source: Source 1 or 2.
        :param signal: Source, BOTH, EXT.
        :see: manual p156
        """
        assert signal in ['BOTH', 'EXT']
        self.tx("SOUR{}:AM:SOUR {}".format(source, signal))


    def getAMSource(self, source):
        """ Query the modulating source.

        :param source: Source 1 or 2.
        :return: Source, BOTH, EXT.
        :see: manual p156
        """
        return float(self.ask("SOUR{}:AM:SOUR?".format(source)))


    def setAMState(self, source, state):
        """ Disable or enable AM.

        :param source: Source 1 or 2.
        :param state: bool, AM state.
        :see: manual p156
        """
        typing.check_class(state, bool)
        self.tx("SOUR{}:AM:STAT {}".format(source, ['OFF', 'ON'][state]))


    def getAMState(self, source):
        """ Query the state of AM.

        :param source: Source 1 or 2.
        :return: bool, AM state.
        :see: manual p156
        """
        return 'ON' if int(self.ask("SOUR{}:AM:STAT?".format(source))) else 'OFF'


    def setFMPeakDeviation(self, source, deviation):
        """ Set the peak frequency deviation.

        :param source: Source 1 or 2
        :param deviation: Frequency deviation, float, MIN, MAX.
        :see: manual p158
        """
        if deviation not in ['MIN', 'MAX']:
            typing.check_class(deviation, float)
            assert deviation >= 10e-3 and deviation <= 75e5
        self.tx("SOUR{}:FM:DEV {}".format(source, deviation))


    def getFMPeakDeviation(self, source, arg):
        """ Query the peak frequency deviation.

        :param source: Source 1 or 2.
        :param arg: Argument, None, MIN, MAX.
        :return: float, Frequency deviation.
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:FM:DEV? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:FM:DEV?".format(source)))


    def setFMInternalFunction(self, source, shape):
        """ Select the shape of the modulating waveform.

        :param source: Source 1 or 2.
        :param shape: Modulating waveform shape, SIN, SQU, TRI, RAMP, NOIS, USER.
        :see: manual p159
        """
        assert shape in ['SIN', 'SQU', 'TRI', 'RAMP', 'NOIS', 'USER']
        self.tx("SOUR{}:FM:INT:FUNC {}".format(source, shape))


    def getFMInternalFunction(self, source):
        """ Query the shape of the modulating waveform.

        :param source: Source 1 or 2.
        :return: Modulating waveform shape.
        :see: manual p159
        """
        return self.ask("SOUR{}:FM:INT:FUNC?".format(source))


    def setFMInternalFrequency(self, source, frequency):
        """ Set the frequency of the modulating waveform.

        :param source: Source 1 or 2.
        :param frequency: float, Modulating waveform frequency.
        :see: manual p159
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
            assert frequency >= 10e-3 and frequency <= 10e3
        self.tx("SOUR{}:FM:INT:FREQ {}".format(source, frequency))


    def getFMInternalFrequency(self, source, arg=None):
        """ Query the modulating frequency.

        :param source: Source 1 or 2.
        :param arg: Argument, None, MIN, MAX.
        :return: float, Modulating waveform frequency.
        :see: manual p159
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:FM:INT:FREQ? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:FM:INT:FREQ?".format(source)))


    def setFMState(self, source, state):
        """ Disable or enable FM.

        :param source: Source 1 or 2.
        :params state: bool, FM state.
        :see: manual p159
        """
        typing.check_class(state, bool)
        self.tx("SOUR{}:FM:STAT {}".format(source, ['OFF', 'ON'][state]))


    def getFMState(self, source):
        """ Query the state of FM.

        :param source: Source 1 or 2.
        :return: bool, FM state.
        :see: manual p159
        """
        return 'ON' if int(self.ask("SOUR{}:FM:STAT?".format(source))) else 'OFF'


    def setBurstCycles(self, source, cycles):
        """ Set the number for cycles to be output per burst.

        :param source: Source 1 or 2.
        :param cycles: Cycles per burst, int, INF, MIN, MAX
        :see: manual p162
        """
        if cycles not in ['INF', 'MIN', 'MAX']:
            typing.check_class(cycles, int)
            assert cycles >= 1 and cycles <= 50000
        self.tx("SOUR{}:BM:NCYC {}".format(source, cycles))


    def getBurstCycles(self, source, arg):
        """ Query the burst count.

        :param source: Source 1 or 2.
        :return: int, Cycles per burst.
        :see: manual p163
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return int(self.ask("SOUR{}:BM:NCYC? {}".format(source, arg)))
        else:
            return int(self.ask("SOUR{}:BM:NCYC?".format(source)))


    def setBurstPhase(self, source, degrees):
        """ Set starting phase for the burst.

        :param source: Source 1 or 2.
        :param degrees: int, Degrees
        :see: manual p163
        """
        if degrees not in ['MIN', 'MAX']:
            typing.check_class(degrees, float)
            assert degrees >= -360 and degrees <= 360
        self.tx("SOUR{}:BM:NCYC {}".format(source, degrees))


    def getBurstPhase(self, source, arg):
        """ Query the starting phase.

        :param source: Source 1 or 2.
        :param arg: Argument, None, MIN, MAX.
        :return: int, Degrees.
        :see: manual p163
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:BM:PHAS? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:BM:PHAS?".format(source)))


    def setBMInternalRate(self, source, rate):
        """ Set the burst rate for internally triggered bursts.

        :param source: Source 1 or 2.
        :param rate: float, Burst Rage.
        :see: manual p164
        """
        if rate not in ['MIN', 'MAX']:
            typing.check_class(rate, float)
            assert rate >= 10e-3 and rate <= 555550e3
        self.tx("SOUR{}:BM:INT:RATE {}".format(source, rate))


    def getBMInternalRate(self, source, arg=None):
        """ Query the burst rate.

        :param source: Source 1 or 2.
        :param arg: Argument, None, MIN, MAX.
        :return: float, Burst Rage.
        :see: manual p154
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:BM:INT:RATE? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:BM:INT:RATE?".format(source)))


    def setBMSource(self, src, source):
        """ Select the burst modulation source.

        :param src: Source 1 or 2.
        :param source: str, Modulation source, INT, EXT.
        :see: manual p164
        """
        assert source in ['INT', 'EXT']
        self.tx("SOUR{}:BM:SOUR {}".format(src, source))


    def getBMSource(self, source):
        """ Query the present burst modulation source.

        :param source: Source 1 or 2.
        :return: str, Modulation source.
        :see: manual p164
        """
        return float(self.ask("SOUR{}:BM:SOUR?".format(source)))


    def setBMState(self, source, state):
        """ Disable or enable burst modulation.

        :param source: Source 1 or 2.
        :param state: bool, Burst modulation state.
        :see: manual p165
        """
        typing.check_class(state, bool)
        self.tx("SOUR{}:BM:STAT {}".format(source, ['OFF', 'ON'][state]))


    def getBMState(self, source):
        """ Query the state of burst modulation.

        :param source: Source 1 or 2.
        :return: bool, Burst modulation state.
        :see: manual p165
        """
        return 'ON' if int(self.ask("SOUR{}:BM:STAT?".format(source))) else 'OFF'


    def setFSKFrequency(self, source, frequency):
        """ Set the FSK "hop" frequency.

        :param source: Source 1 or 2.
        :param frequency: FSK "hop" frequency.
        :see: manual p168
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
            assert frequency >= 10e-3 and frequency <= 15e6
        self.tx("SOUR{}:FSK:FREQ {}".format(source, frequency))


    def getFSKFrequency(self, source, arg=None):
        """ Query the FSK "hop" frequency.

        :param source: Source 1 or 2.
        :param arg: Argument, None, MIN, MAX.
        :return: FSK "hop" frequency.
        :see: manual p168
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:FSK:FREQ? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:FSK:FREQ?".format(source)))


    def setFSKInternalRate(self, source, rate):
        """ Set the rate at which the output frequency "shifts" between carrier
            and hop frequency.

        :param source: Source 1 or 2.
        :param rate: float, FSK rate.
        :see: manual p168
        """
        if rate not in ['MIN', 'MAX']:
            typing.check_class(rate, float)
            assert rate >= 10e-3 and rate <= 50e3
        self.tx("SOUR{}:FSK:INT:RATE {}".format(source, rate))


    def getFSKInternalRate(self, source, arg=None):
        """ Query the FSK rate.

        :param source: Source 1 or 2.
        :return: float, FSK rate.
        :see: manual p168
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:FSK:INT:RATE? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:FSK:INT:RATE?".format(source)))


    def setFSKSource(self, src, source):
        """ Select an internal or external FSK source.

        :param src: Source 1 or 2.
        :param source: str, FSK source, INT, EXT.
        :see: manual p169
        """
        assert source in ['INT', 'EXT']
        self.tx("SOUR{}:FSK:SOUR {}".format(src, source))


    def getFSKSource(self, source):
        """ Query the internal FSK source.

        :param source: Source 1 or 2.
        :return: FSK source.
        :see: manual p169
        """
        return float(self.ask("SOUR{}:FSK:SOUR?".format(source)))


    def setFSKState(self, source, state):
        """ Disable or enable FSK modulation.

        :param source: Source 1 or 2.
        :param state: bool, FSK modulation sate.
        :see: manual p169
        """
        typing.check_class(state, bool)
        self.tx("SOUR{}:FSK:STAT {}".format(source, ['OFF', 'ON'][state]))


    def getFSKState(self, source):
        """ Query the state of the FSK mode.

        :param source: Source 1 or 2.
        :return: FSK modulation sate.
        :see: manual p169
        """
        return 'ON' if int(self.ask("SOUR{}:FSK:STAT?".format(source))) else 'OFF'


    def setSweepStartFrequency(self, source, frequency):
        """ Set the start frequency.

        :param source: Source 1 or 2.
        :param frequency: Sweep start frequency, float, MIN, MAX.
        :see: manual p171
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
            assert frequency >= 10e-3 and frequency <= 15e6
        self.tx("SOUR{}:FREQ:STAR {}".format(source, frequency))


    def getSweepStartFrequency(self, source, arg=None):
        """ Query the start frequency.

        :param source: Source 1 or 2.
        :param arg: Argument, None, MIN, MAX.
        :return: float, Sweep start frequency.
        :see: manual p171
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:FREQ:STAR? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:FREQ:STAR?".format(source)))


    def setSweepStopFrequency(self, source, frequency):
        """ Set the stop frequency.

        :param source: Source 1 or 2.
        :param frequency: Sweep stop frequency, float, MIN, MAX.
        :see: manual p171
        """
        if frequency not in ['MIN', 'MAX']:
            typing.check_class(frequency, float)
            assert frequency >= 10e-3 and frequency <= 15e6
        self.tx("SOUR{}:FREQ:STOP {}".format(source, frequency))


    def getSweepStopFrequency(self, source, arg=None):
        """ Query the stop frequency.

        :param source: Source 1 or 2
        :return: float, Sweep stop frequency.
        :see: manual p171
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:FREQ:STOP? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:FREQ:STOP?".format(source)))


    def setSweepSpacing(self, source, spacing):
        """ Set linear or logarithmic spacing for the sweep.

        :param source: Source 1 or 2.
        :param spacing: str, Sweep spacing, LIN, LOG.
        :see: manual p171
        """
        assert spacing in ['LIN', 'LOG']
        self.tx("SOUR{}:SWE:SPAC {}".format(source, spacing))


    def getSweepSpacing(self, source):
        """ Query the sweep mode.

        :param source: Source 1 or 2.
        :return: str, Sweep spacing, LIN, LOG.
        :see: manual p171
        """
        return float(self.ask("SOUR{}:SWE:SPAC?".format(source)))


    def setSweepTime(self, source, time):
        """ Set the number of seconds required to sweep from start
            frequency to stop frequency.

        :param source: Source 1 or 2.
        :param time: Sweep time, float, MIN, MAX.
        :see: manual p171

        """
        if time not in ['MIN', 'MAX']:
            typing.check_class(time, float)
            assert time >= 1e-3 and time <= 500
        self.tx("SOUR{}:SWE:TIME {}".format(source, time))


    def getSweepTime(self, source, arg=None):
        """ Query the sweep time.

        :param source: Source 1 or 2.
        :return: float, Sweep time.
        :see: manual p172
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg:
            return float(self.ask("SOUR{}:SWE:TIME? {}".format(source, arg)))
        else:
            return float(self.ask("SOUR{}:SWE:TIME?".format(source)))


    def setSweepState(self, source, state):
        """ Disable or enable the sweep mode.

        :param source: Source 1 or 2.
        :param state: bool, Sweep mode state.
        :see: manual p172
        """
        typing.check_class(state, bool)
        self.tx("SOUR{}:SWE:STAT {}".format(source, ['OFF', 'ON'][state]))


    def getSweepState(self, source):
        """ Query the state of the sweep mode.

        :param source: Source 1 or 2.
        :return: bool, Sweep mode state.
        :see: manual p172
        """
        return 'ON' if int(self.ask("SOUR{}:SWE:STAT?".format(source))) else 'OFF'


    def setTriggerSource(self, src, source):
        """ Select the trigger source.

        :param src: Source 1 or 2.
        :param source: str, Trigger source, IMM, EXT, BUS.
        :see: manual p186
        """
        assert source in ['IMM', 'EXT', 'BUS']
        self.tx("TRIG{}:SOUR {}".format(src, source))


    def getTriggerSource(self, source):
        """ Query the present trigger source.

        :param source: Source 1 or 2.
        :return: str, Trigger source.
        :see: manual p186
        """
        return float(self.ask("TRIG{}:SOUR?".format(source)))


    def setTriggerSlope(self, source, slope):
        """ Select the polarity of the Ext Trig terminal.

        :param source: Source 1 or 2.
        :param slope: str, Trigger slope, POS, NEG.
        :see: manual p187
        """
        assert slope in ['POS', 'NEG']
        self.tx("TRIG{}:SLOP {}".format(source, slope))


    def getTriggerSlope(self, source):
        """ Query the present polarity of the Ext Trig terminal.

        :return: str, Trigger slope.
        :see: manual p187
        """
        return float(self.ask("TRIG{}:SLOP?".format(source)))


    def setDisplayOn(self, state):
        """ Turn the front-panel display off or on.

        :param state: bool, Display sate.
        :see: manual p187
        """
        typing.check_class(state, bool)
        self.tx("DISP {}".format(['OFF', 'ON'][state]))


    def getDisplayOn(self):
        """ Query the front-panel display settings.

        :return: bool, Display sate.
        """
        return 'ON' if int(self.ask("DISP?")) else 'OFF'


    def setDisplayText(self, text):
        """ Display a message containing up to 11 characters
            on the front panel.

        :param text: str, Display text.
        :see: manual p188
        """
        typing.check_class(text, str)
        self.tx('DISP:TEXT "{}"'.format(text))


    def getDisplayText(self):
        """ Query the message sent to the front panel and return
            a quoted string.

        :return: str, Display text.
        :see: manual p188
        """
        return self.tx("DISP:TEXT?")


    def setDisplayClear(self):
        """ Clear the message displayed on the front panel. """
        self.tx("DISP:TEXT:CLE")


def main():
    """ User interface program for configuring a waveform generator. """

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('host', nargs='?',
                        type=str,
                        help='hostname of the waveform generator '
                             '(default: %(default)s)',
                        default=DEFAULT_HOST)
    parser.add_argument('port', nargs='?',
                        type=int,
                        help='SCPI port number of the waveform generator '
                             '(default: %(default)s)',
                        default=DEFAULT_PORT)

    args = parser.parse_args()

    agilent = Agilent33522A(host=args.host,
                            port=args.port)

    print("Identification: {}".format(agilent.identification().rstrip()))

    agilent.setApply(1, 'SIN',1.0, 0.1)
    print(agilent.getApply(1))

    agilent.finalize()


if __name__ == '__main__':
    main()
