#!/usr/bin/env python3
#pylint: disable=invalid-name
""" Module for interfacing with an Agilent 34401A 6-1/2 Digit Multimeter. """

import argparse
import logging
import gssw.lib.typing  as typing
import gssw.ucelib.gpib as gpib

logger = logging.getLogger(__name__)

__all__ = ['Agilent34401A']

DEVICE_IDENTITY_PREFIX = "HEWLETT-PACKARD,34401A"

# Defaults
DEFAULT_HOST = "192.168.1.201"
DEFAULT_PORT = 1234
DEFAULT_GPIB_ADDRESS = 22


class Agilent34401A(gpib.GpibTcpIp):
    """ Provides a communication interface to an Agilent 34401A 6-1/2 Digit Multimeter.

    :note: The multimeter must be connected via a Prologix Gpib to Ethernet converter.
    :note: Some function docstrings are excerpts from the Agilent 34401A programming manual.
           Capital words correspond to (soft) keys on the instrument's front panel.
    """

    def __init__(self, host, port, gpibAddr):
        """ Constructor, for LAN communication. """

        super().__init__(host, port, "analyzer", commandInterval=0.01, processingDelay=0.01,
                         explicitRead=True,       # Request readings from the Prologix converter
                         timeout=4)               # Receiving measurement data can take a while

        if self.isConnected:
            if gpibAddr:
                # Set up the Prologix converter
                self.tx('++auto 1')                   # Read response after sending commands
                #                                       to the instrument.
                self.tx('++mode 1')                   # Use Prologix converter in Controller mode
                self.tx('++addr {}'.format(gpibAddr)) # Configure the GPIB address
                self.tx('++eos 2')                    # Append LF to instrument commands
                self.tx('++eoi 1')                    # Use EOI signals after all last characters

            self._checkIdentification()


    def finalize(self):
        """ Cleans up the connection to the power meter and the Prologix converter. """
        if self.isConnected:
            pass


    def _checkIdentification(self):
        """ Verifies that we are connected to a Agilent 34401A. """
        identification = self.identification().strip()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        assert deviceCorrect,\
            "Device with identity '{}' is not supported.".format(identification)


    @staticmethod
    def _parseArgument(arg=None):
        """ Parse argument for Agilent 34401A command. """
        if arg not in [None, 'MIN', 'MAX']:
            typing.check_class(arg, float)
        if arg is None:
            arg = 'DEF'
        return arg


    def getVoltageDC(self, rnge=None, resolution=None):
        """ Make a DC voltage measurement with the specified range
            and resolution.

        :param rnge: Range, None=Default, float, 'MIN', 'MAX'.
        :param resolution: Resolution, None=Default, float, 'MIN', 'MAX'.
        :return: float, DC Voltage.
        :see: manual p117
        """
        rnge = self._parseArgument(rnge)
        resolution = self._parseArgument(resolution)
        return float(self.ask("MEAS:VOLT:DC? {},{}".format(rnge, resolution)))


    def getVoltageDCRatio(self, rnge=None, resolution=None):
        """ Make a DC:DC ratio measurement with the specified range
            and resolution.

        :param rnge: Range, None=Default, float, 'MIN', 'MAX'.
        :param resolution: Resolution, None=Default, float, 'MIN', 'MAX'.
        :return: float, DC:DC ratio.
        :see: manual p117
        """
        rnge = self._parseArgument(rnge)
        resolution = self._parseArgument(rnge)
        return float(self.ask("MEAS:VOLT:DC:RAT? {},{}".format(rnge, resolution)))


    def getVoltageAC(self, rnge=None, resolution=None):
        """ Make a AC voltage measurement with the specified range
            and resolution.

        :param rnge: Range, None=Default, float, 'MIN', 'MAX'
        :param resolution: Resolution, None=Default, float, 'MIN', 'MAX'
        :return: float, AC voltage.
        :see: manual p117
        """
        rnge = self._parseArgument(rnge)
        resolution = self._parseArgument(resolution)
        return float(self.ask("MEAS:VOLT:AC? {},{}".format(rnge, resolution)))


    def getCurrentDC(self, rnge=None, resolution=None):
        """ Make a DC current measurement with the specified range
            and resolution.

        :param rnge: Range, None=Default, float, 'MIN', 'MAX'.
        :param resolution: Resolution, None=Default, float, 'MIN', 'MAX'.
        :return: float, DC current.
        :see: manual p117
        """
        rnge = self._parseArgument(rnge)
        resolution = self._parseArgument(resolution)
        return float(self.ask("MEAS:CURRT:DC? {},{}".format(rnge, resolution)))


    def getCurrentAC(self, rnge=None, resolution=None):
        """ Make an AC current measurement with the specified range
            and resolution.

        :param rnge: Range, None=Default, float, 'MIN', 'MAX'.
        :param resolution: Resolution, None=Default, float, 'MIN', 'MAX'.
        :return: float, AC current.
        :see: manual p118
        """
        rnge = self._parseArgument(rnge)
        resolution = self._parseArgument(resolution)
        return float(self.ask("MEAS:CURRT:AC? {},{}".format(rnge, resolution)))


    def getResistance(self, rnge=None, resolution=None):
        """ Make a 2-wire resistance measurement with the specified range
            and resolution.

        :param rnge: Range, None=Default, float, 'MIN', 'MAX'.
        :param resolution: Resolution, None=Default, float, 'MIN', 'MAX'.
        :return: float, 2-wire resistance.
        :see: manual p118
        """
        rnge = self._parseArgument(rnge)
        resolution = self._parseArgument(resolution)
        return float(self.ask("MEAS:RES? {},{}".format(rnge, resolution)))


    def get4wResistance(self, rnge=None, resolution=None):
        """ Make a 4-wire resistance measurement with the specified range
            and resolution.

        :param rnge: Range, None=Default, float, 'MIN', 'MAX'.
        :param resolution: Resolution, None=Default, float, 'MIN', 'MAX'.
        :return: float, 4-wire resistance.
        :see: manual p118
        """
        rnge = self._parseArgument(rnge)
        resolution = self._parseArgument(resolution)
        return float(self.ask("MEAS:FRES? {},{}".format(rnge, resolution)))


    def getFrequency(self, rnge=None, resolution=None):
        """ Make a frequency measurement with the specified range
            and resolution.

        :param rnge: Range, None=Default, float, 'MIN', 'MAX'.
        :param resolution: Resolution, None=Default, float, 'MIN', 'MAX'.
        :return: float, Frequency.
        :see: manual p118
        """
        rnge = self._parseArgument(rnge)
        resolution = self._parseArgument(resolution)
        return float(self.ask("MEAS:FREQ? {},{}".format(rnge, resolution)))


    def getPeriod(self, rnge=None, resolution=None):
        """ Make a period measurement with the specified range
            and resolution.

        :param rnge: Range, None=Default, float, 'MIN', 'MAX'.
        :param resolution: Resolution, None=Default, float, 'MIN', 'MAX'.
        :return: float, Period.
        :see: manual p118
        """
        rnge = self._parseArgument(rnge)
        resolution = self._parseArgument(resolution)
        return float(self.ask("MEAS:FREQ? {},{}".format(rnge, resolution)))


    def getContinuity(self):
        """ Make a continuity measurement. The range and resolution are
            fixed for continuity tests (1 kOhm range and 5-1/2 digits).

        :return: float, Ohm.
        :see: manual p118
        """
        return self.ask("MEAS:CONT?")


    def getDiode(self):
        """ Make a diode measurement. The range and resolution are
            fixed for diode tests (1 Vdc range with 1 mA current source
            and 5-1/2 digits).

        :return: float.
        :see: manual p118
        """
        return self.ask("MEAS:DIOD?")


def main():
    """ User interface program for configuring a multimeter. """

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('host', nargs='?',
                        type=str,
                        help='hostname of the multimeter '
                             '(default: %(default)s)',
                        default=DEFAULT_HOST)
    parser.add_argument('port', nargs='?',
                        type=int,
                        help='SCPI port number of the multimeter '
                             '(default: %(default)s)',
                        default=DEFAULT_PORT)
    parser.add_argument('gpibAddress', nargs='?',
                        type=int,
                        help='GPIB address of te multimeter '
                             '(default: %(default)s)',
                        default=DEFAULT_GPIB_ADDRESS)

    args = parser.parse_args()

    agilent = Agilent34401A(host=args.host,
                            port=args.port,
                            gpibAddr=args.gpibAddress)

    print("Identification: {}".format(agilent.identification().rstrip()))
    print("Prologix version: {}".format(agilent.converterVersion().rstrip()))

    print("voltage DC: %s" % agilent.getVoltageAC())
    agilent.finalize()


if __name__ == '__main__':
    main()
