#!/usr/bin/env python3
# pylint: disable=too-many-lines
""" Interface class for the Agilent 34970A multiplexed multimeter/data acquisition system """

import csv
import datetime
import logging

from gssw.ucelib.gpib import GpibSerial
from gssw.lib.typing import check_is_a_sequence, check_class

logger = logging.getLogger(__name__)

__all__ = ['Agilent34970A']

DEVICE_IDENTITY_PREFIX = "HEWLETT-PACKARD,34970A"

class Agilent34970A(GpibSerial):
    """
    Interfaces to an Agilent 34970A DAQ system.

    :seealso: Module docstring
    :note: Function docstrings are excerpts from the Agilent 34970A manual
    """
    def __init__(self, serialPort="/dev/ttyUSB0", timeout=1.2):
        super().__init__(serialPort         =serialPort,
                         baudrate           =115200,
                         device             ="Agilent 34970A",
                         timeout            =timeout,
                         processingDelay    =0.01,
                         commandTerminator  ='\n')

        self.checkIdentification()

    ################################################################################################
    # Static properties
    ################################################################################################

    VOLTAGE_RANGES = [# Possible settings for the Volt range
                      "MIN", # 100mV
                      "MAX", # 300 V
                      0.1,   # 100mV
                      1,     # 1   V
                      10,    # 10  V
                      100,   # 100 V
                      300,   # 300 V
                      "100mV",
                      "1V",
                      "10V",
                      "100V",
                      "300V"]

    BAUDRATES = [# Supported by the device, can only be changed via the front-panel
        1200,
        2400,
        4800,
        9600,
        19200,
        38400,
        57600,
        115200]

    SLOTS = [# Possible slot numbers
        100,
        200,
        300]

    CHANNELS_PER_MODULE = {# The number of channels per type of module
        "34902A": 16,
        "34901A": 20}

    PLC_INTEGRATIONTIMES = [# Possible options for ADC integration time (number of Powerline cycles)
        0.02, # 400µs
        0.2,  # 4  ms
        1,    # 20 ms
        2,    # 40 ms
        10,   # 200ms
        20,   # 400ms
        100,  # 2   s
        200]  # 4   s

    ################################################################################################
    # General functions, not directly derived from Agilent 34970A API
    ################################################################################################

    def beep(self):
        """ Causes the Agilent 34970A to beep
        This can be useful for alerting the user while displaying a message.

        :note: Side effect: If the error log is full, one entry will be deleted
        :note: When calling repeatedly, wait 0.25 seconds between each call.
        """
        self.tx("BEEP")     # This is not a valid command, and the unit beeps
                            # when an invalid command is received
        self.getLastError() # Remove the error, caused by the invalid syntax command

    def checkIdentification(self):
        """
        Verifies that the connected device actual is an Agilent 34970A

        :return: bool Result of the verification
        """
        identification = super().identification()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        if not deviceCorrect:
            logger.error("Device identity '%s' does not match '%s'",
                         identification, DEVICE_IDENTITY_PREFIX)
        return deviceCorrect


    def setDateTime(self, timestamp=None):
        """
        Sets the date and time from a python datetime.datetime object

        :param timestamp: datetime.datetime (set to datetime.now() when left to None)
        """
        if timestamp is None:
            timestamp = datetime.datetime.now()
        check_class(timestamp, datetime.datetime)

        self.setDate(timestamp.year, timestamp.month, timestamp.day)
        self.setTime(timestamp.hour,
                     timestamp.minute,
                     timestamp.second,
                     timestamp.microsecond//1000)

    ################################################################################################
    # Helper functions, e.g. for parsing and validating data.
    # These functions do not communicate with the Agilent 34970A
    ################################################################################################

    @classmethod
    def _checkValidResolution(cls, resolution=None, integrationTime=None):
        """
        For more information on the relationship between integration time,
        measurement resolution, number of digits, and number of bits, see the
        table on page 203.

        :param resolution: float, resolution reported by device
        :param integrationTime: float, :see: self.PLC_INTEGRATIONTIMES
        :see: manual p103 and p203, table with resolution table
        """
        # Resolution number is the highest (0.03) at 300 V range, 15 bits
        if resolution is not None:
            assert isinstance(resolution, float) and resolution <= 0.03 or \
                resolution in ["MIN", "MAX", "DEF"]
        elif integrationTime is not None:
            assert integrationTime in cls.PLC_INTEGRATIONTIMES

    @staticmethod
    def _formatChannelList(channelList):
        """
        Creates a serialized string with channels for sending via SCPI

        :param channelList: list of ints
        :return:            str, channel list readable by Agilent 34970A
        """
        check_is_a_sequence(channelList)
        assert all(isinstance(channel, int) for channel in channelList)
        formattedChannelList = ",".join(str(channel) for channel in channelList)
        return "(@{})".format(formattedChannelList)

    @staticmethod
    def _parseReadingsFromResponse(response):
        """
        Parses a raw list of readings received from the Agilent 34970A

        :param response: raw list (comma separated)
        :return: list of floats
        """
        readings = []
        if response:
            for reading in response.split(","):
                overloadValue = "9.90000001E+37" # See manual p98
                if reading[1:] == overloadValue:
                    readings.append(None)
                else:
                    readings.append(float(reading))
            return readings
        return []

    @staticmethod
    def parseResolutionBits(resolution, measuringRange):
        """
        Helper function which converts the measuring resolution to bits

        :param resolution:     float, resolution reported by device
        :param measuringRange: float, measuring range in Volts
        :return: int, number of bits
        :see: manual p103 and p203, table with resolution table
        """
        check_class(resolution,     float)
        check_class(measuringRange, float)
        resolutions = [
            #precision   bits digits PLC
            (0.0001,     15,  "4½",  0.02),
            (0.00001,    18,  "5½",  0.2 ),
            (0.000003,   20,  "5½",  1   ),
            (0.0000022,  21,  "6½",  2   ),
            (0.000001,   24,  "6½",  10  ),
            (0.0000008,  25,  "6½",  20  ),
            (0.0000003,  26,  "6½",  100 ),
            (0.00000022, 26,  "6½",  200 )
            ]

        for precision, bits, _, _ in resolutions:
            rangeResolution = precision * measuringRange
            # We cannot perform a float==float comparison due to float imprecision
            match = "{:.1E}".format(resolution) == "{:.1E}".format(rangeResolution)
            if match:
                return bits
        return None # No supported resolution found

    ################################################################################################
    # CONFigure Command Syntax, manual p211
    ################################################################################################

    def setChannelConfigurationVoltage(self, voltageRange, resolution, channelList):
        """
        Configure the specified channels for dc or ac voltage measurements but
        do not initiate the scan. Note that this command also redefines the scan list.
        For ac measurements, the resolution is actually fixed at 6½ digits;
        the resolution parameter only affects the number of digits shown on the
        front-panel.

        :param voltageRange: str or float, desired range. See list below for accepted values
        :param resolution:   float, resolution in Volts
        :param channelList:  list of integers with channel numbers
        :note: Fixed to DC settings
        :see: manual p212
        """
        assert voltageRange in ["AUTO", "DEF"] + self.VOLTAGE_RANGES
        self._checkValidResolution(resolution)
        self.tx("CONF:VOLT:DC {},{}, {}".format(voltageRange,
                                                resolution,
                                                self._formatChannelList(channelList)))

    def getChannelConfiguration(self):
        """
        Query the present configuration on the specified channels and return a
        series of quoted strings. If you omit the optional ch_list parameter,
        the instrument uses the active scan list. The command returns a series
        of comma-separated fields as shown in the example below. Each field
        contains the function, range, and resolution.
        "FRES +1.000000E+02,+3.000000E-04" , "TEMP TC,K,+1.000000E+00,+3.000000E-06"

        :return: list of tuples, 1 tuple per channel (function, range (in Volt), resolution)
        :note:   Channel numbers are not present in the configuration (access via getRouteScanList)
        :see:    manual p213
        """
        configurationResponse = self.ask("CONF?")

        configuration = []
        for channel in csv.reader(configurationResponse,
                                  delimiter=',',
                                  quotechar='"'):
            if len(channel)!=2: # csv.reader creates lists between the channels, skip them
                function, voltageRangeResolution = channel[0].split()
                voltageRange, resolution = [float(x) for x in voltageRangeResolution.split(',')]
                configuration.append((function,
                                      voltageRange,
                                      resolution))
        return configuration

    def setChannelVoltageRange(self, voltageRange, channelList):
        """
        Select the measurement range for the function selected on the specified
        channels. MIN selects the lowest range. MAX selects the highest range.

        :param voltageRange: :see: self.VOLTAGE_RANGES
        :param channelList:  list of integers with channel numbers
        :see: manual p215
        """
        assert voltageRange in self.VOLTAGE_RANGES
        self.tx("SENS:VOLT:DC:RANG {},{}".format(voltageRange,
                                                 self._formatChannelList(channelList)))

    def setChannelVoltageRangeAuto(self, enable, channelList):
        """
        Disable or enable autoranging on the specified channels. Autoranging
        thresholds: Down ranges at <10% of range; Up range at >120% of range.

        :param enable:      bool Desired state of autoranging
        :param channelList: list of integers with channel numbers
        :see: manual p215
        """
        check_class(enable, bool)
        self.tx("SENS:VOLT:DC:RANG:AUTO {},{}".format("ON" if enable else "OFF",
                                                      self._formatChannelList(channelList)))

    def getChannelVoltageRangeAuto(self):
        """
        Query the autorange setting on the specified channels. Returns “0” (OFF)
        or “1” (ON).

        :return: list with booleans, where True means that autoranging is on for the channel
        :see: manual p216
        """

        autorangeResponse = self.ask("SENS:VOLT:DC:RANG:AUTO?")
        return [bool(int(autoRange)) for autoRange in autorangeResponse.split(",")]

    def setChannelVoltageResolution(self, resolution, channelList):
        """
        Select the resolution for the function selected on the specified channels.
        Specify the resolution in the same units as the measurement function,
        not in number of digits. MIN selects the smallest value accepted for this
        parameter, which gives the most resolution. MAX selects the largest
        value accepted for this parameter, which gives the least resolution.
        For more information on the relationship between integration time,
        measurement resolution, number of digits, and number of bits, see the
        table on page 203.

        :param resolution:  str or float, lower value means more precise resolution
        :param channelList: list of integers with channel numbers
        :see: manual p216
        """
        self._checkValidResolution(resolution)
        self.tx("SENS:VOLT:DC:RES {},{}".format(resolution,
                                                self._formatChannelList(channelList)))

    def setChannelIntegrationTime(self, integrationTime, channelList):
        """
        Set the integration time in number of power line cycles (PLCs) on the
        specified channels. The default is 1 PLC. MIN = 0.02. MAX = 200.
        For more information on the relationship between integration time,
        measurement resolution, number of digits, and number of bits, see the
        table on page 203. For more information on integration time, refer to
        “Custom A/D Integration Time” in chapter 4 starting on page 103.

        :param integrationTime: float, :see: self.PLC_RESOLUTIONS
        :param channelList: list of integers with channel numbers
        :see: manual p218
        """
        self._checkValidResolution(integrationTime=integrationTime)
        self.tx("SENS:VOLT:DC:NPLC {},{}".format(integrationTime,
                                                self._formatChannelList(channelList)))

    def getChannelIntegrationTime(self, channelList):
        """
        Query the integration time on the specified channels. Returns a number
        in the form “+1.00000000E+00”.

        :param channeLList: list of integers with channel numbers
        :see: manual p218
        """
        response = self.ask("SENS:VOLT:DC:NPLC? {}".format(self._formatChannelList(channelList)))
        return [float(integrationTime) for integrationTime in response.split(",")]


    ################################################################################################
    # Voltage Configuration Commands, manual p223
    ################################################################################################

    def setChannelInputImpedanceAuto(self, enable, channelList):
        """
        Disable or enable the automatic input resistance mode for dc voltage
        measurements on the specified channels. With AUTO OFF (default),
        the input resistance is fixed at 10 MΩ for all ranges. With AUTO ON,
        the input resistance is set to >10 GΩ for the 100 mV, 1 V, and 10 V
        ranges.

        :param enabled: True enables automatic mode
        :see: manual p113, p223
        """
        self.tx("INP:IMP:AUTO {},{}".format("ON" if enable else "OFF",
                                                self._formatChannelList(channelList)))

    def getChannelInputImpedanceAuto(self, channelList):
        """
        The :AUTO? query returns the input resistance setting on the
        specified channels. Returns “0” (OFF) or “1” (ON).

        :see: manual p223
        :return: list with bools
        """
        response = self.ask("INP:IMP:AUTO? {}".format(self._formatChannelList(channelList)))
        return [bool(int(channelAutoStatus)) for channelAutoStatus in response.split(",")]

    ################################################################################################
    # Scanning Overview, manual p228
    ################################################################################################

    def setRouteScanList(self, channelList):
        """
        Select the channels to be included in the scan list. To start the scan,
        use the INITiate or READ? command. To remove all channels from the
        scan list, send ROUT:SCAN (@).
        The :SCAN? query returns a list of channel numbers in the SCPI
        definite length block format. The response begins with the “#” character,
        followed by a single character representing the number of succeeding
        characters to interpret as a length specifier, followed by a length
        specifier representing the number of bytes in the block, followed by a
        block of that many bytes. An empty scan list (one with no channels
        selected) will be “#13(@)”.
        For example, if you send ROUT:SCAN (@101:103), the ROUT:SCAN?
        command will return the following:
        #214(@101,102,103)

        :param channelList: list of integers with channel numbers
        :see: manual p228
        """
        check_is_a_sequence(channelList)
        assert all(isinstance(channel, int) for channel in channelList)

        self.tx("ROUT:SCAN {}".format(self._formatChannelList(channelList)))

    def getRouteScanList(self):
        """
        :see: setRouteScanList
        :return: list of integers with channel numbers
        """
        channelListResponse = self.ask("ROUT:SCAN?")

        # Decode the channel list (example: '#214(@101,102,107)')
        #First character is always #
        length = int(channelListResponse[1]) # The 2nd character is the channel list string length

        # Isolate the list #214(@......)
        channelListString = channelListResponse[2+length+2:-1]
        if channelListString:
            return [int(channel) for channel in channelListString.split(",")]
        return []

    def getRouteScanListSize(self):
        """
        Query the number of channels in the scan list. Returns a value between
        0 and 120 channels.

        :return: int, number of enabled channels
        :see: manual p228
        """
        channelListSizeResponse = self.ask("ROUT:SCAN:SIZE?")
        return int(channelListSizeResponse)

    def setTriggerSource(self, source):
        """
        Select the trigger source to control the onset of each sweep through the
        scan list (a sweep is one pass through the scan list). The selected source
        is used for all channels in the scan list. The instrument will accept a
        software (bus) command, an immediate (continuous) scan trigger, an
        external TTL trigger pulse, an alarm-initiated action, or an internallypaced
        timer. The default is “IMMediate”. The :SOUR? query returns the
        present scan trigger source. Returns “BUS”, “IMM”, “EXT”, “ALAR1”,
        “ALAR2”, “ALAR3”, “ALAR4”, or “TIM”.

        :see: manual p228
        """
        assert source in ["BUS", "IMM", "EXT", "ALAR1", "ALAR2", "ALAR3", "ALAR4", "TIM"]

        self.tx("TRIG:SOUR {}".format(source))

    def setTriggerTimer(self, interval):
        """
        Set the scan-to-scan interval (in seconds) for measurements on the
        channels in the scan list. This command defines the time from the start
        of one scan sweep to the start of the next sweep. You can set the interval
        to any value between 0 seconds and 359,999 seconds (99:59:59 hours),
        with 1 ms resolution. MIN = 0 seconds. MAX = 359,999 seconds.
        The :TIM? query returns the scan-to-scan interval in seconds in the
        form “+1.00000000E+01”.

        :see: manual p229
        """
        assert isinstance(interval, int) and interval in range(360000) or interval in ["MIN",
                                                                                          "MAX"]
        self.tx("TRIG:TIM {}".format(interval))

    def setTriggerCount(self, count):
        """
        Select the number of times that the instrument will sweep through the
        scan list (a sweep is one pass through the scan list). When the specified
        number of sweeps have occurred, the scan stops. Select a scan count
        between 1 to 50,000 sweeps, or continuous (INFinity). MIN = 1 sweep.
        MAX = 50,000 sweeps.

        :see: manual p229
        """
        assert isinstance(count, int) and count in range(1, 50001) or count in ["MIN", "MAX", "INF"]
        self.tx("TRIG:COUN {}".format(count))

    def getTriggerCount(self):
        """
        The :COUN? query returns the scan count in the form “+1.00000000E+01”.
        If you specified a continuous scan count, the query command returns
        “9.90000200E+37”.

        :return: int, normally [1,50000] and higher for continuous sweeps
        :see: manual p229
        """
        countResponse = self.ask("TRIG:COUN?")
        return int(float(countResponse))

    def setChannelDelays(self, delay, channelList):
        """
        Add a delay between multiplexer channels in the scan list (useful for
        high-impedance or high-capacitance circuits). The delay is inserted
        between the relay closure and the actual measurement on the channel.
        The programmed channel delay overrides the default channel delay that
        the instrument automatically adds to each channel.
        You can set the channel delay to any value between 0 seconds and
        60 seconds, with 1 ms resolution. You can select a different delay for
        each channel. The default channel delay is automatic; the instrument
        determines the delay based on function, range, integration time,
        and ac filter setting (see “Automatic Channel Delays” on page 89).

        :param delay:       float, delay (in seconds) for all specified channels
        :param channelList: list of integers with channel numbers
        :see: manual p229
        """
        check_class(delay, float)
        assert 0 <= delay <= 60
        self.tx("ROUT:CHAN:DELAY {},{}".format(delay,
                                               self._formatChannelList(channelList)))

    def getChannelDelays(self):
        """
        The :DEL? query reads the channel delay setting in seconds and returns
        a value in the form “+1.00000000E+00”.

        :return: list with floats (delay in seconds)
        :see: manual p229
        """
        delayResponse = self.ask("ROUT:CHAN:DELAY?")
        return [float(x) for x in delayResponse.split(",")]

    def setScanAbort(self):
        """
        Abort a scan in progress.

        :see: manual p230
        """
        self.tx("ABORT")

    def setScanInit(self):
        """
        Change the state of the triggering system from the “idle” state to the
        “wait-for-trigger” state. Scanning will begin when the specified trigger
        conditions are satisfied following the receipt of the INITiate command.
        The readings are placed in the instrument’s internal reading memory
        (up to 50,000 readings can be stored; if memory fills, new readings will
        overwrite the first readings stored). Readings are stored in memory until
        you are able to retrieve them. Use the FETCh? command to retrieve
        reading results.

        :see: manual p230
        """
        self.tx("INIT")

    ################################################################################################
    # Reading Format Commands, manual p231
    ################################################################################################

    def setFormatReadingChannel(self, enable):
        """
        Disable or enable the inclusion of the channel number with
        data produced by READ?, FETCh?, or other queries of scan results.
        This command operates in conjunction with the other FORMat:READing
        commands (they are not mutually exclusive). The :CHAN? query reads
        whether or not the channel is included. Returns “0” (OFF) or “1” (ON).

        :see: manual p231
        """
        self.tx("READ:CHAN {}".format("ON" if enable else "OFF"))

    def setFormatReadingUnit(self, enable):
        """
        Disable (default) or enable the inclusion of measurement units with
        data produced by READ?, FETCh?, or other queries of scan results.
        This command operates in conjunction with the other FORMat:READing
        commands (they are not mutually exclusive). The :UNIT? query reads
        whether or not the units are included. Returns “0” (OFF) or “1” (ON).

        :see: manual p232
        """
        self.tx("READ:UNIT {}".format("ON" if enable else "OFF"))

    ################################################################################################
    # Scan Statistics Commands, manual p233
    ################################################################################################

    def getCalculateAverage(self, channelList):
        """
        Calculate the mathematical average of all readings taken on each of the
        specified channels since the start of the scan. Each channel must be a
        multiplexer channel that has been configured to be part of the scan list.
        If no data is available for the specified channels, “0” is returned.
        Returns a number in the form “+2.61920000E+01”.

        :param channelList: list of ints
        :return: list of: floats and/or Nones, None indicates an out-of-range value
        :see: manual p234
        """

        values = self.ask("CALC:AVER:AVER? {}".format(self._formatChannelList(channelList)))
        return self._parseReadingsFromResponse(values)

    def getCalculatePeakToPeak(self, channelList):
        """
        Calculate the peak-to-peak value of all readings taken on each of the
        specified channel since the start of the scan (the difference between the
        maximum and minimum reading). Each channel must be a multiplexer
        channel that has been configured to be part of the scan list. If no data is
        available for the specified channels, “0” is returned. Returns a number
        in the form “+0.00000000E+00”.

        :param channelList: list of ints
        :return: list of: floats and/or Nones, None indicates an out-of-range value
        :see: manual p234
        """

        values = self.ask("CALC:AVER:PTP? {}".format(self._formatChannelList(channelList)))
        return self._parseReadingsFromResponse(values)

    def getCalculateCount(self, channelList):
        """
        Count the number of readings taken on each of the specified channels
        since the start of the scan. Each channel must be a multiplexer channel
        that has been configured to be part of the scan list. Returns a number in
        the form “+5.00000000E+00”.

        :param channelList: list of ints
        :return: list of ints
        :see: manual p234
        """
        values = self.ask("CALC:AVER:COUN? {}".format(self._formatChannelList(channelList)))
        values = self._parseReadingsFromResponse(values)

        # Count is an integer per channel, thus convert the values to integer
        return [int(value) for value in values]

    ################################################################################################
    # Scan Memory Commands, manual p235
    ################################################################################################

    def getScanDataPoints(self):
        """
        Count the total number of readings (from all channels in the scan list)
        currently stored in reading memory from a scan. Returns a value
        between 0 and 50,000 readings.

        :see: manual p235
        """
        scanDataPointsResponse = self.ask("DATA:POIN?")
        return int(scanDataPointsResponse)

    def setRemoveDataPoints(self, numberOfReadings):
        """
        Read and erase the specified number of readings from non-volatile
        memory. The readings are erased from memory starting with the oldest
        reading first. The purpose of this command is to allow you to periodically
        remove readings from memory during a scan than would normally cause
        reading memory to overflow (for example, during a scan with an infinite
        scan count). The output from this command is affected by the
        FORM:READ commands (see “Reading Format Commands” on page 231).

        :see: manual p235
        """
        check_class(numberOfReadings, int)
        self.tx("DATA:REM? {}".format(numberOfReadings))

    def getScanStarttime(self):
        """
        Read the time at the start of the scan. This command is not affected by
        the FORMat:READ:TIME:TYPE command. Returns the time in the form
        “2002,06,02,18,30,00.000” (June 2, 2002 at 6:30 PM).

        :return: datetime.datetime object
        :see: manual p236
        """
        scanStartimeResponse = self.ask("SYST:TIME:SCAN?")

        # Parse the formatted timestamp
        year, month, day, hour, minute, second = scanStartimeResponse.split(",")
        second, millisecond = second.split(".")
        return datetime.datetime(year       =int(year),
                                 month      =int(month),
                                 day        =int(day),
                                 hour       =int(hour),
                                 minute     =int(minute),
                                 second     =int(second),
                                 microsecond=int(millisecond)*1000)

    ################################################################################################
    # System-Related Commands, manual p264
    ################################################################################################

    def setDate(self, year, month, day):
        """
        Set the instrument calendar. The setting is stored in non-volatile memory.
        When shipped from the factory, instrument is set to the current time
        and date (U.S Mountain Time). For example, the following command
        sets the calendar to June 1, 2002.
        SYST:DATE 2002,06,01

        :param year:  int
        :param month: int
        :param day:   int
        :see: manual p264
        """

        check_class(year, int)
        check_class(month, int)
        check_class(day, int)
        assert year in range(10000)
        assert month in range(1,13)
        assert day in range(1,32)

        self.tx("SYST:DATE {},{},{}".format(year, month, day))

    def setTime(self, hours, minutes, seconds, milliseconds):
        """
        SYSTem:TIME <hh>,<mm>,<ss.sss> (based on 24-hour clock)
        Set the instrument clock. The setting is stored in non-volatile memory.
        When shipped from the factory, instrument is set to the current time
        and date (U.S Mountain Time). For example, the following command
        sets the time to 6:30 PM.
        SYST:TIME 18,30,00

        :param hours:        int
        :param minutes:      int
        :param seconds:      int
        :param milliseconds: int
        :see: manual p264
        """
        check_class(hours, int)
        check_class(minutes, int)
        check_class(seconds, int)
        check_class(milliseconds, int)
        assert hours   in range(24)
        assert minutes in range(60)
        assert seconds in range(60)
        assert milliseconds in range(1000)

        self.tx("SYST:TIME {},{},{}.{}".format(hours, minutes, seconds, milliseconds))

    def getModuleIdentity(self, slot):
        """
        Read the identity of the module in the specified slot. For example, the
        34903A 20-Channel Actuator returns:
        HEWLETT-PACKARD,34903A,0,1.0
        The last number (1.0 in this example) is the firmware revision number
        for the module.

        :param slot: int, number of the slot
        :return:     tuple (manufacturer, identity, firmware revision)
        :see: manual p265
        """
        assert slot in self.SLOTS

        moduleTypeString = self.ask("SYST:CTYP? {}".format(slot))
        manufacturer, model, _, firmwareRevision = moduleTypeString.split(",")
        return manufacturer, model, firmwareRevision

    def setDisplayEnabled(self, enable):
        """
        Disable or enable the front-panel display. When disabled, the entire
        front-panel display goes dark and all display annunciators except
        ERROR are disabled. All keys except are locked out when the
        display is disabled. The display is automatically enabled when power
        is cycled, after a Factory Reset (``*RST`` command), or when you return
        to local by pressing . The DISP? query the front-panel display
        setting. Returns “0” (OFF) or “1” (ON).

        :see: manual p266
        """
        check_class(enable, bool)

        self.tx("DISP {}".format("ON" if enable else "OFF"))

    def setDisplayText(self, text):
        """
        Display a message on the instrument’s front-panel display.
        The instrument can display up to 13 characters on the front panel;
        if you attempt to send more than 13 characters, an error is generated.
        The :TEXT? query reads the message sent to the instrument front panel
        and returns a quoted string.
        You can use letters (A-Z), numbers (0-9), and special characters like
        “@”, “%”, “*”, etc. Use the “#” character to display a degree symbol ( ° ).
        Commas, periods, and semicolons share a display space with the
        preceding character, and are not considered individual characters.
        While a message is displayed on the front panel, readings from a scan or
        monitor are not sent to the display.
        The following command displays a message on the front panel.
        DISP:TEXT ’SCANNING ...’

        :see: manual p266
        """
        self.tx("DISP:TEXT '{}'".format(text))

    def getDisplayText(self):
        """
        :see: setDisplayText
        """
        return self.ask("DISP:TEXT?")

    def setDisplayClear(self):
        """
        Clear the message displayed on the front-panel display.

        :see: manual p266
        """
        self.tx("DISP:TEXT:CLE")

    def getLastError(self):
        """ Retrieves and erases the last entry from the error log

        :return: (code, message)
        """
        return self.ask("SYST:ERR?")

    ################################################################################################
    # Interface Configuration Commands, manual p269
    ################################################################################################

    def setSystemLockKeypanel(self, locked):
        """
        Places the instrument in remote or local mode. In the remote
        mode, the keys on the front panel are locked.

        :param locked: bool
        :see: manual p269
        """
        if locked:
            self.tx("SYST:RWL")
        else:
            self.tx("SYST:REM")

    def setClearOperation(self):
        """
        Clear the operation in progress over the RS-232 interface and
        discard any pending output data. This is equivalent to the IEEE-488
        Device Clear action over the GPIB interface.

        :see: manual p269
        """
        self.tx('\x03') # Send the Ctrl-C character

    ################################################################################################
    # Calibration Commands, manual p292
    ################################################################################################

    def getCalibrationCount(self):
        """
        Query the instrument to determine the number of times it has been
        calibrated. Note that your instrument was calibrated before it left the
        factory. When you receive your instrument, read the count to determine
        its initial value. The calibration count is stored in non-volatile memory
        in the mainframe. The calibration count increments up to a maximum
        of 65,535 after which it rolls over to “0”. Since the value increments by
        one for each calibration point, a complete calibration may increase the
        value by many counts. The calibration count is also incremented by
        calibrations of the DAC channels on the multifunction module.

        :see: manual p292
        :return: int, number of times the unit has been calibrated
        """
        calibrationCountResponse = self.ask("CAL:COUN?")
        return int(calibrationCountResponse)

    def getCalibrationMessage(self):
        """
        Query the calibration message and return a quoted string. Returns a
        string in the form “CAL: 06-01-98”.

        :see: manual p293
        :return: str
        """
        calibrationMessageResponse = self.ask("CAL:STR?")
        return calibrationMessageResponse

    ################################################################################################
    # Service-Related Commands, manual p294
    ################################################################################################

    def getDmmRelayCycles(self):
        """
        Query the cycle count of the three backplane relays on the internal DMM.
        These relays open or close when a function or range is changed on
        a module. Returns three numbers indicating the cycle count on relays
        “1”, “2”, and “3” (which correspond to relays K102, K103, and K104
        respectively).

        :see: manual p294
        """
        relayCyclesResponse = self.ask("DIAG:DMM:CYCL?")
        return [int(x) for x in relayCyclesResponse.split(",")]

    def getChannelRelayCycles(self, channelList):
        """
        Query the cycle count on the specified channels. In addition to the
        channel relays, you can also query the count on backplane relays and
        bank relays. Note that you cannot control the state of these relays from
        the front panel but you can query the count. For more information on
        channel numbering and layout, refer to “Module Overview” starting on
        page 163.

        :see: manual p294
        """
        channelRelayCyclesResponse = self.ask("DIAG:REL:CYCL? {}".format(
            self._formatChannelList(channelList)))
        return [int(x) for x in channelRelayCyclesResponse.split(",")]


class Agilent34970ASimulator(Agilent34970A):
    """ Simulator class for developing scripts using the voltmeter """

    _averages = [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0]
    _scanDataPoints = 1

    def checkIdentification(self):
        return True

    def open(self):
        pass

    def getCalculateAverage(self, channelList):
        for i, _ in enumerate(self._averages):
            self._averages[i] += 0.1
        return self._averages

    def getCalculatePeakToPeak(self, channelList):
        return [5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5]

    def getCalculateCount(self, channelList):
        return [5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5]

    def getScanDataPoints(self):
        self._scanDataPoints += 1
        return self._scanDataPoints


def main():
    """ Test program for data acquisition unit """
    logging.basicConfig(
        level='DEBUG',
        format='%(asctime)s %(filename)25s:%(lineno)-4d : %(levelname)-7s: %(message)s')

    print ("Test program for:" + __doc__)
    agilent = Agilent34970A()
    try:
        #agilent.tx("TEST") # Send an invalid command to create an error entry
        print ("Connected to: {}".format(agilent.identification()))
        if agilent.checkIdentification():
            errors = agilent.getErrorLog()
            if errors:
                print ("Error log:")
                for code, message in errors:
                    print ("Code: {},\tDescription: {}".format(code, message))
            else:
                print ("Error log is empty")
            #agilent.reset()
            print("Scanned channels:", agilent.getRouteScanList())
            channelList = [101,102,107]
            agilent.setRouteScanList(channelList)
            print("Scanned channels:", agilent.getRouteScanList())
            agilent.setDateTime()
            for slot in agilent.SLOTS:
                print("Slot {} type: {}".format(slot, agilent.getModuleIdentity(slot)[1]))
            #agilent.setScanInit()
            #import time
            #time.sleep(1)

            #print("Scan start time:", agilent.getScanStarttime())
            #print("Scan data points:", agilent.getScanDataPoints())
            #print("Scan averages:", agilent.getCalculateAverage(channelList))
            #print("Scan Pk2Pk:", agilent.getCalculatePeakToPeak(channelList))
            #print("Scan Count:", agilent.getCalculateCount(channelList))
            agilent.setChannelConfigurationVoltage(10.0, 0.0001, [101,102])
            agilent.setChannelVoltageRangeAuto(False, [101])
            agilent.setChannelVoltageRange(0.1, [102])

            for function, voltageRange, resolution in agilent.getChannelConfiguration():
                print ("\nfunction:     {}".format(function))

                #print ("raw:",voltageRangeResolution)

                resolutionP = agilent.parseResolutionBits(resolution, voltageRange)
                print("voltageRange: {} V".format(voltageRange))
                print("resolution:   {} V ({} bits)".format(resolution, resolutionP))

            #print("Channel configuration:", agilent.getChannelConfiguration())
            print("Channel autorange:", agilent.getChannelVoltageRangeAuto())

            #agilent.setRemoveDataPoints(3)
            #agilent.setDisplayText("GSSW BUSY")
            #agilent.setSystemLockKeypanel(True)

            #import time
            #for _ in range(4):
            #    for i in range(5):
            #        agilent.setDisplayText("GSSW BUSY" + "."*i)
            #        time.sleep(.2)
            #agilent.setSystemLockKeypanel(False)
            #agilent.setDisplayClear()
            print("Sweep count: {}".format(agilent.getTriggerCount()))
            agilent.setTriggerCount("INF")
            triggerCount = agilent.getTriggerCount()
            print("Sweep count: {}".format(triggerCount if triggerCount <= 50000 else "Continuous"))

            print ("Channel delays: {}".format(agilent.getChannelDelays()))
            agilent.setChannelDelays(1.0, [101])
            print ("Channel delays: {}".format(agilent.getChannelDelays()))
            print ("NPLCs: {}".format(agilent.getChannelIntegrationTime(channelList)))

    finally:
        agilent.close()

if __name__ == '__main__':
    main()
