#!/usr/bin/env python3
#pylint: disable=invalid-name
""" Module for interfacing with an Agilent E3632A DC Power Supply. """

import argparse
import logging
import gssw.lib.typing  as typing
import gssw.ucelib.gpib as gpib

logger = logging.getLogger(__name__)

__all__ = ['Agilent3632A']

DEVICE_IDENTITY_PREFIX = "HEWLETT-PACKARD,E363"

# Defaults
DEFAULT_HOST = "192.168.1.201"
DEFAULT_PORT = 1234
DEFAULT_GPIB_ADDRESS = 5


class Agilent3632A(gpib.GpibTcpIp):
    """ Provides a communication interface to an Agilent E3632A DC Power Supply.

    :note: The power supply must be connected via a Prologix Gpib to Ethernet converter.
    :note: Some function docstrings are excerpts from the Agilent E3632A programming manual.
           Capital words correspond to (soft) keys on the instrument's front panel.
    """

    def __init__(self, host, port, gpibAddr):
        """ Constructor, for LAN communication. """

        super().__init__(host, port, "analyzer", commandInterval=0.01, processingDelay=0.01,
                         explicitRead=True,       # Request readings from the Prologix converter
                         timeout=4)               # Receiving measurement data can take a while

        if self.isConnected:
            if gpibAddr:
                # Set up the Prologix converter
                self.tx('++auto 0')                   # Don't read response after sending commands
                #                                       Otherwise commands which do not send a
                #                                       response will generate a -420 error on the
                #                                       instrument.
                self.tx('++mode 1')                   # Use Prologix converter in Controller mode
                self.tx('++addr {}'.format(gpibAddr)) # Configure the GPIB address
                self.tx('++eos 2')                    # Append LF to instrument commands
                self.tx('++eoi 1')                    # Use EOI signals after all last characters

            self._checkIdentification()


    def finalize(self):
        """ Cleans up the connection to the power supply and the Prologix converter. """
        if self.isConnected:
            self.setDisplayOn(True)


    def _checkIdentification(self):
        """ Verifies that we are connected to an Agilent E3632A. """
        identification = self.identification().strip()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        assert deviceCorrect,\
            "Device with identity '{}' is not supported.".format(identification)


    def setApply(self, voltage=None, current=None, output=None):
        """ This method is a combination of select, voltage and current commands. The values
            of voltage and the current of the specified output are changed as soon as the command
            is executed.

        :param voltage: Voltage, None=Default, float, 'MIN', 'MAX'.
        :param current: Current, None=Default, float, 'MIN', 'MAX'.
        :param output: Select output, P6V, P25V, N25V
        :see: manual p73

        """
        assert output in [None, 'P6V', 'P25V', 'N25V']
        if voltage not in [None, 'MIN', 'MAX']:
            typing.check_class(voltage, float)
        if voltage is None:
            voltage = 'DEF'
        if current not in [None, 'MIN', 'MAX']:
            typing.check_class(current, float)
        if current is None:
            current = 'DEF'
        if output is None:
            self.tx("APPL {},{}".format(voltage, current))
        else:
            self.tx("APPL {},{},{}".format(output, voltage, current))


    def getApply(self):
        """ This method queries the power supply's present voltage and current setting
            values.

        :return: (float, float), Voltage and current values.
        :see: manual p83
        """
        return tuple([float(x.strip().strip('"')) for x in self.ask("APPL?").split(',')])


    def setCurrent(self, current):
        """ This method programs the immediate current level of the power supply. The
            immediate level is the current value of the output terminals.

        :param current: Current, float, 'MIN', 'MAX'.
        :see: manual p77

        """
        if current not in ['MIN', 'MAX']:
            typing.check_class(current, float)
        self.tx("CURR {}".format(current))


    def getCurrent(self, arg=None):
        """ Return the presently programmed current level of the power supply.

        :return: Current, float, 'MIN', 'MAX'.
        :see: manual p77
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg is None:
            return float(self.ask("CURR?"))
        else:
            return float(self.ask("CURR? {}".format(arg)))


    def setInstrumentSelect(self, output):
        """ This command selects the output to be programmed by the
            output identifier. The outputs of the power supply are
            considered as three logical instruments.

        :param output: Select output, 'P6V', 'P25V', 'N25V'.
        :see: manual p74
        """
        assert output in ['P6V', 'P25V', 'N25V']
        self.tx("INST:SEL {}".format(output))


    def getInstrumentSelect(self):
        """ Return the currently selected output. The returned value
            is P6V, P25V, N25V.

        :return: Selected output, 'P6V', 'P25V', 'N25V'.
        :see: manual p74
        """
        return self.ask("INST:SEL?")


    def setInstrumentNSelect(self, output):
        """ This command selects the output to be programmed by a
            numeric value. 1 selects +6V output, 2 selects +25V output,
            and 3 selects -25V output.

        :param output: int, Select output.
        :see: manual p74
        """
        typing.check_class(output, int)
        assert output in [1, 2]
        self.tx("INST:NSEL {}".format(output))


    def getInstrumentNSelect(self):
        """ Return the currently selected output. The returned
            parameter is 1 for +6V output, 2 selects +25V output, and
            3 selects -25V output.

        :see: manual p74
        """
        return int(self.ask("INST:NSEL?"))


    def getMeasureCurrent(self, output=None):
        """ Query the current measured at the sense terminals of the
            power supply.

        :param output: Select output, 'P6V', 'P25V', 'N25V'.
        :see: manual p76
        """
        assert output in [None, 'P6V', 'P25V', 'N25V']
        if output is None:
            return float(self.ask("MEAS:CURR?"))
        else:
            return float(self.ask("MEAS:CURR? {}".format(output)))


    def getMeasureVoltage(self, output=None):
        """ Query the voltage measured at the sense terminals of the
            power supply.

       :param output: Select output, 'P6V', 'P25V', 'N25V'.
       :see: manual p76
        """
        assert output in [None, 'P6V', 'P25V', 'N25V']
        if output is None:
            return float(self.ask("MEAS:VOLT?"))
        else:
            return float(self.ask("MEAS:VOLT? {}".format(output)))


    def setVoltage(self, voltage):
        """ This method programs the immediate voltage level of the power supply. The
            immediate level is the voltage value of the output terminals.

        :param voltage: Voltage, float, 'MIN', 'MAX'.
        :see: manual p78

        """
        if voltage not in ['MIN', 'MAX']:
            typing.check_class(voltage, float)
        self.tx("VOLT {}".format(voltage))


    def getVoltage(self, arg=None):
        """ Return the presently programmed voltage level of the power supply.

        :return: Voltage, float, 'MIN', 'MAX'.
        :see: manual p78
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg is None:
            return float(self.ask("VOLT?"))
        else:
            return float(self.ask("VOLT? {}".format(arg)))


    def setDisplayOn(self, state):
        """ Turn the front panel display off (False) or on (True).

        :param state: bool, Display state.
        :see: manual p82
        """
        typing.check_class(state, bool)
        self.tx("DISP:STAT {}".format(['OFF', 'ON'][state]))


    def getDisplayOn(self):
        """ Query front panel display settings.

        :return: bool, Display state.
        :see: manual p82
        """
        return bool(int(self.ask("DISP:STAT?")))


    def setDisplayText(self, text):
        """ Display a message on the front panel. The power supply
            will display up to 11 characters in a message.

        :param text: str, Display message.
        :see: manual p82
        """
        typing.check_class(text, str)
        self.tx('DISP:TEXT "{}"'.format(text))


    def getDisplayText(self):
        """ Query the message sent to the front panel and returns
            a quoted string.

        :return: str, Display message.
        :see: manual p82
        """
        return self.ask("DISP:TEXT?")


    def setDisplayClear(self):
        """ Clear the message displayed on the front panel.

        :see: manual p82
        """
        self.tx("DISP:TEXT:CLE")


    def setOutputOn(self, state):
        """ Enable or disable all three outputs of the power supply.

        :param state: bool, Output state.
        :see: manual p77
        """
        typing.check_class(state, bool)
        self.tx("OUTP:STAT {}".format(['OFF', 'ON'][state]))


    def getOutputOn(self):
        """ Query the output state of the power supply.

        :return: bool, Output state.
        :see: manual p77
        """
        return bool(int(self.ask("OUTP:STAT?")))


def main():
    """ User interface program for configuring a power supply. """

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('host', nargs='?',
                        type=str,
                        help='hostname of the power supply '
                             '(default: %(default)s)',
                        default=DEFAULT_HOST)
    parser.add_argument('port', nargs='?',
                        type=int,
                        help='SCPI port number of the power supply '
                             '(default: %(default)s)',
                        default=DEFAULT_PORT)
    parser.add_argument('gpibAddress', nargs='?',
                        type=int,
                        help='GPIB address of te power supply '
                             '(default: %(default)s)',
                        default=DEFAULT_GPIB_ADDRESS)

    args = parser.parse_args()

    agilent = Agilent3632A(host=args.host,
                           port=args.port,
                           gpibAddr=args.gpibAddress)

    print("Identification: {}".format(agilent.identification().rstrip()))
    print("Prologix version: {}".format(agilent.converterVersion().rstrip()))

    agilent.setApply(1.0, 0.1)
    print('voltage, current: {}'.format(agilent.getApply()))

    agilent.setDisplayOn(False)
    agilent.setVoltage(0.1)
    print('voltage: {}'.format(agilent.getVoltage()))

    agilent.setCurrent(0.2)
    print('current: {}'.format(agilent.getCurrent()))

    agilent.finalize()


if __name__ == '__main__':
    main()
