#!/usr/bin/env python3
#pylint: disable=invalid-name
""" Module for interfacing with an Agilent E3648A Dual Output DC Power Supply. """

import argparse
import logging
import gssw.lib.typing  as typing
import gssw.ucelib.gpib as gpib

logger = logging.getLogger(__name__)

__all__ = ['Agilent3648A']

DEVICE_IDENTITY_PREFIX = "Agilent Technologies,E364"

# Defaults
DEFAULT_HOST = "192.168.1.201"
DEFAULT_PORT = 1234
DEFAULT_GPIB_ADDRESS = 6


class Agilent3648A(gpib.GpibTcpIp):
    """ Provides a communication interface to an Agilent E3648A Dual Output DC Power Supply.

    :note: The power supply must be connected via a Prologix Gpib to Ethernet converter.
    :note: Some function docstrings are excerpts from the Agilent E3648A programming manual.
           Capital words correspond to (soft) keys on the instrument's front panel.
    """

    def __init__(self, host, port, gpibAddr):
        """ Constructor, for LAN communication. """

        super().__init__(host, port, "analyzer", commandInterval=0.01, processingDelay=0.01,
                         explicitRead=True,       # Request readings from the Prologix converter
                         timeout=4)               # Receiving measurement data can take a while

        if self.isConnected:
            if gpibAddr:
                # Set up the Prologix converter
                self.tx('++auto 0')                   # Don't read response after sending commands
                #                                       Otherwise commands which do not send a
                #                                       response will generate a -420 error on the
                #                                       instrument.
                self.tx('++mode 1')                   # Use Prologix converter in Controller mode
                self.tx('++addr {}'.format(gpibAddr)) # Configure the GPIB address
                self.tx('++eos 2')                    # Append LF to instrument commands
                self.tx('++eoi 1')                    # Use EOI signals after all last characters

            self._checkIdentification()


    def finalize(self):
        """ Cleans up the connection to the power supply and the Prologix converter. """
        if self.isConnected:
            self.setDisplayOn(True)


    def _checkIdentification(self):
        """ Verifies that we are connected to an Agilent E3648A. """
        identification = self.identification().strip()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        assert deviceCorrect,\
            "Device with identity '{}' is not supported.".format(identification)


    def setApply(self, voltage=None, current=None):
        """ This method is a combination of voltage and current commands. This method
            changes the power supply's output to the newly programmed values.

        :param voltage: Voltage, None=Default, float, 'MIN', 'MAX'.
        :param current: Current, None=Default, float, 'MIN', 'MAX'.
        :see: manual p83

        """
        if voltage not in [None, 'MIN', 'MAX']:
            typing.check_class(voltage, float)
        if voltage is None:
            voltage = 'DEF'
        if current not in [None, 'MIN', 'MAX']:
            typing.check_class(current, float)
        if current is None:
            current = 'DEF'
        self.tx("APPL {},{}".format(voltage, current))


    def getApply(self):
        """ This method queries the power supply's present voltage and current setting
            values.

        :return: (float, float), Voltage and current values.
        :see: manual p83
        """
        return tuple([float(x.strip().strip('"')) for x in self.ask("APPL?").split(',')])


    def setCurrent(self, current):
        """ This method programs the immediate current level of the power supply. The
            immediate level is the current value of the output terminals.

        :param current: Current, float, 'MIN', 'MAX', 'UP', 'DOWN'.
        :see: manual p84

        """
        if current not in ['MIN', 'MAX', 'UP', 'DOWN']:
            typing.check_class(current, float)
        self.tx("CURR {}".format(current))


    def getCurrent(self, arg=None):
        """ Return the presently programmed current level of the power supply.

        :return: Current, float, 'MIN', 'MAX'.
        :see: manual p84
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg is None:
            return float(self.ask("CURR?"))
        else:
            return float(self.ask("CURR? {}".format(arg)))


    def setCurrentStep(self, step=None):
        """ Set the step size for current programming with the UP and
            DOWN commands.

        :param step: float, Current step size.
        :see: manual p84
        """
        if step is None:
            step = 'DEF'
        else:
            typing.check_class(step, float)
        self.tx("CURR:STEP {}".format(step))


    def getCurrentStep(self, arg=None):
        """ Return the value of the current step size currently specified.
            DEF gives the minimum resolution of the step size in unit
            of amperes.

        :param arg: Argument, None, DEF.
        :return: Current step size.
        :see: manual p85

        """
        assert arg in [None, 'DEF']
        if arg is None:
            return float(self.ask("CURR:STEP?"))
        else:
            return float(self.ask("CURR:STEP? {}".format(arg)))


    def setCurrentUp(self):
        """ Current step up.

        :see: manual p85
        """
        self.tx("CURR UP")


    def setCurrentDown(self):
        """ Current step down.

        :see: manual p86
        """
        self.tx("CURR DOWN")


    def setInstrumentSelect(self, output):
        """ This command selects the output to be programmed by the
            output identifier. The outputs of the power supply are
            considered as two logical instruments.

        :param output: Select output, 'OUTP1', 'OUTP2', 'OUT1', 'OUT2'.
        :see: manual p86
        """
        assert output in ['OUTP1', 'OUTP2', 'OUT1', 'OUT2']
        self.tx("INST:SEL {}".format(output))


    def getInstrumentSelect(self):
        """ Return the currently selected output. The returned value
            is OUTP1 or OUTP2.

        :return: Selected output, 'OUTP1', 'OUTP2'
        :see: manual p86
        """
        return self.ask("INST:SEL?")


    def setInstrumentNSelect(self, output):
        """ This command selects the output to be programmed by a
            numeric value. 1 selects output1, and 2 selects output2.

        :param output: int, Select output.
        :see: manual p86
        """
        typing.check_class(output, int)
        assert output in [1, 2]
        self.tx("INST:NSEL {}".format(output))


    def getInstrumentNSelect(self):
        """ Return the currently selected output. The returned
            parameter is 1 for output1, and 2 for output2.

        :see: manual p86
        """
        return int(self.ask("INST:NSEL?"))


    def getMeasureCurrent(self):
        """ Query the current measured at the sense terminals of the
            power supply.

        :see: manual p88
        """
        return float(self.ask("MEAS:CURR?"))


    def getMeasureVoltage(self):
        """ Query the voltage measured at the sense terminals of the
            power supply.

        :see: manual p88
        """
        return float(self.ask("MEAS:VOLT?"))


    def setVoltage(self, voltage):
        """ This method programs the immediate voltage level of the power supply. The
            immediate level is the voltage value of the output terminals.

        :param voltage: Voltage, float, 'MIN', 'MAX', 'UP', 'DOWN'.
        :see: manual p84

        """
        if voltage not in ['MIN', 'MAX', 'UP', 'DOWN']:
            typing.check_class(voltage, float)
        self.tx("VOLT {}".format(voltage))


    def getVoltage(self, arg=None):
        """ Return the presently programmed voltage level of the power supply.

        :return: Voltage, float, 'MIN', 'MAX'.
        :see: manual p88
        """
        assert arg in [None, 'MIN', 'MAX']
        if arg is None:
            return float(self.ask("VOLT?"))
        else:
            return float(self.ask("VOLT? {}".format(arg)))


    def setVoltageStep(self, step):
        """ Set the step size for voltage programming with the UP and
            DOWN commands.

        :param step: float, Voltage step size.
        :see: manual p89
        """
        if step != 'DEF':
            typing.check_class(step, float)
        self.tx("VOLT:STEP {}".format(step))


    def getVoltageStep(self, arg=None):
        """ Return the value of the voltage step size currently specified.
            DEF gives the minimum resolution of the step size in unit
            of volts.

        :param arg: Argument, None, DEF.
        :see: manual p89

        """
        assert arg in [None, 'DEF']
        if arg is None:
            return float(self.ask("VOLT:STEP?"))
        else:
            return float(self.ask("VOLT:STEP? {}".format(arg)))


    def setVoltageUp(self):
        """ Voltage step up.

        :see: manual p89
        """
        self.tx("VOLT UP")


    def setVoltageDown(self):
        """ Voltage step down.

        :see: manual p89
        """
        self.tx("VOLT DOWN")


    def setVoltageRange(self, rnge):
        """ Select an output range to be programmed by the rnge identifier.

        :param rnge: Voltage range, P8V, P20V, P35V, P60V, LOW, HIGH.
        :see: manual p91
        """
        assert rnge in ['P8V', 'P20V', 'P35V', 'P60V',
                        'LOW', 'HIGH']
        self.tx("VOLT:RANG {}".format(rnge))


    def getVoltageRange(self):
        """ Query the currently selected voltage range.

        :return: str, Voltage range, P8V, P20V, P35V, P60V
        :see: manual p91
        """
        return self.ask("VOLT:RANG?")


    def setDisplayOn(self, state):
        """ Turn the front panel display off (False) or on (True).

        :param state: bool, Display state.
        :see: manual p95
        """
        typing.check_class(state, bool)
        self.tx("DISP {}".format(['OFF', 'ON'][state]))


    def getDisplayOn(self):
        """ Query front panel display settings.

        :return: bool, Display state.
        :see: manual p95
        """
        return bool(int(self.ask("DISP?")))


    def setDisplayText(self, text):
        """ Display a message on the front panel. The power supply
            will display up to 11 characters in a message.

        :param text: str, Display message.
        :see: manual p95
        """
        typing.check_class(text, str)
        self.tx('DISP:TEXT "{}"'.format(text))


    def getDisplayText(self):
        """ Query the message sent to the front panel and returns
            a quoted string.

        :return: str, Display message.
        :see: manual p96
        """
        return self.ask("DISP:TEXT?")


    def setDisplayClear(self):
        """ Clear the message displayed on the front panel.

        :see: manual p96
        """
        self.tx("DISP:TEXT:CLE")


    def setOutputOn(self, state):
        """ Enable or disable the outputs of the power supply.

        :param state: bool, Output state.
        :see: manual p96
        """
        typing.check_class(state, bool)
        self.tx("OUTP {}".format(['OFF', 'ON'][state]))


    def getOutputOn(self):
        """ Query the output state of the power supply.

        :return: bool, Output state.
        :see: manual p96
        """
        return bool(int(self.ask("OUTP?")))


def main():
    """ User interface program for configuring a power supply. """

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('host', nargs='?',
                        type=str,
                        help='hostname of the power supply '
                             '(default: %(default)s)',
                        default=DEFAULT_HOST)
    parser.add_argument('port', nargs='?',
                        type=int,
                        help='SCPI port number of the power supply '
                             '(default: %(default)s)',
                        default=DEFAULT_PORT)
    parser.add_argument('gpibAddress', nargs='?',
                        type=int,
                        help='GPIB address of te power supply '
                             '(default: %(default)s)',
                        default=DEFAULT_GPIB_ADDRESS)

    args = parser.parse_args()

    agilent = Agilent3648A(host=args.host,
                           port=args.port,
                           gpibAddr=args.gpibAddress)

    print("Identification: {}".format(agilent.identification().rstrip()))
    print("Prologix version: {}".format(agilent.converterVersion().rstrip()))

    agilent.setApply(1.0, 0.1)
    print('voltage, current: {}'.format(agilent.getApply()))

    agilent.setDisplayOn(False)

    agilent.setVoltage(0.1)
    agilent.setVoltageStep(0.1)
    agilent.setVoltageUp()
    agilent.setVoltageDown()

    agilent.finalize()


if __name__ == '__main__':
    main()
