#!/usr/bin/env python3
"""
Module for interfacing with an 4395A Network/Spectrum/Impendance Analyzer

:note: Only the Network Analyzer functionality is provided
"""

import argparse
import logging
import socket
import time
import gssw.lib.typing  as typing
import gssw.ucelib.gpib as gpib

logger = logging.getLogger(__name__)

__all__ = ['Agilent4395A']

# Identity and serial number of supported network analyzer
# Note that the analyzer with serial number JP1KE00940 (SRON nr 1013) is not supported; the GPIB
# commands for downloading measured data don't work correctly.
# Only the analyzer with serial number MY41102094 (SRON nr 2649) works.
DEVICE_IDENTITY_PREFIX = "HEWLETT-PACKARD,4395A,MY41102094"

# Defaults
DEFAULT_HOST = "192.168.2.95"
DEFAULT_PORT = 1234
DEFAULT_GPIB_ADDRESS = 17

class Agilent4395A(gpib.GpibTcpIp):
    """
    Provides a communication interface to an Agilent 4395A Network Analyzer.

    :note: The network analyzer must be connected via a Prologix Gpib to Ethernet converter.
    :note: Some function docstrings are excerpts from the Agilent 4395A programming manual.
           Capital words correspond to (soft) keys on the instrument's front panel.
    """

    def __init__(self, host, port, gpibAddr):
        """Constructor, for LAN communication"""

        super().__init__(host, port, "analyzer", commandInterval=0.01, processingDelay=0.01,
                         explicitRead=True,       # Request readings from the Prologix converter
                         timeout=4)               # Receiving measurement data can take a while

        if self.isConnected:
            if gpibAddr:
                # Set up the Prologix converter
                self.tx('++auto 0')                   # Don't read response after sending commands
                #                                       Otherwise commands which do not send a
                #                                       response will generate a -420 error on the
                #                                       instrument.
                self.tx('++mode 1')                   # Use Prologix converter in Controller mode
                self.tx('++addr {}'.format(gpibAddr)) # Configure the GPIB address
                self.tx('++eos 2')                    # Append LF to instrument commands
                self.tx('++eoi 1')                    # Use EOI signals after all last characters

            self._checkIdentification()

            # All data reading function assume ASCII-formatted numerical data
            self._setDataFormatAscii()


    def finalize(self):
        """ Cleans up the connection to the Network analyzer and the Prologix converter """
        if self.isConnected:
            self.setLockFrontPanelKeys(False) # Unlock the front panel keys
            self.setLocalMode()               # Exit the remote mode


    def _checkIdentification(self):
        """ Verifies that we are connected to a supported Agilent 4395a.
        Not all types are supported, i.e. the variants with a HP-IB instead of GPIB cannot transfer
        measured data successfully over the GPIB bus. """
        identification = self.identification().strip()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        assert deviceCorrect,\
            "Device with identity '{}' is not supported.".format(identification) +\
            " Only the analyzer with serial number MY41102094 (SRON nr 2649) is supported." +\
            " The analyzer with serial number JP1KE00940 (SRON nr 1013) is not supported because" +\
            " it has issues with sending measured data over the GPIB bus."


    def setBandwidth(self, bandwidth):
        """
        Sets the bandwidth value for IF bandwidth reduction, or sets the IF bandwidth of the list
        sweep table. This command is valid only if the automatic IF bandwidth setting is off by
        BWAUTO OFF command. (Network and impedance analyzers) (IF BW under Bw/Avg,
        or IF BW under Sweep)
        Sets the bandwidth value for the resolution bandwidth reduction, or sets the resolution
        bandwidth of the list sweep table. This command is valid only if the automatic resolution
        bandwidth setting is off by BWAUTO OFF command. (Spectrum analyzer) (RES BW under Bw/Avg,
        or RES BW under Sweep)

        :param bandwidth: float, Bandwidth (in Hz). The value is rounded to the nearest
                                 of the following values {2,10,30,100,300,1000,3000,10000,30000}
        :see: manual, p B-3
        """
        typing.check_class(bandwidth, float)
        self.tx("BW {}".format(bandwidth))


    def setBandwidthAutomaticEnabled(self, enable):
        """
        When log frequency sweeping mode is selected, sets either the automatic or manual IF
        bandwidth ON. (Network and impedance analyzers) (RES BW AUTO man under Bw/Avg)
        When linear frequency sweeping mode is selected, sets either the automatic or manual
        resolution bandwidth ON. (Spectrum analyzer only) (RES BW AUTO man under Bw/Avg)

        :param enable: bool, False=Manually sets the IF bandwidth or resolution bandwidth
                             True =Automatically sets the IF bandwidth or resolution bandwidth
        :see: manual, p B-4
        """
        typing.check_class(enable, bool)
        self.tx("BWAUTO {}".format(int(enable)))


    def setModeNetworkAnalyzer(self):
        """
        Puts the Agilent 4395A in network analyzer mode.
        (NETWORK ANALYZER under MEAS)

        :see: manual, p N-1
        """
        self.tx('NA')


    def isModeNetworkAnalyzer(self):
        """
        Determines whether the Agilent 4395A is in network analyzer mode.

        :see: setModeNetworkAnalyzer
        :see: manual, p N-1
        :return: bool
        """
        return bool(int(self.ask('NA?')))


    def setSweepSpan(self, frequency):
        """
        Sets the span of the sweep parameters. This command is not valid when the list sweeping
        mode is selected.
        When editing a list sweep table, the command sets the span of a segment.

        :see: manual, p S-14
        :param frequency: float, Frequency span (in Hz)
        """
        typing.check_class(frequency, float)
        self.tx('SPAN {}'.format(frequency))


    def getSweepSpan(self):
        """
        Retrieves the span of the sweep parameters.

        :see: setSweepSpan
        :see: manual, p S-14
        """
        return float(self.ask('SPAN?'))


    def setAutomaticSweep(self, enable):
        """
        Sets the automatic or manual sweep time.

        :see: manual, p S-20
        """
        typing.check_class(enable, bool)
        if enable:
            self.tx('SWETAUTO 1')
        else:
            self.tx('SWETAUTO 0')


    def setLockFrontPanelKeys(self, lock):
        """
        Enables or disables the front panel keys and the rotary knob.

        :param lock: bool, Lock state of the front panel keys
                           (True=keys are disabled, False=keys are enabled)
        :see: manual, p D-11, E-2
        """
        typing.check_class(lock, bool)
        if lock:
            self.tx('DSKEY')
        else:
            self.tx('ENKEY')


    def setDisplayEnabled(self, enable):
        """
        Controls whether the LCD display is visible or not.

        :param enable: bool, False=Invisible (only softkey labels are displayed.)
                             True =Visible
        :see: manual, p S-7
        """
        typing.check_class(enable, bool)
        self.tx("SCRN {}".format(int(enable)))


    def getDisplayEnabled(self):
        """
        Checks whether the instrument's LCD display is enabled

        :see: setDisplayEnabled
        :see: manual, p S-7
        """
        return bool(int(self.ask("SCRN?")))


    def setSweepCenter(self, frequency):
        """
        Defines the center value of the sweep range, or the center value of the
        segment to be edited in the list sweep table.

        :param frequency: float, Frequency center (in Hz)
        :see: manual, p C-4
        """
        typing.check_class(frequency, float)
        self.tx('CENT {}'.format(frequency))


    def getSweepCenter(self):
        """
        Retrieves the center value of the sweep range.

        :see: manual, p C-4
        :return: float
        """
        return float(self.ask('CENT?'))


    def _parseRawDataList(self, rawData):
        """
        Converts raw float data to list of floats

        :return: list with floats
        """
        if rawData is not None:
            try:
                return [float(x) for x in rawData.split(',')]
            except ValueError:
                logger.error('Unparsable data: %r', rawData)
        logger.debug('Could not parse empty raw data list')


    def getDataErrorCorrected(self):
        """
        Outputs the error corected data.

        :see: manual, p O-9
        :return: list of complex numbers, 1 complex number per point.
        """
        rawData = self.ask("OUTPDATA?")
        floatData = self._parseRawDataList(rawData)
        # Convert the list with pairs to a list of complex numbers
        iterator = iter(floatData)
        complexData = [complex(i,j) for i,j in zip(iterator, iterator)]
        return complexData


    def getDataTrace(self):
        """
        Outputs DATA TRACE data.

        :note: Auxiliary amplitudes are removed.
        :see: manual, p O-10
        """
        rawData = self.ask('OUTPDTRC?')
        floatData = self._parseRawDataList(rawData)
        if floatData is not None:
            return floatData[::2] # Remove the auxiliary amplitudes from the trace data
        logger.warning('No data received')


    def getLastError(self):
        """
        Outputs the error message in the error queue.

        :see: manual, p O-10
        :return: (code, message)
        """
        return self.ask("OUTPERRO?")


    def setNumberOfPoints(self, nrOfPoints):
        """
        Sets the number of points for the segment, or sets the number of points for the list sweep
        table. (In the spectrum analyzer mode, this command can set the number of points for
        zero span measurement only;can be used to query in the other measurement types.)
        (NUMBER OF POINTS under Sweep)

        :param nrOfPoints: int, [2-801]
        :see: manual, p P-5
        """
        typing.check_class(nrOfPoints, int)
        assert nrOfPoints in range(2, 801+1)
        self.tx("POIN {}".format(nrOfPoints))


    def getNumberOfPoints(self):
        """
        Retrieves the number of points for the segment.

        :see: setNumberOfPoints
        :see: manual, p P-5
        """
        return int(self.ask("POIN?"))


    def setPowerLevel(self, dBm):
        """
        Sets the power level segment by segment, or sets the power level for the list sweep table.
        (POWER under Sweep)
        This command is valid when the linear frequency or log frequency sweeping mode is selected
        in the network and impedance analyzer modes, or when measuring on zero span in the
        spectrum analyzer mode.

        :param dBm: float, [-50, 15] .1 resolution
        :see: manual, p P-7
        """
        typing.check_class(dBm, float)
        assert -50 <= dBm <= 15

        self.tx("POWE {}".format(dBm))


    def getPowerLevel(self):
        """
        Retrieves the current power level.

        :see: setPowerLevel
        :see: manual, p P-7
        """
        return float(self.ask("POWE?"))


    def setAnalyzerToDefaults(self):
        """
        Presets the ANALYZER to the preset default values. See the Operation Manual for the
        default values. The PRES command does not preset the HP Instrument BASIC. (PRESET)

        :see: manual, p P-7
        """
        self.tx("PRES")


    def triggerSingleSweep(self):
        """
        Makes one sweep of the data and returns to the hold mode. (SINGLE under TRIGGER)

        :see: manual, p S-12
        """
        self.tx("SING")


    def triggerContinuous(self):
        """
        Triggers sweep automatically and continuously and the trace is updated with each sweep.
        (CONTINUOUS under TRIG)

        :see: manual, p C-10
        """
        self.tx("CONT")


    def setDisplayText(self, text):
        """
        Sends the string to the title area on the display. (TITLE under DISPLAY)

        :param text: str, 53 characters maximum
        :see: manual, p T-2
        """
        typing.check_class(text, str)
        self.tx('TITL "{}"'.format(text[:53]))


    def setChannelActive(self, channelNr):
        """
        Selects channel 1 or 2 as the active channel. (CHAN 1 or CHAN 2)

        :param channelNr: int, 1 or 2
        :see: manua, p C-4
        """
        assert channelNr in [1,2]
        self.tx("CHAN{}".format(channelNr))


    def isChannelActive(self, channelNr):
        """
        Checks whether a given channel is active on the instrument.

        :see: setChannelActive
        :see: manual, p C-4
        :return: bool
        """
        assert channelNr in [1,2]
        return bool(int(self.ask("CHAN{}?".format(channelNr))))


    def pressKey(self, keyCode):
        """
        Sends the key code for a key or a softkey on the front panel. This is equivalent to actually
        pressing a key. See Figure K-1 for key codes.

        :param keyCode: int, [0,52]
        :see: manual, p K-1
        """
        assert keyCode in range(52+1), "Invalid key code"
        self.tx("KEY {}".format(keyCode))


    def setMeasurementInputs(self, parameter):
        """
        Selects the parameters or inputs to be measured.

        * Network analyzer parameters: A/R, B/R, R, A, B,
        * Refl: FWD S11 [A/R],
        * Trans:FWD S21 [B/R],
        * Trans:REV S12 [A/R],
        * Refl: REV S22 [B/R]).

        :param parameter: str, see list below
        :see: manual, p M-2
        """
        assert parameter in ['AR', 'BR',
                             'R', 'A', 'B',
                             'S11', 'S12',
                             'S21', 'S22'], "Invalid measurement parameter '{}'".format(parameter)
        self.tx("MEAS {}".format(parameter))


    def getMeasurementInputs(self):
        """
        Retrieves the instrument's current measurement type.

        :see: manual, p M-3
        :return: str, Short code for the measurement type
        """
        return self.ask("MEAS?")


    def setAttenuationOnPort(self, port, value):
        """
        Set the Attenuation value on port R/A/B in dB [0, 10, 20, 30, 40, 50]

        :param value: int from the array [0, 10, 20, 30, 40, 50]
        :see: manual, p A-2
        """
        assert port in ['R', 'A', 'B'], "Invalid Port given. Only ports allowed are R/A/B."

        assert value in [0, 10, 20, 30, 40, 50], \
            "Invalid Value for Attenuation. Can enter [0, 10, 20, 30, 40, 50] dB."

        if port == 'R':
            self.tx("ATTR {}".format(value))
        elif port == 'A':
            self.tx("ATTA {}".format(value))
        elif port == 'B':
            self.tx("ATTB {}".format(value))


    def getAttenuationOnPort(self, port):
        """
        Get the Attenuation value on port R/A/B in dB [0, 10, 20, 30, 40, 50]

        :see: manual, p A-2
        :return: int from the array [0, 10, 20, 30, 40, 50] as attenuation dB
        """
        assert port in ['R', 'A', 'B'], "Invalid Port given. Only ports allowed are R/A/B."

        if port == 'R':
            return int(self.ask("ATTR?"))
        elif port == 'A':
            return int(self.ask("ATTA?"))
        elif port == 'B':
            return int(self.ask("ATTB?"))


    def setScreenScaleAuto(self):
        """
        Sets the Scale on the Network Analyzer screen automatically to being the trace data into
        view.

        :see: manual, p A-3
        """
        self.tx("AUTO")


    def setAverageFunction(self, value):
        """
        Turns the Averaging Function ON or OFF for the active channel.

        :param value: True or False
        :see: manual, p A-3
        """
        assert value in [True, False], "Input should be Boolean True/False."

        if value:
            self.tx("AVER {}".format(1))
        else:
            self.tx("AVER {}".format(0))


    def getAverageFunction(self):
        """
        Returns the state of the avergaing function on the active channel as ON/OFF.

        :return: bool
        :see: manual, p A-3
        """
        return bool(self.ask("AVER?"))


    def setAveragingFactor(self, value):
        """
        Set the averaging factor the measurements.

        :param value: integer averaging factor
        :see: manual, p A-4
        """
        assert value in range(1,999), "Value not integer or out of range."
        self.tx("AVERFACT {}".format(value))


    def getAveragingFactor(self):
        """
        Returns the current averaging factor on the active Channel.

        :return: integer averaging factor
        :see: manual, p A-4
        """

        return int(self.ask("AVERFACT?"))


    def resetAveragingFactor(self):
        """
        Resets the sweep-to-sweep averaging and restarts the sweep count at 1 at the beginning of
        the next sweep.

        :see: manual, p A-4
        """
        self.tx("AVERREST")


    def setSweepType(self, parameter):
        """
        Selects the Sweep type for the measurement.

        :param parameter: str from list ['LINF', 'LOGF', 'LIST', 'POWE']
        :see: manual, p S-20
        """
        assert parameter in ['LINF', 'LOGF', 'LIST', 'POWE'], "Invalid Sweet Type."
        self.tx("SWPT {}".format(parameter))


    def getSweepType(self):
        """
        Returns the current sweep type

        :return: str
        :see: manual, p S-20
        """
        return str(self.ask("SWPT?"))


    def getSweepParameterData(self):
        """
        Outputs the sweep parameter data.

        :see: manual, p O-15
        """
        rawData = self.ask('OUTPSWPRM?')
        return self._parseRawDataList(rawData)


    def setSweepStartFreq(self, value):
        """
        Sets the start value of the sweep parameters. This command is not valid when the list
        sweeping mode is selected.

        :param value: integer
        :see: manual, p S-17
        """
        assert value in range(10,500000000), "Input Frequency out of range."
        self.tx("STAR {}".format(value))


    def getSweepStartFreq(self):
        """
        Returns the start value of the sweep parameters. This command is not valid when the list
        sweeping mode is selected.

        :return: integer
        :see: manual, p S-17
        """
        return int(self.ask("STAR?"))


    def setSweepStopFreq(self, value):
        """
        Sets the stop value of the sweep parameters. This command is not valid when the list
        sweeping mode is selected.

        :param value: integer
        :see: manual, p S-18
        """
        assert value in range(10,500000000), "Input Frequency out of range."
        self.tx("STOP {}".format(value))


    def getSweepStopFreq(self):
        """
        Returns the stop value of the sweep parameters. This command is not valid when the list
        sweeping mode is selected.

        :return: integer
        :see: manual, p S-18
        """
        return int(float(self.ask("STOP?")))


    def prepareSweepCompleteSignal(self):
        """
        Clears the bits of status and enable registers so that the service request can be triggered
        when measurement sweep is done.

        :see: manual, p 5-14
        """
        # Clear all bits of Status and Enable Registers
        self.tx('CLES')
        # Prepare SRQ for Signalling the End of Sweep
        # the command *SRE 4 sets the Service Request Enable Register to 00000100 (this enables
        # bit 2 of the Status Byte Register).
        self.tx('*SRE 4')
        # The command ESNB 1 sets the Event Status Enable Register B to 0000000000000001
        # (this enables bit 0 of the Event Status Register B).
        self.tx('ESNB 1')


    def checkSweepComplete(self):
        """
        Polls the SRQ pin to check if the measurement sweep is completed. Returns bool True when
        measurement sweep is done. This pin is read by the Prologix Adapter and the value is
        returned.
        No communication is carried out with the network analyzer.

        :return: bool
        :see: manual, p 5-14
        """
        self.tx('++srq')
        return bool(int(self.rx()))

    def getInstalledOptions(self):
        """
        Queries the options installed.

        :see: manual, p O-2
        """
        return self.ask("*OPT?").strip()


    def deleteFile(self, remoteFilename):
        """
        Deletes a file from the instruments storage.

        :param filename: str, remote filename including extentsion
        :see: manual, p P-10
        """
        assert 1 <= len(remoteFilename) <= 8, "Filename length is incorrect"

        self.tx('PURG "{}"'.format(remoteFilename))
        time.sleep(0.1)


    def _createScreenShot(self, dualChannel, color, remoteFilename, showSoftKeys):
        """
        Creates a screenshot of the LCD, which is saved to a file on the network analyzer.

        :param dualChannel:    bool, Show both channels or only the active channel
        :param color:          bool, Use LCD colors or black/white
        :param remoteFilename: str,  Filename to which the screenshot is saved
        :param showSoftKeys:   bool, Show/hide the softkeys displayed on the LCD in the screenshot
        :see: manual, p D-12, P-8, P-9, P-10, S-4
        """
        typing.check_class(dualChannel, bool)
        typing.check_class(color, bool)
        typing.check_class(remoteFilename, str)
        assert 1 <= len(remoteFilename) <= 8, "Filename length is incorrect"

        if color:
            self.tx('PRIC')     # Sets the print command to color
            self.tx('PRICVARI') # Use the colors as on the screen.
        else:
            self.tx('PRIS')     # Sets the print command to black/white
            self.tx('PRICFIXE') # Sets black/white as colors
        self.tx('DUAC {}'.format(int(dualChannel)))     # Show both channels or the active channel
        self.tx('PRSOFT {}'.format(int(showSoftKeys)))  # Show/hide the softkeys in the screenshot
        self.tx('SAVDTIF "{}"'.format(remoteFilename))  # Take screenshot, save to .tif file
        time.sleep(3.5)                                 # Saving the screenshot takes a while


    def downloadFile(self, remoteFilename, localFilename, deleteAfterDownloading=False):
        """
        Downloads a file from the network analyzer to the PC.

        :param remoteFilename:         str,  Name of the file to download from the instrument
        :param localFilename:          str,  File on the PC to save the downloaded file to.
        :param deleteAfterDownloading: bool, Delete the file from the instrument after downloading
        :see: manual, p C-6, R-1, R-2, R-6
        """
        typing.check_class(remoteFilename, str)
        typing.check_class(localFilename, str)
        typing.check_class(deleteAfterDownloading, bool)

        fileData = b'' # Buffer for the file's contents

        with open(localFilename, "wb") as localFile:
            self.tx('ROPEN "{}"'.format(remoteFilename)) # Open the remote file for downloading
            try:
                while True:
                    # The Agilent 4395A sends the file's contents in blocks of 16384 bytes
                    # A block starts with "#6xxxxxx", where xxxxxx is the length of the block
                    self.tx("READ?")      # Request a new block
                    self.tx("++read eoi") # Read the reply (Prologix)
                    blockData = b''
                    while True:
                        try:
                            blockData += self.sock.recv(self.sockBuffSize) # Read the block data
                            if chr(blockData[-1])=='\n':
                                break
                            time.sleep(.01)
                        except socket.timeout:
                            logger.error("Could not download file due to a network timeout")
                            return
                        except socket.error as e:
                            if e.errno != socket.errno.EWOULDBLOCK:
                                logger.error("Could not download file %s: %s", remoteFilename, e)
                                return
                    lengthRaw = blockData[2:8]  # Extract the block length from the header
                    if not lengthRaw:
                        logger.error("Could not download file, no block length received")
                        return
                    blockLength = int(lengthRaw)

                    if not blockLength:
                        break                   # End of the file is reached, stop requesting blocks
                    fileData += blockData[8:-1] # Add the block's data to the buffer
                if fileData:
                    localFile.write(fileData)   # Write the downloaded data to the file
            finally:
                time.sleep(0.1)
                self.tx('CLOSE')                # Close the remote file
                if deleteAfterDownloading:
                    time.sleep(0.1)
                    self.deleteFile(remoteFilename)
        return bool(fileData)


    def captureScreenShot(self, filename, dualChannel=False, color=True, showSoftKeys=False):
        """
        Creates a screenshot of the LCD display, and downloads it as a .tif image.

        :param filename:     str,  Filename to which the screenshot is saved
        :param dualChannel:  bool, Show both channels or only the active channel
        :param color:        bool, Use LCD colors or black/white
        :param showSoftKeys: bool, Show/hide the softkeys displayed on the LCD in the screenshot
        """
        typing.check_class(filename, str)
        typing.check_class(dualChannel, bool)
        typing.check_class(color, bool)

        remoteFilename = 'sc'
        self._createScreenShot(dualChannel, color, remoteFilename, showSoftKeys)
        return self.downloadFile(remoteFilename + '.tif',
                                 localFilename=filename,
                                 deleteAfterDownloading=True)


    def getFileList(self):
        """
        Retrieves a list of the files on the instrument's storage.

        :see: manual, p F-3
        """
        numberofFiles = int(self.ask('FNUM?'))

        fileList = []
        for i in range(1, numberofFiles+1):
            filename = self.ask('FNAME? {}'.format(i)).strip()
            fileList.append(filename)
        return fileList


    def _setDataFormatAscii(self):
        """
        Configures the instrument to use ASCII formatting when outputting data.
        E.g.:  -1.025737E-004,+1.613693E-005,

        :see: manual, p F-3
        """
        self.tx('FORM4')


def main():
    """ User interface program for configuring a network analyzer """

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('host', nargs='?',
                        type=str,
                        help='hostname of the network analyzer '
                             '(default: %(default)s)',
                        default=DEFAULT_HOST)
    parser.add_argument('port', nargs='?',
                        type=int,
                        help='SCPI port number of the network analyzer '
                            '(default: %(default)s)',
                        default=DEFAULT_PORT)
    parser.add_argument('gpibAddress', nargs='?',
                        type=int,
                        help='GPIB address of te network analyzer '
                            '(default: %(default)s)',
                        default=DEFAULT_GPIB_ADDRESS)

    args = parser.parse_args()


    agilent = Agilent4395A(host=args.host,
                           port=args.port,
                           gpibAddr=args.gpibAddress)

    print("Identification: {}".format(agilent.identification()))
    print("Prologix version: {}".format(agilent.converterVersion()))

    for channelNr in [1,2]:
        active = agilent.isChannelActive(channelNr)
        print("Channel {}: {}".format(channelNr, "Active" if active else "Inactive"))

    errors = agilent.getErrorLog()
    if errors:
        print ("Error log:")
        for code, message in errors:
            print ("Code: {},\tDescription: {}".format(code, repr(message)))
    else:
        print ("Error log is empty")

    for channelNr in [1,2]:
        agilent.setChannelActive(channelNr)
        agilent.eventStatus()
        agilent.clearStatus()
        agilent.setDisplayText('')

    #options = agilent.getInstalledOptions()
    #print("Installed options: {}".format(options if options else 'None'))

    try:
        #agilent.setLockFrontPanelKeys(True)
        agilent.setDisplayText("Testing...")
        #agilent.triggerSingleSweep()
    finally:
        agilent.setLockFrontPanelKeys(False) # unlock the front panel
    #agilent.triggerContinuous()

    # Capture a screenshot, and download it
    print ("Taking screenshot...")
    agilent.captureScreenShot('screenshot.tif', dualChannel=False, color=True)

    # Open the downloaded screenshot
    import webbrowser
    webbrowser.open('screenshot.tif')

    agilent.setLocalMode()
    #fileList = agilent.getFileList()
    #for file in fileList:
    #    print("Downloading:", file)
    #    agilent.downloadFile(file, file)
    agilent.finalize()

if __name__=='__main__':
    main()
