#!/usr/bin/env python3
#pylint: disable=invalid-name
""" Module for interfacing with an Agilent E4418B EPM Series Power Meter. """

import argparse
import logging
import gssw.lib.typing  as typing
import gssw.ucelib.gpib as gpib

logger = logging.getLogger(__name__)

__all__ = ['Agilent4418B']

DEVICE_IDENTITY_PREFIX = "HEWLETT-PACKARD,E4418B"

# Defaults
DEFAULT_HOST = "192.168.1.201"
DEFAULT_PORT = 1234
DEFAULT_GPIB_ADDRESS = 13


class Agilent4418B(gpib.GpibTcpIp):
    """ Provides a communication interface to an Agilent 4418B EPM Series Power Meter.

    :note: The power meter must be connected via a Prologix Gpib to Ethernet converter.
    :note: Some function docstrings are excerpts from the Agilent 4418B programming manual.
           Capital words correspond to (soft) keys on the instrument's front panel.
    """

    def __init__(self, host, port, gpibAddr):
        """ Constructor, for LAN communication. """

        super().__init__(host, port, "analyzer", commandInterval=0.01, processingDelay=0.01,
                         explicitRead=True,       # Request readings from the Prologix converter
                         timeout=4)               # Receiving measurement data can take a while

        if self.isConnected:
            if gpibAddr:
                # Set up the Prologix converter
                self.tx('++auto 0')                   # Don't read response after sending commands
                #                                       Otherwise commands which do not send a
                #                                       response will generate a -420 error on the
                #                                       instrument.
                self.tx('++mode 1')                   # Use Prologix converter in Controller mode
                self.tx('++addr {}'.format(gpibAddr)) # Configure the GPIB address
                self.tx('++eos 2')                    # Append LF to instrument commands
                self.tx('++eoi 1')                    # Use EOI signals after all last characters

            self._checkIdentification()


    def finalize(self):
        """ Cleans up the connection to the power meter and the Prologix converter. """
        if self.isConnected:
            self.setDisplayEnabled(True)


    def _checkIdentification(self):
        """ Verifies that we are connected to a Agilent 4418B. """
        identification = self.identification().strip()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        assert deviceCorrect,\
            "Device with identity '{}' is not supported.".format(identification)


    def zeroing(self, channelNr):
        """ Zeroing adjust the power meter's specified channel for a zero power reading with
            no power applied to the power sensor. This method is non-blocking.

        :param channelNr: int, Channel number, 1 or 2.
        """
        assert channelNr in [1, 2]
        self.tx("CAL{}:ZERO:AUTO ONCE".format(channelNr))


    #pylint: disable=too-many-arguments
    def setConfiguration(self, windowNr, mode=None, expectedVal=None,
                         resolution=None, channelNr=None):
        """ This method is used on the specified window to set: the expected power
            level value of the measurement, the resolution of the window, the channels on
            which the measurement is to be made.

        :param windowNr: int, Window number, 1 or 2.
        :param mode: str, Measurement mode, None=Default, 'REL', 'DIFF', 'RAT'
                          or 'RAT:REL'.
        :param expectedVal: float, Expected value in given units.
        :param resolution: float, Resolution, None=Default,
                                  1.0, 0.1, 0.01, 0.001.
        :param channelNr: Channel number, 1 or 2, or list with 2 channels
                          for ratio and relative measurements.
        :see: manual p2-8

        """
        assert windowNr in [1, 2]
        assert mode in [None, 'REL', 'DIFF', 'RAT', 'RAT:REL']
        assert resolution in [None, 1.0, 0.1, 0.01, 0.001]

        cmd = "CONF{}:SCAL:POW:AC".format(windowNr)

        # Mode.
        if mode is None:
            cmd += " "
        else:
            cmd += ":" + mode + " "

        # Expected value.
        if expectedVal is None:
            cmd += "DEF,"
        else:
            cmd += str(expectedVal) +','

        # Resolution.
        if resolution is None:
            cmd += "DEF"
        else:
            cmd += str([1.0, 0.1, 0.01, 0.001].index(resolution) + 1)

        # Channel number.
        if channelNr in [1, 2]:
            cmd += ",(@{})".format(channelNr)
        elif isinstance(channelNr, list):
            if channelNr[0] in [1, 2]:
                cmd += ",(@{})".format(channelNr[0])
            if channelNr[1] in [1, 2]:
                cmd += ",(@{})".format(channelNr[1])

        self.tx(cmd)


    def getPower(self, windowNr):
        """ Get a power measurement on specified window.

        :param windowNr: int, Window number, 1 or 2.
        :return: float, Measured value.
        """
        assert windowNr in [1, 2]
        conf = self.ask("CONF{}?".format(windowNr))
        if '(@1)' in conf:
            self.tx("INIT1:CONT ON")
            self.tx("ABORT1")
        if '(@2)' in conf:
            self.tx("INIT2:CONT ON")
            self.tx("ABORT2")
        return float(self.ask("FETCH{}?".format(windowNr)))


    def setUnit(self, windowNr, unit):
        """ This method is used to set the power meter measurement
            unit to Watts and dBm or for ratio measurements % and dB.

        :param windowNr: int, Window number, 1 or 2.
        :param unit: str, Unit, 'W', 'dBm' or ratio unit '%', 'dB'.
        :see: manual p13-2

        """
        assert windowNr in [1, 2]
        assert unit in ['W', 'dBm', '%', 'dB']
        txUnit = ['W', 'DBM', 'PCT', 'DB'][['W', 'dBm', '%', 'dB'].index(unit)]
        if unit in ['W', 'dBm']:
            self.tx("UNIT{}:POW {}".format(windowNr, txUnit))
        elif unit in ['%', 'dB']:
            self.tx("UNIT{}:POW:RAT {}".format(windowNr, txUnit))


    def setFrequency(self, channelNr, freq):
        """ This method is used to enter a frequency. If the frequency does not
            correspond directly to a frequency in the sensor calibration table, the
            power meter calculates the calibration factor using linear interpolation.

        :param channelNr: int, Channel number, 1 or 2.
        :param freq: Calibration frequency, None=Default, float, 'MIN', 'MAX'
        :see: manual p9-37
        """
        assert channelNr in [1, 2]
        if freq not in [None, 'MIN', 'MAX']:
            typing.check_class(freq, float)
        if freq is None:
            freq = 'DEF'
        self.tx("SENS{}:FREQ:CW {}".format(channelNr, freq))


    def getFrequency(self, channelNr):
        """ This method gets the frequency.

        :param channelNr: int, Channel number, 1 or 2.
        :return: Frequency.
        :see: manual p9-37
        """
        assert channelNr in [1, 2]
        return float(self.ask("SENS{}:FREQ:CW?".format(channelNr)))


    def setAverageCount(self, channelNr, count=None):
        """ This method is used to enter a value for the filter length. Automatic
            averaging will be disabled. Increasing the value of filter length
            increases measurement accuracy, but also increases the time taken to
            make a power measurement.

        :param channelNr: int, Channel number, 1 or 2.
        :param count: Filter length, None=Default, int, 'MIN', 'MAX'
        :see: manual p9-5
        """
        assert channelNr in [1, 2]
        if count not in [None, 'MIN', 'MAX']:
            typing.check_class(count, int)
        if count is None:
            count = 'DEF'
        self.tx("SENS{}:AVER:COUN {}".format(channelNr, count))


    def getAverageCount(self, channelNr):
        """ This method gets the filter length.

        :param channelNr: int, Channel number, 1 or 2.
        :return: Filther length.
        :see: manual p9-5
        """
        assert channelNr in [1, 2]
        return int(self.ask("SENS{}:AVER:COUN?".format(channelNr)))


    def setAverageAuto(self, channelNr, auto):
        """ This method is used to enable and disable automatic averaging. When the
            auto filter mode is enabled, the power meter automatically sets the number
            of readings averaged together to satisfy the averaging requirements for most
            power measurements.

        :param channelNr: int, Channel number, 1 or 2.
        :param auto: bool, Auto averaging.
        :see: manual p9-7
        """
        typing.check_class(auto, bool)
        self.tx("SENS{}:AVER:COUN:AUTO {}".format(channelNr, int(auto)))


    def setDisplayEnabled(self, enable):
        """ Controls whether the LCD display is visible or not.

        :param enable: bool, False=Invisible
                             True=Visible
        :see: manual p5-5
        """
        typing.check_class(enable, bool)
        self.tx("DISP:ENAB {}".format(int(enable)))


    def getDisplayEnabled(self):
        """ Checks whether the instrument's LCD display is enabled.

        :return: bool, LCD enabled
        :see: manual p5-5
        """
        return bool(int(self.ask("DISP:ENAB?")))


def main():
    """ User interface program for configuring a power meter. """

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('host', nargs='?',
                        type=str,
                        help='hostname of the power meter '
                             '(default: %(default)s)',
                        default=DEFAULT_HOST)
    parser.add_argument('port', nargs='?',
                        type=int,
                        help='SCPI port number of the power meter '
                             '(default: %(default)s)',
                        default=DEFAULT_PORT)
    parser.add_argument('gpibAddress', nargs='?',
                        type=int,
                        help='GPIB address of te power meter '
                             '(default: %(default)s)',
                        default=DEFAULT_GPIB_ADDRESS)

    args = parser.parse_args()

    agilent = Agilent4418B(host=args.host,
                           port=args.port,
                           gpibAddr=args.gpibAddress)

    print("Identification: {}".format(agilent.identification().rstrip()))
    print("Prologix version: {}".format(agilent.converterVersion().rstrip()))

    agilent.setDisplayEnabled(False)

    agilent.setAverageAuto(1, True)
    agilent.setUnit(1, 'dBm')
    agilent.setFrequency(1, 10e6)
    agilent.setConfiguration(1, None, None, 0.1, 1)
    print('power: %f' % agilent.getPower(1))

    agilent.finalize()


if __name__ == '__main__':
    main()
