#!/usr/bin/env python3
#pylint: disable=invalid-name

"""SRON generic EGSE Support for the Agilent 8257D programmable source generator"""

import time
import logging
import numpy as np
import gssw.ucelib.gpib as gpib

__all__ = ['Agilent8257D']

###############################################################################
# C O N S T A N T S
###############################################################################
DEFAULT_HOST = "192.168.2.94"
DEFAULT_PORT = 5025

logger = logging.getLogger(__name__)

###############################################################################
# A G I L E N T   E 8 2 5 7 D   S Y N T H E Z I S E R
###############################################################################

class Agilent8257D(gpib.GpibTcpIp):
    """Control for the Agilent programmable source generator. Note:
       only LAN support is implemented!

       example usage:

       try:
           psg = Agilent8257D()

           if psg.xxxx:
               xxx
           else:
               xxx

        except IOError as inst:
            print(inst.args[0])
        except NameError as inst:
            print(inst.args[0])
    """

    def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT):
        """Constructor, for LAN communication"""
        super().__init__(host, port, "synthesizer")
        self.tx("*ESE 1\n") # ebable op complete in standard event status register


    def opReady(self):
        """Send SRQ query"""
        msg  = "STAT:OPER:EVEN?\n"
        done = False

        while not done:
            self.tx(msg)
            rsp = self.rx()

            if rsp is None:
                logger.error("No response from %s (STAT:OPER:EVEN?)", self.device)
                return

            value = int(rsp.rstrip())
            if value & 0x01:
                done = True

        logger.debug("Operation result: %s", rsp.rstrip())


    def setFreq(self, f):
        """Send frequency (Hz) to Agilent"""
        logger.debug("Setting frequency to %f Hz", f)
        msg = ":FREQ:FIX %f HZ\n" % f
        self.tx(msg)


    def getFreq(self):
        """Get current frequency (Hz) from Agilent"""
        msg = ":FREQ?\n"
        self.tx(msg)
        rsp = self.rx()

        if rsp is None:
            logger.error("No response from synthesizer (:FREQ?)")
            return None

        f = float(rsp.rstrip())
        logger.debug("Read frequency %f Hz", f)
        return f


    def setPower(self, p):
        """Send power (dBm) to Agilent"""
        logger.debug("Setting power to %f dBm", p)
        msg = ":POW %0.5f DBM\n" % p
        self.tx(msg)


    def getPower(self):
        """Get current power (dBm) from Agilent"""
        msg = ":POW?\n"
        self.tx(msg)
        rsp = self.rx()

        if rsp is None:
            logger.error("No response from synthesizer (:POW?)")
            return None

        p = float(rsp.rstrip())
        logger.debug("Read power %.2f dBm", p)
        return p


    def setSwitchOnOff(self, m):
        """Set switch mode to Agilent"""
        logger.debug("Setting mode to %d", m)
        msg = ":OUTP %d" % m
        self.tx(msg)


    def getSwitchOnOff(self):
        """Get switch mode from Agilent"""
        msg = ":OUTP?\n"
        self.tx(msg)
        rsp = self.rx()

        if rsp is None:
            logger.error("No response from synthesizer (:OUTP?)")
            return None

        s = int(rsp.rstrip())
        logger.debug("Read switch mode %d", s)
        return s


    def getManual(self):
        """Gets manual settings"""
        msg = "LIST:MODE?\n"
        self.tx(msg)
        rsp = self.rx()

        if rsp != None:
            logger.debug("List manual = %s", rsp.rstrip())

        msg = "FREQ:MODE?\n"
        self.tx(msg)
        rsp = self.rx()

        if rsp != None:
            logger.debug("Freq manual = %s", rsp.rstrip())

        msg = "SWE:MODE?\n"
        self.tx(msg)
        rsp = self.rx()

        if rsp != None:
            logger.debug("Swep manual = %s", rsp.rstrip())


    def startFrequencySweep(self, startF, stopF, nSteps, dwellTime):
        """Perform a frequency sweep from start frequency to stop frequency
           in nSteps. At each step, wait dwellTime.

           startF    start frequency (in GHz)
           stopF     stop frequency (in GHz)
           nSteps    number of steps
           dwellTime dwell time per step (in s/step)
        """

        logger.debug("Performing frequency sweep")
        logger.debug("- from %f GHz", startF)
        logger.debug("- to   %f GHz", stopF)
        logger.debug("- in   %d steps", nSteps)
        logger.debug("- with %s s dwell time", dwellTime)

        # perfrom the following in a try - except block

        self.tx("ABOR\n")                 # abort all previous actions
        self.tx("*CLS\n")                 # clear status byte register
        self.tx("FREQ:MODE FIX\n")        # Sets the sig gen frequency mode to FIX
                                          # LIST mode is not supported by 8257D of KID team!

        for freq in np.arange(startF, stopF, (stopF-startF)/nSteps):
            self.setFreq(freq*1.0e9)
            time.sleep(dwellTime)

        logger.debug("End of sweep")


    def abort(self):
        """Abort all previous actions"""
        self.tx("ABOR")


    def error(self):
        """Get last error from agilent"""

        rsp = self.ask("SYST:ERR?")
        if rsp is None:
            logger.error("No response from %s (*SYST:ERR)", self.device)
            return

        if rsp.rstrip() == "+0,\"No error\"":
            return

        logger.debug("Last error: %s", rsp.rstrip())


    def optionIdentificationQuery(self):
        """Queries the options included in the instrument and returns a list of
        the options installed. The options are separated from each other by means
        of commas."""

        rsp = self.ask("*OPT?")
        if rsp is None:
            logger.error("No response from %s (*OPT?)", self.device)
            return None

        logger.debug("OPT result: %s", rsp)
        return rsp


    def individualStatus(self):
        """Sends Individual Status Query. Returns the contents of the IST flag
        (status bit during a parallel poll)."""

        rsp = self.ask("*IST?")
        if rsp is None:
            logger.error("No response from %s (*IST?)", self.device)
            return None

        logger.debug("IST result: %d", rsp)
        return rsp


    def calibration(self):
        """Send calibration query"""
        rsp = self.ask("*CAL?")
        if rsp is None:
            logger.error("No response from %s (*CAL?)", self.device)
        elif rsp > 0:
            logger.error("Calibration of %s failed with error %d", self.device, rsp)
        else:
            logger.debug("Calibration of %s succeeded", self.device)
