#!/usr/bin/python3

"""
Support the AVS47 (and AVS47A, AVS47B) resistance bridge,
directly connected via serial port (or USB serial port adapter).

The AVS47 computer interface uses the Picobus protocol, a proprietary
interface protocol from PicoWatt. There are two ways to access the
Picobus from a PC:
 - via additional interface adapters from PicoWatt (for example,
   Picobus-to-GPIB adapter);
 - via direct connection of the Picobus port to the RS-232 port of the computer
   (possibly via a RS-232-to-USB converter).

This module supports only direct connection to the Picobus via the serial port.
This approach is not recommended and scarcely documented by PicoWatt.
However, direct serial-port commanding works well in practice and avoids
endless issues caused by additional protocol converters.

The low-level Picobus protocol is documented by PicoWatt
in AVS47B_Picobus_Manual_20131011
  http://www.picowatt.fi/interfacing/usb_picobus/picobus_description.pdf

The Picobus does not use the normal asynchronous communication features
of the RS-232 port; the RX and TX lines are not even connected. Instead,
the Picobus uses 4 modem control lines which are explicitly manipulated
by PC software for each sent/received bit.

The RS-232 port is wired to the Picobus port as follows:
 RS-232 connector (DE9S)  |  AVS-47B connector (DB25P)
 (not connected)          |  pin 1  = cable shield
 pin 7 = RTS (to AVS)     |  pin 4  = CP Clock Pulse
 pin 8 = CTS (from AVS)   |  pin 5  = DI Data from Instrument
 pin 6 = DSR (from AVS)   |  pin 6  = AL Alarm Status Line
 pin 5 = GND              |  pin 7  = GND
 pin 4 = DTR (to AVS)     |  pin 20 = DC Data from Computer

This Python file can be used as a library module as follows:
    import ucelib.avs47serial
    avs = ucelib.avs47serial.Avs47Serial(port='/dev/ttyS0')
    r = avs.measure()

Alternatively, this Python file can be used as a stand-alone program:
    python3 -m ucelib.avs47serial --port /dev/ttyS0 --measure
"""

import sys
import argparse
import logging
import math
import time
import serial

__all__ = [ 'Avs47Serial' ]


logger = logging.getLogger('ucelib.avs47serial')


class AvsConfig():
    """ Helper class representing AVS-47 configuration parameters. """

    __slots__ = ( 'inp', 'chan', 'disp', 'exc', 'rng', 'rem' )

    def __init__(self, inp, chan, disp, exc, rng, rem):
        self.inp    = inp       # INPUT
        self.chan   = chan      # CHANNEL
        self.disp   = disp      # DISPLAY
        self.exc    = exc       # EXCITATION
        self.rng    = rng       # RANGE
        self.rem    = rem       # REMOTE


class Avs47Serial():
    """ Encapsulates the AVS-47 resistance bridge. Connects to the AVS via RS-232. """

    # Wait interval (seconds) between bits on the Picobus.
    BIT_TIME            = 0.00125

    # Constants for setting INPUT mode.
    INPUT_ZERO          = 0
    INPUT_MEAS          = 1
    INPUT_CAL           = 2

    # Constants for setting DISPLAY mode.
    DISPLAY_R           = 0
    DISPLAY_DR          = 1
    DISPLAY_ADJ_REF     = 2
    DISPLAY_REFERENCE   = 3
    DISPLAY_EXC_VOLT    = 4
    DISPLAY_HEAT_V      = 5
    DISPLAY_HEAT_I      = 6
    DISPLAY_SET_PT      = 7

    # Constants for setting EXCITATION level.
    EXCITATION_ZERO     = 0
    EXCITATION_3uV      = 1
    EXCITATION_10uV     = 2
    EXCITATION_30uV     = 3
    EXCITATION_100uV    = 4
    EXCITATION_300uV    = 5
    EXCITATION_1mV      = 6
    EXCITATION_3mV      = 7

    # Tables for converting configuration value to descriptive string.
    INPUT_STR           = [ 'ZERO', 'MEAS', 'CAL' ]
    DISPLAY_STR         = [ 'R', 'DR', 'ADJ_REF', 'REFERENCE', 'EXC_VOLT',
                            'HEAT_V', 'HEAT_I', 'SET_PT' ]
    EXCITATION_STR      = [ '0', '3uV', '10uV', '30uV', '100uV', '300uV',
                            '1mV', '3mV' ]
    EXCITATION_UV       = [ 0, 3, 10, 30, 100, 300, 1000, 3000 ]
    RANGE_STR           = [ 'OPEN', '2 Ohm', '20 Ohm', '200 Ohm', '2 kOhm',
                            '20 kOhm', '200 kOhm', '2 MOhm' ]

    def __init__(self, port, addr=1):
        """ Open RS-232 port to AVS-47.

        :param str port: device name, e.g. '/dev/ttyS0' or /dev/ttyUSB0'
        :param int addr: Picobus device address
        """

        logger.info('Connecting to AVS-47 via %r (addr %d)', port, addr)

        self.port = port
        self.addr = addr

        self.cfg = AvsConfig(inp=self.INPUT_ZERO,
                             chan=0,
                             disp=self.DISPLAY_R,
                             exc=self.EXCITATION_ZERO,
                             rng=0,
                             rem=0)

        # Open serial port.
        self.serdev = serial.Serial(port=port)

        # Set CP=0, DC=0.
        self.serdev.setRTS(0)
        self.serdev.setDTR(0)


    def close(self):
        """ Close serial port.

        Further use of this object after calling .close() is not possible.

        It is recommended to set the AVS-47 back to LOCAL mode before closing.
        If a new instance of this class connects to the instrument while
        already in REMOTE mode, the first access will clear all settings.
        """

        self.serdev.close()
        self.serdev = None

    def setRemote(self, enable):
        """ Enable or disable REMOTE mode.

        In remote mode, the front-panel switches are locked and
        the configuration can be changed by the PC.

        :param bool enable: remote setting
        """

        if not enable or not self.cfg.rem:

            logger.debug('Setting LOCAL mode')

            # Either switching to LOCAL mode,
            # or switching to REMOTE mode from LOCAL mode.
            # Perform a send/receive cycle in LOCAL mode to
            # obtain the current settings.
            (cfg, _, _) = self._doSendRecv(rem=0)

            # Update current configuration.
            self.cfg = cfg

        self.cfg.rem = 1 if enable else 0

        if enable:

            logger.debug('Setting REMOTE mode')

            # Switch to REMOTE mode.
            self._doSendRecv(rem=1)

    def getRemote(self):
        """ Get the AVS remote/local setting

        :return: remote setting; REMOTE = True, LOCAL = False
        :rtype: bool
        """

        (cfg, _, _) = self._doSendRecv()

        return cfg.rem != 0

    def setInput(self, idxInput):
        """ Set the input source.

        This is only useful in REMOTE mode.

        :param int idxInput: One of the INPUT_xxx constants [0 - 2].
        """

        assert isinstance(idxInput, int) and idxInput in range(3)

        logger.debug('Setting input mode %d (%s)',
                  idxInput, self.INPUT_STR[idxInput])

        self.cfg.inp = idxInput
        self._doSendRecv()

    def getInput(self):
        """ Get input mode and return one of the INPUT_xxx constants.

        :return: input setting value
        :rtype: int
        """

        (cfg, _, _) = self._doSendRecv()

        return cfg.inp

    def setChannel(self, channel):
        """ Set input channel.

        This is only useful in REMOTE mode.

        :param int channel: Input channel number in [0 - 7]
        """

        assert isinstance(channel, int) and channel in range(8)

        logger.debug('Setting channel %d', channel)

        self.cfg.chan = channel
        self._doSendRecv()

    def getChannel(self):
        """ Get input channel and return channel number in [0 - 7]

        :return: channel number
        :rtype: int
        """

        (cfg, _, _) = self._doSendRecv()

        return cfg.chan

    def setDisplay(self, idxDisplay):
        """ Set display mode.

        This is only useful in REMOTE mode.

        :param int idxDisplay: One of the DISPLAY_xxx constants in [0 - 7].
        """

        assert isinstance(idxDisplay, int) and idxDisplay in range(8)

        logger.debug('Setting display to %d (%s)', idxDisplay, self.DISPLAY_STR[idxDisplay])

        self.cfg.disp = idxDisplay
        self._doSendRecv()

    def getDisplay(self):
        """ Get display mode and return one of the DISPLAY_xxx constants.

        :return: display setting value
        :rtype: int
        """

        (cfg, _, _) = self._doSendRecv()

        return cfg.disp

    def setRange(self, idxRange):
        """ Set measurement range.

        This is only useful in REMOTE mode.
        The actual range is 0.2 * 10**idxRange, unless idxRange == 0.

        :param int idxRange: 0 = no range (open), 1 = 2 Ohm, 3 = 20 Ohm, ..., 7 = 200 MOhm
        """

        assert isinstance(idxRange, int) and idxRange in range(8)

        logger.debug('Setting range %d (%s)', idxRange, self.RANGE_STR[idxRange])

        self.cfg.rng = idxRange
        self._doSendRecv()

    def getRange(self):
        """ Get actual range and return range index.

        :return: range setting value
        :rtype: int
        """

        (cfg, _, _) = self._doSendRecv()

        return cfg.rng

    def setExcitation(self, idxExcitation):
        """ Set excitation.

        This is only useful in REMOTE mode.

        :param int idxExcitation: One of the EXCITATION_xxx constants in range [0 - 7].
        """

        assert isinstance(idxExcitation, int) and idxExcitation in range(8)

        logger.debug('Setting excitation %d (%s)',
                     idxExcitation, self.EXCITATION_STR[idxExcitation])

        self.cfg.exc = idxExcitation
        self._doSendRecv()

    def getExcitation(self):
        """ Get actual excitation and return EXCITATION_xxx constant.

            :return: excitation setting value
            :rtype: int
        """

        (cfg, _, _) = self._doSendRecv()

        return cfg.exc

    def measureRaw(self, timeout=1.0):
        """ Measure resistance and return raw ADC code.

        Return (adc, ovr, idxRange)
        where adc is the raw adc code in range -19999 .. 19999;
        ovr is 1 if the AVS-47 indicates overrange, otherwise 0;
        idxRange is the current range index.

        :raises IOError: if the AVS-47 does not respond within timeout seconds.
        :return: a tuple containing the ADC value, overflow indication bit and range setting.
        :rtype: (int, int, int)
        """

        # Wait until measurement finished.
        if not self.waitReady(timeout=timeout):
            raise IOError('No response from AVS-47')

        # Read most recent measurement.
        (cfg, adc, ovr) = self._doSendRecv()

        return (adc, ovr, cfg.rng)

    def measure(self, timeout=1.0):
        """ Measure resistance and return resistance in Ohm.

        :return: measured resistance
        :rtype: float
        :raises IOError: if the AVS-47 does not respond within timeout seconds.
        """

        (adc, _, rng) = self.measureRaw(timeout=timeout)

        return adc * 10.0**(rng - 5)

    def waitReady(self, timeout=1.0):
        """ Wait until the AVS-47 indicates conversion complete.

        :return: if the AVS-47 indicates conversion completed succesfully
        :rtype: bool
        """

        logger.debug('Waiting for AL signal (DSR=1)')

        niter = 1 + int(timeout / 0.01)

        for _ in range(niter):
            if self.serdev.getDSR():
                return True
            time.sleep(0.01)

        return False

    def _doSendRecv(self, rem=None):
        """ Send new configuration and receive new conversion result.

        :return: a tuple containing the AVS config, ADC value and overflow indication bit.
        :rtype: (AvsConfig, int, int)
        """

        # Prepare 48-bit configuration word.
        cfg = AvsConfig(inp=self.cfg.inp,
                        chan=self.cfg.chan,
                        disp=self.cfg.disp,
                        exc=self.cfg.exc,
                        rng=self.cfg.rng,
                        rem=self.cfg.rem if rem is None else rem)

        txbits = self._encodeConfig(cfg)

        logger.debug("Sending %s %s",
                  ''.join([ str((self.addr >> p) & 1)
                            for p in range(7, 0, -1) ]),
                  ''.join([ str((txbits >> p) & 1)
                            for p in range(47, 0, -1) ]))

        # Send address.
        rxbits = self._doSendRecvBits(self.addr, 8)

        # Strobe address.
        self._doStrobe()

        # Send and receive data.
        rxbits = self._doSendRecvBits(txbits, 48)

        # Strobe data.
        self._doStrobe()

        logger.debug("Received %s",
                  ''.join([ str((rxbits >> p) & 1)
                            for p in range(47, 0, -1) ]))

        (cfg, adc, ovr) = self._decodeData(rxbits)

        return (cfg, adc, ovr)

    def _doSendRecvBits(self, txbits, nbits):
        """ Send and receive the specified number of bits via the Picobus.

        :return: the response word received from the device
        :rtype: int
        """

        self.serdev.setRTS(0)           # set CP=0
        time.sleep(self.BIT_TIME)

        rxbits = 0
        pos = nbits
        for _ in range(nbits):

            # Read DI line from AVS.
            # NOTE: This must be done BEFORE setting CP=0, otherwise
            # interference from the falling edge on CP may corrupt the
            # signal on DI.
            bit = self.serdev.getCTS()    # CTS = DI
            rxbits = (rxbits << 1) | (1 if bit else 0)

            # Set CP=0 (falling clock edge).
            self.serdev.setRTS(0)       # RTS = CP

            # Write DC line to AVS.
            pos -= 1
            bit = (txbits >> pos) & 1
            self.serdev.setDTR(bit)       # DTR = DC

            # Cycle the clock: 1 ms CP=0, then 1 ms CP=1.
            time.sleep(self.BIT_TIME)
            self.serdev.setRTS(1)       # Set RTS = CP = 1
            time.sleep(self.BIT_TIME)

        # Final falling edge on CP.
        self.serdev.setRTS(0)           # Set RTS = CP = 0

        return rxbits

    def _doStrobe(self):
        """ Strobe the data line while the clock remains low. """

        self.serdev.setRTS(0)           # set CP=0
        self.serdev.setDTR(0)           # set DC=0

        for _ in range(3):
            time.sleep(self.BIT_TIME)
            self.serdev.setDTR(1)       # set DC=1
            time.sleep(self.BIT_TIME)
            self.serdev.setDTR(0)       # set DC=0

    @staticmethod
    def _encodeConfig(cfg):
        """ Encode 48-bit configuration word for AVS-47.

        :return: 48-bit configuration word
        :rtype: int
        """

        # bits 47 .. 32 = DIGITAL REFERENCE / 5 (only when address == 3)
        # bits 31 .. 24 = ADDRESS
        # bits 23 .. 22 = "00"
        # bits 21 .. 20 = INPUT
        # bits 19 .. 17 = CHANNEL
        # bits 16 .. 14 = DISPLAY
        # bits 13 .. 11 = EXCITATION
        # bits 10 ..  8 = RANGE
        # bit 7         = '0'
        # bit 6         = REMOTE
        # bit 5         = '0'
        # bit 4         = DISABLE AL
        # bits 3 .. 0   = "0000"

        word = 0
        word |= (0x0000 << 32)
        word |= (0x00 << 24)
        word |= (cfg.inp << 20)
        word |= (cfg.chan << 17)
        word |= (cfg.disp << 14)
        word |= (cfg.exc << 11)
        word |= (cfg.rng << 8)
        word |= (cfg.rem << 6)
        word |= (0 << 4)

        return word

    @staticmethod
    def _decodeData(rxbits):
        """ Decode 48-bit data word from AVS-47.

        :return: a tuple containing the AVS config, ADC value and overflow indication bit.
        :rtype: (AvsConfig, int, int)
        """

        # bits 47 .. 44 = ??
        # bit 43        = ??
        # bit 42        = ADC OVERRANGE
        # bit 41        = SIGN POSITIVE
        # bit 40        = digit_4
        # bits 39 .. 36 = digit_3
        # bits 35 .. 32 = digit_2
        # bits 31 .. 28 = digit_1
        # bits 27 .. 24 = digit_0
        # bits 23 .. 22 = ??
        # bits 21 .. 20 = INPUT
        # bits 19 .. 17 = CHANNEL
        # bits 16 .. 14 = DISPLAY
        # bits 13 .. 11 = EXCITATION
        # bits 10 ..  8 = RANGE
        # bit 7         = ??
        # bit 6         = REMOTE
        # bit 5         = ??
        # bit 4         = DISABLE AL
        # bits 3 .. 0   = ??

        ovr     = (rxbits >> 42) & 1

        sign    = (rxbits >> 41) & 1
        digit4  = (rxbits >> 40) & 1
        digit3  = (rxbits >> 36) & 15
        digit2  = (rxbits >> 32) & 15
        digit1  = (rxbits >> 28) & 15
        digit0  = (rxbits >> 24) & 15

        adc =  10000 * digit4
        adc +=  1000 * digit3
        adc +=   100 * digit2
        adc +=    10 * digit1
        adc +=     1 * digit0
        adc = adc if sign else (-adc)

        cfg = AvsConfig(
            inp  = (rxbits >> 20) & 3,
            chan = (rxbits >> 17) & 7,
            disp = (rxbits >> 14) & 7,
            exc  = (rxbits >> 11) & 7,
            rng  = (rxbits >> 8)  & 7,
            rem  = (rxbits >> 6) & 1)

        return (cfg, adc, ovr)


def main():
    """ Main program. """
    # pylint: disable=too-many-statements, too-many-boolean-expressions

    parser = argparse.ArgumentParser(
        description='Command-line access to the AVS-47 resistance bridge')

    parser.add_argument('--port', action='store', type=str, required=True,
                        help='Serial port, e.g. /dev/ttyS0')
    parser.add_argument('--addr', action='store', type=int, default=1,
                        help='Picobus address of AVS-47')
    parser.add_argument('--verbose', action='store_true',
                        help='Show debug message')
    parser.add_argument('--show', action='store_true',
                        help='Show current configuration of AVS')
    parser.add_argument('--measure', action='store_true',
                        help='Read one measurement; write resistance in Ohm')
    parser.add_argument('--channel', action='store', type=int,
                        help='Set input channel')
    parser.add_argument('--input', action='store', type=str,
                        choices=('zero', 'meas', 'cal'),
                        help='Set input mode')
    parser.add_argument('--display', action='store', type=str,
                        choices=('r', 'dr', 'adjref', 'ref', 'exc',
                                 'heatv', 'heati', 'setpt'),
                        help='Set display mode')
    parser.add_argument('--excitation', action='store', type=int,
                        choices=(0, 3, 10, 30, 100, 300, 1000, 3000),
                        help='Set excitation in microvolts')
    parser.add_argument('--range', action='store', type=int,
                        choices=(0, 2, 20, 200, 2000, 20000, 200000, 2000000),
                        help='Set range in Ohms')
    parser.add_argument('--average', action='store', type=int,
                        help='Take multiple measurements and show average')

    args = parser.parse_args()

    # Set log levels.
    logging.getLogger('ucelib').setLevel(logging.INFO)
    if args.verbose:
        logging.getLogger('ucelib.avs47serial').setLevel(logging.DEBUG)

    # Send output from root logger to stderr stream.
    rootlog = logging.getLogger('')
    hdlr = logging.StreamHandler(sys.stderr)
    hdlr.setFormatter(logging.Formatter('[%(asctime)s] %(name)-7s %(levelname)-7s: %(message)s'))
    rootlog.addHandler(hdlr)

    # Remove handler installed by ucelib package.
    logging.getLogger('ucelib').handlers = [ ]

    if ( (not args.show) and (not args.measure) and (args.average is None) and
         (args.channel is None) and (args.input is None) and
         (args.display is None) and (args.excitation is None) and
         (args.range is None) ):
        parser.print_help(sys.stderr)
        sys.exit(0)

    # Open connection to AVS.
    avs = Avs47Serial(port=args.port, addr=args.addr)

    # Switch to REMOTE mode before changing settins.
    remote = False
    if (args.input is not None or args.channel is not None or
        args.display is not None or args.excitation is not None or
        args.range is not None):
        remote = True
        avs.setRemote(True)

    # Set input mode.
    if args.input is not None:
        conf = { 'zero': avs.INPUT_ZERO,
                 'meas': avs.INPUT_MEAS,
                 'cal':  avs.INPUT_CAL }[args.input]
        avs.setInput(conf)

    # Set channel.
    if args.channel is not None:
        avs.setChannel(args.channel)

    # Set display mode.
    if args.display is not None:
        conf = { 'r':      avs.DISPLAY_R,
                 'dr':     avs.DISPLAY_DR,
                 'adjref': avs.DISPLAY_ADJ_REF,
                 'ref':    avs.DISPLAY_REFERENCE,
                 'exc':    avs.DISPLAY_EXC_VOLT,
                 'heatv':  avs.DISPLAY_HEAT_V,
                 'heati':  avs.DISPLAY_HEAT_I,
                 'setpt':  avs.DISPLAY_SET_PT }[args.display]
        avs.setDisplay(conf)

    # Set excitation.
    if args.excitation is not None:
        conf = avs.EXCITATION_UV.index(args.excitation)
        avs.setExcitation(conf)

    # Set range.
    if args.range is not None:
        if args.range == 0:
            conf = 0
        else:
            conf = 1 + int(math.log10(args.range))
            assert args.range == 2 * 10**(conf-1)
        avs.setRange(conf)

    # Show current settings.
    if args.show:
        print("Remote:    ", avs.getRemote())
        print("Input:     ", avs.INPUT_STR[avs.getInput()])
        print("Display:   ", avs.DISPLAY_STR[avs.getDisplay()])
        print("Channel:   ", avs.getChannel())
        print("Excitation:", avs.EXCITATION_STR[avs.getExcitation()])
        print("Range:     ", avs.RANGE_STR[avs.getRange()])

    # Measure resistance.
    if args.average is not None:
        rvec = [ ]
        for _ in range(args.average):
            (adc, ovr, rng) = avs.measureRaw()
            res = adc * 10.0**(rng-5)
            logger.info('Measured ADC=%r OVR=%r RANGE=%r (%r)', adc, ovr, rng, res)
            if ovr or rng == 0:
                break
            rvec.append(res)
        if ovr:
            print("OVR")
        elif rng == 0:
            print("OPEN")
        else:
            ravg = sum(rvec) / float(len(rvec))
            print("%.6e" % ravg)
    elif args.measure or args.show:
        (adc, ovr, rng) = avs.measureRaw()
        res = adc * 10.0**(rng-5)
        logger.info('Measured ADC=%r OVR=%r RANGE=%r (%r)', adc, ovr, rng, res)
        if ovr:
            print("OVR")
        elif rng == 0:
            print("OPEN")
        else:
            print("%.4e" % res)

    # Switch back to local mode.
    if remote:
        avs.setRemote(False)

    # Close connection to AVS.
    avs.close()


if __name__ == '__main__':
    main()
