"""
Support for the AVS47(B) resistance bridge and the optionally connected TS530A,
through control with the Prologix Ethernet to GPIB adapter.

 TS530A   (optional, analog PID controller)
    |
 AVS47B   (resistance bridge)
    |
 AVS47-IB (GPIB interface)
    |
 Prologix (ethernet to GPIB adapter)
"""

import logging
import time

from gssw.ucelib.gpib import GpibTcpIp


__all__ = ['Avs47b', 'Avs47b_Ts530a']

logger = logging.getLogger(__name__)


class Avs47b(GpibTcpIp):
    """Encapsulates the AVS47B resistance bridge.
    Connects by Prologix Ethernet to GPIB adapter.
    """

    INPUTS          = ['input', 'measure', 'calibrate']
    MODES           = ['local', 'remote']
    RANGES_STR      = ['Open', '2 ohm', '20 ohm', '200 ohm',
                       '2 kOhm', '20 kOhm', '200 kOhm', '2 MOhm']
    EXCITATIONS_STR = ['0', '3 uV', '10 uV', '30 uV', '100 uV', '300 uV', '1 mV', '3 mV']
    RANGES          = [0, 2, 20, 200, 2000, 20000, 200000, 2000000]
    EXCITATIONS     = [0, 3, 10, 30, 100, 300, 1000, 3000]

    def __init__(self, host, port=1234, gpibDevAddress=20, gpibDevReset=False):
        """
        :param int gpibDevAddress: GPIB address of the prologix adapter.
        :param bool gpibDevReset:  If true the prologix adapter is resetted.
                                   Note that this causes a 5 second delay.
        """

        # base class constructor
        super().__init__(host, port, device='ucelib.avs47b', commandTerminator='\n', timeout=2,
              processingDelay=1, commandInterval=0.2, sockBuffSize=1000*1024,
              oneReadingPerTrigger=True)

        if gpibDevReset:
            # reset prologix ethernet to GPIB adapter
            self.tx('++rst')
            time.sleep(5)

        # configure prologix ethernet to gpib adapter
        #
        # ++mode 1 : controller mode
        # ++addr   : set primary address
        # ++eos 2  : terminate each command with a LF character
        # ++eoi 0  : disable EOI assertion
        # ++auto 1 : automatically address the instrument to talk after sending a command
        #
        self.tx('++mode 1\n' + '++addr ' + str(gpibDevAddress) + '\n++eos 2\n++eoi 0\n++auto 1\n')

        # TODO
        #
        # Prologix GPIB ETHERNET controller can be configured to automatically address instruments
        # to talk after sending them a command in order to read their response. The feature called,
        # Read After Write saves the user from having to issue read commands repeatedly.
        # This command enables or disabled the Read - After - Write feature.
        # In addition, auto command also addresses the instrument at the currently specified
        # address to TALK or LISTEN.
        # ++auto 0 addresses the instrument to LISTEN and
        # ++auto 1 addresses the instrument to TALK
        #
        # LAST SUCCESSFULL TEST I'VE CONDUCTED WAS WITH ++auto 1, PERHAPS THIS IS THE REASON I NEED
        # AN EXTREMELY LARGE TCPIP BUFFER DUE TO THE RECEIVING OF A HUGE AMOUNT OF '\n CHARACTERS
        # WHEN READING RESPONSES OF THE AVS47B. HOWEVER I BELIEVE I'VE TRIED ++auto 0 AND IT WAS
        # NOT SUCCESSFUL. I NEED TO DO MORE TESTS ON THIS.

        self.muxChannel    = None
        self.range         = None
        self.rangeIdx      = None
        self.excitation    = None
        self.excitationIdx = None
        self.outputFile    = None

    def getID(self):
        """Get the AVS47B identifier."""
        return self.ask("*IDN?").strip()

    def reset(self):
        """Reset the device."""
        self.tx('*RST')

    def setRemote(self, idxRemote):
        """Set the remote mode to <idxRemote> Input (0:local or 1:remote)."""
        logger.debug('Set Remote mode %i (%s)', idxRemote, self.MODES[idxRemote])
        self.tx('REM %i' % idxRemote)

    def getRemote(self):
        """Get remote mode, returns <idxRemote> Input (0:local or 1:remote)."""
        rsp = self.ask('REM?').strip()
        idxRemote = int(rsp[4:])
        logger.debug('Get Remote mode: %i (%s)', idxRemote, self.MODES[idxRemote])
        return idxRemote

    def setInput(self, idxInput):
        """Set input source.

        <idxInput> set parameter, represents input source 0, 1 or 2
                   - Input source 0 connects the bridge input to ground.
                   - Input source 1 enables the actual measurement.
                   - Input source 2 connects the bridge to an internal 100 ohm
                     precision reference.
        """
        logger.debug('Set Input %i (%s)', idxInput, self.INPUTS[idxInput])
        self.tx('INP %i' % idxInput)

    def getInput(self):
        """Get input source, returns <idxInput> Input (0, 1 or 2)."""
        rsp = self.ask('INP?').strip()
        idxInput = int(rsp[4:])
        logger.debug('Get Input: %i (%s)', idxInput, self.INPUTS[idxInput])
        return idxInput

    def setChannel(self, channel):
        """Set multiplexer to <channel> (between 0 and 7)."""
        logger.debug('Set Channel %i', channel)
        self.tx('MUX %i' % channel)

    def getChannel(self):
        """Get multiplexer channel, returns <channel> (between 0 and 7)."""
        rsp = self.ask('MUX?').strip()
        channel = int(rsp[4:])
        logger.debug('Get Channel: %i', int(channel))
        return channel

    def setRange(self, idxResRange):
        """Set resistance range to <idxResRange> value (0..7), n=1."""
        logger.debug('Set Range %i (%s)', idxResRange, self.RANGES_STR[idxResRange])
        self.tx('RAN %i' % idxResRange)

    def setAutorange(self, onOff):
        """Set autorange to <onOff> value (0|1), 0=off, 1=on."""
        logger.debug('Set Autorange to %i', onOff)
        self.tx('ARN %i' % onOff)

    def getRange(self):
        """Get resistance range index value (0..7), n = 1.
        In theory it shouldn't be necesary to check for 'RAN' return value,
        however in practice it occured that the AVS returned a different parameter."""
        rsp = self.ask('RAN?').strip()
        if rsp[:3] != 'RAN':
            logger.warning('getRange returned %s value', rsp[:3])
            return None
        else:
            idxResRange = int(rsp[4:])
            logger.debug('Get Range: %i (%s)', idxResRange, self.RANGES_STR[idxResRange])
            return idxResRange

    def setExcitation(self, idxExcitation):
        """Set excitation to <idxExcitation> value (0..7), n=1."""
        logger.debug('Set Excitation %i (%s)', idxExcitation, self.EXCITATIONS_STR[idxExcitation])
        self.tx('EXC %i' % idxExcitation)

    def getExcitation(self):
        """Get excitation index value.
        TODO
        Returns:
            Index of the resistance range (0..7), n = 1.
        """
        rsp = self.ask('EXC?').strip()
        idxExcitation = int(rsp[4:])
        logger.debug('Get Excitation %i (%s)', idxExcitation, self.EXCITATIONS_STR[idxExcitation])
        return idxExcitation

    def measure(self):
        """Measure current resistance.
        Returns:
            Resistance in Ohm.
        """
        self.tx('ADC')
        time.sleep(0.5)
        return self.ask('RES?').strip()[4:]

    def writeNewOutFile(self, outputFileName):
        """Write file header for OUT file into new file.
        TODO
        OUT file format is used by R->T conversion applications.

        Args:
            outputFileName: Filename of new OUT file."""

        self.outputFile = outputFileName
        with open(self.outputFile, 'w') as f:
            f.write(time.strftime('%d %m %Y %k %M %S\n'))
        logger.info('Created new OUT file: %s', self.outputFile)

    def writeResistanceOUTLine(self, resistance):
        """Produce one line containing one resistance measurement.
        TODO
        Args:
            resistance: Resistance value in mOhm"""

        try:
            self.rangeIdx = self.RANGES.index(self.range)
        except ValueError:
            self.rangeIdx = 7

        try:
            self.excitationIdx = self.EXCITATIONS.index(self.excitation)
        except ValueError:
            self.excitationIdx = 7

        with open(self.outputFile, 'a') as f:
            timeStr = time.strftime('%H,%M,%S')
            f.write('%s,%i,%i,%i,%f\n' % (timeStr, self.muxChannel,
                    self.rangeIdx, self.excitationIdx, resistance))

    def setDisplaySelector(self, idx):
        """Set display selector to <idx>

        <idx>   display-item
        -----------------------------------------
        0       R
        1       delta R
        2       adj.ref.
        3       reference
        4       exc.volt
        5       530 heat v
        6       530 heat i
        7       530 set PT
        -----------------------------------------
        """
        logger.debug('Set Excitation %i (%s)', idx, self.EXCITATIONS_STR[idx])
        self.tx('DIS %i' % idx)

    def getDisplaySelector(self):
        """ get display selector """
        rsp = self.ask('DIS?').strip()
        idx = int(rsp[4:])
        return idx


    def operationComplete(self):
        """Poll whether an issued operation is complete.

        Works for autoranging and ADC sample averaging.
        returns a boolea value.
        """

        while True:
            reply = self.ask('*OPC?')
            if reply is None:
                logger.debug('self.ask(\'*OPC?\') timed out')
                time.sleep(0.5)
            else:
                logger.debug('self.ask(\'*OPC?\') returned')
                break

        return bool('1' in reply)

    def waitForOperation(self):
        """Block until the last issued operation is completed."""

        while not self.operationComplete():
            time.sleep(0.5)


class Avs47b_Ts530a(Avs47b): #pylint: disable=invalid-name
    """Encapsulates the TS530A analog PID controller.
    The initial sate of the TS-530A is such that all parameters are in their
    lowest states. Setpoint value may be random until the <ENTER> button is
    pressed (in case of manual operation).

    NB: page 18 of the AVS47-IB manual: If you want to control any one aspect
    of the TS-530A remotely, you must control them all!
    """

    # Pair commands, commands which consist out of respectively a command
    # and a query. NB: The documentation of the TS-530A states: "Allow for a
    # couple of seconds delay between the command and the query.
    PAIR_COMMAND_DELAY = 3

    def __init__(self, host, port=1234, gpibDevAddress=20, gpibDevReset=False):
        # call base class constructor
        super(Avs47b_Ts530a, self).__init__(host=host, port=port, gpibDevAddress=gpibDevAddress,
                gpibDevReset=gpibDevReset)

    def setSetpoint(self, setpoint):
        """Set setpoint voltage to <setpoint> ([0 .. 42000] * 100uV)."""
        logger.debug('set setpoint voltage to %i * 100 = %i [uV]', setpoint, setpoint * 100)
        self.tx('SPT %i' % setpoint)

    def getSetpoint(self):
        """Get setpoint voltage ([0 .. 42000] * 100uV)."""
        rsp = self.ask('SPT?').strip()
        setpoint = int(rsp[4:])
        logger.debug('get setpoint: %i', int(setpoint))
        return setpoint

    def setProportionalGain(self, num):
        """Set proportional gain to <num>

        num decibels  gain  comment
        -----------------------------------------
        15  -         0     ;error forced to zero
        14  60        1000
        10  55        560
        9   50        316
        8   45        180
        7   40        100
        6   35        56
        5   30        31
        4   25        18
        3   20        10
        2   15        6
        1   10        3
        0   5         1.8
        -----------------------------------------
        """

        if not (isinstance(num, int) and num >= 0 and (num < 11 or num == 14 or num == 15)):
            logger.error('invalid gain index value %s', str(num))
            return

        if num >= 0 and num < 11:
            logger.debug('set gain index to %s [dB]', str(5 + num * 5))
        elif num == 14:
            logger.debug('set gain index to 60 [dB]')
        elif num == 15:
            logger.debug('error forced to zero')
        self.tx('PRO %i' % num)

    def getProportionalGain(self):
        """Get gain index (0..11), n=1."""
        rsp = self.ask('PRO?').strip()
        gain = int(rsp[4:])
        logger.debug('Get gain: %i', int(gain))
        return gain

    def forceErrorSignalToZero(self):
        """Forcing the error signal to zero prevents the controller from supplying
        excesive power when the bridge settings are being altered or the sensor
        channel is changes. The SCN procedure uses this comman extensively."""
        self.setProportionalGain(15)

    def setIntegrator(self, i):
        """Set integrator value to <i> (0..11), n=1.

        n   integrator time-constant
        -----------------------------------------
        11  reset (PD-mode)
        10  hold (TI=inf.)
        9   1000 s
        8   500 s
        7   200 s
        6   100 s
        5   50 s
        4   20 s
        3   10 s
        2   5 s
        1   2 s
        0   reset (PD-mode)
        -----------------------------------------
        """
        logger.debug('Set integrator value to %i', i)
        self.tx('ITC %i' % i)

    def getIntegrator(self):
        """Get integrator index value (0..11), n=1."""
        rsp = self.ask('ITC?').strip()
        i = int(rsp[4:])
        logger.debug('Get integrator value: %i', int(i))
        return i

    def setDerivator(self, d):
        """Set derivator value to <i> (0..11), n=1.

        n   derivator time-constant
        -----------------------------------------
        7   100 s
        6   50 s
        5   20 s
        4   10 s
        3   5 s
        2   2 s
        1   1 s
        0   no derivation
        -----------------------------------------
        """
        logger.debug('Set derivator value to %i', d)
        self.tx('DTC %i' % d)

    def getDerivator(self):
        """Get derivator value."""
        rsp = self.ask('DTC?').strip()
        d = int(rsp[4:])
        logger.debug('Get derivator value: %i', int(d))
        return d

    def setPowerBias(self, num):
        """Set power bias.
        -----------------------------------------
        num bias-power  comment
        -----------------------------------------
        5   100%        full range of output power
        4   80%
        3   60%
        2   40%
        1   20%
        0   0
        -----------------------------------------
        """
        logger.debug('Set power-bias index value to %i (= %i%% of the full range output power',
                       num, (0,20,40,60,80,100)[num])
        self.tx('BIA %i' % num)

    def getPowerBias(self):
        """Get power bias, refer to setPowerBias(...) for the actual values."""
        rsp = self.ask('BIA?').strip()
        num = int(rsp[4:])
        logger.debug('Power-bias: %i = %i%% of full range ouptut power.',
                       num, (0,20,40,60,80,100)[num])
        return num

    def setPowerRange(self, num):
        """Set power range. Note that the values below apply for a TS-530A in standard configuration
        and with a heater resistance of 100ohm.
        -----------------------------------------
        num   power-range
        -----------------------------------------
        7   1W
        6   100mW
        5   10mW
        4   1mW
        3   100uW
        2   10uW
        1   1uW
        0   output disabled
        -----------------------------------------
        """
        logger.debug('Set power-range index value to %i', num)
        self.tx('POW %i' % num)

    def getPowerRange(self):
        """Get power range, refer to setPowerRange(...) for the actual values."""
        rsp = self.ask('POW?').strip()
        num = int(rsp[4:])
        logger.debug('Power-range: %i (= %s)',
                       num,
                       ('output disabled','1uW', '10uW','100uW','1mW','10mW','100mW','1W')[num])
        return num

    #
    # Pair commands, commands which consist out of respectively a command and a query.
    # NB: The documentation of the TS-530A states: "Allow for a couple of seconds
    # delay between the command and the query.
    #

    def _pairCommand(self, command):
        self.tx(command)
        time.sleep(self.PAIR_COMMAND_DELAY)
        return self.ask(command + '?').strip()

    def measureSetpointVoltage(self):
        """ measure set point voltage """
        result = self._pairCommand('SPV')[4:].strip()
        logger.debug('Measured setpoint voltage = %s', str(result))
        return result

    def measureHeaterOutputVoltage(self):
        """ meassure heater output voltage """
        result = self._pairCommand('HTV')[4:].strip()
        logger.debug('Measured heater output voltage = %s', str(result))
        return result

    def measureHeaterOutputCurrent(self):
        """ measure heater output current """
        result = self._pairCommand('HTI')[4:].strip()
        logger.debug('Measured heater output current = %s', str(result))
        return result
