""" Support for the Espec temperature cabinet. """

import logging
import serial

__all__ = ['Espec', 'EspecError']

logger = logging.getLogger(__name__)

class EspecError(Exception):
    """ Errors in communication with the temperature cabinet """


class Espec:
    """ Encapsulates the Espec temperature cabinet. Assumes we use a serial connection. """

    TIMEOUT = 1.0

    constant = 'CONSTANT'
    standby = 'STANDBY'
    off = 'OFF'

    def __init__(self, serialdev='/dev/ttyS0'):
        self.ser = serial.Serial(port=serialdev,
                                 timeout=self.TIMEOUT,
                                 baudrate=9600,
                                 parity=serial.PARITY_NONE,
                                 bytesize=serial.EIGHTBITS,
                                 stopbits=serial.STOPBITS_ONE)
        logger.debug('Opened connection to ESPEC via %s', serialdev)

    def _write(self, msg):
        self.ser.write(msg.encode())

    def _read(self):
        msg = self.ser.readline().decode()
        return msg

    def close(self):
        """Close the serial connection"""
        self.ser.close()
        logger.info("Closed serial connection to Espec")

    def version(self):
        """Returns the firmware version, useful for checking the connection"""
        self._write("ROM?\r\n")
        s = self._read()
        if not s:
            raise EspecError('No response from ESPEC')
        s = s.strip()
        return s

    def setPower(self, status):
        """Set the power of the Espec"""

        if status not in [ "ON", "OFF"]:
            raise EspecError("Incorrect status for setPower command to Espec")

        self._write("POWER %s\r\n" % status)
        rsp = self._read()
        if not rsp:
            raise EspecError('No response from ESPEC')
        rsp.strip()
        if not rsp.startswith('OK:'):
            raise EspecError('Error response from ESPEC [%s] (MODE)', rsp)

    def getMode(self):
        """Return the current mode of the Espec"""
        self._write("MODE?\r\n")
        rsp = self._read()
        if not rsp:
            raise EspecError('No response from ESPEC')
        rsp = rsp.strip()
        return rsp

    def setMode(self, mode):
        """Set the operational mode of the Espec"""
        if mode not in [ "OFF", "STANDBY", "CONSTANT", "RUN"]:
            logger.error("Incorrect mode for setMode command to Espec")
            return None

        self._write("MODE,%s\r\n" % mode)
        rsp = self._read()
        if not rsp:
            raise EspecError('No response from ESPEC')
        rsp = rsp.strip()
        if not rsp.startswith('OK:'):
            raise EspecError('Error response from ESPEC [%s] (MODE)', rsp)

    def getMonitor(self):
        """Get some parameters from the Espec cabinet"""
        self._write("MON?\r\n")
        rsp = self._read()
        if not rsp:
            raise EspecError('No response from ESPEC')
        rsp = rsp.strip().split(",")
        return rsp

    def getTemperature(self):
        """Gets the temperature of the Espec cabinet"""
        self._write("TEMP?\r\n")
        rsp = self._read()
        if not rsp:
            raise EspecError('No response from ESPEC')
        rsp = rsp.strip()
        words = rsp.split(',')
        return float(words[0])

    def setTemperature(self, temp):
        """Set temperature setpoint of cabinet."""

        logger.info("Setting temperature setpoint to %f", temp)
        self._write("TEMP,S%f\r\n" % temp)
        rsp = self._read()
        if not rsp:
            raise EspecError('No response from ESPEC')
        rsp = rsp.strip()
        if not rsp.startswith('OK:'):
            raise EspecError('Error response from ESPEC [%s] (TEMP)', rsp)

    def setTemperatureLimits(self, limlow, limhigh):
        """Set maximum allowable temperature range."""

        logger.info("Setting lower temperature limit to %f", limlow)
        self._write("TEMP, L%f\r\n" % limlow)
        rsp = self._read()
        if not rsp:
            raise EspecError('No response from ESPEC')
        rsp = rsp.strip()
        if not rsp.startswith('OK:'):
            raise EspecError('Error response from ESPEC [%s] (TEMP)', rsp)

        logger.info("Setting upper temperature limit to %f", limhigh)
        self._write("TEMP, H%f\r\n" % limhigh)
        rsp = self._read()
        if not rsp:
            raise EspecError('No response from ESPEC')
        rsp = rsp.strip()
        if not rsp.startswith('OK:'):
            raise EspecError('Error response from ESPEC [%s] (TEMP)', rsp)

    def getKeyLock(self):
        """Gets the key lock status"""
        self._write("KEYPROTECT?\r\n")
        rsp = self._read()
        if rsp is not None:
            rsp = rsp.strip()
        return rsp

    def setKeyLock(self, mode):
        """Sets the keylock status to mode (ON or OFF)"""
        if mode not in [ "OFF", "ON" ]:
            logger.error("Incorrect mode for key lock command to Espec")
            return None

        self._write("KEYPROTECT,%s\r\n" % mode)
        rsp = self._read()
        if rsp is not None:
            rsp = rsp.strip()
        return rsp
