#!/usr/bin/env python3
""" Base classes for GPIB-based communication with testing equipment """

import os, time, errno, socket, logging
from abc import ABC, abstractmethod
from gssw.lib.typing import check_class

try:
    import serial
except ImportError:
    print("This module requires the Python serial module, "
          "install it via 'apt install python3-serial'")
    serial = None  #pylint: disable=invalid-name

logger = logging.getLogger(__name__)

__all__ = ['GpibSerial', 'GpibTcpIp']

class Gpib(ABC):
    """ Abstract base class for Gpib classes """

    def __init__(self, device, processingDelay, commandTerminator):
        self.device             = device
        self.processingDelay    = processingDelay
        self.commandTerminator  = commandTerminator

        self.oneReadingPerTrigger = False
        self.explicitRead = False

    def ask(self, command, extraProcessingDelay=0):
        """Send <command> to device and return its response."""
        self.tx(command)
        if self.oneReadingPerTrigger:
            # device must be triggered to perform one reading
            self.tx('++trg')

        if self.explicitRead:
            # read until EOI asserted by instrument
            self.tx('++read eoi')

        delay = self.processingDelay + extraProcessingDelay

        if delay:
            # extra delay before response is read
            time.sleep(delay)

        return self.rx()


    @abstractmethod
    def rx(self):  #pylint: disable=invalid-name
        """Receive data"""
        pass

    @abstractmethod
    def tx(self, msg):  #pylint: disable=invalid-name
        """Transmit data <msg> to device."""
        pass

    #
    # Common (mandatory) GPIB commands
    #
    # According the standard: All SCPI instruments shall implement the common
    # commands declared mandatory by 488.2
    #

    def clearStatus(self):
        """send clear status to set the standard event register (ESR) and the
        event part of the questionable and the operation register to zero"""
        self.tx("*CLS")

    def eventStatusEnable(self, value=None):
        """Sends Event Status Enable. If value is not None (0..255), this value
        will be set, otherwise the value from the event status register will be
        returned"""

        msg = ""
        if value:
            msg = "*ESE %d" % (value & 0xff)
            self.tx(msg)
            return None

        rsp = self.ask("*ESE?")
        if rsp is None:
            logger.error("No response from %s (*ESE?)", self.device)
            return None

        logger.info("ESE result: %d", int(rsp))
        return rsp

    def eventStatus(self):
        """Sends Event Status Query"""

        rsp = self.ask("*ESR?")
        if rsp is None:
            logger.error("No response from %s (*ESR?)", self.device)
            return None

        logger.info("ESR result: %s", rsp)
        return rsp

    def readStatusByte(self):
        """Sends Individual Status Query. Returns the contents of the IST flag
        (status bit during a parallel poll)."""

        rsp = self.ask("*STB?")
        if rsp is None:
            logger.error("No response from %s (*STB?)", self.device)
            return None

        logger.info("STB result: %s", str(rsp))
        return rsp

    def operationComplete(self):
        """Sets bit 0 in the event status register after all preceding commands
        have been executed. This bit can be used to initiate a service request."""
        self.tx("*OPC")

    def operationCompleteQuery(self):
        """Query the status of the operation complete command"""
        return self.ask("*OPC?")

    def identification(self):
        """Send identification query"""
        rsp = self.ask("*IDN?")
        if rsp is None:
            logger.error("No response from %s (*IDN?)", self.device)
            return 'IDN Unknown'
        return rsp

    def reset(self):
        """Sets the instrument to a defined default status. The command essentially
        corresponds to pressing the PRESET key."""
        self.tx("*RST")

    def serviceRequestEnable(self, value=None):
        """Sends Service Request Enable. If value is not None (0..255), this value
        will be set, otherwise the value from the service request register will be
        returned. Note that bit 6 is ignored."""

        msg = ""
        if value:
            msg = "*SRE %d" % (value & 0xff)
            self.tx(msg)
            return None

        rsp = self.ask("*SRE?")
        if rsp is None:
            logger.error("No response from %s (*SRE?)", self.device)
            return None

        logger.info("SRE result: %d", int(rsp))
        return rsp


    def trigger(self):
        """This command causes the instrument to request a single scan. If a scan is
        already in progress, an additional scan is not executed."""
        self.tx("*TRG")

    def test(self):
        """sets the instrument to a defined default status. The command essentially
        corresponds to pressing the PRESET key."""

        rsp = self.ask("*TST?", 15)
        if rsp is None:
            logger.error("No response from %s (*TST?)", self.device)
            return None

        logger.info("TST result: %s", rsp)
        return rsp

    def wait(self):
        """This command causes the instrument to wait."""
        self.tx("*WAI")

    #
    # Useful helper functions 
    #

    def getLastError(self):
        """
        Retrieves and erases, 1 entry from the error queue.
        :return: (int, str) with error code, error message
        """
        return self.ask("SYST:ERR?")

    def getErrorLog(self):
        """ Retrieves and erases all entries from the error log 
        :return: List with entries: (code, message) 
        """

        errors = []
        while True:
            error = self.getLastError()
            if not error:
                break

            # Decode the error entry (example: '+113,Undefined header"' 
            code, msg = error.split(",")
            code = int(code.strip())
            msg  = msg.strip()
            if code==0:
                break
            errors.append((code, msg))
        return errors



class GpibSerial(Gpib):  #pylint: disable=abstract-method
    """Base class for GPIB (IEEE488) devices over Serial"""

    def __init__(self, serialPort, baudrate, device, timeout, processingDelay, commandTerminator):
        """
        Creates a serial connection to a Gpib device.
        :note: The constructor opens the serial connection, calling open() is not required.
        :param serialPort:      str,   Name of the serial port
        :param baudrate:        int,   Baud rate of the serial port
        :param device:          str,   Name of the device, used in log messages
        :param timeout:         float, Timeout for waiting for a reply 
        :param processingDelay: float, Default time waited before receiving a reply to a command
        :param commandTerminator: str, Appended after each transmitted string
        """
        assert serial is not None, "Python serial communication module is not available"
        super().__init__(device=device,
                         processingDelay=processingDelay,
                         commandTerminator=commandTerminator)
        self.timeout = timeout

        self.serial = serial.Serial(timeout=self.timeout,
                                 baudrate=baudrate,
                                 parity=serial.PARITY_NONE,
                                 bytesize=serial.EIGHTBITS,
                                 stopbits=serial.STOPBITS_TWO,
                                 xonxoff = True)

        #By setting the port separately it is not opened during construction of the Serial object
        #If the port cannot be opened we still have a valid Serial object
        #We can still open the port later, after setting a valid port name
        self.serial.port = serialPort
        try:
            self.open()
        except serial.SerialException as ex:
            logger.error("Could not open serial connection to %s: %s", self.device, ex)

    def open(self):
        """"Re-opens the serial connection. Only has to be used after a call to close."""
        if self.serial.isOpen():
            self.serial.close() # Silently close if already opened

        self.serial.open()
        if not self.serial.isOpen():
            raise IOError("Error: Could not open serial connection to {}.".format(self.device))
        else:
            logger.debug("Opened serial connection to %s", self.device)

    def close(self):
        """Close the serial connection"""
        self.serial.close()
        logger.debug("Closed serial connection to %s", self.device)

    def rx(self):
        if self.serial.isOpen():
            response = self.serial.readline()    # read data until \n (LF) or timeout

            if response is not None:
                response = response.strip()       # remove white space
            return response.decode('utf8')

    def tx(self, msg):
        if self.serial.isOpen():
            if isinstance(msg, str):
                if msg[-1] != self.commandTerminator:
                    msg += self.commandTerminator
                msg = msg.encode('utf8')
            self.serial.write(msg)


class GpibTcpIp(Gpib):  #pylint: disable=abstract-method
    """Base class for GPIB (IEEE488) devices over TCP/IP"""

    def __init__(self,  #pylint: disable=too-many-arguments
                 host,
                 port,
                 device="IEEE488",
                 commandTerminator='\n',
                 timeout=0.5,
                 explicitRead=False,
                 oneReadingPerTrigger=False,
                 commandInterval=0,
                 processingDelay=0,
                 sockBuffSize=1024):
        """
        Constructor, for LAN communication.
        Implements common SCPI commands which are mandatory by IEEE 4888.2

        :param host:              Destination IP number.
        :param port:              Destination port.
        :param device:            Device name which is used in log messages.
        :param commandTerminator: Command terminator.
        :param timeout:           Timeout when expecting an answer from device.

        :param explicitRead:   Some instruments (e.g. Keithley2400) generate a
                               "Query Unterminated" or "-420" if they are addressed
                               to talk after sending a command which does not generate
                               a response (non-query commands). In effect the instrument
                               is saying, I have been asked to talk but have nothing
                               to say. The error is often benign and may be ignored.
                               Otherwise the ++read command MUST be used to read
                               the instrument response, In that case explicitRead
                               MUST be initialised to True.

        :param oneReadingPerTrigger:
                               Some instruments (e.g. HP3478A) send data coninuously
                               as soon as it is addressed to talk. (The prologix
                               Ethernet to GPIB adapter crashes on this behavior).
                               If this is the case intialise oneReadingPerTrigger
                               to True. NB: YOU MUST CONFIGURE THE PROLOGIX
                               ETHERNET TO GPIB ADAPTER WITH ++auto 1 IN CASE THIS
                               MODE IS USED.

        :param commandInterval: Command interval delay [seconds].
        :param processingDelay: Delay between send command and ask response (tx / rx) [seconds].

        :param sockBuffSize:   Normally you don't have to change the buffer size,
                               however this is GPIB and there are rarely normal cases.
                               E.g. the AVS47B sends responses which are sometimes
                               extemely large (> 1024 bytes filled with command terminators
                               in which the actual data is hidden somewhere).
        """
        check_class(host, str)
        check_class(port, int)

        super().__init__(device=device,
                         processingDelay=processingDelay,
                         commandTerminator=commandTerminator)

        # Initalise member variables according the given constructor arguments.
        self.host                     = host
        self.port                     = port
        self.timeout                  = timeout
        self.explicitRead             = explicitRead
        self.oneReadingPerTrigger     = oneReadingPerTrigger
        self.commandInterval          = commandInterval
        self.sockBuffSize             = sockBuffSize

        # Intialise member variables
        self.sock                     = None
        self._isConnected             = False

        # try to connect to target
        self.connect()

    def __del__(self):
        """Destructor"""
        self.close()


    @property
    def isConnected(self):
        """ Returns whether a successfull connection to the GPIB device is made """
        return self._isConnected

    def close(self):
        """Closing the connection to the target"""
        if self.sock is not None:
            self.sock.close()
        self.sock = None
        self._isConnected = False

    def connect(self):
        """Connect to the target"""

        if self.sock is not None:
            logger.info("Dropping connection to %s", self.device)
            self.close()

        logger.info("Opening socket for %s", self.device)
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.settimeout(self.timeout)

        logger.info("Try to connect to %s on host %s port %d",
                    self.device, self.host, self.port)

        error    = self.sock.connect_ex((self.host, self.port))
        if error == 0:
            self._isConnected = True
            logger.info("Connected to %s on host %s port %d", self.device, self.host, self.port)
        elif error == errno.EISCONN:
            logger.info("Already connected to %s on host %s port %d",
                        self.device, self.host, self.port)
        elif error == errno.ECONNREFUSED:
            logger.warning("Cannot connect to %s on host %s port %d",
                        self.device, self.host, self.port)
        else:
            logger.warning("Error %i: %s while connecting to host %s port %d",
                        error, os.strerror(error), self.host, self.port)

    def rx(self):
        """Receive data"""

        data = [ ]
        self.sock.settimeout(self.timeout)

        while (not data) or (not data[-1].endswith(self.commandTerminator)):
            try:
                s = self.sock.recv(self.sockBuffSize)
                data.append(s.decode('utf-8'))
            except socket.timeout:
                logger.warning('A socket timeout occurred whilst waiting for data')
                break
            except socket.error as e:
                if e.errno != errno.EWOULDBLOCK:
                    # unexpected error
                    if e.errno is None:
                        logger.error('Unexpected error: %s', e)
                    else:
                        logger.error('Error %i: %s when reading from socket',
                                    e.errno, os.strerror(e.errno))
                    self.sock.settimeout(None)
                    break

        self.sock.settimeout(None)
        return ''.join(data)

    def tx(self, msg):
        """Transmit data <msg> to device."""
        if self.commandInterval:
            time.sleep(self.commandInterval)

        if isinstance(msg, str):
            if self.commandTerminator is not None and msg[-1] != self.commandTerminator:
                msg += self.commandTerminator
            msg = msg.encode('utf8')
        self.sock.send(msg)


    ################################################
    # Prologic interface functions
    # For manual references, see the prologic manual
    ################################################

    def converterVersion(self):
        """
        Retrieves the firmware version of the Prologix converter.

        :see: manual, p15
        """
        return self.ask('++ver')

    def setLocalMode(self):
        """
        This command enables front panel operation of the currently addressed instrument.

        :see: manual, p11
        """
        self.tx('++loc')
