#!/usr/bin/env python3
""" Driver module for Keithley sourcemeter """
import time
import logging

from gssw.ucelib import gpib


__all__ = ['Keithley2400']
logger = logging.getLogger(__name__)


class Keithley2400(gpib.GpibTcpIp):
    '''Encapsulates the Keithley 2400 sourcemeter.
    Connects by Prologix Ethernet to GPIB adapter.'''

    #: possible current source ranges, used when k2400 is configured as a
    #: fixed-range current source
    CURRENT_SOURCE_RANGE = (
        0.00004000, # [0] 0.001..100
        0.00040000, # [1] 0.01..1000
        0.00400000, # [2] 0.1..10000
        0.04000000) # [3] 1..1000000

    def __init__(self, host, port=1234, gpibDevAddress=25, gpibDevReset=False, scaleFactor=1):

        # base class constructor
        super(Keithley2400, self).__init__(host, port, device='ucelib.Keithley2400',
                timeout=1, explicitRead=True)

        if gpibDevReset:
            # reset prologix ethernet to gpib adapter
            self.tx('++rst')
            time.sleep(5)

        # configure the prologix ethernet to gpib adapter
        #
        # ++mode 1      >> controller mode
        # ++addr        >> set primary address
        # ++eos 2       >> terminate each response with an LF and EOI
        # ++eoi 1       >> enable the assertion of the EOI signal with the last character
        # ++auto 0      >> turn off read-after-write and address the instrument to listen
        #                  (else the keithley gives an 'Unterminated Query Error')
        #
        self.tx('++mode 1\n++addr ' + str(gpibDevAddress) + '\n++eos 2\n++eoi 1\n++auto 0\n')

        # scalefactor which is used to convert voltage to ohm
        self.scaleFactor = scaleFactor

    def getID(self):
        '''Get identifier.'''
        return self.ask("*IDN?")[:-1]

    def reset(self):
        '''Return Keithley to default power-on state.'''
        logger.debug('Reset')
        self.tx(':SYST:PRES;')

    def sourceON(self):
        '''Turn voltage/current source ON.'''
        logger.info('Source ON')
        self.tx(':OUTP:STAT ON;')

    def sourceOFF(self):
        '''Turn voltage/current source OFF.'''
        logger.info('Source OFF')
        self.tx(':OUTP:STAT OFF;')

    def currentSourceInit(self, rangeMode, currentRange=None):
        '''
        Initialize as current source.

        :param str rangeMode: either "auto" or "fixed".
        :param float currentRange: in ampere only when range is set to "fixed".
        '''
        # set mode
        self.tx(':SOUR:FUNC:MODE CURR;')
        # set range
        if rangeMode == 'auto':
            logger.info('Initialized Current Source. Setting auto range')
            self.tx(':SOUR:CURR:RANG:AUTO ON;')
        else:
            logger.info('Initialised Current Source. Setting range to %f A', currentRange)
            self.tx(':SOUR:CURR:RANG %f;' % currentRange)
        # Set current amplitude to zero
        self.tx(':SOUR:CURR:LEV:IMM:AMPL 0.0;')

    def calcCurrentSourceRange(self, current_A):
        """Calculate the necessary current source range for the given <current_A>"""
        res = self.CURRENT_SOURCE_RANGE[-1:]
        for currentSourceItem in self.CURRENT_SOURCE_RANGE:
            if current_A < currentSourceItem:
                res = currentSourceItem
                break
        return res

    def setCurrentSourceRange(self, currentRange):
        """

        :param currentRange:
        """
        self.tx(':SOUR:CURR:RANG %f;' % currentRange)

    def voltageSourceInit(self, rangeMode, voltageRange=None):
        '''
        Initialize as voltage source.

        :param str rangeMode: either "auto" or "fixed".
        :param float voltageRange: range in volt only when range is set to "fixed".
        '''
        # set mode
        self.tx(':SOUR:FUNC:MODE VOLT;')
        # set range
        if rangeMode == 'auto':
            logger.debug('Initialised Voltage Source. Setting auto range')
            self.tx(':SOUR:VOLT:RANG:AUTO ON;')
        else:
            logger.debug('Initialised Voltage Source. Setting range to %f A', voltageRange)
            self.tx(':SOUR:VOLT:RANG %f;' % voltageRange)
        # set current amplitude to zero
        self.tx(':SOUR:VOLT:LEV:IMM:AMPL 0.0;')

    def currentMeterInit(self):
        '''Initialize as current meter.'''
        # set mode
        self.tx(':CONF:CURR:DC')
        # autorange provides the highest accuracy
        self.tx(':SENS:CURR:RANG:AUTO ON;')
        logger.debug('Initialised Current Meter')

    def voltageMeterInit(self):
        '''Initialize as voltage meter.'''
        # set mode
        self.tx(':CONF:VOLT:DC')
        # autorange provides the highest accuracy
        self.tx('SENS:VOLT:RANG:AUTO ON;')
        logger.debug('Initialised Voltage Meter')

    def setCurrent(self, current):
        '''
        pre: device is initialised as current source
        Set current (in ampere).
        '''
        logger.info('Set current to %f ampere (= %f uA)', current, current * 1.0e6)
        self.tx(':SOUR:CURR:LEV:IMM:AMPL %E;' % current)

    def getCurrent(self):
        '''
        pre: device is initialised as current meter
        Get current (in ampere).
        '''
        current = float(self.ask(':INIT;FETC?'))
        logger.info('GetCurrent returns %f ampere (= %f uA)', current, current * 1.0e6)
        return current

    def setVoltage(self, voltage):
        '''
        pre: device is initialised as voltage source
        Set voltage (in volt).
        '''
        logger.debug('Set voltage to %f volt', voltage)
        self.tx(':SOUR:VOLT:LEV:IMM:AMPL %f;' % voltage)

    def getVoltage(self):
        '''
        pre: device is initialised as voltage meter
        Get voltage (in volt).
        '''
        voltage = float(self.ask(':INIT;FETC?'))
        logger.info('GetVoltage returns %f volt (= %f uV)', voltage, voltage * 1.0e6)
        return voltage

    def measure(self):
        '''
        pre: device is initialised as voltage meter
        Measures voltage and applies scalefactor to convert to ohm.
        '''
        volts = self.getVoltage()
        volts *= self.scaleFactor
        logger.info('measure: %f ohm', volts)
        return volts

    def setScaleFactor(self, scaleFactor):
        '''Set scalefactor'''
        self.scaleFactor = scaleFactor
        logger.debug('scale factor = %f', scaleFactor)

    def setDisplayDigits(self, digits):
        '''
        In order to prevent the dot moving constantly, resulting in an
        unreadable display of the keithley 2400, we have the option
        to set a fixed number of digits
        NB: This command is not compatible with the Keithley2000
        '''
        self.tx(':DISP:DIG %d;' % digits)
        logger.debug('setDisplayDigits(%d)', digits)

    def setDisplayMessage(self, message):
        '''Set display message.'''
        logger.debug('Display message: %s', message)
        self.tx(":DISP:WIND1:TEXT:DATA '%s'", message)

    def displayMessageShow(self):
        '''Enable showing of display message.'''
        logger.debug('Show Messages')
        self.tx(':DISP:WIND1:TEXT:STAT ON;')

    def displayMessageHide(self):
        '''Disable showing of display message.'''
        logger.debug('Hide Messages')
        self.tx(':DISP:WIND1:TEXT:STAT OFF;')

    def tracData(self):
        '''Read contents of buffer.'''
        return self.ask(':TRAC:DATA?')

    def clearData(self):
        '''Clear contents of buffer.'''
        self.tx(':TRAC:CLE')

    def beep(self, frequency=1500, duration=0.5):
        '''Generate a beep.

        :param int frequency:  Frequency in Hz.
        :param float duration: Duration in seconds.'''
        logger.debug('Beep !')
        self.tx(':SYST:BEEP:IMM %i,%f' % (frequency, duration))
