#!/usr/bin/env python3
""" Module to interface with a LakeShore 218 Temperature Monitoring System. The temperature 
monitor is assumed to be connected via a serial interface to the computer. Eg USB-RS232"""

import logging, time
import serial

from gssw.ucelib.gpib import GpibSerial

logger = logging.getLogger(__name__)



__all__ = ['LakeShore218']
DEVICE_IDENTITY_PREFIX = "LSCI,MODEL218E"


class LakeShore218(GpibSerial):
    """
    Class to provice a communication interface to the LakeShore 218 Temperature Monitor.

    :seealso: Module docstring
    """
    def __init__(self, serialPort="/dev/ttyUSB0", timeout=1.2):
        super().__init__(serialPort         =serialPort,
                         baudrate           =9600,
                         device             ="LakeShore 218",
                         timeout            =timeout,
                         processingDelay    =0.5,
                         commandTerminator  ='\n')

        self.serial.bytesize = serial.SEVENBITS
        self.serial.parity = serial.PARITY_ODD
        self.serial.stopbits = serial.STOPBITS_ONE


    def checkIdentification(self):
        """
        Verifies that the connected device actual is an LakeShore 218

        :return: bool, Result of the verification
        """
        time.sleep(0.1)                 # Sometimes the device needs time after initialization.
        identification = super().identification()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        if not deviceCorrect:
            logger.error("Device identity '%s' does not match '%s'",
                         identification, DEVICE_IDENTITY_PREFIX)
        return deviceCorrect


    def setSystemLockKeypanel(self, locked):
        """
        Places the instrument in remote or local mode. In the remote
        mode, the keys on the front panel are locked.

        :param bool locked: boolean state of keypad lock
        :see: manual p6-28
        """
        if locked:
            self.tx("MODE 1")
        else:
            self.tx("MODE 0")


    def getSingleCelsiusMeas(self, inputNum):
        """
        Query single temperature reading in Celsius.

        :param int inputNum: select input number to query. Valid entries: 1-8.
        :return: str, single temperature measurement in Celsius
        :see: manual p6-22
        """
        assert inputNum in list(range(0,9)), "Inputs should between 1-8"
        return self.ask("CRDG? {}".format(inputNum))


    def getAllCelsiusMeas(self):
        """
        Query All thermister temperatures in Celsius
        
        :return: str, All temperature measurements in Celsius as a comma separated string
        :see: manual p6-22
        """
        return self.getSingleCelsiusMeas(0)


    def getSingleKelvinMeas(self, inputNum):
        """
        Query single temperature reading in Kelvin.

        :param int inputNum: select input to query. Valid entries: 1-8.
        :return: str, single temperature measurement in Kelvin
        :see: manual p6-25
        """
        assert inputNum in list(range(0,9)), "Inputs should between 1-8"
        return self.ask("KRDG? {}".format(inputNum))


    def getAllKelvinMeas(self):
        """
        Query All thermister temperatures in Kelvin
        
        :return: str, All temperature measurements (1-8) in Kelvin as a comma separated string
        :see: manual p6-25
        """
        return self.getSingleKelvinMeas(0)



    def getSingleResistanceMeas(self, inputNum):
        """
        Query single Sensor Units Reading (Assumed to be Resistance in Ohms)

        :param int inputNum: select input resistance to query. Valid entries: 1-8.
        :return: str, single resistance measurement in Ohms as a string. Assumming a resistance \
        thermometer is attached. Check manual for more info.
        :see: manual p6-30
        """
        assert inputNum in list(range(0,9)), "Inputs should between 1-8"
        return self.ask("SRDG? {}".format(inputNum))


    def getAllResistanceMeas(self):
        """
        Query all Sensor Units Reading (Assumed to be Resistance in Ohms)
        
        :return: str, All resistance measurements in Ohms as a comma separated string. Assumming \
        a resistance thermometer is attached. Check manual for more info.
        :see: manual p6-30
        """
        return self.getSingleResistanceMeas(0)


    def deleteCurve(self, curveNum):
        """
        Delete Specific User defined Curve. Accpets inputs as curve number between 21 and 28 
        for inputs 1-8.

        :param int curveNum: Curve number to delete. Valid entries: 21-28
        :see: manual p6-22
        """
        assert curveNum in list(range(21,29)), "Only Curves 21-28 are user defined curves."

        self.tx("CRVDEL {}".format(curveNum))


    def setCurveHeader(self, curveNum, curveName, curveSerialNum, curveDataFormat, limitVal, 
                       tempCoefficient):
        """
        Function to set Curve Header Parameters.

        :param int curveNum: Specifies whcih curve to Configure. Valid entries: 21-28
        :param str curveName: Specifies Curve Name. Limited to 15 Characters
        :param int curveSerialNum: Specifies curve serial number. Limited to 10 digits
        :param int curveDataFormat: Specifies curve data format. Valid entries: 2=V/K, \
                                    3=Ohm/K, 4=log(Ohm)/K
        :param float limitVal: Specifies curve limit temperature in Kelvin.
        :param int tempCoefficient: Set the curve's temperature coefficient. 1=Negative, 2=Positive
        :see: manual p6-22
        """
        # Check inputs
        assert curveNum in list(range(21,29)), "Only Curves 21-29 are user defined curves."
        assert isinstance(curveName, str) is True, "Curve Name Should be a string."
        assert len(curveName) < 15, "Curve Name limited to 15 charaters"
        assert isinstance(curveSerialNum, int) is True, "Curve Serial number must be an int."
        assert curveSerialNum < 9999999999, "Curve Serial Number should be 10 digits"
        assert curveDataFormat in [2, 3, 4], "Curve data format should be 2-4."
        assert isinstance(limitVal, (int, float)) is True, "Temperature (k) should be int or float"
        assert tempCoefficient in [1, 2], "Temperature coefficient can be either 1(-ve) or 2(+ve)"

        self.tx("CRVHDR {},{},{},{},{},{}".format(curveNum, curveName, curveSerialNum, 
                                                  curveDataFormat, limitVal, tempCoefficient))


    def getCurveHeader(self, curveNum):
        """
        Query Curve Header information.

        :param int curveNum: Specifies whcih curve header to query. Valid entries: 1-5: Standard \
                                Diode Curves, 6-9: Standard PLatinum Curves, 21-28: User Curves. \
                                Curve Locations 10-20 are not used.
        :return: list, [curveName(str), curveNum(int) (0 if not defined or Standard Curve), \
                        curveDataFormat(int), limitVal(float, k), tempCoefficeint(int)]
        :see: manual p6-22
        """

        assert curveNum in list(range(1, 28)), "Can query curve number 1-36 only."
        assert curveNum not in list(range(10,21)), "Curve Locations 10-20 are not used."

        headerInfo = self.ask("CRVHDR? {}".format(curveNum)).split(',')
        headerInfo = [headerInfo[i].strip() for i, _ in enumerate(headerInfo)]
        if not headerInfo[1]:
            headerInfo[1] = 0
        elif headerInfo[1] == 'STANDARD':
            headerInfo[1] = 0
        else:
            headerInfo[1] = int(headerInfo[1])

        headerInfo[2] = int(headerInfo[2])
        headerInfo[3] = float(headerInfo[3])
        headerInfo[4] = int(headerInfo[4])

        return headerInfo


    def setCurveDataPoint(self, curveNum, pointIndex, units, temperature):
        """
        Configure a user curve data point.

        :param int curveNum: Specifies which user curve to configure. Valid entries: 21-28.
        :param int pointIndex: Specifies point index in the curve. Valid entries: 1-200.
        :param float units: Specifies sensor units for this point (eg. Ohms, Volts) max 6 digits.
        :param float temperature: Specifies the corresponding temperature in Kelvin for this point \
                                    to 6 digits.
        :see: manual p6-22
        """

        # Check inputs
        assert curveNum in list(range(21,29)), "Only Curves 21-28 are user defined curves."
        assert pointIndex in list(range(1,201)), "Point index should be between 1-200."
        assert isinstance(units, (int, float)) is True, "Sensor Unit should be int or float"
        assert len(str(units)) < 8, "Max 6 Digits for Sensor Units"
        assert isinstance(temperature, (int, float)) is True, "Temperature (k) should be int or \
                                                                float"
        assert len(str(temperature)) < 8, "Max 6 Digits for Temperature in Kelvin"

        self.tx("CRVPT {},{},{},{}".format(curveNum, pointIndex, units, temperature))


    def getCurveDataPoint(self, curveNum, pointIndex):
        """
        Query a curve data point.

        :param int curveNum: Specifies which user curve to query. Valid entries: 1-5: Standard \
                                Diode Curves, 6-9: Standard PLatinum Curves, 21-28: User Curves. \
                                Curve Locations 10-20 are not used.
        :param int pointIndex: Specifies point index in the curve. Valid entries: 1-200.
        :return: list, [sensor unit value, temperature (k)] in float
        :see: manual p6-23
        """
        # Check Inputs
        assert curveNum in list(range(1, 28)), "Can query curve number 1-36 only."
        assert curveNum not in list(range(10,21)), "Curve Locations 10-20 are not used."
        assert pointIndex in list(range(1,201)), "Point index should be between 1-200."

        curvePointData = self.ask("CRVPT? {},{}".format(curveNum, pointIndex)).split(',')
        curvePointData = [float(i) for i in curvePointData]

        return curvePointData


    def setInputCurve(self, inputNum, curveNum):
        """
        Specify the curve an input uses for temperature conversion.

        :param int inputNum: specifies input to configure. Valid entries: 1-8
        :param int curveNum: Specifies which curve the input uses. Defaults to 0 if the specified \
                                curve does not match the input. Valid entries: 1-5: Standard \
                                Diode Curves, 6-9: Standard PLatinum Curves, 21-28: User Curves. \
                                Curve Locations 10-20 are not used.
                                Curves, 21-35:User Defined Curves.
        :see: manual p6-24
        """
        # Check inputs
        assert inputNum in list(range(1,9)), "Inputs should between 1-8"
        assert curveNum in list(range(1, 28)), "Can query curve number 1-36 only."
        assert curveNum not in list(range(10,21)), "Curve Locations 10-20 are not used."

        self.tx("INCRV {},{}".format(inputNum, curveNum))


    def getInputCurve(self, inputNum):
        """
        Query the temperature conversion curve number selected for a given input.

        :param int inputNum: specifies input to query. Valid entries: 1-8
        :return: int, Curve number for the given input. 0 if None. 1-5: Standard Diode Curves, \
                        6-9: Standard PLatinum Curves, 21-28: User Curves. Curve Locations 10-20 \
                        are not used.
        :see: manual p6-24
        """
        # Check inputs
        assert inputNum in list(range(1,9)), "Inputs should between 1-8"

        curveNum = self.ask("INCRV? {}".format(inputNum))

        return int(curveNum)






def main():
    """ Test program for the temperature monitor unit """
    logging.basicConfig(
        level='DEBUG',
        format='%(asctime)s %(filename)25s:%(lineno)-4d : %(levelname)-7s: %(message)s')

    print ("Test program for:" + __doc__)
    lakeShore = LakeShore218()
    try:
        print ("Connected to: {}".format(lakeShore.identification()))
        print('')
        if lakeShore.checkIdentification():
            # Procees with test
            print("Device Identified.")
            print('')
            # Read all inputs in Celsius
            print("------------ Showing all temperatures in Celsius ------------")
            print('')
            print(lakeShore.getAllCelsiusMeas())
            print('')

            print("------------ Showing all Resistances in Ohms ------------")
            print('')
            print(lakeShore.getAllResistanceMeas())
            print('')
            print("------------ Locking Front Panel for 5 Seconds ------------")
            print('')
            lakeShore.setSystemLockKeypanel(True)
            time.sleep(5)
            print("------------ Unlocking Front Panel ------------")
            print('')
            lakeShore.setSystemLockKeypanel(False)
            print("------------ Test Completed. ------------")
            print('')
        else:
            print("Device Identification Failed.")

    finally:
        lakeShore.close()



if __name__ == '__main__':
    main()
