#!/usr/bin/env python3
""" Module to interface with a LakeShore 325 Temperature Controller. The temperature 
controller is assumed to be connected via a serial interface to the computer. Eg USB-RS232"""

import logging, time
import serial

from gssw.ucelib.gpib import GpibSerial

logger = logging.getLogger(__name__)



__all__ = ['LakeShore325']
DEVICE_IDENTITY_PREFIX = "LSCI,MODEL325"


class LakeShore325(GpibSerial):
    """
    Class to provice a communication interface to the LakeShore 325 Temperature Controller.

    :seealso: Module docstring
    """
    def __init__(self, serialPort="/dev/ttyUSB0", timeout=1.2):
        super().__init__(serialPort         =serialPort,
                         baudrate           =9600,
                         device             ="LakeShore 325",
                         timeout            =timeout,
                         processingDelay    =0.5,
                         commandTerminator  ='\n')

        self.serial.bytesize = serial.SEVENBITS
        self.serial.parity = serial.PARITY_ODD
        self.serial.stopbits = serial.STOPBITS_ONE


    def checkIdentification(self):
        """
        Verifies that the connected device actual is an LakeShore 325

        :return: bool, Result of the verification
        """
        time.sleep(0.1)                 # Sometimes the device needs time after initialization.
        identification = super().identification()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        if not deviceCorrect:
            logger.error("Device identity '%s' does not match '%s'",
                         identification, DEVICE_IDENTITY_PREFIX)
        return deviceCorrect


    def setSystemLockKeypanel(self, locked):
        """
        Places the instrument in remote or local mode. In the remote
        mode, the keys on the front panel are locked.

        :param bool locked: Set keylock status
        :see: manual p6-30
        """
        if locked:
            self.tx("MODE 1")
        else:
            self.tx("MODE 0")


    def setControlLoopMode(self, loop, mode):
        """
        Control loop mode command

        :param int loop: specifying which loop to configure (Either 1 or 2)
        :param int mode: specifying which mode to set the loop to (1: Manual PID, \
        2: Zone, 3: Open Loop, 4: Auto Tune PID, 5: Auto Tune PI, 6: Auto Tune P)
        :see: manual p6-25
        """
        # Check Inputs
        assert loop in [1, 2], "Loop Number should be either 1 or 2"
        assert mode in [1, 2, 3, 4, 5, 6], "Mode should be an integer from 1 to 6"


        self.tx("CMODE? {},{}".format(loop, mode))


    def getControlLoopMode(self, loop):
        """
        Get the current control loop mode for a given input loop

        :param int loop: specifying which loop mode to query (Either 1 or 2)
        :return: Integer specifying which mode the queried loop is in (0: Loop Disabled, 1: Manual \
        PID, 2: Zone, 3: Open Loop, 4: Auto Tune PID, 5: Auto Tune PI, 6: Auto Tune P)
        :see: manual p6-25
        """

        # Check Inputs
        assert loop in [1, 2], "Loop Number should be either 1 or 2"

        loopMode =  self.ask("CMODE {}".format(loop))

        if not loopMode:
            return 0
        else:
            return int(loopMode)


    def getSingleCelsiusMeas(self, inputSelect):
        """
        Query single temperature reading in Celsius.

        :param int inputSelect: specifying which input temperature to query in celsius (Either \
        1: input A or 2: input B). 
        :return: str, Single temperature measurement in Celsius as a string
        :see: manual p6-25
        """

        # Check Inputs
        assert inputSelect in [1, 2], "Loop Number should be either 1 or 2"

        if inputSelect == 1:
            return self.ask("CRDG? A")
        elif inputSelect == 2:
            return self.ask("CRDG? B")


    def getSingleKelvinMeas(self, inputSelect):
        """
        Query single temperature reading in Celsius.

        :param int inputSelect: specifying which input temperature to query in kelvin (Either \
        1: input A or 2: input B). 
        :return: str, single temperature measurement in Kelvin as a string
        :see: manual p6-30
        """
        # Check Inputs
        assert inputSelect in [1, 2], "Loop Number should be either 1 or 2"

        if inputSelect == 1:
            return self.ask("KRDG? A")
        elif inputSelect == 2:
            return self.ask("KRDG? B")


    def getSingleResistanceMeas(self, inputSelect):
        """
        Query single Sensor Units Reading (Assumed to be Resistance in Ohms)

        :param int inputSelect: specifying which input sensor units (ohms) to query \
        (Either 1: input A or 2: input B). 
        :return: str, single resistance measurement in Ohms as a string
        :see: manual p6-33
        """
        # Check Inputs
        assert inputSelect in [1, 2], "Loop Number should be either 1 or 2"

        if inputSelect == 1:
            return self.ask("SRDG? A")
        elif inputSelect == 2:
            return self.ask("SRDG? B")


    def setControlLoopParameters(self, loop, inputSelect, setPointUnits, enablePowerUp, outputType):
        """
        Command to set the parameters for a given control loop
        
        :param int loop: specifies which loop to configure (1 or 2)
        :param int inputSelect: specifies which input to use for the given loop (1: Input A, \
        2: Input B)
        :param int setPointUnits: Specifies which units to control from the input (1: Kelvin, \
        2: Celsius, 3: Sensor Units (Ohms))
        :param int enablePowerUp: specifies whether the cotrol loop is on or off after power-up \
        (0: Power-up enable off, 1: Power-up enable on)
        :param int outputType: specifies whether heater ouput displays in current or power \
        (1: Current, 2: Power)
        :see: manual p6-27
        """
        # Check Inputs
        assert loop in [1, 2], "Loop Number should be either 1 or 2"
        assert inputSelect in [1, 2], "Inputs should either be 1 (for A) or 2 (for B)"
        assert setPointUnits in [1, 2, 3], "Setpoint units should be either 1 (for Kelvin) or 2  \
        (for Celsius) or 3 (for Ohms) "
        assert enablePowerUp in [0, 1], "Enable power-up should be either 0 (off) or 1 (on)"
        assert outputType in [1, 2], "Output Type should be either 1 (for Current) or 2 (for Power)"

        if inputSelect == 1:
            self.tx("CSET {},A,{},{},{}".format(loop, setPointUnits, enablePowerUp, outputType))
        elif inputSelect == 2:
            self.tx("CSET {},B,{},{},{}".format(loop, setPointUnits, enablePowerUp, outputType))


    def getControlLoopParameters(self, loop):
        """
        Query the current parameters for a given control loop

        :param int loop: specifies which loop parameters are queried (1 or 2)
        :return: list, [selected input, setpoint units, powerup enabled, output type]
        :see: manual p6-27
        """
        # Check Inputs
        assert loop in [1, 2], "Loop Number should be either 1 or 2"
        
        loopParams = self.ask("CSET? {}".format(loop)).split(",")

        if loopParams[0] == 'A':
            loopParams[0] = 1
        elif loopParams[0] == 'B':
            loopParams[0] = 2

        return list(map(int, loopParams))


    def getCurrentHeaterOutput(self, loop):
        """
        Get the current output of the heater in a loop as a % of current/power as specified in the
        setControlLoopParameters() function.

        :param int loop: specifies which loop heater output is queried (1 or 2)
        :return: float, percentage of the current heater output (power/current)
        :see: manual p6-28
        """
        # Check Inputs
        assert loop in [1, 2], "Loop Number should be either 1 or 2"
        
        return float(self.ask("HTR? {}".format(loop)))


    def setHeaterResistance(self, loop, heaterResistance):
        """
        Set the resistance value of the heaters. Two options are available: 25Ohms and 50Ohms

        :param int loop: specifies which loop heater resistance is set (1 or 2)
        :param int heaterResistance: value of the resistance set (either 25 or 50 Ohms)
        :see: manual p6-28
        """
        # Check Inputs
        assert loop in [1, 2], "Loop Number should be either 1 or 2"
        assert heaterResistance in [25, 50], "Heater resistance should be either 25 or 50 Ohms"

        if heaterResistance == 25:
            heaterResistance = 1
        elif heaterResistance == 50:
            heaterResistance = 2

        self.tx("HTRRES {},{}".format(loop, heaterResistance))


    def getHeaterResistance(self, loop):
        """
        Get the currently set value of heater resistance on a given loop.

        :param int loop: specifies which loop heater resistance is queried (1 or 2)
        :return: int, value of the heater resistance in the given loop. 0 if not configured.
        :see: manual p6-28
        """
        # Check Inputs
        assert loop in [1, 2], "Loop Number should be either 1 or 2"
        
        loopHeaterResistance = self.ask("HTRRES? {}".format(loop))

        if loopHeaterResistance == '1':
            loopHeaterResistance = 25
        elif loopHeaterResistance == '2':
            loopHeaterResistance = 50
        else:
            loopHeaterResistance = 0

        return int(loopHeaterResistance)


    def setInputType(self, inputSelect, sensorType, compensation=0):
        """
        Set input type for a given input (A or B)

        :param int inputSelect: specifies input to configure (1: Input A, 2: Input B)
        :param int sensorType: Specifies input sensor type. Valid Entries: 0: Silicon Diode, \
        1: GaAlAs Diode, 2: 100 Ohm Platinum/250, 3: 100 Ohm Platinum/500, 4: 1000 Ohm Platinum, \
        5: NTC RTD, 6: Thermocouple 25mV, 7: Thermocouple 50mV, 8: 2.5V 1mA, 9: 7.5V 1mA
        :param int compensation: specifies input compensation where 0=off and 1=on. Resersal for \
        thermal EMF compensation if input is resistive, room temperature compensation if input is \
        thermocouple. Always 0 if input is diode.
        :see: manual p6-29
        """
        # Check inputs
        assert inputSelect in [1, 2], "Inputs should either be 1 (for A) or 2 (for B)"
        assert sensorType in list(range(10)), "Sensor type should be given as 0-9"
        assert compensation in [0,1], "Compensation should be either 0=off or 1=on"

        if inputSelect == 1:
            self.tx("INTYPE A,{},{}".format(sensorType, compensation))
        elif inputSelect == 2:
            self.tx("INTYPE B,{},{}".format(sensorType, compensation))


    def getInputType(self, inputSelect):
        """
        Get input type settings for a given input (A or B)

        :param int inputSelect: specifies input to query (1: Input A, 2: Input B)
        :return: list, [sensorType, compensation] 0: Silicon Diode, 1: GaAlAs Diode, \
        2: 100 Ohm Platinum/250, 3: 100 Ohm Platinum/500, 4: 1000 Ohm Platinum, \
        5: NTC RTD, 6: Thermocouple 25mV, 7: Thermocouple 50mV, 8: 2.5V 1mA, 9: 7.5V 1mA
        :see: manual p6-29
        """
        # Check inputs
        assert inputSelect in [1, 2], "Inputs should either be 1 (for A) or 2 (for B)"

        if inputSelect == 1:
            inputParams = self.ask("INTYPE? A").split(",")
        elif inputSelect == 2:
            inputParams = self.ask("INTYPE? B").split(",")

        return list(map(int, inputParams))


    def setManualHeaterPower(self, loop, heaterOutput):
        """
        Control loop Manual Heater Power (MHP) Ouput Command

        :param int loop: specifies which loop heater should be controlled (1 or 2)
        :param float heaterOutput: percentage at which the heater output is set at
        :see: manual p6-31
        """
        # Check inputs
        assert loop in [1, 2], "Loop should either be 1 or 2"
        assert isinstance(heaterOutput, (int, float)) is True, "Heater output Should be a float."
        assert heaterOutput <= 100, "Heater output can only be set less than 100 %"

        self.tx("MOUT {},{}".format(loop, heaterOutput))


    def getManualHeaterPower(self, loop):
        """
        Control loop manual Heater Power (MHP) Output Query

        :param int loop: specifies which loop heater should be queried (1 or 2)
        :return: float, current heater power output percentage
        :see: manual p6-31
        """

        return float(self.ask("MOUT? {}".format(loop)))


    def setPIDParameters(self, loop, p, i, d):
        """
        Set Control loop P, I, D Values command

        :param int loop: specifies which loop PID parameters should be changed
        :param float p: The value for control loop proportional gain (0.1 < P < 1000)
        :param float i: The value for control loop integral reset (0.1 < I < 1000)
        :param float d: The value for control loop derivative rate (0.1 < D < 200)
        :see: manual p6-31
        """
        # Check inputs
        assert loop in [1, 2], "Loop should either be 1 or 2"
        assert isinstance(p, (int, float)) is True, "P value should be either integer or float."
        assert isinstance(i, (int, float)) is True, "I value should be either integer or float."
        assert isinstance(d, int) is True, "D value should be either integer."
        assert 0.1 <= p <= 1000, "P value should be between 0.1 and 1000"
        assert 0.1 <= i <= 1000, "I value should be between 0.1 and 1000"
        assert 0.1 <= d <= 200, "D value should be between 0.1 and 200"

        self.tx("PID {},{},{},{}".format(loop, p, i, d))


    def getPIDParameters(self, loop):
        """
        Query Control Loop PID Parameters

        :param int loop: specifies which loop PID parameters should be queried
        :return: list, a list containing PID as floats [P, I, D]
        :see: manual p6-31
        """
        # Check inputs
        assert loop in [1, 2], "Loop should either be 1 or 2"

        pidParams = self.ask("PID? {}".format(loop)).split(',')

        return list(map(float, pidParams))


    def setHeaterRange(self, loop, heaterRange):
        """
        Command to set control loop heater ranges

        :param int loop: specifies which loop heater range to set
        :param int heaterRange: For loop 1: 0=off, 1=low(2.5W), 2=High(25W). For loop 2: 0=off, \
        1=on
        :see: manual p6-32
        """
        # Check inputs
        assert loop in [1, 2], "Loop should either be 1 or 2"
        assert heaterRange in [0,1,2], "Heater range can be set only as 0, 1, or 2"

        self.tx("RANGE {},{}".format(loop, heaterRange))


    def getHeaterRange(self, loop):
        """
        Query the current heater range in a control loop

        :param int loop: specifies which loop heater range to query
        :return: int, For loop 1: 0=off, 1=low(2.5W), 2=High(25W). For loop 2: 0=off, 1=on
        :see: manual p6-32
        """
        # Check inputs
        assert loop in [1, 2], "Loop should either be 1 or 2"

        return int(self.ask("RANGE? {}".format(loop)))


    def deleteCurve(self, curveNum):
        """
        Delete Specific User defined Curve. Accpets inputs as curve number between 21 and 35.

        :param int curveNum: Curve number to delete. Valid entries: 21-25
        :see: manual p6-25
        """
        assert curveNum in list(range(21,36)), "Only Curves 21-35 are user defined curves."

        self.tx("CRVDEL {}".format(curveNum))


    def setCurveHeader(self, curveNum, curveName, curveSerialNum, curveDataFormat, limitVal, 
                       tempCoefficient):
        """
        Function to set Curve Header Parameters.

        :param int curveNum: Specifies whcih curve to Configure. Valid entries: 21-35
        :param str curveName: Specifies Curve Name. Limited to 15 Characters
        :param int curveSerialNum: Specifies curve serial number. Limited to 10 digits
        :param int curveDataFormat: Specifies curve data format. Valid entries: 1=mV/K, 2=V/K, \
                                    3=Ohm/K, 4=log(Ohm)/K
        :param float limitVal: Specifies curve limit temperature in Kelvin.
        :param int tempCoefficient: Set the curve's temperature coefficient. 1=Negative, 2=Positive
        :see: manual p6-26
        """
        # Check inputs
        assert curveNum in list(range(21,36)), "Only Curves 21-35 are user defined curves."
        assert isinstance(curveName, str) is True, "Curve Name Should be a string."
        assert len(curveName) < 15, "Curve Name limited to 15 charaters"
        assert isinstance(curveSerialNum, int) is True, "Curve Serial number must be an int."
        assert curveSerialNum < 9999999999, "Curve Serial Number should be 10 digits"
        assert curveDataFormat in [1, 2, 3, 4], "Curve data format should be 1-4."
        assert isinstance(limitVal, (int, float)) is True, "Temperature (k) should be int or float"
        assert tempCoefficient in [1, 2], "Temperature coefficient can be either 1(-ve) or 2(+ve)"

        self.tx("CRVHDR {},{},{},{},{},{}".format(curveNum, curveName, curveSerialNum, 
                                                  curveDataFormat, limitVal, tempCoefficient))


    def getCurveHeader(self, curveNum):
        """
        Query Curve Header information.

        :param int curveNum: Specifies whcih curve header to query. Valid entries: 1-36
        :return: list, [curveName(str), curveNum(int) (0 if not defined or Standard Curve), \
                        curveDataFormat(int), limitVal(float, k), tempCoefficeint(int)]
        :see: manual p6-26
        """

        assert curveNum in list(range(1, 37)), "Can query curve number 1-36 only."

        headerInfo = self.ask("CRVHDR? {}".format(curveNum)).split(',')
        headerInfo = [headerInfo[i].strip() for i, _ in enumerate(headerInfo)]
        if not headerInfo[1]:
            headerInfo[1] = 0
        elif headerInfo[1] == 'STANDARD':
            headerInfo[1] = 0
        else:
            headerInfo[1] = int(headerInfo[1])

        headerInfo[2] = int(headerInfo[2])
        headerInfo[3] = float(headerInfo[3])
        headerInfo[4] = int(headerInfo[4])

        return headerInfo


    def setCurveDataPoint(self, curveNum, pointIndex, units, temperature):
        """
        Configure a user curve data point.

        :param int curveNum: Specifies which user curve to configure. Valid entries: 21-35.
        :param int pointIndex: Specifies point index in the curve. Valid entries: 1-200.
        :param float units: Specifies sensor units for this point (eg. Ohms, Volts) max 6 digits.
        :param float temperature: Specifies the corresponding temperature in Kelvin for this point \
                                    to 6 digits.
        :see: manual p6-26
        """

        # Check inputs
        assert curveNum in list(range(21,36)), "Only Curves 21-35 are user defined curves."
        assert pointIndex in list(range(1,201)), "Point index should be between 1-200."
        assert isinstance(units, (int, float)) is True, "Sensor Unit should be int or float"
        assert len(str(units)) < 8, "Max 6 Digits for Sensor Units"
        assert isinstance(temperature, (int, float)) is True, "Temperature (k) should be int or \
                                                                float"
        assert len(str(temperature)) < 8, "Max 6 Digits for Temperature in Kelvin"

        self.tx("CRVPT {},{},{},{}".format(curveNum, pointIndex, units, temperature))


    def getCurveDataPoint(self, curveNum, pointIndex):
        """
        Query a curve data point.

        :param int curveNum: Specifies which user curve to query. Valid entries: 1-35.
        :param int pointIndex: Specifies point index in the curve. Valid entries: 1-200.
        :return: list, [sensor unit value, temperature (k)] in float
        :see: manual p6-26
        """
        # Check Inputs
        assert curveNum in list(range(1, 37)), "Can query curve number 1-36 only."
        assert pointIndex in list(range(1,201)), "Point index should be between 1-200."

        curvePointData = self.ask("CRVPT? {},{}".format(curveNum, pointIndex)).split(',')
        curvePointData = [float(i) for i in curvePointData]

        return curvePointData


    def setInputCurve(self, inputSelect, curveNum):
        """
        Specify the curve an input uses for temperature conversion.

        :param int inputSelect: specifies input to configure (1: Input A, 2: Input B)
        :param int curveNum: Specifies which curve the input uses. Defaults to 0 if the specified
                                curve does not match the input. Valid entries: 0:None, 1-20 Standard
                                Curves, 21-35:User Defined Curves.
        """
        # Check inputs
        assert inputSelect in [1, 2], "Inputs should either be 1 (for A) or 2 (for B)"
        assert curveNum in list(range(37)), "Can select curve number 1-36 only. 0 for None."

        if inputSelect == 1:
            self.tx("INCRV A,{}".format(curveNum))
        elif inputSelect == 2:
            self.tx("INCRV B,{}".format(curveNum))


    def getInputCurve(self, inputSelect):
        """
        Query the temperature conversion curve number selected for a given input.

        :param int inputSelect: specifies input to query (1: Input A, 2: Input B)
        :return: int, Curve number for the given input. 0 if None.
        """
        # Check inputs
        assert inputSelect in [1, 2], "Inputs should either be 1 (for A) or 2 (for B)"

        if inputSelect == 1:
            curveNum = self.ask("INCRV? A")
        elif inputSelect == 2:
            curveNum = self.ask("INCRV? B")

        return int(curveNum)


    def setLoopSetPoint(self, loop, setPoint):
        """
        Configure the setpoint value (in whatever units the loop is using) for the given loop.

        :param int loop: Specifies which loop to configure setpoint.
        :param float setPoint: The value of the setpoint in whatever units the loop is using.
        """
        # Check inputs
        assert loop in [1, 2], "Loop should either be 1 or 2"
        assert isinstance(setPoint, (int, float)) is True, "The setpoint value should be either \
                                                            int or float."

        self.tx("SETP {},{}".format(loop, setPoint))


    def getLoopSetPoint(self, loop):
        """
        Query the setpoint value (in whatever units the loop is using) for the given loop.

        :param int loop: Specifies which loop to query setpoint.
        :return: float, Setpoint of the given loop.
        """
        # Check inputs
        assert loop in [1, 2], "Loop should either be 1 or 2"

        setPoint = self.ask("SETP? {}".format(loop))

        return float(setPoint)


def main():
    """ Test program for the temperature monitor unit """
    logging.basicConfig(
        level='DEBUG',
        format='%(asctime)s %(filename)25s:%(lineno)-4d : %(levelname)-7s: %(message)s')

    print ("Test program for:" + __doc__)
    lakeShore = LakeShore325()
    try:
        print ("Connected to: {}".format(lakeShore.identification()))
        print('')
        if lakeShore.checkIdentification():
            # Procees with test
            print("Device Identified.")
            print('')
            # Read all inputs in Celsius
            print("------------ Showing all temperatures in Kelvin ------------")
            print('')
            print(lakeShore.getSingleKelvinMeas(1))
            print('')
            print(lakeShore.getSingleKelvinMeas(2))

            print("------------ Showing all Resistances in Ohms ------------")
            print('')
            print(lakeShore.getSingleResistanceMeas(1))
            print('')
            print(lakeShore.getSingleResistanceMeas(2))

            print("------------ Setting Control Loop 1 to Manual PID (Mode 1) ------------")
            print('')
            lakeShore.setControlLoopMode(1,1)

            print("------------ Showing Control Loop's Modes ------------")
            print('Loop 1:')
            print(lakeShore.getControlLoopMode(1))
            time.sleep(0.1)
            print('Loop 2:')
            print(lakeShore.getControlLoopMode(2))
            time.sleep(0.1)

            print("------------ Showing Control Loop's Parameters ------------")
            print('Loop 1:')
            print(lakeShore.getControlLoopParameters(1))
            time.sleep(0.1)
            print('Loop 2:')
            print(lakeShore.getControlLoopParameters(2))
            time.sleep(0.1)

            print("------------ Setting Loop 1 Heater resistance to 50 Ohms ------------")
            lakeShore.setHeaterResistance(1, 50)
            time.sleep(0.1)

            print("------------ Showing Current Heater Resistances ------------")
            print('Loop 1:')
            print(lakeShore.getHeaterResistance(1))
            print(type(lakeShore.getHeaterResistance(1)))
            time.sleep(0.1)
            print('Loop 2:')
            print(lakeShore.getHeaterResistance(2))
            print(type(lakeShore.getHeaterResistance(2)))
            time.sleep(0.1)

            print("------------ Locking Front Panel for 5 Seconds ------------")
            lakeShore.setSystemLockKeypanel(True)
            time.sleep(5)
            print("------------ Unlocking Front Panel ------------")
            lakeShore.setSystemLockKeypanel(False)
            print("------------ Test Completed. ------------")
        else:
            print("Device Identification Failed.")

    finally:
        lakeShore.close()



if __name__ == '__main__':
    main()
