#!/usr/bin/env python3
""" This module contains the interface class for the Lakeshore 372 extended with the 3726 scanner
"""

import logging
import socket


logger = logging.getLogger(__name__)


class Lakeshore3726():
    """ Interface class for the Lakeshore 3726 resistance bridge extended with the Lakeshore 3726
        scanner. The device is connected through a tcp connection.
    """

    EXCITATION_VOLTAGE_RANGE = {
        1:  2E-6,
        2:  6.32E-6,
        3:  20E-6,
        4:  63.2E-6,
        5:  200E-6,
        6:  632E-6,
        7:  2E-3,
        8:  6.32E-3,
        9:  20E-3,
        10: 63.2E-3,
        11: 200E-3,
        12: 632E-3
    }

    EXCITATION_CURRENT_RANGE = {
        1:  1E-12,
        2:  3.16E-12,
        3:  10E-12,
        4:  31.6E-12,
        5:  100E-12,
        6:  316E-12,
        7:  1E-9,
        8:  3.16E-9,
        9:  10E-9,
        10: 31.6E-9,
        11: 100E-9,
        12: 316E-9,
        13: 1E-6,
        14: 3.16E-6,
        15: 10E-6,
        16: 31.6E-6,
        17: 100E-6,
        18: 316E-6,
        19: 1E-3,
        20: 3.16E-3,
        21: 10E-3,
        22: 31.6E-3
    }

    RESISTANCE_RANGE = {
        1:  2E-3,
        2:  6.32E-3,
        3:  20E-3,
        4:  63.2E-3,
        5:  200E-3,
        6:  632E-3,
        7:  2E0,
        8:  6.32E0,
        9:  20E0,
        10: 63.2E0,
        11: 200E0,
        12: 632E0,
        13: 2E3,
        14: 6.32E3,
        15: 20E3,
        16: 63.2E3,
        17: 200E3,
        18: 632E3,
        19: 2E6,
        20: 6.32E6,
        21: 20E6,
        22: 63.2E6,
        23: 200E6,
        24: 632E6
    }


    def __init__(self, ip, port=7777, timeout=1):
        """ Constructor

            :param str ip: ip address of the device
            :param int port: port number of the device
            :param int timeout: socket timeout in seconds
        """

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((ip, port))
        self._socket.settimeout(timeout)


    def __del__(self):
        """ Destructor """

        try:
            self._socket.close()
        except AttributeError:
            pass


    def _send(self, command):
        """ Send a command to the device

            :param str command: the command string to send without terminator
        """

        rawCommand = bytes(command, encoding='utf-8')
        rawCommand += '\r\n'

        self._socket.send(rawCommand)


    def _query(self, command):
        """ Send a command and return the reply

            :param str command: command string including arguments and terminator
            :return: response
            :rtype: str
        """

        self._send(command)
        response = b''

        # this should generate a timeout if a non-terminated response is received
        while True:
            response += self._socket.recv(1024)
            if response[-2:] == b'\r\n':
                return str(response[:-2], 'utf-8')


    def getId(self):
        """ Return the ID string of the device

            :return: ID string
            :rtype: str
        """

        return self._query('*IDN?')


    def reset(self):
        """ Reset the device """

        self._send('RST')


    def setChannel(self, channel):
        """ Set the scanner channel

            :param int channel: chennel to select [1 - 16]
        """

        assert channel in range(1, 17), 'channel must be in [1 - 16]'

        self._send('SCAN {},0'.format(channel))


    def getResistance(self, channel):
        """ Read the resistance for a specific channel

            :param int channel: chennel to select [1 - 16]
        """

        assert channel in range(1, 17), 'channel must be in [1 - 16]'

        return float(self._query('RDGR? {}'.format(channel)))


    def inputSetup(self, channel, mode, excitation, autoRange, rangeSetting, csShunt, units):
        """ Configure a channel
            :param int channel: channel to configure [1 - 16]
            :param int mode: 0 - current excitation, 1 - voltage excitation
            :param int excitation: see Lakeshore3726.EXCITATION_[VOLATAGE/CURRENT]_RANGE
            :param int autoRange: 0 = Autorange off, 1 = Autorange Current,
                2 = ROX 102B Autorange (only available for control input)
            :param int rangeSetting: see Lakeshore3726.RESISTANCE_RANGE (range is a python keyword)
            :param int csShunt:  0 = current source not shunted, excitation on.
                1 = current source shunted, excitation off.
            :param int units:  Specifies the preferred units parameter for sensor readings and for
                the control setpoint: 1 = kelvin, 2 = ohms
        """

        assert channel in range(1, 17), 'channel must be in [1 - 16]'
        assert mode in [0, 1], 'mode munst be in [0, 1]'
        assert autoRange in [0, 1], 'autoRange munst be in [0, 1]'
        assert rangeSetting in range(1, 25), 'range must be in [1 - 24]'
        assert csShunt in [0, 1], 'csShunt munst be in [0, 1]'
        assert units in [0, 1, 2], 'units munst be in [0, 1, 2]'
        if mode == 0:
            assert excitation in range(1, 13), \
                'excitation must be in [1 - 12] for current excitation mode'
        else:
            assert excitation in range(1, 23), \
                'excitation must be in [1 - 22] for voltage excitation mode'

        self._send('INSET {},{},{},{},{},{},{}'.format(
            channel, mode, excitation, autoRange, rangeSetting, csShunt, units))
