#!/usr/bin/env python3
# pylint: disable=too-many-lines
""" Driver for the Magnicon XXF-1 SQUID electronics + CSE board

    This module supports device ID 3. Other IDs have different conversions for registers and
    these differences are only accounted for in the current source related funtions.
"""

import logging
import serial
from gssw.ucelib.serialdevice import SerialDevice


logger = logging.getLogger(__name__)


class MagniconError(Exception):
    """ Magnicon protocol errors """


class MagniconXXF1(SerialDevice):
    """ Driver class for controlling a Magnicon XFF-1 device + CSE"""


    ERROR_CODE = {
        0: 'no error',
        1: 'error during transmission',
        2: 'parameter out of range',
        3: 'wrong board type (board not able to handle command)',
        4: 'unknown command',
        5: 'read/write violation',
        6: 'wrong characters in HEX number'
    }

    # selector values for read and write commands
    SEL_READ    = 8
    SEL_WRITE   = 0

    # channel nnumbers
    CH_MAGNICON = 1
    CH_CSE      = 2

    # coefficients for current sources based on device ID
    LOW_RANGE_MAX  = [125E-6, 125E-6, 1E-3, 250E-6]
    HIGH_RANGE_MAX = [500E-6, 500E-6, 2.5E-3, 5E-3]
    LOW_RANGE_R    = [43600, 43600, 5400, 21870]
    HIGH_RANGE_R   = [10895, 10895, 2167.7, 1093.1]


    def __init__(self, load_resistance=0, device_id=3, port='/dev/ttyUSB0',
            vendor_id=None, product_id=None, serial_number=None):
        """ Constructor

            :param float load_resistance: wire resistance in Ohm
            :param int device_id: magnicon device ID (different current source conversions)
            :param str port: device port on host machine
            :param str vendor_id: usb device vendor id
            :param str product_id: usb device product id
            :param str serial_number: usb device serial number
        """

        assert device_id in range(4), 'Device ID must be in [0, 1, 2, 3]'

        super().__init__(port=port, vendorId=int(vendor_id, 16), productId=int(product_id, 16),
            serialNumber=serial_number, baudrate=57600, bytesize=serial.SEVENBITS,
            parity=serial.PARITY_EVEN)

        self.load_resistance = load_resistance
        self.device_id = device_id

        self.open()


    def get_i(self):
        """ Get the I current source current

            :return: I current (A) [-5000 uA - 4997.6 uA]
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_MAGNICON, 'q', self.SEL_READ)

        raw_range = int(reg_value & 0xF)
        raw_current = int(reg_value >> 4 & 0xFFFF)

        return self._current_raw_to_amp(raw_current, raw_range, self.device_id)


    def set_i(self, current):
        """ Set the I current source current

            The range is also set based on the current

            :param int current: current [-5000 uA - 4997.6 uA] in A
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """
        # pylint: disable=invalid-name

        raw_current, raw_range = self._current_amp_to_raw(current, self.device_id)

        reg_value = raw_current << 4 | raw_range

        self._command(self.CH_MAGNICON, 'q', self.SEL_WRITE,
            parameter=reg_value, parameter_digits=5)


    def get_amp_fll(self):
        """ Get the amp / fll switch state

            :return: 0 = amp, 1 = fll
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._command(self.CH_MAGNICON, 'b', self.SEL_READ)


    def set_amp_fll(self, mode):
        """ Set the amp / fll switch state

            :param int mode: 0 = amp, 1 = fll
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert mode in [0, 1]
        self._command(self.CH_MAGNICON, 'b', self.SEL_WRITE, parameter=mode)


    def get_rf_amp(self):
        """ Get the RF setting for amp mode

            :return: 0 = 0 = off, 1 = 0k7, 2 = 0k75, 3 = 0k91, 4 = 1k, 5 = 2k14, 6 = 2k31,
            7 = 2k73, 8 = 3k, 9 = 7k, 10 = 7k5, 11 = 9k1, 12 = 10k, 13 = 23k1, 14 = 30k, 15 = 100k
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._command(self.CH_MAGNICON, 'n', self.SEL_READ)


    def set_rf_amp(self, value):
        """ Set the RF setting for amp mode

            :param int value: 0 = off, 1 = 0k7, 2 = 0k75, 3 = 0k91, 4 = 1k, 5 = 2k14, 6 = 2k31,
            7 = 2k73, 8 = 3k, 9 = 7k, 10 = 7k5, 11 = 9k1, 12 = 10k, 13 = 23k1, 14 = 30k, 15 = 100k
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in range(16)
        self._command(self.CH_MAGNICON, 'n', self.SEL_WRITE, parameter=value)


    def get_rf_fll(self):
        """ Get the RF setting for fll mode

            :return: 0 = off, 1 = 0k7, 2 = 0k75, 3 = 0k91, 4 = 1k, 5 = 2k14, 6 = 2k31,
            7 = 2k73, 8 = 3k, 9 = 7k, 10 = 7k5, 11 = 9k1, 12 = 10k, 13 = 23k1, 14 = 30k, 15 = 100k
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._command(self.CH_MAGNICON, 'n', self.SEL_READ + 1)


    def set_rf_fll(self, value):
        """ Set the RF setting for fll mode

            :param int value: 0 = off, 1 = 0k7, 2 = 0k75, 3 = 0k91, 4 = 1k, 5 = 2k14, 6 = 2k31,
            7 = 2k73, 8 = 3k, 9 = 7k, 10 = 7k5, 11 = 9k1, 12 = 10k, 13 = 23k1, 14 = 30k, 15 = 100k
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in range(16)
        self._command(self.CH_MAGNICON, 'n', self.SEL_WRITE + 1, parameter=value)


    def get_input_termination_amp(self):
        """ Get input termination state for amp

            :return: 0 = off, 1 = on
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_switch_bit(22)


    def set_input_termination_amp(self, value):
        """ Set input termination state for amp

            :param int value: 0 = off, 1 = on
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in [0, 1]
        self._set_switch_bit(22, value)


    def get_phib_disconnect(self):
        """ Get Phib disconnect

            :return: 0 = off, 1 = on
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_switch_bit(21)


    def set_phib_disconnect(self, value):
        """ Set Phib disconnect

            :param int value: 0 = off, 1 = on
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in [0, 1]
        self._set_switch_bit(21, value)


    def get_line_shunt(self):
        """ Get line shunt

            :return: 0 = off, 1 = on
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_switch_bit(20)


    def set_line_shunt(self, value):
        """ Set line shunt

            :param int value: 0 = off, 1 = on
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in [0, 1]
        self._set_switch_bit(20, value)


    def get_phix_disconnect(self):
        """ Get PhiX(IX) disconnect

            :return: 0 = off, 1 = on
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_switch_bit(19)


    def set_phix_disconnect(self, value):
        """ Set PhiX(IX) disconnect

            :param int value: 0 = off, 1 = on
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in [0, 1]
        self._set_switch_bit(19, value)


    def get_ext_flux_termination(self):
        """ Get ext flux termination

            :return: 0 = off, 1 = on
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_switch_bit(18)


    def set_ext_flux_termination(self, value):
        """ Set ext flux termination

            :param int value: 0 = off, 1 = on
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in [0, 1]
        self._set_switch_bit(18, value)


    def get_ext_flux_fll(self):
        """ Get ext flux fll

            :return: 0 = off, 1 = 10 kOhm, 2 = 2.5 kOhm, 3 = 2 kOhm
            (note the bits are swapped w.r.t the documentation)
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self.get_switches() >> 15 & 0x3


    def set_ext_flux_fll(self, value):
        """ Set ext flux fll

            :param int value: 0 = off, 1 = 10 kOhm, 2 = 2.5 kOhm, 3 = 2 kOhm
            (note the bits are swapped w.r.t the documentation)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in range(4)

        switches = self.get_switches()
        switches &= 0xFFE7FFF
        switches |= value << 15
        self.set_switches(switches)



    def get_out_coupling(self):
        """ Get out coupling

            :return: 0 = DC, 1 = AC
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_switch_bit(13)


    def set_out_coupling(self, value):
        """ Set out coupling

            :param int value: 0 = DC, 1 = AC
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in [0, 1]
        self._set_switch_bit(13, value)


    def get_squid_dummy(self):
        """ Get squid dummy switch state

            :return: 0 = off, 1 = on
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_switch_bit(12)


    def set_squid_dummy(self, value):
        """ Set squid dummy switch state

            :param int value: 0 = off, 1 = on
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in [0, 1]
        self._set_switch_bit(12, value)


    def get_ext_flux_amp(self):
        """ Get ext flux amp

            :return: 0 = off, 1 = 10 kOhm, 2 = 2.5 kOhm, 3 = 2 kOhm
            (note the bits are swapped w.r.t the documentation)
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self.get_switches() >> 8 & 0x3


    def set_ext_flux_amp(self, value):
        """ Set ext flux amp

            :param int value: 0 = off, 1 = 10 kOhm, 2 = 2.5 kOhm, 3 = 2 kOhm
            (note the bits are swapped w.r.t the documentation)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in range(4)

        switches = self.get_switches()
        switches &= 0xFFFFCFF
        switches |= value << 8
        self.set_switches(switches)


    def get_input_termination_fll(self):
        """ Get input termination state for fll

            :return: 0 = off, 1 = on
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_switch_bit(9)


    def set_input_termination_fll(self, value):
        """ Set input termination state for fll

            :param int value: 0 = off, 1 = on
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in [0, 1]
        self._set_switch_bit(9, value)


    def get_squid_gain_polarity(self):
        """ Get squid gain setting (positive/negative)

            :return: 0 = positive, 1 = negative <= NOTE THIS
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_switch_bit(8)


    def set_squid_gain_polarity(self, value):
        """ Set squid gain setting (positive/negative)

            :param int value: 0 = positive, 1 = negative <= NOTE THIS
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in [0, 1]
        self._set_switch_bit(8, value)


    def get_amp_gain_polarity(self):
        """ Get amp gain setting (positive/negative)

            :return: 0 = negative, 1 = positive
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_switch_bit(7)


    def set_amp_gain_polarity(self, value):
        """ Set amp gain setting (positive/negative)

            :param int value: 0 = negative, 1 = positive
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in [0, 1]
        self._set_switch_bit(7, value)


    def get_ib_delay(self):
        """Set Ib delay

            :return: 0 = 21 ns, 1 = 29 ns
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_switch_bit(6)


    def set_ib_delay(self, value):
        """ Set Ib delay

            :param int value: 0 = 21 ns, 1 = 29 ns
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in [0, 1]

        self._set_switch_bit(6, value)


    def get_vb_delay(self):
        """ Get Vb delay 1

            :return: 0 = 49 ns, 1 = 27 ns, 2 = 37 ns, 3 = 16 ns
            (note the bits are swapped w.r.t the documentation)
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self.get_switches() >> 4 & 0x3


    def set_vb_delay(self, value):
        """ Set Vb delay

            :param int value: 0 = 49 ns, 1 = 27 ns, 2 = 37 ns, 3 = 16 ns
            (note the bits are swapped w.r.t the documentation)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in range(4)

        switches = self.get_switches()
        switches &= 0xFFFFFCF
        switches |= value << 4
        self.set_switches(switches)


    def get_rfx_amp(self):
        """ Get RfX @ amp 1

            :return: 0 = n.a., 1 = off, 2 = 10 kOhm, 3 = 2.5 kOhm
            (note the bits are swapped w.r.t the documentation)
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self.get_switches() >> 2 & 0x3


    def set_rfx_amp(self, value):
        """ Set RfX @ amp

            :param int value: 0 = n.a., 1 = off, 2 = 10 kOhm, 3 = 2.5 kOhm
            (note the bits are swapped w.r.t the documentation)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in range(4)

        switches = self.get_switches()
        switches &= 0xFFFFFF3
        switches |= value << 2
        self.set_switches(switches)


    def get_rfx_fll(self):
        """ Get RfX @ fll 1

            :return: 0 = n.a., 1 = off, 2 = 10 kOhm, 3 = 2.5 kOhm
            (note the bits are swapped w.r.t the documentation)
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self.get_switches() & 0x3


    def set_rfx_fll(self, value):
        """ Set RfX @ fll

            :param int value: 0 = n.a., 1 = off, 2 = 10 kOhm, 3 = 2.5 kOhm
            (note the bits are swapped w.r.t the documentation)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value in range(4)

        switches = self.get_switches()
        switches &= 0xFFFFFFC
        switches |= value
        self.set_switches(switches)


    def get_amp_gain(self):
        """ Get the amp gain

            gain: 0 = 1100 (220), 1 = 140 (280), 2 = 1700 (340), 3 = 2000 (400)

            :return: gain
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_amp_mode()[0]


    def set_amp_gain(self, gain):
        """ Set the amp gain

            gain: 0 = 1100 (220), 1 = 140 (280), 2 = 1700 (340), 3 = 2000 (400)

            :param int gain: see above
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        _, bandwidth = self._get_amp_mode()
        self._set_amp_mode(gain, bandwidth)


    def get_amp_bandwidth(self):
        """ Get the amp bandwidth

            bandwidth: 0 = 0.2, 1 = 0.7, 2 = 1.5, 3 = 3.5, 4 = 20, 5 = full, 6 = AC amp off (MHz)

            :return: bandwidth: see above
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._get_amp_mode()[1]


    def set_amp_bandwidth(self, bandwidth):
        """ Set the amp bandwidth

            bandwidth: 0 = 0.2, 1 = 0.7, 2 = 1.5, 3 = 3.5, 4 = 20, 5 = full, 6 = AC amp off (MHz)

            :param int bandwidth: see above
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        gain, _ = self._get_amp_mode()
        self._set_amp_mode(gain, bandwidth)


    def get_gain_bandwidth_product(self):
        """ Get the GBWP

            GBWP values: 0 = 0.23, 1 = 0.27, 2 = 0.3, 3 = 0.38, 4 = 0.47, 5 = 0.55, 6 = 0.66,
            7 = 0.82, 8 = 1.04, 9 = 1.28, 10 = 1.5, 11 = 1.8, 12 = 2.25, 13 = 2.8, 14 = 3.3,
            15 = 4, 16 = 5, 17 = 6.2, 18 = 7.2

            :return: GBWP: see above
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._command(self.CH_MAGNICON, 'm', self.SEL_READ)


    def set_gain_bandwidth_product(self, value):
        """ Set the GBWP

            GBWP values: 0 = 0.23, 1 = 0.27, 2 = 0.3, 3 = 0.38, 4 = 0.47, 5 = 0.55, 6 = 0.66,
            7 = 0.82, 8 = 1.04, 9 = 1.28, 10 = 1.5, 11 = 1.8, 12 = 2.25, 13 = 2.8, 14 = 3.3,
            15 = 4, 16 = 5, 17 = 6.2, 18 = 7.2

            :param int value: see above
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        self._command(self.CH_MAGNICON, 'm', self.SEL_WRITE, parameter=value, parameter_digits=2)


    def get_ib(self):
        """ Get the bias current

            :return: Ib (A) [0 - 400.195 uA]
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_MAGNICON, 'c', self.SEL_READ)

        # remove the unused range digit
        reg_value >>= 4

        # convert register value to current
        return 5.467 * reg_value / (2**14 * 13660)


    def set_ib(self, current):
        """ Set the bias current

            :param float current: Ib (A) [0 - 400.195 uA]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        assert 0 <= current <= 400.195E-6

        # convert current to register value
        reg_value = min(int(current * 2**14 * 13660 / 5.467), 0x3FFF)

        # add a 0 for the unused range digit
        reg_value <<= 4

        self._command(self.CH_MAGNICON, 'c', self.SEL_WRITE,
            parameter=reg_value, parameter_digits=5)


    def get_vb(self):
        """ Get the bias voltage

            :return: Vb (V) [0 - 1700.27 uV]
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_MAGNICON, 'f', self.SEL_READ)

        # convert register value to voltage
        return 2.5 * reg_value / (2**12 * 1470)


    def set_vb(self, voltage):
        """ Set the bias voltage

            :param float voltage: Vb (V) [0 - 1700.27 uV]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert 0 <= voltage <= 1700.27E-6

        # convert voltage to register value
        reg_value = min(int(voltage * 2**12 * 1470 / 2.5), 0xFFF)

        self._command(self.CH_MAGNICON, 'f', self.SEL_WRITE,
            parameter=reg_value, parameter_digits=3)


    def get_vo(self):
        """ Get the voltage offset

            Note that the magnicon xxf also has a Vo+ that is ignored in this module.

            :return: Vo (V) [-111.05 - 111.00 uV]
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_MAGNICON, 'g', self.SEL_READ)
        raw_voltage = reg_value >> 3

        # convert register value to voltage
        return 2.5 * (raw_voltage - 482) / 2**12 * 1470


    def set_vo(self, voltage):
        """ Set the voltage offset

            Note that the magnicon xxf also has a Vo+ that is ignored in this module.

            :param float voltage: Vo (V) [-111.05 - 111.00 uV]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert -111.05E-6 <= voltage <= 111.00E-6

        # convert voltage to raw value
        raw_voltage = min(int(voltage * 2**12 * 1470 / 2.5 + 482), 0x3C4)

        # read the `g` register and only modify the Vo part, leaving Vo+ unchanged
        reg_value = self._command(self.CH_MAGNICON, 'g', self.SEL_READ)
        reg_value &= 0x7
        reg_value |= raw_voltage << 3

        self._command(self.CH_MAGNICON, 'g', self.SEL_WRITE,
            parameter=reg_value, parameter_digits=6)


    def get_phib(self):
        """ Get the bias flux

            :return: Phib (A) [-124.56 uA - 124.50 uA]
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_MAGNICON, 'j', self.SEL_READ)

        # convert register value to flux
        # note that the datasheet states an incorrect conversion
        return 5 * (reg_value - 2**11) / (2**12 * 20070)


    def set_phib(self, flux):
        """ Set the bias flux

            :param float flux: Phib (A) [-124.56 uA - 124.50 uA]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert -124.56E-6 <= flux <= 124.5E-6

        # convert flux to register value
        # note that the datasheet states an incorrect conversion
        reg_value = min(int(flux * 2**12 * 20070 / 5 + 2**11), 0xFFF)

        self._command(self.CH_MAGNICON, 'j', self.SEL_WRITE,
            parameter=reg_value, parameter_digits=3)


    def get_monitoring_enable(self):
        """ Get enable for monitoring (generator)

            :param int generator: generator
            :return: whether the generator is enabled
            :rtype: bool
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        _, _, monitoring = self._get_generator_status()

        return monitoring


    def set_monitoring_enable(self, enabled):
        """ set enable for monitoring (generator)

            :param bool enabled: requested state
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        generator_1, generator_2, _ = self._get_generator_status()

        self._set_generator_status(generator_1, generator_2, enabled)


    def get_generator_enable(self, generator):
        """ Get enable for one of the generators

            :param int generator: generator
            :return: whether the generator is enabled
            :rtype: bool
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """
        assert generator in [1, 2]

        generator_1, generator_2, _ = self._get_generator_status()
        if generator == 1:
            return generator_1
        else:
            return generator_2


    def set_generator_enable(self, generator, enabled):
        """ set enable for one of the generators

            :param int generator: generator
            :param bool enabled: requested state
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """
        assert generator in [1, 2]

        generator_1, generator_2, monitoring = self._get_generator_status()
        if generator == 1:
            self._set_generator_status(enabled, generator_2, monitoring)
        else:
            self._set_generator_status(generator_1, enabled, monitoring)


    def get_generator_source(self, generator):
        """ get source for one of the generators

            :param int generator: generator
            :return: source: 0 = Ib, 2 = Vb, 4 = I, 5 = PhiX/IX, 6 = Phib
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        source, _, _, _, _ = self._get_generator_control(generator)

        return source


    def set_generator_source(self, generator, source):
        """ set source for one of the generators

            :param int generator: generator
            :param bool source: 0 = Ib, 2 = Vb, 4 = I, 5 = PhiX/IX, 6 = Phib
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        _, waveform, peak_to_peak, phase, frequency_divider = \
            self._get_generator_control(generator)

        self._set_generator_control(
            generator, source, waveform, peak_to_peak, phase, frequency_divider)


    def get_generator_waveform(self, generator):
        """ get waveform for one of the generators

            :param int generator: generator
            :return: waveform: 0 = triangle, 1 = sawtooth+, 2 = sawtooth-, 3 = square, 4 = sine,
                5 = noise, 8 - F = don't know
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        _, waveform, _, _, _ = self._get_generator_control(generator)

        return waveform


    def set_generator_waveform(self, generator, waveform):
        """ set waveform for one of the generators

            :param int generator: generator
            :param bool waveform: 0 = triangle, 1 = sawtooth+, 2 = sawtooth-, 3 = square, 4 = sine,
                5 = noise, 8 - F = don't know
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        source, _, peak_to_peak, phase, frequency_divider = self._get_generator_control(generator)

        self._set_generator_control(
            generator, source, waveform, peak_to_peak, phase, frequency_divider)


    def get_generator_peak_to_peak(self, generator):
        """ get waveform for one of the generators

            :param int generator: generator
            :return: peak_to_peak: peak-to-peak value [0 - 0xFFFF]
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        _, _, peak_to_peak, _, _ = self._get_generator_control(generator)

        return peak_to_peak


    def set_generator_peak_to_peak(self, generator, peak_to_peak):
        """ set waveform for one of the generators

            :param int generator: generator
            :param bool peak_to_peak: peak-to-peak value [0 - 0xFFFF]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        source, waveform, _, phase, frequency_divider = self._get_generator_control(generator)

        self._set_generator_control(
            generator, source, waveform, peak_to_peak, phase, frequency_divider)


    def get_generator_phase(self, generator):
        """ get phase for one of the generators

            :param int generator: generator
            :return: phase: phase shift [0 - 3] => [0 - 270 degrees]
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        _, _, phase, _, _ = self._get_generator_control(generator)

        return phase


    def set_generator_phase(self, generator, phase):
        """ set phase for one of the generators

            :param int generator: generator
            :param bool phase: phase shift [0 - 3] => [0 - 270 degrees]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        source, waveform, peak_to_peak, _, frequency_divider = self._get_generator_control(
            generator)

        self._set_generator_control(
            generator, source, waveform, peak_to_peak, phase, frequency_divider)


    def get_generator_frequency_divider(self, generator):
        """ get frequency divider for one of the generators

            :param int generator: generator
            :return: frequency_divider: division = 2**value [0 - A]
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        _, _, _, _, frequency_divider = self._get_generator_control(generator)

        return frequency_divider


    def set_generator_frequency_divider(self, generator, frequency_divider):
        """ set frequency divider for one of the generators

            :param int generator: generator
            :param bool frequency_divider: division = 2**value [0 - A]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        source, waveform, peak_to_peak, phase, _ = self._get_generator_control(generator)

        self._set_generator_control(
            generator, source, waveform, peak_to_peak, phase, frequency_divider)


    def get_heater_time(self):
        """ Get heater time

           :return: heating time (s)
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        heating_time, _, _, _ = self._get_heater_control()

        return heating_time


    def set_heater_time(self, heating_time):
        """ Set heater time

            :param float heating_time: heating_time (s)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        _, pwm_duty, pwm_enable, jj_heater = self._get_heater_control()

        self._set_heater_control(heating_time, pwm_duty, pwm_enable, jj_heater)


    def get_heater_pwm_duty(self):
        """ Get heater time

            :return: PWM duty cycle (%)
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        _, pwm_duty, _, _ = self._get_heater_control()

        return pwm_duty


    def set_heater_pwm_duty(self, pwm_duty):
        """ Set heater time

            :param float pwm_duty: pwm duty cycle [1% - 99%]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        heating_time, _, pwm_enable, jj_heater = self._get_heater_control()

        self._set_heater_control(heating_time, pwm_duty, pwm_enable, jj_heater)


    def get_heater_pwm_enable(self):
        """ Get heater time

            :return: PWM enabled
            :rtype: bool
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        _, _, pwm_enable, _ = self._get_heater_control()

        return pwm_enable


    def set_heater_pwm_enable(self, pwm_enable):
        """ en-/disable heater PWM

            :param bool pwm_enable: pwm enable
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        heating_time, pwm_duty, _, jj_heater = self._get_heater_control()

        self._set_heater_control(heating_time, pwm_duty, pwm_enable, jj_heater)


    def get_heater_jj_heater(self):
        """ Get heater time

            :return: josephson junction heater current setting. i(A) = jj_heater * 35 mA
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        _, _, _, jj_heater = self._get_heater_control()

        return jj_heater


    def set_heater_jj_heater(self, jj_heater):
        """ en-/disable heater PWM

            :param int jj_heater: josephson junction heater current setting.
                i(A) = jj_heater * 35 mA
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        heating_time, pwm_duty, pwm_enable, _ = self._get_heater_control()

        self._set_heater_control(heating_time, pwm_duty, pwm_enable, jj_heater)


    def get_switches(self):
        """ Get switch settings

            :return: 28 bits of switch settings
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._command(self.CH_MAGNICON, 'a', self.SEL_READ)


    def set_switches(self, value):
        """ Set switch settings

            :param int value: 28 bits of switch settings
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value < 2**28
        self._command(self.CH_MAGNICON, 'a', self.SEL_WRITE, parameter=value)


    def get_generator_control_base(self):
        """ Get the raw current source settings (current and range)

            :return: base timer and increment (N & M)
            :rtype: (int, int)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_MAGNICON, 'v', self.SEL_READ)

        timer = reg_value >> 8 & 0xFFFF
        increment = reg_value & 0xFF

        return timer, increment


    def set_generator_control_base(self, timer, increment):
        """ Set the raw current source settings (current and range)

            :param int timer: N [0 - 0xFFFF]
            :param int increment: M [0 - 0xFF]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert 0 <= timer <= 0xFFFF
        assert 0 <= increment <= 0xFF

        reg_value = timer << 8 | increment

        self._command(self.CH_MAGNICON, 'v', self.SEL_WRITE,
            parameter=reg_value, parameter_digits=7)


    def _get_amp_mode(self):
        """ Get the amp gain and bandwidth

            gain: 0 = 1100 (220), 1 = 140 (280), 2 = 1700 (340), 3 = 2000 (400)
            bandwidth: 0 = 0.2, 1 = 0.7, 2 = 1.5, 3 = 3.5, 4 = 20, 5 = full, 6 = AC amp off (MHz)

            :return: (gain, bandwitdth)
            :rtype: tuple(int, int)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        mode = self._command(self.CH_MAGNICON, 'S', self.SEL_READ)
        return (mode & 0xF, mode >> 4 & 0xF)


    def _set_amp_mode(self, gain, bandwidth):
        """ Set the amp gain and bandwidth

            bandwidth: 0 = 0.2, 1 = 0.7, 2 = 1.5, 3 = 3.5, 4 = 20, 5 = full, 6 = AC amp off (MHz)

            :param int gain: 0 = 1100 (220), 1 = 140 (280), 2 = 1700 (340), 3 = 2000 (400)
            :param int bandwidth: see above
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert gain in range(4)
        assert bandwidth in range(7)

        parameter = gain | bandwidth << 4
        self._command(self.CH_MAGNICON, 'S', self.SEL_WRITE,
            parameter=parameter, parameter_digits=2)


    def _get_generator_control(self, generator):
        """ Get the raw current source settings (current and range)

            :return: source, waveform, peak_to_peak, phase, frequency_divider
            :rtype: (int, int, int, int, int)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert generator in [1, 2]

        # note that the datasheet incorrectly states selectors 1/9 for both generators
        selector = self.SEL_READ + generator

        reg_value = self._command(self.CH_MAGNICON, 'v', selector)

        source            = reg_value >> 28 & 0xF
        waveform          = reg_value >> 24 & 0xF
        peak_to_peak      = reg_value >> 8 & 0xFFFF
        phase             = reg_value >> 4 & 0xF
        frequency_divider = reg_value & 0xF

        return source, waveform, peak_to_peak, phase, frequency_divider


    def _set_generator_control(
            self, generator, source, waveform, peak_to_peak, phase, frequency_divider):
        """ Set the raw current source settings (current and range)

            :param int generator: select generator 1 or 2
            :param int source: 0 = Ib, 2 = Vb, 4 = I, 5 = PhiX/IX, 6 = Phib
            :param int waveform: 0 = triangle, 1 = sawtooth+, 2 = sawtooth-, 3 = square, 4 = sine,
                5 = noise, 8 - F = don't know
            :param int peak_to_peak: peak-to-peak value [0 - 0xFFFF]
            :param int phase: phase shift [0 - 3] => [0 - 270 degrees]
            :param int frequency_divider: division = 2**value [0 - A]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert generator in [1, 2]
        assert source in [0, 2, 4, 5 ,6, 7]
        assert 0 <= waveform <= 0xF
        assert 0 <= peak_to_peak <= 0xFFFF
        assert phase in range(4)
        assert frequency_divider in range(11)

        # note that the datasheet incorrectly states selectors 1/9 for both generators
        selector = self.SEL_WRITE + generator

        reg_value = \
            source << 28 | waveform << 24 | peak_to_peak << 8 | phase << 4 | frequency_divider

        self._command(self.CH_MAGNICON, 'v', selector, parameter=reg_value, parameter_digits=8)


    def _get_generator_status(self):
        """ Get the raw current source settings (current and range)

            :return: enable bits for generator_1, generator_2, monitoring
            :rtype: (int, int, int)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_MAGNICON, 'v', self.SEL_READ + 4)

        generator_1 = reg_value >> 2 & 0x1
        generator_2 = reg_value >> 1 & 0x1
        monitoring  = reg_value & 0x1

        return generator_1, generator_2, monitoring


    def _set_generator_status(self, generator_1, generator_2, monitoring):
        """ Set the raw current source settings (current and range)

            :param int generator_1: enable generator 1 [0, 1]
            :param int generator_2: enable generator 2 [0, 1]
            :param int monitoring: enable monitoring [0, 1]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert generator_1 in [0, 1]
        assert generator_2 in [0, 1]
        assert monitoring in [0, 1]

        reg_value = monitoring << 2 | generator_2 << 1 | generator_1

        self._command(self.CH_MAGNICON, 'v', self.SEL_WRITE + 4,
            parameter=reg_value, parameter_digits=1)


    def _get_heater_control(self):
        """ Get heater control settings

            Note there is also a heater_auto field but is has no effect on the HW so it is omitted

            :return: heating_time (s), pwm_duty (%), pwm_enable, jj_heater
            :rtype: (float, int, bool, int)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_MAGNICON, 'u', self.SEL_READ)

        heating_time_raw = reg_value >> 24 & 0xFFFF
        pwm_duty         = int((reg_value >> 16 & 0xFF) * 100 / 255)
        pwm_enable       = bool(reg_value >> 12 & 0xF)
        jj_heater        = reg_value >> 8 & 0xF

        if pwm_enable:
            heating_time = heating_time_raw * 145.633 / 2**16
        else:
            heating_time = heating_time_raw * 65.403 / 2**16

        return (heating_time, pwm_duty, pwm_enable, jj_heater)


    def _set_heater_control(
            self, heating_time, pwm_duty, pwm_enable, jj_heater):
        """ Set heater control settings

            Note there is also a heater_auto field but is has no effect on the HW so it is omitted

            :param float heating_time: heating time (s)
            :param int pwm_duty: pwm duty cycle (%s)
            :param bool pwm_enable: pwm enable
            :param int jj_heater: josephson junction heater current setting.
                i(A) = jj_heater * 35 mA
            :return: heating_time (s), pwm_duty (%), pwm_enable, jj_heater, heater_auto
            :rtype: (float, float, bool, int, int)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert isinstance(pwm_duty, int) and 0 < pwm_duty < 100 # 0 and 100% are not alowed
        assert isinstance(pwm_enable, bool)
        assert jj_heater in range(4)

        if pwm_enable:
            assert 0 < heating_time < 145.633
        else:
            assert 0 < heating_time < 65.403

        heating_time_raw = int(heating_time * 2**16 / 145.633)

        reg_value = 0x0
        reg_value |= jj_heater << 8
        reg_value |= pwm_enable << 12
        reg_value |= int(pwm_duty * 255 / 100) << 16
        reg_value |= heating_time_raw << 24

        self._command(self.CH_MAGNICON, 'u', self.SEL_WRITE,
            parameter=reg_value, parameter_digits=10)


    def _get_switch_bit(self, index):
        """ Get specific bit from switch settings

            :param in index: bit index (max 27)
            :return: single bit setting value
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert 0 <= index < 28
        switches = self.get_switches()
        return switches >> index & 1


    def _set_switch_bit(self, index, value):
        """ Set switch settings

            :param int index: bit position (max 27)
            :param int value: bit value [0, 1]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert 0 <= index < 28
        assert value in [0, 1]

        switches = self.get_switches()
        switches &= ~(1 << index)  # erase old bit
        switches |= value << index # set new bit
        self.set_switches(switches)


    def cse_get_lncs(self):
        """ Get CSE-1 LNCS current

            :return: Ilncs (A) [-50005.3 uA - 49999.2 uA]
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_CSE, 'q', self.SEL_READ)
        raw_current = reg_value >> 4 # remove the unused range character

        return 5 * 6.001 * (raw_current - 2**13) / (2**14 * 300)


    def cse_set_lncs(self, current):
        """ Set CSE-1 LNCS current

            :param float current: (A) [-50005.3 uA - 49999.2 uA]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert -50005.3E-6 <= current <= 49999.2E-6

        raw_current = int(current * 2**14 * 300 / (5 * 6.001) + 2**13)
        reg_value = raw_current << 4 # add unused range character

        self._command(self.CH_CSE, 'q', self.SEL_WRITE, reg_value, parameter_digits=5)


    def cse_get_i1(self):
        """ Get CSE-1 I1 current

            :return: I1 (A) [-250 uA - 249.9 uA]
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_CSE, 'q', self.SEL_READ + 1)
        raw_current = reg_value >> 4 # remove the unused character

        # This conversion function has an error in the docs
        return 2.5 * (raw_current - 2**11) / (2**11 * (10000 + self.load_resistance))

    def cse_set_i1(self, current):
        """ Set CSE-1 I1 current

            :param float current: (A) [-250 uA - 249.9 uA]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert -250E-6 <= current <= 249.9E-6

        raw_current = int(current * 2**11 * (10000 + self.load_resistance) / 2.5 + 2**11)
        reg_value = raw_current << 4

        # This conversion function has an error in the docs
        self._command(self.CH_CSE, 'q', self.SEL_WRITE + 1, reg_value, parameter_digits=5)


    def cse_get_i2(self):
        """ Get CSE-1 I2 current

            :return: I2 (A) [-5000 uA - 4997.6 uA]
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_CSE, 'q', self.SEL_READ + 2)
        raw_current = reg_value >> 4 # remove the unused character

        # This conversion function has an error in the docs
        return 10 * (raw_current - 2**11) / (2**11 * (2000 + self.load_resistance))


    def cse_set_i2(self, current):
        """ Set CSE-1 I2 current

            :param float current: (A) [-5000 uA - 4997.6 uA]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert -5000E-6 <= current <= 4997.6E-6

        raw_current = int(current * 2**11 * (2000 + self.load_resistance) / 10 + 2**11)
        reg_value = raw_current << 4 # add unused range character

        # This conversion function has an error in the docs
        self._command(self.CH_CSE, 'q', self.SEL_WRITE + 2, reg_value, parameter_digits=5)


    def cse_get_tes1(self):
        """ Get CSE-1 TES1 current

            :return: TES1 (A) [0 - 500 uA]
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_CSE, 'q', self.SEL_READ + 4)
        raw_current = int(reg_value >> 4 & 0xFFFF)
        raw_range = self._cse_get_switch_bit(11)

        return self._current_raw_to_amp(raw_current, raw_range, 3)


    def cse_set_tes1(self, current):
        """ Set the CSE TES-1 current source current

            The range is also set based on the current

            :param int current: (A) [0 - 500 uA]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert -5000E-6 <= current <= 5000E-6

        raw_current, raw_range = self._current_amp_to_raw(current, 3)

        # set range bit separately in switch settings
        self._cse_set_switch_bit(11, raw_range)

        reg_value = raw_current << 4

        self._command(self.CH_CSE, 'q', self.SEL_WRITE + 4,
            parameter=reg_value, parameter_digits=5)


    def cse_get_tes2(self):
        """ Get CSE-1 TES-2 current

            :return: TES2 (A) [0 - 500 uA]
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        reg_value = self._command(self.CH_CSE, 'q', self.SEL_READ + 6)
        raw_current = int(reg_value >> 4 & 0xFFFF)
        raw_range = self._cse_get_switch_bit(13)

        return self._current_raw_to_amp(raw_current, raw_range, 3)


    def cse_set_tes2(self, current):
        """ Set the CSE TES2 current source current

            The range is also set based on the current

            :param int current: (A) [0 - 500 uA]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert -5000E-6 <= current <= 5000E-6

        raw_current, raw_range = self._current_amp_to_raw(current, 3)

        # set range bit separately in switch settings
        self._cse_set_switch_bit(13, raw_range)

        reg_value = raw_current << 4

        self._command(self.CH_CSE, 'q', self.SEL_WRITE + 6,
            parameter=reg_value, parameter_digits=5)


    def cse_get_switches(self):
        """ Get switch settings

            :return: 28 bits of switch settings
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
        """

        return self._command(self.CH_CSE, 'a', self.SEL_READ)


    def cse_set_switches(self, value):
        """ Set switch settings

            :param int value: 28 bits of switch settings
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert value < 2**16
        self._command(self.CH_CSE, 'a', self.SEL_WRITE, parameter=value)


    def _cse_get_switch_bit(self, index):
        """ Get specific bit from CSE switch settings

            :param in index: bit index (max 27)
            :return: single bit setting value
            :rtype: int
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert 0 <= index < 16
        switches = self.cse_get_switches()
        return switches >> index & 1


    def _cse_set_switch_bit(self, index, value):
        """ Set CSE switch bit by doing a read-modify-write operation

            :param int index: bit position (max 27)
            :param int value: bit value [0, 1]
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert 0 <= index < 16
        assert value in [0, 1]

        switches = self.cse_get_switches()
        switches &= ~(1 << index)  # erase old bit
        switches |= value << index # set new bit
        self.cse_set_switches(switches)


    def _current_amp_to_raw(self, current, device_id):
        """ Convert current in A to raw register values (current and range)

            The conversion if dependant on the magnicon device ID.
            The CSE uses ID 0 or 1 for its current sources TES-1 and TES-2

            :param float current: (A) range depends on device_id
            :return: raw_current [0 - 0x3FFF], raw_range [0, 1]
            :param int device_id: magnicon device ID [0 - 3]
            :rtype: (int, int)
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """
        # pylint: disable=invalid-name

        assert device_id in range(4)
        assert 0 <= current <= self.HIGH_RANGE_MAX[device_id]

        if current > self.LOW_RANGE_MAX[device_id]:
            raw_range = 1
            r = self.HIGH_RANGE_R[device_id]
        else:
            raw_range = 0
            r = self.LOW_RANGE_R[device_id]

        if self.load_resistance == 0:
            coeff = 1
        else:
            rp = 1 / (1 / self.load_resistance + 1 / 22143)
            coeff = r / (r + rp) # note that this deviates from the datasheet (which is wrong)

        raw_current = int(current * 2**14 * (r + self.load_resistance) / (10.934 * coeff) + 2**13)

        assert raw_current <= 0x3FFF, 'selected current too high (Rl > 0 decreases range)'

        return raw_current, raw_range


    def _current_raw_to_amp(self, raw_current, raw_range, device_id):
        """ Convert raw register values (current and range) to current in A.

            The conversion if dependant on the magnicon device ID.
            The CSE uses ID 0 or 1 for its current sources TES-1 and TES-2

            :param int raw_current: current register value [0 - 0x3FFF]
            :param int raw_range: range register value [0, 1]
            :param int device_id: magnicon device ID [0 - 3]
            :return: current (A) range depends on device_id
            :rtype: float
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """
        # pylint: disable=invalid-name

        assert device_id in range(4)
        assert raw_range in [0, 1]
        assert 0 < raw_current <= 0x3FFF

        r = self.HIGH_RANGE_R[device_id] if raw_range else self.LOW_RANGE_R[device_id]

        if self.load_resistance == 0:
            coeff = 1
        else:
            rp = 1 / (1 / self.load_resistance + 1 / 22143)
            coeff = r / (r + rp) # note that this deviates from the datasheet (which is wrong)

        current = (10.934 * (raw_current - 2**13)
            / (2**14 * (r + self.load_resistance)) * coeff)

        return current


    def _get_checksum(self, message):
        """ Return the checksum of a message as hex string

            :param str message: command message from starting character up to the checksum
            :return: 1 byte checksum
            :rtype: str
        """

        assert isinstance(message, str)

        checksum = 0
        for char in message:
            checksum += ord(char)

        return hex(checksum).upper()[-2:]


    def _command(self, channel, opcode, selector, parameter=None, parameter_digits=1):
        """ Send a command and check the response for errors

            :param int channel: channel number [0 - 0xFF]
            :param str opcode: command identifier (single character)
            :param int selector: sub command [0 - 16]
            :param int parameter: write data (binary)
            :param int parameter_digits: length of parameter in hex digits or 4-bit nibbles
            :return: read data (binary) or None for writes
            :rtype: int, None
            :raises SerialException: communication interface error
            :raises MagniconError: protocol error
            :raises AssertionError: invalid argument
        """

        assert isinstance(channel, int) and 0 <= channel <= 0xFF, 'channel argument out of range'
        assert isinstance(opcode, str) and len(opcode) == 1, 'command argument must be char'
        assert isinstance(selector, int) and 0 <= selector <= 0xF, 'selector argument out of range'
        assert not selector < 8 or not parameter is None, \
            'no parameter argument given for write command'

        # construct message from channel opcode and selector
        message = '<{:02X}{:1s}{:1X}'.format(channel, opcode, selector)

        # add (variable length) optional parameter
        if not parameter is None:
            message += '{:0{:d}x}'.format(parameter, parameter_digits)

        #  add checksum and termination character
        message += self._get_checksum(message) + '\r'

        logger.debug('send: %s', message)

        # send the command
        self._serial.write(message.encode())

        # broadcasts don't get a response
        if channel != 0xFF:

            response = self._serial.read(1024).decode()
            logger.debug('recv: %s', response)

            if response == '':
                raise serial.SerialException(
                    'Got empty response (did you select the right channel?)')

            # check response terminator
            if response[-1] != '\r':
                raise serial.SerialException(
                    'Incorrect response termination character ({:s})'.format(response[-1]))

            # check response checksum
            if response[-3:-1] != self._get_checksum(response[:-3]):
                raise serial.SerialException(
                    'Incorrect response checksum (got: 0x{:x}, expected: 0x{:x})'.format(
                        response[-3:-1], self._get_checksum(response[:-3])))

            #logger.debug('recv data length = {:d}'.format(len(response) - 4))

            # check for error frames
            if response[0] == '|':
                if response[1] != '0':
                    raise MagniconError(self.ERROR_CODE[int(response[1])])

            # handle read answer
            elif response[0] == '>':
                return int(response[1:-3], 16)

            else:
                raise serial.SerialException(
                    'Unknown staring character for Magnicon response ({:s})'.format(response[0]))

        return None


def main():
    """ Module test function """
    # pylint: disable=protected-access, too-many-statements

    # set up logging
    logging.basicConfig()
    logger.setLevel(logging.DEBUG)

    # create magnicon object
    magnicon = MagniconXXF1(
        vendor_id='0856', product_id='ac03', serial_number='BBP2LQ50')

    # # perform some commands
    # logger.info('Set amp/fll to 1')
    # magnicon.set_amp_fll(1)
    # logger.info('Get amp/fll')
    # logger.info('result = {}'.format(magnicon.get_amp_fll()))

    # print('')
    # logger.info('Set rf@amp to 2')
    # magnicon.set_rf_amp(2)
    # logger.info('Get rf@amp')
    # logger.info('result = {}'.format(magnicon.get_rf_amp()))
    #
    # print('')
    # logger.info('Set rf@fll to 3')
    # magnicon.set_rf_fll(3)
    # logger.info('Get rf@fll')
    # logger.info('result = {}'.format(magnicon.get_rf_fll()))
    #
    # print('')
    # logger.info('Set input termination amp to 1')
    # magnicon.set_input_termination_amp(1)
    # logger.info('Get input termination amp')
    # logger.info('result = {}'.format(magnicon.get_input_termination_amp()))
    #
    # print('')
    # logger.info('Set input termination fll to 1')
    # magnicon.set_input_termination_fll(1)
    # logger.info('Get input termination fll')
    # logger.info('result = {}'.format(magnicon.get_input_termination_fll()))
    #
    # print('')
    # logger.info('Set amp gain polarity to 1')
    # magnicon.set_amp_gain_polarity(1)
    # logger.info('Get amp gain polarity')
    # logger.info('result = {}'.format(magnicon.get_amp_gain_polarity()))
    #
    # print('')
    # logger.info('Set amp gain to 1')
    # magnicon.set_amp_gain(1)
    # logger.info('Get amp gain')
    # logger.info('result = {}'.format(magnicon.get_amp_gain()))
    #
    # print('')
    # logger.info('Set amp bandwidth to 1')
    # magnicon.set_amp_bandwidth(1)
    # logger.info('Get amp bandwidth')
    # logger.info('result = {}'.format(magnicon.get_amp_bandwidth()))
    #
    # print('')
    # logger.info('Set GBWP to 16')
    # magnicon.set_gain_bandwidth_product(16)
    # logger.info('Get GBWP')
    # logger.info('result = {}'.format(magnicon.get_gain_bandwidth_product()))
    #
    # print('')
    # logger.info('Set Ib to 40 uA')
    # magnicon.set_ib(40E-6)
    # logger.info('Get Ib')
    # logger.info('result = {}'.format(magnicon.get_ib()))
    #
    # print('')
    # logger.info('Set Vb to 130 uA')
    # magnicon.set_vb(130E-6)
    # logger.info('Get Vb')
    # logger.info('result = {}'.format(magnicon.get_vb()))
    #
    # print('')
    # logger.info('Set Phib to 12.45 uA')
    # magnicon.set_phib(12.45E-6)
    # logger.info('Get Phib')
    # logger.info('result = {}'.format(magnicon.get_phib()))
    #
    # print('')
    # logger.info('Set generator base')
    # magnicon.set_generator_control_base(0x2000, 0x40)
    # logger.info('Get generator base')
    # logger.info('result = {}'.format(magnicon.get_generator_control_base()))
    #
    # print('')
    # logger.info('Set generator 1')
    # magnicon._set_generator_control(1, 2, 1, 0x2000, 1, 2)
    # logger.info('Get generator 1')
    # logger.info('result = {}'.format(magnicon._get_generator_control(1)))
    #
    # print('')
    # logger.info('Set generator enable')
    # magnicon._set_generator_status(0, 1, 1)
    # logger.info('Get generator enable')
    # logger.info('result = {}'.format(magnicon._get_generator_status()))
    #
    # print('')
    # logger.info('Set heater control')
    # magnicon._set_heater_control(1, 10, True, 0)
    # logger.info('Get heater control')
    # logger.info('result = {}'.format(magnicon._get_heater_control()))
    #
    # print('')
    # logger.info('Set I to 25 uA (10%% of low range)')
    # magnicon.set_i(25E-6)
    # logger.info('Get I')
    # logger.info('result = {}'.format(magnicon.get_i()))
    #
    # print('')
    # logger.info('Set I to 500 uA (10%% of high range)')
    # magnicon.set_i(500E-6)
    # logger.info('Get I')
    # logger.info('result = {}'.format(magnicon.get_i()))

    # # CSE functions
    # print('')
    # logger.info('Set CSE LNCS to -50 uA')
    # magnicon.cse_set_lncs(-50E-6)
    # logger.info('Get CSE LNCS')
    # logger.info('result = {}'.format(magnicon.cse_get_lncs()))

    print('')
    logger.info('Set CSE I1 to 200 uA')
    magnicon.cse_set_i1(200E-6)
    logger.info('Get CSE I1')
    logger.info('result = {} uA'.format(1E6 * magnicon.cse_get_i1()))

    print('')
    logger.info('Set CSE I2 to 700 uA')
    magnicon.cse_set_i2(700E-6)
    logger.info('Get CSE I2')
    logger.info('result = {} uA'.format(1E6 * magnicon.cse_get_i2()))

    # print('')
    # logger.info('Set CSE TES-1 to 190 uA')
    # magnicon.cse_set_tes1(190E-6)
    # logger.info('Get CSE TES-1')
    # logger.info('result = {} uA'.format(1E6 * magnicon.cse_get_tes1()))
    # #
    # print('')
    # logger.info('Set CSE TES-1 to 300 uA ')
    # magnicon.cse_set_tes1(300E-6)
    # logger.info('Get CSE TES-1')
    # logger.info('result = {} uA'.format(1E6 * magnicon.cse_get_tes1()))
    #
    # print('')
    # logger.info('Set CSE TES-2 to 500 uA')
    # magnicon.cse_set_tes2(500E-6)
    # logger.info('Get CSE TES-2')
    # logger.info('result = {} uA'.format(1E6 * magnicon.cse_get_tes2()))

if __name__ == '__main__':
    main()
