#!/usr/bin/env python3
# pylint: disable=missing-docstring
""" :note: The documentation is in :class:`.Maxpos`. """
import time
import logging
import struct

try:
    # pysoem library is not available as Debian package
    import pysoem
except ImportError:
    print('pysoem module is not available')
    pysoem = None

logger = logging.getLogger(__name__)


class MotorType:
    """ Enumerates the different motor types supported by the maxpos. """

    TYPE_DC   = 1
    TYPE_BLDC = 10


class SensorType:
    """ Enumerates the possible sensor options for the motor.
        Multiple sensor types can be combined by bitwise OR-ing them.
    """

    NONE                  = 0x00000000
    DIG_INC_ENCODER_1     = 0x00000001 # port X5
    DIG_INC_ENCODER_2     = 0x00000100 # port X6
    ANALOG_INC_ENCODER    = 0x00000200 # port X6
    SSI_ENCODER           = 0x00000300 # port X6
    BISS_ENCODER          = 0x00000500 # port X6
    DIG_HALL_SENSOR       = 0x00100000 # port X4


class Maxpos:
    """ Maxpos ethercat motor controller interface.

        This module gets the ethercat slave object from the contructor.
        This is necesarry to manage multiple ethercat slaves connected to the same interface.
        A nice side effect is that this keeps the ethercat dependency out of the egress repository

        Note that the controller can be configured for different types of motors. Currently this
        module assumes the default which is BLDC.
    """

    # register name                index        subindex    type        length (bytes)
    REG_DEVICETYPE              = (0x1000,      0x00,       'uint',     4)
    REG_MAN_DEV_NAME            = (0x1008,      0x00,       'str',      None)
    REG_SENSOR_CONFIG           = (0x3000,      0x01,       'uint',     4)
    REG_CONTROL_STRUCTURE       = (0x3000,      0x02,       'uint',     4)
    REG_TORQUE_CONSTANT         = (0x3001,      0x05,       'uint',     4)
    REG_ENCODER_1_PULSE_NUMBER  = (0x3010,      0x01,       'uint',     4)
    REG_ENCODER_1_TYPE          = (0x3010,      0x02,       'uint',     2)
    REG_ERROR_CODE              = (0x603F,      0x00,       'uint',     2)
    REG_CONTROLWORD             = (0x6040,      0x00,       'uint',     2)
    REG_STATUSWORD              = (0x6041,      0x00,       'uint',     2)
    REG_MODES_OF_OPERATION      = (0x6060,      0x00,       'uint',     1)
    REG_TARGET_POSITION         = (0x607A,      0x00,       'int',      4)
    REG_MAX_PROFILE_VELOCITY    = (0x607F,      0x00,       'uint',     4)
    REG_MAX_MOTOR_SPEED         = (0x6080,      0x00,       'uint',     4)

    REG_PROFILE_VELOCITY        = (0x6081,      0x00,       'uint',     4)
    REG_PROFILE_ACCELERATION    = (0x6083,      0x00,       'uint',     4)
    REG_PROFILE_DECELERATION    = (0x6084,      0x00,       'uint',     4)
    REG_QUICK_STOP_DECELERATION = (0x6085,      0x00,       'uint',     4)
    REG_MOTION_PROFILE_TYPE     = (0x6086,      0x00,       'int',      2)
    REG_HOMING_METHOD           = (0x6098,      0x00,       'int',      1)
    REG_DIGITAL_INPUTS          = (0x60FD,      0x00,       'uint',     4)
    REG_TARGET_VELOCITY         = (0x60FF,      0x00,       'int',      4)
    REG_MOTOR_TYPE              = (0x6402,      0x00,       'uint',     2)
    REG_SUPPORTED_DRIVE_MODES   = (0x6502,      0x00,       'uint',     4)

    CONTROL_SWITCH_ON           = 0x0001
    CONTROL_ENABLE_VOLTAGE      = 0x0002
    CONTROL_QUICK_STOP          = 0x0004
    CONTROL_ENABLE_OPERATION    = 0x0008
    CONTROL_FAULT_RESET         = 0x0080

    # struct format for little endian signed integers
    FORMAT = {1: '<b', 2: '<h', 4: '<l'}


    def __init__(self, slave, motortype, sensortype, torque):
        """ Constructor.

            :param slave: pysoem slave object corresponding to the maxpos device
            :param int motortype: select which type of motor is connected to the controller
            :param float torque: motor torque in Nm/A
        """

        self.slave = slave
        self.setRegister(self.REG_MOTOR_TYPE, motortype)
        self.setRegister(self.REG_SENSOR_CONFIG, sensortype)
        self.setRegister(self.REG_TORQUE_CONSTANT, int(torque * 1E6))


    def _del__(self):
        """ Destructor. """

        self.disable()


    def enable(self):
        """ The motor must be enabled before it can run. Accessing control related registers
            while the controller is enabled generates an exception.
        """

        self.setControlword(self.CONTROL_FAULT_RESET)
        self.setControlword(self.CONTROL_ENABLE_VOLTAGE | self.CONTROL_QUICK_STOP)
        self.setControlword(self.CONTROL_ENABLE_OPERATION
                            | self.CONTROL_QUICK_STOP
                            | self.CONTROL_ENABLE_VOLTAGE
                            | self.CONTROL_SWITCH_ON)

        time.sleep(0.01)


    def disable(self):
        """ Disable the controller """

        self.setControlword(0)


    def getRegister(self, regInfo):
        """ Read a register from the device and parse the raw response

            :param regInfo: tuple (register address, subaddress, datatype, byte length)
        """

        index, subindex, datatype, length = regInfo

        # read from the device
        rawdata = self.slave.sdo_read(index, subindex)

        # parse and return the raw response
        if datatype == 'str':
            return rawdata.decode('utf-8')
        elif datatype == 'int':
            return struct.unpack(self.FORMAT[length], rawdata)[0]
        else:
            return struct.unpack(self.FORMAT[length].upper(), rawdata)[0]


    def setRegister(self, regInfo, value):
        """ Write a register from the device

            :param regInfo: tuple (register address, subaddress, datatype, byte length)
            :param type data: data to write
        """

        index, subindex, datatype, length = regInfo # assume write data is always an integer

        # we never write strings to the device
        if datatype == 'int':
            rawdata = struct.pack(self.FORMAT[length], value)
        else:
            rawdata = struct.pack(self.FORMAT[length].upper(), value)

        self.slave.sdo_write(index, subindex, rawdata)


    def setVelocityMode(self, acceleration):
        """ Set the motor to profile velocity mode

            :param int acceleration: motor acceleration (and deceleration) in rpm/s
        """

        self.setRegister(self.REG_PROFILE_ACCELERATION, acceleration)
        self.setRegister(self.REG_PROFILE_DECELERATION, acceleration)
        self.setRegister(self.REG_MODES_OF_OPERATION, 3)


    def setPositionMode(self, acceleration, velocity):
        """ Set the motor to profile position mode

            :param int acceleration: motor acceleration (and deceleration) in rpm/s
        """

        self.setRegister(self.REG_PROFILE_ACCELERATION, acceleration)
        self.setRegister(self.REG_PROFILE_DECELERATION, acceleration)
        self.setRegister(self.REG_PROFILE_VELOCITY, velocity)
        self.setRegister(self.REG_MODES_OF_OPERATION, 1)


    def setTargetVelocity(self, velocity):
        """ Set the target velocity for profile velocity mode

            :param int velocity: motor velocity setpoint in rpm
        """

        self.setRegister(self.REG_TARGET_VELOCITY, velocity)
        self._performOperation()


    def setTargetPosition(self, position):
        """ Set the target positon for profile position mode

            :param int position: motor target setpoint in encoder steps
        """

        self.setRegister(self.REG_TARGET_POSITION, position)
        self._performOperation()


    def getManufacturerDeviceName(self):
        return self.getRegister(self.REG_MAN_DEV_NAME)

    def getControlword(self):
        return self.getRegister(self.REG_CONTROLWORD)

    def getStatusword(self):
        return self.getRegister(self.REG_STATUSWORD)

    def getDigitalInputs(self):
        return self.getRegister(self.REG_DIGITAL_INPUTS)

    def getErrorCode(self):
        return self.getRegister(self.REG_ERROR_CODE)

    def setControlword(self, value):
        self.setRegister(self.REG_CONTROLWORD, value)

    def setControlStructure(self, value):
        """ Select which sensor to use for the control loop.

            :param int value: select sensor port, 1 - X5, 2 - X6, 3 - X4
        """
        self.setRegister(self.REG_CONTROL_STRUCTURE, value)

    def resetFault(self):
        """ Reset faults detected by the controller. """
        self.setRegister(self.REG_CONTROLWORD, 0x80)


    def _performOperation(self):
        """ Operations are executed by this write to the control word """

        self.setControlword(Maxpos.CONTROL_ENABLE_OPERATION
                            | Maxpos.CONTROL_QUICK_STOP
                            | Maxpos.CONTROL_ENABLE_VOLTAGE
                            | Maxpos.CONTROL_SWITCH_ON)


def main():
    logging.basicConfig()

    if pysoem is None:
        print('EtherCat wrapper module not available, quitting...')
        return

    # Initialize ethercat
    master = pysoem.Master()
    master.open('enp4s0')
    master.config_init()

    # Create device object
    maxpos = Maxpos(master.slaves[2], MotorType.TYPE_DC, SensorType.DIG_INC_ENCODER_1, 50.3E-3)

    # Configure device
    maxpos.setRegister(Maxpos.REG_CONTROL_STRUCTURE, 0x11) # control using sensor 1
    maxpos.setRegister(Maxpos.REG_ENCODER_1_PULSE_NUMBER, 256) # 256 counts per turn

    # Read some registers
    print('device name          = {:s}'.format(maxpos.getManufacturerDeviceName()))
    print('control              = 0x{:04X}'.format(maxpos.getControlword()))
    print('status               = 0x{:04X}'.format(maxpos.getStatusword()))
    print('digital inputs       = 0x{:08X}'.format(maxpos.getDigitalInputs()))
    print('error code           = 0x{:04X}'.format(maxpos.getErrorCode()))
    print('sensor type          = 0x{:04X}'.format(maxpos.getRegister(Maxpos.REG_SENSOR_CONFIG)))
    print('supported drive modes= 0x{:04X}'.format(
        maxpos.getRegister(Maxpos.REG_SUPPORTED_DRIVE_MODES)))

    maxpos.enable()

    # # Test position profile mode
    # maxpos.setPositionMode(1000, 6);
    # maxpos.setTargetPosition(10000);
    # print('status               = 0x{:04X}'.format(maxpos.getStatusword()))
    # time.sleep(2);
    # maxpos.setTargetPosition(-10000);
    # print('status               = 0x{:04X}'.format(maxpos.getStatusword()))

    # Test profile velocity mode
    maxpos.setVelocityMode(1000)
    maxpos.setTargetVelocity(100)
    print('status               = 0x{:04X}'.format(maxpos.getStatusword()))
    time.sleep(2)
    maxpos.setTargetVelocity(-100)
    time.sleep(2)
    maxpos.setTargetVelocity(0)

    maxpos.disable()
    master.close()


if __name__ == '__main__':
    main()
