#!/usr/bin/env python3
""" Interface module for Measurement Computing E-1608.

    Ethernet data acquisition device, with 8 SE/4 DIFF analog inputs,
    16-bit resolution, up to 250 kS/s sample rate, two 16-bit analog outputs,
    one 32-bit counter input, and eight digital I/O.
"""

import socket
import struct
import argparse
import logging

logger = logging.getLogger(__name__)

__all__ = ['MccE1608']

class MccSocket:
    """ Class for Measurement Computing sockets. """
    _DISCOVER_PORT = 54211
    _COMMAND_PORT = 54211
    _SCAN_PORT = 54212

    def __init__(self, addr, timeout=None, connectCode=0x00):
        self._addr = addr
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sock.settimeout(timeout)
        self._send(MccSocket._DISCOVER_PORT,
            b'C'+connectCode.to_bytes(5, byteorder='little'))
        data, addr = self._recv()
        if data != b'C\x00' or addr != (self._addr, MccSocket._DISCOVER_PORT):
            logger.error('Failed to discover device.')
            raise IOError('failed to discover device')
        self._sock.close()

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.settimeout(timeout)
        self._sock.connect((self._addr, MccSocket._COMMAND_PORT))
        self._frameId = 0x00
        self._scanSock = None

    def __del__(self):
        try:
            self._sock.close()
            self._scanSock.close()
        except AttributeError:
            pass

    def _openScan(self, timeout):
        self._closeScan()
        self._scanSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._scanSock.settimeout(timeout)
        self._scanSock.connect((self._addr, MccSocket._SCAN_PORT))

    def _closeScan(self):
        try:
            self._scanSock.close()
        except AttributeError:
            pass

    def _sendScan(self, data):
        self._scanSock.sendto(data, (self._addr, MccSocket._SCAN_PORT))

    def _recvScan(self, length):
        return self._scanSock.recv(length)

    def _send(self, port, data):
        self._sock.sendto(data, (self._addr, port))

    def _recv(self):
        return self._sock.recvfrom(64)

    @staticmethod
    def _checksum(data):
        return 0xFF - (sum(data) % 0x100)

    def _query(self, cmd, data=b''):
        # Datagram: start (0xDB), command, frame id, status, data length (n) - low,
        #           data length (n) - high, data (n-bytes), checksum.
        sendBuffer = b'\xDB'
        sendBuffer += cmd.to_bytes(1, byteorder='little')
        self._frameId += 1 # Increment frame id with every send.
        self._frameId %= 0x100
        sendBuffer += self._frameId.to_bytes(1, byteorder='little')
        sendBuffer += b'\x00'
        sendBuffer += len(data).to_bytes(2, byteorder='little')
        sendBuffer += data
        sendBuffer += self._checksum(sendBuffer).to_bytes(1, byteorder='little')

        self._send(self._COMMAND_PORT, sendBuffer)
        recvBuffer, _ = self._recv()

        recvLength = (len(recvBuffer) - 7).to_bytes(2, byteorder='little')
        if recvBuffer[0] != sendBuffer[0]:
            logger.error('Incorrect start-byte received.')
            raise IOError('incorrect start-byte received')
        if recvBuffer[1] != (sendBuffer[1] | 0x80):
            logger.error('Incorrect command-byte received.')
            raise IOError('incorrect command-byte received')
        if recvBuffer[2] != sendBuffer[2]:
            logger.error('Incorrect frame-id received.')
            raise IOError('incorrect frame-id received')
        if recvBuffer[3] == 0x01:
            logger.error('Improper protocol.')
            raise IOError('improper protocol')
        elif recvBuffer[3] == 0x02:
            logger.error('Invalid parameters.')
            raise IOError('invalid parameters')
        elif recvBuffer[3] == 0x03:
            logger.error('Resource busy.')
            raise IOError('resource busy')
        elif recvBuffer[3] == 0x04:
            logger.error('Resource not ready.')
            raise IOError('resource not ready')
        elif recvBuffer[3] == 0x05:
            logger.error('Resource timeout.')
            raise IOError('resource timeout')
        elif recvBuffer[3] == 0x06 or recvBuffer[3] != 0:
            logger.error('Other error.')
            raise IOError('other error')
        if recvBuffer[4] != recvLength[0] or recvBuffer[5] != recvLength[1]:
            logger.error('Incorrect data-length received.')
            raise IOError('incorrect data-length received')
        if  recvBuffer[-1] != self._checksum(recvBuffer[:-1]):
            logger.error('Incorrect checksum received.')
            raise IOError('incorrect checksum received')
        return recvBuffer[6:-1]


class MccE1608(MccSocket):
    """ Class for Measurement Computing E-1608. """
    _AIN_CHNLS = ['CH0H', 'CH0L', 'CH1H', 'CH1L', 'CH2H', 'CH2L',
                   'CH3H', 'CH3L', 'CH0', 'CH1', 'CH2', 'CH3']
    _AIN_RANGES = ['10V', '5V', '2V', '1V']

    def __init__(self, addr, timeout=1):
        super().__init__(addr, timeout)
        self._buildGainTables()
        self._queue = []

    def getDIn(self, raw=False):
        """ This method reads the current state of the DIO pins. A 0 in
        a bit position indicates the corresponing pin is reading a
        low state, and a 1 indicates a high state.

        :param bool raw: Raw byte value or list of bits [DIO0, ... , DIO7].
        """
        val = int.from_bytes(self._query(0x00), byteorder='little')
        if raw:
            return val
        else:
            return [(val >> idx) & 0x01 for idx in range(8)]

    def setDOut(self, val):
        """ This method writes the DIO latch value. The factory power
        on default is all 1 (pins are floating). Writing a 0 to a
        bit will set the corresponding pin driver low, writing a 1
        sets it high.

        :param val: Value of state of DIO pins.
                    Raw byte value or list of bits [DIO0, ... , DIO7].
        """
        if isinstance(val, list):
            val = sum([(val[idx] & 0x01) << idx for idx in range(8)])
        self._query(0x03, bytes([val]))

    def getDOut(self, raw=False):
        """ This method reads the DIO output latch value. The factory power
        on default is all 0. A 0 in bit position indicates the
        corresponding pin driver is low a 1 indicates it is high.

        :param bool raw: Raw byte value or list of bits [DIO0, ... , DIO7].
        """
        val = int.from_bytes(self._query(0x02), byteorder='little')
        if raw:
            return val
        else:
            return [(val >> idx) & 0x01 for idx in range(8)]

    def setDConfig(self, val):
        """ This method writes the DIO configuration value. A 1 in a bit
        position sets the corresponding pin to an input, a 0 sets it to an
        output. The power on default is all 1 (input).

        :param val: Value of DIO configration.
                    Raw byte value or list of bits [DIO0, ... , DIO7].
        """
        if isinstance(val, list):
            val = sum([(val[idx] & 0x01) << idx for idx in range(8)])
        self._query(0x05, bytes([val]))

    def getDConfig(self, raw=False):
        """ This method reads the DIO configuration value. A 1 in a bit
        position indicates the corresponding pin is set to an input a
        0 indicates it is set to an output. The power on default is al 1
        (input).

        :param bool raw: Raw byte value or list of bits [DIO0, ... , DIO7].
        """
        val = int.from_bytes(self._query(0x04), byteorder='little')
        if raw:
            return val
        else:
            return [(val >> idx) & 0x01 for idx in range(8)]

    def getAIn(self, channel, rnge, raw=False):
        """ Reads the value of analog input channel. This method will not
        return valid data if AIn scan is currently running.

        :param str channel: single-ended 'CH[0-3][H-L]' or differential 'CH[0-3]'.
        :param str rnge: Input range: +/-'10V, 5V, 2V, 1V'.
        :param bool raw: Raw value or floating-point voltage value.
        :return: Value of analog input channel.
        """
        idxRange = self._AIN_RANGES.index(rnge)
        idxChnl = self._AIN_CHNLS.index(channel)
        val = int.from_bytes(self._query(0x10,
            bytes([idxChnl, idxRange])), byteorder='little')
        if idxChnl < 8:
            val = int(val * self._table_AInSE[idxRange][0] + self._table_AInSE[idxRange][1])
        else:
            val = int(val * self._table_AInDF[idxRange][0] + self._table_AInDF[idxRange][1])
        if val > 0xFFFF:
            val = 0xFFFF
        elif val < 0:
            val = 0
        if raw:
            return val
        else:
            return self._voltage(val, rnge)

    def getAOut(self, raw=False):
        """ This method reads the values of the analog output channels.

        :param bool raw: Raw values or floating-point voltage values.
        :return: Values analog output channels (AOUT0, AOUT1).
        """
        val = self._query(0x20)
        valChnl0 = int.from_bytes(val[0:2], byteorder='little')
        valChnl1 = int.from_bytes(val[2:4], byteorder='little')
        valChnl0 = int(valChnl0 / self._table_AOut[0][0] - \
            self._table_AOut[0][1] / self._table_AOut[0][0])
        valChnl1 = int(valChnl1 / self._table_AOut[1][0] - \
            self._table_AOut[1][1] / self._table_AOut[1][0])
        if valChnl0 > 0xFFFF:
            valChnl0 = 0xFFFF
        elif valChnl0 < 0:
            valChnl0 = 0
        if valChnl1 > 0xFFFF:
            valChnl1 = 0xFFFF
        elif valChnl1 < 0:
            valChnl1 = 0
        if raw:
            return valChnl0, valChnl1
        else:
            return self._voltage(valChnl0, '10V'), self._voltage(valChnl1, '10V')

    def setAOut(self, channel, val, raw=False):
        """ This method writes the value of analog output channel.

        :param str channel: Channel to write 'AOUT0' or 'AOUT1'.
        :param val: Value of analog output channel (-10V-10V).
        :param bool raw: Raw value or floating-point voltage value.
        """
        idxChnl = ['AOUT0', 'AOUT1'].index(channel)
        if not raw:
            val = self._raw(val)
        val = int(val * self._table_AOut[idxChnl][0] + self._table_AOut[idxChnl][1])
        if val > 0xFFFF:
            val = 0xFFFF
        elif val < 0:
            val = 0
        data = idxChnl.to_bytes(1, byteorder='little')
        data += val.to_bytes(2, byteorder='little')
        self._query(0x21, data)

    def scanStartAIn(self, count, frequency, trigger='NONE'):
        """ This method starts an analog input scan. The channel ordering
        and number of channels per scan is set by the channel gain queue,
        which must first be set with the setQueueAIn() method.

        :param count: Total number of scans to acquire, 0 for continuous scan.
        :param frequency: Sampling frequency; max. 250 kHz.
                          Use 0 for external clock on 'AICKI' pin.
        :param trigger: External trigger to start scan 'NONE, RISE, FALL, HIGH, LOW'.
                        (default: NONE)
        """
        if frequency < 0 or frequency > 250000:
            raise ValueError('frequency out of range')
        if frequency == 0:
            pacerPeriod = 0
        else:
            pacerPeriod = int((80E6 / frequency) - 1.0)
        data = count.to_bytes(4, byteorder='little')
        data += pacerPeriod.to_bytes(4, byteorder='little')
        data += (['NONE', 'RISE', 'FALL', 'HIGH', 'LOW'].index(trigger) \
            << 2).to_bytes(1, byteorder='little')
        self._openScan(1.0 + 1/frequency)
        self._query(0x11, data)
        self._sendScan(bytes([0x0]))

    def scanReadAIn(self, count, raw=False):
        """ This method reads scan data obtained with analog input scan.

        :param count: Total number of scans to read.
        :param bool raw: Raw value or floating-point voltage value.
        :return:
        """
        try:
            replyCount = count * len(self._queue) * 2
        except AttributeError:
            logger.error('Scan not configured.')
            raise
        byteData = b''

        while len(byteData) < replyCount:
            byteData += self._recvScan(replyCount - len(byteData))

        valData = [[] for _ in range(len(self._queue))]
        for idxScan in range(0, len(byteData), len(self._queue)*2):
            for idxQueue in range(0, len(self._queue)):
                val = int.from_bytes(byteData[idxScan+idxQueue*2:idxScan+idxQueue*2+2], \
                    byteorder='little')
                idxRange = self._AIN_RANGES.index(self._queue[idxQueue][1])
                idxChnl = self._AIN_CHNLS.index(self._queue[idxQueue][0])
                if idxChnl < 8:
                    val = int(val * self._table_AInSE[idxRange][0] + self._table_AInSE[idxRange][1])
                else:
                    val = int(val * self._table_AInDF[idxRange][0] + self._table_AInDF[idxRange][1])
                if val > 0xFFFF:
                    val = 0XFFFF
                elif val < 0:
                    val = 0
                if raw:
                    valData[idxQueue].append(val)
                else: valData[idxQueue].append( \
                    self._voltage(val, self._queue[idxQueue][1]))
        return valData

    def setQueueAIn(self, setQueue):
        """ This method write the analog input scan channel gain queue for
        AInScan. This method will result in an error response if an AIn
        scan is currently running.

        :param list setQueue: List of tuples (channel, range); max. 8.
                           e.g. [('CH0H', '5V'), ('CH0L', '2V')]
        """
        if len(setQueue) > 8:
            raise IndexError('too many queue entries')
        self._queue = setQueue
        data = len(setQueue).to_bytes(1, byteorder='little')
        for item in setQueue:
            data += self._AIN_CHNLS.index(item[0]).to_bytes(1, byteorder='little')
            data += self._AIN_RANGES.index(item[1]).to_bytes(1, byteorder='little')
        self._query(0x15, data)

    def getQueueAIn(self):
        """ This method reads the analog input scan channel gain queue.

        :return: List of tuples (channel, range); max. 8.
                 e.g. [('CH0H', '5V'), ('CH0L', '2V')]
        """
        data = self._query(0x14)
        return [(self._AIN_CHNLS[data[idx]], self._AIN_RANGES[data[idx+1]]) \
            for idx in range(1, data[0]+1)]

    def scanStopAIn(self, closeSocket=False):
        """ This method stops the analog input scan (if running).
        It will clear the scan data FIFO.

        :param bool closeSocket: Close the data socket (default: False).
        """
        if closeSocket:
            data = bytes([1])
        else:
            data = bytes([0])
        self._query(0x13, data)
        if closeSocket:
            self._closeScan()

    def getCounter(self):
        """ This method reads the event counter. """
        return int.from_bytes(self._query(0x30), byteorder='little')

    def resetCounter(self):
        """ This method resets the event counter. On a write, the counter will be reset to 0. """
        self._query(0x31)

    def blinkLED(self, count):
        """ Blink the device power-LED.

        :param count: Number of times to blink LED.
        """
        self._query(0x50, bytes([count]))

    def reset(self):
        """ Reset the device. """
        self._query(0x51)

    def getStatus(self):
        """ Get the device status.

        :return: Status integer.
        """
        return int.from_bytes(self._query(0x52), byteorder='little')

    def getNetworkConfig(self):
        """ Get the current network configuration.

        :return: Address, subnet mask and gateway.
        """
        netConf = self._query(0x54)
        return list(netConf[0:4]), list(netConf[4:8]), list(netConf[8:12])

    def _buildGainTables(self):
        """ Builds a lookup table of calibration coefficients to translate into voltages:
        'calibrated code = code * slope + intercept'.
        """
        self._table_AInDF = [self._getCalMemory(idx) for idx in range(0,4)]
        self._table_AInSE = [self._getCalMemory(idx) for idx in range(4,8)]
        self._table_AOut = [self._getCalMemory(idx) for idx in range(8,10)]

    def _getCalMemory(self, idx):
        """ Get calibration from nonvolatile memory.

        :param idx: Index: (0-3) Analog input differential, (10V, 5V, 2V, 1V) gain.
                           (4-7) Analog input single-ended, (10V, 5V, 2V, 1V) gain.
                           (8-9) Two channel analog output.
        :return: Slope and intercept.
        """
        data = (idx * 8).to_bytes(2, byteorder='little')
        data += b'\x04\x00'
        slope = struct.unpack('f', self._query(0x40, data))[0]
        data = (idx * 8 + 4).to_bytes(2, byteorder='little')
        data += b'\x04\x00'
        intercept = struct.unpack('f', self._query(0x40, data))[0]
        return slope, intercept

    @staticmethod
    def _voltage(raw, rnge):
        """ Converts a raw value to voltage value for a given range. """
        if rnge == '10V':
            return (raw - 0x8000) * 10.0 / 0x8000
        elif rnge == '5V':
            return (raw - 0x8000) * 5.0 / 0x8000
        elif rnge == '2V':
            return (raw - 0x8000) * 2.0 / 0x8000
        elif rnge == '1V':
            return (raw - 0x8000) * 1.0 / 0x8000

    @staticmethod
    def _raw(val):
        """ Converts voltage value to a 16 bit raw value for +/-10V output. """
        if val >= 10.0:
            return 0xFFFF
        elif val <= -10.0:
            return 0x00
        else:
            return int(val * 0x8000 / 10 + 0x8000)

def parseMccE1608():
    """ Parse arguments for Measurement Computing E-1608. """
    parser = argparse.ArgumentParser(prog='mcc_e1608',
        description=__doc__)
    parser.add_argument('addr', type=str, help='IP Address.')
    subparsers = parser.add_subparsers(dest='cmd', help='Command.')

    subparsers.add_parser('reset', help='Reset device')
    subparser = subparsers.add_parser('blink', help='Blink power-LED.')
    subparser.add_argument('count', type=int, help='Times to blink LED.')
    subparsers.add_parser('status', help='Get device status.')
    subparsers.add_parser('config', help='Get network configuration.')
    subparser = subparsers.add_parser('get-ain', help='Get analog inputs.')
    subparser.add_argument('channel', type=str, help='Channel: CH[0-3][H-L] or CH[0-3].')
    subparser.add_argument('range', type=str, help='Range: +/- 10V, 5V, 2V, 1V.')
    subparser = subparsers.add_parser('set-aout', help='Set analog output.')
    subparser.add_argument('channel', type=str, help='Channel: AOUT0 or AOUT1.')
    subparser.add_argument('val', type=float, help='Output value: -10V - 10V.')
    subparsers.add_parser('get-aout', help='Get analog outputs.')
    subparser = subparsers.add_parser('set-dconfig', help='Set digital config.')
    subparser.add_argument('val', type=str, help='Config byte, 1 = input, 0 = output.')
    subparsers.add_parser('get-dconfig', help='Get digital config byte, 1 = input, 0 = output.')
    subparser = subparsers.add_parser('set-dout', help='Set digital outputs.')
    subparser.add_argument('val', type=str, help='Digital output byte.')
    subparsers.add_parser('get-din', help='Get digital output byte.')
    subparser = subparsers.add_parser('scan-ain', help='Analog input scan.')
    subparser.add_argument('channel', type=str, help='Channel: CH[0-3]H-L] or CH[0-3].')
    subparser.add_argument('range', type=str, help='Range: +/- 10V, 5V, 2V, 1V.')
    subparser.add_argument('freq', type=int, help='Frequency: 0-250000 Hz.')
    subparser.add_argument('count', type=int, help='Number of samples.')

    return parser.parse_args()

def runMccE1608(args):
    """ Run command on Measurement Computing E-1608. """
    mccE1608 = MccE1608(args.addr)
    if args.cmd == 'reset':
        mccE1608.reset()
    elif args.cmd == 'blink':
        mccE1608.blinkLED(args.count)
    elif args.cmd == 'status':
        print(mccE1608.getStatus())
    elif args.cmd == 'config':
        conf = mccE1608.getNetworkConfig()
        print('addr: ', '.'.join(map(str, conf[0])))
        print('mask: ', '.'.join(map(str, conf[1])))
        print('gateway: ', '.'.join(map(str, conf[2])))
    elif args.cmd == 'get-ain':
        print(mccE1608.getAIn(args.channel, args.range))
    elif args.cmd == 'set-aout':
        mccE1608.setAOut(args.channel, args.val)
    elif args.cmd == 'get-aout':
        print(mccE1608.getAOut()[0])
        print(mccE1608.getAOut()[1])
    elif args.cmd == 'set-dconfig':
        mccE1608.setDConfig(int(args.val, 2))
    elif args.cmd == 'get-dconfig':
        print(''.join(map(str, reversed(mccE1608.getDConfig()))))
    elif args.cmd == 'set-dout':
        mccE1608.setDOut(int(args.val, 2))
    elif args.cmd == 'get-din':
        print(''.join(map(str, reversed(mccE1608.getDIn()))))
    elif args.cmd == 'scan-ain':
        mccE1608.scanStopAIn()
        mccE1608.setQueueAIn([(args.channel, args.range)])
        queue = mccE1608.getQueueAIn()
        print(queue[0][0], queue[0][1])
        mccE1608.scanStartAIn(args.count, args.freq)
        print(', '.join(map(str, mccE1608.scanReadAIn(args.count)[0])))

if __name__ == '__main__':
    runMccE1608(parseMccE1608())
