"""
Support library for Keysight PNA-X communication with SCPI via TCP/IP.
"""


import logging
import socket
import numpy

logger = logging.getLogger(__name__)

class Error(Exception):
    """Raised when a communication error with the PNA-X occurs."""


class PnaX:
    """Control the Keysight PNA-X with SCPI via TCP/IP."""

    def __init__(self, host, port=5025):
        """Construct object and connect to analyzer.

        host    :: String containing host name or IP address.
        port    :: Integer representing TCP port number.
        """

        self.host = host
        self.port = port

        logger.info('Connecting to PNA-X, host=%r, port=%r', host, port)

        self.sock = socket.create_connection(address=(host, port))

    def close(self):
        """Close connection to device.

        This object can no longer be used after the connection is closed.
        """

        if self.sock is not None:
            logger.info('Disconnecting from PNA-X')
            self.sock.close()
        self.sock = None

    def cmd(self, cmd):
        """Send SCPI command to device.

        cmd     :: String containing SCPI command WITHOUT line terminator.
        """

        assert self.sock is not None

        logger.debug('Sending to PNA-X: %r', cmd)

        if not cmd.endswith('\n'):
            cmd = cmd + '\n'

        self.sock.send(cmd.encode('ascii'))

#pylint: disable=invalid-name
    def rx(self):
        """Receive single line response from device.

        Return a string containing the response from the device without
        line terminator.
        """

        assert self.sock is not None

        data = []
        while (not data) or (not data[-1].endswith('\n')):
            s = self.sock.recv(4096)
            logger.debug('Got from PNA-X: %r', s)
            if not s:
                logger.error('Socket closed while waiting for data')
                raise Error("Socket closed while waiting for data")
            data.append(s.decode('ascii'))

        s = ''.join(data)
        return s.strip("\r\n")
#pylint: enable=invalid-name

    def qry(self, cmd):
        """Send SCPI command to device and receive single line response.

        cmd     :: String containing SCPI command WITHOUT line terminator.

        Returns a string containing the answer from the device.
        """

        assert self.sock is not None

        self.cmd(cmd)
        return self.rx()

    def rxdata(self, dataformat):
        """Receive numerical data from the device in ASCII format.

        This function assumes that the device has been configured to
        transfer data in the expected format and that the SCPI command
        to initiate data transfer has already been sent.

        dataformat  :: Expected data format, 'ascii' or 'real32' or 'real64'

        Return a numpy array with received values.
        """

        assert self.sock is not None
        assert dataformat in ('ascii', 'real32', 'real64')

        if dataformat == 'ascii':
            # Receive one line of comma-separated values.
            s = self.rx()
            if s.startswith('#'):
                logger.error('Unexpected block data from device in ASCII mode')
                raise Error('Unexpected block data from device in ASCII mode')

            return numpy.fromstring(s, sep=',')

        else:
            # Binary data block transfer.

            # Receive marker byte '#'.
            s = self.sock.recv(1)
            if s != b'#':
                logger.error("Expected '#' from device but got %r", s)
                raise Error("Invalid block data header from device")

            # Receive byte count.
            s = self.sock.recv(1)
            if s not in b"123456789":
                logger.error('Expected nr of size digits from device but got %r', s)
                raise Error('Invalid block data header from device')

            ndigit = int(s)
            digits = b''

            while len(digits) < ndigit:
                digits += self.sock.recv(ndigit - len(digits))

            if [s for s in digits if s not in b"0123456789"]:
                logger.error('Got invalid block size from device %r', digits)
                raise Error('Invalid block data header from device')

            nbyte = int(digits)
            buf = bytearray(nbyte)

            # Receive binary data.
            p = 0
            while p < nbyte:
                s = self.sock.recv(nbyte - p)
                buf[p:p+len(s)] = s
                p += len(s)

            # Receive end-of-line marker.
            s = self.sock.recv(1)
            if s != b'\n':
                logger.error('Expected ENDL at end of block data but got %r', s)
                raise Error("Invalid block data terminator from device")

            # Convert binary data to floating point numbers.
            if dataformat == 'real32':
                wsize = 4
                dtype = numpy.dtype('>f4')
            elif dataformat == 'real64':
                wsize = 8
                dtype = numpy.dtype('>f8')

            if nbyte % wsize != 0:
                logger.error('Expected multiple of %d bytes for %s data ' +
                             'but got %d bytes', wsize, dataformat, nbyte)
                raise Error("Invalid block data size from device")

            return numpy.frombuffer(memoryview(buf), dtype=dtype)

    def qrydata(self, cmd, dataformat):
        """Send SCPI command to device and receive data response.

        cmd         :: String containing SCPI command WITHOUT line terminator.
        dataformat  :: Preferred data format, 'ascii' or 'real32' or 'real64'

        Return a numpy array with numerical values received from the device.
        """

        assert self.sock is not None
        assert dataformat in ('ascii', 'real32', 'real64')

        if dataformat == 'ascii':
            self.cmd('FORM:DATA ASCII')
        elif dataformat == 'real32':
            self.cmd('FORM:DATA REAL,32')
        elif dataformat == 'real64':
            self.cmd('FORM:DATA REAL,64')

        self.cmd(cmd)

        return self.rxdata(dataformat)

    def identification(self):
        """Query device identification string.

        Returns the device identification string.
        """

        return self.qry('*IDN?')

    def reset(self):
        """Reset device to default state."""

        self.cmd('*RST')

    def clearStatus(self):
        """Clear status, discarding error queue and event status."""

        self.cmd('*CLS')

    def operationComplete(self):
        """Query operation completion status.

        Wait until pending operations completed, then return 1.
        """

        s = self.qry('*OPC?')
        return int(s)

#pylint: disable=invalid-name
    def getError(self):
        """Query device error queue.

        Return (error code, error string) for the oldest pending error,
        or return None if the error queue is empty.
        """

        s = self.qry('SYST:ERR?')
        w = s.split(',')
        if len(w) != 2:
            logger.error('Unexpected device response to SYST:ERR: %r', s)
            raise Error("Unexpected device response to SYST:ERR")

        errcode = int(w[0])
        errstr = w[1]

        if errcode == 0:
            return None

        if errstr.startswith('"') and errstr.endswith('"'):
            errstr = errstr[1:-1].replace('""', '"')

        return (errcode, errstr)
#pylint: enable=invalid-name


