#!/usr/bin/env python3

"""
SRON generic EGSE Support for the Rohde Schwarz network/signal analyzer ZVL.
Also works with Rohde&Schwarz spectrum analyzer FSV.
"""

import logging

from   gssw.lib.textui       import inputChoice, inputNumber, inputYesNo, MenuOption, MenuText,\
                                    showMenu
from   gssw.lib.typing       import check_class
from   gssw.ucelib.gpib      import GpibTcpIp

__all__ = ['RohdeSchwarzZVL']

###############################################################################
# C O N S T A N T S
###############################################################################

logger         = logging.getLogger(__name__)

DEFAULT_PEER_ADDRESS    = "192.168.2.95"  # default address
DEFAULT_PEER_PORT       = 5025            # default port

###############################################################################
# R O H D E - S C H W A R Z   N E T W O R K / S I G N A L   A N A L Y Z E R
###############################################################################

class RohdeSchwarzZVL(GpibTcpIp):
    """Control for the RohdeSchwarz network/signal analyzer ZVL
       only LAN support is implemented!

       example usage:

       try:
           na = RohdeSchwarz_ZVL

           if na.xxxx:
               xxx
           else:
               xxx

        except IOError as inst:
            print inst.args[0]
        except NameError as inst:
            print inst.args[0]
    """

    def __init__(self, host, port, gpibAddr=None):
        """Constructor, for LAN communication"""

        super().__init__(host, port, "analyzer")

        if gpibAddr is not None:
            check_class(gpibAddr, int)
            self.tx('++mode 1\n++addr %d\n++eos 2\n++eoi 1\n' % gpibAddr)


    def enableDisplay(self, enable):
        """Enable/disable interactive display on device."""

        if enable:
            self.tx('SYST:DISP:UPD ON\n'.encode('utf-8'))
        else:
            self.tx('SYST:DISP:UPD OFF\n'.encode('utf-8'))


    def mode(self, mode):
        """This command switches between the measurement modes by means
           of text parameters."""
        if mode[0:3] == "SAN" or mode[0:3] == "san" or \
           mode[0:3] == "NWA" or mode[0:3] == "nwa":
            msg = "INST %s\n" % mode
            self.tx(msg)
        else:
            logger.error("Incorrect mode: %s. Must be SAN or NWA", mode)


    def getCenterFrequency(self):
        """This command queries the center frequency of the analyzer
           or the measuring frequency for span = 0."""
        msg = "SENS:FREQ:CENT?"
        return float(self.ask(msg)) / 1.e6


    def getFrequencySpan(self):
        """This command queries the frequency span in the
           Spectrum Analyzer mode."""
        msg = "SENS:FREQ:SPAN?"
        return float(self.ask(msg)) / 1.e6

    def centerFrequency(self, freq):
        """This command defines the center frequency of the analyzer
           or the measuring frequency for span = 0."""
        msg = "SENS:FREQ:CENT %fMHz\n" % freq
        self.tx(msg)


    def frequencySpan(self, span):
        """This command defines the frequency span in the
           Spectrum Analyzer mode."""
        msg = "SENS:FREQ:SPAN %fMHz\n" % span
        self.tx(msg)


    def contInit(self, enable=None):
        """Enable/disable continuous measurements.

        contInit(True)      Enable continuous measurements
        contInit(False)     Disable continuous measurements
        r = contInit()      Return current status
        """

        if enable is None:
            response = self.ask('INIT:CONT?\n')
            return bool(int(response.strip()))
        elif enable:
            self.tx('INIT:CONT ON\n')
        else:
            self.tx('INIT:CONT OFF\n')


    def measure(self):
        """Initiate a new sweep and wait for completion.
        Waiting does not work when continuous measurements are enabled.
        """

        self.tx('INIT;*WAI\n')


    def refLevel(self, level):
        """Set display reference level in dBm."""

        self.tx('DISP:TRAC:Y:RLEV %gdBm\n' % level)


    def resBandwidth(self, bandwidth, auto=0):
        """Set resolution bandwidth in MHz."""

        if auto:
            self.tx('BAND:RES:AUTO ON\n')
        else:
            self.tx('BAND:RES %gMHz\n' % bandwidth)


    def vidBandwidth(self, bandwidth):
        """Set video bandwidth in MHz."""

        self.tx('BAND:VID %gMHz\n' % bandwidth)


    def sweepPoints(self, points=None):
        """This command defines the number of measurement points to be collected
           during one sweep.
        """

        if points is None:
            msg = ":SWE:POIN?\n"
            self.tx(msg)
            rsp = self.rx()

            if rsp is None:
                logger.error("No response from analyzer")
            else:
                logger.info("Number of sweep points: " + rsp.rstrip())
                return int(rsp)
        else:
            msg = ":SWE:POIN %d\n" % points
            self.tx(msg)

    def inputAttenuation(self, attenuation=None):
        """ This command defines the input attenuation used [dB]
        """

        if attenuation is None:
            msg = ":INP:ATT?\n"
            rsp = self.ask(msg)

            if rsp is None:
                logger.error("No response from analyzer")
            else:
                logger.info("Input attenuation: %s dB", rsp.rstrip())
                return int(rsp)
        else:
            msg = ":INP:ATT %d\n" % attenuation
            self.tx(msg)

    def averages(self, averages=0):
        """Set number of averages to be taken by the instrument.
           Use averages=0 to turn off averaging."""

        if averages:
            self.tx("SENS:AVER:COUNT %d\n" % averages)
            self.tx("SENS:AVER:TYPE POW\n")
            self.tx("SENS:AVER ON\n")
        else:
            self.tx("SENS:AVER OFF\n")


    def trace(self, traceMem=1):
        """This command reads trace data out of the instrument.

           The number of returned results depends on the sweep points
           that have been set. By default 501 results are returned.
           The returned values are scaled in the current level unit.
        """

        msg = "TRAC? TRACE%d\n" % traceMem
        self.tx(msg)

        rsp = self.rx()
        if rsp is None:
            return
        return list(map(float, rsp.strip().split(',')))


    def timeReference(self, source):
        """Set time reference source.

        source :: 'INT' for internal oscillator
                  'EXT' for external time reference
        """

        # Switch to chosen source
        self.tx('ROSC:SOUR ' + source + '\n')

######  ACTIONS #####

class RohdeSchwarzZVLMenu:
    """ User interface menu for network analyzer """

    def __init__(self, rohde, config):
        check_class(rohde, RohdeSchwarzZVL)
        self.rohde = rohde

        from gssw.lib.sockets import PubSocket # Explicit local import
        self.pubSocket = PubSocket()
        proxyConfig  = config.componentParameters('proxy')
        proxyHost    = proxyConfig['hostname']
        proxyPubport = proxyConfig['sockets']['xsub']
        self.pubSocket = PubSocket()
        self.pubSocket.createZmqSocket()
        self.pubSocket.connectSocket("tcp://{}:{}".format(proxyHost, proxyPubport))

    def askDisplay(self):
        """Ask user to enable of disable display."""

        self.rohde.enableDisplay(inputYesNo('Enable display'))

    def askMode(self):
        """Ask user to choose mode type."""
        modes = [('NWA','NWA'), ('SAN', 'SAN')]

        self.rohde.mode(inputChoice('Choose a mode', modes, 'NWA'))

    def askCenterFrequency(self):
        """Ask user to choose a center frequency."""
        center = self.rohde.getCenterFrequency()

        self.rohde.centerFrequency(inputNumber('Choose a center frequency', float, unit='MHz',
                                               defaultValue=center))

    def askFrequencySpan(self):
        """Ask user to choose a frequency span."""
        span = self.rohde.getFrequencySpan()

        self.rohde.frequencySpan(inputNumber('Choose a span frequency', float, unit='MHz',
                                             defaultValue=span))

    def askContInit(self):
        """Ask to enable/disable continuous measurements."""

        self.rohde.contInit(inputYesNo('Enable continuous measurements'))

    def askRefLevel(self):
        """Ask user to choose a reference level."""

        self.rohde.refLevel(inputNumber('Choose a reference level', float, unit='dBm'))

    def askResBandwidth(self):
        """Ask user to choose a res bandwidth."""

        self.rohde.resBandwidth(inputNumber('Choose a res bandwidth', float, unit='MHz'))

    def askVidBandwidth(self):
        """Ask user to choose a vid bandwidth"""

        self.rohde.vidBandwidth(inputNumber('Choose a vid bandwidth', float, unit='MHz'))

    def askSweepPoints(self):
        """Ask user to choose the number of sweep points."""
        sweepPoints = self.rohde.sweepPoints()

        self.rohde.sweepPoints(inputNumber('Choose the number of sweep points', int,
                                           defaultValue=sweepPoints))

    def askInputAttenuation(self):
        """Ask user to configure input attenuation."""
        attenuation = self.rohde.inputAttenuation()

        self.rohde.inputAttenuation(inputNumber('Choose the input attenuation', int,
                                           defaultValue=attenuation, unit='dB',
                                           minValue=0, maxValue=30))

    def askAverages(self):
        """Ask user to choose the number of averages."""

        self.rohde.averages(inputNumber('Choose the number of averages (0 = off)', int,))

    def askTrace(self):
        """Prints trace data on the screen."""

        print(self.rohde.trace(inputNumber('Give number of trace', int,)))

    def askTimeReference(self):
        """Ask user for the time reference."""
        modes = [('INT','internal oscilator'), ('EXT', 'external time source')]

        self.rohde.timeReference(inputChoice('Choose a mode', modes, 'NWA'))

    def askEventStatusEnable(self):
        """Ask user for the even status."""

        print(self.rohde.eventStatusEnable(
            inputNumber('Choose the event status', int, minValue=0, maxValue=255)))


    def askServiceRequestEnable(self):
        """Ask user for the even status."""

        print(self.rohde.serviceRequestEnable(
            inputNumber('Choose the service request', int, minValue=0, maxValue=255)))

    def openGpibMenu(self):
        """Opens the GPIB menu."""

        while True:
            # Prepare menu options.
            items = [
                MenuOption( 1,  'Clear status',                action=self.rohde.clearStatus),
                MenuOption( 2,  'Set the event status enable', action=self.askEventStatusEnable),
                MenuOption( 3,  'Send event status query',     action=self.rohde.eventStatus),
                MenuOption( 4,  'Read status byte',            action=self.rohde.readStatusByte),
                MenuOption( 5,  'Operation complete',          action=self.rohde.operationComplete),
                MenuOption( 6,  'Operation complete query',
                            action=self.rohde.operationCompleteQuery),
                MenuOption( 7,  'Identification',              action=self.rohde.identification),
                MenuOption( 8,  'Reset',                       action=self.rohde.reset),
                MenuOption( 9,  'Service request enable',      action=self.askServiceRequestEnable),
                MenuOption( 10, 'Trigger',                     action=self.rohde.trigger),
                MenuOption( 11, 'Test',                        action=self.rohde.test),
                MenuOption( 12, 'Wait',                        action=self.rohde.wait),
                MenuText(''),
                MenuOption(-2,  'To previous menu',            action='return'),
            ]

            # Show menu and ask user input
            action = showMenu('Main', items)

            if action == 'return':
                break

            # Handle other menu actions
            if action is not None:
                action()

    def askCustomCommands(self):
        """ Lets the user send custom commands, and prints the response """
        while True:
            command = input('Enter a custom command string (enter to return): ')
            if not command:
                break
            print(self.rohde.ask(command))

    ######  MAIN MENU #####


    def measure(self):
        """ Measure and publish continuously """

        # Local import to reduce loading module time
        from gssw.lib.arrays.gridded import GriddedArray
        from gssw.lib.arrays.semantic import SemanticArray
        from gssw.lib.sockets import tagName
        import time
        try:
            logger.info('Starting measurement loop, stop with Ctrl-C')
            self.rohde.contInit(0)
            while True:
                #self.rohde.contInit(0)
                center = self.rohde.getCenterFrequency()
                span   = self.rohde.getFrequencySpan()
                logger.info("center: %s, span: %s", center, span)
                points = self.rohde.sweepPoints()

                # set nr of sweep points
                self.rohde.sweepPoints(points)
                self.rohde.averages(0)

                # start measure
                self.rohde.measure()
                time.sleep(.7)
                trace = self.rohde.trace()         # get data

                attenuation = self.rohde.inputAttenuation()


                # Create gridded array
                gArray = GriddedArray(trace, 'zvl_{}'.format(attenuation))
                import numpy as np
                xdata = np.linspace(center-span/2, center+span/2, points, endpoint=True)
                xaxis = {0: SemanticArray(xdata)}
                gArray.coords = xaxis

                # Create end send protobuf
                msg = gArray.asProtoBuf()
                self.pubSocket.publishProtoBuf("/atp/", tagName(msg), msg.SerializeToString())
        except KeyboardInterrupt:
            logger.info('Stopped measurement loop')
        finally:
            self.rohde.contInit(1)

    def menu(self):
        """ User interface for the ZVL network analyser """

        while True:
            # Prepare menu options.
            items = [
                MenuOption( 1,  'Set display',                  action=self.askDisplay),
                MenuOption( 2,  'Set mode',                     action=self.askMode),
                MenuOption( 3,  'Set center frequency',         action=self.askCenterFrequency),
                MenuOption( 4,  'Set span frequency',           action=self.askFrequencySpan),
                MenuOption( 5,  'Enable/Disable continuous measurements',
                            action=self.askContInit),
                MenuOption( 6,  'Set reference level',          action=self.askRefLevel),
                MenuOption( 7,  'Set res bandwith',             action=self.askResBandwidth),
                MenuOption( 8,  'Set vid bandwith',             action=self.askVidBandwidth),
                MenuOption( 9,  'Set number of sweep points',   action=self.askSweepPoints),
                MenuOption( 10, 'Set number of averages',       action=self.askAverages),
                MenuOption( 11, 'Print trace data',             action=self.askTrace),
                MenuOption( 12, 'Set the time source',          action=self.askTimeReference),
                MenuOption( 13, 'Set the input attenuation',    action=self.askInputAttenuation),
                MenuOption( 20, 'Open GPIB menu',               action=self.openGpibMenu),
                MenuOption( 21, 'Send custom commands',         action=self.askCustomCommands),
                MenuText(''),
                MenuOption( 100,  'Initiate Sweep measurement', action=self.rohde.measure),
                MenuOption('M', 'Measure and publish', action=self.measure),
                MenuText(''),
                MenuOption(-2,  'End script',                   action='exit'),
            ]

            # Show menu and ask user input
            action = showMenu('Main', items)

            if action == 'exit':
                break

            # Handle other menu actions
            if action is not None:
                action()

def main():
    """ User interface program for controlling the network analyzer """
    # Do local imports which are only relevant to the menu
    from gssw.common    import initGssw
    from gssw.lib.misc  import createGsswArgumentParser
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(name)s: %(levelname)s: %(message)s')

    parser = createGsswArgumentParser(__doc__)
    parser.add_argument('host', nargs='?',
                        type=str,
                        help='hostname of the ZVL (default: {})'.format(DEFAULT_PEER_ADDRESS),
                        default=DEFAULT_PEER_ADDRESS)
    parser.add_argument('port', nargs='?',
                        type=int,
                        help='SCPI port number on the ZVL (default: {})'.format(DEFAULT_PEER_PORT),
                        default=DEFAULT_PEER_PORT)

    args = parser.parse_args()

    config = initGssw(args)
    logger.info('Start test')

    rohde = RohdeSchwarzZVL(host=args.host, port=args.port)
    menu = RohdeSchwarzZVLMenu(rohde, config)
    menu.menu()

if __name__ == '__main__':
    main()
