""" General driver for serial devices """

import logging
import serial
from serial.tools.list_ports    import comports

from gssw.lib.errors            import GsswError

logger = logging.getLogger(__name__)

class SerialDevice:
    """ A general class for instruments to communicate through a serial interface """

    def __init__(self  #pylint: disable=too-many-arguments
               , port=None
               , vendorId=None
               , productId=None
               , serialNumber=None
               , baudrate=115200
               , bytesize=serial.EIGHTBITS
               , parity=serial.PARITY_NONE
               , stopbits=serial.STOPBITS_ONE
               , inputsep='\r'
               , outputsep='\r'
               , readbufsz=1024
               , timeout=0.2
               , **kwargs):
        """ Constructor

            :param port:      Serial port to use, e.g. /dev/ttyUSB0
            :param inputsep:  Character(s) to end an input command with
            :param outputsep: Character(s) separating output lines
            :param readbufsz: Maximum bytes to read on the output
            :param timeout:   Maximum time to block when reading serial port
            :param **kwargs:  Paramers to pass to the Serial class
        """

        assert port is not None or (vendorId is not None and \
                                    productId is not None and \
                                    serialNumber is not None), \
                                    "Specify either a port or vendor, productid, and serial number"

        if vendorId is not None: # defined port is fallback
            # Try to determine usb port from given vendorid, productid and serial number
            for serialPort in comports():
                if serialPort.vid==vendorId \
                and serialPort.pid==productId \
                and str(serialPort.serial_number)==str(serialNumber):
                    # Device match found, use this one
                    port = serialPort.device
                    break
            else:
                logger.warning("The serial port of the device was not found "
                               "(vendor: %s, product: %s, serial number: %s)",
                               vendorId, productId, serialNumber)

        if port is None: # i.e. the port is still None
            # This means that is was not found in the list of serial ports
            raise GsswError("The serial port of the device was not found "
                            "(vendor: {}, product: {}, serial number: {}".format(
                                vendorId, productId, serialNumber))

        self._serial = serial.Serial(**kwargs)
        self._serial.port = port
        self._serial.baudrate= baudrate
        self._serial.timeout= timeout
        self._serial.bytesize = bytesize
        self._serial.parity = parity 
        self._serial.stopbits = stopbits

        self.inputsep  = inputsep
        self.outputsep = outputsep
        self.readbufsz = readbufsz

    def close(self):
        """ Close underying serial port """
        self._serial.close()

    def open(self):
        """ Open underying serial port """
        self._serial.open()

    def cmd(self, command):
        """ Send a command and get the response

            :param command: The command as a string
            :return:        Returns lines as a list, separated by the set output separator
        """
        self._serial.write((command + self.inputsep).encode())

        resp = self._serial.read(self.readbufsz).decode()

        return resp.split(self.outputsep)
