#!/usr/bin/env python3
#pylint: disable=invalid-name

"""
Python module and command-line utility to control the
SignalCore SC5503B programmable RF syntesizer via USB.
"""

import array
import collections
import logging
import struct
import time
import usb

from gssw.lib.errors import GsswError


__all__ = [ 'AttachedDevice', 'DeviceStatus',
            'searchUsbDevices', 'SignalCore5503B' ]

###############################################################################
# Constants
###############################################################################

SC5503B_VENDOR_ID       = 0x277c
SC5503B_PRODUCT_ID_1    = 0x0015
SC5503B_PRODUCT_ID_2    = 0x0023

###############################################################################
# Logging
###############################################################################

logger = logging.getLogger(__name__)

###############################################################################
# Named tuples
###############################################################################

AttachedDevice = collections.namedtuple('AttachedDevice',
                                        'model location serialnr')

DeviceStatus = collections.namedtuple('DeviceStatus',
    'status rfEnable standbyEnable fastTuneEnable alcOpen ' +
    'extRefLockEnable refClkOutEnable extRefDetected ' +
    'coarsePllLock finePllLock sumPllLock vcxoPllLock tcxoPllLock')

###############################################################################
# Internal helper functions
###############################################################################

def _searchUsb():
    """ Search for attached SC5503B USB devices.

        Return a list of (model, location, serialnr, dev).
    """

    found_devices = [ ]

    usb_busses = usb.busses()

    for usb_bus in usb_busses:
        for usb_dev in usb_bus.devices:
            if ( usb_dev.idVendor == SC5503B_VENDOR_ID and
                 usb_dev.idProduct in (SC5503B_PRODUCT_ID_1,
                                       SC5503B_PRODUCT_ID_2) ):
                model = 'SC5503B'
                loc   = ( 'usb:' + str(usb_bus.location) +
                          ':' + str(usb_dev.devnum) )
                logger.debug('found model=%r, location=%r', model, loc)
                h = usb_dev.open()
                try:
                    snr = h.getString(usb_dev.iSerialNumber, 63)
                except usb.USBError:
                    snr = None
                logger.debug('... snr=%r', snr)
                found_devices.append((model, loc, snr, usb_dev))

    return found_devices


###############################################################################
# Function searchUsbDevices
###############################################################################

def searchUsbDevices():
    """Search for attached SC5503B devices.

    Return a list of tuples (dev_model, dev_location, dev_serialnr) where
        model    is a string identifying the device model,
                 currently always 'SC5503B';
        location is a string representing the USB port where the device
                 is attached;
        serialnr is a string containing the SC5503B serial number.
    """

    found_devices = _searchUsb()

    return [ AttachedDevice(model, location, serialnr)
             for (model, location, serialnr, _) in found_devices ]


###############################################################################
# class SignalCoreSC5503B
###############################################################################

class SignalCore5503B:
    """Provide access to the SC5503B internal registers."""

    # USB endpoint numbers
    ENDPOINT_OUT = 0x04
    ENDPOINT_IN  = 0x83

    # Timeout for USB read/write in milliseconds.
    TIMEOUT_MS   = 1000

    # Register addresses.
    INITIALIZE              = 0x01
    SET_SYSTEM_ACTIVE       = 0x02
    DEVICE_STANDBY          = 0x05
    RF_FREQUENCY            = 0x10
    RF_POWER                = 0x11
    SYNTH_MODE              = 0x12
    RF_ALC_MODE             = 0x13
    SET_ALC_DAC_VALUE       = 0x14
    REFERENCE_MODE          = 0x15
    REFERENCE_DAC_SETTING   = 0x16
    USER_EEPROM_WRITE       = 0x23
    RF_OUT_ENABLE           = 0x26
    STORE_STARTUP_STATE     = 0x28
    AUTO_POWER_DISABLE      = 0x29
    GET_DEVICE_STATUS       = 0x17
    GET_TEMPERATURE         = 0x18
    CAL_EEPROM_READ         = 0x20
    USER_EEPROM_READ        = 0x22
    GET_ALC_DAC_VALUE       = 0x2A

    # Fine tuning constants.
    FINETUNE_1MHZ           = 0
    FINETUNE_25KHZ          = 1
    FINETUNE_1HZ            = 2

    def __init__(self, location=None, serialnr=None):
        """Open the USB connection to the SC5503B device.

        Specify either "location" or "serialnr" of the device.

        location    -- string of the form "usb:<busnum>:<devnum>"
        serialnr    -- string with serial number of the SC5503B device
        """

        # Find specified device on USB bus.
        found_devices = _searchUsb()

        if location is not None:
            if not location.startswith('usb:'):
                raise ValueError('Invalid format of location string %r' % location)

            found_devices = [ (model, loc, snr, dev)
                              for (model, loc, snr, dev) in found_devices
                              if loc == location ]

        if serialnr is not None:
            have_devices = bool(found_devices)

            found_devices = [ (model, loc, snr, dev)
                              for (model, loc, snr, dev) in found_devices
                              if snr == serialnr ]

            if have_devices and location is not None and not found_devices:
                raise GsswError('Wrong serial number of device at %r: expecting %r' %
                                (location, serialnr))

        s1 = ' at %r' % (location,) if location is not None else ''
        s2 = ' with serialnr %r' % (serialnr,) if serialnr is not None else ''
        if not found_devices:
            raise GsswError('No SC5503B device found' + s1 + s2)
        if len(found_devices) > 1:
            raise GsswError('Multiple SC5503B devices found' + s1 + s2)

        (self.model,
         self.location,
         self.serialnr,
         self.usb_dev) = found_devices[0]

        logger.debug('found device location=%r, serialnr=%r',
                  self.location,
                  self.serialnr)

        # Open USB device and claim interface.
        ok = False
        self.handle = self.usb_dev.open()
        self.handle.claimInterface(0)
        try:
            self.handle.setAltInterface(1)
            ok = True
        finally:
            if not ok:
                self.handle.releaseInterface()

        logger.debug('claimed interface')

        self._freq = 5000000000.0 # Frequency [Hz] (Power-up default)
        self._refout100 = False   # Reference output clock set to 10MHz (Power-up default)

    def close(self):
        """Close USB connection to device."""

        logger.debug('releasing interface and closing device (serialnr=%r)',
                  self.serialnr)

        self.handle.releaseInterface()

    def writeReg(self, reg, wval, writeBytes):
        """Write to internal register in device."""

        logger.debug('writeReg(reg=%d, wval=%d, writeBytes=%d)',
                  reg, wval, writeBytes)

        assert reg >= 0 and reg <= 255
        assert writeBytes >= 0 and writeBytes <= 5
        assert wval >= 0 and wval <= (1<<(8*writeBytes))-1

        cmd = array.array('B', (writeBytes+1) * [ 0 ])
        cmd[0] = reg
        for i in range(writeBytes):
            cmd[i+1] = (wval >> (8*(writeBytes-i-1))) & 0xff
        cmdstr = cmd.tostring()

        self.handle.bulkWrite(self.ENDPOINT_OUT, cmdstr, self.TIMEOUT_MS)
        repstr = self.handle.bulkRead(self.ENDPOINT_IN, 6, self.TIMEOUT_MS)

        logger.debug('got %d bytes', len(repstr))

    def readReg(self, reg, wval, writeBytes, readBytes):
        """Read from internal register in device."""

        logger.debug('readReg(reg=%d, wval=%d, writeBytes=%d, readBytes=%d)',
                  reg, wval, writeBytes, readBytes)

        assert reg >= 0 and reg <= 255
        assert writeBytes >= 0 and writeBytes <= 5
        assert readBytes >= 0 and readBytes <= 5
        assert wval >= 0 and wval <= (1<<(8*writeBytes))-1

        cmd = array.array('B', (writeBytes+1) * [ 0 ])
        cmd[0] = reg
        for i in range(writeBytes):
            cmd[i+1] = (wval >> (8*(writeBytes-i-1))) & 0xff
        cmdstr = cmd.tostring()

        self.handle.bulkWrite(self.ENDPOINT_OUT, cmdstr, self.TIMEOUT_MS)
        repstr = self.handle.bulkRead(self.ENDPOINT_IN, 6, self.TIMEOUT_MS)

        logger.debug('got %d bytes', len(repstr))

        if len(repstr) < readBytes:
            raise GsswError('Got %d bytes from device while expecting %d' %
                        (len(repstr), readBytes))

        rep = array.array('B', repstr)
        rval = 0
        for i in range(readBytes):
            rval = (rval << 8) | rep[i]

        logger.debug(' ... rval=%d', rval)
        return rval

    def initDevice(self, keepstate):
        """Reset device.

        keepstate   -- True to keep current state,
                       False to reset to power-on state.
        """

        wval = 1 if keepstate else 0
        time.sleep(0.1)
        self.writeReg(self.INITIALIZE, wval, 1)
        time.sleep(0.1)

    def setActiveLed(self, ledon):
        """Set "active" LED on front panel.

        ledon       -- True to turn LED on, False to turn off.
        """

        wval = 1 if ledon else 0
        self.writeReg(self.SET_SYSTEM_ACTIVE, wval, 1)

    def setStandby(self, standby):
        """Set device standby mode.

        standby     -- True to power down RF circuits,
                       False to return to active state.
        """

        wval = 1 if standby else 0
        self.writeReg(self.DEVICE_STANDBY, wval, 1)

    def setSwitchOnOff(self, status):
        """ Set device standby mode. This function is introduced
            to mimic behaviour of the agilent8257d.

        status      -- False to power down RF circuits
                       True to return to active state.
        """
        self.setStandby(not bool(status))

    def setFreq(self, freq):
        """Set RF frequency in Hz."""

        if freq < 0 or freq > 10.0e9:
            raise ValueError('Invalid frequency %d Hz' % freq)

        if not isinstance(freq, int):
            freq = int(round(freq))

        self.writeReg(self.RF_FREQUENCY, freq, 5)

        self._freq = float(freq) # remember last written frequency

    def getFreq(self):
        """ Not supported by the signal core. Instead return the frequency [Hz] last configured, or
        the power-up default. """
        return self._freq

    def setPowerLevel(self, power):
        """Set RF output power in dB."""

        rval = int(round(100.0 * power))
        if rval > 32767:
            raise ValueError('Invalid power level %g dB' % power)
        rval = max(-32767, rval)

        wval = rval if rval >= 0 else (0x8000 - rval)
        self.writeReg(self.RF_POWER, wval, 2)

    def setSynthesizerMode(self, fasttune, finetune):
        """Set synthesizer mode.

        fasttune    -- True to enable fast tuning,
                       False to disable fast tuning.
        finetune    -- fine tuning mode, 0 = 1 MHz steps (PLL),
                                         1 = 25 kHz steps (PLL),
                                         2 = 1 Hz steps (DDS).

        The default mode is 1 Hz fine tuning steps, but larger tuning
        steps result in lower phase spurs.
        The default mode is fast tuning disabled (better noise suppression).
        """

        assert finetune >= 0 and finetune <= 2

        wval = finetune & 3
        if fasttune:
            wval |= 0x4

        self.writeReg(self.SYNTH_MODE, wval, 1)

    def setAlcMode(self, openloop):
        """Set RF Automatic Level Control mode.

        openloop    -- True to disable ALC closed loop operation;
                       False to enable ALC closed loop operation (default).
        """

        wval = 1 if openloop else 0
        self.writeReg(self.RF_ALC_MODE, wval, 1)

    def setReferenceClock(self, extref, refout, refout100):
        """Set Reference Clock mode.

        extref      -- True to enable locking to external reference input;
                       False to use internal reference source.
        refout      -- True to enable the output reference signal.
        refout100   -- True to select 100 MHz output reference signal;
                       False to select 10 MHz output reference signal.
        """

        wval = 0
        if extref:
            wval |= 1
        if refout:
            wval |= 2
        if refout100:
            wval |= 4
        self.writeReg(self.REFERENCE_MODE, wval, 1)
        self._refout100 = refout100

    def getReferenceOutput100(self):
        """ Not supported by the signal core. Instead return status of the 100/10MHz last
        configured, or the power-up default. """
        return self._refout100

    def setRfOut(self, enable):
        """Set RF output enable.

        enable      -- True to enable RF output; False to disable.
        """

        wval = 1 if enable else 0
        self.writeReg(self.RF_OUT_ENABLE, wval, 1)

    def storeCurrentState(self):
        """Store current state as default power-up state."""

        self.writeReg(self.STORE_STARTUP_STATE, 1, 1)

    def getDeviceInfo(self):
        """Read device information."""

        eeprom_data = array.array('B', 64 * [ 0 ])
        for i in range(64):
            rval = self.readReg(self.CAL_EEPROM_READ, i, 2, 1)
            eeprom_data[i] = rval

        (productsnr,)   = struct.unpack('<I', eeprom_data[4:8])
        (rfmodulesnr,)  = struct.unpack('<I', eeprom_data[8:12])
        (mandate,)      = struct.unpack('<I', eeprom_data[12:16])
        (caldate,)      = struct.unpack('<I', eeprom_data[16:20])
        (firmwarerev,)  = struct.unpack('<f', eeprom_data[44:48])
        (synthrev,)     = struct.unpack('<f', eeprom_data[48:52])
        (signalrev,)    = struct.unpack('<f', eeprom_data[52:56])

        return { 'productSerialNumber':     productsnr,
                 'rfModuleSerialNumber':    rfmodulesnr,
                 'manDate':                 mandate,
                 'calDate':                 caldate,
                 'firmwareRevision':        firmwarerev,
                 'synthHardwareRevision':   synthrev,
                 'signalHardwareRevision':  signalrev }

    def getDeviceStatus(self):
        """Read device status."""

        rval = self.readReg(self.GET_DEVICE_STATUS, 0, 1, 2)
        return DeviceStatus(status          = rval,
                            rfEnable        = ((rval >> 1) & 1),
                            standbyEnable   = ((rval >> 2) & 1),
                            fastTuneEnable  = ((rval >> 3) & 1),
                            alcOpen         = ((rval >> 4) & 1),
                            extRefLockEnable = ((rval >> 5) & 1),
                            refClkOutEnable = ((rval >> 6) & 1),
                            extRefDetected  = ((rval > 7) & 1),
                            coarsePllLock   = ((rval >> 9) & 1),
                            finePllLock     = ((rval >> 10) & 1),
                            sumPllLock      = ((rval >> 13) & 1),
                            vcxoPllLock     = ((rval >> 14) & 1),
                            tcxoPllLock     = ((rval >> 15) & 1))

    def getTemperature(self):
        """Read device temperature (Celcius)."""

        rval = self.readReg(self.GET_TEMPERATURE, 0, 1, 2)
        temp = ((rval & 0x1fff) - (rval & 0x2000)) / 32.0
        return temp

