#!/usr/bin/env python3
""" This module provides the class for interfacing with the SRS 785.

:see: Srs785 class description for details.
"""
#pylint: disable=invalid-name

import logging
import codecs
import argparse
from time import sleep
from datetime import datetime
from collections import OrderedDict
import numpy as np
from gssw.ucelib.gpib import GpibTcpIp


class Srs785(GpibTcpIp):
    """ Stanford Research Systems 785 Spectrum Analyzer interface module.
        This module uses a proglogix "yellow brick" to convert ethernet to GPIB.
        The user must configure the device by hand and can then use this module to perform 
        measurements and extract the resulting data. Data can be dumped to a csv file with a
        configuration header.
    """

    CONFIG = {'measurement':            'MEAS?',
              'FFT span':               'FSPN?',
              'FFT resolution':         'FLIN?',
              'window':                 'FWIN?',
              'average method':         'FAVM?',  
              'average number':         'NAVG?', # number of completed averages not the setting
              'time record increment':  'FOVL?',
              'unit':                   'UNIT?'} 

    SOURCE_CONFIG = {'source enabled':  'SRCO?',
                     'source type':     'STYP?'}

    SINE_CONFIG = {'sine 1 frequency':  'S1FR?',
                   'sine 1 amplitude':  'S1AM?',
                   'sine 2 frequency':  'S2FR?',
                   'sine 2 amplitude':  'S2AM?',
                   'sine offset':       'SOFF?'}

    CHIRP_CONFIG = {'chirp amplitude':           'CAMP?',
                    'chirp burst percentage':    'CBUR?',
                    'source display':            'CSRC?'}

    SWEPT_SINE_CONFIG = {'swept-sine start frequency': 'SSTR? {}',
                         'swept-sine stop frequency': 'SSTP? {}',
                         'swept-sine logarithmic': 'SSTY? {}',
                         'swept-sine slower treshold': 'SSLO? {}',
                         'swept-sine faster treshold': 'SFST? {}',
                         'swept-sine offset': 'SOFF?',
                         'swept-sine lower limit': 'SSLL?',
                         'swept-sine upper limit': 'SSUL?',
                         'swept-sine max level': 'SMAX?',
                         'swept-sine amplitude': 'SSAM?',
                         'swept-sine ramp rate': 'SRAT?'}


    # This would be nice to support but will not be used in the forseeable future
    # NOISE_CONFIG = {'noise amplitude':           'NAMP?',
    #                 'noise type':                'NTYP?',
    #                 'noise burst percentage':    'NBUR?',
    #                 'noise burst source period': 'NPER?',
    #                 'source display':            'CSRC?'}

    # ARBITRARY_CONFIG = {}

    INPUT_CH1 = {'input grounding':     'I1GD?',
                 'input coupling':      'I1CP?',
                 'input range':         'I1RG?'} 

    INPUT_CH2 = {'input grounding':     'I2GD?',
                 'input coupling':      'I2CP?',
                 'input range':         'I2RG?'} 

    # note: there are more, but these are not added (yet)
    TRANSLATE = {'measurement':         {0:'FFT 1', 1:'FFT 2', 2:'Power Spectrum 1',   
                                         3:'Power Spectrum 2', 4:'Time 1', 5:'Time 2', 
                                         6:'WinTime 1', 7:'WinTime 2', 8:'Orbit',
                                         9:'Coherence', 10:'Cross Spec.', 11:'Freq. Resp.',
                                         42:'Line Spec 1', 43:'Line Spec 2'},
                 'FFT resolution':      ['100', '200', '400', '800'],
                 'window':              ['Uniform', 'Flattop', 'Hanning', 'BMH', 
                                         'Kaiser', 'Force/Exponential', 'User', 
                                         '[-T/2..T/2]', '[0..T/2]', '[-T/4..T/4]'],
                 'average method':      ['None', 'Vector', 'RMS', 'Peak Hold'],
                 'input grounding':     ['Float', 'Ground'],
                 'input coupling':      ['DC', 'AC', 'ICP'],
                 'source type':         ['sine', 'chirp', 'noise', 'arbitrary']}

    INPUT_RANGE_UNIT = ['dBVpk', 'dBVpp', 'dBVrms', 'Vpk', 'Vpp', 'Vrms', 'dBEUpk', 
                        'dBEUpp', 'dBEUrms', 'EUpk', 'EUpp', 'EUrms']

    SOURCE_AMPLITUDE_UNIT = ['mVpk', 'mVpp', 'mVrms', '[Vpk]', 'Vrms', 'dBVpk', 'dBVpp', 
                             'dBVrms']
        

    DISPLAY = ['A', 'B']


    def __init__(self, host, port=1234, gpibAddress=10, logLevel=logging.INFO):

        super(Srs785, self).__init__(host, port, timeout=10)

        # initialize logging
        self.log = logging.getLogger('SRS785')
        self.log.setLevel(logLevel)

        # configure the ethernet-gpib converter
        self.tx('++mode 1\n') # controller mode
        self.tx('++addr %d\n' % gpibAddress) # set gpib address
        self.tx('++auto 1\n') # auto addressing

        # direct the devices output to the GPIB interface
        self.tx('OUTX0')
        self.rx() # clear buffer


    def getId(self):
        """ get the device ID string """

        return self.ask('*IDN?').strip()


    def reset(self):
        """ Reset the device and poll the gpib channel until it is online. """

        self.log.info('resetting the device...')
        self.tx('*RST')

        # test the communication 
        self.log.setLevel(logging.CRITICAL) # avoid time out errors
        while True:
            rsp = self.ask('*STB?')
            if rsp:
                self.tx('++clr')
                self.log.setLevel(logging.INFO)
                self.log.info('device online')
                self.log.info('IDN: ' + self.ask('*IDN?'))
                break
            else:
                sleep(1)


    def start(self):
        """ Start a new measurement """

        self.tx('STRT')


    def waitSineSweep(self, display):
        """ Wait and perform a sine sweep """
        sleep(10)
        nops = int(self.ask('SNPS? {}'.format(display)))
        while int(self.ask('SSFR?')) + 1 < nops:
            sleep(10)


    def readDisplay(self, display):
        """ Read the data from display A or B. 

            This function is generic and can be used for all measurement types
        """

        # pause to keep the spectrum stable
        self.tx('PAUS')

        # transfer data 
        data = self.ask('DSPY? %d' % display)

        self.log.info('reading display %s completed', self.DISPLAY[display])

        return data


    def readTimeSeries(self, display):
        """ Read a timeseries from the display and returns it together with the configuration. """

        assert display in [0, 1]

        # switch the active display
        self.tx('ACTD %d' % display)

        self.log.info('reading display %s', self.DISPLAY[display])

        # Poll NEWA (new data available on the display)
        if not int(self.ask('DSPS?')) & 1: 
            self.log.info('waiting for time series...')
            while not int(self.ask('DSPS?')) & 1: 
                sleep(0.5)

        # get the data
        data = [float(x) for x in self.readDisplay(display).split(',')]
        config = self.readConfiguration(display)

        # return the time series and config dict
        return data, config


    def dumpTimeSeries(self, filename, timestamp, display=0):
        """ Read a timeseries from the display and dump it to a file """

        timeseries, config = self.readTimeSeries(display)

        with codecs.open('%s_%s.log' % (filename, timestamp), 'w', 'utf-8') as f:

            f.write(u'# Time series taken at: %s\n#\n' % timestamp)

            f.write(u'# SR785 configuration:\n')
            for param in config:
                f.write(u'# %-30s: %s\n' % (param, config[param].splitlines()[0]))

            f.write(u'#\n# index, amplitude [%s]\n' % config['unit'].splitlines()[0])

            for i, x in enumerate(timeseries):
                f.write(u'%d, %e\n' % (i, x))

        return timeseries, config


    def readSpectrum(self, display):
        """ Read the spectrum from display A or B 

            the data is returned as an array of tuples of x,y data
        """

        assert display in [0, 1]

        # switch the active display
        self.tx('ACTD %d' % display)

        self.log.info('reading display %s', self.DISPLAY[display])

        # Poll NEWA (new data available on the display)
        if not int(self.ask('DSPS?')) & 1: 
            self.log.info('waiting for spectrum...')
            while not int(self.ask('DSPS?')) & 1: 
                sleep(0.5)

        # Wait for averaging to complete
        if  int(self.ask('FAVM? %d' % display)) != 0: # if averaging shown on display
            if int(self.ask('NAVG? %d' % display)) < int(self.ask('FAVN? %d' % display)):
                self.log.info('waiting for averaging to complete...')
            while int(self.ask('NAVG? %d' % display)) < int(self.ask('FAVN? %d' % display)):
                sleep(0.5)

        # get the data
        data = self.readDisplay(display)
        config = self.readConfiguration(display)

        yData = [float(x) for x in data.split(',')]
        xData = np.linspace(0, float(config['FFT span']), len(yData))

        # convert the spectrum to an array and return
        return zip(xData, yData), config


    def readSweptSine(self, display):
        """ Read the swept-sine from display A or B

            the data is returned as an array of tuples of x,y data
        """

        assert display in [0, 1]

        # switch the active display
        self.tx('ACTD %d' % display)

        self.log.info('reading display %s', self.DISPLAY[display])

        # get the data
        data = self.readDisplay(display)
        config = self.readConfiguration(display)

        yData = [float(x) for x in data.split(',')]
        if '1' in config['swept-sine logarithmic']:
            xData = np.logspace(np.log10(float(config['swept-sine start frequency'])),
                                np.log10(float(config['swept-sine stop frequency'])),
                                len(yData))
        else:
            xData = np.linspace(float(config['swept-sine start frequency']),
                                float(config['swept-sine stop frequency']),
                                len(yData))

        # convert the spectrum to an array and return
        return zip(xData, yData), config


    def dumpSweptSine(self, filename, timestamp, display=0):
        """ Read a swept-sinze and dump to a file ('filename_datetime.log'). """ 

        sweptsine, config = self.readSweptSine(display)

        with codecs.open('%s_%s.log' % (filename, timestamp), 'w', 'utf-8') as f:

            f.write(u'# Swept-sine taken at: %s\n#\n' % timestamp)

            f.write(u'# SRS785 configuration:\n')
            for param in config:
                f.write(u'# %-30s: %s\n' % (param, config[param].splitlines()[0]))

            f.write(u'#\n# index, bin [Hz], amplitude [%s]\n' % config['unit'].splitlines()[0])

            for i, (x, y) in enumerate(sweptsine):
                f.write(u'%d, %e, %e\n' % (i, x, y))

        return sweptsine, config


    def dumpSpectrum(self, filename, timestamp, display=0):
        """ Read a spectrum and dump to a file ('filename_datetime.log'). """ 

        spectrum, config = self.readSpectrum(display)

        with codecs.open('%s_%s.log' % (filename, timestamp), 'w', 'utf-8') as f:

            f.write(u'# Spectrum taken at: %s\n#\n' % timestamp)

            f.write(u'# SRS785 configuration:\n')
            for param in config:
                f.write(u'# %-30s: %s\n' % (param, config[param].splitlines()[0]))

            f.write(u'#\n# index, bin [Hz], amplitude [%s]\n' % config['unit'].splitlines()[0])

            for i, (x, y) in enumerate(spectrum):
                f.write(u'%d, %e, %e\n' % (i, x, y))

        return spectrum, config


    def dump(self, filename=None, timestamp=None, display=0):
        """ dump a spectrum or a timeseries depending on the device settings. 
            other modes are not supported. 
        """

        if not timestamp:
            timestamp = self.dateTimeString() 

        mode = int(self.ask('MEAS? %d' % display).strip())

        if mode in [0, 1]:
            if not filename:
                filename = '../../output/spectrum'
            self.dumpSpectrum(filename, timestamp)
        elif mode in [4, 5]:
            if not filename:
                filename = '../../output/timeseries'
            self.dumpTimeSeries(filename, timestamp)
        else:
            self.log.warning('Unsupported mode (%d). Supported are FFT and timeseries.', mode)
            #if not filename:
            #    filename = '../../output/unsupported'
            #self.dumpSpectrum(filename, timestamp)


    def setSpan(self, base, exp, display=2):
        """ Set the span of the spectrum. 

            base   - FFT base frequency selection (0 - 100 kHz, 1 - 102.4 kHz).
            exp    - exponent used in span calculation [-19, 0]
            dispay - apply span to this display (0, 1 or 2 for both)

            Span: base * 2^exp.

            Note that invoking this function might change the base frequency for both displays.
        """

        assert base in [0, 1]
        assert exp in range(-19, 1)
        assert display in [0, 1, 2]

        base = 100E3 if base == 0 else 102.4E3
        freq = base * 2**exp
        self.log.info('setting display %d span to %f Hz', display, freq)

        self.tx('FBAS 2, %d' % base)
        self.tx('FSPN %d, %f' % (display, freq))


    def readSourceConfiguration(self):
        """ Read the configurarion of the source subsystem. 

            For now only sine in supported
        """

        config = OrderedDict()
        config['source enabled'] = self.ask('SRCO?').strip()

        if config['source enabled'] == '0':
            self.log.debug('No source info in header because the function is disabled.')

        else:
            typeIdx = self.ask('STYP?').strip()
            config['source type'] = self.TRANSLATE['source type'][int(typeIdx)]

            if config['source type'] == 'sine':
                for param in self.SINE_CONFIG:
                    reply = self.ask(self.SINE_CONFIG[param]).strip()

                    # some parameters have a unit which must be translated
                    if len(reply.split(',')) == 2:
                        value, unitIdx = reply.split(',')
                        config[param] = value + ' ' + self.SOURCE_AMPLITUDE_UNIT[int(unitIdx)]
                    else:
                        config[param] = reply
            elif config['source type'] == 'chirp':
                for param in self.CHIRP_CONFIG:
                    reply = self.ask(self.CHIRP_CONFIG[param]).strip()

                    # some parameters have a unit which must be translated
                    if len(reply.split(',')) == 2:
                        value, unitIdx = reply.split(',')
                        config[param] = value + ' ' + self.SOURCE_AMPLITUDE_UNIT[int(unitIdx)]
                    else:
                        config[param] = reply
            else:
                self.log.warning(
                    'Unsupported source mode (%s). Source parameters will not be written to header'
                        , config['source type'])

        return config


    def readSweptSineConfiguration(self, display):
        """ Read the configuration of swept-sine. """

        config = OrderedDict()

        for param in self.SWEPT_SINE_CONFIG:
            if '{}' in self.SWEPT_SINE_CONFIG[param]:
                reply = self.ask(self.SWEPT_SINE_CONFIG[param].format(display)).strip()
            else:
                reply = self.ask(self.SWEPT_SINE_CONFIG[param]).strip()
            
            # some parameters have a unit which must be translated
            if len(reply.split(',')) == 2:
                value, unitIdx = reply.split(',')
                config[param] = value + ' ' + self.SOURCE_AMPLITUDE_UNIT[int(unitIdx)]
            else:
                config[param] = reply

        return config


    def readConfiguration(self, display):
        """ Read the devices configuration into a dict. """

        config = OrderedDict()

        # read configuration parameters from the device
        for param in self.CONFIG:
            config[param] = self.ask(self.CONFIG[param] + ' %d' % display).strip()

        # select the right dict of channel specific commands based on the active channel
        if int(config['measurement']) in [0, 4, 42]:
            channelConfig = self.INPUT_CH1
        elif int(config['measurement']) in [1, 5, 11, 43]:
            channelConfig = self.INPUT_CH2
        else:
            self.log.error('Unsupported measurement selected (%d)', int(config['measurement']))
            return {}

        # read channel specific commands 
        for param in channelConfig:
            config[param] = self.ask(channelConfig[param])

        # translate indices to strings 
        for name in self.TRANSLATE:
            if name in config:
                config[name] = self.TRANSLATE[name][int(config[name])] + '\n'

        # calculate derived parameters
        config['line width'] = str(float(config['FFT span']) / float(config['FFT resolution']))
        config['acquisition time'] = str(1 / float(config['line width']))

        # translate the input range 
        value, unitIdx = config['input range'].split(',')
        unit = self.INPUT_RANGE_UNIT[int(unitIdx)]
        config['input range'] = '%s %s\n' % (value, unit)

        # get source parameters if enabled
        if 'Line Spec' in config['measurement']:
            config.update(self.readSweptSineConfiguration(display))
        
        config.update(self.readSourceConfiguration())

        # convert cp437 to unicode
        ##for name in config:
        ##    config[name] = config[name].decode('cp437')
        return config


    def dateTimeString(self, timestamp=None):
        """ Create a string containing the current date and string. """

        timestamp = datetime.now() if timestamp is None else timestamp
        timestampString = '%04d-%02d-%02d_%02d:%02d:%02d' % \
            (timestamp.year, timestamp.month, timestamp.day, 
             timestamp.hour, timestamp.minute, timestamp.second)

        return timestampString


def main():
    """ Test function. 

        - print the devide ID 
        - start a new measurement
        - dump the data to file

        Only supports FFT and time series measurements.
        configring the measurement should be done manually before starting this script 
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("prefix", type=str)

    args = parser.parse_args()

    # Set up logging
    logging.basicConfig()

    # Create device object
    srs = Srs785('192.168.0.3', logLevel=logging.INFO) 

    # Print device ID 
    print(srs.getId())

    # Start a new measurement and dump the data to file
    srs.start()
    srs.waitSineSweep(0)
    timestamp = srs.dateTimeString().replace(':','_')
    srs.dumpSweptSine(args.prefix + '_dsp0', timestamp, 0)
    srs.dumpSweptSine(args.prefix + '_dsp1', timestamp, 1)



if __name__ == '__main__':
    main()
