#!/usr/bin/env python3
#pylint: disable=invalid-name
""" Module for interfacing with an SRS 830 DSP Lock-In Amplifier. """

import argparse
import logging
import math
import time
import socket
import struct
import gssw.lib.typing  as typing
import gssw.ucelib.gpib as gpib

logger = logging.getLogger(__name__)

__all__ = ['SRS830']

DEVICE_IDENTITY_PREFIX = "Stanford_Research_Systems,SR830"

# Defaults
DEFAULT_HOST = "192.168.1.201"
DEFAULT_PORT = 1234
DEFAULT_GPIB_ADDRESS = 23


class SRS830(gpib.GpibTcpIp):
    """ Provides a communication interface to an SRS 830 DSP Lock-In Amplifier.

    :note: The lock-in amplifier must be connected via a Prologix Gpib to Ethernet converter.
    :note: Some function docstrings are excerpts from the SRS 830 programming manual.
           Capital words correspond to (soft) keys on the instrument's front panel.
    """

    def __init__(self, host, port, gpibAddr):
        """ Constructor, for LAN communication. """

        super().__init__(host, port, "analyzer", commandInterval=0.01, processingDelay=0.01,
                         explicitRead=True,       # Request readings from the Prologix converter
                         timeout=4)               # Receiving measurement data can take a while

        if self.isConnected:
            if gpibAddr:
                # Set up the Prologix converter
                self.tx('++auto 0')  # Please listen
                self.tx('FAST 0')  # Stop stream

                self.tx('++auto 1')                   # Read response after sending commands
                #                                       to the instrument.
                self.tx('++mode 1')                   # Use Prologix converter in Controller mode
                self.tx('++addr {}'.format(gpibAddr)) # Configure the GPIB address
                self.tx('++eos 2')                    # Append LF to instrument commands
                self.tx('++eoi 1')                    # Use EOI signals after all last characters

            self._checkIdentification()


    def finalize(self):
        """ Cleans up the connection to the lock-in amplifier and the Prologix converter. """
        if self.isConnected:
            pass


    def _checkIdentification(self):
        """ Verifies that we are connected to a SRS 830. """
        identification = self.identification().strip()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        assert deviceCorrect,\
            "Device with identity '{}' is not supported.".format(identification)


    def setRefPhase(self, phase):
        """ This method sets the reference phase shift. The phase is given in
            real number of degrees. The phase may be programmed from -360.00 till 729.99
            and will be wrapped around +/- 180.

        :param phase: float, Reference phase shift.
        :see: manual p5-4
        """
        typing.check_class(phase, float)
        assert phase >= -360.0 and phase <= 729.99
        self.tx("PHAS {}".format(phase))


    def getRefPhase(self):
        """ This method gets the reference phase shift.

        :return: float, Reference phase shift.
        :see: manual p5-4
        """

        return float(self.ask("PHAS?"))


    def setRefInternalSource(self, source):
        """ This method sets the reference source. The source can be
            set internal (True) or external (False).

        :param source: bool, Reference source.
        :see: manual p5-4
        """
        typing.check_class(source, bool)
        self.tx("FMOD {}".format(int(source)))


    def getRefSource(self):
        """ This method gets the reference source. The source can be
            internal (True) or external (False).

        :return: bool, Reference source.
        :see: manual p5-4
        """
        return bool(int(self.tx("FMOD?")))


    def setRefFrequency(self, freq):
        """ This method sets the reference frequency of the internal oscillator.
            This command is allowed only if the reference source is internal. The
            parameter freq is a real number in Hz. The value of freq is limited to
            0.001 till 102000.

        :param freq: float, Reference frequency.
        :see: manual p5-4
        """
        typing.check_class(freq, float)
        assert freq >= 0.001 and freq <= 102000
        self.tx("FREQ {}".format(freq))


    def getRefFrequency(self):
        """ This method gets the reference frequency of the internal oscillator.

        :return: float, Reference frequency.
        :see: manual p5-4
        """
        return float(self.ask("FREQ?"))


    def setRefSlope(self, trigger):
        """ This method sets the reference trigger when using the external reference source.
            The parameter trigger selects sine zero crossing (ZERO), or TTL rising edge (RISE)
            or TTL falling edge (FALL). At frequencies below 1 Hz, a TTL referene must be used.

        :param trigger: str, Reference trigger.
        :see: manual p5-4
        """
        assert trigger in ['ZERO', 'RISE', 'FALL']
        self.tx("RSLP {}".format(['ZERO', 'RISE', 'FALL'].index(trigger)))


    def getRefSlope(self):
        """ This method gets the reference trigger when using the external reference source.

        :return: str, Reference trigger.
        :see: manual p5-4
        """
        return ['ZERO', 'RISE', 'FALL'][int(self.ask("RSLP?"))]


    def setHarmonic(self, harmonic):
        """ This method sets the detection harmonic. The parameter harmonic is an integer
            from 1 to 19999. This method set the lock-in to detect at the i^th harmonic of
            the reference frequency.

        :param harmonic: int, Detection harmonic.
        :see: manual p5-4
        """
        typing.check_class(harmonic, int)
        assert harmonic >= 1 and harmonic <= 19999
        self.tx("HARM {}".format(harmonic))


    def getHarmonic(self):
        """ This method gets the detection harmonic.

        :return: int, Detection harmonic.
        :see: manual p5-4
        """
        return int(self.ask("HARM?"))


    def setSineLevel(self, level):
        """ This method sets the amplitude of the sine output. The parameter level
            is a voltage in real number of Volts. The value of level is limited to
            0.004 till 5.00.

        :param level: float, Amplitude sine output.
        :see: manual p5-4
        """
        typing.check_class(level, float)
        assert level >= 0.004 and level <= 5.0
        self.tx("SLVL {}".format(level))


    def getSineLevel(self):
        """ This method gets the amplitude of the sine output.

        :return: float, Amplitude sine output.
        :see: manual p5-4
        """
        return float(self.ask("SLVL?"))


    def setInputSource(self, source):
        """ This method sets the input configuration. The parameter source
            selects A, A-B, I-1M (1 MOhm), I-100M (100 MOhm).

        :param source: str, Input configuration.
        :see: manual p5-5
        """
        assert source in ['A', 'A-B', 'I-1M', 'I-100M']
        self.tx("ISRC {}".format(['A', 'A-B', 'I-1M', 'I-100M'].index(source)))


    def getInputSource(self):
        """ This method gets the input configuration.

        :return: str, Input configuration.
        :see: manual p5-5
        """
        return ['A', 'A-B', 'I-1M', 'I-100M'][int(self.ask("ISRC?"))]


    def setInputGrounded(self, ground):
        """ This method sets the input shield grounding. The parameter ground
            selects Float (False) or Ground (True).

        :param ground: bool, Shield grounding.
        :see: manual p5-5
        """
        typing.check_class(ground, bool)
        self.tx("IGND {}".format(ground))


    def getInputGrounded(self):
        """ This method gets the input shield grounding.

        :return: bool, Shield grounding.
        :see: manual p5-5
        """
        return bool(int(self.ask("IGND?")))


    def setInputCoupling(self, coupling):
        """ This method sets the input coupling. The parameter coupling
            selects AC or DC.

        :param coupling: str, Input coupling.
        :see: manual p5-5
        """
        assert coupling in ['AC', 'DC']
        self.tx("ICPL {}".format(['AC', 'DC'].index(coupling)))


    def getInputCoupling(self):
        """ This method gets the input coupling.

        :return: str, Input coupling.
        :see: manual p5-5
        """
        return 'DC' if int(self.ask("ICPL?")) else 'AC'


    def setInputNotchFilter(self, notch):
        """ This method sets the input line notch filter. The parameter notch
            selects NONE, LINE notch, 2LINE notch or BOTH notch filters.

        :param notch: str, Notch filter.
        :see: manual p5-5
        """
        assert notch in ['NONE', 'LINE', '2LINE', 'BOTH']
        self.tx("ILIN {}".format(['NONE', 'LINE', '2LINE', 'BOTH'].index(notch)))


    def getInputNotchFilter(self):
        """ This method gets the input line notch filter.

        :return: str, Notch filter.
        :see: manual p5-5
        """
        return ['NONE', 'LINE', '2LINE', 'BOTH'][int(self.ask("ILIN?"))]


    def setSensitivity(self, sensitivity):
        """ This method sets the sensitivity. The parameter sensitivity
            selects a sensitivity from 2e-9 till 1.0 in V or muA.

        :param sensitivity: float, Sensitivity.
        :see: manual p5-6
        """
        typing.check_class(sensitivity, float)
        assert sensitivity >= 2e-9 and sensitivity <= 1
        self.tx("SENS {}".format(round(26 + 3 * math.log10(sensitivity))))


    def getSensitivity(self):
        """ This method gets the sensitivity.

        :return: float, Sensitivity.
        :see: manual p5-6
        """
        return 10**((float(self.ask("SENS?")) - 26) / 3)


    def setReserveMode(self, reserve):
        """ This method sets the reserve mode. The parameter reserve
            selects HIGH reserve, NORMAL or LOW noise.

        :param reserve: str, Reserve mode.
        :see: manual p5-6
        """
        assert reserve in ['HIGH', 'NORMAL', 'LOW']
        self.tx("RMOD {}".format(['HIGH', 'NORMAL', 'LOW'].index(reserve)))


    def getReserveMode(self):
        """ This method gets the reserve mode.

        :return: str, Reserve mode.
        :see: manual p5-6
        """
        return ['HIGH', 'NORMAL', 'LOW'][int(self.ask("RMOD?"))]


    def setTimeConstant(self, tconstant):
        """ This method sets the time constant. The parameter tconstant
            selects a time constant from 1e-5 till 3e4 in seconds.

        :param tconstant: float, Time constant.
        :see: manual p5-6
        """
        typing.check_class(tconstant, float)
        assert tconstant >= 1e-5 and tconstant <= 3e4
        self.tx("OFLT {}".format(math.ceil(10-2*math.log10(1/tconstant))))


    def getTimeConstant(self):
        """ This method gets the time constant.

        :return: float, Time constant.
        :see: manual p5-6
        """
        return 1/(10**((10 - float(self.ask("OFLT?"))) / 2))


    def setLowPassSlope(self, slope):
        """ This method sets the low pass filter slope. The parameter slope
            selects 6, 12, 18 or 24 dB/oct.

        :param slope: int, Low pass filter slope.
        :see: manual p5-6
        """
        assert slope in [6, 12, 18, 24]
        self.tx("OFSL {}".format([6, 12, 18, 24].index(slope)))


    def getLowPassSlope(self):
        """ This method sets the low pass filter slope.

        :return: int, Low pass filter slope.
        :see: manual p5-6
        """
        return [6, 12, 18, 24][int(self.ask("OFSL?"))]


    def setSyncFilter(self, sync):
        """ This method sets the synchronous filter status. The parameter sync
            selects Off (False) or synchronous filtering below 200 Hz (True).

        :param sync: bool, Synchronous filter status.
        :see: manual p5-6
        """
        typing.check_class(sync, bool)
        self.tx("SYNC {}".format(int(sync)))


    def getSyncFilter(self):
        """ This method gets the synchronous filter status.

        :return: bool, Synchronous filter status.
        :see: manual p5-7
        """
        bool(self.ask("SYNC?"))


    def getOutputParameter(self, parameter):
        """ This method reads the value of X, Y, R or THETA. The parameter
            selects X, Y, R or THETA. Values are returned as floating point
            numbers in units of Volts or degrees.

        :param parameter: str, Output parameter.
        :see: manual p5-15
        """
        assert parameter in ['X', 'Y', 'R', 'THETA']

        tx = "OUTP? {}\n".format(['X', 'Y', 'R', 'THETA'].index(parameter) + 1)
        self.sock.send(tx.encode('utf8'))
        rx = self.sock.recv(self.sockBuffSize)
        return float(rx.decode('utf8'))


    def setDisplayDefinition(self, channel, display, ratio):
        """ This method selects the CH1 and CH2 displays. The parameter channel
            selects CH1 or CH2. The display parameter sets the display to
            X/Y, R/THETA, X-NOISE/Y-NOISE, AUX-IN-1/3, AUX-IN-2/4. The ratio
            parameter is set to NONE, AUX-IN-1/3, AUX-IN-2/4.

        :param channel: str, Channel parameter.
        :param display: str, Display parameter.
        :param ratio: str, Ratio parameter.
        :see: manual p5-8
        """
        assert channel in ['CH1','CH2']
        assert display in ['X', 'Y', 'R', 'THETA', 'X-NOISE', 'Y-NOISE',
                           'AUX-IN-1', 'AUX-IN-2', 'AUX-IN-3', 'AUX-IN-4']
        assert ratio in ['NONE', 'AUX-IN-1', 'AUX-IN-2', 'AUX-IN-2', 'AUX-IN-3']
        channel = ['CH1','CH2'].index(channel)
        if channel:
            display = ['Y', 'THETA', 'Y-NOISE', 'AUX-IN-3', 'AUX-IN-4'].index(display)
            ratio = ['NONE', 'AUX-IN-3', 'AUX-IN-4'].index(ratio)
        else:
            display = ['X', 'R', 'X-NOISE', 'AUX-IN-1', 'AUX-IN-2'].index(display)
            ratio = ['NONE', 'AUX-IN-1', 'AUX-IN-2'].index(ratio)
        self.tx("DDEF {},{},{}".format(channel + 1, display, ratio))


    def getDisplayDefinition(self, channel):
        """ This method selects the CH1 and CH2 displays. The parameter channel
            selects CH1 or CH2. The parameters display and ratio are returned.

        :param channel: str, Channel parameter.
        :return: str,str, Display, ratio parameters.
        :see: manual p5-8
        """
        channel = ['CH1', 'CH2'].index(channel)
        display, ratio = self.ask("DDEF? {}".format(channel + 1)).split(',')
        if channel:
            display = ['Y', 'THETA', 'Y-NOISE', 'AUX-IN-3', 'AUX-IN-4'][int(display)]
            ratio = ['NONE', 'AUX-IN-3', 'AUX-IN-4'][int(ratio)]
        else:
            display = ['X', 'R', 'X-NOISE', 'AUX-IN-1', 'AUX-IN-2'][int(display)]
            ratio = ['NONE', 'AUX-IN-1', 'AUX-IN-2'][int(ratio)]
        return display, ratio


    def setScanMode(self, mode):
        """ This method sets the end of buffer mode. The parameter mode
            selects SHOT or LOOP. If loop mode is used, make sure to pause
            data storage before reading the data to avoid confusion about
            which point is most recent.

        :param mode: str, Mode parameter.
        :see: manual p5-13

        """
        assert mode in ['SHOT', 'LOOP']
        self.tx("SEND {}".format(['SHOT', 'LOOP'].index(mode)))


    def getScanMode(self):
        """ This method gets the end of buffer mode. The parameter mode
            SHOT or LOOP is returned.

        :return: str, Mode parameter.
        :see: manual p5-13
        """
        return 'LOOP' if int(self.ask("SEND?")) else 'SHOT'


    def setScanRate(self, rate):
        """ This method sets the sample rate. The parameter rate selects
            the sample rate from 62.5e-3 up till 512 Hz.

        :param: float, Sample rate.
        :see: manual p5-13
        """
        assert rate >= 62.5e-3 and rate <= 512
        self.tx("SRAT {}".format(round(4 + math.log2(rate))))


    def getScanRate(self):
        """ This method gets the sample rate.

        :return: float Sample rate.
        :see: manual p5-13
        """
        return 2 ** (int(self.ask("SRAT?")) - 4)


    def startScan(self):
        """ This method starts or resumes data storage. The method is ignored
            if storage is already in progress.

        :see: manual p5-13
        """
        self.tx("STRD")

    def pauseScan(self):
        """ This method pauses data storage. If storage is already paused
            or reset then this command is ignored.

        :see: manual p5-14
        """
        self.tx("PAUS")


    def resetScan(self):
        """ This method resets the data buffer. This method can be called any
            time, any storage in progress, paused or not, will be reset.

        :see: manual p5-14
        """
        self.tx("REST")


    def getScanLength(self):
        """ This method gets the number of points stored in the buffer.
            Both displays have the same number of points. If the buffer is
            reset, then is returned. Remember, this method returns N
            where N is the number of points - the points are numbered
            from 0 (oldest) to N-1 (most recent).

        :return: int, Number of points stored.
        :see: manual p5-16
        """
        return int(self.ask("SPTS?"))


    def getScanPoints(self, channel, start, length):
        """ This method gets the points stored in the buffer selected, CH1 or CH2,
            by channel parameter. Point are read from the buffer starting at
            parameter start (start≥0). The length parameter selects the total
            number of points read (length≥1). To read a single point, set length=1.
            If start+length exceeds the number of stored points (as returned by
            the `getScanLength` method), then an error occurs. This method numbers
            the points from 0 (oldest) to N-1 (most recent). If data storage is set
            to Loop mode, make sure that storage is paused before reading any data.
            This is because the points are indexed relative to the most recent point,
            which is continually changing.

        :param: str, Channel parameter.
        :param: int, Start parameter.
        :param: int, Length parameter.
        :return: list(floats), Points in buffer.
        """
        assert channel in ['CH1', 'CH2']
        pnts = self.ask("TRCA? {},{},{}".format(['CH1','CH2'].index(channel) + 1, start, length))
        pnts = pnts.rstrip(',\n')
        return [float(pnt) for pnt in pnts.split(',')]


    def autoGain(self):
        """ This method performs the Auto Gain function. This method is
            the same as pressing the [Auto Gain] key. Auto Gain may take
            some time if the time constant is long.

        :see: manual p5-11
        """
        self.tx("AGAN")


    def autoRange(self, iterations):
        """ This method performs Auto Ranging. The parameter iterations
            specifies how many iterations the range is adjusted.

        :param iterations: int, Number of iterations.
        """
        tconstant = self.getTimeConstant()
        for _ in range(iterations):
            ampl = self.getOutputParameter('R')
            self.setSensitivity(1.1 * ampl)
            time.sleep(tconstant)


    def autoAverage(self, parameter, length):
        """ This method performs Auto Average. The parameter
            selects output 'X','Y','R' or 'THETA'. The length
            parameter sets the number of points of the average.

         :param parameter: str, Output parameter.
         :param length:, int, Length parameter.
         :return: float, Average.
        """
        assert parameter in ['X', 'Y', 'R', 'THETA']

        tc = self.getTimeConstant()
        sfreq = 1 / tc
        sfreq = 512 if sfreq > 512 else sfreq
        sfreq = 62.5e-3 if sfreq < 62.5e-3 else sfreq

        if length < 75:
            avr = 0
            for _ in range(length):
                avr += self.getOutputParameter(parameter) / length
            return avr
        else:
            # Average using data stream.
            sens = self.getSensitivity()

            self.resetScan()
            self.setScanRate(sfreq)
            self.setScanMode('SHOT')
            self.tx('FAST 2')

            self.startScan()
            data = b''
            try:
                self.sock.settimeout(max(tc * 2, self.timeout))
                while not data or len(data) < (length * 4):
                    s = self.sock.recv(self.sockBuffSize)
                    data += s
            finally:
                self.tx('++auto 0')  # Please listen
                self.tx('FAST 0')  # Stop stream
                self.tx('++auto 1')
                self.pauseScan()

                self.sock.settimeout(max(tc * 2, self.timeout / 2))

                # Flush remaining packets from stream.
                flush = True
                while flush:
                    try:
                        self.sock.recv(self.sockBuffSize)
                    except socket.timeout:
                        flush = False

            datLength = len(data)
            datLength = datLength - datLength % 4
            data = [struct.unpack('hh', data[idx:idx + 4]) for idx in range(0, datLength, 4)]
            data = [(sens * x / 32768, sens * y / 32768) for x, y in data]

            length = min(length, datLength)
            if parameter == 'X':
                return sum([x for x, _ in data[-length:]]) / length
            elif parameter  == 'Y':
                return sum([y for _, y in data[-length:]]) / length
            elif parameter == 'R':
                return sum([math.sqrt(x ** 2 + y ** 2) for x, y in data[-length:]]) / length
            else:
                return sum([math.atan2(y, x) for x, y in data[-length:]]) / length


    def autoReserve(self):
        """ This method performs Auto Reserve. This method is
            the same as pressing the [Auto Reserve] key. Auto Reserve
            may take some time.
        :see: manual p5-11

        """
        self.tx("ARSV")


    def autoPhase(self):
        """ This method performs Auto Phase. This method is
            the same as pressing the [Auto Phase] key. The outputs will
            take many time constants to reach their new values.

        :see: manual p5-11

        """
        self.tx("APHS")


    def autoOffset(self, parameter):
        """ This method automatically offsets X, Y or R to zero. The parameter
            selects X, Y, or R. The method is equivalent to pressing the [Auto Offset] keys.

        :param parameter: str, Offset parameter.
        """
        assert parameter in ['X', 'Y', 'R']
        self.tx("AOFF {}".format(['X', 'Y', 'R'].index(parameter) + 1))


def main():
    """ User interface program for configuring a multimeter. """

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('host', nargs='?',
                        type=str,
                        help='hostname of the multimeter '
                             '(default: %(default)s)',
                        default=DEFAULT_HOST)
    parser.add_argument('port', nargs='?',
                        type=int,
                        help='SCPI port number of the multimeter '
                             '(default: %(default)s)',
                        default=DEFAULT_PORT)
    parser.add_argument('gpibAddress', nargs='?',
                        type=int,
                        help='GPIB address of te multimeter '
                             '(default: %(default)s)',
                        default=DEFAULT_GPIB_ADDRESS)

    args = parser.parse_args()

    srs = SRS830(host=args.host,
                 port=args.port,
                 gpibAddr=args.gpibAddress)

    print("Identification: {}".format(srs.identification().rstrip()))
    print("Prologix version: {}".format(srs.converterVersion().rstrip()))

    srs.setTimeConstant(3e-4)
    srs.autoRange(5)
    print("phase: %f" % srs.getOutputParameter('THETA'))
    srs.finalize()


    main()
