#!/usr/bin/env python3
#pylint: disable=invalid-name
""" Module for interfacing with an SRS SIM960 Analog PID Controller. """

import argparse
import logging
import gssw.lib.typing  as typing
import gssw.ucelib.gpib as gpib

logger = logging.getLogger(__name__)

__all__ = ['SRSSim960']

DEVICE_IDENTITY_PREFIX = "Stanford_Research_Systems,SIM900"

# Defaults
DEFAULT_HOST = "192.168.1.201"
DEFAULT_PORT = 1234
DEFAULT_GPIB_ADDRESS = 1


#pylint: disable=too-many-public-methods
class SRSSim960(gpib.GpibTcpIp):
    """ Provides a communication interface to an SRS SIM960 Analog PID Controller.

    :note: The PID controller must be connected via a Prologix Gpib to Ethernet converter.
    :note: Some function docstrings are excerpts from the SRS SIM960 programming manual.
           Capital words correspond to (soft) keys on the instrument's front panel.
    """

    def __init__(self, host, port, gpibAddr):
        """ Constructor, for LAN communication. """

        super().__init__(host, port, "analyzer", commandInterval=0.2, processingDelay=0.01,
                         explicitRead=True,       # Request readings from the Prologix converter
                         timeout=4)               # Receiving measurement data can take a while

        if self.isConnected:
            if gpibAddr:
                # Set up the Prologix converter
                self.tx('++auto 1')                   # Read response after sending commands
                #                                       to the instrument.
                self.tx('++mode 1')                   # Use Prologix converter in Controller mode
                self.tx('++addr {}'.format(gpibAddr)) # Configure the GPIB address
                self.tx('++eos 2')                    # Append LF to instrument commands
                self.tx('++eoi 1')                    # Use EOI signals after all last characters

            self._checkIdentification()
            self.tx('FLOQ')
            self.tx('FLSH 1')


    def finalize(self):
        """ Cleans up the connection to the power meter and the Prologix converter. """
        if self.isConnected:
            self.setDisplayOn(True)


    def _checkIdentification(self):
        """ Verifies that we are connected to a SRS SIM960. """
        identification = self.identification().strip()
        deviceCorrect = identification.startswith(DEVICE_IDENTITY_PREFIX)
        assert deviceCorrect,\
            "Device with identity '{}' is not supported.".format(identification)


    @staticmethod
    def _parseArgument(arg=None):
        """ Parse argument for SRS SIM960 command. """
        if arg not in [None, 'MIN', 'MAX']:
            typing.check_class(arg, float)
        if arg is None:
            arg = 'DEF'
        return arg


    def setProportionalActionOn(self, action):
        """ Proportional action ON/OFF. Set the proportional control; Off (False), On (True).
            When On, the PID Control path includes the proportional control term.

        :param action: bool, Proportional action ON/OFF.
        :see: manual p3-10
        """
        typing.check_class(action, bool)
        self.tx('SNDT 1,"PCTL {}"'.format(int(action)))


    def getProportionalActionOn(self):
        """ Proportional action ON/OFF. Query the proportional control; Off (False), On (True).

        :return: bool, Proportional action ON/OFF.
        :see: manual p3-10
        """
        self.tx('SNDT 1,"PCTL?"')
        return bool(int(self.ask("GETN? 1,80")[5:].rstrip()))


    def setIntegralActionOn(self, action):
        """ Integral action ON/OFF. Set the integral control; Off (False), On (True).
            When On, the PID Control path includes the integral control therm.

        :param action: bool, Integral action ON/OFF.
        :see: manual p3-10
        """
        typing.check_class(action, bool)
        self.tx('SNDT 1,"ICTL {}"'.format(int(action)))


    def getIntegralActionOn(self):
        """ Integral action ON/OFF. Query the integral control; Off (False), On (True).

        :return: bool, Integral action ON/OFF.
        :see: manual p3-10
        """
        self.tx('SNDT 1,"ICTL?"')
        return bool(int(self.ask("GETN? 1,80")[5:].rstrip()))


    def setDerivativeActionOn(self, action):
        """ Derivative action ON/OFF. Set the derivative control; Off (False), On (True).
            When On, the PID Control path includes the derivative control therm.

        :param action: bool, Derivative action ON/OFF.
        :see: manual p3-10
        """
        typing.check_class(action, bool)
        self.tx('SNDT 1,"DCTL {}"'.format(int(action)))


    def getDerivativeActionOn(self):
        """ Derivative action ON/OFF. Query the derivative control; Off (False), On (True).

        :return: bool, Derivative action ON/OFF.
        :see: manual p3-10
        """
        self.tx('SNDT 1,"DCTL?"')
        return bool(int(self.ask("GETN? 1,80")[5:].rstrip()))


    def setOffsetOn(self, action):
        """ Offset ON/OFF. Set the offset control; Off (False), On (True).
            When On, the PID Control path includes the constant output offset.

        :param action: bool, Constant offset ON/OFF.
        :see: manual p3-10
        """
        typing.check_class(action, bool)
        self.tx('SNDT 1,"OCTL {}"'.format(int(action)))


    def getOffsetOn(self):
        """ Offset ON/OFF. Query the offset control; Off (False), On (True).

        :return: bool, Constant offset ON/OFF.
        :see: manual p3-10
        """
        self.tx('SNDT 1,"OCTL?"')
        return bool(int(self.ask("GETN? 1,80")[5:].rstrip()))


    def setGain(self, gain):
        """ Proportional gain. Set the proportional gain in V/V. The allowed range
            for gain is 1e-1 till 1e3.

        :param gain: float, Proportional gain.
        :see: manual p3-10
        """
        typing.check_class(gain, float)
        assert gain >= 1e-1 and gain <= 1e3
        self.tx('SNDT 1,"GAIN {}"'.format(gain))


    def getGain(self):
        """ Proportional gain. Query the proportional gain.

        :return: float, Proportional gain.
        :see: manual p3-10
        """
        self.tx('SNDT 1,"GAIN?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def setPolarityPos(self, polarity):
        """ Controller polarity. Set the proportional gain polarity; Pos (True), Neg (False).

        :param polarity: bool, Controller polarity POS/NEG.
        :see: manual p3-11
        """
        typing.check_class(polarity, bool)
        self.tx('SNDT 1,"APOL {}"'.format(int(polarity)))


    def getPolarityPos(self):
        """ Controller polarity POS/NEG. Query the proportional gain polarity;
            Pos (True), Neg (False).

        :return: bool, Controller polarity POS/NEG.
        :see: manual p3-11
        """
        self.tx('SNDT 1,"APOL?"')
        return bool(int(self.ask("GETN? 1,80")[5:].rstrip()))


    def setIntegralGain(self, gain):
        """ Integral gain. Set the integral gain in V/(V*s). The allowed range
            for gain is 1e-2 till 1e5.

        :param gain: float, Integral gain.
        :see: manual p3-11
        """
        typing.check_class(gain, float)
        assert gain >= 1e-2 and gain <= 1e5
        self.tx('SNDT 1,"INTG {}"'.format(gain))


    def getIntegralGain(self):
        """ Integral gain. Query the integral gain.

        :return: float, Integral gain.
        :see: manual p3-11
        """
        self.tx('SNDT 1,"INTG?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def setDerivativeGain(self, gain):
        """ Derivative gain. Set the derivative gain in V/(V/s). The allowed range
            for gain is 1e-6 till 10.

        :param gain: float, Derivative gain.
        :see: manual p3-11
        """
        typing.check_class(gain, float)
        assert gain >= 1e-2 and gain <= 1e5
        self.tx('SNDT 1,"DERV {}"'.format(gain))


    def getDerivativeGain(self):
        """ Derivative gain. Query the derivative gain.

        :return: float, Derivative gain.
        :see: manual p3-11
        """
        self.tx('SNDT 1,"DERV?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def setOutputOffset(self, offset):
        """ Output offset. Set the output offset in V. The allowed range
            for offset is -10.0 till 10.0.

        :param gain: float, Output offset.
        :see: manual p3-11
        """
        typing.check_class(offset, float)
        assert offset >= -10.0 and offset <= 10.0
        self.tx('SNDT 1,"OFST {}"'.format(offset))


    def getOutputOffset(self):
        """ Output offset. Query the output offset.

        :return: float, Output offset.
        :see: manual p3-11
        """
        self.tx('SNDT 1,"OFST?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def setOutputState(self, state):
        """ Output state. Set the output state; Manual (False), PID (True).

        :param state: bool, Output state MAN/PID.
        :see: manual p3-12
        """
        typing.check_class(state, bool)
        self.tx('SNDT 1,"AMAN {}"'.format(int(state)))


    def getOutputState(self):
        """ Output state. Query the output state; Manual (False), PID (True).

        :return: bool, Output state MAN/PID.
        :see: manual p3-12
        """
        self.tx('SNDT 1,"AMAN?"')
        return bool(int(self.ask("GETN? 1,80")[5:].rstrip()))


    def setSetpointState(self, state):
        """ Setpoint state. Set the setpoint input state; Internal (False), External (True).

        :param state: bool, Output state INT/EXT.
        :see: manual p3-12
        """
        typing.check_class(state, bool)
        self.tx('SNDT 1,"INPT {}"'.format(int(state)))


    def getSetpointState(self):
        """ Setpoint state. Query the setpoint input state; Internal (False), External (True).

        :return: bool, Output state INT/EXT.
        :see: manual p3-12
        """
        self.tx('SNDT 1,"INPT?"')
        return bool(int(self.ask("GETN? 1,80")[5:].rstrip()))


    def setSetpoint(self, setpoint):
        """ New setpoint. Set the setpoint value in V. The allowed range
            for setpoint is -10.0 till 10.0.

        :param setpoint: float, Setpoint.
        :see: manual p3-12
        """
        typing.check_class(setpoint, float)
        assert setpoint >= -10.0 and setpoint <= 10.0
        self.tx('SNDT 1,"SETP {}"'.format(setpoint))


    def getSetpoint(self):
        """ Setpoint. Query the setpoint.

        :return: float, Setpoint.
        :see: manual p3-12
        """
        self.tx('SNDT 1,"SETP?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def setSetpointRampingOn(self, action):
        """ Internal setpoint ramping. Set the internal setpoint ramping; OFF (False), ON (True).

        :param action: bool, Internal setpoint ramping.
        :see: manual p3-12
        """
        typing.check_class(action, bool)
        self.tx('SNDT 1,"RAMP {}"'.format(int(action)))


    def getSetpointRampingOn(self):
        """ Internal setpoint ramping. Query the internal setpoint ramping.

        :return: bool, Internal setpoint ramping OFF/ON.
        :see: manual p3-12
        """
        self.tx('SNDT 1,"RAMP?"')
        return bool(int(self.ask("GETN? 1,80")[5:].rstrip()))


    def setSetpointRate(self, rate):
        """ Setpoint ramping rate. Set the setpoint rate in V/s. The allowed range
            for rate is 1e-3 till 1e4.

        :param gain: float, Setpoint ramping rate.
        :see: manual p3-12
        """
        typing.check_class(rate, float)
        assert rate >= 1e-3 and rate <= 1e4
        self.tx('SNDT 1,"RATE {}"'.format(rate))


    def getSetpointRate(self):
        """ Setpoint ramping rate. Query the setpoint rate.

        :return: float, Setpoint ramping rate.
        :see: manual p3-12
        """
        self.tx('SNDT 1,"RATE?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def getSetpointRampingStatus(self):
        """ Setpoint ramping status. Query the setpoint status.
            The response is one of the following token values: IDLE,
            PENDING, RAMPING, PAUSED

        :return: str, Setpoint ramping status.
        :see: manual p3-13
        """
        self.tx('SNDT 1,"RMPS?"')
        return ['IDLE', 'PENDING', 'RAMPING', 'PAUSED'][self.ask("GETN? 1,80")[5:].rstrip()]


    def setSetpointRampingState(self, state):
        """ Pause or continue ramping. Cause a setpoint ramping event
            in progress to pause (False) or continue (True).

        :param state: bool, Setpoint ramping state.
        :see: manual p3-13
        """
        typing.check_class(state, bool)
        self.tx('SNDT 1,"STRT {}"'.format(int(state)))


    def setManualOutput(self, output):
        """ Manual output. Set the manual output value in V. The allowed range
            for output is -10.0 till 10.0.

        :param output: float, Manual output.
        :see: manual p3-13
        """
        typing.check_class(output, float)
        assert output >= -10.0 and output <= 10.0
        self.tx('SNDT 1,"MOUT {}"'.format(output))


    def getManualOutput(self):
        """ Manual output. Query the manual output.

        :return: float, Manual output.
        :see: manual p3-13
        """
        self.tx('SNDT 1,"MOUT?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def setUpperOutputLimit(self, limit):
        """ Upper output limit. Set the upper output limit in V. The allowed range
            for limit is from lower output limit till 10.0.

        :param limit: float, Upper output limit.
        :see: manual p3-13
        """
        typing.check_class(limit, float)
        assert limit >= -10.0 and limit <= 10.0
        self.tx('SNDT 1,"MOUT {}"'.format(limit))


    def getUpperOutputLimit(self):
        """ Upper output limit. Query the upper output limit.

        :return: float, Upper output limit.
        :see: manual p3-13
        """
        self.tx('SNDT 1,"ULIM?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def setLowerOutputLimit(self, limit):
        """ Lower output limit. Set the lower output limit in V. The allowed range
            for limit is -10.0 till upper output limit.

        :param limit: float, Lower output limit.
        :see: manual p3-13
        """
        typing.check_class(limit, float)
        assert limit >= -10.0 and limit <= 10.0
        self.tx('SNDT 1,"LLIM {}"'.format(limit))


    def getLowerOutputLimit(self):
        """ Lower output limit. Query the lower output limit.

        :return: float, Lower output limit.
        :see: manual p3-14
        """
        self.tx('SNDT 1,"LLIM?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def getSetpointInputMonitor(self):
        """ Setpoint input monitor. Query the setpoint input voltage to the error
            amplifier in volts.

        :return: float, Setpoint input.
        :see: manual p3-14
        """
        self.tx('SNDT 1,"SETP?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def getMeasureInputMonitor(self):
        """ Measure input monitor. Query the measure input voltage to the error
            amplifier in volts.

        :return: float, Measure input.
        :see: manual p3-14
        """
        self.tx('SNDT 1,"MMON?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def getErrorMonitor(self):
        """ Amplified error monitor. Query the P x epsilon voltage in volts.

        :return: float, Amplified error monitor.
        :see: manual p3-15
        """
        self.tx('SNDT 1,"EMON?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def getOutputMonitor(self):
        """ Output monitor. Query the output voltage in volts.

        :return: float, Output monitor.
        :see: manual p3-15
        """
        self.tx('SNDT 1,"OMON?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def setFrequencyPowerLineCycle(self, freq):
        """ Frequency of power line cycle. Set the power line cycle frequency
            to 50 or 60 Hz.

        :param freq: int, Frequency of power line cycle.
        :see: manual p3-16
        """
        typing.check_class(freq, int)
        assert freq in [50, 60]
        self.tx('SNDT 1,"FPLC {}"'.format(freq))


    def getFrequencyPowerLineCycle(self):
        """ Frequency of power line cyle. Query the power line cycle frequency.

        :return: int, Frequency of power line cycle.
        :see: manual p3-16
        """
        self.tx('SNDT 1,"FGLC?"')
        return float(self.ask("GETN? 1,80")[5:].rstrip())


    def setDisplayField(self, field):
        """ Select field. Set the field level to be displayed. Allowed values
            of field are: PRP, IGL, DER, OFS, RTE, STP, MNL, ULM, LLM, SMN,
            MMN, EMN, OMN.

        :param field: str, Field level.
        :see: manual p3-16

        """
        assert field in ['PRP', 'IGL', 'DER', 'OFS', 'RTE', 'STP', 'MNL', 'ULM',
                         'LLM', 'SMN', 'MMN', 'EMN', 'OMN']
        self.tx('SNDT 1, "DISP {}"'.format(['PRP', 'IGL', 'DER', 'OFS', 'RTE',
                                            'STP', 'MNL', 'ULM', 'LLM', 'SMN',
                                            'MMN', 'EMN', 'OMN'].index(field)))

    def getDisplayField(self):
        """ Select field. Query the field level to be displayed.

        :return: str, Field level.
        :see: manual p3-16
        """
        self.tx('SNDT 1, "DISP?"')
        return ['PRP', 'IGL', 'DER', 'OFS', 'RTE', 'STP', 'MNL', 'ULM',
                'LLM', 'SMN', 'MMN', 'EMN', 'OMN'][int(self.ask("GETN? 1,80")[5:].rstrip())]


    def setShiftStatus(self, status):
        """ Shift status ON/OFF. Set the current shift status; Off (False), On (True).

        :param status: bool, Shift status ON/OFF.
        :see: manual p3-16
        """
        typing.check_class(status, bool)
        self.tx('SNDT 1,"SHFT {}"'.format(int(status)))


    def getShiftStatus(self):
        """ Shift status ON/OFF. Query the current shift status; Off (False), On (True).

        :return: bool, Sift status ON/OFF.
        :see: manual p3-16
        """
        self.tx('SNDT 1,"SHFT?"')
        return bool(int(self.ask("GETN? 1,80")[5:].rstrip()))


    def setDisplayOn(self, status):
        """ Front panel display enable. Set the front panel display status;
            Off (False), On (True).

        :param status: bool, Display status ON/OFF.
        :see: manual p3-17
        """
        typing.check_class(status, bool)
        self.tx('SNDT 1,"DISX {}"'.format(int(status)))


    def getDisplayOn(self):
        """ Front panel display enable. Query the front panel display status;
            Off (False), On (True).

        :return: bool, Display status ON/OFF.
        :see: manual p3-17
        """
        self.tx('SNDT 1,"DISX?"')
        return bool(int(self.ask("GETN? 1,80")[5:].rstrip()))



def main():
    """ User interface program for configuring a multimeter. """

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('host', nargs='?',
                        type=str,
                        help='hostname of the multimeter '
                             '(default: %(default)s)',
                        default=DEFAULT_HOST)
    parser.add_argument('port', nargs='?',
                        type=int,
                        help='SCPI port number of the multimeter '
                             '(default: %(default)s)',
                        default=DEFAULT_PORT)
    parser.add_argument('gpibAddress', nargs='?',
                        type=int,
                        help='GPIB address of te multimeter '
                             '(default: %(default)s)',
                        default=DEFAULT_GPIB_ADDRESS)

    args = parser.parse_args()

    srs = SRSSim960(host=args.host,
                    port=args.port,
                    gpibAddr=args.gpibAddress)

    print("Identification: {}".format(srs.identification().rstrip()))
    print("Prologix version: {}".format(srs.converterVersion().rstrip()))

    srs.setProportionalActionOn(True)
    print('P_on: {}'.format(srs.getProportionalActionOn()))

    srs.setGain(1.0)
    print('P: {}'.format(srs.getGain()))

    srs.setDisplayField('PRP')

    srs.finalize()


if __name__ == '__main__':
    main()
