#!/usr/bin/env python3
""" Driver for the ThorLabs Multi-Channel Fiber-Coupled Laser Source """

import logging
import serial

from gssw.ucelib.serialdevice import SerialDevice

logger = logging.getLogger(__name__)


class ThorlabsMCLS(SerialDevice):
    """ Multi Channel Laser Source driver

        The serial port is found automatically

        If you have installed the udev rules file under setups/fastreadout/system,
        there should also be a symlink for the serial port.
    """

    def __init__(self, port='/dev/ttyMCLS'):
        super().__init__(port     =port,
                         vendorId =0x0403,
                         productId=0x6001,
                         serialNumber='17042090202',
                         baudrate =115200,
                         outputsep='\r',
                         inputsep ='\r')

        # Active channel unknown
        self._channel = -1
        try:
            self.open()
        except serial.SerialException:
            logger.error('Could not connect to port: %s', port)
        else:
            logger.info('Connected to MCLS at port: %s', self._serial.port)

    def close(self):
        """ Close serial port """
        self._channel = -1
        super().close()

    def enable_channel(self, channel, enable=True):
        """ Turn one laser channel on/off """
        self.channel = channel
        self.set_reg('enable', 1 if enable else 0)

    def get_enable_channel(self, channel):
        """ Turn one laser channel on/off """
        self.channel = channel
        resp = self.get_reg('enable')
        return bool(int(resp[-1]))

    def enable_system(self, enable=True):
        """ Enable to enable output from enabled channels

            :param enable: True to enable, False to disable
        """
        self.set_reg('system', 1 if enable else 0)

    def get_enable_system(self):
        """ Check whether outputs from selected channels are enabled.

            :return: bool, status off whole system output (on/off).
        """
        resp = self.get_reg('system')
        return bool(int(resp[-1]))

    def set_max_current(self, channel):
        """ Set a channel's current to its own maximum level. """
        self.channel = channel

        specs = self.get_specs(channel)
        currentSetting = float(specs['IOp'])

        self.set_reg('current', currentSetting)

    def get_max_current(self, channel):
        """ Query a channel's own maximum current level. """
        self.channel = channel

        specs = self.get_specs(channel)
        return float(specs['IOp'])

    def get_specs(self, channel):
        """ Get the specifications for a channel

            :return: The specs as a dictionary (of strings)
        """
        self.channel = channel

        resplist = self.get_reg('specs')
        return dict([resp.split(' = ') for resp in resplist])

    @property
    def channel(self):
        """ Change selected laser channel on the device """
        return self._channel

    @channel.setter
    def channel(self, channel):

        if channel not in [1, 2, 3, 4]:
            raise RuntimeError('Invalid channel number')

        if self._channel != channel:
            self.set_reg('channel', channel)
            self._channel = channel

    def set_reg(self, register, value):
        """ Write register """
        cmd = '{reg}={val}'.format(reg=register, val=value)
        logger.debug('Sending : %s', cmd)

        # Return command echo and prompt, ignore output
        self.cmd(cmd)

    def get_reg(self, register):
        """ Read register """
        cmd = '{reg}?'.format(reg=register)
        logger.debug('Sending : %s', cmd)

        # Return command echo, response and prompt
        resplist = self.cmd(cmd)

        logger.debug('Returned %d lines', len(resplist))

        # Response can be multiple lines
        # Return everything except command echo and prompt
        return resplist[1:-1]

def main():
    """ Quick self-test """

    laser = ThorlabsMCLS()
    try:
        resp = laser.get_reg('id')
        if len(resp) == 1:
            print('Got ID {}'.format(resp[0]))
        else:
            print('Unexpected response: {}'.format(resp))
    finally:
        laser.close()

if __name__ == '__main__':
    main()
