#!/usr/bin/env python3
""" Driver for the Vacscan. """

import logging
import serial
from gssw.ucelib.serialdevice import SerialDevice


logger = logging.getLogger(__name__)


class VacscanError(Exception):
    """ Vacscan protocol errors. """


class Vacscan(SerialDevice):
    """ Driver class for controlling the vacsan through the serial interface. """


    def __init__(self, port='/dev/ttyUSB0', vendorId=None, productId=None, serialNumber=None):
        """ Constructor

            :param str port: device port on host machine
            :param int vendorId: usb device vendor id
            :param int productId: usb device product id
            :param str serialNumber: usb device serial number
        """

        # initialize the serial base class
        super().__init__(port=port, vendorId=vendorId, productId=productId,
            serialNumber=serialNumber, baudrate=9600, bytesize=serial.EIGHTBITS,
            parity=serial.PARITY_NONE, stopbits=serial.STOPBITS_ONE, timeout=10)

        # open the serial interface
        self.open()


    def reset(self):
        """ Perform a software reset on the device. """

        self._command('(R)', checkResponse=False)


    def enableAcknowledge(self):
        """ Enable response messages from the device. """

        self._command('(A1)')


    def getStatus(self):
        """ Read the device status. """

        return self._command('(S)')


    def setUserPeaks(self):
        """ Put the gas analyzer in User peaks mode.

        :see: Manual, section 8.5.10, Set mode m."""

        self._command('(M1)')


    def setBarChart(self):
        """ Put the gas analyzer in Bar chart mode.

        :see: Manual, section 8.5.10, Set mode m."""

        self._command('(M2)')


    def setAutoRange(self, enable):
        """ Set auto range on/off (for non-Bar chart modes).

        :param bool enable: Auto range state.

        :see: Manual, section 8.5.18, Parameter set Y.
        """

        self._command('(Y7,{})'.format(int(enable)))


    def setTotalPressure(self, enable):
        """ Set total pressure on/off (not available in leak-check mode).

        :param bool enable: On/off state.

        :see: Manual, section 8.5.18, Parameter set Y.
        """

        self._command('(Y2,{})'.format(int(enable)))


    def setGain(self, gain):
        """ Set range (gain).

        :param int gain: Gain 0-9.

        :see: Manual, section 8.5.18, Parameter set Y.
        """

        self._command('(Y6,{})'.format(gain))


    def getData(self, span=50):
        """ Read the complete scan.

        :param int span: Number of expected data lines.

        :see: Manual, section 8.5.12, Read complete scan.
        """

        return self._command('(P)', dataLines=span)


    def enableFilament(self, index, enable):
        """ turn filament on or off
            index - filament index [0. 1]
            enable - (bool)
        """

        self._command('(X{},{})'.format(index + 1, int(enable)))


    def _command(self, cmd, checkResponse=True, dataLines=1):
        """ Issue a command string to the device and check the response for errors. """

        print('send: {}'.format(cmd))
        self._serial.write((cmd + '\r\n').encode())
        retData = ''

        if checkResponse:

            response = self._serial.readline().decode()
            print('recv: {}'.format(response.strip()))

            if not response: # pylint was complaining here: Don't use len(x)
                raise VacscanError('No response received for command: {}'.format(cmd))
            if response[0] == 'N':
                raise VacscanError('NACK response received for command: {}'.format(cmd))
            if response[0] != 'A':
                raise VacscanError('Invalid response received: {}'.format(response))
            if response[-2:] != '\r\n':
                raise VacscanError('Invalid response terminators received: {}'.format(response))

            if len(response) > 3:
                retData += response[1:-2]

            for _ in range(1, dataLines):
                retData += ',' + self._serial.readline().decode()[:-2]

            return retData

        return None


def main():
    """ Module test function. """
    import time
    #vacscan = Vacscan(vendorId=0x067b, productId=0x2303, serialNumber=None)
    vacscan = Vacscan(port='/dev/ttyUSB8')
    vacscan.reset()
    vacscan.enableAcknowledge()
    vacscan.setBarChart()
    vacscan.setTotalPressure(True)
    vacscan.setAutoRange(True)
    #vacscan.setGain(7)
    vacscan.enableFilament(0, True)
    time.sleep(1)
    print(vacscan.getStatus())


    for _ in range(3):
        time.sleep(10)
        print(vacscan.getData())

    #vacscan.enableFilament(0, False)
    print(vacscan.getStatus())

if __name__ == '__main__':
    main()
